/*LINTLIBRARY*/
/*  atm.c
    This code implements ATM support for Linux

    Copyright (C) 1998-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/
/*

    This file contains the various utility routines for manipulating Karma
    connections.


    Written by      Richard Gooch   12-FEB-1998

    Updated by      Richard Gooch   19-FEB-1998

    Updated by      Richard Gooch   8-OCT-1998: Took account of change to
  <<karma_qos>> structure.

    Updated by      Richard Gooch   3-NOV-1998: Fixes for glibc 2 pedantry.

    Updated by      Richard Gooch   13-DEC-1998: Took account of change to
  <<karma_qos>> structure: separate Tx and Rx traffic parameters.

    Updated by      Richard Gooch   25-DEC-1998: Created read, write and close
  wrapper functions.

    Updated by      Richard Gooch   9-FEB-1999: Updated documentation to say
  that address strings are binary not textual.

    Updated by      Richard Gooch   20-FEB-1999: Removed work-around for bug in
  ENI driver limiting max_sdu to 43690 bytes: ATM-0.53 fixes the problem.

    Updated by      Richard Gooch   2-NOV-1999: Compile fixes for glibc 2.0.

    Last updated by Richard Gooch   4-MAY-2001: Moved __STRICT_ANSI__ down to
  work around changes in kernel 2.4 headers.


*/
#define _LOOSE_KERNEL_NAMES
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#define __STRICT_ANSI__
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#undef __STRICT_ANSI__
#include <asm/types.h>
#if __GLIBC__ >= 2  /*  Glibc 2 has caused everybody no end of trouble  */
#  if __GLIBC_MINOR__ < 1
typedef         __u32           uint32_t;
#  endif
/*#include <linux/types.h>*/
#else     /*  The arrogance of the Glibc 2 bastards!  */
typedef         int             socklen_t;
#include <linux/socket.h>
#endif
#ifndef AF_ATMSVC
#  define AF_ATMSVC 20
#  define PF_ATMSVC AF_ATMSVC
#endif
#ifndef SOL_ATM
#  define SOL_ATM   264
#endif
#define _LINUX_TYPES_H
#include <linux/atm.h>
#include <linux/atmdev.h>
#include <linux/atmsvc.h>
#include <karma.h>
#include <karma_r.h>

#define DEFAULT_SDU 65535
#define BYTES_PER_CELL 48  /*  Number of payload bytes per cell     */


/*  Private functions  */
STATIC_FUNCTION (void _r_atm_karma_to_atm,
		 (struct atm_trafprm *atm,
		  CONST struct karma_traffic_parameters *karma) );
STATIC_FUNCTION (void _r_atm_to_karma,
		 (struct karma_traffic_parameters *karma,
		  struct atm_trafprm *atm) );
STATIC_FUNCTION (flag _r_atm_get_qos, (int fd, struct karma_qos *qos) );


/*  Public functions follow  */


/*UNPUBLISHED_FUNCTION*/
int r_atm_alloc (char *addr_string, unsigned *addr_len)
/*  [SUMMARY] Create an ATMSVC socket and bind to a random port number.
    <addr_string> The binary representation of the ATM address is written here
    <addr_len> The length of the address string. This is updated.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [RETURNS] The file descriptor of the socket on success, else -1.
*/
{
    int sock, count;
    struct sockaddr_atmsvc addr;
    struct atm_qos atmqos;
    struct atmif_sioc req;
    static char function_name[] = "r_atm_alloc";

    if (*addr_len < ATM_ESA_LEN)
    {
	fprintf (stderr,
		 "%s: address length: %u is too short, must be at least: %u\n",
		 function_name, *addr_len, ATM_ESA_LEN);
	return (-1);
    }
    if ( ( sock = socket (PF_ATMSVC, SOCK_DGRAM, 0) ) < 0 )
    {
	fprintf (stderr, "%s: error creating ATMSVC socket\t%s\n",
		 function_name, ERRSTRING);
	return (-1);
    }
    memset (&addr, 0, sizeof addr);
    req.number = 0;
    req.arg = &addr;
    req.length = sizeof addr;
    if (ioctl (sock, ATM_GETADDR, &req) < 0)
    {
	fprintf (stderr, "%s: error getting address\t%s\n",
		 function_name, ERRSTRING);
	return (-1);
    }
    memset (&atmqos, 0, sizeof atmqos);
    atmqos.txtp.traffic_class = ATM_ANYCLASS;
    atmqos.rxtp = atmqos.txtp;
    atmqos.aal = ATM_AAL5;
    if (setsockopt (sock, SOL_ATM, SO_ATMQOS, &atmqos, sizeof atmqos) != 0)
    {
	fprintf (stderr, "%s: error setting ATMSVC socket options\t%s\n",
		 function_name, ERRSTRING);
	close (sock);
	return (-1);
    }
    /*  Try to bind to a random port (SAP)  */
    addr.sas_family = AF_ATMSVC;
    for (count = 0; count < 256; ++count)
    {
	addr.sas_addr.prv[ATM_ESA_LEN - 1] = count;
	if (bind (sock, (struct sockaddr *) &addr, (int) sizeof addr) != 0)
	{
	    fprintf (stderr, "%s: error binding ATMSVC socket\t%s\n",
		     function_name, ERRSTRING);
	    if (close (sock) != 0)
	    {
		fprintf (stderr, "%s: error closing ATMSVC socket\t%s\n",
			 function_name, ERRSTRING);
	    }
	    exit (RV_SYS_ERROR);
	}
	if (listen (sock, 2) == 0) break;
	if (errno == EADDRINUSE) continue;
	fprintf (stderr, "%s: error setting ATMSVC socket to listening\t%s\n",
		 function_name, ERRSTRING);
	close (sock);
	return (-1);
    }
    if (count >= 256)
    {
	fprintf (stderr, "%s: no free ATMSVC SAPs\n", function_name);
	errno = 0;
	close (sock);
	return (-1);
    }
    memcpy (addr_string, addr.sas_addr.prv, ATM_ESA_LEN);
    *addr_len = ATM_ESA_LEN;
    /*  Set close-on-exec flag  */
    if (fcntl (sock, F_SETFD, 1) == -1)
    {
	fprintf (stderr,
		 "Error setting close-on-exec flag for ATMSVC socket\t%s\n",
		 ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing ATMSVC socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    return (sock);
}   /*  End Function r_atm_alloc  */

/*UNPUBLISHED_FUNCTION*/
int r_atm_connect (CONST char *addr_string, unsigned int addr_len,
		   struct karma_qos *qos)
/*  [SUMMARY] Make a connection using ATM SVC with AAL5.
    <addr_string> The binary representation of the ATM address to connect to.
    <addr_len> The length of the address string.
    <qos> The Quality of Service parameters. This will be updated to reflect
    the parameters provided. This may be NULL.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [NOTE] This function is not implemented on platforms without ATM support.
    [RETURNS] The file descriptor of the opened connection on success, else -1.
*/
{
    int fd;
    struct sockaddr_atmsvc addr;
    struct atm_qos atmqos;
    struct karma_qos default_qos;
    static char function_name[] = "r_atm_connect";

    /*  Set up addressing structure  */
    if (addr_len < ATM_ESA_LEN)
    {
	fprintf (stderr,
		 "%s: address length: %u is too short, must be at least: %u\n",
		 function_name, addr_len, ATM_ESA_LEN);
	return (-1);
    }
    memset (&addr, 0, sizeof addr);
    addr.sas_family = AF_ATMSVC;
    memcpy (addr.sas_addr.prv, addr_string, ATM_ESA_LEN);
    /*  Set up QOS structure  */
    if (qos == NULL)
    {
	memset (&default_qos, 0, sizeof default_qos);
	default_qos.tx.traffic_class = K_ATM_TRAFFIC_CLASS_UBR;
	default_qos.rx.traffic_class = K_ATM_TRAFFIC_CLASS_UBR;
	qos = &default_qos;
    }
    memset (&atmqos, 0, sizeof atmqos);
    _r_atm_karma_to_atm (&atmqos.txtp, &qos->tx);
    _r_atm_karma_to_atm (&atmqos.rxtp, &qos->rx);
    atmqos.aal = ATM_AAL5;
    if ( ( fd = socket (PF_ATMSVC, SOCK_DGRAM, 0) ) < 0 )
    {
	fprintf (stderr, "%s: error creating ATMSVC socket\t%s\n",
		 function_name, ERRSTRING);
	return (-1);
    }
    if (setsockopt (fd, SOL_ATM, SO_ATMQOS, &atmqos, sizeof atmqos) != 0)
    {
	fprintf (stderr, "%s: error setting ATMSVC socket options\t%s\n",
		 function_name, ERRSTRING);
	close (fd);
	return (-1);
    }
    if (connect (fd, (struct sockaddr *) &addr, sizeof addr) != 0)
    {
	fprintf (stderr, "%s: error connecting\t%s\n",
		 function_name, ERRSTRING);
	close (fd);
	return (-1);
    }
    if ( !_r_atm_get_qos (fd, qos) )
    {
	close (fd);
	return (-1);
    }
    return (fd);
}   /*  End Function r_atm_connect  */

/*UNPUBLISHED_FUNCTION*/
int r_atm_accept (int dock, struct karma_qos *qos)
/*  [SUMMARY] Accept a connection on an ATM dock.
    <dock> The dock.
    <qos> The QoS parameters will be written here.
    [RETURNS] The file descriptor of the opened connection on success, else -1.
*/
{
    int fd;
    /*static char function_name[] = "r_atm_accept";*/

    if ( ( fd = accept (dock, NULL, NULL) ) < 0 ) return (-1);
    if ( !_r_atm_get_qos (fd, qos) )
    {
	close (fd);
	return (-1);
    }
    return (fd);
}   /*  End Function r_atm_accept  */

/*UNPUBLISHED_FUNCTION*/
iaddr r_atm_read (int fd, char *buf, uaddr nbytes)
/*  [SUMMARY] Read bytes from a file descriptor.
    [PURPOSE] This routine is similar to the system <<read(2)>> call, except
    that interrupted system calls are restarted.
    <fd> The file descriptor to read from. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to read.
    [RETURNS] The number of bytes requested on success, the number of bytes
    read on end of file (or closure) and -1 on error.
*/
{
    uaddr total_bytes_read = 0;
    iaddr bytes_read;

    while (total_bytes_read < nbytes)
    {
	errno = 0;
	if ( ( bytes_read = read (fd, buf + total_bytes_read,
				  nbytes - total_bytes_read) ) < 0 )
	{
	    /*  Error  */
	    if (errno != EINTR)
	    {
		/*  Unrecoverable error  */
		return (-1);
	    }
	    continue;
	}
	/*  No error condition  */
	total_bytes_read += bytes_read;
	return (total_bytes_read);
    }
    return (total_bytes_read);
}   /*  End Function r_atm_read  */

/*UNPUBLISHED_FUNCTION*/
iaddr r_atm_write (int fd, CONST char *buf, uaddr nbytes)
/*  [SUMMARY] Write bytes to a file descriptor.
    <fd> The file descriptor to write to. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to write.
    [RETURNS] The number of bytes requested on success, else -1 indicating
    error.
*/
{
    return write (fd, buf, nbytes);
}   /*  End Function r_atm_write  */

/*UNPUBLISHED_FUNCTION*/
flag r_atm_close (int fd)
/*  [SUMMARY] Close a file descriptor.
    <fd> The file descriptor to close.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    if (close (fd) == 0) return (TRUE);
    fprintf (stderr, "Error closing descriptor: %d\t%s\n", fd, ERRSTRING);
    return (FALSE);
}   /*  End Function r_atm_close  */


/*  Private functions follow  */

static void _r_atm_karma_to_atm (struct atm_trafprm *atm,
				 CONST struct karma_traffic_parameters *karma)
/*  [SUMMARY] Convert traffic parameters from Karma to ATM.
    <atm> The ATM traffic parameters will be written here.
    <karma> The Karma traffic parameters.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_r_atm_karma_to_atm";

    switch (karma->traffic_class)
    {
      case K_ATM_TRAFFIC_CLASS_UBR:
	atm->traffic_class = ATM_UBR;
	break;
      case K_ATM_TRAFFIC_CLASS_CBR:
	atm->traffic_class = ATM_CBR;
	break;
      default:
	fprintf (stderr, "%s: illegal traffic class: %u\n",
		 function_name, karma->traffic_class);
	abort ();
	/*break;*/
    }
    atm->min_pcr = karma->rate_min / BYTES_PER_CELL;
    atm->pcr = karma->rate_typ / BYTES_PER_CELL;
    atm->max_pcr = karma->rate_max / BYTES_PER_CELL;
    atm->max_sdu = karma->sdu_max;
    if (atm->max_sdu < 1) atm->max_sdu = DEFAULT_SDU;
    atm->max_cdv = karma->dv_max * BYTES_PER_CELL / 1000;
}   /*  End Function _r_atm_karma_to_atm  */

static void _r_atm_to_karma (struct karma_traffic_parameters *karma,
			     struct atm_trafprm *atm)
/*  [SUMMARY] Convert traffic parameters from ATM to Karma.
    <karma> The Karma traffic parameters. This will be updated.
    <atm> The ATM traffic parameters.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_r_atm_to_karma";

    switch (atm->traffic_class)
    {
      case ATM_UBR:
	karma->traffic_class = K_ATM_TRAFFIC_CLASS_UBR;
	break;
      case ATM_CBR:
	karma->traffic_class = K_ATM_TRAFFIC_CLASS_CBR;
	break;
      default:
	fprintf (stderr, "%s: illegal traffic class: %u\n",
		 function_name, atm->traffic_class);
	abort ();
	/*break;*/
    }
    karma->rate_min = atm->min_pcr * BYTES_PER_CELL;
    karma->rate_typ = atm->pcr * BYTES_PER_CELL;
    karma->rate_max = atm->max_pcr * BYTES_PER_CELL;
    karma->sdu_max = atm->max_sdu;
    karma->dv_max = atm->max_cdv * 1000 / BYTES_PER_CELL;
}   /*  End Function _r_atm_to_karma  */

static flag _r_atm_get_qos (int fd, struct karma_qos *qos)
/*  [SUMMARY] Get the Quality of Service parameters for an ATM connection.
    <fd> The connected socket.
    <qos> The QoS parameters will be written here.
    [RETURNS] TRUE on success, else FALSE to indicate failure.
*/
{
    struct atm_qos atmqos;
    socklen_t size = sizeof atmqos;
    static char function_name[] = "_r_atm_get_qos";

    if (getsockopt (fd, SOL_ATM, SO_ATMQOS, &atmqos, &size) != 0)
    {
	fprintf (stderr, "%s: error getting ATMSVC socket options\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
    _r_atm_to_karma (&qos->tx, &atmqos.txtp);
    _r_atm_to_karma (&qos->rx, &atmqos.rxtp);
    return (TRUE);
}   /*  End Function _r_atm_get_qos  */
