/*LINTLIBRARY*/
/*  misc8.c

    This code provides simple colourmap generation routines.

    Copyright (C) 1998 Richard Gooch  API design and general maintenance
    Copyright (C) 1998 Steven Gibson  Colour tables

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Steven Gibson may be reached my email at  gibson@ras.ucalgary.ca
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for writing simple colour
    values.


    Written by      Richard Gooch   19-MAY-1998

    Last updated by Richard Gooch   19-MAY-1998


*/
#include <stdio.h>
#include <math.h>
#include <karma_cf.h>
#include <karma_a.h>
#include <karma_n.h>
#include <karma.h>

#define LUT_SIZE      256

/* Three color tables: Fire, Spring, and Sea, based upon permutations 
   of the SAOimage BlackBody ("BB") table, called "Fire" here:

	     1.0|      ______
		|     /  /  /
		|    /  /  /
		|  R/ G/ B/
		|  /  /  /
		| /  /  /
		|/  /  /
	      0	+------------
		0         1.0

   These are half of a family of six similar tables in which the order of
   the colors is swapped:

		Rose   Fire		Fire:     RGB
		   \   /		Spring:   GRB
		    \ /			Sea:      GBR
	 Midnight----*----Spring	Ice:      BGR
		    / \			Midnight: BRG
		   /   \		Rose:     RBG
		 Ice   Sea		

   Only those on the right half of the circle are given here, since the
   others can be replicated by reversing and inverting them.  

   The format of the tables below is based upon those found in 
   $KARMAROOT/source/lib/karma/master/cf_misc2.c, where LUT_SIZE = 256

   Steven Gibson, 1998-05-18
*/

static Colour_struct fire[LUT_SIZE] = 
{
    {0.00000, 0.00000, 0.00000},
    {0.00781, 0.00000, 0.00000},
    {0.01562, 0.00000, 0.00000},
    {0.02344, 0.00000, 0.00000},
    {0.03125, 0.00000, 0.00000},
    {0.03906, 0.00000, 0.00000},
    {0.04688, 0.00000, 0.00000},
    {0.05469, 0.00000, 0.00000},
    {0.06250, 0.00000, 0.00000},
    {0.07031, 0.00000, 0.00000},
    {0.07812, 0.00000, 0.00000},
    {0.08594, 0.00000, 0.00000},
    {0.09375, 0.00000, 0.00000},
    {0.10156, 0.00000, 0.00000},
    {0.10938, 0.00000, 0.00000},
    {0.11719, 0.00000, 0.00000},
    {0.12500, 0.00000, 0.00000},
    {0.13281, 0.00000, 0.00000},
    {0.14062, 0.00000, 0.00000},
    {0.14844, 0.00000, 0.00000},
    {0.15625, 0.00000, 0.00000},
    {0.16406, 0.00000, 0.00000},
    {0.17188, 0.00000, 0.00000},
    {0.17969, 0.00000, 0.00000},
    {0.18750, 0.00000, 0.00000},
    {0.19531, 0.00000, 0.00000},
    {0.20312, 0.00000, 0.00000},
    {0.21094, 0.00000, 0.00000},
    {0.21875, 0.00000, 0.00000},
    {0.22656, 0.00000, 0.00000},
    {0.23438, 0.00000, 0.00000},
    {0.24219, 0.00000, 0.00000},
    {0.25000, 0.00000, 0.00000},
    {0.25781, 0.00000, 0.00000},
    {0.26562, 0.00000, 0.00000},
    {0.27344, 0.00000, 0.00000},
    {0.28125, 0.00000, 0.00000},
    {0.28906, 0.00000, 0.00000},
    {0.29688, 0.00000, 0.00000},
    {0.30469, 0.00000, 0.00000},
    {0.31250, 0.00000, 0.00000},
    {0.32031, 0.00000, 0.00000},
    {0.32812, 0.00000, 0.00000},
    {0.33594, 0.00000, 0.00000},
    {0.34375, 0.00000, 0.00000},
    {0.35156, 0.00000, 0.00000},
    {0.35938, 0.00000, 0.00000},
    {0.36719, 0.00000, 0.00000},
    {0.37500, 0.00000, 0.00000},
    {0.38281, 0.00000, 0.00000},
    {0.39062, 0.00000, 0.00000},
    {0.39844, 0.00000, 0.00000},
    {0.40625, 0.00000, 0.00000},
    {0.41406, 0.00000, 0.00000},
    {0.42188, 0.00000, 0.00000},
    {0.42969, 0.00000, 0.00000},
    {0.43750, 0.00000, 0.00000},
    {0.44531, 0.00000, 0.00000},
    {0.45312, 0.00000, 0.00000},
    {0.46094, 0.00000, 0.00000},
    {0.46875, 0.00000, 0.00000},
    {0.47656, 0.00000, 0.00000},
    {0.48438, 0.00000, 0.00000},
    {0.49219, 0.00000, 0.00000},
    {0.50000, 0.00781, 0.00000},
    {0.50781, 0.01562, 0.00000},
    {0.51562, 0.02344, 0.00000},
    {0.52344, 0.03125, 0.00000},
    {0.53125, 0.03906, 0.00000},
    {0.53906, 0.04688, 0.00000},
    {0.54688, 0.05469, 0.00000},
    {0.55469, 0.06250, 0.00000},
    {0.56250, 0.07031, 0.00000},
    {0.57031, 0.07812, 0.00000},
    {0.57812, 0.08594, 0.00000},
    {0.58594, 0.09375, 0.00000},
    {0.59375, 0.10156, 0.00000},
    {0.60156, 0.10938, 0.00000},
    {0.60938, 0.11719, 0.00000},
    {0.61719, 0.12500, 0.00000},
    {0.62500, 0.13281, 0.00000},
    {0.63281, 0.14062, 0.00000},
    {0.64062, 0.14844, 0.00000},
    {0.64844, 0.15625, 0.00000},
    {0.65625, 0.16406, 0.00000},
    {0.66406, 0.17188, 0.00000},
    {0.67188, 0.17969, 0.00000},
    {0.67969, 0.18750, 0.00000},
    {0.68750, 0.19531, 0.00000},
    {0.69531, 0.20312, 0.00000},
    {0.70312, 0.21094, 0.00000},
    {0.71094, 0.21875, 0.00000},
    {0.71875, 0.22656, 0.00000},
    {0.72656, 0.23438, 0.00000},
    {0.73438, 0.24219, 0.00000},
    {0.74219, 0.25000, 0.00000},
    {0.75000, 0.25781, 0.00000},
    {0.75781, 0.26562, 0.00000},
    {0.76562, 0.27344, 0.00000},
    {0.77344, 0.28125, 0.00000},
    {0.78125, 0.28906, 0.00000},
    {0.78906, 0.29688, 0.00000},
    {0.79688, 0.30469, 0.00000},
    {0.80469, 0.31250, 0.00000},
    {0.81250, 0.32031, 0.00000},
    {0.82031, 0.32812, 0.00000},
    {0.82812, 0.33594, 0.00000},
    {0.83594, 0.34375, 0.00000},
    {0.84375, 0.35156, 0.00000},
    {0.85156, 0.35938, 0.00000},
    {0.85938, 0.36719, 0.00000},
    {0.86719, 0.37500, 0.00000},
    {0.87500, 0.38281, 0.00000},
    {0.88281, 0.39062, 0.00000},
    {0.89062, 0.39844, 0.00000},
    {0.89844, 0.40625, 0.00000},
    {0.90625, 0.41406, 0.00000},
    {0.91406, 0.42188, 0.00000},
    {0.92188, 0.42969, 0.00000},
    {0.92969, 0.43750, 0.00000},
    {0.93750, 0.44531, 0.00000},
    {0.94531, 0.45312, 0.00000},
    {0.95312, 0.46094, 0.00000},
    {0.96094, 0.46875, 0.00000},
    {0.96875, 0.47656, 0.00000},
    {0.97656, 0.48438, 0.00000},
    {0.98438, 0.49219, 0.00000},
    {0.99219, 0.50000, 0.00000},
    {1.00000, 0.50781, 0.00781},
    {1.00000, 0.51562, 0.01562},
    {1.00000, 0.52344, 0.02344},
    {1.00000, 0.53125, 0.03125},
    {1.00000, 0.53906, 0.03906},
    {1.00000, 0.54688, 0.04688},
    {1.00000, 0.55469, 0.05469},
    {1.00000, 0.56250, 0.06250},
    {1.00000, 0.57031, 0.07031},
    {1.00000, 0.57812, 0.07812},
    {1.00000, 0.58594, 0.08594},
    {1.00000, 0.59375, 0.09375},
    {1.00000, 0.60156, 0.10156},
    {1.00000, 0.60938, 0.10938},
    {1.00000, 0.61719, 0.11719},
    {1.00000, 0.62500, 0.12500},
    {1.00000, 0.63281, 0.13281},
    {1.00000, 0.64062, 0.14062},
    {1.00000, 0.64844, 0.14844},
    {1.00000, 0.65625, 0.15625},
    {1.00000, 0.66406, 0.16406},
    {1.00000, 0.67188, 0.17188},
    {1.00000, 0.67969, 0.17969},
    {1.00000, 0.68750, 0.18750},
    {1.00000, 0.69531, 0.19531},
    {1.00000, 0.70312, 0.20312},
    {1.00000, 0.71094, 0.21094},
    {1.00000, 0.71875, 0.21875},
    {1.00000, 0.72656, 0.22656},
    {1.00000, 0.73438, 0.23438},
    {1.00000, 0.74219, 0.24219},
    {1.00000, 0.75000, 0.25000},
    {1.00000, 0.75781, 0.25781},
    {1.00000, 0.76562, 0.26562},
    {1.00000, 0.77344, 0.27344},
    {1.00000, 0.78125, 0.28125},
    {1.00000, 0.78906, 0.28906},
    {1.00000, 0.79688, 0.29688},
    {1.00000, 0.80469, 0.30469},
    {1.00000, 0.81250, 0.31250},
    {1.00000, 0.82031, 0.32031},
    {1.00000, 0.82812, 0.32812},
    {1.00000, 0.83594, 0.33594},
    {1.00000, 0.84375, 0.34375},
    {1.00000, 0.85156, 0.35156},
    {1.00000, 0.85938, 0.35938},
    {1.00000, 0.86719, 0.36719},
    {1.00000, 0.87500, 0.37500},
    {1.00000, 0.88281, 0.38281},
    {1.00000, 0.89062, 0.39062},
    {1.00000, 0.89844, 0.39844},
    {1.00000, 0.90625, 0.40625},
    {1.00000, 0.91406, 0.41406},
    {1.00000, 0.92188, 0.42188},
    {1.00000, 0.92969, 0.42969},
    {1.00000, 0.93750, 0.43750},
    {1.00000, 0.94531, 0.44531},
    {1.00000, 0.95312, 0.45312},
    {1.00000, 0.96094, 0.46094},
    {1.00000, 0.96875, 0.46875},
    {1.00000, 0.97656, 0.47656},
    {1.00000, 0.98438, 0.48438},
    {1.00000, 0.99219, 0.49219},
    {1.00000, 1.00000, 0.50000},
    {1.00000, 1.00000, 0.50781},
    {1.00000, 1.00000, 0.51562},
    {1.00000, 1.00000, 0.52344},
    {1.00000, 1.00000, 0.53125},
    {1.00000, 1.00000, 0.53906},
    {1.00000, 1.00000, 0.54688},
    {1.00000, 1.00000, 0.55469},
    {1.00000, 1.00000, 0.56250},
    {1.00000, 1.00000, 0.57031},
    {1.00000, 1.00000, 0.57812},
    {1.00000, 1.00000, 0.58594},
    {1.00000, 1.00000, 0.59375},
    {1.00000, 1.00000, 0.60156},
    {1.00000, 1.00000, 0.60938},
    {1.00000, 1.00000, 0.61719},
    {1.00000, 1.00000, 0.62500},
    {1.00000, 1.00000, 0.63281},
    {1.00000, 1.00000, 0.64062},
    {1.00000, 1.00000, 0.64844},
    {1.00000, 1.00000, 0.65625},
    {1.00000, 1.00000, 0.66406},
    {1.00000, 1.00000, 0.67188},
    {1.00000, 1.00000, 0.67969},
    {1.00000, 1.00000, 0.68750},
    {1.00000, 1.00000, 0.69531},
    {1.00000, 1.00000, 0.70312},
    {1.00000, 1.00000, 0.71094},
    {1.00000, 1.00000, 0.71875},
    {1.00000, 1.00000, 0.72656},
    {1.00000, 1.00000, 0.73438},
    {1.00000, 1.00000, 0.74219},
    {1.00000, 1.00000, 0.75000},
    {1.00000, 1.00000, 0.75781},
    {1.00000, 1.00000, 0.76562},
    {1.00000, 1.00000, 0.77344},
    {1.00000, 1.00000, 0.78125},
    {1.00000, 1.00000, 0.78906},
    {1.00000, 1.00000, 0.79688},
    {1.00000, 1.00000, 0.80469},
    {1.00000, 1.00000, 0.81250},
    {1.00000, 1.00000, 0.82031},
    {1.00000, 1.00000, 0.82812},
    {1.00000, 1.00000, 0.83594},
    {1.00000, 1.00000, 0.84375},
    {1.00000, 1.00000, 0.85156},
    {1.00000, 1.00000, 0.85938},
    {1.00000, 1.00000, 0.86719},
    {1.00000, 1.00000, 0.87500},
    {1.00000, 1.00000, 0.88281},
    {1.00000, 1.00000, 0.89062},
    {1.00000, 1.00000, 0.89844},
    {1.00000, 1.00000, 0.90625},
    {1.00000, 1.00000, 0.91406},
    {1.00000, 1.00000, 0.92188},
    {1.00000, 1.00000, 0.92969},
    {1.00000, 1.00000, 0.93750},
    {1.00000, 1.00000, 0.94531},
    {1.00000, 1.00000, 0.95312},
    {1.00000, 1.00000, 0.96094},
    {1.00000, 1.00000, 0.96875},
    {1.00000, 1.00000, 0.97656},
    {1.00000, 1.00000, 0.98438},
    {1.00000, 1.00000, 0.99219},
    {1.00000, 1.00000, 1.00000},
};

static Colour_struct spring[LUT_SIZE] = 
{
    {0.00000, 0.00000, 0.00000},
    {0.00000, 0.00781, 0.00000},
    {0.00000, 0.01562, 0.00000},
    {0.00000, 0.02344, 0.00000},
    {0.00000, 0.03125, 0.00000},
    {0.00000, 0.03906, 0.00000},
    {0.00000, 0.04688, 0.00000},
    {0.00000, 0.05469, 0.00000},
    {0.00000, 0.06250, 0.00000},
    {0.00000, 0.07031, 0.00000},
    {0.00000, 0.07812, 0.00000},
    {0.00000, 0.08594, 0.00000},
    {0.00000, 0.09375, 0.00000},
    {0.00000, 0.10156, 0.00000},
    {0.00000, 0.10938, 0.00000},
    {0.00000, 0.11719, 0.00000},
    {0.00000, 0.12500, 0.00000},
    {0.00000, 0.13281, 0.00000},
    {0.00000, 0.14062, 0.00000},
    {0.00000, 0.14844, 0.00000},
    {0.00000, 0.15625, 0.00000},
    {0.00000, 0.16406, 0.00000},
    {0.00000, 0.17188, 0.00000},
    {0.00000, 0.17969, 0.00000},
    {0.00000, 0.18750, 0.00000},
    {0.00000, 0.19531, 0.00000},
    {0.00000, 0.20312, 0.00000},
    {0.00000, 0.21094, 0.00000},
    {0.00000, 0.21875, 0.00000},
    {0.00000, 0.22656, 0.00000},
    {0.00000, 0.23438, 0.00000},
    {0.00000, 0.24219, 0.00000},
    {0.00000, 0.25000, 0.00000},
    {0.00000, 0.25781, 0.00000},
    {0.00000, 0.26562, 0.00000},
    {0.00000, 0.27344, 0.00000},
    {0.00000, 0.28125, 0.00000},
    {0.00000, 0.28906, 0.00000},
    {0.00000, 0.29688, 0.00000},
    {0.00000, 0.30469, 0.00000},
    {0.00000, 0.31250, 0.00000},
    {0.00000, 0.32031, 0.00000},
    {0.00000, 0.32812, 0.00000},
    {0.00000, 0.33594, 0.00000},
    {0.00000, 0.34375, 0.00000},
    {0.00000, 0.35156, 0.00000},
    {0.00000, 0.35938, 0.00000},
    {0.00000, 0.36719, 0.00000},
    {0.00000, 0.37500, 0.00000},
    {0.00000, 0.38281, 0.00000},
    {0.00000, 0.39062, 0.00000},
    {0.00000, 0.39844, 0.00000},
    {0.00000, 0.40625, 0.00000},
    {0.00000, 0.41406, 0.00000},
    {0.00000, 0.42188, 0.00000},
    {0.00000, 0.42969, 0.00000},
    {0.00000, 0.43750, 0.00000},
    {0.00000, 0.44531, 0.00000},
    {0.00000, 0.45312, 0.00000},
    {0.00000, 0.46094, 0.00000},
    {0.00000, 0.46875, 0.00000},
    {0.00000, 0.47656, 0.00000},
    {0.00000, 0.48438, 0.00000},
    {0.00000, 0.49219, 0.00000},
    {0.00781, 0.50000, 0.00000},
    {0.01562, 0.50781, 0.00000},
    {0.02344, 0.51562, 0.00000},
    {0.03125, 0.52344, 0.00000},
    {0.03906, 0.53125, 0.00000},
    {0.04688, 0.53906, 0.00000},
    {0.05469, 0.54688, 0.00000},
    {0.06250, 0.55469, 0.00000},
    {0.07031, 0.56250, 0.00000},
    {0.07812, 0.57031, 0.00000},
    {0.08594, 0.57812, 0.00000},
    {0.09375, 0.58594, 0.00000},
    {0.10156, 0.59375, 0.00000},
    {0.10938, 0.60156, 0.00000},
    {0.11719, 0.60938, 0.00000},
    {0.12500, 0.61719, 0.00000},
    {0.13281, 0.62500, 0.00000},
    {0.14062, 0.63281, 0.00000},
    {0.14844, 0.64062, 0.00000},
    {0.15625, 0.64844, 0.00000},
    {0.16406, 0.65625, 0.00000},
    {0.17188, 0.66406, 0.00000},
    {0.17969, 0.67188, 0.00000},
    {0.18750, 0.67969, 0.00000},
    {0.19531, 0.68750, 0.00000},
    {0.20312, 0.69531, 0.00000},
    {0.21094, 0.70312, 0.00000},
    {0.21875, 0.71094, 0.00000},
    {0.22656, 0.71875, 0.00000},
    {0.23438, 0.72656, 0.00000},
    {0.24219, 0.73438, 0.00000},
    {0.25000, 0.74219, 0.00000},
    {0.25781, 0.75000, 0.00000},
    {0.26562, 0.75781, 0.00000},
    {0.27344, 0.76562, 0.00000},
    {0.28125, 0.77344, 0.00000},
    {0.28906, 0.78125, 0.00000},
    {0.29688, 0.78906, 0.00000},
    {0.30469, 0.79688, 0.00000},
    {0.31250, 0.80469, 0.00000},
    {0.32031, 0.81250, 0.00000},
    {0.32812, 0.82031, 0.00000},
    {0.33594, 0.82812, 0.00000},
    {0.34375, 0.83594, 0.00000},
    {0.35156, 0.84375, 0.00000},
    {0.35938, 0.85156, 0.00000},
    {0.36719, 0.85938, 0.00000},
    {0.37500, 0.86719, 0.00000},
    {0.38281, 0.87500, 0.00000},
    {0.39062, 0.88281, 0.00000},
    {0.39844, 0.89062, 0.00000},
    {0.40625, 0.89844, 0.00000},
    {0.41406, 0.90625, 0.00000},
    {0.42188, 0.91406, 0.00000},
    {0.42969, 0.92188, 0.00000},
    {0.43750, 0.92969, 0.00000},
    {0.44531, 0.93750, 0.00000},
    {0.45312, 0.94531, 0.00000},
    {0.46094, 0.95312, 0.00000},
    {0.46875, 0.96094, 0.00000},
    {0.47656, 0.96875, 0.00000},
    {0.48438, 0.97656, 0.00000},
    {0.49219, 0.98438, 0.00000},
    {0.50000, 0.99219, 0.00000},
    {0.50781, 1.00000, 0.00781},
    {0.51562, 1.00000, 0.01562},
    {0.52344, 1.00000, 0.02344},
    {0.53125, 1.00000, 0.03125},
    {0.53906, 1.00000, 0.03906},
    {0.54688, 1.00000, 0.04688},
    {0.55469, 1.00000, 0.05469},
    {0.56250, 1.00000, 0.06250},
    {0.57031, 1.00000, 0.07031},
    {0.57812, 1.00000, 0.07812},
    {0.58594, 1.00000, 0.08594},
    {0.59375, 1.00000, 0.09375},
    {0.60156, 1.00000, 0.10156},
    {0.60938, 1.00000, 0.10938},
    {0.61719, 1.00000, 0.11719},
    {0.62500, 1.00000, 0.12500},
    {0.63281, 1.00000, 0.13281},
    {0.64062, 1.00000, 0.14062},
    {0.64844, 1.00000, 0.14844},
    {0.65625, 1.00000, 0.15625},
    {0.66406, 1.00000, 0.16406},
    {0.67188, 1.00000, 0.17188},
    {0.67969, 1.00000, 0.17969},
    {0.68750, 1.00000, 0.18750},
    {0.69531, 1.00000, 0.19531},
    {0.70312, 1.00000, 0.20312},
    {0.71094, 1.00000, 0.21094},
    {0.71875, 1.00000, 0.21875},
    {0.72656, 1.00000, 0.22656},
    {0.73438, 1.00000, 0.23438},
    {0.74219, 1.00000, 0.24219},
    {0.75000, 1.00000, 0.25000},
    {0.75781, 1.00000, 0.25781},
    {0.76562, 1.00000, 0.26562},
    {0.77344, 1.00000, 0.27344},
    {0.78125, 1.00000, 0.28125},
    {0.78906, 1.00000, 0.28906},
    {0.79688, 1.00000, 0.29688},
    {0.80469, 1.00000, 0.30469},
    {0.81250, 1.00000, 0.31250},
    {0.82031, 1.00000, 0.32031},
    {0.82812, 1.00000, 0.32812},
    {0.83594, 1.00000, 0.33594},
    {0.84375, 1.00000, 0.34375},
    {0.85156, 1.00000, 0.35156},
    {0.85938, 1.00000, 0.35938},
    {0.86719, 1.00000, 0.36719},
    {0.87500, 1.00000, 0.37500},
    {0.88281, 1.00000, 0.38281},
    {0.89062, 1.00000, 0.39062},
    {0.89844, 1.00000, 0.39844},
    {0.90625, 1.00000, 0.40625},
    {0.91406, 1.00000, 0.41406},
    {0.92188, 1.00000, 0.42188},
    {0.92969, 1.00000, 0.42969},
    {0.93750, 1.00000, 0.43750},
    {0.94531, 1.00000, 0.44531},
    {0.95312, 1.00000, 0.45312},
    {0.96094, 1.00000, 0.46094},
    {0.96875, 1.00000, 0.46875},
    {0.97656, 1.00000, 0.47656},
    {0.98438, 1.00000, 0.48438},
    {0.99219, 1.00000, 0.49219},
    {1.00000, 1.00000, 0.50000},
    {1.00000, 1.00000, 0.50781},
    {1.00000, 1.00000, 0.51562},
    {1.00000, 1.00000, 0.52344},
    {1.00000, 1.00000, 0.53125},
    {1.00000, 1.00000, 0.53906},
    {1.00000, 1.00000, 0.54688},
    {1.00000, 1.00000, 0.55469},
    {1.00000, 1.00000, 0.56250},
    {1.00000, 1.00000, 0.57031},
    {1.00000, 1.00000, 0.57812},
    {1.00000, 1.00000, 0.58594},
    {1.00000, 1.00000, 0.59375},
    {1.00000, 1.00000, 0.60156},
    {1.00000, 1.00000, 0.60938},
    {1.00000, 1.00000, 0.61719},
    {1.00000, 1.00000, 0.62500},
    {1.00000, 1.00000, 0.63281},
    {1.00000, 1.00000, 0.64062},
    {1.00000, 1.00000, 0.64844},
    {1.00000, 1.00000, 0.65625},
    {1.00000, 1.00000, 0.66406},
    {1.00000, 1.00000, 0.67188},
    {1.00000, 1.00000, 0.67969},
    {1.00000, 1.00000, 0.68750},
    {1.00000, 1.00000, 0.69531},
    {1.00000, 1.00000, 0.70312},
    {1.00000, 1.00000, 0.71094},
    {1.00000, 1.00000, 0.71875},
    {1.00000, 1.00000, 0.72656},
    {1.00000, 1.00000, 0.73438},
    {1.00000, 1.00000, 0.74219},
    {1.00000, 1.00000, 0.75000},
    {1.00000, 1.00000, 0.75781},
    {1.00000, 1.00000, 0.76562},
    {1.00000, 1.00000, 0.77344},
    {1.00000, 1.00000, 0.78125},
    {1.00000, 1.00000, 0.78906},
    {1.00000, 1.00000, 0.79688},
    {1.00000, 1.00000, 0.80469},
    {1.00000, 1.00000, 0.81250},
    {1.00000, 1.00000, 0.82031},
    {1.00000, 1.00000, 0.82812},
    {1.00000, 1.00000, 0.83594},
    {1.00000, 1.00000, 0.84375},
    {1.00000, 1.00000, 0.85156},
    {1.00000, 1.00000, 0.85938},
    {1.00000, 1.00000, 0.86719},
    {1.00000, 1.00000, 0.87500},
    {1.00000, 1.00000, 0.88281},
    {1.00000, 1.00000, 0.89062},
    {1.00000, 1.00000, 0.89844},
    {1.00000, 1.00000, 0.90625},
    {1.00000, 1.00000, 0.91406},
    {1.00000, 1.00000, 0.92188},
    {1.00000, 1.00000, 0.92969},
    {1.00000, 1.00000, 0.93750},
    {1.00000, 1.00000, 0.94531},
    {1.00000, 1.00000, 0.95312},
    {1.00000, 1.00000, 0.96094},
    {1.00000, 1.00000, 0.96875},
    {1.00000, 1.00000, 0.97656},
    {1.00000, 1.00000, 0.98438},
    {1.00000, 1.00000, 0.99219},
    {1.00000, 1.00000, 1.00000},
};

static Colour_struct sea[LUT_SIZE] = 
{
    {0.00000, 0.00000, 0.00000},
    {0.00000, 0.00781, 0.00000},
    {0.00000, 0.01562, 0.00000},
    {0.00000, 0.02344, 0.00000},
    {0.00000, 0.03125, 0.00000},
    {0.00000, 0.03906, 0.00000},
    {0.00000, 0.04688, 0.00000},
    {0.00000, 0.05469, 0.00000},
    {0.00000, 0.06250, 0.00000},
    {0.00000, 0.07031, 0.00000},
    {0.00000, 0.07812, 0.00000},
    {0.00000, 0.08594, 0.00000},
    {0.00000, 0.09375, 0.00000},
    {0.00000, 0.10156, 0.00000},
    {0.00000, 0.10938, 0.00000},
    {0.00000, 0.11719, 0.00000},
    {0.00000, 0.12500, 0.00000},
    {0.00000, 0.13281, 0.00000},
    {0.00000, 0.14062, 0.00000},
    {0.00000, 0.14844, 0.00000},
    {0.00000, 0.15625, 0.00000},
    {0.00000, 0.16406, 0.00000},
    {0.00000, 0.17188, 0.00000},
    {0.00000, 0.17969, 0.00000},
    {0.00000, 0.18750, 0.00000},
    {0.00000, 0.19531, 0.00000},
    {0.00000, 0.20312, 0.00000},
    {0.00000, 0.21094, 0.00000},
    {0.00000, 0.21875, 0.00000},
    {0.00000, 0.22656, 0.00000},
    {0.00000, 0.23438, 0.00000},
    {0.00000, 0.24219, 0.00000},
    {0.00000, 0.25000, 0.00000},
    {0.00000, 0.25781, 0.00000},
    {0.00000, 0.26562, 0.00000},
    {0.00000, 0.27344, 0.00000},
    {0.00000, 0.28125, 0.00000},
    {0.00000, 0.28906, 0.00000},
    {0.00000, 0.29688, 0.00000},
    {0.00000, 0.30469, 0.00000},
    {0.00000, 0.31250, 0.00000},
    {0.00000, 0.32031, 0.00000},
    {0.00000, 0.32812, 0.00000},
    {0.00000, 0.33594, 0.00000},
    {0.00000, 0.34375, 0.00000},
    {0.00000, 0.35156, 0.00000},
    {0.00000, 0.35938, 0.00000},
    {0.00000, 0.36719, 0.00000},
    {0.00000, 0.37500, 0.00000},
    {0.00000, 0.38281, 0.00000},
    {0.00000, 0.39062, 0.00000},
    {0.00000, 0.39844, 0.00000},
    {0.00000, 0.40625, 0.00000},
    {0.00000, 0.41406, 0.00000},
    {0.00000, 0.42188, 0.00000},
    {0.00000, 0.42969, 0.00000},
    {0.00000, 0.43750, 0.00000},
    {0.00000, 0.44531, 0.00000},
    {0.00000, 0.45312, 0.00000},
    {0.00000, 0.46094, 0.00000},
    {0.00000, 0.46875, 0.00000},
    {0.00000, 0.47656, 0.00000},
    {0.00000, 0.48438, 0.00000},
    {0.00000, 0.49219, 0.00000},
    {0.00000, 0.50000, 0.00781},
    {0.00000, 0.50781, 0.01562},
    {0.00000, 0.51562, 0.02344},
    {0.00000, 0.52344, 0.03125},
    {0.00000, 0.53125, 0.03906},
    {0.00000, 0.53906, 0.04688},
    {0.00000, 0.54688, 0.05469},
    {0.00000, 0.55469, 0.06250},
    {0.00000, 0.56250, 0.07031},
    {0.00000, 0.57031, 0.07812},
    {0.00000, 0.57812, 0.08594},
    {0.00000, 0.58594, 0.09375},
    {0.00000, 0.59375, 0.10156},
    {0.00000, 0.60156, 0.10938},
    {0.00000, 0.60938, 0.11719},
    {0.00000, 0.61719, 0.12500},
    {0.00000, 0.62500, 0.13281},
    {0.00000, 0.63281, 0.14062},
    {0.00000, 0.64062, 0.14844},
    {0.00000, 0.64844, 0.15625},
    {0.00000, 0.65625, 0.16406},
    {0.00000, 0.66406, 0.17188},
    {0.00000, 0.67188, 0.17969},
    {0.00000, 0.67969, 0.18750},
    {0.00000, 0.68750, 0.19531},
    {0.00000, 0.69531, 0.20312},
    {0.00000, 0.70312, 0.21094},
    {0.00000, 0.71094, 0.21875},
    {0.00000, 0.71875, 0.22656},
    {0.00000, 0.72656, 0.23438},
    {0.00000, 0.73438, 0.24219},
    {0.00000, 0.74219, 0.25000},
    {0.00000, 0.75000, 0.25781},
    {0.00000, 0.75781, 0.26562},
    {0.00000, 0.76562, 0.27344},
    {0.00000, 0.77344, 0.28125},
    {0.00000, 0.78125, 0.28906},
    {0.00000, 0.78906, 0.29688},
    {0.00000, 0.79688, 0.30469},
    {0.00000, 0.80469, 0.31250},
    {0.00000, 0.81250, 0.32031},
    {0.00000, 0.82031, 0.32812},
    {0.00000, 0.82812, 0.33594},
    {0.00000, 0.83594, 0.34375},
    {0.00000, 0.84375, 0.35156},
    {0.00000, 0.85156, 0.35938},
    {0.00000, 0.85938, 0.36719},
    {0.00000, 0.86719, 0.37500},
    {0.00000, 0.87500, 0.38281},
    {0.00000, 0.88281, 0.39062},
    {0.00000, 0.89062, 0.39844},
    {0.00000, 0.89844, 0.40625},
    {0.00000, 0.90625, 0.41406},
    {0.00000, 0.91406, 0.42188},
    {0.00000, 0.92188, 0.42969},
    {0.00000, 0.92969, 0.43750},
    {0.00000, 0.93750, 0.44531},
    {0.00000, 0.94531, 0.45312},
    {0.00000, 0.95312, 0.46094},
    {0.00000, 0.96094, 0.46875},
    {0.00000, 0.96875, 0.47656},
    {0.00000, 0.97656, 0.48438},
    {0.00000, 0.98438, 0.49219},
    {0.00000, 0.99219, 0.50000},
    {0.00781, 1.00000, 0.50781},
    {0.01562, 1.00000, 0.51562},
    {0.02344, 1.00000, 0.52344},
    {0.03125, 1.00000, 0.53125},
    {0.03906, 1.00000, 0.53906},
    {0.04688, 1.00000, 0.54688},
    {0.05469, 1.00000, 0.55469},
    {0.06250, 1.00000, 0.56250},
    {0.07031, 1.00000, 0.57031},
    {0.07812, 1.00000, 0.57812},
    {0.08594, 1.00000, 0.58594},
    {0.09375, 1.00000, 0.59375},
    {0.10156, 1.00000, 0.60156},
    {0.10938, 1.00000, 0.60938},
    {0.11719, 1.00000, 0.61719},
    {0.12500, 1.00000, 0.62500},
    {0.13281, 1.00000, 0.63281},
    {0.14062, 1.00000, 0.64062},
    {0.14844, 1.00000, 0.64844},
    {0.15625, 1.00000, 0.65625},
    {0.16406, 1.00000, 0.66406},
    {0.17188, 1.00000, 0.67188},
    {0.17969, 1.00000, 0.67969},
    {0.18750, 1.00000, 0.68750},
    {0.19531, 1.00000, 0.69531},
    {0.20312, 1.00000, 0.70312},
    {0.21094, 1.00000, 0.71094},
    {0.21875, 1.00000, 0.71875},
    {0.22656, 1.00000, 0.72656},
    {0.23438, 1.00000, 0.73438},
    {0.24219, 1.00000, 0.74219},
    {0.25000, 1.00000, 0.75000},
    {0.25781, 1.00000, 0.75781},
    {0.26562, 1.00000, 0.76562},
    {0.27344, 1.00000, 0.77344},
    {0.28125, 1.00000, 0.78125},
    {0.28906, 1.00000, 0.78906},
    {0.29688, 1.00000, 0.79688},
    {0.30469, 1.00000, 0.80469},
    {0.31250, 1.00000, 0.81250},
    {0.32031, 1.00000, 0.82031},
    {0.32812, 1.00000, 0.82812},
    {0.33594, 1.00000, 0.83594},
    {0.34375, 1.00000, 0.84375},
    {0.35156, 1.00000, 0.85156},
    {0.35938, 1.00000, 0.85938},
    {0.36719, 1.00000, 0.86719},
    {0.37500, 1.00000, 0.87500},
    {0.38281, 1.00000, 0.88281},
    {0.39062, 1.00000, 0.89062},
    {0.39844, 1.00000, 0.89844},
    {0.40625, 1.00000, 0.90625},
    {0.41406, 1.00000, 0.91406},
    {0.42188, 1.00000, 0.92188},
    {0.42969, 1.00000, 0.92969},
    {0.43750, 1.00000, 0.93750},
    {0.44531, 1.00000, 0.94531},
    {0.45312, 1.00000, 0.95312},
    {0.46094, 1.00000, 0.96094},
    {0.46875, 1.00000, 0.96875},
    {0.47656, 1.00000, 0.97656},
    {0.48438, 1.00000, 0.98438},
    {0.49219, 1.00000, 0.99219},
    {0.50000, 1.00000, 1.00000},
    {0.50781, 1.00000, 1.00000},
    {0.51562, 1.00000, 1.00000},
    {0.52344, 1.00000, 1.00000},
    {0.53125, 1.00000, 1.00000},
    {0.53906, 1.00000, 1.00000},
    {0.54688, 1.00000, 1.00000},
    {0.55469, 1.00000, 1.00000},
    {0.56250, 1.00000, 1.00000},
    {0.57031, 1.00000, 1.00000},
    {0.57812, 1.00000, 1.00000},
    {0.58594, 1.00000, 1.00000},
    {0.59375, 1.00000, 1.00000},
    {0.60156, 1.00000, 1.00000},
    {0.60938, 1.00000, 1.00000},
    {0.61719, 1.00000, 1.00000},
    {0.62500, 1.00000, 1.00000},
    {0.63281, 1.00000, 1.00000},
    {0.64062, 1.00000, 1.00000},
    {0.64844, 1.00000, 1.00000},
    {0.65625, 1.00000, 1.00000},
    {0.66406, 1.00000, 1.00000},
    {0.67188, 1.00000, 1.00000},
    {0.67969, 1.00000, 1.00000},
    {0.68750, 1.00000, 1.00000},
    {0.69531, 1.00000, 1.00000},
    {0.70312, 1.00000, 1.00000},
    {0.71094, 1.00000, 1.00000},
    {0.71875, 1.00000, 1.00000},
    {0.72656, 1.00000, 1.00000},
    {0.73438, 1.00000, 1.00000},
    {0.74219, 1.00000, 1.00000},
    {0.75000, 1.00000, 1.00000},
    {0.75781, 1.00000, 1.00000},
    {0.76562, 1.00000, 1.00000},
    {0.77344, 1.00000, 1.00000},
    {0.78125, 1.00000, 1.00000},
    {0.78906, 1.00000, 1.00000},
    {0.79688, 1.00000, 1.00000},
    {0.80469, 1.00000, 1.00000},
    {0.81250, 1.00000, 1.00000},
    {0.82031, 1.00000, 1.00000},
    {0.82812, 1.00000, 1.00000},
    {0.83594, 1.00000, 1.00000},
    {0.84375, 1.00000, 1.00000},
    {0.85156, 1.00000, 1.00000},
    {0.85938, 1.00000, 1.00000},
    {0.86719, 1.00000, 1.00000},
    {0.87500, 1.00000, 1.00000},
    {0.88281, 1.00000, 1.00000},
    {0.89062, 1.00000, 1.00000},
    {0.89844, 1.00000, 1.00000},
    {0.90625, 1.00000, 1.00000},
    {0.91406, 1.00000, 1.00000},
    {0.92188, 1.00000, 1.00000},
    {0.92969, 1.00000, 1.00000},
    {0.93750, 1.00000, 1.00000},
    {0.94531, 1.00000, 1.00000},
    {0.95312, 1.00000, 1.00000},
    {0.96094, 1.00000, 1.00000},
    {0.96875, 1.00000, 1.00000},
    {0.97656, 1.00000, 1.00000},
    {0.98438, 1.00000, 1.00000},
    {0.99219, 1.00000, 1.00000},
    {1.00000, 1.00000, 1.00000},
};

/*EXPERIMENTAL_FUNCTION*/
void cf_fire (unsigned int num_cells, unsigned short *reds,
	      unsigned short *greens, unsigned short *blues,
	      unsigned int stride, double x, double y, void *var_param)
/*  [SUMMARY] Compute a "Fire" colourmap.
    <num_cells> The number of colour cells to modify.
    <reds> The red intensity values.
    <greens> The green intensity values.
    <blues> The blue intensity values.
    <stride> The stride (in unsigned shorts) between intensity values.
    <x> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <y> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <var_param> A parameter used to compute the colour values. Ignored.
    [RETURNS] Nothing.
*/
{
    cf_colour_table (num_cells, reds, greens, blues, stride, x, y, fire,
		     LUT_SIZE);
}   /*  End Function cf_fire  */

/*EXPERIMENTAL_FUNCTION*/
void cf_spring (unsigned int num_cells, unsigned short *reds,
		unsigned short *greens, unsigned short *blues,
		unsigned int stride, double x, double y, void *var_param)
/*  [SUMMARY] Compute a "Spring" colourmap.
    <num_cells> The number of colour cells to modify.
    <reds> The red intensity values.
    <greens> The green intensity values.
    <blues> The blue intensity values.
    <stride> The stride (in unsigned shorts) between intensity values.
    <x> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <y> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <var_param> A parameter used to compute the colour values. Ignored.
    [RETURNS] Nothing.
*/
{
    cf_colour_table (num_cells, reds, greens, blues, stride, x, y, spring,
		     LUT_SIZE);
}   /*  End Function cf_spring  */

/*EXPERIMENTAL_FUNCTION*/
void cf_sea (unsigned int num_cells, unsigned short *reds,
	     unsigned short *greens, unsigned short *blues,
	     unsigned int stride, double x, double y, void *var_param)
/*  [SUMMARY] Compute a "Sea" colourmap.
    <num_cells> The number of colour cells to modify.
    <reds> The red intensity values.
    <greens> The green intensity values.
    <blues> The blue intensity values.
    <stride> The stride (in unsigned shorts) between intensity values.
    <x> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <y> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <var_param> A parameter used to compute the colour values. Ignored.
    [RETURNS] Nothing.
*/
{
    cf_colour_table (num_cells, reds, greens, blues, stride, x, y, sea,
		     LUT_SIZE);
}   /*  End Function cf_sea  */
