/*LINTLIBRARY*/
/*  descriptor.c

    This code provides Channel objects.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the creation, manipulation and
  deletion of channel objects.


    Written by      Richard Gooch   14-AUG-1997: Copied bits from main.c

    Updated by      Richard Gooch   15-AUG-1997

    Updated by      Richard Gooch   9-SEP-1997: Supported reading for non-disc
  files opened with <ch_open_file>.

    Updated by      Richard Gooch   17-NOV-1997: Bug fix in <ch_test_for_mmap>:
  did not trap NULL hook pointer.

    Updated by      Richard Gooch   24-DEC-1997: Created experimental
  <ch_get_buffer_sizes> and <ch_set_buffer_sizes> functions.

    Updated by      Richard Gooch   17-FEB-1997: Created experimental
  <ch_udp_alloc> and <ch_udp_connect> functions.

    Updated by      Richard Gooch   19-FEB-1998: Created freelist. Created
  experimental <ch_atm_alloc> and <ch_atm_connect> functions.

    Updated by      Richard Gooch   22-FEB-1998: Removed casts for memory
  functions.

    Updated by      Richard Gooch   26-FEB-1998: Created <skip_func>.

    Updated by      Richard Gooch   23-APR-1998: Supported bzip2 decompression.

    Updated by      Richard Gooch   12-MAY-1998: Worked around IRIX bug when
  getting bytes readable on pipes. Moved <ch_get_descriptor> to generic.c and
  implemented <get_descriptor> hook function.

    Updated by      Richard Gooch   14-MAY-1998: Created experimental function
  <ch_create_socketpair>.

    Updated by      Richard Gooch   18-MAY-1998: Pretend character files are
  local connections under Solaris.

    Updated by      Richard Gooch   23-SEP-1998: Created experimental function
  <ch_set_nodelay>.

    Updated by      Richard Gooch   12-DEC-1998: Added timeout when accepting
  connections. Created experimental function <ch_get_qos>.

    Updated by      Richard Gooch   25-DEC-1998: Made use of <r_atm_read>,
  <r_atm_write> and <r_atm_close>.

    Updated by      Richard Gooch   8-FEB-1999: Fixed bug in function
  <_ch_read_datagram>: small reads were draining the buffer.

    Updated by      Richard Gooch   13-FEB-1999: Coped with systems which
  default to stupid text mode when opening files.

    Updated by      Richard Gooch   26-FEB-1999: Made use of new experimental
  function <r_open_stdin2>.

    Updated by      Richard Gooch   5-APR-1999: Fixed breakage due to Linux
  glibc2 port.

    Updated by      Richard Gooch   18-AUG-1999: Took account of change to read
  hook interface.

    Updated by      Richard Gooch   23-JAN-2000: Supported 's' option in
  <ch_open_file>.

    Updated by      Richard Gooch   26-JAN-2000: Supported 'l' and 'L' options
  in <ch_open_file>. Documentation fix.

    Updated by      Richard Gooch   14-FEB-2000: Fixed bug in allocating last
  entry of freelist.

    Updated by      Richard Gooch   17-FEB-2000: Improved diagnostics.

    Last updated by Richard Gooch   9-JUN-2000: Only define __EXTENSIONS__ for
  OS_Solaris.


*/
#define __STRICT_ANSI__
#ifdef OS_Linux                  /*  Breaks IRIX 6.2                         */
#  define _POSIX_C_SOURCE 199506 /*  Linux/glibc 2: struct sigaction         */
#endif
#ifdef OS_Solaris
#  define __EXTENSIONS__         /*  Needed for Solaris 2: struct itimerval  */
#endif
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <os.h>
#if defined(HAS_SIGNALS) && defined(HAS_ALARM)
#  define CAN_TIMEOUT
#  include <signal.h>
#  ifdef HAS_ITIMER
#    include <sys/time.h>
#  endif
#endif
#include <karma.h>
#include <karma_ch.h>
#include <karma_ch_hooks.h>
#include <karma_m.h>
#include <karma_r.h>
#include <karma_a.h>

#define ALLOC_SIZE 64
#define USED_MAGIC_NUMBER 65487347
#define FREE_MAGIC_NUMBER 1459872432

#define VERIFY_USED(info) if (info == NULL) \
{fprintf (stderr, "NULL channel subclass info passed\n"); \
 a_prog_bug (function_name); } \
if (info->magic_number != USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid channel subclass object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FREE(info) if (info == NULL) \
{fprintf (stderr, "NULL channel subclass info passed\n"); \
 a_prog_bug (function_name); } \
if (info->magic_number != FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed channel subclass object\n"); \
 a_prog_bug (function_name); }

#define CONNECTION_BUF_SIZE 4096    /*  Bytes    */
#define DEFAULT_BLOCK_SIZE  4096    /*  Bytes    */
#define UDP_PACKET_SIZE     65507   /*  Bytes    */
#define TIMEOUT             5       /*  Seconds  */

#define READ (*info->read)
#define WRITE (*info->write)

#define CHANNEL_TYPE_UNDEFINED        0
#define CHANNEL_TYPE_CONNECTION       1
#define CHANNEL_TYPE_DOCK_STREAM      2
#define CHANNEL_TYPE_DOCK_UDP         3
#define CHANNEL_TYPE_DOCK_ATM         4
#define CHANNEL_TYPE_CHARACTER        5
#define CHANNEL_TYPE_FIFO             6
#define CHANNEL_TYPE_UDP_UNCONNECTED  7


struct channel_hook_type
{
    unsigned int magic_number;
    Channel channel;              /*  Not everyone sets this            */
    unsigned int type;
    flag local;
    flag call_io_tap_functions;
    flag locked;                  /*  Whether the file has been locked  */
    int fd;
    int ch_errno;
    char *unlink_file;            /*  File to unlink on close           */
    char *read_buffer;
    uaddr read_buf_len;
    uaddr read_buf_pos;
    uaddr bytes_read;
    char *write_buffer;
    uaddr write_buf_len;
    uaddr write_buf_pos;
    uaddr write_start_pos;
    struct karma_qos qos;
    iaddr (*read) (int fd, char *buf, uaddr nbytes);
    iaddr (*write) (int fd, CONST char *buf, uaddr nbytes);
    flag (*close) (int fd);
    KChannelHook next;
};


/*  Private data  */
static KChannelHook first_free = NULL;
#ifdef CAN_TIMEOUT
static volatile flag timed_out = FALSE;
#endif


/*  Private functions  */
STATIC_FUNCTION (KChannelHook alloc_hook, () );
#ifdef CAN_TIMEOUT
STATIC_FUNCTION (void sig_alarm_handler, (int sig) );
#endif
/*  Hook functions  */
STATIC_FUNCTION (flag close_func, (KChannelHook info) );
STATIC_FUNCTION (flag flush_func, (KChannelHook info) );
STATIC_FUNCTION (unsigned int _ch_read_disc,
		 (KChannelHook info, char *buffer, CONST char **newbuf,
		  unsigned char alignment, unsigned int length) );
STATIC_FUNCTION (unsigned int _ch_read_connection,
		 (KChannelHook info, char *buffer, CONST char **newbuf,
		  unsigned char alignment, unsigned int length) );
STATIC_FUNCTION (unsigned int _ch_read_datagram,
		 (KChannelHook info, char *buffer, CONST char **newbuf,
		  unsigned char alignment, unsigned int length) );
STATIC_FUNCTION (unsigned int _ch_write_descriptor,
		 (KChannelHook info, CONST char *buffer,unsigned int length) );
STATIC_FUNCTION (int mywrite_raw,
		 (KChannelHook info, CONST char *buffer,unsigned int length) );
STATIC_FUNCTION (flag seek_func, (KChannelHook info, unsigned long position) );
STATIC_FUNCTION (flag get_bytes_readable, (KChannelHook info) );
STATIC_FUNCTION (flag skip_func, (KChannelHook info, unsigned long position) );
STATIC_FUNCTION (int get_descriptor, (KChannelHook info) );
STATIC_FUNCTION (iaddr iread, (int fd, char *buf, uaddr nbytes) );
STATIC_FUNCTION (iaddr lwrite, (int fd, CONST char *buf, uaddr nbytes) );
STATIC_FUNCTION (iaddr sparse_write, (int fd, CONST char *buf, uaddr nbytes) );
STATIC_FUNCTION (flag iclose, (int fd) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
Channel ch_open_file (CONST char *filename, CONST char *type)
/*  [SUMMARY] Open a file.
    [PURPOSE] This routine will open a file channel. The file may be a regular
    disc file, a named FIFO, a character special device, a Unix domain
    connection or a TCP/IP connection. The channel may be later tested
    to determine what the true channel type is by calling routines such as:
    [<ch_test_for_asynchronous>] and [<ch_test_for_io>].
    <filename> The pathname of the file to open. This parameter has the same
    meaning as the first parameter to <<open(2)>>. Filenames of the form
    "//tcpIP/{hostname}:{port}" indicate a connection to a TCP/IP port on host
    <<hostname>> with raw port number <<port>> is requested.
    <type> The mode of the file. See [<ch_FILE_MODES>] for a list of allowable
    modes.
    [NOTE] For character special files and named FIFOs, these modes
    degenerate into read-write, read-only and write-only.
    [RETURNS] A channel object on success, else NULL.
*/
{
    flag read_flag = FALSE;
    flag write_flag = FALSE;
    flag grab_lock = FALSE;
    flag wait_for_lock = FALSE;
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;
    int flags = 0;  /*  Initialised to keep compiler happy  */
    int mode, ch, name_length;
    unsigned int filetype;
    unsigned int blocksize;
    CONST char *ptr;
    char *decompress_programme = NULL;
    static char function_name[] = "ch_open_file";

    if ( (filename == NULL) || (type == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if ( ( info = alloc_hook () ) == NULL ) return (NULL);
    info->call_io_tap_functions = TRUE;
    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.write = _ch_write_descriptor;
    hooks.get_bytes_readable = get_bytes_readable;
    hooks.skip = skip_func;
    hooks.get_descriptor = get_descriptor;
    name_length = strlen (filename);
#ifdef S_IRUSR
    mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
#else
    mode = 0664;  /*  Octal!  */
#endif
    /*  Determine access mode  */
    ptr = type + 1;
    switch (type[0])
    {
      case 'r':
	read_flag = TRUE;
	if (*ptr == '+')
	{
	    write_flag = TRUE;                          /*  "r+"  */
	    flags = O_RDWR;
	    ++ptr;
	}
	else flags = O_RDONLY;                          /*  "r"   */
	break;
      case 'w':
	write_flag = TRUE;
	if (*ptr == '+')
	{
	    flags = O_CREAT | O_TRUNC | O_RDWR;         /*  "w+"  */
	    read_flag = TRUE;
	    ++ptr;
	}
	else flags = O_CREAT | O_TRUNC | O_WRONLY;      /*  "w"   */
	break;
      case 'a':
	write_flag = TRUE;
	if (*ptr == '+')
	{
	    flags = O_CREAT | O_APPEND | O_RDWR;        /*  "a+"  */
	    read_flag = TRUE;
	    ++ptr;
	}
	else flags = O_CREAT | O_APPEND | O_WRONLY;     /*  "a"   */
	break;
      case 'W':
	flags = O_WRONLY;
	write_flag = TRUE;
	break;
      case 's':
	write_flag = TRUE;
	flags = O_CREAT | O_TRUNC | O_WRONLY;
	break;
      default:
	fprintf (stderr, "Illegal access mode: \"%s\"\n", type);
	a_prog_bug (function_name);
    }
    /*  Force mode to be binary (raw) on fucked systems which make the silly
	distinction between binary and text mode. Binary mode is the One True
	Mode  */
#ifdef O_BINARY
    flags |= O_BINARY;
#endif
    for (; *ptr != '\0'; ++ptr)
    {
	ch = *ptr;
	switch (ch)
	{
	  case 'z':
	    if (!read_flag || write_flag)
	    {
		fprintf (stderr, "Illegal access mode: \"%s\"\n", type);
		a_prog_bug (function_name);
	    }
#ifdef CAN_FORK
	    if ( (strcmp (filename + name_length - 3, ".gz") == 0) ||
		 (strcmp (filename + name_length - 2, ".z") == 0) )
		decompress_programme = "gzip";
	    else if (strcmp (filename + name_length - 4, ".bz2") == 0)
		decompress_programme = "bzip2";
#else
	    fprintf (stderr, "%s: decompression not supported\n",
		     function_name);
	    close_func (info);
	    return (NULL);
#endif
	    break;
	  case 't':
	    /*  Set sticky bit if defined  */
#ifdef S_ISVTX
	    mode |= S_ISVTX;
#endif
	    break;
	  case 'L':
	    wait_for_lock = TRUE;
	    /*  Fall through  */
	  case 'l':
	    grab_lock = TRUE;
	    break;
	  default:
	    fprintf (stderr, "Illegal access mode: \"%s\"\n", type);
	    a_prog_bug (function_name);
	}
    }
    if (grab_lock)
    {
	if ( ( info->unlink_file = m_alloc (name_length + 6) ) == NULL )
	{
	    m_error_notify (function_name, "lockfile name");
	    close_func (info);
	    return (NULL);
	}
	m_copy (info->unlink_file, filename, name_length);
	strcpy (info->unlink_file + name_length, ".lock");
	if ( !r_create_exclusive_file (info->unlink_file, wait_for_lock) )
	{
	    m_free (info->unlink_file);
	    info->unlink_file = NULL;
	    close_func (info);
	    return (NULL);
	}
    }
    /*  Open file descriptor  */
    if ( ( info->fd = r_open_file (filename, flags, mode, &filetype,
				   &blocksize) ) < 0 )
    {
	close_func (info);
	return (NULL);
    }
    /*  Set channel type  */
    switch (filetype)
    {
      case KFTYPE_DISC:
	hooks.read = _ch_read_disc;
	hooks.seek = seek_func;
	hooks.get_bytes_readable = NULL;
	info->write = (type[0] == 's') ? sparse_write : lwrite;
	info->close = iclose;
	break;
      case KFTYPE_CHARACTER:
	info->type = CHANNEL_TYPE_CHARACTER;
	hooks.read = _ch_read_connection;
	info->write = r_write;
	info->close = iclose;
	break;
      case KFTYPE_FIFO:
	info->type = CHANNEL_TYPE_FIFO;
	hooks.read = _ch_read_connection;
	info->write = r_write;
	info->close = iclose;
	break;
      case KFTYPE_UNIX_CONNECTION:
	info->type = CHANNEL_TYPE_CONNECTION;
	blocksize = CONNECTION_BUF_SIZE;
	hooks.read = _ch_read_connection;
	info->write = r_write;
	info->close = r_close_connection;
	break;
      case KFTYPE_LOCAL_tcpIP_CONNECTION:
	info->type = CHANNEL_TYPE_CONNECTION;
	info->local = TRUE;
	blocksize = CONNECTION_BUF_SIZE;
	hooks.read = _ch_read_connection;
	info->write = r_write;
	info->close = r_close_connection;
	break;
      case KFTYPE_REMOTE_tcpIP_CONNECTION:
	info->type = CHANNEL_TYPE_CONNECTION;
	info->local = FALSE;
	blocksize = CONNECTION_BUF_SIZE;
	hooks.read = _ch_read_connection;
	info->write = r_write;
	info->close = r_close_connection;
	break;
      default:
	fprintf (stderr, "Illegal filetype: %u\n", filetype);
	a_prog_bug (function_name);
	break;
    }
    if (grab_lock)
    {
	if ( !r_set_file_lock (info->fd, TRUE, wait_for_lock) )
	{
	    close_func (info);
	    return (NULL);
	}
	info->locked = TRUE;
    }
    if (blocksize == 0) blocksize = DEFAULT_BLOCK_SIZE;
    if (read_flag)
    {
	/*  Allocate read buffer  */
	if ( ( info->read_buffer = m_alloc (blocksize) ) == NULL )
	{
	    m_error_notify (function_name, "read buffer");
	    close_func (info);
	    return (NULL);
	}
	info->read_buf_len = blocksize;
    }
    if (write_flag)
    {
	/*  Allocate write buffer  */
	if ( ( info->write_buffer = m_alloc (blocksize) ) == NULL )
	{
	    m_error_notify (function_name, "write buffer");
	    close_func (info);
	    return (NULL);
	}
	info->write_buf_len = blocksize;
    }
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	close_func (info);
	return (NULL);
    }
    info->channel = channel;
#ifdef CAN_FORK
    if (decompress_programme != NULL)
    {
	fprintf (stderr, "attempting decompression with %s...\n",
		 decompress_programme);
	if (ch_add_uncompress (channel, decompress_programme) == NULL)
	{
	    ch_close (channel);
	    return (NULL);
	}
    }
#endif
    return (channel);
}   /*  End Function ch_open_file  */

/*PUBLIC_FUNCTION*/
Channel ch_open_connection (unsigned long host_addr, unsigned int port_number)
/*  [SUMMARY] Open a connection to a Karma server.
    [PURPOSE] This routine will open a full-duplex connection channel to a
    Karma server running on another host machine.
    <host_addr> The Internet address of the host machine. If this is 0 the
    connection is made to a server running on the same machine using the most
    efficient transport available.
    <port_number> The Karma port number to connect to. This should not be
    confused with Internet port numbers.
    [NOTE] The use of this routine is not recommended, see the [<conn>]
    package for more powerful functionality.
    [RETURNS] A channel object on success, else NULL.
*/
{
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_open_connection";

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.read = _ch_read_connection;
    hooks.write = _ch_write_descriptor;
    hooks.get_bytes_readable = get_bytes_readable;
    hooks.get_descriptor = get_descriptor;
    if ( ( info = alloc_hook () ) == NULL ) return (NULL);
    /*  Set channel type  */
    info->type = CHANNEL_TYPE_CONNECTION;
    info->write = r_write;
    info->close = r_close_connection;
    info->call_io_tap_functions = TRUE;
    /*  Open connection descriptor  */
    if ( ( info->fd = r_connect_to_port (host_addr, port_number,
					 &info->local) ) < 0 )
    {
	close_func (info);
	return (NULL);
    }
    /*  Allocate read buffer  */
    if ( ( info->read_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "read buffer");
	close_func (info);
	return (NULL);
    }
    info->read_buf_len = CONNECTION_BUF_SIZE;
    /*  Allocate write buffer  */
    if ( ( info->write_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "write buffer");
	close_func (info);
	return (NULL);
    }
    info->write_buf_len = CONNECTION_BUF_SIZE;
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	close_func (info);
	return (NULL);
    }
    return (channel);
}   /*  End Function ch_open_connection  */

/*PUBLIC_FUNCTION*/
Channel ch_accept_on_dock (Channel dock, unsigned long *addr)
/*  [SUMMARY] Accept a connection.
    [PURPOSE] This routine will open a full-duplex connection channel to the
    first connection on a waiting dock.
    <dock> The dock.
    <addr> The address of the host connecting to the dock will be written here.
    [RETURNS] A channel object on success, else NULL.
*/
{
    Channel channel;
    KChannelHook info, dock_info;
    KChannelHookFuncs hooks;
    int errnum = 0;
    unsigned int read_buf_len = 0;
    unsigned int write_buf_len = 0;
#ifdef CAN_TIMEOUT
    struct sigaction new_action, old_action;
#  ifdef HAS_ITIMER
    struct itimerval new_timer, old_timer;
    struct timeval current_time;
#  else
    unsigned int old_timeout;
#  endif
#endif
    static char function_name[] = "ch_accept_on_dock";

    if (addr == NULL)
    {
	fprintf (stderr, "NULL address pointer passed\n");
	a_prog_bug (function_name);
    }
    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.write = _ch_write_descriptor;
    hooks.get_bytes_readable = get_bytes_readable;
    hooks.get_descriptor = get_descriptor;
    dock_info = ch_get_hook_info (dock);
    if ( ( info = alloc_hook () ) == NULL )
    {
	return (NULL);
    }
    /*  Check if channel is a dock  */
    switch (dock_info->type)
    {
      case CHANNEL_TYPE_DOCK_STREAM:
	info->write = r_write;
	info->close = r_close_connection;
	info->call_io_tap_functions = TRUE;
	hooks.read = _ch_read_connection;
	break;
      case CHANNEL_TYPE_DOCK_UDP:
	info->read = iread;
	info->write = r_write;
	info->close = iclose;
	hooks.read = _ch_read_datagram;
	break;
      case CHANNEL_TYPE_DOCK_ATM:
	info->read = r_atm_read;
	info->write = r_atm_write;
	info->close = r_atm_close;
	hooks.read = _ch_read_datagram;
	break;
      default:
	fprintf (stderr, "Illegal dock type: %u\n", dock_info->type);
	a_prog_bug (function_name);
	break;
    }
    /*  Set channel type  */
    info->type = CHANNEL_TYPE_CONNECTION;
    /*  Accept connection. First set up a timeout if possible  */
#ifdef CAN_TIMEOUT
#  ifdef HAS_ITIMER
    memset (&new_timer, 0, sizeof new_timer);
    new_timer.it_value.tv_sec = TIMEOUT;
    if (setitimer (ITIMER_REAL, &new_timer, &old_timer) != 0)
    {
	fprintf (stderr, "%s: error setting interval timer\t%s\n",
		 function_name, ERRSTRING);
	close_func (info);
	return (NULL);
    }
    if ( (old_timer.it_value.tv_sec > 0) && (old_timer.it_value.tv_usec > 0) )
    {
	/*  Convert old timer value to absolute time  */
	if (gettimeofday (&current_time, NULL) != 0)
	{
	    fprintf (stderr, "%s: error getting time of day\t%s\n",
		     function_name, ERRSTRING);
	    close_func (info);
	    return (NULL);
	}
	old_timer.it_value.tv_sec += current_time.tv_sec;
	old_timer.it_value.tv_usec += current_time.tv_usec;
	if (old_timer.it_value.tv_usec >= 1000000)
	{
	    ++old_timer.it_value.tv_sec;
	    old_timer.it_value.tv_usec -= 1000000;
	}
    }
#  else
    old_timeout = alarm (TIMEOUT);
#  endif
    sigemptyset (&new_action.sa_mask);
    new_action.sa_flags = 0;
    new_action.sa_handler = sig_alarm_handler;
    if (sigaction (SIGALRM, &new_action, &old_action) != 0)
    {
	fprintf (stderr, "%s: error setting SIGALRM handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    timed_out = FALSE;
#endif
    switch (dock_info->type)
    {
      case CHANNEL_TYPE_DOCK_STREAM:
	if ( ( info->fd = r_accept_connection_on_dock (dock_info->fd, addr,
						       &info->local) ) < 0 )
	{
	    errnum = errno;
	    close_func (info);
	    break;
	}
	read_buf_len = CONNECTION_BUF_SIZE;
	write_buf_len = CONNECTION_BUF_SIZE;
	break;
      case CHANNEL_TYPE_DOCK_UDP:
	if ( ( info->fd = accept (dock_info->fd, NULL, NULL) ) < 0 )
	{
	    errnum = errno;
	    close_func (info);
	    break;
	}
	read_buf_len = UDP_PACKET_SIZE;
	write_buf_len = UDP_PACKET_SIZE;
	info->qos.tx.sdu_max = UDP_PACKET_SIZE;
	info->qos.rx.sdu_max = UDP_PACKET_SIZE;
	break;
      case CHANNEL_TYPE_DOCK_ATM:
	if ( ( info->fd = r_atm_accept (dock_info->fd, &info->qos) ) < 0 )
	{
	    errnum = errno;
	    close_func (info);
	    break;
	}
	read_buf_len = info->qos.rx.sdu_max;
	write_buf_len = info->qos.tx.sdu_max;
	break;
      default:
	fprintf (stderr, "Illegal dock type: %u\n", dock_info->type);
	a_prog_bug (function_name);
	break;
    }
#ifdef CAN_TIMEOUT
    /*  Restore old handler  */
    if (sigaction (SIGALRM, &old_action, NULL) != 0)
    {
	fprintf (stderr, "%s: error restoring old SIGALRM handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
#  ifdef HAS_ITIMER
    if ( (old_timer.it_value.tv_sec > 0) && (old_timer.it_value.tv_usec > 0) )
    {
	/*  Convert old timer value back to relative time, taking into account
	    the time we've stolen  */
	if (gettimeofday (&current_time, NULL) != 0)
	{
	    fprintf (stderr, "%s: error getting time of day\t%s\n",
		     function_name, ERRSTRING);
	    close_func (info);
	    return (NULL);
	}
	old_timer.it_value.tv_sec -= current_time.tv_sec;
	old_timer.it_value.tv_usec -= current_time.tv_usec;
	if (old_timer.it_value.tv_usec < 0)
	{
	    --old_timer.it_value.tv_sec;
	    old_timer.it_value.tv_usec += 1000000;
	}
	if (old_timer.it_value.tv_sec < 0)
	{
	    old_timer.it_value.tv_sec = 0;
	    old_timer.it_value.tv_usec = 1;
	}
    }
    if (setitimer (ITIMER_REAL, &old_timer, NULL) != 0)
    {
	fprintf (stderr, "%s: error restoring interval timer\t%s\n",
		 function_name, ERRSTRING);
	close_func (info);
	return (NULL);
    }
#  else
    if (old_timeout > 0)
    {
	if (old_timeout <= TIMEOUT) alarm (1);
	else alarm (old_timeout - TIMEOUT);
    }
    else alarm (0);
#  endif
    if ( (errnum == EINTR) && timed_out )
    {
	fprintf (stderr, "%s: timed out waiting for accept\n", function_name);
	return (NULL);
    }
#endif
    if (errnum != 0)
    {
	errno = errnum;
	fprintf (stderr, "%s: error accepting connection\t%s\n",
		 function_name, ERRSTRING);
	return (NULL);
    }
    /*  Allocate read buffer  */
    if (read_buf_len > 0)
    {
	if ( ( info->read_buffer = m_alloc (read_buf_len) ) == NULL )
	{
	    m_error_notify (function_name, "read buffer");
	    close_func (info);
	    return (NULL);
	}
	info->read_buf_len = read_buf_len;
    }
    /*  Allocate write buffer  */
    if ( ( info->write_buffer = m_alloc (write_buf_len) ) == NULL )
    {
	m_error_notify (function_name, "write buffer");
	close_func (info);
	return (NULL);
    }
    info->write_buf_len = write_buf_len;
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	close_func (info);
	return (NULL);
    }
    return (channel);
}   /*  End Function ch_accept_on_dock  */

/*PUBLIC_FUNCTION*/
Channel *ch_alloc_port (unsigned int *port_number, unsigned int retries,
			unsigned int *num_docks)
/*  [SUMMARY] Allocate a port.
    [PURPOSE] This routine will allocate a Karma port for the module so that it
    can operate as a server (able to receive network connections).
    <port_number> A pointer to the port number to allocate. The routine will
    write the actual port number allocated to this address. This must point to
    an address which lies on an <<int>> boundary.
    <retries> The number of succsessive port numbers to attempt to allocate
    before giving up. If this is 0, then the routine will give up immediately
    if the specified port number is in use.
    <num_docks> The routine will create a number of docks for one port. Each
    dock is an alternative access point for other modules to connect to this
    port. The number of docks allocated will be written here. This must point
    to an address which lies on an <<int>> boundary.
    [NOTE] The close-on-exec flags of the docks are set such that the docks
    will close on a call to execve(2V).
    [NOTE] The docks are placed into blocking mode.
    [RETURNS] A pointer to an array of channel docks on success, else NULL.
*/
{
    KChannelHook info;
    KChannelHookFuncs hooks;
    unsigned int dock_count;
    int *docks;
    Channel *ch_docks;
    static char function_name[] = "ch_alloc_port";

    if ( (port_number == NULL) || (num_docks == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.get_descriptor = get_descriptor;
    /*  Check alignment of  port_number  */
    if ( !IS_ALIGNED (port_number, sizeof *port_number) )
    {
	fprintf (stderr,
		 "Pointer to port number storage does not lie on an  int  boundary\n");
	a_prog_bug (function_name);
    }
    /*  Check alignment of  num_docks  */
    if ( !IS_ALIGNED (num_docks, sizeof *num_docks) )
    {
	fprintf (stderr,
		 "Pointer to number of docks storage does not lie on an  int  boundary\n");
	a_prog_bug (function_name);
    }
    if ( ( docks = r_alloc_port (port_number, retries, num_docks) )
	 == NULL )
    {
	return (NULL);
    }
    /*  Allocate array of channel pointers  */
    if ( ( ch_docks = m_alloc (sizeof *ch_docks * *num_docks) ) == NULL )
    {
	m_error_notify (function_name, "array of channel pointers");
	/*  Close docks  */
	for (dock_count = 0; dock_count < *num_docks; ++dock_count)
	{
	    r_close_dock (docks[dock_count]);
	}
	return (NULL);
    }
    /*  Create a channel object for each dock  */
    for (dock_count = 0; dock_count < *num_docks; ++dock_count)
    {
	if ( ( info = alloc_hook () ) == NULL )
	{
	    /*  Close docks  */
	    for (; dock_count > 0; --dock_count)
	    {
		ch_close (ch_docks[dock_count - 1]);
	    }
	    m_free (ch_docks);
	    return (NULL);
	}
	info->type = CHANNEL_TYPE_DOCK_STREAM;
	info->close = r_close_dock;
	if ( ( ch_docks[dock_count] = ch_create_generic (info, hooks) )
	     == NULL )
	{
	    m_error_notify (function_name, "channel object");
	    close_func (info);
	    /*  Close docks  */
	    for (; dock_count > 0; --dock_count)
	    {
		ch_close (ch_docks[dock_count - 1]);
	    }
	    m_free (ch_docks);
	    return (NULL);
	}
	info->fd = docks[dock_count];
    }
    return (ch_docks);
}   /*  End Function ch_alloc_port  */

/*PUBLIC_FUNCTION*/
Channel ch_attach_to_asynchronous_descriptor (int fd)
/*  [SUMMARY] Create a channel object from an asynchronous descriptor.
    <fd> The descriptor.
    [RETURNS] A channel object on success, else NULL.
*/
{
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_attach_to_asynchronous_descriptor";

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.get_descriptor = get_descriptor;
    if ( ( info = alloc_hook () ) == NULL ) return (NULL);
    /*  Set channel type  */
    info->fd = fd;
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	close_func (info);
	m_abort (function_name, "channel object");
    }
    return (channel);
}   /*  End Function ch_attach_to_asynchronous_descriptor  */

/*PUBLIC_FUNCTION*/
void ch_open_stdin ()
/*  [SUMMARY] Create starndard input channel.
    [PURPOSE] This routine will create a channel object for the standard input
    descriptor (typically 0 on Unix systems).
    [NOTE] The standard input channel will be written to the external variable
    <<ch_stdin>>.
    [RETURNS] Nothing.
*/
{
    unsigned int filetype;
    KChannelHook info;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_open_stdin";

    if (ch_stdin != NULL)
    {
	fprintf (stderr, "ch_stdin already open\n");
	a_prog_bug (function_name);
    }
    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.get_descriptor = get_descriptor;
    if ( ( info = alloc_hook () ) == NULL )
    {
	m_abort (function_name, "ch_stdin");
    }
    if ( ( info->fd = r_open_stdin2 (&filetype) ) < 0 )
    {
	fprintf (stderr, "Error getting input descriptor\n");
	exit (RV_UNDEF_ERROR);
    }
    /*  Allocate read buffer  */
    if ( ( info->read_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_abort (function_name, "read buffer");
    }
    info->read_buf_len = CONNECTION_BUF_SIZE;
    /*  Tag type  */
    switch (filetype)
    {
      case KFTYPE_DISC:
	hooks.read = _ch_read_disc;
	hooks.seek = seek_func;
	info->write = lwrite;
	info->close = iclose;
	break;
      case KFTYPE_CHARACTER:
	info->type = CHANNEL_TYPE_CHARACTER;
	hooks.read = _ch_read_connection;
	hooks.get_bytes_readable = get_bytes_readable;
	info->write = r_write;
	info->close = iclose;
	break;
      case KFTYPE_FIFO:
	info->type = CHANNEL_TYPE_FIFO;
	hooks.read = _ch_read_connection;
	hooks.get_bytes_readable = get_bytes_readable;
	info->write = r_write;
	info->close = iclose;
	break;
      default:
	fprintf (stderr, "Illegal filetype: %u\n", filetype);
	a_prog_bug (function_name);
	break;
    }
    info->call_io_tap_functions = TRUE;
    if ( ( ch_stdin = ch_create_generic (info, hooks) ) == NULL )
    {
	close_func (info);
	m_abort (function_name, "channel object");
    }
}   /*  End Function ch_open_stdin  */

/*EXPERIMENTAL_FUNCTION*/
void ch_open_stdout ()
/*  [SUMMARY] Create starndard output channel.
    [PURPOSE] This routine will create a channel object for the standard output
    descriptor (typically 1 on Unix systems).
    [NOTE] The standard output channel will be written to the external variable
    <<ch_stdout>>.
    [RETURNS] Nothing.
*/
{
    KChannelHook info;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_open_stdout";

    if (ch_stdout != NULL)
    {
	fprintf (stderr, "ch_stdout already open\n");
	a_prog_bug (function_name);
    }
    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.write = _ch_write_descriptor;
    hooks.get_bytes_readable = get_bytes_readable;
    hooks.get_descriptor = get_descriptor;
    if ( ( info = alloc_hook () ) == NULL )
    {
	m_abort (function_name, "ch_stdout");
    }
    if ( !r_get_fd_filetype (1, &info->type) ) exit (RV_SYS_ERROR);
    info->fd = 1;
    /*  Allocate write buffer  */
    if ( ( info->write_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_abort (function_name, "read buffer");
    }
    info->write_buf_len = CONNECTION_BUF_SIZE;
    info->call_io_tap_functions = TRUE;
    switch (info->type)
    {
      case KFTYPE_DISC:
	info->write = lwrite;
	info->close = iclose;
	hooks.seek = seek_func;
	break;
      case KFTYPE_CHARACTER:
	info->type = CHANNEL_TYPE_CHARACTER;
	info->write = r_write;
	info->close = iclose;
	break;
      case KFTYPE_FIFO:
	info->type = CHANNEL_TYPE_FIFO;
	info->write = r_write;
	info->close = iclose;
	break;
      case KFTYPE_SOCKET:
	info->type = CHANNEL_TYPE_CONNECTION;
	info->local = TRUE;  /*  This is an assumption  */
	info->write = r_write;
	info->close = r_close_connection;
	break;
      default:
	fprintf (stderr, "%s: illegal descriptor type: %u\n",
		 function_name, info->type);
	a_prog_bug (function_name);
	break;
    }
    if ( ( ch_stdout = ch_create_generic (info, hooks) ) == NULL )
    {
	close_func (info);
	m_abort (function_name, "channel object");
    }
}   /*  End Function ch_open_stdout  */

/*PUBLIC_FUNCTION*/
flag ch_create_pipe (Channel *read_ch, Channel *write_ch)
/*  [SUMMARY] Create a pipe.
    [PURPOSE] This routine will create an un-named pipe (see <<pipe(2)>> for
    details on un-named pipes).
    <read_ch> The channel corresponding to the read end of the pipe will be
    written here.
    <write_ch> The channel corresponding to the write end of the pipe will be
    written here.
    [RETURNS] TRUE on success, else FALSE and sets <<errno>> with the error
    code.
*/
{
    Channel rch, wch;
    KChannelHook rinfo, winfo;
    KChannelHookFuncs hooks;
    int read_fd, write_fd;
    static char function_name[] = "ch_create_pipe";

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.get_bytes_readable = get_bytes_readable;
    hooks.get_descriptor = get_descriptor;
    if (r_create_pipe (&read_fd, &write_fd) < 0) return (FALSE);
    errno = 0;
    /*  Read channel  */
    if ( ( rinfo = alloc_hook () ) == NULL )
    {
	close (read_fd);
	close (write_fd);
	return (FALSE);
    }
    rinfo->type = CHANNEL_TYPE_FIFO;
    rinfo->fd = read_fd;
    /*  Allocate read buffer  */
    if ( ( rinfo->read_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "read buffer");
	close_func (rinfo);
	close (write_fd);
	return (FALSE);
    }
    rinfo->read_buf_len = CONNECTION_BUF_SIZE;
    hooks.read = _ch_read_connection;
    rinfo->write = r_write;
    rinfo->close = iclose;
    rinfo->call_io_tap_functions = TRUE;
    if ( ( rch = ch_create_generic (rinfo, hooks) ) == NULL )
    {
	close_func (rinfo);
	close (write_fd);
	m_abort (function_name, "channel object");
    }
    hooks.read = NULL;
    /*  Write channel  */
    if ( ( winfo = alloc_hook () ) == NULL )
    {
	ch_close (rch);
	close (write_fd);
	return (FALSE);
    }
    winfo->type = CHANNEL_TYPE_FIFO;
    winfo->fd = write_fd;
    /*  Allocate write buffer  */
    if ( ( winfo->write_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "write buffer");
	ch_close (rch);
	close_func (winfo);
	return (FALSE);
    }
    winfo->write_buf_len = CONNECTION_BUF_SIZE;
    hooks.write = _ch_write_descriptor;
    winfo->write = r_write;
    winfo->close = iclose;
    winfo->call_io_tap_functions = TRUE;
    if ( ( wch = ch_create_generic (winfo, hooks) ) == NULL )
    {
	ch_close (rch);
	close_func (winfo);
	m_abort (function_name, "channel object");
    }
    *read_ch = rch;
    *write_ch = wch;
    return (TRUE);
}   /*  End Function ch_create_pipe  */

/*EXPERIMENTAL_FUNCTION*/
flag ch_create_socketpair (Channel *ch1, Channel *ch2)
/*  [SUMMARY] Create a pair of un-named connected sockets.
    <ch1> The channel corresponding to one end of the pair is written here.
    <ch2> The channel corresponding to the other end of the pair is written
    here.
    [RETURNS] TRUE on success, else FALSE and sets <<errno>> with the error
    code.
*/
{
    Channel ch_1, ch_2;
    KChannelHook info1, info2;
    KChannelHookFuncs hooks;
    int fd1, fd2;
    static char function_name[] = "ch_create_socketpair";

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.read = _ch_read_connection;
    hooks.write = _ch_write_descriptor;
    hooks.get_bytes_readable = get_bytes_readable;
    hooks.get_descriptor = get_descriptor;
    if (r_create_socketpair (&fd1, &fd2) < 0) return (FALSE);
    errno = 0;
    /*  One channel  */
    if ( ( info1 = alloc_hook () ) == NULL )
    {
	close (fd1);
	close (fd2);
	return (FALSE);
    }
    info1->type = CHANNEL_TYPE_CONNECTION;
    info1->local = TRUE;
    info1->fd = fd1;
    info1->write = r_write;
    info1->close = r_close_connection;
    info1->call_io_tap_functions = TRUE;
    if ( ( info1->read_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "read buffer");
	close_func (info1);
	close (fd2);
	return (FALSE);
    }
    info1->read_buf_len = CONNECTION_BUF_SIZE;
    if ( ( info1->write_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "write buffer");
	close_func (info1);
	close (fd2);
	return (FALSE);
    }
    info1->write_buf_len = CONNECTION_BUF_SIZE;
    if ( ( ch_1 = ch_create_generic (info1, hooks) ) == NULL )
    {
	close_func (info1);
	close (fd2);
	m_abort (function_name, "channel object");
    }
    /*  Other channel  */
    if ( ( info2 = alloc_hook () ) == NULL )
    {
	ch_close (ch_1);
	close (fd2);
	return (FALSE);
    }
    info2->type = CHANNEL_TYPE_CONNECTION;
    info1->local = TRUE;
    info2->fd = fd2;
    info2->write = r_write;
    info2->close = r_close_connection;
    info2->call_io_tap_functions = TRUE;
    if ( ( info2->read_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "read buffer");
	ch_close (ch_1);
	close_func (info2);
	return (FALSE);
    }
    info2->read_buf_len = CONNECTION_BUF_SIZE;
    if ( ( info2->write_buffer = m_alloc (CONNECTION_BUF_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "write buffer");
	ch_close (ch_1);
	close_func (info2);
	return (FALSE);
    }
    info2->write_buf_len = CONNECTION_BUF_SIZE;
    if ( ( ch_2 = ch_create_generic (info2, hooks) ) == NULL )
    {
	ch_close (ch_1);
	close_func (info2);
	m_abort (function_name, "channel object");
    }
    *ch1 = ch_1;
    *ch2 = ch_2;
    return (TRUE);
}   /*  End Function ch_create_socketpair  */

/*PUBLIC_FUNCTION*/
flag ch_test_for_asynchronous (Channel channel)
/*  [SUMMARY] Test if a channel object is an asynchronous channel.
    [PURPOSE] This routine will test if a channel object is an asynchronous
    channel, i.e. a character special file, named FIFO, connection, a dock
    channel or one created by a call to
    <<ch_attach_to_asynchronous_descriptor>> or <<ch_create_pipe>>.
    <channel> The channel object.
    [RETURNS] TRUE if the channel is asynchronous, else FALSE.
*/
{
    KChannelHook info;
    /*static char function_name[] = "ch_test_for_asynchronous";*/

    info = ch_get_hook_info (channel);
    if (info == NULL) return (FALSE);
    if (info->magic_number != USED_MAGIC_NUMBER) return (FALSE);
    return (TRUE);
}   /*  End Function ch_test_for_asynchronous  */

/*PUBLIC_FUNCTION*/
flag ch_test_for_connection (Channel channel)
/*  [SUMMARY] Test if a channel object is a connection channel.
    <channel> The channel object.
    [RETURNS] TRUE if the channel object is a connection, else FALSE.
*/
{
    KChannelHook info;
    /*static char function_name[] = "ch_test_for_connection";*/

    info = ch_get_hook_info (channel);
    if (info == NULL) return (FALSE);
    if (info->magic_number != USED_MAGIC_NUMBER) return (FALSE);
    if (info->type == CHANNEL_TYPE_CONNECTION) return (TRUE);
    else return (FALSE);
}   /*  End Function ch_test_for_connection  */

/*PUBLIC_FUNCTION*/
flag ch_test_for_local_connection (Channel channel)
/*  [SUMMARY] Test if a connection channel object is a local connection.
    <channel> The channel object.
    [RETURNS] TRUE if the channel object is a local connection, else FALSE.
*/
{
    KChannelHook info;
    /*static char function_name[] = "ch_test_for_local_connection";*/

    info = ch_get_hook_info (channel);
    if (info == NULL) return (FALSE);
    if (info->magic_number != USED_MAGIC_NUMBER) return (FALSE);
    switch (info->type)
    {
      case CHANNEL_TYPE_FIFO:
#ifdef OS_Solaris  /*  Hack for socketpairs  */
      case CHANNEL_TYPE_CHARACTER:
#endif
	return (TRUE);
	/*break;*/
      case CHANNEL_TYPE_CONNECTION:
	return (info->local);
	/*break;*/
      default:
	break;
    }
    return (FALSE);
}   /*  End Function ch_test_for_local_connection  */

/*EXPERIMENTAL_FUNCTION*/
flag ch_get_buffer_sizes (Channel channel, uaddr *read_size, uaddr *write_size)
/*  [SUMMARY] Get the read and write buffer sizes for a channel.
    <channel> The channel object.
    <read_size> The read buffer size is written here.
    <write_size> The write buffer size is written here.
    [RETURNS] TRUE on success, else FALSE indicating the channel does not have
    read or write buffers.
*/
{
    KChannelHook info;

    info = ch_get_hook_info (channel);
    if (info == NULL) return (FALSE);
    if (info->magic_number != USED_MAGIC_NUMBER) return (FALSE);
    *read_size = info->read_buf_len;
    *write_size = info->write_buf_len;
    return (TRUE);
}   /*  End Function ch_get_buffer_sizes  */

/*EXPERIMENTAL_FUNCTION*/
flag ch_set_buffer_sizes (Channel channel, unsigned long read_size,
			  unsigned long write_size)
/*  [SUMMARY] Set the read and write buffer sizes for a channel.
    <channel> The channel object.
    <read_size> The read buffer size.
    <write_size> The write buffer size.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KChannelHook info;
    char *ptr;
    static char function_name[] = "ch_set_buffer_sizes";

    info = ch_get_hook_info (channel);
    if (info == NULL) return (FALSE);
    if (info->magic_number != USED_MAGIC_NUMBER)
    {
	fprintf (stderr, "Channel does not have buffers\n");
	a_prog_bug (function_name);
    }
    if ( !ch_flush (channel) ) return (FALSE);
    if (info->read_buf_len > 0)
    {
	/*  Has existing read buffer  */
	if (read_size < 1)
	{
	    fprintf (stderr, "%s: cannot remove a read buffer\n",
		     function_name);
	    return (FALSE);
	}
	if ( (info->read_buf_pos > read_size) &&
	     (info->bytes_read < info->read_buf_pos) )
	{
	    fprintf (stderr,
		     "%s: cannot shrink read buffer as data would be lost\n",
		     function_name);
	    return (FALSE);
	}
	if ( ( ptr = m_alloc (read_size) ) == NULL )
	{
	    m_error_notify (function_name, "read buffer");
	    return (FALSE);
	}
	m_copy (ptr, info->read_buffer + info->bytes_read,
		info->read_buf_pos - info->bytes_read);
	m_free (info->read_buffer);
	info->read_buffer = ptr;
	info->read_buf_len = read_size;
	info->read_buf_pos = info->read_buf_pos - info->bytes_read;
	info->bytes_read = 0;
    }
    else
    {
	/*  No existing read buffer  */
	if (read_size > 0)
	{
	    fprintf (stderr, "%s: cannot create a read buffer\n",
		     function_name);
	    return (FALSE);
	}
    }
    if (info->write_buf_len > 0)
    {
	/*  Has existing write buffer  */
	if (write_size < 1)
	{
	    fprintf (stderr, "%s: cannot remove a write buffer\n",
		     function_name);
	    return (FALSE);
	}
	if ( ( ptr = m_alloc (write_size) ) == NULL )
	{
	    m_error_notify (function_name, "write buffer");
	    return (FALSE);
	}
	info->write_buffer = ptr;
	info->write_buf_len = write_size;
    }
    else
    {
	/*  No existing read buffer  */
	if (write_size > 0)
	{
	    fprintf (stderr, "%s: cannot create a write buffer\n",
		     function_name);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function ch_set_buffer_sizes  */

/*UNPUBLISHED_FUNCTION*/
Channel ch_udp_alloc (unsigned int *port_number)
/*  [SUMMARY] Create a UDP socket and bind to a random port number.
    <port_number> The port number the socket was bound to.
    [RETURNS] The channel object on success, else NULL.
*/
{
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_udp_alloc";

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.get_descriptor = get_descriptor;
    if ( ( info = alloc_hook () ) == NULL ) return (NULL);
    /*  Set channel type  */
    info->type = CHANNEL_TYPE_UDP_UNCONNECTED;
    info->close = iclose;
    /*  Open connection descriptor  */
    if ( ( info->fd = r_alloc_udp (port_number) ) < 0 )
    {
	close_func (info);
	return (NULL);
    }
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	close_func (info);
	return (NULL);
    }
    return (channel);
}   /*  End Function ch_udp_alloc  */

/*UNPUBLISHED_FUNCTION*/
Channel ch_udp_connect (Channel channel, unsigned long addr,
			unsigned int port_number)
/*  [SUMMARY] Make a connection to a UDP port on some machine.
    <channel> A UDP socket channel object. If this is NULL, a socket is created
    <addr> The address of the machine to connect to.
    <port_number> The port number to connect to.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [RETURNS] The channel object of the opened connection on success, else
    NULL.
*/
{
    KChannelHook info;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_udp_connect";

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.read = _ch_read_datagram;
    hooks.write = _ch_write_descriptor;
    hooks.get_bytes_readable = get_bytes_readable;
    hooks.get_descriptor = get_descriptor;
    if (channel == NULL)
    {
	if ( ( info = alloc_hook () ) == NULL ) return (NULL);
	if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
	{
	    m_error_notify (function_name, "channel object");
	    close_func (info);
	    return (NULL);
	}
    }
    else
    {
	info = ch_get_hook_info (channel);
	ch_change_hooks (channel, hooks);
	if (info->type != CHANNEL_TYPE_UDP_UNCONNECTED)
	{
	    fprintf (stderr, "Channel is not an unconnected UDP socket\n");
	    a_prog_bug (function_name);
	}
    }
    /*  Set channel type  */
    info->read = iread;
    info->write = lwrite;
    info->close = iclose;
    /*  Open connection descriptor  */
    if ( ( info->fd = r_connect_to_udp (info->fd, addr, port_number) ) < 0 )
    {
	ch_close (channel);
	return (NULL);
    }
    /*  Allocate read buffer  */
    if ( ( info->read_buffer = m_alloc (UDP_PACKET_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "read buffer");
	ch_close (channel);
	return (NULL);
    }
    info->read_buf_len = UDP_PACKET_SIZE;
    /*  Allocate write buffer  */
    if ( ( info->write_buffer = m_alloc (UDP_PACKET_SIZE) ) == NULL )
    {
	m_error_notify (function_name, "write buffer");
	ch_close (channel);
	return (NULL);
    }
    info->write_buf_len = UDP_PACKET_SIZE;
    info->qos.tx.sdu_max = UDP_PACKET_SIZE;
    info->qos.rx.sdu_max = UDP_PACKET_SIZE;
    return (channel);
}   /*  End Function ch_udp_connect  */

/*UNPUBLISHED_FUNCTION*/
Channel ch_atm_alloc (char *addr_string, unsigned int *addr_len)
/*  [SUMMARY] Allocate an ATMSVC socket and bind to a random port number.
    <addr_string> The textual representation of the ATM address is written here
    <addr_len> The length of the address string. This is updated.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [RETURNS] The channel for the socket on success, else NULL.
*/
{
    static char function_name[] = "ch_atm_alloc";
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.get_descriptor = get_descriptor;
    if ( ( info = alloc_hook () ) == NULL ) return (NULL);
    if ( ( info->fd = r_atm_alloc (addr_string, addr_len) ) < 0 )
    {
	close_func (info);
	return (NULL);
    }
    info->type = CHANNEL_TYPE_DOCK_ATM;
    info->close = r_atm_close;
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	close_func (info);
    }
    return (channel);
}   /*  End Function ch_atm_alloc  */

/*UNPUBLISHED_FUNCTION*/
Channel ch_atm_connect (CONST char *addr_string, unsigned int addr_len,
			struct karma_qos *qos)
/*  [SUMMARY] Make a connection using ATM SVC with AAL5.
    <addr_string> The textual representation of the ATM address to connect to.
    <addr_len> The length of the address string.
    <qos> The Quality Of Service parameters. This will be updated to reflect
    the parameters provided. This may be NULL.
    [RETURNS] The channel object of the opened connection on success, else
    NULL.
*/
{
    static char function_name[] = "ch_atm_connect";
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.read = _ch_read_datagram;
    hooks.write = _ch_write_descriptor;
    hooks.get_bytes_readable = get_bytes_readable;
    hooks.get_descriptor = get_descriptor;
    if ( ( info = alloc_hook () ) == NULL ) return (NULL);
    info->read = r_atm_read;
    info->write = r_atm_write;
    info->close = r_atm_close;
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	close_func (info);
	return (NULL);
    }
    /*  Open connection descriptor  */
    if ( ( info->fd = r_atm_connect (addr_string, addr_len, qos) ) < 0 )
    {
	ch_close (channel);
	return (NULL);
    }
    info->qos = *qos;
    /*  Allocate read buffer  */
    if ( ( info->read_buffer = m_alloc (qos->rx.sdu_max) ) == NULL )
    {
	m_error_notify (function_name, "read buffer");
	ch_close (channel);
	return (NULL);
    }
    info->read_buf_len = qos->rx.sdu_max;
    /*  Allocate write buffer  */
    if ( ( info->write_buffer = m_alloc (qos->tx.sdu_max) ) == NULL )
    {
	m_error_notify (function_name, "write buffer");
	ch_close (channel);
	return (NULL);
    }
    info->write_buf_len = qos->tx.sdu_max;
    return (channel);
}   /*  End Function ch_atm_connect  */

/*EXPERIMENTAL_FUNCTION*/
flag ch_set_nodelay (Channel channel)
/*  [SUMMARY] Set the no-delay flag for a channel if appropriate.
    <channel> The channel object.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KChannelHook info;
    static char function_name[] = "ch_set_nodelay";

    info = ch_get_hook_info (channel);
    VERIFY_USED (info);
    return r_set_nodelay (info->fd);
}   /*  End Function ch_set_nodelay  */

/*EXPERIMENTAL_FUNCTION*/
void ch_get_qos (Channel channel, struct karma_qos *qos)
/*  [SUMMARY] Get the Quality of Service parameters for a connection.
    <channel> The channel object.
    <qos> The QoS parameters will be written here.
    [RETURNS] Nothing.
*/
{
    KChannelHook info;
    static char function_name[] = "ch_get_qos";

    info = ch_get_hook_info (channel);
    VERIFY_USED (info);
    *qos = info->qos;
}   /*  End Function ch_get_qos  */

/*UNPUBLISHED_FUNCTION*/
void ch_disable_tap_functions (Channel channel)
/*  [SUMMARY] Disable tap functions for a channel.
    <channel> The channel object.
    <qos> The QoS parameters will be written here.
    [RETURNS] Nothing.
*/
{
    KChannelHook info;
    static char function_name[] = "ch_disable_tap_functions";

    info = ch_get_hook_info (channel);
    VERIFY_USED (info);
    info->call_io_tap_functions = FALSE;
}   /*  End Function ch_disable_tap_functions  */


/*  Private functions follow  */

static KChannelHook alloc_hook ()
/*  This routine will allocate and initialise a channel data structure.
    The routine returns a channel object structure on success, 
    else it returns NULL.
*/
{
    KChannelHook new;
    unsigned int count;
    static char function_name[] = "_ch_descriptor_alloc_hook";

    /*  Get new entry  */
    if (first_free == NULL)
    {
	/*  Have to allocate some more for the free list  */
	if ( ( new = m_alloc (ALLOC_SIZE * sizeof *new) ) == NULL )
	{
	    m_error_notify (function_name, "new descriptor channel entries");
	    return (NULL);
	}
	m_clear (new, ALLOC_SIZE * sizeof *new);
	for (count = 0; count < ALLOC_SIZE - 1; ++count)
	{
	    new[count].magic_number = FREE_MAGIC_NUMBER;
	    new[count].next = new + count + 1;
	}
	new[count].magic_number = FREE_MAGIC_NUMBER;
	/*  Add them to the free list  */
	first_free = new;
    }
    /*  Take one from the free list  */
    new = first_free;
    VERIFY_FREE (new);
    first_free = new->next;
    /*  Fill in entry  */
    m_clear (new, sizeof *new);
    new->magic_number = USED_MAGIC_NUMBER;
    new->fd = -1;
    return (new);
}   /*  End Function alloc_hook  */

static void free_hook (KChannelHook info)
/*  [SUMMARY] Add a KChannelHook object to the free list.
    <info> The KChannelHook object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_ch_descriptor_free_hook";

    VERIFY_USED (info);
    m_clear (info, sizeof *info);
    info->magic_number = FREE_MAGIC_NUMBER;
    info->next = first_free;
    first_free = info;
}   /*  End Function free_hook  */

#ifdef CAN_TIMEOUT
static void sig_alarm_handler (int sig)
{
    timed_out = TRUE;
}   /*  End Function sig_alarm_handler  */
#endif


/*  Hook functions follow  */

static flag close_func (KChannelHook info)
/*  [SUMMARY] Close a channel. The channel must already be flushed.
    <info> The channel object subclass information pointer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag return_value = TRUE;
    /*static char function_name[] = "_ch_descriptor_close_func";*/

    if (info->locked) r_set_file_lock (info->fd, FALSE, FALSE);
    if ( (info->fd > -1) && (info->close != NULL) )
    {
	/*  Have open file descriptor: close descriptor  */
	return_value = (*info->close) (info->fd);
    }
    if (info->unlink_file != NULL)
    {
	unlink (info->unlink_file);
	m_free (info->unlink_file);
    }
    /*  Deallocate buffers  */
    if (info->read_buffer != NULL) m_free (info->read_buffer);
    if (info->write_buffer != NULL) m_free (info->write_buffer);
    free_hook (info);
    return (return_value);
}   /*  End Function close_func  */

static flag flush_func (KChannelHook info)
/*  [SUMMARY] Flush the write buffer of a channel object.
    <info> The channel object subclass information pointer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int bytes_to_write;
    /*static char function_name[] = "_ch_descriptor_flush_func";*/

    if (info->fd < 0) return (TRUE);
    if (info->write_buf_pos > info->write_start_pos)
    {
	/*  Flush write buffer  */
	bytes_to_write = info->write_buf_pos - info->write_start_pos;
	if (mywrite_raw (info, info->write_buffer, bytes_to_write)
	    < bytes_to_write)
	{
	    info->write_start_pos = 0;
	    return (FALSE);
	}
	info->write_start_pos = 0;
	info->write_buf_pos = 0;
    }
    return (TRUE);
}   /*  End Function flush_func  */

static unsigned int _ch_read_disc (KChannelHook info, char *buffer,
				   CONST char **newbuf,
				   unsigned char alignment,unsigned int length)
/*  [SUMMARY] Read from a channel.
    [PURPOSE] This routine will read a number of bytes from a channel and
    places them into a buffer.
    <channel> The channel object.
    <buffer> A suggested buffer to write the data into.
    <newbuf> The buffer pointer used is written here. This may be a read-only
    buffer. If this is NULL, the suggested <<buffer>> is always used.
    <alignment> The number of bytes the buffer must be aligned to.
    <length> The number of bytes to write into the buffer.
    [RETURNS] The number of bytes read.
*/
{
    unsigned int read_pos = 0;
    unsigned int count;
    int bytes_to_read;
    int bytes_read;
    CONST char *source;
    /*static char function_name[] = "_ch_read_disc";*/

    if ( (info->bytes_read - info->read_buf_pos) >= length )
    {
	/*  Read buffer has enough data in it  */
	source = info->read_buffer + info->read_buf_pos;
	if (length < 16)
	{
	    /*  Avoid function call overhead  */
	    for (count = 0; count < length; ++count) *buffer++ = *source++;
	}
	else m_copy (buffer, source, length);
	info->read_buf_pos += length;
	return (length);
    }
    /*  Need to read from disc  */
    /*  Copy what is left in the read buffer  */
    read_pos = info->bytes_read - info->read_buf_pos;
    m_copy (buffer, info->read_buffer + info->read_buf_pos, read_pos);
    /*  Test if an error occurred  */
    if (info->ch_errno != 0)
    {
	errno = (info->ch_errno > 0) ? info->ch_errno : 0;
	info->read_buf_pos = 0;
	info->bytes_read = 0;
	return (read_pos);
    }
    /*  Test if that's all there is in the file  */
    if ( (info->bytes_read > 0) &&
	 (info->bytes_read < info->read_buf_len) )
    {
	/*  Must have hit EOF  */
	return (read_pos);
    }
    info->read_buf_pos = 0;
    info->bytes_read = 0;
    bytes_to_read = length - read_pos;
    if (bytes_to_read > info->read_buf_len)
    {
	/*  Call any tap functions that were registered  */
	if (info->call_io_tap_functions) ch_call_tap_callbacks ();
	/*  Read an intregal number of blocks directly  */
	bytes_to_read -= bytes_to_read % (int) info->read_buf_len;
	if ( ( bytes_read = read (info->fd, buffer + read_pos,
				  bytes_to_read) )
	    < 0 )
	{
	    /*  Error occurred  */
	    info->ch_errno = errno;
	    if (read_pos > 0)
	    {
		errno = 0;
	    }
	    return (read_pos);
	}
	if (bytes_read < bytes_to_read)
	{
	    /*  Only read some bytes: hit End-Of-File  */
	    read_pos += bytes_read;
	    return (read_pos);
	}
	read_pos += bytes_read;
    }
    /*  Call any tap functions that were registered  */
    if (info->call_io_tap_functions) ch_call_tap_callbacks ();
    /*  Read one block into read buffer  */
    bytes_to_read = info->read_buf_len;
    if ( ( bytes_read = read (info->fd, info->read_buffer, bytes_to_read) )
	 < 0 )
    {
	/*  Error reading complete block  */
	info->ch_errno = errno;
    }
    if (bytes_read < bytes_to_read)
    {
	info->ch_errno = -1;
    }
    info->bytes_read = bytes_read;
    /*  Read what is needed from the read buffer  */
    read_pos += _ch_read_disc (info, buffer + read_pos, NULL, 0,
			       length - read_pos);
    errno = 0;
    return (read_pos);
}   /*  End Function _ch_read_disc  */

static unsigned int _ch_read_connection (KChannelHook info, char *buffer,
					 CONST char **newbuf,
					 unsigned char alignment,
					 unsigned int length)
/*  [SUMMARY] Read from a channel.
    [PURPOSE] This routine will read a number of bytes from a channel and
    places them into a buffer.
    <channel> The channel object.
    <buffer> A suggested buffer to write the data into.
    <newbuf> The buffer pointer used is written here. This may be a read-only
    buffer. If this is NULL, the suggested <<buffer>> is always used.
    <alignment> The number of bytes the buffer must be aligned to.
    <length> The number of bytes to write into the buffer.
    [RETURNS] The number of bytes read.
*/
{
    flag was_error = TRUE;
    int bytes_available;
    unsigned int read_pos;
    int bytes_to_read;
    int bytes_read;
    static char function_name[] = "_ch_read_connection";

    if ( (info->bytes_read - info->read_buf_pos) >= length )
    {
	/*  Read buffer has enough data in it  */
	m_copy (buffer, info->read_buffer + info->read_buf_pos, length);
	info->read_buf_pos += length;
	return (length);
    }
    /*  Will need to read from connection  */
    /*  Copy what is left in the read buffer  */
    read_pos = info->bytes_read - info->read_buf_pos;
    m_copy (buffer, info->read_buffer + info->read_buf_pos, read_pos);
    info->read_buf_pos = 0;
    info->bytes_read = 0;
    /*  Test if an error occurred  */
    if (info->ch_errno > 0)
    {
	errno = info->ch_errno;
	return (read_pos);
    }
    /*  Call any tap functions that were registered  */
    if (info->call_io_tap_functions) ch_call_tap_callbacks ();
    /*  Read the remainder directly  */
    bytes_to_read = length - read_pos;
    if ( ( bytes_read = r_read (info->fd, buffer + read_pos, bytes_to_read) )
	 < 0 )
    {
	/*  Error reading  */
	info->ch_errno = errno;
	return (read_pos);
    }
    if (bytes_read < bytes_to_read)
    {
	/*  Only read some bytes  */
	switch (info->type)
	{
	  case CHANNEL_TYPE_CONNECTION:
	    fprintf (stderr, "%s: connection channel closed while reading\n",
		     function_name);
	    break;
	  case CHANNEL_TYPE_CHARACTER:
	    fprintf (stderr,
		     "%s: character special channel closed while reading\n",
		     function_name);
	    break;
	  case CHANNEL_TYPE_FIFO:
	    was_error = FALSE;
	    break;
	  default:
	    fprintf (stderr, "Bad channel type: %u\n", info->type);
	    a_prog_bug (function_name);
	    break;
	}
	if (was_error) fprintf (stderr,
				"%s: wanted: %d bytes  got: %d bytes\n",
				function_name, bytes_to_read, bytes_read);
	return (read_pos);
    }
    read_pos += bytes_read;
    /*  Drain data on connection into buffer  */
    if ( ( bytes_available = r_get_bytes_readable (info->fd) ) < 0 )
    {
	/*  ENOSYS is required because IRIX is broken  */
	if ( (errno == EINVAL) || (errno == ENOSYS) )
	{
	    errno = 0;
	    return (read_pos);
	}
	info->ch_errno = errno;
	fprintf (stderr, "%s: error getting bytes readable\t%s\n",
		 function_name, ERRSTRING);
	errno = 0;
	return (read_pos);
    }
    /*  Call any tap functions that were registered  */
    if (info->call_io_tap_functions) ch_call_tap_callbacks ();
    bytes_to_read = ( (bytes_available <= info->read_buf_len) ?
		      bytes_available : info->read_buf_len );
    if ( ( bytes_read = r_read (info->fd, info->read_buffer,
				bytes_to_read) )
	< bytes_to_read )
    {
	/*  Only read some bytes  */
	switch (info->type)
	{
	  case CHANNEL_TYPE_CONNECTION:
	    fprintf (stderr, "Connection channel closed while draining\t%s\n",
		     ERRSTRING);
	    break;
	  case CHANNEL_TYPE_CHARACTER:
	    fprintf (stderr,
		     "Character special channel closed while draining\t%s\n",
		     ERRSTRING);
	    break;
	  case CHANNEL_TYPE_FIFO:
	    fprintf (stderr, "FIFO channel closed while draining\t%s\n",
		     ERRSTRING);
	    break;
	  default:
	    fprintf (stderr, "Bad channel type: %u\n", info->type);
	    a_prog_bug (function_name);
	    break;
	}
	exit (RV_SYS_ERROR);
    }
    info->bytes_read = bytes_read;
    errno = 0;
    return (read_pos);
}   /*  End Function _ch_read_connection  */

static unsigned int _ch_read_datagram (KChannelHook info, char *buffer,
				       CONST char **newbuf,
				       unsigned char alignment,
				       unsigned int length)
/*  [SUMMARY] Read from a channel.
    [PURPOSE] This routine will read a number of bytes from a channel and
    places them into a buffer.
    <channel> The channel object.
    <buffer> A suggested buffer to write the data into.
    <newbuf> The buffer pointer used is written here. This may be a read-only
    buffer. If this is NULL, the suggested <<buffer>> is always used.
    <alignment> The number of bytes the buffer must be aligned to.
    <length> The number of bytes to write into the buffer.
    [RETURNS] The number of bytes read.
*/
{
    uaddr read_pos;
    iaddr bytes_to_copy, bytes_read;
    /*static char function_name[] = "_ch_read_datagram";*/

    /*  Do not call tap functions as datagrams need minimum delay  */
    /*  Copy what is left in the read buffer  */
    read_pos = info->bytes_read - info->read_buf_pos;
    if (read_pos > length) read_pos = length;
    m_copy (buffer, info->read_buffer + info->read_buf_pos, read_pos);
    info->read_buf_pos += read_pos;
    if (info->read_buf_pos >= info->bytes_read)
    {
	info->read_buf_pos = 0;
	info->bytes_read = 0;
    }
    if (read_pos >= length) return (read_pos);
    /*  Need to do some real reading  */
    if (length - read_pos >= info->read_buf_len)
    {
	/*  Want more than one packet: give just one packet  */
	bytes_read = READ (info->fd, buffer + read_pos, info->read_buf_len);
	if (bytes_read < 0) return (-1);
	if (bytes_read == 0)
	{
	    fprintf (stderr, "Datagram channel closed while reading\t%s\n",
		     ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	return (read_pos + bytes_read);
    }
    /*  Want less than one packet: read one packet into buffer and copy what's
	available and required  */
    bytes_read = READ (info->fd, info->read_buffer, info->read_buf_len);
    if (bytes_read < 0) return (-1);
    if (bytes_read == 0)
    {
	fprintf (stderr, "Datagram channel closed while draining\t%s\n",
		 ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    info->bytes_read = bytes_read;
    bytes_to_copy = length - read_pos;
    if (bytes_to_copy > bytes_read) bytes_to_copy = bytes_read;
    m_copy (buffer + read_pos, info->read_buffer, bytes_to_copy);
    info->read_buf_pos = bytes_to_copy;
    return (read_pos + bytes_to_copy);
}   /*  End Function _ch_read_datagram  */

static unsigned int _ch_write_descriptor (KChannelHook info,
					  CONST char *buffer,
					  unsigned int length)
/*  This routine will write a number of bytes from a buffer to a descriptor
    channel.
    The channel object must be given by  channel  .
    The buffer to read the data from must be pointed to by  buffer  .
    The number of bytes to write from the buffer must be pointed to by  length
    The routine returns the number of bytes written.
*/
{
    int bytes_written;
    int bytes_to_write;
    unsigned int write_pos = 0;
    /*static char function_name[] = "_ch_write_descriptor";*/

    if (info->ch_errno > 0)
    {
	/*  Error occurred previously during writing  */
	errno = info->ch_errno;
	return (0);
    }
    if ( (info->write_buf_pos == 0) && (length == info->write_buf_len) )
    {
	/*  The buffer is empty and we are writing a block the same size as the
	    buffer: bypass the buffer entirely  */
	return mywrite_raw (info, buffer, length);
    }
    if ( length < (info->write_buf_len - info->write_buf_pos) )
    {
	/*  Write buffer will not be filled up  */
	m_copy (info->write_buffer + info->write_buf_pos, buffer, length);
	info->write_buf_pos += length;
	return (length);
    }
    /*  Copy over as much as possible into write buffer  */
    write_pos = info->write_buf_len - info->write_buf_pos;
    m_copy (info->write_buffer + info->write_buf_pos, buffer, write_pos);
    info->write_buf_pos += write_pos;
    /*  Flush write buffer  */
    bytes_to_write = info->write_buf_pos - info->write_start_pos;
    if ( ( bytes_written = mywrite_raw (info, info->write_buffer,
					bytes_to_write) )
	< bytes_to_write )
    {
	info->write_start_pos = 0;
	return (FALSE);
    }
    info->write_start_pos = 0;
    /*  Write integral number of blocks directly  */
    bytes_to_write = length - write_pos;
    bytes_to_write -= bytes_to_write % (int) info->write_buf_len;
    if ( ( bytes_written = mywrite_raw (info, buffer + write_pos,
					bytes_to_write) )
	< bytes_to_write )
    {
	/*  Error writing  */
	info->ch_errno = errno;
	return (write_pos + bytes_written);
    }
    write_pos += bytes_written;
    /*  Place remainder into write buffer  */
    m_copy (info->write_buffer, buffer + write_pos, length - write_pos);
    info->write_buf_pos = length - write_pos;
    return (length);
}  /*  End Function _ch_write_descriptor  */

static int mywrite_raw (KChannelHook info, CONST char *buffer,
			unsigned int length)
/*  This routine will write a number of bytes from a buffer to a channel. This
    routine takes no account of buffering, nor a  write_converter_func  .
    The channel object must be given by  channel  .
    The buffer to read the data from must be pointed to by  buffer  .
    The number of bytes to write from the buffer must be pointed to by  length
    The routine returns the number of bytes written.
*/
{
    static char function_name[] = "mywrite_raw";

    if (buffer == NULL)
    {
	fprintf (stderr, "NULL buffer pointer passed\n");
	a_prog_bug (function_name);
    }
    if (length < 1) return (0);
    if (info->call_io_tap_functions) ch_call_tap_callbacks ();
    return WRITE (info->fd, buffer, length);
}   /*  End Function mywrite_raw  */

static flag seek_func (KChannelHook info, unsigned long position)
/*  [SUMMARY] Move read/write pointer.
    [PURPOSE] This routine will position the read and write pointers for a
    channel.
    <info> The channel object subclass information pointer.
    <position> The position (relative to the start of the channel data).
    [NOTE] This routine cannot be used for connection channels.
    [RETURNS] TRUE on success, else FALSE (indicating a seek request beyond the
    channel limits)
*/
{
    char dummy;
    unsigned long newpos, block_pos, block_len, count;
    static char function_name[] = "_ch_descriptor_seek_func";

    if (info->ch_errno > 0) return (FALSE);
    if ( (block_len = info->read_buf_len) > 0 )
    {
	if ( (info->write_buffer != NULL) &&
	     (block_len != info->write_buf_len) )
	{
	    fprintf (stderr, "Read buffer length: %lu not equal to write ",
		     (unsigned long) info->read_buf_len);
	    fprintf (stderr, "buffer length: %lu\n",
		     (unsigned long) info->write_buf_len);
	    a_prog_bug (function_name);
	}
    }
    else block_len = info->write_buf_len;
    block_pos = position % block_len;
    newpos = position - block_pos;
    if (lseek (info->fd, newpos, SEEK_SET) == -1)
    {
	info->ch_errno = errno;
	return (FALSE);
    }
    info->ch_errno = 0;
    if (info->read_buffer != NULL)
    {
	info->read_buf_pos = 0;
	info->bytes_read = 0;
	for (count = 0; count < block_pos; ++count)
	{
	    if (_ch_read_disc (info, &dummy, NULL, 0, 1) < 1)
	    {
		fprintf (stderr, "%s: error reading dummy character\t%s\n",
			 function_name, ERRSTRING);
		return (FALSE);
	    }
	}
    }
    if (info->write_buffer != NULL)
    {
	info->write_buf_pos = block_pos;
	info->write_start_pos = block_pos;
    }
    info->ch_errno = 0;
    return (TRUE);
}   /*  End Function seek_func  */

static flag get_bytes_readable (KChannelHook info)
/*  [SUMMARY] Count unread bytes.
    [PURPOSE] This routine will determine the number of bytes currently
    readable on a connection channel. This is equal to the maximum number of
    bytes that could be read from the channel at this time without blocking
    waiting for more input to be transmitted from the other end of the
    connection.
    <info> The channel object subclass information pointer.
    [RETURNS] The number of bytes readable on success, else -1 indicating error
*/
{
    int bytes_available;
    /*static char function_name[] = "_ch_descriptor_get_bytes_readable";*/

    if (info->ch_errno > 0)
    {
	errno = info->ch_errno;
	return (-1);
    }
    if ( ( bytes_available = r_get_bytes_readable (info->fd) ) < 0 )
    {
	info->ch_errno = errno;
	return (-1);
    }
    return (bytes_available + (int) info->bytes_read -
	    (int) info->read_buf_pos);
}   /*  End Function get_bytes_readable  */

static flag skip_func (KChannelHook info, uaddr num_bytes)
/*  [SUMMARY] Skip bytes in a channel, seeking if possible.
    [PURPOSE] This function will skip a specified number of bytes in a channel,
    jumping to the desired point without reading (if possible). If the channel
    does not support seeking (jumping), then the specified number of bytes are
    read from the channel.
    <channel> The channel object.
    <num_bytes> The number of bytes skip.
    [NOTE] Seeking may change the write pointer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned long read_pos, write_pos;

    if ( (info->bytes_read - info->read_buf_pos) >= num_bytes )
    {
	/*  Read buffer has enough data in it  */
	info->read_buf_pos += num_bytes;
	return (TRUE);
    }
    if ( !ch_tell (info->channel, &read_pos, &write_pos) ) return (FALSE);
    return ch_seek (info->channel, read_pos + num_bytes);
}   /*  End Function skip_func  */

static int get_descriptor (KChannelHook info)
/*  [SUMMARY] Get the file descriptor associated with a channel.
    <channel> The channel object.
    [RETURNS] The file descriptor on success, else -1 indicating error.
*/
{
    /*static char function_name[] = "ch_get_descriptor";*/

    return (info->fd);
}   /*  End Function get_descriptor  */

static iaddr iread (int fd, char *buf, uaddr nbytes)
/*  [SUMMARY] Read bytes from a file descriptor.
    [PURPOSE] This routine is similar to the system <<read(2)>> call, except
    that interrupted system calls are restarted.
    <fd> The file descriptor to read from. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to read.
    [RETURNS] The number of bytes requested on success, the number of bytes
    read on end of file (or closure) and -1 on error.
*/
{
    uaddr total_bytes_read = 0;
    iaddr bytes_read;

    while (total_bytes_read < nbytes)
    {
	errno = 0;
	if ( ( bytes_read = read (fd, buf + total_bytes_read,
				  nbytes - total_bytes_read) ) < 0 )
	{
	    if (errno == EINTR) continue;
	    return (-1);  /*  Unrecoverable error  */
	}
	/*  No error condition  */
	total_bytes_read += bytes_read;
	return (total_bytes_read);
    }
    return (total_bytes_read);
}   /*  End Function iread  */

static iaddr lwrite (int fd, CONST char *buf, uaddr nbytes)
/*  [SUMMARY] Write bytes to a file descriptor.
    <fd> The file descriptor to write to. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to write.
    [RETURNS] The number of bytes requested on success, else -1 indicating
    error.
*/
{
    return write (fd, buf, nbytes);
}   /*  End Function lwrite  */

static iaddr sparse_write (int fd, CONST char *buf, uaddr nbytes)
/*  [SUMMARY] Write bytes to a file descriptor, with holes for zero blocks.
    <fd> The file descriptor to write to. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to write.
    [RETURNS] The number of bytes requested on success, else -1 indicating
    error.
*/
{
    uaddr count;

    for (count = 0; count < nbytes; ++count)
	if (buf[count]) return write (fd, buf, nbytes);
    if (lseek (fd, nbytes, SEEK_CUR) == -1) return (-1);
    return (nbytes);
}   /*  End Function sparse_write  */

static flag iclose (int fd)
/*  [SUMMARY] Close a file descriptor.
    <fd> The file descriptor to close.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    if (close (fd) == 0) return (TRUE);
    fprintf (stderr, "Error closing descriptor: %d\t%s\n", fd, ERRSTRING);
    return (FALSE);
}   /*  End Function iclose  */
