/*LINTLIBRARY*/
/*  generic.c

    This code provides Channel objects.

    Copyright (C) 1992-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the creation, manipulation and
  deletion of channel objects.


    Written by      Richard Gooch   14-AUG-1992

    Updated by      Richard Gooch   13-DEC-1992

    Updated by      Richard Gooch   22-DEC-1992: Fixed bug in  ch_read_disc
  which hangs when reading zero length files.

    Updated by      Richard Gooch   1-JAN-1993: Added memory mapped disc files
  and changed parameters to  ch_gets  .

    Updated by      Richard Gooch   2-JAN-1993: Added  ch_tell  support for
  disc and connection channels.

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   1-MAR-1993: Added support for Ultrix
  mmap(2)

    Updated by      Richard Gooch   22-MAR-1993: Made reference to  conn_
  package in documentation on  ch_open_connection  .

    Updated by      Richard Gooch   28-MAR-1993: Created  ch_map_disc  routine.

    Updated by      Richard Gooch   12-APR-1993: Fixed  ch_read_connection  so
  that closure during read does not abort process.

    Updated by      Richard Gooch   10-MAY-1993: Added patch for SGImips.

    Updated by      Richard Gooch   24-JUL-1993: Added support for opening
  character special devices and FIFOs in  ch_open_file  and created the
  ch_test_for_io  routine.

    Updated by      Richard Gooch   26-JUL-1993: Fixed bug in  ch_tell  .

    Updated by      Richard Gooch   20-AUG-1993: Moved  ch_gets  and  ch_puts
  to  ch_misc.c

    Updated by      Richard Gooch   16-SEP-1993: Fixed memory mapping for
  Convex.

    Updated by      Richard Gooch   3-OCT-1993: Improved diagnostic messages
  when reading from closed asynchronous I/O channel.

    Updated by      Richard Gooch   19-FEB-1994: Made  ch_flush  more robust,
  prevented  ch_close  from flushing for asynchronous descriptor type and fixed
  bug with checking for NFS/local filesystems under Linux.

    Updated by      Richard Gooch   1-APR-1994: Added  ch_register_converters

    Updated by      Richard Gooch   13-APR-1994: Redefined interface for
  ch_register_converters  and changed implementation.

    Updated by      Richard Gooch   16-APR-1994: Disabled converter code for
  the VX/MVX because of some obscure bug (not mine!).

    Updated by      Richard Gooch   11-MAY-1994: Added un-named pipe support by
  creating  ch_create_pipe  .

    Updated by      Richard Gooch   16-MAY-1994: Added data sink support by
  creating  ch_create_sink  .

    Updated by      Richard Gooch   20-MAY-1994: Added  CONST  declaration
  where appropriate.

    Updated by      Richard Gooch   7-JUL-1994: Used  MAP_VARIABLE  in
  ch_map_disc  if available (ie. on Alpha/OSF1).

    Updated by      Richard Gooch   14-AUG-1994: Extended functionality of
  ch_open_file  .

    Updated by      Richard Gooch   24-AUG-1994: Created  ch_tap_events  .

    Updated by      Richard Gooch   12-SEP-1994: Completed implementation of
  ch_tap_events  .

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   14-NOV-1994: Supported disc channels in
  ch_seek  .

    Updated by      Richard Gooch   15-NOV-1994: Added  #include <unistd.h>  .

    Updated by      Richard Gooch   21-NOV-1994: Tidied up  VERIFY_CHANNEL
  macro.

    Updated by      Richard Gooch   25-NOV-1994: Renamed  ch_tap_events  to
  ch_tap_io_events  .

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/ch/main.c

    Updated by      Richard Gooch   27-NOV-1994: Made use of  c_  package.

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   25-DEC-1994: Removed group-write enable
  when writing files.

    Updated by      Richard Gooch   31-JAN-1995: Fixed descriptor leak in
  <ch_map_disc> when channels not mapped.

    Updated by      Richard Gooch   5-MAY-1995: Placate SGI compiler.

    Updated by      Richard Gooch   15-JUN-1995: Made use of IS_ALIGNED macro.

    Updated by      Richard Gooch   31-MAR-1996: Changed documentation style.

    Updated by      Richard Gooch   15-JUN-1996: Inlined memory copy for small
  transfers in <ch_read_disc> and <ch_read_memory>. Created
  <ch_read_and_swap_blocks>.

    Updated by      Richard Gooch   29-JUN-1996: Created
  <ch_swap_and_write_blocks>.

    Updated by      Richard Gooch   10-AUG-1996: Moved
  <ch_read_and_swap_blocks> and <ch_swap_and_write_blocks> routines to misc.c.

    Updated by      Richard Gooch   14-MAY-1997: Supported 'z' and 't' options
  in <ch_open_file>.

    Updated by      Richard Gooch   28-JUL-1997: Recognise FIFOs as local
  connections in <ch_test_for_local_connection>. Created <ch_open_stdout>.
  Fixed bug in <ch_seek> when seeking to within one block of EndOfFile.

    Updated by      Richard Gooch   2-AUG-1997: Made use of new function
  <r_fork_and_execvp>.

    Updated by      Richard Gooch   14-AUG-1997: Moved over to subclassing
  system for specific channel types.

    Updated by      Richard Gooch   20-AUG-1997: Moved gunzip bits to gunzip.c.

    Updated by      Richard Gooch   17-NOV-1997: Delayed calling close_func
  hook until converter close functions called in <ch_close>. Pass channel to
  converter close functions.

    Updated by      Richard Gooch   29-DEC-1997: Created experimental
  <ch_test_valid> function and made use of it in VERIFY_CHANNEL().

    Updated by      Richard Gooch   12-FEB-1998: Changed VERIFY_CHANNEL back to
  a macro because <ch_test_valid> eats performance. Added alignment check.

    Updated by      Richard Gooch   19-FEB-1998: Created freelist.

    Updated by      Richard Gooch   22-FEB-1998: Removed casts for memory
  functions.

    Updated by      Richard Gooch   26-FEB-1998: Moved <ch_skip> from generic.c
  and implemented hook interface.

    Updated by      Richard Gooch   12-MAY-1998: Moved <ch_get_descriptor> from
  descriptor.c

    Updated by      Richard Gooch   10-JUN-1999: Took account of change to
  <m_test_access>.

    Updated by      Richard Gooch   13-AUG-1999: Make egcs happy.

    Updated by      Richard Gooch   18-AUG-1999: Created experimental function
  <ch_read2>.

    Updated by      Richard Gooch   11-SEP-1999: Added calls to function
  <s_check_for_int>.

    Last updated by Richard Gooch   14-FEB-2000: Fixed bug in allocating last
  entry of freelist.


*/
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_ch_hooks.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_c.h>
#include <karma_s.h>
#include <os.h>


#define ALLOC_SIZE 64
#define CHANNEL_USED_MAGIC_NUMBER 3498653275U
#define CHANNEL_FREE_MAGIC_NUMBER 2429874538U
#define CONVERTER_MAGIC_NUMBER 2495849834U

#define VERIFY_USED_CHANNEL(channel) if (channel == NULL) \
{fprintf (stderr, "NULL channel passed\n"); \
 a_prog_bug (function_name); } \
if ((uaddr) channel % sizeof (void *) != 0) \
{fprintf (stderr, "Channel pointer not aligned properly\n"); \
 a_prog_bug (function_name); } \
if (channel->magic_number != CHANNEL_USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid channel object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FREE_CHANNEL(channel) if (channel == NULL) \
{fprintf (stderr, "NULL channel passed\n"); \
 a_prog_bug (function_name); } \
if ((uaddr) channel % sizeof (void *) != 0) \
{fprintf (stderr, "Channel pointer not aligned properly\n"); \
 a_prog_bug (function_name); } \
if (channel->magic_number != CHANNEL_FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed channel object\n"); \
 a_prog_bug (function_name); }

#define CONV_BUF_SIZE (unsigned int) 4096

/*  Internal definition of ChConverter object structure type  */
struct _converter_struct
{
    unsigned int magic_number;
    Channel channel;
    unsigned int (*size_func) (Channel channel, void **info);
    unsigned int (*read_func) (Channel channel, char *buffer,
			       unsigned int length, void **info);
    unsigned int (*write_func) (Channel channel, char *buffer,
				unsigned int length, void **info);
    flag (*flush_func) (Channel channel, void **info);
    void (*close_func) (void *info, Channel channel);
    void *info;
    struct _converter_struct *prev;
    struct _converter_struct *next;
};

/*  Internal definition of Channel object structure type  */
struct channel_type
{
    unsigned int magic_number;
    uaddr abs_read_pos;
    uaddr abs_write_pos;
    KChannelHook hook_info;
    KChannelHookFuncs hooks;
    ChConverter top_converter;
    ChConverter next_converter;
    struct channel_type *prev;
    struct channel_type *next;
};


/*  Private data follows  */
static Channel first_channel = NULL;
static Channel first_free = NULL;
static KCallbackList tap_list = NULL;


/*  Private functions  */


/*  Public functions follow  */

/*UNPUBLISHED_FUNCTION*/
Channel ch_create_generic (KChannelHook hook_info, KChannelHookFuncs hooks)
/*  [SUMMARY] Create a generic channel object.
    <hook_info> The subclass-specific information pointer.
    <hooks> The function hooks.
    [RETURNS] A channel object on success, else NULL.
*/
{
    Channel new;
    unsigned int count;
    static flag registered_exit_func = FALSE;
    static char function_name[] = "ch_create_generic";

    if (!registered_exit_func)
    {
	/*  Register function to be called upon processing of  exit(3)  */
#ifdef HAS_ATEXIT
	atexit (ch_close_all_channels);
#endif
#ifdef HAS_ON_EXIT
	on_exit (ch_close_all_channels, (caddr_t) NULL);
#endif
	registered_exit_func = TRUE;
    }
    /*  Get new entry  */
    if (first_free == NULL)
    {
	/*  Have to allocate some more for the free list  */
	if ( ( new = m_alloc (ALLOC_SIZE * sizeof *new) ) == NULL )
	{
	    m_error_notify (function_name, "new channel entries");
	    return (NULL);
	}
	m_clear (new, ALLOC_SIZE * sizeof *new);
	/*  Add them to the free list  */
	first_free = new;
	for (count = 0; count < ALLOC_SIZE - 1; ++count)
	{
	    new[count].magic_number = CHANNEL_FREE_MAGIC_NUMBER;
	    new[count].next = new + count + 1;
	}
	new[count].magic_number = CHANNEL_FREE_MAGIC_NUMBER;
    }
    /*  Take one from the free list  */
    new = first_free;
    VERIFY_FREE_CHANNEL (new);
    first_free = new->next;
    /*  Fill in entry  */
    new->magic_number = CHANNEL_USED_MAGIC_NUMBER;
    new->next = NULL;
    new->abs_read_pos = 0;
    new->abs_write_pos = 0;
    new->hook_info = hook_info;
    new->hooks = hooks;
    new->top_converter = NULL;
    new->next_converter = NULL;
    /*  Place channel object into list  */
    new->prev = NULL;
    new->next = first_channel;
    if (first_channel != NULL) first_channel->prev = new;
    first_channel = new;
    return (new);
}   /*  End Function ch_create_generic  */

/*PUBLIC_FUNCTION*/
flag ch_close (Channel channel)
/*  [SUMMARY] Close a channel.
    [PURPOSE] This routine will close a channel object. The write buffer will
    be flushed prior to closure.
    <channel> The channel object.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag return_value = TRUE;
    ChConverter converter;
    static char function_name[] = "ch_close";

    VERIFY_USED_CHANNEL (channel);
    if (channel->hooks.flush != NULL)
	if ( !ch_flush (channel) ) return_value = FALSE;
    if (channel->hooks.close == NULL)
    {
	fprintf (stderr, "Channel closure not supported\n");
	a_prog_bug (function_name);
    }
    while ( (converter = channel->top_converter) != NULL )
    {
	/*  Call the close routine  */
	channel->next_converter = converter->next;
	channel->top_converter = converter->next;
	if (converter->close_func != NULL)
	{
	    (*converter->close_func) (converter->info, channel);
	}
	converter->magic_number = 0;
	m_free (converter);
    }
    if ( !(*channel->hooks.close) (channel->hook_info) ) return_value = FALSE;
    /*  Remove channel object from list  */
    if (channel->next != NULL)
    {
	/*  Another entry further in the list  */
	channel->next->prev = channel->prev;
    }
    if (channel->prev != NULL)
    {
	/*  Another entry previous in the list  */
	channel->prev->next = channel->next;
    }
    if (channel == first_channel)
    {
	/*  Channel is first in list: make next entry the first  */
	first_channel = channel->next;
    }
    /*  Kill magic number entry and everything else for security  */
    m_clear (channel, sizeof *channel);
    channel->magic_number = CHANNEL_FREE_MAGIC_NUMBER;
    channel->next = first_free;
    first_free = channel;
    return (return_value);
}   /*  End Function ch_close  */

/*PUBLIC_FUNCTION*/
flag ch_flush (Channel channel)
/*  [SUMMARY] Flush the write buffer of a channel object.
    <channel> The channel object.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    ChConverter converter;
    static char function_name[] = "ch_flush";

    /*  THIS ROUTINE MUST NOT BE CALLED FROM ANY OF THE  ch_write  SUPPORT
	ROUTINES  */
    VERIFY_USED_CHANNEL (channel);
    if (channel->hooks.flush == NULL)
    {
	fprintf (stderr, "Channel flushing not supported\n");
	a_prog_bug (function_name);
    }
    for (converter = channel->top_converter; converter != NULL;
	 converter = converter->next)
    {
	/*  Call the flush routine  */
	channel->next_converter = converter->next;
	if ( !(*converter->flush_func) (channel, &converter->info) )
	{
	    channel->next_converter = channel->top_converter;
	    return (FALSE);
	}
    }
    channel->next_converter = channel->top_converter;
    return (*channel->hooks.flush) (channel->hook_info);
}   /*  End Function ch_flush  */

/*PUBLIC_FUNCTION*/
uaddr ch_read (Channel channel, char *buffer, uaddr length)
/*  [SUMMARY] Read from a channel.
    [PURPOSE] This routine will read a number of bytes from a channel and
    places them into a buffer.
    <channel> The channel object.
    <buffer> The buffer to write the data into.
    <length> The number of bytes to write into the buffer.
    [NOTE] If the channel is a connection and the number of bytes readable from
    the connection is equal to or more than <<length>> the routine will NOT
    block.
    [RETURNS] The number of bytes read.
*/
{
    uaddr num_read = 0;  /*  Initialised to keep compiler happy  */
    ChConverter converter, old_converter;
    static char function_name[] = "ch_read";

    VERIFY_USED_CHANNEL (channel);
    if (buffer == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    if (channel->hooks.read == NULL)
    {
	fprintf (stderr, "Read operation not supported\n");
	a_prog_bug (function_name);
    }
    if ( (converter = channel->next_converter) != NULL )
    {
	/*  Pass it on to a read converter  */
	old_converter = channel->next_converter;
	channel->next_converter = converter->next;
	if ( ( num_read = (*converter->read_func) (channel, buffer, length,
						   &converter->info) )
	     < length )
	{
	    fprintf (stderr,
		     "%s: error in convertor, expected: %lu bytes, got: %lu\n",
		     function_name, (unsigned long) length,
		     (unsigned long) num_read);
	    channel->next_converter = channel->top_converter;
	    if (converter == channel->top_converter)
		channel->abs_read_pos += num_read;
	    return (num_read);
	}
	/*  Read converter has produced all data  */
	if (converter == channel->top_converter)
	    channel->abs_read_pos += num_read;
	channel->next_converter = old_converter;
	return (num_read);
    }
    if ( s_check_for_int () )
    {
	fprintf (stderr, "%s: interrupt detected\n", function_name);
	errno = EINTR;
	return (0);
    }
    num_read = (*channel->hooks.read) (channel->hook_info, buffer, NULL, 0,
				       length);
    if (channel->top_converter == NULL) channel->abs_read_pos += num_read;
    return (num_read);
}   /*  End Function ch_read  */

/*EXPERIMENTAL_FUNCTION*/
uaddr ch_read2 (Channel channel, char *buffer, CONST char **newbuf,
		unsigned char alignment, uaddr length)
/*  [SUMMARY] Read from a channel.
    [PURPOSE] This routine will read a number of bytes from a channel and
    places them into a buffer.
    <channel> The channel object.
    <buffer> A suggested buffer to write the data into.
    <newbuf> The buffer pointer used is written here. This may be a read-only
    buffer. If this is NULL, the suggested <<buffer>> is always used.
    <alignment> The number of bytes the buffer must be aligned to.
    <length> The number of bytes to write into the buffer.
    [NOTE] If the channel is a connection and the number of bytes readable from
    the connection is equal to or more than <<length>> the routine will NOT
    block.
    [RETURNS] The number of bytes read.
*/
{
    uaddr num_read = 0;  /*  Initialised to keep compiler happy  */
    static char function_name[] = "ch_read2";

    VERIFY_USED_CHANNEL (channel);
    if (channel->hooks.read == NULL)
    {
	fprintf (stderr, "Read operation not supported\n");
	a_prog_bug (function_name);
    }
    if (newbuf != NULL) *newbuf = buffer;
    if (channel->next_converter != NULL)
	return ch_read (channel, buffer, length);
    if ( s_check_for_int () )
    {
	fprintf (stderr, "%s: interrupt detected\n", function_name);
	errno = EINTR;
	return (0);
    }
    num_read = (*channel->hooks.read) (channel->hook_info, buffer, newbuf,
				       alignment, length);
    switch (alignment)
    {
      case 0:
      case 1:
	break;
      case 2:
      case 4:
      case 8:
	if ( (uaddr) *newbuf & (alignment - 1) )
	{
	    fprintf (stderr, "Buffer was not %u bytes aligned\n", alignment);
	    a_prog_bug (function_name);
	}
	break;
      default:
	if ( (uaddr) *newbuf % alignment != 0 )
	{
	    fprintf (stderr, "Buffer was not %u bytes aligned\n", alignment);
	    a_prog_bug (function_name);
	}
	break;
    }
    channel->abs_read_pos += num_read;
    return (num_read);
}   /*  End Function ch_read2  */

/*PUBLIC_FUNCTION*/
uaddr ch_write (Channel channel, CONST char *buffer, uaddr length)
/*  [SUMMARY] Write to a channel.
    [PURPOSE] This routine will write a number of bytes from a buffer to a
    channel.
    <channel> The channel object.
    <buffer> The buffer to read the data from.
    <length> The number of bytes to read from the buffer and write to the
    channel.
    [RETURNS] The number of bytes written.
*/
{
    ChConverter converter, old_converter;
    uaddr bytes_to_write, bytes_written;
    uaddr num_written = 0;  /*  Initialised to keep compiler happy  */
    static char function_name[] = "ch_write";

    VERIFY_USED_CHANNEL (channel);
    if (buffer == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    if (channel->hooks.write == NULL)
    {
	fprintf (stderr, "Write operation not supported\n");
	a_prog_bug (function_name);
    }
    if ( (converter = channel->next_converter) != NULL )
    {
	/*  Pass it on to a write converter  */
	char tmp_buf[CONV_BUF_SIZE];

	old_converter = channel->next_converter;
	channel->next_converter = converter->next;
	for (num_written = 0; length > 0;
	     length -= bytes_written, buffer += bytes_written,
	     num_written += bytes_written)
	{
	    bytes_to_write = (length > CONV_BUF_SIZE) ? CONV_BUF_SIZE : length;
	    m_copy (tmp_buf, buffer, bytes_to_write);
	    if ( ( bytes_written =
		   (*converter->write_func) (channel, tmp_buf, bytes_to_write,
					     &converter->info) )
		 < bytes_to_write )
	    {
		channel->next_converter = channel->top_converter;
		if (converter == channel->top_converter)
		{
		    channel->abs_write_pos += num_written;
		}
		return (num_written);
	    }
	}
	/*  Write converter has consumed all data  */
	if (converter == channel->top_converter)
	{
	    channel->abs_write_pos += num_written;
	}
	channel->next_converter = old_converter;
	return (num_written);
    }
    if ( s_check_for_int () )
    {
	fprintf (stderr, "%s: interrupt detected\n", function_name);
	errno = EINTR;
	return (0);
    }
    num_written = (*channel->hooks.write) (channel->hook_info, buffer, length);
    if (channel->top_converter == NULL) channel->abs_write_pos += num_written;
    return (num_written);
}   /*  End Function ch_write  */

/*PUBLIC_FUNCTION*/
void ch_close_all_channels ()
/*  [SUMMARY] Close all open channels.
    [PURPOSE] The routine will close all open channels. The routine is meant to
    be called from the exit(3) function.
    [RETURNS] Nothing.
*/
{
    while (first_channel != NULL) ch_close (first_channel);
}   /*  End Function ch_close_all_channels  */

/*PUBLIC_FUNCTION*/
flag ch_seek (Channel channel, unsigned long position)
/*  [SUMMARY] Move read/write pointer.
    [PURPOSE] This routine will position the read and write pointers for a
    channel.
    <channel> The channel object.
    <position> The position (relative to the start of the channel data).
    [NOTE] This routine cannot be used for connection channels.
    [RETURNS] TRUE on success, else FALSE (indicating a seek request beyond the
    channel limits)
*/
{
    static char function_name[] = "ch_seek";

    VERIFY_USED_CHANNEL (channel);
    if (channel->top_converter != NULL)
    {
	fprintf (stderr, "Cannot seek channels with converter functions\n");
	a_prog_bug (function_name);
    }
    if (channel->hooks.seek == NULL)
    {
	fprintf (stderr, "Seek operation not supported\n");
	a_prog_bug (function_name);
    }
    if ( !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error flushing\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
    if ( !(*channel->hooks.seek) (channel->hook_info, position) ) return FALSE;
    channel->abs_read_pos = position;
    channel->abs_write_pos = position;
    return (TRUE);
}   /*  End Function ch_seek  */

/*PUBLIC_FUNCTION*/
int ch_get_bytes_readable (Channel channel)
/*  [SUMMARY] Count unread bytes.
    [PURPOSE] This routine will determine the number of bytes currently
    readable on a connection channel. This is equal to the maximum number of
    bytes that could be read from the channel at this time without blocking
    waiting for more input to be transmitted from the other end of the
    connection.
    <channel> The channel object.
    [RETURNS] The number of bytes readable on success, else -1 indicating error
*/
{
    ChConverter converter;
    int bytes_available;
    int conv_bytes = 0;
    static char function_name[] = "ch_get_bytes_readable";

    VERIFY_USED_CHANNEL (channel);
    if (channel->hooks.get_bytes_readable == NULL)
    {
	fprintf (stderr, "Function: %s not supported\n", function_name);
	return (-1);
    }
    for (converter = channel->top_converter, conv_bytes = 0;
	 converter != NULL;
	 converter = converter->next)
    {
	/*  Call the close routine  */
	channel->next_converter = converter->next;
	conv_bytes += (*converter->size_func) (channel, &converter->info);
    }
    channel->next_converter = channel->top_converter;
    if ( ( bytes_available =
	   channel->hooks.get_bytes_readable (channel->hook_info) ) < 0 )
    {
	return (-1);
    }
    return (conv_bytes + bytes_available);
}   /*  End Function ch_get_bytes_readable  */

/*PUBLIC_FUNCTION*/
flag ch_test_for_io (Channel channel)
/*  [SUMMARY] Test if I/O possible on channel.
    [PURPOSE] This routine will test if a channel object is capable of
    supporting reading and writing operations. Most channels fall under this
    category. The notable exceptions are the dock channel and channels created
    by a call to <<ch_attach_to_asynchronous_descriptor>>.
    <channel> The channel object.
    [RETURNS] TRUE if the channel is capable of reading and writing, else FALSE
*/
{
    static char function_name[] = "ch_test_for_io";

    VERIFY_USED_CHANNEL (channel);
    if ( (channel->hooks.read == NULL) && (channel->hooks.write == NULL) )
	return (FALSE);
    return (TRUE);
}   /*  End Function ch_test_for_io  */

/*EXPERIMENTAL_FUNCTION*/
flag ch_test_for_converters (Channel channel)
/*  [SUMMARY] Test if a channel object has converters.
    <channel> The channel object.
    [RETURNS] TRUE if the channel object is has converters, else FALSE.
*/
{
    static char function_name[] = "ch_test_for_converters";

    VERIFY_USED_CHANNEL (channel);
    return (channel->top_converter == NULL) ? FALSE : TRUE;
}   /*  End Function ch_test_for_converters  */

/*PUBLIC_FUNCTION*/
flag ch_tell (Channel channel, unsigned long *read_pos,
	      unsigned long *write_pos)
/*  [SUMMARY] Get the read and write pointers for a channel.
    <channel> The channel object.
    <read_pos> The read position (relative to the start of the channel data)
    will be written here.
    <write_pos> The write position (relative to the start of the channel data)
    will be written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "ch_tell";

    VERIFY_USED_CHANNEL (channel);
    if ( !ch_test_for_io (channel) )
    {
	fprintf (stderr, "I/O not possible on channel\n");
	a_prog_bug (function_name);
    }
    *read_pos = channel->abs_read_pos;
    *write_pos = channel->abs_write_pos;
    return (TRUE);
}   /*  End Function ch_tell  */

/*PUBLIC_FUNCTION*/
ChConverter ch_register_converter (Channel channel,
				   unsigned int (*size_func) (),
				   unsigned int (*read_func) (),
				   unsigned int (*write_func) (),
				   flag (*flush_func) (),
				   void (*close_func) (),
				   void *info)
/*  [SUMMARY] Register channel converter function.
    [PURPOSE] This routine will register a set of converter functions which
    will be called when channel data is read or written. The operation of these
    routines is transparent. Converter functions are useful for automatic
    compression and encryption of data streams.
    It is permissable to register multiple converter functions with a channel.
    Converter functions are pushed down from the top (application) level. In
    other words, the last converter functions registered are called first.
    [NOTE] Converters impose restrictions on channel operations (i.e.
    [<ch_seek>] cannot be called).
    [NOTE] Converter functions are expected to provide their own buffering as
    needed.
    <channel> The channel object.
    <size_func> The function which will determine (approximately) how many
    bytes would be returned by the <<read_func>>. This routine is called when
    [<ch_get_bytes_readable>] is called for the channel. The prototype function
    is [<ch_PROTO_size_func>].
    <read_func> The function which will convert data when reading.
    The prototype function is [<ch_PROTO_read_func>].
    <write_func> The function which will convert data when writing. If this is
    NULL, no write conversion is performed. The prototype function is
    [<ch_PROTO_write_func>].
    <flush_func> The function which is called when the channel is to be flushed
    The prototype function is [<ch_PROTO_flush_func>].
    <close_func> The function which is called when the channel is closed. If
    this is NULL, no special action is taken upon channel closure.
    The prototype function is [<ch_PROTO_close_func>].
    [NOTE] The <<flush_func>> will be called prior to the <<close_func>> upon
    channel closure.
    [NOTE]
    The sequence of events when the application level calls [<ch_write>] is:
      The last registered write converter is popped from the stack and called.
      This write converter may buffer some or all of the data. It may call
        [<ch_write>] with some converted data.
      When [<ch_write>] is called from a write converter, the next last
        registered write converter is popped from the stack and called.
      This sequence is continued until data is actually transferred into the
      channel write buffer.
    A similar sequence of events occurs when  ch_read  is called.
    The sequence of events when the application level calls  ch_flush  is:
      The last registered flush converter is popped from the stack and called.
      This flush converter MUST write all data in it's buffer using  ch_write
      When  ch_write  is called from a flush converter, the next last
        registered write converter is popped from the stack and called.
      When the last registered flush converter returns, the sequence is
      repeated with the next last flush converter, and so on, until all data
      in all write buffers is flushed, including the channel write buffer.
    [RETURNS] A ChConverter object on success (which may be used in a call to
    [<ch_unregister_converter>]), else NULL.
*/
{
    ChConverter new;
    static char function_name[] = "ch_register_converter";

    VERIFY_USED_CHANNEL (channel);
    if ( !ch_test_for_io (channel) )
    {
	fprintf (stderr, "I/O not possible on channel\n");
	a_prog_bug (function_name);
    }
    if ( (size_func == NULL) || (read_func == NULL) || (write_func == NULL) ||
	 (flush_func == NULL) )
    {
	fprintf (stderr, "NULL functions pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "converter structure");
	return (NULL);
    }
    new->magic_number = CONVERTER_MAGIC_NUMBER;
    new->channel = channel;
    new->size_func = size_func;
    new->read_func = read_func;
    new->write_func = write_func;
    new->flush_func = flush_func;
    new->close_func = close_func;
    new->info = info;
    new->prev = NULL;
    new->next = channel->top_converter;
    if (channel->top_converter != NULL) channel->top_converter->prev = new;
    channel->top_converter = new;
    channel->next_converter = channel->top_converter;
    return (new);
}   /*  End Function ch_register_converter  */

/*PUBLIC_FUNCTION*/
void ch_unregister_converter (ChConverter converter)
/*  [SUMMARY] Unregister converter.
    [PURPOSE] This routine will unregister a set of converter functions
    previously registered with [<ch_register_converter>]. This will cause the
    registered flush and close functions to be called.
    <converter> The ChConverter object.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    ChConverter old_converter;
    static char function_name[] = "ch_unregister_converter";

    if (converter == NULL)
    {
	fprintf (stderr, "NULL converter passed\n");
	a_prog_bug (function_name);
    }
    if (converter->magic_number != CONVERTER_MAGIC_NUMBER)
    {
	fprintf (stderr, "Invalid converter object\n");
	a_prog_bug (function_name);
    }
    channel = converter->channel;
    /*  Preserve the stack  */
    old_converter = channel->next_converter;
    /*  Call flush and close functions  */
    channel->next_converter = converter->next;
    (*converter->flush_func) (channel, &converter->info);
    if (converter->close_func != NULL)
    {
	(*converter->close_func) (converter->info, channel);
    }
    /*  Restore the stack  */
    channel->next_converter = old_converter;
    /*  Unlink from list  */
    if (converter->prev != NULL) converter->prev->next = converter->next;
    if (converter->next != NULL) converter->next->prev = converter->prev;
    if (converter == channel->next_converter) 
    {
	channel->next_converter = converter->next;
    }
    if (converter == channel->top_converter)
    {
	channel->top_converter = converter->next;
    }
    converter->magic_number = 0;
    m_free (converter);
}   /*  End Function ch_unregister_converter  */

/*PUBLIC_FUNCTION*/
KCallbackFunc ch_tap_io_events (void (*tap_func) (), void *info)
/*  [SUMMARY] Register I/O tap function.
    [PURPOSE] This routine will tap I/O events by calling a registered function
    whenever data is transferred to/from a disc, connection or FIFO channel.
    Reading and writing memory mapped or memory channels does *not* constitute
    an I/O event.
    Multiple tap functions may be registered, with the first one registered
    being the first one called upon a channel I/O event.
    <tap_func> The function which is called when I/O occurs. The prototype
    function is [<ch_PROTO_tap_func>].
    <info> The arbitrary information passed to the tap function. This may be
    NULL.
    [RETURNS] A KCallbackFunc. On failure, the process aborts.
*/
{
    return c_register_callback (&tap_list, ( flag (*) () ) tap_func,
				info, NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function ch_tap_io_events  */

/*UNPUBLISHED_FUNCTION*/
KChannelHook ch_get_hook_info (Channel channel)
/*  [SUMMARY] Get the channel subclass hook information pointer.
    <channel> The channel object.
    [RETURNS] The hook information pointer.
*/
{
    static char function_name[] = "ch_get_hook_info";

    VERIFY_USED_CHANNEL (channel);
    return (channel->hook_info);
}   /*  End Function ch_get_hook_info  */

/*UNPUBLISHED_FUNCTION*/
void ch_call_tap_callbacks ()
/*  [SUMMARY] Call any registered tap callbacks.
    [RETURNS] Nothing.
*/
{
    c_call_callbacks (tap_list, NULL);
}   /*  End Function ch_call_tap_callbacks  */

/*EXPERIMENTAL_FUNCTION*/
flag ch_test_valid (Channel channel, CONST char *function_name)
/*  [SUMMARY] Test if a channel is valid.
    <channel> The channel object.
    <function_name> If not NULL diagnostic messages are written to the standard
    error prepended with <<function_name>>.
    [NOTE] Frequent use may degrade performance.
    [RETURNS] TRUE if the channel is valid, else FALSE.
*/
{
    if (channel == NULL)
    {
	if (function_name == NULL) return (FALSE);
	fprintf (stderr, "%s: NULL channel object\n", function_name);
	return (FALSE);
    }
    if ( (uaddr) channel % sizeof (double) != 0)
    {
	if (function_name == NULL) return (FALSE);
	fprintf (stderr, "%s: channel object is not on a channel boundary\n",
		 function_name);
	return (FALSE);
    }
    if ( !m_test_access (channel, sizeof *channel, FALSE) )
    {
	if (function_name == NULL) return (FALSE);
	fprintf (stderr, "%s: channel object is not readable\n",function_name);
	return (FALSE);
    }
    if (channel->magic_number != CHANNEL_USED_MAGIC_NUMBER)
    {
	if (function_name == NULL) return (FALSE);
	fprintf (stderr, "%s: invalid channel object\n", function_name);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function ch_test_valid  */

/*UNPUBLISHED_FUNCTION*/
void ch_change_hooks (Channel channel, KChannelHookFuncs hooks)
/*  [SUMMARY] Change the function hooks for a channel.
    <hooks> The function hooks.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "ch_change_hooks";

    VERIFY_USED_CHANNEL (channel);
    channel->hooks = hooks;
}   /*  End Function ch_change_hooks  */

/*EXPERIMENTAL_FUNCTION*/
flag ch_skip (Channel channel, uaddr num_bytes)
/*  [SUMMARY] Skip bytes in a channel, seeking if possible.
    [PURPOSE] This function will skip a specified number of bytes in a channel,
    jumping to the desired point without reading (if possible). If the channel
    does not support seeking (jumping), then the specified number of bytes are
    read from the channel.
    <channel> The channel object.
    <num_bytes> The number of bytes skip.
    [NOTE] Seeking may change the write pointer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned long old_read_pos;
    static char function_name[] = "ch_skip";

    VERIFY_USED_CHANNEL (channel);
    if (num_bytes < 1) return (TRUE);
    if (channel->top_converter != NULL)
    {
	if (ch_drain (channel, num_bytes) < num_bytes) return (FALSE);
	return (TRUE);
    }
    if (channel->hooks.skip == NULL)
	return ch_seek (channel, channel->abs_read_pos + num_bytes);
    old_read_pos = channel->abs_read_pos;
    if ( !(*channel->hooks.skip) (channel->hook_info, num_bytes) )
	return (FALSE);
    if (channel->abs_read_pos > old_read_pos) return (TRUE);
    channel->abs_read_pos += num_bytes;
    return (TRUE);
}   /*  End Function ch_skip  */

/*PUBLIC_FUNCTION*/
int ch_get_descriptor (Channel channel)
/*  [SUMMARY] Get the file descriptor associated with a channel.
    <channel> The channel object.
    [RETURNS] The file descriptor on success, else -1 indicating error.
*/
{
    static char function_name[] = "ch_get_descriptor";

    VERIFY_USED_CHANNEL (channel);
    if (channel->hooks.get_descriptor == NULL) return (-1);
    return (*channel->hooks.get_descriptor) (channel->hook_info);
}   /*  End Function ch_get_descriptor  */


/*  Private functions follow  */
