/*LINTLIBRARY*/
/*  memory.c

    This code provides Channel objects.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the creation, manipulation and
  deletion of channel objects.


    Written by      Richard Gooch   15-AUG-1997: Moved bits from descriptor.c

    Updated by      Richard Gooch   2-SEP-1997: Switched to ERRSTRING.

    Updated by      Richard Gooch   16-SEP-1997: Bug fix in close function:
  don't unmap memory buffer if NULL.

    Updated by      Richard Gooch   17-NOV-1997: Bug fix in <ch_test_for_mmap>:
  did not trap NULL hook pointer.

    Updated by      Richard Gooch   22-FEB-1998: Removed casts for memory
  functions.

    Updated by      Richard Gooch   11-MAY-1998: Created experimental function
  <ch_map_fd>.

    Updated by      Richard Gooch   21-JUL-1998: Add 'z' option when falling
  back to plain open in <ch_map_disc>.

    Updated by      Richard Gooch   13-NOV-1998: Made use of <m_map_fd> and
  <m_unmap> functions.

    Updated by      Richard Gooch   13-JAN-1999: Documented "CH_MMAP_DISABLE"
  environment variable.

    Updated by      Richard Gooch   18-AUG-1999: Modified <read_func> to
  support passing back a read-only buffer.

    Last updated by Richard Gooch   26-JAN-2000: Documentation fix.


*/
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_ch_hooks.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_r.h>
#include <os.h>
#ifdef OS_Linux
#  include <sys/vfs.h>
     /*  This is in  linux/nfs_fs.h  but contains declarations for Kernel
	 internals. Hope this will be fixed one day.  */
#  define NFS_SUPER_MAGIC 0x6969
#endif

#define INFO_MAGIC_NUMBER 439873532

#define VERIFY_INFO(info) if (info == NULL) \
{fprintf (stderr, "NULL channel subclass info passed\n"); \
 a_prog_bug (function_name); } \
if (info->magic_number != INFO_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid channel subclass object\n"); \
 a_prog_bug (function_name); }

#define MMAP_LARGE_SIZE 1048576


struct channel_hook_type
{
    unsigned int magic_number;
    int fd;                          /*  If > -1 then mmapped  */
    char *memory_buffer;
    unsigned int mmap_access_count;
    unsigned int mem_buf_len;
    unsigned int mem_buf_read_pos;
    unsigned int mem_buf_write_pos;
    flag mem_buf_allocated;
};


/*  Private functions  */
STATIC_FUNCTION (KChannelHook info_alloc, () );
STATIC_FUNCTION (flag _ch_memory_close_func, (KChannelHook info) );
STATIC_FUNCTION (flag _ch_memory_flush_func, (KChannelHook info) );
STATIC_FUNCTION (unsigned int _ch_memory_read_func,
		 (KChannelHook info, char *buffer,
		  CONST char **newbuf, unsigned char alignment,
		  unsigned int length) );
STATIC_FUNCTION (unsigned int _ch_memory_write_func,
		 (KChannelHook info, CONST char *buffer,unsigned int length) );
STATIC_FUNCTION (flag _ch_memory_seek_func,
		 (KChannelHook info, unsigned long position) );
STATIC_FUNCTION (int _ch_memory_get_descriptor, (KChannelHook info) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
Channel ch_map_disc (CONST char *filename, unsigned int option, flag writable,
		     flag update_on_write)
/*  [SUMMARY] Map a disc file.
    [PURPOSE] This routine will open a memory channel with the memory pages
    being mapped from a disc file. The disc file must already exist.
    <filename> The pathname of the file to open.
    <option> Control value which determines whether the channel is opened as an
    ordinary disc file or is mapped. See [<ch_MAP_CONTROLS>] for legal values.
    If the file is not mapped then the routine will attempt to open an ordinary
    disc channel. If the file is opened as a disc channel the access mode is:
    "r".
    <writable> If the mapped pages are to be writable, this must be TRUE. If
    this is FALSE and the memory pages are written to, a segmentation fault
    occurs.
    <update_on_write> If the disc file should be updated when the memory pages
    are written to, this must be TRUE. If this is FALSE, then a write to a
    memory page causes the page to be copied into swap space and the process
    retains a private copy of the page from this point on.
    [NOTE] If <<update_on_write>> is FALSE and <<writable>> is TRUE, then some
    systems require the allocation of normal virtual memory equal to the size
    of the disc file at the time of mapping, while others will dynamically
    allocate memory from swap space as pages are written into. In the latter
    case, some systems will cause a segmentation fault if swap space is
    exhausted, while other systems wait for space to be freed.
    [NOTE] The channel may be queried to determine if it has been memory mapped
    using the call <<ch_test_for_mmap>>.
    [NOTE] If the "CH_MMAP_DISABLE" environment variable is defined, then
    memory mapping is disabled.
    [RETURNS] A channel object on success, else NULL.
*/
{
    int open_flags, errnum;
    flag memory_map = FALSE;
#ifdef OS_Linux
    struct statfs statfs_buf;
#endif
    struct stat statbuf;
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;
    char *env_name = "CH_MMAP_DISABLE";
    static flag first_time = TRUE;
    static flag disable_mmap = FALSE;
    static char function_name[] = "ch_map_disc";

    if (first_time)
    {
	first_time = FALSE;
	if (r_getenv (env_name) != NULL)
	{
	    disable_mmap = TRUE;
	    fprintf (stderr, "%s: \"%s\" environment variable found\n",
		     function_name, env_name);
	    fprintf (stderr, "disabling memory mapping\n");
	}
    }
    if (filename == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    m_clear (&hooks, sizeof hooks);
    hooks.close = _ch_memory_close_func;
    hooks.flush = _ch_memory_flush_func;
    hooks.read = _ch_memory_read_func;
    hooks.write = _ch_memory_write_func;
    hooks.seek = _ch_memory_seek_func;
    hooks.get_descriptor = _ch_memory_get_descriptor;
    if ( ( info = info_alloc () ) == NULL ) return (NULL);
    /*  Open disc file descriptor  */
    if (writable && update_on_write) open_flags = O_RDWR;
    else open_flags = O_RDONLY;
    if ( ( info->fd = open (filename, open_flags, 0) ) < 0 )
    {
	_ch_memory_close_func (info);
	return (NULL);
    }
    /*  Get stats on file  */
    if (fstat (info->fd, &statbuf) != 0)
    {
	fprintf (stderr, "%s: error getting file stats\t%s\n",
		 function_name, ERRSTRING);
	_ch_memory_close_func (info);
	return (NULL);
    }
#ifdef OS_Linux
    if (fstatfs (info->fd, &statfs_buf) != 0)
    {
	fprintf (stderr, "%s: error getting filesystem stats\t%s\n",
		 function_name, ERRSTRING);
	_ch_memory_close_func (info);
	return (NULL);
    }
#endif
    /*  Determine if file should be memory mapped  */
    switch (option)
    {
      case K_CH_MAP_NEVER:
	break;
      case K_CH_MAP_LARGE_LOCAL:
	/*  Check if large, local file  */
#ifdef OS_Linux
	if ( (statbuf.st_size >= MMAP_LARGE_SIZE) &&
	     (statfs_buf.f_type != NFS_SUPER_MAGIC) )
#else
	if ( (statbuf.st_size >= MMAP_LARGE_SIZE) && (statbuf.st_dev >= 0) )
#endif
	{
	    memory_map = TRUE;
	}
	break;
      case K_CH_MAP_LOCAL:
        /*  Check if local  */
#ifdef OS_Linux
        if (statfs_buf.f_type != NFS_SUPER_MAGIC)
#else
	if (statbuf.st_dev >= 0)
#endif
	{
	    memory_map = TRUE;
	}
	break;
      case K_CH_MAP_LARGE:
	/*  Check if large  */
	if (statbuf.st_size >= MMAP_LARGE_SIZE)
	{
	    memory_map = TRUE;
	}
	break;
      case K_CH_MAP_ALWAYS:
	if (disable_mmap)
	{
	    _ch_memory_close_func (info);
	    return (NULL);
	}
	/*  Fall through  */
      case K_CH_MAP_IF_AVAILABLE:
	memory_map = TRUE;
	break;
      default:
	fprintf (stderr, "Illegal value of: option : %u\n", option);
	a_prog_bug (function_name);
	break;
    }
    if (disable_mmap) memory_map = FALSE;
    if (!memory_map)
    {
	/*  Mapping not required  */
	_ch_memory_close_func (info);
	return ch_open_file (filename, "rz");
    }
    /*  Memory mapping desired  */
    if ( ( info->memory_buffer =
	   m_map_fd (&info->fd, statbuf.st_size, writable, update_on_write,
		     TRUE) ) == NULL )
    {
	/*  Failed  */
	errnum = errno;
	_ch_memory_close_func (info);
	errno = errnum;
	if (option == K_CH_MAP_ALWAYS) return (NULL);
	return ch_open_file (filename, "rz");
    }
    /*  Set channel type  */
    info->mem_buf_allocated = FALSE;
    info->mem_buf_len = statbuf.st_size;
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	_ch_memory_close_func (info);
	return (NULL);
    }
    return (channel);
}   /*  End Function ch_map_disc  */

/*EXPERIMENTAL_FUNCTION*/
Channel ch_map_fd (int fd, uaddr size, flag writable, flag update_on_write)
/*  [SUMMARY] Map a disc file.
    [PURPOSE] This routine will open a memory channel with the memory pages
    being mapped from a disc file. The disc file must already exist.
    <fd> The file descriptor to map. This will be closed on error or when the
    channel closes.
    <size> The size of the region to map.
    <writable> If the mapped pages are to be writable, this must be TRUE. If
    this is FALSE and the memory pages are written to, a segmentation fault
    occurs.
    <update_on_write> If the disc file should be updated when the memory pages
    are written to, this must be TRUE. If this is FALSE, then a write to a
    memory page causes the page to be copied into swap space and the process
    retains a private copy of the page from this point on.
    [NOTE] If <<update_on_write>> is FALSE and <<writable>> is TRUE, then some
    systems require the allocation of normal virtual memory equal to the size
    of the disc file at the time of mapping, while others will dynamically
    allocate memory from swap space as pages are written into. In the latter
    case, some systems will cause a segmentation fault if swap space is
    exhausted, while other systems wait for space to be freed.
    [NOTE] The channel may be queried to determine if it has been memory mapped
    using the call <<ch_test_for_mmap>>.
    [RETURNS] A channel object on success, else NULL.
*/
{
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_map_fd";

    m_clear (&hooks, sizeof hooks);
    hooks.close = _ch_memory_close_func;
    hooks.flush = _ch_memory_flush_func;
    hooks.read = _ch_memory_read_func;
    hooks.write = _ch_memory_write_func;
    hooks.seek = _ch_memory_seek_func;
    hooks.get_descriptor = _ch_memory_get_descriptor;
    /*  Can memory map  */
    if ( ( info = info_alloc () ) == NULL )
    {
	close (fd);
	return (NULL);
    }
    info->fd = fd;
    if ( ( info->memory_buffer =
	   m_map_fd (&info->fd, size, writable, update_on_write, TRUE) )
	 == NULL )
    {
	/*  Failed  */
	return (NULL);
    }
    /*  Set channel type  */
    info->mem_buf_allocated = FALSE;
    info->mem_buf_len = size;
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	_ch_memory_close_func (info);
	return (NULL);
    }
    return (channel);
}   /*  End Function ch_map_fd  */

/*PUBLIC_FUNCTION*/
Channel ch_open_memory (char *buffer, unsigned int size)
/*  [SUMMARY] Open a memory channel.
    [PURPOSE] This routine will open a memory channel. A memory channel behaves
    like a disc channel with a limited (specified) file (device) size. Data is
    undefined when reading before writing has occurred.
    <buffer> The buffer to use. If this is NULL, the routine will allocate a
    buffer of the specified size which is automatically deallocated upon
    closure of the channel.
    <size> The size of the buffer to allocate.
    [RETURNS] A channel object on success, else NULL.
*/
{
    Channel channel;
    KChannelHook info;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_open_memory";

    m_clear (&hooks, sizeof hooks);
    hooks.close = _ch_memory_close_func;
    hooks.flush = _ch_memory_flush_func;
    hooks.read = _ch_memory_read_func;
    hooks.write = _ch_memory_write_func;
    hooks.seek = _ch_memory_seek_func;
    if ( ( info = info_alloc () ) == NULL ) return (NULL);
    if (buffer == NULL)
    {
	/*  Must allocate a buffer  */
	if ( ( info->memory_buffer = m_alloc (size) ) == NULL )
	{
	    m_error_notify (function_name, "channel memory buffer");
	    _ch_memory_close_func (info);
	    return (NULL);
	}
	info->mem_buf_allocated = TRUE;
    }
    else
    {
	/*  Use specified buffer  */
	info->memory_buffer = buffer;
	info->mem_buf_allocated = FALSE;
    }
    info->mem_buf_len = size;
    if ( ( channel = ch_create_generic (info, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	_ch_memory_close_func (info);
	return (NULL);
    }
    return (channel);
}   /*  End Function ch_open_memory  */

/*PUBLIC_FUNCTION*/
flag ch_test_for_mmap (Channel channel)
/*  [SUMMARY] Test if a channel object is a memory mapped disc channel.
    <channel> The channel object.
    [RETURNS] TRUE if the channel object is memory mapped, else FALSE.
*/
{
    KChannelHook info;
    /*static char function_name[] = "ch_test_for_mmap";*/

    info = ch_get_hook_info (channel);
    if (info == NULL) return (FALSE);
    if (info->magic_number != INFO_MAGIC_NUMBER) return (FALSE);
    return (info->fd > -1) ? TRUE : FALSE;
}   /*  End Function ch_test_for_mmap  */

/*PUBLIC_FUNCTION*/
char *ch_get_mmap_addr (Channel channel)
/*  [SUMMARY] Get memory mapped address.
    [PURPOSE] This routine will get the starting address of the data for a
    memory mapped disc channel. The channel MUST be a memory mapped disc
    channel.
    <channel> The channel object.
    [RETURNS] The address of the memory mapped data.
    [NOTE] If the memory mapped address space is read-only, any attempt to
    write to this address space will cause a segmentation fault.
*/
{
    KChannelHook info;
    static char function_name[] = "ch_get_mmap_addr";

    info = ch_get_hook_info (channel);
    VERIFY_INFO (info);
    if (info->fd < 0)
    {
	fprintf (stderr, "Channel is not a memory mapped disc file\n");
	a_prog_bug (function_name);
    }
    ++info->mmap_access_count;
    return (info->memory_buffer);
}   /*  End Function ch_get_mmap_addr  */

/*PUBLIC_FUNCTION*/
unsigned int ch_get_mmap_access_count (Channel channel)
/*  [SUMMARY] Get memory mapped access count.
    [PURPOSE] This routine will get the number of times a memory mapped disc
    channel has been queried for the mapping address using <<ch_get_mmap_addr>>
    <channel> The channel object.
    [NOTE] The channel MUST be a memory mapped disc channel.
    [RETURNS] The number of address queries.
*/
{
    KChannelHook info;
    static char function_name[] = "ch_get_mmap_access_count";

    info = ch_get_hook_info (channel);
    VERIFY_INFO (info);
    if (info->fd < 0)
    {
	fprintf (stderr, "Channel is not a memory mapped disc file\n");
	a_prog_bug (function_name);
    }
    return (info->mmap_access_count);
}   /*  End Function ch_get_mmap_access_count  */

/*EXPERIMENTAL_FUNCTION*/
uaddr ch_get_mem_size (Channel channel)
/*  [SUMMARY] Get memory buffer size.
    <channel> The channel object.
    [NOTE] The channel MUST be a memory channel.
    [RETURNS] The number of bytes in the memory channel.
*/
{
    KChannelHook info;
    static char function_name[] = "ch_get_mem_size";

    info = ch_get_hook_info (channel);
    VERIFY_INFO (info);
    return (info->mem_buf_len);
}   /*  End Function ch_get_mem_size  */


/*  Private functions follow  */

static KChannelHook info_alloc ()
/*  This routine will allocate and initialise a channel data structure.
    The routine returns a channel object structure on success, 
    else it returns NULL.
*/
{
    KChannelHook info;
    static char function_name[] = "info_alloc";

    /*  Allocate channel structure  */
    if ( ( info = m_alloc (sizeof *info) ) == NULL )
    {
	m_error_notify (function_name, "channel object info");
	return (NULL);
    }
    info->fd = -1;
    info->memory_buffer = NULL;
    info->mmap_access_count = 0;
    info->mem_buf_len = 0;
    info->mem_buf_read_pos = 0;
    info->mem_buf_write_pos = 0;
    info->mem_buf_allocated = FALSE;
    info->magic_number = INFO_MAGIC_NUMBER;
    return (info);
}   /*  End Function info_alloc  */


/*  Hook functions follow  */

static flag _ch_memory_close_func (KChannelHook info)
/*  [SUMMARY] Close a channel. The channel must already be flushed.
    <info> The channel object subclass information pointer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag return_value = TRUE;
    static char function_name[] = "_ch_memory_close_func";

    if (info->fd > -1)
    {
	/*  Have open file descriptor: flush buffer and close descriptor  */
	if (info->memory_buffer != NULL)
	{
	    if ( !m_unmap (info->memory_buffer, info->mem_buf_len) )
	    {
		fprintf (stderr, "%s: error unmapping\t%s\n",
			 function_name, ERRSTRING);
		abort ();
	    }
	    info->memory_buffer = NULL;
	}
	if (close (info->fd) != 0)
	{
	    fprintf (stderr, "Error closing descriptor: %d\t%s\n",
		     info->fd, ERRSTRING);
	    return_value = FALSE;
	}
    }
    /*  Deallocate buffers  */
    if ( (info->memory_buffer != NULL) && info->mem_buf_allocated )
    {
	m_free (info->memory_buffer);
    }
    m_clear (info, sizeof *info);
    m_free (info);
    return (return_value);
}   /*  End Function _ch_memory_close_func  */

static flag _ch_memory_flush_func (KChannelHook info)
/*  [SUMMARY] Flush the write buffer of a channel object.
    <info> The channel object subclass information pointer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    /*static char function_name[] = "_ch_memory_flush_func";*/

    return (TRUE);
}   /*  End Function _ch_memory_flush_func  */

static unsigned int _ch_memory_read_func (KChannelHook info, char *buffer,
					  CONST char **newbuf,
					  unsigned char alignment,
					  unsigned int length)
/*  [SUMMARY] Read from a channel.
    [PURPOSE] This routine will read a number of bytes from a channel and
    places them into a buffer.
    <channel> The channel object.
    <buffer> A suggested buffer to write the data into.
    <newbuf> The buffer pointer used is written here. This may be a read-only
    buffer. If this is NULL, the suggested <<buffer>> is always used.
    <alignment> The number of bytes the buffer must be aligned to.
    <length> The number of bytes to write into the buffer.
    [RETURNS] The number of bytes read.
*/
{
    unsigned int bytes_to_read;
    unsigned int bytes_left;
    CONST char *source;
    /*static char function_name[] = "_ch_memory_read_func";*/

    bytes_left = info->mem_buf_len - info->mem_buf_read_pos;
    bytes_to_read = (length <= bytes_left) ? length : bytes_left;
    source = info->memory_buffer + info->mem_buf_read_pos;
    if (newbuf != NULL)
    {
	switch (alignment)
	{
	  case 0:
	  case 1:
	    break;
	  case 2:
	  case 4:
	  case 8:
	    if ( (uaddr) source & (alignment - 1) ) newbuf = NULL;
	    break;
	  default:
	    if ( (uaddr) source % alignment != 0 ) newbuf = NULL;
	    break;
	}
	if (newbuf != NULL)
	{
	    *newbuf = source;
	    info->mem_buf_read_pos += bytes_to_read;
	    return (bytes_to_read);
	}
    }
    if (bytes_to_read < 16)
    {
	unsigned int count;

	/*  Avoid function call overhead  */
	for (count = 0; count < bytes_to_read; ++count) *buffer++ = *source++;
    }
    else m_copy (buffer, source, bytes_to_read);
    info->mem_buf_read_pos += bytes_to_read;
    return (bytes_to_read);
}   /*  End Function _ch_memory_read_func  */

static unsigned int _ch_memory_write_func (KChannelHook info,
					   CONST char *buffer,
					   unsigned int length)
/*  This routine will write a number of bytes from a buffer to a channel. This
    routine takes no account of buffering, nor a  write_converter_func  .
    The channel object must be given by  channel  .
    The buffer to read the data from must be pointed to by  buffer  .
    The number of bytes to write from the buffer must be pointed to by  length
    The routine returns the number of bytes written.
*/
{
    int bytes_to_write;
    unsigned int bytes_left;
    static char function_name[] = "_ch_memory_write_func";

    if (buffer == NULL)
    {
	fprintf (stderr, "NULL buffer pointer passed\n");
	a_prog_bug (function_name);
    }
    if (length < 1) return (0);
    bytes_left = info->mem_buf_len - info->mem_buf_write_pos;
    bytes_to_write = (length <= bytes_left) ? length : bytes_left;
    m_copy (info->memory_buffer + info->mem_buf_write_pos,
	    buffer, bytes_to_write);
    info->mem_buf_write_pos += bytes_to_write;
    return (bytes_to_write);
}   /*  End Function _ch_memory_write_func  */

static flag _ch_memory_seek_func (KChannelHook info, unsigned long position)
/*  [SUMMARY] Move read/write pointer.
    [PURPOSE] This routine will position the read and write pointers for a
    channel.
    <info> The channel object subclass information pointer.
    <position> The position (relative to the start of the channel data).
    [NOTE] This routine cannot be used for connection channels.
    [RETURNS] TRUE on success, else FALSE (indicating a seek request beyond the
    channel limits)
*/
{
    /*static char function_name[] = "_ch_memory_seek_func";*/

    if (position > info->mem_buf_len)
    {
	/*  Illegal seek position  */
	return (FALSE);
    }
    info->mem_buf_read_pos = position;
    info->mem_buf_write_pos = position;
    return (TRUE);
}   /*  End Function _ch_memory_seek_func  */

static int _ch_memory_get_descriptor (KChannelHook info)
/*  [SUMMARY] Get the file descriptor associated with a channel.
    <channel> The channel object.
    [RETURNS] The file descriptor on success, else -1 indicating error.
*/
{
    /*static char function_name[] = "_ch_memory_get_descriptor";*/

    return (info->fd);
}   /*  End Function _ch_memory_get_descriptor  */
