/*LINTLIBRARY*/
/*  uncompress.c

    This code provides Channel objects.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the creation, manipulation and
  deletion of channel objects.


    Written by      Richard Gooch   20-AUG-1997: Moved bits from generic.c

    Updated by      Richard Gooch   22-FEB-1998: Removed casts for memory
  functions.

    Updated by      Richard Gooch   23-APR-1998: Changed to generic
  uncompression to support bzip2.

    Updated by      Richard Gooch   21-JUN-1998: Switched from CAN_ FORK to
  HAS_SIGNALS.

    Last updated by Richard Gooch   28-JUN-2000: Removed unnecessary #include.


*/
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_m.h>
#include <karma_r.h>
#include <karma_a.h>
#include <os.h>
#ifdef HAS_SIGNALS
#  include <signal.h>
#  include <sys/wait.h>
#endif


#define UNCOMPRESS_MAGIC_NUMBER 1498723534

#define VERIFY_UNCOMPRESS_INFO(st) if (st == NULL) \
{fprintf (stderr, "NULL info passed\n"); \
 a_prog_bug (function_name); } \
if (st->magic_number != UNCOMPRESS_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid info object\n"); \
 a_prog_bug (function_name); }


struct uncompress_info_type
{
    unsigned int magic_number;
    int fd;
    pid_t child;
};

/*  Private functions  */
STATIC_FUNCTION (unsigned int _ch_uncompress_size_func,
		 (Channel channel, void **info) );
STATIC_FUNCTION (unsigned int _ch_uncompress_read_func,
		 (Channel channel, char *buffer,
		  unsigned int length, void **info) );
STATIC_FUNCTION (unsigned int _ch_uncompress_write_func,
		 (Channel channel, char *buffer,
		  unsigned int length, void **info) );
STATIC_FUNCTION (flag _ch_uncompress_flush_func,
		 (Channel channel, void **info) );
STATIC_FUNCTION (void _ch_uncompress_close_func, (void *info) );
STATIC_FUNCTION (void _ch_uncompress_cleanup,
		 (struct uncompress_info_type *info) );


/*  Public functions follow  */

/*UNPUBLISHED_FUNCTION*/
ChConverter ch_add_uncompress (Channel channel, CONST char *programme)
/*  [SUMMARY] Add decompression to a channel object.
    [PURPOSE] This routine will register converter functions for a channel so
    that all read operations will be decompressed.
    <channel> The channel object to add decompression to.
    <programme> The decompression programme. This will be passed the "-d"
    option and it's standard input will be the compressed data source.
    [RETURNS] A ChConverter object on success (which may be used in a call to
    [<ch_unregister_converter>]), else NULL.
*/
{
    ChConverter converter;
    int read_fd, write_fd;
    pid_t child;
    struct uncompress_info_type *info;
    char *argv[3];
    static char function_name[] = "ch_add_uncompress";

    if ( ( info = m_alloc (sizeof *info) ) == NULL )
    {
	m_error_notify (function_name, "channel converter info");
	return (NULL);
    }
    info->fd = -1;
    info->child = -1;
    if (r_create_pipe (&read_fd, &write_fd) != 0)
    {
	fprintf (stderr, "%s: error creating pipe\t%s\n",
		 function_name, ERRSTRING);
	m_free (info);
	return (NULL);
    }
    argv[0] = (char *) programme;
    argv[1] = "-d";
    argv[2] = NULL;
    if ( ( child = r_fork_and_execvp (ch_get_descriptor (channel), write_fd, 2,
				      1, &read_fd, argv[0], argv) ) == -1 )
    {
	fprintf (stderr, "%s: error forking\t%s\n", function_name, ERRSTRING);
	close (read_fd);
	close (write_fd);
	m_free (info);
	return (NULL);
    }
    info->magic_number = UNCOMPRESS_MAGIC_NUMBER;
    info->child = child;
    info->fd = read_fd;
    close (write_fd);
    if ( ( converter =
	   ch_register_converter (channel, _ch_uncompress_size_func,
				  _ch_uncompress_read_func,
				  _ch_uncompress_write_func,
				  _ch_uncompress_flush_func,
				  _ch_uncompress_close_func,
				  info) ) == NULL )
    {
	_ch_uncompress_cleanup (info);
	return (NULL);
    }
    return (converter);
}   /*  End Function ch_add_uncompress  */


/*  Private functions follow  */

/*  Functions to support decompression follow  */

static unsigned int _ch_uncompress_size_func (Channel channel, void **info)
/*  This routine will determine the approximate number of bytes that the
    read_func  will return.
    The channel object will be given by  channel  .
    The arbitrary information pointer will be pointed to by  info  .The
    information pointer may be modified.
    The routine returns the number of bytes the  read_func  will return.
*/
{
    int bytes;
    struct uncompress_info_type *uncompress_info;
    static char function_name[] = "_ch_uncompress_size_func";

    uncompress_info = (struct uncompress_info_type *) *info;
    VERIFY_UNCOMPRESS_INFO (uncompress_info);
    if ( ( bytes = r_get_bytes_readable (uncompress_info->fd) ) < 0 )
    {
	fprintf (stderr, "%s: error getting bytes readable\t%s\n",
		 function_name, ERRSTRING);
	return (0);
    }
    return (bytes);
}   /*  End Function _ch_uncompress_size_func  */

static unsigned int _ch_uncompress_read_func (Channel channel, char *buffer,
					      unsigned int length, void **info)
/*   This routine will convert bytes being read from a channel object.
     The channel object will be given by  channel  .It is permissable for
     the routine to call  ch_read  with this channel. If this is done,
     this  read_func  will be bypassed.
     The buffer to write the data into will be pointed to by  buffer  .
     The number of bytes to write into the buffer will be pointed to by
     length  .
     The arbitrary information pointer will be pointed to by  info  .The
     information pointer may be modified.
     The routine returns the number of bytes actually written to the buffer.
*/
{
    struct uncompress_info_type *uncompress_info;
    static char function_name[] = "_ch_uncompress_read_func";

    uncompress_info = (struct uncompress_info_type *) *info;
    VERIFY_UNCOMPRESS_INFO (uncompress_info);
    if (r_read (uncompress_info->fd, buffer, length) < length) return (0);
    return (length);
}   /*  End Function _ch_uncompress_read_func  */

static unsigned int _ch_uncompress_write_func (Channel channel, char *buffer,
					       unsigned int length,void **info)
/*  This routine will convert bytes being written to a channel object.
    The channel object will be given by  channel  .It is permissable for
    the routine to call  ch_write  with this channel. If this is done,
    this  write_func  will be bypassed.
    The buffer to read the data from will be pointed to by  buffer  .The
    contents of this buffer may be modified if needed.
    The number of bytes to read from the buffer will be given by  length
    The arbitrary information pointer will be pointed to by  info  .The
    information pointer may be modified.
    The routine returns the number of bytes read from the buffer.
*/
{
    struct uncompress_info_type *uncompress_info;
    static char function_name[] = "_ch_uncompress_write_func";

    uncompress_info = (struct uncompress_info_type *) *info;
    VERIFY_UNCOMPRESS_INFO (uncompress_info);
    fprintf (stderr, "Illegal write to uncompressing channel!\n");
    a_prog_bug (function_name);
    return (0);
}   /*  End Function _ch_uncompress_write_func  */

static flag _ch_uncompress_flush_func (Channel channel, void **info)
/*  This routine will process a flush request for a channel object.
    The channel object will be given by  channel  .It is permissable for
    the routine to call  ch_write  with this channel. If this is done,
    this  write_func  will be bypassed.
    The arbitrary information pointer will be pointed to by  info  .The
    information pointer may be modified.
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    return (TRUE);
}   /*  End Function _ch_uncompress_flush_func  */

static void _ch_uncompress_close_func (void *info)
/*  This routine is called when a channel is closed.
    The arbitrary information pointer will be given by  info  .
    The routine returns nothing.
*/
{
    struct uncompress_info_type *uncompress_info;
    static char function_name[] = "_ch_uncompress_close_func";

    uncompress_info = (struct uncompress_info_type *) info;
    VERIFY_UNCOMPRESS_INFO (uncompress_info);
    _ch_uncompress_cleanup (uncompress_info);
}   /*  End Function _ch_uncompress_close_func  */

static void _ch_uncompress_cleanup (struct uncompress_info_type *info)
/*  [SUMMARY] Clean up.
    <info> The info.
    [RETURNS] Nothing.
*/
{
    int status;

    if (info->fd > -1) close (info->fd);
#ifdef HAS_SIGNALS
    if (info->child > -1)
    {
	kill (info->child, SIGKILL);
	waitpid (info->child, &status, 0);
    }
#endif
    m_free (info);
}   /*  End Function _ch_uncompress_cleanup  */
