/*LINTLIBRARY*/
/*  virtual.c

    This code provides Channel objects.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the creation, manipulation and
  deletion of channel objects.


    Written by      Richard Gooch   15-AUG-1997: Copied bits from main.c

    Updated by      Richard Gooch   22-FEB-1998: Removed casts for memory
  functions.

    Updated by      Richard Gooch   15-DEC-1998: Documentation fix.

    Last updated by Richard Gooch   26-JAN-2000: Documentation fix.


*/
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_ch_hooks.h>
#include <karma_m.h>

struct channel_hook_type
{
    int dummy;
};


/*  Private functions  */
STATIC_FUNCTION (flag close_func, (KChannelHook info) );
STATIC_FUNCTION (flag flush_func, (KChannelHook info) );
STATIC_FUNCTION (unsigned int read_func,
		 (KChannelHook info, char *buffer, CONST char **newbuf,
		  unsigned char alignment, unsigned int length) );
STATIC_FUNCTION (unsigned int write_func,
		 (KChannelHook info, CONST char *buffer,unsigned int length) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
Channel ch_create_sink ()
/*  [SUMMARY] Create data sink.
    [PURPOSE] This routine will create a data sink channel. All writes to the
    channel are discarded (and reported as having succeeded) and all reads
    return an End-Of-File condition. Read and write operations modify the
    absolute read and write pointers (obtainable with [<ch_tell>]) as expected.
    [RETURNS] The channel object on succes, else NULL.
*/
{
    Channel channel;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_create_sink";

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.flush = flush_func;
    hooks.write = write_func;
    if ( ( channel = ch_create_generic (NULL, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	return (NULL);
    }
    return (channel);
}   /*  End Function ch_create_sink  */

/*EXPERIMENTAL_FUNCTION*/
Channel ch_create_zero ()
/*  [SUMMARY] Create data source.
    [PURPOSE] This routine will create a data source channel. All reads from
    the channel will yield values of zero. Write operations fail.
    [RETURNS] The channel object on succes, else NULL.
*/
{
    Channel channel;
    KChannelHookFuncs hooks;
    static char function_name[] = "ch_create_zero";

    m_clear (&hooks, sizeof hooks);
    hooks.close = close_func;
    hooks.read = read_func;
    if ( ( channel = ch_create_generic (NULL, hooks) ) == NULL )
    {
	m_error_notify (function_name, "channel object");
	return (NULL);
    }
    return (channel);
}   /*  End Function ch_create_zero  */


/*  Private functions follow  */


/*  Hook functions follow  */

static flag close_func (KChannelHook info)
/*  [SUMMARY] Close a channel. The channel must already be flushed.
    <info> The channel object subclass information pointer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    /*static char function_name[] = "_ch_virtual_close_func";*/

    return (TRUE);
}   /*  End Function close_func  */

static flag flush_func (KChannelHook info)
/*  [SUMMARY] Flush the write buffer of a channel object.
    <info> The channel object subclass information pointer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    /*static char function_name[] = "_ch_virtual_flush_func";*/

    return (TRUE);
}   /*  End Function flush_func  */

static unsigned int read_func (KChannelHook info, char *buffer,
			       CONST char **newbuf,
			       unsigned char alignment, unsigned int length)
/*  [SUMMARY] Read from a channel.
    [PURPOSE] This routine will read a number of bytes from a channel and
    places them into a buffer.
    <channel> The channel object.
    <buffer> A suggested buffer to write the data into.
    <newbuf> The buffer pointer used is written here. This may be a read-only
    buffer. If this is NULL, the suggested <<buffer>> is always used.
    <alignment> The number of bytes the buffer must be aligned to.
    <length> The number of bytes to write into the buffer.
    [RETURNS] The number of bytes read.
*/
{
    /*static char function_name[] = "_ch_virtual_read_func";*/

    m_clear (buffer, length);
    return (length);
}   /*  End Function read_func  */

static unsigned int write_func (KChannelHook info, CONST char *buffer,
				unsigned int length)
/*  [SUMMARY] Write to a channel.
    [PURPOSE] This routine will write a number of bytes from a buffer to a
    channel.
    <info> The channel object subclass information pointer.
    <buffer> The buffer to read the data from.
    <length> The number of bytes to read from the buffer and write to the
    channel.
    [RETURNS] The number of bytes written.
*/
{
    /*static char function_name[] = "_ch_virtual_write_func";*/

    return (length);
}   /*  End Function write_func  */
