/*LINTLIBRARY*/
/*  main.c

    This code provides Channel Management routines.

    Copyright (C) 1992-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the management of channel
  objects.


    Written by      Richard Gooch   12-SEP-1992

    Updated by      Richard Gooch   14-NOV-1992

    Updated by      Richard Gooch   15-DEC-1992: Added complete read processing

    Updated by      Richard Gooch   19-DEC-1992: Tested for  chm_poll re-entry.

    Updated by      Richard Gooch   28-DEC-1992: Added closure detection for
  asynchronous channels.

    Updated by      Richard Gooch   31-DEC-1992: Added call to  close_func  if
  event functions return FALSE.

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   24-JAN-1993: Disabled code which sets
  descriptor to send SIGIO, since it was not required.

    Updated by      Richard Gooch   24-JUL-1993: Made use of new
  ch_test_for_io  routine.

    Updated by      Richard Gooch   3-OCT-1993: Improved diagnostics in
  read_channel  .

    Updated by      Richard Gooch   2-NOV-1994: Changed from  rs6000  to
  rs6000_AIX

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/chs/main.c

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno

    Updated by      Richard Gooch   2-JAN-1995: Fixed some comments.

    Updated by      Richard Gooch   5-MAY-1995: Placate SGI compiler.

    Updated by      Richard Gooch   1-APR-1996: Moved remaing functions to new
  documentation style.

    Updated by      Richard Gooch   2-DEC-1996: Added scheduling of work
  functions.

    Updated by      Richard Gooch   18-DEC-1996: Switched to <dm> package.

    Updated by      Richard Gooch   24-AUG-1997: Minor cosmetic changes.

    Updated by      Richard Gooch   17-FEB-1998: Fixed documentation error in
  <chm_unmanage>.

    Updated by      Richard Gooch   19-FEB-1998: Created freelist.

    Updated by      Richard Gooch   22-FEB-1998: Removed casts for memory
  functions.

    Updated by      Richard Gooch   28-JUN-1999: Took account of change to
  <dm_manage>

    Updated by      Richard Gooch   9-FEB-2000: Fixed bug in <input_func>:
  wasn't always testing for NULL function pointer.

    Updated by      Richard Gooch   14-FEB-2000: Pass NULL input function to
  <dm_manage> if NULL input function passed to <chm_manage>. Fixed bug in
  allocating last entry of freelist.

    Last updated by Richard Gooch   15-FEB-2000: Created <call_func>.


*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <karma.h>
#include <karma_chm.h>
#include <karma_ch.h>
#include <karma_dm.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_r.h>


#define ALLOC_SIZE 64
#define USED_MAGIC_NUMBER 1987534832
#define FREE_MAGIC_NUMBER 1249873549

#define VERIFY_USED(entry) if (entry == NULL) \
{fprintf (stderr, "NULL entry passed\n"); \
 a_prog_bug (function_name); } \
if ((uaddr) entry % sizeof (void *) != 0) \
{fprintf (stderr, "Entry pointer not aligned properly\n"); \
 a_prog_bug (function_name); } \
if (entry->magic_number != USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid entry object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FREE(entry) if (entry == NULL) \
{fprintf (stderr, "NULL entry passed\n"); \
 a_prog_bug (function_name); } \
if ((uaddr) entry % sizeof (void *) != 0) \
{fprintf (stderr, "Entry pointer not aligned properly\n"); \
 a_prog_bug (function_name); } \
if (entry->magic_number != FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed entry object\n"); \
 a_prog_bug (function_name); }


typedef flag (*func_t) (Channel channel, void **info);

/*  Structure and private data to manage channels  */
struct managed_channel_type
{
    unsigned int magic_number;
    Channel channel;
    int fd;
    void *info;
    func_t input_func;
    void (*close_func) (Channel channel, void *info);
    func_t output_func;
    func_t exception_func;
    KManagedChannel next;
    KManagedDescriptor md;
};


/*  Private data  */
static KManagedChannel first_free = NULL;


/*  Private functions  */
STATIC_FUNCTION (flag _chm_input_func, (int fd, void **info) );
STATIC_FUNCTION (void _chm_close_func, (int fd, void *info) );
STATIC_FUNCTION (flag _chm_output_func, (int fd, void **info) );
STATIC_FUNCTION (flag _chm_exception_func, (int fd, void **info) );
STATIC_FUNCTION (void _chm_free_entry, (KManagedChannel entry) );
STATIC_FUNCTION (flag _chm_call_func, (KManagedChannel entry, func_t func) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
KManagedChannel chm_manage ( Channel channel, void *info,
			     flag (*input_func) (), void (*close_func) (),
			     flag (*output_func) (),flag (*exception_func) () )
/*  [SUMMARY] Manage a channel for activity by registering callback routines.
    <channel> The channel object to manage.
    <info> The arbitrary information pointer associated with the managed
    channel. This pointer may be modified by the callback routines.
    <input_func> This routine is called when new input occurs on the channel.
    If this is NULL, no callback routine is installed. The prototype function
    is [<chm_PROTO_input_func>].
    <close_func> This routine is called when the channel closes. If this is
    NULL, no callback routine is installed. The prototype function is
    [<chm_PROTO_close_func>]. The channel object MUST be capable of detecting
    closure if this routine is supplied (i.e. this routine cannot be supplied
    for dock channels). Any unread buffered data in the channel will be lost
    upon closure. The call to this function is the last chance to read this
    buffered data.
    <output_func> This routine is called when the channel becomes ready for
    output. If this is NULL, no callback routine is installed. The prototype
    function is [<chm_PROTO_output_func>].
    <exception_func> This routine is called when exceptions occur on the
    channel. If this is NULL, no callback routine is installed. The prototype
    function is [<chm_PROTO_exception_func>].
    [RETURNS] A KManagedChannel object on success, else NULL.
*/
{
    int fd;
    unsigned int count;
    KManagedChannel new;
    static char function_name[] = "chm_manage";

    /*  Check if channel is a connection  */
    if ( !ch_test_for_asynchronous (channel) )
    {
	fprintf (stderr,"Cannot manage a channel if it is not asynchronous\n");
	a_prog_bug (function_name);
    }
    /*  Get file descriptor  */
    if ( ( fd = ch_get_descriptor (channel) ) < 0 )
    {
	fprintf (stderr, "Error getting file descriptor for channel object\n");
	return (FALSE);
    }
    if (close_func != NULL)
    {
	/*  Check if closure detectable  */
	if (r_get_bytes_readable (fd) < 0)
	{
	    fprintf (stderr,
		     "close_func  supplied and closure not detectable\n");
	    a_prog_bug (function_name);
	}
    }
    /*  Get new entry  */
    if (first_free == NULL)
    {
	/*  Have to allocate some more for the free list  */
	if ( ( new = m_alloc (ALLOC_SIZE * sizeof *new) ) == NULL )
	{
	    m_error_notify (function_name, "new managed channel entries");
	    return (NULL);
	}
	m_clear (new, ALLOC_SIZE * sizeof *new);
	/*  Add them to the free list  */
	first_free = new;
	for (count = 0; count < ALLOC_SIZE - 1; ++count)
	{
	    new[count].magic_number = FREE_MAGIC_NUMBER;
	    new[count].next = new + count + 1;
	}
	new[count].magic_number = FREE_MAGIC_NUMBER;
    }
    /*  Take one from the free list  */
    new = first_free;
    VERIFY_FREE (new);
    first_free = new->next;
    /*  Fill in entry  */
    new->magic_number = USED_MAGIC_NUMBER;
    new->next = NULL;
    new->channel = channel;
    new->fd = fd;
    new->info = info;
    new->input_func = input_func;
    new->close_func = close_func;
    new->output_func = output_func;
    new->exception_func = exception_func;
    new->md = dm_manage (fd, new, FALSE,
			 KARMA_FD_EVENT_INPUT,
			 (input_func == NULL) ? NULL : _chm_input_func,
			 KARMA_FD_EVENT_CLOSE,
			 (close_func == NULL) ? NULL : _chm_close_func,
			 KARMA_FD_EVENT_OUTPUT,
			 (output_func == NULL) ? NULL : _chm_output_func,
			 KARMA_FD_EVENT_EXCEPTION,
			 (exception_func == NULL) ? NULL : _chm_exception_func,
			 KARMA_FD_EVENT_END);
    if (new->md == NULL)
    {
	_chm_free_entry (new);
	return (NULL);
    }
    return (new);
}   /*  End Function chm_manage  */

/*PUBLIC_FUNCTION*/
void chm_unmanage (Channel channel)
/*  [SUMMARY] Terminate the management of a channel for activity.
    <channel> The channel object to unmanage. The channel must be open.
    [NOTE] This routine will NOT close the channel.
    [RETURNS] Nothing.
*/
{
    KManagedChannel mch;
    static char function_name[] = "chm_unmanage";

    mch = (KManagedChannel) dm_unmanage (ch_get_descriptor (channel), NULL);
    VERIFY_USED (mch);
    _chm_free_entry (mch);
}   /*  End Function chm_unmanage  */

/*EXPERIMENTAL_FUNCTION*/
void chm_unmanage_by_handle (KManagedChannel mch)
/*  [SUMMARY] Terminate the management of a channel for activity.
    <mch> The KManagedChannel object.
    [NOTE] This routine will NOT close the channel.
    [RETURNS] Nothing.
*/
{
    void *info;
    static char function_name[] = "chm_unmanage_by_handle";

    VERIFY_USED (mch);
    info = dm_unmanage (mch->fd, mch->md);
    if (info != (void *) mch)
    {
	fprintf (stderr, "Pointer: %p does not match pointer in entry: %p\n",
		 mch, info);
	a_prog_bug (function_name);
    }
    _chm_free_entry (mch);
}   /*  End Function chm_unmanage_by_handle  */

/*OBSOLETE_FUNCTION*/
void chm_poll (long timeout_ms)
/*  [SUMMARY] Poll all managed channels for any activity.
    <timeout_ms> The time (in milliseconds) to poll. If this is less than 0 the
    routine will poll forever (until some activity occurs or a signal is
    caught).
    [RETURNS] Nothing.
*/
{
    dm_native_poll (timeout_ms);
}   /*  End Function chm_poll  */


/*  Private functions follow  */

static flag _chm_input_func (int fd, void **info)
/*  [SUMMARY] Input event callback.
    [PURPOSE] This routine is called when new input occurs on a descriptor.
    <fd> The file descriptor.
    <info> A pointer to the arbitrary information pointer. The pointer
    written here will persist until the descriptor is unmanaged (or a
    subsequent callback routine changes it).
    [RETURNS] TRUE if the descriptor is to remain managed and open, else it
    returns FALSE (indicating that the descriptor is to be unmanaged and
    closed). This routine MUST NOT unmanage or close the descriptor.
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
    int bytes_readable = 1;
    KManagedChannel entry = (KManagedChannel) *info;
    static char function_name[] = "_chm_input_func";

    VERIFY_USED (entry);
    if (entry->input_func == NULL)
    {
	fprintf (stderr, "Input on channel but no callback\n");
	a_prog_bug (function_name);
    }
    if ( !ch_test_for_io (entry->channel) )
	return _chm_call_func (entry, entry->input_func);
    while (bytes_readable > 0)
    {
	/*  More input  */
	if ( !_chm_call_func (entry, entry->input_func) ) return (FALSE);
	/*  Test channel, not descriptor, to ensure channel read buffer as
	    well as descriptor drained  */
	if ( ( bytes_readable = ch_get_bytes_readable (entry->channel) ) < 0 )
	    a_prog_bug (function_name);
    }
    return (TRUE);
}   /*  End Function _chm_input_func  */

static void _chm_close_func (int fd, void *info)
/*  [SUMMARY] Close event callback.
    [PURPOSE] This routine is called when a descriptor closes.
    <fd> The file descriptor. The descriptor MUST be capable of detecting
    closure if this routine is supplied (ie. this routine cannot be
    supplied for docks).
    <info> The arbitrary pointer for the managed descriptor.
    [NOTE] This routine MUST NOT unmanage the descriptor. The descriptor
    will be automatically unmanaged and deleted upon closure (even if no
    <<close_func>> is specified).
    [NOTE] Any unread data in the descriptor will be lost upon closure. The
    call to this function is the last chance to read this data.
    [RETURNS] Nothing.
*/
{
    KManagedChannel entry = (KManagedChannel) info;
    static char function_name[] = "_chm_close_func";

    VERIFY_USED (entry);
    (*entry->close_func) (entry->channel, entry->info);
    ch_close (entry->channel);
    _chm_free_entry (entry);
}   /*  End Function _chm_close_func  */

static flag _chm_output_func (int fd, void **info)
/*  [SUMMARY] Output event callback.
    [PURPOSE] This routine is called when a descriptor becomes ready for
    writing.
    <fd> The file descriptor.
    <info> A pointer to the arbitrary information pointer. The pointer
    written here will persist until the descriptor is unmanaged (or a
    subsequent callback routine changes it).
    [RETURNS] TRUE if the descriptor is to remain managed and open, else it
    returns FALSE (indicating that the descriptor is to be unmanaged and
    closed). This routine MUST NOT unmanage or close the descriptor.
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
    KManagedChannel entry = (KManagedChannel) *info;
    static char function_name[] = "_chm_output_func";

    VERIFY_USED (entry);
    return _chm_call_func (entry, entry->output_func);
}   /*  End Function _chm_output_func  */

static flag _chm_exception_func (int fd, void **info)
/*  [SUMMARY] Exception event callback.
    [PURPOSE] This routine is called when an exception occurrs on a descriptor.
    <fd> The file descriptor.
    <info> A pointer to the arbitrary information pointer. The pointer
    written here will persist until the descriptor is unmanaged (or a
    subsequent callback routine changes it).
    [RETURNS] TRUE if the descriptor is to remain managed and open, else it
    returns FALSE (indicating that the descriptor is to be unmanaged and
    closed). This routine MUST NOT unmanage or close the descriptor.
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
    KManagedChannel entry = (KManagedChannel) *info;
    static char function_name[] = "_chm_exception_func";

    VERIFY_USED (entry);
    return _chm_call_func (entry, entry->exception_func);
}   /*  End Function _chm_exception_func  */

static void _chm_free_entry (KManagedChannel entry)
/*  [SUMMARY] Add a KManagedChannel object to the free list.
    <entry> The KManagedChannel object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_chm_free_entry";

    VERIFY_USED (entry);
    m_clear (entry, sizeof *entry);
    entry->magic_number = FREE_MAGIC_NUMBER;
    entry->next = first_free;
    first_free = entry;
}   /*  End Function _chm_free_entry  */

static flag _chm_call_func (KManagedChannel entry, func_t func)
/*  [SUMMARY] Call a callback, with optional cleanup on failure.
    <entry> The KManagedChannel object.
    <func> The callback function.
    [RETURNS] The return value from the callback.
*/
{
    if ( (*func) (entry->channel, &entry->info) ) return (TRUE);
    if (entry->close_func == NULL)
    {
	ch_close (entry->channel);
	_chm_free_entry (entry);
    }
    return (FALSE);
}   /*  End Function _chm_call_func  */
