/*LINTLIBRARY*/
/*  main.c

    This code provides enhanced directory scanning routines.

    Copyright (C) 1992-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the enhanced, portable scanning
  of directories..


    Written by      Richard Gooch   1-JUN-1993

    Updated by      Richard Gooch   2-JUN-1993

    Updated by      Richard Gooch   16-AUG-1994: Minor comment changes.

    Updated by      Richard Gooch   26-AUG-1994: Moved  typedef  of KDir
  class to header.

    Updated by      Richard Gooch   7-SEP-1994: Made allowance for SunOS 4.1.3
  int  array  rather than  gid_t  array  as argument 2 to  getgroups(2V)  .

    Updated by      Richard Gooch   4-OCT-1994: Added CONST declaration to
  dir_open  .

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/dir/main.c

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno

    Updated by      Richard Gooch   6-MAY-1995: Placate gcc -Wall

    Updated by      Richard Gooch   7-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   3-JUN-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   24-JUN-1996: Added <dirname> field to
  KFileInfo structure.

    Updated by      Richard Gooch   28-JUN-1999: Created experimental function
  <dir_rewind>.

    Updated by      Richard Gooch   19-MAY-2000: Created experimental function
  <dir_get_size>.

    Last updated by Richard Gooch   13-SEP-2000: Cosmetic changes.


*/
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <karma.h>
#include <karma_dir.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_p.h>
#include <karma_a.h>

/*  Check for brain damaged platforms which don't define  S_ISSOCK
*/
#ifndef S_ISSOCK
#  if defined(S_IFMT) && defined(S_IFSOCK)
#    define S_ISSOCK(m)     (((m) & S_IFMT) == S_IFSOCK)
#  else
!!!! ERROR !!! *** Do not know how to check for sockets ****
#  endif
#endif

/*  God-awful hacks  */
#ifndef S_ISLNK
#  define S_ISLNK(m)      (((m)&0170000) == 0120000)
#endif


#if __STDC__ == 1
#  define MAGIC_NUMBER 2942390429u
#else
#  define MAGIC_NUMBER (unsigned int) 2942390429
#endif

#define VERIFY_DIR(dir) if (dir == NULL) \
{fprintf (stderr, "NULL dir passed\n"); \
 a_prog_bug (function_name); } \
if ((uaddr) dir % sizeof (void *) != 0) \
{fprintf (stderr, "Dir pointer not aligned properly\n"); \
 a_prog_bug (function_name); } \
if (dir->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed dir object\n"); \
 a_prog_bug (function_name); }

/*  Internal definition of KDir object structure type  */
struct dir_type
{
    unsigned int magic_number;
    char *dirname;
    DIR *dirp;
    KFileInfo file;
    char sym_path[STRING_LENGTH];
};


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
KDir dir_open (CONST char *dirname)
/*  [SUMMARY] Open a directory for reading (scanning).
    <dirname> The directory name.
    [RETURNS] A KDir object on success, else NULL.
*/
{
    KDir dir;
    DIR *dirp;
    struct stat statbuf;
    static char function_name[] = "dir_open";

    if (stat (dirname, &statbuf) == -1)
    {
	fprintf (stderr, "%s: error statting path: \"%s\"\t%s\n",
		 function_name, dirname, ERRSTRING);
	return (NULL);
    }
    if ( !S_ISDIR (statbuf.st_mode) )
    {
	fprintf (stderr, "%s: path: \"%s\" is not a directory\n",
		 function_name, dirname);
	return (NULL);
    }
    if ( ( dirp = opendir (dirname) ) == NULL )
    {
	fprintf (stderr, "Error opening directory: \"%s\"\t%s\n",
		 dirname, ERRSTRING);
	return (NULL);
    }
    if ( ( dir = m_alloc (sizeof *dir) ) == NULL )
    {
	m_error_notify (function_name, "directory object");
	return (NULL);
    }
    if ( ( dir->dirname = st_dup (dirname) ) == NULL )
    {
	m_error_notify (function_name, "directory name");
	m_free (dir);
	return (NULL);
    }
    dir->magic_number = MAGIC_NUMBER;
    dir->dirp = dirp;
    dir->file.dirname = dir->dirname;
    return (dir);
}   /*  End Function dir_open  */

/*PUBLIC_FUNCTION*/
KFileInfo *dir_read (KDir dir, unsigned int skip_control)
/*  [SUMMARY] Read (scan) a directory for files.
    <dir> The directory object.
    <skip_control> Determines whether or not to skip the special files: "."
    and "..". See [<dir_SKIP_VALUES>] for a list of legal values.
    [RETURNS] A pointer to a <<KFileInfo>> structure on success, else NULL.
    The data in this structure is valid until the next call to [<dir_read>] or
    [<dir_close>] with this directory object.
*/
{
    flag scan_another = TRUE;
    flag other = TRUE;
    int sym_length;
    long tmp_numgrp;
    long count;
    struct stat statbuf;
    KFileInfo *file;
    struct dirent *dp = NULL;  /*  Initialised to keep compiler happy  */
    char pathname[STRING_LENGTH];
    static uid_t euid = -1;
    /*  I was wondering where memory block overrun was happening! Have a look
	at the man page for  getgroups(2V)  under SunOS 4.1.3 for a
	demonstration of stupidity! At least they could have given a warning in
	the BUGS section...  */
#ifdef OS_SunOS
#  define GID_T int
#else
#  define GID_T gid_t
#endif
    static GID_T *groups = NULL;
    static long num_groups = -1;
    static char function_name[] = "dir_read";

    VERIFY_DIR (dir);
    /*  Initialisation  */
    if (num_groups < 0)
    {
	euid = geteuid ();
	if ( ( tmp_numgrp = sysconf (_SC_NGROUPS_MAX) ) == -1 )
	{
	    fprintf (stderr, "Error getting NGROUPS_MAX\t%s\n", ERRSTRING);
	    return (NULL);
	}
	if ( ( groups = m_alloc (sizeof *groups * tmp_numgrp) ) == NULL )
	{
	    m_abort (function_name, "array of process group IDs");
	}
	/*  Under SunOS 4.1.3 with GCC 2.5.8, the following line generates a
	    warning about incompatible pointer type (arg 2). This is because
	    the GCC  unistd.h  prototype is broken.
	    */
	if ( ( num_groups = getgroups (tmp_numgrp, groups) ) < 0 )
	{
	    fprintf (stderr, "Error getting process group IDs\t%s\n",
		     ERRSTRING);
	    return (NULL);
	}
    }
    while (scan_another)
    {
	if ( ( dp = readdir (dir->dirp) ) == NULL ) return (NULL);
	switch (skip_control)
	{
	  case KDIR_DOT:
	    if (strcmp ("..", dp->d_name) == 0) continue;
	    break;
	  case KDIR_DOTDOT:
	    if (strcmp (".", dp->d_name) == 0) continue;
	    break;
	  case KDIR_DOT_AND_DOTDOT:
	    break;
	  case KDIR_NO_DOTS:
	    if (strcmp (".", dp->d_name) == 0) continue;
	    if (strcmp ("..", dp->d_name) == 0) continue;
	    break;
	  default:
	    fprintf (stderr, "Illegal value of  skip_control: %u\n",
		     skip_control);
	    a_prog_bug (function_name);
	    break;
	}
	scan_another = FALSE;
    }
    file = &dir->file;
    file->filename = dp->d_name;
    /*  Construct full pathname  */
    strcpy (pathname, dir->dirname);
    strcat (pathname, "/");
    strcat (pathname, file->filename);
    /*  Get stats on file  */
    if (lstat (pathname, &statbuf) != 0)
    {
	fprintf (stderr, "Error getting stats on file: \"%s\"\t%s\n",
		 pathname, ERRSTRING);
	return (NULL);
    }
    file->type = KFILETYPE_REGULAR;
    if ( S_ISLNK (statbuf.st_mode) )
    {
	/*  Symbolic link  */
	if (stat (pathname, &statbuf) != 0)
	{
	    if (errno == ENOENT)
	    {
		/*  Symbolic link does not point to anything  */
		file->type = KFILETYPE_DANGLING_SYMLINK;
	    }
	    else
	    {
		fprintf (stderr,"Error getting stats on symlink: \"%s\"\t%s\n",
			 pathname, ERRSTRING);
		return (NULL);
	    }
	}
	if ( ( sym_length = readlink (pathname, dir->sym_path,
				      STRING_LENGTH - 1) ) == -1 )
	{
	    fprintf (stderr, "Error reading symbolic link: \"%s\"\t%s\n",
		     pathname, ERRSTRING);
	    return (NULL);
	}
	dir->sym_path[sym_length] = '\0';
	file->sym_path = dir->sym_path;
	file->is_symlink = TRUE;
	if (file->type == KFILETYPE_DANGLING_SYMLINK) return (file);
    }
    else
    {
	file->sym_path = NULL;
	file->is_symlink = FALSE;
    }
    if ( S_ISDIR (statbuf.st_mode) )       file->type = KFILETYPE_DIRECTORY;
    else if ( S_ISCHR (statbuf.st_mode) )  file->type = KFILETYPE_CHAR;
    else if ( S_ISBLK (statbuf.st_mode) )  file->type = KFILETYPE_BLOCK;
    else if ( S_ISREG (statbuf.st_mode) )  file->type = KFILETYPE_REGULAR;
    else if ( S_ISSOCK (statbuf.st_mode) ) file->type = KFILETYPE_SOCKET;
    else if ( S_ISFIFO (statbuf.st_mode) ) file->type = KFILETYPE_FIFO;
    else
    {
	fprintf (stderr, "Unknown file mode: %d\n", (int) statbuf.st_mode);
	return (NULL);
    }
    file->uid = statbuf.st_uid;
    file->gid = statbuf.st_gid;
    file->mode = statbuf.st_mode;
    file->num_links = statbuf.st_nlink;
    file->dev_num = statbuf.st_dev;
    file->inode = statbuf.st_ino;
    file->size = statbuf.st_size;
    file->atime = statbuf.st_atime;
    file->mtime = statbuf.st_mtime;
    file->ctime = statbuf.st_ctime;
/*  NOT PORTABLE (not in POSIX)
    file->blocksize = statbuf.st_blksize;
*/
    /*  Determine file access permissions for this process  */
    file->can_read = FALSE;
    file->can_write = FALSE;
    file->can_execute = FALSE;
    if (euid == file->uid)
    {
	/*  Owner  */
	if (S_IRUSR & file->mode) file->can_read = TRUE;
	if (S_IWUSR & file->mode) file->can_write = TRUE;
	if (S_IXUSR & file->mode) file->can_execute = TRUE;
	other = FALSE;
    }
    else
    {
	/*  See if in group  */
	for (count = 0; count < num_groups; ++count)
	{
	    if (groups[count] == file->gid)
	    {
		/*  In the group  */
		if (S_IRGRP & file->mode) file->can_read = TRUE;
		if (S_IWGRP & file->mode) file->can_write = TRUE;
		if (S_IXGRP & file->mode) file->can_execute = TRUE;
		other = FALSE;
		count = num_groups;
	    }
	}
    }
    if (other)
    {
	if (S_IROTH & file->mode) file->can_read = TRUE;
	if (S_IWOTH & file->mode) file->can_write = TRUE;
	if (S_IXOTH & file->mode) file->can_execute = TRUE;
    }
    /*  Determine if local filesystem  */
    file->local_fs = ( (signed long) file->dev_num < 0 ) ? FALSE : TRUE;
    return (file);
}   /*  End Function dir_read  */

/*EXPERIMENTAL_FUNCTION*/
unsigned long dir_get_size (KDir dir, unsigned int skip_control,
			    unsigned long *max_width)
/*  [SUMMARY] Find number of entries in a directory.
    <dir> The directory object. This is automatically rewound.
    <skip_control> Determines whether or not to skip the special files: "."
    and "..". See [<dir_SKIP_VALUES>] for a list of legal values.
    <max_width> The maximum width of the entries is written here. This may be
    NULL.
    [RETURNS] The number of entries in the directory.
*/
{
    struct dirent *dp;
    unsigned long num_entries = 0;
    unsigned long l_max_width = 0;
    static char function_name[] = "dir_get_size";

    VERIFY_DIR (dir);
    rewinddir (dir->dirp);
    while ( ( dp = readdir (dir->dirp) ) != NULL )
    {
	switch (skip_control)
	{
	  case KDIR_DOT:
	    if (strcmp ("..", dp->d_name) == 0) continue;
	    break;
	  case KDIR_DOTDOT:
	    if (strcmp (".", dp->d_name) == 0) continue;
	    break;
	  case KDIR_DOT_AND_DOTDOT:
	    break;
	  case KDIR_NO_DOTS:
	    if (strcmp (".", dp->d_name) == 0) continue;
	    if (strcmp ("..", dp->d_name) == 0) continue;
	    break;
	  default:
	    fprintf (stderr, "Illegal value of  skip_control: %u\n",
		     skip_control);
	    a_prog_bug (function_name);
	    break;
	}
	++num_entries;
	if (max_width != NULL) l_max_width += strlen (dp->d_name);
    }
    if (max_width != NULL) *max_width = l_max_width;
    return (num_entries);
}   /*  End Function dir_get_size  */

/*PUBLIC_FUNCTION*/
void dir_close (KDir dir)
/*  [SUMMARY] Close a directory.
    <dir> The directory object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "dir_close";

    VERIFY_DIR (dir);
    if (closedir (dir->dirp) != 0)
    {
	fprintf (stderr, "Error closing directory: \"%s\"\t%s\n",
		 dir->dirname, ERRSTRING);
    }
    m_free (dir->dirname);
    m_free (dir);
}   /*  End Function dir_close  */

/*EXPERIMENTAL_FUNCTION*/
void dir_rewind (KDir dir)
/*  [SUMMARY] Rewind a directory.
    <dir> The directory object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "dir_rewind";

    VERIFY_DIR (dir);
    rewinddir (dir->dirp);
}   /*  End Function dir_rewind  */
