/*LINTLIBRARY*/
/*  main.c

    This code provides code to load shared objects.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains code to load shared objects.


    Written by      Richard Gooch   26-DEC-1996

    Updated by      Richard Gooch   29-DEC-1996: Improved debugging messages.

    Updated by      Richard Gooch   11-FEB-1997: Added initial test for shared
  object. Removed requirement for dependency file.

    Updated by      Richard Gooch   19-JUN-1997: Removed some error messages.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <r> package.

    Updated by      Richard Gooch   10-MAR-1999: Cope with NULL return from
  <r_get_karmabase>.

    Updated by      Richard Gooch   17-MAR-1999: Improved error message from
  <dl_load_one_object>.

    Last updated by Richard Gooch   26-JUL-2001: Spelling fix.


*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <karma.h>
#include <os.h>
#ifdef HAS_DL
#  include <dlfcn.h>
#  include <karma_dl.h>
#  include <karma_r.h>

#  define BUF_LEN 4096

/*  Private functions  */


/*  Private data  */


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
void *dl_load_object (CONST char *name, CONST char *dirs,
		      char failed_object[STRING_LENGTH])
/*  [SUMMARY] Load shared object.
    <name> The name of the shared object to load (without the ".so" extension).
    <dirs> An optional list of colon-separated directories in which to search
    for the shared object, and any other shared objects depended upon. This may
    be NULL.
    <failed_object> The name of the object which failed to load is written here
    (without the ".so" extension).
    [RETURNS] A handle to the loaded object on success, else NULL.
*/
{
    int len;
    FILE *fp;
    char *start_ptr, *stop_ptr;
    CONST char *karmabase;
    void *handle;
    char object_path[STRING_LENGTH];
    char obj_dir[STRING_LENGTH], dep_file[STRING_LENGTH];
    char dep_libs[BUF_LEN], dep_lib[BUF_LEN];
    char search_dirs[BUF_LEN];
    char tmp[BUF_LEN];
    static int load_flags = 0;
    static flag first_time = TRUE;
    /*static char function_name[] = "dl_load_object";*/

    if (first_time)
    {
#  ifdef RTLD_GLOBAL
	load_flags |= RTLD_GLOBAL;
#  endif
#  ifdef RTLD_LAZY
	load_flags |= RTLD_LAZY;
#  endif
	first_time = FALSE;
    }
    failed_object[0] = '\0';
    if ( ( karmabase = r_get_karmabase () ) == NULL )
    {
	strcpy (failed_object, "KARMABASE");
	return (NULL);
    }
    sprintf (object_path, "%s/lib/shared-objects/%s.so", karmabase, name);
    /*  First a quick check for the shared object  */
    if (access (object_path, R_OK) != 0)
    {
	strcpy (failed_object, name);
	return (NULL);
    }
    sprintf (obj_dir, "%s/lib/shared-objects", karmabase);
    sprintf (dep_file, "%s/%s.libs", obj_dir, name);
    /*  Open dependency file if available  */
    if ( ( fp = fopen (dep_file, "r") ) != NULL )
    {
	/*  Copy search directories to amalgamated list  */
	if (dirs == NULL) search_dirs[0] = '\0';
	else
	{
	    strcpy (search_dirs, dirs);
	    strcat (search_dirs, ":");
	}
	/*  First read search directories  */
	if (fgets (tmp, BUF_LEN, fp) == NULL)
	{
	    fclose (fp);
	    return (NULL);
	}
	tmp[strlen (tmp) - 1] = '\0';
	/*  Append to amalgamated list  */
	strcat (search_dirs, tmp);
	/*  Append $KARMALIBPATH:$KARMALIBPATH/shared-objects  */
	len = strlen (search_dirs);
	if ( (len > 0) && (search_dirs[len - 1] != ':') )
	    strcat (search_dirs, ":");
	sprintf (tmp, "%s/lib:%s/lib/shared-objects", karmabase, karmabase);
	strcat (search_dirs, tmp);
	/*  Then read depend libraries  */
	if (fgets (dep_libs, BUF_LEN, fp) == NULL)
	{
	    fclose (fp);
	    return (NULL);
	}
	fclose (fp);
	dep_libs[strlen (dep_libs) - 1] = '\0';
	/*  Loop through each depend library  */
	for (start_ptr = dep_libs; start_ptr[0] != '\0'; start_ptr = stop_ptr)
	{
	    stop_ptr = strchr (start_ptr, ' ');
	    if (stop_ptr == NULL) stop_ptr = start_ptr + strlen (start_ptr);
	    /*  Construct library name  */
	    strcpy (dep_lib, "lib");
	    strncat (dep_lib, start_ptr, stop_ptr - start_ptr);
	    if (dl_load_one_object (dep_lib, search_dirs, load_flags) == NULL)
	    {
		strcpy (failed_object, dep_lib);
		return (NULL);
	    }
	    if (*stop_ptr == ' ') ++stop_ptr;
	}
    }
    if ( ( handle = dlopen (object_path, load_flags) ) == NULL )
    {
	fputs (dlerror (), stderr);
	putc ('\n', stderr);
	strcpy (failed_object, name);
	return (NULL);
    }
    return (handle);
}   /*  End Function dl_load_object  */

/*UNPUBLISHED_FUNCTION*/
void *dl_load_one_object (CONST char *object, CONST char *search_dirs,
			  int lflags)
/*  [SUMMARY] Load a single shared object.
    <object> The name of the shared object.
    <search_dirs> A colon-separated directories in which to search for the
    shared object.
    <lflags> The load flags.
    [RETURNS] A handle to the loaded object on success, else NULL.
*/
{
    int count;
    CONST char *ptr = search_dirs;
    void *handle;
    char dir[STRING_LENGTH], path[STRING_LENGTH];
    static char function_name[] = "dl_load_one_object";

    /*  Loop through each search directory  */
    while (ptr[0] != '\0')
    {
	/*  Construct directory name  */
	for (count = 0; ( (*ptr != '\0') && (*ptr != ':') );
	     dir[count++] = *ptr++);
	dir[count] = '\0';
	if (*ptr == ':') ++ptr;
	sprintf (path, "%s/%s.so", dir, object);
	if (access (path, R_OK) == -1) continue;
	handle = dlopen (path, lflags);
	if (handle != NULL) return (handle);
	fprintf ( stderr, "%s: error loading: \"%s\"\t%s\n",
		  function_name, path, dlerror () );
	return (NULL);
    }
    return (NULL);
}   /*  End Function dl_load_one_object  */


/*  Private functions follow  */

#else  /*  HAS_DL  */


/*  Stub functions follow  */

/*STUB_FUNCTION*/
void *dl_load_object (CONST char *name, CONST char *dirs,
		      char failed_object[STRING_LENGTH])
/*  [SUMMARY] Load shared object.
    <name> The name of the shared object to load (without the ".so" extension).
    <dirs> An optional list of colon-separated directories in which to search
    for the shared object, and any other shared objects depended upon. This may
    be NULL.
    <failed_object> The name of the object which failed to load is written here
    (without the ".so" extension).
    [RETURNS] A handle to the loaded object on success, else NULL.
*/
{
    strcpy (failed_object, "SHARED OBJECTS NOT SUPPORTED");
    return (NULL);
}   /*  End Function dl_load_object  */

/*STUB_FUNCTION*/
void *dl_load_one_object (CONST char *object, CONST char *search_dirs,
			  int lflags)
/*  [SUMMARY] Load a single shared object.
    <object> The name of the shared object.
    <search_dirs> An colon-separated directories in which to search for the
    shared object.
    <lflags> The load flags.
    [RETURNS] A handle to the loaded object on success, else NULL.
*/
{
    fprintf (stderr, "SHARED OBJECTS NOT SUPPORTED\n");
    return (NULL);
}   /*  End Function dl_load_one_object  */

#endif  /*  !HAS_DL  */
