/*LINTLIBRARY*/
/*  interface.c

    This code provides file descriptor Management routines.

    Copyright (C) 1995-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the management of file
  descriptors. This file contains the generic interface.


    Written by      Richard Gooch   3-OCT-1995

    Updated by      Richard Gooch   7-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   3-JUN-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   15-DEC-1996: Split into interface.c and
  native.c files.

    Updated by      Richard Gooch   18-FEB-1998: Changed interfaces to support
  new KManagedDescriptor object.

    Updated by      Richard Gooch   13-APR-1998: Published experimental
  function <dm_register_managers>.

    Last updated by Richard Gooch   28-JUN-1999: Changed interface to
  <dm_manage>.


*/
#include <stdio.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_dm.h>
#include <karma_a.h>
#include <karma_r.h>


/*  Private functions  */


/*  Private data  */
static KManagedDescriptor (*manage) (int fd, void *info, flag auto_close,
				     va_list argp) = NULL;
static void *(*unmanage) (int fd, KManagedDescriptor md) = NULL;


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void dm_register_managers ( KManagedDescriptor (*manage_func)
			    (int fd, void *info, flag auto_close,va_list argp),
			    void *(*unmanage_func)
			    (int fd, KManagedDescriptor md) )
/*  [SUMMARY] Register the managers.
    <manage_func> The function which will be called to manage descriptors. The
    interface to this function is the same as [<dm_manage>].
    <unmanage_func> The function which will be called to manage descriptors.
    The interface to this function is the same as [<dm_unmanage>].
    [RETURNS] Nothing. On failure the process aborts.
*/
{
    static char function_name[] = "dm_register_managers";

    if ( (manage != NULL) || (unmanage != NULL) )
    {
	fprintf (stderr, "Existing managers registered\n");
	a_prog_bug (function_name);
    }
    if (manage_func == NULL)
    {
	fprintf (stderr, "NULL manage_func passed\n");
	a_prog_bug (function_name);
    }
    if (unmanage_func == NULL)
    {
	fprintf (stderr, "NULL unmanage_func passed\n");
	a_prog_bug (function_name);
    }
    manage = manage_func;
    unmanage = unmanage_func;
}   /* End Function dm_register_managers  */

/*UNPUBLISHED_FUNCTION*/
flag dm_test_if_setup ()
/*  [SUMMARY] Test if descriptor managers setup.
    [RETURNS] TRUE if descriptor managers setup.
*/
{
    return (manage == NULL) ? FALSE : TRUE;
}   /* End Function dm_test_if_setup  */

/*EXPERIMENTAL_FUNCTION*/
KManagedDescriptor dm_manage (int fd, void *info, flag auto_close, ...)
/*  [SUMMARY] Manage a file descriptor for events.
    [PURPOSE] This routine will manage a file descriptor for activity by
    registering callback routines. Only one set of callbacks may be registered
    per file descriptor.
    <fd> The file descriptor to manage.
    <info> The initial value of the arbitrary pointer associated with the
    managed descriptor. This pointer may be modified by the callback routines.
    <auto_close> If TRUE, the descriptor is automatically closed upon a closure
    event.
    [VARARGS] The optional list of parameter event-key event-function pairs
    must follow. This list must be terminated with the value
    KARMA_FD_EVENT_END. See [<dm_EVENTS>] for the list of attributes.
    [RETURNS] A KManagedDescriptor object on success, else NULL.
*/
{
    va_list argp;
    KManagedDescriptor md;
    unsigned int att_key;
    void *close_func = NULL;
    static char function_name[] = "dm_manage";

    va_start (argp, auto_close);
    if (manage == NULL)
    {
	fprintf (stderr, "WARNING in %s: <dm> package not setup. ",
		 function_name);
	fprintf (stderr, "Using <dm_native_setup>\n");
	dm_native_setup ();
    }
    while ( ( att_key = va_arg (argp, unsigned int) ) != KARMA_FD_EVENT_END )
    {
	switch (att_key)
	{
	  case KARMA_FD_EVENT_CLOSE:
	    close_func = va_arg (argp, void *);
	    break;
	}
    }
    va_end (argp);
    /*  Check if descriptor is a connection  */
    if (close_func != NULL)
    {
	/*  Check if closure detectable  */
	if (r_get_bytes_readable (fd) < 0)
	{
	    fprintf (stderr,
		     "close_func  supplied and closure not detectable\n");
	    a_prog_bug (function_name);
	}
    }
    va_start (argp, auto_close);
    md = (*manage) (fd, info, auto_close, argp);
    va_end (argp);
    return (md);
}   /*  End Function dm_manage  */

/*PUBLIC_FUNCTION*/
void *dm_unmanage (int fd, KManagedDescriptor md)
/*  [SUMMARY] Terminate the management of a file descriptor for activity.
    [NOTE] The routine will NOT close the descriptor (nor does it assume the
    descriptor is open).
    <fd> The descriptor to unmanage. This must always be valid.
    <md> The entry to unmanage. If this is NULL then <<fd>> is used.
    [RETURNS] The arbitrary pointer passed when the descriptor was managed.
*/
{
    static char function_name[] = "dm_unmanage";

    if (unmanage == NULL)
    {
	fprintf (stderr, "No manager registered\n");
	a_prog_bug (function_name);
    }
    return (*unmanage) (fd, md);
}   /*  End Function dm_unmanage  */
