/*LINTLIBRARY*/
/*  native.c

    This code provides file descriptor Management routines.

    Copyright (C) 1995-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the management of file
  descriptors. This file contains the native support.


    Written by      Richard Gooch   3-OCT-1995

    Updated by      Richard Gooch   7-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   3-JUN-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   15-DEC-1996: Split into interface.c and
  native.c files.

    Updated by      Richard Gooch   22-MAR-1997: Compute MAX_FD + 1 for call to
  <select>: slightly faster than using FD_SETSIZE.

    Updated by      Richard Gooch   22-AUG-1997: Switched to use of poll(2)
  system call. Emulate if not available.

    Updated by      Richard Gooch   27-AUG-1997: Fixed bug in <dm_native_poll>:
  did not check for POLLHUP (hangup) and fixed memory leak for active list (was
  not declaring length and pointer as static).

    Updated by      Richard Gooch   5-SEP-1997: Added message about bug in
  IRIX 6.2 when draining closed descriptor gives incorrect result.

    Updated by      Richard Gooch   9-SEP-1997: Reduced jitter in synchronous
  event dispatch.

    Updated by      Richard Gooch   22-NOV-1997: Placed use of POLLHUP inside
  #ifdef POLLHUP so code compiles on machines without the poll(2) system call.

    Updated by      Richard Gooch   5-DEC-1997: Added message about bug in
  Solaris 2.x when draining closed descriptor gives incorrect result.

    Updated by      Richard Gooch   31-DEC-1997: Move attempt to drain data to
  <_dm_native_read_entry> from <_dm_native_close_entry>.

    Updated by      Richard Gooch   17-FEB-1998: Clear revents field for
  entries placed on unmanage list.

    Updated by      Richard Gooch   18-FEB-1998: Changed interfaces to support
  new KManagedDescriptor object.

    Updated by      Richard Gooch   19-FEB-1998: Ripped out usage of new
  KManagedDescriptor object and create a fake instead. Have no choice but to do
  searches by fd.

    Updated by      Richard Gooch   21-FEB-1998: Added bit array for fds.

    Updated by      Richard Gooch   27-JUN-1998: Took account of variations in
  possible return values from select(2).

    Updated by      Richard Gooch   17-OCT-1998: Created experimental function
  <dm_native_get_activity_count>.

    Updated by      Richard Gooch   19-OCT-1998: Created experimental functions
  <dm_native_clear_idle_counter> and <dm_native_get_idle_time>.

    Updated by      Richard Gooch   21-OCT-1998: Renamed function
  <dm_native_clear_idle_counter> to <dm_native_clear_idle_timeer> and created
  experimental function <dm_native_clear_activity_counter>.

    Updated by      Richard Gooch   12-JAN-1999: Made use of <w_test_supported>
  to take account of disabling of work function support.

    Updated by      Richard Gooch   19-FEB-1999: Display FD when closed
  descriptor is readable. Set FD bit in <_dm_native_manage>.

    Updated by      Richard Gooch   20-FEB-1999: Test for cleared FD bit rather
  than set FD fit in <_dm_native_unmanage>.

    Updated by      Richard Gooch   3-MAR-1999: Created experimental function
  <dm_native_wait_input>.

    Updated by      Richard Gooch   28-JUN-1999: Took account of interface
  change for manage function.

    Updated by      Richard Gooch   20-AUG-1999: Fixed bug in function
  <dm_native_manage>: <<managed_fd_type>> structure was not being initialised.

    Updated by      Richard Gooch   15-FEB-2000: Moved test for POLLHUP after
  other event tests so that input followed by closure won't lose the input
  event, where both events come from the one call to poll(2). Don't test for
  POLLHUP if <<close_func>> not given.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Updated by      Richard Gooch   7-OCT-2000: Changed interface to function
  <dm_native_wait_input>.

    Last updated by Richard Gooch   20-SEP-2001: Cope with location of poll.h
  under Darwin.


*/
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <os.h>
#ifdef HAS_SOCKETS
#  include <sys/stat.h>
#  include <sys/time.h>
#  include <fcntl.h>
#  include <sys/ioctl.h>
#  ifdef OS_AIX
#    include <sys/select.h>
#  endif
#endif
#ifdef HAS_POLL
#  ifdef OS_Darwin      /*  Fucking gratuitous difference  */
#    include <poll.h>
#  else                 /*  Everybody else in the world    */
#    include <sys/poll.h>
#  endif
#endif
#ifdef OS_VXMVX
#  include <vx/vx.h>
#endif
#include <karma.h>
#include <karma_dm.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_r.h>
#include <karma_e.h>
#include <karma_w.h>

#define MIN_TABLE_SIZE 256
#define INITIAL_FD_BITS 65536
#define INITIAL_FD_BYTES INITIAL_FD_BITS >> 3

#if defined(HAS_SOCKETS) /* || defined(HAS_COMMUNICATIONS_EMULATION) */
#  define COMMUNICATIONS_AVAILABLE
#endif

#ifdef OS_Linux
#  define POLLWRINODE  0x8000
#endif

#ifndef HAS_POLL    /*  poll(2) emulation  */
struct pollfd
{
    int fd;			/*  File descriptor to poll                 */
    short int events;		/*  Types of events poller cares about      */
    short int revents;		/*  Types of events that actually occurred  */
};

#  define POLLIN		1         /* There is data to read          */
#  define POLLPRI		2         /* There is urgent data to read   */
#  define POLLOUT		4         /* Writing now will not block.    */

STATIC_FUNCTION (int poll, (struct pollfd *ufds, unsigned int nfds,
			    int timeout) );
#endif  /*  !HAS_POLL  */

/*  Structure and private data to manage descriptors  */
struct managed_fd_type
{
    void *info;
    flag (*input_func) ();
    void (*close_func) ();
    flag (*output_func) ();
    flag (*exception_func) ();
    flag (*inode_write_func) ();
    flag auto_close;
};

struct managed_descriptor_type
{
    char dummy;
};

static struct pollfd *pollfd_array = NULL;         /*  Ordered list by fd's  */
static struct managed_fd_type *cbk_array = NULL;   /*  Matched to above      */
static unsigned int num_allocated = 0;
static unsigned int num_managed = 0;
static unsigned int *unmanage_list = NULL;
static unsigned int unmanage_list_size = 0;
static unsigned int num_to_unmanage = 0;
static flag inside_poll = FALSE;
static struct managed_descriptor_type fake_md;
static unsigned long activity_count = 0;
static flag count_idle_time = FALSE;
static unsigned long idle_us = 0;


static unsigned char *fd_array = NULL;             /*  Array of bits         */
static unsigned int fd_array_size = 0;             /*  In bytes              */

#define TEST_FD_BIT(fd) ( fd_array[(fd)>>3] & ( 1 << ((fd)&7) ) )
#define CLEAR_FD_BIT(fd) ( fd_array[(fd)>>3] &= ~( 1 << ((fd)&7) ) )
#define SET_FD_BIT(fd) ( fd_array[(fd)>>3] |= ( 1 << ((fd)&7) ) )


/*  Private functions  */
STATIC_FUNCTION (void *_dm_native_unmanage_entry, (unsigned int index) );
STATIC_FUNCTION (flag _dm_native_read_entry,
		 (int fd, struct managed_fd_type *entry) );
STATIC_FUNCTION (void _dm_native_close_entry, (unsigned int index) );
STATIC_FUNCTION (flag _dm_native_increase_table_size,
		 (unsigned int num_extra) );
STATIC_FUNCTION (void _dm_native_add_to_unmanage_list,
		 (unsigned int index) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void dm_native_setup ()
/*  [SUMMARY] Setup descriptor management using native implementation.
    [RETURNS] Nothing.
*/
{
    dm_register_managers (dm_native_manage, dm_native_unmanage);
}   /*  End Function dm_native_setup  */

/*EXPERIMENTAL_FUNCTION*/
void dm_native_poll (long timeout_ms)
/*  [SUMMARY] Poll all managed descriptors for any activity.
    <timeout_ms> The time (in milliseconds) to poll. If this is less than 0
    the routine will poll forever (until some activity occurs or a signal is
    caught).
    [RETURNS] Nothing.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    int count, index, num_ready, errnum;
    signed long time_taken;
    struct timeval before, after;
    struct pollfd *pollfd_ptr;
    struct managed_fd_type *cbk_ptr;
    static unsigned int index_arr_size = 0;
    static unsigned int *active_array = NULL;
    static char function_name[] = "dm_native_poll";

    if (inside_poll)
    {
	fprintf (stderr, "Code non-reentrant\n");
	a_prog_bug (function_name);
    }
    inside_poll = TRUE;
    if ( w_test_supported () && w_work_to_be_done () )
    {
	timeout_ms = 0;
	w_do_work ();
    }
    /*  Minimise race window, to reduce jitter for synchronous events. Note
	that events are not lost, just delayed  */
    e_unix_dispatch_events (DISPATCH_SYNCHRONOUS);
    if (count_idle_time)
    {
	if (gettimeofday (&before, NULL) != 0)
	{
	    fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
    num_ready = poll (pollfd_array, num_managed, timeout_ms);
    errnum = errno;
    if (count_idle_time)
    {
	if (gettimeofday (&after, NULL) != 0)
	{
	    fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	time_taken = (after.tv_sec - before.tv_sec) * 1000000;
	time_taken += (after.tv_usec - before.tv_usec);
	idle_us += time_taken;
    }
    e_unix_dispatch_events (DISPATCH_SYNCHRONOUS);
    switch (num_ready)
    {
      case 0:
	inside_poll = FALSE;
	return;
	/*break;*/
      case -1:
	inside_poll = FALSE;
	if (errnum == EINTR) return;
	/*  Failure  */
	fprintf ( stderr, "Error calling  poll(2)\t%s\n", STRERROR (errnum) );
	return;
	/*break;*/
      default:
	/*  Success: there is activity  */
	break;
    }
    ++activity_count;
    if (index_arr_size < num_ready)
    {
	if (active_array != NULL) m_free (active_array);
	index_arr_size *= 2;
	if (index_arr_size < MIN_TABLE_SIZE) index_arr_size = MIN_TABLE_SIZE;
	if ( ( active_array = m_alloc (index_arr_size * sizeof *active_array) )
	     == NULL )
	{
	    m_abort (function_name, "active index array");
	}
    }
    /*  Extract list of active descriptors into index array  */
    count = 0;
    for (pollfd_ptr = pollfd_array; TRUE; ++pollfd_ptr)
    {
	if (pollfd_ptr->revents == 0) continue;
	/*  Have an active descriptor  */
	active_array[count] = pollfd_ptr - pollfd_array;
	if (++count >= num_ready) break;
    }
    /*  Now process the list of active descriptors  */
    for (count = 0; count < num_ready; ++count)
    {
	index = active_array[count];
	pollfd_ptr = pollfd_array + index;
	cbk_ptr = cbk_array + index;
	if ( (cbk_ptr->exception_func != NULL) &&
	     (pollfd_ptr->revents & POLLPRI) )
	{
	    /*  Exception occurred  */
	    if ( !(*cbk_ptr->exception_func) (pollfd_ptr->fd, &cbk_ptr->info) )
	    {
		/*  Descriptor to be unmanaged and closed  */
		_dm_native_close_entry (index);
		continue;
	    }
	}
	if (pollfd_ptr->revents & POLLIN)
	{
	    /*  Input/ close occurred  */
	    if ( !_dm_native_read_entry (pollfd_ptr->fd, cbk_ptr) )
	    {
		/*  Close and unmanage  */
		_dm_native_close_entry (index);
		continue;
	    }
	}
	if ( (cbk_ptr->output_func != NULL) &&
	     (pollfd_ptr->revents & POLLOUT) )
	{
	    /*  Output occurred  */
	    if ( !(*cbk_ptr->output_func) (pollfd_ptr->fd, &cbk_ptr->info) )
	    {
		/*  Descriptor to be unmanaged and closed  */
		_dm_native_close_entry (index);
		continue;
	    }
	}
#ifdef POLLWRINODE
	if ( (cbk_ptr->inode_write_func != NULL) &&
	     (pollfd_ptr->revents & POLLWRINODE) )
	{
	    /*  Inode write occurred  */
	    if ( !(*cbk_ptr->inode_write_func) (pollfd_ptr->fd,
						&cbk_ptr->info) )
	    {
		/*  Descriptor to be unmanaged and closed  */
		_dm_native_close_entry (index);
		continue;
	    }
	}
#endif
#ifdef POLLHUP
	if ( (cbk_ptr->close_func != NULL) && (pollfd_ptr->revents & POLLHUP) )
	{
	    /*  Hangup (close) occurred  */
	    _dm_native_close_entry (index);
	    continue;
	}
#endif
    }
    /*  Processed all activity (we hope)  */
    inside_poll = FALSE;
    /*  Now unmanage anything requiring it  */
    while (num_to_unmanage > 0)
    {
	/*  Have to unmanage the entry with the largest index first  */
	for (count = 0, index = 0; count < num_to_unmanage; ++count)
	{
	    if (unmanage_list[count] > index) index = unmanage_list[count];
	}
	_dm_native_unmanage_entry (index);
	--num_to_unmanage;
    }

#else  /*  COMMUNICATIONS_AVAILABLE  */
    fprintf (stderr, "Operating system does not support communications\n");
    inside_poll = FALSE;
#endif  /*  COMMUNICATIONS_AVAILABLE  */
}   /*  End Function dm_native_poll  */

/*EXPERIMENTAL_FUNCTION*/
void dm_native_clear_activity_counter ()
/*  [SUMMARY] Clear the idle counter.
    [RETURNS] Nothing.
*/
{
    activity_count = 0;
}   /*  End Function dm_native_clear_activity_counter  */

/*EXPERIMENTAL_FUNCTION*/
unsigned long dm_native_get_activity_count ()
/*  [SUMMARY] Get the current activity count.
    [PURPOSE] This function will get the current activity count. This is the
    count of the number of times the OS has detected activity on one or more
    file descriptors.
    [RETURNS] The activity count.
*/
{
    return (activity_count);
}   /*  End Function dm_native_get_activity_count  */

/*EXPERIMENTAL_FUNCTION*/
void dm_native_clear_idle_timer ()
/*  [SUMMARY] Clear the idle timer.
    [RETURNS] Nothing.
*/
{
    count_idle_time = TRUE;
    idle_us = 0;
}   /*  End Function dm_native_clear_idle_timer  */

/*EXPERIMENTAL_FUNCTION*/
unsigned long dm_native_get_idle_time ()
/*  [SUMMARY] Get the idle time in microseconds.
    [RETURNS] The idle time.
*/
{
    return (idle_us);
}   /*  End Function dm_native_get_idle_time  */

/*EXPERIMENTAL_FUNCTION*/
flag dm_native_wait_input (int fd, int timeout)
/*  [SUMMARY] Block waiting for input on a file descriptor.
    <fd> The file descriptor.
    <timeout> The time to wait for (in milliseconds). If this is less than 0
    the function waits indefinately.
    [RETURNS] TRUE on activity, else FALSE on error/timeout.
*/
{
    int num_ready;
    struct pollfd fds;

    fds.fd = fd;
    fds.events = POLLIN;
    while ( ( num_ready = poll (&fds, 1, timeout) ) < 1 )
    {
	if (num_ready == 0) return (FALSE);
	if (errno == EINTR) continue;
	/*  Failure  */
	fprintf (stderr, "Error calling  poll(2)\t%s\n", ERRSTRING);
	return (FALSE);
    }
    /*  Success: there is activity  */
    return (TRUE);
}   /*  End Function dm_native_wait_input  */

/*UNPUBLISHED_FUNCTION*/
KManagedDescriptor dm_native_manage (int fd, void *info,
				     flag auto_close, va_list argp)
/*  [SUMMARY] Manage a file descriptor for events.
    [PURPOSE] This routine will manage a file descriptor for activity by
    registering callback routines. Only one set of callbacks may be registered
    per file descriptor.
    <fd> The file descriptor to manage.
    <info> The initial value of the arbitrary pointer associated with the
    managed descriptor. This pointer may be modified by the callback routines.
    <auto_close> If TRUE, the descriptor is automatically closed upon an
    unmanage event.
    <argp> The optional list of parameter event-key event-function pairs.
    [RETURNS] A KManagedDescriptor object on success, else NULL.
*/
{
    static char function_name[] = "dm_native_manage";
#ifdef COMMUNICATIONS_AVAILABLE
    short events;
    unsigned int tmp, att_key;
    unsigned char *fdarr;
    void **ptr;

    if (fd < 0)
    {
	fprintf (stderr, "Illegal fd: %d\n", fd);
	a_prog_bug (function_name);
    }
    /*  See if descriptor is already managed  */
    if (fd >= fd_array_size << 3)
    {
	/*  Need to increase array size  */
	tmp = ( 1 + (fd >> 3) ) / INITIAL_FD_BYTES;
	++tmp;
	tmp *= INITIAL_FD_BYTES;
	if ( ( fdarr = m_alloc (tmp) ) == NULL )
	{
	    m_abort (function_name, "fd array");
	}
	if (fd_array != NULL)
	{
	    m_copy (fdarr, fd_array, fd_array_size);
	    m_free (fd_array);
	}
	m_clear (fdarr + fd_array_size, tmp - fd_array_size);
	fd_array = fdarr;
	fd_array_size = tmp;
    }
    if ( TEST_FD_BIT (fd) )
    {
	fprintf (stderr, "Descriptor: %d is already managed\n", fd);
	a_prog_bug (function_name);
    }
    /*  Descriptor is not already managed  */
    SET_FD_BIT (fd);
    if ( !_dm_native_increase_table_size (1) ) return (NULL);
    pollfd_array[num_managed].fd = fd;
    events = 0;
    m_clear (cbk_array + num_managed, sizeof *cbk_array);
    while ( ( att_key = va_arg (argp, unsigned int) ) != KARMA_FD_EVENT_END )
    {
	switch (att_key)
	{
	  case KARMA_FD_EVENT_INPUT:
	    ptr = (void **) &cbk_array[num_managed].input_func;
	    *ptr = va_arg (argp, void *);
	    break;
	  case KARMA_FD_EVENT_OUTPUT:
	    ptr = (void **) &cbk_array[num_managed].output_func;
	    *ptr = va_arg (argp, void *);
	    break;
	  case KARMA_FD_EVENT_EXCEPTION:
	    ptr = (void **) &cbk_array[num_managed].exception_func;
	    *ptr = va_arg (argp, void *);
	    break;
	  case KARMA_FD_EVENT_CLOSE:
	    ptr = (void **) &cbk_array[num_managed].close_func;
	    *ptr = va_arg (argp, void *);
	    break;
	  case KARMA_FD_EVENT_INODE_WRITE:
	    ptr = (void **) &cbk_array[num_managed].inode_write_func;
	    *ptr = va_arg (argp, void *);
	    break;
	}
    }
    if (cbk_array[num_managed].input_func != NULL) events |= POLLIN;
    if (cbk_array[num_managed].output_func != NULL) events |= POLLOUT;
    if (cbk_array[num_managed].exception_func != NULL) events |= POLLPRI;
    if (cbk_array[num_managed].close_func != NULL) events |= POLLIN;
#ifdef POLLWRINODE
    if (cbk_array[num_managed].inode_write_func != NULL) events |= POLLWRINODE;
#endif
    pollfd_array[num_managed].events = events;
    pollfd_array[num_managed].revents = 0;
    cbk_array[num_managed].info = info;
    cbk_array[num_managed].auto_close = auto_close;
    ++num_managed;
    return (&fake_md);

#else  /*  COMMUNICATIONS_AVAILABLE  */
    fprintf (stderr, "%s: communications support not available\n",
	     function_name);
    return (NULL);
#endif  /*  COMMUNICATIONS_AVAILABLE  */
}   /*  End Function dm_native_manage  */

/*UNPUBLISHED_FUNCTION*/
void *dm_native_unmanage (int fd, KManagedDescriptor md)
/*  [SUMMARY] Terminate the management of a file descriptor for activity.
    [NOTE] The routine will NOT close the descriptor (nor does it assume the
    descriptor is open).
    <fd> The descriptor to unmanage. This must always be valid.
    <md> The entry to unmanage. If this is NULL then <<fd>> is used.
    [NOTE] This implementation ignores <<md>>, and relies solely upon <<fd>>.
    [RETURNS] The arbitrary pointer passed when the descriptor was managed.
*/
{
    unsigned int index;
    struct pollfd *pollfd_ptr;
    static char function_name[] = "dm_native_unmanage";

    if ( (md != NULL) && (md != &fake_md) )
    {
	fprintf (stderr, "KManagedDescriptor pointer: %p is not: %p\n",
		 md, &fake_md);
	a_prog_bug (function_name);
    }
    if (fd < 0)
    {
	fprintf (stderr, "Illegal fd: %d\n", fd);
	a_prog_bug (function_name);
    }
    if ( (fd >= fd_array_size << 3) || !TEST_FD_BIT (fd) )
    {
	fprintf (stderr, "Descriptor: %d not managed\n", fd);
	a_prog_bug (function_name);
    }
    /*  Get index  */
    for (index = 0, pollfd_ptr = pollfd_array; index < num_managed;
	 ++index, ++pollfd_ptr)
    {
	if (fd != pollfd_ptr->fd) continue;
	return _dm_native_unmanage_entry (index);
    }
    /*  Descriptor not found  */
    fprintf (stderr, "Descriptor: %d not managed, passed first test\n", fd);
    a_prog_bug (function_name);
    return (NULL);
}   /*  End Function dm_native_unmanage  */


/*  Private functions follow  */

static void *_dm_native_unmanage_entry (unsigned int index)
/*  [SUMMARY] Terminate the management of a file descriptor for activity.
    [NOTE] The routine will NOT close the descriptor (nor does it assume the
    descriptor is open).
    <index> The index of the descriptor in the table.
    [RETURNS] The arbitrary pointer passed when the descriptor was managed.
*/
{
    void *info;
    /*static char function_name[] = "_dm_native_unmanage_entry";*/

    info = cbk_array[index].info;
    CLEAR_FD_BIT (pollfd_array[index].fd);
    if (inside_poll)
    {
	_dm_native_add_to_unmanage_list (index);
	return (info);
    }
    if (num_managed > 1)
    {
	/*  Swap last entry with this one  */
	m_copy (pollfd_array + index, pollfd_array + num_managed - 1,
		 sizeof *pollfd_array);
	m_copy (cbk_array + index, cbk_array + num_managed - 1,
		sizeof *cbk_array);
    }
    --num_managed;
    return (info);
}   /*  End Function _dm_native_unmanage_entry  */

static flag _dm_native_read_entry (int fd, struct managed_fd_type *entry)
/*  [PURPOSE] This routine will read a descriptor (if appropriate) until all
    data is drained from the descriptor.
    <fd> The file descriptor.
    <entry> The descriptor entry.
    [RETURNS] TRUE if the descriptor is to remain managed, else FALSE.
*/
{
    char drain_buffer;
    int bytes_readable;
    static char function_name[] = "_dm_native_read_entry";

    if (entry->close_func == NULL)
    {
	if (entry->input_func == NULL)
	{
	    fprintf (stderr,
		     "Input on asynchronous descriptor but no callback\n");
	    a_prog_bug (function_name);
	}
	return (*entry->input_func) (fd, &entry->info);
    }
    /*  Can test for input/ closure  */
    if ( ( bytes_readable = r_get_bytes_readable (fd) ) < 0 )
    {
	fprintf (stderr, "Error getting bytes readable\t%s\n", ERRSTRING);
	a_prog_bug (function_name);
    }
    if (bytes_readable < 1)
    {
	/*  Close occurred  */
	/*  Drain data (close event)  */
	switch ( r_read (fd, &drain_buffer, 1) )
	{
	  case -1:
	    /*  Error  */
	    fprintf (stderr, "%s: error draining descriptor: %d\t%s\n",
		     function_name, fd, ERRSTRING);
	    break;
	  case 0:
	    /*  This is what we expect  */
	    break;
	  default:
	    fprintf (stderr, "Data readable on closed descriptor: %d\n", fd);
#ifdef OS_IRIX
	    fputs ("There is a known bug in IRIX 6.2 that can cause this: contact SGI!\n",
		   stderr);
#endif
#ifdef OS_Solaris
	    fputs ("There is a known bug in Solaris 2.x that can cause this: contact Sun!\n",
		   stderr);
#endif
	    a_prog_bug (function_name);
	    break;
	}
	return (FALSE);
    }
    /*  Input event occurred  */
    if (entry->input_func == NULL)
    {
	fprintf (stderr, "%d bytes input on descriptor but no callback\n",
		 bytes_readable);
	a_prog_bug (function_name);
    }
    while (bytes_readable > 0)
    {
	/*  More input  */
	if ( !(*entry->input_func) (fd, &entry->info) )
	{
	    /*  Descriptor to be unmanaged and closed  */
	    return (FALSE);
	}
	if ( ( bytes_readable = r_get_bytes_readable (fd) ) < 0 )
	{
	    fprintf (stderr, "%s: error getting bytes readable\t%s\n",
		     function_name, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
    return (TRUE);
}   /*  End Function _dm_native_read_entry  */

static void _dm_native_close_entry (unsigned int index)
/*  [PURPOSE] This routine will call the registered <<close_func>> for a
    descriptor, and will then close and unmanage the descriptor.
    <index> The index into the table.
    [RETURNS] Nothing.
*/
{
    struct managed_fd_type *entry;
    struct pollfd *pollfd_ptr;
    /*static char function_name[] = "_dm_native_close_entry";*/

    entry = cbk_array + index;
    pollfd_ptr = pollfd_array + index;
    /*  Descriptor drainable  */
    if (entry->close_func != NULL)
	(*entry->close_func) (pollfd_ptr->fd, entry->info);
    if (entry->auto_close) close (pollfd_ptr->fd);
    _dm_native_unmanage_entry (index);
}   /*  End Function _dm_native_close_entry  */

static flag _dm_native_increase_table_size (unsigned int num_extra)
/*  [SUMMARY] Increase the table size so at least one more entry can be added.
    <num_extra> The number of entries to increase the table size by.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int new_size;
    struct pollfd *new_pollfd;
    struct managed_fd_type *new_cbk;

    if (num_managed + num_extra <= num_allocated) return (TRUE);
    /*  Not enough space  */
    if (num_extra < num_allocated) new_size = num_allocated * 2;
    else new_size = (num_allocated + num_extra) * 2;
    if (new_size < MIN_TABLE_SIZE) new_size = MIN_TABLE_SIZE;
    if ( ( new_pollfd = m_alloc (new_size * sizeof *new_pollfd) ) == NULL )
	return (FALSE);
    if ( ( new_cbk = m_alloc (new_size * sizeof *new_cbk) ) == NULL )
    {
	m_free (new_pollfd);
	return (FALSE);
    }
    if (pollfd_array != NULL)
    {
	m_copy (new_pollfd, pollfd_array, num_managed * sizeof *new_pollfd);
	m_free (pollfd_array);
    }
    pollfd_array = new_pollfd;
    if (cbk_array != NULL)
    {
	m_copy (new_cbk, cbk_array, num_managed * sizeof *new_cbk);
	m_free (cbk_array);
    }
    cbk_array = new_cbk;
    num_allocated = new_size;
    return (TRUE);
}   /*  End Function _dm_native_increase_table_size  */

static void _dm_native_add_to_unmanage_list (unsigned int index)
/*  [SUMMARY] Add an entry to the unmanage list.
    <index> The entry to add.
    [RETURNS] Nothing.
*/
{
    unsigned int new_size;
    unsigned int *new_list;
    static char function_name[] = "_dm_native_add_to_unmanage_list";

    if (num_to_unmanage + 1 > unmanage_list_size)
    {
	new_size = unmanage_list_size * 2;
	if (new_size < MIN_TABLE_SIZE) new_size = MIN_TABLE_SIZE;
	if ( ( new_list = m_alloc (new_size * sizeof *new_list) ) == NULL )
	{
	    m_abort (function_name, "unmanage list");
	}
	unmanage_list_size = new_size;
	if (unmanage_list != NULL)
	{
	    m_copy (new_list, unmanage_list,
		    num_to_unmanage * sizeof *new_list);
	    m_free (unmanage_list);
	}
	unmanage_list = new_list;
    }
    unmanage_list[num_to_unmanage] = index;
    pollfd_array[index].revents = 0;
    ++num_to_unmanage;
}   /*  End Function _dm_native_add_to_unmanage_list  */


/*  Emulation functions follow  */

#ifndef HAS_POLL

#  ifdef HAS_SOCKETS

static int poll (struct pollfd *ufds, unsigned int nfds, int timeout)
/*  [SUMMARY] Emulate poll(2) using select(2).
    [MT-LEVEL] Safe.
*/
{
    int count, num_ready;
    int max_fd = 0;
    fd_set input_fds, output_fds, exception_fds;
    struct timeval tv;
    struct timeval *timeout_ptr;
    struct pollfd *fds_ptr;

    FD_ZERO (&input_fds);
    FD_ZERO (&output_fds);
    FD_ZERO (&exception_fds);
    for (count = 0, fds_ptr = ufds; count < nfds; ++count, ++fds_ptr)
    {
	if (fds_ptr->fd < 0) continue;
	if (fds_ptr->events & POLLIN) FD_SET (fds_ptr->fd, &input_fds);
	if (fds_ptr->events & POLLOUT) FD_SET (fds_ptr->fd, &output_fds);
	if (fds_ptr->events & POLLPRI) FD_SET (fds_ptr->fd, &exception_fds);
	if ( (fds_ptr->fd > max_fd) &&
	     ( fds_ptr->events & (POLLIN | POLLOUT | POLLPRI) ) )
	    max_fd = fds_ptr->fd;
    }
    /*  Now call  select(2)  routine  */
    if (timeout < 0)
    {
	/*  Wait forever for something to happen  */
	timeout_ptr = NULL;
    }
    else
    {
	/*  Block for a finite time  */
	timeout_ptr = &tv;
	tv.tv_sec = (unsigned long) timeout / 1000;
	timeout -= (long) (tv.tv_sec * 1000);
	tv.tv_usec = (unsigned long) timeout * 1000;
    }
    /*  Note that some systems return the number of FDs that have *any* bits
	set (i.e. any activity) while others return the number of bits set.
	Because of this, the return value cannot be properly utilised  */
    num_ready = select (max_fd + 1, &input_fds, &output_fds, &exception_fds,
			timeout_ptr);
    if (num_ready < 1) return (num_ready);
    num_ready = 0;
    for (count = 0, fds_ptr = ufds; count < nfds; ++count, ++fds_ptr)
    {
	fds_ptr->revents = 0;
	if (fds_ptr->fd < 0) continue;
	if ( FD_ISSET (fds_ptr->fd, &input_fds) ) fds_ptr->revents |= POLLIN;
	if ( FD_ISSET (fds_ptr->fd, &output_fds) ) fds_ptr->revents |= POLLOUT;
	if ( FD_ISSET (fds_ptr->fd, &exception_fds) )
	    fds_ptr->revents |= POLLPRI;
	if (fds_ptr->revents != 0) ++num_ready;
    }
    return (num_ready);
}   /*  End Function poll  */

#  else  /*  HAS_SOCKETS  */

static int poll (struct pollfd *ufds, unsigned int nfds, int timeout)
/*  [SUMMARY] Emulate poll(2) using [<r_test_input_event>]
*/
{
    int count;
    int num_ready = 0;
    struct pollfd *fds_ptr;

    for (count = 0, fds_ptr = ufds; count < nfds; ++count, ++fds_ptr)
    {
	fds_ptr->revents = 0;
	if (fds_ptr->fd < 0) continue;
	if (fds_ptr->events & POLLIN)
	{
	    if ( r_test_input_event (entry->fd) )
	    {
		fds_ptr->revents |= POLLIN;
		++num_ready;
	    }
	}
    }
#    ifdef OS_VXMVX
    if ( (num_ready < 1) && (timeout > 0) )
    {
	task_sleep (timeout);
    }
#    endif  /*  OS_VXMVX  */
    return (num_ready);
}   /*  End Function poll  */

#  endif  /*  !HAS_SOCKETS  */

#endif  /*  !HAS_POLL  */
