/*LINTLIBRARY*/
/*  alloc.c

    This code provides data structure allocation routines.

    Copyright (C) 1992-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for allocating data
  structures for the general data structure supported in Karma.


    Written by      Richard Gooch   13-SEP-1992

    Updated by      Richard Gooch   3-DEC-1992

    Updated by      Richard Gooch   22-DEC-1992: Fixed memory leak in functions
  ds_easy_alloc_array  ,  ds_dealloc_array_desc  and  ds_dealloc_packet  .

    Updated by      Richard Gooch   1-JAN-1993: Added support for memory mapped
  arrays.

    Updated by      Richard Gooch   3-JAN-1993: Added support for  K_ARRAY
  (padded array) and moved deallocation routines into  ds_dealloc.c

    Updated by      Richard Gooch   4-JAN-1993: Added preliminary support for
  tiled arrays.

    Updated by      Richard Gooch   5-JAN-1993: Fixed bug in  ds_alloc_array  .

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   30-JAN-1993: Added  ds_alloc_tiling_info  .

    Updated by      Richard Gooch   1-FEB-1993: Added setting of
  (*arr_desc).lengths[*]  in  ds_easy_alloc_array  .

    Updated by      Richard Gooch   9-FEB-1993: Improved diagnostics in
  ds_alloc_array  .

    Updated by      Richard Gooch   24-MAR-1993: Fixed bug in
  ds_alloc_contiguous_list  with stepping through list entries to allocate.

    Updated by      Richard Gooch   25-MAR-1993: Took account of changes to
  list_header  structure and added  ds_alloc_packet_subdata  .

    Updated by      Richard Gooch   17-MAY-1993: Added support for string data
  types.

    Updated by      Richard Gooch   10-AUG-1993: Changed initial attachment
  count for  multi_array  descriptor from one to zero to reflect recent changes
  to attachment implementation.

    Updated by      Richard Gooch   12-AUG-1993: Added
  ds_easy_alloc_n_element_array  .

    Updated by      Richard Gooch   25-AUG-1993: Improved documentation for
  ds_alloc_list_head  .

    Updated by      Richard Gooch   20-MAY-1994: Added  CONST  declarations
  where appropriate.

    Updated by      Richard Gooch   14-JUN-1994: Fixed syntax error in
  ds_alloc_packet  which caused problems on some platforms.

    Updated by      Richard Gooch   31-JUN-1994: Wrote in magic number in
  ds_alloc_multi  .

    Updated by      Richard Gooch   6-AUG-1994: Fixed bug in  ds_alloc_array
  which did not zero atomic array for Dec Alpha.

    Updated by      Richard Gooch   8-AUG-1994: Changed to  uaddr  type.

    Updated by      Richard Gooch   12-OCT-1994: Created
  ds_wrap_preallocated_n_element_array  routine.

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/ds/alloc.c

    Updated by      Richard Gooch   8-DEC-1994: Created
  ds_easy_alloc_array_desc  routine.

    Updated by      Richard Gooch   17-JAN-1995: Changed to  destroy_callbacks
  field for  multi_array  structure.

    Updated by      Richard Gooch   23-FEB-1995: Made use of CAN_ZERO_CLEAR.

    Updated by      Richard Gooch   19-APR-1995: Cleaned some code.

    Updated by      Richard Gooch   4-AUG-1995: Added more tests for misaligned
  data.

    Updated by      Richard Gooch   7-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   3-JUN-1996: Took account of new fields in
  dimension descriptor for first and last co-ordinate.

    Updated by      Richard Gooch   14-JUN-1996: Changed more pointers to
  CONST.

    Updated by      Richard Gooch   28-JUN-1996: Changed more pointers to
  CONST.

    Updated by      Richard Gooch   5-AUG-1996: Initialised new reference field
  for dimension descriptor.

    Updated by      Richard Gooch   15-AUG-1996: No longer initialise
  <<reference>> field of dimension descriptor.

    Updated by      Richard Gooch   29-AUG-1996: Allowed
  <ds_easy_alloc_array_desc> to accept mixture of dimension names and NULLs.

    Updated by      Richard Gooch   14-SEP-1996: Removed tiling warning message
  in <ds_alloc_array_desc>.

    Updated by      Richard Gooch   1-OCT-1996: Made <ds_easy_alloc_array_desc>
  a little more flexible by not requiring packet information.

    Updated by      Richard Gooch   20-FEB-1998: Switched to array_pointer type

    Updated by      Richard Gooch   22-FEB-1998: Moved easy array allocation
  functions to easy_alloc.c

    Updated by      Richard Gooch   24-FEB-1998: Switch to new array_pointer
  which has free function.

    Last updated by Richard Gooch   15-OCT-2001: Only test for <<data_bytes>>
  in <ds_alloc_array> if allocation needed.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#include <os.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_c.h>
#include <karma_j.h>


/*PUBLIC_FUNCTION*/
multi_array *ds_alloc_multi (unsigned int num_arrays)
/*  [SUMMARY] Allocate a multi_array descriptor.
    [PURPOSE] This routine will allocate a multi_array descriptor. The memory
    for the array of pointers to the array names, headers and data arrays is
    also allocated.
    <num_arrays> The number of independent general data structures arrays.
    [RETURNS] A pointer to the header on success, else NULL.
*/
{
    unsigned int count;
    multi_array *new;
    static char function_name[] = "ds_alloc_multi";

    if ( ( new = m_alloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "multi_desc header");
        return (NULL);
    }
    new->num_arrays = num_arrays;
    new->array_names = NULL;
    new->first_hist = NULL;
    new->last_hist = NULL;
    new->destroy_callbacks = NULL;
    new->attachments = 0;
    new->jlist = NULL;
    if (num_arrays < 1)
    {
	new->headers = NULL;
        new->data = NULL;
	new->jlists = NULL;
        return (new);
    }
    new->magic_number = MAGIC_MULTI_ARRAY;
    if ( ( new->headers = m_alloc (sizeof *new->headers * num_arrays) )
	 == NULL )
    {
	ds_dealloc_multi (new);
        m_error_notify (function_name, "multi_desc packet descriptors");
        return (NULL);
    }
    if ( ( new->data = m_alloc (sizeof *new->data * num_arrays) ) == NULL )
    {
	ds_dealloc_multi (new);
        m_error_notify (function_name, "array of data pointers");
        return (NULL);
    }
    if ( ( new->jlist = j_create (NULL, strcmp, ( void *(*) () ) NULL,
				  ( void (*) () ) NULL) ) == NULL )
    {
	ds_dealloc_multi (new);
        m_error_notify (function_name, "main joined list");
        return (NULL);
    }
    if ( ( new->jlists = m_alloc (num_arrays * sizeof *new->jlists) ) == NULL )
    {
	ds_dealloc_multi (new);
        m_error_notify (function_name, "array of joined lists");
        return (NULL);
    }
    if (num_arrays > 1)
    {
	/*  Many arrays: allocate memory for names  */
        if ( ( new->array_names =
	       m_alloc (sizeof *new->array_names * num_arrays) )
	     == NULL )
        {
	    ds_dealloc_multi (new);
            m_error_notify (function_name,
			    "array of pointers for array names");
            return (NULL);
        }
    }
    for (count = 0; count < num_arrays; ++count)
    {
	if (num_arrays > 1)
        {
	    new->array_names[count] = NULL;
        }
        new->headers[count] = NULL;
        new->data[count] = NULL;
	if ( ( new->jlists[count] =
	       j_create (NULL, strcmp, ( void *(*) () ) NULL,
			 ( void (*) () ) NULL) ) == NULL )
	{
	    ds_dealloc_multi (new);
	    m_error_notify (function_name, "main joined list");
	    return (NULL);
	}
    }
    return (new);
}   /*  End Function ds_alloc_multi  */

/*PUBLIC_FUNCTION*/
packet_desc *ds_alloc_packet_desc (unsigned int num_elem)
/*  [SUMMARY] Allocate a packet descriptor.
    [PURPOSE] This routine will allocate a packet descriptor. The memory for
    the array of element types and descriptors is also allocated.
    <num_elem> The number of elements to allocate.
    [RETURNS] A pointer to the descriptor on success, else NULL.
*/
{
    unsigned int element_count = 0;
    packet_desc *return_value;
    static char function_name[] = "ds_alloc_packet_desc";

    if ( ( return_value = m_alloc (sizeof *return_value) ) == NULL )
    {
	m_error_notify (function_name, "pack_descriptor");
        return (NULL);
    }
    if ( ( return_value->element_types =
	   m_alloc (sizeof *return_value->element_types * num_elem) ) == NULL )
    {
	m_free (return_value);
        m_error_notify (function_name, "array of element types");
        return (NULL);
    }
    if ( ( return_value->element_desc =
	  m_alloc (sizeof *return_value->element_desc * num_elem) ) == NULL )
    {
	m_free (return_value->element_types);
        m_free (return_value);
        m_error_notify(function_name,"array of element descriptor pointers");
        return (NULL);
    }
    return_value->num_elements = num_elem;
    while (element_count < num_elem)
    {
	return_value->element_types[element_count] = NONE;
        return_value->element_desc[element_count++] = NULL;
    }
    return (return_value);
}   /*  End Function ds_alloc_packet_desc  */

/*PUBLIC_FUNCTION*/
char *ds_alloc_data (packet_desc *pack_desc, flag clear, flag array_alloc)
/*  [SUMMARY] Allocate packet data recursively.
    [PURPOSE] This routine will allocate all memory required to store data in a
    packet. The routine will recursively allocate space for packets, sub arrays
    of packets and linked list headers.
    The routine is quite robust, cleanly bypassing missing sections of the
    descriptor hierarchy, and deallocating any memory allocated after an
    error occurs.
    <pack_desc> The packet descriptor.
    <clear> If TRUE, then the routine will initialise (set to zero) all the
    data.
    <array_alloc> If FALSE, and an array is an atomic array, then instances of
    that array will NOT be allocated, and a NULL pointer will be written into
    the parent element, else the array will be allocated.
    [RETURNS] A pointer to the data memory if all memory could be allocated,
    else NULL.
*/
{
    char *return_value;
    static char function_name[] = "ds_alloc_data";

    FLAG_VERIFY (clear);
    FLAG_VERIFY (array_alloc);
    if (pack_desc == NULL)
    {
	return (NULL);
    }
    if ( ( return_value = ds_alloc_packet (pack_desc) ) == NULL )
    {
	m_error_notify (function_name, "packet");
        return (NULL);
    }
    if ( !ds_alloc_packet_subdata (pack_desc, return_value, clear,
				   array_alloc) )
    {
	m_free (return_value);
	m_error_notify (function_name, "packet subdata");
	return (NULL);
    }
    return (return_value);
}   /*  End Function ds_alloc_data  */

/*PUBLIC_FUNCTION*/
flag ds_alloc_packet_subdata (CONST packet_desc *pack_desc, char *packet,
			      flag clear, flag array_alloc)
/*  [SUMMARY] Allocate packet sub-data.
    [PURPOSE] This routine will recursively allocate space for sub arrays of
    packets and linked list headers for a packet. The data space for the packet
    is NOT allocated, it must be supplied.
    The routine is quite robust, cleanly bypassing missing sections of the
    descriptor hierarchy, and deallocating any memory allocated after an
    error occurs.
    <pack_desc> The packet descriptor.
    <packet> The packet data.
    <clear> If TRUE, then the routine will initialise (set to zero) all the
    sub-structure data.
    <array_alloc> If FALSE, and an array is an atomic array, then instances of
    that array will NOT be allocated, and a NULL pointer will be written into
    the parent element, else the array will be allocated.
    [RETURNS] TRUE if all memory could be allocated, else FALSE.
*/
{
    unsigned int elem_count;
    unsigned int type;
    char *element;
    list_header *head;
    static char function_name[] = "ds_alloc_packet_subdata";

    FLAG_VERIFY (clear);
    FLAG_VERIFY (array_alloc);
    if (pack_desc == NULL) return (FALSE);
    if (packet == NULL) return (FALSE);
    for (elem_count = 0; elem_count < pack_desc->num_elements; ++elem_count)
    {
	type = pack_desc->element_types[elem_count];
	if ( !ds_element_is_named (type) )
	{
	    /*  Recursive data type  */
	    element = packet + ds_get_element_offset (pack_desc, elem_count);
	    switch (type)
	    {
	      case K_ARRAY:
		/*  Array of packets to allocate memory for  */
		if (pack_desc->element_desc[elem_count] == NULL) break;
		if ( !ds_alloc_array ( (array_desc *)
				       pack_desc->element_desc[elem_count],
				       element, clear, array_alloc ) )
		{
		    ds_dealloc_packet_subdata (pack_desc, packet);
		    m_error_notify (function_name, "array of packets");
		    return (FALSE);
		}
		break;
	      case LISTP:
		/*  Linked list: allocate header  */
		if ( ( head = ds_alloc_list_head () ) == NULL )
		{
		    ds_dealloc_packet_subdata (pack_desc, packet);
		    m_error_notify (function_name, "linked list header");
		    return (FALSE);
		}
#ifdef NEED_ALIGNED_DATA
		if ( IS_ALIGNED (element, sizeof head) )
		{
		    *(list_header **) element = head;
		}
		else m_copy (element, &head, sizeof head);
#else
		*(list_header **) element = head;
#endif
		break;
	      default:
		/*  Bad data type  */
		fprintf (stderr, "Bad data type: %u\n",
			 pack_desc->element_types[elem_count]);
		a_prog_bug (function_name);
		break;
	    }
	}
    }
    return (TRUE);
}   /*  End Function ds_alloc_packet_subdata  */

/*PUBLIC_FUNCTION*/
char *ds_alloc_packet (packet_desc *pack_desc)
/*  [SUMMARY] Allocate packet.
    [PURPOSE] This routine will allocate memory for a packet. This routine is
    NOT recursive (i.e. sub arrays and linked lists are not allocated).
    The elements of the packet will be set to zero (for all types).
    <pack_desc> The packet descriptor.
    [RETURNS] A pointer to the packet on success, else NULL.
*/
{
    unsigned int packet_size;
#ifndef CAN_ZERO_CLEAR
    FString *fstring;
    unsigned int element_count;
    char *element;
    extern char host_type_sizes[NUMTYPES];
#endif
    char *packet;
    static char function_name[] = "ds_alloc_packet";

    packet_size = ds_get_packet_size (pack_desc);
    if (packet_size < 1)
    {
	/*  Zero sized packet: don't bother  */
        return (NULL);
    }
    if ( ( packet = m_alloc (packet_size) ) == NULL )
    {
	/*  Packet could not be allocated  */
        m_error_notify (function_name, "packet");
        return (NULL);
    }
#ifdef CAN_ZERO_CLEAR
    m_clear (packet, packet_size);
#else
    element = packet;
    /*  Zero packet  */
    for (element_count = 0; element_count < pack_desc->num_elements;
	 ++element_count)
    {
	/*  Zero element  */
        switch (pack_desc->element_types[element_count])
        {
	  case K_FLOAT:
	    *( (float *) element ) = 0.0;
	    break;
	  case K_DOUBLE:
	    *( (double *) element ) = 0.0;
	    break;
	  case K_BYTE:
	    *element = NULL;
	    break;
	  case K_INT:
	    *( (int *) element ) = NULL;
	    break;
	  case K_SHORT:
	    *( (short *) element ) = NULL;
	    break;
	  case K_ARRAY:
	    m_clear ( element, sizeof (array_pointer) );
	    break;
	  case LISTP:
	    *( (char **) element ) = NULL;
	    break;
	  case K_COMPLEX:
	    *( (float *) element ) = 0.0;
	    *( (float *) element + 1 ) = 0.0;
	    break;
	  case K_DCOMPLEX:
	    *( (double *) element ) = 0.0;
	    *( (double *) element + 1 ) = 0.0;
	    break;
	  case K_BCOMPLEX:
	    *element = NULL;
	    *(element + 1) = NULL;
	    break;
	  case K_ICOMPLEX:
	    *( (int *) element ) = NULL;
	    *( (int *) element + 1 ) = NULL;
	    break;
	  case K_SCOMPLEX:
	    *( (short *) element ) = NULL;
	    *( (short *) element + 1 ) = NULL;
	    break;
	  case K_LONG:
	    *( (long *) element ) = NULL;
	    break;
	  case K_LCOMPLEX:
	    *( (long *) element ) = NULL;
	    *( (long *) element + 1 ) = NULL;
	    break;
	  case K_UBYTE:
	    *( (unsigned char *) element ) = NULL;
	    break;
	  case K_UINT:
	    *( (unsigned int *) element ) = NULL;
	    break;
	  case K_USHORT:
	    *( (unsigned short *) element ) = NULL;
	    break;
	  case K_ULONG:
	    *( (unsigned long *) element ) = NULL;
	    break;
	  case K_UBCOMPLEX:
	    *( (unsigned char *) element ) = NULL;
	    *( (unsigned char *) element + 1 ) = NULL;
	    break;
	  case K_UICOMPLEX:
	    *( (unsigned int *) element ) = NULL;
	    *( (unsigned int *) element + 1 ) = NULL;
	    break;
	  case K_USCOMPLEX:
	    *( (unsigned short *) element ) = NULL;
	    *( (unsigned short *) element + 1 ) = NULL;
	    break;
	  case K_ULCOMPLEX:
	    *( (unsigned long *) element ) = NULL;
	    *( (unsigned long *) element + 1 ) = NULL;
	    break;
	  case K_VSTRING:
	    *(char **) element = NULL;
	    break;
	  case K_FSTRING:
	    fstring = (FString *) element;
	    fstring->max_len = 0;
	    fstring->string = NULL;
	    break;
	  default:
	    *element = NULL;
        }
        element += host_type_sizes[pack_desc->element_types[element_count] ];
    }
#endif
    return (packet);
}   /*  End Function ds_alloc_packet  */

/*PUBLIC_FUNCTION*/
array_desc *ds_alloc_array_desc (unsigned int num_dimensions,
				 unsigned int num_levels)
/*  [SUMMARY] Allocate array descriptor.
    [PURPOSE] This function will allocate a header for a multi-dimensional
    tiled array of data packets.
    <num_dimensions> The number of dimensions in the array.
    <num_levels> The number of levels of tiling. If this is 0, the array is not
    tiled.
    [NOTE] Note that the dimension descriptors are not allocated, however, the
    array of pointers for them is allocated.
    [RETURNS] A pointer to the descriptor on success, else NULL.
*/
{
    unsigned int dim_count;
    array_desc *return_value;
    static char function_name[] = "ds_alloc_array_desc";

    if (num_dimensions < 1)
    {
	fprintf (stderr, "Cannot allocate zero-dimensioned array\n");
	a_prog_bug (function_name);
    }
    /*  Allocate descriptor  */
    if ( ( return_value = m_alloc (sizeof *return_value) ) == NULL )
    {
	m_error_notify (function_name, "array descriptor");
        return (NULL);
    }
    return_value->num_dimensions = num_dimensions;
    return_value->num_levels = 0;
    /*  Alocate array of dimension descriptor pointers  */
    if ( ( return_value->dimensions =
	  m_alloc (sizeof *return_value->dimensions * num_dimensions) )
	 == NULL )
    {
	m_free (return_value);
        m_error_notify (function_name,
			"array of dimension descriptor pointers");
	fprintf (stderr, "  num_dimensions: %u\n", num_dimensions);
        return (NULL);
    }
    return_value->tile_lengths = NULL;
    /*  Allocate array of lengths  */
    if ( ( return_value->lengths =
	   m_alloc (sizeof *return_value->lengths * num_dimensions) )
	 == NULL )
    {
	m_error_notify (function_name, "array of lengths");
	m_free (return_value->dimensions);
	m_free (return_value);
	return (NULL);
    }
    for (dim_count = 0; dim_count < num_dimensions; ++dim_count)
    {
	return_value->dimensions[dim_count] = NULL;
	return_value->lengths[dim_count] = 0;
    }
    if (ds_alloc_tiling_info (return_value, num_levels) != TRUE)
    {
	m_error_notify (function_name, "tiling information");
	m_free (return_value->dimensions);
	m_free (return_value);
	return (NULL);
    }
    return_value->packet = NULL;
    return_value->offsets = NULL;
    return_value->padded = 2 * abs (TRUE) + 2 * abs (FALSE);
    return (return_value);
}   /*  End Function ds_alloc_array_desc  */

/*PUBLIC_FUNCTION*/
flag ds_alloc_tiling_info (array_desc *arr_desc, unsigned int num_levels)
/*  [SUMMARY] Allocate array tiling information.
    [PURPOSE] This routine will allocate tiling information for an array
    descriptor which does not have any existing tiling information.
    <arr_desc> The array descriptor.
    <num_levels> The number of levels of tiling. If this is 0, the array is not
    tiled.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int dim_count;
    static char function_name[] = "ds_alloc_tiling_info";

    if (arr_desc->num_levels > 0)
    {
	fprintf (stderr, "Existing tiling information\n");
	a_prog_bug (function_name);
    }
    
    if (num_levels < 1)
    {
	return (TRUE);
    }
    /*  Allocate array of tile length pointers  */
    if ( ( arr_desc->tile_lengths =
	  m_alloc (sizeof *arr_desc->tile_lengths *
		   arr_desc->num_dimensions) )
	== NULL )
    {
	m_error_notify (function_name, "array of tile length pointers");
	return (FALSE);
    }
    for (dim_count = 0; dim_count < arr_desc->num_dimensions; ++dim_count)
    {
	/*  Allocate array of tile lengths for this dimension  */
	if ( ( arr_desc->tile_lengths[dim_count] =
	       m_alloc (sizeof **arr_desc->tile_lengths * num_levels) )
	     == NULL )
	{
	    /*  Abort here because I can't be bothered cleaning up  */
	    m_abort (function_name, "array of tile lengths");
	}
	m_clear (arr_desc->tile_lengths[dim_count],
		 sizeof **arr_desc->tile_lengths *num_levels);
    }
    arr_desc->num_levels = num_levels;
    return (TRUE);
}   /*  End Function ds_alloc_tiling_info  */

/*PUBLIC_FUNCTION*/
dim_desc *ds_alloc_dim_desc (CONST char *dim_name, uaddr length,
			     double first, double last, flag regular)
/*  [SUMMARY] Allocate a dimension descriptor.
    <dim_name> The name of the dimension. The name is copied, thus the input
    character string may be subsequently deallocated.
    <length> The length (number of co-ordinates) of the dimension.
    <first> The first dimension co-ordinate.
    <last> The last dimension co-ordinate.
    <regular> If TRUE the co-ordinates of the dimension are regularly spaced,
    else the co-ordinates are not regularly spaced. If the co-ordinates are
    regularly spaced the <<coordinates>> field of the array descriptor will be
    set to NULL, else it will point to an array of doubles of length <<length>>
    These co-ordinates will be set to 0.0.
    [RETURNS] A pointer to the descriptor on success, else NULL.
*/
{
    unsigned int count;
    dim_desc *return_value;
    static char function_name[] = "ds_alloc_dim_desc";

    if (dim_name == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    if (length < 1)
    {
	fprintf (stderr, "Illegal dimension length: %lu passed\n",
			length);
	a_prog_bug (function_name);
    }
    if ( ( return_value = m_alloc (sizeof *return_value) ) == NULL )
    {
	m_error_notify (function_name, "dimension_descriptor");
        return (NULL);
    }
    if ( ( return_value->name = m_dup (dim_name, strlen (dim_name) + 1) )
	 == NULL )
    {
	m_free (return_value);
        m_error_notify (function_name, "dimension name");
        return (NULL);
    }
    return_value->length = length;
    return_value->first_coord = first;
    return_value->last_coord = last;
    if (first < last)
    {
	return_value->minimum = first;
	return_value->maximum = last;
    }
    else
    {
	return_value->minimum = last;
	return_value->maximum = first;
    }
    if ( regular || (length < 1) )
    {
	return_value->coordinates = NULL;
        return (return_value);
    }
    if ( ( return_value->coordinates =
	   m_alloc (sizeof *return_value->coordinates * length) )
	 == NULL )
    {
	m_free (return_value->name);
        m_free (return_value);
        m_error_notify (function_name, "coordinate array");
        return (NULL);
    }
    for (count = 0; count < length; ++count)
    {
	return_value->coordinates[count] = 0.0;
    }
    return (return_value);
}   /*  End Function ds_alloc_dim_desc */

/*PUBLIC_FUNCTION*/
list_header *ds_alloc_list_head ()
/*  [SUMMARY] Allocate a linked list header.
    [PURPOSE] This routine will allocate a linked list header.
    The length of the linked list specified in the header will be 0, and the
    <<contiguous_until>> entry will be set to 0.
    The <<sort_type>> field in the header will be set to SORT_UNDEF. This MUST
    be set to some other value (i.e. SORT_RANDOM) prior to use with other
    library routines.
    The <<list_start>> and <<list_end>> pointers will be set to NULL.
    [RETURNS] A pointer to the header on success, else NULL.
*/
{
    list_header *return_value;
    static char function_name[] = "ds_alloc_list_head";

    if ( ( return_value = m_alloc (sizeof *return_value) ) == NULL )
    {
	m_error_notify (function_name, "linked list header");
        return (NULL);
    }
    return_value->magic = MAGIC_LIST_HEADER;
    return_value->length = 0;
    return_value->contiguous_length = 0;
    return_value->sort_type = SORT_UNDEF;
    return_value->sort_elem_num = 0;
    return_value->contiguous_data = NULL;
    return_value->first_frag_entry = NULL;
    return_value->last_frag_entry = NULL;
    return (return_value);
}   /*  End Function ds_alloc_list_head  */

/*PUBLIC_FUNCTION*/
list_entry *ds_alloc_list_entry (packet_desc *list_desc, flag array_alloc)
/*  [SUMMARY] Allocate a linked list entry.
    [PURPOSE] This routine will allocate an entry in a linked list (it will
    NOT insert it in the list: see [<ds_list_insert>]).
    The list pointers will be set to NULL.
    The routine will initialise (set to zero) the data in the entry.
    The routine will recursively allocate memory for sub arrays and linked
    lists.
    <list_desc> The descriptor for the linked list.
    <array_alloc> If FALSE, and an array is an atomic array, then instances of
    that array will NOT be allocated, and a NULL pointer will be written into
    the parent element, else the array will be allocated.
    [RETURNS] A pointer to the entry on success, else NULL.
*/
{
    list_entry *return_value;
    static char function_name[] = "ds_alloc_list_entry";

    FLAG_VERIFY (array_alloc);
    if ( ( return_value = m_alloc (sizeof *return_value) ) == NULL )
    {
	m_error_notify (function_name, "linked list entry");
        return (NULL);
    }
    return_value->prev = NULL;
    return_value->next = NULL;
    if (ds_get_packet_size (list_desc) < 1)
    {
	return_value->data = NULL;
        return (return_value);
    }
    if ( ( return_value->data = ds_alloc_data (list_desc, TRUE,
					       array_alloc) ) == NULL )
    {
	m_free (return_value);
	m_error_notify (function_name, "list data entry");
        return (NULL);
    }
    return (return_value);
}   /*  End Function ds_alloc_list_entry  */

/*PUBLIC_FUNCTION*/
flag ds_alloc_array (CONST array_desc *arr_desc, char *element, flag clear,
		     flag array_alloc)
/*  [SUMMARY] Allocate an array.
    [PURPOSE] This routine will allocate memory for an array. The routine will
    recursively allocate sub arrays and linked lists.
    Any memory which is allocated will be deallocated if an error occurs.
    <arr_desc> The array descriptor.
    <element> The element to write the array pointer to.
    <clear> If TRUE, then the routine will initialise (set to zero) all the
    data.
    <array_alloc> If FALSE, and an array is an atomic array, then instances of
    that array will NOT be allocated, and a NULL pointer will be written into
    the parent element, else the array will be allocated.
    [RETURNS] A pointer to the array on success, else NULL.
*/
{
    flag atomic;
    unsigned int data_bytes;
    unsigned int packet_size;
    unsigned int array_size;
    unsigned int packet_count = 0;
    array_pointer arrayp;
    char *data;
    static char function_name[] = "ds_alloc_array";

    FLAG_VERIFY (clear);
    FLAG_VERIFY (array_alloc);
    if (arr_desc == NULL)
    {
	a_func_abort (function_name, "NULL array descriptor");
	return (FALSE);
    }
    if (arr_desc->packet == NULL)
    {
	a_func_abort (function_name, "NULL packet descriptor for array");
	return (FALSE);
    }
    packet_size = ds_get_packet_size (arr_desc->packet);
    array_size = ds_get_array_size (arr_desc);
    data_bytes = packet_size * array_size;
    atomic = ds_packet_all_data (arr_desc->packet);
    m_clear (&arrayp, sizeof arrayp);
    if (atomic && !array_alloc)
    {
	/*  Array need not be allocated  */
	arrayp.array = NULL;
    }
    else
    {
	/*  Array must be allocated  */
	if (data_bytes < 1) return (FALSE);
	if ( ( arrayp.array = m_alloc (data_bytes) ) == NULL )
	{
	    m_error_notify (function_name, "array");
	    return (FALSE);
	}
	arrayp.free = ( void (*) () ) m_free;
	arrayp.info = arrayp.array;
    }
    m_copy (element, &arrayp, sizeof arrayp);
    if (atomic && !array_alloc) return (TRUE);
    if (atomic)
    {
	/*  No sub-arrays or linked lists  */
	if (!clear)
	{
	    /*  Initialisation not required  */
	    return (TRUE);
	}
	/*  Do fast array initialize  */
	/*  NOTE: this will break on platforms which don't have 0 as the zero
	    value (for say floating point numbers).  */
	m_clear (arrayp.array, data_bytes);
	return (TRUE);
    }
    /*  Initialize the array and allocate sub arrays and linked lists  */
    for (data = arrayp.array; packet_count < array_size;
	 ++packet_count, data += packet_size)
    {
	/*  Allocate a packet  */
        if ( !ds_alloc_packet_subdata (arr_desc->packet, data, clear,
				       array_alloc) )
        {
	    m_error_notify (function_name, "packet subdata");
            return (FALSE);
        }
    }
    return (TRUE);
}   /*  End Function ds_alloc_array  */

/*PUBLIC_FUNCTION*/
flag ds_alloc_contiguous_list (CONST packet_desc *list_desc,
			       list_header *list_head, unsigned int length,
			       flag clear, flag array_alloc)
/*  [SUMMARY] Allocate a contiguous list of linked list entries.
    [PURPOSE] This routine will allocate a contiguous block of linked list
    entry data  packets. No <<list_entry>> structures are allocated, they are
    implied. The routine will recursively allocate memory for sub arrays and
    linked lists. The list must be empty.
    The <<contiguous_length>> value in the list header will be set to the list
    length.
    <list_desc> The packet descriptor for the linked list.
    <list_head> The list header.
    <length> The number of list entries to allocate.
    <clear> If TRUE, then the routine will initialise (set to zero) all the
    data.
    <array_alloc> If FALSE, and an array is an atomic array, then instances of
    that array will NOT be allocated, and a NULL pointer will be written into
    the parent element, else the array will be allocated.
    [RETURNS] TRUE on success, else FALSE.
    [NOTE] On failure, the list header is deallocated.
*/
{
    unsigned int count;
    unsigned int pack_size;
    char *data;
    static char function_name[] = "ds_alloc_contiguous_list";

    FLAG_VERIFY (clear);
    FLAG_VERIFY (array_alloc);
    if ( (list_desc == NULL) || (list_head == NULL) )
    {
	return (FALSE);
    }
    if (list_head->magic != MAGIC_LIST_HEADER)
    {
	fprintf (stderr, "List header has bad magic number\n");
	a_prog_bug (function_name);
    }
    if (list_head->length > 0)
    {
	fprintf (stderr, "List has: %lu entries: must be empty!\n",
		 list_head->length);
	a_prog_bug (function_name);
    }
    if (length < 1) return (TRUE);
    pack_size = ds_get_packet_size (list_desc);
    /*  Allocate a contiguous block of list data packets  */
    if ( ( list_head->contiguous_data = m_alloc (pack_size * length) )
	 == NULL )
    {
	m_error_notify (function_name, "linked list data");
        return (FALSE);
    }
    /*  Clear block of list data packets  */
    m_clear (list_head->contiguous_data, pack_size * length);
    /*  Write in data for length, etc. into header  */
    list_head->length = length;
    list_head->contiguous_length = length;
    if ( ds_packet_all_data (list_desc) )
    {
	/*  No sub-structure to allocate for: return OK here  */
	return (TRUE);
    }
    /*  Now go through each entry's data and allocate sub-structure  */
    for (count = 0, data = list_head->contiguous_data; count < length;
	 ++count, data += pack_size)
    {
	if ( !ds_alloc_packet_subdata (list_desc, data, clear, array_alloc) )
	{
	    ds_dealloc_list_entries (list_desc, list_head);
            return (FALSE);
        }
    }
    return (TRUE);
}   /*  End Function ds_alloc_contiguous_list  */
