/*LINTLIBRARY*/
/*  centroid.c

    This code provides centroiding routines.

    Copyright (C) 1997-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for manipulating arrays
    in the general data structure supported in Karma.


    Written by      Richard Gooch   7-AUG-1997: Took code from
  source/modules/koords/event.c

    Updated by      Richard Gooch   17-SEP-1997: Fixed algorithm for finding
  value to subtrace in <find_centroid> when peak is positive: now finds minimum
  value in box rather than using 0.

    Updated by      Richard Gooch   1-DEC-1997: Added <<background>> parameter
  to <ds_find_centroid>.

    Updated by      Richard Gooch   15-FEB-1998: Estimate background by taking
  median, iteratively.

    Updated by      Richard Gooch   16-FEB-1998: Estimate background by taking
  true median, by sorting values and taking midpoint.

    Updated by      Richard Gooch   22-FEB-1998: Removed casts for memory
  functions.

    Updated by      Richard Gooch   4-MAR-1998: Fixed bug in <find_centroid>
  which did not cope with zero sum, causing divide by zero errors.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   4-JUN-1998: Changed interface to
  <ds_find_centroid>.

    Last updated by Richard Gooch   1-SEP-1999: Added code to cope with ellipse
  falling off the image.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>

#define SIZE_THRESHOLD 0.5
#define LIST_FACTOR 1.1


/*  Private structures  */
typedef struct sarray_struct  /*  Simple array  */
{
    CONST char *data;
    unsigned int type;
    uaddr xlen;
    CONST uaddr *xoffsets;
    uaddr ylen;
    CONST uaddr *yoffsets;
} *sarray;


/*  Private functions  */
STATIC_FUNCTION (double get_value, (sarray array, iaddr y, iaddr x) );
STATIC_FUNCTION (flag find_peak,
		 (sarray array, iaddr *xpos, iaddr *ypos,
		  double radius_x, double radius_y, double background) );
STATIC_FUNCTION (void walk_up_hill, (sarray array, iaddr *xpos, iaddr *ypos,
				     flag inverted) );
STATIC_FUNCTION (int find_size,
		 (sarray array, iaddr xpos, iaddr ypos, double threshold,
		  double background, flag inverted) );
STATIC_FUNCTION (void find_centroid, (sarray array, double *xpos, double *ypos,
				      double max_radius, flag inverted) );
STATIC_FUNCTION (double find_background,
		 (sarray array, double centre_x, double centre_y,
		  double inner_radius_x, double inner_radius_y,
		  double outer_radius_x, double outer_radius_y) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag ds_find_centroid (CONST char *image, unsigned int type,
		       double *background, flag *inverted,
		       uaddr xlen, CONST uaddr *xoffsets,
		       double *xpos, uaddr xradius,
		       uaddr ylen, CONST uaddr *yoffsets,
		       double *ypos, uaddr yradius)
/*  [SUMMARY] Find the centroid inside an ellipse.
    [PURPOSE] This routine will search for the centroid of an object inside a
    specified ellipse. The background (median) value is computed for an annulus
    outside the ellipse. The annulus is 10% larger than the ellipse.
    <image> The image data.
    <type> The type of the image data.
    <background> The background value is written here. This may be NULL.
    <inverted> The value TRUE will be written here if the image values are
    inverted (i.e. the object of interested is in a local minimum rather than a
    maximum).
    <xlen> The horizontal image width.
    <xoffsets> The array of horizontal byte offsets
    <xpos> The horizontal data pixel position. The original value must be a
    reasonable approximation. This is modified to the centroid position.
    <xradius> The horizontal radius of the ellipse.
    <ylen> The vertical image width.
    <yoffsets> The array of vertical byte offsets
    <ypos> The vertical data pixel position. The original value must be a
    reasonable approximation. This is modified to the centroid position.
    <yradius> The vertical radius of the ellipse.
    [NOTE] (0.0, 0.0) corresponds to the centre of the first pixel.
    [RETURNS] TRUE on succes, else FALSE.
*/
{
    iaddr ix, iy, size;
    flag linverted;
    double lbackground, value, scale, outer_radius_x, outer_radius_y;
    struct sarray_struct array;

    array.data = image;
    array.type = type;
    array.xlen = xlen;
    array.xoffsets = xoffsets;
    array.ylen = ylen;
    array.yoffsets = yoffsets;
    /*  First compute background (median)  */
    scale = ( (double) xradius + 2.0 ) / (double) xradius;
    if (scale < 1.1) scale = 1.1;
    outer_radius_x = (double) xradius * scale;
    scale = ( (double) yradius + 2.0 ) / (double) yradius;
    if (scale < 1.1) scale = 1.1;
    outer_radius_y = (double) yradius * scale;
    lbackground = find_background (&array, *xpos, *ypos, xradius, yradius,
				   outer_radius_x, outer_radius_y);
    /*  Find peak/trough in this ellipse  */
    ix = *xpos;
    iy = *ypos;
    linverted = find_peak (&array, &ix, &iy, xradius, yradius, lbackground);
    walk_up_hill (&array, &ix, &iy, linverted);
    value = get_value (&array, iy, ix);
    if (linverted) value = lbackground - value;
    else value -= lbackground;
    size = find_size (&array, ix, iy, value * SIZE_THRESHOLD, lbackground,
		      linverted);
    *xpos = ix;
    *ypos = iy;
    find_centroid (&array, xpos, ypos, size, linverted);
    if (background != NULL) *background = lbackground;
    if (inverted != NULL) *inverted = linverted;
    return (TRUE);
}   /*  End Function find_object  */


/*  Private functions follow  */

static double get_value (sarray array, iaddr y, iaddr x)
/*  [SUMMARY] Get value from a simple array.
    <array> The simple array.
    <y> The vertical pixel position.
    <x> The horizontal pixel position.
    [RETURNS] The pixel value.
*/
{
    double value[2];
    CONST char *ptr;

    if (x < 0) return (TOOBIG);
    if (x >= array->xlen) return (TOOBIG);
    if (y < 0) return (TOOBIG);
    if (y >= array->ylen) return (TOOBIG);
    ptr = array->data + array->xoffsets[x] + array->yoffsets[y];
    ds_get_element (ptr, array->type, value, NULL);
    return (value[0]);
}   /*  End Function get_value  */

static flag find_peak (sarray array, iaddr *xpos, iaddr *ypos,
		       double radius_x, double radius_y, double background)
/*  [SUMMARY] Find the peak/trough inside an ellipse.
    [PURPOSE] This routine will search for the peak/trough inside an ellipse.
    It does this by finding the value with the greatest absolute difference
    from the background value.
    <array> The simple array.
    <xpos> The horizontal pixel position. This is modified to the centroid
    position.
    <ypos> The vertical pixel position. This is modified to the centroid
    position.
    <radius_x> The horizontal radius of the ellipse.
    <radius_y> The vertical radius of the ellipse.
    <background> The background value.
    <inverted> If TRUE, the algorithms should search for a minimum, rather than
    a maximum.
    [NOTE] (0.0, 0.0) corresponds to the centre of the first pixel.
    [RETURNS] TRUE if the value is a trough, otherwise FALSE is returned for a
    peak.
*/
{
    iaddr x, y, old_x, old_y;
    double distance, tmp;
    double diff, old_diff;

    old_x = *xpos;
    old_y = *ypos;
    old_diff = fabs (get_value (array, old_y, old_x) - background);
    for (y = *ypos - (iaddr) radius_y; y < *ypos + (iaddr) radius_y; ++y)
    {
	for (x = *xpos - (iaddr) radius_x; x < *xpos + (iaddr) radius_x; ++x)
	{
	    diff = get_value (array, y, x);
	    if ( DOUBLE_IS_BLANK (diff) ) continue;
	    tmp = x - *xpos;
	    distance = tmp * tmp / (radius_x * radius_x);
	    tmp = y - *ypos;
	    distance += tmp * tmp / (radius_y * radius_y);
	    if (distance > 1.0) continue;
	    diff = fabs (diff - background);
	    if (diff <= old_diff) continue;
	    old_x = x;
	    old_y = y;
	    old_diff = diff;
	}
    }
    *xpos = old_x;
    *ypos = old_y;
    return (get_value (array, old_y, old_x) >= background) ? FALSE : TRUE;
}   /*  End Function find_peak  */

static void walk_up_hill (sarray array, iaddr *xpos, iaddr *ypos,flag inverted)
/*  [SUMMARY] Find a nearby peak.
    [PURPOSE] This routine will search for the peak of the nearest object
    to a specified pixel position. It does this by walking "up the hill" from
    an initial point until it can go no higher.
    <array> The simple array.
    <xpos> The horizontal pixel position. This is modified to the centroid
    position.
    <ypos> The vertical pixel position. This is modified to the centroid
    position.
    <inverted> If TRUE, the algorithms should search for a minimum, rather than
    a maximum.
    [NOTE] (0.0, 0.0) corresponds to the centre of the first pixel.
    [RETURNS] Nothing.
*/
{
    flag keep_going = TRUE;
    iaddr x, y, old_x, old_y, count;
    double value, old_value, tmp;
    iaddr xlist[8], ylist[8];

    /*  Set initial position  */
    old_x = *xpos;
    old_y = *ypos;
    old_value = get_value (array, old_y, old_x);
    while (keep_going)
    {
	x = old_x;
	y = old_y;
	value = old_value;
	xlist[0] = old_x - 1;
	ylist[0] = old_y - 1;
	xlist[1] = old_x;
	ylist[1] = old_y - 1;
	xlist[2] = old_x + 1;
	ylist[2] = old_y - 1;
	xlist[3] = old_x + 1;
	ylist[3] = old_y;
	xlist[4] = old_x + 1;
	ylist[4] = old_y + 1;
	xlist[5] = old_x;
	ylist[5] = old_y + 1;
	xlist[6] = old_x - 1;
	ylist[6] = old_y + 1;
	xlist[7] = old_x - 1;
	ylist[7] = old_y;
	for (count = 0; count < 8; ++count)
	{
	    tmp = get_value (array, ylist[count], xlist[count]);
	    if ( DOUBLE_IS_BLANK (tmp) ) continue;
	    if (inverted)
	    {
		if (tmp < value)
		{
		    x = xlist[count];
		    y = ylist[count];
		    value = tmp;
		}
	    }
	    else
	    {
		if (tmp > value)
		{
		    x = xlist[count];
		    y = ylist[count];
		    value = tmp;
		}
	    }
	}
	if (value > old_value)
	{
	    old_x = x;
	    old_y = y;
	    old_value = value;
	}
	else keep_going = FALSE;
    }
    *xpos = old_x;
    *ypos = old_y;
}   /*  End Function walk_up_hill  */

static int find_size (sarray array, iaddr xpos, iaddr ypos, double threshold,
		      double background, flag inverted)
/*  [SUMMARY] Find radial size of an object.
    [PURPOSE] This routine will find the size of an object. It does this by
    constructing circles of ever increasing radius and checking if all values
    just outside that circle lie below a threshold. If all values just outside
    do fall below the threshold, the radius of that circle is deemed to be the
    radial size of the object.
    <array> The simple array.
    <xpos> The horizontal pixel position.
    <ypos> The vertical pixel position.
    <threshold> The threshold value.
    <background> The average background value.
    <inverted> If TRUE, the algorithms should search for a minimum, rather than
    a maximum.
    [NOTE] (0.0, 0.0) corresponds to the centre of the first pixel.
    [RETURNS] The radial size.
*/
{
    flag keep_going = TRUE;
    iaddr radius, sq_radius, radiusP1, sq_radiusP1;
    iaddr x, y, sq_distance;
    double value;
    /*double sum, nval;*/

    for (radius = 1; keep_going; ++radius)
    {
	sq_radius = radius * radius;
	radiusP1 = radius + 1;
	sq_radiusP1 = radiusP1 * radiusP1;
	keep_going = FALSE;
	/*  Scan through a box just bigger than the circle  */
	for (y = ypos - radiusP1; y < ypos + radiusP1; ++y)
	{
	    for (x = xpos - radiusP1; x < xpos + radiusP1; ++x)
	    {
		sq_distance = (x - xpos) * (x - xpos) + (y - ypos) * (y -ypos);
		if (sq_distance < sq_radius) continue;
		if (sq_distance > sq_radiusP1) continue;
		/*  Should be just outside the circle  */
		value = get_value (array, y, x);
		if ( DOUBLE_IS_VALID (value) )
		{
		    if (inverted) value = background - value;
		    else value -= background;
		    if (value > threshold) keep_going = TRUE;
		}
	    }
	}
	/*if (sum / nval > threshold) keep_going = TRUE;*/
	if (radius > 50)
	{
	    fprintf (stderr, "Enormous object: setting size to 50\n");
#if 0
	    fprintf (stderr, "thres: %e  sum: %e  nval: %e  aver: %e\n",
		     threshold, sum, nval, sum / nval);
#endif
	    return (radius);
	}
    }
    /*fprintf (stderr, "radius: %d\n", radius);*/
    return (radius);
}   /*  End Function find_size  */

static void find_centroid (sarray array, double *xpos, double *ypos,
			   double max_radius, flag inverted)
/*  [SUMMARY] Find the centroid of a feature.
    [PURPOSE] This routine will search for the centroid of an object. It does
    this by computing weighted means.
    <array> The simple array.
    <xpos> The horizontal data pixel position. This is modified to the centroid
    position.
    <ypos> The vertical data pixel position. This is modified to the centroid
    position.
    <max_radius> The maximum radius the computation is performed to.
    <inverted> If TRUE, the algorithms should search for a minimum, rather than
    a maximum.
    [NOTE] (0.0, 0.0) corresponds to the centre of the first pixel.
    [RETURNS] Nothing.
*/
{
    iaddr startx, stopx, x, starty, stopy, y;
    double dx, dy, sq_radius, sq_distance;
    double weighted_sum_x, weighted_sum_y, sum, value, sub;

    startx = *xpos - max_radius;
    stopx = *xpos + max_radius + 1.0;
    starty = *ypos - max_radius;
    stopy = *ypos + max_radius + 1.0;
    sq_radius = max_radius * max_radius;
    weighted_sum_x = 0.0;
    weighted_sum_y = 0.0;
    sum = 0.0;
    if (inverted)
    {
	sub = -TOOBIG;
	for (y = starty; y < stopy; ++y) for (x = startx; x < stopx; ++x)
	{
	    value = get_value (array, y, x);
	    if ( DOUBLE_IS_BLANK (value) ) continue;
	    if (value > sub) sub = value;
	}
    }
    else
    {
	sub = TOOBIG;
	for (y = starty; y < stopy; ++y) for (x = startx; x < stopx; ++x)
	{
	    value = get_value (array, y, x);
	    if ( DOUBLE_IS_BLANK (value) ) continue;
	    if (value < sub) sub = value;
	}
    }
    for (y = starty; y < stopy; ++y) for (x = startx; x < stopx; ++x)
    {
	dx = x - *xpos;
	dy = y - *ypos;
	sq_distance = dx * dx + dy * dy;
	if (sq_distance > sq_radius) continue;
	value = get_value (array, y, x) - sub;
	if ( DOUBLE_IS_BLANK (value) ) continue;
	weighted_sum_x += dx * value;
	weighted_sum_y += dy * value;
	sum += value;
    }
    if (sum == 0.0) return;
    *xpos += weighted_sum_x / sum;
    *ypos += weighted_sum_y / sum;
}   /*  End Function find_centroid  */

static double find_background (sarray array, double centre_x, double centre_y,
			       double inner_radius_x, double inner_radius_y,
			       double outer_radius_x, double outer_radius_y)
/*  [SUMMARY] Find the background value inside an annulus. This is the median.
    <array> The simple array.
    <centre_x> The horizontal centre of the concentric circles defining the
    annulus.
    <centre_y> The vertical centre of the concentric circles.
    <inner_radius_x> The horizontal radius of the inner circle.
    <inner_radius_y> The vertical radius of the inner circle.
    <outer_radius_x> The horizontal radius of the outer circle.
    <outer_radius_y> The vertical radius of the outer circle.
    [RETURNS] The background estimate.
*/
{
    iaddr x, xstart, xstop, y, ystart, ystop;
    unsigned int length, count;
    double value, tmp, distance, inner_area, outer_area;
    double *list;
    static char function_name[] = "_ds_centroid_find_background";

    /*  First extract all the values in the annulus and put in a list  */
    inner_area = PI * inner_radius_x * inner_radius_y;
    outer_area = PI * outer_radius_x * outer_radius_y;
    length = ( (outer_area - inner_area) * LIST_FACTOR );
    list = m_alloc_scratch (length * sizeof *list, function_name);
    xstart = centre_x - outer_radius_x;
    xstop = centre_x + outer_radius_x;
    ystart = centre_y - outer_radius_y;
    ystop = centre_y + outer_radius_y;
    count = 0;
    for (y = ystart; y < ystop; ++y) for (x = xstart; x < xstop; ++x)
    {
	tmp = (double) x - centre_x;
	distance = tmp * tmp / (inner_radius_x * inner_radius_x);
	tmp = (double) y - centre_y;
	distance += tmp * tmp / (inner_radius_y * inner_radius_y);
	if (distance < 1.0) continue;
	tmp = (double) x - centre_x;
	distance = tmp * tmp / (outer_radius_x * outer_radius_x);
	tmp = (double) y - centre_y;
	distance += tmp * tmp / (outer_radius_y * outer_radius_y);
	if (distance > 1.0) continue;
	value = get_value (array, y, x);
	if ( DOUBLE_IS_BLANK (value) ) continue;
	if (count >= length)
	{
	    fprintf (stderr, "Values array too small, length is: %u\n",length);
	    a_prog_bug (function_name);
	}
	list[count++] = value;
    }
    length = count;
    /*  Now sort the list: primitive bubble sort  */
    while (TRUE)
    {
	flag changed = FALSE;

	for (count = 0; count < length - 1; ++count)
	{
	    if (list[count] > list[count + 1])
	    {
		changed = TRUE;
		swap (list[count], list[count + 1], tmp);
	    }
	}
	if (!changed) break;
    }
    /*  The median is the midpoint of the sorted list  */
    if (length & 1)
    {
	length = length >> 1;
	value = list[length + 1];
    }
    else
    {
	length = length >> 1;
	value = 0.5 * (list[length] + list[length + 1]);
    }
    m_free_scratch ();
    return (value);
}   /*  End Function find_background  */
