/*LINTLIBRARY*/
/*  contour.c

    This code provides contour extraction routines.

    Copyright (C) 1996-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for extracting contours
    from images in the general data structure supported in Karma.


    Written by      Richard Gooch   20-JUL-1996

    Updated by      Richard Gooch   7-JUN-1997: Supported non-float data types.

    Updated by      Richard Gooch   30-JUL-1997: Avoid drawing contours around
  blanks.

    Updated by      Richard Gooch   18-SEP-1997: Switched to <ds_get_element>
  from <ds_convert_atomic>, since latter takes absolute value!

    Updated by      Richard Gooch   12-DEC-1997: Reduced buffer increment which
  was too big when using large images.

    Updated by      Richard Gooch   22-FEB-1998: Removed casts for memory
  functions.

    Updated by      Richard Gooch   5-MAR-1998: Put back in cast for
  initialising number of segments. Otherwise get SEGV.

    Updated by      Richard Gooch   19-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   13-AUG-1998: Documentation fix.

    Last updated by Richard Gooch   11-SEP-1999: Added call to function
  <s_check_for_int>.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_s.h>


/*  Private functions  */
STATIC_FUNCTION (void reallocate_coords,
		 (uaddr buf_size, uaddr num_to_copy,
		  double **x0, double **y0, double **x1, double **y1) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag ds_contour (CONST char *image, unsigned int elem_type,
		 CONST dim_desc *hdim, CONST uaddr *hoffsets,
		 CONST dim_desc *vdim, CONST uaddr *voffsets,
		 unsigned int num_levels, CONST double *contour_levels,
		 uaddr *buffer_sizes, uaddr size_stride,
		 double **x0_arr, double **y0_arr,
		 double **x1_arr, double **y1_arr,
		 uaddr *num_segments, uaddr seg_stride)
/*  [SUMMARY] Extract contours from a 2-dimensional array.
    [PURPOSE] This routine will extract contours from a 2-dimensional
    Intelligent Array, producing a list of line segments that approximate the
    countours. The co-ordinates of the line segments are in linear world
    co-ordinates.
    <image> The start of the image slice data.
    <elem_type> The type of the data.
    <hdim> The horizontal dimension descriptor.
    <hoffsets> The address offsets for data along the horizontal dimension.
    <vdim> The vertical dimension descriptor.
    <voffsets> The address offsets for data along the vertical dimension.
    <num_levels> The number of contour levels.
    <contour_levels> The array of contour levels.
    <buffer_sizes> A pointer to an array of co-ordinate buffer sizes. This is
    modified.
    <size_stride> The stride (in bytes) between consecutive elements in the
    co-ordinate buffer size array.
    <x0_arr> A pointer to an array of co-ordinate array pointers.
    The co-ordinate arrays may be internally reallocated, hence the array
    pointer may be modified.
    <y0_arr> A pointer to an array of co-ordinate array pointers. The
    co-ordinate arrays may be internally reallocated, hence the array pointer
    may be modified.
    <x1_arr> A pointer to an array of co-ordinate array pointers. The
    co-ordinate arrays may be internally reallocated, hence the array pointer
    may be modified.
    <y1_arr> A pointer to an array of co-ordinate array pointers. The
    co-ordinate arrays may be internally reallocated, hence the array pointer
    may be modified.
    <num_segments> A pointer to an array where the the number of segments per
    level will be written.
    <seg_stride> The stride (in bytes) between consecutive elements in the
    segment count array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int icase;
    unsigned int xlen, ylen, x, y, lev_count;
    uaddr buf_inc;
    double xpos, ypos, first_x, xdelt, ydelt;
    double val_00, val_10, val_01, val_11;
    double cval, x0, y0, x1, y1;
    CONST char *line1_ptr, *line2_ptr;
    uaddr *buf_size, *seg_count;
    double value[2];
    static char function_name[] = "ds_contour";

    if ( (image == NULL) || (hoffsets == NULL) || (voffsets == NULL) )
    {
	fputs ("NULL pointer(s) passed\n", stderr);
	a_prog_bug (function_name);
    }
    xlen = hdim->length;
    ylen = vdim->length;
    buf_inc = xlen + ylen;
    if (buf_inc < 256) buf_inc = 256;
    if ( (hdim->coordinates != NULL) || (vdim->coordinates != NULL) )
    {
	fprintf (stderr, "Co-ordinate array not supported\n");
	a_prog_bug (function_name);
    }
    for (lev_count = 0; lev_count < num_levels; ++lev_count)
    {
	seg_count = (uaddr *) ( (char *) num_segments +lev_count *seg_stride );
	*seg_count = 0;
    }
    /*  Process image pixel-by-pixel  */
    xdelt = (hdim->last_coord - hdim->first_coord) /
	(double) (hdim->length - 1);
    ydelt = (vdim->last_coord - vdim->first_coord) /
	(double) (vdim->length - 1);
    first_x = hdim->first_coord;  /*Use a variable to reduce memory accesses */
    for (y = 0; y < ylen - 1; ++y)
    {
	if ( s_check_for_int () )
	{
	    fprintf (stderr, "%s: interrupt detected\n", function_name);
	    return (FALSE);
	}
	ypos = ds_get_coordinate (vdim, y);
	line1_ptr = image + voffsets[y];
	line2_ptr = image + voffsets[y + 1];
	for (x = 0; x < xlen - 1; ++x)
	{
	    if (elem_type == K_FLOAT)
	    {
		val_00 = *(float *) (line1_ptr + hoffsets[x]);
		val_10 = *(float *) (line1_ptr + hoffsets[x + 1]);
		val_01 = *(float *) (line2_ptr + hoffsets[x]);
		val_11 = *(float *) (line2_ptr + hoffsets[x + 1]);
	    }
	    else
	    {
		ds_get_element (line1_ptr + hoffsets[x], elem_type, value,
				NULL);
		val_00 = value[0];
		ds_get_element (line1_ptr + hoffsets[x + 1], elem_type, value,
				NULL);
		val_10 = value[0];
		ds_get_element (line2_ptr + hoffsets[x], elem_type, value,
				NULL);
		val_01 = value[0];
		ds_get_element (line2_ptr + hoffsets[x + 1], elem_type, value,
				NULL);
		val_11 = value[0];
	    }
	    if ( DOUBLE_IS_BLANK (val_00) || DOUBLE_IS_BLANK (val_10) ||
		 DOUBLE_IS_BLANK (val_01) || DOUBLE_IS_BLANK (val_11) )
		continue;
	    xpos = (double) x * xdelt + first_x;
	    /*  Search contour levels for an intersection or two  */
	    for (lev_count = 0; lev_count < num_levels;
		 ++lev_count)
	    {
		cval = contour_levels[lev_count];
		buf_size = (uaddr *) ( (char *) buffer_sizes +
				       lev_count * size_stride );
		seg_count = (uaddr *) ( (char *) num_segments +
					lev_count * seg_stride );
		icase = 1;
		if (cval > val_00) icase = icase + 1;
		if (cval > val_10) icase = icase + 2;
		if (cval > val_01) icase = icase + 4;
		if (cval > val_11) icase = 9 - icase;
		if (icase < 2) continue;
		switch (icase)
		{
		  case 2:
		    x0 = xpos +  xdelt * (cval - val_00) / (val_10 - val_00);
		    y0 = ypos;
		    x1 = xpos;
		    y1 = ypos + ydelt * (cval - val_00) / (val_01 - val_00);
		    break;
		  case 3:
		    x0 = xpos + xdelt * (cval - val_00) / (val_10 - val_00);
		    y0 = ypos;
		    x1 = xpos + xdelt;
		    y1 = ypos + ydelt * (cval - val_10) / (val_11 - val_10);
		    break;
		  case 4:
		    x0 = xpos;
		    y0 = ypos + ydelt * (cval - val_00) / (val_01 - val_00);
		    x1 = xpos + xdelt;
		    y1 = ypos + ydelt * (cval - val_10) / (val_11 - val_10);
		    break;
		  case 5:
		    x0 = xpos;
		    y0 = ypos + ydelt * (cval - val_00) / (val_01 - val_00);
		    x1 = xpos + xdelt * (cval - val_01) / (val_11 - val_01);
		    y1 = ypos + ydelt;
		    break;
		  case 6:
		    x0 = xpos + xdelt * (cval - val_00) / (val_10 - val_00);
		    y0 = ypos;
		    x1 = xpos + xdelt * (cval - val_01) / (val_11 - val_01);
		    y1 = ypos + ydelt;
		    break;
		  case 7:
		    x0 = xpos + xdelt * (cval - val_00) / (val_10 - val_00);
		    y0 = ypos;
		    x1 = xpos;
		    y1 = ypos + ydelt * (cval - val_00) / (val_01 - val_00);
		    if (*seg_count >= *buf_size)
		    {
			reallocate_coords (*buf_size + buf_inc, *buf_size,
					   x0_arr + lev_count,
					   y0_arr + lev_count,
					   x1_arr + lev_count,
					   y1_arr + lev_count);
			*buf_size += buf_inc;
		    }
		    x0_arr[lev_count][*seg_count] = x0;
		    y0_arr[lev_count][*seg_count] = y0;
		    x1_arr[lev_count][*seg_count] = x1;
		    y1_arr[lev_count][*seg_count] = y1;
		    ++*seg_count;
		    x0 = xpos + xdelt * (cval - val_01) / (val_11 - val_01);
		    y0 = ypos + ydelt;
		    x1 = xpos + xdelt;
		    y1 = ypos + ydelt * (cval - val_10) / (val_11 - val_10);
		    break;
		  case 8:
		    x0 = xpos + xdelt * (cval - val_01) / (val_11 - val_01);
		    y0 = ypos + ydelt;
		    x1 = xpos + xdelt;
		    y1 = ypos + ydelt * (cval - val_10) / (val_11 - val_10);
		    break;
		  default:
		    continue;
		}
		if (*seg_count >= *buf_size)
		{
		    reallocate_coords (*buf_size + buf_inc, *buf_size,
				       x0_arr + lev_count, y0_arr + lev_count,
				       x1_arr + lev_count, y1_arr + lev_count);
		    *buf_size += buf_inc;
		}
		x0_arr[lev_count][*seg_count] = x0;
		y0_arr[lev_count][*seg_count] = y0;
		x1_arr[lev_count][*seg_count] = x1;
		y1_arr[lev_count][*seg_count] = y1;
		++*seg_count;
	    }
	}
    }
    return (TRUE);
}   /*  End Function ds_contour  */


/*  Private functions follow  */

static void reallocate_coords (uaddr buf_size, uaddr num_to_copy,
			       double **x0, double **y0,
			       double **x1, double **y1)
/*  [SUMMARY] Reallocate co-ordinate buffers.
    <buf_size> The new desired buffer size.
    <num_to_copy> The number of values to copy from the old buffers.
    <x0> A pointer to a co-ordinate array pointer. The co-ordinate array may be
    internally reallocated, hence the array pointer may be modified.
    <y0> A pointer to a co-ordinate array pointer. The co-ordinate array may be
    internally reallocated, hence the array pointer may be modified.
    <x1> A pointer to a co-ordinate array pointer. The co-ordinate array may be
    internally reallocated, hence the array pointer may be modified.
    <y1> A pointer to a co-ordinate array pointer. The co-ordinate array may be
    internally reallocated, hence the array pointer may be modified.
    [RETURNS] Nothing. On failure the process aborts.
*/
{
    double *ptr;
    static char function_name[] = "_ds_contour_reallocate_coords";

    if (num_to_copy < 1)
    {
	if (*x0 != NULL) m_free (*x0);
	if (*y0 != NULL) m_free (*y0);
	if (*x1 != NULL) m_free (*x1);
	if (*y1 != NULL) m_free (*y1);
	if ( ( *x0 = m_alloc (sizeof **x0 * buf_size) ) == NULL )
	{
	    m_abort (function_name, "x0 array");
	}
	if ( ( *y0 = m_alloc (sizeof **y0 * buf_size) ) == NULL )
	{
	    m_abort (function_name, "y0 array");
	}
	if ( ( *x1 = m_alloc (sizeof **x1 * buf_size) ) == NULL )
	{
	    m_abort (function_name, "x1 array");
	}
	if ( ( *y1 = m_alloc (sizeof **y1 * buf_size) ) == NULL )
	{
	    m_abort (function_name, "y1 array");
	}
	return;
    }
    /*  Pity we have to do this  */
    /*  Copy x0  */
    if ( ( ptr = m_alloc (sizeof *ptr * buf_size) ) == NULL )
    {
	m_abort (function_name, "x0 array");
    }
    m_copy (ptr, (char *) *x0, sizeof *ptr * num_to_copy);
    m_free (*x0);
    *x0 = ptr;
    /*  Copy y0  */
    if ( ( ptr = m_alloc (sizeof *ptr * buf_size) ) == NULL )
    {
	m_abort (function_name, "y0 array");
    }
    m_copy (ptr, *y0, sizeof *ptr * num_to_copy);
    m_free (*y0);
    *y0 = ptr;
    /*  Copy x1  */
    if ( ( ptr = m_alloc (sizeof *ptr * buf_size) ) == NULL )
    {
	m_abort (function_name, "x1 array");
    }
    m_copy (ptr, *x1, sizeof *ptr * num_to_copy);
    m_free (*x1);
    *x1 = ptr;
    /*  Copy y1  */
    if ( ( ptr = m_alloc (sizeof *ptr * buf_size) ) == NULL )
    {
	m_abort (function_name, "y1 array");
    }
    m_copy (ptr, *y1, sizeof *ptr * num_to_copy);
    m_free (*y1);
    *y1 = ptr;
}   /*  End Function reallocate_coords  */
