/*LINTLIBRARY*/
/*  easy_alloc.c

    This code provides data structure allocation routines.

    Copyright (C) 1992-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for allocating data
  structures for the general data structure supported in Karma.


    Written by      Richard Gooch   22-FEB-1998: Copied bits from alloc.c

    Updated by      Richard Gooch   24-FEB-1998: Switch to new array_pointer
  which has free function. Created <ds_alloc_shm_array> experimental function.

    Updated by      Richard Gooch   30-APR-1998: Added #ifdef for function
  <ds_alloc_shm_array>.

    Updated by      Richard Gooch   9-MAY-1998: Changed interface to function
  <ds_alloc_shm_array>.

    Updated by      Richard Gooch   13-MAY-1998: Finished implementing function
  <ds_alloc_mmap_array>.

    Updated by      Richard Gooch   23-JUN-1998: No longer clear array in
  <ds_alloc_mmap_array> if no FD passed.

    Last updated by Richard Gooch  15-OCT-2001: Fixed bug in function
  <ds_easy_alloc_array_from_array_desc>: detach <<arr_desc>> from
  <<multi_desc>> prior to calling <ds_dealloc_multi>.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <karma.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>


/*  Private functions  */
STATIC_FUNCTION (void mmap_close, (void *info,struct _array_pointer *arrayp) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
char *ds_easy_alloc_array (multi_array **multi_desc, unsigned int num_dim,
			   CONST uaddr *lengths, CONST double *first_arr,
			   CONST double *last_arr,
			   CONST char **names, unsigned int data_type,
			   CONST char *data_name)
/*  [SUMMARY] Allocate a data structure with a simple array.
    [PURPOSE] This routine will allocate memory for a multi-dimensional,
    regular array, and the required headers and the multi-array header. The
    array is NOT tiled nor are any address offsets computed. The data packet
    that may be stored in the array is a single, atomic datum.
    [NOTE] This routine does NOT create a Karma arrayfile. This must be done
    with a call to [<dsxfr_put_multi>].
    <multi_desc> The multi-array structure pointer to the created structure is
    written here.
    <num_dim> The number of dimensions of the array.
    <lengths> The array of dimension lengths.
    <first_arr> An array of first co-ordinate values for each dimension.
    <last_arr> An array of last co-ordinate values for each dimension.
    [NOTE] If either of <<first_arr>> or <<last_arr>> is NULL, the range of the
    co-ordinates will be the lengths of the dimensions minus 1. If one of the
    pointers is NULL, the other array is used to tie one end of the range for
    each dimension. If both pointers are NULL,the minima are 0.0 and the maxima
    are the lengths minus 1.
    <names> The array of dimension names. If this is NULL, the names "Axis 0",
    "Axis 1", ...etc will be used. The character arrays are copied, so the
    arrays of characters and the array of pointers may be subsequently
    deallocated.
    <data_type> The type of the elements in the array. See
    [<ds_KARMA_DATA_TYPES>] for a list of legal values.
    <data_name> The name of the data type. If this is NULL, then the name
    "Data Value" will be used. The name string is copied, thus the memory used
    for the input string may be subsequently deallocated.
    [RETURNS] A pointer to the start of the array on success, else NULL.
*/
{
    static char *def_data_name = "Data Value";
    static char function_name[] = "ds_easy_alloc_array";

    if ( (multi_desc == NULL) || (lengths == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if (data_name == NULL)
    {
	data_name = def_data_name;
    }
    return ds_easy_alloc_n_element_array (multi_desc, num_dim, lengths,
					  first_arr, last_arr, names,
					  1, &data_type,
					  (CONST char **) &data_name);
}   /*  End Function ds_easy_alloc_array  */

/*PUBLIC_FUNCTION*/
char *ds_easy_alloc_n_element_array (multi_array **multi_desc,
				     unsigned int num_dim,
				     CONST uaddr *lengths,
				     CONST double *first_arr,
				     CONST double *last_arr,
				     CONST char **names,
				     unsigned int num_elements,
				     CONST unsigned int *data_types,
				     CONST char **data_names)
/*  [SUMMARY] Allocate a data structure with a simple, multi element array.
    [PURPOSE] This routine will allocate memory for a multi-dimensional,
    regular array, and the required headers and the multi-array header. The
    array is NOT tiled nor are any address offsets computed. The data packet
    that may be stored in the array contains many atomic data.
    [NOTE] This routine does NOT create a Karma arrayfile. This must be done
    with a call to [<dsxfr_put_multi>].
    <multi_desc> The multi-array structure pointer to the created structure is
    written here.
    <num_dim> The number of dimensions of the array.
    <lengths> The array of dimension lengths.
    <first_arr> An array of first co-ordinate values for each dimension.
    <last_arr> An array of last co-ordinate values for each dimension.
    [NOTE] If either of <<first_arr>> or <<last_arr>> is NULL, the range of the
    co-ordinates will be the lengths of the dimensions minus 1. If one of the
    pointers is NULL, the other array is used to tie one end of the range for
    each dimension. If both pointers are NULL,the minima are 0.0 and the maxima
    are the lengths minus 1.
    <names> The array of dimension names. If this is NULL, the names "Axis 0",
    "Axis 1", ...etc will be used. The character arrays are copied, so the
    arrays of characters and the array of pointers may be subsequently
    deallocated.
    <num_elements> The number of elements in the array packets.
    <data_types> The types of the elements in the array. See
    [<ds_KARMA_DATA_TYPES>] for a list of legal values.
    <data_names> The names of the data type. The name strings are copied, thus
    the memory used for the input string may be subsequently deallocated.
    [RETURNS] A pointer to the start of the array on success, else NULL.
*/
{
    array_pointer arrayp;
    static char function_name[] = "ds_easy_alloc_n_element_array";

    if ( (multi_desc == NULL) || (lengths == NULL) || (data_types == NULL) ||
	(data_names == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if ( ( *multi_desc =
	  ds_wrap_preallocated_n_element_array (NULL, num_dim, lengths,
						first_arr, last_arr, NULL,
						names, num_elements,
						data_types, data_names) )
	== NULL ) return (NULL);
    m_copy (&arrayp, (**multi_desc).data[0], sizeof arrayp);
    return (arrayp.array);
}   /*  End Function ds_easy_alloc_n_element_array  */

/*PUBLIC_FUNCTION*/
multi_array *ds_wrap_preallocated_n_element_array
    (char *array, unsigned int num_dim, CONST uaddr *lengths,
     CONST double *first_arr, CONST double *last_arr,
     CONST double **coordinates, CONST char **names,
     unsigned int num_elements, CONST unsigned int *data_types,
     CONST char **data_names)
/*  [SUMMARY] Wrap a data structure around an array.
    [PURPOSE] This routine will "wrap" an externally allocated array by
    allocating the required descriptors. The array is NOT tiled nor are any
    address offsets computed. The data packet that may be stored in the array
    contains many atomic data.
    [NOTE] This routine does NOT create a Karma arrayfile. This must be done
    with a call to [<dsxfr_put_multi>].
    <array> The externally allocated array. This data must be externally
    deallocated when no longer needed. If this is NULL, then the array data is
    internally allocated (and should not be externally deallocated).
    <num_dim> The number of dimensions of the array.
    <lengths> The array of dimension lengths.
    <first_arr> An array of first co-ordinate values for each dimension.
    <last_arr> An array of last co-ordinate values for each dimension.
    [NOTE] If either of <<first_arr>> or <<last_arr>> is NULL, the range of the
    co-ordinates will be the lengths of the dimensions minus 1. If one of the
    pointers is NULL, the other array is used to tie one end of the range for
    each dimension. If both pointers are NULL,the minima are 0.0 and the maxima
    are the lengths minus 1.
    <coordinates> The array of co-ordinate array pointers. If this is NULL,
    all dimensions are assumed to be regularly spaced. If any co-ordinate array
    pointer is NULL, the corresponding dimension is assumed to be regularly
    spaced.
    <names> The array of dimension names. If this is NULL, the names "Axis 0",
    "Axis 1", ...etc will be used. The character arrays are copied, so the
    arrays of characters and the array of pointers may be subsequently
    deallocated.
    <num_elements> The number of elements in the array packets.
    <data_types> The types of the elements in the array. See
    [<ds_KARMA_DATA_TYPES>] for a list of legal values.
    <data_names> The names of the data type. The name strings are copied, thus
    the memory used for the input string may be subsequently deallocated.
    [RETURNS] A pointer the multi_array descriptor on success, else NULL.
*/
{
    array_pointer arrayp;
    multi_array *multi_desc;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    static char function_name[] = "ds_wrap_preallocated_n_element_array";

    if ( ( multi_desc = ds_alloc_multi ( (unsigned int) 1 ) ) == NULL )
    {
	m_error_notify (function_name, "multi_desc descriptor");
        return (NULL);
    }
    if ( ( pack_desc = ds_alloc_packet_desc ( (unsigned) 1 ) ) == NULL )
    {
	ds_dealloc_multi (multi_desc);
	m_error_notify (function_name, "packet descriptor");
        return (NULL);
    }
    multi_desc->headers[0] = pack_desc;
    if ( ( arr_desc = ds_easy_alloc_array_desc (num_dim, lengths,
						first_arr, last_arr,
						coordinates, names,
						num_elements, data_types,
						data_names) ) == NULL )
    {
	ds_dealloc_multi (multi_desc);
	m_error_notify (function_name, "array descriptor");
        return (NULL);
    }
    pack_desc->element_desc[0] = (char *) arr_desc;
    pack_desc->element_types[0] = K_ARRAY;
    /*  Allocate the data space  */
    if ( ( multi_desc->data[0] =
	   ds_alloc_data (multi_desc->headers[0], TRUE,
			  (array == NULL) ? TRUE : FALSE) )
	 == NULL )
    {
	ds_dealloc_multi (multi_desc);
	m_error_notify (function_name, "top level packet");
        return (NULL);
    }
    if (array != NULL)
    {
	m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
	arrayp.array = array;
	arrayp.free = NULL;
	arrayp.info = NULL;
	m_copy (multi_desc->data[0], &arrayp, sizeof arrayp);
    }
    return (multi_desc);
}   /*  End Function ds_wrap_preallocated_n_element_array  */

/*PUBLIC_FUNCTION*/
array_desc *ds_easy_alloc_array_desc (unsigned int num_dim,
				      CONST uaddr *lengths,
				      CONST double *first_arr,
				      CONST double *last_arr,
				      CONST double **coordinates,
				      CONST char **names,
				      unsigned int num_elements,
				      CONST unsigned int *data_types,
				      CONST char **data_names)
/*  [SUMMARY] Allocate a simple array descriptor.
    [PURPOSE] This routine will allocate an array descriptor and its
    associated packet descriptor.
    <num_dim> The number of dimensions in the array.
    <lengths> An array of lengths for each dimension (axis).
    <first_arr> An array of first co-ordinate values for each dimension.
    <last_arr> An array of last co-ordinate values for each dimension.
    If either of these pointers is NULL, the range of the co-ordinates will be
    the lengths of the dimensions minus 1. If one of the pointers is NULL, the
    other array is used to tie one end of the range for each dimension. If both
    pointers are NULL,the minima are 0.0 and the maxima are the lengths minus 1
    <coordinates> The array of co-ordinate array pointers. If this is NULL,
    all dimensions are assumed to be regularly spaced. If any co-ordinate array
    pointer is NULL, the corresponding dimension is assumed to be regularly
    spaced.
    <names> An array of string pointers to the names of each dimension. If this
    is NULL, the names "Axis 0", "Axis 1", ...etc will be used. Note: the
    character arrays are copied, so the arrays of characters and the array of
    pointers may be subsequently deallocated.
    <num_elements> The number of atomic elements in the array packet. If this
    is 0, no packet descriptor is allocated.
    <data_types> The types of the elements in the array. See
    [<ds_KARMA_DATA_TYPES>] for a list of legal values. This may be NULL if
    <<num_elements>> is 0.
    <data_names> The array of element names. The name strings are copied, thus
    the memory used for the input strings may be subsequently deallocated. This
    may be NULL if <<num_elements>> is 0.
    [RETURNS] An array descriptor pointer on success, else NULL.
*/
{
    flag this_dim_has_coords;
    unsigned int dim_count, elem_count, max_dim;
    double first_coord, last_coord;
    CONST char *tmp_pointer;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    dim_desc *dimension;
    char tmp_name[129];
    static char function_name[] = "ds_easy_alloc_array_desc";

    if (lengths == NULL)
    {
	fprintf (stderr, "NULL lengths pointer passed\n");
	a_prog_bug (function_name);
    }
    if (num_elements > 0)
    {
	if ( (data_types == NULL) || (data_names == NULL) )
	{
	    fprintf (stderr, "NULL data info pointer(s) passed\n");
	    a_prog_bug (function_name);
	}
    }
    max_dim = ~0;
    --max_dim;
    if (num_dim >= max_dim)
    {
	fprintf (stderr, "Too many dimensions: %u\n", num_dim);
	a_prog_bug (function_name);
    }
    if ( ( arr_desc = ds_alloc_array_desc (num_dim, 0) ) == NULL )
    {
	m_error_notify (function_name, "array descriptor");
        return (NULL);
    }
    for (dim_count = 0; dim_count < num_dim; ++dim_count)
    {
	if ( (coordinates != NULL) && (coordinates[dim_count] != NULL) )
	{
	    this_dim_has_coords = TRUE;
	}
	else
	{
	    this_dim_has_coords = FALSE;
	}
	/*  Allocate and add adimension descriptor  */
        if ( (names == NULL) || (names[dim_count] == NULL) )
        {
	    /*  Create temporary dimension name  */
#ifdef OS_VXMVX
            sprintf (tmp_name, "Axis %d", (int) dim_count);
#else
            sprintf (tmp_name, "Axis %u", dim_count);
#endif
            tmp_pointer = tmp_name;
        }
        else
        {
	    /*  Copy over character pointer  */
            tmp_pointer = names[dim_count];
        }
	if (this_dim_has_coords)
	{
	    first_coord = coordinates[dim_count][0];
	    last_coord = coordinates[dim_count][lengths[dim_count] - 1];
	}
	else
	{
	    if ( (first_arr == NULL) && (last_arr == NULL) )
	    {
		/*  Range is from  0.0  to  lengths - 1  */
		first_coord = 0.0;
		last_coord = (lengths[dim_count] - 1);
	    }
	    else
	    {
		/*  Range has at least one boundary supplied  */
		if (first_arr == NULL)
		{
		    /*  Last co-ords only supplied  */
		    last_coord = last_arr[dim_count];
		    first_coord = last_coord - (lengths[dim_count] - 1);
		}
		else
		{
		    /*  First co-ords supplied  */
		    first_coord = first_arr[dim_count];
		    if (last_arr == NULL)
		    {
			/*  First co-ords only supplied  */
			last_coord = first_coord + (lengths[dim_count] - 1);
		    }
		    else
		    {
			/*  Last co-ords also supplied  */
			last_coord = last_arr[dim_count];
		    }
		}
	    }
	}
        if ( ( dimension =
	      ds_alloc_dim_desc (tmp_pointer, lengths[dim_count],
				 first_coord, last_coord,
				 this_dim_has_coords ? FALSE : TRUE) )
	    == NULL )
        {
	    ds_dealloc_array_desc (arr_desc);
	    m_error_notify (function_name, "dimension descriptor");
            return (NULL);
        }
        arr_desc->dimensions[dim_count] = dimension;
	arr_desc->lengths[dim_count] = dimension->length;
	if (this_dim_has_coords)
	{
	    m_copy (dimension->coordinates, coordinates[dim_count],
		    sizeof **coordinates * lengths[dim_count]);
	}
    }
    if (num_elements < 1) return (arr_desc);
    if ( ( pack_desc = ds_alloc_packet_desc (num_elements) ) == NULL )
    {
	ds_dealloc_array_desc (arr_desc);
	m_error_notify (function_name, "array packet descriptor");
        return (NULL);
    }
    arr_desc->packet = pack_desc;
    for (elem_count = 0; elem_count < num_elements; ++elem_count)
    {
	if (ds_element_is_atomic (data_types[elem_count]) != TRUE)
	{
	    /*  Bad data type  */
	    fprintf (stderr, "Bad data type: %u\n",
			    data_types[elem_count]);
	    a_prog_bug (function_name);
	}
	pack_desc->element_types[elem_count] = data_types[elem_count];
	if (data_names[elem_count] == NULL)
	{
	    fprintf (stderr, "NULL data name pointer in array\n");
	    a_prog_bug (function_name);
	}
	if ( ( pack_desc->element_desc[elem_count] =
	       m_alloc (strlen (data_names[elem_count]) + 1) ) == NULL )
	{
	    ds_dealloc_array_desc (arr_desc);
	    m_error_notify (function_name, "element name");
	    return (NULL);
	}
	strcpy (pack_desc->element_desc[elem_count], data_names[elem_count]);
    }
    return (arr_desc);
}   /*  End Function ds_easy_alloc_array_desc  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *ds_easy_alloc_array_from_array_desc (CONST array_desc *arr_desc,
						  CONST array_pointer *arrayp,
						  flag clear)
/*  [SUMMARY] Wrap a data structure around an array.
    [PURPOSE] This routine will "wrap" an externally allocated array by
    allocating the required descriptors. The array is NOT tiled nor are any
    address offsets computed. The data packet that may be stored in the array
    contains many atomic data.
    [NOTE] This routine does NOT create a Karma arrayfile. This must be done
    with a call to [<dsxfr_put_multi>].
    <arr_desc> The array descriptor describing the array data.
    <arrayp> The array data pointer. If this is NULL the array data is
    internally allocated.
    <clear> If TRUE, then the routine will initialise (set to zero) all the
    data, if that data is internally allocated.
    [RETURNS] A pointer the multi_array descriptor on success, else NULL.
*/
{
    multi_array *multi_desc;
    packet_desc *pack_desc;
    static char function_name[] = "ds_easy_alloc_array_from_array_desc";

    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor\n");
	a_prog_bug (function_name);
    }
    if ( ( multi_desc = ds_alloc_multi (1) ) == NULL )
    {
	m_error_notify (function_name, "multi_desc descriptor");
        return (NULL);
    }
    if ( ( pack_desc = ds_alloc_packet_desc (1) ) == NULL )
    {
	ds_dealloc_multi (multi_desc);
	m_error_notify (function_name, "packet descriptor");
        return (NULL);
    }
    multi_desc->headers[0] = pack_desc;
    pack_desc->element_desc[0] = (char *) arr_desc;
    pack_desc->element_types[0] = K_ARRAY;
    if (arrayp == NULL)
    {
	multi_desc->data[0] = ds_alloc_data (multi_desc->headers[0], clear,
					     TRUE);
	if (multi_desc->data[0] == NULL)
	{
	    pack_desc->element_desc[0] = NULL;
	    pack_desc->element_types[0] = NONE;
	    ds_dealloc_multi (multi_desc);
	    m_error_notify (function_name, "array data");
	    return (NULL);
	}
    }
    else
    {
	multi_desc->data[0] = ds_alloc_data (multi_desc->headers[0], TRUE,
					     FALSE);
	if (multi_desc->data[0] == NULL)
	{
	    pack_desc->element_desc[0] = NULL;
	    pack_desc->element_types[0] = NONE;
	    ds_dealloc_multi (multi_desc);
	    m_error_notify (function_name, "array data");
	    return (NULL);
	}
	m_copy (multi_desc->data[0], arrayp, sizeof *arrayp);
    }
    return (multi_desc);
}   /*  End Function ds_easy_alloc_array_from_array_desc  */

/*EXPERIMENTAL_FUNCTION*/
flag ds_alloc_vm_array (array_pointer *arrayp, CONST array_desc *arr_desc,
			flag clear, flag warn)
/*  [SUMMARY] Allocate storage for an array using ordinary virtual memory.
    <arrayp> Information describing the shared memory region is written here.
    <arr_desc> The array descriptor describing the array data. This is used to
    determine the amount of memory require to store the array.
    <clear> If TRUE the array data is initialised to zero.
    <warn> If TRUE, warning messages are displayed if there are any problems.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    uaddr size;
    void *addr;
    static char function_name[] = "ds_alloc_vm_array";

    size = ds_get_packet_size (arr_desc->packet) *ds_get_array_size (arr_desc);
    addr = m_alloc (size);
    if (addr == NULL)
    {
	if (warn) fprintf (stderr,
			   "%s: error allocating %lu bytes of array data\n",
			   function_name, (unsigned long) size);
	return (FALSE);
    }
    m_clear (arrayp, sizeof *arrayp);
    if (clear) m_clear (addr, size);
    arrayp->array = addr;
    arrayp->free = ( void (*) () ) m_free;
    arrayp->info = addr;
    arrayp->mem_type = K_ARRAY_MEM_TYPE_NORMAL;
    return (TRUE);
}   /*  End Function ds_alloc_vm_array  */

/*EXPERIMENTAL_FUNCTION*/
flag ds_alloc_shm_array (array_pointer *arrayp, CONST array_desc *arr_desc,
			 flag clear, flag warn)
/*  [SUMMARY] Allocate storage for an array using shared memory.
    <arrayp> Information describing the shared memory region is written here.
    <arr_desc> The array descriptor describing the array data. This is used to
    determine the amount of memory require to store the array.
    <clear> If TRUE the array data is initialised to zero.
    <warn> If TRUE, warning messages are displayed if there are any problems.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int shmid;
    uaddr size;
    void *addr;
    static char function_name[] = "ds_alloc_shm_array";

    size = ds_get_packet_size (arr_desc->packet) *ds_get_array_size (arr_desc);
    if ( ( shmid = m_shm_alloc (size, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) )
	 < 0 )
    {
	if (warn) fprintf (stderr, "%s: error creating SHM segment\t%s\n",
			   function_name, ERRSTRING);
	return (FALSE);
    }
    addr = m_shm_attach (shmid, TRUE, TRUE);
#ifdef OS_Linux
    /*  Advanced operating systems support attaching after a segment has been
	marked for deletion  */
    m_shm_delete (shmid);
#endif
    if (addr == NULL)
    {
	if (warn) fprintf (stderr, "%s: error attaching SHM segment\t%s\n",
			   function_name, ERRSTRING);
#ifndef OS_Linux
	m_shm_delete (shmid);
#endif
	return (FALSE);
    }
    m_clear (arrayp, sizeof *arrayp);
    if (clear) m_clear (addr, size);
    arrayp->array = addr;
    arrayp->free = ( void (*) () ) m_shm_detach;
    arrayp->info = addr;
    arrayp->mem_type = K_ARRAY_MEM_TYPE_SHM;
    arrayp->u.shm.shmid = shmid;
    arrayp->u.shm.base = addr;
    return (TRUE);
}   /*  End Function ds_alloc_shm_array  */

/*EXPERIMENTAL_FUNCTION*/
flag ds_alloc_mmap_array (int fd, uaddr offset, uaddr size, flag writable,
			  array_pointer *arrayp, CONST array_desc *arr_desc,
			  flag clear, flag warn)
/*  [SUMMARY] Allocate storage for an array using shared anon mmap.
    <fd> The file descriptor to map. If this is less than zero, "/dev/zero" is
    mapped.
    <offset> The offset from the start of the file where the array data should
    reside.
    <size> The size of the region to map. If this is 0 the size is taken from
    the array size.
    <writable> If the mapped pages are to be writable, this must be TRUE. If
    this is FALSE and the memory pages are written to, a segmentation fault
    occurs.
    <arrayp> Information describing the shared memory region is written here.
    <arr_desc> The array descriptor describing the array data. This is used to
    determine the amount of memory require to store the array.
    <clear> If TRUE the array data is initialised to zero.
    <warn> If TRUE, warning messages are displayed if there are any problems.
    [NOTE] The file is mapped from its start.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    uaddr arr_size;
    struct stat statbuf;
    void *addr;
    static char function_name[] = "ds_alloc_mmap_array";

    arr_size = ds_get_packet_size (arr_desc->packet) *
	ds_get_array_size (arr_desc);
    if (size < 1) size = arr_size;
    if (fd < 0) clear = FALSE;
    if ( ( addr = m_map_fd (&fd, size, writable, TRUE, FALSE) ) == NULL )
    {
	if (warn) fprintf (stderr, "%s: error mapping\t%s\n",
			   function_name, ERRSTRING);
	return (FALSE);
    }
    if (fstat (fd, &statbuf) != 0)
    {
	fprintf (stderr, "%s: error statting FD=%d\t%s\n",
		 function_name, fd, ERRSTRING);
	m_unmap (addr, size);
	close (fd);
	return (FALSE);
    }
    m_clear (arrayp, sizeof *arrayp);
    arrayp->array = (char *) addr + offset;
    arrayp->free = mmap_close;
    arrayp->info = NULL;
    arrayp->mem_type = S_ISREG (statbuf.st_mode) ?
	K_ARRAY_MEM_TYPE_MMAP : K_ARRAY_MEM_TYPE_NORMAL;
    arrayp->u.mmap.fd = fd;
    arrayp->u.mmap.size = size;
    arrayp->u.mmap.base = addr;
    if (clear) m_clear (arrayp->array, arr_size);
    return (TRUE);
}   /*  End Function ds_alloc_mmap_array  */


/*  Private functions follow  */

static void mmap_close (void *info, struct _array_pointer *arrayp)
/*  [SUMMARY] Close a mmaped region.
    <info> The arbitrary pointer.
    <arrayp> The array_pointer.
    [RETURNS] Nothing.
*/
{
    m_unmap (arrayp->u.mmap.base, arrayp->u.mmap.size);
    close (arrayp->u.mmap.fd);
}   /*  End Function mmap_close  */
