/*LINTLIBRARY*/
/*  gauss.c

    This code provides gaussian fitting routines.

    Copyright (C) 1997-1998  Karl Glazebrook and Alison Offer Real code
    Copyright (C) 1997-2001  Richard Gooch                    API design&
                                                              maintenance

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Karl Glazebrook may be reached by email at  kgb@aaoepp2.aao.gov.au
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for getting info on the
    general data structure supported in Karma.


    Written by      Richard Gooch   12-AUG-1997: Received code from Karl.

    Updated by      Karl Glazebrook 3-JUN-1998: Added magnitude.

    Updated by      Karl Glazebrook 8-JUN-1998: Added negative peak and use
  supplied background estimate.

    Last updated by Richard Gooch   8-JUN-2001: Allow NULL <<offsets>> in
  <ds_gausscurve>.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_ds.h>
#include <karma_m.h>

#define NPAR 3
#define MAXITER 1000
#define malloc     m_alloc
#define free       m_free


/*  Private functions  */
STATIC_FUNCTION (double **malloc2D, (int nx, int ny) );
STATIC_FUNCTION (void free2D, (double **p, int nx, int ny) );
STATIC_FUNCTION (void seta,
		 (int npar, int npoints, double a[NPAR][NPAR], double **d,
		  double *sig) );
STATIC_FUNCTION (void setb,
		 (int npar, int npoints, double b[NPAR], double **d, 
		  double *y, double *yfit, double *sig) );
STATIC_FUNCTION (void funct, (int npoints, int npar, double *x, double *yfit,
			      double **d, double par[NPAR]) );
STATIC_FUNCTION (void decomp, (int n, int ndim, double x[NPAR][NPAR],
			       int iorder[NPAR]) );
STATIC_FUNCTION (void lineq,
		 (int n, int ndim, double x[NPAR][NPAR], double b[NPAR],
		  double d[NPAR], int iorder[NPAR]) );
STATIC_FUNCTION (int marquardt, (int npoints, int npar, double*x, double *y,
				 double* sig, double par[NPAR], double* r,
				 double a[NPAR][NPAR]) );

 
/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag ds_fitgauss (CONST float *array, CONST uaddr *offsets,
		  unsigned int num_points, double background, flag inverted,
		  double xcentre, double *peak_height, double *fwhm,  
		  double *magnitude)
/*  [SUMMARY] Fit a gaussian to a profile.
    [PURPOSE] This routine will fit a gaussian to a 1-dimensional profile. The
    algorithm uses a non-linear least squares method.
    <array> The 1-dimensional array of data.
    <offsets> The array of byte offsets for the array data.
    <num_points> The length of the array.
    <background> The background sky level..
    <inverted> If TRUE the object values are inverted (i.e. is negative) else
    the object is positive.
    <xcentre> The centre of the profile. If this is 0 it is assumed the profile
    is radial (i.e. the centre is at 0).
    <peak_height> The height of the fitted peak is written here.
    <fwhm> The Full Width Half Maximum of the fit is written here.
    <magnitude> The instrumental magnitude of the fitted profile
                (= 30-2.5log10( total counts )) is written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    double *x, *y, *sig, par[NPAR], *r, a[NPAR][NPAR];
    double ymax, xmax, xmin, val, pi;
    int i, ierr, np;
    double sign = inverted ? -1 : 1;

    ierr = 0;

    /* Memory allocation - alright its copied but this is probably best done
       as double anyway */
    x = malloc (num_points*sizeof(double));
    if (x == NULL)
    {
	fprintf (stderr, "Error allocating X array\n");
	return (FALSE);
    }
    y = malloc(num_points*sizeof(double));
    if (y == NULL)
    {
	fprintf (stderr, "Error allocating Y array\n");
	free (x);
	return (FALSE);
    }
   
  
    np = 0;
    for (i = 0; i < num_points; i++ )
    {
	val = (double) ( *(float *) ( (char *) array + offsets[i] ));
	if (val != TOOBIG) 
	{
	    x[np] = (double) i;
	    y[np] =  val;
	    np++;
	}
    }
    sig = malloc(num_points*sizeof(double));
    if (sig == NULL)
    {
	fprintf (stderr, "Error allocating sig array\n");
	free(x);
	free(y);
	return (FALSE);
    }
      
    for (i = 0; i < np; i++)  /* Room for expansion */
	sig[i]= 1.0;
      
    r = malloc (num_points*sizeof(double));
    if (r == NULL)
    {
	fprintf (stderr, "Error allocating r array\n");
	free(x);
	free(y);
	free(sig); 
	return (FALSE);
    }
    /* Subtract background and fix sign */
    for (i = 0; i < np; i++) y[i] = sign * (y[i] - background);

    /* Estimate Parameters */
    ymax = -1e-30; xmax = -1e-30; xmin = 1e30;
    for (i = 0; i < np; i++)
    {
	if (y[i]>ymax) ymax = y[i];
	if (x[i]>xmax) xmax = x[i];
	if (x[i]<xmin) xmin = x[i];
    }
    par[2] = xcentre;
    par[1] = ymax;
    par[0] = (xmax-xmin)/3; /* 1/3 of given box */
    
    /*fprintf (stderr, "gauss...1  %f\n", par[0]);*/
    /* Just fit 2 pars out of 3 */
    ierr = marquardt (np, 2, x, y, sig, par, r, a); 
    *fwhm = (fabs(*par))*2.0*sqrt(log(2.0)); /* Ret Values */
    /*fprintf (stderr, "gauss 2...%d  %f\n", ierr,par[0]);*/
    *peak_height = sign*(*(par+1))+background; 
    
    /* Magnitude */
    pi = 4.0*atan2(1.0,1.0);    
    *magnitude = 30.0 - 2.5*log10(pi * (*(par+1)) * fabs(*par)*fabs(*par));

#if 0  /*  Doesn't actually do anything  */
    /* Add background back */
    for (i = 0; i < np; i++) y[i] = y[i] * sign + background;   
#endif  
    free (x); free (y); free (sig); free (r); 
    return (ierr == 0) ? TRUE : FALSE;
}   /*  End Function ds_fitgauss  */

/*EXPERIMENTAL_FUNCTION*/
void ds_gausscurve (double *array, CONST uaddr *offsets,
                    unsigned int num_points, double xcentre,
                    double peak_height, double fwhm,
		    double background, flag inverted)
/*  [SUMMARY] Write gaussian curve to an array.
    [PURPOSE] This routine will a gaussian curve to a 1-dimensional array. This
    array may be used for plotting over real data to which a gaussian has been
    fitted with [<ds_fitgauss>].
    <array> The 1-dimensional array of data.
    <offsets> The array of byte offsets for the array data. If this is NULL,
    the array is assumed to be contiguous.
    <num_points> The length of the array.
    <xcentre> The centre of the profile. If this is 0 it is assumed the profile
    is radial (i.e. the centre is at 0).
    <peak_height> The height of the fitted peak.
    <fwhm> The Full Width Half Maximum of the fit.
    <background> The background level.
    <inverted> If TRUE the object values are inverted (i.e. is negative) else
    the object is positive.
    [RETURNS] Nothing.
*/
{
    int i;
    double a = 2.0 * sqrt ( 2.0 * log (2.0) );
    double arg;

    for (i = 0; i < num_points; i++)
    {
	arg = ( (double) i - xcentre ) / (fwhm / a);
	arg = (peak_height - background) * exp (-0.5 * arg * arg) + background;
	if (offsets == NULL) array[i] = arg;
	else *(double *) ( (char *) array + offsets[i] ) = arg;
    }
}   /*  End Function ds_gausscurve  */
 

/* Malloc 2D ptr array e.g. a[nx][ny] */

static double **malloc2D (int nx, int ny)
{
    int i;
    double **p;

    p = malloc( nx*sizeof(double*) ); /* 1D array of ptrs p[i] */
    if (p==NULL)
	return NULL;
    for (i=0;i<nx;i++) {
	p[i] = malloc( ny*sizeof(double) );
	if (p[i] == NULL) {
	    free(p);
	    return(NULL);
	}
    }
    return p;
}   /*  End Function malloc2D  */

/* Free a 2D ptr array */

static void free2D (double **p, int nx, int ny)
{
    int i;

    for (i=0;i<nx;i++)
	free(p[i]);

    free(p);
}   /*  End Function free2D  */


/*    ======================================================================== */

static void seta (int npar, int npoints, double a[NPAR][NPAR], double **d,
		  double *sig)
{
    int i, j, k;

    for (i = 0; i < npar; i++)
    {
	for (j = 0; j < npar; j++)
	{
	    a[i][j] = 0.0;
	    for (k = 0; k < npoints; k++) 
		a[i][j] += d[k][i] * d[k][j] / sig[k];
	}
    }
}   /*  End Function seta  */

static void setb (int npar, int npoints, double b[NPAR], double **d, 
		  double *y, double *yfit, double *sig)
{
    int i, k;

    for(i = 0; i < npar; i++) { 
	b[i] = 0.0;
	for (k = 0; k < npoints; k++)
	    b[i] += d[k][i] * (y[k] - yfit[k]) / sig[k];
    }
}   /*  End Function setb  */

/* ========================================================================

     returns the fitted function in yfit and the partial derivatives in d

     here f = b exp ( - ((x - c)/a)^2)
*/

static void funct (int npoints, int npar, double *x, double *yfit,
		   double **d, double par[NPAR])
{
    int i;
    double a,b,c,arg;

    a=par[0]; b=par[1]; c=par[2];
    for (i = 0; i < npoints; i++) {
	arg = (x[i]-c) /a; arg *= arg;
	yfit[i] = b * exp(-arg);
	d[i][0] = 2.0 * arg * yfit[i] / a;
	d[i][1] = yfit[i] / b;
	d[i][2] = 2.0 * arg * yfit[i] / (x[i]-c);
    }
}   /*  End Function funct  */


/* These have been checked */

/* =======================================================================

   LU decomposition of matrix x using Crouts algotithm with partial
   pivoting.

   Returns decomposition in x and permutation of rows in iorder
   
*/

static void decomp (int n, int ndim, double x[NPAR][NPAR], int iorder[NPAR])
{
   int icol, irow, isum, iexc, ipivot;
   double store, xmax, sum, eps=1e-16;
   
   /* initialise iorder */
   
   for (iexc = 1; iexc <= n; iexc++) 
      iorder[iexc-1] = iexc;
      
   /* loop over columns */
   
   for (icol = 1; icol <= n; icol++)
   {
      /* U - matrix */
      if (icol > 1)
      {
         for (irow = 1; irow <= icol - 1; irow++)
	 {
            sum = x[irow-1][icol-1];
            for (isum = 1; isum <= irow-1; isum++) 
		sum -= x[irow-1][isum-1]*x[isum-1][irow-1];
            x[irow-1][icol-1] = sum;
         }
      }
      
      /* L - matrix plus diagonal element of U matrix */
      xmax = 0; ipivot = icol;
      for (irow = icol; irow <= n; irow++)
      {
         sum = x[irow-1][icol-1];
         if (icol > 1)
	 {
            for (isum = 1; isum <= icol-1; isum++) 
               sum -= x[irow-1][isum-1] * x[isum-1][icol-1];
          }
          if (fabs (sum) > xmax)
	  {
             xmax = sum;
             ipivot = irow;
          }
          x[irow-1][icol-1] = sum;
       }
       
       /* if xmax is very small replace by epsilon to avoid
          dividing by zero */
   
       if (fabs (xmax) < eps) 
          x[ipivot-1][icol-1] = eps;
          
       /* Partial Pivoting - look for maximum x(i,j<=i) and interchange 
		   rows if necessary to put this on the diagonal */
		   
       if (ipivot!=icol) {
          iexc = iorder[ipivot-1];
          iorder[ipivot-1] = iorder[icol-1];
          iorder[icol-1] = iexc;
          for (iexc = 1; iexc <= n; iexc++)
	  {
             store = x[ipivot-1][iexc-1];
             x[ipivot-1][iexc-1] = x[icol-1][iexc-1];
             x[icol-1][iexc-1] = store;
          }
       }
       
       /* divide new L-matrix elements by new diagonal element */
       
       if (icol < n)
       {
          for (irow = icol + 1; irow <= n; irow++)
	      x[irow-1][icol-1] /= x[icol-1][icol-1];
       }
    }
}   /*  End Function decomp  */
		   
/* ======================================================================== 

    x[i<=j[]j] is U-matrix from LU-decomposition
    x[i>j][j]  is L-matrix (diagonal elements are unity)
    iorder is the permutation of the rows

    b is the input vector, d is the solution vector
*/

static void lineq (int n, int ndim, double x[NPAR][NPAR], double b[NPAR],
		   double d[NPAR], int iorder[NPAR])
{
   int i,isum;
   double sum;
   
   /* solving X.b = d  ==> (L.U).b = d 
      or	L.(U.b) = d 
      
      first re-order the vector   */
      
   for (i=1; i<=n; i++) 
       d[i-1] = b[iorder[i-1]-1];
      
   /* first find (U.b) */
   
   for(i=2; i<=n; i++)
   {
      sum = d[i-1];
      for (isum = 1; isum <= i - 1; isum++) 
	  sum -= x[i-1][isum-1] * d[isum-1];
      d[i-1] = sum;
   }
   
   /* Now fill out d (solution of X.b) by back substitution */

   d[n-1] /= x[n-1][n-1];
   for (i = n-1; i >= 1; i--)
   {
      sum = d[i-1];
      for (isum = i + 1; isum <= n; isum++) 
	  sum -= x[i-1][isum-1] * d[isum-1];
      d[i-1] = sum / x[i-1][i-1];
   }
   
}   /*  End Function lineq  */
      
      
/* ========================================================================

   My C version of Alison's subroutine to fit a non-linear functions using the 
   Levenberg-Marquardt algorithm 
   
     input: npoints = number of data points
	   npar    = number of parameters in fit
	   par     = initial estimates of parameters
	   sigma   = errors on data (sigma^2)

    output: par    = output parameters
	    r	   = residuals (y(i) - yfit(i))
	    a  = estimated covariance matrix of std errs
		     in fitted params.  
*/
    
    
static int marquardt (int npoints, int npar, double*x, double *y,
		      double* sig, double par[NPAR], double* r,
		      double a[NPAR][NPAR])
{
   int i,k, done, decrease, niter;
   int iorder[NPAR];
   double *yfit, **d, **d2, tmp;
   double  par2[NPAR], delta[NPAR], b[NPAR], aprime[NPAR][NPAR];
   
   double lambda, chisq, chisq2, eps=0.001, lamfac=2.0;
      
   /* Memory allocation */
   
   yfit = malloc( npoints*sizeof(double));
   if (yfit==NULL)
      return(1);
   
   d = malloc2D( npoints, NPAR);
   if (d==NULL) {
      free(yfit); return(1);
   }
   
   d2 = malloc2D( npoints, NPAR);
   if (d2==NULL) {
      free(yfit); free2D(d,npoints,NPAR); return(1);
   }
   
   /* Not enough points */
   
   if (npoints < npar) {
      free(yfit); free2D(d,npoints,NPAR); free2D(d2,npoints,NPAR);
      return(2);
   }
   
   lambda = 0.001; done = 0; decrease = 0; niter = 1;
   
   /* Get the value for the initial fit and the value of the derivatives
   for the current estimate of the parameters */
   
   funct(npoints, npar, x, yfit, d, par);
   
   /* Calculate chi^2 */
   
   chisq = 0;
   for (k=0; k<npoints; k++) {
      tmp = y[k] - yfit[k];
      chisq += tmp*tmp / sig[k];
   }
    
   /*  Set up initial matrices; curvature matrix a and b */
   
   seta(npar,npoints,a,d,sig); 
   setb(npar,npoints,b,d,y,yfit,sig);
   
   /* Main Loop */
   
   while(!done) {
   
      /* Set up a' */
      
      for (k=0; k<npar; k++) {
         for (i=0; i<npar; i++) {
            aprime[i][k] = a[i][k]; 
         }
         aprime[k][k] = a[k][k] * (1.0 + lambda);
      }
      
      /* Solve Matrix Equation: a'.delta = b */
      
      decomp(npar ,npar,  aprime, iorder);
      lineq(npar, npar,  aprime, b, delta, iorder); 
      
      /* Increment parameters in work space and evaluate new chisq */
      
      for (i=0; i<NPAR; i++){
         if (i<npar) 
            par2[i] = par[i] + delta[i];
         else
            par2[i] = par[i];
      }
         
      funct(npoints, npar, x, r, d2, par2);
      chisq2 = 0.0;
      
      for (k=0; k<npoints; k++) {
            tmp = y[k] - r[k];
            chisq2 += tmp*tmp / sig[k];
      }      
      
      /* SUccess ? */ 
      
      /* printf ("diag: %lf %lf %lf\n", chisq, chisq2, 
         fabs(((chisq - chisq2)/chisq))); */
      
      if (chisq2<=chisq) {
      
         /* Test convergence */
         
         if (chisq == chisq2) 
            done = 1;
         if (fabs(((chisq - chisq2)/chisq)) < eps && decrease ) 
            done = 1;
         for (k=0; k<npoints; k++) 
            yfit[k] = r[k];
                
         for (i=0; i<NPAR; i++)  {
            par[i] = par2[i];
            for (k=0; k<npoints; k++) {    
              d[k][i] = d2[k][i];
            }
         }
          
         if (!done) {
            seta(npar,npoints,a,d,sig); setb(npar,npoints,b,d,y,yfit,sig);
            chisq = chisq2; lambda = lambda / lamfac; decrease = 1;
         }
      }
      else {
         
         decrease = 0;
         lambda *= lamfac;
      }
      niter++;
      
      /* Failure to converge after numerous iterations */
      
      if (niter>MAXITER) {
         free(yfit); free2D(d,npoints,NPAR); free2D(d2,npoints,NPAR);
         return(3);
      }
   }      
         
   /* Success!!! - compute residual and covariance matrix then return
 
      first calculating inverse of aprime */
 
    for (i=0; i<npar; i++)
       delta[i] = 0.0;
    for (i=0; i<npar; i++) {
       delta[i] = 1.0;
       lineq (npar, NPAR, aprime, delta, b, iorder);
       for (k=0; k<npar; k++) 
          a[k][i] = b[k];
    }
 
    for (k=0; k<npoints; k++)
       r[k] = y[k] - yfit[k];
 
    free(yfit); free2D(d,npoints,NPAR); free2D(d2,npoints,NPAR);
 
    return(0);
}   /*  End Function marquardt  */
