/*LINTLIBRARY*/
/*  misc.c

    This code provides miscellaneous data structure manipulation routines.

    Copyright (C) 1992-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the miscellaneous utility routines for the
    general data structure supported in Karma.


    Written by      Richard Gooch   2-NOV-1996

    Updated by      Richard Gooch   2-JAN-1997: Also recognise "METR/SEC" in
  <ds_format_unit> function.

    Updated by      Richard Gooch   31-JAN-1997: Created <ds_add_named_object>
  and <ds_get_named_object> functions.

    Updated by      Richard Gooch   13-APR-1997: Changed interface to
  <ds_format_unit> function.

    Updated by      Richard Gooch   16-APR-1997: Fixed bug in <ds_format_unit>
  function where unit format was not initialised for default case.

    Updated by      Richard Gooch   10-JUN-1997: Created experimental
  <ds_remove_named_elements> function.

    Updated by      Richard Gooch   19-JUN-1997: Supported "ANGLE" unit in
  <ds_format_value>.

    Updated by      Richard Gooch   21-JUN-1997: Changed default format from
  "%.2f" to "%.2e".

    Updated by      Richard Gooch   24-JUL-1997: Switched to %g formatting for
  unknown unit.

    Updated by      Richard Gooch   4-AUG-1997: Supported "JY/PIX" unit.

    Updated by      Richard Gooch   6-AUG-1997: Supported "JY" unit.

    Updated by      Richard Gooch   14-AUG-1997: Completed switch to %g
  formatting for unknown unit. Some places were missed 24-JUL-1997.

    Updated by      Richard Gooch   24-NOV-1997: Tried a few experimental
  hacks to make function pointer passing more C++ friendly.

    Updated by      Richard Gooch   9-MAY-1998: Added CONST to <<multi_desc>>
  parameter to <ds_get_named_object> function.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   16-JUN-2000: Added <<value>> parameter to
  <ds_format_unit>.

    Updated by      Richard Gooch   10-JUL-2000: Optionally use uJy/Beam units
  in <ds_format_unit>.

    Last updated by Richard Gooch   22-SEP-2000: Added <<format_value>>
  parameter to <ds_format_value>.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#define KARMA_DS_MISC_C
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_j.h>


/*EXPERIMENTAL_FUNCTION*/
void ds_format_unit (char unit[STRING_LENGTH], char format[STRING_LENGTH],
		     double *scale, CONST char *value_name, double value)
/*  [SUMMARY] Format a unit string.
    <unit> The unit string will be written here.
    <format> The formatting string is written here.
    <scale> The scale value to apply to data to convert to displayed units will
    be written here. If this is NULL nothing is written here.
    <value_name> The name of the value to format.
    <value> The (scaled) value. Used for beautifying the unit name.
    [RETURNS] Nothing.
*/
{
    double multiplier = 1.0;
    static char function_name[] = "ds_format_unit";

    if ( (unit == NULL) || (value_name == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    strcpy (format, "%.2f");
    if ( (strncmp (value_name, "M/S", 3) == 0) ||
	 (strncmp (value_name, "METR/SEC", 8) == 0) )
    {
	if (fabs (value) < 1e3) strcpy (unit, "m/s");
	else
	{
	    strcpy (unit, "km/s");
	    multiplier = 1e-3;
	}
    }
    else if (strncmp (value_name, "KM/S", 4) == 0)
    {
	if (fabs (value) < 1.0)
	{
	    strcpy (unit, "m/s");
	    multiplier = 1e+3;
	}
	else strcpy (unit, "km/s");
    }
    else if (strncmp (value_name, "JY/BEAM", 7) == 0)
    {
	if (fabs (value) < 1e-6)
	{
	    strcpy (unit, "nJy/Beam");
	    multiplier = 1e+9;
	}
	else if (fabs (value) < 1e-3)
	{
	    strcpy (unit, "uJy/Beam");
	    multiplier = 1e+6;
	}
	else if (fabs (value) < 1.0)
	{
	    strcpy (unit, "mJy/Beam");
	    multiplier = 1e+3;
	}
	else strcpy (unit, "Jy/Beam");
    }
    else if (strncmp (value_name, "JY/PIX", 7) == 0)
    {
	strcpy (unit, "mJy/Pixel");
	multiplier = 1e+3;
    }
    else if (strncmp (value_name, "JY", 7) == 0)
    {
	strcpy (unit, "mJy");
	multiplier = 1e+3;
    }
    else if (strncmp (value_name, "FREQ", 4) == 0)
    {
	strcpy (unit, "MHz");
	multiplier = 1e-6;
    }
    else if (strcmp (value_name, "HZ") == 0)
    {
	strcpy (unit, "MHz");
	strcpy (format, "%.3f");
	multiplier = 1e-6;
    }
    else if (strncmp (value_name, "FELO", 4) == 0)
    {
	strcpy (unit, "km/s");
	multiplier = 1e-3;
    }
    else if (strncmp (value_name, "VELO", 4) == 0)
    {
	strcpy (unit, "km/s");
	multiplier = 1e-3;
    }
    else if (strcmp (value_name, "BEAM") == 0)
    {
	strcpy (unit, "Beam Number");
	strcpy (format, "%.0f");
    }
    else
    {
	strcpy (unit, value_name);
	strcpy (format, "%g");
    }
    if (scale != NULL) *scale = multiplier;
}   /*  End Function ds_format_unit  */

/*EXPERIMENTAL_FUNCTION*/
void ds_format_value (char string[STRING_LENGTH], double value,
		      CONST char *value_name, double scale, double offset,
		      double format_value,
		      CONST packet_desc *pack_desc, CONST char *packet)
/*  [SUMMARY] Format a data value into a string.
    <string> The string to write to.
    <value> The (raw) value to format.
    <value_name> The name of the value to format.
    <scale> The scale value to apply to the data. If this is TOOBIG and
    <<pack_desc>> and <<packet>> are non-NULL, the routine uses the scale and
    offset attached to the data. See [<ds_get_data_scaling>] for details.
    <offset> The offset to apply after scaling the data.
    <format_value> The (raw) value to use for beautifying the unit name.
    <pack_desc> The packet descriptor containing associated data scaling.
    <packet> The packet containing associated data scaling.
    [RETURNS] Nothing.
*/
{
    double scaled_value, unit_scale, int_raw;
    char unit_string[STRING_LENGTH], unit_format[STRING_LENGTH];
    char format[STRING_LENGTH];
    static char function_name[] = "ds_format_value";

    if ( (string == NULL) || (value_name == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if ( DOUBLE_IS_BLANK (value) )
    {
	sprintf (string, "value: blank");
	return;
    }
    if ( DOUBLE_IS_BLANK (scale) )
    {
	if ( (pack_desc != NULL) && (packet != NULL) )
	{
	    ds_get_data_scaling (value_name, pack_desc, packet,
				 &scale, &offset);
	}
	else
	{
	    scale = 1.0;
	    offset = 0.0;
	}
    }
    /*  Compute value  */
    scaled_value = scale * value + offset;
    if (strcmp (value_name, "Data Value") == 0)
    {
	/*  No useful unit name  */
	if ( (scale == 1.0) && (offset == 0.0) )
	{
	    sprintf (string, "value: %g", value);
	}
	else sprintf (string, "raw: %g  sc: %g", value, scaled_value);
	return;
    }
    /*  Have a useful value name: first check if it should be scaled and the
	units fiddled to make (some) users happier  */
    ds_format_unit (unit_string, unit_format, &unit_scale, value_name,
		    format_value * scale + offset);
    if ( (scale == 1.0) && (offset == 0.0) )
    {
	if (strcmp (value_name, "ANGLE") == 0)
	{
	    int degrees, minutes;
	    double seconds;

	    seconds = scaled_value * unit_scale + 1e-10;
	    degrees = seconds;
	    seconds -= degrees;
	    seconds *= 60.0;
	    minutes = seconds;
	    seconds -= minutes;
	    seconds *= 60.0;
	    sprintf (string, "Offset: %.2dd %.2dm %5.2fs",
		     degrees, minutes, seconds);
	}
	else
	{
	    sprintf (format, "value: %s %s", unit_format, unit_string);
	    sprintf (string, format, scaled_value * unit_scale);
	}
    }
    else
    {
	int_raw = floor (value);
	if (int_raw == value)
	{
	    sprintf (format, "raw: %%ld  sc: %s %s", unit_format, unit_string);
	    sprintf (string, format, (long)int_raw, scaled_value * unit_scale);
	}
	else
	{
	    sprintf (format, "raw: %%g  sc: %s %s", unit_format, unit_string);
	    sprintf (string, format, value, scaled_value * unit_scale);
	}
    }
}   /*  End Function ds_format_value  */

/*UNPUBLISHED_FUNCTION*/
flag ds_add_named_object (multi_array *multi_desc, unsigned int index,
			  CONST char *name,
			  void *object, void (*destroy_func) ())
/*  [SUMMARY] Add a named object to a data structure.
    <multi_desc> The multi_array descriptor.
    <index> The index of the general data structure to associate the object to.
    If this is equal to or greater than the number of general data structures,
    the object is associated with the entire descriptor.
    <name> The name of the object.
    <object> The object. This may be NULL.
    <destroy_func> The object destroy function. This is called automatically
    when the descriptor is destroyed. The prototype function is
    [<ds_PROTO_object_destroy_func>].
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KJoinedPairList jlist;
    static char function_name[] = "ds_add_named_object";

    if ( (multi_desc == NULL) || (name == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    jlist = (index < multi_desc->num_arrays) ?
	multi_desc->jlists[index] : multi_desc->jlist;
    if (j_put_pair (jlist, name, strlen (name) + 1,
		    object, 0, ( void *(*) () ) NULL, destroy_func,
		    KJ_REPLACEMENT_POLICY_NEW, FALSE) == NULL)
    {
	m_error_notify (function_name, "adding pair");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function ds_add_named_object  */

/*UNPUBLISHED_FUNCTION*/
flag ds_get_named_object (CONST multi_array *multi_desc, unsigned int index,
			  CONST char *name, void **object)
/*  [SUMMARY] Get a named object from a data structure.
    <multi_desc> The multi_array descriptor.
    <index> The index of the general data structure to associate the object
    with. If this is equal to or greater than the number of general data
    structures, the object is associated with the entire descriptor.
    <name> The name of the object.
    <object> The object is written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KJoinedPairList jlist;
    static char function_name[] = "ds_get_named_object";

    if ( (multi_desc == NULL) || (name == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    jlist = (index < multi_desc->num_arrays) ?
	multi_desc->jlists[index] : multi_desc->jlist;
    if (j_get_pair (jlist, name, object) == NULL) return (FALSE);
    return (TRUE);
}   /*  End Function ds_get_named_object  */

/*EXPERIMENTAL_FUNCTION*/
flag ds_remove_named_elements (packet_desc *pack_desc, char **packet,
			       CONST char **element_names)
/*  [SUMMARY] Remove a list of named elements from a packet.
    <pack_desc> The packet descriptor.
    <packet> A pointer to the packet pointer.
    <element_names> The NULL-terminated array of element names to be removed.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag keep;
    unsigned int count, num_new, new_size, elem_size, elem_type;
    FString fstring;
    char *new_packet, *inp_ptr, *out_ptr, *string_ptr;
    CONST char **name;
    char **new_desc;
    unsigned int *new_types;
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "ds_remove_named_elements";

    /*  First find how many elements in new packet  */
    new_size = 0;
    for (count = 0, num_new = 0; count < pack_desc->num_elements; ++count)
    {
	keep = FALSE;
	if ( ds_element_is_named (pack_desc->element_types[count]) )
	{
	    /*  Search for element in list  */
	    for (name = element_names; *name != NULL; ++name)
	    {
		if (strcmp (*name, pack_desc->element_desc[count]) == 0) break;
	    }
	    if (*name == NULL) keep = TRUE;
	}
	else keep = TRUE;
	if (keep)
	{
	    ++num_new;
	    elem_type = pack_desc->element_types[count];
	    new_size += host_type_sizes[elem_type];
	}
    }
    if ( ( new_desc = (char **) m_alloc (sizeof *new_desc * num_new) )
	 == NULL )
    {
	m_error_notify (function_name, "new element descriptor array");
	return (FALSE);
    }
    if ( ( new_types = (unsigned int *) m_alloc (sizeof *new_types * num_new) )
	 == NULL )
    {
	m_error_notify (function_name, "new element type array");
	return (FALSE);
    }
    if ( ( new_packet = m_alloc (new_size) ) == NULL )
    {
	m_error_notify (function_name, "new packet");
	return (FALSE);
    }
    /*  Copy over some elements and free others  */
    inp_ptr = *packet;
    out_ptr = new_packet;
    for (count = 0, num_new = 0; count < pack_desc->num_elements;
	 ++count, inp_ptr += elem_size)
    {
	keep = FALSE;
	elem_type = pack_desc->element_types[count];
	elem_size = host_type_sizes[elem_type];
	if ( ds_element_is_named (pack_desc->element_types[count]) )
	{
	    /*  Search for element in list  */
	    for (name = element_names; *name != NULL; ++name)
	    {
		if (strcmp (*name, pack_desc->element_desc[count]) == 0) break;
	    }
	    if (*name == NULL) keep = TRUE;
	}
	else keep = TRUE;
	if (keep)
	{
	    new_desc[num_new] = pack_desc->element_desc[count];
	    new_types[num_new] = pack_desc->element_types[count];
	    m_copy (out_ptr, inp_ptr, elem_size);
	    out_ptr += elem_size;
	    ++num_new;
	}
	else
	{
	    switch (elem_type)
	    {
	      case K_VSTRING:
		m_copy ( (char *) &string_ptr, inp_ptr, elem_size );
		m_free (string_ptr);
		break;
	      case K_FSTRING:
		m_copy ( (char *) &fstring, inp_ptr, elem_size );
		m_free (fstring.string);
		break;
	    }
	}
    }
    pack_desc->num_elements = num_new;
    m_free ( (char *) pack_desc->element_types );
    pack_desc->element_types = new_types;
    m_free ( (char *) pack_desc->element_desc );
    pack_desc->element_desc = new_desc;
    m_free (*packet);
    *packet = new_packet;
    return (TRUE);
}   /*  End Function ds_remove_named_elements  */
