/*LINTLIBRARY*/
/*  filter.c

    This code provides data structure IO routines.

    Copyright (C) 1998-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for writing and reading
    the general data structure supported in Karma to and from channel objects.


    Written by      Richard Gooch   11-MAY-1998

    Updated by      Richard Gooch   12-MAY-1998: Supported shared anon. mmapped
  regions.

    Updated by      Richard Gooch   19-MAY-1998: Fixed bug in function
  <dsrw_filter_process>: only last bunch of packets was written.

    Updated by      Richard Gooch   21-APR-1999: Made use of new function
  <dsrw_write_history>.

    Last updated by Richard Gooch   9-JUN-2000: Removed unused variable.


*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <karma_dsrw.h>
#include <karma_pio.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_m.h>


#define BUFFER_BYTES 1048576
#define ARRAY_BOUNDARY (unsigned int) 16  /*  Copied from  main.c  */


/*  Private data  */
static char dummy_array;


/*  Private functions  */
STATIC_FUNCTION (flag _dsrw_open,
		 (Channel channel, CONST multi_array *multi_desc) );
STATIC_FUNCTION (flag _dsrw_write_tail,
		 (Channel channel, multi_array *multi_desc) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *dsrw_filter_setup (Channel channel, CONST array_desc *arr_desc,
				CONST char *infile, CONST char *outfile)
/*  [SUMMARY] Set up a Karma data filter.
    <channel> The channel to write to.
    <arr_desc> The array descriptor.
    <infile> The input filename.
    <outfile> The output Karma filename.
    [RETURNS] A multi_array pointer on success, else NULL.
*/
{
    array_pointer arrayp;
    multi_array *multi_desc;
    char txt[STRING_LENGTH];
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    extern char module_lib_version[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];
    static char function_name[] = "dsrw_filter_setup";

    m_clear (&arrayp, sizeof arrayp);
    if ( !ch_test_for_local_connection (channel) ||
	 ( !ds_alloc_shm_array (&arrayp, arr_desc, FALSE, FALSE) &&
	   !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp, arr_desc, FALSE,
				 FALSE) ) )
    {
	/*  Failure allocating shared memory: create a dummy array for
	    streaming purposes  */
	m_clear (&arrayp, sizeof arrayp);
	arrayp.array = &dummy_array;
    }
    if ( ( multi_desc = ds_easy_alloc_array_from_array_desc (arr_desc,
							     &arrayp, FALSE) )
	 == NULL )
    {
	fprintf (stderr, "%s: error allocating descriptors\n", function_name);
	if (arrayp.free != NULL) (*arrayp.free) (arrayp.info, &arrayp);
	return (NULL);
    }
    /*  Append a bit of history  */
    sprintf (txt, "%s: Module version %s  Karma v%s  compiled with v%s",
	     module_name, module_version_date,
	     karma_library_version, module_lib_version);
    ds_history_append_string (multi_desc, txt, TRUE);
    sprintf (txt, "%s: infile: %s  outfile: %s",
	     module_name, infile, outfile);
    ds_history_append_string (multi_desc, txt, TRUE);
    return (multi_desc);
}   /*  End Function dsrw_filter_setup  */

/*EXPERIMENTAL_FUNCTION*/
flag dsrw_filter_process (Channel channel, multi_array *multi_desc,
			  flag (*read_func) (), void *info)
/*  [SUMMARY] Read in data and write in Karma format.
    <channel> The channel to write the Karma data to.
    <multi_desc> The multi_array descriptor to write.
    <func> The function to call when new data should be read. The prototype
    function is [<dsrw_PROTO_read_func>].
    <info> An arbitrary pointer which will be passed to <<func>>.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    array_pointer arrayp;
    unsigned long array_size, packet_size, buffer_size,num_values;
    array_desc *arr_desc;
    char buffer[BUFFER_BYTES];
    /*static char function_name[] = "dsrw_filter_process";*/

    m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
    arr_desc = (array_desc *) multi_desc->headers[0]->element_desc[0];
    array_size = ds_get_array_size (arr_desc);
    packet_size = ds_get_packet_size (arr_desc->packet);
    buffer_size = BUFFER_BYTES / packet_size;
    if (arrayp.array != &dummy_array)
    {
	/*  Must be in shared memory: read the all data in first  */
	if ( !(*read_func) (arrayp.array, array_size, info) ) return (FALSE);
	/*  Now transfer it in one go  */
	if ( !dsrw_write_multi (channel, multi_desc) ) return (FALSE);
    }
    else
    {
	/*  Need to write in a streaming mode. First write the Karma "header"
	 */
	if ( !_dsrw_open (channel, multi_desc) ) return (FALSE);
	/*  Now read data values from the input and write them out in blocks */
	for (num_values = 0; array_size > 0; array_size -= num_values)
	{
	    num_values = (array_size > buffer_size) ? buffer_size : array_size;
	    if ( !(*read_func) (buffer, num_values, info) ) return (FALSE);
	    if ( !dsrw_write_packets (channel, arr_desc->packet, buffer,
				      num_values) ) return (FALSE);
	}
	if ( !_dsrw_write_tail (channel, multi_desc) ) return (FALSE);
    }
    return ch_flush (channel);
}   /*  End Function dsrw_filter_process  */


/*  Private functions follow  */

static flag _dsrw_open (Channel channel, CONST multi_array *multi_desc)
/*  [SUMMARY] Write part of a Karma data stucture.
    [PURPOSE] This routine will write the multi_array descriptor header, the
    first packet descriptor and some array padding information for a Karma data
    structure. The array data may then be written.
    <channel> The channel to write to.
    <multi_desc> The multi_array descriptor.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int bytes_to_pad;
    unsigned long read_pos;
    unsigned long write_pos;

    /*  Write magic string and version number  */
    if ( !dsrw_write_multi_header (channel, multi_desc) ) return (FALSE);
    if ( !dsrw_write_packet_desc (channel, multi_desc->headers[0]) )
	return (FALSE);
    if ( ch_test_for_local_connection (channel) )
    {
	/*  Write a special value saying the array was sent the easy way  */
	if ( !pio_write32 (channel, 0) )
	{
	    fprintf (stderr, "Error writing transfer type\n");
	    return (FALSE);
	}
    }
    else
    {
	/*  Array will be sent the hard way: first do some padding  */
	if ( !ch_tell (channel, &read_pos, &write_pos) )
	{
	    fprintf (stderr, "Error getting channel position\n");
	    return (FALSE);
	}
	/*  Add 4 bytes for pad size  */
	write_pos += 4;
	bytes_to_pad = ARRAY_BOUNDARY - write_pos % ARRAY_BOUNDARY;
	if ( !pio_write32 (channel, bytes_to_pad) )
	{
	    fprintf (stderr, "Error writing pad size\n");
	    return (FALSE);
	}
	if (bytes_to_pad > 0)
	{
	    if (ch_fill (channel, bytes_to_pad, 0) < bytes_to_pad)
	    {
		fprintf (stderr, "Error padding: %u bytes\t%s\n",
			 bytes_to_pad, ERRSTRING);
		return (FALSE);
	    }
	}
    }
    return (TRUE);
}   /*  End Function _dsrw_open  */

static flag _dsrw_write_tail (Channel channel, multi_array *multi_desc)
/*  [SUMMARY] Write the tail part of a Karma data structure.
    <channel> The channel object to write to.
    <multi_desc> The multi_array descriptor.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count, elem_type;
    packet_desc *pack_desc;
    char *packet;
    extern char host_type_sizes[NUMTYPES];

    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0] + host_type_sizes[K_ARRAY];
    for (count = 1; count < pack_desc->num_elements; ++count)
    {
	/*  Write element   */
        elem_type = pack_desc->element_types[count];
        dsrw_write_element (channel, elem_type,
			    pack_desc->element_desc[count],
			    packet);
        packet += host_type_sizes[elem_type];
    }
    return dsrw_write_history (channel, multi_desc->first_hist);
}   /*  End Function _dsrw_write_tail  */
