/*LINTLIBRARY*/
/*  main.c

    This code provides data structure IO routines.

    Copyright (C) 1992-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for writing and reading
    the general data structure supported in Karma to and from channel objects.


    Written by      Richard Gooch   13-SEP-1992

    Updated by      Richard Gooch   9-NOV-1992

    Updated by      Richard Gooch   31-DEC-1992: Added fast local copy of
  arrays when using local connections.

    Updated by      Richard Gooch   2-JAN-1992: Added memory mapped array
  support.

    Updated by      Richard Gooch   3-JAN-1992: Added tiled array support and
  obsoleted  ARRAYP  .

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   13-FEB-1993: Added support for old ARRAYP
  in  dsrw_write_element  and changed to 64 bit dimension lengths and list
  lengths.

    Updated by      Richard Gooch   23-FEB-1993: Removed diagnostics when
  transferring arrays between nodes on VX/MVX. Hardware problem fixed.

    Updated by      Richard Gooch   1-MAR-1993: Fixed bug in  dsrw_read_element
  changed  K_BYTE_SWAPPER  back to  BYTE_SWAPPER  .Problem with sed scripts to
  change from  BYTE  to  K_BYTE  .

    Updated by      Richard Gooch   21-MAR-1993: Added clearing of padding
  array in  dsrw_write_array  to guarantee identical files for identical
  structures.

    Updated by      Richard Gooch   25-MAR-1993: Took account of changes to
  list_header  structure.

    Updated by      Richard Gooch   11-APR-1993: Fixed bug in  dsrw_write_list

    Updated by      Richard Gooch   16-MAY-1993: Moved reading and writing of
  arraynames strings after all structures. Should not affect anyone as existing
  files have only needed one general data structure per file.

    Updated by      Richard Gooch   17-MAY-1993: Added support for string data
  types.

    Updated by      Richard Gooch   5-AUG-1993: Added test for arrays not
  aligned on a 16 byte boundary when transfering between VX/MVX nodes.

    Updated by      Richard Gooch   27-AUG-1993: Fixed bug in  dsrw_read_list
  which could cause problems when reading in a zero length list on machines
  with network data format.

    Updated by      Richard Gooch   14-SEP-1993: Fixed bug in
  dsrw_write_element  which did not write out K_VSTRINGs correctly.

    Updated by      Richard Gooch   1-DEC-1993: Fixed bug in
  dsrw_read_array  which caused reading of memory mapped channels to fail on
  byte swapper machines for byte-type arrays.

    Updated by      Richard Gooch   3-APR-1994: Added more diagnostics when
  reading strings.

    Updated by      Richard Gooch   16-APR-1994: Chopped one line of code for
  the VX/MVX because of some obscure bug (not mine!).

    Updated by      Richard Gooch   21-MAY-1994: Added  #include <karma_ch.h>

    Updated by      Richard Gooch   7-JUL-1994: Fixed bug in  dsrw_read_array
  which used sizeof (int) to find size of pointers. Problem on Alpha.
  
    Updated by      Richard Gooch   6-AUG-1994: Added support for more data
  types in  dsrw_read_element  for non matching sizes platforms.

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/dsrw/main.c

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   23-FEB-1995: Made use of
  <ds_can_transfer_packet_as_block> routine.

    Updated by      Richard Gooch   9-APR-1995: Fixed bug in transferring
  non-byte complex values on little-endian machines: whole value was byte
  swapped rather than separately byte-swapping the real and imaginary
  components.

    Updated by      Richard Gooch   5-MAY-1995: Placate SGI compiler.

    Updated by      Richard Gooch   22-AUG-1995: Made <dsrw_write_element> and
  <dsrw_read_element> more aware of alignment problems (hopefully fully aware)

    Updated by      Richard Gooch   11-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   30-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   3-JUN-1996: Took account of new fields in
  dimension descriptor for first and last co-ordinate.

    Updated by      Richard Gooch   16-JUN-1996: Optimised <dsrw_read_array>
  when reading single-valued arrays on little-endian machines under certain
  conditions. 

    Updated by      Richard Gooch   28-JUN-1996: Created <dsrw_read_packets>
  and used in <dsrw_read_array> and <dsrw_read_list>. 

    Updated by      Richard Gooch   29-JUN-1996: Created <dsrw_write_packets>
  and used in <dsrw_write_array> and <dsrw_write_list>. <dsrw_write_list> had
  a bug which could cause the fragmented section of list not to be written.

    Updated by      Richard Gooch   6-AUG-1996: Made use of
  <ds_history_append_string> routine.

    Updated by      Richard Gooch   10-AUG-1996: All write routines now return
  a status, rather than calling exit(3).

    Updated by      Richard Gooch   16-AUG-1996: Created
  <dsrw_write_multi_header>.

    Updated by      Richard Gooch   27-SEP-1996: Do not write empty history
  strings otherwise the reader will be fooled into thinking no more history.

    Updated by      Richard Gooch   7-JAN-1997: os.h no longer defines
  <network_type_bytes>.

    Updated by      Richard Gooch   23-SEP-1997: Optimised sending arrays over
  local connections: always write raw data.

    Updated by      Richard Gooch   21-FEB-1998: Switched to array_pointer type

    Updated by      Richard Gooch   24-FEB-1998: Switch to new array_pointer
  which has free function. Moved local array transfer before padding. Added
  support for transferring via SHM.

    Updated by      Richard Gooch   30-APR-1998: Improved diagnostics in
  <receive_array_local>.

    Updated by      Richard Gooch   11-MAY-1998: Fixed bug in fast array
  transfer: sending a response back over a pipe is not possible.

    Updated by      Richard Gooch   16-JAN-1999: Moved notice of reading tiled
  array to when array data is read, and included offset.

    Updated by      Richard Gooch   8-FEB-1999: Improved diagnostics in
  <dsrw_read_flag> and <dsrw_write_flag>.

    Updated by      Richard Gooch   21-APR-1999: Created experimental function
  <dsrw_write_history>.

    Updated by      Richard Gooch   22-JUN-1999: Work around bug in
  LinuxThreads where SIGUSR1 and SIGUSR2 are stolen.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   11-SEP-1999: Improved debugging in
  <dsrw_read_packets>.

    Last updated by Richard Gooch   9-JUN-2000: Removed unused variables.


*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <karma.h>
#include <karma_dsrw.h>
#include <karma_pio.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_r.h>
#include <os.h>


#define MAGIC_STRING_LENGTH (unsigned int) 16
#define VERSION_NUMBER (unsigned long) 0
#define ARRAY_BOUNDARY (unsigned int) 16
#define VXMVX_REMOTE_MEMCPY_BOUNDARY 16

#define ARRAYP        6    /*Obsolete*/

/*  The following constants define the various control values used when
    transmitting atomic arrays over local connections.
*/
#define FA_NORMAL  0  /*  Cannot change this because people rely on it in their
			  data filters  */
#define FA_VX    1
#define FA_SHM   2
#define FA_MMAP  3

#define FA_RESPONSE_OK      0x01
#define FA_RESPONSE_FAILED  0x02

/*  LinuxThreads steals SIGUSR1 and SIGUSR2  */
#if defined(OS_Linux) && defined(_POSIX_THREADS)
#  define SIG_FA_OK      SIGINT
#  define SIG_FA_FAILED  SIGTERM
#else
#  define SIG_FA_OK      SIGUSR1
#  define SIG_FA_FAILED  SIGUSR2
#endif

/*  The following describes what is written by the transmitter when an atomic
    array is sent over a local connection:

    FA_NORMAL     :    the array data
    FA_VX         :    sender's task ID
                       sender's parent process ID
                       starting virtual address of array
    FA_SHM        :    the SHM id
                       the offset from the base
		       the PID of the sender


    The following describes what response is sent from the receiving process:

    FA_NORMAL     :    no response is given
    FA_VX         :    TRUE if receiver is happy, FALSE if data must be sent
    FA_SHM        :    SIG_FA_OK if OK, else SIG_FA_FAILED

*/


/*  Private functions  */
STATIC_FUNCTION (flag _dsrw_transmit_array_local,
		 (Channel channel, CONST array_pointer *arrayp,
		  unsigned int length) );
STATIC_FUNCTION (flag _dsrw_fa_get_response, (Channel channel) );
STATIC_FUNCTION (flag _dsrw_receive_array_local,
		 (Channel channel, CONST array_desc *arr_desc,
		  char *element, unsigned int length) );
STATIC_FUNCTION (flag read_array_padding, (Channel channel) );
STATIC_FUNCTION (void signal_handler, (int sig) );


/*  Private data  */
static char magic_string[] = "KarmaRHD Version";
static volatile unsigned int fa_response = 0;


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
flag dsrw_write_multi (Channel channel, CONST multi_array *multi_desc)
/*  [SUMMARY] Write a multi_array data structure.
    [PURPOSE] This routine will write a binary multi_array descriptor to a
    channel object. All descriptors and data, as well as history strings are
    written.
    <channel> The channel object.
    <multi_desc> The data structure.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int array_count;
    char *array_name;
    static char function_name[] = "dsrw_write_multi";

    if (channel == NULL) return (TRUE);
    if ( !dsrw_write_multi_header (channel, multi_desc) ) return (FALSE);
    if (multi_desc->num_arrays < 1)
    {
	fprintf (stderr, "No arrays in multi-array header\n");
	a_prog_bug (function_name);
    }
    /*  Write packet descriptors and data for each data structure  */
    for (array_count = 0; array_count < multi_desc->num_arrays; 
	 ++array_count)
    {
	if ( !dsrw_write_packet_desc (channel,
				      multi_desc->headers[array_count]) )
	{
	    return (FALSE);
	}
	if ( !dsrw_write_packet (channel, multi_desc->headers[array_count],
				 multi_desc->data[array_count]) )
	{
	    return (FALSE);
	}
    }
    /*  May need to write arraynames  */
    if (multi_desc->num_arrays > 1)
    {
	if (multi_desc->array_names == NULL)
        {
	    fprintf (stderr, "Multiple arrays and no array of array names\n");
            a_prog_bug (function_name);
        }
        /*  Write array names   */
        for (array_count = 0; array_count < multi_desc->num_arrays;
	     ++array_count)
        {
	    array_name = multi_desc->array_names[array_count];
            if ( (array_name == NULL) || (*array_name == '\0') )
            {
		fprintf (stderr, "Array name must not be a null string\n");
                a_prog_bug (function_name);
            }
	    if ( !pio_write_string (channel, array_name) )
	    {
		return (FALSE);
	    }
        }
    }
    return dsrw_write_history (channel, multi_desc->first_hist);
}   /*  End Function dsrw_write_multi  */

/*PUBLIC_FUNCTION*/
flag dsrw_write_multi_header (Channel channel, CONST multi_array *multi_desc)
/*  [SUMMARY] Write a multi_array data structure header.
    [PURPOSE] This routine will write a binary multi_array descriptor header to
    a channel object. The various top-level packet descriptors, data and
    history are not written.
    <channel> The channel object.
    <multi_desc> The data structure.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "dsrw_write_multi_header";

    if (channel == NULL) return (TRUE);
    if (multi_desc == NULL)
    {
	fprintf (stderr, "NULL descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Write magic string and version number  */
    if (ch_write (channel, magic_string, MAGIC_STRING_LENGTH)
	< MAGIC_STRING_LENGTH)
    {
	fprintf (stderr, "Error writing magic string\t%s\n",
		 ERRSTRING);
	return (FALSE);
    }
    if ( !pio_write32 (channel, VERSION_NUMBER) ) return (FALSE);
    /*  Write number of data structures  */
    if ( !pio_write32 (channel, multi_desc->num_arrays) ) return (FALSE);
    return (TRUE);
}   /*  End Function dsrw_write_multi_header  */

/*PUBLIC_FUNCTION*/
flag dsrw_write_packet_desc (Channel channel, CONST packet_desc *pack_desc)
/*  [SUMMARY] Write a packet descriptor in binary format.
    [PURPOSE] This routine will write the binary format of a packet descriptor
    to a channel object. The routine will also write all lower levels of array
    or linked list descriptors.
    <channel> The channel object.
    <pack_desc> The packet descriptor.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int element_count = 0;
    static char function_name[] = "dsrw_write_packet_desc";

    if (channel == NULL) return (TRUE);
    if (pack_desc == NULL)
    {
	fprintf (stderr, "NULL descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Write number of elements    */
    if ( !pio_write32 (channel, (unsigned long) pack_desc->num_elements ) )
    {
	return (FALSE);
    }
    if (pack_desc->num_elements < 1)
    {
	fprintf (stderr, "No elements in packet descriptor\n");
	a_prog_bug (function_name);
    }
    if (pack_desc->element_types == NULL)
    {
	fprintf (stderr, "Packet descriptor has no array of element types\n");
        a_prog_bug (function_name);
    }
    if (pack_desc->element_desc == NULL)
    {
	fprintf (stderr,
		 "Packet descriptor has no array of element descriptor pointers\n");
        a_prog_bug (function_name);
    }
    /*  Write element descriptors   */
    while (element_count < pack_desc->num_elements)
    {
	/*  Write an element descriptor */
        if ( !dsrw_write_element_desc (channel,
				       pack_desc->element_types[element_count],
				       pack_desc->element_desc[element_count]))
	{
	    return (FALSE);
	}
        ++element_count;
    }
    return (TRUE);
}   /*  End Function dsrw_write_packet_desc  */

/*PUBLIC_FUNCTION*/
flag dsrw_write_element_desc (Channel channel, unsigned int type,
			      CONST char *desc)
/*  [SUMMARY] Write an element descriptor in binary format.
    [PURPOSE] This routine will write the binary format of an element
    descriptor to a channel object. If the element is an array pointer or a
    linked list pointer, that descriptor will also be written.
    <channel> The channel object.
    <type> The type of the element.
    <desc> The element descriptor.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "dsrw_write_element_desc";

    if (channel == NULL) return (TRUE);
    if (desc == NULL)
    {
	fprintf (stderr, "NULL descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (type == ARRAYP)
    {
	fprintf (stderr, "Should not be writing type ARRAYP\n");
	a_prog_bug (function_name);
    }
    /*  Write element type  */
#ifdef RETYPE_NEEDED
    /*  Retyping needed  */
    if ( !pio_write32 (channel, (unsigned long) host_to_net_retype[type]) )
    {
	return (FALSE);
    }
#else
    /*  No retyping needed  */
    if ( !pio_write32 (channel, (unsigned long) type) ) return (FALSE);
#endif

    if ( ds_element_is_named (type) )
    {
	if ( (desc == NULL) || (*desc == '\0') )
	{
	    fprintf (stderr,"Element name must not be a null string\n");
	    a_prog_bug (function_name);
	}
	/*  Write element name  */
	if ( !pio_write_string (channel, desc) ) return (FALSE);
    }
    else
    {
	switch (type)
	{
	  case K_ARRAY:
	    return dsrw_write_array_desc (channel, (array_desc *) desc);
	    /*break;*/
	  case LISTP:
	    return dsrw_write_packet_desc (channel, (packet_desc *) desc);
	    /*break;*/
	}
    }
    return (TRUE);
}   /*  End Function dsrw_write_element_desc */

/*PUBLIC_FUNCTION*/
flag dsrw_write_array_desc (Channel channel, CONST array_desc *arr_desc)
/*  [SUMMARY] Write an array descriptor in binary format.
    [PURPOSE] This routine will write the binary representation of an array
    descriptor to a channel object. The packet descriptor for the array will
    also be written (this is a recursive process).
    <channel> The channel object.
    <arr_desc> The array descriptor.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int dim_count;
    unsigned int level_count;
    unsigned int product;
    static char function_name[] = "dsrw_write_array_desc";

    if (channel == NULL) return (TRUE);
    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Write number of dimensions  */
    if ( !pio_write32 (channel, (unsigned long) arr_desc->num_dimensions) )
    {
	return (FALSE);
    }
    if (arr_desc->num_dimensions < 1)
    {
	fprintf (stderr, "No dimensions in array descriptor\n");
	a_prog_bug (function_name);
    }
    /*  Write tiling information  */
    if (arr_desc->num_levels > 0)
    {
	fprintf (stderr, "Writing tiled array\n");
    }
    if ( !pio_write32 (channel, (unsigned long) arr_desc->num_levels) )
    {
	fprintf (stderr, "Error writing number of tiling levels\n");
	return (FALSE);
    }
    if (arr_desc->dimensions == NULL)
    {
	fprintf (stderr,
		 "Array descriptor has no array of dimension descriptor pointers\n");
        a_prog_bug (function_name);
    }
    /*  Write dimension descriptors */
    for (dim_count = 0; dim_count < arr_desc->num_dimensions; ++dim_count)
    {
	dsrw_write_dim_desc (channel, arr_desc->dimensions[dim_count]);
	/*  Write tile lengths if appropriate  */
	for (level_count = 0, product = 1;
	     level_count < arr_desc->num_levels;
	     ++level_count)
	{
	    if ( !pio_write32(channel, (unsigned long)
			      arr_desc->tile_lengths[dim_count][level_count]) )
	    {
		fprintf (stderr, "Error writing tile length[%u][%u]\n",
			 dim_count, level_count);
		return (FALSE);
	    }
	    product *= arr_desc->tile_lengths[dim_count][level_count];
	}
	/*  Check if tile lengths appropriate  */
	if (product * arr_desc->lengths[dim_count] !=
	    arr_desc->dimensions[dim_count]->length)
	{
	    fprintf (stderr,
		     "Tile product * bottom length (%u * %lu): %lu not equal to length: %lu\n",
		     product, arr_desc->lengths[dim_count],
		     product * arr_desc->lengths[dim_count],
		     arr_desc->dimensions[dim_count]->length);
	    a_prog_bug (function_name);
	}
    }
    /*  Write array packet descriptor   */
    return ( dsrw_write_packet_desc (channel, arr_desc->packet) );
}   /*  End Function dsrw_write_array_desc   */

/*PUBLIC_FUNCTION*/
flag dsrw_write_dim_desc (Channel channel, CONST dim_desc *dimension)
/*  [SUMMARY] Write a dimension descriptor in binary format.
    [PURPOSE] This routine will write the binary representation of a dimension
    descriptor to a channel object.
    <channel> The channel object.
    <dimension> The dimension descriptor.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int coord_count = 0;
    static char function_name[] = "dsrw_write_dim_desc";

    if (channel == NULL) return (TRUE);
    if (dimension == NULL)
    {
	fprintf (stderr, "NULL descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (dimension->name == NULL)
    {
	fprintf (stderr, "Dimension name must not be a null string\n");
        a_prog_bug (function_name);
    }
    if (*dimension->name == '\0')
    {
	fprintf (stderr, "Dimension name must not be an empty string\n");
        a_prog_bug (function_name);
    }
    /*  Write dimension name    */
    if ( !pio_write_string (channel, dimension->name) )
    {
	return (FALSE);
    }
    if (dimension->length < 1)
    {
	fprintf (stderr, "Dimension: \"%s\" has zero length\n",
		 dimension->name);
	a_prog_bug (function_name);
    }
    /*  Write dimension length  */
    if ( !pio_write64 (channel, dimension->length) ) return (FALSE);
    if ( (dimension->first_coord == dimension->last_coord) &&
	 (dimension->length != 1) )
    {
	fprintf (stderr,
		 "Dimension: \"%s\" first_coord is equal to last_coord: %e\n",
		 dimension->name, dimension->first_coord);
	fprintf (stderr, "and length: %lu is not 1\n", dimension->length);
	a_prog_bug (function_name);
    }
    /*  Write REGULAR flag  */
    if ( !dsrw_write_flag (channel,
			   (dimension->coordinates == NULL) ? TRUE : FALSE) )
    {
	return (FALSE);
    }
    if (dimension->coordinates == NULL)
    {
	/*  Write first and last co-ordinates   */
	if ( !pio_write_double (channel, dimension->first_coord) )
	{
	    return (FALSE);
	}
	if ( !pio_write_double (channel, dimension->last_coord) )
	{
	    return (FALSE);
	}
    }
    else
    {
	/*  Write co-ordinates  */
        while (coord_count < dimension->length)
        {
	    if ( !pio_write_double (channel,
				    dimension->coordinates[coord_count]) )
	    {
		return (FALSE);
	    }
	    ++coord_count;
        }
    }
    return (TRUE);
}   /*  End Function dsrw_write_dim_desc  */

/*EXPERIMENTAL_FUNCTION*/
flag dsrw_write_history (Channel channel, CONST history *hist)
/*  [SUMMARY] Write history information.
    <channel> The channel to write to.
    <hist> A pointer to the first history entry.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    for (; hist != NULL; hist = hist->next)
    {
	if (*hist->string == '\0') continue;
	if ( !pio_write_string (channel, hist->string) )
	{
	    fprintf (stderr, "Error writing history string: \"%s\"\n",
		     hist->string);
	    return (FALSE);
	}
    }
    if ( !pio_write_string (channel, NULL) )
    {
	fprintf (stderr, "Error writing NULL history string\n");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function dsrw_write_history  */

/*PUBLIC_FUNCTION*/
flag dsrw_write_packet (Channel channel, CONST packet_desc *pack_desc,
			CONST char *packet)
/*  [SUMMARY] Write packet data in binary format.
    [PURPOSE] This routine will write the binary representation of a data
    packet to a channel object. The routine will recursively write sub arrays
    and linked lists of packets.
    <channel> The channel object.
    <pack_desc> The packet descriptor.
    <packet> The packet data.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int element_count = 0;
    unsigned int bytes_to_write;
    unsigned int type;
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "dsrw_write_packet";

    if (channel == NULL) return (TRUE);
    if (pack_desc == NULL)
    {
	fprintf (stderr, "NULL descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (packet == NULL)
    {
	fprintf (stderr, "NULL data pointer passed\n");
	a_prog_bug (function_name);
    }
    if ( ds_can_transfer_packet_as_block (pack_desc) )
    {
	/*  Block write  */
	bytes_to_write = ds_get_packet_size (pack_desc);
	if (ch_write (channel, packet, bytes_to_write) < bytes_to_write)
	{
	    fprintf (stderr, "Error writing packet\t%s\n", ERRSTRING);
	    return (FALSE);
	}
	return (TRUE);
    }
    /*  Write packet data   */
    while (element_count < pack_desc->num_elements)
    {
	/*  Write element   */
        type = pack_desc->element_types[element_count];
        if ( !dsrw_write_element (channel, type,
				  pack_desc->element_desc[element_count],
				  packet) ) return (FALSE);
        packet += host_type_sizes[type];
        ++element_count;
    }
    return (TRUE);
}   /*  End Function dsrw_write_packet   */

/*PUBLIC_FUNCTION*/
flag dsrw_write_element (Channel channel, unsigned int type, CONST char *desc,
			 CONST char *element)
/*  [SUMMARY] Write an element in binary format.
    [PURPOSE] This routine will write the binary representation of an element
    to a channel object.
    <channel> The channel object.
    <type> The type of the element.
    <desc> The descriptor for the element.
    <element> A pointer to the element data.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok = FALSE;
    unsigned int size;
    unsigned int fstring_len;
    FString *fstring;
#ifdef MACHINE_LITTLE_ENDIAN
    int byte_count;
    extern char network_type_bytes[NUMTYPES];
#endif
#if defined(NEED_ALIGNED_DATA) || defined(MACHINE_LITTLE_ENDIAN)
    char buffer[64];
#endif
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "dsrw_write_element";

    if (channel == NULL) return (TRUE);
    if (element == NULL)
    {
	fprintf (stderr, "No element data to write\n");
        a_prog_bug (function_name);
    }
    size = host_type_sizes[type];
    if ( ds_can_transfer_element_as_block (type) )
    {
	if (ch_write (channel, element, size) != size)
	{
	    fprintf (stderr, "Error writing atomic element to channel\t%s\n",
		     ERRSTRING);
	    return (FALSE);
	}
	return (TRUE);
    }
#if defined(NEED_ALIGNED_DATA) || defined(MACHINE_LITTLE_ENDIAN)
    if (size > 64)
    {
	fprintf (stderr, "Data type size: %u too large for buffer\n", size);
	a_prog_bug (function_name);
    }
#endif
#ifdef MACHINE_LITTLE_ENDIAN
    switch (type)
    {
      case K_FLOAT:
      case K_DOUBLE:
      case K_INT:
      case K_SHORT:
      case K_LONG:
      case K_UINT:
      case K_USHORT:
      case K_ULONG:
	/*  Swap bytes (cannot do this for complex types)  */
	if (host_type_sizes[type] == network_type_bytes[type])
	{
	    for (byte_count = 0; byte_count < (int) size; ++byte_count)
	    {
		buffer[byte_count] = element[size - byte_count - 1];
	    }
	    if (ch_write (channel, buffer, size) != size)
	    {
		fprintf (stderr,
			 "Error writing atomic element to channel\t%s\n",
			 ERRSTRING);
		return (FALSE);
	    }
	    return (TRUE);
	}
        break;
    }
#endif
    /*  Now we read from the element using non-byte C types. We have to be
	careful about data alignment.  */
    /*  Data will need to be converted  */
#ifdef NEED_ALIGNED_DATA
    m_copy (buffer, element, size);
    element = buffer;
#endif
    switch (type)
    {
      case K_FLOAT:
	ok = pio_write_float (channel, *(float *) element);
        break;
      case K_DOUBLE:
	ok = pio_write_double (channel, *(double *) element);
        break;
      case K_COMPLEX:
	if ( !pio_write_float (channel, *(float *) element) ) return (FALSE);
	ok = pio_write_float ( channel, *( (float *) element + 1 ) );
        break;
      case K_DCOMPLEX:
	if ( !pio_write_double (channel, *(double *) element) ) return (FALSE);
	ok = pio_write_double ( channel, *( (double *) element + 1 ) );
        break;
      case K_INT:
	ok = pio_write32s (channel, *(int *) element);
	break;
      case K_SHORT:
	ok = pio_write16s (channel, *(short *) element);
	break;
      case K_ICOMPLEX:
	if ( !pio_write32s (channel, *(int *) element) ) return (FALSE);
	ok = pio_write32s ( channel, *( (int *) element + 1 ) );
	break;
      case K_SCOMPLEX:
	if ( !pio_write16s (channel, *(short *) element) ) return (FALSE);
	ok = pio_write16s ( channel, *( (short *) element + 1 ) );
	break;
      case K_LONG:
	fprintf (stderr, "64 bit quantity not yet supported\n");
	a_prog_bug (function_name);
	break;
      case K_LCOMPLEX:
	fprintf (stderr, "64 bit quantity not yet supported\n");
	a_prog_bug (function_name);
	break;
      case K_UINT:
	ok = pio_write32 (channel, *(unsigned int *) element);
	break;
      case K_USHORT:
	ok = pio_write16 (channel, *(unsigned short *) element);
	break;
      case K_ULONG:
	ok = pio_write64 (channel, *(unsigned long *) element);
	break;
      case K_UICOMPLEX:
	if ( !pio_write32 (channel, *(unsigned int *) element) ) return FALSE;
	ok = pio_write32 ( channel, *( (unsigned int *) element + 1 ) );
	break;
      case K_USCOMPLEX:
	if ( !pio_write16 (channel,*(unsigned short *) element) ) return FALSE;
	ok = pio_write16 ( channel, *( (unsigned short *) element +1 ) );
	break;
      case K_ULCOMPLEX:
	if ( !pio_write64 (channel, *(unsigned long *) element) ) return FALSE;
	ok = pio_write64 ( channel, *( (unsigned long *) element + 1 ) );
	break;
      case ARRAYP:
	fprintf (stderr, "Writing unpadded array\n");
	ok = dsrw_write_array (channel, (array_desc *) desc, element, FALSE);
	break;
      case K_ARRAY:
	ok = dsrw_write_array (channel, (array_desc *) desc, element, TRUE);
	break;
      case LISTP:
	ok = dsrw_write_list (channel, (packet_desc *) desc,
			      *(list_header **) element);
	break;
      case K_VSTRING:
	ok = pio_write_string (channel, *(char **) element);
	break;
      case K_FSTRING:
	fstring = (FString *) element;
	if ( !pio_write32 (channel, fstring->max_len) ) return (FALSE);
	/*  Clear tail end of string  */
	fstring_len = strlen (fstring->string);
	m_clear (fstring_len + fstring->string,
		 fstring->max_len - fstring_len);
	if (ch_write (channel, fstring->string, fstring->max_len) <
	    fstring->max_len)
	{
	    fprintf (stderr, "Error writing string to channel\t%s\n",
		     ERRSTRING);
	    return (FALSE);
	}
	ok = TRUE;
        break;
      default:
	fprintf (stderr, "Bad element type: %u\n", type);
	a_prog_bug (function_name);
	break;
    }
    return (ok);
}   /*  End Function dsrw_write_element  */

/*PUBLIC_FUNCTION*/
flag dsrw_write_array (Channel channel, CONST array_desc *arr_desc,
		       CONST char *element, flag pad)
/*  [SUMMARY] Write an array in binary format.
    [PURPOSE] This routine will write the binary representation of an array to
    a channel object. The routine will recursively write any sub arrays or
    linked lists of packets.
    <channel> The channel object.
    <arr_desc> The array descriptor.
    <element> The element to write the array pointer to.
    <pad> If TRUE the array will be padded on a selected boundary (this is the
    case when writing .kf files).
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag local;
    unsigned int bytes_to_write;
    unsigned int array_size;
    unsigned int packet_size;
    unsigned int bytes_to_pad;
    unsigned long read_pos;
    unsigned long write_pos;
    array_pointer arrayp;
    char padding[ARRAY_BOUNDARY];
    static char function_name[] = "dsrw_write_array";

    FLAG_VERIFY (pad);
    if (channel == NULL) return (TRUE);
    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (arr_desc->packet == NULL)
    {
	fprintf (stderr, "NULL pointer to packet in array descriptor\n");
	a_prog_bug (function_name);
    }
    if (element == NULL)
    {
	fprintf (stderr, "NULL element pointer passed\n");
	a_prog_bug (function_name);
    }
    m_copy (&arrayp, element, sizeof arrayp);
    if (arrayp.array == NULL)
    {
	fprintf (stderr, "NULL data pointer passed\n");
	a_prog_bug (function_name);
    }
    array_size = ds_get_array_size (arr_desc);
    packet_size = ds_get_packet_size (arr_desc->packet);
    local = ch_test_for_local_connection (channel);
    bytes_to_write = packet_size * array_size;
    if (ds_packet_all_data (arr_desc->packet) && local)
    {
	return _dsrw_transmit_array_local (channel, &arrayp, bytes_to_write);
    }
    /*  Pad array if needed  */
    if (pad)
    {
	/*  Clear padding array (to avoid random file differences)  */
	m_clear (padding, ARRAY_BOUNDARY);
	if ( !ch_tell (channel, &read_pos, &write_pos) )
	{
	    fprintf (stderr, "Error getting channel position\n");
	    return (FALSE);
	}
	/*  Add 4 bytes for pad size  */
	write_pos += 4;
	bytes_to_pad = ARRAY_BOUNDARY - write_pos % ARRAY_BOUNDARY;
	if ( !pio_write32 (channel, (unsigned long) bytes_to_pad) )
	{
	    fprintf (stderr, "Error writing pad size\n");
	    return (FALSE);
	}
	if (bytes_to_pad > 0)
	{
	    if (ch_write (channel, padding, bytes_to_pad) < bytes_to_pad)
	    {
		fprintf (stderr, "Error padding: %u bytes\t%s\n",
			 bytes_to_pad, ERRSTRING);
		return (FALSE);
	    }
	}
    }
    if ( !dsrw_write_packets (channel, arr_desc->packet, arrayp.array,
			      array_size) ) return (FALSE);
    return (TRUE);
}   /*  End Function dsrw_write_array  */

/*PUBLIC_FUNCTION*/
flag dsrw_write_list (Channel channel, CONST packet_desc *pack_desc,
		      CONST list_header *list_head)
/*  [SUMMARY] Write a linked list in binary format.
    [PURPOSE] This routine will write the binary representation of the data in
    a linked list to a channel object. The routine will recursively write any
    sub arrays or linked lists of packets.
    <channel> The channel object.
    <pack_desc> The packet descriptor for the list entries.
    <list_head> The linked list header.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    list_entry *curr_entry;
    static char function_name[] = "dsrw_write_list";

    if (channel == NULL) return (TRUE);
    if (pack_desc == NULL)
    {
	fprintf (stderr, "NULL descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (list_head == NULL)
    {
	fprintf (stderr, "NULL list header pointer passed\n");
	a_prog_bug (function_name);
    }
    if (list_head->magic != MAGIC_LIST_HEADER)
    {
	fprintf (stderr, "List header has bad magic number\n");
	a_prog_bug (function_name);
    }
    /*  Write list length   */
    if ( !pio_write64 (channel, list_head->length) ) return (FALSE);
    switch (list_head->sort_type)
    {
      case SORT_INCREASING:
      case SORT_DECREASING:
      case SORT_RANDOM:
	break;
      default:
	fprintf (stderr, "Sort type: %u not recognised\n",
		 list_head->sort_type);
	a_prog_bug (function_name);
	break;
    }
    /*  Write list sort type    */
    if ( !pio_write32 (channel, list_head->sort_type) ) return (FALSE);
    if (list_head->sort_elem_num >= pack_desc->num_elements)
    {
	fprintf(stderr,
		"List sort element number: %u is not less than num_elem: %u\n",
		list_head->sort_elem_num, pack_desc->num_elements);
        a_prog_bug (function_name);
    }
    /*  Write list sort element number  */
    if ( !pio_write32 (channel, (unsigned long) list_head->sort_elem_num) )
    {
	return (FALSE);
    }
    /*  Write list packets  */
    /*  Write contiguous section of list  */
    if (list_head->contiguous_length > 0)
    {
	if ( !dsrw_write_packets (channel, pack_desc,
				  list_head->contiguous_data,
				  list_head->contiguous_length) )
	    return (FALSE);
    }
    /*  Write out fragmented section of list  */
    for (curr_entry = list_head->first_frag_entry; curr_entry != NULL;
	 curr_entry = curr_entry->next)
    {
	if ( !dsrw_write_packet (channel, pack_desc, curr_entry->data) )
	{
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function dsrw_write_list */

/*PUBLIC_FUNCTION*/
flag dsrw_write_packets (Channel channel, CONST packet_desc *descriptor,
			 CONST char *source, unsigned long num_packets)
/*  [SUMMARY] Write the binary representation of many data packets.
    [PURPOSE] This routine will write the binary representation many data
    packets to a channel object. The routine will recursively write out sub
    arrays and linked lists.
    <channel> The channel object.
    <descriptor> The descriptor for the packets.
    <source> The packet data.
    <num_packets> The number of packets to write.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int packet_size, type;
    unsigned long bytes_to_write, bytes_written, count;
    /*static char function_name[] = "dsrw_write_packets";*/

    packet_size = ds_get_packet_size (descriptor);
    bytes_to_write = packet_size * num_packets;
    if ( ds_can_transfer_packet_as_block (descriptor) ||
	 ( ch_test_for_local_connection (channel) &&
	   ds_packet_all_data (descriptor) ) )
    {
	/*  Do fast save  */
	if ( ( bytes_written = ch_write (channel, source, bytes_to_write) )
	     < bytes_to_write)
	{
	    fprintf (stderr, "Error writing packets to channel\t%s\n",
		     ERRSTRING);
	    fprintf (stderr, "Wanted: %lu bytes, wrote: %lu bytes\n",
		     bytes_to_write, bytes_written);
	    return (FALSE);
	}
	return (TRUE);
    }
    /*  It has not been possible to simply write the data in a single block,
	more work is needed to write the packets  */
    type = descriptor->element_types[0];
    if ( (descriptor->num_elements == 1) &&
	 ds_can_swaptransfer_element (type) )
    {
	/*  Because there is only one atomic element in the packets, and that
	    element can be transferred simply by writing and byte-swapping it
	    is possible to swap bytes and then write the packets in one block
	    */
	if ( ds_element_is_complex (type) )
	{
	    packet_size /= 2;
	    num_packets *= 2;
	}
	if ( ( bytes_written =
	       ch_swap_and_write_blocks (channel, source, num_packets,
					 packet_size) ) != bytes_to_write )
	{
	    fprintf (stderr, "Error writing packets\t%s\n",
		     ERRSTRING);
	    fprintf (stderr, "Wanted: %lu bytes, wrote: %lu bytes\n",
		     bytes_to_write, bytes_written);
	    return (FALSE);
	}
	return (TRUE);
    }
    /*  Do recursive save  */
    for (count = 0; count < num_packets; ++count, source += packet_size)
    {
	/*  Write a packet in the array */
	if ( !dsrw_write_packet (channel, descriptor, source) ) return (FALSE);
    }
    return (TRUE);
}   /*  End Function dsrw_write_packets  */

/*PUBLIC_FUNCTION*/
flag dsrw_write_flag (Channel channel, flag logical)
/*  [SUMMARY] Write a boolean value in binary format to a channel object.
    <channel> The channel object.
    <logical> The boolean value.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char value;
    static char function_name[] = "dsrw_write_flag";

    if (channel == NULL) return (TRUE);
    if ( (logical != TRUE) && (logical != FALSE) )
    {
	fprintf (stderr, "flag value: %u is not TRUE or FALSE\n",
		 (unsigned int) logical);
        a_prog_bug (function_name);
    }
    value = logical;
    if (ch_write (channel, &value, 1) != 1)
    {
	fprintf (stderr, "%s: error writing flag to channel\t%s\n",
		 function_name, ERRSTRING);
        return (FALSE);
    }
    return (TRUE);
}   /*  End Function dsrw_write_flag  */

/*PUBLIC_FUNCTION*/
multi_array *dsrw_read_multi (Channel channel)
/*  [SUMMARY] Read the binary representation of a multi_array structure.
    [PURPOSE] This routine will read the binary representation of a
    multi_array general data structure header from a channel object and will
    allocate the required descriptors. The routine will then read in the data
    from the channel object and will write it into the data structure(s)
    created.
    <channel> The channel object.
    [RETURNS] A pointer to the multi array header on success, else NULL.
*/
{
    flag array_alloc = TRUE;
    unsigned int num_arrays;
    unsigned int array_count;
    unsigned int length;
    char *string;
    char *data;
    multi_array *multi_desc;
    packet_desc *pack_desc;
    static char function_name[] = "dsrw_read_multi";

    if (channel == NULL)
    {
	fprintf (stderr,
		 "No channel to read multi array data structure from\n");
        return (NULL);
    }
    if ( ( multi_desc = dsrw_read_multi_header (channel) ) == NULL )
    {
	return (NULL);
    }
    num_arrays = multi_desc->num_arrays;
    /*  Read the structures  */
    if ( ch_test_for_mmap (channel) ) array_alloc = FALSE;
    else if ( ch_test_for_local_connection (channel) ) array_alloc = FALSE;
    for (array_count = 0; array_count < num_arrays; ++array_count)
    {
	/*  Load the descriptor for one array   */
        if ( ( pack_desc = dsrw_read_packet_desc (channel) ) == NULL )
        {
	    ds_dealloc_multi (multi_desc);
            return (NULL);
        }
        multi_desc->headers[array_count] = pack_desc;
	/*  Allocate the required data space   */
	if ( ( data = ds_alloc_data (multi_desc->headers[array_count],
				     FALSE, array_alloc) )
	    == NULL )
        {
	    m_error_notify (function_name, "general data structure");
            ds_dealloc_multi (multi_desc);
            return (NULL);
        }
        multi_desc->data[array_count] = data;
	/*  Read the data   */
        if ( !dsrw_read_packet (channel, multi_desc->headers[array_count],
				multi_desc->data[array_count]) )
        {
	    fprintf (stderr, "Error reading array number %u\n", array_count);
            ds_dealloc_multi (multi_desc);
            return (NULL);
        }
    }
    /*  May need to read arraynames  */
    for (array_count = 0; (array_count < num_arrays) && (num_arrays > 1);
	 ++array_count)
    {
	/*  Load the names of the arrays    */
	if ( ( multi_desc->array_names[array_count] =
	      pio_read_string (channel, &length) ) == NULL )
        {
	    a_func_abort (function_name, "Error reading array name");
            ds_dealloc_multi (multi_desc);
            return (NULL);
        }
        if (length == 0)
        {
	    a_func_abort (function_name,
			  "Array name must not be a null string");
            ds_dealloc_multi (multi_desc);
            return (NULL);
        }
    }
    /*  Read history strings  */
    if ( ( string = pio_read_string (channel, &length) )
	== NULL )
    {
	a_func_abort (function_name, "Error reading first history string");
	ds_dealloc_multi (multi_desc);
	return (NULL);
    }
    while (length > 0)
    {
	if ( !ds_history_append_string (multi_desc, string, FALSE) )
	{
	    m_abort (function_name, "history entry");
	}
	if ( ( string = pio_read_string (channel, &length) )
	    == NULL )
	{
	    a_func_abort (function_name, "Error reading history string");
	    ds_dealloc_multi (multi_desc);
	    return (NULL);
	}
    }
    m_free (string);
    return (multi_desc);
}   /*  End Function dsrw_read_multi */

/*PUBLIC_FUNCTION*/
multi_array *dsrw_read_multi_header (Channel channel)
/*  [SUMMARY] Read the binary representation of a multi_array header.
    [PURPOSE] This routine will read the binary representation of a
    multi_array general data structure header from a channel object. The
    various top-level packet descriptors, data and history are not read.
    <channel> The channel object.
    [RETURNS] A pointer to the multi array header on success, else NULL.
*/
{
    unsigned long num_arrays;
    unsigned long version;
    char magic_str_buf[MAGIC_STRING_LENGTH];
    multi_array *multi_desc;
    static char function_name[] = "dsrw_read_multi";

    if (channel == NULL)
    {
	fprintf (stderr,
		 "%s: no channel to read multi array data structure from\n",
		 function_name);
        return (NULL);
    }
    /*  Read magic string and version number  */
    if (ch_read (channel, magic_str_buf, MAGIC_STRING_LENGTH)
	< MAGIC_STRING_LENGTH)
    {
	fprintf (stderr, "%s: error reading magic string\n", function_name);
        return (NULL);
    }
    if (strncmp (magic_string, magic_str_buf, MAGIC_STRING_LENGTH) != 0)
    {
	fprintf (stderr, "%s: magic string not present\n", function_name);
	return (NULL);
    }
    if ( !pio_read32 (channel, &version) )
    {
	fprintf (stderr, "%s: error reading version number\n", function_name);
        return (NULL);
    }
    if (version != VERSION_NUMBER)
    {
	fprintf (stderr, "%s: unknown version number: %lu\n",
		 function_name, version);
        return (NULL);
    }
    if ( !pio_read32 (channel, &num_arrays) )
    {
	fprintf (stderr, "%s: error reading number of arrays\n",function_name);
        return (NULL);
    }
    if (num_arrays < 1)
    {
	fprintf (stderr, "%s: must have at least one array\n", function_name);
	return (NULL);
    }
    if ( ( multi_desc = ds_alloc_multi (num_arrays) ) == NULL )
    {
	m_error_notify (function_name, "multi_desc descriptor");
        return (NULL);
    }
    return (multi_desc);
}   /*  End Function dsrw_read_multi_header  */

/*PUBLIC_FUNCTION*/
packet_desc *dsrw_read_packet_desc (Channel channel)
/*  [SUMMARY] Read the binary representation of a packet descriptor.
    [PURPOSE] This routine will read the binary representation of a packet
    descriptor from a channel object. The routine will recursively read in
    array and linked list descriptors if required.
    <channel> The channel object.
    [RETURNS] A pointer to the packet descriptor on success, else NULL.
*/
{
    unsigned int num_elements = 0;
    unsigned int element_count;
    unsigned long data;
    unsigned int elem_type;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    packet_desc *list_desc;
    static char function_name[] = "dsrw_read_packet_desc";

    if (channel == NULL)
    {
	fprintf (stderr, "No channel to read packet descriptor from\n");
        return (NULL);
    }
    /*  Get number of elements  */
    if ( !pio_read32 (channel, &data) )
    {
	fprintf (stderr, "Error reading number of elements\n");
        return (NULL);
    }
    num_elements = data;
    if (num_elements < 1)
    {
	fprintf (stderr, "Bad number of elements: %u\n", num_elements);
        return (NULL);
    }
    /*  Allocate packet descriptor  */
    if ( ( pack_desc = ds_alloc_packet_desc (num_elements) ) == NULL )
    {
	m_error_notify (function_name, "packet descriptor");
        return (NULL);
    }
    for (element_count = 0; element_count < num_elements; ++element_count)
    {
	if ( !dsrw_read_type (channel, &elem_type) )
        {
	    ds_dealloc_packet (pack_desc, NULL);
            return (NULL);
        }
	pack_desc->element_types[element_count] = elem_type;
	if ( ds_element_is_named (elem_type) )
	{
	    /*  Atomic data type    */
	    if ( ( pack_desc->element_desc[element_count] =
		  pio_read_string (channel, (unsigned int *) NULL) )
		== NULL )
	    {
		a_func_abort (function_name, "Error reading element_name");
		ds_dealloc_packet (pack_desc, NULL);
		return (NULL);
	    }
	}
	else
	{
	    switch (elem_type)
	    {
	      case ARRAYP:
		pack_desc->element_types[element_count] = K_ARRAY;
		/*  Fall through  */
	      case K_ARRAY:
		if ( ( arr_desc = dsrw_read_array_desc (channel, elem_type) )
		    == NULL )
		{
		    ds_dealloc_packet (pack_desc, NULL);
		    return (NULL);
		}
		pack_desc->element_desc[element_count] = (char *) arr_desc;
		break;
	      case LISTP:
		if ( ( list_desc = dsrw_read_packet_desc (channel) ) == NULL )
		{
		    ds_dealloc_packet (pack_desc, NULL);
		    return (NULL);
		}
		pack_desc->element_desc[element_count] = (char *) list_desc;
		break;
	      default:
		/*  This should NEVER happen!   */
		fprintf (stderr, "Illegal element type: %u\n", elem_type);
		fprintf (stderr, "Returned from function: dsrw_read_type\n");
		a_prog_bug (function_name);
		break;
	    }
	}
    }
    return (pack_desc);
}   /*  End Function dsrw_read_packet_desc  */

/*PUBLIC_FUNCTION*/
array_desc *dsrw_read_array_desc (Channel channel, unsigned int type)
/*  [SUMMARY] Read the binary representation of an array descriptor.
    [PURPOSE] This routine will read the binary representation of an array
    descriptor from a channel object. The routine will recursively read the
    packet descriptor for the array.
    <channel> The channel object.
    <type> The type of the array. Legal values are:
        ARRAYP   (old format)
	K_ARRAY  (new format)
    [RETURNS] A pointer to the array descriptor on success, else NULL.
*/
{
    unsigned int num_dim = 0;
    unsigned int dim_count;
    unsigned int level_count;
    unsigned int num_levels;
    unsigned int product;
    unsigned long data;
    array_desc *arr_desc;
    dim_desc *dimension;
    packet_desc *pack_desc;
    static char function_name[] = "dsrw_read_array_desc";

    if (channel == NULL)
    {
	fprintf (stderr, "No channel to read array descriptor from\n");
        return (NULL);
    }
    if ( (type != ARRAYP) && (type != K_ARRAY) )
    {
	fprintf (stderr, "Bad type: %u\n", type);
	a_prog_bug (function_name);
    }
    /*  Get number of dimensions    */
    if ( !pio_read32 (channel, &data) )
    {
	fprintf (stderr, "Error reading number of dimensions\n");
        return (NULL);
    }
    num_dim = data;
    if (num_dim < 1)
    {
	fprintf (stderr, "Bad number of dimensions: %u\n", num_dim);
        return (NULL);
    }
    if (type == K_ARRAY)
    {
	/*  Get number of tiling levels  */
	if ( !pio_read32 (channel, &data) )
	{
	    fprintf (stderr, "Error reading number of tiling levels\n");
	    return (NULL);
	}
	num_levels = data;
    }
    else
    {
	num_levels = 0;
    }
    if ( ( arr_desc = ds_alloc_array_desc (num_dim, num_levels) ) == NULL )
    {
	m_error_notify (function_name, "array descriptor");
        return (NULL);
    }
    if (type == ARRAYP)
    {
	arr_desc->padded = FALSE;
    }
    else
    {
	arr_desc->padded = TRUE;
    }
    /*  Get dimension descriptors   */
    for (dim_count = 0; dim_count < num_dim; ++dim_count)
    {
	if ( ( dimension = dsrw_read_dim_desc (channel) ) == NULL )
        {
	    ds_dealloc_array_desc (arr_desc);
            return (NULL);
        }
        arr_desc->dimensions[dim_count] = dimension;
	/*  Get tile lengths if appropriate  */
	for (level_count = 0, product = 1; level_count < num_levels;
	     ++level_count)
	{
	    if ( !pio_read32 (channel, &data) )
	    {
		fprintf (stderr, "Error reading tile length[%u][%u]\n",
			 dim_count, level_count);
		ds_dealloc_array_desc (arr_desc);
		return (NULL);
	    }
	    arr_desc->tile_lengths[dim_count][level_count] = data;
	    product *= data;
	}
	/*  Check if tile lengths appropriate  */
	if (dimension->length % product != 0)
	{
	    fprintf (stderr, "Tile product: %u not a factor of length: %lu\n",
		     product, dimension->length);
	    a_func_abort (function_name, "bad data");
	    ds_dealloc_array_desc (arr_desc);
	    return (NULL);
	}
	arr_desc->lengths[dim_count] = dimension->length / product;
    }
    /*  Get packet descriptor for array */
    if ( ( pack_desc = dsrw_read_packet_desc (channel) ) == NULL )
    {
	ds_dealloc_array_desc (arr_desc);
        return (NULL);
    }
    arr_desc->packet = pack_desc;
    return (arr_desc);
}   /*  End Function dsrw_read_array_desc    */

/*PUBLIC_FUNCTION*/
dim_desc *dsrw_read_dim_desc (Channel channel)
/*  [SUMMARY] Read the binary representation of a dimension descriptor.
    [PURPOSE] This routine will read the binary representation of a dimension
    descriptor from a channel object.
    <channel> The channel object.
    [RETURNS] A pointer to the dimension descriptor on success, else NULL.
*/
{
    unsigned int length;
    unsigned int dim_length = 0;
    unsigned int coord_count = 0;
    unsigned long data;
    double first_coord, last_coord;
    double coordinate;
    flag regular;
    char *dim_name;
    dim_desc *dimension;
    static char function_name[] = "dsrw_read_dim_desc";

    if (channel == NULL)
    {
	fprintf (stderr, "No channel to read dimension descriptor from\n");
        return (NULL);
    }
    /*  Get name    */
    if ( ( dim_name = pio_read_string (channel, &length) ) == NULL )
    {
	a_func_abort (function_name, "Error reading dimension name");
        return (NULL);
    }
    if (length == 0)
    {
	a_func_abort (function_name,
		      "Dimension name must not be a null string");
	m_free (dim_name);
        return (NULL);
    }
    /*  Get length  */
    if ( !pio_read64 (channel, &data) )
    {
	fprintf (stderr, "Error reading dimension length\n");
	m_free (dim_name);
        return (NULL);
    }
    dim_length = data;
    if (dim_length < 1)
    {
	fprintf (stderr, "Bad dimension length: %u\n", dim_length);
	m_free (dim_name);
        return (NULL);
    }
    /*  Get REGULAR flag  */
    if ( !dsrw_read_flag (channel, &regular) )
    {
	fprintf (stderr, "Error reading REGULAR flag\n");
	m_free (dim_name);
        return (NULL);
    }
    if (regular)
    {
	/*  Dimension co-ordinates are regularly spaced */
        /*  Get first_coord */
        if ( !pio_read_double (channel, &first_coord) )
        {
	    fprintf (stderr, "Error reading first co-ordinate\n");
	    m_free (dim_name);
            return (NULL);
        }
        /*  Get maximum */
        if ( !pio_read_double (channel, &last_coord) )
        {
	    fprintf (stderr, "Error reading last co-ordinate\n");
	    m_free (dim_name);
            return (NULL);
        }
	if ( (first_coord == last_coord) && (dim_length != 1) )
	{
	    fprintf (stderr,
		     "Dimension: \"%s\" first_coord is equal to last_coord: %e\n",
		     dim_name, first_coord);
	    fprintf (stderr, "and length: %u is not 1\n", dim_length);
	    a_prog_bug (function_name);
	}
        if ( ( dimension = ds_alloc_dim_desc (dim_name, dim_length,
					      first_coord, last_coord,
					      TRUE) ) == NULL )
        {
	    m_error_notify (function_name, "dimension descriptor");
	    m_free (dim_name);
            return (NULL);
        }
	m_free (dim_name);
        return (dimension);
    }
    /*  Dimension co-ordinates are to be supplied   */
    if ( ( dimension = ds_alloc_dim_desc (dim_name, dim_length,
					  0.0, 1.0,  /*  Dummy values  */
					  FALSE) ) == NULL )
    {
	m_error_notify (function_name, "dimension descriptor");
	m_free (dim_name);
        return (NULL);
    }
    m_free (dim_name);
    /*  Get co-ordinates and compute minimum and maximum    */
    while (coord_count < dim_length)
    {
	if ( !pio_read_double (channel, &coordinate) )
        {
	    fprintf (stderr, "Error reading Co-ordinate number: %u\n",
		     coord_count);
            m_free (dimension->name);
            m_free (dimension->coordinates);
            m_free (dimension);
            return (NULL);
        }
	dimension->coordinates[coord_count] = coordinate;
	last_coord = coordinate;
        ++coord_count;
    }
    dimension->first_coord = dimension->coordinates[0];
    dimension->last_coord = dimension->coordinates[dim_length - 1];
    if (dimension->first_coord < dimension->last_coord)
    {
	dimension->minimum = dimension->first_coord;
	dimension->maximum = dimension->last_coord;
    }
    else
    {
	dimension->minimum = dimension->last_coord;
	dimension->maximum = dimension->first_coord;
    }
    return (dimension);
}   /*  End Function dsrw_read_dim_desc  */

/*PUBLIC_FUNCTION*/
flag dsrw_read_packet (Channel channel, CONST packet_desc *descriptor,
		       char *packet)
/*  [SUMMARY] Read the binary representation of a data packet.
    [PURPOSE] This routine will read the binary representation of a data packet
    from a channel object. The routine will recursively read in sub arrays and
    linked lists.
    <channel> The channel object.
    <descriptor> The descriptor for the packet.
    <packet> The packet data will be written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int elem_count = 0;
    unsigned int type;
    unsigned int bytes_to_read;
    extern char host_type_sizes[NUMTYPES];
    /*static char function_name[] = "dsrw_read_packet";*/

    if (channel == NULL)
    {
	fprintf (stderr, "No channel to read packet from\n");
        return (FALSE);
    }
    if (descriptor == NULL)
    {
	fprintf (stderr, "No packet descriptor for packet to be read\n");
        return (FALSE);
    }
    if (packet == NULL)
    {
	fprintf (stderr, "No packet to write to\n");
        return (FALSE);
    }
    if ( ds_can_transfer_packet_as_block (descriptor) )
    {
	/*  Block read  */
	bytes_to_read = ds_get_packet_size (descriptor);
	if (ch_read (channel, packet, bytes_to_read) < bytes_to_read)
	{
	    fprintf (stderr, "Error reading packet\t%s\n", ERRSTRING);
	    return (FALSE);
	}
	return (TRUE);
    }
    while (elem_count < descriptor->num_elements)
    {
	type = descriptor->element_types[elem_count];
        if ( !dsrw_read_element (channel, type,
				 descriptor->element_desc[elem_count],
				 packet) )
        {
	    return (FALSE);
        }
        packet += host_type_sizes[type];
        ++elem_count;
    }
    return (TRUE);
}   /*  End Function dsrw_read_packet  */

/*PUBLIC_FUNCTION*/
flag dsrw_read_element (Channel channel, unsigned int type, char *desc,
			char *element)
/*  [SUMMARY] Read the binary representation of an element.
    [PURPOSE] This routine will read the binary representation of an element
    from a channel object. If the element type is a pointer to an array or
    linked lists, the routine will recursively read in the array or linked
    list of packets.
    <channel> The channel object.
    <type> The type of the element.
    <desc> The element descriptor.
    <element> A pointer to the element storage. This will be written to.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    FString *fstring;
#ifdef MACHINE_LITTLE_ENDIAN
    int byte_count;
    extern char network_type_bytes[NUMTYPES];
#endif
    unsigned int size;
    unsigned long fstring_len;
    array_desc *arr_desc;
    char *ptr;
#if defined(NEED_ALIGNED_DATA) || defined(MACHINE_LITTLE_ENDIAN)
    char buffer[64];
#endif
    long data_s[2];
    unsigned long data_u[2];
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "dsrw_read_element";

    if (channel == NULL)
    {
	fprintf (stderr, "No channel to read element from\n");
        return (FALSE);
    }
    if (element == NULL)
    {
	fprintf (stderr, "No element to write to\n");
        return (FALSE);
    }
    size = host_type_sizes[type];
    if ( ds_can_transfer_element_as_block (type) )
    {
	if (ch_read (channel, element, size) != size)
	{
	    fprintf (stderr, "%s: error reading atomic element\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	return (TRUE);
    }
#if defined(NEED_ALIGNED_DATA) || defined(MACHINE_LITTLE_ENDIAN)
    if (size > 64)
    {
	fprintf (stderr, "Data type size: %u too large for buffer\n", size);
	a_prog_bug (function_name);
    }
#endif
#ifdef MACHINE_LITTLE_ENDIAN
    switch (type)
    {
#  ifdef HAS_IEEE
      case K_FLOAT:
      case K_DOUBLE:
#  endif
      case K_INT:
      case K_SHORT:
      case K_LONG:
      case K_UINT:
      case K_USHORT:
      case K_ULONG:
	/*  NOTE: cannot simply byte-swap a complex value!  */
	if (size == network_type_bytes[type])
	{
	    if (ch_read (channel, buffer, size) != size)
	    {
		fprintf (stderr,
			 "%s: error reading atomic element from channel\t%s\n",
			 function_name, ERRSTRING);
		return (FALSE);
	    }
	    /*  Swap bytes  */
	    for (byte_count = 0; byte_count < (int) size; ++byte_count)
	    {
		element[byte_count] = buffer[size - byte_count - 1];
	    }
	    return (TRUE);
	}
	break;
    }
#endif
    /*  Now we write into the element using non-byte C types. We have to be
	careful about data alignment.  */
#ifdef NEED_ALIGNED_DATA
    ptr = buffer;
    m_copy (buffer, element, size);
#else
    ptr = element;
#endif
    switch (type)
    {
      case K_FLOAT:
	if ( !pio_read_float (channel, (float *) ptr) ) return (FALSE);
	break;
      case K_DOUBLE:
	if ( !pio_read_double (channel, (double *) ptr) ) return (FALSE);
	break;
      case K_INT:
	if ( !pio_read32s (channel, data_s) ) return (FALSE);
	*(int *) ptr = data_s[0];
	break;
      case K_SHORT:
	if ( !pio_read16s (channel, data_s) ) return (FALSE);
	*(short *) ptr = data_s[0];
	break;
      case K_COMPLEX:
	if ( !pio_read_float (channel, (float *) ptr) ) return(FALSE);
	if ( !pio_read_float (channel, (float *) ptr + 1) ) return (FALSE);
	break;
      case K_DCOMPLEX:
	if ( !pio_read_double (channel, (double *) ptr) ) return (FALSE);
	if ( !pio_read_double (channel, (double *) ptr + 1) ) return FALSE;
	break;
      case K_ICOMPLEX:
	if ( !pio_read32s (channel, data_s) ) return (FALSE);
	if ( !pio_read32s (channel, data_s + 1) ) return (FALSE);
	*(int *) ptr = data_s[0];
	*( (int *) ptr + 1 ) = data_s[1];
	break;
      case K_SCOMPLEX:
	if ( !pio_read16s (channel, data_s) ) return (FALSE);
	if ( !pio_read16s (channel, data_s + 1) ) return (FALSE);
	*(short *) ptr = data_s[0];
	*( (short *) ptr + 1 ) = data_s[1];
	break;
      case K_LONG:
      case K_LCOMPLEX:
	fprintf (stderr,
		 "Atomic data type: %u not supported on this architecture yet\n",
		 type);
	a_prog_bug (function_name);
	break;
      case K_UINT:
	if ( !pio_read32 (channel, data_u) ) return (FALSE);
	*(unsigned int *) ptr = data_u[0];
	break;
      case K_USHORT:
	if ( !pio_read16 (channel, data_u) ) return (FALSE);
	*(unsigned short *) ptr = data_u[0];
	break;
      case K_ULONG:
	fprintf (stderr,
		 "Atomic data type: %u not supported on this architecture yet\n",
		 type);
	a_prog_bug (function_name);
	break;
      case K_UICOMPLEX:
	if ( !pio_read32 (channel, data_u) ) return (FALSE);
	if ( !pio_read32 (channel, data_u + 1) ) return (FALSE);
	*(unsigned int *) ptr = data_u[0];
	*( (unsigned int *) ptr + 1 ) = data_u[1];
	break;
      case K_USCOMPLEX:
	if ( !pio_read16 (channel, data_u) ) return (FALSE);
	if ( !pio_read16 (channel, data_u + 1) ) return (FALSE);
	*(unsigned short *) ptr = data_u[0];
	*( (unsigned short *) ptr + 1 ) = data_u[1];
	break;
      case K_ULCOMPLEX:
	fprintf (stderr,
		 "Atomic data type: %u not supported on this architecture yet\n",
		 type);
	a_prog_bug (function_name);
	break;
      case K_ARRAY:
	arr_desc = (array_desc *) desc;
	if ( !dsrw_read_array (channel, arr_desc, ptr, arr_desc->padded) )
	{
	    return (FALSE);
	}
	break;
      case LISTP:
	if ( !dsrw_read_list ( channel, (packet_desc *) desc,
			      *( (list_header **) ptr ) ) ) return (FALSE);
	break;
      case K_VSTRING:
	if (*(char **) ptr != NULL)
	{
	    fprintf (stderr, "Variable string already allocated\n");
	    a_prog_bug (function_name);
	}
	if ( ( *(char **) ptr = pio_read_string (channel,
						 (unsigned int *) NULL) )
	     == NULL )
	{
	    /*  I'm hitting some obscure limit with the VX/MVX: so I've been
		chopping out the more recent additions until it works.
		Fuck I'm fustrated by this!  */
#ifndef OS_VXMVX
	    fprintf (stderr, "Could not read VSTRING for element: \"%s\"\n",
		     desc);
#endif
	    return (FALSE);
	}
	break;
      case K_FSTRING:
	fstring = (FString *) ptr;
	if ( (fstring->max_len > 0) || (fstring->string != NULL) )
	{
	    fprintf (stderr, "Fixed string already allocated\n");
	    a_prog_bug (function_name);
	}
	if ( !pio_read32 (channel, &fstring_len) )
	{
	    a_func_abort (function_name, "Error reading FSTRING length");
	    return (FALSE);
	}
	if ( ( fstring->string = m_alloc (fstring_len) ) == NULL )
	{
	    m_error_notify (function_name, "fixed string length");
	    return (FALSE);
	}
	if (ch_read (channel, fstring->string, fstring_len) < fstring_len)
	{
	    fprintf (stderr, "Error reading fixed string element\n");
	    m_free (fstring->string);
	    fstring->string = NULL;
	    return (FALSE);
	}
	fstring->max_len = fstring_len;
	break;
      default:
	fprintf (stderr, "Bad data type: %u\n", type);
	a_prog_bug (function_name);
	break;
    }
#ifdef NEED_ALIGNED_DATA
    m_copy (element, buffer, size);
#endif
    return (TRUE);
}   /*  End Function dsrw_read_element   */

/*PUBLIC_FUNCTION*/
flag dsrw_read_array (Channel channel, CONST array_desc *arr_desc,
		      char *element, flag pad)
/*  [SUMMARY] Read the binary representation of an array.
    [PURPOSE] This routine will read the binary representation of an array
    from a channel object. The routine will recursively read in array packets.
    <channel> The channel object.
    <arr_desc> The array descriptor.
    <element> The array pointer will be written here.
    <pad> If TRUE the array will be padded on a selected boundary (this is the
    case when reading .kf files).
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag block_transfer;
    flag local;
    unsigned int bytes_to_read;
    unsigned int array_size;
    unsigned int packet_size;
    unsigned long read_pos;
    unsigned long write_pos;
    array_pointer arrayp;
    packet_desc *pack_desc;
    static char function_name[] = "dsrw_read_array";

    FLAG_VERIFY (pad);
    if (channel == NULL)
    {
	a_func_abort (function_name, "No channel to read array from");
        return (FALSE);
    }
    if (arr_desc == NULL)
    {
	fprintf (stderr, "No array descriptor for array to be read\n");
        return (FALSE);
    }
    if ( (pack_desc = arr_desc->packet) == NULL )
    {
	fprintf (stderr, "No packet descriptor for array to be read\n");
        return (FALSE);
    }
    if (element == NULL)
    {
	fprintf (stderr, "No element containing array pointer\n");
        return (FALSE);
    }
    m_copy (&arrayp, element, sizeof arrayp);
    array_size = ds_get_array_size (arr_desc);
    packet_size = ds_get_packet_size (pack_desc);
    block_transfer = ds_can_transfer_packet_as_block (pack_desc);
    local = ch_test_for_local_connection (channel);
    bytes_to_read = packet_size * array_size;
    if (ds_packet_all_data (arr_desc->packet) && local)
    {
	/*  Read from local connection  */
	return _dsrw_receive_array_local (channel, arr_desc, element,
					  bytes_to_read);
    }
    if (pad)
    {
	if ( !read_array_padding (channel) )
	{
	    fprintf (stderr, "Error reading array padding\n");
	    return (FALSE);
	}
    }
    if ( !ch_tell (channel, &read_pos, &write_pos) )
    {
	fprintf (stderr, "%s: error getting channel positions\n",
		 function_name);
	return (FALSE);
    }
    if (arr_desc->num_levels > 0)
    {
	fprintf (stderr, "Reading tiled array at offset %lu bytes\n",read_pos);
    }
    if ( ch_test_for_mmap (channel) )
    {
	if (block_transfer)
	{
	    /*  Channel was memory mapped  */
	    if (arrayp.array != NULL)
	    {
		/*  No data should have been allocated  */
		fprintf (stderr, "Memory mapped array already allocated\n");
		fprintf (stderr, "Top dimension of array: \"%s\"\n",
			 arr_desc->dimensions[0]->name);
		a_prog_bug (function_name);
	    }
	    /*  Can memory map array  */
	    if (!pad)
	    {
		fprintf (stderr, "Memory mapping unpadded array\n");
	    }
	    arrayp.array = ch_get_mmap_addr (channel) + read_pos;
	    arrayp.free = NULL;
	    arrayp.info = NULL;
	    arrayp.mem_type = K_ARRAY_MEM_TYPE_MMAP;
	    arrayp.u.mmap.fd = ch_get_descriptor (channel);
	    arrayp.u.mmap.size = ch_get_mem_size (channel);
	    arrayp.u.mmap.base = ch_get_mmap_addr (channel);
	    m_copy (element, &arrayp, sizeof arrayp);
	    if ( !ch_seek (channel, read_pos + bytes_to_read) )
	    {
		fprintf (stderr,
			 "%s: error seeking channel from %lu to %lu\t%s\n",
			 function_name,
			 read_pos, read_pos + bytes_to_read, ERRSTRING);
		return (FALSE);
	    }
	    return (TRUE);
	}
	/*  Cannot memory map array: allocate for later if not already
	    allocated  */
	if (arrayp.array == NULL)
	{
	    if ( !ds_alloc_array (arr_desc, element, FALSE, TRUE) )
	    {
		m_error_notify (function_name, "array data");
		return (FALSE);
	    }
	}
    }
    /*  Must get that pointer again, as things may have been allocated now  */
    m_copy (&arrayp, element, sizeof arrayp);
    /*  Read from channel  */
    if (arrayp.array == NULL)
    {
	fprintf (stderr, "No array to write to\n");
	return (FALSE);
    }
    return dsrw_read_packets (channel, pack_desc, arrayp.array, array_size);
}   /*  End Function dsrw_read_array  */

/*PUBLIC_FUNCTION*/
flag dsrw_read_list (Channel channel, CONST packet_desc *descriptor,
		     list_header *header)
/*  [SUMMARY] Read the binary representation of a linked list.
    [PURPOSE] This routine will read in the binary representation of a linked
    list of data from a channel object. The routine will recursively read in
    sub arrays and linked lists.
    <channel> The channel object.
    <descriptor> The packet descriptor for the list entries.
    <header> The linked list header. The linked list entries and data will be
    contiguous in memory.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag array_alloc;
    unsigned int length;
    unsigned long data;
    static char function_name[] = "dsrw_read_list";

    if (channel == NULL)
    {
	fprintf (stderr, "No channel to read linked list from\n");
        return (FALSE);
    }
    if (descriptor == NULL)
    {
	fprintf (stderr, "No descriptor for list to be read\n");
        return (FALSE);
    }
    if (header == NULL)
    {
	fprintf (stderr, "No header for list to be read\n");
    }
    if ( header->magic != MAGIC_LIST_HEADER )
    {
	fprintf (stderr, "List header has bad magic number\n");
	a_prog_bug (function_name);
    }
    /*  Get linked list length  */
    if ( !pio_read64 (channel, &data) )
    {
	fprintf (stderr, "Error reading linked list length\n");
        return (FALSE);
    }
    length = data;
    array_alloc = ch_test_for_mmap (channel) ? FALSE : TRUE;
    if ( !ds_alloc_contiguous_list (descriptor, header, length, FALSE,
				    array_alloc) )
    {
	fprintf (stderr, "Error allocating memory for linked list\n");
        return (FALSE);
    }
    /*  Get sort type   */
    if ( !pio_read32 (channel, &data) )
    {
	fprintf (stderr, "Error reading list sort type\n");
        ds_dealloc_list_entries (descriptor, header);
        return (FALSE);
    }
    header->sort_type = data;
    switch (header->sort_type)
    {
      case SORT_INCREASING:
      case SORT_DECREASING:
      case SORT_RANDOM:
	break;
      default:
	fprintf (stderr, "Bad sort type value: %u\n", header->sort_type);
	ds_dealloc_list_entries (descriptor, header);
	return (FALSE);
	/*break;*/
    }
    /*  Get sort element number     */
    if ( !pio_read32 (channel, &data) )
    {
	fprintf (stderr, "Error reading list sort element number\n");
        ds_dealloc_list_entries (descriptor, header);
        return (FALSE);
    }
    header->sort_elem_num = data;
    if (header->sort_elem_num >= descriptor->num_elements)
    {
	fprintf (stderr,
		 "List sort element number: %u is not less than num_elem: %u\n",
		 header->sort_elem_num, descriptor->num_elements);
	fprintf (stderr, "Error detected in function: %s\n", function_name);
        ds_dealloc_list_entries (descriptor, header);
        return (FALSE);
    }
    if (length < 1) return (TRUE);
    /*  Get list  */
    return dsrw_read_packets (channel, descriptor, header->contiguous_data,
			      length);
}   /*  End Function dsrw_read_list  */

/*PUBLIC_FUNCTION*/
flag dsrw_read_packets (Channel channel, CONST packet_desc *descriptor,
			char *dest, unsigned long num_packets)
/*  [SUMMARY] Read the binary representation of many data packets.
    [PURPOSE] This routine will read the binary representation many data
    packets from a channel object. The routine will recursively read in sub
    arrays and linked lists.
    <channel> The channel object.
    <descriptor> The descriptor for the packets.
    <dest> The packet data will be written here.
    <num_packets> The number of packets to read.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int packet_size, type;
    unsigned long got_bytes;
    unsigned long bytes_to_read, count;
    static char function_name[] = "dsrw_read_packets";

    packet_size = ds_get_packet_size (descriptor);
    bytes_to_read = packet_size * num_packets;
    if ( ds_can_transfer_packet_as_block (descriptor) ||
	 ( ch_test_for_local_connection (channel) &&
	   ds_packet_all_data (descriptor) ) )
    {
	/*  Can do a fast read  */
	if ( ( got_bytes = ch_read (channel, dest, bytes_to_read) )
	     < bytes_to_read )
	{
	    fprintf (stderr, "%s: error reading packets\t%s\n",
		     function_name, ERRSTRING);
	    fprintf (stderr, "Wanted: %lu bytes, got: %lu bytes\n",
		     bytes_to_read, got_bytes);
	    return (FALSE);
	}
	return (TRUE);
    }
    /*  It has not been possible to simply read the data in a single block,
	more work is needed to read the packets  */
    type = descriptor->element_types[0];
    if ( (descriptor->num_elements == 1) &&
	 ds_can_swaptransfer_element (type) )
    {
	/*  Because there is only one atomic element in the packets, and that
	    element can be transferred simply by reading and byte-swapping it
	    is possible to read the packets in one block and then swap all the
	    bytes  */
	if ( ds_element_is_complex (type) )
	{
	    packet_size /= 2;
	    num_packets *= 2;
	}
	if ( ( got_bytes = ch_read_and_swap_blocks (channel, dest, num_packets,
						    packet_size) )
	     != bytes_to_read )
	{
	    fprintf (stderr, "%s: error reading packets\t%s\n",
		     function_name, ERRSTRING);
	    fprintf (stderr, "Wanted: %lu bytes, got: %lu bytes\n",
		     bytes_to_read, got_bytes);
	    return (FALSE);
	}
	return (TRUE);
    }
    /*  Do recursive read   */
    for (count = 0; count < num_packets; ++count, dest += packet_size)
    {
	if ( !dsrw_read_packet (channel, descriptor, dest) )
	{
	    fprintf (stderr, "Error reading packet: %lu\n", count);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function dsrw_read_packets  */

/*PUBLIC_FUNCTION*/
flag dsrw_read_flag (Channel channel, flag *logical)
/*  [SUMMARY] Read the binary representation of a boolean.
    [PURPOSE] This routine will read in the binary representation of a boolean
    flag from a channel object.
    <channel> The channel object.
    <logical> The boolean value is written here.
    [RETURNS] TRUE if the flag was read in without error, else FALSE.
*/
{
    char value;
    static char function_name[] = "dsrw_read_flag";

    if (channel == NULL)
    {
	fprintf (stderr, "No channel to read flag from\n");
        return (FALSE);
    }
    if (ch_read (channel, &value, 1) != 1)
    {
	fprintf (stderr, "%s: error reading flag from channel\t%s\n",
		 function_name, ERRSTRING);
        return (FALSE);
    }
    *logical = value;
    if (*logical == TRUE)
    {
	return (TRUE);
    }
    if (*logical == FALSE)
    {
	return (TRUE);
    }
    fprintf (stderr, "%s: error reading flag, bad value: %u\n",
	     function_name, (int) *logical);
    return (FALSE);
}   /*  End Function dsrw_read_flag  */

/*PUBLIC_FUNCTION*/
flag dsrw_read_type (Channel channel, unsigned int *type)
/*  [SUMMARY] Read the binary representation of a data type.
    [PURPOSE] This routine will read in the binary representation of a data
    type from a channel object.
    <channel> The channel object.
    <type> The type value is written here.
    [RETURNS] TRUE if the type was read without error, else FALSE.
*/
{
    unsigned long data;
    static char function_name[] = "dsrw_read_type";

    if (channel == NULL)
    {
	a_func_abort (function_name, "No channel to read type from");
        return (FALSE);
    }
    if ( !pio_read32 (channel, &data) )
    {
	return (FALSE);
    }
    if ( !ds_element_is_legal (data) )
    {
	fprintf (stderr, "Function: %s\tbad data type value: %lu\n",
		 function_name, data);
	return (FALSE);
    }
#ifdef RETYPE_NEEDED
    /*  Retyping needed  */
    *type = net_to_host_retype[data];
#else
    /*  No retyping needed  */
    *type = data;
#endif
    return (TRUE);
}   /*  End Function dsrw_read_type  */


/*  Private functions follow  */

static flag _dsrw_transmit_array_local (Channel channel,
					CONST array_pointer *arrayp,
					unsigned int length)
/*  This routine will transmit an atomic array over a local connection, using
    the most efficient transport available.
    The channel must be given by  channel  .
    <arrayp> The array pointer.
    The array data must be pointed to by  array  .
    The length (in bytes) of the array must be given by  length  .
    [RETURNS] TRUE on success, else FALSE.
*/
{
#ifdef OS_VXMVX
    flag ok;
#endif
    uaddr offset;
    unsigned int bytes_written;
    unsigned int control;
    static char function_name[] = "_dsrw_transmit_array_local";

    control = FA_NORMAL;
#ifdef OS_VXMVX
    if ( (int) array % VXMVX_REMOTE_MEMCPY_BOUNDARY == 0 ) control = FA_VX;
    else
    {
	fprintf (stderr,
		 "%s: array: %p not on a %d byte boundary: slow transfer\n",
		 array, VXMVX_REMOTE_MEMCPY_BOUNDARY);
	control = FA_NORMAL;
    }
#endif
    if (arrayp->mem_type == K_ARRAY_MEM_TYPE_SHM) control = FA_SHM;
#ifdef CAN_PASS_FD
    if (arrayp->mem_type == K_ARRAY_MEM_TYPE_MMAP) control = FA_MMAP;
#endif
    /*  Write control value  */
    if ( !pio_write32 (channel, control) )
    {
	fprintf (stderr, "Error writing control value\n");
	return (FALSE);
    }
    /*  Flush now to ensure that subsequent control information will require an
	explicit flush (i.e. write buffer will be empty)  */
    if ( !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error flushing channel\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
#ifdef OS_VXMVX
    if (control == FA_VX)
    {
	/*  Try to use  remote_memcpy  */
	if ( !pio_write32s ( channel, task_get_id () ) )
	{
	    fprintf (stderr, "Error writing task ID\n");
	    return (FALSE);
	}
	if ( !pio_write32s ( channel, r_getppid () ) )
	{
	    fprintf (stderr, "Error writing parent process ID\n");
	    return (FALSE);
	}
	if ( !pio_write32 (channel, array) )
	{
	    fprintf (stderr, "Error writing array address ID\n");
	    return (FALSE);
	}
	if ( !ch_flush (channel) )
	{
	    fprintf (stderr, "Error flushing channel\t%s\n", ERRSTRING);
	    return (FALSE);
	}
	/*  Get response  */
	if ( !dsrw_read_flag (channel, &ok) )
	{
	    fprintf (stderr, "Error reading response flag\n");
	    return (FALSE);
	}
	if (ok) return (TRUE);
    }
#endif  /*  OS_VXMVX  */
    if (control == FA_SHM)
    {
	if ( !pio_write32 (channel, arrayp->u.shm.shmid) )
	{
	    fprintf (stderr, "%s: error writing SHM ID\n", function_name);
	    return (FALSE);
	}
	offset = (char *) arrayp->array - (char *) arrayp->u.shm.base;
	if (ch_write (channel, (char *) &offset,sizeof offset) < sizeof offset)
	{
	    fprintf (stderr, "%s: error writing SHM offset\n", function_name);
	    return (FALSE);
	}
	if ( !pio_write32 ( channel, getpid () ) )
	{
	    fprintf (stderr, "%s: error writing PID\n", function_name);
	    return (FALSE);
	}
	/*  Get response: do it with signals because the connection might be a
	    pipe and hence not bi-directional  */
	if ( _dsrw_fa_get_response (channel) )
	{
	    fprintf (stderr, "Transferred using SHM, ID=%d\n",
		     arrayp->u.shm.shmid);
	    return (TRUE);
	}
    }
#ifdef CAN_PASS_FD
    if (control == FA_MMAP)
    {
	char dummy;

	offset = (char *) arrayp->array - (char *) arrayp->u.mmap.base;
	if (ch_write (channel, (char *) &offset,sizeof offset) < sizeof offset)
	{
	    fprintf (stderr, "%s: error writing mmap offset\n", function_name);
	    return (FALSE);
	}
	if (ch_write (channel, (char *) &arrayp->u.mmap.size,
		      sizeof arrayp->u.mmap.size) < sizeof arrayp->u.mmap.size)
	{
	    fprintf (stderr, "%s: error writing mmap size\n", function_name);
	    return (FALSE);
	}
	if ( !pio_write32 ( channel, getpid () ) )
	{
	    fprintf (stderr, "%s: error writing PID\n", function_name);
	    return (FALSE);
	}
	/*  First flush and then wait for response  */
	_dsrw_fa_get_response (channel);
	/*  Send the file descriptor  */
	if ( !r_send_fd (ch_get_descriptor (channel), arrayp->u.mmap.fd) )
	{
	    fprintf (stderr, "%s: error passing FD=%d\t%s\n",
		     function_name, arrayp->u.mmap.fd, ERRSTRING);
	}
	/*  Send a synchronisation byte  */
	if (ch_write (channel, &dummy, 1) < 1)
	{
	    fprintf (stderr, "%s: error writing sync byte\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	/*  Get response: do it with signals because the connection might be a
	    pipe and hence not bi-directional  */
	if ( _dsrw_fa_get_response (channel) )
	{
	    fprintf (stderr, "Transferred using mmapped FD=%d\n",
		     arrayp->u.mmap.fd);
	    return (TRUE);
	}
    }
#endif
    /*  Simple transfer  */
    if ( ( bytes_written = ch_write (channel, arrayp->array, length) )
	 < length)
    {
	fprintf (stderr, "Error writing array to channel\t%s\n",
		 ERRSTRING);
	fprintf (stderr, "%s: wanted: %u bytes, wrote: %u bytes\n",
		 function_name, length, bytes_written);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _dsrw_transmit_array_local  */

static flag _dsrw_fa_get_response (Channel channel)
/*  [SUMMARY] Get FA response.
    <channel> The channel to flush after the signal handlers are installed.
    [RETURNS] TRUE if the destination is happy, else FALSE.
*/
{
    struct sigaction new_action, old_sa_usr1, old_sa_usr2;
    static char function_name[] = "_dsrw_fa_get_response";

    fa_response = 0;
    new_action.sa_flags = 0;
#ifdef SA_RESTART
    new_action.sa_flags |= SA_RESTART;
#endif
    new_action.sa_handler = signal_handler;
    /*  Set up SIG_FA_OK handler  */
    sigemptyset (&new_action.sa_mask);
    if (sigaction (SIG_FA_OK, &new_action, &old_sa_usr1) != 0)
    {
	fprintf (stderr, "%s: error setting OK handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    sigemptyset (&new_action.sa_mask);
    if (sigaction (SIG_FA_FAILED, &new_action, &old_sa_usr2) != 0)
    {
	fprintf (stderr, "%s: error setting failed handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    if (channel != NULL)
    {
	if ( !ch_flush (channel) )
	{
	    fprintf (stderr, "%s: error flushing channel\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
    }
    while (fa_response == 0) pause ();
    /*  Restore old handlers  */
    if (sigaction (SIG_FA_OK, &old_sa_usr1, NULL) != 0)
    {
	fprintf (stderr, "%s: error restoring old USR1 handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    if (sigaction (SIG_FA_FAILED, &old_sa_usr2, NULL) != 0)
    {
	fprintf (stderr, "%s: error restoring old USR1 handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    return (fa_response & FA_RESPONSE_OK) ? TRUE : FALSE;
}   /*  End Function _dsrw_fa_get_response  */

static flag _dsrw_receive_array_local (Channel channel,
				       CONST array_desc *arr_desc,
				       char *element, unsigned int length)
/*  This routine will receive an atomic array over a local connection, using
    the most efficient transport available.
    The channel must be given by  channel  .
    <arr_desc> The array descriptor.
    <element> The array element pointer.
    The length (in bytes) of the array must be given by  length  .
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag happy = FALSE;
    char dummy;
    int fd;
    long sender_id;
    long sender_ppid;
    unsigned int bytes_read;
    unsigned long control;
    unsigned long sender_addr;
    unsigned long shmid;
    unsigned long pid = 0;  /*  Initialised to keep compilers happy  */
    uaddr offset, size;
    array_pointer arrayp;
    void *addr;
    static char function_name[] = "_dsrw_receive_array_local";

    m_copy (&arrayp, element, sizeof arrayp);
    /*  Read control value  */
    if ( !pio_read32 (channel, &control) )
    {
	fprintf (stderr, "%s: error reading control value\n", function_name);
	return (FALSE);
    }
    /*  Read special information depending on transfer type  */
    switch (control)
    {
      case FA_NORMAL:
	/*  Do nothing  */
	break;
      case FA_VX:
	/*  Try to use  remote_memcpy  */
	if ( !pio_read32s (channel, &sender_id) )
	{
	    fprintf (stderr, "%s: error reading task ID\n", function_name);
	    return (FALSE);
	}
	if ( !pio_read32s (channel, &sender_ppid) )
	{
	    fprintf (stderr, "%s: error reading parent process ID\n",
		     function_name);
	    return (FALSE);
	}
	if ( !pio_read32 (channel, &sender_addr) )
	{
	    fprintf (stderr, "%s: error reading array address ID\n",
		     function_name);
	    return (FALSE);
	}
#ifdef OS_VXMVX
	if (r_getppid () == sender_ppid)
	{
	    if ( (int) arrayp.array % VXMVX_REMOTE_MEMCPY_BOUNDARY == 0 )
	    {
		remote_memcpy (task_get_id (), arrayp.array,
			       sender_id, (char *) sender_addr, length);
		happy = TRUE;
	    }
	    else
	    {
		fprintf (stderr,
			 "%s: array: %p not on a %d byte boundary: slow transfer\n",
			 function_name,
			 arrayp.array, VXMVX_REMOTE_MEMCPY_BOUNDARY);
		happy = FALSE;
	    }
	}
#endif  /*  OS_VXMVX  */
	break;
      case FA_SHM:
	if ( !pio_read32 (channel, &shmid) )
	{
	    fprintf (stderr, "%s: error reading SHM ID\n", function_name);
	    return (FALSE);
	}
	if (ch_read (channel, (char *) &offset, sizeof offset) < sizeof offset)
	{
	    fprintf (stderr, "%s: error reading offset\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	if ( !pio_read32 (channel, &pid) )
	{
	    fprintf (stderr, "%s: error reading PID\n", function_name);
	    return (FALSE);
	}
	if ( ( addr = m_shm_attach (shmid, TRUE, FALSE) ) == NULL )
	{
	    fprintf (stderr, "%s: error attaching SHM segment: %lu\t%s\n",
		     function_name, shmid, ERRSTRING);
	    break;
	}
	if (arrayp.array == NULL)
	{
	    arrayp.array = (char *) addr + offset;
	    arrayp.free = ( void (*) () ) m_shm_detach;
	    arrayp.info = addr;
	    arrayp.mem_type = K_ARRAY_MEM_TYPE_SHM;
	    arrayp.u.shm.shmid = shmid;
	    arrayp.u.shm.base = addr;
	    m_copy (element, &arrayp, sizeof arrayp);
	}
	else
	{
	    fprintf (stderr, "%s: array already allocated: doing bulk copy\n",
		     function_name);
	    m_copy (arrayp.array, addr, length);
	    m_shm_detach (addr);
	    break;
	}
	happy = TRUE;
	break;
      case FA_MMAP:
	if (ch_read (channel, (char *) &offset, sizeof offset) < sizeof offset)
	{
	    fprintf (stderr, "%s: error reading offset\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	if (ch_read (channel, (char *) &size, sizeof size) < sizeof size)
	{
	    fprintf (stderr, "%s: error reading size\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	if ( !pio_read32 (channel, &pid) )
	{
	    fprintf (stderr, "%s: error reading PID\n", function_name);
	    return (FALSE);
	}
	if (kill (pid, SIG_FA_OK) != 0)
	{
	    fprintf (stderr, "%s: error sending signal\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	if (arrayp.array != NULL)
	{
	    fprintf (stderr, "%s: array already allocated\n", function_name);
	    break;
	}
	fd = r_recv_fd ( ch_get_descriptor (channel) );
	if (ch_read (channel, &dummy, 1) < 1)
	{
	    fprintf (stderr, "%s: error reading sync byte\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	if (fd < 0) break;
	if ( !ds_alloc_mmap_array (fd, offset, size, FALSE, &arrayp, arr_desc,
				   FALSE, TRUE) )
	    break;
	m_copy (element, &arrayp, sizeof arrayp);
	happy = TRUE;
	break;
      default:
	fprintf (stderr, "%s: illegal control value: %lu\n",
		 function_name, control);
	return (FALSE);
	/*break;*/
    }
    /*  Send a response message  */
    switch (control)
    {
      case FA_VX:
	if ( !dsrw_write_flag (channel, happy) ) return (FALSE);
	if ( !ch_flush (channel) )
	{
	    fprintf (stderr, "%s: error flushing channel\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	break;
      case FA_SHM:
      case FA_MMAP:
	if (kill (pid, happy ? SIG_FA_OK : SIG_FA_FAILED) != 0)
	{
	    fprintf (stderr, "%s: error sending signal\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	break;
    }
    if (happy) return (TRUE);
    /*  Simple transfer  */
    if (arrayp.array == NULL)
    {
	if ( !ds_alloc_array (arr_desc, element, FALSE, TRUE) )
	{
	    m_error_notify (function_name, "array data");
	    return (FALSE);
	}
	m_copy (&arrayp, element, sizeof arrayp);
    }
    if ( ( bytes_read = ch_read (channel, arrayp.array, length) ) < length)
    {
	fprintf (stderr, "%s: error reading array from channel\t%s\n",
		 function_name, ERRSTRING);
	fprintf (stderr, "%s: wanted: %u bytes, got: %u bytes\n",
		 function_name, length, bytes_read);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _dsrw_receive_array_local  */

static flag read_array_padding (Channel channel)
/*  This routine will read array padding data from a channel.
    The channel must be given by  channel  .
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    unsigned long pad_bytes;

    if ( !pio_read32 (channel, &pad_bytes) )
    {
	fprintf (stderr, "Error reading pad size\n");
	return (FALSE);
    }
    if (ch_drain (channel, pad_bytes) < pad_bytes)
    {
	fprintf (stderr, "Error reading byte\t%s\n", ERRSTRING);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function read_array_padding  */

static void signal_handler (int sig)
{
    switch (sig)
    {
      case SIG_FA_OK:
	fa_response = FA_RESPONSE_OK;
	break;
      case SIG_FA_FAILED:
	fa_response = FA_RESPONSE_FAILED;
	break;
    }
}   /*  End Function signal_handler  */
