/*LINTLIBRARY*/
/*  main.c

    This code provides high level event dispatching routines.

    Copyright (C) 1998-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for dispatching events to
    other programmes.


    Written by      Richard Gooch   29-MAR-1998

    Updated by      Richard Gooch   13-APR-1998: Changed to generic event
  structure (using embedded union) and added keypress event support.

    Updated by      Richard Gooch   17-APR-1998: Bug fix in <event_wait> when
  reading non-plain-point events.

    Updated by      Richard Gooch   12-AUG-1998: Support new event
  K_CANVAS_EVENT_POINTER_WARP.

    Updated by      Richard Gooch   23-NOV-1999: Dispatch events locally in
  <event_dispatch>.

    Last updated by Richard Gooch   28-JUN-2000: Made <<event_list>> static.


*/
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_event.h>
#include <karma_conn.h>
#include <karma_pio.h>
#include <karma_ch.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_c.h>


#define PROTOCOL_VERSION 2


/*  Private structures  */
typedef struct
{
    unsigned long mask;
    long count;
} client_state_at_server;

typedef struct
{
    KEvent data;
    Connection connection;
} event_data;

typedef flag (*EventFunc) (KEvent event, Connection connection, void **f_info);


/*  Private functions  */
STATIC_FUNCTION (flag _event_read_coordinate,
		 (Channel channel, KEventCoordinate *coordinate) );
STATIC_FUNCTION (flag _event_read_single_coordinate,
		 (Channel channel, KEventSingleCoordinate *coordinate) );
STATIC_FUNCTION (flag _event_write_coordinate,
		 (Channel channel, KEventCoordinate coordinate) );
STATIC_FUNCTION (flag _event_write_single_coordinate,
		 (Channel channel, KEventSingleCoordinate coordinate) );
STATIC_FUNCTION (flag _event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (flag _event_read, (Channel channel, KEvent *event) );
STATIC_FUNCTION (flag _event_write, (Channel channel, KEvent event) );
STATIC_FUNCTION (flag _event_server_open_func,
		 (Connection connection, void **info) );
STATIC_FUNCTION (flag _event_server_read_func,
		 (Connection connection, void **info, Channel channel) );
STATIC_FUNCTION (void _event_server_close_func,
		 (Connection connection, void *info) );
STATIC_FUNCTION (flag _event_client_read_func,
		 (Connection connection, void **info, Channel channel) );


/*  Private data  */
static KCallbackList event_list = NULL;


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void event_initialise ()
/*  [SUMMARY] Initialise the package.
    [RETURNS] Nothing.
*/
{
    static flag initialised = FALSE;

    if (initialised) return;
    initialised = TRUE;
    conn_register_server_protocol ("generic_event", PROTOCOL_VERSION, 0,
				   _event_server_open_func,
				   _event_server_read_func,
				   _event_server_close_func);
    conn_register_client_protocol ("generic_event", PROTOCOL_VERSION, 0,
				   NULL, NULL, _event_client_read_func, NULL);
}   /*  End Function event_initialise  */

/*EXPERIMENTAL_FUNCTION*/
void event_dispatch (KEvent event)
/*  [SUMMARY] Dispatch an event.
    <event> The event.
    [RETURNS] Nothing.
*/
{
    Connection connection;
    Channel channel;
    event_data ed;
    unsigned int num_conn, count;
    unsigned long bit_in_mask = 0;
    client_state_at_server *state;
    static char function_name[] = "event_dispatch";

    ed.data = event;
    ed.connection = NULL;
    c_call_callbacks (event_list, &ed);
    num_conn = conn_get_num_serv_connections ("generic_event");
    if (num_conn < 1) return;
    switch (event.type)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
	bit_in_mask = K_EVENT_MASK_MOUSE_CLICK;
	break;
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
	bit_in_mask = K_EVENT_MASK_MOUSE_RELEASE;
	break;
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
	bit_in_mask = K_EVENT_MASK_MOUSE_DRAG;
	break;
      case K_CANVAS_EVENT_POINTER_MOVE:
	bit_in_mask = K_EVENT_MASK_POINTER_MOVE;
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	bit_in_mask = K_EVENT_MASK_KEYPRESS;
	break;
      case K_CANVAS_EVENT_POINTER_WARP:
	bit_in_mask = K_EVENT_MASK_POINTER_WARP;
	break;
      default:
	return;
	/*break;*/
    }
    for (count = 0; count < num_conn; ++count)
    {
	if ( ( connection = conn_get_serv_connection ("generic_event", count) )
	     == NULL )
	{
	    fprintf (stderr, "Connection: %u not found\n", count);
	    a_prog_bug (function_name);
	}
	state = conn_get_connection_info (connection);
	if ( !(state->mask & bit_in_mask) ) continue;
	if (state->count == 0) continue;
	channel = conn_get_channel (connection);
	if ( !_event_write (channel, event) ) continue;
	if ( !ch_flush (channel) ) continue;
	if (state->count > 0) --state->count;
    }
}   /*  End Function event_dispatch  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc event_register_event_func (flag (*func) (), void *f_info)
/*  [SUMMARY] Register event callback.
    <func> The function that is called when a position event occurs. The
    prototype function is [<event_PROTO_event_func>].
    <f_info> The initial arbitrary function information pointer.
    [RETURNS] A handle to a KCallbackFunc object.
*/
{
    /*static char function_name[] = "event_register_event_func";*/

    return c_register_callback (&event_list, _event_func,
				NULL, f_info, TRUE, (void *) func, FALSE,
				TRUE);
}   /*  End Function event_register_event_func  */

/*EXPERIMENTAL_FUNCTION*/
flag event_set_mask (unsigned long mask, long num_events,
		     Connection connection)
/*  [SUMMARY] Set the event mask
    <mask> The event mask. See [<event_MASKS>] for a list of event masks.
    <num_events> The number of events to receive per connection. If this is
    less than 0, an unlimited number of events may be received.
    <connection> The connection to apply the mask to. If this is NULL, all
    connections will have the mask applied.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    Channel channel;
    flag ok = TRUE;
    unsigned int num_conn, count;
    static char function_name[] = "event_set_mask";

    if (connection != NULL)
    {
	channel = conn_get_channel (connection);
	if ( !pio_write32 (channel, mask) ) return (FALSE);
	if ( !pio_write32s (channel, num_events) ) return (FALSE);
	return ch_flush (channel);
    }
    num_conn = conn_get_num_client_connections ("generic_event");
    if (num_conn < 1) return (TRUE);
    for (count = 0; count < num_conn; ++count)
    {
	if ( ( connection = conn_get_client_connection ("generic_event",
							count) ) == NULL )
	{
	    fprintf (stderr, "Connection: %u not found\n", count);
	    a_prog_bug (function_name);
	}
	channel = conn_get_channel (connection);
	if ( !pio_write32 (channel, mask) ||
	     !pio_write32s (channel, num_events) ||
	     !ch_flush (channel) )
	{
	    ok = FALSE;
	    continue;
	}
    }
    return (ok);
}   /*  End Function event_set_mask  */

/*EXPERIMENTAL_FUNCTION*/
flag event_wait (unsigned long mask, Connection connection, KEvent *event)
/*  [SUMMARY] Wait for an event.
    <mask> The event mask. After an event is received the event mask is set to
    0. See [<event_MASKS>] for a list of event masks.
    <connection> The connection to receive the event from.
    <event> The event data will be written here
    [RETURNS] TRUE on success, else FALSE.
*/
{
    Channel channel;
    /*static char function_name[] = "event_wait";*/

    channel = conn_get_channel (connection);
    if ( !event_set_mask (mask, 1, connection) ) return (FALSE);
    /*  Wait  */
    return _event_read (channel, event);
}   /*  End Function event_wait  */


/*  Private functions follow  */

static flag _event_read_coordinate (Channel channel,
				    KEventCoordinate *coordinate)
/*  [SUMMARY] Read an event co-ordinate.
    <channel> The channel to read from.
    <coordinate> The co-ordinate.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    if ( !_event_read_single_coordinate (channel, &coordinate->screen_pixel) )
	return (FALSE);
    if ( !_event_read_single_coordinate (channel, &coordinate->linear) )
	return (FALSE);
    if ( !_event_read_single_coordinate (channel, &coordinate->world) )
	return (FALSE);
    if ( !_event_read_single_coordinate (channel, &coordinate->image_pixel) )
	return (FALSE);
    return (TRUE);
}   /*  End Function _event_read_coordinate  */

static flag _event_read_single_coordinate (Channel channel,
					   KEventSingleCoordinate *coordinate)
/*  [SUMMARY] Read an event co-ordinate.
    <channel> The channel to read from.
    <coordinate> The co-ordinate.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_event_read_single_coordinate";

    if (ch_read (channel, &coordinate->defined, 1) < 1)
    {
	fprintf (stderr, "%s: error writing defined flag\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
    if (coordinate->defined == FALSE)
    {
	coordinate->x = TOOBIG;
	coordinate->y = TOOBIG;
    }
    if (coordinate->defined != TRUE)
    {
	fprintf (stderr, "%s: illegal value for defined flag: %d\n",
		 function_name, coordinate->defined);
	return (FALSE);
    }
    if ( !pio_read_double (channel, &coordinate->x) )
    {
	fprintf (stderr, "%s: error reading X co-ordinate\n", function_name);
	return (FALSE);
    }
    if ( !pio_read_double (channel, &coordinate->y) )
    {
	fprintf (stderr, "%s: error reading Y co-ordinate\n", function_name);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _event_read_single_coordinate  */

static flag _event_write_coordinate (Channel channel,
				     KEventCoordinate coordinate)
/*  [SUMMARY] Write an event co-ordinate.
    <channel> The channel to write to.
    <coordinate> The co-ordinate.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    if ( !_event_write_single_coordinate (channel, coordinate.screen_pixel) )
	return (FALSE);
    if ( !_event_write_single_coordinate (channel, coordinate.linear) )
	return (FALSE);
    if ( !_event_write_single_coordinate (channel, coordinate.world) )
	return (FALSE);
    if ( !_event_write_single_coordinate (channel, coordinate.image_pixel) )
	return (FALSE);
    return (TRUE);
}   /*  End Function _event_write_coordinate  */

static flag _event_write_single_coordinate (Channel channel,
					    KEventSingleCoordinate coordinate)
/*  [SUMMARY] Write an event co-ordinate.
    <channel> The channel to write to.
    <coordinate> The co-ordinate.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_event_write_single_coordinate";

    if ( (coordinate.defined != TRUE) && (coordinate.defined != FALSE) )
    {
	fprintf (stderr, "Illegal value for defined flag: %d\n",
		 coordinate.defined);
	a_prog_bug (function_name);
    }
    if (ch_write (channel, &coordinate.defined, 1) < 1)
    {
	fprintf (stderr, "%s: error writing defined flag\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
    if (!coordinate.defined) return (TRUE);
    if ( !pio_write_double (channel, coordinate.x) )
    {
	fprintf (stderr, "%s: error writing X co-ordinate\n", function_name);
	return (FALSE);
    }
    if ( !pio_write_double (channel, coordinate.y) )
    {
	fprintf (stderr, "%s: error writing Y co-ordinate\n", function_name);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _event_write_single_coordinate  */

static flag _event_func (void *object, void *client1_data,
			 void *call_data, void *client2_data)
/*  This routine is called when object callbacks are called.
    The object information pointer will be given by  object  .
    The first client information pointer will be given by  client1_data  .
    The call information pointer will be given by  call_data  .
    The second client information pointer will be given by  client2_data  .
    The routine returns TRUE if further callbacks should not be called.
*/
{
    EventFunc func = (EventFunc) client2_data;
    event_data *event = call_data;
    /*static char function_name[] = "_event_func";*/

    return (*func) (event->data, event->connection, client1_data);
}   /*  End Function _event_func  */


/*  Read and write a single event  */

static flag _event_read (Channel channel, KEvent *event)
/*  [SUMMARY] Read an event.
    <channel> The channel to read from.
    <event> The event data is written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned long type;
    /*static char function_name[] = "_event_read";*/

    if ( !pio_read32 (channel, &type) ) return (FALSE);
    event->type = type;
    switch (type)
    {
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	event->data.keypress.modifier_mask = 0;
	break;
    }
    switch (type)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_POINTER_WARP:
	if ( !_event_read_coordinate (channel, &event->data.point) )
	    return (FALSE);
	break;
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	if ( !pio_read32 (channel, &event->data.keypress.modifier_mask) )
	    return (FALSE);
	/*  Fall through  */
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	if ( !_event_read_coordinate (channel,&event->data.keypress.position) )
	    return (FALSE);
	event->data.keypress.string = pio_read_string (channel, NULL);
	if (event->data.keypress.string == NULL) return (FALSE);
	break;
    }
    return (TRUE);
}   /*  End Function _event_read  */

static flag _event_write (Channel channel, KEvent event)
/*  [SUMMARY] Write an event.
    <channel> The channel to write to.
    <event> The event.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    if ( !pio_write32 (channel, event.type) ) return (FALSE);
    switch (event.type)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_POINTER_WARP:
	if ( !_event_write_coordinate (channel, event.data.point) )
	    return (FALSE);
	break;
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	if ( !pio_write32 (channel, event.data.keypress.modifier_mask) )
	    return (FALSE);
	/*  Fall through  */
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	if ( !_event_write_coordinate (channel, event.data.keypress.position) )
	    return (FALSE);
	if ( !pio_write_string (channel, event.data.keypress.string) )
	    return (FALSE);
	break;
    }
    return (TRUE);
}   /*  End Function _event_write  */


/*  Server connection callbacks follow  */

static flag _event_server_open_func (Connection connection, void **info)
/*  [SUMMARY] Connection open event callback.
    [PURPOSE] This routine is called when a connection from a client opens.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    [RETURNS] TRUE on successful registration, else FALSE (indicating the
    connection should be closed).
    [NOTE] The <<close_func>> will not be called if this routine returns
    FALSE.
*/
{
    client_state_at_server *new;
    static char function_name[] = "_event_server_open_func";

    if ( ( new = m_alloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "state data");
	return (FALSE);
    }
    m_clear (new, sizeof *new);
    *info = new;
    return (TRUE);
}   /*  End Function _event_server_open_func  */

static flag _event_server_read_func (Connection connection, void **info,
				     Channel channel)
/*  [SUMMARY] Connection read event callback.
    [PURPOSE] This routine is called when data is ready to be read from a
    connection.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    <channel> The channel object on which the read event occurred. This may be
    the normal or datagram channel.
    [RETURNS] TRUE on successful reading, else FALSE (indicating the connection
    should be closed).
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
    client_state_at_server *new = *info;
    static char function_name[] = "_event_server_read_func";

    if ( !pio_read32 (channel, &new->mask) )
    {
	fprintf (stderr, "%s: error reading mask\n", function_name);
	return (FALSE);
    }
    if ( !pio_read32s (channel, &new->count) )
    {
	fprintf (stderr, "%s: error reading event count\n", function_name);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _event_server_read_func  */

static void _event_server_close_func (Connection connection, void *info)
/*  [SUMMARY] Connection close event callback.
    [PURPOSE] This routine is called when a connection closed.
    When this routine is called, this is the last chance to read any
    buffered data from the channel associated with the connection object.
    <connection> The connection object.
    <info> The arbitrary connection information pointer.
    [RETURNS] Nothing.
*/
{
    client_state_at_server *new = info;
    /*static char function_name[] = "_event_server_close_func";*/

    m_free (new);
}   /*  End Function _event_server_close_func  */


/*  Client connection callbacks follow  */

static flag _event_client_read_func (Connection connection, void **info,
				     Channel channel)
/*  [SUMMARY] Connection read event callback.
    [PURPOSE] This routine is called when data is ready to be read from a
    connection.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    <channel> The channel object on which the read event occurred. This may be
    the normal or datagram channel.
    [RETURNS] TRUE on successful reading, else FALSE (indicating the connection
    should be closed).
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
    event_data event;
    /*static char function_name[] = "_event_client_read_func";*/

    if ( !_event_read (channel, &event.data) ) return (FALSE);
    event.connection = connection;
    c_call_callbacks (event_list, &event);
    switch (event.data.type)
    {
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	m_free (event.data.data.keypress.string);
	break;
    }
    return (TRUE);
}   /*  End Function _event_client_read_func  */
