/*LINTLIBRARY*/
/*  main.c

    This file provides code to support extensions.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains code to support dynamic extensions.


    Written by      Richard Gooch   10-OCT-1997

    Updated by      Richard Gooch   26-NOV-1997: Changed interface to
  <ext_call> to support limits and getting basic return information. And added
  support for config files in $KARMABASE/site/share

    Updated by      Richard Gooch   26-MAY-1998: Removed cast of <m_alloc>
  return value.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <r> package.

    Updated by      Richard Gooch   10-MAR-1999: Cope with NULL return from
  <r_get_karmabase>.

    Updated by      Richard Gooch   8-APR-1999: Added "EXT_VERBOSE" environment
  variable.

    Updated by      Richard Gooch   8-JUN-1999: Took account of change to
  <st_expr_expand>.

    Last updated by Richard Gooch   28-JUL-2001: Fixed bug in function
  <_ext_add_service>: trailing/multiple whitespace caused spurious warnings.


*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <karma.h>
#include <os.h>
#ifdef HAS_DL
#  include <dlfcn.h>
#  define KARMA_EXT_MAIN_C
#  include <karma_ext.h>
#  include <karma_dl.h>
#  include <karma_ch.h>
#  include <karma_st.h>
#  include <karma_r.h>
#  include <karma_m.h>
#  include <karma_a.h>

#  define BUF_LEN 4096


/*  Private structures  */
struct extension_type
{
    char service[STRING_LENGTH];
    char object_name[STRING_LENGTH];
    void *object;
    struct extension_type *next;
};
typedef flag (*Func) (void *arg1, void *arg2, void *arg3, void *arg4,
		      void *arg5, void *arg6);


/*  Private functions  */
STATIC_FUNCTION (void _ext_initialise, (void) );
STATIC_FUNCTION (void _ext_read_config, (CONST char *filename) );
STATIC_FUNCTION (void _ext_add_directories, (char *dirlist) );
STATIC_FUNCTION (void _ext_add_service, (CONST char *service, char *line) );
STATIC_FUNCTION (char *skip_white, (char *start) );
STATIC_FUNCTION (char *skip_to_white, (char *start) );


/*  Private data  */
static struct extension_type *first_extension = NULL;
static struct extension_type *last_extension = NULL;
static char directories[BUF_LEN];


/*  Public functions follow  */

/*UNPUBLISHED_FUNCTION*/
unsigned int ext_call (CONST char *service, CONST char *function,
		       unsigned int max_func, unsigned int *num_ok,
		       void *arg1, void *arg2, void *arg3, void *arg4,
		       void *arg5, void *arg6)
/*  [SUMMARY] Call an extension function.
    <service> The service name for the extension.
    <function> The function to call.
    <max_func> The maximum number of functions to call. If this is 0 there is
    no limit.
    <num_ok> The number of functions which returned TRUE is written here. This
    may be NULL.
    <arg1> An optional argument.
    <arg2> An optional argument.
    <arg3> An optional argument.
    <arg4> An optional argument.
    <arg5> An optional argument.
    <arg6> An optional argument.
    [RETURNS] The number of functions called. The value 0 is returned if no
    functions were called.
*/
{
    Func func;
    flag ok;
    unsigned int num_called = 0;
    unsigned int _num_ok = 0;
    struct extension_type *entry;
    static char function_name[] = "ext_call";

    _ext_initialise ();
    for (entry = first_extension;
	 (entry != NULL) && ( (num_called < max_func) || (max_func < 1) );
	 entry = entry->next)
    {
	if (strcmp (service, entry->service) != 0) continue;
	/*  Found it!  */
	if ( ( func = (Func) dlsym (entry->object, function) ) == NULL )
	    continue;
	++num_called;
	ok = (*func) (arg1, arg2, arg3, arg4, arg5, arg6);
	if (ok == TRUE) ++_num_ok;
	else if (ok != FALSE)
	{
	    fprintf(stderr,
		    "Function: \"%s\" for service: \"%s\" in object: \"%s\"\n",
		    function, entry->service, entry->object_name);
	    fprintf (stderr, "did not return a boolean!\n");
	    a_prog_bug (function_name);
	}
    }
    if (num_ok != NULL) *num_ok = _num_ok;
    return (num_called);
}   /*  End Function ext_call  */


/*  Private functions follow  */

static void _ext_initialise (void)
/*  [SUMMARY] Initialise.
    [RETURNS] Nothing.
*/
{
    CONST char *home, *karmabase;
    char config_file[STRING_LENGTH];
    static flag first_time = TRUE;
    /*static char function_name[] = "_ext_initialise";*/

    if (!first_time) return;
    first_time = FALSE;
    directories[0] = '\0';
    if ( ( home = r_getenv ("HOME") ) != NULL )
    {
	sprintf (config_file, "%s/.karma/extensions", home);
	_ext_read_config (config_file);
    }
    if ( ( karmabase = r_get_karmabase () ) == NULL ) return;
    sprintf (config_file, "%s/site/share/extensions", karmabase);
    _ext_read_config (config_file);
    sprintf (config_file, "%s/share/extensions", karmabase);
    _ext_read_config (config_file);
}   /*  End Function _ext_initialise  */

static void _ext_read_config (CONST char *filename)
/*  [SUMMARY] Read config from a file.
    <filename> The file to read from.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    char *ptr;
    char p1[STRING_LENGTH], p2[STRING_LENGTH], line[STRING_LENGTH];
    static char function_name[] = "_ext_read_config";

    if (access (filename, R_OK) != 0) return;
    if ( ( channel = ch_open_file (filename, "r") ) == NULL )
    {
	fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		 function_name, filename, ERRSTRING);
	return;
    }
    while ( ch_getl (channel, line, STRING_LENGTH) )
    {
	ptr = skip_white (line);
	if ( (*ptr == '\0') || (*ptr == '#') ) continue;
	sscanf (line, "%s %s", p1, p2);
	if (strcmp (p1, "SO_DIRS") == 0)
	{
	    _ext_add_directories (p2);
	    continue;
	}
	else if (strcmp (p1, "SERVICE") != 0)
	{
	    fprintf (stderr, "%s: illegal line: \"%s\"\n", function_name,line);
	    ch_close (channel);
	    return;
	}
	_ext_add_service (p2, line);
    }
    ch_close (channel);
}   /*  End Function _ext_read_config  */

static void _ext_add_directories (char *dirlist)
/*  [SUMMARY] Add a list of directories to the main list.
    <dirlist> The new directories.
    [RETURNS] Nothing.
*/
{
    if ( !st_expr_expand (dirlist, STRING_LENGTH, dirlist, NULL, NULL, NULL) )
	return;
    if (directories[0] == '\0') strcpy (directories, dirlist);
    else
    {
	strcat (directories, ":");
	strcat (directories, dirlist);
    }
}   /*  End Function _ext_add_directories  */

static void _ext_add_service (CONST char *service, char *line)
/*  [SUMMARY] Add service entry to the end of the list of extensions.
    <service> The service name.
    <line> The configuration file line. This is modified.
    [RETURNS] Nothing.
*/
{
    int lflags = 0;
    char *ptr;
    void *obj;
    struct extension_type *new;
    static flag first_time = FALSE;
    static CONST char *ext_verbose = NULL;
    static char function_name[] = "_ext_add_service";

    if (first_time)
    {
	first_time = FALSE;
	ext_verbose = r_getenv ("EXT_VERBOSE");
    }
#  ifdef RTLD_LAZY
    lflags |= RTLD_LAZY;
#  endif
    ptr = skip_white (line);
    ptr = skip_to_white (ptr);
    ptr = skip_white (ptr);
    ptr = skip_to_white (ptr);
    ptr = skip_white (ptr);
    if (*ptr == '\0')
    {
	fprintf (stderr, "%s: illegal line: \"%s\"\n", function_name, line);
	return;
    }
    for ( line = ptr; *line != '\0'; line = skip_white (ptr) )
    {
	ptr = skip_to_white (line);
	if (*ptr != '\0') *ptr++ = '\0';
	if ( ( obj = dl_load_one_object (line, directories, lflags) ) == NULL )
	{
	    fprintf (stderr,
		     "%s: error loading extension: \"%s\" for service: \"%s\"\n",
		     function_name, line, service);
	    continue;
	}
	if (ext_verbose) fprintf (stderr, "service: \"%s\" object: \"%s\"\n",
				  service, line);
	if ( ( new = m_alloc (sizeof *new) ) == NULL )
	{
	    m_error_notify (function_name, "filter entry");
	    return;
	}
	strcpy (new->service, service);
	strcpy (new->object_name, line);
	new->object = obj;
	new->next = NULL;
	if (last_extension != NULL) last_extension->next = new;
	last_extension = new;
	if (first_extension == NULL) first_extension = new;
    }
}   /*  End Function _ext_add_service  */

static char *skip_white (char *start)
/*  [SUMMARY] Skip past whitespace.
    <start> The starting position.
    [RETURNS] A pointer to the first non-whitespace pointer on success, else
    a pointer to the '\0' terminator.
*/
{
    for (; isspace (*start) && (*start != '\0'); ++start);
    return (start);
}   /*  End Function skip_white  */

static char *skip_to_white (char *start)
/*  [SUMMARY] Skip to next whitespace.
    <start> The starting position.
    [RETURNS] A pointer to the first whitespace pointer on success, else
    a pointer to the '\0' terminator.
*/
{
    for (; !isspace (*start) && (*start != '\0'); ++start);
    return (start);
}   /*  End Function skip_to_white  */


#else  /*  HAS_DL  */


/*  Stub functions follow  */

/*STUB_FUNCTION*/
unsigned int ext_call (CONST char *service, CONST char *function,
		       unsigned int max_func, unsigned int *num_ok,
		       void *arg1, void *arg2, void *arg3, void *arg4,
		       void *arg5, void *arg6)
/*  [SUMMARY] Call an extension function.
    <service> The service name for the extension.
    <function> The function to call.
    <max_func> The maximum number of functions to call. If this is 0 there is
    no limit.
    <num_ok> The number of functions which returned TRUE is written here. This
    may be NULL.
    <arg1> An optional argument.
    <arg2> An optional argument.
    <arg3> An optional argument.
    <arg4> An optional argument.
    <arg5> An optional argument.
    <arg6> An optional argument.
    [RETURNS] The number of functions called. The value 0 is returned if no
    functions were called.
*/
{
    return (0);
}   /*  End Function ext_call  */

#endif  /*  !HAS_DL  */
