/*LINTLIBRARY*/
/*  aips1_read.c

    This code provides an AIPS1 (old AIPS) read facility.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading data files in
  the AIPS1 format (Astronomical Image Processing System, old AIPS, not to be
  confused with AIPS++).


    Written by      Richard Gooch   30-JUL-1997

    Updated by      Richard Gooch   31-JUL-1997

    Updated by      Richard Gooch   11-AUG-1997: Fixed bugs in function
 <foreign_guess_and_read>: directory name not used in constructing filenames.

    Updated by      Richard Gooch   25-SEP-1997: No longer clear array when
  allocating.

    Updated by      Richard Gooch   17-NOV-1997: Fixed bug when slot number or
  sequence number was greater than 9.

    Updated by      Richard Gooch   30-JAN-1998: Ignore user sequence number.

    Updated by      Richard Gooch   24-FEB-1998: Switch to new array_pointer
  which has free function.

    Updated by      Richard Gooch   13-MAY-1998: Added fallback to using anon.
  mmapped array.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <r> package.

    Updated by      Richard Gooch   23-NOV-1999: Search for catalogue files if
  AIPS ID is not known, in order to guess it.

    Updated by      Richard Gooch   24-NOV-1999: No longer ignore user sequence
  number (was previously ignored because I was misusing it). Automatically
  handle byte swapping. No longer guess AIPS ID, just show all catalogues in
  directory if ID not given.

    Updated by      Richard Gooch   24-FEB-2000: Ignore "documented" offsets
  for BMAJ, BMIN and BPA values and used empirically derived values. Sigh.

    Updated by      Richard Gooch   22-MAY-2000: Do not show datasets in
  catalogue with missing headers.

    Updated by      Richard Gooch   26-MAY-2000: Added streaming support.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Last updated by Richard Gooch   20-JUL-2001: Added support for version 'C'
  format, and possibly earlier versions.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_dir.h>
#include <karma_pio.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_p.h>
#include <karma_r.h>
#include <karma_a.h>

#define RECORD_LEN   128     /*  doubles: for alignment purposes  */
#define RECORD_BYTES 1024
#define BUFFER_INC   4060
#define SLOT_BYTES   40      /*  a single slot in the catalogue  */
#define NUM_SLOTS_PER_RECORD (RECORD_BYTES / SLOT_BYTES)
#define MAX_DIMENSIONS 7
#define GET_VALUE(out, offset) get_value ((out), (char *) slot + (offset), \
					  sizeof *(out), swap_bytes)
#define MAGIC_NUMBER 487932729

#define VERIFY_CONTEXT(ctx) if (ctx == NULL) \
{fprintf (stderr, "NULL context passed\n"); \
 a_prog_bug (function_name); } \
if (ctx->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid context object\n"); \
 a_prog_bug (function_name); }

/*  Offsets in 32 bit words for various header entries  */
#define KHOBJ  0
#define KHTEL  2
#define KHINS  4 
#define KHOBS  6
#define KHDOB  8
#define KHDMP  10
#define KHBUN  12
#define KHPTP  14
#define KIPTPN 14  /*  NOT AN OFFSET  */
#define KHCTP  42
#define KICTPN 7   /*  NOT AN OFFSET  */
#define KDCRV  56
#define KRCIC  70
#define KRCRP  77
#define KRCRT  84
#define KREPO  91
#define KRDMX  92
#define KRDMN  93
#define KRBLK  94
#define KIGCN  95
#define KIPCN  96
#define KIDIM  97
#define KINAX  98
#if 0  /*  The follow block has been found to not be ordered as listed  */
#define KIIMS  105
#define KHIMN  106
#define KHIMNO 1   /*  NOT AN OFFSET  */
#define KHIMC  109
#define KHIMCO 13   /*  NOT AN OFFSET  */
#define KHPTY  111
#define KHPTYO 19   /*  NOT AN OFFSET  */
#define KIIMU  112
#define KINIT  113
#define KRBMJ  114
#define KRBMN  115
#define KRBPA  116
#define KITYP  117
#define KIALT  118
#define KDORA  119
#define KDODE  121
#define KDRST  123
#define KDARV  125
#define KRARP  127
#define KRXSH  128
#define KRYSH  129
#define KHEXT  130
#define KHEXTN 20   /*  NOT AN OFFSET  */
#define KIVER  150
#endif

/*  Determined empirically  */
#define KRBMJ  113
#define KRBMN  114
#define KRBPA  115


/*  Private structures  */
struct aips1_data_context_type
{
    unsigned int magic_number;
    Channel channel;
    flag swap_bytes;
    flag have_blanking;
    int blank_value;
    unsigned int num_dim;
    unsigned long blank_count;
    uaddr read_pos;
    uaddr dim_lengths[MAX_DIMENSIONS];
    char datafile[STRING_LENGTH];
};


/*  Private functions  */
STATIC_FUNCTION (CONST char *_foreign_aips1_read_catalogue,
		 (CONST char *dirname, char version, CONST char *user_string,
		  unsigned int *num_datasets, unsigned int *buf_usage) );
STATIC_FUNCTION (void id_to_string, (char *string, int aips_id) );
STATIC_FUNCTION (void strip_trailing_whitespace, (char *string) );
STATIC_FUNCTION (void get_aips_id, () );
STATIC_FUNCTION (void get_value,
		 (void *dest, CONST void *source, uaddr size, flag swap) );
STATIC_FUNCTION (uaddr read_blocks,
		 (Channel channel, char *buffer, uaddr num, uaddr size,
		  flag swap_bytes) );
STATIC_FUNCTION (void generate_filename,
		 (char filename[STRING_LENGTH], CONST char *dirname,
		  CONST char *separator, CONST char *filetype,
		  char version, int slot_number,
		  CONST char *user_string) );


/*  Private data  */
static flag have_aips_id = FALSE;
static char user_string[4];


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
CONST char *foreign_aips1_read_dataset_names (CONST char *dirname,
					      unsigned int *num_datasets)
/*  [SUMMARY] Get the names of AIPS1 datasets in the directory catalogue.
    [PURPOSE] This routine will read the "CA*.???;*" catalogue file in
    the specified directory and extracts the names of the datasets. The
    "AIPS_ID" environment variable should be set to your AIPS user ID.
    <dirname> The directory containing the catalogue file.
    <num_datasets> The number of datasets found is written here.
    [RETURNS] A pointer to a buffer packed with dataset names. Each dataset
    name is separated by a '\0' character. The buffer is internally allocated
    and must not be freed. If no datasets are found NULL is returned.
*/
{
    KDir kdir;
    unsigned int buf_usage = 0;
    KFileInfo *kfile;
    char *ptr;
    CONST char *names = NULL;
    char filename[STRING_LENGTH];
    char id_string[4];
    /*static char function_name[] = "foreign_aips1_read_dataset_names";*/

    *num_datasets = 0;
    get_aips_id ();
    sprintf (filename, "%s/SPACE", dirname);
    if (access (filename, F_OK) != 0) return (NULL);
    if ( (sizeof (int) != 4) || (sizeof (float) != 4) ||
	 (sizeof (double) != 8) ) return (NULL);
    if (have_aips_id)
    {
	names = _foreign_aips1_read_catalogue (dirname, '\0', user_string,
					       num_datasets, &buf_usage);
	if (*num_datasets < 1) return (NULL);
	return (names);
    }
    if ( ( kdir = dir_open (dirname) ) == NULL ) return (NULL);
    while ( ( kfile = dir_read (kdir, KDIR_NO_DOTS) ) != NULL )
    {
	if (strncmp (kfile->filename, "CA", 2) != 0) continue;
	if (strncmp (kfile->filename + 3, "00000", 5) != 0) continue;
	if ( ( ptr = strchr (kfile->filename + 8, '.') ) == NULL ) continue;
	/*  Found one  */
	m_copy (id_string, ptr + 1, 3);
	id_string[3] = '\0';
	names = _foreign_aips1_read_catalogue (dirname, kfile->filename[2],
					       id_string, num_datasets,
					       &buf_usage);
    }
    dir_close (kdir);
    if (*num_datasets < 1) return (NULL);
    return (names);
}   /*  End Function foreign_aips1_read_dataset_names  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_aips1_read (CONST char *setname)
/*  [SUMMARY] Read an AIPS dataset.
    <setname> The AIPS dataset name.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    KAIPS1DataContext context;
    uaddr arr_size;
    array_pointer arrayp;
    multi_array *multi_desc;
    /*static char function_name[] = "foreign_aips1_read";*/

    if ( ( multi_desc = foreign_aips1_read_header (setname, TRUE, &context) )
	 == NULL ) return (NULL);
    /*  Now read data  */
    arr_size =
	ds_get_array_size ( (array_desc *)
			    multi_desc->headers[0]->element_desc[0] );
    m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
    if ( !foreign_aips1_read_data (context, arrayp.array, arr_size) )
    {
	ds_dealloc_multi (multi_desc);
	foreign_aips1_close_data_context (context);
	return (NULL);
    }
    if (context->blank_count > arr_size >> 4)
    {
	fprintf (stderr,
		 "WARNING: %lu out of %lu values are blank: bad data?\n",
		 context->blank_count, arr_size);
    }
    foreign_aips1_close_data_context (context);
    return (multi_desc);
}   /*  End Function foreign_aips1_read  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_aips1_read_header (CONST char *setname, flag data_alloc,
					KAIPS1DataContext *context)
/*  [SUMMARY] Read an AIPS dataset header.
    <setname> The AIPS dataset name.
    <data_alloc> If TRUE, the data space is allocated.
    <context> The data reading context is written here, to be used in
    subsequent calls to [<foreign_aips1_read_data>]. This may be NULL.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    Channel channel;
    flag swap_bytes;
    char version;
    int len, slot_count, slot_offset, user_id;
    unsigned long num_slots;
    char *ptr, *record, *slot;
    double drecord[RECORD_LEN];
    char dirname[STRING_LENGTH], catfile[STRING_LENGTH];
    char sname[STRING_LENGTH], id_string[4];
    static char function_name[] = "foreign_aips1_read_header";

    record = (char *) drecord;
    if ( (sizeof (int) != 4) || (sizeof (float) != 4) ||
	 (sizeof (double) != 8) ) return (NULL);
    strcpy (dirname, setname);
    len = strlen (dirname);
    if (strcmp (dirname + len - 5, ".aips") != 0)
    {
	fprintf (stderr,
		 "%s: dataset \"%s\" does not have \".aips\" extension\n",
		 function_name, setname);
	return (NULL);
    }
    len -= 5;
    dirname[len] = '\0';
    /*  Search backwards for third set of spaces (just before sequence #).
	This convoluted scheme is needed to cope with catalogue names which
	have '/' characters in them  */
    ptr = dirname + len - 1;
    while ( !isspace (*ptr) ) --ptr;
    while ( isspace (*ptr) ) --ptr;
    while ( !isspace (*ptr) ) --ptr;
    user_id = atoi (ptr + 1);
    while ( isspace (*ptr) ) --ptr;
    while ( !isspace (*ptr) ) --ptr;
    ptr -= 19;
    if (ptr < dirname)
    {
	strcpy (sname, dirname);
	dirname[0] = '\0';
    }
    else
    {
	strcpy (sname, ptr + 1);
	ptr[1] = '\0';
    }
    id_to_string (id_string, user_id);
    /*  Find version  */
    for (version = 'D'; version >= 'A'; --version)
    {
	generate_filename (catfile, dirname, "", "CA", version, 0, id_string);
	if (access (catfile, R_OK) == 0) break;
    }
    if (version < 'A')
    {
	fprintf (stderr, "%s: error opening catalogue for set: \"%s\"\n",
		 function_name, setname);
	return (NULL);
    }
    if ( ( channel = ch_open_file (catfile, "r") ) == NULL )
    {
	fprintf (stderr, "%s: error opening catalogue index file: \"%s\"\n",
		 function_name, catfile);
	return (NULL);
    }
    /*  Read the first "header" record  */
    if (ch_read (channel, record, RECORD_BYTES) < RECORD_BYTES)
    {
	fprintf (stderr, "%s: error reading catalogue header\t%s\n",
		 function_name, ERRSTRING);
	ch_close (channel);
	return (NULL);
    }
    slot = record;
    swap_bytes = FALSE;
    GET_VALUE (&num_slots, 8);
    if (num_slots > 100000)
    {
	/*  Assume it's byte-swapped  */
	swap_bytes = TRUE;
	GET_VALUE (&num_slots, 8);
    }
    /*  Now loop through the slots, searching for the dataset  */
    slot_offset = NUM_SLOTS_PER_RECORD;
    for (slot_count = 0; slot_count < num_slots; ++slot_count, ++slot_offset)
    {
	array_pointer arrayp;
	int axis_length, seq, naxis, ival;
	unsigned int dim_count, axis_count;
	unsigned int ktype = NONE;
	float bmin, bmaj, fval;
	struct aips1_data_context_type  ctx;
	multi_array *multi_desc;
	array_desc *arr_desc;
	CONST char *dim_names[MAX_DIMENSIONS];
	char aipsname[STRING_LENGTH], txt[STRING_LENGTH];
	char headername[STRING_LENGTH];
	char bunit[STRING_LENGTH], coord_names[MAX_DIMENSIONS][STRING_LENGTH];
	double value[2];

	if (slot_offset >= NUM_SLOTS_PER_RECORD)
	{
	    if (ch_read (channel, record, RECORD_BYTES) < RECORD_BYTES)
	    {
		fprintf (stderr, "%s: error reading catalogue header\t%s\n",
			 function_name, ERRSTRING);
		ch_close (channel);
		return (NULL);
	    }
	    slot_offset = 0;
	}
	/*  Have a new slot  */
	slot = record + SLOT_BYTES * slot_offset;
	ctx.have_blanking = FALSE;
	ctx.blank_value = 0;
	GET_VALUE (&ival, 0);
	if (user_id != ival) continue;
	if ( (slot[38] != 'M') || (slot[39] != 'A') ) continue;
	GET_VALUE (&seq, 16);
	m_copy (aipsname, slot + 20, 18);
	sprintf (aipsname + 18, " %d %d %d", slot_count + 1, user_id, seq);
	if (strcmp (sname, aipsname) != 0) continue;
	/*  At this point assume any further failures result in the function to
	    return rather than continue the loop. NOTE: catalogue file is
	    closed here  */
	ch_close (channel);
	/*  Construct header filename  */
	generate_filename (headername, dirname, "", "CB", version, slot_count,
			   id_string);
	generate_filename (ctx.datafile, dirname, "", "MA", version,
			   slot_count, id_string);
	if ( (channel = ch_open_file (headername, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening header file: \"%s\"\t%s\n",
		     function_name, headername, ERRSTRING);
	    return (NULL);
	}
	if (ch_read (channel, record, RECORD_BYTES) < RECORD_BYTES)
	{
	    fprintf (stderr, "%s: error reading map header\t%s\n",
		     function_name, ERRSTRING);
	    ch_close (channel);
	    return (NULL);
	}
	/*  At this point assume any further failures result in the function to
	    return rather than continue the loop. NOTE: header file is
	    closed here  */
	ch_close (channel);
	slot = record;
	GET_VALUE (&naxis, KIDIM * 4);
	if (naxis < 2)
	{
	    fprintf (stderr, "%s: one dimensional data not supported\n",
		     function_name);
	    return (NULL);
	}
	for (axis_count = 0, ctx.num_dim = 0; axis_count < naxis; ++axis_count)
	{
	    m_copy (coord_names[axis_count],
		    record + KHCTP * 4 + axis_count * 8, 8);
	    coord_names[axis_count][8] = '\0';
	    strip_trailing_whitespace (coord_names[axis_count]);
	    GET_VALUE (&axis_length, (KINAX + axis_count) * 4);
	    if (axis_length > 1) ++ctx.num_dim;
	}
	for (axis_count = 0, dim_count = ctx.num_dim - 1; axis_count < naxis;
	     ++axis_count)
	{
	    GET_VALUE (&axis_length, (KINAX + axis_count) * 4);
	    if (axis_length < 2) continue;
	    ctx.dim_lengths[dim_count] = axis_length;
	    dim_names[dim_count] = coord_names[axis_count];
	    --dim_count;
	}
	m_copy (bunit, record + KHBUN * 4, 8);
	bunit[8] = '\0';
	strip_trailing_whitespace (bunit);
	ktype = K_FLOAT;
	ptr = bunit;
	arr_desc = ds_easy_alloc_array_desc (ctx.num_dim, ctx.dim_lengths,
					     NULL, NULL, NULL, dim_names,
					     1, &ktype, (CONST char **) &ptr);
	if (arr_desc == NULL) return (NULL);
	if (data_alloc)
	{
	    if ( !ds_alloc_vm_array (&arrayp, arr_desc, FALSE, FALSE) &&
		 !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp, arr_desc,
				       FALSE, FALSE) )
	    {
		m_error_notify (function_name, "array data");
		ds_dealloc_array_desc (arr_desc);
		return (NULL);
	    }
	}
	else m_clear (&arrayp, sizeof arrayp);
	if ( ( multi_desc =
	       ds_easy_alloc_array_from_array_desc (arr_desc, &arrayp, TRUE) )
	     == NULL )
	{
	    ds_dealloc_array_desc (arr_desc);
	    if (arrayp.free != NULL) (*arrayp.free) (arrayp.info, &arrayp);
	}
	/*  Attach the axis information  */
	value[1] = 0.0;
	value[0] = naxis;
	ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				   "NAXIS", K_INT, value, FALSE);
	for (axis_count = 0; axis_count < naxis; ++axis_count)
	{
	    GET_VALUE (&axis_length, (KINAX + axis_count) * 4);
	    sprintf (txt, "NAXIS%d", axis_count + 1);
	    value[0] = axis_length;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_INT, value, FALSE);
	}
	/*  Attach the co-ordinate information  */
	for (axis_count = 0; axis_count < naxis; ++axis_count)
	{
	    /*  CTYPEn  */
	    sprintf (txt, "CTYPE%d", axis_count + 1);
	    ds_put_unique_named_string (multi_desc->headers[0],
					multi_desc->data,
					txt, coord_names[axis_count],
					FALSE);
	    /*  CRVALn  */
	    sprintf (txt, "CRVAL%d", axis_count + 1);
	    GET_VALUE (value, KDCRV * 4 + axis_count * 8);
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_DOUBLE, value, FALSE);
	    /*  CRPIXn  */
	    sprintf (txt, "CRPIX%d", axis_count + 1);
	    GET_VALUE (&fval, KRCRP * 4 + axis_count * 4);
	    value[0] = fval;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_DOUBLE, value, FALSE);
	    /*  CDELTn  */
	    sprintf (txt, "CDELT%d", axis_count + 1);
	    GET_VALUE (&fval, KRCIC * 4 + axis_count * 4);
	    value[0] = fval;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_DOUBLE, value, FALSE);
	    /*  CROTAn  */
	    sprintf (txt, "CROTA%d", axis_count + 1);
	    GET_VALUE (&fval, KRCRT * 4 + axis_count * 4);
	    value[0] = fval;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_DOUBLE, value, FALSE);
	}
	/*  Attach other keywords from first record  */
	/*  BUNIT  */
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "BUNIT", bunit, FALSE);
	/*  OBJECT  */
	m_copy (txt, record + KHOBJ * 4, 8);
	txt[8] = '\0';
	strip_trailing_whitespace (txt);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "OBJECT", txt, FALSE);
	/*  TELESCOP  */
	m_copy (txt, record + KHTEL * 4, 8);
	txt[8] = '\0';
	strip_trailing_whitespace (txt);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "TELESCOP", txt, FALSE);
	/*  INSTRUME  */
	m_copy (txt, record + KHINS * 4, 8);
	txt[8] = '\0';
	strip_trailing_whitespace (txt);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "INSTRUME", txt, FALSE);
	/*  OBSERVER  */
	m_copy (txt, record + KHOBS * 4, 8);
	txt[8] = '\0';
	strip_trailing_whitespace (txt);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "OBSERVER", txt, FALSE);
	/*  EPOCH  */
	GET_VALUE (&fval, KREPO * 4);
	value[0] = fval;
	ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				   "EPOCH", K_DOUBLE, value, FALSE);
	/*  DATAMIN  */
	GET_VALUE (&fval, KRDMN * 4);
	value[0] = fval;
	ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				   "DATAMIN", K_DOUBLE, value, FALSE);
	/*  DATAMAX  */
	GET_VALUE (&fval, KRDMX * 4);
	value[0] = fval;
	ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				   "DATAMAX", K_DOUBLE, value, FALSE);
	GET_VALUE (&ival, KRBLK * 4);
	if (ival != 0)
	{
	    ctx.have_blanking = TRUE;
	    ctx.blank_value = ival;
	}
	GET_VALUE (&bmin, KRBMN * 4);
	GET_VALUE (&bmaj, KRBMJ * 4);
	if ( (bmin != 0.0) && (bmaj != 0.0) )
	{
	    /*  BMIN  */
	    value[0] = bmin;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       "BMIN", K_DOUBLE, value, FALSE);
	    /*  BMAJ  */
	    value[0] = bmaj;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       "BMAJ", K_DOUBLE, value, FALSE);
	    /*  BPA  */
	    GET_VALUE (&fval, KRBPA * 4);
	    value[0] = fval;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       "BPA", K_DOUBLE, value, FALSE);
	}
	if (context == NULL) return (multi_desc);
	if ( ( *context = m_dup (&ctx, sizeof ctx) ) == NULL )
	{
	    m_error_notify (function_name, "data context");
	    ds_dealloc_multi (multi_desc);
	    return (NULL);
	}
	(*context)->magic_number = MAGIC_NUMBER;
	(*context)->channel = NULL;
	(*context)->swap_bytes = swap_bytes;
	(*context)->blank_count = 0;
	(*context)->read_pos = 0;
	return (multi_desc);
    }
    ch_close (channel);
    fprintf (stderr, "%s: dataset \"%s\" not found in catalogue\n",
	     function_name, sname);
    return (NULL);
}   /*  End Function foreign_aips1_read_header  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_aips1_read_data (KAIPS1DataContext context, char *data,
			      uaddr num_values)
/*  [SUMMARY] Read data in an AIPS file.
    [PURPOSE] This routine will read the data of an AIPS dataset from a
    KAIPS1DataContext object. The header section is NOT read.
    <context> The context to read from.
    <data> The array to write the data into. If this is NULL, input data is
    skipped and nothing is written.
    <num_values> The number of values to read.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int rows_per_block;
    unsigned int ktype = K_FLOAT;
    uaddr values_per_row, bytes_per_row, values_per_image;
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "foreign_aips1_read_data";

    VERIFY_CONTEXT (context);
    if (context->channel == NULL)
    {
	if ( ( context->channel = ch_open_file (context->datafile, "rz") )
	     == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, context->datafile, ERRSTRING);
	    return (FALSE);
	}
    }
    values_per_row = context->dim_lengths[context->num_dim - 1];
    values_per_image = values_per_row*context->dim_lengths[context->num_dim-2];
    bytes_per_row = values_per_row * host_type_sizes[ktype];
    if (bytes_per_row > RECORD_BYTES) rows_per_block = 1;
    else rows_per_block = RECORD_BYTES / bytes_per_row;
    while (num_values > 0)
    {
	flag drain = FALSE;
	uaddr pos_in_image = context->read_pos % values_per_image;
	uaddr pos_in_row = pos_in_image % values_per_row;
	uaddr values_to_read, bytes_read;

	values_to_read = num_values;
	if (values_to_read + pos_in_row > values_per_row)
	    values_to_read = values_per_row - pos_in_row;
	if ( ( bytes_read =
	       read_blocks (context->channel, data, values_to_read,
			    host_type_sizes[ktype], context->swap_bytes) )
	     < host_type_sizes[ktype] * values_to_read )
	{
	    fprintf (stderr, "%s: error reading: %lu bytes, got: %lu\t%s\n",
		     function_name, host_type_sizes[ktype] * values_to_read,
		     bytes_read, ERRSTRING);
	    return (FALSE);
	}
	if ( context->have_blanking && (data != NULL) )
	{
	    int val_count;
	    int *iptr;

	    for (val_count = 0, iptr = (int *) data;
		 val_count < values_to_read;
		 ++val_count, ++iptr)
	    {
		if (*iptr == context->blank_value)
		{
		    *(float *) iptr = TOOBIG;
		    ++context->blank_count;
		}
	    }
	}
	if ( (pos_in_row + values_to_read == values_per_row) &&
	     (rows_per_block < 2) ) drain = TRUE;
	if (pos_in_image + values_to_read == values_per_image) drain = TRUE;
	if (drain)
	{
	    if ( !ch_drain_to_boundary (context->channel, RECORD_BYTES) )
	    {
		fprintf (stderr, "%s: error reading to record boundary\n",
			 function_name);
		return (FALSE);
	    }
	}
	num_values -= values_to_read;
	context->read_pos += values_to_read;
	if (data != NULL) data += bytes_read;
    }
    return (TRUE);
}   /*  End Function foreign_aips1_read_data  */

/*EXPERIMENTAL_FUNCTION*/
void foreign_aips1_close_data_context (KAIPS1DataContext context)
/*  [SUMMARY] Close a KAIPS1DataContext object.
    <context> The context.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "foreign_aips1_close_data_context";

    VERIFY_CONTEXT (context);
    if (context->channel != NULL) ch_close (context->channel);
    context->magic_number = 0;
    m_free (context);
}   /*  End Function foreign_aips1_close_data_context  */


/*  Private functions follow  */

static CONST char *_foreign_aips1_read_catalogue (CONST char *dirname,
						  char version,
						  CONST char *user_string,
						  unsigned int *num_datasets,
						  unsigned int *buf_usage)
/*  [SUMMARY] Get the names of AIPS1 datasets in directory catalogue.
    [PURPOSE] This routine will read the "CA*.???;*" catalogue file in
    the specified directory and extracts the names of the datasets.
    <dirname> The directory containing the catalogue file.
    <version> The version of the catalogue format. If '\0', all supported
    versions are tried.
    <user_string> The AIPS user ID string.
    <num_datasets> The number of datasets found is written here.
    <buf_usage> The number of bytes used from the buffer so far.
    [RETURNS] A pointer to a buffer packed with dataset names. Each dataset
    name is separated by a '\0' character. The buffer is internally allocated
    and must not be freed. If no datasets are found NULL is returned.
*/
{
    Channel channel;
    flag swap_bytes;
    int len, slot_count, slot_offset, user_id, seq;
    unsigned int tmp;
    unsigned long num_slots;
    char *ptr, *record, *slot;
    char catfile[STRING_LENGTH], aipsname[STRING_LENGTH];
    double dbuffer[RECORD_LEN];
    static unsigned int buf_size = 0;
    static char *buffer = NULL;
    static char function_name[] = "_foreign_aips1_read_catalogue";

    record = (char *) dbuffer;
    if (version == '\0')
    {
	for (version = 'D'; version >= 'A'; --version)
	{
	    generate_filename (catfile, dirname, "/", "CA", version, 0,
			       user_string);
	    if (access (catfile, R_OK) == 0) break;
	}
	if (version < 'A') return (buffer);
    }
    else generate_filename (catfile, dirname, "/", "CA", version, 0,
			    user_string);
    if ( ( channel = ch_open_file (catfile, "r") ) == NULL ) return (buffer);
    /*  Read the first "header" record  */
    if (ch_read (channel, record, RECORD_BYTES) < RECORD_BYTES)
    {
	fprintf (stderr, "%s: error reading catalogue header\t%s\n",
		 function_name, ERRSTRING);
	ch_close (channel);
	return (buffer);
    }
    slot = record;
    swap_bytes = FALSE;
    GET_VALUE (&num_slots, 8);
#ifdef DEBUG
    fprintf (stderr, "num_slots: %lu\n", num_slots);
#endif
    if (num_slots > 100000)
    {
	/*  Assume it's byte-swapped  */
	swap_bytes = TRUE;
	GET_VALUE (&num_slots, 8);
#ifdef DEBUG
	fprintf (stderr, "Byte swap. num_slots: %lu\n", num_slots);
#endif
    }
    /*  Now loop through the slots  */
    slot_offset = NUM_SLOTS_PER_RECORD;
    for (slot_count = 0; slot_count < num_slots; ++slot_count, ++slot_offset)
    {
	char headername[STRING_LENGTH];

	if (slot_offset >= NUM_SLOTS_PER_RECORD)
	{
	    if (ch_read (channel, record, RECORD_BYTES) < RECORD_BYTES)
	    {
		fprintf (stderr, "%s: error reading catalogue header\t%s\n",
			 function_name, ERRSTRING);
		ch_close (channel);
		*num_datasets = 0;
		return (buffer);
	    }
	    slot_offset = 0;
	}
	/*  Have a new slot  */
	slot = record + SLOT_BYTES * slot_offset;
	GET_VALUE (&user_id, 0);
#ifdef DEBUG
	fprintf (stderr, "  slot: %d  user_id: %d  type: %c%c\n",
		 slot_count, user_id, slot[38], slot[39]);
#endif
	if (user_id < 0) continue;
	if ( (slot[38] != 'M') || (slot[39] != 'A') ) continue;
	GET_VALUE (&seq, 16);
	/*  Construct header filename  */
	generate_filename (headername, dirname, "/", "CB", version, slot_count,
			   user_string);
#ifdef DEBUG
	fprintf (stderr, "  headername: \"%s\"\n", headername);
#endif
	if (access (headername, R_OK) != 0) continue;
	m_copy (aipsname, slot + 20, 18);
	sprintf (aipsname + 18, " %d %d %d", slot_count + 1, user_id, seq);
	len = strlen (aipsname);
	if (*buf_usage + len + 1 >= buf_size)
	{
	    /*  Need to reallocate buffer  */
	    tmp = len % BUFFER_INC;
	    tmp = BUFFER_INC * (tmp + 1) + buf_size;
	    if ( ( ptr = m_alloc (tmp) ) == NULL )
	    {
		m_error_notify (function_name, "name buffer");
		ch_close (channel);
		return (buffer);
	    }
	    if (buffer != NULL)
	    {
		m_copy (ptr, buffer, *buf_usage);
		m_free (buffer);
	    }
	    buffer = ptr;
	    buf_size = tmp;
	}
	strcpy (buffer + *buf_usage, aipsname);
	*buf_usage += len + 1;
	++*num_datasets;
    }
    ch_close (channel);
    return (buffer);
}   /*  End Function read_catalogue  */

static void id_to_string (char *string, int aips_id)
/*  [SUMMARY] Convert an AIPS user ID to a string.
    <string> The string is written here. This will be 4 characters including
    the '\0' terminator.
    <aips_id> The AIPS user ID.
    [RETURNS] Nothing.
*/
{
    int ival;

    string[3] = '\0';
    ival = aips_id % 36;
    string[2] = ival + ( (ival < 10) ? '0' : 'A' - 10 );
    ival = (aips_id / 36) % 36;
    string[1] = ival + ( (ival < 10) ? '0' : 'A' - 10 );
    ival = (aips_id / 36 / 36) % 36;
    string[0] = ival + ( (ival < 10) ? '0' : 'A' - 10 );
}   /*  End Function id_to_string  */

static void strip_trailing_whitespace (char *string)
/*  [SUMMARY] Strip trailing whitespace.
    <string> The string to strip.
    [RETURNS] Nothing.
*/
{
    char *ptr;

    for (ptr = string + strlen (string) - 1; (ptr > string) && isspace (*ptr);
	 --ptr) *ptr = '\0';
}   /*  End Function strip_trailing_whitespace  */

static void get_aips_id ()
{
    CONST char *ptr;
    static flag first_time = TRUE;
    /*static char function_name[] = "_foreign_aips1_read_get_aips_id";*/

    if (!first_time) return;
    first_time = FALSE;
    if ( ( ptr = r_getenv ("AIPS_ID") ) != NULL )
    {
	have_aips_id = TRUE;
	id_to_string ( user_string, atoi (ptr) );
    }
}   /*  End Function get_aips_id  */

static void get_value (void *dest, CONST void *source, uaddr size, flag swap)
{
    if (swap) m_copy_and_swap_blocks (dest, source, size, size, size, 1);
    else m_copy (dest, source, size);
}   /*  End Function get_value  */

static uaddr read_blocks (Channel channel, char *buffer, uaddr num, uaddr size,
			  flag swap_bytes)
/*  [SUMMARY] Read blocks of data from a channel.
    <channel> The channel object.
    <buffer> The buffer to write the data into. If NULL, data are skipped.
    <num> The number of data blocks to read.
    <size> The size of the data blocks.
    <swap_bytes> If TRUE, data is byte-swapped after reading.
    [RETURNS] The number of bytes read.
*/
{
    if (buffer == NULL) return ch_skip (channel, num *size) ? (num * size) : 0;
    if (swap_bytes) return ch_read_and_swap_blocks (channel, buffer, num,size);
    else return ch_read (channel, buffer, num * size);
}   /*  End Function read_blocks  */

static void generate_filename (char filename[STRING_LENGTH],
			       CONST char *dirname, CONST char *separator,
			       CONST char *filetype, char version,
			       int slot_number, CONST char *user_string)
/*  [SUMMARY] Generate filename.
    <filename> The filename is written here.
    <dirname> The directory name.
    <separator> The directory separator.
    <filetype> The file type (i.e. "CA", "CB", "MA", etc.).
    <version> The version.
    <slot_number> The slot number
    <user_string> The user ID string
    [RETURNS] Nothing.
*/
{
    char slot_string[4];

    id_to_string (slot_string, slot_number + 1);
    switch (version)
    {
      case 'A':
      case 'B':
      case 'C':
	if (strcmp (filetype, "CA") == 0)
	    sprintf (filename, "%s%s%s%c00000.%s;1",
		     dirname, separator, filetype, version, user_string);
	else
	    sprintf (filename, "%s%s%s%c%s01.%s;1",
		     dirname, separator, filetype, version, slot_string,
		     user_string);
	break;
      default:
	if (strcmp (filetype, "CA") == 0)
	    sprintf (filename, "%s%s%s%c000000.%s;",
		     dirname, separator, filetype, version, user_string);
	else
	    sprintf (filename, "%s%s%s%c%s001.%s;",
		     dirname, separator, filetype, version, slot_string,
		     user_string);
	break;
    }
}   /*  End Function generate_filename  */
