/*LINTLIBRARY*/
/*  drao_read.c

    This code provides a DRAO read facility.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading data files in
  the DRAO (Dominion Radio Astronomy Observatory) format.


    Written by      Richard Gooch   12-JUL-1997

    Updated by      Richard Gooch   16-JUL-1997

    Updated by      Richard Gooch   7-AUG-1997: Apply negative increment for
  RA/GLON axes.

    Updated by      Richard Gooch   12-AUG-1997: Drag object name from comment
  field if it exists.

    Updated by      Richard Gooch   25-SEP-1997: No longer clear array when
  allocating.

    Updated by      Richard Gooch   26-SEP-1997: Now properly handle subsets.
  Make use of scale factor for float & double data.

    Updated by      Richard Gooch   22-OCT-1997: Added compile-time switchable
  debugging. Now use blank control flag and blank value. Trap factor of 0.

    Updated by      Richard Gooch   23-OCT-1997: Set factor to 1 for unitless.

    Updated by      Richard Gooch   11-NOV-1997: Fixed bug when reading cubes:
  did not read correct number of files when Z axis was subsetted.

    Updated by      Richard Gooch   24-FEB-1998: Switch to new array_pointer
  which has free function.

    Updated by      Richard Gooch   26-FEB-1998: Added catalogue number to
  ensure uniqueness.

    Updated by      Richard Gooch   13-MAY-1998: Added fallback to using anon.
  mmapped array.

    Updated by      Richard Gooch   1-JUN-1999: Fixed handling of RV, VD and XY
  files.

    Updated by      Richard Gooch   3-JUN-1999: Automatically handle byte
  swapping. Fixed bug in RA/GLON scale introduced in previous change.

    Updated by      Richard Gooch   19-AUG-1999: Support catalogue names with
  directory components.

    Updated by      Richard Gooch   26-AUG-1999: Added <<max_width>> parameter
  to function <foreign_drao_read_dataset_names>. Changed format of list of
  names (".f=%03u" is now appended).

    Updated by      Richard Gooch   31-AUG-1999: Fixed bug in function
  <write_sequence_chars>: off by one when <<index>> larger than 358.

    Updated by      Richard Gooch   6-DEC-1999: Support catalogue names with
  full path specification in <foreign_drao_read_dataset_names>.

    Updated by      Richard Gooch   8-DEC-1999: Support catalogue names with
  full path specification in <foreign_drao_read>.

    Updated by      Richard Gooch   9-DEC-1999: Apply scaling factor to integer
  datatypes. Compute data range where convenient.

    Updated by      Richard Gooch   20-APR-2000: Apply scaling factor for
  floats before testing for Karma allowed range. Also test for > -TOOBIG. Added
  warning if too many values are blank.

    Updated by      Richard Gooch   10-MAY-2000: Apply axis stepping factor to
  generated "CRPIXn" values.

    Updated by      Richard Gooch   21-MAY-2000: Added streaming support.

    Updated by      Richard Gooch   14-JUN-2000: Fixed bug introduced by
  streaming support: read_pos was not being updated for image reads.

    Last updated by Richard Gooch   15-AUG-2001: Fixed function
  <foreign_drao_read_header> to continue to next record if datatype is invalid.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_pio.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_p.h>

/*  Work around overly fussy compilers  */
#define ISSPACE(a) isspace((int)(a))
#define ISDIGIT(a) isdigit((int)(a))

#define RECORD_LEN 65     /*  doubles  */
#define RECORD_BYTES 520
#define BUFFER_INC  4060
#define GET_VALUE(out, offset) get_value ((out), (char *) orecord + (offset), \
					  sizeof *(out), swap_bytes)
#define MAGIC_NUMBER 452389324

#define VERIFY_CONTEXT(ctx) if (ctx == NULL) \
{fprintf (stderr, "NULL context passed\n"); \
 a_prog_bug (function_name); } \
if (ctx->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid context object\n"); \
 a_prog_bug (function_name); }


/*  Private structures  */
struct axis_info
{
    int length;
    int sublength;
    int stride;
    unsigned int offset;
};

struct drao_data_context_type
{
    unsigned int magic_number;
    unsigned int elem_size;
    unsigned int ktype;
    Channel channel;
    int fam_num, first_file, fam_off, file_step, num_file, current_file;
    struct axis_info x;
    struct axis_info y;
    uaddr read_pos;
    float factor;
    flag has_blanks;
    flag swap_bytes;
    float blankval;
    float minimum;
    float maximum;
    unsigned long blank_count;
    char datafile[STRING_LENGTH];
};


/*  Private functions  */
STATIC_FUNCTION (uaddr _foreign_drao_read_image,
		 (KDRAODataContext ctx, char *array, uaddr num_values) );
STATIC_FUNCTION (void _foreign_drao_read_scan,
		 (KDRAODataContext ctx, char *array, uaddr num_values) );
STATIC_FUNCTION (flag _foreign_drao_read_cube,
		 (KDRAODataContext ctx, int fam_off, int fam_num,
		  unsigned int first_file, unsigned int num_images,
		  unsigned int file_step, char *array, uaddr num_values) );
STATIC_FUNCTION (void write_sequence_chars,
		 (char *filename, unsigned int index,
		  unsigned int fam_off, unsigned int fam_num) );
STATIC_FUNCTION (void get_value,
		 (void *dest, CONST void *source, uaddr size, flag swap) );
STATIC_FUNCTION (uaddr read_blocks,
		 (Channel channel, char *buffer, uaddr num, uaddr size,
		  flag swap_bytes) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
CONST char *foreign_drao_read_dataset_names (CONST char *dirname,
					     unsigned int *num_datasets,
					     unsigned int *max_width)
/*  [SUMMARY] Get the names of DRAO datasets in the directory catalogue.
    [PURPOSE] This routine will read the "deffil.mad" catalogue file in the
    specified directory and extracts the names of the datasets.
    <dirname> The directory containing the catalogue file.
    <num_datasets> The number of datasets found is written here.
    <max_width> The width of the longest dataset name is written here. This
    may be NULL.
    [RETURNS] A pointer to a buffer packed with dataset names. Each dataset
    name is separated by a '\0' character. The buffer is internally allocated
    and must not be freed. If no datasets are found NULL is returned.
*/
{
    Channel channel;
    int len;
    unsigned int tmp, filenum;
    unsigned int buf_usage = 0;
    char *ptr;
    CONST char *orecord;
    char catfile[STRING_LENGTH], fname[STRING_LENGTH];
    char record[RECORD_BYTES];
    static unsigned int buf_size = 0;
    static char *buffer = NULL;
    static char function_name[] = "foreign_drao_read_dataset_names";

    *num_datasets = 0;
    if (max_width != NULL) *max_width = 0;
    if ( (sizeof (int) != 4) || (sizeof (float) != 4) ||
	 (sizeof (double) != 8) ) return (NULL);
    sprintf (catfile, "%s/deffil.mad", dirname);
    if ( ( channel = ch_open_file (catfile, "r") ) == NULL ) return (NULL);
    orecord = record - 1;
    for (filenum = 1;
	 ch_read (channel, (char *) record, RECORD_BYTES) >= RECORD_BYTES;
	 ++filenum)
    {
	char datafile[STRING_LENGTH];

	/*  Have a new record  */
	flag swap_bytes;
	int first_elem, first_rec, first_file;
	int datatype;
	int fam_off, fam_num, num_elem, num_rec;
	int filetype;

	swap_bytes = FALSE;
	/*  Auto-detect byte order  */
	GET_VALUE (&datatype, 181);
	if ( !(datatype & 0xff) )
	{
	    swap_bytes = TRUE;
	    GET_VALUE (&datatype, 181);
	}
	switch (datatype)
	{
	  case 1:
	  case 2:
	  case 3:
	    break;
	  default:
	    continue;
	    /*break;*/
	}
	GET_VALUE (&filetype, 213);
	if (filetype == 4) continue;
	GET_VALUE (&first_elem, 137);
	GET_VALUE (&first_rec, 149);
	GET_VALUE (&first_file, 161);
	GET_VALUE (&num_elem, 197);
	GET_VALUE (&num_rec, 201);
	/* It appears to be an image or a cube  */
	if ( (first_elem < 1) || (first_rec < 1) || (first_file < 1) ||
	     (num_elem < 1) || (num_rec < 1) )
	{
	    /*  Stupid filename is copied even for empty catalogues: check for
		starting position and size instead  */
	    continue;
	}
	m_copy (datafile, orecord + 377, 52);
	datafile[52] = '\0';
	/*  Strip trailing spaces  */
	for (ptr = datafile + 51; ISSPACE (*ptr); --ptr) *ptr = '\0';
	GET_VALUE (&fam_off, 173);
	/*  Prepend directory if catalogue entry is not absolute path  */
	if (datafile[0] == '/') fname[0] = '\0';
	else
	{
	    strcpy (fname, dirname);
	    strcat (fname, "/");
	    fam_off += strlen (dirname) + 1;
	}
	strcat (fname, datafile);
	sprintf (++ptr, ".f=%03u ", filenum);
	GET_VALUE (&fam_num, 177);
	if (fam_num > 0)
	    write_sequence_chars (fname, first_file - 1, fam_off - 1,
				  fam_num);
	/*  Test for readability of file  */
	if (access (fname, R_OK) != 0) continue;
	len = strlen (datafile);
	if ( (max_width != NULL) && (len > *max_width) ) *max_width = len;
	if (buf_usage + len + 1 >= buf_size)
	{
	    /*  Need to reallocate buffer  */
	    tmp = len % BUFFER_INC;
	    tmp = BUFFER_INC * (tmp + 1) + buf_size;
	    if ( ( ptr = m_alloc (tmp) ) == NULL )
	    {
		m_error_notify (function_name, "name buffer");
		ch_close (channel);
		return (NULL);
	    }
	    if (buffer != NULL)
	    {
		m_copy (ptr, buffer, buf_usage);
		m_free (buffer);
	    }
	    buffer = ptr;
	    buf_size = tmp;
	}
	strcpy (buffer + buf_usage, datafile);
	buf_usage += len + 1;
	++*num_datasets;
    }
    ch_close (channel);
    return (buffer);
}   /*  End Function foreign_drao_read_dataset_names  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_drao_read (CONST char *setname)
/*  [SUMMARY] Read a DRAO dataset.
    <setname> The DRAO dataset name.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    KDRAODataContext context;
    uaddr arr_size;
    array_pointer arrayp;
    multi_array *multi_desc;
    double value[2];
    /*static char function_name[] = "foreign_drao_read";*/

    if ( ( multi_desc = foreign_drao_read_header (setname, TRUE, &context) )
	 == NULL ) return (NULL);
    /*  Now read data  */
    arr_size =
	ds_get_array_size ( (array_desc *)
			    multi_desc->headers[0]->element_desc[0] );
    m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
    if ( !foreign_drao_read_data (context, arrayp.array, arr_size) )
    {
	ds_dealloc_multi (multi_desc);
	foreign_drao_close_data_context (context);
	return (NULL);
    }
    switch (context->ktype)
    {
      case K_FLOAT:
	/*  Scaling factor has been applied  */
	context->factor = 1.0;
	break;
      case K_SHORT:
      case K_INT:
	break;
    }
    if (context->minimum < context->maximum)
    {
	value[0] = context->minimum * context->factor;
	ds_put_unique_named_value (multi_desc->headers[0],
				   multi_desc->data,
				   "DATAMIN", K_DOUBLE, value, FALSE);
	value[0] = context->maximum * context->factor;
	ds_put_unique_named_value (multi_desc->headers[0],
				   multi_desc->data,
				   "DATAMAX", K_DOUBLE, value, FALSE);
    }
    if (context->blank_count > arr_size >> 4)
    {
	fprintf (stderr,
		 "WARNING: %lu out of %lu values are blank: bad data?\n",
		 context->blank_count, arr_size);
    }
    foreign_drao_close_data_context (context);
    return (multi_desc);
}   /*  End Function foreign_drao_read  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_drao_read_header (CONST char *setname, flag data_alloc,
				       KDRAODataContext *context)
/*  [SUMMARY] Read a DRAO dataset header.
    <setname> The DRAO dataset name.
    <data_alloc> If TRUE, the data space is allocated.
    <context> The data reading context is written here, to be used in
    subsequent calls to [<foreign_drao_read_data>]. This may be NULL.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    Channel channel;
    int len, dirlen;
    unsigned int fcount;
    unsigned int filenum = 0;
    char *ptr;
    CONST char *orecord, *comment;
    double record[RECORD_LEN];
    char dirname[STRING_LENGTH], catfile[STRING_LENGTH], sname[STRING_LENGTH];
    static char function_name[] = "foreign_drao_read_header";

    if ( (sizeof (int) != 4) || (sizeof (float) != 4) ||
	 (sizeof (double) != 8) ) return (NULL);
    len = strlen (setname) - 5;
    if (strcmp (setname + len, ".drao") != 0)
    {
	fprintf (stderr,
		 "%s: dataset \"%s\" does not have \".drao\" extension\n",
		 function_name, setname);
	return (NULL);
    }
    strncpy (dirname, setname, len);
    dirname[len] = '\0';
    /*  Extract file number and discard from name  */
    for (ptr = dirname + len - 1; (ptr > dirname) && ISDIGIT (*ptr); --ptr);
    ptr -= 2;
    if ( (ptr > dirname) && (strncmp (ptr, ".f=", 3) == 0) )
    {
	/*  There is a file number encoded in the filename  */
	filenum = atoi (ptr + 3);
	*ptr = '\0';
	len -= 3;
    }
    /*  Extract directory name and construct catalogue file  */
    if ( ( ptr = strrchr (dirname, '/') ) == NULL )
    {
	strcpy (sname, dirname);
	dirname[0] = '\0';
	dirlen = 0;
    }
    else
    {
	strcpy (sname, ptr + 1);
	ptr[1] = '\0';
	dirlen = ptr - dirname + 1;
    }
    strcpy (catfile, dirname);
    strcpy (catfile + dirlen, "deffil.mad");
    if ( ( channel = ch_open_file (catfile, "r") ) == NULL )
    {
	fprintf (stderr, "%s: error opening catalogue file: \"%s\"\t%s\n",
		 function_name, catfile, ERRSTRING);
	return (NULL);
    }
    orecord = (char *) record - 1;
    comment = (CONST char *) (orecord + 297);
    for (fcount = 1;
	 ch_read (channel, (char *) record, RECORD_BYTES) >= RECORD_BYTES;
	 ++fcount)
    {
	/*  Have a new record  */
	flag swap_bytes;
	array_pointer arrayp;
	int count, zval_defined;
	int first_elem;
	int first_rec;
	int tot_file;
	int datatype, filetype;
	int unit, iproj, epochc, velref, stokes, hasblanks;
	unsigned int num_axes;
	float fact, bmaj, bmin, bpos;
	double zval;
	struct drao_data_context_type ctx;
	multi_array *multi_desc;
	array_desc *arr_desc;
	char name_in_catalogue[53];
	float refpix[3], delta[3];
	double refval[3], value[2];
	CONST char *dim_names[3];
	char unit_name[STRING_LENGTH], coord_names[3][STRING_LENGTH];
	char object[STRING_LENGTH];
	uaddr axis_lengths[3];
	double axis_scales[3];
	extern char host_type_sizes[NUMTYPES];

	if ( (filenum != 0) && (fcount != filenum) ) continue;
	swap_bytes = FALSE;
	/*  Auto-detect byte order  */
	GET_VALUE (&datatype, 181);
	if ( !(datatype & 0xff) )
	{
	    swap_bytes = TRUE;
	    GET_VALUE (&datatype, 181);
	}
	switch (datatype)
	{
	  case 1:
	    ctx.ktype = K_SHORT;
	    break;
	  case 2:
	    ctx.ktype = K_FLOAT;
	    break;
	  case 3:
	    ctx.ktype = K_LONG;
	    break;
	  default:
	    continue;
	    /*break;*/
	}
	GET_VALUE (&filetype, 213);
	if (filetype == 4) continue;
	GET_VALUE (&refval[0], 1);
	GET_VALUE (&refval[1], 9);
	GET_VALUE (&refval[2], 17);
	GET_VALUE (&zval, 41);
	GET_VALUE (&refpix[0], 65);
	GET_VALUE (&refpix[1], 69);
	GET_VALUE (&refpix[2], 73);
	GET_VALUE (&delta[0], 77);
	GET_VALUE (&delta[1], 81);
	GET_VALUE (&delta[2], 85);
	GET_VALUE (&ctx.blankval, 105);
	GET_VALUE (&fact, 109);
	GET_VALUE (&bmaj, 113);
	GET_VALUE (&bmin, 117);
	GET_VALUE (&bpos, 121);
	GET_VALUE (&first_elem, 137);
	GET_VALUE (&ctx.x.stride, 145);  /*  elem_step  */
	GET_VALUE (&first_rec, 149);
	GET_VALUE (&ctx.y.stride, 157);  /*  rec_step  */
	GET_VALUE (&ctx.first_file, 161);
	GET_VALUE (&ctx.file_step, 169);
	GET_VALUE (&ctx.fam_off, 173);
	GET_VALUE (&ctx.fam_num, 177);
	GET_VALUE (&datatype, 181);
	GET_VALUE (&ctx.x.length, 185);  /*  tot_elem  */
	GET_VALUE (&ctx.y.length, 189);  /*  tot_rec  */
	GET_VALUE (&tot_file, 193);
	GET_VALUE (&ctx.x.sublength, 197);      /*  num_elem  */
	GET_VALUE (&ctx.y.sublength, 201);  /*  num_rec  */
	GET_VALUE (&ctx.num_file, 205);
	GET_VALUE (&unit, 209);
	GET_VALUE (&iproj, 221);
	GET_VALUE (&epochc, 225);
	GET_VALUE (&velref, 229);
	GET_VALUE (&stokes, 233);
	GET_VALUE (&hasblanks, 273);
	GET_VALUE (&zval_defined, 281);
	axis_scales[0] = 1.0;
	axis_scales[1] = 1.0;
	axis_scales[2] = 1.0;
	/* It appears to be an image or a cube  */
	if ( (first_elem < 1) || (first_rec < 1) || (ctx.first_file < 1) ||
	     (ctx.x.sublength < 1) || (ctx.y.sublength < 1) )
	{
	    /*  Stupid filename is copied even for empty catalogues: check for
		size instead  */
	    continue;
	}
	m_copy (name_in_catalogue, orecord + 377, 52);
	name_in_catalogue[52] = '\0';
	/*  Strip trailing spaces  */
	for (ptr = name_in_catalogue + 51; ISSPACE (*ptr); --ptr) *ptr = '\0';
	/*  Do filename comparison without directory component  */
	if ( ( ptr = strrchr (name_in_catalogue, '/') ) == NULL )
	    ptr = name_in_catalogue;
	else ++ptr;
	if (strcmp (sname, ptr) != 0) continue;
	/*  At this point assume any further failures result in the function to
	    return rather than continue the loop. NOTE: catalogue file is
	    closed here  */
	ch_close (channel);
	if (name_in_catalogue[0] == '/')
	    strcpy (ctx.datafile, name_in_catalogue);
	else
	{
	    strcpy (ctx.datafile, dirname);
	    strcat (ctx.datafile + dirlen, name_in_catalogue);
	    ctx.fam_off += dirlen;
	}
#ifdef DEBUG
	fprintf (stderr, "%s: reading dataset: \"%s\" index: %u\n",
		 function_name, setname, fcount);
#endif
	/*  Get unit string  */
	switch (unit)
	{
	  case 0:
	    strcpy (unit_name, "Data Value");
	    break;
	  case 1:
	    strcpy (unit_name, "JY/BEAM");
	    break;
	  case 2:
	    strcpy (unit_name, "JY/PIX");
	    break;
	  case 3:
	    strcpy (unit_name, "JY");
	    break;
	  case 4:
	    strcpy (unit_name, "K");
	    break;
	  case 5:
	    strcpy (unit_name, "RADIANS");
	    break;
	  case 6:
	    strcpy (unit_name, "DEGREES");
	    break;
	  case 7:
	    m_copy (unit_name, orecord + 429, 32);
	    unit_name[32] = '\0';
	    for (ptr = unit_name + 31; ISSPACE (*ptr); --ptr) *ptr = '\0';
	    break;
	  default:
	    fprintf (stderr, "Illegal unit type: %d\n", unit);
	    continue;
	    /*break;*/
	}
#ifdef DEBUG
	fprintf (stderr,
		 "%s: ktype: %u factor: %g hasblanks: %d blankval: %g\n",
		 function_name, ctx.ktype, fact, hasblanks, ctx.blankval);
#endif
	if (filetype == 5)  /*  XY: user defined  */
	{
	    for (count = 0; count < 3; ++count)
	    {
		m_copy (coord_names[count], orecord + 461 + count * 8, 8);
		coord_names[count][8] = '\0';
		for (ptr = coord_names[count] + 7;
		     ISSPACE (*ptr) && (ptr >= coord_names[count]);
		     --ptr) *ptr = '\0';
	    }
	}
	else  /*  RD/RV/VD: sky co-ordinates  */
	{
	    double spec_scale;
	    char *lon_name, *lat_name, *spec_name;

	    /*  Setup X and Y co-ordinate axes  */
	    switch (iproj)
	    {
	      case 1:
		/*  S  */
		lon_name = "RA---NCP";
		lat_name = "DEC--NCP";
		break;
	      case 2:
		/*  E  */
		lon_name = "RA---FLT";
		lat_name = "DEC--FLT";
		break;
	      case 3:
		/*  G  */
		lon_name = "GLON-FLT";
		lat_name = "GLAT-FLT";
		break;
	      case 4:
		/*  T  */
		lon_name = "RA---TAN";
		lat_name = "DEC--TAN";
		break;
	      case 5:
		/*  V  */
		lon_name = "RA---SIN";
		lat_name = "DEC--SIN";
		break;
	      case 6:
		/*  A  */
		lon_name = "RA---ARC";
		lat_name = "DEC--ARC";
		break;
	      case 7:
		/*  C  */
		lon_name = "RA---CYL";
		lat_name = "DEC--CYL";
		break;
	      case 8:
		/*  F  */
		lon_name = "RA---AIT";
		lat_name = "DEC--AIT";
		break;
	      default:
		fprintf (stderr, "%s: unknown projection: %d\n",
			 function_name, iproj);
		return (NULL);
		/*break;*/
	    }
	    /* Setup Z co-ordinate axis  */
	    switch (velref)
	    {
	      case 0:
		spec_name = "FREQ";
		spec_scale = 1e6;
		break;
	      case 1:
		spec_name = "VELO-LSR";
		spec_scale = 1e3;
		break;
	      case 2:
		spec_name = "VELO-HELO";
		spec_scale = 1e3;
		break;
	      default:
		fprintf (stderr, "%s: unknown velocity type: %d\n",
			 function_name, velref);
		return (NULL);
		/*break;*/
	    }
	    switch (filetype)
	    {
	      case 1:  /*  RD: ra/glon,  dec/glat, vel/freq (optional)  */
		strcpy (coord_names[0], lon_name);
		strcpy (coord_names[1], lat_name);
		strcpy (coord_names[2], spec_name);
		axis_scales[0] = 180.0 / PI;
		axis_scales[1] = 180.0 / PI;
		axis_scales[2] = spec_scale;
		delta[0] *= -1.0;
		break;
	      case 2:  /*  RV: ra/glon,  vel/freq, dec/glat (optional)  */
		strcpy (coord_names[0], lon_name);
		strcpy (coord_names[1], spec_name);
		strcpy (coord_names[2], lat_name);
		axis_scales[0] = 180.0 / PI;
		axis_scales[1] = spec_scale;
		axis_scales[2] = 180.0 / PI;
		delta[0] *= -1.0;
		break;
	      case 3:  /*  VD: vel/freq, dec/glat, ra/glon  (optional)  */
		strcpy (coord_names[0], spec_name);
		strcpy (coord_names[1], lat_name);
		strcpy (coord_names[2], lon_name);
		axis_scales[0] = spec_scale;
		axis_scales[1] = 180.0 / PI;
		axis_scales[2] = 180.0 / PI;
		delta[2] *= -1.0;
		break;
	    }
	}
	num_axes = (ctx.num_file < 2) ? 2 : 3;
	if (num_axes == 2)
	{
	    axis_lengths[0] = ctx.y.sublength;
	    axis_lengths[1] = ctx.x.sublength;
	    dim_names[0] = coord_names[1];
	    dim_names[1] = coord_names[0];
	}
	else
	{
	    axis_lengths[0] = ctx.num_file;
	    axis_lengths[1] = ctx.y.sublength;
	    axis_lengths[2] = ctx.x.sublength;
	    dim_names[0] = coord_names[2];
	    dim_names[1] = coord_names[1];
	    dim_names[2] = coord_names[0];
	}
	ptr = unit_name;
	/*  Allocate the array manaully so that clearing is not done  */
	arr_desc = ds_easy_alloc_array_desc (num_axes, axis_lengths,
					     NULL, NULL, NULL, dim_names,
					     1, &ctx.ktype,
					     (CONST char **) &ptr);
	if (arr_desc == NULL) return (NULL);
#ifdef DEBUG
	fprintf ( stderr, "%s: allocating array of size: %lu bytes\n",
		  function_name, ds_get_array_size (arr_desc) *
		  ds_get_packet_size (arr_desc->packet) );
#endif
	if (data_alloc)
	{
	    if ( !ds_alloc_vm_array (&arrayp, arr_desc, FALSE, FALSE) &&
		 !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp, arr_desc,
				       FALSE, FALSE) )
	    {
		m_error_notify (function_name, "array data");
		ds_dealloc_array_desc (arr_desc);
		return (NULL);
	    }
	}
	else m_clear (&arrayp, sizeof arrayp);
	if ( ( multi_desc =
	       ds_easy_alloc_array_from_array_desc (arr_desc, &arrayp, TRUE) )
	     == NULL )
	{
	    ds_dealloc_array_desc (arr_desc);
	    if (arrayp.free != NULL) (*arrayp.free) (arrayp.info, &arrayp);
	    return (NULL);
	}
	/*  Setup data context structure  */
	ctx.elem_size = host_type_sizes[ctx.ktype];
	ctx.factor = (unit == 0) ? 1.0 : fact;
	if (ctx.factor == 0.0)
	{
	    fprintf (stderr, "%s: data scale factor is zero, forcing to 1\n",
		     function_name);
	    ctx.factor = 1.0;
	}
	ctx.has_blanks = hasblanks ? TRUE : FALSE;
	ctx.x.offset = first_elem - 1;
	ctx.y.offset = first_rec - 1;
	ctx.swap_bytes = swap_bytes;
	/*  Write header information  */
	value[1] = 0.0;
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "CTYPE1", coord_names[0], FALSE);
	value[0] = refval[0] * axis_scales[0];
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "CRVAL1", K_DOUBLE, value, FALSE);
	value[0] = (refpix[0] - 1.0 - ctx.x.offset) /
	    (double) ctx.x.stride + 1.0;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "CRPIX1", K_DOUBLE, value, FALSE);
	value[0] = delta[0] * axis_scales[0] * ctx.x.stride;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "CDELT1", K_DOUBLE, value, FALSE);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "CTYPE2", coord_names[1], FALSE);
	value[0] = refval[1] * axis_scales[1];
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "CRVAL2", K_DOUBLE, value, FALSE);
	value[0] = (refpix[1] - 1.0 - ctx.y.offset) /
	    (double) ctx.y.stride + 1.0;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "CRPIX2", K_DOUBLE, value, FALSE);
	value[0] = delta[1] * axis_scales[1] * ctx.y.stride;
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   "CDELT2", K_DOUBLE, value, FALSE);
	if ( zval_defined || (num_axes > 2) )
	{
	    ds_put_unique_named_string (multi_desc->headers[0],
					multi_desc->data,
					"CTYPE3", coord_names[2], FALSE);
	    if (num_axes < 3)
	    {
		value[0] = zval * axis_scales[2];
		ds_put_unique_named_value (multi_desc->headers[0],
					   multi_desc->data,
					   "CRVAL3", K_DOUBLE, value, FALSE);
		value[0] = 1.0;
		ds_put_unique_named_value (multi_desc->headers[0],
					   multi_desc->data,
					   "CRPIX3", K_DOUBLE, value, FALSE);
		value[0] = 0.0;
		ds_put_unique_named_value (multi_desc->headers[0],
					   multi_desc->data,
					   "CDELT3", K_DOUBLE, value, FALSE);
	    }
	    else
	    {
		value[0] = refval[2] * axis_scales[2];
		ds_put_unique_named_value (multi_desc->headers[0],
					   multi_desc->data,
					   "CRVAL3", K_DOUBLE, value, FALSE);
		value[0] = (refpix[2] -
			    ctx.first_file) / (double) ctx.file_step + 1.0;
		ds_put_unique_named_value (multi_desc->headers[0],
					   multi_desc->data,
					   "CRPIX3", K_DOUBLE, value, FALSE);
		value[0] = delta[2] * axis_scales[2] * ctx.file_step;
		ds_put_unique_named_value (multi_desc->headers[0],
					   multi_desc->data,
					   "CDELT3", K_DOUBLE, value, FALSE);
	    }
	}
	m_copy (object, comment, 80);
	object[80] = '\0';
	for (ptr = object + 79; (ptr > object) && ISSPACE (*ptr); --ptr)
	    *ptr = '\0';
	if (strncmp (object, "OBJECT = ", 9) == 0)
	    ds_put_unique_named_string (multi_desc->headers[0],
					multi_desc->data,
					"OBJECT", object + 9, FALSE);
	/*  Apply data scaling information if appropriate  */
	switch (ctx.ktype)
	{
	  case K_FLOAT:
	    /*  Performed on read  */
	    break;
	  case K_SHORT:
	  case K_INT:
	    if (ctx.factor != 1.0)
	    {
		value[0] = ctx.factor;
		ds_put_unique_named_value (multi_desc->headers[0],
					   multi_desc->data,
					   "BSCALE", K_DOUBLE, value, FALSE);
	    }
	    break;
	}
	if (context == NULL) return (multi_desc);
	if ( ( *context = m_dup (&ctx, sizeof ctx) ) == NULL )
	{
	    m_error_notify (function_name, "data context");
	    ds_dealloc_multi (multi_desc);
	    return (NULL);
	}
	(*context)->magic_number = MAGIC_NUMBER;
	(*context)->channel = NULL;
	foreign_drao_rewind_data_context (*context);
	return (multi_desc);
    }
    ch_close (channel);
    fprintf (stderr, "%s: dataset \"%s\" not found in catalogue\n",
	     function_name, sname);
    return (NULL);
}   /*  End Function foreign_drao_read_header  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_drao_read_data (KDRAODataContext context, char *data,
			     uaddr num_values)
/*  [SUMMARY] Read data in a DRAO file.
    [PURPOSE] This routine will read the data of a DRAO dataset from a
    KDRAODataContext object. The header section is NOT read.
    <context> The context to read from.
    <data> The array to write the data into. If this is NULL, input data is
    skipped and nothing is written.
    <num_values> The number of values to read.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "foreign_drao_read_data";

    VERIFY_CONTEXT (context);
    if (context->fam_num < 1)
    {
	/*  Single 2D file  */
	if (context->channel == NULL)
	{
	    if ( ( context->channel = ch_open_file (context->datafile, "rz") )
		 == NULL )
	    {
		fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
			 function_name, context->datafile, ERRSTRING);
		return (FALSE);
	    }
	}
	while (num_values > 0)
	{
	    uaddr read_values;

	    if ( ( read_values =
		   _foreign_drao_read_image (context, data,num_values) ) == 0 )
	    {
		fprintf (stderr, "%s: error reading image\n", function_name);
		return (FALSE);
	    }
	    _foreign_drao_read_scan (context, data, read_values);
	    num_values -= read_values;
	    if (data != NULL) data += read_values * context->elem_size;
	}
    }
    else
    {
	/*  3D dataset is a collection of files: stupid design  */
	if ( !_foreign_drao_read_cube (context, context->fam_off - 1,
				       context->fam_num,
				       context->first_file - 1,
				       context->num_file,
				       context->file_step, data, num_values) )
	{
	    fprintf (stderr, "%s: error reading image files for cube\n",
		     function_name);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function foreign_drao_read_data  */

/*EXPERIMENTAL_FUNCTION*/
void foreign_drao_rewind_data_context (KDRAODataContext context)
/*  [SUMMARY] Rewind a KDRAODataContext object.
    <context> The context.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "foreign_drao_rewind_data_context";

    VERIFY_CONTEXT (context);
    if (context->channel != NULL)
    {
	ch_close (context->channel);
	context->channel = NULL;
    }
    context->read_pos = 0;
    context->current_file = -1;
    context->minimum = TOOBIG;
    context->maximum = -TOOBIG;
    context->blank_count = 0;
}   /*  End Function foreign_drao_rewind_data_context  */

/*EXPERIMENTAL_FUNCTION*/
void foreign_drao_close_data_context (KDRAODataContext context)
/*  [SUMMARY] Close a KDRAODataContext object.
    <context> The context.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "foreign_drao_close_data_context";

    VERIFY_CONTEXT (context);
    if (context->channel != NULL) ch_close (context->channel);
    context->magic_number = 0;
    m_free (context);
}   /*  End Function foreign_drao_close_data_context  */


/*  Private functions follow  */

static uaddr _foreign_drao_read_image (KDRAODataContext ctx, char *array,
				       uaddr num_values)
/*  [SUMMARY] Read a single image in DRAO format.
    <ctx> The KDRAODataContext object, specifying image parameters (size,
    stride, etc).
    <array> The array to write the data into. If this is NULL, input data is
    skipped and nothing is written.
    <num_values> The number of values to read/skip. The function may read less
    than this, in which case further calls should be made.
    [RETURNS] The number of bytes read on success, else 0.
*/
{
    uaddr pos_in_image, pos_in_line, image_size;
    unsigned int x, bytes_to_read, bytes_read;
    static char function_name[] = "_foreign_drao_read_image";

    image_size = ctx->x.sublength * ctx->y.sublength;
    pos_in_image = ctx->read_pos % image_size;
    pos_in_line = pos_in_image % ctx->x.sublength;
    if (pos_in_image == 0)
    {
	/*  Skip to first record in file  */
	if ( !ch_skip (ctx->channel,
		       ctx->y.offset * ctx->x.length * ctx->elem_size) )
	{
	    fprintf (stderr,
		     "%s: error skipping to start of first file record\n",
		     function_name);
	    return (0);
	}
    }
    if ( (ctx->x.sublength == ctx->x.length) && (ctx->y.stride == 1) )
    {
	/*  Records (lines) are contiguous: do it the fast way  */
	bytes_to_read = ctx->elem_size * num_values;
	if ( ( bytes_read = read_blocks (ctx->channel, array, num_values,
					 ctx->elem_size, ctx->swap_bytes) )
	     < bytes_to_read )
	{
	    fprintf (stderr, "%s: error reading %u bytes, got: %u\t%s\n",
		     function_name, bytes_to_read, bytes_read, ERRSTRING);
	    return (0);
	}
	ctx->read_pos += num_values;
	return (num_values);
    }
    /*  Records (lines) are not contiguous: do it the slow way  */
    if (num_values + pos_in_line > ctx->x.sublength)
	num_values = ctx->x.sublength - pos_in_line;
    if (pos_in_line == 0)
    {
	/*  Skip to start of record  */
	if ( !ch_skip (ctx->channel, ctx->x.offset * ctx->elem_size) )
	{
	    fprintf (stderr,
		     "%s: error skipping %u bytes to start of subrecord\n",
		     function_name, ctx->x.offset * ctx->elem_size);
	    return (0);
	}
    }
    if (ctx->x.stride == 1)
    {
	/*  Elements are contiguous  */
	if ( ( bytes_read = read_blocks (ctx->channel, array, num_values,
					 ctx->elem_size, ctx->swap_bytes) )
	     < num_values * ctx->elem_size )
	{
	    fprintf (stderr,
		     "%s: error reading %lu bytes, got: %u\t%s\n",
		     function_name, num_values * ctx->elem_size, bytes_read,
		     ERRSTRING);
	    return (0);
	}
    }
    else
    {
	/*  Elements are spaced  */
	for (x = 0; x < num_values; ++x)
	{
	    if ( ( bytes_read =
		   read_blocks (ctx->channel, array, 1, ctx->elem_size,
				ctx->swap_bytes) ) < ctx->elem_size )
	    {
		fprintf (stderr,
			 "%s: error reading %u bytes, got: %u\t%s\n",
			 function_name, ctx->elem_size, bytes_read,
			 ERRSTRING);
		return (0);
	    }
	    if (pos_in_line + x + 1 < ctx->x.sublength) /*  Not end of line  */
	    {
		/*  Skip to next element  */
		if ( !ch_skip (ctx->channel,
			       (ctx->x.stride - 1) * ctx->elem_size) )
		{
		    fprintf (stderr, "%s: error skipping to next element\n",
			     function_name);
		    return (0);
		}
	    }
	    if (array != NULL) array += ctx->elem_size;
	}
    }
    if (pos_in_line + num_values >= ctx->x.sublength)
    {
	/*  Skip to end of record (line)  */
	if ( !ch_skip (ctx->channel,
		       ( ctx->x.length -
			 (ctx->x.offset + (ctx->x.sublength - 1) *
			  ctx->x.stride + 1) ) * ctx->elem_size) )
	{
	    fprintf (stderr,"%s: error skipping to end of record\n",
		     function_name);
	    return (0);
	}
	if ( (pos_in_image + num_values < image_size) && (ctx->y.stride > 1) )
	{
	    /*  Skip to start of next record (line)  */
	    if ( !ch_skip (ctx->channel, (ctx->y.stride - 1) * ctx->x.length *
			   ctx->elem_size) )
	    {
		fprintf (stderr,"%s: error skipping to next record\n",
			 function_name);
		return (0);
	    }
	}
    }
    ctx->read_pos += num_values;
    return (num_values);
}   /*  End Function _foreign_drao_read_image  */

static void _foreign_drao_read_scan (KDRAODataContext ctx, char *array,
				     uaddr num_values)
/*  [SUMMARY] Scan data values and apply scaling.
    <ctx> The KDRAODataContext object, specifying image parameters (size,
    stride, etc).
    <array> The array. This may be NULL.
    <num_values> The number of values to scan.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_foreign_drao_read_scan";

    if (array == NULL) return;
    if ( (ctx->factor == 1.0) && !ctx->has_blanks ) return;
    /*  May need to convert blanks or apply scaling factor  */
    if (ctx->ktype == K_FLOAT)
    {
	float val;
	float blankval = ctx->blankval;
	float factor = ctx->factor;
	float toobig = TOOBIG;
	float *ptr;

	if (ctx->has_blanks)
	{
	    for (ptr = (float *) array; num_values > 0; --num_values, ++ptr)
	    {
		val = *ptr;
		if (val == blankval)
		{
		    *ptr = toobig;
		    ++ctx->blank_count;
		}
		else
		{
		    val *= factor;
		    if ( (val < toobig) && (val > -toobig) )
		    {
			if (val < ctx->minimum) ctx->minimum = val;
			if (val > ctx->maximum) ctx->maximum = val;
			*ptr = val;
		    }
		    else
		    {
			*ptr = toobig;
			++ctx->blank_count;
		    }
		}
	    }
	}
	else
	{
	    for (ptr = (float *) array; num_values > 0; --num_values, ++ptr)
	    {
		val = *ptr *  factor;
		if ( (val < toobig) && (val > -toobig) )
		{
		    if (val < ctx->minimum) ctx->minimum = val;
		    if (val > ctx->maximum) ctx->maximum = val;
		    *ptr = val;
		}
		else
		{
		    *ptr = toobig;
		    ++ctx->blank_count;
		}
	    }
	}
	return;
    }
    /*  Integer types: only need to search for blanks  */
    if (!ctx->has_blanks) return;
    if (ctx->ktype == K_SHORT)
    {
	short val;
	short blankval = ctx->blankval;
	short *ptr;

	for (ptr = (short *) array; num_values > 0; --num_values, ++ptr)
	{
	    val = *ptr;
	    if (val == blankval)
	    {
		*ptr = -32768;
		++ctx->blank_count;
	    }
	    else
	    {
		float fval = val;

		if (fval < ctx->minimum) ctx->minimum = fval;
		if (fval > ctx->maximum) ctx->maximum = fval;
	    }
	}
    }
    else if (ctx->ktype == K_INT)
    {
	int val;
	int blankval = ctx->blankval;
	int *ptr;

	for (ptr = (int *) array; num_values > 0; --num_values, ++ptr)
	{
	    val = *ptr;
	    if (val == blankval)
	    {
		*(unsigned int *) ptr = 0x80000000;
		++ctx->blank_count;
	    }
	    else
	    {
		float fval = val;

		if (fval < ctx->minimum) ctx->minimum = fval;
		if (fval > ctx->maximum) ctx->maximum = fval;
	    }
	}
    }
    else
    {
	fprintf (stderr, "%s: illegal data type: %u\n",
		 function_name, ctx->ktype);
    }
}   /*  End Function _foreign_drao_read_scan  */

static flag _foreign_drao_read_cube (KDRAODataContext ctx,
				     int fam_off, int fam_num,
				     unsigned int first_file,
				     unsigned int num_images,
				     unsigned int file_step, char *array,
				     uaddr num_values)
/*  [SUMMARY] Read a DRAO cube from a "family" of files.
    <ctx> The KDRAODataContext object, specifying image parameters (size,
    stride, etc).
    <fam_off> The offset of the first family character, counting from 0.
    <fam_num> The number of family characters.
    <first_file> The first file in the family, counting from 0.
    <num_images> The number of images.
    <file_step> The step between files.
    <array> The array into which to write the data into.
    <num_values> The number of values to read/skip.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    uaddr image_size;
    static char function_name[] = "_foreign_drao_read_cube";

    image_size = ctx->x.sublength * ctx->y.sublength;
    while (num_values > 0)
    {
	unsigned int image_num = ctx->read_pos / image_size;
	int fnum = first_file + image_num * file_step;
	uaddr read_values;
	char fname[STRING_LENGTH];

	if (image_num >= 359 * 36)
	{
	    fprintf (stderr, "%s: number of images: %u not supported\n",
		     function_name, image_num);
	    return (FALSE);
	}
	if ( (array == NULL) && (num_values >= image_size) )
	{
	    num_values -= image_size;
	    ctx->read_pos += image_size;
	    continue;
	}
	strcpy (fname, ctx->datafile);
	write_sequence_chars (fname, fnum, fam_off, fam_num);
	if (fnum != ctx->current_file)
	{
	    if (ctx->channel != NULL)
	    {
		ch_close (ctx->channel);
		ctx->channel = NULL;
	    }
	    ctx->current_file = -1;
	}
	if (ctx->channel == NULL)
	{
	    if ( ( ctx->channel = ch_open_file (fname, "rz") ) == NULL )
	    {
		fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
			 function_name, fname, ERRSTRING);
		return (FALSE);
	    }
	    ctx->current_file = fnum;
	}
	read_values = (ctx->read_pos + num_values) / image_size;
	if (read_values == image_num) read_values = num_values;
	else read_values = (image_num + 1) * image_size - ctx->read_pos;
	if ( ( read_values = _foreign_drao_read_image (ctx, array,
						       read_values) ) == 0 )
	{
	    fprintf (stderr, "%s: error reading image file: \"%s\"\n",
		     function_name, fname);
	    ch_close (ctx->channel);
	    ctx->channel = NULL;
	    ctx->current_file = -1;
	    return (FALSE);
	}
	_foreign_drao_read_scan (ctx, array, read_values);
	num_values -= read_values;
	if (array != NULL) array += read_values * ctx->elem_size;
    }
    return (TRUE);
}   /*  End Function _foreign_drao_read_cube  */

static void write_sequence_chars (char *filename, unsigned int index,
				  unsigned int fam_off, unsigned int fam_num)
/*  [SUMMARY] Write the sequence characters to a filename.
    <filename> The filename to write to.
    <index> The index in the sequence counting from 0.
    <fam_off> The offset of the first family character, counting from 0.
    <fam_num> The number of family characters.
    [RETURNS] Nothing.
*/
{
    int digit1, digit2, digit3;
    unsigned int count;

    /*  Initialise counter  */
    for (count = 0; count < fam_num; ++count) filename[fam_off + count] = '0';
    digit3 = index / 359;
    digit2 = index - digit3 * 359 + ( (digit3 < 1) ? 1 : 0 );
    digit3 += (digit3 > 9) ? 'A' - 10 : '0';
    if (fam_num > 2) filename[fam_off + fam_num - 3] = digit3;
    if (digit2 > 99)
    {
	digit1 = (digit2 - 100) % 10;
	digit2 = (digit2 - 100) / 10;
	filename[fam_off + fam_num - 2] = digit2 + 'A';
	filename[fam_off + fam_num - 1] = digit1 + '0';
    }
    else
    {
	digit1 = digit2 % 10;
	digit2 = digit2 / 10;
	filename[fam_off + fam_num - 2] = digit2 + '0';
	filename[fam_off + fam_num - 1] = digit1 + '0';
    }
}   /*  End Function write_sequence_chars  */

static void get_value (void *dest, CONST void *source, uaddr size, flag swap)
{
    if (swap) m_copy_and_swap_blocks (dest, source, size, size, size, 1);
    else m_copy (dest, source, size);
}   /*  End Function get_value  */

static uaddr read_blocks (Channel channel, char *buffer, uaddr num, uaddr size,
			  flag swap_bytes)
/*  [SUMMARY] Read blocks of data from a channel.
    <channel> The channel object.
    <buffer> The buffer to write the data into. If NULL, data are skipped.
    <num> The number of data blocks to read.
    <size> The size of the data blocks.
    <swap_bytes> If TRUE, data is byte-swapped after reading.
    [RETURNS] The number of bytes read/skipped.
*/
{
    if (buffer == NULL) return ch_skip (channel, num *size) ? (num * size) : 0;
    if (swap_bytes) return ch_read_and_swap_blocks (channel, buffer, num,size);
    else return ch_read (channel, buffer, num * size);
}   /*  End Function read_blocks  */
