/*LINTLIBRARY*/
/*  filter.c

    This code provides filter support functions for the <foreign> package.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various filter routines for the <foreign> package.


    Written by      Richard Gooch   29-JUL-1997

    Updated by      Richard Gooch   2-AUG-1997: Made use of new function
  <r_fork_and_execvp>.

    Updated by      Richard Gooch   23-SEP-1997: Made <foreign_filter_read>
  report bad child termination status.

    Updated by      Richard Gooch   10-OCT-1997: Fixed bug in <read_config>
  where last_filter->next was not updated: only first filter would ever be used

    Updated by      Richard Gooch   17-OCT-1997: Changed config file format to
  support tester programmes.

    Updated by      Richard Gooch   6-NOV-1997: Fixed bug in <get_filter>:
  changed to use lstat(2) rather than stat(2).

    Updated by      Richard Gooch   17-NOV-1997: Created
  <foreign_filter_get_channel>.

    Updated by      Richard Gooch   26-NOV-1997: Added support for config files
  in $KARMABASE/site/share

    Updated by      Richard Gooch   1-DEC-1997: Added support for full pathname
  expressions for filter and testers.

    Updated by      Richard Gooch   3-DEC-1997: Made use of new distinction
  between executable not found and other system errors.

    Updated by      Richard Gooch   5-DEC-1997: Improved parsing of config
  files so that format name could be recorded. Now discard duplicate entries
  based on format name.

    Updated by      Richard Gooch   4-FEB-1998: Improved child wait algorithm.

    Updated by      Richard Gooch   5-MAY-1998: Added support for PNM filters.

    Updated by      Richard Gooch   12-MAY-1998: Fixed bugs in functions
  <foreign_filter_get_channel> and <foreign_filter_read>: not initialising.

    Updated by      Richard Gooch   14-MAY-1998: Switched to using function
  <ch_create_socketpair>. This is better for passing file descriptors.

    Updated by      Richard Gooch   21-JUN-1998: Switched from CAN_ FORK to
  HAS_SIGNALS.

    Updated by      Richard Gooch   23-JUN-1998: Switched from setitimer(2) to
  alarm(2).

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> and <r> packages.

    Updated by      Richard Gooch   10-MAR-1999: Cope with NULL return from
  <r_get_karmabase>.

    Updated by      Richard Gooch   27-MAY-1999: Created experimental function
  <foreign_filter_register>.

    Updated by      Richard Gooch   8-JUN-1999: Took account of change to
  <st_expr_expand>.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Last updated by Richard Gooch   23-JUL-2001: Removed debugging fprintf(3).


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_dsrw.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_ex.h>
#include <karma_r.h>
#include <karma_m.h>
#include <karma_a.h>
#include <os.h>
#ifdef HAS_SIGNALS
#  include <signal.h>
#  include <sys/wait.h>
#endif


#define FILTER_MAGIC_NUMBER 34984732

#define VERIFY_FILTER_INFO(st) if (st == NULL) \
{fprintf (stderr, "NULL info passed\n"); \
 a_prog_bug (function_name); } \
if (st->magic_number != FILTER_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid info object\n"); \
 a_prog_bug (function_name); }


/*  Private structures  */
struct filter_type
{
    flag (*tester_func) (CONST char *dirname, flag test_dir);
    pid_t (*run) (CONST char *filename, int stdout_fd, int close_fd);
    char extension[STRING_LENGTH];
    char filter_prog[STRING_LENGTH];
    char tester_prog[STRING_LENGTH];
    char name[STRING_LENGTH];
    unsigned int output_type;
    struct filter_type *next;
};

struct filter_info_type
{
    unsigned int magic_number;
    pid_t child;
    char filter_prog[STRING_LENGTH];
};

/*  Private data  */
static struct filter_type *first_filter = NULL;


/*  Private functions  */
STATIC_FUNCTION (void _foreign_filter_initialise, () );
STATIC_FUNCTION (void _foreign_filter_read_config, (CONST char *filename) );
STATIC_FUNCTION (struct filter_type *_foreign_filter_get_filter,
		 (CONST char *filename) );
STATIC_FUNCTION (flag _foreign_filter_try_tester,
		 (CONST char *filename, CONST struct filter_type *filter) );
STATIC_FUNCTION (unsigned int _foreign_filter_size_func,
		 (Channel channel, void **info) );
STATIC_FUNCTION (unsigned int _foreign_filter_read_func,
		 (Channel channel, char *buffer, unsigned int length,
		  void **info) );
STATIC_FUNCTION (unsigned int _foreign_filter_write_func,
		 (Channel channel, char *buffer, unsigned int length,
		  void **info) );
STATIC_FUNCTION (flag _foreign_filter_flush_func,
		 (Channel channel, void **info) );
STATIC_FUNCTION (void _foreign_filter_close_func,
		 (void *info, Channel channel) );
#if defined(HAS_SIGNALS) && defined(HAS_ALARM)
STATIC_FUNCTION (void sig_handler, () );
#endif


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int foreign_filter_get_format (CONST char *filename)
/*  [SUMMARY] Get the filter format
    <filename> The name of the file.
    [RETURNS] A value indicating the format of the file. The value
    FOREIGN_FILE_FORMAT_FILTER_KARMA is returned if a data filter to Karma for
    the file exists. The value FOREIGN_FILE_FORMAT_FILTER_FITS is returned if a
    data filter to FITS exists. Otherwise FOREIGN_FILE_FORMAT_UNKNOWN is
    returned.
*/
{
    struct filter_type *filter;

    _foreign_filter_initialise ();
    if (first_filter == NULL) return (FOREIGN_FILE_FORMAT_UNKNOWN);
    if ( ( filter = _foreign_filter_get_filter (filename) ) == NULL )
	return (FOREIGN_FILE_FORMAT_UNKNOWN);
    return (filter->output_type);
}   /*  End Function foreign_filter_get_format  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_filter_test_directory_dataset (CONST char *dirname)
/*  [SUMMARY] Test if a directory is a dataset.
    <dirname> The name of the directory.
    [RETURNS] TRUE if the directory is a dataset, else FALSE.
*/
{
    struct filter_type *filter;

    _foreign_filter_initialise ();
    if ( (strcmp (dirname, ".") == 0) || (strcmp (dirname, "..") == 0) )
	return (FALSE);
    if (first_filter == NULL) return (FALSE);
    for (filter = first_filter; filter != NULL; filter = filter->next)
    {
	if (strcmp (filter->extension, "DIRECTORY") != 0) continue;
	/*  Found a directory dataset  */
	if ( _foreign_filter_try_tester (dirname, filter) ) return (TRUE);
    }
    return (FALSE);
}   /*  End Function foreign_filter_test_directory_dataset  */

/*EXPERIMENTAL_FUNCTION*/
Channel foreign_filter_get_channel (CONST char *filename)
/*  [SUMMARY] Get the channel for a filtered file.
    <filename> The filename to read.
    [RETURNS] A channel object on success, else NULL.
*/
{
    Channel ch1, ch2;
    int fd;
    pid_t child;
    struct filter_type *filter;
    struct filter_info_type *info;
    static char function_name[] = "foreign_filter_get_channel";

    _foreign_filter_initialise ();
    if ( ( filter = _foreign_filter_get_filter (filename) ) == NULL )
	return (NULL);
    if ( !ch_create_socketpair (&ch1, &ch2) )
    {
	fprintf (stderr, "%s: error creating socketpair\t%s\n",
		 function_name, ERRSTRING);
	return (NULL);
    }
    if ( ( info = m_alloc (sizeof *info) ) == NULL )
    {
	m_error_notify (function_name, "channel converter info");
	ch_close (ch1);
	ch_close (ch2);
	return (NULL);
    }
    info->child = -1;
    if (ch_register_converter (ch1, _foreign_filter_size_func,
			       _foreign_filter_read_func,
			       _foreign_filter_write_func,
			       _foreign_filter_flush_func,
			       _foreign_filter_close_func,
			       info) == NULL)
    {
	ch_close (ch1);
	ch_close (ch2);
	m_free (info);
	return (NULL);
    }
    fd = ch_get_descriptor (ch1);
    if (filter->run == NULL)
    {
	char *argv[3];

	argv[0] = filter->filter_prog;
	argv[1] = (char *) filename;
	argv[2] = NULL;
	child = r_fork_and_execvp (0, ch_get_descriptor (ch2), 2,
				   1, &fd, argv[0], argv);
    }
    else child = (*filter->run) (filename, ch_get_descriptor (ch2), fd);
    if (child == -1)
    {
	fprintf (stderr, "%s: error forking\t%s\n", function_name, ERRSTRING);
	ch_close (ch1);
	ch_close (ch2);
	return (NULL);
    }
    info->magic_number = FILTER_MAGIC_NUMBER;
    info->child = child;
    strcpy (info->filter_prog, filter->filter_prog);
    ch_close (ch2);
    return (ch1);
}   /*  End Function foreign_filter_get_channel  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_filter_read (CONST char *filename)
/*  [SUMMARY] Read a filtered file.
    <filename> The filename to read.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    Channel channel;
    multi_array *multi_desc = NULL;
    struct filter_type *filter;
    static char function_name[] = "foreign_filter_read";

    _foreign_filter_initialise ();
    if ( ( filter = _foreign_filter_get_filter (filename) ) == NULL )
	return (NULL);
    if ( ( channel = foreign_filter_get_channel (filename) ) == NULL )
	return (NULL);
    switch (filter->output_type)
    {
      case FOREIGN_FILE_FORMAT_FILTER_KARMA:
	multi_desc = dsrw_read_multi (channel);
	break;
      case FOREIGN_FILE_FORMAT_FILTER_FITS:
	if ( ( multi_desc =
	       foreign_fits_read_header (channel, TRUE, FALSE, TRUE,
					 FA_FITS_READ_HEADER_END) ) != NULL )
	{
	    if ( !foreign_fits_read_data (channel, multi_desc, NULL, 0,
					  FA_FITS_READ_DATA_END) )
	    {
		ds_dealloc_multi (multi_desc);
		multi_desc = NULL;
	    }
	}
	break;
      case FOREIGN_FILE_FORMAT_FILTER_PNM:
	multi_desc = foreign_pnm_read (channel);
	break;
      default:
	fprintf (stderr, "Illegal filter type: %u\n", filter->output_type);
	a_prog_bug (function_name);
	break;
    }
    ch_close (channel);
    if (multi_desc == NULL)
    {
	fprintf (stderr, "%s: error reading filtered file: \"%s\"\n",
		 function_name, filename);
    }
    return (multi_desc);
}   /*  End Function foreign_filter_read  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_filter_register (CONST char *extension,CONST char *converter_prog,
			      pid_t (*run) (CONST char *filename,
					    int stdout_fd, int close_fd),
			      CONST char *tester_prog,
			      flag (*tester_func) (CONST char *dirname,
						   flag test_dir),
			      unsigned int output_type, CONST char *name)
/*  [SUMMARY] Register a filter.
    <extension> The filename extension. If NULL, the file format is a directory
    dataset.
    <converter_prog> The programme used to convert data.
    <run> The function used to run a new programme.
    <tester_prog> The programme used to test the data.
    <tester_func> The function used to test the data.
    <output_type> The output type of the converter.
    <name> The name of the data format.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag duplicate;
    struct filter_type filt;
    struct filter_type *new, *curr, *last_filter;
    static char function_name[] = "foreign_filter_register";

    if (extension == NULL) extension = "DIRECTORY";
    if (tester_prog == NULL) tester_prog = "-";
    if ( (strcmp (extension, "DIRECTORY") == 0) &&
	 (strcmp (tester_prog, "-") == 0) && (tester_func == NULL) )
    {
	fprintf (stderr, "%s: no tester for directory type\n",
		 function_name);
	return (FALSE);
    }
    strcpy (filt.extension, extension);
    strcpy (filt.filter_prog, converter_prog);
    filt.run = run;
    filt.tester_func = tester_func;
    strcpy (filt.tester_prog, tester_prog);
    filt.output_type = output_type;
    if (name == NULL) filt.name[0] = '\0';
    else strcpy (filt.name, name);
    if ( !st_expr_expand (filt.filter_prog, STRING_LENGTH, filt.filter_prog,
			  NULL, NULL, NULL) )
	return (FALSE);
    if ( !st_expr_expand (filt.tester_prog, STRING_LENGTH, filt.tester_prog,
			  NULL, NULL, NULL) ) return (FALSE);
    /*  Search through the existing list of filters to find any duplicates
	and also to find the last filter  */
    for (curr = first_filter, last_filter = NULL, duplicate = FALSE;
	 curr != NULL; curr = curr->next)
    {
	last_filter = curr;
	if (curr->name[0] == '\0') continue;
	if (filt.name[0] == '\0') continue;
	if (strcmp (filt.name, curr->name) == 0)
	{
	    duplicate = TRUE;
	    break;
	}
    }
    if (duplicate) return (FALSE);
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "filter entry");
	return (FALSE);
    }
    *new = filt;
    new->next = NULL;
    if (last_filter != NULL) last_filter->next = new;
    if (first_filter == NULL) first_filter = new;
    return (TRUE);
}   /*  End Function foreign_filter_register  */


/*  Private functions follow  */

static void _foreign_filter_initialise ()
/*  [SUMMARY] Initialise the filter list.
    [RETURNS] Nothing.
*/
{
    CONST char *home, *karmabase;
    char config_file[STRING_LENGTH];
    static flag first_time = TRUE;

    if (!first_time) return;
    first_time = FALSE;
    if ( ( home = r_getenv ("HOME") ) != NULL )
    {
	sprintf (config_file, "%s/.karma/data-filters", home);
	_foreign_filter_read_config (config_file);
    }
    if ( ( karmabase = r_get_karmabase () ) != NULL )
    {
	sprintf (config_file, "%s/site/share/data-filters", karmabase);
	_foreign_filter_read_config (config_file);
	sprintf (config_file, "%s/share/data-filters", karmabase);
	_foreign_filter_read_config (config_file);
    }
    /*  Special built-in filter rules  */
    foreign_filter_register (NULL, "image2fits",
			     foreign_aips2_run,
			     NULL, foreign_aips2_test,
			     FOREIGN_FILE_FORMAT_FILTER_FITS, "AIPS++ Image");
}   /*  End Function _foreign_filter_initialise  */

static void _foreign_filter_read_config (CONST char *filename)
/*  [SUMMARY] Read config from a file.
    <filename> The file to read from.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    char line[STRING_LENGTH];
    static char function_name[] = "_foreign_filter_read_config";

    if (access (filename, R_OK) != 0) return;
    if ( ( channel = ch_open_file (filename, "r") ) == NULL )
    {
	fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		 function_name, filename, ERRSTRING);
	return;
    }
    while ( chs_get_line (channel, line, STRING_LENGTH) )
    {
	unsigned int output_type;
	char *ptr;
	CONST char *p;
	char extension[STRING_LENGTH], filter[STRING_LENGTH];
	char tester[STRING_LENGTH], output[STRING_LENGTH];

	if ( ( ptr = ex_word (line, &p) ) == NULL )
	{
	    fprintf (stderr, "%s: extension field not found in line: \"%s\"\n",
		     function_name, line);
	    ch_close (channel);
	    return;
	}
	strcpy (extension, ptr);
	m_free (ptr);
	if ( ( ptr = ex_word (p, &p) ) == NULL )
	{
	    fprintf (stderr, "%s: filter field not found in line: \"%s\"\n",
		     function_name, line);
	    ch_close (channel);
	    return;
	}
	strcpy (filter, ptr);
	m_free (ptr);
	if ( ( ptr = ex_word (p, &p) ) == NULL )
	{
	    fprintf (stderr, "%s: tester field not found in line: \"%s\"\n",
		     function_name, line);
	    ch_close (channel);
	    return;
	}
	strcpy (tester, ptr);
	m_free (ptr);
	if ( ( ptr = ex_word (p, &p) ) == NULL )
	{
	    fprintf (stderr, "%s: output field not found in line: \"%s\"\n",
		     function_name, line);
	    ch_close (channel);
	    return;
	}
	strcpy (output, ptr);
	m_free (ptr);
	if (st_icmp (output, "karma") == 0)
	    output_type = FOREIGN_FILE_FORMAT_FILTER_KARMA;
	else if (st_icmp (output, "fits") == 0)
	    output_type = FOREIGN_FILE_FORMAT_FILTER_FITS;
	else if (st_icmp (output, "pnm") == 0)
	    output_type = FOREIGN_FILE_FORMAT_FILTER_PNM;
	else
	{
	    fprintf (stderr, "%s: illegal output type: \"%s\"\n",
		     function_name, output);
	    ch_close (channel);
	    return;
	}
	foreign_filter_register (extension, filter, ( pid_t (*) () ) NULL,
				 tester, ( flag (*) () ) NULL, output_type, p);
    }
    ch_close (channel);
}   /*  End Function _foreign_filter_read_config  */

static struct filter_type *_foreign_filter_get_filter (CONST char *filename)
/*  [SUMMARY] Get the filter for a file.
    <filename> The filename.
    [RETURNS] A pointer to the filter on success, else NULL.
*/
{
    struct stat statbuf;
    char *ext;
    struct filter_type *filter;
    static char function_name[] = "_foreign_filter_get_filter";

    if ( ( ext = strrchr (filename, '.') ) != NULL )
    {
	++ext;
	for (filter = first_filter; filter != NULL; filter = filter->next)
	{
	    if (strcmp (ext, filter->extension) != 0) continue;
	    /*  Found it!  */
	    if ( !_foreign_filter_try_tester (filename, filter) ) continue;
	    return (filter);
	}
    }
    /*  No extension match: keep trying if a directory  */
    if (lstat (filename, &statbuf) == -1)
    {
	fprintf (stderr, "%s: error statting file: \"%s\"\t%s\n",
		 function_name, filename, ERRSTRING);
	return (NULL);
    }
    if ( !S_ISDIR (statbuf.st_mode) ) return (NULL);
    for (filter = first_filter; filter != NULL; filter = filter->next)
    {
	if (strcmp (filter->extension, "DIRECTORY") != 0) continue;
	/*  Found a directory dataset  */
	if ( _foreign_filter_try_tester (filename, filter) ) return (filter);
    }
    return (NULL);
}   /*  End Function _foreign_filter_get_filter  */

static flag _foreign_filter_try_tester (CONST char *filename,
					CONST struct filter_type *filter)
/*  [SUMMARY] Try the tester for a format.
    <filename> The filename.
    <filter> The filter.
    [RETURNS] TRUE if the tester succeeds, else FALSE.
*/
{
    int status, rval;
    pid_t child;
    char *argv[3];
    static char function_name[] = "_foreign_filter_try_tester";

    if (filter->tester_func != NULL)
	return (*filter->tester_func) (filename, FALSE);
    if (strcmp (filter->tester_prog, "-") == 0) return (TRUE);
    argv[0] = (char *) filter->tester_prog;
    argv[1] = (char *) filename;
    argv[2] = NULL;
    if ( ( child = r_fork_and_execvp (0, 1, 2, 0, NULL, argv[0], argv) )== -1 )
    {
	fprintf (stderr, "%s: error forking\t%s\n", function_name, ERRSTRING);
	return (FALSE);
    }
    waitpid (child, &status, 0);
    if ( !WIFEXITED (status) )
    {
	fprintf (stderr, "%s: filter programme: %s did not exit cleanly\n",
		 function_name, filter->tester_prog);
	return (FALSE);
    }
    rval = WEXITSTATUS (status);
    if (rval == RV_OK) return (TRUE);
    if (rval == RV_UNKNOWN_FORMAT) return (FALSE);
    if (rval == RV_EXECUTABLE_NOT_FOUND)
    {
	fprintf (stderr, "%s: could not find executable programme: \"%s\"\n",
		 function_name, filter->tester_prog);
	return (FALSE);
    }
    if (rval == RV_SYS_ERROR)
    {
	fprintf (stderr, "%s: system error running filter programme: %s\n",
		 function_name, filter->tester_prog);
	return (FALSE);
    }
    fprintf (stderr, "%s: tester programme: %s returned failure status: %d\n",
	     function_name, filter->tester_prog, rval);
    return (FALSE);
}   /*  End Function _foreign_filter_try_tester  */

static unsigned int _foreign_filter_size_func (Channel channel, void **info)
/*  This routine will determine the approximate number of bytes that the
    read_func  will return.
    The channel object will be given by  channel  .
    The arbitrary information pointer will be pointed to by  info  .The
    information pointer may be modified.
    The routine returns the number of bytes the  read_func  will return.
*/
{
    struct filter_info_type *filter_info;
    static char function_name[] = "_foreign_filter_size_func";

    filter_info = (struct filter_info_type *) *info;
    VERIFY_FILTER_INFO (filter_info);
    return ch_get_bytes_readable (channel);
}   /*  End Function _foreign_filter_size_func  */

static unsigned int _foreign_filter_read_func (Channel channel, char *buffer,
					       unsigned int length,
					       void **info)
/*   This routine will convert bytes being read from a channel object.
     The channel object will be given by  channel  .It is permissable for
     the routine to call  ch_read  with this channel. If this is done,
     this  read_func  will be bypassed.
     The buffer to write the data into will be pointed to by  buffer  .
     The number of bytes to write into the buffer will be pointed to by
     length  .
     The arbitrary information pointer will be pointed to by  info  .The
     information pointer may be modified.
     The routine returns the number of bytes actually written to the buffer.
*/
{
    struct filter_info_type *filter_info;
    static char function_name[] = "_foreign_filter_read_func";

    filter_info = (struct filter_info_type *) *info;
    VERIFY_FILTER_INFO (filter_info);
    return ch_read (channel, buffer, length);
}   /*  End Function _foreign_filter_read_func  */

static unsigned int _foreign_filter_write_func (Channel channel, char *buffer,
						unsigned int length,
						void **info)
/*  This routine will convert bytes being written to a channel object.
    The channel object will be given by  channel  .It is permissable for
    the routine to call  ch_write  with this channel. If this is done,
    this  write_func  will be bypassed.
    The buffer to read the data from will be pointed to by  buffer  .The
    contents of this buffer may be modified if needed.
    The number of bytes to read from the buffer will be given by  length
    The arbitrary information pointer will be pointed to by  info  .The
    information pointer may be modified.
    The routine returns the number of bytes read from the buffer.
*/
{
    struct filter_info_type *filter_info;
    static char function_name[] = "_foreign_filter_write_func";

    filter_info = (struct filter_info_type *) *info;
    VERIFY_FILTER_INFO (filter_info);
    fprintf (stderr, "Illegal write to filtered channel!\n");
    a_prog_bug (function_name);
    return (0);
}   /*  End Function _foreign_filter_write_func  */

static flag _foreign_filter_flush_func (Channel channel, void **info)
/*  This routine will process a flush request for a channel object.
    The channel object will be given by  channel  .It is permissable for
    the routine to call  ch_write  with this channel. If this is done,
    this  write_func  will be bypassed.
    The arbitrary information pointer will be pointed to by  info  .The
    information pointer may be modified.
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    return (TRUE);
}   /*  End Function _foreign_filter_flush_func  */

static void _foreign_filter_close_func (void *info, Channel channel)
/*  [SUMMARY] Close.
    [PURPOSE] This routine is called when a channel is closed.
    <info> The arbitrary information pointer.
    <channel> The channel object.
    [RETURNS] Nothing.
*/
{
#if defined(HAS_SIGNALS) && defined(HAS_ALARM)
    flag ok;
    int status;
    struct sigaction sigchld_old, sigalrm_old, new_action;
    struct filter_info_type *filter_info;
    static char function_name[] = "_foreign_filter_close_func";

    filter_info = (struct filter_info_type *) info;
    VERIFY_FILTER_INFO (filter_info);
    if (filter_info->child < 0)
    {
	m_clear (filter_info, sizeof *filter_info);
	m_free (filter_info);
	return;
    }
    /*  Need to clean up child  */
    if (ch_get_bytes_readable (channel) > 0)
    {
	/*  Data unread: kill quickly  */
	kill (filter_info->child, SIGKILL);
	waitpid (filter_info->child, &status, 0);
	m_clear (filter_info, sizeof *filter_info);
	m_free (filter_info);
	return;
    }
    switch ( waitpid (filter_info->child, &status, WNOHANG) )
    {
      case -1:
	/*  Error  */
	fprintf (stderr, "%s: error reaping child: %d\t%s\n",
		 function_name, (int) filter_info->child, ERRSTRING);
	m_clear (filter_info, sizeof *filter_info);
	m_free (filter_info);
	return;
	/*break;*/
      case 0:
	/*  Child not dead yet: wait for up to 1 second for it to die  */
	sigemptyset (&new_action.sa_mask);
	new_action.sa_flags = 0;
	new_action.sa_handler = sig_handler;
	if (sigaction (SIGCHLD, &new_action, &sigchld_old) != 0)
	{
	    fprintf (stderr, "Error setting SIGCHLD handler\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (sigaction (SIGALRM, &new_action, &sigalrm_old) != 0)
	{
	    fprintf (stderr, "Error setting SIGALRM handler\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	alarm (1);
	pause ();
	/*  Restore everything  */
	if (sigaction (SIGCHLD, &sigchld_old, NULL) != 0)
	{
	    fprintf (stderr, "Error restoring SIGCHLD handler\t%s\n",
		     ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (sigaction (SIGALRM, &sigalrm_old, NULL) != 0)
	{
	    fprintf (stderr, "Error restoring SIGALRM handler\t%s\n",
		     ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	alarm (0);
	/*  Check again to see if child has died  */
	if (waitpid (filter_info->child, &status, WNOHANG) == 0)
	{
	    fprintf (stderr, "child not dead yet\n");
	    /*  Still not dead: give up  */
	    kill (filter_info->child, SIGKILL);
	    waitpid (filter_info->child, &status, 0);
	    m_clear (filter_info, sizeof *filter_info);
	    m_free (filter_info);
	    return;
	}
	break;
      default:
	/*  Child is dead  */
	break;
    }
    if ( WIFEXITED (status) )
    {
	ok = (WEXITSTATUS (status) == RV_OK) ? TRUE : FALSE;
	if (!ok)
	{
	    fprintf ( stderr,
		      "%s: filter programme: %s returned failure status: %d\n",
		      function_name, filter_info->filter_prog,
		      WEXITSTATUS (status) );
	}
    }
    else if ( WIFSIGNALED (status) )
    {
	fprintf ( stderr,
		  "%s: filter programme: %s killed by signal: %d\n",
		  function_name, filter_info->filter_prog, WTERMSIG (status) );
    }
    else
    {
	ok = FALSE;
	fprintf (stderr, "%s: filter programme: %s did not exit cleanly\n",
		 function_name, filter_info->filter_prog);
    }
    m_clear (filter_info, sizeof *filter_info);
    m_free (filter_info);
#endif
}   /*  End Function _foreign_filter_close_func  */

#if defined(HAS_SIGNALS) && defined(HAS_ALARM)
static void sig_handler ()
/*  [SUMMARY] Generic signal handler.
    [RETURNS] Nothing.
*/
{
}   /*  End Function sig_handler  */
#endif
