/*LINTLIBRARY*/
/*  fits_read.c

    This code provides a FITS read facility.

    Copyright (C) 1995-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading data files in
  FITS format.


    Written by      Richard Gooch   17-APR-1995: Copied skeleton from
  ppm_read.c  and guts from  modules/fits/fits2karma.c

    Updated by      Richard Gooch   20-APR-1995

    Updated by      Richard Gooch   5-MAY-1995: Placate SGI compiler.

    Updated by      Richard Gooch   17-MAY-1995: Fixed bug when BUNIT is an
  empty string.

    Updated by      Richard Gooch   1-JUN-1995: Tolerate header lines with
  tab in first column: print warning and skip them.

    Updated by      Richard Gooch   4-JUN-1995: Allow for floating point values
  with no 'E', only '.' character.

    Updated by      Richard Gooch   18-JUL-1995: Translate '\0' characters to
  ' ' characters.

    Updated by      Richard Gooch   31-JUL-1995: Made use of
  <pio_read_float_nantrap> and <pio_read_double_nantrap> and junked massive
  #if blocks testing for NaN.

    Updated by      Richard Gooch   29-FEB-1996: Cope with multiple instances
  of keywords if at all possible.

    Updated by      Richard Gooch   12-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   3-JUN-1996: No longer fiddle reversed axes.

    Updated by      Richard Gooch   15-JUN-1996: Add *__SCALE and *__OFFSET
  keywords if necessary.

    Updated by      Richard Gooch   17-JUN-1996: Subtracted 1.0 from CRPIXn
  values.

    Updated by      Richard Gooch   24-JUN-1996: Trapped "NAXISn" and other
  "?????n" keywords coming before "NAXIS" keyword.

    Updated by      Richard Gooch   6-AUG-1996: Copied "CRVALn" value to
  reference field of dimension descriptor and now add header keywords in the
  same order as in the FITS file. Now also copy history.

    Updated by      Richard Gooch   15-AUG-1996: No longer update <<reference>>
  field of dimension descriptor.

    Updated by      Richard Gooch   29-AUG-1996: Allowed some axes to not have
  names but still use as many supplied names as possible. Fail if an axis does
  not have length specified. Issue warnings if CRVAL, CRPIX or CDELT not found.

    Updated by      Richard Gooch   5-SEP-1996: Made use of <pio_read_floats>.

    Updated by      Richard Gooch   7-SEP-1996: Made use of <pio_read_doubles>.

    Updated by      Richard Gooch   25-SEP-1996: Changed to implement the new
  FITS-style co-ordinate handling, where dimension co-ordinates range from 0
  to length - 1.

    Updated by      Richard Gooch   15-OCT-1996: No longer throw away last
  column in each line of the header.

    Updated by      Richard Gooch   30-DEC-1996: Fixed memory leak: card info
  was not being freed.

    Updated by      Richard Gooch   31-DEC-1996: Fixed bug in previous bugfix:
  when "BUNIT" not supplied a non-allocated pointer was freed.

    Updated by      Richard Gooch   22-FEB-1997: Modify duplicate axis names
  to contain a ":n" where n is the axis number.

    Updated by      Richard Gooch   25-FEB-1997: Fixed bug introduced in
  previous modification: NULL axis names were not trapped.

    Updated by      Richard Gooch   18-MAR-1997: Removed interpretation of
  "CDELTn" keywords: simply passed along like anything else.

    Updated by      Richard Gooch   19-MAR-1997: Improved debugging when
  extracting "BUNIT" value.

    Updated by      Richard Gooch   26-MAR-1997: Fixed bug when BSCALE != 0.0
  and BZERO = 0.0 and when reading integer data without conversion to floats.

    Updated by      Richard Gooch   22-APR-1997: Optimised reading of integer
  data.

    Updated by      Richard Gooch   23-APR-1997: Added skip attribute.

    Updated by      Richard Gooch   5-AUG-1997: No longer bother with checking
  "CRVALn" and "CRPIXn" keywords.

    Updated by      Richard Gooch   8-AUG-1997: Fixed bug in function
  <foreign_fits_read_data> when blank was -2147483648 (0x80000000).

    Updated by      Richard Gooch   25-SEP-1997: No longer clear array when
  allocating.

    Updated by      Richard Gooch   20-FEB-1998: Ignore keywords which re-use
  an existing name.

    Updated by      Richard Gooch   23-FEB-1998: Trap "CUNITn" keywords.

    Updated by      Richard Gooch   24-FEB-1998: Switch to new array_pointer
  which has free function.

    Updated by      Richard Gooch   11-APR-1998: Fixed bug when have "CUNITn"
  with "NAXIS=m" and n > m.

    Updated by      Richard Gooch   13-MAY-1998: Added fallback to using anon.
  mmapped array.

    Updated by      Richard Gooch   16-MAY-1998: Removed *__SCALE and *__OFFSET
  keywords. Created experimental function <foreign_fits_read>.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   19-MAY-1998: Fixed bug in function
  <foreign_fits_read> when not memory mapping: array was not allocated.

    Updated by      Richard Gooch   26-MAY-1998: Trap mmapped files too small
  in <foreign_fits_read>.

    Updated by      Richard Gooch   27-MAY-1998: Added attribute
  FA_FITS_READ_HEADER_ALLOW_TRUNC.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   14-OCT-1998: Discard scaling keywords if
  data is scaled.

    Updated by      Richard Gooch   6-NOV-1998: Created unpublished function
  <foreign_fits_copy_special>.

    Updated by      Richard Gooch   26-NOV-1998: Trap truncated files in
  <foreign_fits_read>.

    Updated by      Richard Gooch   29-JUN-1999: Added support for simple FITS.

    Updated by      Richard Gooch   7-AUG-1999: Fixed bugs in function
  <foreign_fits_read_data>: assumed automatic char arrays are word aligned.

    Updated by      Richard Gooch   13-AUG-1999: Fixed bug in function
  <extract_info_from_header>: keywords starting with "END" caused premature end

    Updated by      Richard Gooch   18-AUG-1999: Made use of new experimental
  function <ch_read2>.

    Updated by      Richard Gooch   8-DEC-1999: Fixed bug on little-endian
  machines when reading from mmapped file: caused by use of new <ch_read2>.

    Updated by      Richard Gooch   14-MAY-2000: Drop empty strings in function
  <process_keywords>.  

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Updated by      Richard Gooch   20-JUN-2000: Supported 'e' when checking
  for floating-point values.

    Updated by      Richard Gooch   16-MAR-2001: Sanitise "EQUINOX" keywords
  with string values.

    Updated by      Richard Gooch   4-MAY-2001: Created <isalldigit> function
  and use it in <extract_info_from_header> for CTYPE processing.

    Updated by      Richard Gooch   22-JUN-2001: Fixed bug in function
  <foreign_fits_read_header>: sanitising NAXIS1 wrote past array bounds.

    Updated by      Richard Gooch   15-OCT-2001: Trapped for zero-length array
  in <foreign_fits_read_header>.

    Updated by      Richard Gooch   1-NOV-2001: Added calls to
  <ds_check_array_can_fit> and removed previous change.

    Last updated by Richard Gooch   13-NOV-2001: Fixed bug in function
  <foreign_fits_read_header> when data allocation failed: debugging message
  used unallocated structure.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_pio.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_p.h>


#define CARD_WIDTH 80
#define CARD_LENGTH 36
#define EQUALS_POSITION 8
#define BUF_SIZE 65536

struct keyword_type
{
    char *keyword;
    char *value;
    struct keyword_type *next;
};

struct data_info_type
{
    double bzero;
    double bscale;
    int bitpix;
};

struct card_info_type
{
    struct data_info_type data_info;
    int naxis;
    int *dim_lengths;
    char **dim_names;     /*  Trailing whitespace is removed      */
    char *elem_name;
    double *axis_scales;  /*  These are stored in FITS-order      */
    flag simple;
    flag end_found;
    flag hide_scaling;
    struct keyword_type *first_keyword;
    struct keyword_type *last_keyword;
    multi_array *history_holder;
};


/*  Declarations of private functions follow  */
STATIC_FUNCTION (flag extract_info_from_header,
		 (char card[CARD_LENGTH][CARD_WIDTH],
		  struct card_info_type *info,
		  flag strip_trailing_whitespace) );
STATIC_FUNCTION (void redeclaration_message, (char *string) );
STATIC_FUNCTION (void conversion_error_message,
		 (char *parameter, char *value) );
STATIC_FUNCTION (void cleanup_string, (char *string) );
STATIC_FUNCTION (void free_keywords, (struct keyword_type *keywords) );
STATIC_FUNCTION (flag process_keywords,
		 (struct keyword_type *keywords, multi_array *multi_desc,
		  struct card_info_type finfo) );
STATIC_FUNCTION (void free_card_info, (struct card_info_type *info) );
STATIC_FUNCTION (flag isalldigit, (CONST char *string) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
multi_array *foreign_fits_read_header (Channel channel, flag data_alloc,
				       flag convert_int_to_float,
				       flag sanitise, ...)
/*  [SUMMARY] Read a FITS header.
    [PURPOSE] This routine will read the header of a FITS file from a channel.
    The data section is NOT read.
    <channel> The channel to read from.
    <data_alloc> If TRUE, the data space is allocated.
    <convert_int_to_float> If TRUE, integer FITS data is converted to floating
    point data.
    <sanitise> If TRUE, FITS axes with length 1 are ignored. This is highly
    recommended.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_FITS_READ_HEADER_END. See [<foreign_ATT_FITS_READ_HEADER>] for a list of
    defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    va_list argp;
    flag keep_reading = TRUE;
    flag allow_truncated = FALSE;
    flag simple = FALSE;
    flag duplicate;
    unsigned int att_key;
    unsigned int num_dim;
    unsigned int dim_count, count;
    unsigned int elem_type;
    unsigned int bytes_read;
    struct card_info_type finfo;
    array_pointer arrayp;
    struct data_info_type *data_info;
    uaddr *dim_lengths;
    char **dim_names;
    multi_array *multi_desc;
    array_desc *arr_desc;
    char txt[STRING_LENGTH], name[STRING_LENGTH];
    char card[CARD_LENGTH][CARD_WIDTH];
    /*double value[2];*/
    static char function_name[] = "foreign_fits_read_header";

    va_start (argp, sanitise);
    if (channel == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    FLAG_VERIFY (data_alloc);
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp, unsigned int) )
	    != FA_FITS_READ_HEADER_END )
    {
	switch (att_key)
	{
	  case FA_FITS_READ_HEADER_ALLOW_TRUNC:
	    allow_truncated = va_arg (argp, flag);
	    FLAG_VERIFY (allow_truncated);
	    break;
	  case FA_FITS_READ_HEADER_SIMPLE:
	    simple = va_arg (argp, flag);
	    FLAG_VERIFY (simple);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    /*  Initialise card info  */
    finfo.end_found = FALSE;
    finfo.simple = 2 * abs (TRUE) + 2 * abs (FALSE);
    finfo.naxis = -1;
    finfo.data_info.bitpix = 0;
    finfo.data_info.bzero = TOOBIG;
    finfo.data_info.bscale = TOOBIG;
    finfo.dim_lengths = NULL;
    finfo.dim_names = NULL;
    finfo.elem_name = NULL;
    finfo.axis_scales = NULL;
    finfo.hide_scaling = TRUE;
    finfo.first_keyword = NULL;
    finfo.last_keyword = NULL;
    if ( ( finfo.history_holder = ds_alloc_multi (1) ) == NULL )
    {
	m_abort (function_name, "history holder");
    }
    if (simple)
    {
	/*  Read simple formatted header  */
	char space = ' ';
	int len;
	unsigned line_count = 0;

	m_fill (card, 1, &space, 1, CARD_WIDTH * CARD_LENGTH);
	while (keep_reading)
	{
	    if ( ch_getl (channel, card[line_count], CARD_WIDTH) )
	    {
		len = strlen (card[line_count]);
		if (strcmp (card[line_count], "END") == 0)
		{
		    line_count = CARD_LENGTH;
		    card[line_count][len] = ' ';
		    keep_reading = FALSE;
		}
		else
		{
		    card[line_count][len] = ' ';
		    ++line_count;
		}
	    }
	    else
	    {
		line_count = CARD_LENGTH;
		keep_reading = FALSE;
	    }
	    if (line_count >= CARD_LENGTH)
	    {
		if ( !extract_info_from_header (card, &finfo, TRUE) )
		{
		    fprintf(stderr,
			    "Error extracting information from header card\n");
		    free_keywords (finfo.first_keyword);
		    free_card_info (&finfo);
		    return (NULL);
		}
		m_fill (card, 1, &space, 1, CARD_WIDTH * CARD_LENGTH);
		line_count = 0;
	    }
	    if (finfo.end_found) keep_reading = FALSE;
	}
	if (!finfo.end_found)
	{
	    fprintf (stderr, "No END keyword in simple header\n");
	    free_keywords (finfo.first_keyword);
	    free_card_info (&finfo);
	    return (NULL);
	}
	keep_reading = FALSE;
    }
    /*  Read header cards  */
    while (keep_reading)
    {
	bytes_read = ch_read (channel, (char *) card, CARD_WIDTH *CARD_LENGTH);
	if ( (bytes_read < CARD_WIDTH * CARD_LENGTH) && !allow_truncated )
	{
	    fprintf (stderr, "%s: error reading: %u bytes\t%s\n",
		     function_name, CARD_WIDTH * CARD_LENGTH,
		     (errno == 0) ? "" : ERRSTRING);
	    fprintf (stderr, "Error reading another header card\n");
	    free_keywords (finfo.first_keyword);
	    free_card_info (&finfo);
	    return (NULL);
	}
	if ( !extract_info_from_header (card, &finfo, TRUE) )
	{
	    fprintf (stderr,"Error extracting information from header card\n");
	    free_keywords (finfo.first_keyword);
	    free_card_info (&finfo);
	    return (NULL);
	}
	if (finfo.end_found)
	{
	    keep_reading = FALSE;
	}
	else if (bytes_read < CARD_WIDTH * CARD_LENGTH)
	{
	    fprintf(stderr,
		    "%s: no END keyword in truncated card, length: %u bytes\n",
		    function_name, bytes_read);
	    keep_reading = FALSE;
	}
    }
    if (!finfo.simple)
    {
	fprintf (stderr, "Can only read simple FITS files\n");
	free_keywords (finfo.first_keyword);
	free_card_info (&finfo);
	return (NULL);
    }
    if ( DOUBLE_IS_BLANK (finfo.data_info.bscale) )
	finfo.data_info.bscale = 1.0;
    if ( DOUBLE_IS_BLANK (finfo.data_info.bzero) ) finfo.data_info.bzero = 0.0;
    if ( (finfo.elem_name != NULL) && (strlen (finfo.elem_name) < 1) )
    {
	fprintf (stderr, "WARNING: BUNIT is an empty string: defaulting to ");
	fprintf (stderr, "\"Data Value\"\n");
	m_free (finfo.elem_name);
	finfo.elem_name = NULL;
    }
    if (finfo.elem_name == NULL)
    {
	if ( ( finfo.elem_name = st_dup ("Data Value") ) == NULL )
	{
	    m_abort (function_name, "default element name");
	}
    }
    /*  Determine output data type  */
    switch (finfo.data_info.bitpix)
    {
      case 8:
	if (convert_int_to_float) elem_type = K_FLOAT;
	else elem_type = K_UBYTE;
	break;
      case 16:
	if (convert_int_to_float) elem_type = K_FLOAT;
	else elem_type = K_SHORT;
	break;
      case 32:
	elem_type = convert_int_to_float ? K_FLOAT : K_INT;
	break;
      case -32:
	elem_type = K_FLOAT;
	finfo.hide_scaling = FALSE;
	break;
      case -64:
	elem_type = K_DOUBLE;
	break;
      default:
	fprintf (stderr, "Illegal value for BITPIX: %d\n",
		 finfo.data_info.bitpix);
	free_keywords (finfo.first_keyword);
	free_card_info (&finfo);
	return (NULL);
	/*break;*/
    }
    if (elem_type != K_FLOAT) finfo.hide_scaling = FALSE;
    /*  Modify dimension names if there are duplicates  */
    for (dim_count = 0; dim_count < finfo.naxis; ++dim_count)
    {
	/*  A first pass to check if duplicates exist  */
	if (finfo.dim_names[dim_count] == NULL) continue;
	duplicate = FALSE;
	for (count = 0; !duplicate && (count < finfo.naxis); ++count)
	{
	    if (count == dim_count) continue;
	    if (finfo.dim_names[count] == NULL) continue;
	    if (strcmp (finfo.dim_names[count], finfo.dim_names[dim_count])
		== 0) duplicate = TRUE;
	}
	if (!duplicate) continue;
	strcpy (name, finfo.dim_names[dim_count]);
	/*  Now modify all duplicates  */
	for (count = 0; count < finfo.naxis; ++count)
	{
	    if (strcmp (finfo.dim_names[count], name) != 0) continue;
	    sprintf (txt, "%s:%u", finfo.dim_names[count], finfo.naxis -count);
	    fprintf (stderr, "Modifying duplicate: \"%s\" to: \"%s\"\n",
		     finfo.dim_names[count], txt);
	    strcpy (finfo.dim_names[count], txt);
	}
    }
    if (sanitise)
    {
	/*  Strip dimensions with length of 1  */
	for (dim_count = 0, num_dim = 0; dim_count < finfo.naxis; ++dim_count)
	{
	    if (finfo.dim_lengths[dim_count] > 1) ++num_dim;
	}
    }
    else
    {
	/*  All dimensions must be allocated  */
	num_dim = finfo.naxis;
    }
    /*  Create array of dimension lengths  */
    if ( ( dim_lengths = m_alloc (sizeof *dim_lengths * num_dim) ) == NULL )
    {
	m_error_notify (function_name, "dimension lengths");
	free_keywords (finfo.first_keyword);
	free_card_info (&finfo);
	return (NULL);
    }
    /*  Create array of dimension name pointers  */
    if ( ( dim_names = m_alloc (sizeof *dim_names * num_dim) ) == NULL )
    {
	m_error_notify (function_name, "dimension name pointers");
	m_free (dim_lengths);
	free_keywords (finfo.first_keyword);
	free_card_info (&finfo);
	return (NULL);
    }
    for (dim_count = 0, num_dim = 0; dim_count < finfo.naxis; ++dim_count)
    {
	if ( sanitise && (finfo.dim_lengths[dim_count] < 2) ) continue;
	if (finfo.dim_lengths[dim_count] < 1)
	{
	    fprintf (stderr, "NAXIS%u length: %d invalid\n",
		     finfo.naxis - dim_count, finfo.dim_lengths[dim_count]);
	    m_free (dim_lengths);
	    m_free (dim_names);
	    free_keywords (finfo.first_keyword);
	    free_card_info (&finfo);
	    return (NULL);
	}
	dim_lengths[num_dim] = finfo.dim_lengths[dim_count];
	dim_names[num_dim] = finfo.dim_names[dim_count];
	++num_dim;
    }
    /*  Now allocate the Karma data structure descriptors  */
    arr_desc = ds_easy_alloc_array_desc (num_dim, dim_lengths,
					 NULL, NULL, NULL,
					 (CONST char **) dim_names,
					 1, &elem_type,
					 (CONST char **) &finfo.elem_name);
    m_free (dim_lengths);
    /*  No need to deallocate each dimension name, since we only took a pointer
	copy: <free_card_info> will clean up later  */
    m_free (dim_names);
    if (arr_desc == NULL)
    {
	free_keywords (finfo.first_keyword);
	free_card_info (&finfo);
	m_error_notify (function_name, "array_descriptor");
	return (NULL);
    }
    m_clear (&arrayp, sizeof arrayp);
    if (data_alloc)
    {
	if ( !ds_check_array_can_fit (arr_desc) )
	{
	    free_keywords (finfo.first_keyword);
	    free_card_info (&finfo);
	    ds_dealloc_array_desc (arr_desc);
	    return (NULL);
	}
	/*  Allocate the array manaully so that clearing is not done  */
	if ( !ds_alloc_vm_array (&arrayp, arr_desc, FALSE, FALSE) &&
	     !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp, arr_desc, FALSE,
				   FALSE) )
	{
	    fprintf (stderr, "%s: error allocating %lu bytes for array data\n",
		     function_name,
		     ds_get_array_size (arr_desc) *
		     ds_get_packet_size (arr_desc->packet));
	    free_keywords (finfo.first_keyword);
	    free_card_info (&finfo);
	    ds_dealloc_array_desc (arr_desc);
	    return (NULL);
	}
    }
    if ( ( multi_desc =
	   ds_easy_alloc_array_from_array_desc (arr_desc, &arrayp, TRUE) )
	 == NULL )
    {
	free_keywords (finfo.first_keyword);
	free_card_info (&finfo);
	ds_dealloc_array_desc (arr_desc);
	if (arrayp.free != NULL) (*arrayp.free) (arrayp.info, &arrayp);
	m_error_notify (function_name, "multi_array descriptor");
	return (NULL);
    }
    if ( ( data_info = m_alloc (sizeof *data_info) ) == NULL )
	m_abort (function_name, "data info");
    *data_info = finfo.data_info;
    if ( !ds_add_named_object (multi_desc, 1, "FITS_READ", data_info, m_free) )
    {
	m_abort (function_name, "FITS_READ attachment");
    }
    /*  Add FITS keywords to top-level packet  */
    if ( !process_keywords (finfo.first_keyword, multi_desc, finfo) )
    {
	free_keywords (finfo.first_keyword);
	free_card_info (&finfo);
	ds_dealloc_multi (multi_desc);
	return (NULL);
    }
    free_keywords (finfo.first_keyword);
    finfo.first_keyword = NULL;
    /*  Copy over any history  */
    multi_desc->first_hist = finfo.history_holder->first_hist;
    multi_desc->last_hist = finfo.history_holder->last_hist;
    /*  Disconnect from temporary container for history  */
    finfo.history_holder->first_hist = NULL;
    finfo.history_holder->last_hist = NULL;
    free_card_info (&finfo);
    return (multi_desc);
}   /*  End Function foreign_fits_read_header  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_fits_read (CONST char *filename,
				flag convert_int_to_float, flag sanitise,
				...)
/*  [SUMMARY] Read a FITS file.
    <filename> The name of the file.
    <convert_int_to_float> If TRUE, integer FITS data is converted to floating
    point data.
    <sanitise> If TRUE, FITS axes with length 1 are ignored. This is highly
    recommended.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_FITS_READ_END. See [<foreign_ATT_FITS_READ>] for a list of defined
    attributes.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    va_list argp;
    Channel channel;
    flag writable = FALSE;
    flag have_blank = FALSE;
    flag simple = FALSE;
    flag can_mmap;
    signed long blank = 0;
    unsigned int att_key, type;
    unsigned int mmap_option = 0;
    unsigned long read_pos, write_pos;
    array_pointer arrayp;
    struct data_info_type *dinfo;
    unsigned long *blank_count = NULL;
    multi_array *multi_desc;
    array_desc *arr_desc;
    double value[2];
    static char function_name[] = "foreign_fits_read";

    va_start (argp, sanitise);
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp, unsigned int) ) != FA_FITS_READ_END )
    {
	switch (att_key)
	{
	  case FA_FITS_READ_NUM_BLANKS:
	    blank_count = va_arg (argp, unsigned long *);
	    break;
	  case FA_FITS_READ_MMAP_OPTION:
	    mmap_option = va_arg (argp, unsigned int);
	    break;
	  case FA_FITS_READ_WRITABLE:
	    writable = va_arg (argp, flag);
	    FLAG_VERIFY (writable);
	    break;
	  case FA_FITS_READ_SIMPLE:
	    simple = va_arg (argp, flag);
	    FLAG_VERIFY (simple);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    if ( ( channel = ch_map_disc (filename, mmap_option, writable, TRUE) )
	 == NULL )
    {
	fprintf (stderr, "Error opening: \"%s\"\t%s\n", filename, ERRSTRING);
	return (NULL);
    }
    /*  Do a quick test for plain read  */
    can_mmap = simple ? FALSE : ch_test_for_mmap (channel);
    /*  Ensure "SIMPLE" is found. If not, close and reopen file so that
	decompression can work  */
    if (can_mmap)
    {
	if (ch_get_mem_size (channel) < CARD_WIDTH)
	{
	    fprintf (stderr, "%s: file too small\n", function_name);
	    ch_close (channel);
	    return (NULL);
	}
	if (strncmp (ch_get_mmap_addr (channel), "SIMPLE", 6) != 0)
	{
	    can_mmap = FALSE;
	    ch_close (channel);
	    if ( ( channel = ch_open_file (filename, "rz") ) == NULL )
	    {
		fprintf (stderr, "%s: error opening: \"%s\"\t%s\n",
			 function_name, filename, ERRSTRING);
		return (NULL);
	    }
	}
    }
    if (blank_count != NULL) can_mmap = FALSE;
    if (!can_mmap)
    {
	if ( ( multi_desc =
	       foreign_fits_read_header (channel, TRUE, convert_int_to_float,
					 sanitise,
					 FA_FITS_READ_HEADER_SIMPLE, simple,
					 FA_FITS_READ_HEADER_END) ) == NULL )
	{
	    ch_close (channel);
	    return (NULL);
	}
	if ( !foreign_fits_read_data (channel, multi_desc, NULL, 0,
				      FA_FITS_READ_DATA_NUM_BLANKS,
				      blank_count,
				      FA_FITS_READ_DATA_ASCII, simple,
				      FA_FITS_READ_DATA_END) )
	{
	    ds_dealloc_multi (multi_desc);
	    ch_close (channel);
	    return (NULL);
	}
	ch_close (channel);
	return (multi_desc);
    }
    /*  Channel was mmapped and block transfer may be OK  */
    multi_desc = foreign_fits_read_header (channel, FALSE,
					   convert_int_to_float, sanitise,
					   FA_FITS_READ_HEADER_END);
    if (multi_desc == NULL)
    {
	ch_close (channel);
	return (NULL);
    }
    arr_desc = (array_desc *) multi_desc->headers[0]->element_desc[0];
    if ( !ds_check_array_can_fit (arr_desc) )
    {
	ds_dealloc_multi (multi_desc);
	ch_close (channel);
	return (NULL);
    }
    type = arr_desc->packet->element_types[0];
    can_mmap = ds_can_transfer_element_as_block (type);
    if ( ds_get_unique_named_value (multi_desc->headers[0],
				    multi_desc->data[0], "BLANK", &type,
				    value) )
    {
	if ( DOUBLE_IS_BLANK (value[0]) ) blank = 0x80000000;
	else blank = value[0];
	have_blank = TRUE;
    }
    else have_blank = FALSE;
    if ( !ds_get_named_object (multi_desc, 1, "FITS_READ", (void **) &dinfo) )
    {
	fprintf (stderr, "FITS_READ attachment not found\n");
	a_prog_bug (function_name);
    }
    switch (dinfo->bitpix)
    {
      case -32:
      case -64:
	break;
      case 8:
	if (convert_int_to_float) can_mmap = FALSE;
	if ( (have_blank) && (blank != 255) ) can_mmap = FALSE;
	break;
      case 16:
	if (convert_int_to_float) can_mmap = FALSE;
	if ( (have_blank) && (blank != -32768) ) can_mmap = FALSE;
	break;
      case 32:
	if (convert_int_to_float) can_mmap = FALSE;
	if ( (have_blank) && (blank != 0x80000000) ) can_mmap = FALSE;
	break;
    }
    if (writable) can_mmap = FALSE;
    if ( !ch_tell (channel, &read_pos, &write_pos) )
    {
	fprintf (stderr, "%s: error getting file position\n", function_name);
	ds_dealloc_multi (multi_desc);
	ch_close (channel);
	return (NULL);
    }
    if (ch_get_mem_size (channel) <
	ds_get_array_size (arr_desc) * abs (dinfo->bitpix) / 8 + read_pos)
    {
	fprintf (stderr, "%s: file: \"%s\" short by %lu bytes\n",
		 function_name, filename,
		 ds_get_array_size (arr_desc) * abs (dinfo->bitpix) / 8 -
		 ch_get_mem_size (channel) + read_pos);
	ds_dealloc_multi (multi_desc);
	ch_close (channel);
	return (NULL);
    }
    if (!can_mmap)
    {
	array_pointer arrayp;

	if ( !ds_alloc_vm_array (&arrayp, arr_desc, FALSE, FALSE) &&
	     !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp, arr_desc, FALSE,
				   FALSE) )
	{
	    m_error_notify (function_name, "array");
	    ds_dealloc_multi (multi_desc);
	    ch_close (channel);
	    return (NULL);
	}
	m_copy (multi_desc->data[0], &arrayp, sizeof arrayp);
	if ( !foreign_fits_read_data (channel, multi_desc, NULL, 0,
				      FA_FITS_READ_DATA_NUM_BLANKS,
				      blank_count,
				      FA_FITS_READ_DATA_END) )
	{
	    ds_dealloc_multi (multi_desc);
	    ch_close (channel);
	    return (NULL);
	}
	ch_close (channel);
	return (multi_desc);
    }
    m_clear (&arrayp, sizeof arrayp);
    arrayp.array = ch_get_mmap_addr (channel) + read_pos;
    arrayp.free = ( void (*) () ) ch_close;
    arrayp.info = channel;
    arrayp.mem_type = K_ARRAY_MEM_TYPE_MMAP;
    arrayp.u.mmap.fd = ch_get_descriptor (channel);
    arrayp.u.mmap.size = ch_get_mem_size (channel);
    arrayp.u.mmap.base = ch_get_mmap_addr (channel);
    m_copy (multi_desc->data[0], &arrayp, sizeof arrayp);
    return (multi_desc);
}   /*  End Function foreign_fits_read  */

/*PUBLIC_FUNCTION*/
flag foreign_fits_read_data (Channel channel, multi_array *multi_desc,
			     char *data, uaddr num_values, ...)
/*  [SUMMARY] Read data in a FITS file.
    [PURPOSE] This routine will read the data of a FITS file from a channel.
    The header section is NOT read.
    <channel> The channel to read from.
    <multi_desc> The Karma data structure to write the data into.
    <data> An alternate data array to write the FITS data into. If this is
    NULL, the routine will write the data into the Karma data structure.
    <num_values> The number of values to write into the data array. This is
    used when data is not NULL or skipping is requested.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_FITS_READ_DATA_END. See [<foreign_ATT_FITS_READ_DATA>] for a list of
    defined attributes.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    va_list argp;
    flag skip = FALSE;
    flag have_blank = FALSE;
    flag ascii = FALSE;
    uaddr toobig_count = 0;
    signed long long_data;
    signed long blank = 0;
    unsigned char uchar_data;
    unsigned int att_key;
    unsigned int value_count;
    unsigned int elem_type, elem_size;
    unsigned int type;
    float f_toobig = TOOBIG;
    double value[2];
    struct data_info_type *dinfo;
    char *packet;
    unsigned long *blank_count = NULL;
    float f_data;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "foreign_fits_read_data";

    va_start (argp, num_values);
    if ( (channel == NULL) || (multi_desc == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if ( !ds_get_named_object (multi_desc, 1, "FITS_READ", (void **) &dinfo) )
    {
	fprintf (stderr, "FITS_READ attachment not found\n");
	a_prog_bug (function_name);
    }
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp,unsigned int) ) != FA_FITS_READ_DATA_END )
    {
	switch (att_key)
	{
	  case FA_FITS_READ_DATA_NUM_BLANKS:
	    blank_count = va_arg (argp, unsigned long *);
	    break;
	  case FA_FITS_READ_DATA_SKIP_BLOCKS:
	    skip = va_arg (argp, flag);
	    FLAG_VERIFY (skip);
	    break;
	  case FA_FITS_READ_DATA_ASCII:
	    ascii = va_arg (argp, flag);
	    FLAG_VERIFY (ascii);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0];
    if (pack_desc->element_types[0] != K_ARRAY)
    {
	fprintf (stderr,"First element in top level packet must be K_ARRAY\n");
	a_prog_bug (function_name);
    }
    arr_desc = (array_desc *) pack_desc->element_desc[0];
    elem_type = arr_desc->packet->element_types[0];
    elem_size = host_type_sizes[elem_type];
    if ( ds_get_unique_named_value (pack_desc, packet, "BLANK", &type, value) )
    {
	if ( DOUBLE_IS_BLANK (value[0]) ) blank = 0x80000000;
	else blank = value[0];
	have_blank = TRUE;
    }
    else have_blank = FALSE;
    if (skip && !ascii)
    {
	unsigned long size;

	size = abs (dinfo->bitpix) / 8;
	if ( !ch_skip (channel, size * num_values) )
	{
	    fprintf (stderr, "Error skipping in FITS file\n");
	    return (FALSE);
	}
	return (TRUE);
    }
    if (data == NULL)
    {
	array_pointer arrayp;

	m_copy (&arrayp, packet, sizeof arrayp);
	data = arrayp.array;
	if (data == NULL)
	{
	    fprintf (stderr, "No array to write data into!\n");
	    a_prog_bug (function_name);
	}
	num_values = ds_get_array_size (arr_desc);
    }
    /*  Read data  */
    if (ascii)
    {
	char line[STRING_LENGTH];
	double value[2];

	value[1] = 0.0;
	for (value_count = 0; value_count < num_values;
	     ++value_count, data += elem_size)
	{
	    if ( !ch_getl (channel, line, STRING_LENGTH) )
	    {
		fprintf (stderr, "%s: error reading simple FITS file\n",
			 function_name);
		return (FALSE);
	    }
	    if (skip) continue;
	    sscanf (line, "%le", value);
	    if ( have_blank && (value[0] == blank) )
	    {
		value[0] = TOOBIG;
		++toobig_count;
	    }
	    else value[0] = value[0] * dinfo->bscale + dinfo->bzero;
	    ds_put_element (data, elem_type, value);
	}
	return (TRUE);
    }
    if (dinfo->bitpix == -32)
    {
	if (pio_read_floats (channel, num_values, (float *) data,
			     &toobig_count) < num_values)
	{
	    fprintf (stderr, "%s: error reading FITS file\n", function_name);
	    return (FALSE);
	}
	if (blank_count != NULL) *blank_count = toobig_count;
	return (TRUE);
    }
    if (dinfo->bitpix == -64)
    {
	if (pio_read_doubles (channel, num_values, (double *) data,
			      &toobig_count) < num_values)
	{
	    fprintf (stderr, "%s: error reading FITS file\n", function_name);
	    return (FALSE);
	}
	if (blank_count != NULL) *blank_count = toobig_count;
	return (TRUE);
    }
    if ( (dinfo->bitpix == 8) && (elem_type == K_UBYTE) )
    {
	if (ch_read (channel, data, num_values) < num_values)
	{
	    fprintf (stderr, "%s: error reading from file\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	if (blank_count != NULL) *blank_count = toobig_count;
	return (TRUE);
    }
    if (dinfo->bitpix == 8)
    {
	/*  Need to convert to float  */
	unsigned int block_size, count;
	char buffer[BUF_SIZE];
	CONST unsigned char *realbuf;

	for (value_count = 0; value_count < num_values;
	     value_count += block_size)
	{
	    block_size = num_values - value_count;
	    if (block_size > BUF_SIZE) block_size = BUF_SIZE;
	    if (ch_read2 (channel, buffer, (CONST char **) &realbuf, 0,
			  block_size) < block_size)
	    {
		fprintf (stderr, "%s: error reading from file\t%s\n",
			 function_name, ERRSTRING);
		return (FALSE);
	    }
	    for (count = 0; count < block_size; ++count, data += elem_size)
	    {
		uchar_data = realbuf[count];
		if ( have_blank && (uchar_data == blank) )
		{
		    f_data = f_toobig;
		    ++toobig_count;
		}
		else f_data = (float) uchar_data * dinfo->bscale +dinfo->bzero;
		*(float *) data = f_data;
	    }
	}
	if (blank_count != NULL) *blank_count = toobig_count;
	return (TRUE);
    }
    if (dinfo->bitpix == 16)
    {
	flag block_transfer, swap_transfer;
	unsigned int block_size, count;
	CONST char *ptr;
	short buffer[BUF_SIZE * 2 / sizeof (short)];  /*  Force alignment  */
	CONST short *realbuf;

	block_transfer = ds_can_transfer_element_as_block (K_SHORT);
	swap_transfer = ds_can_swaptransfer_element (K_SHORT);
	for (value_count = 0; value_count < num_values;
	     value_count += block_size)
	{
	    block_size = num_values - value_count;
	    if (block_size > BUF_SIZE) block_size = BUF_SIZE;
	    if (ch_read2 (channel, (char *) buffer,
			  (CONST char **) &realbuf, sizeof *realbuf,
			  block_size * 2) < block_size * 2)
	    {
		fprintf (stderr, "%s: error reading from file\t%s\n",
			 function_name, ERRSTRING);
		return (FALSE);
	    }
	    if (swap_transfer)
	    {
		m_copy_and_swap_blocks (buffer, realbuf, 2, 2, 2, block_size);
		realbuf = buffer;
	    }
	    for (count = 0, ptr = (CONST char *) realbuf;
		 count < block_size; ++count, data += elem_size, ptr += 2)
	    {
		if (block_transfer || swap_transfer)
		    long_data = *(CONST signed short *) ptr;
		else p_read_buf16s (ptr, &long_data);
		if (elem_type == K_SHORT)
		{
		    if ( have_blank && (long_data == blank) )
		    {
			long_data = -32768;
			++toobig_count;
		    }
		    *(signed short *) data = long_data;
		}
		else
		{
		    /*  Convert to float  */
		    if ( have_blank && (long_data == blank) )
		    {
			f_data = f_toobig;
			++toobig_count;
		    }
		    else f_data = (float) long_data * dinfo->bscale +
			dinfo->bzero;
		    *(float *) data = f_data;
		}
	    }
	}
	if (blank_count != NULL) *blank_count = toobig_count;
	return (TRUE);
    }
    if (dinfo->bitpix == 32)
    {
	flag block_transfer, swap_transfer;
	unsigned int block_size, count;
	CONST char *ptr;
	signed int buffer[BUF_SIZE * 4 / sizeof (int)]; /*  Force alignment  */
	CONST int *realbuf;

	block_transfer = ds_can_transfer_element_as_block (K_INT);
	swap_transfer = ds_can_swaptransfer_element (K_INT);
	for (value_count = 0; value_count < num_values;
	     value_count += block_size)
	{
	    block_size = num_values - value_count;
	    if (block_size > BUF_SIZE) block_size = BUF_SIZE;
	    if (ch_read2 (channel, (char *) buffer,
			  (CONST char **) &realbuf, sizeof *realbuf,
			  block_size * 4) < block_size * 4)
	    {
		fprintf (stderr, "%s: error reading from file\t%s\n",
			 function_name, ERRSTRING);
		return (FALSE);
	    }
	    if (swap_transfer)
	    {
		m_copy_and_swap_blocks (buffer, realbuf, 4, 4, 4, block_size);
		realbuf = buffer;
	    }
	    for (count = 0, ptr = (CONST char *) realbuf;
		 count < block_size; ++count, data += elem_size, ptr += 4)
	    {
		if (block_transfer || swap_transfer)
		    long_data = *(CONST signed int *) ptr;
		else p_read_buf32s (ptr, &long_data);
		if (elem_type == K_INT)
		{
		    if ( have_blank && (long_data == blank) )
		    {
			long_data = 0x80000000;
			++toobig_count;
		    }
		    *(signed int *) data = long_data;
		}
		else
		{
		    /*  Convert to float  */
		    if ( have_blank && (long_data == blank) )
		    {
			f_data = f_toobig;
			++toobig_count;
		    }
		    else f_data = (float) long_data * dinfo->bscale +
			dinfo->bzero;
		    *(float *) data = f_data;
		}
	    }
	}
	if (blank_count != NULL) *blank_count = toobig_count;
	return (TRUE);
    }
    fprintf (stderr, "Illegal BITPIX: %d\n", dinfo->bitpix);
    a_prog_bug (function_name);
    return (FALSE);
}   /*  End Function foreign_fits_read_data  */

/*UNPUBLISHED_FUNCTION*/
void foreign_fits_copy_special (multi_array *new, CONST multi_array *old)
/*  [SUMMARY] Copy special data from one multi_array to another.
    <new> The data structure to copy the data to.
    <old> The data structure to copy the data from.
    [RETURNS] Nothing. On failure the process aborts.
*/
{
    struct data_info_type *dinfo;
    void *old_info;
    static char function_name[] = "foreign_fits_copy_special";

    if ( !ds_get_named_object (old, 1, "FITS_READ", &old_info) )
    {
	fprintf (stderr, "FITS_READ attachment not found\n");
	a_prog_bug (function_name);
    }
    if ( ( dinfo = m_dup (old_info, sizeof *dinfo) ) == NULL )
    {
	m_abort (function_name, "FITS_READ attachment");
    }
    if ( !ds_add_named_object (new, 1, "FITS_READ", dinfo, m_free) )
    {
	m_abort (function_name, "FITS_READ attachment");
    }
}   /*  End Function foreign_fits_copy_special  */


/*  Private functions follow  */

static flag extract_info_from_header (char card[CARD_LENGTH][CARD_WIDTH],
				      struct card_info_type *info,
				      flag strip_trailing_whitespace)
/*  This routine will extract information from a single FITS card.
    The card must be pointed to by  card  .
    The information extracted will be written to the storage pointed to by
    info  .
    If the flag  strip_trailing_whitespace  is TRUE, then converted strings
    with trailing whitespace will have the whitespace removed.
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    flag quote;
    flag keep_searching;
    flag blank_keyword;
    char *char_ptr, *value_ptr;
    unsigned int line_count, char_count;
    int naxis, dim_index;
    char line[STRING_LENGTH], tmp[STRING_LENGTH];
    struct keyword_type *new_keyword, *curr;
    static char function_name[] = "_fits_read_extract_info_from_header";

    for (line_count = 0; line_count < CARD_LENGTH; ++line_count)
    {
	m_copy (line, card[line_count], CARD_WIDTH);
	line[CARD_WIDTH] = '\0';
	/*  Change any '\0' characters to ' '  */
	for (char_count = 0; char_count < CARD_WIDTH - 1; ++char_count)
	{
	    if (line[char_count] == '\0') line[char_count] = ' ';
	}
	/*  Strip trailing whitespace  */
	for (char_ptr = line + CARD_WIDTH - 1;
	     isspace (*char_ptr) && (char_ptr >= line); --char_ptr)
	{
	    *char_ptr = '\0';
	}
	if (line[0] == '\0') continue;
	if (strcmp (line, "END") == 0)
	{
	    info->end_found = TRUE;
	    return (TRUE);
	}
	if (strncmp (line, "HISTORY", 7) == 0)
	{
	    /*  Append history. First clean up line  */
	    char_ptr = line + 7;
	    /*  Skip leading whitespace  */
	    while ( isspace (*char_ptr) ) ++char_ptr;
	    strncpy (tmp, char_ptr, line + CARD_WIDTH - char_ptr);
	    tmp[line + CARD_WIDTH - char_ptr] = '\0';
	    /*  Remove comment  */
	    for (quote = FALSE, char_count = 0; char_count < CARD_WIDTH;
		 ++char_count)
	    {
		switch (tmp[char_count])
		{
		  case '/':
		    if (!quote) tmp[char_count] = '\0';
		    break;
		  case '\'':
		    quote = quote ? FALSE : TRUE;
		    break;
		}
	    }
	    if ( !ds_history_append_string (info->history_holder, tmp, TRUE) )
	    {
		m_abort (function_name, "history string");
	    }
	    continue;
	}
	if (strncmp (line, "COMMENT", 7) == 0) continue;
	blank_keyword = TRUE;
	for (char_count = 0; blank_keyword && (char_count < 8); ++char_count)
	{
	    if ( !isspace (line[char_count]) ) blank_keyword = FALSE;
	}
	if (blank_keyword) continue;
	if (line[0] == '\t')
	{
	    fprintf (stderr, "WARNING: tab in first column: ignoring line\n");
	    continue;
	}
	/*  Tidy up line  */
	/*  Remove '=' and any whitespace before '='  */
	if (line[EQUALS_POSITION] != '=')
	{
	    line[CARD_WIDTH] = '\0';
	    fprintf (stderr,
		     "WARNING: Expected '=' at column: %u in line: \"%s\"\n",
		     EQUALS_POSITION + 1, line);
	    fprintf (stderr, "Skipping line\n");
	    continue;
	}
	line[EQUALS_POSITION] = '\0';
	char_ptr = line + EQUALS_POSITION - 1;
	while ( isspace (*char_ptr) )
	{
	    *char_ptr-- = '\0';
	}
	/*  Search for value field  */
	value_ptr = line + EQUALS_POSITION + 1;
	while ( isspace (*value_ptr) ) ++value_ptr;
	/*  Search for '/' (comment) or '\'' (quote) character  */
	char_ptr = value_ptr;
	quote = FALSE;
	keep_searching = TRUE;
	while (keep_searching)
	{
	    switch (*char_ptr)
	    {
	      case '/':
		/*  Comment  */
		if (quote)
		{
		    /*  Ignore  */
		    ++char_ptr;
		}
		else
		{
		    /*  Comment (unquoted)  */
		    keep_searching = FALSE;
		}
		break;
	      case '\'':
		quote = quote ? FALSE : TRUE;
		++char_ptr;
		break;
	      default:
		++char_ptr;
		break;
	    }
	    if (char_ptr - line >= CARD_WIDTH) keep_searching = FALSE;
	}
	if (*char_ptr == '/')
	{
	    /*  Found comment: kill it  */
	    *char_ptr = '\0';
	}
	/*  Remove trailing whitespace  */
	--char_ptr;
	while ( isspace (*char_ptr) )
	{
	    *char_ptr-- = '\0';
	}
	/*  Now start extacting info  */
	if (strcmp (line, "SIMPLE") == 0)
	{
	    if ( (info->simple == TRUE) || (info->simple == FALSE) )
	    {
		redeclaration_message (line);
		continue;
	    }
	    if (*value_ptr == 'T')
	    {
		info->simple = TRUE;
	    }
	    else if (*value_ptr == 'F')
	    {
		info->simple = FALSE;
	    }
	    else
	    {
		conversion_error_message (line, value_ptr);
		info->simple = FALSE;
		continue;
	    }
	}
	else if (strcmp (line, "BITPIX") == 0)
	{
	    if (info->data_info.bitpix != 0)
	    {
		redeclaration_message (line);
	    }
	    if (sscanf (value_ptr, "%d", &info->data_info.bitpix) != 1)
	    {
		conversion_error_message (line, value_ptr);
		return (FALSE);
	    }
	}
	else if (strcmp (line, "BZERO") == 0)
	{
	    if ( DOUBLE_IS_VALID (info->data_info.bzero) )
	    {
		redeclaration_message (line);
	    }
	    if (sscanf (value_ptr, "%le", &info->data_info.bzero) != 1)
	    {
		conversion_error_message (line, value_ptr);
		info->data_info.bzero = 0.0;
		continue;
	    }
	}
	else if (strcmp (line, "BSCALE") == 0)
	{
	    if ( DOUBLE_IS_VALID (info->data_info.bscale) )
	    {
		redeclaration_message (line);
	    }
	    if (sscanf (value_ptr, "%le", &info->data_info.bscale) != 1)
	    {
		conversion_error_message (line, value_ptr);
		info->data_info.bscale = 1.0;
	    }
	}
	else if (strcmp (line, "BUNIT") == 0)
	{
	    if (info->elem_name != NULL)
	    {
		redeclaration_message (line);
	    }
	    if ( ( info->elem_name = ex_str (value_ptr, (CONST char **) NULL) )
		 == NULL )
	    {
		fprintf (stderr,
			 "%s: error extracting BUNIT value for line: \"%s\"\n",
			 function_name, line);
		return (FALSE);
	    }
	    if (strip_trailing_whitespace)
	    {
		cleanup_string (info->elem_name);
	    }
	}
	else if (strcmp (line, "NAXIS") == 0)
	{
	    if (sscanf (value_ptr, "%d", &naxis) != 1)
	    {
		conversion_error_message (line, value_ptr);
		return (FALSE);
	    }
	    if (info->naxis != -1)
	    {
		redeclaration_message (line);
		if (info->naxis != naxis)
		{
		    fprintf (stderr, "NAXIS: old value: %d  new value: %d\n",
			     info->naxis, naxis);
		    continue;
		}
	    }
	    info->naxis = naxis;
	    /*  Allocate dimension info arrays  */
	    if ( ( info->dim_lengths =
		   m_alloc (sizeof *info->dim_lengths * info->naxis) )
		 == NULL )
	    {
		m_abort (function_name, "dimension lengths");
	    }
	    if ( ( info->dim_names =
		   m_alloc (sizeof *info->dim_names * info->naxis) )
		 == NULL )
	    {
		m_abort (function_name, "dimension names");
	    }
	    if ( ( info->axis_scales =
		   m_alloc (sizeof *info->axis_scales * info->naxis) )
		 == NULL )
	    {
		m_abort (function_name, "axis scales");
	    }
	    /*  Clear dimension lengths and set minima and maxima  */
	    for (dim_index = 0; dim_index < info->naxis; ++dim_index)
	    {
		info->dim_lengths[dim_index] = -1;
		info->dim_names[dim_index] = NULL;
		info->axis_scales[dim_index] = 1.0;
	    }
	}
	else if (strncmp (line, "NAXIS", 5) == 0)
	{
	    if (info->dim_lengths == NULL)
	    {
		fprintf (stderr, "Keyword: \"%s\" came before \"NAXIS\"\n",
			 line);
		return (FALSE);
	    }
	    /*  Get info for this dimension  */
	    if (sscanf (line + 5, "%d", &naxis) != 1)
	    {
		conversion_error_message (line, line + 5);
		continue;
	    }
	    if ( (naxis < 0) || (naxis > info->naxis) )
	    {
		fprintf (stderr, "NAXIS%d goes beyond NAXIS=%d, dropping\n",
			 naxis, info->naxis);
		continue;
	    }
	    dim_index = info->naxis - naxis;
	    if (info->dim_lengths[dim_index] > -1)
	    {
		redeclaration_message (line);
	    }
	    if (sscanf (value_ptr, "%d", &info->dim_lengths[dim_index]) != 1)
	    {
		conversion_error_message (line, value_ptr);
		return (FALSE);
	    }
	}
	else if ( (strncmp (line, "CTYPE", 5) == 0) && isalldigit (line + 5) )
	{
	    if (info->dim_names == NULL)
	    {
		fprintf (stderr, "Keyword: \"%s\" came before \"NAXIS\"\n",
			 line);
		return (FALSE);
	    }
	    /*  Get info for this dimension  */
	    if (sscanf (line + 5, "%d", &naxis) != 1)
	    {
		conversion_error_message (line, line + 5);
		continue;
	    }
	    /*  Record in Karma array descriptor if naxis is legal, otherwise
		just leave it for the keyword section  */
	    if ( (naxis >= 1) && (naxis <= info->naxis) )
	    {
		dim_index = info->naxis - naxis;
		if (info->dim_names[dim_index] != NULL)
		{
		    redeclaration_message (line);
		}
		if ( ( char_ptr = ex_str (value_ptr, (CONST char **) NULL) )
		     == NULL )
		{
		    m_abort (function_name, "dimension name");
		    return (FALSE);
		}
		if ( ( info->dim_names[dim_index] = m_alloc (STRING_LENGTH) )
		     == NULL )
		{
		    m_abort (function_name, "dimension name buffer");
		    return (FALSE);
		}
		strcpy (info->dim_names[dim_index], char_ptr);
		m_free (char_ptr);
		if (strip_trailing_whitespace)
		{
		    cleanup_string (info->dim_names[dim_index]);
		}
	    }
	}
	else if (strncmp (line, "CUNIT", 5) == 0)
	{
	    flag bad = FALSE;
	    char unit_string[STRING_LENGTH];

	    if (info->axis_scales == NULL)
	    {
		fprintf (stderr, "Keyword: \"%s\" came before \"NAXIS\"\n",
			 line);
		return (FALSE);
	    }
	    /*  Get info for this dimension  */
	    if (sscanf (line + 5, "%d", &naxis) != 1)
	    {
		conversion_error_message (line, line + 5);
		continue;
	    }
	    if ( (naxis < 0) || (naxis > info->naxis) )
	    {
		fprintf (stderr,
			 "Non-standard \"CUNIT%d\" goes beyond NAXIS=%d, dropping\n",
			 naxis, info->naxis);
		continue;
	    }
	    if ( ( char_ptr = ex_str (value_ptr, (CONST char **) NULL) )
		 == NULL )
	    {
		m_abort (function_name, "dimension name");
		return (FALSE);
	    }
	    strcpy (unit_string, char_ptr);
	    m_free (char_ptr);
	    /*  Remove trailing whitespace  */
	    for (char_ptr = unit_string + strlen (unit_string) - 1;
		 (char_ptr > unit_string) && isspace (*char_ptr);
		 --char_ptr) *char_ptr = '\0';
	    if ( (st_icmp (unit_string, "m/s") == 0) ||
		 (st_icmp (unit_string, "degree") == 0) ||
		 (st_icmp (unit_string, "hz") == 0) )
	    {
		info->axis_scales[naxis - 1] = 1.0;
	    }
	    else if (st_icmp (unit_string, "km/s") == 0)
	    {
		info->axis_scales[naxis - 1] = 1e3;
		bad = TRUE;
	    }
	    else
	    {
		fprintf (stderr,
			 "Unknown unit: \"%s\" for: \"CUNIT%d\", dropping\n",
			 unit_string, naxis);
		continue;
	    }
	    if (bad)
	    {
		fprintf (stderr,
			 "Non-standard \"CUNIT%d\" has non-standard unit: \"%s\"\n",
			 naxis, unit_string);
		fprintf (stderr, "Modifying standard header entries\n");
	    }
	    else
	    {
		fprintf (stderr,
			 "Non-standard \"CUNIT%d\" has no effect: dropping\n",
			 naxis);
	    }
	    /*  Do not allow CUNITn to leak out  */
	    continue;
	}
	/*  Save all keywords  */
	/*  If keyword is duplicated, update  */
	new_keyword = NULL;
	for (curr = info->first_keyword;
	     (curr != NULL) && (new_keyword == NULL);
	     curr = curr->next)
	{
	    if (strcmp (line, curr->keyword) == 0) new_keyword = curr;
	}
	/*  Append to list only if this is a new keyword  */
	if (new_keyword == NULL)
	{
	    if ( ( new_keyword = m_alloc (sizeof *new_keyword) ) == NULL )
	    {
		m_abort (function_name, "keyword struct");
	    }
	    /*  Append  */
	    new_keyword->next = NULL;
	    if (info->first_keyword == NULL) info->first_keyword = new_keyword;
	    else info->last_keyword->next = new_keyword;
	    info->last_keyword = new_keyword;
	}
	else
	{
	    /*  Deallocate old name and value  */
	    m_free (new_keyword->keyword);
	    m_free (new_keyword->value);
	}
	if ( ( new_keyword->keyword = st_dup (line) ) == NULL )
	{
	    m_abort (function_name, "keyword name");
	}
	if ( ( new_keyword->value = st_dup (value_ptr) ) == NULL )
	{
	    m_abort (function_name, "keyword value");
	}
	/*  Scan next line  */
    }
    return (TRUE);
}   /*  End Function extract_info_from_header  */

static void redeclaration_message (char *string)
/*  This routine will display a message stating that the parameter with name
    pointed to by  string  was re-declared.
    The routine returns nothing.
*/
{
    fprintf (stderr, "FITS keyword: \"%s\" redeclared\n", string);
}   /*  End Function redeclaration_message  */

static void conversion_error_message (char *parameter, char *value)
/*  This routine will print an error message indicating that there was an
    error scanning the value string for the parameter.
    The parameter name must be pointed to by  parameter  .
    The value string must be pointed to by  value  .
    The routine returns nothing.
*/
{
    fprintf (stderr,
	     "Error converting value string: \"%s\" for FITS keyword: \"%s\"\n",
	     value, parameter);
}   /*  End Function conversion_error_message  */

static void cleanup_string (char *string)
/*  This routine will strip trailing whitespace in a string by replacing that
    whitespace with NULL characters.
    The string must be pointed to by  string  .
    The routine returns nothing.
*/
{
    char *ptr;

    ptr = string + strlen (string) - 1;
    while (ptr >= string)
    {
	if ( isspace (*ptr) )
	{
	    /*  Some whitespace: kill it  */
	    *ptr = '\0';
	    --ptr;
	}
	else
	{
	    /*  Not whitespace: finished  */
	    return;
	}
    }
}   /*  End Function cleanup_string  */

static void free_keywords (struct keyword_type *keywords)
/*  This routine will free the storage associated with unknown keywords.
    The keywords must be pointed to by  keywords  .
    The routine returns nothing.
*/
{
    struct keyword_type *next;

    while (keywords != NULL)
    {
	if (keywords->keyword != NULL) m_free (keywords->keyword);
	if (keywords->value != NULL) m_free (keywords->value);
	next = keywords->next;
	m_free (keywords);
	keywords = next;
    }
}   /*  End Function free_keywords  */

static flag process_keywords (struct keyword_type *keywords,
			      multi_array *multi_desc,
			      struct card_info_type finfo)
/*  This routine will process the unknown keywords and add them to a Karma
    data structure.
    The keywords must be pointed to by  keywords  .
    The multi_array data structure must be pointed to by  multi_desc  .
    <finfo> The card information.
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    int bool, naxis;
    double scale;
    double value[2];
    char *string, *ptr;

    for (; keywords != NULL; keywords = keywords->next)
    {
	if (strcmp (keywords->keyword, "SIMPLE") == 0) continue;
	if (strcmp (keywords->keyword, "BITPIX") == 0) continue;
	if (finfo.hide_scaling)
	{
	    if (strcmp (keywords->keyword, "BSCALE") == 0) continue;
	    if (strcmp (keywords->keyword, "BZERO") == 0) continue;
	}
	/*  Try to determine type of keyword  */
	scale = 1.0;
	if (ds_f_name_in_packet (multi_desc->headers[0], keywords->keyword,
				 NULL, NULL) != IDENT_NOT_FOUND)
	{
	    fprintf (stderr,
		     "Identifier: \"%s\" already used, ignoring keyword\n",
		     keywords->keyword);
	    continue;
	}
	switch (keywords->value[0])
	{
	  case 'T':
	  case 'F':
	    /*  TRUE or FALSE: assign to K_INT  */
	    bool = ( keywords->value[0] == 'T' ) ? TRUE : FALSE;
	    value[0] = bool;
	    value[1] = 0.0;
	    if ( !ds_put_unique_named_value (multi_desc->headers[0],
					     &multi_desc->data[0],
					     keywords->keyword, K_INT,
					     value, FALSE) ) return (FALSE);
	    break;
	  case '\'':
	    /*  String  */
	    if (keywords->value[1] == '\'') continue; /*  Drop empty string  */
	    if ( ( string = ex_str (keywords->value, (CONST char **) &ptr) )
		 == NULL )
	    {
		fprintf (stderr, "Error extracting string from: \"%s\"\n",
			 keywords->value);
		fprintf (stderr, "Ignoring keyword: \"%s\"\n",
			 keywords->keyword);
		continue;
	    }
	    /*  Strip trailing whitespace  */
	    for (ptr = string + strlen (string) - 1;
		 isspace (*ptr) && ptr >= string; --ptr) *ptr = '\0';
	    /*  Some broken FITS writers put in a string for EQUINOX. Fix  */
	    if (strcmp (keywords->keyword, "EQUINOX") == 0)
	    {
		fprintf (stderr,
			 "WARNING: string: \"%s\" found for keyword: \"%s\", expected value\n",
			 string, keywords->keyword);
		if (st_nicmp (string, "B1950", 5) == 0) value[0] = 1950.0;
		else if (st_nicmp (string, "J", 1) == 0)
		    value[0] = atoi (string + 1);
		else
		{
		    fprintf (stderr, "         bogus string, ignoring\n");
		    continue;
		}
		fprintf (stderr, "         assuming value %g\n", value[0]);
		ds_put_unique_named_value (multi_desc->headers[0],
					   &multi_desc->data[0],
					   keywords->keyword, K_DOUBLE,
					   value, FALSE);
		continue;
	    }
	    if ( !ds_put_unique_named_string (multi_desc->headers[0],
					      multi_desc->data,
					      keywords->keyword,
					      string, FALSE ) )
	    {
		m_free (string);
		return (FALSE);
	    }
	    m_free (string);
	    break;
	  default:
	    /*  You better be numeric: you ...  */
	    if (isdigit (keywords->value[0]) == 0)
	    {
		/*  Not numeric!  */
		fprintf (stderr,
			 "WARNING: bad value: \"%s\" for keyword: \"%s\" assuming string\n",
			 keywords->value, keywords->keyword);
		if ( !ds_put_unique_named_string (multi_desc->headers[0],
						  multi_desc->data,
						  keywords->keyword,
						  keywords->value,
						  FALSE ) ) return (FALSE);
		break;
	    }
	    /*  Drop through  */
	  case '-':
	  case '.':
	    /*  Get scaling information  */
	    if ( (strncmp (keywords->keyword, "CRVAL", 5) == 0) ||
		 (strncmp (keywords->keyword, "CDELT", 5) == 0) )
	    {
		sscanf (keywords->keyword + 5, "%d", &naxis);
		if (naxis <= finfo.naxis)
		{
		    scale = finfo.axis_scales[naxis - 1];
		    if (scale != 1.0)
		    {
			fprintf (stderr, "Scaling \"%s\" by %g\n",
				 keywords->keyword, scale);
		    }
		}
		else scale = 1.0;
	    }
	    /*  Check for 'E', 'e' or '.' (indicates float value)  */
	    if ( (strchr (keywords->value, 'E') == NULL) &&
		 (strchr (keywords->value, 'e') == NULL) &&
		 (strchr (keywords->value, '.') == NULL) )
	    {
		/*  Integer  */
		value[0] = ex_int (keywords->value, (CONST char **) &ptr);
		value[0] *= scale;
		value[1] = 0.0;
		if ( !ds_put_unique_named_value (multi_desc->headers[0],
						 &multi_desc->data[0],
						 keywords->keyword, K_INT,
						 value,FALSE) ) return (FALSE);
		continue;
	    }
	    /*  Floating point  */
	    sscanf (keywords->value, "%le", value);
	    value[0] *= scale;
	    value[1] = 0.0;
	    if ( !ds_put_unique_named_value (multi_desc->headers[0],
					     &multi_desc->data[0],
					     keywords->keyword, K_DOUBLE,
					     value, FALSE) ) return (FALSE);
	    break;
	}
    }
    return (TRUE);
}   /*  End Function process_keywords  */

static void free_card_info (struct card_info_type *info)
/*  [SUMMARY] Free info in a card.
    <info> The card info structure.
    [RETURNS] Nothing.
*/
{
    int count;

    if (info->dim_lengths != NULL) m_free (info->dim_lengths);
    for (count = 0; count < info->naxis; ++count)
    {
	if (info->dim_names[count] != NULL) m_free (info->dim_names[count]);
    }
    if (info->dim_names != NULL) m_free (info->dim_names);
    if (info->elem_name != NULL) m_free (info->elem_name);
    if (info->axis_scales != NULL) m_free (info->axis_scales);
    if (info->history_holder != NULL) ds_dealloc_multi (info->history_holder);
}   /*  End Function free_card_info  */

static flag isalldigit (CONST char *string)
/*  [SUMMARY] Test if a string is all digits.
    <string> The string to test.
    [RETURNS] TRUE if the string is all digits, else FALSE.
*/
{
    while (*string != '\0')
    {
	if ( !isdigit (*string) ) return (FALSE);
	++string;
    }
    return (TRUE);
}   /*  End Function isalldigit  */
