/*LINTLIBRARY*/
/*  guess_read.c

    This code provides miscellaneous functions for the <foreign_> package.

    Copyright (C) 1995-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for the <foreign_> package.


    Written by      Richard Gooch   23-APR-1995

    Updated by      Richard Gooch   24-APR-1995: Created
  FA_GUESS_READ_FITS_TO_FLOAT  attribute.

    Updated by      Richard Gooch   6-MAY-1995: Placate gcc -Wall

    Updated by      Richard Gooch   21-MAY-1995: Added support for Sun
  rasterfile format.

    Updated by      Richard Gooch   28-SEP-1995: Added support for Miriad Image
  format.

    Updated by      Richard Gooch   12-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   21-JUN-1996: Added support for GIPSY format

    Updated by      Richard Gooch   12-OCT-1996: Created
  <foreign_read_and_setup> routine.

    Updated by      Richard Gooch   3-DEC-1996: Added support for PGM files.

    Updated by      Richard Gooch   27-DEC-1996: Made use of
  <iarray_create_and_setup>.

    Updated by      Richard Gooch   14-MAY-1997: Supported compressed FITS
  files.

    Updated by      Richard Gooch   14-JUL-1997: Supported DRAO datasets.

    Updated by      Richard Gooch   29-JUL-1997: Supported filtered files.

    Updated by      Richard Gooch   30-JUL-1997: Supported AIPS datasets.

    Updated by      Richard Gooch   8-AUG-1997: Passed mmap_option and
  writable arguments to Miriad reader.

    Updated by      Richard Gooch   21-MAR-1998: Added test for Miriad format
  when always mmapping.

    Updated by      Richard Gooch   14-APR-1998: Changed to include
  karma_wcs_def.h instead.

    Updated by      Richard Gooch   5-MAY-1998: Made use of new experimantal
  function <foreign_pnm_read_type>.

    Updated by      Richard Gooch   9-MAY-1998: Attached "filename" object.

    Updated by      Richard Gooch   16-MAY-1998: Made use of new experimental
  function <foreign_fits_read>.

    Updated by      Richard Gooch   27-MAY-1998: Added support for empty FITS
  files.

    Updated by      Richard Gooch   29-JUN-1999: Added support for simple FITS
  files.

    Updated by      Richard Gooch   30-NOV-2000: Added support for IDF datasets

    Last updated by Richard Gooch   13-NOV-2001: Added FITS to list of formats
  that work when mmap==K_CH_MAP_ALWAYS.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_dsxfr.h>
#include <karma_wcs_def.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>


/*  Private functions  */


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
multi_array *foreign_guess_and_read (CONST char *filename,
				     unsigned int mmap_option, flag writable,
				     unsigned int *ftype, ...)
/*  [SUMMARY] Guess file type and read.
    [PURPOSE] This routine will attempt to guess the filetype of a file and
    in the file, converting to the Karma data format if possible.
    <filename> The name of the file to read.
    <mmap_option> This has the same meaning as for the <dsxfr_get_multi>
    routine.
    <writable> This has the same meaning as for the <dsxfr_get_multi> routine.
    <ftype> The type of the file that was read in is written here. This may be
    NULL.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must terminated with FA_GUESS_READ_END.
    See [<foreign_ATT_GUESS>] for a list of defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    Channel inp;
    flag fits_convert_to_float = FALSE;
    unsigned int att_key, filetype;
    va_list argp;
    multi_array *multi_desc = NULL;  /*  Initialised to keep compiler happy  */
    void *object = NULL;
    static char function_name[] = "foreign_guess_and_read";

    va_start (argp, ftype);
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp, unsigned int) ) != FA_GUESS_READ_END )
    {
	switch (att_key)
	{
	  case FA_GUESS_READ_FITS_TO_FLOAT:
	    fits_convert_to_float = va_arg (argp, flag);
	    FLAG_VERIFY (fits_convert_to_float);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    filetype = foreign_guess_format_from_filename (filename);
    if (ftype != NULL) *ftype = filetype;
    if (filetype == FOREIGN_FILE_FORMAT_UNKNOWN) return (NULL);
    if (mmap_option == K_CH_MAP_ALWAYS)
    {
	switch (filetype)
	{
	  case FOREIGN_FILE_FORMAT_KARMA:
	  case FOREIGN_FILE_FORMAT_FITS:
	  case FOREIGN_FILE_FORMAT_MIRIAD:
	    break;
	  default:
	    return (NULL);
	    /*break;*/
	}
    }
    switch (filetype)
    {
      case FOREIGN_FILE_FORMAT_KARMA:
	if ( ( multi_desc = dsxfr_get_multi (filename, FALSE,
					     mmap_option, writable) )
	     == NULL )
	{
	    fprintf (stderr, "%s: error getting arrayfile: \"%s\"\n",
		     function_name, filename);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_PNM:
      case FOREIGN_FILE_FORMAT_PGM:
      case FOREIGN_FILE_FORMAT_PPM:
	/*  Read PNM file  */
	if ( ( inp = ch_open_file (filename, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, filename, ERRSTRING);
	    return (NULL);
	}
	if ( ( multi_desc = foreign_pnm_read (inp) ) == NULL )
	{
	    ch_close (inp);
	    return (NULL);
	}
	ch_close (inp);
	break;
      case FOREIGN_FILE_FORMAT_FITS:
      case FOREIGN_FILE_FORMAT_SIMPLE_FITS:
	if ( ( multi_desc =
	       foreign_fits_read (filename, fits_convert_to_float, TRUE,
				  FA_FITS_READ_MMAP_OPTION, mmap_option,
				  FA_FITS_READ_WRITABLE, writable,
				  FA_FITS_READ_SIMPLE,
				  (filetype ==
				   FOREIGN_FILE_FORMAT_SIMPLE_FITS) ?
				  TRUE : FALSE,
				  FA_FITS_READ_END) )
	     == NULL )
	{
	    fprintf (stderr, "%s: error reading FITS file\n", function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_EMPTY_FITS:
	if ( ( inp = ch_open_file (filename, "rz") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, filename, ERRSTRING);
	    return (NULL);
	}
	if ( ( multi_desc =
	       foreign_fits_read_header (inp, TRUE, fits_convert_to_float,TRUE,
					 FA_FITS_READ_HEADER_ALLOW_TRUNC, TRUE,
					 FA_FITS_READ_END) )
	     == NULL )
	{
	    fprintf (stderr, "%s: error reading FITS header\n", function_name);
	    ch_close (inp);
	    return (NULL);
	}
	ch_close (inp);
	break;
      case FOREIGN_FILE_FORMAT_SUNRAS:
	/*  Read Sun rasterfile  */
	if ( ( inp = ch_open_file (filename, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, filename, ERRSTRING);
	    return (NULL);
	}
	if ( ( multi_desc = foreign_sunras_read (inp,
						 FA_SUNRAS_READ_END) )
	     == NULL )
	{
	    fprintf (stderr, "%s: error reading Sun rasterfile\n",
		     function_name);
	    ch_close (inp);
	    return (NULL);
	}
	ch_close (inp);
	break;
      case FOREIGN_FILE_FORMAT_MIRIAD:
	if ( ( multi_desc =
	       foreign_miriad_read (filename, TRUE,
				    FA_MIRIAD_READ_MMAP_OPTION, mmap_option,
				    FA_MIRIAD_READ_WRITABLE, writable,
				    FA_MIRIAD_READ_END) )
	     == NULL )
	{
	    fprintf (stderr, "%s: error reading Miriad Image file\n",
		     function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	if ( ( multi_desc = foreign_gipsy_read (filename, TRUE,
						FA_GIPSY_READ_END) )
	     == NULL )
	{
	    fprintf (stderr, "%s: error reading GIPSY file\n",
		     function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_DRAO:
	if ( ( multi_desc = foreign_drao_read (filename) ) == NULL )
	{
	    fprintf (stderr, "%s: error reading DRAO file\n",
		     function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_AIPS1:
	if ( ( multi_desc = foreign_aips1_read (filename) ) == NULL )
	{
	    fprintf (stderr, "%s: error reading AIPS file\n", function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_FILTER_KARMA:
      case FOREIGN_FILE_FORMAT_FILTER_FITS:
      case FOREIGN_FILE_FORMAT_FILTER_PNM:
	if ( ( multi_desc = foreign_filter_read (filename) ) == NULL )
	{
	    fprintf (stderr, "%s: error reading filtered file\n",
		     function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_IDF:
	if ( ( multi_desc = foreign_idf_read (filename, TRUE,
					      FA_IDF_READ_END) ) == NULL )
	{
	    fprintf (stderr, "%s: error reading IDF file\n", function_name);
	    return (NULL);
	}
	break;
      default:
	fprintf (stderr, "Illegal filetype: %u\n", filetype);
	a_prog_bug (function_name);
	break;
    }
    ds_get_named_object (multi_desc, multi_desc->num_arrays, "filename",
			 &object);
    if (object == NULL)
	ds_add_named_object (multi_desc, multi_desc->num_arrays,
			     "filename", st_dup (filename), m_free);
    return (multi_desc);
}   /*  End Function foreign_guess_and_read  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_read_and_setup (CONST char *filename, unsigned int mmap_option,
			     flag writable, unsigned int *ftype, flag inform,
			     unsigned int num_dim,
			     unsigned int preferred_type, flag force_type,
			     iarray *array, double *min, double *max,
			     flag discard_zero_range, KwcsAstro *ap)
/*  [SUMMARY] Read a file and perform some setup.
    [PURPOSE] This routine will attempt to guess the filetype of a file and
    in the file, converting to an Intelligent Array if possible. The routine
    then performs some simple checks and some other convenience functions.
    <filename> The name of the file to read.
    <mmap_option> This has the same meaning as for the <dsxfr_get_multi>
    routine.
    <writable> This has the same meaning as for the <dsxfr_get_multi> routine.
    <ftype> The type of the file that was read in is written here. This may be
    NULL.
    <inform> If TRUE, the routine displays some informative messages.
    <num_dim> The number of dimensions required. If this is 0, any number of
    dimensions is allowed.
    <preferred_type> The preferred data type. If this is NONE, then no type is
    preferred.
    <force_type> If TRUE, the routine fails if the preferred data type was not
    available.
    <array> The Intelligent Array is written here. An existing array pointed to
    by this is deallocated.
    <min> The minimum data value in the array is written here. If this is NULL
    nothing is written here.
    <max> The maximum data value in the array is written here. If this is NULL
    nothing is written here.
    <discard_zero_range> If TRUE, and the range of the data is zero, the
    routine fails.
    <ap> The KwcsAstro object is written here. This may be NULL.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag convert_to_float = FALSE;
    flag ok;
    multi_array *multi_desc;
    static char function_name[] = "foreign_read_and_setup";

    /*  Check if floating point data is preferred  */
    if (preferred_type == K_FLOAT) convert_to_float = TRUE;
    if ( ( multi_desc =
	   foreign_guess_and_read(filename, mmap_option, writable, ftype,
				  FA_GUESS_READ_FITS_TO_FLOAT,convert_to_float,
				  FA_GUESS_READ_END) ) == NULL )
    {
	fprintf (stderr, "%s: error reading file: \"%s\"\n",
		 function_name, filename);
	return (FALSE);
    }
    ok = iarray_create_and_setup (array, multi_desc, inform, num_dim,
				  force_type ? preferred_type : NONE,
				  min, max, discard_zero_range, ap);
    ds_dealloc_multi (multi_desc);
    return (ok);
}   /*  End Function foreign_read_and_setup  */


/*  Private functions follow  */
