/*LINTLIBRARY*/
/*  gipsy_read.c

    This code provides an IDF (Image Descriptor File) read facility.

    Copyright (C) 2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading data files in
  IDF format (some random medical format).


    Written by      Richard Gooch   27-NOV-2000: Copied from gipsy_read.c

    Last updated by Richard Gooch   30-NOV-2000: First working version.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_dsrw.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_r.h>


#define MAX_DIMENSIONS 10


/*  Private data follows  */
static flag verbose = FALSE;


/*  Declarations of private functions follow  */


/*  Public functions follow  */


/*EXPERIMENTAL_FUNCTION*/
flag foreign_idf_test (CONST char *filename)
/*  [SUMMARY] Test if a file is part of an IDF file set.
    <filename> The name of the ".idf" header file in the IDF file set.
    [RETURNS] TRUE if the file is part of an IDF file set, else FALSE.
*/
{
    struct stat statbuf;
    char *ptr;
    char image_name[STRING_LENGTH];

    if ( ( ptr = strrchr (filename, '.') ) == NULL ) return (FALSE);
    if (strcmp (ptr + 1, "idf") != 0) return (FALSE);
    strncpy (image_name, filename, ptr - filename + 1);
    ptr = image_name + (ptr - filename);
    strcpy (ptr + 1, "int2");
    /*  Check for the header file  */
    if (stat (filename, &statbuf) != 0)
    {
	if (errno == ENOENT) return (FALSE);
	fprintf (stderr, "Error statting file: \"%s\"\t%s\n",
		 filename, ERRSTRING);
	return (FALSE);
    }
    /*  Well, the header file exists. Check for the image file  */
    if (stat (image_name, &statbuf) != 0)
    {
	if (errno == ENOENT) return (FALSE);
	fprintf (stderr, "Error statting file: \"%s\"\t%s\n",
		 image_name, ERRSTRING);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function foreign_idf_test  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_idf_read_header (Channel channel, flag data_alloc,
				      flag sanitise, ...)
/*  [SUMMARY] Read an IDF file header.
    [PURPOSE] This routine will read the header of an IDF file from a channel.
    The data section is NOT read.
    <channel> The channel to read from.
    <data_alloc> If TRUE, the data space is allocated.
    <sanitise> If TRUE, IDF axes with length 1 are ignored. This is highly
    recommended.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_IDF_READ_HEADER_END. See [<foreign_ATT_IDF_READ_HEADER>] for a
    list of defined attributes.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    va_list argp;
    int curr_dim_index = -1;
    unsigned int att_key;
    unsigned int count, ndim;
    unsigned int type = K_USHORT;
    array_pointer arrayp;
    multi_array *multi_desc;
    array_desc *arr_desc;
    char *p1;
    char *data_name = "Data Value";
    char line[STRING_LENGTH];
    uaddr raw_dim_lengths[MAX_DIMENSIONS], dim_lengths[MAX_DIMENSIONS];
    double crval[MAX_DIMENSIONS], crpix[MAX_DIMENSIONS], cdelt[MAX_DIMENSIONS];
    char ctype[MAX_DIMENSIONS][STRING_LENGTH];
    char *dim_names[MAX_DIMENSIONS];
    static flag first_time = TRUE;
    static char function_name[] = "foreign_idf_read_header";

    va_start (argp, sanitise);
    if (first_time)
    {
	if (r_getenv ("FOREIGN_IDF_DEBUG") != NULL) verbose = TRUE;
	first_time = FALSE;
    }
    if (channel == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    FLAG_VERIFY (data_alloc);
    if (sizeof (short) != 2)
    {
	a_func_abort (function_name, "short integers are not 16 bit");
	return (NULL);
    }
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp, unsigned int) )
	    != FA_IDF_READ_HEADER_END )
    {
	switch (att_key)
	{
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    if ( !chs_get_line (channel, line, STRING_LENGTH) )
    {
	fprintf (stderr, "%s: error reading\t%s\n", function_name, ERRSTRING);
	return (NULL);
    }
    if (strcmp (line, "IMAGE DESCRIPTOR FILE version 8") != 0)
    {
	fprintf (stderr, "%s: bad first line\n", function_name);
	return (NULL);
    }
    m_clear (raw_dim_lengths, sizeof raw_dim_lengths);
    while ( chs_get_line (channel, line, STRING_LENGTH) )
    {
	if (strncmp (line, "dimension:", 10) == 0)
	{
	    curr_dim_index = ex_int (line + 10, (CONST char **) &p1);
	}
	else if (strncmp (line, "npix:", 4) == 0)
	{
	    char *p2;

	    raw_dim_lengths[curr_dim_index] = ex_int (line + 4,
						      (CONST char **) &p1);
	    crval[curr_dim_index] = TOOBIG;
	    crpix[curr_dim_index] = TOOBIG;
	    cdelt[curr_dim_index] = TOOBIG;
	    ctype[curr_dim_index][0] = '\0';
	    if ( ( p1 = strstr (p1, "center(") ) == NULL ) continue;
	    if ( ( p2 = strchr (p1, ')') ) == NULL ) continue;
	    *p2 = '\0';
	    sprintf (ctype[curr_dim_index], "Dim %u (%s)",
		     curr_dim_index, p1 + 7);
	    crval[curr_dim_index] = ex_float (p2 + 2, (CONST char **) &p1);
	    if ( ( p2 = strchr (p1, ':') ) == NULL ) continue;
	    cdelt[curr_dim_index] = ex_float (p2 + 1, (CONST char **) &p1);
	    crpix[curr_dim_index] = raw_dim_lengths[curr_dim_index] * 0.5;
	}
    }
    for (count = 0, ndim = 0; count < MAX_DIMENSIONS; ++count)
    {
	if (raw_dim_lengths[count] < 1) continue;
	if (sanitise && raw_dim_lengths[count] < 2) continue;
	++ndim;
    }
    for (count = 0, curr_dim_index = ndim - 1; count < MAX_DIMENSIONS; ++count)
    {
	if (raw_dim_lengths[count] < 1) continue;
	if (sanitise && raw_dim_lengths[count] < 2) continue;
	dim_lengths[curr_dim_index] = raw_dim_lengths[count];
	dim_names[curr_dim_index] = ctype[count];
	--curr_dim_index;
    }
    if ( ( arr_desc =
	   ds_easy_alloc_array_desc (ndim, dim_lengths, NULL, NULL, NULL,
				     (CONST char **) dim_names,
				     1, &type, (CONST char **) &data_name) )
	 == NULL )
    {
	m_error_notify (function_name, "array descriptor");
	return (NULL);
    }
    m_clear (&arrayp, sizeof arrayp);
    if (data_alloc)
    {
	/*  Allocate the array manaully so that clearing is not done  */
	if ( !ds_alloc_vm_array (&arrayp, arr_desc, FALSE, FALSE) &&
	     !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp, arr_desc, FALSE,
				   FALSE) )
	{
	    ds_dealloc_array_desc (arr_desc);
	    fprintf (stderr, "%s: error allocating %lu bytes for array data\n",
		     function_name,
		     ds_get_array_size (arr_desc) *
		     ds_get_packet_size (arr_desc->packet));
	    return (NULL);
	}
    }
    if ( ( multi_desc =
	   ds_easy_alloc_array_from_array_desc (arr_desc, &arrayp, TRUE) )
	 == NULL )
    {
	ds_dealloc_array_desc (arr_desc);
	if (arrayp.free != NULL) (*arrayp.free) (arrayp.info, &arrayp);
	m_error_notify (function_name, "multi_array descriptor");
	return (NULL);
    }
    for (count = 0; count < MAX_DIMENSIONS; ++count)
    {
	if (raw_dim_lengths[count] < 1) continue;
	if ( DOUBLE_IS_BLANK (cdelt[count]) ) continue;
	sprintf (line, "CTYPE%u", count);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    line, ctype[count], FALSE);
	sprintf (line, "CRVAL%u", count);
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   line, K_DOUBLE, crval + count, FALSE);
	sprintf (line, "CRPIX%u", count);
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   line, K_DOUBLE, crpix + count, FALSE);
	sprintf (line, "CDELT%u", count);
	ds_put_unique_named_value (multi_desc->headers[0], multi_desc->data,
				   line, K_DOUBLE, cdelt + count, FALSE);
    }
    return (multi_desc);
}   /*  End Function foreign_idf_read_header  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_idf_read_data (Channel channel, multi_array *multi_desc,
			    char *data, uaddr num_values, ...)
/*  [SUMMARY] Read data in an IDF image file.
    [PURPOSE] This routine will read the data of an IDF image file from a
    channel.
    <channel> The channel to read from.
    <multi_desc> The Karma data structure to write the data into.
    <data> An alternate data array to write the FITS data into. If this is
    NULL, the routine will write the data into the Karma data structure.
    <num_values> The number of values to write into the data array. This is
    used when data is not NULL or skipping is requested.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_IDF_READ_DATA_END. See [<foreign_ATT_IDF_READ_DATA>] for a list of
    defined attributes.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    va_list argp;
    flag skip = FALSE;
    unsigned int att_key;
    unsigned long toobig_count;
    char *packet;
    unsigned long *blank_count = NULL;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    static char function_name[] = "foreign_idf_read_data";

    va_start (argp, num_values);
    if ( (channel == NULL) || (multi_desc == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp,unsigned int) ) !=
	    FA_IDF_READ_DATA_END )
    {
	switch (att_key)
	{
	  case FA_IDF_READ_DATA_NUM_BLANKS:
	    blank_count = va_arg (argp, unsigned long *);
	    break;
	  case FA_IDF_READ_DATA_SKIP_BLOCKS:
	    skip = va_arg (argp, flag);
	    FLAG_VERIFY (skip);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0];
    if (pack_desc->element_types[0] != K_ARRAY)
    {
	fprintf (stderr,"First element in top level packet must be K_ARRAY\n");
	a_prog_bug (function_name);
    }
    arr_desc = (array_desc *) pack_desc->element_desc[0];
    if (skip)
    {
	if ( !ch_skip (channel,
		       ds_get_packet_size (arr_desc->packet) * num_values) )
	{
	    fprintf (stderr, "Error skipping in IDF file\n");
	    return (FALSE);
	}
	return (TRUE);
    }
    if (data == NULL)
    {
	data = *(char **) packet;
	if (data == NULL)
	{
	    fprintf (stderr, "No array to write data into!\n");
	    a_prog_bug (function_name);
	}
	num_values = ds_get_array_size (arr_desc);
    }
    toobig_count = 0;
    if ( !dsrw_read_packets (channel, arr_desc->packet, data, num_values) )
    {
	fprintf (stderr, "Error reading IDF data\t%s\n", ERRSTRING);
	return (FALSE);
    }
    if (blank_count != NULL) *blank_count = toobig_count;
    return (TRUE);
}   /*  End Function foreign_idf_read_data  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_idf_read (CONST char *filename, flag sanitise, ...)
/*  [SUMMARY] Read an IDF file set.
    <filename> The name of the ".idf" header file in the IDF file set.
    <sanitise> If TRUE, IDF axes with length 1 are ignored. This is highly
    recommended.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with
    FA_IDF_READ_END. See [<foreign_ATT_IDF_READ>] for a list of defined
    attributes.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    va_list argp;
    Channel channel;
    unsigned int att_key;
    unsigned long blank_count_local;
    unsigned long *blank_count = NULL;
    multi_array *multi_desc;
    char *ptr;
    char image_name[STRING_LENGTH];
    static char function_name[] = "foreign_idf_read";

    va_start (argp, sanitise);
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp, unsigned int) ) != FA_IDF_READ_END )
    {
	switch (att_key)
	{
	  case FA_IDF_READ_NUM_BLANKS:
	    blank_count = va_arg (argp, unsigned long *);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    if (blank_count == NULL) blank_count = &blank_count_local;
    if ( ( ptr = strrchr (filename, '.') ) == NULL ) return (FALSE);
    if (strcmp (ptr + 1, "idf") != 0) return (FALSE);
    strncpy (image_name, filename, ptr - filename + 1);
    ptr = image_name + (ptr - filename);
    strcpy (ptr + 1, "int2");
    if ( ( channel = ch_open_file (filename, "r") ) == NULL )
    {
	fprintf (stderr, "Error opening: \"%s\"\t%s\n", filename, ERRSTRING);
	return (NULL);
    }
    multi_desc = foreign_idf_read_header (channel, TRUE, sanitise,
					  FA_IDF_READ_HEADER_END);
    ch_close (channel);
    if (multi_desc == NULL) return (NULL);
    if ( ( channel = ch_open_file (image_name, "r") ) == NULL )
    {
	fprintf (stderr, "Error opening: \"%s\"\t%s\n", image_name, ERRSTRING);
	ds_dealloc_multi (multi_desc);
	return (NULL);
    }
    if ( !foreign_idf_read_data (channel, multi_desc, NULL, 0,
				 FA_IDF_READ_DATA_NUM_BLANKS, blank_count,
				 FA_IDF_READ_DATA_END) )
    {
	ds_dealloc_multi (multi_desc);
	return (NULL);
    }
    return (multi_desc);
}   /*  End Function foreign_idf_read  */


/*  Private functions follow  */
