/*LINTLIBRARY*/
/*  misc.c

    This code provides miscellaneous functions for the <foreign> package.

    Copyright (C) 1995-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for the <foreign> package.


    Written by      Richard Gooch   20-APR-1995

    Updated by      Richard Gooch   20-APR-1995

    Updated by      Richard Gooch   6-MAY-1995: Added  #include <karma_st.h>

    Updated by      Richard Gooch   21-MAY-1995: Added support for Sun
  rasterfile format.

    Updated by      Richard Gooch   28-SEP-1995: Added support for Miriad Image
  format.

    Updated by      Richard Gooch   10-JAN-1996: Recognise *.mt files as FITS
  files (blunder in GIPSY which thinks all FITS is on tape).

    Updated by      Richard Gooch   12-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   21-JUN-1996: Added support for GIPSY files.

    Updated by      Richard Gooch   3-DEC-1996: Added support for PGM files.

    Updated by      Richard Gooch   14-MAY-1997: Supported compressed Karma and
  FITS files.

    Updated by      Richard Gooch   16-MAY-1997: Supported PNM files.

    Updated by      Richard Gooch   14-JUL-1997: Supported DRAO datasets.

    Updated by      Richard Gooch   29-JUL-1997: Supported filtered files.

    Updated by      Richard Gooch   30-JUL-1997: Supported AIPS datasets.

    Updated by      Richard Gooch   23-OCT-1997: Added test for *.fit.gz files.

    Updated by      Richard Gooch   5-DEC-1997: Added test AIPS++ Images.

    Updated by      Richard Gooch   12-MAR-1998: Added fallback test for .kf
  files when extension not given.

    Updated by      Richard Gooch   13-APR-1998: Added test for *.kf.bz2,
  *.fit.bz2 and *.fits.bz2 files.

    Updated by      Richard Gooch   5-MAY-1998: Moved support for generic PNM
  files elsewhere.

    Updated by      Richard Gooch   27-MAY-1998: Added test for .efits files.

    Updated by      Richard Gooch   30-JAN-1999: Added support for split Karma
  files (".desc" and ".dat" pairs).

    Updated by      Richard Gooch   29-JUN-1999: Added test for .sfits files.

    Last updated by Richard Gooch   30-NOV-2000: Added test for .idf files.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_dsxfr.h>
#include <karma_st.h>


/*  Private functions  */


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
unsigned int foreign_guess_format_from_filename (CONST char *filename)
/*  [SUMMARY] Attempt to guess the format of a file by examining its filename.
    <filename> The name of the file.
    [RETURNS] A value indicating the format of the file. The value
    FOREIGN_FILE_FORMAT_KARMA is returned if the extension is ".kf". See
    [<foreign_TYPES>] for a list of possible values.
*/
{
    flag is_dir, exists;
    int len;
    struct stat statbuf;
    CONST char *end;
    char fname[STRING_LENGTH];

    len = strlen (filename);
    if (strcmp (filename, "connection") == 0) return FOREIGN_FILE_FORMAT_KARMA;
    end = filename + len;
    if (strcmp  (end - 3, ".kf")   == 0) return (FOREIGN_FILE_FORMAT_KARMA);
    if (st_icmp (end - 3, ".mt")  == 0) return (FOREIGN_FILE_FORMAT_FITS);
    if (strcmp  (end - 4, ".pgm")  == 0) return (FOREIGN_FILE_FORMAT_PGM);
    if (strcmp  (end - 4, ".ppm")  == 0) return (FOREIGN_FILE_FORMAT_PPM);
    if (strcmp  (end - 4, ".pnm")  == 0) return (FOREIGN_FILE_FORMAT_PNM);
    if (st_icmp (end - 4, ".fts")  == 0) return (FOREIGN_FILE_FORMAT_FITS);
    if (st_icmp (end - 4, ".fit")  == 0) return (FOREIGN_FILE_FORMAT_FITS);
    if (strcmp  (end - 4, ".ras")  == 0) return (FOREIGN_FILE_FORMAT_SUNRAS);
    if (strcmp  (end - 4, ".idf")  == 0) return (FOREIGN_FILE_FORMAT_IDF);
    if (st_icmp (end - 5, ".fits") == 0) return (FOREIGN_FILE_FORMAT_FITS);
    if (strcmp  (end - 5, ".drao")  == 0) return (FOREIGN_FILE_FORMAT_DRAO);
    if (strcmp  (end - 5, ".aips")  == 0) return (FOREIGN_FILE_FORMAT_AIPS1);
    if (strcmp  (end - 6, ".kf.gz")   == 0) return (FOREIGN_FILE_FORMAT_KARMA);
    if (strcmp  (end - 6, ".efits")   == 0) return
						FOREIGN_FILE_FORMAT_EMPTY_FITS;
    if (strcmp  (end - 6, ".sfits")   == 0)
	return FOREIGN_FILE_FORMAT_SIMPLE_FITS;
    if (strcmp  (end - 7, ".kf.bz2")   == 0) return FOREIGN_FILE_FORMAT_KARMA;
    if (st_icmp (end - 7, ".fit.gz") == 0) return (FOREIGN_FILE_FORMAT_FITS);
    if (st_icmp (end - 8, ".fit.bz2") == 0) return (FOREIGN_FILE_FORMAT_FITS);
    if (st_icmp (end - 8, ".fits.gz") == 0) return (FOREIGN_FILE_FORMAT_FITS);
    if (st_icmp (end - 9, ".fits.bz2") == 0) return (FOREIGN_FILE_FORMAT_FITS);
    if ( dsxfr_test_split (filename) ) return (FOREIGN_FILE_FORMAT_KARMA);
    if ( foreign_gipsy_test (filename) ) return (FOREIGN_FILE_FORMAT_GIPSY);
    /*  All possible directory dataset formats follow  */
    if (stat (filename, &statbuf) == 0)
    {
	exists = TRUE;
	is_dir = S_ISDIR (statbuf.st_mode) ? TRUE : FALSE;
    }
    else
    {
	exists = FALSE;
	is_dir = FALSE;
	if (errno != ENOENT)
	{
	    fprintf (stderr, "Error statting file: \"%s\"\t%s\n",
		     filename, ERRSTRING);
	    return (FOREIGN_FILE_FORMAT_UNKNOWN);
	}
    }
    if (is_dir)
    {
	if ( foreign_miriad_test2 (filename, FALSE) )
	    return (FOREIGN_FILE_FORMAT_MIRIAD);
	if ( foreign_aips2_test (filename, FALSE) )
	    return (FOREIGN_FILE_FORMAT_FILTER_FITS);
    }
    if (exists)
    {
	return foreign_filter_get_format (filename);
    }
    /*  Nothing else seems to work. Since the filename doesn't exist, try
	adding ".kf" extension and see if that exists  */
    if (len + 4 > STRING_LENGTH) return (FOREIGN_FILE_FORMAT_UNKNOWN);
    strcpy (fname, filename);
    strcpy (fname + len, ".kf");
    if (stat (fname, &statbuf) == 0)
    {
	if ( S_ISREG (statbuf.st_mode) ) return (FOREIGN_FILE_FORMAT_KARMA);
    }
    if ( dsxfr_test_split (fname) ) return (FOREIGN_FILE_FORMAT_KARMA);
    return (FOREIGN_FILE_FORMAT_UNKNOWN);
}   /*  End Function foreign_guess_format_from_filename  */


/*  Private functions follow  */
