/*LINTLIBRARY*/
/*  ppm_write.c

    This code provides a PPM write facility.

    Copyright (C) 1995-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for writing colour images
    in PPM format.


    Written by      Richard Gooch   15-APR-1995

    Updated by      Richard Gooch   16-APR-1995

    Updated by      Richard Gooch   12-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   4-JUN-1996: Created experimental functions
  <foreign_ppm_write_pseudo> and <foreign_ppm_write_rgb>.

    Updated by      Richard Gooch   17-SEP-1996: Fixed bug when writing binary
  PPM files: maxval was not immediately followed by a whitespace and then image
  data.

    Updated by      Richard Gooch   19-JUN-1997: Support PseudoColour images
  with colourmap in <foreign_ppm_write>.

    Updated by      Richard Gooch   18-FEB-1998: Added comment in PPM file
  to say what module produced the file.

    Updated by      Richard Gooch   5-MAY-1998: Added minimum image size
  parameters with blank-filling.

    Updated by      Richard Gooch   7-MAY-1998: Changed minimum image size
  parameters to output image size parameters, which can also truncate.

    Updated by      Richard Gooch   11-JUL-1998: Fixed bugs when output image
  size given as 0: did not default to input image size.

    Updated by      Richard Gooch   15-JUN-1999: Added intensity scaling
  parameters to <foreign_ppm_write_pseudo>.

    Updated by      Richard Gooch   12-NOV-1999: Cope with NaNs in function
  <foreign_ppm_write_pseudo>.

    Last updated by Richard Gooch   11-MAR-2001: Cope with min>max after iscale
  transform.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>


/*  Private functions  */
STATIC_FUNCTION (double *alloc_values_buffer, (unsigned int num_values) );
STATIC_FUNCTION (flag _foreign_ppm_write_blanks,
		 (Channel channel,unsigned int num_blanks, flag binary) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
flag foreign_ppm_write (Channel channel, multi_array *multi_desc, flag binary,
			...)
/*  [SUMMARY] Write a colour image to a channel in PPM format
    <channel> The channel to write to. The channel is not flushed.
    <multi_desc> The multi_array descriptor pointer. The routine will find a
    TrueColour image or a PseudoColour image within the data structure.
    <binary> If TRUE, the pixels will be written in binary mode.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. This list must be terminated with FA_PPM_WRITE_END.
    See [<foreign_ATT_PPM_WRITE>] for a list of defined attributes.
    [RETURNS] TRUE on succes, else FALSE.
*/
{
    iarray image_pseudo;
    iarray image_red, image_green, image_blue;
    flag ok;
    int width, height;
    unsigned int cmap_index, cmap_size;
    unsigned int att_key;
    double min, max;
    CONST unsigned short *cmap;
    va_list argp;
    static char function_name[] = "foreign_ppm_write";

    va_start (argp, binary);
    FLAG_VERIFY (binary);
    if ( ( channel == NULL) || (multi_desc == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    /*  Process attributes  */
    while ( ( att_key = va_arg (argp, unsigned int) ) != FA_PPM_WRITE_END )
    {
	switch (att_key)
	{
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    if ( !iarray_get_image_from_multi (multi_desc, &image_pseudo,
				       &image_red, &image_green, &image_blue,
				       &cmap_index) )
    {
	return (FALSE);
    }
    /*  Sanity checks  */
    if (image_red != NULL)
    {
	/*  Have RGB image  */
	if ( (iarray_type (image_red) != K_UBYTE) ||
	     (iarray_type (image_green) != K_UBYTE) ||
	     (iarray_type (image_blue) != K_UBYTE) )
	{
	    fprintf (stderr,"TrueColour image data must be of type K_UBYTE\n");
	    iarray_dealloc (image_red);
	    iarray_dealloc (image_green);
	    iarray_dealloc (image_blue);
	    return (FALSE);
	}
	width = iarray_dim_length (image_red, 1);
	height = iarray_dim_length (image_red, 0);
	ok = foreign_ppm_write_rgb (channel, binary,
				    (CONST unsigned char *) image_red->data,
				    (CONST unsigned char *) image_green->data,
				    (CONST unsigned char *) image_blue->data,
				    image_red->offsets[1],
				    image_red->offsets[0],
				    width, height, 0, 0, NULL, NULL, NULL, 0);
	iarray_dealloc (image_red);
	iarray_dealloc (image_green);
	iarray_dealloc (image_blue);
	return (ok);
    }
    if (cmap_index >= multi_desc->num_arrays)
    {
	a_func_abort (function_name,  "Greyscale images not supported");
	iarray_dealloc (image_pseudo);
	return (FALSE);
    }
    if ( ( cmap = ds_cmap_find_colourmap (multi_desc->headers[cmap_index],
					  multi_desc->data[cmap_index],
					  &cmap_size, NULL, NULL, NULL, 0) )
	 == NULL )
    {
	a_func_abort (function_name,  "colourmap not found");
	iarray_dealloc (image_pseudo);
	return (FALSE);
    }
    iarray_min_max (image_pseudo, CONV_CtoR_REAL, &min, &max);
    ok = foreign_ppm_write_pseudo (channel, binary,
				   (CONST char *) image_pseudo->data,
				   iarray_type (image_pseudo),
				   image_pseudo->offsets[1],
				   image_pseudo->offsets[0],
				   iarray_dim_length (image_pseudo, 1),
				   iarray_dim_length (image_pseudo, 0),
				   0, 0,
				   cmap, cmap + 1, cmap + 2, cmap_size, 3,
				   min, max, NULL, NULL);
    iarray_dealloc (image_pseudo);
    return (ok);
}   /*  End Function foreign_ppm_write  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_ppm_write_pseudo (Channel channel, flag binary,
			       CONST char *image, unsigned int type,
			       uaddr *hoffsets, uaddr *voffsets,
			       unsigned int inp_width, unsigned int inp_height,
			       unsigned int out_width, unsigned int out_height,
			       CONST unsigned short *cmap_reds,
			       CONST unsigned short *cmap_greens,
			       CONST unsigned short *cmap_blues,
			       unsigned int cmap_size,unsigned int cmap_stride,
			       double i_min, double i_max,
			       flag (*iscale_func) (), void *iscale_info)
/*  [SUMMARY] Write a PseudoColor image to a channel in PPM format.
    <channel> The channel to write to. The channel is not flushed.
    <binary> If TRUE, the pixels will be written in binary mode.
    <image> The image data.
    <type> The type of the image data.
    <hoffsets> The array of horizontal byte offsets.
    <voffsets> The array of vertical byte offsets.
    <inp_width> The width of the input image.
    <inp_height> The height of the input image.
    <out_width> The width of the output image. If the input image is
    smaller than this, then the image will be centred with blank-filling. If
    the input image is larger than this, the image will be truncated. If this
    is 0, the output image will be the same width as the input image.
    <out_height> The height of the output image. If the input image is smaller
    than this, then the image will be centred with blank-filling. If the input
    image is larger than this, the image will be truncated. If this is 0, the
    output image will be the same height as the input image.
    <cmap_reds> The red colourmap values.
    <cmap_greens> The green colourmap values.
    <cmap_blues> The blue colourmap values.
    <cmap_size> The number of colourmap entries.
    <cmap_stride> The stride (in unsigned shorts) between colourmap values.
    <i_min> The minimum image value. Image values below this will be clipped.
    <i_max> The maximum image value. Image values above this will be clipped.
    <iscale_func> The function to be called when non-linear intensity scaling
    is required. If NULL, linear intensity scaling is used. The prototype
    function is [<imw_PROTO_iscale_func>].
    <iscale_info> A pointer to arbitrary information for <<iscale_func>>.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on succes, else FALSE.
*/
{
    flag complex;
    int ival;
    int hcount, vcount;
    unsigned int left_pad, right_pad, bottom_pad, top_pad;
    unsigned int use_width, use_height;
    uaddr hoff, voff;
    double d_mul, d_data, scaled_min, scaled_max, tmp;
    CONST char *inp_line;
    double *values, *val_ptr;
    unsigned char pixel[3];
    extern char module_lib_version[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    static char function_name[] = "foreign_ppm_write_pseudo";

    if ( (hoffsets == NULL) || (voffsets == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if (cmap_size > 256)
    {
	fprintf (stderr, "Too many colour values: %u\n", cmap_size);
	a_prog_bug (function_name);
    }
    if (i_min >= i_max)
    {
	fprintf (stderr, "i_max: %e  is not greater than i_min: %e\n",
		 i_max, i_min);
	a_prog_bug (function_name);
    }
    if ( !ds_element_is_atomic (type) )
    {
	fprintf (stderr, "Input image must be atomic\n");
	a_prog_bug (function_name);
    }
    if ( ds_element_is_complex (type) )
    {
	fprintf (stderr, "Input image must not be complex\n");
	a_prog_bug (function_name);
    }
    if (out_width == 0) out_width = inp_width;
    if (out_height == 0) out_height = inp_height;
    if (out_width < inp_width)
    {
	left_pad = 0;
	right_pad = 0;
	hoff = (inp_width - out_width) / 2;
	use_width = out_width;
    }
    else if (out_width > inp_width)
    {
	left_pad = (out_width - inp_width) / 2;
	right_pad = out_width - inp_width - left_pad;
	hoff = 0;
	use_width = inp_width;
    }
    else
    {
	left_pad = 0;
	right_pad = 0;
	hoff = 0;
	use_width = inp_width;
    }
    if (out_height < inp_height)
    {
	bottom_pad = 0;
	top_pad = 0;
	voff = (inp_height - out_height) / 2;
	use_height = out_height;
    }
    else if (out_height > inp_height)
    {
	bottom_pad = (out_height - inp_height) / 2;
	top_pad = out_height - inp_height - bottom_pad;
	voff = 0;
	use_height = inp_height;
    }
    else
    {
	bottom_pad = 0;
	top_pad = 0;
	voff = 0;
	use_height = inp_height;
    }
    /*  Allocate values buffer  */
    if ( ( values = alloc_values_buffer (use_width) ) == NULL ) return (FALSE);
    if (binary)
    {
	if ( !ch_puts (channel, "P6\n# Binary", FALSE) ) return (FALSE);
    }
    else
    {
	if ( !ch_puts (channel, "P3\n# ASCII", FALSE) ) return (FALSE);
    }
    if ( !ch_printf (channel, " PPM file written by <%s>\n",
		     function_name) ) return (FALSE);
    if ( !ch_printf (channel, "# for module: \"%s\" version: \"%s\"\n",
		     module_name, module_version_date) ) return (FALSE);
    if ( !ch_printf (channel, "# Karma library version: %s\n",
		     karma_library_version) ) return (FALSE);
    if ( !ch_printf (channel, "# Module compiled with library version: %s\n",
		     module_lib_version) ) return (FALSE);
    if ( !ch_printf (channel, "%u %u # width height\n",
		     out_width, out_height) ) return (FALSE);
    if ( !ch_printf (channel,
		     "# max value follows, then comes the image data\n") )
	return (FALSE);
    if ( !ch_printf (channel, "255\n") ) return (FALSE);
    /*  First do the top pad  */
    if ( !_foreign_ppm_write_blanks (channel, top_pad * out_width, binary) )
    {
	fprintf (stderr, "%s: error writing top blanks\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
    if (iscale_func == NULL)
    {
	scaled_min = i_min;
	scaled_max = i_max;
    }
    else
    {
	if ( !(*iscale_func) (&scaled_min, 0, &i_min, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if ( !(*iscale_func) (&scaled_max, 0, &i_max, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity maximum\n",
		     function_name);
	}
	if (scaled_max < scaled_min) swap (scaled_min, scaled_max, tmp);
    }
    /*  Loop through the image lines  */
    if (cmap_size > 2) d_mul = (cmap_size - 1) / (scaled_max - scaled_min);
    else d_mul = 255.0 / (scaled_max - scaled_min);
    for (vcount = use_height - 1; vcount >= 0; --vcount)
    {
	/*  First do the left pad  */
	if ( !_foreign_ppm_write_blanks (channel, left_pad, binary) )
	{
	    fprintf (stderr, "%s: error writing left blanks\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	inp_line = image + voffsets[vcount + voff];
	/*  inp_line  may be modified now  */
	/*  Now have pointer to input image line  */
	/*  Convert data to generic data type  */
	if ( !ds_get_scattered_elements (inp_line, type, hoffsets + hoff,
					 values, &complex, use_width) )
	{
	    fprintf (stderr, "Error converting data\n");
	    return (FALSE);
	}
	if (iscale_func != NULL)
	{
	    if ( !(*iscale_func) (values, 2, values, 2, use_width,
				  i_min, i_max, iscale_info) )
	    {
		fprintf (stderr, "Error applying intensity scale\n");
		return (FALSE);
	    }
	}
	/*  Loop for each value  */
	for (hcount = 0, val_ptr = values; hcount < use_width;
	     ++hcount, val_ptr += 2)
	{
	    if ( DOUBLE_IS_BLANK (d_data = *val_ptr) ) ival = cmap_size - 1;
	    else if ( (d_data = *val_ptr) < scaled_min ) ival = 0;
	    else if (d_data > scaled_max) ival = cmap_size - 1;
	    else ival = (int) ( (d_data - scaled_min) * d_mul + 0.5 );
	    if (cmap_size > 2)
	    {
		pixel[0] = (cmap_reds[ival * cmap_stride] >> 8) & 0xff;
		pixel[1] = (cmap_greens[ival * cmap_stride] >> 8) & 0xff;
		pixel[2] = (cmap_blues[ival * cmap_stride] >> 8) & 0xff;
	    }
	    else
	    {
		pixel[0] = ival;
		pixel[1] = ival;
		pixel[2] = ival;
	    }
	    if (binary)
	    {
		if (ch_write (channel, (char *) pixel, 3) < 3) return (FALSE);
	    }
	    else
	    {
		if ( !ch_printf (channel, "%u %u %u\n",
				 pixel[0], pixel[1],pixel[2]) ) return (FALSE);
	    }
	}
	/*  Finish off with the right pad  */
	if ( !_foreign_ppm_write_blanks (channel, right_pad, binary) )
	{
	    fprintf (stderr, "%s: error writing right blanks\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
    }
    /*  Finish off with the bottom pad  */
    if ( !_foreign_ppm_write_blanks (channel, bottom_pad * out_width, binary) )
    {
	fprintf (stderr, "%s: error writing bottom blanks\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function foreign_ppm_write_pseudo  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_ppm_write_rgb (Channel channel, flag binary,
			    CONST unsigned char *image_red,
			    CONST unsigned char *image_green,
			    CONST unsigned char *image_blue,
			    uaddr *hoffsets, uaddr *voffsets,
			    unsigned int inp_width, unsigned int inp_height,
			    unsigned int out_width, unsigned int out_height,
			    CONST unsigned short *cmap_red,
			    CONST unsigned short *cmap_green,
			    CONST unsigned short *cmap_blue,
			    unsigned int cmap_stride)
/*  [SUMMARY] Write a TrueColor image to a channel in Sun rasterfile format
    <channel> The channel to write to. The channel is not flushed.
    <binary> If TRUE, the pixels will be written in binary mode.
    <red_image> The red image data.
    <green_image> The green image data.
    <blue_image> The blue image data.
    <hoffsets> The array of horizontal byte offsets.
    <voffsets> The array of vertical byte offsets.
    <inp_width> The width of the input image.
    <inp_height> The height of the input image.
    <out_width> The width of the output image. If the input image is
    smaller than this, then the image will be centred with blank-filling. If
    the input image is larger than this, the image will be truncated. If this
    is 0, the output image will be the same width as the input image.
    <out_height> The height of the output image. If the input image is smaller
    than this, then the image will be centred with blank-filling. If the input
    image is larger than this, the image will be truncated. If this is 0, the
    output image will be the same height as the input image.
    <cmap_red> The red component colourmap entries. 256 entries required. If
    this is NULL a linear mapping is assumed.
    <cmap_green> The green component colourmap entries. 256 entries required.
    If this is NULL a linear mapping is assumed.
    <cmap_blue> The blue component colourmap entries. 256 entries required. If
    this is NULL a linear mapping is assumed.
    <cmap_stride> The stride (in unsigned shorts) between colourmap values.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on succes, else FALSE.
*/
{
    int hcount, vcount;
    int red, green, blue;
    unsigned int left_pad, right_pad, bottom_pad, top_pad;
    unsigned int use_width, use_height;
    uaddr hoff, voff;
    uaddr voffset;
    unsigned char pixel[3];
    extern char module_lib_version[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    static char function_name[] = "foreign_ppm_write_rgb";

    if ( (hoffsets == NULL) || (voffsets == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if (out_width == 0) out_width = inp_width;
    if (out_height == 0) out_height = inp_height;
    if (out_width < inp_width)
    {
	left_pad = 0;
	right_pad = 0;
	hoff = (inp_width - out_width) / 2;
	use_width = out_width;
    }
    else if (out_width > inp_width)
    {
	left_pad = (out_width - inp_width) / 2;
	right_pad = out_width - inp_width - left_pad;
	hoff = 0;
	use_width = inp_width;
    }
    else
    {
	left_pad = 0;
	right_pad = 0;
	hoff = 0;
	use_width = inp_width;
    }
    if (out_height < inp_height)
    {
	bottom_pad = 0;
	top_pad = 0;
	voff = (inp_height - out_height) / 2;
	use_height = out_height;
    }
    else if (out_height > inp_height)
    {
	bottom_pad = (out_height - inp_height) / 2;
	top_pad = out_height - inp_height - bottom_pad;
	voff = 0;
	use_height = inp_height;
    }
    else
    {
	bottom_pad = 0;
	top_pad = 0;
	voff = 0;
	use_height = inp_height;
    }
    if (binary)
    {
	if ( !ch_puts (channel, "P6\n# Binary", FALSE) ) return (FALSE);
    }
    else
    {
	if ( !ch_puts (channel, "P3\n# ASCII", FALSE) ) return (FALSE);
    }
    if ( !ch_printf (channel, " PPM file written by <%s>\n",
		     function_name) ) return (FALSE);
    if ( !ch_printf (channel, "# for module: \"%s\" version: \"%s\"\n",
		     module_name, module_version_date) ) return (FALSE);
    if ( !ch_printf (channel, "# Karma library version: %s\n",
		     karma_library_version) ) return (FALSE);
    if ( !ch_printf (channel, "# Module compiled with library version: %s\n",
		     module_lib_version) ) return (FALSE);
    if ( !ch_printf (channel, "%u %u # width height\n",
		     out_width, out_height) ) return (FALSE);
    if ( !ch_printf (channel,
		     "# max value follows, then comes the image data\n") )
	return (FALSE);
    if ( !ch_printf (channel, "255\n") ) return (FALSE);
    /*  First do the top pad  */
    if ( !_foreign_ppm_write_blanks (channel, top_pad * out_width, binary) )
	return (FALSE);
    /*  Loop through the image lines  */
    for (vcount = use_height - 1; vcount >= 0; --vcount)
    {
	/*  First do the left pad  */
	if ( !_foreign_ppm_write_blanks (channel, left_pad, binary) )
	    return (FALSE);
	voffset = voffsets[vcount] + voff;
	for (hcount = 0; hcount < use_width; ++hcount)
	{
	    red = image_red[hoffsets[hcount + hoff] + voffset];
	    if (cmap_red != NULL) red = (cmap_red[red * cmap_stride] >>
					 8) & 0xff;
	    pixel[0] = red;
	    green = image_green[hoffsets[hcount + hoff] + voffset];
	    if (cmap_green != NULL) green = (cmap_green[green * cmap_stride] >>
					     8) & 0xff;
	    pixel[1] = green;
	    blue = image_blue[hoffsets[hcount + hoff] + voffset];
	    if (cmap_blue != NULL) blue = (cmap_blue[blue * cmap_stride] >>
					   8) & 0xff;
	    pixel[2] = blue;
	    if (binary)
	    {
		if (ch_write (channel, (char *) pixel, 3) < 3) return (FALSE);
	    }
	    else
	    {
		if ( !ch_printf (channel, "%u %u %u\n",
				 pixel[0], pixel[1],pixel[2]) ) return (FALSE);
	    }
	}
	/*  Finish off with the right pad  */
	if ( !_foreign_ppm_write_blanks (channel, right_pad, binary) )
	    return (FALSE);
    }
    /*  Finish off with the bottom pad  */
    if ( !_foreign_ppm_write_blanks (channel, bottom_pad * out_width, binary) )
	return (FALSE);
    return (TRUE);
}   /*  End Function foreign_ppm_write_rgb  */


/*  Private functions follow  */

static double *alloc_values_buffer (unsigned int num_values)
/*  This routine will allocate a buffer space for generic data values.
    The number of (DCOMPLEX) values to allocate must be given by  num_values  .
    The routine will return a pointer to the buffer space on success,
    else it returns NULL. The buffer is global and must NOT be deallocated.
*/
{
    static unsigned int value_buf_len = 0;
    static double *values = NULL;
    static char function_name[] = "alloc_values_buffer";

    /*  Make sure value buffer is big enough  */
    if (value_buf_len < num_values)
    {
	/*  Buffer too small  */
	if (values != NULL)
	{
	    m_free ( (char *) values );
	}
	value_buf_len = 0;
	if ( ( values = m_alloc (sizeof *values * 2 * num_values) ) == NULL )
	{
	    m_error_notify (function_name, "values buffer");
	    return (NULL);
	}
	value_buf_len = num_values;
    }
    return (values);
}   /*  End Function alloc_values_buffer  */

static flag _foreign_ppm_write_blanks (Channel channel,unsigned int num_blanks,
				       flag binary)
/*  [SUMMARY] Write blanks.
    <channel> The channel to write to.
    <num_blanks> The number of blanks to write.
    <binary> If TRUE the blanks will be written in binary format.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char line[] = "0 0 0\n";

    if (num_blanks < 1) return (TRUE);
    if (binary)
    {
	unsigned int num_char = num_blanks * 3;

	return (ch_fill (channel, num_char, 0) < num_char) ? FALSE : TRUE;
    }
    /*  Do it in ASCII  */
    while (num_blanks-- > 0)
	if (ch_write (channel, line, sizeof line) < sizeof line) return FALSE;
    return (TRUE);
}   /*  End Function _foreign_ppm_write_blanks  */
