/*LINTLIBRARY*/
/*  interface.c

    This file provides Glish interface routines.

    Copyright (C) 1996-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for interfacing to Glish.


    Written by      Richard Gooch   21-DEC-1996

    Updated by      Richard Gooch   27-DEC-1996

    Updated by      Richard Gooch   29-DEC-1996: Improved debugging messages.

    Updated by      Richard Gooch   5-JAN-1997: Created
  <gls_register_glish_event_handler> function.

    Updated by      Richard Gooch   24-JAN-1997: Created <gls_send_plain_event>
  function.

    Updated by      Richard Gooch   2-SEP-1997: Added verbose flag to
  initialise function.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <r> package.

    Updated by      Richard Gooch   10-MAR-1999: Cope with NULL return from
  <r_get_karmabase>.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   9-DEC-1999: Changed <<data>> pointer type
  for <gls_send_plain_event>.

    Last updated by Richard Gooch   18-FEB-2000: Disabled use of AIPSPATH.


*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <karma.h>
#include <os.h>
#ifdef HAS_DL  /*  Glish requires dlopen() and friends  */
#  include <dlfcn.h>
#endif
#include <karma_gls.h>
#include <karma_dm.h>
#include <karma_dl.h>
#include <karma_a.h>
#include <karma_c.h>
#include <karma_r.h>


/*  Private data  */
static flag initialised = FALSE;
static flag running_under_glish = FALSE;


/*  Private functions  */
STATIC_FUNCTION (flag initialise, (flag verbose) );


/*  Functions in the shared object  */


/*  Functions in the shared object  */
STATIC_FUNCTION (flag (*FPTR_gls_initialise), (int *argc, char **argv) );
STATIC_FUNCTION (KCallbackFunc (*FPTR_gls_register_glish_event_handler),
		 (flag (*func) (), CONST char *name, void *object) );
STATIC_FUNCTION (KCallbackFunc (*FPTR_gls_register_plain_event_handler),
		 (flag (*func) (), CONST char *name, void *object) );
STATIC_FUNCTION (void (*FPTR_gls_send_plain_event),
		 (CONST char *name, CONST void *data,
		  unsigned int type, unsigned int num_elements,
		  unsigned int num_axes, CONST int *axis_lengths) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
flag gls_test_available ()
/*  [SUMMARY] Test if Glish available.
    [RETURNS] TRUE if Glish available, else FALSE.
*/
{
    return initialise (FALSE);
}   /*  End Function gls_available  */

/*UNPUBLISHED_FUNCTION*/
flag gls_initialise (int *argc, char **argv)
/*  [SUMMARY] Initialise Glish support.
    <argc> As passed to the <<main>> function.
    <argv> As passed to the <<main>> function.
    [NOTE] This function must not be called until the [<dm>] package is
    initialised.
    [RETURNS] TRUE if running under Glish, else FALSE.
*/
{
    static char function_name[] = "gls_initialise";

    if (initialised)
    {
	fprintf (stderr, "Glish already initialised\n");
	a_prog_bug (function_name);
    }
    initialised = TRUE;
    if (*argc < 4) return (FALSE);
    if (strcmp (argv[1], "-id") != 0) return (FALSE);
    if ( !initialise (TRUE) )
    {
	fprintf (stderr,
		 "%s: found Glish command line but Glish not available\n",
		 function_name);
	return (FALSE);
    }
    running_under_glish = (*FPTR_gls_initialise) (argc, argv);
    return (running_under_glish);
}   /*  End Function gls_initialise  */

/*UNPUBLISHED_FUNCTION*/
KCallbackFunc gls_register_glish_event_handler (flag (*func) (),
						CONST char *name, void *object)
/*  [SUMMARY] Register a Glish event handler for Glish.
    <func> The function that is called.
    <name> The name of the Glish events the handler will receive. If this is
    NULL, all events will be received.
    <object> The object.
    [RETURNS] A KCallbackFunc object on success, else NULL indicating Glish is
    not available.
*/
{
    if ( !initialise (FALSE) ) return (NULL);
    /*  Don't bother if we know we are not running under Glish  */
    if (initialised && !running_under_glish) return (NULL);
    /*  Register the handler. Note that we might not be running under Glish
	anyway, so this just wastes a bit of memory  */
    return (*FPTR_gls_register_glish_event_handler) (func, name, object);
}   /*  End Function gls_register_glish_event_handler  */

/*UNPUBLISHED_FUNCTION*/
KCallbackFunc gls_register_plain_event_handler (flag (*func) (),
						CONST char *name, void *object)
/*  [SUMMARY] Register a plain event handler for Glish.
    <func> The function that is called.
    <name> The name of the Glish events the handler will receive. If this is
    NULL, all events will be received.
    <object> The object.
    [RETURNS] A KCallbackFunc object on success, else NULL indicating Glish is
    not available.
*/
{
    if ( !initialise (FALSE) ) return (NULL);
    /*  Don't bother if we know we are not running under Glish  */
    if (initialised && !running_under_glish) return (NULL);
    /*  Register the handler. Note that we might not be running under Glish
	anyway, so this just wastes a bit of memory  */
    return (*FPTR_gls_register_plain_event_handler) (func, name, object);
}   /*  End Function gls_register_plain_event_handler  */

/*UNPUBLISHED_FUNCTION*/
void gls_send_plain_event (CONST char *name, CONST void *data,
			   unsigned int type, unsigned int num_elements,
			   unsigned int num_axes, CONST int *axis_lengths)
/*  [SUMMARY] Send a plain Glish event.
    <name> The event name.
    <data> The data to send (this is a vector).
    <type> The type of the data.
    <num_elements> The number of elements in the vector.
    <num_axes> The number of axes. This may be 0.
    <axis_lengths> The length of each axis.
    [RETURNS] Nothing.
*/
{
    /*  First check if running under Glish: no point sending if not  */
    if (!running_under_glish) return;
    if ( !initialise (FALSE) ) return;
    (*FPTR_gls_send_plain_event) (name, data, type, num_elements,
				  num_axes, axis_lengths);
}   /*  End Function gls_send_plain_event  */


/*  Private functions follow  */

static flag initialise (flag verbose)
/*  [SUMMARY] Initialise. May be called many times.
    <verbose> If TRUE, verbose error messages are displayed.
    [RETURNS] TRUE if Glish is available, else FALSE.
*/
{
#ifdef HAS_DL
    void *handle;
    CONST char *aips2lib;
#  if 0
    FILE *fp;
    CONST char *ptr, *stop_ptr, *karmabase;
    char txt[STRING_LENGTH], dep_file[STRING_LENGTH];
    char aips2root[STRING_LENGTH];
#  endif
    char failed_object[STRING_LENGTH];
#endif
    static flag initialised = FALSE;
    static flag ok = FALSE;
    static char function_name[] = "_gls_initialise";

    if (initialised) return (ok);
    initialised = TRUE;
#ifdef HAS_DL
    /*  Load gls_main.so  */
#  if 0  /*  Disable AIPSPATH: it fouls people up who are developing AIPS++
	     applications with Karma and the latest versions of AIPS++, but
	     Karma wants a stable interface to Glish  */
    if ( ( ptr = r_getenv ("AIPSPATH") ) == NULL ) aips2lib = NULL;
    else
    {
	sscanf (ptr, "%s", aips2root);
	/*  Read first part of gls_main.libs file to determine the compiler
	    used  */
	if ( ( karmabase = r_get_karmabase () ) == NULL ) return (FALSE);
	sprintf (dep_file, "%s/lib/shared-objects/gls_main.libs", karmabase);
	if ( ( fp = fopen (dep_file, "r") ) == NULL ) return (FALSE);
	if (fgets (txt, STRING_LENGTH, fp) == NULL)
	{
	    fclose (fp);
	    return (FALSE);
	}
	fclose (fp);
	if ( ( ptr = strchr (txt, ':') ) == NULL ) ptr = txt + strlen (txt);
	stop_ptr = ptr;
	/*  Step before the beginning of the trailing "/lib"  */
	ptr -= 5;
	/*  Search back for the previous '/'  */
	while (*ptr != '/') --ptr;
	strncat (aips2root, ptr, stop_ptr - ptr);
	aips2lib = aips2root;
    }
#  else
    aips2lib = NULL;
#  endif
    if ( ( handle = dl_load_object ("gls_main", aips2lib, failed_object) )
	 == NULL )
    {
	if (verbose) fprintf (stderr, "%s: failed to load object: \"%s\"\n",
			      function_name, failed_object);
	return (FALSE);
    }
    if ( ( FPTR_gls_initialise = ( flag (*) () )
	   dlsym (handle, "DL_gls_initialise") ) == NULL )
    {
	fprintf ( stderr, "%s\n", dlerror () );
	return (FALSE);
    }
    if ( ( FPTR_gls_register_glish_event_handler = ( KCallbackFunc (*) () )
	   dlsym (handle, "DL_gls_register_glish_event_handler") ) == NULL )
    {
	fprintf ( stderr, "%s\n", dlerror () );
	return (FALSE);
    }
    if ( ( FPTR_gls_register_plain_event_handler = ( KCallbackFunc (*) () )
	   dlsym (handle, "DL_gls_register_plain_event_handler") ) == NULL )
    {
	fprintf ( stderr, "%s\n", dlerror () );
	return (FALSE);
    }
    if ( ( FPTR_gls_send_plain_event = ( void (*) () )
	   dlsym (handle, "DL_gls_send_plain_event") ) == NULL )
    {
	fprintf ( stderr, "%s\n", dlerror () );
	return (FALSE);
    }
    ok = TRUE;
#endif
    return (ok);
}   /*  End Function initialise  */
