/*LINTLIBRARY*/
/*  contour.c

    This code provides contour extraction from Intelligent Arrays.

    Copyright (C) 1996-1998  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the simple processing of
  n-dimensional data structures.


    Written by      Richard Gooch   18-JUL-1996

    Updated by      Richard Gooch   20-JUL-1996: Made use of <ds_contour>

    Updated by      Richard Gooch   7-JUL-1997: Changed interface to match the
  new interface for <ds_contour>.

    Updated by      Richard Gooch   24-FEB-1998: Changed to specific ds include
  files.

    Updated by      Richard Gooch   13-AUG-1998: Documentation fix.

    Last updated by Richard Gooch   12-SEP-1998: Introduced test for magic
  number.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_ds_func.h>
#include <karma_a.h>
#include <karma_m.h>


#define MAGIC_NUMBER 939032982  /*  The other files depend on this too!  */

#define VERIFY_IARRAY(array) if (array == NULL) \
{fprintf (stderr, "NULL iarray passed\n"); a_prog_bug (function_name); } \
if (array->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid iarray\n"); a_prog_bug (function_name); }


/*  Private functions  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void iarray_contour (iarray array, unsigned int num_levels,
		     CONST double *contour_levels,
		     uaddr *buffer_sizes, uaddr size_stride,
		     double **x0_arr, double **y0_arr,
		     double **x1_arr, double **y1_arr,
		     uaddr *num_segments, uaddr seg_stride)
/*  [SUMMARY] Extract contours from a 2-dimensional Intelligent Array.
    [PURPOSE] This routine will extract contours from a 2-dimensional
    Intelligent Array, producing a list of line segments that approximate the
    countours. The co-ordinates of the line segments are in linear world
    co-ordinates.
    <array> The array.
    <num_levels> The number of contour levels.
    <contour_levels> The array of contour levels.
    <buffer_sizes> A pointer to an array of co-ordinate buffer sizes. This is
    modified.
    <size_stride> The stride (in bytes) between consecutive elements in the
    co-ordinate buffer size array.
    <x0_arr> A pointer to an array of co-ordinate array pointers.
    The co-ordinate arrays may be internally reallocated, hence the array
    pointer may be modified.
    <y0_arr> A pointer to an array of co-ordinate array pointers. The
    co-ordinate arrays may be internally reallocated, hence the array pointer
    may be modified.
    <x1_arr> A pointer to an array of co-ordinate array pointers. The
    co-ordinate arrays may be internally reallocated, hence the array pointer
    may be modified.
    <y1_arr> A pointer to an array of co-ordinate array pointers. The
    co-ordinate arrays may be internally reallocated, hence the array pointer
    may be modified.
    <num_segments> A pointer to an array where the the number of segments per
    level will be written.
    <seg_stride> The stride (in bytes) between consecutive elements in the
    segment count array.
    [RETURNS] Nothing.
*/
{
    dim_desc *xdim, *ydim;
    static char function_name[] = "iarray_contour";

    VERIFY_IARRAY (array);
    if (iarray_type (array) != K_FLOAT)
    {
	fprintf (stderr, "Intelligent array is not floating point\n");
	a_prog_bug (function_name);
    }
    xdim = iarray_get_dim_desc (array, 1);
    ydim = iarray_get_dim_desc (array, 0);
    ds_contour (array->data, iarray_type (array),
		xdim, array->offsets[1], ydim, array->offsets[0],
		num_levels, contour_levels,
		buffer_sizes, size_stride,
		x0_arr, y0_arr, x1_arr, y1_arr, num_segments, seg_stride);
}   /*  End Function iarray_contour  */
