/*LINTLIBRARY*/
/*  misc.c

    This code provides miscellaneous Intelligent Array routines.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the simple processing of
  n-dimensional data structures.


    Written by      Richard Gooch   28-FEB-1996

    Updated by      Richard Gooch   1-APR-1996: Moved remaing functions to new
  documentation style.

    Updated by      Richard Gooch   16-JUN-1996: Created
  <iarray_get_data_scaling> routine.

    Updated by      Richard Gooch   19-JUN-1996: Created <iarray_format_value>
  routine.

    Updated by      Richard Gooch   31-JUL-1996: Created
  <iarray_create_from_template> routine.

    Updated by      Richard Gooch   2-AUG-1996: Created <iarray_get_coordinate>
  routine.

    Updated by      Richard Gooch   5-AUG-1996: Created <iarray_set_dim_name>
  routine.

    Updated by      Richard Gooch   5-AUG-1996: Created
  <iarray_append_history_string> routine.

    Updated by      Richard Gooch   8-AUG-1996: Created
  <iarray_copy_named_element> routine.

    Updated by      Richard Gooch   9-AUG-1996: Added <fail_if_not_found>
  parameter to <iarray_copy_named_element> routine.

    Updated by      Richard Gooch   15-AUG-1996: Created <iarray_get_fits_axis>
  routine.

    Updated by      Richard Gooch   29-SEP-1996: Made use of <ds_get_fits_axis>
  routine.

    Updated by      Richard Gooch   17-OCT-1996: "Cosmetic" name change of
  <<template>> parameter in <iarray_create_from_template> because of stupid C++
  theft from namespace.

    Updated by      Richard Gooch   2-NOV-1996: Made use of <ds_format_value>.

    Updated by      Richard Gooch   27-DEC-1996: Created
  <iarray_create_and_setup>.

    Updated by      Richard Gooch   29-DEC-1996: Switched to
  <iarray_put_double>.

    Updated by      Richard Gooch   2-FEB-1997: Made use of new
  <wcs_astro_get_associated> function.

    Updated by      Richard Gooch   15-FEB-1997: Created
  <iarray_get_associated_wcs_astro> function.

    Updated by      Richard Gooch   2-MAR-1997: Created experimental
  <iarray_set_value_name> function.

    Updated by      Richard Gooch   7-AUG-1997: Created experimental
  <iarray_find_centroid> function.

    Updated by      Richard Gooch   1-DEC-1997: Added <<background>> parameter
  to <iarray_find_centroid>.

    Updated by      Richard Gooch   20-FEB-1998: Added trap for min or max
  outside {-TOOBIG,TOOBIG}.

    Updated by      Richard Gooch   24-FEB-1998: Changed to specific ds include
  files.

    Updated by      Richard Gooch   16-MAY-1998: Added support for using FITS
  "BSCALE" and "BZERO" in <iarray_set_data_scaling>. Created experimental
  function <iarray_get_data_range>.

    Updated by      Richard Gooch   19-MAY-1998: Made use of new function
  <ds_element_is_fits_compatible>.

    Updated by      Richard Gooch   26-MAY-1998: Made use of function
  <iarray_get_data_scaling> in <iarray_create_and_setup>.

    Updated by      Richard Gooch   31-MAY-1998: Made use of new function
  <ds_set_data_range> in <iarray_set_data_range>.

    Updated by      Richard Gooch   1-JUN-1998: Made use of new function
  <ds_set_data_scaling> in <iarray_set_data_scaling>.

    Updated by      Richard Gooch   4-JUN-1998: Changed interface to
  <iarray_find_centroid>.

    Updated by      Richard Gooch   28-JUL-1998: Ignore stated range if bogus
  (i.e. min >= max) in <iarray_create_and_setup>.

    Updated by      Richard Gooch   12-SEP-1998: Introduced test for magic
  number.

    Updated by      Richard Gooch   14-OCT-1998: Handle negative data scaling
  in <iarray_create_and_setup>.

    Updated by      Richard Gooch   2-AUG-1999: Tolerate NULL <<multi_desc>>
  in function <iarray_create_and_setup>.

    Updated by      Richard Gooch   23-NOV-1999: Created experimental function
  <iarray_copy_attached>.

    Updated by      Richard Gooch   26-APR-2000: Trap data with computed min >=
  max.

    Updated by      Richard Gooch   22-SEP-2000: Took account of interface
  change in <ds_format_value>.

    Updated by      Richard Gooch   22-MAR-2001: Improved error message in
  <iarray_create_and_setup>.

    Last updated by Richard Gooch   8-OCT-2001: Fixed documentation links.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_wcs.h>
#include <karma_ds_func.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>


#define MAGIC_NUMBER 939032982  /*  The other files depend on this too!  */

#define VERIFY_IARRAY(array) if (array == NULL) \
{fprintf (stderr, "NULL iarray passed\n"); a_prog_bug (function_name); } \
if (array->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid iarray\n"); a_prog_bug (function_name); }


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
unsigned int iarray_dim_index (iarray array, CONST char *name)
/*  [SUMMARY] This routine will find the index of a named dimension.
    <array> The array.
    <name> The name of the dimension.
    [RETURNS] The dimension index if found, else the number of dimension in the
    array is returned.
*/
{
    unsigned int count, num_dim;
    static char function_name[] = "iarray_dim_index";

    VERIFY_IARRAY (array);
    num_dim = iarray_num_dim (array);
    for (count = 0; count < num_dim; ++count)
    {
	if (strcmp (iarray_dim_name (array, count), name) == 0) return (count);
    }
    return (num_dim);
}   /*  End Function iarray_dim_index  */

/*PUBLIC_FUNCTION*/
flag iarray_get_data_scaling (iarray array, double *scale, double *offset)
/*  [SUMMARY] Get the scale and offset for data in an Intelligent Array.
    [PURPOSE] This routine will determine the scale and offset for data in an
    Intelligent Array. This may be important when a floating-point array has
    been converted to an integer array to save space. Scaling information
    should be attached to the array so that the original data values may be
    reconstructed (aside from quantisation effects). The following expression
    may be used to convert scaled values to real values:
    (output = input * scale + offset). The scaling and offset values should
    previously have been attached to the Intelligent Array using the
    [<iarray_set_data_scaling>] routine.
    <array> The array.
    <scale> The scaling value will be written here. The name of the scaling
    value is constructed by appending "__SCALE" to the array value name (see
    [<iarray_value_name>]). If no scaling value is found, 1.0 is written
    here. As a fallback, if the FITS-style "BSCALE" keyword is found and is
    appropriate, this is used.
    <offset> The offset value will be written here. The name of the offset
    value is constructed by appending "__OFFSET" to the array value name (see
    [<iarray_value_name>]). If no offset value is found, 0.0 is written
    here. As a fallback, if the FITS-style "BZERO" keyword is found and is
    appropriate, this is used.
    [RETURNS] TRUE if either the scaling or offset value were found, else FALSE
*/
{
    static char function_name[] = "iarray_get_data_scaling";

    VERIFY_IARRAY (array);
    return ds_get_data_scaling (iarray_value_name (array),
				array->top_pack_desc, *array->top_packet,
				scale, offset);
}   /*  End Function iarray_get_data_scaling  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_set_data_scaling (iarray array, double scale, double offset)
/*  [SUMMARY] Set the scale and offset for data in an Intelligent Array.
    [PURPOSE] This routine will set the scale and offset for data in an
    Intelligent Array. This may be important when a floating-point array has
    been converted to an integer array to save space. The scaling information
    will be attached to the array so that the original data values may be
    reconstructed (aside from quantisation effects). The following expression
    may be used to convert scaled values to real values:
    (output = input * scale + offset).
    <array> The array.
    <scale> The scaling value. The name of the scaling value is constructed by
    appending "__SCALE" to the array value name (see [<iarray_value_name>])
    If appropriate, the FITS-style "BSCALE" keyword is written instead.
    <offset> The offset value. The name of the offset value is constructed by
    appending "__OFFSET" to the array value name(see [<iarray_value_name>])
    If appropriate, the FITS-style "BZERO" keyword is written instead.
    [RETURNS] TRUE if the scaling information is different from what was
    already attached to the array, else FALSE.
*/
{
    static char function_name[] = "iarray_set_data_scaling";

    VERIFY_IARRAY (array);
    return ds_set_data_scaling (iarray_value_name (array),
				array->multi_desc->headers[0],
				array->multi_desc->data,
				scale, offset);
}   /*  End Function iarray_set_data_scaling  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_get_data_range (iarray array, double *minimum, double *maximum)
/*  [SUMMARY] Get the range of data in an Intelligent Array.
    [PURPOSE] This routine will determine the range of data in an Intelligent
    Array. Range information should be attached to the header.
    <array> The array.
    <minimum> The minimum value will be written here. The name of the minimum
    value is constructed by appending "__MINIMUM" to the element name. If no
    minimum value is found, -TOOBIG is written here. As a fallback, if the
    FITS-style "DATAMIN" keyword is found and is appropriate, this is used.
    <maximum> The minimum value will be written here. The name of the minimum
    value is constructed by appending "__MINIMUM" to the element name. If no
    minimum value is found, TOOBIG is written here. As a fallback, if the
    FITS-style "DATAMAX" keyword is found and is appropriate, this is used.
    [RETURNS] TRUE if either the minimum or maximum value were found,
    else FALSE
*/
{
    static char function_name[] = "iarray_get_data_range";

    VERIFY_IARRAY (array);
    return ds_get_data_range (iarray_value_name (array),
			      array->top_pack_desc, *array->top_packet,
			      minimum, maximum);
}   /*  End Function iarray_get_data_range  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_set_data_range (iarray array, double minimum, double maximum)
/*  [SUMMARY] Set the range of data in an Intelligent Array.
    [PURPOSE] This routine will set the range of data in an Intelligent
    Array. The range information will be attached to the header.
    <array> The array.
    <minimum> The minimum value. The name of the minimum value is constructed
    by appending "__MINIMUM" to the array value name (see
    [<iarray_value_name>]).
    If appropriate, the FITS-style "DATAMIN" keyword is written instead.
    <maximum> The maximum value. The name of the maximum value is constructed
    by appending "__MAXIMUM" to the array value name
    (see [<iarray_value_name>]).
    If appropriate, the FITS-style "DATAMAX" keyword is written instead.
    [RETURNS] TRUE if the range information is different from what was
    already attached to the array, else FALSE.
*/
{
    static char function_name[] = "iarray_set_data_range";

    VERIFY_IARRAY (array);
    return ds_set_data_range (iarray_value_name (array),
			      array->multi_desc->headers[0],
			      array->multi_desc->data,
			      minimum, maximum);
}   /*  End Function iarray_set_data_range  */

/*EXPERIMENTAL_FUNCTION*/
void iarray_format_value (iarray array, char string[STRING_LENGTH],
			  double value, double scale, double offset)
/*  [SUMMARY] Format a data value into a string.
    <array> The Intelligent Array the value is associated with.
    <string> The string to write to.
    <value> The value to format.
    <scale> The scale value to apply to the data. If this is TOOBIG the routine
    uses the scale and offset attached to the array. See
    [<iarray_set_data_scaling>] for details.
    <offset> The offset to apply after scaling the data.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "iarray_format_value";

    VERIFY_IARRAY (array);
    ds_format_value (string, value, iarray_value_name (array), scale, offset,
		     value, array->top_pack_desc, *array->top_packet);
}   /*  End Function iarray_format_value  */

/*PUBLIC_FUNCTION*/
iarray iarray_create_from_template (iarray template_arr,unsigned int elem_type,
				    flag copy_world_coords, flag copy_names,
				    flag copy_aux_info)
/*  [SUMMARY] Create an Intelligent Array from a template array.
    [PURPOSE] This routine will create an Intelligent Array of a specified type
    using an existing Intelligent Array as the template for the size,
    dimensionality and other attributes.
    <template_arr> The template Intelligent Array.
    <elem_type> The element type for the new Intelligent Array.
    <copy_world_coords> If TRUE, the world co-ordinates for each dimension in
    the template array are copied, otherwise the defaults are used.
    <copy_names> If TRUE, the dimension names and the value name of the
    template array are copied, else the defaults are used.
    <copy_aux_info> If TRUE, auxilary information (attachments) in the template
    array is copied to the new array.
    [RETURNS] An Intelligent Array on success, else NULL.
*/
{
    iarray new;
    unsigned int count, num_dim;
    double first_coord, last_coord;
    char *elem_name;
    char **dim_names;
    unsigned long *dim_lengths;
    static char function_name[] = "iarray_create_from_template";

    VERIFY_IARRAY (template_arr);
    num_dim = iarray_num_dim (template_arr);
    if ( ( dim_lengths = m_alloc (num_dim * sizeof *dim_lengths) ) == NULL )
    {
	m_abort (function_name, "dimension lengths");
    }
    for (count = 0; count < num_dim; ++count)
    {
	dim_lengths[count] = iarray_dim_length (template_arr, count);
    }
    if (copy_names)
    {
	elem_name = iarray_value_name (template_arr);
	if ( ( dim_names = m_alloc (num_dim * sizeof *dim_names) ) == NULL )
	{
	    m_abort (function_name, "dimension names array");
	}
	for (count = 0; count < num_dim; ++count)
	{
	    dim_names[count] = (char *) iarray_dim_name (template_arr, count);
	}
    }
    else
    {
	elem_name = NULL;
	dim_names = NULL;
    }
    new = iarray_create (elem_type, num_dim, (CONST char **) dim_names,
			 dim_lengths, elem_name,
			 copy_aux_info ? template_arr : NULL);
    m_free (dim_lengths);
    if (dim_names != NULL) m_free (dim_names);
    if (new == NULL) return (NULL);
    if (copy_world_coords)
    {
	for (count = 0; count < num_dim; ++count)
	{
	    iarray_get_world_coords (template_arr, count,
				     &first_coord, &last_coord);
	    iarray_set_world_coords (new, count, first_coord, last_coord);
	}
    }
    return (new);
}   /*  End Function iarray_create_from_template  */

/*PUBLIC_FUNCTION*/
double iarray_get_coordinate (iarray array, unsigned int dim_index,
			      double coord_index)
/*  [SUMMARY] Get a co-ordinate along a dimension.
    <array> The Intelligent Array.
    <dim_index> The dimension index.
    <coord_index> The co-ordinate index.
    [RETURNS] The co-ordinate on success, else TOOBIG.
*/
{
    flag found;
    uaddr count;
    unsigned int orig_dim_index;
    dim_desc *dim;
    static char function_name[] = "iarray_get_coordinate";

    VERIFY_IARRAY (array);
    dim = iarray_get_dim_desc (array, dim_index);
    if (array->offsets == array->arr_desc->offsets)
    {
	/*  Not an alias array  */
	return ds_get_coordinate (dim, coord_index);
    }
    /*  It is an alias array: bugger. Search for offset along dimension
	descriptor which equals first offset along iarray dimension  */
    orig_dim_index = array->orig_dim_indices[dim_index];
    count = 0;
    found = FALSE;
    while ( (count < dim->length) && !found )
    {
	if (array->offsets[dim_index][0] ==
	    array->arr_desc->offsets[orig_dim_index][count]) found = TRUE;
	else ++count;
    }
    if (!found)
    {
	fprintf (stderr, "Could not find offset index!\n");
	a_prog_bug (function_name);
    }
    return ds_get_coordinate (dim, coord_index + count);
}   /*  End Function iarray_get_coordinate  */

/*PUBLIC_FUNCTION*/
void iarray_set_dim_name (iarray array, unsigned int index, CONST char *name,
			  flag new_alloc)
/*  [SUMMARY] Change the name of a dimension of an Intelligent Array.
    <array> The Intelligent Array.
    <index> The index of the dimension.
    <name> The new dimension name.
    <new_alloc> If TRUE, the routine will allocate a new copy of the dimension
    name, else it will copy the pointer (in which case the name must never be
    externally deallocated or changed).
    [RETURNS] Nothing.
*/
{
    char *ptr;
    dim_desc *dim;
    static char function_name[] = "iarray_set_dim_name";

    VERIFY_IARRAY (array);
    dim = iarray_get_dim_desc (array, index);
    if (!new_alloc)
    {
	m_free (dim->name);
	dim->name = (char *) name;
	return;
    }
    if ( ( ptr = st_dup (name) ) == NULL )
    {
	m_abort (function_name, "dimension name");
    }
    m_free (dim->name);
    dim->name = ptr;
}   /*  End Function iarray_set_dim_name  */

/*PUBLIC_FUNCTION*/
flag iarray_append_history_string (iarray array, CONST char *string,
				   flag new_alloc)
/*  [SUMMARY] Add a history string to an Intelligent Array.
    <array> The Intelligent Array.
    <string> The history string to add.
    <new_alloc> If TRUE, the routine will allocate a new copy of the history
    string, else it will copy the pointer (in which case the string must never
    be externally deallocated or changed).
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "iarray_append_history_string";

    VERIFY_IARRAY (array);
    return ds_history_append_string (array->multi_desc, string, new_alloc);
}   /*  End Function iarray_append_history_string  */

/*PUBLIC_FUNCTION*/
flag iarray_copy_named_element (iarray out, iarray in, CONST char *name,
				flag fail_if_not_found,
				flag fail_on_duplicate, flag replace)
/*  [SUMMARY] Copy a named element from one Intelligent Array to another.
    <out> The output Intelligent Array.
    <in> The input Intelligent Array.
    <name> The name of the element to copy.
    <fail_if_not_found> If TRUE, the routine will fail if the element does not
    exist in the input packet.
    <fail_on_duplicate> If TRUE, the routine will fail if the element already
    exists in the output packet.
    <replace> If TRUE and the element already exists in the output packet, it
    is replaced.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    return ds_copy_unique_named_element (out->top_pack_desc, out->top_packet,
					 in->top_pack_desc, *in->top_packet,
					 name, fail_if_not_found,
					 fail_on_duplicate, replace);
}   /*  End Function iarray_copy_named_element  */

/*PUBLIC_FUNCTION*/
unsigned int iarray_get_fits_axis (iarray array, unsigned int index)
/*  [SUMMARY] Get the FITS axis number of a dimension.
    <array> The Intelligent Array.
    <index> The index of the dimension.
    [RETURNS] The FITS axis number on success, else 0.
*/
{
    static char function_name[] = "iarray_get_fits_axis";

    VERIFY_IARRAY (array);
    return ds_get_fits_axis ( array->top_pack_desc, *array->top_packet,
			      iarray_dim_name (array, index) );
}   /*  End Function iarray_get_fits_axis  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_create_and_setup (iarray *array, multi_array *multi_desc,
			      flag inform, unsigned int num_dim,
			      unsigned int required_type,
			      double *min, double *max,
			      flag discard_zero_range, KwcsAstro *ap)
/*  [SUMMARY] Create an Intelligent Array and setup.
    [PURPOSE] This routine will attempt to guess the filetype of a file and
    in the file, converting to an Intelligent Array if possible. The routine
    then performs some simple checks and some other convenience functions.
    <array> The Intelligent Array is written here. An existing array pointed to
    by this is deallocated.
    <multi_desc> The multi_array descriptor. If this is NULL, any existing data
    is deallocated and no further action is taken.
    <inform> If TRUE, the routine displays some informative messages.
    <num_dim> The number of dimensions required. If this is 0, any number of
    dimensions is allowed.
    <required_type> The required data type. If this is NONE, then no type is
    required.
    <min> The minimum data value in the array is written here. If this is NULL
    nothing is written here.
    <max> The maximum data value in the array is written here. If this is NULL
    nothing is written here.
    <discard_zero_range> If TRUE, and the range of the data is zero, the
    routine fails.
    <ap> The KwcsAstro object is written here. This may be NULL.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    iarray arr;
    double min_val, max_val, tmp;
    static char function_name[] = "iarray_create_and_setup";

    if (multi_desc == NULL)
    {
	if (*array != NULL) iarray_dealloc (*array);
	*array = NULL;
	if (ap != NULL) *ap = NULL;
	return (TRUE);
    }
    /*  Try to get array  */
    if ( ( arr = iarray_get_from_multi_array (multi_desc, NULL, num_dim,
					      NULL, NULL) ) == NULL )
    {
	if (num_dim < 1)
	    fprintf (stderr, "Could not find any array in data\n");
	else fprintf (stderr, "Could not find %u-dimensional array in data\n",
		      num_dim);
	return (FALSE);
    }
    if ( (required_type != NONE) && (iarray_type (arr) != required_type) )
    {
	fprintf (stderr, "Data type: %u not found\n", required_type);
	iarray_dealloc (arr);
	return (FALSE);
    }
    if (inform)
    {
	if (iarray_num_dim (arr) == 2)
	{
	    fprintf ( stderr, "Loaded image of %lu * %lu\n",
		      iarray_dim_length (arr, 1), iarray_dim_length (arr, 0) );
	}
	else if (iarray_num_dim (arr) == 3)
	{
	    fprintf ( stderr, "Loaded cube of %lu * %lu * %lu\n",
		      iarray_dim_length (arr, 2), iarray_dim_length (arr, 1),
		      iarray_dim_length (arr, 0) );
	}
    }
    /*  Compute the minimum and maximum  */
    if ( discard_zero_range || (min != NULL) || (max != NULL) )
    {
	flag scaled;
	flag compute_range = TRUE;
	double scale, offset;

	scaled = iarray_get_data_scaling (arr, &scale, &offset);
	if ( iarray_get_data_range (arr, &min_val, &max_val) )
	{
	    /*  Range was provided  */
	    if (inform) fprintf (stderr, "Stated scaled min: %g  max: %g\n",
				 min_val, max_val);
	    if (min_val >= max_val)
	    {
		fprintf (stderr,"Min is not less than max: ignoring header\n");
	    }
	    else
	    {
		compute_range = FALSE;
		if (scaled)
		{
		    min_val = (min_val - offset) / scale;
		    max_val = (max_val - offset) / scale;
		    if (min_val > max_val) swap (min_val, max_val, tmp);
		    if (inform) fprintf (stderr,
					 "Stated unscaled min: %g  max: %g\n",
					 min_val, max_val);
		}
	    }
	}
	if (compute_range)
	{
	    if ( !iarray_min_max (arr, CONV1_REAL,
				  &min_val, &max_val) )
	    {
		fprintf (stderr, "Error computing min-max\n");
		iarray_dealloc (arr);
		return (FALSE);
	    }
	    if (min_val >= max_val)
	    {
		fprintf (stderr,
			 "Computed min is not less than max: bad data\n");
		iarray_dealloc (arr);
		return (FALSE);
	    }
	    if (inform)
	    {
		if (scaled) fprintf (stderr,
				     "Computed scaled min: %g  max: %g\n",
				     min_val * scale + offset,
				     max_val* scale + offset);
		fprintf (stderr, "Computed unscaled min: %g  max: %g\n",
			 min_val, max_val);
	    }
	    iarray_set_data_range (arr,
				   min_val * scale + offset,
				   max_val * scale + offset);
	}
	if (min_val <= -TOOBIG)
	{
	    fprintf (stderr, "min: %e  is not greater than %e\n",
		     min_val, -TOOBIG);
	    iarray_dealloc (arr);
	    return (FALSE);
	}
	if ( DOUBLE_IS_BLANK (max_val) )
	{
	    fprintf (stderr, "max: %e  is not less than %e\n", max_val,TOOBIG);
	    iarray_dealloc (arr);
	    return (FALSE);
	}
	if (min != NULL) *min = min_val;
	if (max != NULL) *max = max_val;
	if ( discard_zero_range && (min_val == max_val) )
	{
	    if (inform) fprintf (stderr, "zero (all values: %e): pointless!\n",
				 min_val);
	    else fprintf (stderr, "min: %e is same as max: pointless!\n",
			  min_val);
	    iarray_dealloc (arr);
	    return (FALSE);
	}
    }
    if (*array != NULL) iarray_dealloc (*array);
    *array = arr;
    if (ap != NULL)
    {
	*ap = wcs_astro_get_associated (arr->multi_desc, arr->array_num);
	fprintf (stderr, "%s: ap: %p\n", function_name, *ap);
    }
    return (TRUE);
}   /*  End Function iarray_create_and_setup  */

/*EXPERIMENTAL_FUNCTION*/
KwcsAstro iarray_get_associated_wcs_astro (iarray array)
/*  [SUMMARY] Get the KwcsAstro object associated with an Intelligent Array.
    <array> The Intelligent Array.
    [RETURNS] The KwcsAstro object if associated, else NULL.
*/
{
    return wcs_astro_get_associated (array->multi_desc, array->array_num);
}   /*  End Function iarray_get_associated_wcs_astro  */

/*EXPERIMENTAL_FUNCTION*/
void iarray_set_value_name (iarray array, CONST char *name, flag new_alloc)
/*  [SUMMARY] Change the data name of an Intelligent Array.
    <array> The Intelligent Array.
    <name> The new data value name.
    <new_alloc> If TRUE, the routine will allocate a new copy of the value
    name, else it will copy the pointer (in which case the name must never be
    externally deallocated or changed).
    [RETURNS] Nothing.
*/
{
    char *ptr;
    static char function_name[] = "iarray_set_value_name";

    VERIFY_IARRAY (array);
    if (!new_alloc)
    {
	m_free ( iarray_value_name (array) );
	iarray_value_name (array) = (char *) name;
	return;
    }
    if ( ( ptr = st_dup (name) ) == NULL )
    {
	m_abort (function_name, "dimension name");
    }
    m_free ( iarray_value_name (array) );
    iarray_value_name (array) = ptr;
}   /*  End Function iarray_set_value_name  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_find_centroid (iarray array, double *background, flag *inverted,
			   double *xpos, uaddr xradius,
			   double *ypos, uaddr yradius)
/*  [SUMMARY] Find the centroid inside an ellipse.
    [PURPOSE] This routine will search for the centroid of an object inside a
    specified ellipse.
    <array> The Intelligent Array.
    <inverted> If TRUE the algorithms should search for a minimum, rather than
    a maximum.
    <background> The background value is written here. This may be NULL.
    <inverted> The value TRUE will be written here if the image values are
    inverted (i.e. the object of interested is in a local minimum rather than a
    maximum).
    <xpos> The horizontal data pixel position. The original value must be a
    reasonable approximation. This is modified to the centroid position.
    <xradius> The horizontal radius of the ellipse.
    <ypos> The vertical data pixel position. The original value must be a
    reasonable approximation. This is modified to the centroid position.
    <yradius> The vertical radius of the ellipse.
    [NOTE] (0.0, 0.0) corresponds to the centre of the first pixel.
    [RETURNS] TRUE on succes, else FALSE.
*/
{
    static char function_name[] = "iarray_find_centroid";

    if (iarray_num_dim (array) != 2)
    {
	fprintf ( stderr, "Array dimensions: %u is not 2\n",
		  iarray_num_dim (array) );
	a_prog_bug (function_name);
    }
    return ds_find_centroid (array->data, iarray_type (array),
			     background, inverted,
			     iarray_dim_length (array, 1), array->offsets[1],
			     xpos, xradius,
			     iarray_dim_length (array, 0), array->offsets[0],
			     ypos, yradius);
}   /*  End Function iarray_find_centroid  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_copy_attached (iarray output, iarray input, flag fail_on_duplicate,
			   flag replace)
/*  [SUMMARY] Copy attached data from one Intelligent Array to another.
    <output> The output Intelligent Array.
    <input> The input Intelligent Array.
    <fail_on_duplicate> If TRUE, the routine will fail if the element already
    exists in the output packet.
    <replace> If TRUE and the element already exists in the output packet, it
    is replaced.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    /*static char function_name[] = "iarray_copy_attached";*/

    for (count = 0; count < input->top_pack_desc->num_elements; ++count)
    {
	if ( !ds_element_is_named(input->top_pack_desc->element_types[count]) )
	    continue;
	if ( !ds_copy_unique_named_element
	     (output->top_pack_desc, output->top_packet,
	      input->top_pack_desc, *input->top_packet,
	      input->top_pack_desc->element_desc[count],
	      TRUE, fail_on_duplicate, replace) )
	    return (FALSE);
    }
    return (TRUE);
}   /*  End Function iarray_copy_attached  */
