/*LINTLIBRARY*/
/*  regrid.c

    This code provides regridding support for Intelligent Arrays.

    Copyright (C) 2000-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the regridding of 2 and 3
    dimensional data structures.


    Written by      Richard Gooch   15-JUN-2000: Copied from
  modules/image/kregrid.c

    Last updated by Richard Gooch   9-AUG-2001: Minor documentation update.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_ds_func.h>
#include <karma_wcs.h>
#include <karma_a.h>
#include <karma_m.h>


#define MAGIC_NUMBER 939032982  /*  The other files depend on this too!  */

#define VERIFY_IARRAY(array) if (array == NULL) \
{fprintf (stderr, "NULL iarray passed\n"); a_prog_bug (function_name); } \
if (array->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid iarray\n"); a_prog_bug (function_name); }


/*  Private functions  */
STATIC_FUNCTION (flag compute_region,
		 (iarray out_arr, KwcsAstro out_ap,
		  iarray in_arr, KwcsAstro in_ap,
		  unsigned int *startx, unsigned int *stopx,
		  unsigned int *starty, unsigned int *stopy) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_regrid_2D (iarray out_arr, KwcsAstro out_ap,
		       iarray in_arr, KwcsAstro in_ap,
		       unsigned int sample_option, float *min, float *max)
/*  [SUMMARY] Regrid an array by iterating over the output array.
    <out_arr> The output array. The new grid must already be defined.
    <out_ap> The output KwcsAstro object.
    <in_arr> The input array.
    <n_ap> The input KwcsAstro object.
    <sample_option> The sample option to use.
    <min> The minimum value is written here. This is updated. This may be NULL.
    <max> The maximum value is written here. This is updated. This may be NULL.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    iarray ra_arr, dec_arr;
    unsigned int xlen, ylen, x, y;
    unsigned int startx, stopx, starty, stopy;
    float xmax, ymax;
    float zero = 0.0;
    float one = 1.0;
    float thres = 1e-3;
    float lmin, lmax;
    dim_desc *xdim, *ydim;
    static char function_name[] = "iarray_regrid_2D";

    VERIFY_IARRAY (out_arr);
    VERIFY_IARRAY (in_arr);
    if ( (iarray_type (in_arr) != K_FLOAT) ||
	 (iarray_type (out_arr) != K_FLOAT) )
    {
	fprintf (stderr, "Only floating-point arrays supported\n");
	return (FALSE);
    }
    if (min == NULL) lmin = TOOBIG;
    else lmin = *min;
    if (max == NULL) lmax = -TOOBIG;
    else lmax = *max;
    /*  Setup co-ordinate arrays. These co-ordinate arrays are used to
	translate (integral) co-ordinate indices in the output array to
	co-ordinate indices in the input array. To do this, the co-ordinate
	arrays are filled with linear world co-ordinates of the output array,
	then converted to proper non-linear world co-ordinates using the
	projection system of the output array, then converted back to linear
	world co-ordinates using the projection system of the input array, and
	finally to co-ordinate indices in the input array  */
    if ( !compute_region (out_arr, out_ap, in_arr, in_ap, &startx, &stopx,
			  &starty, &stopy) )
    {
	fprintf (stderr, "No overlap between input and output grids\n");
	return (FALSE);
    }
    xlen = stopx - startx;
    ylen = stopy - starty;
    if ( ( ra_arr = iarray_create_2D (ylen, xlen, K_DOUBLE) ) == NULL )
    {
	m_abort (function_name, "RA array");
    }
    if ( ( dec_arr = iarray_create_2D (ylen, xlen, K_DOUBLE) ) == NULL )
    {
	m_abort (function_name, "DEC array");
    }
    fprintf (stderr, "filling co-ordinate arrays...\n");
    xdim = iarray_get_dim_desc (out_arr, 1);
    ydim = iarray_get_dim_desc (out_arr, 0);
    for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x)
	D2 (ra_arr, y, x) = startx + x;
    ds_convert_coordinates (xdim, xlen * ylen, (double *) ra_arr->data, FALSE,
			    FALSE, FALSE);
    for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x)
	D2 (dec_arr, y, x) = starty + y;
    ds_convert_coordinates (ydim, xlen * ylen, (double *) dec_arr->data, FALSE,
			    FALSE, FALSE);
    fprintf (stderr, "unprojecting...\t");
    wcs_astro_transform (out_ap, xlen * ylen,
			 (double *) ra_arr->data, FALSE,
			 (double *) dec_arr->data, FALSE,
			 NULL, FALSE,
			 0, NULL, NULL);
    fprintf (stderr, "reprojecting...\t");
    wcs_astro_transform (in_ap, xlen * ylen,
			 (double *) ra_arr->data, TRUE,
			 (double *) dec_arr->data, TRUE,
			 NULL, FALSE,
			 0, NULL, NULL);
    fprintf (stderr, "converting to co-ordinate indices...\n");
    xdim = iarray_get_dim_desc (in_arr, 1);
    ydim = iarray_get_dim_desc (in_arr, 0);
    ds_convert_coordinates (xdim, xlen * ylen, (double *) ra_arr->data, TRUE,
			    TRUE,
			    (sample_option ==
			     REGRID_SAMPLE_DATA_COPY) ? TRUE : FALSE);
    ds_convert_coordinates (ydim, xlen * ylen, (double *) dec_arr->data, TRUE,
			    TRUE,
			    (sample_option ==
			     REGRID_SAMPLE_DATA_COPY) ? TRUE : FALSE);
    fprintf (stderr, "regridding...");
    xmax = xdim->length - 1;
    ymax = ydim->length - 1;
    /*  Loop through the lines  */
    for (y = starty; y < stopy; ++y)
    {
	switch (sample_option)
	{
	  case REGRID_SAMPLE_DATA_COPY:
	    for (x = startx; x < stopx; ++x)
	    {
		float xg, yg, val;

		xg = D2 (ra_arr, y - starty, x - startx);
		yg = D2 (dec_arr, y - starty, x - startx);
		if ( DOUBLE_IS_BLANK (xg) || DOUBLE_IS_BLANK (yg) ) continue;
		val = F2 (in_arr, (int) yg, (int) xg);
		F2 (out_arr, y, x) = val;
		if (val < lmin) lmin = val;
		else if (val > lmax) lmax = val;
	    }
	    break;
	  case REGRID_SAMPLE_BILINEAR_INTERPOLATION:
	    for (x = startx; x < stopx; ++x)
	    {
		int toobig_count;
		float val_00, val_10, val_01, val_11;
		float xg, yg, x0, x1, y0, y1, dx, dy, dx_r, dy_r, val;

		xg = D2 (ra_arr, y - starty, x - startx);
		yg = D2 (dec_arr, y - starty, x - startx);
		if ( DOUBLE_IS_BLANK (xg) || DOUBLE_IS_BLANK (yg) ) continue;
		if ( (xg < thres) || (xg > xmax - thres) ||
		     (yg < thres) || (yg > ymax - thres) ) continue;
		x0 = floor (xg);
		x1 = ceil (xg);
		dx = xg - x0;
		dx_r = one - dx;
		y0 = floor (yg);
		y1 = ceil (yg);
		dy = yg - y0;
		dy_r = one - dy;
		/*  Take contribution from 4 neighbouring points  */
		toobig_count = 0;
		val_00 = F2 (in_arr, (int) y0, (int) x0);
		if ( DOUBLE_IS_BLANK (val_00) )
		{
		    val_00 = zero;
		    ++toobig_count;
		}
		val_10 = F2 (in_arr, (int) y0, (int) x1);
		if ( DOUBLE_IS_BLANK (val_10) )
		{
		    val_10 = zero;
		    ++toobig_count;
		}
		val_01 = F2 (in_arr, (int) y1, (int) x0);
		if ( DOUBLE_IS_BLANK (val_01) )
		{
		    val_01 = zero;
		    ++toobig_count;
		}
		val_11 = F2 (in_arr, (int) y1, (int) x1);
		if ( DOUBLE_IS_BLANK (val_11) )
		{
		    val_11 = zero;
		    ++toobig_count;
		}
		if (toobig_count > 3) continue;
		val = val_00 * dx_r * dy_r + val_10 * dx * dy_r +
		    val_01 * dx_r * dy + val_11 * dx * dy;
		F2 (out_arr, y, x) = val;
		if (val < lmin) lmin = val;
		else if (val > lmax) lmax = val;
	    }
	    break;
	  default:
	    fprintf (stderr, "Illegal sample_option: %u\n", sample_option);
	    a_prog_bug (function_name);
	    break;
	}
    }
    iarray_dealloc (ra_arr);
    iarray_dealloc (dec_arr);
    fprintf (stderr, "\tregridded\n");
    if (min != NULL) *min = lmin;
    if (max != NULL) *max = lmax;
    return (TRUE);
}   /*  End Function iarray_regrid_2D  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_regrid_3D (iarray out_arr, KwcsAstro out_ap,
		       iarray in_arr, KwcsAstro in_ap,
		       unsigned int sample_option, float *min, float *max)
/*  [SUMMARY] Regrid an array by iterating over the output array.
    <out_arr> The output array. The new grid must already be defined.
    <out_ap> The output KwcsAstro object.
    <in_arr> The input array.
    <in_ap> The input KwcsAstro object.
    <sample_option> The sample option to use.
    <min> The minimum value is written here. This is updated. This may be NULL.
    <max> The maximum value is written here. This is updated. This may be NULL.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    iarray ra_arr, dec_arr;
    int toobig_count;
    unsigned int xlen, ylen, zlen, x, y, z;
    unsigned int startx, stopx, starty, stopy;
    float val_00, val_10, val_01, val_11;
    float xg, yg, x0, x1, y0, y1, dx, dy, dx_r, dy_r, val;
    float xmax, ymax;
    float zero = 0.0;
    float one = 1.0;
    float thres = 1e-3;
    float lmin, lmax;
    dim_desc *xdim, *ydim;
    static char function_name[] = "iarray_regrid_3D";

    VERIFY_IARRAY (out_arr);
    VERIFY_IARRAY (in_arr);
    if ( (iarray_type (in_arr) != K_FLOAT) ||
	 (iarray_type (out_arr) != K_FLOAT) )
    {
	fprintf (stderr, "Only floating-point arrays supported\n");
	return (FALSE);
    }
    if (min == NULL) lmin = TOOBIG;
    else lmin = *min;
    if (max == NULL) lmax = -TOOBIG;
    else lmax = *max;
    /*  Setup co-ordinate arrays. These co-ordinate arrays are used to
	translate (integral) co-ordinate indices in the output array to
	co-ordinate indices in the input array. To do this, the co-ordinate
	arrays are filled with linear world co-ordinates of the output array,
	then converted to proper non-linear world co-ordinates using the
	projection system of the output array, then converted back to linear
	world co-ordinates using the projection system of the input array, and
	finally to co-ordinate indices in the input array  */
    zlen = iarray_dim_length (out_arr, 0);
    if ( !compute_region (out_arr, out_ap, in_arr, in_ap, &startx, &stopx,
			  &starty, &stopy) )
    {
	fprintf (stderr, "No overlap between input and output grids\n");
	return (FALSE);
    }
    xlen = stopx - startx;
    ylen = stopy - starty;
    if ( ( ra_arr = iarray_create_2D (ylen, xlen, K_DOUBLE) ) == NULL )
    {
	m_abort (function_name, "RA array");
    }
    if ( ( dec_arr = iarray_create_2D (ylen, xlen, K_DOUBLE) ) == NULL )
    {
	m_abort (function_name, "DEC array");
    }
    fprintf (stderr, "filling co-ordinate arrays...\n");
    xdim = iarray_get_dim_desc (out_arr, 2);
    ydim = iarray_get_dim_desc (out_arr, 1);
    for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x)
	D2 (ra_arr, y, x) = startx + x;
    ds_convert_coordinates (xdim, xlen * ylen, (double *) ra_arr->data, FALSE,
			    FALSE, FALSE);
    for (y = 0; y < ylen; ++y) for (x = 0; x < xlen; ++x)
	D2 (dec_arr, y, x) = starty + y;
    ds_convert_coordinates (ydim, xlen * ylen, (double *) dec_arr->data, FALSE,
			    FALSE, FALSE);
    fprintf (stderr, "unprojecting...\t");
    wcs_astro_transform (out_ap, xlen * ylen,
			 (double *) ra_arr->data, FALSE,
			 (double *) dec_arr->data, FALSE,
			 NULL, FALSE,
			 0, NULL, NULL);
    fprintf (stderr, "reprojecting...\t");
    wcs_astro_transform (in_ap, xlen * ylen,
			 (double *) ra_arr->data, TRUE,
			 (double *) dec_arr->data, TRUE,
			 NULL, FALSE,
			 0, NULL, NULL);
    fprintf (stderr, "converting to co-ordinate indices...\n");
    xdim = iarray_get_dim_desc (in_arr, 2);
    ydim = iarray_get_dim_desc (in_arr, 1);
    ds_convert_coordinates (xdim, xlen * ylen, (double *) ra_arr->data, TRUE,
			    TRUE,
			    (sample_option ==
			     REGRID_SAMPLE_DATA_COPY) ? TRUE : FALSE);
    ds_convert_coordinates (ydim, xlen * ylen, (double *) dec_arr->data, TRUE,
			    TRUE,
			    (sample_option ==
			     REGRID_SAMPLE_DATA_COPY) ? TRUE : FALSE);
    fprintf (stderr, "regridding");
    xmax = xdim->length - 1;
    ymax = ydim->length - 1;
    /*  Loop through the planes  */
    for (z = 0; z < zlen; ++z)
    {
	switch (sample_option)
	{
	  case REGRID_SAMPLE_DATA_COPY:
	    for (y = starty; y < stopy; ++y) for (x = startx; x < stopx; ++x)
	    {
		xg = D2 (ra_arr, y - starty, x - startx);
		yg = D2 (dec_arr, y - starty, x - startx);
		if ( DOUBLE_IS_BLANK (xg) || DOUBLE_IS_BLANK (yg) ) continue;
		val = F3 (in_arr, z, (int) yg, (int) xg);
		F3 (out_arr, z, y, x) = val;
		if (val < lmin) lmin = val;
		else if (val > lmax) lmax = val;
	    }
	    break;
	  case REGRID_SAMPLE_BILINEAR_INTERPOLATION:
	    for (y = starty; y < stopy; ++y) for (x = startx; x < stopx; ++x)
	    {
		xg = D2 (ra_arr, y - starty, x - startx);
		yg = D2 (dec_arr, y - starty, x - startx);
		if ( DOUBLE_IS_BLANK (xg) || DOUBLE_IS_BLANK (yg) ) continue;
		if ( (xg < thres) || (xg > xmax - thres) ||
		     (yg < thres) || (yg > ymax - thres) ) continue;
		x0 = floor (xg);
		x1 = ceil (xg);
		dx = xg - x0;
		dx_r = one - dx;
		y0 = floor (yg);
		y1 = ceil (yg);
		dy = yg - y0;
		dy_r = one - dy;
		/*  Take contribution from 4 neighbouring points  */
		toobig_count = 0;
		val_00 = F3 (in_arr, z, (int) y0, (int) x0);
		if ( DOUBLE_IS_BLANK (val_00) )
		{
		    val_00 = zero;
		    ++toobig_count;
		}
		val_10 = F3 (in_arr, z, (int) y0, (int) x1);
		if ( DOUBLE_IS_BLANK (val_10) )
		{
		    val_10 = zero;
		    ++toobig_count;
		}
		val_01 = F3 (in_arr, z, (int) y1, (int) x0);
		if ( DOUBLE_IS_BLANK (val_01) )
		{
		    val_01 = zero;
		    ++toobig_count;
		}
		val_11 = F3 (in_arr, z, (int) y1, (int) x1);
		if ( DOUBLE_IS_BLANK (val_11) )
		{
		    val_11 = zero;
		    ++toobig_count;
		}
		if (toobig_count > 3) continue;
		val = val_00 * dx_r * dy_r + val_10 * dx * dy_r +
		    val_01 * dx_r * dy + val_11 * dx * dy;
		F3 (out_arr, z, y, x) = val;
		if (val < lmin) lmin = val;
		else if (val > lmax) lmax = val;
	    }
	    break;
	  default:
	    fprintf (stderr, "Illegal sample_option: %u\n", sample_option);
	    a_prog_bug (function_name);
	    break;
	}
    }
    iarray_dealloc (ra_arr);
    iarray_dealloc (dec_arr);
    fprintf (stderr, "\tregridded\n");
    if (min != NULL) *min = lmin;
    if (max != NULL) *max = lmax;
    return (TRUE);
}   /*  End Function iarray_regrid_3D  */


/*  Private functions follow  */

static flag compute_region (iarray out_arr, KwcsAstro out_ap,
			    iarray in_arr, KwcsAstro in_ap,
			    unsigned int *startx, unsigned int *stopx,
			    unsigned int *starty, unsigned int *stopy)
/*  [SUMMARY] Compute the region the input image covers.
    <out_arr> The output array.
    <out_ap> The output KwcsAstro object.
    <in_arr> The input array.
    <in_ap> The input KwcsAstro object.
    <startx> The starting X position in the output image.
    <stopx> The stop X position in the output image.
    <starty> The starting Y position in the output image.
    <stopy> The stop Y position in the output image.
    [RETURNS] TRUE if there is overlap, else FALSE.
*/
{
    flag wholly_within;
    unsigned int in_dim, out_dim, num_coords, count;
    unsigned long in_xlen, in_ylen;
    double out_xlen, out_ylen;
    double xmin = TOOBIG;
    double xmax = -TOOBIG;
    double ymin = TOOBIG;
    double ymax = -TOOBIG;
    double zero = 0.0;
    double ra, dec;
    double *ptr, *ra_arr, *dec_arr;
    static char function_name[] = "compute_region";

    in_dim = iarray_num_dim (in_arr);
    out_dim = iarray_num_dim (out_arr);
    in_xlen = iarray_dim_length (in_arr, in_dim - 1);
    in_ylen = iarray_dim_length (in_arr, in_dim - 2);
    out_xlen = iarray_dim_length (out_arr, out_dim - 1);
    out_ylen = iarray_dim_length (out_arr, out_dim - 2);
    num_coords = in_xlen * 2 + in_ylen * 2;
    if (out_xlen * out_ylen <= num_coords)
    {
	/*  Not likely to be profitable computing sub-region  */
	*startx = 0;
	*stopx = out_xlen;
	*starty = 0;
	*stopy = out_ylen;
	return (TRUE);
    }
    ptr = (double *) m_alloc_scratch (num_coords * sizeof *ptr * 2,
				      function_name);
    ra_arr = ptr;
    dec_arr = ra_arr + num_coords;
    /*  Now do a quick check to see if output image lies wholly within input
	image. If each corner of the imput image is inside the output image,
	the input is considered to lie within the output  */
    /*  X = 0, Y = 0  */
    ra_arr[0] = 0.0;
    dec_arr[0] = 0.0;
    /*  X = max, Y = 0  */
    ra_arr[1] = out_xlen - 1;
    dec_arr[1] = 0.0;
    /*  X = 0, Y = max  */
    ra_arr[2] = 0.0;
    dec_arr[2] = out_ylen - 1;
    /*  X = max, Y = max  */
    ra_arr[3] = out_xlen - 1;
    dec_arr[3] = out_ylen - 1;
    /*  Transform into world co-ordinates  */
    wcs_astro_transform (out_ap, 4,
			 ra_arr, FALSE, dec_arr, FALSE, NULL, FALSE,
			 0, NULL, NULL);
    /*  Transform into input pixel co-ordinates  */
    wcs_astro_transform (in_ap, 4,
			 ra_arr, TRUE, dec_arr, TRUE, NULL, FALSE,
			 0, NULL, NULL);
    /*  Now test the limits  */
    for (count = 0, wholly_within = TRUE; count < 4; ++count)
    {
	if (ra_arr[count] < zero) wholly_within = FALSE;
	if (ra_arr[count] > in_xlen - 1) wholly_within = FALSE;
	if (dec_arr[count] < zero) wholly_within = FALSE;
	if (dec_arr[count] > in_ylen - 1) wholly_within = FALSE;
    }
    if (wholly_within)
    {
	fprintf (stderr,
		 "Output image lies wholly within input: not efficient\n");
	*startx = 0;
	*stopx = out_xlen;
	*starty = 0;
	*stopy = out_ylen;
	return (TRUE);
    }
    /*  Fill arrays with input pixel co-ordinates  */
    /*  First along Y = 0  */
    for (count = 0; count < in_xlen; ++count)
    {
	ra_arr[count] = iarray_get_coordinate (in_arr, in_dim - 1, count);
	dec_arr[count] = 0.0;
    }
    ra_arr += count;
    dec_arr += count;
    /*  Next along Y = max  */
    for (count = 0; count < in_xlen; ++count)
    {
	ra_arr[count] = iarray_get_coordinate (in_arr, in_dim - 1, count);
	dec_arr[count] = in_ylen - 1;
    }
    ra_arr += count;
    dec_arr += count;
    /*  Now along X = 0  */
    for (count = 0; count < in_ylen; ++count)
    {
	ra_arr[count] = 0.0;
	dec_arr[count] = iarray_get_coordinate (in_arr, in_dim - 2, count);
    }
    ra_arr += count;
    dec_arr += count;
    /*  Now along X = max  */
    for (count = 0; count < in_ylen; ++count)
    {
	ra_arr[count] = in_xlen - 1;
	dec_arr[count] = iarray_get_coordinate (in_arr, in_dim - 2, count);
    }
    /*  Transform into world co-ordinates  */
    ra_arr = ptr;
    dec_arr = ra_arr + num_coords;
    wcs_astro_transform (in_ap, num_coords,
			 ra_arr, FALSE, dec_arr, FALSE, NULL, FALSE,
			 0, NULL, NULL);
    /*  Transform into output pixel co-ordinates  */
    wcs_astro_transform (out_ap, num_coords,
			 ra_arr, TRUE, dec_arr, TRUE, NULL, FALSE,
			 0, NULL, NULL);
    /*  Now test the limits  */
    out_xlen -= 1.0;
    out_ylen -= 1.0;
    for (count = 0; count < num_coords; ++count)
    {
	ra = ra_arr[count];
	dec = dec_arr[count];
	if ( DOUBLE_IS_BLANK (ra) || DOUBLE_IS_BLANK (dec) ) continue;
	if ( (ra < zero) || (ra > out_xlen) ) continue;
	if ( (dec < zero) || (dec > out_ylen) ) continue;
	if (ra < xmin) xmin = ra;
	if (ra > xmax) xmax = ra;
	if (dec < ymin) ymin = dec;
	if (dec > ymax) ymax = dec;
    }
    if ( (xmin > xmax) || (ymin > ymax) ) return (FALSE);
    *startx = floor (xmin);
    *stopx = (unsigned int) floor (xmax) + 1;
    *starty = floor (ymin);
    *stopy = (unsigned int) floor (ymax) + 1;
    m_free_scratch ();
    return (TRUE);
}   /*  End Function compute_region  */
