/*LINTLIBRARY*/
/*  transform.c

    This code provides miscellaneous Intelligent Array transormation routines.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for various transformations for
  n-dimensional data structures.


    Written by      Richard Gooch   6-AUG-1997

    Last updated by Richard Gooch   9-AUG-2001: Created experimental function
  <iarray_transform_3D>.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_wcs.h>
#include <karma_a.h>
#include <karma_m.h>


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void iarray_rect_to_polar (iarray out, iarray in,
			   double centre_x, double centre_y,
			   double radius_major, double radius_minor,
			   double rotation)
/*  [SUMMARY] Transform an image from rectangular to polar co-ordinates.
    [PURPOSE] This routine will transform an input image with a rectangular
    co-ordinate system to an output image with a polar co-ordinate system.
    Lines of constant radius in the output image correspond to the points
    around an ellipse in the input image. The X axis in the output image will
    correspond to different radii bins, while the Y axis will correspond to
    different angle bins. The last X axis index will correspond to the outer
    ellipse.
    <out> The output 2D Intelligent Array of type K_FLOAT. The polar image is
    written here. This must be externally allocated.
    <in> The input 2D Intelligent Array of type K_FLOAT. This must contain the
    input rectangular image.
    <centre_x> The X centre of the new polar co-ordinate system in the input.
    <centre_y> The Y centre of the new polar co-ordinate system in the input.
    <radius_major> The radius of the major axis of the outer ellipse.
    <radius_minor> The radius of the minor axis of the outer ellipse.
    <rotation> The rotation of the ellipse, in degrees. An angle of 0 would
    mean the major axis is parallel to the input image X axis.
    [RETURNS] Nothing.
*/
{
    unsigned long x, y, xlen, ylen;
    float val;
    double radius, radius_step, angle, angle_step;
    double zero = 0.0;
    double xi, yi, xp, yp, xmax, ymax;
    double cos_rot, sin_rot, cos_ang, sin_ang;
    static char function_name[] = "iarray_rect_to_polar";

    if (iarray_type (in) != K_FLOAT)
    {
	fprintf (stderr, "Input image is not of type K_FLOAT\n");
	return;
    }
    if (iarray_type (out) != K_FLOAT)
    {
	fprintf (stderr, "Input image is not of type K_FLOAT\n");
	a_prog_bug (function_name);
    }
    xlen = iarray_dim_length (out, 1);
    ylen = iarray_dim_length (out, 0);
    radius_step = 1.0 / (double) (xlen - 1);
    angle_step = TWOPI / (double) ylen;
    xmax = iarray_dim_length (in, 1) - 1;
    ymax = iarray_dim_length (in, 0) - 1;
    cos_rot = cos (rotation * PION180);
    sin_rot = sin (rotation * PION180);
    for (y = 0, angle = 0.0; y < ylen; ++y, angle += angle_step)
    {
	cos_ang = cos (angle);
	sin_ang = sin (angle);
	for (x = 0, radius = 0.0; x < xlen; ++x, radius += radius_step)
	{
	    xp = radius_major * radius * cos_ang;
	    yp = radius_minor * radius * sin_ang;
	    xi = centre_x + xp * cos_rot - yp * sin_rot;
	    yi = centre_y + yp * cos_rot + xp * sin_rot;
	    if ( (xi >= zero) && (xi <= xmax) && (yi >= zero) && (yi < ymax) )
		val = F2 (in, (int) yi, (int) xi);
	    else val = TOOBIG;
	    F2 (out, y, x) = val;
	}
    }
}   /*  End Function iarray_rect_to_polar  */

/*EXPERIMENTAL_FUNCTION*/
flag iarray_transform_3D (iarray out_arr, KwcsAstro out_ap,
			  iarray in_arr, KwcsAstro in_ap,
			  void (*func) (void *info, unsigned long length,
					double *x, double *y, double *z),
			  void *info)
/*  [SUMMARY] Transform an array by iterating over the input array.
    <out_arr> The output array. The new grid must already be defined.
    <out_ap> The output KwcsAstro object.
    <in_arr> The input array.
    <in_ap> The input KwcsAstro object.
    <func> The function used to convert from input world co-ordinates to output
    world co-ordinates.
    <info> An arbitrary information pointer passed to <<func>>.
    [NOTE] Aliasing artefacts may appear if the output grid resolution is too
    high. This is an inherent problem with forward transform algorithms.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned long in_x, in_y, in_z;
    double *buffer;
    dim_desc *in_xdim, *in_ydim, *in_zdim, *out_xdim, *out_ydim, *out_zdim;
    static char function_name[] = "iarray_transform_3D";

    if (iarray_type (in_arr) != K_FLOAT)
    {
	fprintf (stderr, "%s: input image is not of type K_FLOAT\n",
		 function_name);
	return (FALSE);
    }
    if (iarray_type (out_arr) != K_FLOAT)
    {
	fprintf (stderr, "Input image is not of type K_FLOAT\n");
	a_prog_bug (function_name);
    }
    in_xdim = iarray_get_dim_desc (in_arr, 2);
    in_ydim = iarray_get_dim_desc (in_arr, 1);
    in_zdim = iarray_get_dim_desc (in_arr, 0);
    out_xdim = iarray_get_dim_desc (out_arr, 2);
    out_ydim = iarray_get_dim_desc (out_arr, 1);
    out_zdim = iarray_get_dim_desc (out_arr, 0);
    buffer = m_alloc_scratch (sizeof *buffer * 3 * in_xdim->length,
			      function_name);
    for (in_z = 0; in_z < in_zdim->length; ++in_z)
	for (in_y = 0; in_y < in_ydim->length; ++in_y)
	{
	    double *x, *y, *z;

	    /*  Transform in batches of length inX  */
	    x = buffer;
	    y = x + in_xdim->length;
	    z = y + in_xdim->length;
	    /*  Generate input pixel co-ordinates  */
	    for (in_x = 0; in_x < in_xdim->length; ++in_x, ++x, ++y, ++z)
	    {
		*x = in_x;
		*y = in_y;
		*z = in_z;
	    }
	    x = buffer;
	    y = x + in_xdim->length;
	    z = y + in_xdim->length;
	    /*  Transform to input world co-ordinates  */
	    wcs_astro_transform3 (in_ap, in_xdim->length,
				  in_xdim->name, x, FALSE,
				  in_ydim->name, y, FALSE,
				  in_zdim->name, z, FALSE,
				  0, NULL, NULL);
	    /*  Transform to output world co-ordinates  */
	    (*func) (info, in_xdim->length, x, y, z);
	    /*  Transform to output pixel co-ordinates  */
	    wcs_astro_transform3 (out_ap, in_xdim->length,
				  out_xdim->name, x, TRUE,
				  out_ydim->name, y, TRUE,
				  out_zdim->name, z, TRUE,
				  0, NULL, NULL);
	    /*  Loop through and copy pixels from input to output  */
	    for (in_x = 0; in_x < in_xdim->length; ++in_x, ++x, ++y, ++z)
	    {
		signed long out_x, out_y, out_z;
		float val = F3 (in_arr, in_z, in_y, in_x);
		float oldval;

		if ( DOUBLE_IS_BLANK (val) || DOUBLE_IS_BLANK (*x) ||
		     DOUBLE_IS_BLANK (*y) || DOUBLE_IS_BLANK (*z) ) continue;
		out_x = *x;
		out_y = *y;
		out_z = *z;
		if ( (out_x < 0) || (out_x >= out_xdim->length) ||
		     (out_y < 0) || (out_y >= out_ydim->length) ||
		     (out_z < 0) || (out_z >= out_zdim->length) ) continue;
		oldval = F3 (out_arr, out_z, out_y, out_x);
		if ( !DOUBLE_IS_BLANK (oldval) ) val += oldval;
		F3 (out_arr, out_z, out_y, out_x) = val;
	    }
	}
    m_free_scratch (buffer);
    return (TRUE);
}   /*  End Function iarray_transform_3D  */
