/*LINTLIBRARY*/
/*  bi-linear.c

    This code provides routines to compute a raw image from a data structure.

    Copyright (C) 1999-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains various utility routines for drawing into a raw
    image.


    Written by      Richard Gooch   12-JUL-1999: Copied from magnify.c

    Updated by      Richard Gooch   5-SEP-1999: Documentation update.

    Updated by      Richard Gooch   18-MAY-2000: Added <<flip>> parameter.

    Last updated by Richard Gooch   11-MAR-2001: Cope with min>max after iscale
  transform.


*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <os.h>
#ifdef HAS_GETRUSAGE
#  include <sys/time.h>
#  include <sys/resource.h>
#endif
#include <karma.h>
#include <karma_imw.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>


#define BUF_LENGTH 2048


/*  Private functions  */
STATIC_FUNCTION (double *alloc_values_buffer, (unsigned int num_values) );


/*  Private data  */


/*  Public functions follow  */


/*EXPERIMENTAL_FUNCTION*/
flag imw_to8_bilinear (unsigned char *out_image, flag flip,
		       iaddr out_hstride, iaddr out_vstride,
		       int out_width, int out_height, CONST char *inp_image,
		       CONST iaddr *inp_hoffsets, CONST iaddr *inp_voffsets,
		       int inp_width, int inp_height,
		       unsigned int inp_type, unsigned int conv_type,
		       unsigned int num_pixels,
		       CONST unsigned char *pixel_values,
		       unsigned char blank_pixel, unsigned char min_sat_pixel,
		       unsigned char max_sat_pixel,
		       double i_min, double i_max,
		       flag (*iscale_func) (), void *iscale_info)
/*  [SUMMARY] Convert generic image to 8 bit image, with resizing.
    [PURPOSE] This routine will convert an image from one format to an 8 bit
    image of pixels, permitting the input and output sizes to differ. If the
    input image is effectively shrunk, the input data is subsampled.
    <out_image> The output image will be written here.
    <flip> If TRUE, the output image is flipped vertically relative to the
    input image.
    <out_hstride> The stride between successive horizontal pixels (in bytes).
    <out_vstride> The stride between successive vertical pixels (in bytes).
    <out_width> The width of the output image.
    <out_height> The height of the output image.
    <inp_image> The input image data.
    <inp_hoffsets> The array of horizontal byte offsets.
    <inp_voffsets> The array of vertical byte offsets.
    <inp_width> The width of the input image.
    <inp_height> The height of the input image.
    <inp_type> The type of the input data.
    <conv_type> The input conversion type (when the input is complex). See
    [<ds_COMPLEX_CONVERSIONS>] for the list of allowed values.
    <num_pixels> The number of pixels in the pixel array.
    <pixel_values> The array of pixel values.
    <blank_pixel> The pixel value to be used when the intensity value is an
    undefined value.
    <min_sat_pixel> The pixel value to be used when the intensity value is
    below the minimum value.
    <max_sat_pixel> The pixel value to be used when the intensity value is
    above the maximum value.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <iscale_func> The function to be called when non-linear intensity scaling
    is required. If NULL, linear intensity scaling is used. The prototype
    function is [<imw_PROTO_iscale_func>].
    <iscale_info> A pointer to arbitrary information for <<iscale_func>>.
    <magnify_func> The function to call when magnification is required. This
    may be NULL, in which case a default pixel replication function is used.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag complex;
    int out_x, out_y, inp_x, inp_y;
    unsigned long pixval;
    float h_factor, v_factor;
    float zero = 0.0;
    float offset = 0.5;
    float one = 1.0;
    double d_mul;
    double d_toobig = TOOBIG;
    double scaled_min, scaled_max, tmp;
    double *inp_line;
    double *inp_values;
#ifdef HAS_GETRUSAGE
    long cputime_used;
    struct rusage rusage_before, rusage_after;
#endif
    double line_values[BUF_LENGTH * 2];
    static char function_name[] = "imw_to8_bilinear";

    if ( (inp_hoffsets == NULL) || (inp_voffsets == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if (i_min >= i_max)
    {
	fprintf (stderr, "i_max: %e  is not greater than i_min: %e\n",
		 i_max, i_min);
	a_prog_bug (function_name);
    }
    if (i_min < -TOOBIG)
    {
	fprintf (stderr, "i_min: %e  is less than %e\n", i_min, -TOOBIG);
	a_prog_bug (function_name);
    }
    if (i_max > TOOBIG)
    {
	fprintf (stderr, "i_max: %e  is greater than %e\n", i_max, TOOBIG);
	a_prog_bug (function_name);
    }
    if ( !ds_element_is_atomic (inp_type) )
    {
	fprintf (stderr, "Input image must be atomic\n");
	a_prog_bug (function_name);
    }
    if (inp_width > BUF_LENGTH)
    {
	fprintf (stderr, "Line width: %u to large\n", inp_width);
	a_prog_bug (function_name);
    }
    if ( imw_test_verbose () )
    {
	 fprintf (stderr, "%s started...\n", function_name);
	 fprintf (stderr, "out_size: %d %d  inp_size: %d %d\n",
		  out_width, out_height, inp_width, inp_height);
#ifdef HAS_GETRUSAGE
	 if (getrusage (RUSAGE_SELF, &rusage_before) == -1)
	 {
	     fprintf (stderr, "%s: error getting resource usage\t%s\n",
		      function_name, ERRSTRING);
	     exit (RV_SYS_ERROR);
	 }
#endif
    }
    /*  Allocate value buffers  */
    if ( ( inp_values = alloc_values_buffer (inp_width * inp_height) )== NULL )
	return (FALSE);
    h_factor = (float) inp_width / (float) out_width;
    v_factor = (float) inp_height / (float) out_height;
    if (iscale_func == NULL)
    {
	scaled_min = i_min;
	scaled_max = i_max;
    }
    else
    {
	if ( !(*iscale_func) (&scaled_min, 0, &i_min, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if ( !(*iscale_func) (&scaled_max, 0, &i_max, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity maximum\n",
		     function_name);
	}
	if (scaled_max < scaled_min) swap (scaled_min, scaled_max, tmp);
    }
    /*  First copy input image into buffer  */
    inp_line = inp_values;
    for (inp_y = 0; inp_y < inp_height; ++inp_y, inp_line += inp_width)
    {
	if ( !ds_get_scattered_elements (inp_image + inp_voffsets[inp_y],
					 inp_type, (uaddr *) inp_hoffsets,
					 line_values, &complex, inp_width) )
	{
	    fprintf (stderr, "Error converting data\n");
	    return (FALSE);
	}
	if (complex)
	{
	    ds_complex_to_real_1D (inp_line, 1, line_values,
				   inp_width, conv_type);
	    if (iscale_func != NULL)
	    {
		if ( !(*iscale_func) (inp_line, 1, inp_line, 1,
				      inp_width, i_min, i_max, iscale_info) )
		{
		    fprintf (stderr, "Error applying intensity scale\n");
		    return (FALSE);
		}
	    }
	}
	else if (iscale_func != NULL)
	{
	    if ( !(*iscale_func) (inp_line, 1, line_values, 2,
				  inp_width, i_min, i_max, iscale_info) )
		{
		    fprintf (stderr, "Error applying intensity scale\n");
		    return (FALSE);
		}
	}
	else
	{
	    for (inp_x = 0; inp_x < inp_width; ++inp_x)
	    {
		inp_line[inp_x] = line_values[inp_x * 2];
	    }
	}
    }
    d_mul = (num_pixels - 1) / (scaled_max - scaled_min);
    for (out_y = 0; out_y < out_height; ++out_y)
    {
	float xg, yg, x0, x1, y0, y1, dx, dy, dx_r, dy_r;
	double d_data;
	unsigned char *out_line;
	double *line0, *line1;

	if (flip)
	    out_line = out_image + (out_height - out_y - 1) * out_vstride;
	else out_line = out_image + out_y * out_vstride;
	/*  out_line  may be modified now  */
	yg = v_factor * (float) out_y;
	y0 = floor (yg);
	y1 = ceil (yg);
	if (y0 < zero) y0 = zero;
	if ( y1 > (float) (inp_height - 1) ) y1 = inp_height - 1;
	line0 = inp_values + (int) y0 * inp_width;
	line1 = inp_values + (int) y1 * inp_width;
	dy = yg - y0;
	dy_r = one - dy;
	for (out_x = 0; out_x < out_width; ++out_x, out_line += out_hstride)
	{
	    int toobig_count;
	    float val_00, val_10, val_01, val_11;

	    xg = h_factor * (float) out_x;
	    x0 = floor (xg);
	    x1 = ceil (xg);
	    if (x0 < zero) x0 = zero;
	    if ( x1 > (float) (inp_width - 1) ) x1 = inp_width - 1;
	    dx = xg - x0;
	    dx_r = one - dx;
	    /*  Take contribution from 4 neighbouring points  */
	    toobig_count = 0;
	    val_00 = line0[(int) x0];
	    if ( DOUBLE_IS_BLANK (val_00) )
	    {
		val_00 = zero;
		++toobig_count;
	    }
	    val_10 = line0[(int) x1];
	    if ( DOUBLE_IS_BLANK (val_10) )
	    {
		val_10 = zero;
		++toobig_count;
	    }
	    val_01 = line1[(int) x0];
	    if ( DOUBLE_IS_BLANK (val_01) )
	    {
		val_01 = zero;
		++toobig_count;
	    }
	    val_11 = line1[(int) x1];
	    if ( DOUBLE_IS_BLANK (val_11) )
	    {
		val_11 = zero;
		++toobig_count;
	    }
	    if (toobig_count > 3) d_data = TOOBIG;
	    else d_data = val_00 * dx_r * dy_r + val_10 * dx * dy_r +
		     val_01 * dx_r * dy + val_11 * dx * dy;
	    if (d_data < d_toobig)
	    {
		if (d_data < scaled_min) pixval = min_sat_pixel;
		else if (d_data > scaled_max) pixval = max_sat_pixel;
		else pixval = pixel_values[(int) ( (d_data - scaled_min)
						   * d_mul + offset )];
	    }
	    else pixval = blank_pixel;
	    *out_line = pixval;
	}
    }
#ifdef HAS_GETRUSAGE
    if ( imw_test_verbose () )
    {
	if (getrusage (RUSAGE_SELF, &rusage_after) == -1)
	{
	    fprintf (stderr, "%s: error getting resource usage\t%s\n",
		     function_name, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	cputime_used = 1000 * (rusage_after.ru_utime.tv_sec -
			       rusage_before.ru_utime.tv_sec);
	cputime_used += (rusage_after.ru_utime.tv_usec -
			 rusage_before.ru_utime.tv_usec) / 1000;
	fprintf (stderr, "%s: CPU time used: %ld milliseconds\n",
		 function_name, cputime_used);
    }
#endif
    return (TRUE);
}   /*  End Function imw_to8_bilinear  */


/*  Private functions follow  */

static double *alloc_values_buffer (unsigned int num_values)
{
    static unsigned int value_buf_len = 0;
    static double *values = NULL;
    static char function_name[] = "alloc_values_buffer";

    /*  Make sure value buffer is big enough  */
    if (value_buf_len < num_values)
    {
	/*  Buffer too small  */
	if (values != NULL)
	{
	    m_free (values);
	}
	value_buf_len = 0;
	if ( ( values = m_alloc (sizeof *values * num_values) ) == NULL )
	{
	    m_error_notify (function_name, "values buffer");
	    return (NULL);
	}
	value_buf_len = num_values;
    }
    return (values);
}   /*  End Function alloc_values_buffer  */
