/*LINTLIBRARY*/
/*  scmap_to24_lossy.c

    This code provides routines to compute a raw image from a data structure.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains various utility routines for drawing into a raw
    image.


    Written by      Richard Gooch   13-MAY-1997

    Updated by      Richard Gooch   18-MAY-1997

    Updated by      Richard Gooch   17-MAY-1998: More efficient conversion and
  silent handling of illegal values.

    Updated by      Richard Gooch   5-SEP-1999: Documentation update.

    Updated by      Richard Gooch   18-MAY-2000: Documentation update. Added
  <<flip>> parameter.

    Last updated by Richard Gooch   11-MAR-2001: Cope with min>max after iscale
  transform.


*/
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <karma.h>
#include <karma_imw.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>


/*  Private routines  */
STATIC_FUNCTION (double *alloc_inp_values_buffer, (unsigned int num_values) );
STATIC_FUNCTION (unsigned long *alloc_out_pixels_buffer,
		 (unsigned int num_pixels) );


/*  Public routines follow  */

/*UNPUBLISHED_FUNCTION*/
flag imw_scmap_to24_lossy (unsigned char *out_red_image,
			   unsigned char *out_green_image,
			   unsigned char *out_blue_image,
			   flag preserve_pad, flag flip,
			   iaddr out_hstride, iaddr out_vstride,
			   int out_width, int out_height,CONST char *inp_image,
			   CONST iaddr *inp_hoffsets,CONST iaddr *inp_voffsets,
			   int inp_width, int inp_height,
			   unsigned int inp_type, unsigned int conv_type,
			   unsigned int cmap_size, CONST unsigned long *pixels,
			   unsigned int red_offset, unsigned int green_offset,
			   unsigned int blue_offset,
			   unsigned int blank_pixel,
			   unsigned int min_sat_pixel,
			   unsigned int max_sat_pixel,
			   double i_min, double i_max,
			   flag (*iscale_func) (), void *iscale_info)
/*  [SUMMARY] Write generic image to a 24bit image using a software colourmap.
    [PURPOSE] This routine will write a generic PseudoColour image into a 24bit
    TrueColour image using a software colourmap to convert values to 24bit RGB
    values. The input and output sizes may be different.
    <out_red_image> The output red image components will be written here.
    <out_green_image> The output green image components will be written here.
    <out_blue_image> The output blue image components will be written here.
    <preserve_pad> If TRUE, padding bytes between output pixels will be
    preserved, else they may be overwritten for efficiency.
    <flip> If TRUE, the output image is flipped vertically relative to the
    input image.
    <out_hstride> The stride between successive horizontal pixels (in bytes).
    <out_vstride> The stride between successive vertical pixels (in bytes).
    <out_width> The width of the output image.
    <out_height> The height of the output image.
    <inp_image> The input image data.
    <inp_hoffsets> The array of horizontal byte offsets.
    <inp_voffsets> The array of vertical byte offsets.
    <inp_width> The width of the input image.
    <inp_height> The height of the input image.
    <inp_type> The type of the input data.
    <conv_type> The input conversion type (when the input is complex). See
    [<ds_COMPLEX_CONVERSIONS>] for the list of allowed values.
    <cmap_size> The number of colourmap entries.
    <pixels> The array of pixel values.
    <red_offset> The byte offset for the red component in the pixel values.
    <green_offset> The byte offset for the green component in the pixel values.
    <blue_offset> The byte offset for the blue component in the pixel values.
    <blank_pixel> The pixel value to be used when the intensity value is an
    undefined value.
    <min_sat_pixel> The pixel value to be used when the intensity value is
    below the minimum value.
    <max_sat_pixel> The pixel value to be used when the intensity value is
    above the maximum value.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <iscale_func> The function to be called when non-linear intensity scaling
    is required. If NULL, linear intensity scaling is used. The prototype
    function is [<imw_PROTO_iscale_func>].
    <iscale_info> A pointer to arbitrary information for <<iscale_func>>.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag complex;
    int out_x, out_y, inp_x, inp_y;
    unsigned long pixval;
    float h_factor, v_factor;
    float offset = 0.5;
    float tiny_offset = 1e-6;
    double d_mul, d_data;
    double d_toobig = TOOBIG;
    double scaled_min, scaled_max, tmp;
    CONST char *inp_line;
    CONST char *last_inp_line = NULL;
    CONST unsigned char *pixel_ptr;
    unsigned char *out_red_ptr, *out_green_ptr, *out_blue_ptr;
    unsigned long *out_pixels;
    double *inp_values;
    static char function_name[] = "imw_scmap_to24_lossy";

    if ( (inp_hoffsets == NULL) || (inp_voffsets == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if (i_min >= i_max)
    {
	fprintf (stderr, "i_max: %e  is not greater than i_min: %e\n",
		 i_max, i_min);
	a_prog_bug (function_name);
    }
    if ( !ds_element_is_atomic (inp_type) )
    {
	fprintf (stderr, "Input image must be atomic\n");
	a_prog_bug (function_name);
    }
    if ( imw_test_verbose () )
    {
	 fprintf (stderr, "%s started...\n", function_name);
	 fprintf (stderr, "out_size: %d %d  inp_size: %d %d\n",
		  out_width, out_height, inp_width, inp_height);
    }
    /*  Allocate value buffers  */
    if ( ( inp_values = alloc_inp_values_buffer (inp_width) ) == NULL )
    {
	return (FALSE);
    }
    if ( ( out_pixels = alloc_out_pixels_buffer (out_width) ) == NULL )
    {
	return (FALSE);
    }
    h_factor = (float) inp_width / (float) out_width;
    v_factor = (float) inp_height / (float) out_height;
    if (iscale_func == NULL)
    {
	scaled_min = i_min;
	scaled_max = i_max;
    }
    else
    {
	if ( !(*iscale_func) (&scaled_min, 0, &i_min, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if ( !(*iscale_func) (&scaled_max, 0, &i_max, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if (scaled_max < scaled_min) swap (scaled_min, scaled_max, tmp);
    }
    d_mul = (cmap_size - 1) / (scaled_max - scaled_min);
    for (out_y = 0; out_y < out_height; ++out_y)
    {
	inp_y = (int) (v_factor * (float) out_y + tiny_offset);
	inp_line = inp_image + inp_voffsets[inp_y];
	/*  The following test will prevent the reading and conversion of
	    duplicated input image lines. A time saver  */
	if (inp_line != last_inp_line)
	{
	    /*  Convert line of input data to generic data type  */
	    /*  Note the cast from (iaddr *) to (uaddr *) for the offset
		arrays. This is dodgy, but it should work.  */
	    if ( !ds_get_scattered_elements (inp_line, inp_type,
					     (uaddr *) inp_hoffsets,
					     inp_values, &complex,
					     (unsigned int) inp_width) )
	    {
		fprintf (stderr, "Error converting data\n");
		return (FALSE);
	    }
	    if (complex) ds_complex_to_real_1D (inp_values, 2, inp_values,
						(unsigned int) inp_width,
						conv_type);
	    if (iscale_func != NULL)
	    {
		if ( !(*iscale_func) (inp_values, 2, inp_values, 2,
				      (unsigned int) inp_width, i_min, i_max,
				      iscale_info) )
		{
		    fprintf (stderr, "Error applying intensity scale\n");
		    return (FALSE);
		}
	    }
	    /*  Now take converted input data and generate an output line. This
		algorithm should work for subsampled data (zoomed out) and
		pixel replicated data (zoomed in).  */
	    for (out_x = 0; out_x < out_width; ++out_x)
	    {
		inp_x = (int) (h_factor * (float) out_x + tiny_offset);
		d_data = inp_values[inp_x * 2];
		if (d_data < d_toobig)
		{
		    if (d_data < scaled_min) pixval = min_sat_pixel;
		    else if (d_data > scaled_max) pixval = max_sat_pixel;
		    else pixval = pixels[(int) ( (d_data - scaled_min)
						 * d_mul + offset )];
		    out_pixels[out_x] = pixval;
		}
		else out_pixels[out_x] = blank_pixel;
	    }
	}
	last_inp_line = inp_line;
	/*  Now ready to write the line  */
	if (flip) out_x = (out_height - out_y - 1) * out_vstride;
	else out_x = out_y * out_vstride;
	out_red_ptr = out_red_image + out_x;
	out_green_ptr = out_green_image + out_x;
	out_blue_ptr = out_blue_image + out_x;
	/*  Loop for each pixel  */
	for (out_x = 0; out_x < out_width;
	     ++out_x, out_red_ptr += out_hstride, out_green_ptr += out_hstride,
		 out_blue_ptr += out_hstride)
	{
	    pixel_ptr = (CONST unsigned char *) (out_pixels + out_x);
	    *out_red_ptr = pixel_ptr[red_offset];
	    *out_green_ptr = pixel_ptr[green_offset];
	    *out_blue_ptr = pixel_ptr[blue_offset];
	}
    }
    return (TRUE);
}   /*  End Function imw_scmap_to24_lossy  */

static double *alloc_inp_values_buffer (unsigned int num_values)
/*  This routine will allocate a buffer space for generic data values.
    The number of (DCOMPLEX) values to allocate must be given by  num_values  .
    The routine will return a pointer to the buffer space on success,
    else it returns NULL. The buffer is global and must NOT be deallocated.
*/
{
    static unsigned int value_buf_len = 0;
    static double *values = NULL;
    static char function_name[] = "alloc_inp_values_buffer";

    /*  Make sure value buffer is big enough  */
    if (value_buf_len < num_values)
    {
	/*  Buffer too small  */
	if (values != NULL)
	{
	    m_free ( (char *) values );
	}
	value_buf_len = 0;
	if ( ( values = (double *) m_alloc (sizeof *values * 2 * num_values) )
	     == NULL )
	{
	    m_error_notify (function_name, "values buffer");
	    return (NULL);
	}
	value_buf_len = num_values;
    }
    return (values);
}   /*  End Function alloc_inp_values_buffer  */

static unsigned long *alloc_out_pixels_buffer (unsigned int num_pixels)
/*  This routine will allocate a buffer space for output pixels.
    The number of pixels to allocate must be given by  num_pixels  .
    The routine will return a pointer to the buffer space on success,
    else it returns NULL. The buffer is global and must NOT be deallocated.
*/
{
    static unsigned int pixel_buf_len = 0;
    static unsigned long *pixels = NULL;
    static char function_name[] = "alloc_out_pixels_buffer";

    /*  Make sure pixel buffer is big enough  */
    if (pixel_buf_len < num_pixels)
    {
	/*  Buffer too small  */
	if (pixels != NULL)
	{
	    m_free ( (char *) pixels );
	}
	pixel_buf_len = 0;
	if ( ( pixels = (unsigned long *) m_alloc (sizeof *pixels * num_pixels)
	       ) == NULL )
	{
	    m_error_notify (function_name, "pixels buffer");
	    return (NULL);
	}
	pixel_buf_len = num_pixels;
    }
    return (pixels);
}   /*  End Function alloc_out_pixels_buffer  */
