/*LINTLIBRARY*/
/*  main.c

    This code provides Associative Array support.

    Copyright (C) 1995-1997  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the management of associative
  arrays.


    Written by      Richard Gooch   14-OCT-1995

    Updated by      Richard Gooch   17-DEC-1995: Created <aa_get_all_pairs> and
  <aa_get_pair_info> routines.

    Updated by      Richard Gooch   21-DEC-1995: Fixed bug in <aa_get_pair>
  routine: value was not returned.

    Updated by      Richard Gooch   22-DEC-1995: Added <<front>> parameter to
  <aa_put_pair> routine.

    Updated by      Richard Gooch   31-MAR-1996: Changed documentation style.

    Updated by      Richard Gooch   29-OCT-1996: Tidied up macros to keep
  Solaris 2 compiler happy.

    Last updated by Richard Gooch   29-JAN-1997: Renamed to <j> package.


*/
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <karma.h>
#include <karma_j.h>

#define LIST_MAGIC_NUMBER 298776298
#define PAIR_MAGIC_NUMBER 2084295498

#define VERIFY_LIST(list) {if (list == NULL) \
{fprintf (stderr, "NULL associative array passed\n"); \
 prog_bug (function_name); } \
if (list->magic_number != LIST_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid associative array object\n"); \
 prog_bug (function_name); } }

#define VERIFY_PAIR(apair) {if (apair == NULL) \
{fprintf (stderr, "NULL associative array pair passed\n"); \
 prog_bug (function_name); } \
if (apair->magic_number != PAIR_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid associative array pair object\n"); \
 prog_bug (function_name); } }


/*  Private structures  */
struct assocarray_type
{
    unsigned int magic_number;
    void *info;
    int (*key_compare_func) (CONST void *key1, CONST void *key2);
    void *(*key_copy_func) (CONST void *key, uaddr length, flag *ok);
    void (*key_destroy_func) (void *key);
    KJoinedPair first_pair;
    KJoinedPair last_pair;
};

struct assocarraypair_type
{
    unsigned int magic_number;
    KJoinedPairList array;
    void *key;
    uaddr key_length;
    void *value;
    uaddr value_length;
    void *(*value_copy_func) (CONST void *value, uaddr length, flag *ok);
    void (*value_destroy_func) (void *value);
    KJoinedPair next;
    KJoinedPair prev;
};


/*  Private functions  */
STATIC_FUNCTION (void merror_notify, (char *function_name, char *purpose) );
STATIC_FUNCTION (void prog_bug, (char *function_name) );
STATIC_FUNCTION (void mcopy, (char *dest, CONST char *source, uaddr length) );
STATIC_FUNCTION (void mclear, (char *memory, uaddr length) );
STATIC_FUNCTION (char *mdup, (CONST char *original, uaddr size) );


/*  Public functions follow  */


/*PUBLIC_FUNCTION*/
KJoinedPairList j_create (void *info,
			  int (*key_compare_func) (CONST void *key1,
						   CONST void *key2),
			  void *(*key_copy_func) (CONST void *key,
						  uaddr length, flag *ok),
			  void (*key_destroy_func) (void *key))
/*  [SUMMARY] Create a general purpose associative array of key-value pairs.
    <info> Arbitrary information to be stored with the associative array.
    <key_compare_func> The function used to compare two keys. The prototype
    function is [<j_PROTO_key_compare_func>].
    <key_copy_func> The function used to copy keys. The prototype function is
    [<j_PROTO_key_copy_func>].
    <key_destroy_func> The function used to destroy keys. The prototype
    function is [<j_PROTO_key_destroy_func>].
    [RETURNS] An associative array on success, else NULL.
*/
{
    KJoinedPairList list;
    static char function_name[] = "j_create";

    if ( ( list = (KJoinedPairList) malloc (sizeof *list) ) == NULL )
    {
	merror_notify (function_name, "associative array");
    }
    list->info = info;
    list->key_compare_func = key_compare_func;
    list->key_copy_func = key_copy_func;
    list->key_destroy_func = key_destroy_func;
    list->first_pair = NULL;
    list->last_pair = NULL;
    list->magic_number = LIST_MAGIC_NUMBER;
    return (list);
}   /*  End Function j_create  */

/*PUBLIC_FUNCTION*/
void j_destroy (KJoinedPairList list)
/*  [SUMMARY] Destroy an associative array.
    <list> The associative array.
    [RETURNS] Nothing.
*/
{
    KJoinedPair curr, next;
    static char function_name[] = "j_destroy";

    VERIFY_LIST (list);
    for (curr = list->first_pair; curr != NULL; curr = next)
    {
	next = curr->next;
	j_destroy_pair (curr);
    }
    free ( (char *) list );
}   /*  End Function j_destroy  */

/*PUBLIC_FUNCTION*/
void *j_get_info (KJoinedPairList list)
/*  [SUMMARY] Get information attached to associative array.
    [PURPOSE] This routine will get the arbitrary information pointer stored
    with an associative array.
    <list> The associative array.
    [RETURNS] The arbitrary information pointer.
*/
{
    static char function_name[] = "j_get_info";

    VERIFY_LIST (list);
    return (list->info);
}   /*  End Function j_get_info  */

/*PUBLIC_FUNCTION*/
KJoinedPair j_put_pair (KJoinedPairList list,
			CONST void *key, uaddr key_length,
			CONST void *value, uaddr value_length,
			void *(*value_copy_func) (CONST void *value,
						  uaddr length, flag *ok),
			void (*value_destroy_func) (void *value),
			unsigned int replacement_policy, flag front)
/*  [SUMMARY] Add a key-value pair to an associative array.
    <list> The associative array.
    <key> The key.
    <key_length> The length of the key, in bytes. If this is 0 and the
    <<key_copy_func>> is NULL, the key pointer is copied directly. If this is
    greater than zero and the <<key_copy_func>> is NULL, the specified number
    of bytes are copied.
    <value> The value.
    <value_length> The length of the value, in bytes. If this is 0 and the
    <<value_copy_func>> is NULL, the value pointer is copied directly. If this
    is greater than zero and the <<value_copy_func>> is NULL, the specified
    number of bytes are copied.
    <value_copy_func> The function used to copy values. The prototype function
    is [<j_PROTO_value_copy_func>].
    <value_destroy_func> The function used to destroy values. The prototype
    function is [<j_PROTO_value_destroy_func>].
    <replacement_policy> The policy to use when adding the pair. See
    [<j_REPLACEMENT_POLICIES>] for a list of legal values.
    <front> If TRUE and the key is new, place the pair at the front of the
    array. If FALSE and the key is new, place the pair at the end of the
    array.
    [RETURNS] A KJoinedPair object on success, else NULL.
*/
{
    KJoinedPair old, new;
    struct assocarraypair_type tmp_pair;
    flag ok;
    void *old_value;
    static char function_name[] = "j_put_pair";

    VERIFY_LIST (list);
    old = j_get_pair (list, key, &old_value);
    if (old == NULL)
    {
	if (replacement_policy == KJ_REPLACEMENT_POLICY_UPDATE) return (NULL);
    }
    else
    {
	if (replacement_policy == KJ_REPLACEMENT_POLICY_NEW) return (NULL);
    }
    if (old == NULL)
    {
	/*  Create new  */
	if ( ( new = (KJoinedPair) malloc (sizeof *new) ) == NULL )
	{
	    merror_notify (function_name, "new pair");
	    return (NULL);
	}
    }
    else
    {
	/*  Update old  */
	fprintf (stderr, "%s: unimplemented update\n", function_name);
	mcopy ( (char *) &tmp_pair, (char *) old, sizeof tmp_pair );
	tmp_pair.key = NULL;
	tmp_pair.key_length = 0;
	tmp_pair.value = NULL;
	tmp_pair.value_length = 0;
	new = &tmp_pair;
    }
    new->array = list;
    new->key_length = key_length;
    new->value_length = value_length;
    new->value_copy_func = value_copy_func;
    new->value_destroy_func = value_destroy_func;
    /*  Copy key  */
    if (list->key_copy_func == NULL)
    {
	if (key_length == 0)
	{
	    new->key = (void *) key;
	    ok = TRUE;
	}
	else
	{
	    if ( ( new->key = (void *) mdup (key, key_length) ) == NULL )
	    {
		ok = FALSE;
	    }
	    else ok = TRUE;
	}
    }
    else
    {
	new->key = (*list->key_copy_func) (key, key_length, &ok);
    }
    if (!ok)
    {
	merror_notify (function_name, "key");
	if (old == NULL) free ( (char *) new );
	return (NULL);
    }
    /*  Copy value  */
    if (new->value_copy_func == NULL)
    {
	if (value_length == 0)
	{
	    new->value = (void *) value;
	    ok = TRUE;
	}
	else
	{
	    if ( ( new->value = (void *) mdup (value,value_length) ) == NULL )
	    {
		ok = FALSE;
	    }
	    else ok = TRUE;
	}
    }
    else
    {
	new->value = (*new->value_copy_func) (value, value_length, &ok);
    }
    if (!ok)
    {
	merror_notify (function_name, "value");
	if (list->key_destroy_func == NULL)
	{
	    if (key_length != 0) free ( (char *) new->key );
	}
	else (*list->key_destroy_func) (new->key);
	if (old == NULL) free ( (char *) new );
	return (NULL);
    }
    if (old != NULL)
    {
	/*  Deallocate the old key and value  */
	if (old->array->key_destroy_func == NULL)
	{
	    if (old->key_length != 0) free ( (char *) old->key );
	}
	else (*old->array->key_destroy_func) (old->key);
	if (old->value_destroy_func == NULL)
	{
	    if (old->value_length != 0) free ( (char *) old->value );
	}
	else (*old->value_destroy_func) (old->value);
	old->key = tmp_pair.key;
	old->key_length = tmp_pair.key_length;
	old->value = tmp_pair.value;
	old->value_length = tmp_pair.value_length;
	return (old);
    }
    /*  New key: add to array  */
    new->magic_number = PAIR_MAGIC_NUMBER;
    if (front)
    {
	new->prev = NULL;
	new->next = list->first_pair;
	if (list->first_pair == NULL)
	{
	    /*  Create the first entry  */
	    list->last_pair = new;
	}
	else
	{
	    /*  Insert at beginning  */
	    list->first_pair->prev = new;
	}
	list->first_pair = new;
    }
    else
    {
	new->prev = list->last_pair;
	new->next = NULL;
	if (list->first_pair == NULL)
	{
	    /*  Create the first entry  */
	    list->first_pair = new;
	}
	else
	{
	    /*  Append to array  */
	    list->last_pair->next = new;
	}
	list->last_pair = new;
    }
    return (new);
}   /*  End Function j_put_pair  */

/*PUBLIC_FUNCTION*/
KJoinedPair j_get_pair (KJoinedPairList list, CONST void *key, void **value)
/*  [SUMMARY] Find a key-value pair in an associative array.
    <list> The associative array.
    <key> The key.
    <value> The value found will be written here on success.
    [RETURNS] A KJoinedPair object if the key was found, else NULL.
*/
{
    KJoinedPair pair;
    static char function_name[] = "j_get_pair";

    VERIFY_LIST (list);
    if (list->first_pair == NULL) return (NULL);
    for (pair = list->first_pair; pair != NULL; pair = pair->next)
    {
	if ( (*list->key_compare_func) (key, pair->key) == 0 )
	{
	    *value = pair->value;
	    return (pair);
	}
    }
    return (NULL);
}   /*  End Function j_get_pair  */

/*PUBLIC_FUNCTION*/
void j_destroy_pair (KJoinedPair pair)
/*  [SUMMARY] Remove a key-value pair from an associative array.
    [PURPOSE] This routine will remove a key-value pair from an associative
    array and will destroy all storage allocated for the pair.
    <pair> The key-value pair.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "j_destroy_pair";

    VERIFY_PAIR (pair);
    if (pair->prev != NULL) pair->prev->next = pair->next;
    if (pair->next != NULL) pair->next->prev = pair->prev;
    if (pair->array->first_pair == pair) pair->array->first_pair = pair->next;
    if (pair->array->last_pair == pair) pair->array->last_pair = pair->prev;
    if (pair->array->key_destroy_func != NULL)
	(*pair->array->key_destroy_func) (pair->key);
    if ( (pair->array->key_copy_func == NULL) && (pair->key_length != 0) )
	free ( (char *) pair->key );
    if (pair->value_destroy_func != NULL)
	(*pair->value_destroy_func) (pair->value);
    if ( (pair->value_copy_func == NULL) && (pair->value_length != 0) )
	free ( (char *) pair->value );
    mclear ( (char *) pair, sizeof *pair );
    free ( (char *) pair );
}   /*  End Function j_destroy_pair  */

/*PUBLIC_FUNCTION*/
KJoinedPair *j_get_all_pairs (KJoinedPairList list, unsigned int *num_pairs)
/*  [SUMMARY] Get all the key-value pairs on an associative array.
    <list> The associative array.
    <num_pairs> The number of pairs found is written here. If no pairs are
    found NULL is written here.
    [RETURNS] An array of key-value pairs on success, else NULL. If
    <<num_pairs>> is non-zero, an error occurred.
    [NOTE] The return value must be freed with the <<free>> function, not the
    [<m_free>] function.
*/
{
    KJoinedPair pair;
    unsigned int num;
    KJoinedPair *pairs;
    static char function_name[] = "j_get_all_pairs";

    VERIFY_LIST (list);
    *num_pairs = 0;
    if (list->first_pair == NULL) return (NULL);
    for (pair = list->first_pair, num = 0; pair != NULL;
	 pair = pair->next, ++num);
    if ( ( pairs = (KJoinedPair *) malloc (sizeof *pairs * num) ) == NULL )
    {
	merror_notify (function_name, "array of pairs");
	*num_pairs = 1;
	return (NULL);
    }
    for (pair = list->first_pair, num = 0; pair != NULL;
	 pair = pair->next, ++num)
    {
	pairs[num] = pair;
    }
    *num_pairs = num;
    return (pairs);
}   /*  End Function j_get_all_pairs  */

/*PUBLIC_FUNCTION*/
void j_get_pair_info (KJoinedPair pair,
		      void **key, uaddr *key_length,
		      void **value, uaddr *value_length)
/*  [SUMMARY] Get key and value information for a key-value pair.
    [PURPOSE] This routine will get the key and value information for a
    key-value pair in an associative array.
    <pair> The key-value pair.
    <key> The key will be written here.
    <key_length> The length of the key will be written here.
    <value> The value will be written here.
    <value_length> The length of the value will be written here.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "j_get_pair_info";

    VERIFY_PAIR (pair);
    *key = pair->key;
    *key_length = pair->key_length;
    *value = pair->value;
    *value_length = pair->value_length;
}   /*  End Function j_get_pair_info  */


/*  Private functions follow  */

static void merror_notify (char *function_name, char *purpose)
/*  [SUMMARY] Print memory error notification message.
    <function_name> The name of the function requesting the memory.
    <purpose> The purpose for the memory allocation.
    [RETURNS] Nothing.
*/
{
    fprintf (stderr, "Error allocating memory for: %s in function: %s\n",
	     purpose, function_name);
}   /*  End Function merror_notify   */

static void prog_bug (char *function_name)
/*  [SUMMARY] Print a bug message.
    [PURPOSE] This routine will print an error message to the standard error
    indicating that a bug occurred.
    <function_name> The name of the function in which the bug was noted.
    [RETURNS] The routine does not return. The process is aborted.
*/
{
    fprintf (stderr, "Program bug noted in function: %s%c\n",
	     function_name, BEL);
    /*fprintf (stderr, "Aborting.\n");*/
    abort ();
}   /*  End Function prog_bug   */

static void mcopy (char *dest, CONST char *source, uaddr length)
/*  [SUMMARY] Copy a block of memory.
    <dest> The destination block of memory.
    <source> The source block of memory.
    <length> The number of bytes to transfer. If this is zero, it is
    permissable for <<source>> and <<dest>> to be NULL.
    [NOTE] The routine copies long integers and chars.
    [RETURNS] Nothing.
*/
{
    while (length-- > 0) *dest++ = *source++;
}   /*  End Function mcopy  */

static void mclear (char *memory, uaddr length)
/*  [SUMMARY] Clear a block of memory
    <memory> The memory block to clear.
    <length> The length of the block (in bytes) to clear.
    [NOTE] The memory is cleared in long integers and chars.
    [RETURNS] Nothing.
*/
{
    while (length-- > 0) *memory++ = 0;
}   /*  End Function mclear  */

static char *mdup (CONST char *original, uaddr size)
/*  [SUMMARY] Duplicate a block of memory into a freshly allocated block.
    <original> The original block of memory.
    <size> The size in bytes of the block.
    [RETURNS] A pointer to a freshly allocated block which contains identical
    data as the orginal on success, else NULL.
*/
{
    char *duplicate;
    static char function_name[] = "mdup";

    if (original == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	prog_bug (function_name);
    }
    if ( ( duplicate = malloc (size) ) == NULL ) return (NULL);
    mcopy (duplicate, original, size);
    return (duplicate);
}   /*  End Function mdup  */
