/*LINTLIBRARY*/
/*  misc.c

    This code provides Kcolourmap objects.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for managing read/write
    colourcells in pseudocolour colourmaps.


    Written by      Richard Gooch   18-JUN-1997

    Updated by      Richard Gooch   13-SEP-1998: Used more efficient bit
  shifting operations instead of floating point. Also avoids roundoff errors.

    Last updated by Richard Gooch   20-MAR-2001: Made use of new function
  <kcmap_register_destroy_func>.


*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <karma.h>
#include <karma_kcmap.h>
#include <karma_m.h>
#include <karma_a.h>

#define MAGIC_NUMBER (unsigned int) 349083234

struct kdisplay_handle_type
{
    unsigned int magic_number;
    Kcolourmap cmap;
    unsigned long red_mask, green_mask, blue_mask;
    int red_shift, green_shift, blue_shift;  /*  Left shift 16bit to mask  */
};


/*  Private functions  */
STATIC_FUNCTION (unsigned int alloc_colours,
		 (unsigned int num_cells, unsigned long *pixel_values,
		  unsigned int min_cells, Kdisplay dpy_handle) );
STATIC_FUNCTION (void free_colours,
		 (unsigned int num_cells, unsigned long *pixel_values,
		  Kdisplay dpy_handle) );
STATIC_FUNCTION (void store_colours,
		 (unsigned int num_cells, unsigned long *pixel_values,
		  unsigned short *reds, unsigned short *greens,
		  unsigned short *blues, unsigned int stride,
		  Kdisplay dpy_handle) );
STATIC_FUNCTION (void get_location,
		 (Kdisplay dpy_handle, unsigned long *serv_hostaddr,
		  unsigned long *serv_display_num) );
STATIC_FUNCTION (void _kcmap_scmap_destroy, (Kdisplay dpy_handle) );


/*  Public functions follow  */

/*UNPUBLISHED_FUNCTION*/
Kcolourmap kcmap_create_scmap (CONST char *name, unsigned int num_cells,
			       flag tolerant,
			       unsigned long red_mask,unsigned long green_mask,
			       unsigned long blue_mask)
/*  [SUMMARY] Create a software colourmap.
    <name> The name of the function used to initialise the colour values. If
    this is NULL, the default "Greyscale1" function is used.
    <num_cells> The initial number of colourcells to allocate. This must not be
    less than 2.
    <tolerant> If TRUE the routine will try to allocate as many colourcells as
    possible (up to <<num_cells>>), else it will fail if it could not
    allocate all required colourcells.
    <red_mask> The red pixel mask.
    <green_mask> The green pixel mask.
    <blue_mask> The blue pixel mask.
    [RETURNS] A Kcolormap on success, else NULL.
*/
{
    Kcolourmap kcmap;
    Kdisplay dpy_handle;
    int rs, gs, bs;
    unsigned int count;
    static char function_name[] = "kcmap_create_scmap";

    if ( ( dpy_handle = m_alloc (sizeof *dpy_handle) ) == NULL )
    {
	m_error_notify (function_name, "display handle");
	return (NULL);
    }
    dpy_handle->magic_number = MAGIC_NUMBER;
    dpy_handle->cmap = NULL;
    dpy_handle->red_mask = red_mask;
    dpy_handle->green_mask = green_mask;
    dpy_handle->blue_mask = blue_mask;
    /*  Compute left shifts required to shift a 16bit colour into each mask.
	First compute bit position of 0 bit just left of masks  */
    for (count = 0, rs = 0, gs = 0, bs = 0; count < sizeof red_mask * 8;
	 ++count, red_mask = red_mask >> 1, green_mask = green_mask >> 1,
	     blue_mask = blue_mask >> 1)
    {
	if (red_mask != 0) ++rs;
	if (green_mask != 0) ++gs;
	if (blue_mask != 0) ++bs;
    }
    dpy_handle->red_shift = rs - 16;
    dpy_handle->green_shift = gs - 16;
    dpy_handle->blue_shift = bs - 16;
    /*  Can create the generic Kcolourmap, now that the shift data has been
	computed  */
    if ( ( kcmap = kcmap_va_create (name, num_cells, TRUE, dpy_handle,
				    alloc_colours, free_colours,
				    store_colours, get_location,
				    KCMAP_ATT_END) ) == NULL )
    {
	fprintf (stderr, "%s: error creating Karma PseudoColour cmap\n",
		 function_name);
	return (NULL);
    }
    dpy_handle->cmap = kcmap;
    kcmap_register_destroy_func (kcmap, _kcmap_scmap_destroy, dpy_handle);
    return (kcmap);
}   /*  End Function kcmap_create_scmap  */


/*  Private functions follow  */

static unsigned int alloc_colours (unsigned int num_cells,
				   unsigned long *pixel_values,
				   unsigned int min_cells, Kdisplay dpy_handle)
/*  [SUMMARY] Allocate colourcells.
    [PURPOSE] This routine will allocate a number of colourcells in a low level
    colourmap (e.g. using the Xlib routine XAllocColorCells).
    <num_cells> The number of colourcells to allocate.
    <pixel_values> A pointer to the array where the pixel values allocated will
    be written.
    <min_cells> The minimum number of colourcells to allocate. The routine will
    try to allocate at least this number of colourcells.
    <dpy_handle> The low level display handle. The meaning of this value
    depends on the lower level graphics library used.
    [RETURNS] The number of colourcells allocated.
*/
{
    return (num_cells);
}   /*  End Function alloc_colours  */

static void free_colours (unsigned int num_cells, unsigned long *pixel_values,
			  Kdisplay dpy_handle)
/*  [SUMMARY] Free a number of colourcells in a low level colourmap.
    <num_cells> The number of colourcells to free.
    <pixel_values> The array of pixel values (colourcells) to free.
    <dpy_handle> The low level display handle. The meaning of this value
    depends on the lower level graphics library used.
    [RETURNS] Nothing.
*/
{
}   /*  End Function free_colours  */

static void store_colours (unsigned int num_cells, unsigned long *pixel_values,
			   unsigned short *reds, unsigned short *greens,
			   unsigned short *blues, unsigned int stride,
			   Kdisplay dpy)
/*  [SUMMARY] Store colours into a low level colourmap.
    <num_cells> The number of colourcells to store.
    <pixel_values> The array of pixel values.
    <reds> The array of red intensity values.
    <greens> The array of green intensity values.
    <blues> The array of blue intensity values.
    <stride> The stride (in unsigned shorts) between intensity values in each
    array.
    <dpy> The low level display handle. The meaning of this value depends on
    the lower level graphics library used.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    unsigned long red, green, blue;
    static char function_name[] = "_kcmap_scmap_store_colours";

    if (dpy->magic_number != MAGIC_NUMBER)
    {
	fprintf (stderr, "Invalid display object\n");
	a_prog_bug (function_name);
    }
    for (count = 0; count < num_cells;
	 ++count, reds += stride, greens += stride, blues += stride)
    {
	red = *reds;
	green = *greens;
	blue = *blues;
	red = (dpy->red_shift < 0) ?
	    (red >> -dpy->red_shift) : (red << dpy->red_shift);
	green = (dpy->green_shift < 0) ?
	    (green >> -dpy->green_shift) : (green << dpy->green_shift);
	blue = (dpy->blue_shift < 0) ?
	    (blue >> -dpy->blue_shift) : (blue << dpy->blue_shift);
	pixel_values[count] = (red & dpy->red_mask) |
	    (green & dpy->green_mask) | (blue & dpy->blue_mask);
    }
    if (dpy->cmap == NULL) return;
    kcmap_notify_pixels_changed (dpy->cmap);
}   /*  End Function store_colours  */

static void get_location (Kdisplay dpy_handle, unsigned long *serv_hostaddr,
			  unsigned long *serv_display_num)
/*  [SUMMARY] Determine the location of the graphics display being used.
    <dpy_handle> The low level display handle. The meaning of this value
    depends on the lower level graphics library used.
    <serv_hostaddr> The Internet address of the host on which the display is
    running will be written here.
    <serv_display_num> The number of the display will be written here.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_kcmap_scmap_get_location";

    if (dpy_handle->magic_number != MAGIC_NUMBER)
    {
	fprintf (stderr, "Invalid display object\n");
	a_prog_bug (function_name);
    }
}   /*  End Function get_location  */

static void _kcmap_scmap_destroy (Kdisplay dpy_handle)
{
    static char function_name[] = "_kcmap_scmap_destroy";

    if (dpy_handle->magic_number != MAGIC_NUMBER)
    {
	fprintf (stderr, "Invalid display object\n");
	a_prog_bug (function_name);
    }
    dpy_handle->magic_number = 0;
    m_free (dpy_handle);
}   /*  End Function _kcmap_scmap_destroy  */
