/*LINTLIBRARY*/
/*  fault.c

    This code provides simple memory access rights testing routines.

    Copyright (C) 1997-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains routines for checking the accessibility of memory.


    Written by      Richard Gooch   29-DEC-1997

    Updated by      Richard Gooch   12-FEB-1998: Added note about performance.

    Last updated by Richard Gooch   10-JUN-1999: Always perform read test and
  restore values after write test.


*/
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <karma.h>
#include <karma_m.h>


/*  Private functions  */
STATIC_FUNCTION (void segv_handler, (int sig) );


/*  Private data  */
static volatile flag caught_segv = FALSE;
static sigjmp_buf env;


/*EXPERIMENTAL_FUNCTION*/
flag m_test_access (void *start, uaddr size, flag write)
/*  [SUMMARY] Test if a block of memory is accessible.
    <start> The starting address of the memory block.
    <size> The size (in bytes) of the memory block.
    <write> If TRUE a write test is performed in addition to a read test. This
    will temporarily change the contents of the memory block. The contents will
    be restored. If FALSE, only a read test is performed.
    [NOTE] The SIGSEGV signal handler is temporarily changed.
    [NOTE] The sigsetjmp(3) and siglongjmp(3) functions are used.
    [NOTE] Since this function makes calls to the Operating System, frequent
    use may degrade performance.
    [RETURNS] TRUE if the all requested tests succeeded, else FALSE if any test
    failed.
*/
{
    char val1, val2, dummy;
    struct sigaction new_action, old_action;
    char *c_start = (char *) start;
    static char function_name[] = "m_test_access";

    /*  Modify the SEGV signal handler  */
    sigemptyset (&new_action.sa_mask);
    new_action.sa_flags = 0;
#ifdef SA_RESTART
    new_action.sa_flags |= SA_RESTART;
#endif
    /*  Set up SIGSEGV handler  */
    new_action.sa_handler = segv_handler;
    if (sigaction (SIGSEGV, &new_action, &old_action) != 0)
    {
	fprintf (stderr, "%s: error setting SEGV handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    caught_segv = FALSE;
    sigsetjmp (env, 1);  /*  On SEGV we end back here  */
    if (!caught_segv)
    {
	/*  Do the read test  */
	/*  Use a function call to defeat optimisations  */
	m_copy (&val1, c_start, 1);
	m_copy (&val2, c_start + size - 1, 1);
	if (write)
	{
	    /*  Do the write test  */
	    /*  Use a function call to defeat optimisations  */
	    m_copy (c_start, &dummy, 1);
	    m_copy (c_start, &val1, 1);
	    m_copy (c_start + size - 1, &dummy, 1);
	    m_copy (c_start + size - 1, &val2, 1);
	}
    }
    /*  Restore old handler  */
    if (sigaction (SIGSEGV, &old_action, NULL) != 0)
    {
	fprintf (stderr, "%s: error restoring old SEGV handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    return (caught_segv ? FALSE : TRUE);
}   /*  End Function m_test_access  */


/*  Private functions follow  */

static void segv_handler (int sig)
{
    caught_segv = TRUE;
    siglongjmp (env, 0);
}   /*  End Function segv_handler  */
