/*LINTLIBRARY*/
/*  misc.c

    This code provides simple memory manipulation routines.

    Copyright (C) 1992-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains various memory operation routines for clearing, copying
    and comparing memory.


    Written by      Richard Gooch   12-AUG-1992

    Updated by      Richard Gooch   19-NOV-1992: Added  m_fill  routine.

    Updated by      Richard Gooch   2-JAN-1993: Fixed bug in  m_copy  routine.

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   25-MAY-1993: Added  m_cmp  .

    Updated by      Richard Gooch   19-SEP-1993: Fixed bug in  m_copy_blocks  .

    Updated by      Richard Gooch   20-MAY-1994: Added  CONST  declaration
  where appropriate.

    Updated by      Richard Gooch   15-OCT-1994: Fixed bug in <m_clear> which
  could result in misaligned accesses.

    Updated by      Richard Gooch   6-NOV-1994: Created  m_dup  .

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/m/misc.c

    Updated by      Richard Gooch   7-JAN-1995: Advertised <m_free_scratch>
  routine to documentation generator.

    Updated by      Richard Gooch   7-MAY-1995: Placate gcc -Wall

    Updated by      Richard Gooch   9-JUN-1995: Added #ifdef MACHINE_crayPVP.

    Updated by      Richard Gooch   12-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   15-JUN-1996: Created
  <m_copy_and_swap_blocks>.

    Updated by      Richard Gooch   13-OCT-1996: Trapped NULL pointer in
  <m_dup>.

    Updated by      Richard Gooch   21-FEB-1998: Changed to void *.

    Updated by      Richard Gooch   4-MAY-1998: Allow NULL pointer to <m_clear>
  if <<size>> less than 1.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   19-AUG-1999: Optimised function
  <m_copy_and_swap_blocks> for 2, 4 and 8 byte blocks.

    Last updated by Richard Gooch   8-MAY-2000: Created <m_calloc> function.


*/
#include <stdio.h>
#include <string.h>
#include <karma.h>
#include <karma_m.h>


/*  Private data  */
static flag scratch_block_reserved = FALSE;
static uaddr scratch_block_size = 0;
static char *scratch_block_ptr = NULL;


/*  Private functions  */
STATIC_FUNCTION (void prog_bug, (char *function_name) );


/*PUBLIC_FUNCTION*/
void m_clear (void *ptr, uaddr length)
/*  [SUMMARY] Clear a block of memory
    <ptr> The memory block to clear.
    <length> The length of the block (in bytes) to clear. If this is less than
    1 the routine does nothing. This is effective as a no-op.
    [NOTE] The memory is cleared in long integers and chars.
    [RETURNS] Nothing.
*/
{
    uaddr front_pad;
    uaddr ulong_count;
    uaddr uchar_count;
    uaddr num_ulong;
    uaddr num_uchar;
    char *memory = ptr;
    unsigned long *ulong_ptr;
    unsigned char *uchar_ptr;
    static char function_name[] = "m_clear";

    if (length < 1) return;
    if (memory == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	prog_bug (function_name);
    }
#ifdef MACHINE_crayPVP
    for (; length > 0; --length) *memory++ = 0;
#else
    /*  First clear stray bytes at start of block  */
    front_pad = (uaddr) memory % sizeof (unsigned long);
    front_pad = (front_pad > 0) ? (sizeof (unsigned long) - front_pad) : 0;
    for (uchar_count = front_pad; (uchar_count > 0) && (length > 0);
	 --uchar_count, --length)
	*memory++ = (char) NULL;
    num_ulong = length / sizeof (unsigned long);
    num_uchar = length - num_ulong * sizeof (unsigned long);
    ulong_ptr = (unsigned long *) memory;
    for (ulong_count = 0; ulong_count < num_ulong; ulong_count++)
        *ulong_ptr++ = (unsigned long) NULL;
    uchar_ptr = (unsigned char *) ulong_ptr;
    for (uchar_count = 0; uchar_count < num_uchar; ++uchar_count)
        *uchar_ptr++ = (unsigned char) NULL;
#endif
}   /*  End Function m_clear  */

/*PUBLIC_FUNCTION*/
void *m_calloc (uaddr size)
/*  [SUMMARY] Allocate Virtual Memory and initialise with zeros.
    <size> The number of bytes to allocate.
    [MT-LEVEL] Safe under POSIX threads.
    [RETURNS] A pointer to the memory on success, else NULL.
*/
{
    void *ptr;

    if ( ( ptr = m_alloc (size) ) == NULL ) return (NULL);
    m_clear (ptr, size);
    return (ptr);
}   /*  End Function m_calloc  */

/*PUBLIC_FUNCTION*/
void m_copy (void *dest, CONST void *source, uaddr length)
/*  [SUMMARY] Copy a block of memory.
    <dest> The destination block of memory.
    <source> The source block of memory.
    <length> The number of bytes to transfer. If this is zero, it is
    permissable for <<source>> and <<dest>> to be NULL.
    [NOTE] The routine copies long integers and chars.
    [RETURNS] Nothing.
*/
{
    uaddr ulong_count;
    unsigned int char_count;
    uaddr num_ulong;
    unsigned int num_uchar;
    CONST unsigned long *source_ulong;
    unsigned long *dest_ulong;
    CONST unsigned char *source_uchar = source;
    unsigned char *dest_uchar = dest;
    static char function_name[] = "m_copy";

    if (length < 1)
    {
	return;
    }
    if (dest_uchar == NULL)
    {
	fprintf (stderr, "NULL destination pointer passed\n");
	prog_bug (function_name);
    }
    if (source_uchar == NULL)
    {
	fprintf (stderr, "NULL source pointer passed\n");
	prog_bug (function_name);
    }
#ifdef MACHINE_crayPVP
    for (; length > 0; --length) *dest_uchar++ = *source_uchar++;
    return;
#else
    if (length < 16)
    {
	/*  Save a few % operations  */
	for (; length > 0; --length) *dest_uchar++ = *source_uchar++;
	return;
    }
    if ( (uaddr) dest_uchar % sizeof (unsigned long) !=
	 (uaddr) source_uchar % sizeof (unsigned long) )
    {
	/*  Blocks have different misalignment: byte copy  */
	for (; length > 0; --length)
	    *dest_uchar++ = *source_uchar++;
	return;
    }
    /*  First copy stray bytes at start of blocks  */
    for (char_count = sizeof (unsigned long) -
	 (uaddr) dest_uchar % sizeof (unsigned long);
	 (char_count > 0) && (length > 0);
	 --char_count, --length)
	*dest_uchar++ = *source_uchar++;
    num_ulong = length / sizeof (unsigned long);
    num_uchar = length - num_ulong * sizeof (unsigned long);
    source_ulong = (unsigned long *) source_uchar;
    dest_ulong = (unsigned long *) dest_uchar;
    for (ulong_count = 0; ulong_count < num_ulong; ++ulong_count)
        *dest_ulong++ = *source_ulong++;
    source_uchar = (unsigned char *) source_ulong;
    dest_uchar = (unsigned char *) dest_ulong;
    for (char_count = 0; char_count < num_uchar; ++char_count)
        *dest_uchar++ = *source_uchar++;
#endif
}   /*  End Function m_copy   */

/*PUBLIC_FUNCTION*/
void m_copy_blocks (void *dest, CONST void *source, unsigned int dest_stride,
		    unsigned int source_stride, unsigned int block_size,
		    unsigned int num_blocks)
/*  [SUMMARY] Copy multiple blocks of data.
    <dest> The destination for the first block copy.
    <source> The source for the first block copy.
    <dest_stride> The spacing (in bytes) between destintion blocks.
    <source_stride> The spacing (in bytes) between source blocks.
    <block_size> The size of each block (in bytes).
    <num_blocks> The number of blocks to copy.
    [RETURNS] Nothing.
*/
{
    unsigned int block_count;
    unsigned int blk_size;
    unsigned int ulong_count;
    unsigned int char_count;
    unsigned int num_ulong;
    unsigned int num_char;
    CONST char *c_source = source;
    char *c_dest = dest;
    CONST unsigned long *source_ulong;
    unsigned long *dest_ulong;
    CONST char *source_char = source;
    char *dest_char = dest;
    static char function_name[] = "m_copy_blocks";

    if ( (source_char == NULL) || (dest_char == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	prog_bug (function_name);
    }
    if ( (dest_stride < 1) || (source_stride < 1) )
    {
	fprintf (stderr, "Strides must be greater than zero\n");
	prog_bug (function_name);
    }
#ifdef MACHINE_crayPVP
    for (block_count = 0; block_count < num_blocks;
	 ++block_count, dest_char += dest_stride,
	     source_char += source_stride)
    {
	for (char_count = 0; char_count < block_size; ++char_count)
	{
	    dest_char[char_count] = source_char[char_count];
	}
    }
#else
    /*  Iterate through the blocks to be copied  */
    for (block_count = 0; block_count < num_blocks;
	 ++block_count, c_dest += dest_stride, c_source += source_stride)
    {
	source_char = c_source;
	dest_char = c_dest;
	if ( (unsigned long) dest_char % sizeof (unsigned long) !=
	     (unsigned long) source_char % sizeof (unsigned long) )
	{
	    /*  Blocks have different misalignment: byte copy  */
	    for (char_count = 0; char_count < block_size; ++char_count)
	         *dest_char++ = *source_char++;
	    continue;
	}
	/*  First copy stray bytes at start of block  */
	blk_size = block_size;
	for (char_count = sizeof (unsigned long) -
	     (unsigned long) dest_char % sizeof (unsigned long);
	     (char_count > 0) && (blk_size > 0);
	     --char_count, --blk_size)
	    *dest_char++ = *source_char++;
	/*  Copy next section of block using long copies  */
	num_ulong = blk_size / sizeof (unsigned long);
	num_char = blk_size - num_ulong * sizeof (unsigned long);
	source_ulong = (unsigned long *) source_char;
	dest_ulong = (unsigned long *) dest_char;
	for (ulong_count = 0; ulong_count < num_ulong; ulong_count++)
	    *dest_ulong++ = *source_ulong++;
	source_char = (char *) source_ulong;
	dest_char = (char *) dest_ulong;
	/*  Copy last section of block using byte copies  */
	for (char_count = 0; char_count < num_char; char_count++)
	    *dest_char++ = *source_char++;
    }
#endif
}   /*  End Function m_copy_blocks  */

/*PUBLIC_FUNCTION*/
void m_copy_and_swap_blocks (void *dest, CONST void *source, uaddr dest_stride,
			     uaddr source_stride, uaddr block_size,
			     uaddr num_blocks)
/*  [SUMMARY] Copy and byte-swap multiple blocks of data.
    <dest> The destination for the first block copy.
    <source> The source for the first block copy. If this is NULL or the same
    as <<dest>> the swap is performed in-situ, and <<source_stride>> is ignored
    <dest_stride> The spacing (in bytes) between destintion blocks.
    <source_stride> The spacing (in bytes) between source blocks.
    <block_size> The size of each block (in bytes).
    <num_blocks> The number of blocks to copy and swap.
    [RETURNS] Nothing.
*/
{
    uaddr count, iter;
    char tmp;
    CONST char *c_source = source;
    char *c_dest = dest;
    char *dest_ptr, *d_up_ptr;
    CONST char *source_ptr;
    static char function_name[] = "m_copy_and_swap_blocks";

    if (c_dest == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	prog_bug (function_name);
    }
    if (dest_stride < 1)
    {
	fprintf (stderr, "dest_stride must be greater than zero\n");
	prog_bug (function_name);
    }
    if (c_source == c_dest) c_source = NULL;
    if (c_source == NULL)
    {
	/*  In-situ  */
	iter = block_size / 2;
	/*  Loop over blocks  */
	for (; num_blocks > 0; --num_blocks, c_dest += dest_stride)
	{
	    dest_ptr = c_dest;
	    d_up_ptr = c_dest + block_size - 1;
	    /*  Swap-copy bytes  */
	    for (count = 0; count < iter; ++count, ++dest_ptr, --d_up_ptr)
	    {
		tmp = *dest_ptr;
		*dest_ptr = *d_up_ptr;
		*d_up_ptr = tmp;
	    }
	}
	return;
    }
    /*  Swap different blocks  */
    if (source_stride < 1)
    {
	fprintf (stderr, "source_stride must be greater than zero\n");
	prog_bug (function_name);
    }
    switch (block_size)
    {
      case 2:
	/*  Loop over blocks  */
	for (; num_blocks > 0;
	     --num_blocks, c_dest += dest_stride, c_source += source_stride)
	{
	    c_dest[1] = c_source[0];
	    c_dest[0] = c_source[1];
	}
	break;
      case 4:
	/*  Loop over blocks  */
	for (; num_blocks > 0;
	     --num_blocks, c_dest += dest_stride, c_source += source_stride)
	{
	    c_dest[3] = c_source[0];
	    c_dest[2] = c_source[1];
	    c_dest[1] = c_source[2];
	    c_dest[0] = c_source[3];
	}
	break;
      case 8:
	/*  Loop over blocks  */
	for (; num_blocks > 0;
	     --num_blocks, c_dest += dest_stride, c_source += source_stride)
	{
	    c_dest[7] = c_source[0];
	    c_dest[6] = c_source[1];
	    c_dest[5] = c_source[2];
	    c_dest[4] = c_source[3];
	    c_dest[3] = c_source[4];
	    c_dest[2] = c_source[5];
	    c_dest[1] = c_source[6];
	    c_dest[0] = c_source[7];
	}
	break;
      default:
	/*  Loop over blocks  */
	for (; num_blocks > 0;
	     --num_blocks, c_dest += dest_stride, c_source += source_stride)
	{
	    dest_ptr = c_dest;
	    source_ptr = c_source + block_size - 1;
	    /*  Swap-copy bytes  */
	    for (count = 0; count < block_size; ++count)
		*dest_ptr++ = *source_ptr--;
	}
	break;
    }
}   /*  End Function m_copy_and_swap_blocks  */

/*PUBLIC_FUNCTION*/
void m_fill (void *dest, uaddr stride, CONST void *source,
	     uaddr size, unsigned int num)
/*  [SUMMARY] Fill memory blocks with a specified value.
    <dest> The destination.
    <stride> The stride (in bytes) of destination blocks.
    <source> The fill block.
    <size> The size (in bytes) of the fill value block.
    <num> The number of destination blocks to fill.
    [RETURNS] Nothing.
*/
{
    uaddr byte_count;
    CONST char *c_source = source;
    char *c_dest = dest;
    static char function_name[] = "m_fill";

    if ( (c_dest == NULL) || (c_source == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	prog_bug (function_name);
    }
    if (stride < size)
    {
	fprintf (stderr, "stride: %lu must not be less than size: %lu\n",
		 stride, size);
	prog_bug (function_name);
    }
    while (num-- > 0)
    {
	for (byte_count = 0; byte_count < size; ++byte_count)
	{
	    c_dest[byte_count] = c_source[byte_count];
	}
	c_dest += stride;
    }
}   /*  End Function m_fill  */

/*PUBLIC_FUNCTION*/
flag m_cmp (CONST void *block1, CONST void *block2, uaddr length)
/*  [SUMMARY] Compare two blocks of memory.
    <block1> The first memory block.
    <block2> The second memory block.
    <length> The number of bytes to compare.
    [NOTE] The routine compares long integers and chars.
    [RETURNS] TRUE if the blocks are equal, else FALSE.
*/
{
    uaddr ulong_count;
    unsigned int char_count;
    uaddr num_ulong;
    unsigned int num_uchar;
    CONST char *c_block1 = block1;
    CONST char *c_block2 = block2;
    CONST unsigned long *block2_ulong;
    CONST unsigned long *block1_ulong;
    CONST unsigned char *block2_uchar;
    CONST unsigned char *block1_uchar;
    static char function_name[] = "m_cmp";

    if ( (c_block1 == NULL) || (c_block2 == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	prog_bug (function_name);
    }
    if (length < 1)
    {
	fprintf (stderr, "length  must be greater than zero\n");
	prog_bug (function_name);
    }
#ifdef MACHINE_crayPVP
    if (memcmp (c_block1, c_block2, length) == 0) return (TRUE);
    return (FALSE);
#else
    if ( (uaddr) c_block1 % sizeof (unsigned long) !=
	 (uaddr) c_block2 % sizeof (unsigned long) )
    {
	/*  Blocks have different misalignment: byte compare  */
	for (; length > 0; --length)
	{
	    if (*c_block1++ != *c_block2++) return (FALSE);
	}
	return (TRUE);
    }
    /*  First compare stray bytes at start of blocks  */
    for (char_count = sizeof (unsigned long) -
	 (uaddr) c_block1 % sizeof (unsigned long);
	 (char_count > 0) && (length > 0);
	 --char_count, --length)
    {
	if (*c_block1++ != *c_block2++) return (FALSE);
    }
    num_ulong = length / sizeof (unsigned long);
    num_uchar = length - num_ulong * sizeof (unsigned long);
    block2_ulong = (unsigned long *) c_block2;
    block1_ulong = (unsigned long *) c_block1;
    for (ulong_count = 0; ulong_count < num_ulong; ++ulong_count)
    {
	if (*block1_ulong++ != *block2_ulong++) return (FALSE);
    }
    block2_uchar = (unsigned char *) block2_ulong;
    block1_uchar = (unsigned char *) block1_ulong;
    for (char_count = 0; char_count < num_uchar; ++char_count)
    {
	if (*block1_uchar++ != *block2_uchar++) return (FALSE);
    }
    return (TRUE);
#endif
}   /*  End Function m_cmp   */

/*PUBLIC_FUNCTION*/
void *m_dup (CONST void *original, uaddr size)
/*  [SUMMARY] Duplicate a block of memory into a freshly allocated block.
    <original> The original block of memory.
    <size> The size in bytes of the block.
    [RETURNS] A pointer to a freshly allocated block which contains identical
    data as the orginal on success, else NULL.
*/
{
    char *duplicate;
    static char function_name[] = "m_dup";

    if (original == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	prog_bug (function_name);
    }
    if ( ( duplicate = m_alloc (size) ) == NULL ) return (NULL);
    m_copy (duplicate, original, size);
    return (duplicate);
}   /*  End Function m_dup   */

/*PUBLIC_FUNCTION*/
void *m_alloc_scratch (uaddr size, char *function_name)
/*  [SUMMARY] Allocate a block of scratch memory.
    [PURPOSE] This routine will allocate a block of scratch memory, which may
    be re-used by many different routines. The block is reserved until a call
    is made to [<m_free_scratch>].
    <size> The minimum size in bytes of the scratch block.
    <function_name> If the memory block is already reserved and this is not
    NULL the string is printed and the process aborts.
    If the memory block is already reserved and this is NULL the routine fails
    normally.
    [RETURNS] A pointer on success to a dynamically allocated block of memory
    which is valid until the next call to this routine, else NULL.
*/
{
    if (scratch_block_reserved)
    {
	fprintf (stderr, "Scratch block already reserved\n");
	if (function_name == NULL) return (NULL);
	prog_bug (function_name);
    }
    if (scratch_block_ptr == NULL)
    {
	/*  First time allocation  */
	if ( ( scratch_block_ptr = m_alloc (size) ) == NULL ) return (NULL);
	scratch_block_size = size;
	scratch_block_reserved = TRUE;
	return (scratch_block_ptr);
    }
    if (scratch_block_size >= size)
    {
	scratch_block_reserved = TRUE;
	return (scratch_block_ptr);
    }
    /*  Existing block is too small  */
    m_free (scratch_block_ptr);
    scratch_block_size = 0;
    if ( ( scratch_block_ptr = m_alloc (size) ) == NULL ) return (NULL);
    scratch_block_size = size;
    scratch_block_reserved = TRUE;
    return (scratch_block_ptr);
}   /*  End Function m_alloc_scratch  */

/*PUBLIC_FUNCTION*/
void m_free_scratch ()
/*  [SUMMARY] Free the scratch memory.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "m_free_scratch";

    if (!scratch_block_reserved)
    {
	fprintf (stderr, "Scratch block not reserved\n");
	prog_bug (function_name);
    }
    scratch_block_reserved = FALSE;
}   /*  End Function m_free_scratch  */


/*  Private routines follow  */

static void prog_bug (char *function_name)
{
    fprintf (stderr, "Program bug noted in function: %s\n", function_name);
    abort ();
}   /*  End Function prog_bug   */
