/*LINTLIBRARY*/
/*  mmap.c

    This code provides memory mapping/allocation routines.

    Copyright (C) 1998-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*


    Written by      Richard Gooch   12-MAY-1998

    Updated by      Richard Gooch   13-MAY-1998: Added support for creating
  file in $TMPDIR.

    Updated by      Richard Gooch   23-JUN-1998: Added support for POSIX SHM
  objects.

    Updated by      Richard Gooch   2-NOV-1998: Removed cast to caddr_t in call
  to <<mmap>>.

    Last updated by Richard Gooch   13-NOV-2001: Added test for OS_Darwin.


*/
#define _POSIX_C_SOURCE 199506  /*  Needed for Solaris 2.6 for shm_open()    */
#ifndef __EXTENSIONS__
#  define __EXTENSIONS__        /*  Needed for Solaris 2.5 to compile at all */
#endif
#ifdef OS_OSF1                  /*  Stupid bastards                          */
#  define _OSF_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <karma.h>
#include <karma_m.h>
#include <os.h>
#if defined(OS_SunOS) || defined(OS_ConvexOS) || defined(OS_Linux)
#  include <sys/vfs.h>
#endif
#if defined(OS_ULTRIX) || defined(OS_OSF1) || defined(OS_Darwin)
#  include <sys/mount.h>
#endif
#if defined(OS_AIX)
#  include <sys/statfs.h>
#endif
#if defined(OS_Solaris) || defined(OS_IRIX)
#  include <sys/statvfs.h>
#endif
#ifdef HAS_MMAP
#  include <sys/mman.h>
#  ifdef OS_ULTRIX
caddr_t mmap ();
#  endif
#endif


#if defined(OS_Solaris) || defined(OS_IRIX)
#  define statfs statvfs
#endif


/*  Private functions  */
#if defined(HAS_MMAP) && defined(CAN_PASS_FD)
STATIC_FUNCTION (int _m_mmap_create_file, (uaddr size) );
#endif

/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void *m_map_fd (int *fd, uaddr size, flag writable, flag update_on_write,
		flag warn_unsupported)
/*  [SUMMARY] Map a file descriptor
    <fd> A pointer to the file descriptor to map. If the descriptor is less
    than 0 then an appropriately chosen file is opened and the newly opened
    file descriptor is written here.
    <size> The size of the region to map.
    <writable> If the mapped pages are to be writable, this must be TRUE. If
    this is FALSE and the memory pages are written to, a segmentation fault
    occurs.
    <update_on_write> If the disc file should be updated when the memory pages
    are written to, this must be TRUE. If this is FALSE, then a write to a
    memory page causes the page to be copied into swap space and the process
    retains a private copy of the page from this point on.
    [NOTE] If <<update_on_write>> is FALSE and <<writable>> is TRUE, then some
    systems require the allocation of normal virtual memory equal to the size
    of the disc file at the time of mapping, while others will dynamically
    allocate memory from swap space as pages are written into. In the latter
    case, some systems will cause a segmentation fault if swap space is
    exhausted, while other systems wait for space to be freed.
    <warn_unsupported> If TRUE, attempts to use unsupported features will
    generate a warning message.
    [RETURNS] A pointer to the mapped file on success, else NULL.
*/
{
#ifdef HAS_MMAP
    flag opened_fd = FALSE;
    int lfd = *fd;
    int mmap_flags = 0;
    int mmap_prot = PROT_READ;
#  ifdef OS_ConvexOS
    unsigned int map_len;
#  endif
    void *addr;
    /*static char function_name[] = "m_map_fd";*/

    if (lfd < 0)
    {
	/*  No file provided: need to open one  */
#  ifdef CAN_PASS_FD
	if (update_on_write)
	{
	    /*  Prefer not to use /dev/zero, as no-one implements it usefully
	     */
	    lfd = _m_mmap_create_file (size);
	}
	if (lfd < 0)
#  endif
	    lfd = open ("/dev/zero", O_RDWR, 0);
	if (lfd < 0) return (NULL);
	opened_fd = TRUE;
    }
#  ifdef MAP_VARIABLE
    mmap_flags |= MAP_VARIABLE;
#  endif
    if (writable)
    {
	/*  Write enable pages  */
	mmap_prot |= PROT_WRITE;
	if (update_on_write)
	{
	    /*  Shared  */
#  ifdef MAP_SHARED
	    mmap_flags |= MAP_SHARED;
#  else
	    if (warn_unsupported)
		fprintf (stderr, "Share-on-write mapping not supported\n");
	    return (NULL);
#  endif
	}
	else
	{
	    /*  Copy-on-write mode required  */
#  ifdef MAP_PRIVATE
	    mmap_flags |= MAP_PRIVATE;
#  else
	    if (warn_unsupported)
		fprintf (stderr, "Copy-on-write mapping not supported\n");
	    return (NULL);
#  endif
	}
    }
    else
    {
	/*  Read only  */
#  ifdef MAP_SHARED
	mmap_flags |= MAP_SHARED;
#  endif
#  ifdef MAP_FILE
	mmap_flags |= MAP_FILE;
#  endif
    }
#  ifdef OS_ConvexOS
    map_len = size;
#  endif
    if ( ( addr = mmap ( NULL,
#  ifdef OS_ConvexOS  /*  Why the fuck to they have to be different?  */
			 &map_len,
#  else
			 (size_t) size,
#  endif
			 mmap_prot, mmap_flags, lfd, (off_t) 0 ) )
#  ifdef MAP_FAILED
	 == MAP_FAILED )
#  else
	 == (caddr_t) -1 )
#  endif
    {
	/*  Failed  */
	if (opened_fd) close (lfd);
	return (NULL);
    }
    *fd = lfd;
    return (addr);
#else  /*  HAS_MMAP  */
    if (warn_unsupported) fprintf (stderr, "Memory mapping not supported\n");
    return (NULL);
#endif  /*  !HAS_MMAP  */
}   /*  End Function m_map_fd  */

/*EXPERIMENTAL_FUNCTION*/
flag m_unmap (void *addr, uaddr size)
/*  [SUMMARY] Unmap a region of memory.
    <addr> The start address of the region.
    <size> The size of the region.
    [RETURNS] TRUE on success, else FALSE.
*/
{
#ifdef HAS_MMAP
    return (munmap (addr, size) < 0) ? FALSE : TRUE;
#else
    return (FALSE);
#endif
}   /*  End Function m_unmap  */


/*  Private functions follow  */

#if defined(HAS_MMAP) && defined(CAN_PASS_FD)
static CONST char *tmpdir = NULL;

static int _m_mmap_open_posix_shm (uaddr size)
/*  [SUMMARY] Create a unique temporary POSIX SHM file. This is unlinked.
    <size> The size of the file.
    [RETURNS] A file descriptor on success, else -1.
*/
{
#  ifdef _POSIX_SHARED_MEMORY_OBJECTS
    int fd, pid, count, errnum;
    char file[STRING_LENGTH];

    pid = getpid ();
    /*  Loop creating temporary file until an exclusive one is created  */
    for (count = 0; count >= 0; ++count)
    {
	sprintf (file, "/karma-mmap-%d-%d", pid, count);
	if ( ( fd = shm_open (file, O_RDWR | O_CREAT | O_EXCL,
			      S_IRUSR | S_IWUSR) ) < 0 )
	{
	    if (errno == EEXIST) continue;
	    if (errno != EACCES) return (-1);
	    /*  Some stupid implementations (IRIX) have the file sitting in the
		normal FS namespace, resulting in silly protection problems. So
		if this happens, try again in TMPDIR  */
	    sprintf (file, "%s/karma-mmap-%d-%d", tmpdir, pid, count);
	    if ( ( fd = shm_open (file, O_RDWR | O_CREAT | O_EXCL,
				  S_IRUSR | S_IWUSR) ) < 0 )
	    {
		if (errno == EEXIST) continue;
		return (-1);
	    }
	}
	if ( (unlink (file) != 0) || (ftruncate (fd, size) == -1) )
	{
	    errnum = errno;
	    close (fd);
	    errno = errnum;
	    return (-1);
	}
	return (fd);
    }
#  endif
    errno = 0;
    return (-1);
}

static int _m_mmap_create_file (uaddr size)
/*  [SUMMARY] Create a unique temporary file. This is unlinked.
    <size> The size of the file.
    [RETURNS] A file descriptor on success, else -1.
*/
{
    char dummy;
    int fd, pid, count, errnum;
    long nblocks, bavail;
    struct statfs statfsbuf;
    char file[STRING_LENGTH];

    if (tmpdir == NULL)
    {
	tmpdir = getenv ("TMPDIR");
	if (tmpdir == NULL) tmpdir = "/tmp";
    }
    /*  Prefer POSIX SHM  */
    fd = _m_mmap_open_posix_shm (size);
    if (fd >= 0) return (fd);
    /*  Otherwise create something in TMPDIR  */
    if (statfs (tmpdir, &statfsbuf) != 0) return (-1);
    errno = ENOSPC;
    nblocks = size / statfsbuf.f_bsize;
    if (nblocks >= statfsbuf.f_bfree) return (-1);
    bavail = statfsbuf.f_bfree - nblocks;
    if ( bavail < (statfsbuf.f_blocks / 4) ) return (-1);
    pid = getpid ();
    /*  Loop creating temporary file until an exclusive one is created  */
    for (count = 0; count >= 0; ++count)
    {
	sprintf (file, "%s/karma-mmap-%d-%d", tmpdir, pid, count);
	if ( ( fd = open (file, O_RDWR |O_CREAT | O_EXCL, S_IRUSR | S_IWUSR) )
	     < 0 )
	{
	    if (errno == EEXIST) continue;
	    return (-1);
	}
	if ( (unlink (file) != 0) || (lseek (fd, size - 1, SEEK_SET) == -1) ||
	     (write (fd, &dummy, 1) < 1) )
	{
	    errnum = errno;
	    close (fd);
	    errno = errnum;
	    return (-1);
	}
	return (fd);
    }
    errno = 0;
    return (-1);
}   /*  End Function _m_mmap_create_file  */

#endif
