/*LINTLIBRARY*/
/*  shm.c

    This code provides shared memory allocation routines.

    Copyright (C) 1998-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*


    Written by      Richard Gooch   18-FEB-1998

    Updated by      Richard Gooch   21-FEB-1998: Added attached list.

    Updated by      Richard Gooch   24-JUN-1998: Removed Solaris threading
  include.

    Last updated by Richard Gooch   14-FEB-2000: Fixed bug in allocating last
  entry of freelist.


*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <os.h>
#include <karma.h>
#include <karma_m.h>
#ifdef HAS_SYSV_SHARED_MEMORY
#  include <sys/ipc.h>
#  include <sys/shm.h>


#define SHARED_MEMORY_TEST_BUFFER_SIZE 16384
#define ALLOC_SIZE 64


struct segment
{
    int shmid;
    struct segment *prev;
    struct segment *next;
};

struct attached_segment
{
    void *addr;
    int shmid;
    unsigned int count;
    struct attached_segment *prev;
    struct attached_segment *next;
};

/*  Private data  */
static struct segment *first_segment = NULL;
static struct segment *first_free_segment = NULL;
static struct attached_segment *first_attached = NULL;
static struct attached_segment *first_free_attached = NULL;


/*  Private functions  */
STATIC_FUNCTION (void delete_segments, () );
STATIC_FUNCTION (void prog_bug, (char *function_name) );


/*  External functions: looks like SunOS is the odd one out  */
#ifdef OS_SunOS
extern char *shmat ();
#endif


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag m_shm_test_if_available ()
/*  [SUMMARY] Test if shared memory available.
    [RETURNS] TRUE if shared memory is available, else FALSE.
*/
{
    int shmid;
    static flag initialised = FALSE;
    static flag available = FALSE;

    if (initialised) return (available);
    initialised = TRUE;
    if ( ( shmid = shmget (IPC_PRIVATE, SHARED_MEMORY_TEST_BUFFER_SIZE,
			   IPC_CREAT | 0777) ) < 0 )
    {
	return (FALSE);
    }
    /*  Shared memory is turned on in this kernel  */
    if (shmctl (shmid, IPC_RMID, NULL) != 0)
    {
	fprintf (stderr, "Error removing shared memory segment\t%s\n",
		 ERRSTRING);
	return (FALSE);
    }
    available = TRUE;
    return (available);
}   /*  End Function m_shm_test_if_available  */

/*EXPERIMENTAL_FUNCTION*/
int m_shm_alloc (uaddr size, mode_t prot)
/*  [SUMMARY] Allocate a shared memory segment.
    <size> The size in bytes of the segment.
    <prot> The protection mask.
    [RETURNS] The shared memory ID on success, else -1.
*/
{
    int shmid;
    unsigned int count;
    struct segment *new;
    static flag registered_exit_func = FALSE;
    static char function_name[] = "m_shm_alloc";

    if (!registered_exit_func)
    {
	/*  Register function to be called upon processing of  exit(3)  */
#ifdef HAS_ATEXIT
	atexit (delete_segments);
#endif
#ifdef HAS_ON_EXIT
	on_exit (delete_segments, (caddr_t) NULL);
#endif
	registered_exit_func = TRUE;
    }
    /*  Ensure free list is big enough  */
    if (first_free_segment == NULL)
    {
	/*  Have to allocate some more for the free list  */
	if ( ( new = m_alloc (ALLOC_SIZE * sizeof *new) ) == NULL )
	{
	    m_error_notify (function_name, "new segment entries");
	    return (-1);
	}
	m_clear (new, ALLOC_SIZE * sizeof *new);
	/*  Add them to the free list  */
	first_free_segment = new;
	for (count = 0; count < ALLOC_SIZE - 1; ++count)
	{
	    new[count].shmid = -1;
	    new[count].next = new + count + 1;
	}
	new[count].shmid = -1;
    }
    if ( ( shmid = shmget (IPC_PRIVATE, size, IPC_CREAT | prot) ) < 0 )
    {
	return (-1);  /*  Failed  */
    }
    /*  Grab from free list  */
    new = first_free_segment;
    new->shmid = shmid;
    /*  Remove it from the free list  */
    first_free_segment = new->next;
    /*  Add to list  */
    if (first_segment != NULL) first_segment->prev = new;
    new->next = first_segment;
    first_segment = new;
    return (new->shmid);
}   /*  End Function m_shm_alloc  */

/*EXPERIMENTAL_FUNCTION*/
void m_shm_delete (int shmid)
/*  [SUMMARY] Delete a shared memory segment.
    <shmid> The ID of the shared memory segment to mark for deletion.
    [RETURNS] Nothing.
*/
{
    struct segment *curr;
    static char function_name[] = "m_shm_delete";

    /*  Search for it  */
    for (curr = first_segment; curr != NULL; curr = curr->next)
    {
	if (curr->shmid == shmid) break;
    }
    if (curr == NULL)
    {
	fprintf (stderr, "SHM segment: %d is not allocated\n", shmid);
	prog_bug (function_name);
    }
    if (shmctl (curr->shmid, IPC_RMID, NULL) != 0)
    {
	fprintf (stderr, "Error removing shared memory segment\t%s\n",
		 ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Remove entry from segment list  */
    if (curr->prev == NULL) first_segment = curr->next;
    else curr->prev->next = curr->next;
    if (curr->next != NULL) curr->next->prev = curr->prev;
    /*  Clear and add to the free list  */
    curr->shmid = -1;
    curr->prev = NULL;
    curr->next = first_free_segment;
    first_free_segment = curr;
}   /*  End Function m_shm_delete  */

/*EXPERIMENTAL_FUNCTION*/
void *m_shm_attach (int shmid, flag reattach, flag writable)
/*  [SUMMARY] Attach a shared memory segment to the process address space.
    <shmid> The ID of the shared memory segment to mark for deletion.
    <reattach> If TRUE an existing attachment of the segment is returned, if
    available.
    <writable> If TRUE the segment is given write access. This is ignored if
    <<reattach>> is TRUE.
    [RETURNS] The address at which the segment was attached on success, else
    NULL.
*/
{
    unsigned int count;
    void *addr;
    struct attached_segment *new, *curr;
    static char function_name[] = "m_shm_attach";

    if (reattach)
    {
	/*  Search for an existing attachment  */
	for (curr = first_attached; curr != NULL; curr = curr->next)
	{
	    if (shmid == curr->shmid)
	    {
		++curr->count;
		return (curr->addr);
	    }
	}
    }
    /*  Have to make a new attachment  */
    /*  Ensure free list is big enough  */
    if (first_free_attached == NULL)
    {
	/*  Have to allocate some more for the free list  */
	if ( ( new = m_alloc (ALLOC_SIZE * sizeof *new) ) == NULL )
	{
	    m_error_notify (function_name, "new attach entries");
	    return (NULL);
	}
	m_clear (new, ALLOC_SIZE * sizeof *new);
	/*  Add them to the free list  */
	first_free_attached = new;
	for (count = 0; count < ALLOC_SIZE - 1; ++count)
	{
	    new[count].shmid = -1;
	    new[count].next = new + count + 1;
	}
	new[count].shmid = -1;
    }
    addr = shmat (shmid, NULL, writable ? 0 : SHM_RDONLY);
    if (-1 == (signed long) addr) return (NULL);
    /*  Grab from free list  */
    new = first_free_attached;
    new->addr = addr;
    new->count = 1;
    /*  Remove it from the free list  */
    first_free_attached = new->next;
    /*  Add to list  */
    if (first_attached != NULL) first_attached->prev = new;
    new->next = first_attached;
    first_attached = new;
    return (new->addr);
}   /*  End Function m_shm_attach  */

/*EXPERIMENTAL_FUNCTION*/
flag m_shm_detach (void *addr)
/*  [SUMMARY] Detach a shared memory segment from the process address space.
    <addr> The address of the segment, as returned by [<m_shm_attach>].
    [RETURNS] TRUE on success, else FALSE.
*/
{
    struct attached_segment *curr;
    static char function_name[] = "m_shm_detach";

    /*  Search for it  */
    for (curr = first_attached; curr != NULL; curr = curr->next)
    {
	if (curr->addr == addr) break;
    }
    if (curr == NULL)
    {
	fprintf (stderr, "No SHM segment attached at: %p\n", addr);
	prog_bug (function_name);
    }
    if (curr->count < 1)
    {
	fprintf (stderr, "Attachment count already 0\n");
	prog_bug (function_name);
    }
    if (--curr->count > 0) return (TRUE);
    /*  Remove entry from attach list  */
    if (curr->prev == NULL) first_attached = curr->next;
    else curr->prev->next = curr->next;
    if (curr->next != NULL) curr->next->prev = curr->prev;
    /*  Clear and add to the free list  */
    curr->shmid = -1;
    curr->addr = NULL;
    curr->prev = NULL;
    curr->next = first_free_attached;
    first_free_attached = curr;
    return (shmdt (addr) == 0) ? TRUE : FALSE;
}   /*  End Function m_shm_detach  */


/*  Private functions follow  */

static void delete_segments ()
/*  [SUMMARY] Delete all shared memory segments.
    [NOTE] This only deletes the segments, the bookkeeping structures are kept.
    [RETURNS] Nothing.
*/
{
    struct segment *curr;

    for (curr = first_segment; curr != NULL; curr = curr->next)
    {
	if (shmctl (curr->shmid, IPC_RMID, NULL) != 0)
	{
	    fprintf (stderr, "Error removing shared memory segment\t%s\n",
		     ERRSTRING);
	}
	curr->shmid = -1;
    }
}   /*  End Function delete_segments  */

static void prog_bug (char *function_name)
{
    fprintf (stderr, "Program bug noted in function: %s\n", function_name);
    fprintf (stderr, "Aborting.%c\n", BEL);
    abort ();
}   /*  End Function prog_bug   */

#else  /*HAS_SYSV_SHARED_MEMORY*/


/*  Stub functions follow  */

/*STUB_FUNCTION*/
flag m_shm_test_if_available ()
/*  [SUMMARY] Test if shared memory available.
    [RETURNS] TRUE if shared memory is available, else FALSE.
*/
{
    return (FALSE);
}   /*  End Function m_shm_test_if_available  */

/*STUB_FUNCTION*/
int m_shm_alloc (uaddr size, mode_t prot)
/*  [SUMMARY] Allocate a shared memory segment.
    [PURPOSE] This routine will create a shared memory segment. On most systems
    this segment will be automatically scheduled for deletion when the process
    exits.
    <size> The size in bytes of the segment.
    <prot> The protection mask.
    [RETURNS] The shared memory ID on success, else -1.
*/
{
    return (-1);
}   /*  End Function m_shm_alloc  */

/*STUB_FUNCTION*/
void m_shm_delete (int shmid)
/*  [SUMMARY] Delete a shared memory segment.
    <shmid> The ID of the shared memory segment to mark for deletion.
    [RETURNS] Nothing.
*/
{
}   /*  End Function m_shm_delete  */

/*STUB_FUNCTION*/
void *m_shm_attach (int shmid, flag reattach, flag writable)
/*  [SUMMARY] Attach a shared memory segment to the process address space.
    <shmid> The ID of the shared memory segment to mark for deletion.
    <reattach> If TRUE an existing attachment of the segment is returned, if
    available.
    <writable> If TRUE the segment is given write access. This is ignored if
    <<reattach>> is TRUE.
    [RETURNS] The address at which the segment was attached on success, else
    NULL.
*/
{
    return (NULL);
}   /*  End Function m_shm_attach  */

/*STUB_FUNCTION*/
flag m_shm_detach (void *addr)
/*  [SUMMARY] Detach a shared memory segment from the process address space.
    <addr> The address of the segment, as returned by [<m_shm_attach>].
    [RETURNS] TRUE on success, else FALSE.
*/
{
    return (FALSE);
}   /*  End Function m_shm_detach  */

#endif  /*HAS_SYSV_SHARED_MEMORY*/
