/*LINTLIBRARY*/
/*  main.c

    This code provides Multi Threading support.

    Copyright (C) 1995-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the management of a pool of
  threads which may execute concurrently.


    Written by      Richard Gooch   9-JAN-1995

    Updated by      Richard Gooch   13-JAN-1995

    Updated by      Richard Gooch   24-JAN-1995: Increased number of call
  arguments to 4.

    Updated by      Richard Gooch   1-FEB-1995: Added check for MT_MAX_THREADS
  environment variable.

    Updated by      Richard Gooch   10-FEB-1995: IRIX5 support.

    Updated by      Richard Gooch   23-FEB-1995: IRIX6 support and bugfix.

    Updated by      Richard Gooch   5-MAY-1995: Placate SGI compiler.

    Updated by      Richard Gooch   23-JUL-1995: Created <mt_destroy_all_pools>

    Updated by      Richard Gooch   9-AUG-1995: Partially worked around a bug
  in thr_kill(3): sending SIGKILL kill the entire process.

    Updated by      Richard Gooch   21-AUG-1995: Above mentioned bug has been
  defined a "feature" by Sun support: used SIGTERM handler to kill thread to
  cope with this lossage.

    Updated by      Richard Gooch   31-AUG-1995: Made use of  __ateachexit
  routine for IRIX. Yuk!

    Updated by      Richard Gooch   24-JAN-1996: Created <mt_setlock> routine.
  Added thread_info parameter to thread function and created
  <mt_new_thread_info> routine.

    Updated by      Richard Gooch   25-JAN-1996: Increased arena space from
  64k to 256k for IRIX. Stupid limitations.

    Updated by      Richard Gooch   18-FEB-1996: Made child process die when
  parent dies with IRIX.

    Updated by      Richard Gooch   12-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   2-MAY-1996: Created <mt_get_shared_pool>
  routine.

    Updated by      Richard Gooch   3-JUN-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   29-JUL-1996: Added support for Linux
  through use of <sproc> routine in karmathread library and mutex emulation.

    Updated by      Richard Gooch   26-AUG-1996: Fixed bug in IRIX support
  where pool lock not deallocated when unthreaded. Fixed bugs where undefined
  threads and synchronisation variables in partial pools were incorrectly
  destroyed. IRIX and Linux threads now reaped.

    Updated by      Richard Gooch   15-NOV-1996: Tolerate old Linux kernels
  which don't report processor number.

    Updated by      Richard Gooch   26-NOV-1996: Created <mt_num_processors>.

    Updated by      Richard Gooch   4-FEB-1997: Created <mt_get_notify_pipe>
  function.

    Updated by      Richard Gooch   4-JUN-1997: Suspend threads under IRIX in
  <mt_wait_for_all_jobs> so they don't compete with niced processes.

    Updated by      Richard Gooch   18-JUN-1997: Fixed previous change: threads
  were suspended twice, causing the waitpid(2) to hang.

    Updated by      Richard Gooch   12-AUG-1997: Added death pipe for Linux so
  that clone threads know to _exit(2) when parent dies.

    Updated by      Richard Gooch   15-AUG-1997: Ensure newly created pools are
  idle.

    Updated by      Richard Gooch   16-AUG-1997: Use new clone(2) system call
  implemented in assembler (from LinuxThreads-0.6) for Linux and munmap(2)
  stack region under Linux.

    Updated by      Richard Gooch   22-AUG-1997: Added support (but have not
  enabled) for new prctl(2) system call for Linux which I've submitted for
  v2.1.51 but has not yet been included. Currently requires a patched C library

    Updated by      Richard Gooch   3-OCT-1997: Switched on use of Linux
  prctl() system call. Requires kernel 2.1.57 or later and libc 5.4.40 or later

    Updated by      Richard Gooch   4-MAY-1998: Fixed race condition under IRIX
  when thread is stopped after releasing the finished lock but before raising
  the pool semaphore. Simply continue all threads before waiting for semaphore
  in <mt_launch_job>.

    Updated by      Richard Gooch   6-MAY-1998: Worked around bug in IRIX where
  a process which is stopped before user code starts executing can be left in
  a strange state.

    Updated by      Richard Gooch   11-SEP-1998: Gracefully fallback if
  /proc/cpuinfo is not readable under Linux. Added check for MT_NUM_CPUS
  environment variable.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <r> package.

    Updated by      Richard Gooch   2-NOV-1998: Added POSIX thread support.

    Updated by      Richard Gooch   22-FEB-1999: Changed notify pipe semantics
  to send an unsigned int with the thread index.

    Updated by      Richard Gooch   7-JUN-1999: Fixed race condition caused by
  <mt_create_pool>.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Last updated by Richard Gooch   10-MAY-2000: Prefer sproc(2)-based threads
  for IRIX for portability reasons.


*/
#include <unistd.h>
#if defined(OS_IRIX5) || defined(OS_IRIX6)
#  define OS_IRIX
#endif
#ifdef OS_Solaris  /*  pthreads broken under Solaris 2.5.1            */
#  undef _POSIX_THREADS
#endif
#ifdef OS_IRIX     /*  Prefer sproc(2)-based threads for portability  */
#  undef _POSIX_THREADS
#endif
#ifdef _POSIX_THREADS
#  ifdef OS_OSF1  /*  Workaround for gcc on OSF1  */
#    define __C_ASM_H
#  endif
#  ifndef _REENTRANT
#    define _REENTRANT
#  endif
#  ifndef _POSIX_THREAD_SAFE_FUNCTIONS
#    define _POSIX_THREAD_SAFE_FUNCTIONS
#  endif
#  include <pthread.h>
#  include <semaphore.h>
#else  /*  No POSIX threads available: see what else is available  */
#  ifdef OS_Solaris
#    include <thread.h>
#  endif
#  ifdef OS_IRIX
#    include <ulocks.h>
#    include <sys/types.h>
#    include <sys/prctl.h>
#  endif
#  ifdef OS_Linux
#    define HAS_PRCTL
#    include <sys/mman.h>
#    include <linux/time.h>
#    include <asm/page.h>
#    include <linux/sched.h>
#    ifdef HAS_PRCTL
#      include <sys/prctl.h>
#    endif
#  endif
#endif
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <karma.h>
#include <karma_mt.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include <karma_c.h>
#include <os.h>

#define MAGIC_NUMBER 590322598

#define VERIFY_POOL(pl) {if (pl == NULL) \
{fprintf (stderr, "NULL thread pool passed\n"); \
 a_prog_bug (function_name); } \
if (pl->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid thread pool object\n"); \
 a_prog_bug (function_name); } }

#ifdef _POSIX_THREADS
#  define LOCK_TYPE pthread_mutex_t
#  define SEMAPHORE_TYPE sem_t
#  define TID_TYPE pthread_t
#  define LOCK_POOL(pl) {if (pthread_mutex_trylock (&pl->lock) != 0) \
{fprintf (stderr, "Recursive operation on pool not permitted\n"); \
 a_prog_bug (function_name);} }
#  define UNLOCK_POOL(pl) pthread_mutex_unlock (&pl->lock)
#  define FUNC_LOCK pthread_mutex_lock (&func_lock)
#  define FUNC_UNLOCK pthread_mutex_unlock (&func_lock)
#  define HAS_FUNC_LOCKS
#  define kLOCK(lock) pthread_mutex_lock (&lock)
#  define kUNLOCK(lock) pthread_mutex_unlock (&lock)
#  define TRYLOCK(lock) if (pthread_mutex_trylock (&lock) == 0)
#  define THREAD_RETURN void *
#  define HAS_THREADS
#  define HAS_THREAD_LIB
#else  /*  No POSIX threads available  */

#  ifdef OS_Solaris
#    define LOCK_TYPE mutex_t
#    define SEMAPHORE_TYPE sema_t
#    define TID_TYPE thread_t
#    define LOCK_POOL(pl) {if (mutex_trylock (&pl->lock) != 0) \
{fprintf (stderr, "Recursive operation on pool not permitted\n"); \
 a_prog_bug (function_name);} }
#    define UNLOCK_POOL(pl) mutex_unlock (&pl->lock)
#    define FUNC_LOCK mutex_lock (&func_lock)
#    define FUNC_UNLOCK mutex_unlock (&func_lock)
#    define HAS_FUNC_LOCKS
#    define kLOCK(lock) mutex_lock (&lock)
#    define kUNLOCK(lock) mutex_unlock (&lock)
#    define TRYLOCK(lock) if (mutex_trylock (&lock) == 0)
#    define THREAD_RETURN void *
#    define HAS_THREADS
#    define HAS_THREAD_LIB
#  endif

#  ifdef OS_IRIX
#    define LOCK_TYPE ulock_t
#    define SEMAPHORE_TYPE usema_t *
#    define TID_TYPE pid_t
#    define LOCK_POOL(pl) {if (uscsetlock (pl->lock, 1) != 1) \
{fprintf (stderr, "Recursive operation on pool not permitted\n"); \
 a_prog_bug (function_name);} }
#    define UNLOCK_POOL(pl) usunsetlock (pl->lock)
/*  Because IRIX locks are created we cannot use the compiler to initialise
    them. Have to forego function locks.
#    define FUNC_LOCK uswsetlock (func_lock)
#    define FUNC_UNLOCK usunsetlock (func_lock)
*/
#    define FUNC_LOCK
#    define FUNC_UNLOCK
#    define kLOCK(lock) uswsetlock (lock, 5)
#    define kUNLOCK(lock) usunsetlock (lock)
#    define TRYLOCK(lock) if (uscsetlock (lock, 5) == 1)
#    define THREAD_RETURN void
#    define HAS_THREADS
#    define HAS_THREAD_LIB
#  endif

#  ifdef OS_Linux
#    define TID_TYPE pid_t
#    define THREAD_RETURN int
#    define HAS_THREADS
#    define EMULATE_MUTEXES
#    define STACK_SIZE 1048576
#    ifndef PR_SET_PDEATHSIG
#      define NEEDS_DEATH_PIPE
#      define SEPARATE_FDs
#    endif
#  endif

#endif  /*  _POSIX_THREADS  */

#ifndef LOCK_POOL
#  define LOCK_POOL(pl)
#  define UNLOCK_POOL(pl)
#  define FUNC_LOCK
#  define FUNC_UNLOCK
#endif

#if defined(kLOCK) && defined(EMULATE_MUTEXES)
    !!!! ERROR !!! *** Redundant thread support ****
#endif

/*  Private structures  */
struct threadpool_type
{
    unsigned int magic_number;
    unsigned int num_threads;  /*  Number of worker threads: 0 or 2 or more  */
    struct thread_type *threads;
    KCallbackFunc callback_handle;
    void *info;
    char *thread_info_buffer;
    uaddr thread_info_buf_size;
    uaddr thread_info_size;
#ifdef LOCK_TYPE
    LOCK_TYPE lock;
    LOCK_TYPE synclock;
    SEMAPHORE_TYPE semaphore;
#endif
#ifdef EMULATE_MUTEXES
    int sync_pipe_read_fd;
    int sync_pipe_write_fd;
#endif
#ifdef NEEDS_DEATH_PIPE
    int death_pipe_read_fd;
    int death_pipe_write_fd;
#endif
    volatile flag notify_pipe_enabled;
    volatile int notify_pipe_read_fd;
    volatile int notify_pipe_write_fd;
};

typedef void (*JobFunc) (void *pool_info, void *call_info1, void *call_info2,
			 void *call_info3, void *call_info4,
			 void *thread_info);

struct thread_type
{
    KThreadPool pool;
#ifdef HAS_THREADS
    TID_TYPE tid;
#endif
#ifdef LOCK_TYPE
    LOCK_TYPE startlock;
    LOCK_TYPE finishedlock;
#endif
    JobFunc func;
    void *info1;
    void *info2;
    void *info3;
    void *info4;
    void *thread_info;
    unsigned int thread_number;
#ifdef EMULATE_MUTEXES
    volatile flag idle;
#endif
#if !defined(_POSIX_THREADS) && defined(OS_Solaris)
    volatile flag exit;
#endif
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
    flag suspended;
#endif
#if !defined(_POSIX_THREADS) && defined(OS_Linux)
    char *mmap_region;
#endif
};


/*  Private data  */
static KCallbackList destroy_list = NULL;
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
static usptr_t *arena = NULL;
#endif
static KThreadPool shared_pool = NULL;


/*  Private functions  */
#ifdef HAS_THREADS
STATIC_FUNCTION (THREAD_RETURN thread_main, (void *arg) );
#endif
STATIC_FUNCTION (flag destroy_callback,
		 (KThreadPool pool, void *client1_data,
		  flag *interrupt, void *client2_data) );
STATIC_FUNCTION (void exit_callback, () );
#if !defined(_POSIX_THREADS) && defined(OS_Linux)
EXTERN_FUNCTION (int clone, (int (*func) (void *info), void *child_stack,
			     int flags, void *info) );
STATIC_FUNCTION (pid_t sproc, (int (*func) (void *info), void *info,
			       char **mmap_region) );
STATIC_FUNCTION (void sigint_handler, (int sig) );
#endif


/*  Public functions follow  */


/*PUBLIC_FUNCTION*/
KThreadPool mt_create_pool (void *pool_info)
/*  [SUMMARY] Create a pool of threads which may have jobs launched onto it.
    <pool_info> An arbitrary pointer passed to work functions.
    [NOTE] The environment variable MT_MAX_THEADS may be used to limit the
    number of threads used.
    [MT-LEVEL] Safe under Solaris 2.
    [RETURNS] A KThreadPool object on success, else NULL.
*/
{
    KThreadPool pool;
#ifdef HAS_THREADS
    unsigned int count;
#endif
    CONST char *env;
#ifdef HAS_FUNC_LOCKS
    static LOCK_TYPE func_lock;
#endif
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
    char txt[STRING_LENGTH];
#endif
    static flag first_time = TRUE;
    static unsigned int max_threads = 0;
    static char function_name[] = "mt_create_pool";

    FUNC_LOCK;
    if (first_time)
    {
	first_time = FALSE;
	/*  Register function to be called upon processing of  exit(3)  */
#ifdef OS_IRIX
	if (__ateachexit (exit_callback) != 0)
	{
	    fprintf (stderr, "%s: error registering exit function\n",
		     function_name);
	}
#else
#  ifdef HAS_ATEXIT
	if (atexit (exit_callback) != 0)
	{
	    fprintf (stderr, "%s: error registering exit function\n",
		     function_name);
	}
#  endif
#  ifdef HAS_ON_EXIT
	on_exit (exit_callback, (caddr_t) NULL);
#  endif
#endif
	/*  Get "MT_MAX_THREADS" environment variable if it exists  */
	if ( ( env = r_getenv ("MT_MAX_THREADS") ) == NULL )
	{
	    max_threads = 0;
	}
	else
	{
	    max_threads = atoi (env);
	    if (max_threads < 1) max_threads = 1;
	    fprintf (stderr, "Forcing maximum number of threads to: %u\n",
		     max_threads);
	}
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
	/*  IRIX-specific initialisations  */
	if (usconfig (CONF_LOCKTYPE, US_NODEBUG) == -1)
	{
	    fprintf (stderr, "Error configuring\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (usconfig (CONF_ARENATYPE, US_SHAREDONLY) == -1)
	{
	    fprintf (stderr, "Error configuring\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (usconfig (CONF_INITSIZE, 262144) == -1)
	{
	    fprintf (stderr, "Error configuring\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (usconfig (CONF_INITUSERS, 256) == -1)
	{
	    fprintf (stderr, "Error configuring\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	sprintf ( txt, "/tmp/karma_arena-%d", getpid () );
	if ( ( arena = usinit (txt) ) == NULL )
	{
	    fprintf (stderr, "Error creating arena\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
#endif
    }
    FUNC_UNLOCK;
    if ( ( pool = m_alloc (sizeof *pool) ) == NULL )
    {
	m_abort (function_name, "thread pool");
    }
    pool->magic_number = MAGIC_NUMBER;
    pool->threads = NULL;
    pool->info = pool_info;
    pool->thread_info_buffer = NULL;
    pool->thread_info_buf_size = 0;
    pool->thread_info_size = 0;
    pool->notify_pipe_enabled = FALSE;
    pool->notify_pipe_read_fd = -1;
    pool->notify_pipe_write_fd = -1;
    /*  Determine number of processors available  */
    pool->num_threads = mt_num_processors ();
    if (max_threads > 0)
    {
	if (pool->num_threads > max_threads) pool->num_threads = max_threads;
    }
    if (pool->num_threads < 2) pool->num_threads = 0;
    /*  Must create main lock for pool even if number of threads forced to
	zero  */
#ifdef _POSIX_THREADS
    pthread_mutex_init (&pool->lock, NULL);
#endif
#if !defined(_POSIX_THREADS) && defined(OS_Solaris)
    mutex_init (&pool->lock, USYNC_THREAD, NULL);
#endif
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
    if ( ( pool->lock = usnewlock (arena) ) == NULL )
    {
	fprintf (stderr, "Error creating pool lock\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
#endif
    FUNC_LOCK;
    pool->callback_handle = c_register_callback (&destroy_list,
						 destroy_callback, pool,
						 NULL, FALSE,
						 NULL, FALSE, FALSE);
    FUNC_UNLOCK;
    if (pool->num_threads < 2) return (pool);
    /*  Now we know there is more than one thread, create other
	synchronisation objects and the threads  */
    if ( ( pool->threads =
	   m_alloc (sizeof *pool->threads * pool->num_threads) ) == NULL )
    {
	m_abort (function_name, "thread array");
    }
    m_clear (pool->threads, sizeof *pool->threads * pool->num_threads);
#ifdef HAS_THREADS
    for (count = 0; count < pool->num_threads; ++count)
    {
	pool->threads[count].pool = pool;
	pool->threads[count].thread_number = count;
    }
#endif
#ifdef _POSIX_THREADS
    pthread_mutex_init (&pool->synclock, NULL);
    sem_init (&pool->semaphore, 0, pool->num_threads);
    for (count = 0; count < pool->num_threads; ++count)
    {
	pthread_mutex_init (&pool->threads[count].startlock, NULL);
	pthread_mutex_init (&pool->threads[count].finishedlock, NULL);
	kLOCK (pool->threads[count].startlock);
	if (pthread_create (&pool->threads[count].tid, NULL, thread_main,
			    pool->threads + count) != 0)
	{
	    pool->threads[count].tid = 0;
	    fprintf (stderr, "Error creating thread\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
#endif  /*  _POSIX_THREADS  */
#if !defined(_POSIX_THREADS) && defined(OS_Solaris)
    mutex_init (&pool->synclock, USYNC_THREAD, NULL);
    sema_init (&pool->semaphore, pool->num_threads, USYNC_THREAD, NULL);
    for (count = 0; count < pool->num_threads; ++count)
    {
	mutex_init (&pool->threads[count].startlock, USYNC_THREAD, NULL);
	mutex_init (&pool->threads[count].finishedlock, USYNC_THREAD, NULL);
	kLOCK (pool->threads[count].startlock);
	pool->threads[count].exit = FALSE;
	if (thr_create (NULL, 0, thread_main, pool->threads + count,
			THR_NEW_LWP, &pool->threads[count].tid) != 0)
	{
	    pool->threads[count].tid = 0;
	    fprintf (stderr, "Error creating thread\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
#endif  /*  OS_Solaris  */
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
    if ( ( pool->synclock = usnewlock (arena) ) == NULL )
    {
	fprintf (stderr, "%s: error creating synclock\t%s\n",
		 function_name, ERRSTRING);
	mt_destroy_pool (pool, TRUE);
	return (NULL);
    }
    if ( ( pool->semaphore = usnewsema (arena, pool->num_threads) ) == NULL )
    {
	fprintf (stderr, "%s: error creating semaphore\t%s\n",
		 function_name, ERRSTRING);
	mt_destroy_pool (pool, TRUE);
	return (NULL);
    }
    for (count = 0; count < pool->num_threads; ++count)
    {
	if ( ( pool->threads[count].startlock = usnewlock (arena) ) == NULL )
	{
	    fprintf (stderr, "%s: error creating startlock\t%s\n",
		     function_name, ERRSTRING);
	    mt_destroy_pool (pool, TRUE);
	    return (NULL);
	}
	if ( ( pool->threads[count].finishedlock = usnewlock (arena) ) ==NULL )
	{
	    fprintf (stderr, "%s: error creating finishedlock\t%s\n",
		     function_name, ERRSTRING);
	    mt_destroy_pool (pool, TRUE);
	    return (NULL);
	}
	kLOCK (pool->threads[count].startlock);
	/*  Grab finishedlock now, so that <mt_wait_for_all_jobs> will be
	    forced to wait for the thread to unlock finishedlock. This is
	    required to work around a bug in IRIX where a process gets left in
	    a strange state if it is STOPped before it starts executing code */
	kLOCK (pool->threads[count].finishedlock);
	pool->threads[count].suspended = FALSE;
	if ( ( pool->threads[count].tid = sproc (thread_main, PR_SALL,
						 &pool->threads[count]) )
	     == -1 )
	{
	    pool->threads[count].tid = 0;
	    fprintf (stderr, "%s: error creating thread\t%s\n",
		     function_name, ERRSTRING);
	    mt_destroy_pool (pool, TRUE);
	    return (NULL);
	}
    }
#endif  /*  OS_IRIX  */
#ifdef NEEDS_DEATH_PIPE
    if (r_create_pipe (&pool->death_pipe_read_fd, &pool->death_pipe_write_fd)
	!= 0)
    {
	fprintf (stderr, "Error creating death pipe\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
#endif
#ifdef EMULATE_MUTEXES
    if (r_create_pipe (&pool->sync_pipe_read_fd, &pool->sync_pipe_write_fd)
	!= 0)
    {
	fprintf (stderr, "Error creating synchronisation pipe\t%s\n",
		 ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    for (count = 0; count < pool->num_threads; ++count)
    {
	/*  Must set initial state to busy so that <mt_wait_for_all_jobs> will
	    wait for thread to start. Otherwise there is a nasty race condition
	    where <mt_launch_job> clears the idle flag and sends the signal
	    before the thread sets the idle flag, so the thread will loop
	    waiting for a signal that's come and gone  */
	pool->threads[count].idle = FALSE;
	if ( ( pool->threads[count].tid =
	       sproc (thread_main, pool->threads + count,
		      &pool->threads[count].mmap_region) ) == -1 )
	{
	    pool->threads[count].tid = 0;
	    fprintf (stderr, "Error creating thread\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
#endif  /*  EMULATE_MUTEXES  */
    /*  Ensure threads start off idle  */
    mt_wait_for_all_jobs (pool);
    return (pool);
}   /*  End Function mt_create_pool  */

/*PUBLIC_FUNCTION*/
KThreadPool mt_get_shared_pool ()
/*  [SUMMARY] Create or get the shared thread pool.
    [PURPOSE] This routine will get a pool of threads which may be shared
    between several services within a process. This shared pool may be used to
    reduce the number of thread pools created. Creating large numbers of thread
    pools may consume significant system resources on some platforms. The first
    time this routine is called the shared pool is created.
    [NOTE] The environment variable MT_MAX_THEADS may be used to limit the
    number of threads used.
    [NOTE] This shared pool cannot be used with the [<mt_new_thread_info>] and
    [<mt_get_thread_info>] routines.
    [MT-LEVEL] Safe under Solaris 2.
    [RETURNS] The shared KThreadPool object on success. On failure the process
    aborts.
*/
{
#ifdef HAS_FUNC_LOCKS
    static LOCK_TYPE func_lock;
#endif
    static char function_name[] = "mt_get_shared_pool";

    FUNC_LOCK;
    if (shared_pool == NULL)
    {
	if ( ( shared_pool = mt_create_pool (NULL) ) == NULL )
	{
	    m_abort (function_name, "shared thread pool");
	}
    }
    FUNC_UNLOCK;
    return (shared_pool);
}   /*  End Function mt_get_shared_pool  */

/*PUBLIC_FUNCTION*/
void mt_destroy_pool (KThreadPool pool, flag interrupt)
/*  [SUMMARY] This routine will destroy a thread pool.
    <pool> The thread pool.
    <interrupt> If TRUE, any jobs not yet completed will be killed, else the
    function will wait for uncompleted jobs to finish prior to destroying the
    pool.
    [MT-LEVEL] Safe.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
#endif
#ifdef HAS_FUNC_LOCKS
    static LOCK_TYPE func_lock;
#endif
    static char function_name[] = "mt_destroy_pool";

    VERIFY_POOL (pool);
    FUNC_LOCK;
    if (pool == shared_pool) shared_pool = NULL;
    FUNC_UNLOCK;
    FLAG_VERIFY (interrupt);
    if (!interrupt) mt_wait_for_all_jobs (pool);
    LOCK_POOL (pool);
    for (count = 0; count < pool->num_threads; ++count)
    {
#ifdef _POSIX_THREADS
	if (pool->threads[count].tid != 0)
	{
	    if (pthread_cancel (pool->threads[count].tid) != 0)
	    {
		fprintf (stderr, "Error killing thread\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	}
	pthread_mutex_destroy (&pool->threads[count].startlock);
	pthread_mutex_destroy (&pool->threads[count].finishedlock);	
#endif
#if !defined(_POSIX_THREADS) && defined(OS_Solaris)
	/*  A bug in Solaris 2.4 caused thr_kill(3) to bugger the whole
	    process, so I try a thr_exit(3) instead.  */
	if (pool->threads[count].tid != 0)
	{
	    TRYLOCK (pool->threads[count].finishedlock)
	    {
		/*  Thread not busy  */
		pool->threads[count].exit = TRUE;
		/*  Fire away  */
		kUNLOCK (pool->threads[count].startlock);
		thr_join (pool->threads[count].tid, NULL, NULL);
	    }
	    else
	    {
		if (thr_kill (pool->threads[count].tid, SIGTERM) != 0)
		{
		    fprintf (stderr, "Error killing thread\t%s\n", ERRSTRING);
		    exit (RV_SYS_ERROR);
		}
		thr_join (pool->threads[count].tid, NULL, NULL);
	    }
	}
	mutex_destroy (&pool->threads[count].startlock);
	mutex_destroy (&pool->threads[count].finishedlock);	
#endif
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
	if (pool->threads[count].tid != 0)
	{
	    if (pool->threads[count].suspended)
	    {
		if (kill (pool->threads[count].tid, SIGCONT) != 0)
		{
		    fprintf (stderr, "%s: error CONTinuing thread\t%s\n",
			     function_name, ERRSTRING);
		    exit (RV_SYS_ERROR);
		}
		pool->threads[count].suspended = FALSE;
	    }
	    if (kill (pool->threads[count].tid, SIGKILL) != 0)
	    {
		fprintf (stderr, "%s: error KILLing thread\t%s\n",
			 function_name, ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    waitpid (pool->threads[count].tid, NULL, 0);
	}
	if (pool->threads[count].startlock != NULL)
	    usfreelock (pool->threads[count].startlock, arena);
	if (pool->threads[count].finishedlock != NULL)
	    usfreelock (pool->threads[count].finishedlock, arena);
#endif
#ifdef EMULATE_MUTEXES
	if (pool->threads[count].tid != 0)
	{
	    if (kill (pool->threads[count].tid, SIGKILL) != 0)
	    {
		fprintf (stderr, "%s: error killing thread\t%s\n",
			 function_name, ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    waitpid (pool->threads[count].tid, NULL, 0);
	}
#endif
#if !defined(_POSIX_THREADS) && defined(OS_Linux)
	munmap (pool->threads[count].mmap_region, STACK_SIZE);
#endif
#ifdef HAS_THREADS
	if (pool->threads[count].tid == 0)
	{
	    /*  This must be the last of the threads in this pool: there are no
		more threads or synchronisation variables  */
	    count = pool->num_threads;
	}
#endif
    }
    /*  Destroy pool lock  */
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
    usfreelock (pool->lock, arena);
#endif
    if (pool->num_threads > 1)
    {
#ifdef _POSIX_THREADS
	sem_destroy (&pool->semaphore);
#endif
#if !defined(_POSIX_THREADS) && defined(OS_Solaris)
	sema_destroy (&pool->semaphore);
#endif
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
	usfreesema (pool->semaphore, arena);
	usfreelock (pool->synclock, arena);
#endif
#ifdef EMULATE_MUTEXES
	close (pool->sync_pipe_read_fd);
	close (pool->sync_pipe_write_fd);
#endif
#ifdef NEEDS_DEATH_PIPE
	close (pool->death_pipe_read_fd);
	close (pool->death_pipe_write_fd);
#endif
	m_free (pool->threads);
    }
    if (pool->thread_info_buf_size > 0) m_free (pool->thread_info_buffer);
    pool->magic_number = 0;
    c_unregister_callback (pool->callback_handle);
    pool->callback_handle = NULL;
    m_free (pool);
}   /*  End Function mt_destroy_pool  */

/*PUBLIC_FUNCTION*/
void mt_destroy_all_pools (flag interrupt)
/*  [SUMMARY] This routine will destroy all thread pools.
    <interrupt> If TRUE, any jobs not yet completed will be killed, else the
    function will wait for uncompleted jobs to finish prior to destroying the
    pools.
    [MT-LEVEL] Safe.
    [RETURNS] Nothing.
*/
{
    c_call_callbacks (destroy_list, (void *) &interrupt);
    if (interrupt) c_destroy_list (destroy_list);
}   /*  End Function mt_destroy_all_pools  */

/*PUBLIC_FUNCTION*/
unsigned int mt_num_threads (KThreadPool pool)
/*  [SUMMARY] Get the number of threads in a thread pool.
    [PURPOSE] This function will determine the number of threads that may be
    run concurrently in a thread pool.
    <pool> The thread pool.
    [MT-LEVEL] Safe.
    [RETURNS] The number of concurrent threads.
*/
{
    static char function_name[] = "mt_num_threads";

    VERIFY_POOL (pool);
    if (pool->num_threads < 2) return (1);
    return (pool->num_threads);
}   /*  End Function mt_num_threads  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int mt_num_processors ()
/*  [SUMMARY] Get the number of processors available on the system.
    [RETURNS] The number of processors available.
*/
{
#if !defined(_SC_NPROCESSORS_ONLN) && !defined (_SC_NPROC_ONLN) && defined(OS_Linux)
    FILE *fp;
    char txt[STRING_LENGTH];
#endif
    CONST char *env;
    static unsigned int num_cpus = 0;

    if (num_cpus > 0) return (num_cpus);
    /*  Get "MT_NUM_CPUS" environment variable if it exists  */
    if ( ( env = r_getenv ("MT_NUM_CPUS") ) == NULL ) num_cpus = 1;
    else
    {
	num_cpus = atoi (env);
	if (num_cpus < 1) num_cpus = 1;
	fprintf (stderr, "Forcing number of CPUs to: %u\n", num_cpus);
	return (num_cpus);
    }
#ifdef _SC_NPROCESSORS_ONLN
    return sysconf (_SC_NPROCESSORS_ONLN);
#else
#  ifdef _SC_NPROC_ONLN
    return sysconf (_SC_NPROC_ONLN);
#  else
#    ifdef OS_Linux
    num_cpus = 0;
    if ( ( fp = fopen ("/proc/cpuinfo", "r") ) == NULL )
    {
	fprintf (stderr, "Error opening: \"/proc/cpuinfo\"\t%s\n", ERRSTRING);
	fprintf (stderr, "Forcing number of CPUs to: 1\n");
	num_cpus = 1;
	return (1);
    }
    while (fgets (txt, STRING_LENGTH, fp) != NULL)
    {
	if (strncmp (txt, "processor", 9) == 0) ++num_cpus;
    }
    fclose (fp);
    if (num_cpus < 1)
    {
	/*  Old kernel  */
	fprintf (stderr,
		 "\"/proc/cpuinfo\" reports no CPUs! Upgrade your kernel\n");
	fprintf (stderr, "Forcing number of CPUs to: 1\n");
	num_cpus = 1;
	return (1);
    }
    if (num_cpus > 256)
    {
	fprintf (stderr,
		 "mt_num_processors: too many CPUs: %u, limiting to 256\n",
		 num_cpus);
	num_cpus = 256;
    }
#    endif
#  endif
#endif
    return (num_cpus);
}   /*  End Function mt_num_processors  */

/*PUBLIC_FUNCTION*/
void mt_launch_job (KThreadPool pool,
		    void (*func) (void *pool_info,
				  void *call_info1, void *call_info2,
				  void *call_info3, void *call_info4,
				  void *thread_info),
		    void *call_info1, void *call_info2,
		    void *call_info3, void *call_info4)
/*  [SUMMARY] Launch a job onto a thread pool.
    [PURPOSE] This function will launch a job to a pool of threads, running the
    job on the first available thread.
    <pool> The thread pool.
    <func> The function to execute. The prototype function is [<mt_PROTO_func>]
    <call_info1> An arbitrary argument to <<func>>.
    <call_info2> An arbitrary argument to <<func>>.
    <call_info3> An arbitrary argument to <<func>>.
    <call_info4> An arbitrary argument to <<func>>.
    [NOTE] Jobs must not modify any signal actions or masks.
    [MT-LEVEL] Safe.
    [RETURNS] Nothing.
*/
{
    char ch = '\0';
    int fd;
#ifdef HAS_THREAD_LIB
    unsigned int count;
#endif
#ifdef HAS_THREADS
    char *thread_info;
    struct thread_type *thread;
#endif
#ifdef EMULATE_MUTEXES
    unsigned char thread_num;
#endif
    static char function_name[] = "mt_launch_job";

    VERIFY_POOL (pool);
    if (func == NULL) return;
    LOCK_POOL (pool);
    if (pool->num_threads < 2)
    {
	(*func) (pool->info, call_info1, call_info2, call_info3, call_info4,
		 (void *) pool->thread_info_buffer);
	fd = pool->notify_pipe_enabled ? pool->notify_pipe_write_fd : -1;
	UNLOCK_POOL (pool);
	if (fd > -1) r_write (fd, &ch, 1);
	return;
    }
    /*  Wait for any thread to become available  */
#ifdef _POSIX_THREADS
    while (sem_wait (&pool->semaphore) == EINTR);
#endif
#if !defined(_POSIX_THREADS) && defined(OS_Solaris)
    while (sema_wait (&pool->semaphore) == EINTR);
#endif
#if !defined(_POSIX_THREADS) && defined(OS_IRIX)
    /*  First ensure all threads are woken up: this eliminates a race. I tried
	raising the semaphore before releasing the finished flag/lock, but that
	meant the main thread could busy-wait for the finished flag  */
    for (count = 0; count < pool->num_threads; ++count)
    {
	if (pool->threads[count].suspended)
	{
	    if (kill (pool->threads[count].tid, SIGCONT) != 0)
	    {
		fprintf (stderr, "%s: error CONTinuing thread\t%s\n",
			 function_name, ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    pool->threads[count].suspended = FALSE;
	}
    }
    uspsema (pool->semaphore);
#endif
#ifdef EMULATE_MUTEXES
    /*  Read byte from pipe to get thread number of an idle thread. This blocks
	until a thread is idle  */
    if (r_read (pool->sync_pipe_read_fd, (char *) &thread_num, 1) != 1)
    {
	fprintf (stderr, "%s: error reading from synchronisation pipe\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    thread = pool->threads + thread_num;
    if (!thread->idle)
    {
	fprintf (stderr, "Thread: %u is not idle!\n", thread_num);
	a_prog_bug (function_name);
    }
#endif
    /*  One of the threads must be ready now  */
#ifdef TRYLOCK
    for (count = 0, thread = NULL; count < pool->num_threads; ++count)
    {
	TRYLOCK (pool->threads[count].finishedlock)
	{
	    /*  This thread ready for work  */
	    thread = pool->threads + count;
	    break;
	}
    }
#endif
#ifdef HAS_THREADS
    if (thread == NULL)
    {
	fprintf (stderr, "Error launching job: no free thread!\n");
	a_prog_bug (function_name);
    }
    /*  Set up the job information  */
    thread->func = func;
    thread->info1 = call_info1;
    thread->info2 = call_info2;
    thread->info3 = call_info3;
    thread->info4 = call_info4;
    /*  Ensure thread has correct private data pointer  */
    if (pool->thread_info_buffer == NULL)
    {
	thread->thread_info = NULL;
    }
    else
    {
	thread_info = pool->thread_info_buffer;
	thread_info += thread->thread_number * pool->thread_info_size;
	thread->thread_info = thread_info;
    }
#endif
    /*  Start thread running  */
#ifdef kUNLOCK
    kUNLOCK (thread->startlock);
    UNLOCK_POOL (pool);
    return;
#endif
#ifdef EMULATE_MUTEXES
    thread->idle = FALSE;
    if (kill (thread->tid, SIGINT) != 0)
    {
	fprintf (stderr, "%s: error resuming thread\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
#endif
}   /*  End Function mt_launch_job  */

/*PUBLIC_FUNCTION*/
void mt_setlock (KThreadPool pool, flag lock)
/*  [SUMMARY] Set a lock on a thread pool.
    [PURPOSE] This function will lock a thread pool such that no other thread
    can lock the pool at the same time. This does not prevent other threads
    from running, nor new jobs from being launched, it merely prevents them
    from aquiring the lock.
    <pool> The thread pool to lock.
    <lock> If TRUE the pool is locked. If FALSE the pool is unlocked and any
    other threads wishing to lock the pool may do so (one at a time of course).
    [MT-LEVEL] Safe.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "mt_setlock";

    VERIFY_POOL (pool);
    FLAG_VERIFY (lock);
    if (pool->num_threads < 2) return;
#ifdef HAS_THREADS
#  ifdef kLOCK
    if (lock) kLOCK (pool->synclock);
    else kUNLOCK (pool->synclock);
#  else
    fprintf (stderr, "Locks not supported\n");
    a_prog_bug (function_name);
#  endif
#endif
}   /*  End Function mt_setlock  */

/*PUBLIC_FUNCTION*/
void mt_new_thread_info (KThreadPool pool, void *info, uaddr size)
/*  [SUMMARY] Register new thread information for the threads in a pool.
    <pool> The thread pool.
    <info> A pointer to the thread information array. If NULL and <<size>> is
    not 0 then the routine will allocate an array of sufficient size.
    <size> The size (per thread) in bytes of the thread information. When
    threads are executing each is guaranteed to have a private working space of
    this size.
    [MT-LEVEL] Safe.
    [RETURNS] Nothing. On failure the process aborts.
*/
{
    uaddr num_threads;
    static char function_name[] = "mt_new_thread_info";

    VERIFY_POOL (pool);
    if (size < 1)
    {
	fprintf (stderr, "Illegal size: %lu\n", size);
	a_prog_bug (function_name);
    }
    if (pool == shared_pool)
    {
	fprintf (stderr, "Operation illegal for shared thread pool\n");
	a_prog_bug (function_name);
    }
    num_threads = mt_num_threads (pool);
    LOCK_POOL (pool);
    pool->thread_info_size = size;
    if (info == NULL)
    {
	/*  Need an internally allocated array  */
	if (size * num_threads <= pool->thread_info_buf_size)
	{
	    UNLOCK_POOL (pool);
	    return;
	}
	/*  Existing buffer not big enough: free and allocate  */
	if (pool->thread_info_buf_size > 0) m_free (pool->thread_info_buffer);
	if ( ( info = m_alloc (size * num_threads) ) == NULL )
	{
	    m_abort (function_name, "thread information array");
	}
	pool->thread_info_buf_size = size * num_threads;
	pool->thread_info_buffer = (char *) info;
	UNLOCK_POOL (pool);
	return;
    }
    /*  Array supplied  */
    if (pool->thread_info_buf_size > 0)
    {
	 m_free (pool->thread_info_buffer);
	 pool->thread_info_buf_size = 0;
     }
    pool->thread_info_buffer = (char *) info;
    UNLOCK_POOL (pool);
}   /*  End Function mt_new_thread_info  */

/*PUBLIC_FUNCTION*/
void *mt_get_thread_info (KThreadPool pool)
/*  [SUMMARY] Get the thread information pointer for a pool of threads.
    <pool> The thread pool.
    [MT-LEVEL] Safe.
    [RETURNS] A pointer to the thread information array.
*/
{
    static char function_name[] = "mt_get_thread_info";

    VERIFY_POOL (pool);
    if (pool == shared_pool)
    {
	fprintf (stderr, "Operation illegal for shared thread pool\n");
	a_prog_bug (function_name);
    }
    return ( (char *) pool->thread_info_buffer );
}   /*  End Function mt_get_thread_info  */

/*PUBLIC_FUNCTION*/
void mt_wait_for_all_jobs (KThreadPool pool)
/*  [SUMMARY] Wait for all previously launched jobs to complete.
    <pool> The thread pool.
    [MT-LEVEL] Safe.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    static char function_name[] = "mt_wait_for_all_jobs";

    VERIFY_POOL (pool);
    LOCK_POOL (pool);
    for (count = 0; count < pool->num_threads; ++count)
    {
#ifdef kLOCK
	kLOCK (pool->threads[count].finishedlock);
	kUNLOCK (pool->threads[count].finishedlock);
#  ifdef OS_IRIX
	if (!pool->threads[count].suspended)
	{
	    if (kill (pool->threads[count].tid, SIGSTOP) != 0)
	    {
		fprintf (stderr, "%s: error STOPping thread\t%s\n",
			 function_name, ERRSTRING);
	    }
	    pool->threads[count].suspended = TRUE;
	    if (waitpid (pool->threads[count].tid, NULL, WUNTRACED) == -1)
	    {
		fprintf (stderr, "%s: error waiting for thread\t%s\n",
			 function_name, ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	}
#  endif
#endif
#ifdef EMULATE_MUTEXES
	/*  It would be preferable to read a byte per thread from the pipe,
	    but this would then drain the pipe for later use (in particular, by
	    the job launcher). Instead a low CPU load test is performed  */
	while (!pool->threads[count].idle) usleep (10000);
#endif
    }
    UNLOCK_POOL (pool);
}   /*  End Function mt_wait_for_all_jobs  */

/*EXPERIMENTAL_FUNCTION*/
int mt_get_notify_pipe (KThreadPool pool)
/*  [SUMMARY] Create/get the notification pipe.
    [PURPOSE] This routine gets the read-end of the notification pipe for a
    thread pool. If the pipe does not yet exist, it is created. Every time a
    job completes, an unsigned integer is written to the write-end of the pipe
    (if it exists). The value if the integer is equal to the thread index.
    [NOTE] Once the pipe is created/gotten (and it has not been disabled by
    [<mt_disable_notify_pipe>], the notification bytes <<must>> be consumed,
    otherwise threads in the pool will eventually become deadlocked.
    <pool> The thread pool.
    [MT-LEVEL] Safe.
    [RETURNS] The read-end of the pipe. On failure the process aborts.
*/
{
    static char function_name[] = "mt_get_notify_pipe";

    VERIFY_POOL (pool);
    LOCK_POOL (pool);
    pool->notify_pipe_enabled = TRUE;
    if (pool->notify_pipe_read_fd > -1)
    {
	UNLOCK_POOL (pool);
	return (pool->notify_pipe_read_fd);
    }
    if (r_create_pipe ( (int *) &pool->notify_pipe_read_fd,
			(int *) &pool->notify_pipe_write_fd ) != 0)
    {
	UNLOCK_POOL (pool);
	fprintf (stderr, "Error creating notify pipe\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    UNLOCK_POOL (pool);
    return (pool->notify_pipe_read_fd);
}   /*  End Function mt_get_notify_pipe  */

/*EXPERIMENTAL_FUNCTION*/
void mt_disable_notify_pipe (KThreadPool pool)
/*  [SUMMARY] Disable the notification pipe.
    [MT-LEVEL] Safe.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "mt_disable_notify_pipe";

    VERIFY_POOL (pool);
    LOCK_POOL (pool);
    pool->notify_pipe_enabled = FALSE;
    UNLOCK_POOL (pool);
}   /*  End Function mt_disable_notify_pipe  */


/*  Private functions follow  */

#ifdef HAS_THREADS
static THREAD_RETURN thread_main (void *arg)
/*  [PURPOSE] This function is the entry point for a thread.
    [RETURNS] NULL on error, else never returns.
*/
{
#  ifdef EMULATE_MUTEXES
    pid_t ppid;
    unsigned char thread_num;
#  endif
#  ifdef NEEDS_DEATH_PIPE
    fd_set input_fds, output_fds, exception_fds;
#  endif
    struct sigaction action;
    struct thread_type *thread = (struct thread_type *) arg;
    KThreadPool parent = thread->pool;
    static char function_name[] = "_mt_thread_main";

    /*  Set up structure for sigaction(2)  */
    m_clear (&action, sizeof action);
    sigemptyset (&action.sa_mask);
    action.sa_flags = 0;
#  ifdef SA_RESTART
    action.sa_flags |= SA_RESTART;
#  endif
    /*  Operating-system specific initialisations  */
#  if !defined(_POSIX_THREADS) && defined(OS_Solaris)
    /*  Set up sigTERM handler  */
    action.sa_handler = ( void (*) () ) thr_exit;
    if (sigaction (SIGTERM, &action, NULL) == -1)
    {
	fprintf (stderr, "Error setting sigTERM handler\t%s\n", ERRSTRING);
	_exit (RV_SYS_ERROR);
    }
#  endif
#  if !defined(_POSIX_THREADS) && defined(OS_IRIX)
    /*  Make child die when parent dies  */
    if (prctl (PR_TERMCHILD) == -1)
    {
	fprintf (stderr, "Error controlling process\t%s\n", ERRSTRING);
	_exit (RV_SYS_ERROR);
    }
    /*  This is used to notify <mt_wait_for_all_jobs> that I've started running
	properly, and that it is safe to send a SIGSTOP now. This works around
	an IRIX bug  */
    kUNLOCK (thread->finishedlock);
#  endif
#  if !defined(_POSIX_THREADS) && defined(OS_Linux)
    /*  Set up sigTERM handler  */
    action.sa_handler = ( void (*) () ) _exit;
    if (sigaction (SIGTERM, &action, NULL) == -1)
    {
	fprintf (stderr, "Error setting sigTERM handler\t%s\n", ERRSTRING);
	_exit (RV_SYS_ERROR);
    }
    /*  Set up sigINT handler  */
    action.sa_flags = 0;
    action.sa_handler = ( void (*) () ) sigint_handler;
    if (sigaction (SIGINT, &action, NULL) == -1)
    {
	fprintf (stderr, "Error setting sigINT handler\t%s\n", ERRSTRING);
	_exit (RV_SYS_ERROR);
    }
#    ifdef PR_SET_PDEATHSIG
    if (prctl (PR_SET_PDEATHSIG, SIGTERM) == -1)
    {
	fprintf (stderr, "Error calling prctl(2)\t%s\n", ERRSTRING);
	_exit (RV_SYS_ERROR);
    }
#    endif
#  endif
#  ifdef EMULATE_MUTEXES
    ppid = getppid ();
#    ifdef SEPARATE_FDs
    close (parent->sync_pipe_read_fd);
#    endif
#  endif
#  ifdef NEEDS_DEATH_PIPE
    if (close (parent->death_pipe_write_fd) == -1)
    {
	fprintf (stderr, "%s: error closing death pipe write\t%s\n",
		 function_name, ERRSTRING);
	_exit (RV_SYS_ERROR);
    }
#  endif
    while (TRUE)
    {
#  ifdef _POSIX_THREADS
	kLOCK (thread->startlock);
	(*thread->func) (parent->info, thread->info1, thread->info2,
			 thread->info3, thread->info4, thread->thread_info);
	kUNLOCK (thread->finishedlock);
	sem_post (&parent->semaphore);
#  endif
#  if !defined(_POSIX_THREADS) && defined(OS_Solaris)
	kLOCK (thread->startlock);
	if (thread->exit) thr_exit (NULL);
	(*thread->func) (parent->info, thread->info1, thread->info2,
			 thread->info3, thread->info4, thread->thread_info);
	if (thread->exit) thr_exit (NULL);
	kUNLOCK (thread->finishedlock);
	sema_post (&parent->semaphore);
#  endif
#  if !defined(_POSIX_THREADS) && defined(OS_IRIX)
	kLOCK (thread->startlock);
	(*thread->func) (parent->info, thread->info1, thread->info2,
			 thread->info3, thread->info4, thread->thread_info);
	kUNLOCK (thread->finishedlock);
	usvsema (parent->semaphore);
#  endif
#  ifdef EMULATE_MUTEXES
	/*  Notify that I am ready for more work  */
	thread_num = thread->thread_number;
	thread->idle = TRUE;
	if (r_write (parent->sync_pipe_write_fd, (char *) &thread_num, 1) != 1)
	{
	    fprintf (stderr, "%s: error writing synchronisation byte\t%s\n",
		     function_name, ERRSTRING);
	    _exit (RV_SYS_ERROR);
	}
	/*  Block until launcher has a new job for me, or the parent dies  */
	while (thread->idle)
	{
#    ifdef NEEDS_DEATH_PIPE
	    FD_ZERO (&input_fds);
	    FD_ZERO (&output_fds);
	    FD_ZERO (&exception_fds);
	    FD_SET (parent->death_pipe_read_fd, &input_fds);
	    select (parent->death_pipe_read_fd + 1, &input_fds,
		    &output_fds, &exception_fds, NULL);
#    else
	    pause ();
#    endif
	    if (getppid () != ppid) _exit (0);
	}
	/*  Have a new job to do  */
	(*thread->func) (parent->info, thread->info1, thread->info2,
			 thread->info3, thread->info4, thread->thread_info);
#  endif
	if ( parent->notify_pipe_enabled &&
	     (parent->notify_pipe_write_fd > -1) )
	{
	    errno = 0;
	    if (write (parent->notify_pipe_write_fd,
		       (void *) &thread->thread_number,
		       sizeof thread->thread_number) !=
		sizeof thread->thread_number)
	    {
		fprintf (stderr, "%s: error writing to notify pipe: %d\t%s\n",
			 function_name, parent->notify_pipe_write_fd,
			 ERRSTRING);
	    }
	}
    }
}   /*  End Function thread_main  */
#endif  /*  HAS_THREADS  */

static flag destroy_callback (KThreadPool pool, void *client1_data,
			      flag *interrupt, void *client2_data)
/*  [PURPOSE] This routine is the destroy callback.
    <pool> The thread pool to destroy.
    <client1_data> Ignored.
    <interrupt> Pointer to value passed to <<mt_destroy_pool>>.
    <client2_data> Ignored.
    [RETURNS] FALSE.
*/
{
    static char function_name[] = "destroy_callback";

    VERIFY_POOL (pool);
    mt_destroy_pool (pool, *interrupt);
    return (FALSE);
}   /*  End Function destroy_callback  */

static void exit_callback ()
/*  [PURPOSE] This routine is the exit callback, called when exit(3) is called.
    [RETURNS] Nothing.
*/
{
    mt_destroy_all_pools (TRUE);
}   /*  End Function exit_callback  */


/*  Linux-specific functions follow  */
#if !defined(_POSIX_THREADS) && defined(OS_Linux)

static void sigint_handler (int sig)
{
}   /*  End Function sigint_handler  */

#ifndef PAGE_SIZE
#  define PAGE_SIZE 4096
#endif

static pid_t sproc (int (*func) (void *info), void *info, char **mmap_region)
{
    int flags = 0;
    char *child_stack;

    /*  Need to create some stack space for the child process-to-be. Don't want
	to use malloc, otherwise the heap will be cluttered up with stacks, so
	mmap(2) is used to create the segment. Must also make sure that the
	segment is not unmapped until the child is finished  */
    if ( ( child_stack = (char *) mmap (NULL, STACK_SIZE,
					PROT_READ | PROT_WRITE,
					MAP_ANON | MAP_PRIVATE, -1, 0) )
	 == (char *) -1 )
    {
	fprintf (stderr, "Error creating stack\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Cause a SEGV if the bottom of the stack is reached  */
    if (mprotect (child_stack, PAGE_SIZE, PROT_NONE) == -1)
    {
	fprintf (stderr, "Error protecting memory\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Cause a SEGV if the top of the stack is reached  */
    if (mprotect (child_stack + STACK_SIZE - PAGE_SIZE, PAGE_SIZE, PROT_NONE)
	== -1)
    {
	fprintf (stderr, "Error protecting memory\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    flags = CLONE_VM | CLONE_FS | SIGCHLD;
#ifndef SEPARATE_FDs
    flags |= CLONE_FILES;
#endif
    return clone (func, child_stack + STACK_SIZE - PAGE_SIZE, flags, info);
}   /*  End Function sproc  */

#endif  /*  !defined(_POSIX_THREADS) && defined(OS_Linux)  */
