/*LINTLIBRARY*/
/*  fork.c

    This code implements forking support.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/
/*

    This file contains the various utility routines for forking processes.


    Written by      Richard Gooch   2-AUG-1997

    Updated by      Richard Gooch   8-OCT-1997: Fixed problems with races under
  OS_IRIX: stack persistence problems.

    Updated by      Richard Gooch   1-DEC-1997: Added support for full pathname

    Updated by      Richard Gooch   3-DEC-1997: Added distinction between
  executable not found and other system errors.

    Last updated by Richard Gooch   14-JUL-2000: Trivial cosmetic change.


*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(OS_IRIX5) || defined(OS_IRIX6)
#  define OS_IRIX
#  include <sys/prctl.h>
#endif
#include <os.h>
#include <karma.h>
#include <karma_r.h>


#ifdef OS_Solaris
#  define FORK vfork
#endif

#ifndef FORK
#  define FORK fork
#endif


/*  Private structures  */
struct info_type
{
    int stdin_fd;
    int stdout_fd;
    int stderr_fd;
    unsigned int num_close_fds;
    int *close_fds;
    CONST char *file;
    char *CONST *argv;
};


/*  Private functions  */
STATIC_FUNCTION (void child_func, (void *info_ptr) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
pid_t r_fork_and_execvp (int stdin_fd, int stdout_fd, int stderr_fd,
			 unsigned int num_close_fds, int *close_fds,
			 CONST char *file, char *CONST *argv)
/*  [SUMMARY] Fork and exec a process in a virtual memory efficient way.
    [PURPOSE] This function will fork and exec a new process in such as way as
    to minimise virtual memory consumption. The traditional use of fork(2) can
    consume large amounts of virtual memory under many operating systems (such
    as Solaris and IRIX), especially when the calling process has large arrays
    allocated. Some operating systems provide the vfork(2) call to work around
    this, but vfork(2) is not always available and may in fact not provide the
    Berkely vfork(2) semantics which preserved virtual memory. This function
    will make use of whatever Operating System specific functions are available
    in order to obtain the desired effect.
    <stdin_fd> The file descriptor which should become the standard input for
    the child process.
    <stdout_fd> The file descriptor which should become the standard output for
    the child process.
    <stderr_fd> The file descriptor which should become the standard error
    output for the child process.
    <num_close_fds> The number of file descriptors to close in the child.
    <close_fds> The array of file descriptors to close in the child.
    <file> The file to execvp(3).
    <argv> The command-line arguments to pass to execvp(3).
    [RETURNS] The child process ID on success, else -1.
*/
{
    pid_t child;
    struct info_type info;
#ifndef CAN_FORK
    static char function_name[] = "r_fork_and_execvp";
#endif

    /*  BEWARE: this information is kept on the caller's stack: if the
	implementation uses threads then the contents of the stack may change
	before the execvp(2) call. IRIX has a convenient PR_BLOCK flag to take
	care of this  */
    info.stdin_fd = stdin_fd;
    info.stdout_fd = stdout_fd;
    info.stderr_fd = stderr_fd;
    info.num_close_fds = num_close_fds;
    info.close_fds = close_fds;
    info.file = file;
    info.argv = argv;
#ifdef CAN_FORK
#  ifdef OS_IRIX
    if ( ( child = sproc (child_func, PR_BLOCK | PR_SADDR | PR_NOLIBC, &info) )
	 == -1 )
	return (-1);
#  else
    switch ( child = FORK () )
    {
      case 0:
	/*  Child  */
	child_func (&info);
	_exit (RV_SYS_ERROR);
	break;
      case -1:
	/*  Error  */
	return (-1);
	break;
      default:
	/*  Parent  */
	break;
    }
#  endif
    return (child);
#else
    fprintf (stderr, "%s: forking not supported\n", function_name);
    return (-1);
#endif
}   /*  End Function r_fork_and_execvp  */


/*  Private functions follow  */

static void child_func (void *info_ptr)
/*  [SUMMARY] Child function.
    <info_ptr> A pointer to info.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    struct info_type *info = (struct info_type *) info_ptr;

#ifdef OS_IRIX
    prctl ( PR_UNBLKONEXEC, getppid () );
#endif
    if (info->stdin_fd != 0) dup2 (info->stdin_fd, 0);
    if (info->stdout_fd != 1) dup2 (info->stdout_fd, 1);
    if (info->stderr_fd != 2) dup2 (info->stderr_fd, 2);
    for (count = 0; count < info->num_close_fds; ++count)
	close (info->close_fds[count]);
    if (info->file[0] == '/') execv (info->file, info->argv);
    else execvp (info->file, info->argv);
    _exit ( (errno == ENOENT) ? RV_EXECUTABLE_NOT_FOUND : RV_SYS_ERROR );
}   /*  End Function child_func  */
