/*LINTLIBRARY*/
/*  main.c

    This code provides signal blocking routines.

    Copyright (C) 1992-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various Karma utility routines which relate to
  blocking and unblocking signals.


    Written by      Richard Gooch   12-SEP-1992

    Updated by      Richard Gooch   14-NOV-1992

    Updated by      Richard Gooch   12-DEC-1992: Fixed some comments.

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   12-APR-1993: Added  s_ignore  .

    Updated by      Richard Gooch   22-MAY-1994: Added  return (NULL)  to
  s_ignore  when signals unsupported.

    Updated by      Richard Gooch   9-NOV-1994: Changed return cast in
  s_ignore  .

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/s/main.c

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   5-MAY-1995: Placate SGI compiler.

    Updated by      Richard Gooch   13-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   2-SEP-1997: Switched to ERRSTRING.

    Updated by      Richard Gooch   21-JUN-1998: Switched from CAN_ FORK to
  HAS_SIGNALS.

    Updated by      Richard Gooch   23-JUN-1998: Used sigaction(2) instead of
  signal(3) for <s_ignore>.

    Last updated by Richard Gooch   15-AUG-1999: Made egcs happy.


*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <os.h>
#ifdef HAS_SIGNALS
#  include <unistd.h>
#  include <sys/ioctl.h>
#  include <signal.h>
#endif
#include <karma.h>
#include <karma_s.h>

/*  Private functions  */
#ifdef HAS_SIGNALS
static void prog_bug ();
#endif  /*  HAS_SIGNALS  */

/*  Private data  */
static flag initialised = FALSE;
static sigset_t blocked_mask;

/*PUBLIC_FUNCTION*/
void s_block (int signal)
/*  [SUMMARY] Block a signal.
    [PURPOSE] This routine will cause a signal to be blocked until a subsequent
    call to [<s_unblock>] or [<s_unblock_all>]. If the signal is already
    blocked, no action is taken.
    <signal> The signal to block.
    [RETURNS] Nothing.
*/
{
#ifdef HAS_SIGNALS
    sigset_t old_mask;
    sigset_t new_mask;
    /*static char function_name[] = "s_block";*/

    if (initialised != TRUE)
    {
	/*  Initialise signal set  */
	if (sigemptyset (&blocked_mask) != 0)
	{
	    fprintf (stderr, "Error clearing signal mask\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	initialised = TRUE;
    }
    /*  Get old signal mask  */
    if (sigprocmask (SIG_BLOCK, (sigset_t *) NULL, &old_mask) != 0)
    {
	fprintf (stderr, "Error getting signal mask\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Block signal  */
    /*  Note that signals which are already blocked are not blocked again,
	as this could cause the call to  sigprocmask  to block (on a Convex).
	This seems to be due to the way multiple processors handle signals  */
    switch ( sigismember (&old_mask, signal) )
    {
      case 0:
	/*  Signal not already blocked  */
	if (sigaddset (&blocked_mask, signal) != 0)
	{
	    fprintf (stderr, "Error adding signal: %d to signal mask\t%s\n",
		     signal, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	/*  Initialise new signal set  */
	if (sigemptyset (&new_mask) != 0)
	{
	    fprintf (stderr, "Error clearing new signal mask\t%s\n",ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (sigaddset (&new_mask, signal) != 0)
	{
	    fprintf (stderr,"Error adding signal: %d to new signal mask\t%s\n",
		     signal, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	/*  Now block signal  */
	if (sigprocmask (SIG_BLOCK, &new_mask, (sigset_t *) NULL) != 0)
	{
	    fprintf (stderr, "Error setting signal mask\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	break;
      case 1:
	/*  Signal already blocked  */
	break;
      default:
	fprintf (stderr, "Error testing if signal: %d is a member\t%s\n",
		 signal, ERRSTRING);
	exit (RV_SYS_ERROR);
	break;
    }

#else  /*  HAS_SIGNALS  */
    fprintf (stderr, "Operating system does not support signal blocking\n");
#endif  /*  HAS_SIGNALS  */
}   /*  End Function s_block  */

/*PUBLIC_FUNCTION*/
void s_unblock (int signal)
/*  [SUMMARY] Unblock a signal.
    [PURPOSE] This routine will cause a signal to be unblocked. If the signal
    is not already blocked, no action is taken.
    <signal> The signal to unblock.
    [NOTE] The signal will only be unblocked if previously blocked by a call
    to [<s_block>].
    [RETURNS] Nothing.
*/
{
#ifdef HAS_SIGNALS
    sigset_t old_mask;
    sigset_t new_mask;
    static char function_name[] = "s_unblock";

    if (initialised != TRUE)
    {
	fprintf (stderr, "No signals blocked yet\n");
	prog_bug (function_name);
    }
    /*  Test supposed signal mask  */
    switch ( sigismember (&blocked_mask, signal) )
    {
      case 0:
	/*  Signal not already blocked  */
	return;
	/*break;*/
      case 1:
	/*  Signal already blocked  */
	break;
      default:
	fprintf (stderr, "Error testing if signal: %d is a member\t%s\n",
		 signal, ERRSTRING);
	exit (RV_SYS_ERROR);
	break;
    }
    /*  Get old signal mask  */
    if (sigprocmask (SIG_BLOCK, (sigset_t *) NULL, &old_mask) != 0)
    {
	fprintf (stderr, "Error getting signal mask\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    switch ( sigismember (&old_mask, signal) )
    {
      case 0:
	/*  Signal not blocked  */
	fprintf (stderr, "Signal: %d has been unblocked via other means\n",
		 signal);
	prog_bug (function_name);
	break;
      case 1:
	/*  Signal blocked: unblock  */
	if (sigdelset (&blocked_mask, signal) != 0)
	{
	    fprintf (stderr, "Error adding signal: %d to signal mask\t%s\n",
		     signal, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	/*  Initialise new signal set  */
	if (sigemptyset (&new_mask) != 0)
	{
	    fprintf (stderr, "Error clearing new signal mask\t%s\n",ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (sigaddset (&new_mask, signal) != 0)
	{
	    fprintf (stderr,"Error adding signal: %d to new signal mask\t%s\n",
		     signal, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (sigprocmask (SIG_UNBLOCK, &new_mask, (sigset_t *) NULL) != 0)
	{
	    fprintf (stderr, "Error restoring old signal mask\t%s\n",
		     ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	break;
      default:
	fprintf (stderr, "Error testing if signal: %d is a member\t%s\n",
		 signal, ERRSTRING);
	exit (RV_SYS_ERROR);
	break;
    }

#else  /*  HAS_SIGNALS  */
    fprintf (stderr, "Operating system does not support signal blocking\n");
#endif  /*  HAS_SIGNALS  */
}   /*  End Function s_unblock  */

/*PUBLIC_FUNCTION*/
void s_unblock_all ()
/*  [SUMMARY] Unblock all signals.
    [PURPOSE] This routine will cause all signals blocked by [<s_block>] to be
    unblocked. If no signals were blocked, no action is taken.
    [RETURNS] Nothing.
*/
{
#ifdef HAS_SIGNALS
    static char function_name[] = "s_unblock_all";

    if (initialised != TRUE)
    {
	fprintf (stderr, "No signals blocked yet\n");
	prog_bug (function_name);
    }
    /*  Unblock signals  */
    if (sigprocmask (SIG_UNBLOCK, &blocked_mask, (sigset_t *) NULL) != 0)
    {
	fprintf (stderr, "Error restoring old signal mask\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Initialise signal set  */
    if (sigemptyset (&blocked_mask) != 0)
    {
	fprintf (stderr, "Error clearing signal mask\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }

#else  /*  HAS_SIGNALS  */
    fprintf (stderr, "Operating system does not support signal blocking\n");
#endif  /*  HAS_SIGNALS  */
}   /*  End Function s_unblock_all  */

/*PUBLIC_FUNCTION*/
void *s_ignore (int sig)
/*  [SUMMARY] Cause a signal to be ignored.
    <sig> The signal to ignore.
    [RETURNS] The previous signal handler.
*/
{
#ifdef HAS_SIGNALS
    struct sigaction new_action, old_action;
    static char function_name[] = "s_ignore";

    new_action.sa_handler = SIG_IGN;
    sigemptyset (&new_action.sa_mask);
    new_action.sa_flags = 0;
    if (sigaction (sig, &new_action, &old_action) != 0)
    {
	fprintf (stderr, "%s: error setting signal handler\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    return (void *) old_action.sa_handler;

#else  /*  HAS_SIGNALS  */
    fprintf (stderr, "Operating system does not support signal ignoring\n");
    return (NULL);
#endif  /*  HAS_SIGNALS  */
}   /*  End Function s_ignore  */

#ifdef HAS_SIGNALS

static void prog_bug (function_name)
char *function_name;
{
    fprintf (stderr, "Program bug noted in function: %s\n", function_name);
    fprintf (stderr, "Aborting.%c\n", BEL);
    exit (RV_PROGRAM_BUG);
}   /*  End Function prog_bug   */

#endif  /*  HAS_SIGNALS  */
