/*LINTLIBRARY*/
/*  main.c

    This code provides KWorkFunc objects.

    Copyright (C) 1995-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the creation, manipulation and
  deletion of KWorkFunc objects.


    Written by      Richard Gooch   30-DEC-1995

    Updated by      Richard Gooch   13-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   15-DEC-1996: Renamed to <w> package and
  changed <w_register_notify_func> so that it does not return a KCallbackFunc.

    Updated by      Richard Gooch   8-JAN-1999: Created <w_unregister_support>.

    Updated by      Richard Gooch   12-JAN-1999: Do not do work if work
  functions disabled.

    Last updated by Richard Gooch   15-AUG-1999: Made egcs happy.


*/
#include <stdio.h>
#include <stdlib.h>
#include <karma.h>
#include <karma_w.h>


#define FUNC_MAGIC_NUMBER 1987342877

#define VERIFY_FUNC(func) if (func == NULL) \
{fprintf (stderr, "NULL KWorkFunc passed\n"); \
 prog_bug (function_name); } \
if (func->magic_number != FUNC_MAGIC_NUMBER ) \
{fprintf (stderr, "Invalid KWorkFunc object\n"); \
 prog_bug (function_name); }

struct work_func_type
{
    unsigned int magic_number;
    flag (*func) (void **info);
    void *info;
    unsigned int priority;
    KWorkFunc next;
    KWorkFunc prev;
};


/*  Private functions  */
STATIC_FUNCTION (void prog_bug, (char *function_name) );
STATIC_FUNCTION (void mem_abort, (char *name, char *reason) );


/*  Private data  */
static flag work_funcs_supported = FALSE;
static KWorkFunc first_func = NULL;
static KWorkFunc last_func = NULL;
static KWorkFunc executing_func = NULL;
static void (*notify_func) (void *info) = NULL;
static void *notify_info = NULL;


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
flag w_test_supported ()
/*  [SUMMARY] Test if application supports work functions.
    [PURPOSE] This routine will determine if work functions have been supported
    by the application.
    [RETURNS] TRUE if work functions are supported, else FALSE.
*/
{
    return (work_funcs_supported);
}   /*  End Function w_test_supported  */

/*PUBLIC_FUNCTION*/
void w_register_support ()
/*  [SUMMARY] Register support for work functions.
    [PURPOSE] This routine will inform the package that the application will
    support work functions. This routine may only be called once unless an
    intervening call to <w_unregister_support> is made.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "w_register_support";

    if (work_funcs_supported)
    {
	fprintf (stderr, "Work functions already supported\n");
	prog_bug (function_name);
    }
    work_funcs_supported = TRUE;
}   /*  End Function w_register_support  */

/*EXPERIMENTAL_FUNCTION*/
void w_unregister_support ()
/*  [SUMMARY] Unregister support for work functions.
    [PURPOSE] This routine will inform the package that the application no
    longer supports work functions. This routine may only be called once unless
    an intervening call to <w_register_support> is made.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "w_unregister_support";

    if (!work_funcs_supported)
    {
	fprintf (stderr, "Work functions not already supported\n");
	prog_bug (function_name);
    }
    work_funcs_supported = FALSE;
}   /*  End Function w_unregister_support  */

/*PUBLIC_FUNCTION*/
void w_register_notify_func (void (*func) (void *info), void *info)
/*  [SUMMARY] Register new work function callback.
    [PURPOSE] This routine will register a routine which is called whenever a
    new work function is registered.
    <func> The function that is called when a new work function is registered.
    The prototype function is [<w_PROTO_new_func>].
    <info> The initial arbitrary information pointer.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "w_register_notify_func";

    if (notify_func != NULL)
    {
	fprintf (stderr, "notify_func already registered");
	prog_bug (function_name);
    }
    notify_func = func;
    notify_info = info;
}   /*  End Function w_register_notify_func  */

/*PUBLIC_FUNCTION*/
flag w_work_to_be_done ()
/*  [SUMMARY] Test for pending work functions waiting to be called.
    [PURPOSE] This routine will determine if there are any pending work
    functions waiting to be called.
    [NOTE] Disabling support for work functions via <w_unregister_support> has
    no effect on this function.
    [RETURNS] TRUE if any work functions are pending, else FALSE.
*/
{
    return (first_func == NULL) ? FALSE : TRUE;
}   /*  End Function w_work_to_be_done  */

/*PUBLIC_FUNCTION*/
flag w_do_work ()
/*  [SUMMARY] Do some work.
    [PURPOSE] This routine will call any pending work functions. The work
    function at the top of the list is called once only.
    [RETURNS] TRUE if more work is left to be done, else FALSE.
*/
{
    flag more_to_do;
    static char function_name[] = "w_do_work";

    if (!work_funcs_supported) return (FALSE);
    if (first_func == NULL) return (FALSE);
    if (executing_func != NULL)
    {
	fprintf (stderr, "Not re-entrant!\n");
	prog_bug (function_name);
    }
    executing_func = first_func;
    more_to_do = (*executing_func->func) (&executing_func->info);
    executing_func = NULL;
    if (more_to_do) return (TRUE);
    w_unregister_func (first_func);
    return (first_func == NULL) ? FALSE : TRUE;
}   /*  End Function w_do_work  */

/*PUBLIC_FUNCTION*/
KWorkFunc w_register_func (flag (*func) (void **info), void *info,
			   unsigned int priority)
/*  [SUMMARY] Register a work function.
    [PURPOSE] This routine will register a work function. The work function at
    the top of the list is the one that is always called. Only when the work
    function at the top of the list is unregistered will the next work function
    be called.
    <func> The routine to be called. The prototype function is
    [<w_PROTO_work_func>].
    <info> The initial arbitrary work function information pointer.
    <priority> The priority level of the work function. See [<w_PRIORITIES>]
    for a list of legal values.
    [RETURNS] A KWorkFunc object. On failure, the process aborts.
*/
{
    KWorkFunc new;
    static char function_name[] = "w_register_func";

    if (!work_funcs_supported)
    {
	fprintf (stderr, "Work functions not supported by application\n");
	prog_bug (function_name);
    }
    if ( ( new = (KWorkFunc) malloc (sizeof *new) ) == NULL )
    {
	mem_abort (function_name, "KWorkFunc object");
    }
    new->func = func;
    new->info = info;
    new->priority = priority;
    new->magic_number = FUNC_MAGIC_NUMBER;
    if ( (executing_func == NULL) ||
	 (executing_func->priority != KWF_PRIORITY_HIGHEST) )
    {
	/*  Add to start of list  */
	new->prev = NULL;
	new->next = first_func;
	first_func = new;
	if (last_func == NULL) last_func = new;
	if (notify_func != NULL) (*notify_func) (notify_info);
	return (new);
    }
    /*  We have been called from inside a work function, which will not allow
	functions registered during its execution to be placed above it in the
	list. Hence place the new function second  */
    new->prev = executing_func;
    new->next = executing_func->next;
    executing_func->next = new;
    if (notify_func != NULL) (*notify_func) (notify_info);
    return (new);
}   /*  End Function w_register_func  */

/*PUBLIC_FUNCTION*/
void w_unregister_func (KWorkFunc wf)
/*  [SUMMARY] Unregister a work function.
    [PURPOSE] This routine will unregister a work function. The second function
    in the list will moved to the top of the list. A work function cannot
    unregister itself in this manner.
    <wf> The KWorkFunc object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "w_unregister_func";

    VERIFY_FUNC (wf);
    if (wf == executing_func)
    {
	fprintf (stderr, "Work procedure cannot unregister itself this way\n");
	prog_bug (function_name);
    }
    if (wf->prev == NULL) first_func = wf->next;
    else wf->prev->next = wf->next;
    if (wf->next == NULL) last_func = wf->prev;
    else wf->next->prev = wf->prev;
    wf->magic_number = 0;
    free ( (char *) wf );
}   /*  End Function w_unregister_func  */


/*  Private functions follow  */

static void prog_bug (char *function_name)
{
    fprintf (stderr, "Program bug noted in function: %s\n", function_name);
    fprintf (stderr, "Aborting.%c\n", BEL);
    abort ();
}   /*  End Function prog_bug   */

static void mem_abort (char *name, char *reason)
{
    fprintf (stderr, "Error allocating memory for: %s  for function: %s%c\n",
	     reason, name, BEL);
    fprintf (stderr, "Aborting.\n");
    exit (RV_MEM_ERROR);
}   /*  End Function mem_abort  */  
