/*LINTLIBRARY*/
/*  main.c

    This code provides astronomical projection support.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file provides astronomical projection support.


    Written by      Richard Gooch   3-MAY-1996

    Updated by      Richard Gooch   12-JUN-1996

    Updated by      Richard Gooch   14-JUN-1996: Moved into library as <wcs>
  package.

    Updated by      Richard Gooch   15-JUN-1996: Created <wcs_astro_destroy>
  and fixed bug in <transform_ra_dec_sin> when converting (A,D) to (X,Y).

    Updated by      Richard Gooch   21-JUN-1996: Created <wcs_astro_format_all>

    Updated by      Richard Gooch   28-JUN-1996: Fixed bug in
  <wcs_astro_transform> when either RA or DEC values not supplied.

    Updated by      Richard Gooch   5-AUG-1996: Fixed bug in
  <wcs_astro_format_dec> when dealing with {-1 < dec < 0}.

    Updated by      Richard Gooch   29-AUG-1996: Created
  <wcs_astro_get_preferred_units> routine.

    Updated by      Richard Gooch   25-SEP-1996: Changed to implement the new
  FITS-style co-ordinate handling, where dimension co-ordinates range from 0
  to length - 1.

    Updated by      Richard Gooch   28-SEP-1996: Supported offset angle axes.
  Fixed name for <wcs_astro_format_all> so that "astro" is a part of the name.

    Updated by      Richard Gooch   6-OCT-1996: Added support for Cartesian
  (*-CAR) projection.

    Updated by      Richard Gooch   7-OCT-1996: Added support for Digital Sky
  Survey plate solutions (i.e. x,y -> RA,DEC).

    Updated by      Richard Gooch   7-OCT-1996: Added support for Digital Sky
  Survey inverse plate solutions (i.e. RA,DEC -> x,y).

    Updated by      Richard Gooch   15-OCT-1996: Added support for Rectangular
  (*-ARC) projection.

    Updated by      Richard Gooch   14-NOV-1996: Fixed Digital Sky Survey
  pixel positions (was out by 0.5 pixels).

    Updated by      Richard Gooch   20-NOV-1996: Partial trapping of TOOBIG
  co-ordinates when converting RA,DEC to x,y. Threaded transformations. Fixed
  x,y <-> RA,DEC transformations around north and south poles.

    Updated by      Richard Gooch   27-NOV-1996: ZEA projection supplied by
  Vincent McIntyre.

    Updated by      Richard Gooch   11-DEC-1996: Replaced support for offset
  angle axes with generic linear co-ordinate support.

    Updated by      Richard Gooch   31-DEC-1996: Created
  <wcs_astro_format_time> function and used for "TIME" axes.

    Updated by      Richard Gooch   2-JAN-1997: Also recognise "METR/SEC" in
  <wcs_astro_get_preferred_units> function.

    Updated by      Richard Gooch   15-JAN-1997: Use default value if "CRVALn",
  "CRPIXn" or "CDELTn" keywords not present, in order to get around broken FITS
  files written by IRAF.

    Updated by      Richard Gooch   28-JAN-1997: Fixed dealing with how "extra"
  RA or DEC axes are treated: report pixel position of 0.0. Removed
  <wcs_astro_format_extra>.

    Updated by      Richard Gooch   29-JAN-1997: Fixed printing of "TIME" axis
  so that it deals with Modified Julian Date (in seconds).

    Updated by      Richard Gooch   30-JAN-1997: Fixed bug in
  <wcs_astro_format_all> where <transform_linear> was not called for non
  RA/DEC/VEL restrictions.

    Updated by      Richard Gooch   1-FEB-1997: Created
  <wcs_astro_get_associated>.

    Updated by      Richard Gooch   13-FEB-1997: Created
  <wcs_astro_get_equinox> and <wcs_astro_get_rotation> functions.

    Updated by      Richard Gooch   14-FEB-1997: Trapped TOOBIG values for
  velocity/frequency and linear transformations.

    Updated by      Richard Gooch   15-FEB-1997: Created
  <wcs_astro_get_reference> function.

    Updated by      Richard Gooch   16-FEB-1997: Created
  <wcs_astro_breakdown_ra> and <wcs_astro_breakdown_dec> functions.

    Updated by      Richard Gooch   22-FEB-1997: Modify duplicate CTYPEn's
  to contain a ":n" where n is the axis number.

    Updated by      Richard Gooch   23-FEB-1997: Fixed bug in
  <wcs_astro_transform3> where linear flags were not being properly passed to
  <transform_linear>.

    Updated by      Richard Gooch   24-FEB-1997: Created
  <wcs_astro_breakdown_time>.

    Updated by      Richard Gooch   25-FEB-1997: Added support for Digital Sky
  Survey mixed transformations (i.e. x,DEC -> RA,y and RA,y -> x,DEC).

    Updated by      Richard Gooch   27-FEB-1997: Fixed bug in
  <wcs_astro_format_all> where reference value was printed for "extra"
  information, rather than using a pixel position of 0.0 and transforming.

    Updated by      Richard Gooch   4-MAR-1997: Created
  <wcs_astro_update_location> function.

    Updated by      Richard Gooch   23-MAR-1997: Added preliminary support for
  CD matrix. RA,y -> x,DEC and x,DEC -> RA,y transforms not yet supported.

    Updated by      Richard Gooch   25-MAR-1997: Added support for Galactic
  co-ordinates.

    Updated by      Richard Gooch   26-MAR-1997: Fixed bug in x,y -> RA,DEC
  conversion for SIN projection.

    Updated by      Richard Gooch   9-APR-1997: Fixed bug when co-ordinates are
  from DSS.

    Updated by      Richard Gooch   11-APR-1997: Fixed bug when CROTA != 0
  because of flaw in Hanisch and Wells paper of Feb 1988. Refutation paper by
  Calabretta and Greisen confirmed my own assumptions.

    Updated by      Richard Gooch   13-APR-1997: Fixed bug in derivation of
  LYtoXM CD matrix.

    Updated by      Richard Gooch   18-APR-1997: Made use of <ds_format_unit>
  in <wcs_astro_format>.

    Updated by      Richard Gooch   22-JUN-1997: Took account of scaling of
  RA/DEC cell size with frequency.

    Updated by      Richard Gooch   2-JUL-1997: Use new "CELLSCAL" keyword
  to check for Miriad chromatic aberration.

    Updated by      Richard Gooch   3-JUL-1997: Use "VOBS" in correction for
  Miriad chromatic aberration.

    Updated by      Richard Gooch   5-AUG-1997: Added support for mixed
  conversions for CAR projection.

    Updated by      Richard Gooch   11-AUG-1997: Make use of "VELREF" keyword
  for "VELO" axes, distinguishes between radio and optical doppler conventions.
  Not needed except for determining RA/DEC cell scaling for datasets with
  chromatic aberration (i.e. Miriad, which only has radio convention for "VELO"
  axes anyway). Not needed now, but tidies things up for the future.

    Updated by      Richard Gooch   20-AUG-1997: Added tests to see if an axis
  is a longitude or latitude axis.

    Updated by      Richard Gooch   13-FEB-1998: Fixed bug in TAN projection
  when crossing south over the equator. Support mixed RA/DEC/GLON/GLAT,VEL
  images.

    Updated by      Richard Gooch   28-FEB-1998: Disable equinox for Galactic
  co-ordinates.

    Updated by      Richard Gooch   30-MAR-1998: Changed to EQUINOX_NONE for
  unrecognised equinox, rather than EQUINOX_J2000.

    Updated by      Richard Gooch   14-APR-1998: Changed from RA & DEC
  conversion functions to HMS & DMS.

    Updated by      Richard Gooch   23-APR-1998: Changed to signed integers for
  breakdown functions.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   14-JUL-1998: Display Galactic co-ordinates
  in decimal notation.

    Updated by      Richard Gooch   15-JUL-1998: Disabled Miriad CAR projection
  since it's non-standard.

    Updated by      Richard Gooch   17-JUL-1998: GLS projection supplied by
  Steven Gibson.

    Updated by      Richard Gooch   23-JUL-1998: No longer perform RA wrapping
  in projection functions, since this can cause problems with drawing tick
  marks in axis labels.

    Updated by      Richard Gooch   24-JUL-1998: Do not wrap output for Glon.

    Updated by      Richard Gooch   14-OCT-1998: Added timeout to function
  <transform_ra_dec_arc>.

    Updated by      Richard Gooch   26-NOV-1998: Added trap for blank values in
  <wcs_astro_format>: caused hanging.

    Updated by      Richard Gooch   3-FEB-1999: Removed experimental function
  <wcs_astro_get_rotation>: not used and not a robust solution to any problem.

    Updated by      Richard Gooch   2-JUN-1999: Fixed bug in TAN projection
  for x,y -> RA,DEC when crossing south over the equator. Added extra digit of
  precision for seconds in <wcs_astro_format_hms>.

    Updated by      Richard Gooch   8-JUL-1999: Fixed memory leak: extra
  information was not being deallocated.

    Updated by      Richard Gooch   1-OCT-1999: Work around degenerate sky axes
  with zero increment.

    Updated by      Richard Gooch   4-OCT-1999: Fixed <process_sky> to set
  equinox to EQUINOX_NONE if either sky axis is not RA or DEC.

    Updated by      Richard Gooch   8-JUN-2000: Converted DSS mixed conversion
  functions into generic functions and make use of these in projection
  functions when they can't do it themselves.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Updated by      Richard Gooch   16-JUN-2000: Took account of interface
  change to function <ds_format_unit>.

    Updated by      Richard Gooch   29-SEP-2000: Ensure output RA does not
  exceed 360 for NCP projection.

    Last updated by Richard Gooch   13-MAR-2001: Removed previously disabled
  Miriad CAR projection since it's non-standard. Miriad reader now removes this
  bogus (flat with longitude scaling) projection.


*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <errno.h>
#include <karma.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_mt.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>


#define MAX_AXES 100
#define LIGHTSPEED 299792458

/*  Projection types  */
#define PROJ_INIT    -1    /*  Initialised                                */
#define PROJ_UNKNOWN 0     /*  Unknown                                    */
#define PROJ_DSS     1     /*  Digital Sky Survey plate solutions         */
#define PROJ_SIN     2     /*  Orthographic                               */
#define PROJ_TAN     3     /*  Gnomonic                                   */
#define PROJ_NCP     4     /*  North Celestial Pole                       */
#define PROJ_FLT     5     /*  Flat                                       */
#define PROJ_ARC     6     /*  Rectangular                                */
#define PROJ_ZEA     7     /*  Zenithal Equal-Area                        */
#define PROJ_GLS     8     /*  Global Sinusoidal                          */

/*  Direction types  */
#define DIR_ADtoXY  0       /*  alpha,delta -> x,y          */
#define DIR_XDtoAY  1       /*  x,delta     -> alpha,y      */
#define DIR_AYtoXD  2       /*  alpha,y     -> x,delta      */
#define DIR_XYtoAD  3       /*  x,y         -> alpha,delta  */

/*  Spectral types  */
#define SPECTRAL_UNKNOWN           0  /*  Unknown                           */
#define SPECTRAL_VELn_FreqOptical  1  /*  m/s, step in F, optical, non-lin  */
#define SPECTRAL_VELl_FreqRadio    2  /*  m/s, step in F, radio, linear     */
#define SPECTRAL_VELl_WaveOptical  3  /*  m/s, step in wave., radio, linear */
#define SPECTRAL_FREQ              4  /*  Hz                                */

/*  Equinox types  */
#define EQUINOX_NONE   0
#define EQUINOX_J2000  1   /*  Julian 2000           */
#define EQUINOX_B1950  2   /*  Bessellian 1950       */

/*  DSS stuff  */
#define NUM_PLATE_COEFFS 20

#define ANGLE_LIMIT 3600.0  /*  Degrees (10 rotations)  */
#define TIME_LIMIT (ANGLE_LIMIT / 15.0)

#define MAGIC_NUMBER (unsigned int) 2076765342

#define VERIFY_AP(ap) if (ap == NULL) \
{fprintf (stderr, "NULL astro context passed\n"); \
 a_prog_bug (function_name); } \
if (ap->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid astro context object\n"); \
 a_prog_bug (function_name); }


/*  Structure declarations follow  */

struct sky_coord_type
{
    char dim_name[STRING_LENGTH];
    double reference;        /*  Degrees  (CRVALn) in degrees               */
    double ref_pos;          /*  Pixel position of reference (CRPIXn - 1)   */
    double increment;        /*  Increment between pixels (CDELTn)          */
    double pix_to_rad;       /*  Conversion from pixel to radian            */
    double rad_to_pix;       /*  Conversion from radian to pixel            */
    double sin_ref;
    double cos_ref;
    unsigned int axis_number;/*  The FITS axis number                       */
    flag is_sky;             /*  TRUE if sky co-ordinate, FALSE if galactic */
};

struct cd_matrix_type
{
    flag defined;
    double cd_11;
    double cd_12;
    double cd_21;
    double cd_22;
};

typedef void (*RaDecTransformFunc) (KwcsAstro ap, unsigned int num_coords,
				    double *ra, double *dec,
				    unsigned int direction);

struct sky_type
{
    double sin_rotation;
    double cos_rotation;
    int projection;
    struct cd_matrix_type LMtoXY;
    struct cd_matrix_type XMtoLY;
    struct cd_matrix_type LYtoXM;
    struct cd_matrix_type XYtoLM;
    unsigned int equinox;
    RaDecTransformFunc transform_func;
    struct sky_coord_type ra;
    struct sky_coord_type dec;
    struct dss_plate_type *dss;
    /*  Buffer for unspecified RA or DEC co-ordinates  */
    double *radec_buffer;
    unsigned int radec_buf_len;  /*  Number of doubles  */
};

struct spectral_coord_type
{
    char dim_name[STRING_LENGTH];
    unsigned int type;
    double reference;          /*  Reference pixel (CRVALn)                  */
    double ref_pos;            /*  Pixel position of reference (CRPIXn - 1)  */
    double increment;          /*  Increment between pixels (CDELTn)         */
    double restfreq;           /*  Rest frequency (Hz)                       */
    double vobs;               /*  Average velocity of observation (m/s)     */
    unsigned int axis_number;  /*  The FITS axis number                      */
    flag chromatic_aberration; /* If the dataset has Miriad chromatic aberr. */
    /*  Buffer for scaling RA and DEC cells depending on frequency  */
    double *radec_scale_buffer;
    unsigned int radec_scale_buf_len;  /*  Number of doubles  */
};

struct linear_coord_type
{
    char dim_name[STRING_LENGTH];
    double reference;        /*  Reference pixel (CRVALn)                   */
    double ref_pos;          /*  Pixel position of reference (CRPIXn - 1)   */
    double increment;        /*  Increment between pixels (CDELTn)          */
    unsigned int axis_number;/*  The FITS axis number                       */
    struct linear_coord_type *next;
};

struct extra_type
{
    char ctype[STRING_LENGTH]; /* Axis name (CTYPEn), possibly with "name:n" */
    double refval;             /*  Reference pixel (CRVALn)                  */
    double refpos;             /*  Pixel position of reference (CRPIXn - 1)  */
    unsigned int axis_number;  /*  The FITS axis number                      */
    struct extra_type *next;
};

struct dss_plate_type  /*  Digital Sky Survey plate solution  */
{
    double corner_pixel_x1;            /*  Pixels (CNPIX1 + 0.5)             */
    double corner_pixel_y1;            /*  Pixels (CNPIX2 + 0.5)             */
    double scale;                      /*  arcsec/mm                         */
    double pixel_size_x;               /*  Microns                           */
    double pixel_size_y;               /*  Microns                           */
    double centre_ra;                  /*  Degrees                           */
    double centre_dec;                 /*  Degrees                           */
    double centre_off_x;               /*  Microns                           */
    double centre_off_y;               /*  Microns                           */
    unsigned int num_amdx;
    double amdx[NUM_PLATE_COEFFS];     /*  Plate solution Xi co-efficients   */
    unsigned int num_amdy;
    double amdy[NUM_PLATE_COEFFS];     /*  Plate solution Eta co-efficients  */
};

typedef struct
{
    KwcsAstro ap;
    unsigned int direction;
    unsigned int num_coords;
} CommonJobInfo;

struct astro_projection_type
{
    /*  Common information  */
    unsigned int magic_number;
    /*  Sky co-ordinate information  */
    struct sky_type sky;
    /*  Velocity axis  */
    struct spectral_coord_type vel;
    /*  Linear co-ordinate axes  */
    struct linear_coord_type *linear_axes;
    /*  Extra information  */
    struct extra_type *first_extra;
};

/*  Note that the header information taken from a Karma data structure is
    assumed to be in FITS format. The readers for foreign data (e.g. Miriad)
    perform on-the-fly conversions to FITS, so what is left in the Karma data
    structure appears to have come from FITS.
    Below are some units used in FITS:
        Right Ascension         Degrees
	Declination             Degrees
	Velocity                m/s (meters per second)
	Temperature             Kelvin
	LL-tangent plane E-W    Degrees
	MM-tangent plane N-S    Degrees
*/


/*  Private data  */


/*  Private functions  */
STATIC_FUNCTION (KwcsAstro create_new, () );
STATIC_FUNCTION (flag process_axis,
		 (KwcsAstro ap,
		  CONST packet_desc *pack_desc, CONST char *packet,
		  CONST char *axis_name, unsigned int index) );
STATIC_FUNCTION (void add_extra,
		 (KwcsAstro ap, CONST char *ctype, double refval,
		  double refpos, unsigned int axis_number) );
STATIC_FUNCTION (void get_axis_names,
		 (CONST packet_desc *pack_desc, CONST char *packet,
		  char axis_names[MAX_AXES][STRING_LENGTH]) );
STATIC_FUNCTION (flag process_sky, (KwcsAstro ap, CONST packet_desc *pack_desc,
				    CONST char *packet) );
STATIC_FUNCTION (flag fill_radec_buffer,
		 (KwcsAstro ap, CONST char *dim_name, unsigned int num_coords,
		  unsigned int num_restr, CONST char **restr_names,
		  CONST double *restr_values) );
STATIC_FUNCTION (void get_radec_scale,
		 (KwcsAstro ap, unsigned int num_coords, double *scales,
		  CONST double *velocities) );
STATIC_FUNCTION (CONST char* find_unspecified,
		 (KwcsAstro ap, CONST char *dim_name, double *value,
		  unsigned int num_restr, CONST char **restr_names,
		  CONST double *restr_values) );
STATIC_FUNCTION (void coords_to_radecvel,
		 (KwcsAstro ap,
		  CONST char *name0, double *coords0, flag to_lin0,
		  CONST char *name1, double *coords1, flag to_lin1,
		  CONST char *name2, double *coords2, flag to_lin2,
		  unsigned int num_restr, CONST char **restr_names,
		  CONST double *restr_values,
		  double **ra, flag *ra_to_linear,
		  double **dec, flag *dec_to_linear,
		  double **vel, flag *vel_to_linear) );
STATIC_FUNCTION (void transform_ra_dec_sin,
		 (KwcsAstro ap, unsigned int num_coords,
		  double *ra, double *dec, unsigned int direction) );
STATIC_FUNCTION (void transform_ra_dec_tan,
		 (KwcsAstro ap, unsigned int num_coords,
		  double *ra, double *dec, unsigned int direction) );
STATIC_FUNCTION (void transform_ra_dec_ncp,
		 (KwcsAstro ap, unsigned int num_coords,
		  double *ra, double *dec, unsigned int direction) );
STATIC_FUNCTION (void transform_ra_dec_flat,
		 (KwcsAstro ap, unsigned int num_coords,
		  double *ra, double *dec, unsigned int direction) );
STATIC_FUNCTION (void transform_ra_dec_arc,
		 (KwcsAstro ap, unsigned int num_coords,
		  double *ra, double *dec, unsigned int direction) );
STATIC_FUNCTION (void transform_ra_dec_zea,
		 (KwcsAstro ap, unsigned int num_coords,
		  double *ra, double *dec, unsigned int direction) );
STATIC_FUNCTION (void transform_ra_dec_gls,
		 (KwcsAstro ap, unsigned int num_coords,
		  double *ra, double *dec, unsigned int direction) );
STATIC_FUNCTION (void transform_vel,
		 (KwcsAstro ap, unsigned int num_coords, double *vel,
		  flag to_linear) );
STATIC_FUNCTION (void transform_linear,
		 (KwcsAstro ap, unsigned int num_coords,
		  CONST char *name, double *coords, flag to_pix) );
STATIC_FUNCTION (flag scan_dss_header,
		 (KwcsAstro ap, CONST packet_desc *pack_desc,
		  CONST char *packet) );
STATIC_FUNCTION (void transform_ra_dec_dss,
		 (KwcsAstro ap, unsigned int num_coords,
		  double *ra, double *dec, unsigned int direction) );
STATIC_FUNCTION (void dss_xy2radec, (KwcsAstro ap, unsigned int num_coords,
				     double *ra, double *dec) );
STATIC_FUNCTION (void dss_radec2xy, (KwcsAstro ap, unsigned int num_coords,
				     double *ra, double *dec) );
STATIC_FUNCTION (void generic_xdec2ray, (KwcsAstro ap, unsigned int num_coords,
					 double *ra_arr, double *dec_arr) );
STATIC_FUNCTION (void generic_ray2xdec, (KwcsAstro ap, unsigned int num_coords,
					 double *ra_arr, double *dec_arr) );
STATIC_FUNCTION (void job_func,
		 (void *pool_info, void *call_info1, void *call_info2,
		  void *call_info3, void *call_info4, void *thread_info) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
KwcsAstro wcs_astro_get_associated (multi_array *multi_desc,unsigned int index)
/*  [SUMMARY] Create/get associated Astronomical Sky Co-ordinate projections.
    [PURPOSE] This routine will attempt create/get the associated Astronomical
    Sky Co-ordinate projection information for a general data structure. If the
    projection information is found, it is associated with the general data
    structure in the multi_array descriptor, to speed subsequent accesses.
    <multi_desc> The multi_array descriptor.
    <index> The index of the general data structure in the descriptor.
    [RETURNS] A KwcsAstro object on success, else NULL.
*/
{
    KwcsAstro ap;
    void *object;
    static char function_name[] = "wcs_astro_get_associated";

    if (multi_desc->magic_number != MAGIC_MULTI_ARRAY)
    {
	fprintf (stderr, "Bad magic number\n");
	a_prog_bug (function_name);
    }
    if ( ds_get_named_object (multi_desc, index, "astro_projection", &object) )
    {
	ap = object;
	return (ap);
    }
    /*  First time for this data structure  */
    ap = wcs_astro_setup (multi_desc->headers[index], multi_desc->data[index]);
    if (ap == NULL)
    {
	if ( !ds_add_named_object (multi_desc, index, "astro_projection",
				   ap, ( void (*) () ) NULL) )
	{
	    m_abort (function_name, "projection object");
	}
    }
    else
    {
	if ( !ds_add_named_object (multi_desc, index, "astro_projection",
				   ap, wcs_astro_destroy) )
	{
	    m_abort (function_name, "projection object");
	}
    }
    return (ap);
}   /*  End Function wcs_astro_get_associated  */

/*EXPERIMENTAL_FUNCTION*/
KwcsAstro wcs_astro_setup (CONST packet_desc *pack_desc, CONST char *packet)
/*  [SUMMARY] Setup for Astronomical Sky Co-ordinate projections.
    <pack_desc> The packet descriptor containing the FITS-style keywords.
    <packet> The packet data containing the FITS-stype keyword data.
    [RETURNS] A KwcsAstro object on success, else NULL.
*/
{
    KwcsAstro new;
    int equinox_val;
    unsigned int count;
    char axis_names[MAX_AXES][STRING_LENGTH];
    double value[2];
    static char function_name[] = "wcs_astro_setup";

    if ( (pack_desc == NULL) || (packet == NULL) )
    {
	fputs ("NULL pointer(s) passed\n", stderr);
	a_prog_bug (function_name);
    }
    /*  Theoretically, one should use the "NAXIS" keyword to determine how many
	axes there are to check, but since perhaps not everyone writes this
	keyword, I just keep searching until there are no more "CTYPEn"
	keywords. If "CTYPEn" does not exist, assume "CTYPEm" (where m > n)
	does not exist.
	Note that this whole scheme depends on the Karma dimension names being
	the same as those in the FITS header. For Digital Sky Survey data,
	which doesn't use "CTYPEn" and friends, this assumption is irrelevant
	*/
    get_axis_names (pack_desc, packet, axis_names);
    if (axis_names[0][0] == '\0')
    {
	/*  It's not a conventional header: perhaps it's DSS data?  */
	if ( !ds_get_unique_named_value (pack_desc, packet, "PLTSCALE", NULL,
					 value) ) return (NULL);
    }
    if ( ( new = create_new () ) == NULL ) return (NULL);
    equinox_val = 2000;
    if ( ds_get_unique_named_value (pack_desc, packet, "EQUINOX", NULL,value) )
    {
	/*  EQUINOX: preferred FITS keyword  */
	equinox_val = value[0];
    }
    else
    {
	/*  Not EQUINOX  */
	if ( ds_get_unique_named_value (pack_desc, packet, "EPOCH", NULL,
					value) )
	{
	    /*  EPOCH: old FITS keyword  */
	    equinox_val = value[0];
	}
    }
    if (equinox_val == 2000) new->sky.equinox = EQUINOX_J2000;
    else if (equinox_val == 1950) new->sky.equinox = EQUINOX_B1950;
    else
    {
	fprintf (stderr, "Unknown equinox: %d\n", equinox_val);
	new->sky.equinox = EQUINOX_NONE;
    }
    /*  Start searching for and processing RA and DEC information  */
    for (count = 0; axis_names[count][0] != '\0'; ++count)
    {
	if ( !process_axis (new, pack_desc, packet, axis_names[count],
			    count + 1) )
	{
	    m_free (new);
	    return (NULL);
	}
    }
    if ( !process_sky (new, pack_desc, packet) )
    {
	m_free (new);
	return (NULL);
    }
    if ( (new->sky.projection == PROJ_UNKNOWN) &&
	 (new->vel.type == SPECTRAL_UNKNOWN) && (new->linear_axes == NULL) )
    {
	m_free (new);
	return (NULL);
    }
    new->magic_number = MAGIC_NUMBER;
    return (new);
}   /*  End Function wcs_astro_setup  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_destroy (KwcsAstro ap)
/*  [SUMMARY] Destroy KwcsAstro object.
    <ap> The KwcsAstro object.
    [RETURNS] Nothing.
*/
{
    struct linear_coord_type *l_curr, *l_next;
    struct extra_type *e_curr, *e_next;
    static char function_name[] = "wcs_astro_destroy";

    VERIFY_AP (ap);
    for (l_curr = ap->linear_axes; l_curr != NULL; l_curr = l_next)
    {
	l_next = l_curr->next;
	m_free (l_curr);
    }
    for (e_curr = ap->first_extra; e_curr != NULL; e_curr = e_next)
    {
	e_next = e_curr->next;
	m_free (e_curr);
    }
    m_clear (ap, sizeof *ap);
    m_free (ap);
}   /*  End Function wcs_astro_destroy  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_test_radec (KwcsAstro ap)
/*  [SUMMARY] Test if a KwcsAstro object has RA/GLON and DEC/GLAT defined.
    <ap> The KwcsAstro object. This may be NULL.
    [RETURNS] TRUE if the KwcsAstro object has RA and DEC defined, else FALSE.
*/
{
    static char function_name[] = "wcs_astro_test_radec";

    if (ap == NULL) return (FALSE);
    VERIFY_AP (ap);
    if (ap->sky.projection == PROJ_UNKNOWN) return (FALSE);
    return (TRUE);
}   /*  End Function wcs_astro_test_radec  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_test_velocity (KwcsAstro ap)
/*  [SUMMARY] Test if a KwcsAstro object has velocity or frequency defined.
    <ap> The KwcsAstro object. This may be NULL.
    [RETURNS] TRUE if the KwcsAstro object has velocity or frequency defined,
    else FALSE.
*/
{
    static char function_name[] = "wcs_astro_test_velocity";

    if (ap == NULL) return (FALSE);
    VERIFY_AP (ap);
    if (ap->vel.type == SPECTRAL_UNKNOWN) return (FALSE);
    return (TRUE);
}   /*  End Function wcs_astro_test_velocity  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_test_if_ra (KwcsAstro ap, CONST char *axis_name)
/*  [SUMMARY] Test if an axis is a Right Ascension axis.
    <ap> The KwcsAstro object. This may be NULL.
    <axis_name> The name of the axis.
    [RETURNS] TRUE the axis is a Right Ascension axis, else FALSE.
*/
{
    static char function_name[] = "wcs_astro_test_if_ra";

    if (ap == NULL)
    {
	/*  Basic test  */
	if (strncmp (axis_name, "RA", 2) == 0) return (TRUE);
	return (FALSE);
    }
    VERIFY_AP (ap);
    if (!ap->sky.ra.is_sky) return (FALSE);
    if (strcmp (axis_name, ap->sky.ra.dim_name) == 0) return (TRUE);
    return (FALSE);
}   /*  End Function wcs_astro_test_if_ra  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_test_if_dec (KwcsAstro ap, CONST char *axis_name)
/*  [SUMMARY] Test if an axis is a Declination axis.
    <ap> The KwcsAstro object. This may be NULL.
    <axis_name> The name of the axis.
    [RETURNS] TRUE the axis is a Declination axis, else FALSE.
*/
{
    static char function_name[] = "wcs_astro_test_if_dec";

    if (ap == NULL)
    {
	/*  Basic test  */
	if (strncmp (axis_name, "DEC", 3) == 0) return (TRUE);
	return (FALSE);
    }
    VERIFY_AP (ap);
    if (!ap->sky.dec.is_sky) return (FALSE);
    if (strcmp (axis_name, ap->sky.dec.dim_name) == 0) return (TRUE);
    return (FALSE);
}   /*  End Function wcs_astro_test_if_dec  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_test_if_glon (KwcsAstro ap, CONST char *axis_name)
/*  [SUMMARY] Test if an axis is a Galactic Longitude axis.
    <ap> The KwcsAstro object. This may be NULL.
    <axis_name> The name of the axis.
    [RETURNS] TRUE the axis is a Galactic Longitude axis, else FALSE.
*/
{
    static char function_name[] = "wcs_astro_test_if_glon";

    if (ap == NULL)
    {
	/*  Basic test  */
	if (strncmp (axis_name, "GLON", 4) == 0) return (TRUE);
	return (FALSE);
    }
    VERIFY_AP (ap);
    if (ap->sky.ra.is_sky) return (FALSE);
    if (strcmp (axis_name, ap->sky.ra.dim_name) == 0) return (TRUE);
    return (FALSE);
}   /*  End Function wcs_astro_test_if_glon  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_test_if_glat (KwcsAstro ap, CONST char *axis_name)
/*  [SUMMARY] Test if an axis is a Galactic Latitude axis.
    <ap> The KwcsAstro object. This may be NULL.
    <axis_name> The name of the axis.
    [RETURNS] TRUE the axis is a Galactic Latitude axis, else FALSE.
*/
{
    static char function_name[] = "wcs_astro_test_if_flat";

    if (ap == NULL)
    {
	/*  Basic test  */
	if (strncmp (axis_name, "GLAT", 4) == 0) return (TRUE);
	return (FALSE);
    }
    VERIFY_AP (ap);
    if (ap->sky.dec.is_sky) return (FALSE);
    if (strcmp (axis_name, ap->sky.dec.dim_name) == 0) return (TRUE);
    return (FALSE);
}   /*  End Function wcs_astro_test_if_glat  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_test_if_lon (KwcsAstro ap, CONST char *axis_name)
/*  [SUMMARY] Test if an axis is a longitude axis.
    <ap> The KwcsAstro object. This may be NULL.
    <axis_name> The name of the axis.
    [RETURNS] TRUE the axis is a longitude axis, else FALSE.
*/
{
    static char function_name[] = "wcs_astro_test_if_lon";

    if (ap == NULL)
    {
	/*  Basic tests  */
	if (strncmp (axis_name, "RA", 2) == 0) return (TRUE);
	if (strncmp (axis_name, "GLON", 4) == 0) return (TRUE);
	return (FALSE);
    }
    VERIFY_AP (ap);
    if (strcmp (axis_name, ap->sky.ra.dim_name) == 0) return (TRUE);
    return (FALSE);
}   /*  End Function wcs_astro_test_if_lon  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_test_if_lat (KwcsAstro ap, CONST char *axis_name)
/*  [SUMMARY] Test if an axis is a latitude axis.
    <ap> The KwcsAstro object. This may be NULL.
    <axis_name> The name of the axis.
    [RETURNS] TRUE the axis is a latitude axis, else FALSE.
*/
{
    static char function_name[] = "wcs_astro_test_if_lat";

    if (ap == NULL)
    {
	/*  Basic tests  */
	if (strncmp (axis_name, "DEC", 3) == 0) return (TRUE);
	if (strncmp (axis_name, "GLAT", 4) == 0) return (TRUE);
	return (FALSE);
    }
    VERIFY_AP (ap);
    if (strcmp (axis_name, ap->sky.dec.dim_name) == 0) return (TRUE);
    return (FALSE);
}   /*  End Function wcs_astro_test_if_lat  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_transform (KwcsAstro ap, unsigned int num_coords,
			  double *ra, flag ra_to_linear,
			  double *dec, flag dec_to_linear,
			  double *vel, flag vel_to_linear,
			  unsigned int num_restr, CONST char **restr_names,
			  CONST double *restr_values)
/*  [SUMMARY] Transform between linear and projected co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <ra_to_linear> If TRUE the right ascension values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <dec> A pointer to the declination values. These will be modified.
    <dec_to_linear> If TRUE the declination values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <vel> A pointer to the velocity values. These will be modified.
    <vel_to_linear> If TRUE the velocity values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <num_restr> The number of restrictions.
    <restr_names> The array of restriction names.
    <restr_values> The restriction values.
    [RETURNS] Nothing.
*/
{
    flag freq_scaling = FALSE;
    KThreadPool shared_pool;
    unsigned int num_threads;
    unsigned int direction, count;
    CommonJobInfo common_info;
    double const_vel;
    double *radec_scale = NULL;
    static char function_name[] = "wcs_astro_transform";

    VERIFY_AP (ap);
    if ( ap->vel.chromatic_aberration &&
	 ( (ra != NULL) || (dec != NULL) ) &&
	 (ap->vel.type != SPECTRAL_UNKNOWN) )
    {
	/*  Have to scale RA/DEC values by frequency  */
	/*  First make sure buffer is large enough  */
	if (num_coords > ap->vel.radec_scale_buf_len)
	{
	    if (ap->vel.radec_scale_buffer != NULL)
		m_free (ap->vel.radec_scale_buffer);
	    if ( ( ap->vel.radec_scale_buffer =
		   m_alloc (num_coords * sizeof *ap->vel.radec_scale_buffer) )
		 == NULL )
	    {
		m_abort (function_name, "RA/DEC scale buffer");
	    }
	    ap->vel.radec_scale_buf_len = num_coords;
	}
	freq_scaling = TRUE;
	radec_scale = ap->vel.radec_scale_buffer;
    }
    if (vel == NULL)
    {
	if (freq_scaling)
	{
	    if (find_unspecified (ap, ap->vel.dim_name, &const_vel,
				  num_restr, restr_names,restr_values) == NULL)
	    {
		freq_scaling = FALSE;
	    }
	    else
	    {
		for (count = 0; count < num_coords; ++count)
		    radec_scale[count] = const_vel;
		transform_vel (ap, num_coords, radec_scale, FALSE);
		get_radec_scale (ap, num_coords, radec_scale, radec_scale);
	    }
	}
    }
    else
    {
	/*  Transform velocity  */
	if (vel_to_linear && freq_scaling)
	    get_radec_scale (ap, num_coords, radec_scale, vel);
	transform_vel (ap, num_coords, vel, vel_to_linear);
	if (!vel_to_linear && freq_scaling)
	    get_radec_scale (ap, num_coords, radec_scale, vel);
    }
    if ( (ra == NULL) && (dec == NULL) ) return;
    /*  Transform RA and DEC  */
    /*  At this point at least one of ra and dec are not NULL  */
    if (ra == NULL)
    {
	/*  No RA array: construct a fake array filled with the reference
	    value in the "extra" section  */
	if ( !fill_radec_buffer (ap, ap->sky.ra.dim_name, num_coords,
				 num_restr, restr_names, restr_values) )
	{
	    fprintf (stderr, "%s: RA information missing\n", function_name);
	    return;
	}
	ra = ap->sky.radec_buffer;
	ra_to_linear = FALSE;
    }
    else if (!ra_to_linear && freq_scaling)
    {
	/*  Pixel -> RA and callee needs RA info: apply scaling first  */
	for (count = 0; count < num_coords; ++count)
	{
	    ra[count] = (ra[count] - ap->sky.ra.ref_pos) *
		radec_scale[count] + ap->sky.ra.ref_pos;
	}
    }
    if (dec == NULL)
    {
	/*  No DEC array: construct a fake array filled with the reference
	    value in the "extra" section  */
	if ( !fill_radec_buffer (ap, ap->sky.dec.dim_name, num_coords,
				 num_restr, restr_names, restr_values) )
	{
	    fprintf (stderr, "%s: DEC information missing\n", function_name);
	    return;
	}
	dec = ap->sky.radec_buffer;
	dec_to_linear = FALSE;
    }
    else if (!dec_to_linear && freq_scaling)
    {
	/*  Pixel -> DEC and callee needs DEC info: apply scaling first  */
	for (count = 0; count < num_coords; ++count)
	{
	    dec[count] = (dec[count] - ap->sky.dec.ref_pos) *
		radec_scale[count] + ap->sky.dec.ref_pos;
	}
    }
    /*  Convert flags to direction. Must wait until RA and DEC arrays have been
	checked  */
    if (!ra_to_linear && !dec_to_linear) direction = DIR_XYtoAD;
    else if (!ra_to_linear && dec_to_linear)
    {
	direction = DIR_XDtoAY;
	if ( !ap->sky.XMtoLY.defined && (ap->sky.projection != PROJ_DSS) )
	{
	    fprintf (stderr, "%s: no XMtoLY matrix\n", function_name);
	    return;
	}
	if (ap->sky.cos_rotation == 0.0)
	{
	    fprintf (stderr, "%s: bad rotation\n", function_name);
	    return;
	}
    }
    else if (ra_to_linear && !dec_to_linear)
    {
	direction = DIR_AYtoXD;
	if ( !ap->sky.LYtoXM.defined && (ap->sky.projection != PROJ_DSS) )
	{
	    fprintf (stderr, "%s: no LYtoXM matrix\n", function_name);
	    return;
	}
	if (ap->sky.cos_rotation == 0.0)
	{
	    fprintf (stderr, "%s: bad rotation\n", function_name);
	    return;
	}
    }
    else
    {
	direction = DIR_ADtoXY;
	if ( !ap->sky.LMtoXY.defined && (ap->sky.projection != PROJ_DSS) )
	{
	    fprintf (stderr, "%s: no LMtoXY matrix\n", function_name);
	    return;
	}
    }
    if (ap->sky.transform_func == NULL)
    {
	fprintf (stderr, "%s: unknown projection: %u\n",
		 function_name, ap->sky.projection);
	return;
    }
    shared_pool = mt_get_shared_pool ();
    num_threads = mt_num_threads (shared_pool);
    if ( (num_coords < 100) || (num_threads < 2) ||
	 (num_threads > num_coords) )
    {
	/*  Don't bother threading  */
	(*ap->sky.transform_func) (ap, num_coords, ra, dec, direction);
    }
    else
    {
	/*  Threaded  */
	common_info.ap = ap;
	common_info.direction = direction;
	common_info.num_coords = num_coords / num_threads;
	for (count = 0; count < num_threads; ++count)
	{
	    mt_launch_job (shared_pool, job_func, &common_info,
			   ra, dec, NULL);
	    ra += common_info.num_coords;
	    dec += common_info.num_coords;
	    num_coords -= common_info.num_coords;
	}
	mt_wait_for_all_jobs (shared_pool);
	if (num_coords > 0)
	{
	    /*  Cleanup residual  */
	    (*ap->sky.transform_func) (ap, num_coords, ra, dec, direction);
	}
    }
    if (ra_to_linear && (ra != ap->sky.radec_buffer) && freq_scaling)
    {
	/*  RA -> pixel and callee needs pixel info: apply scaling now  */
	for (count = 0; count < num_coords; ++count)
	{
	    ra[count] = (ra[count] - ap->sky.ra.ref_pos) /
		radec_scale[count] + ap->sky.ra.ref_pos;
	}
    }
    if (dec_to_linear && (dec != ap->sky.radec_buffer) && freq_scaling)
    {
	/*  DEC -> pixel and callee needs pixel info: apply scaling now  */
	for (count = 0; count < num_coords; ++count)
	{
	    dec[count] = (dec[count] - ap->sky.dec.ref_pos) /
		radec_scale[count] + ap->sky.dec.ref_pos;
	}
    }
}   /*  End Function wcs_astro_transform  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_transform3 (KwcsAstro ap, unsigned int num_coords,
			   CONST char *name0, double *coords0, flag to_lin0,
			   CONST char *name1, double *coords1, flag to_lin1,
			   CONST char *name2, double *coords2, flag to_lin2,
			   unsigned int num_restr, CONST char **restr_names,
			   CONST double *restr_values)
/*  [SUMMARY] Transform between linear and projected co-ordinates.
    [PURPOSE] This routine will convert up to three arrays of co-ordinates
    using an astronomical projection system. The co-ordinate arrays are named.
    If a co-ordinate array is not defined the co-ordinates are unchanged.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <name0> The name of co-ordinate array 0.
    <coords0> Co-ordinate array 0.
    <to_lin0> If TRUE the co-ordinate 0 values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <name1> The name of co-ordinate array 1.
    <coords1> Co-ordinate array 1.
    <to_lin1> If TRUE the co-ordinate 1 values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <name2> The name of co-ordinate array 2.
    <coords2> Co-ordinate array 2.
    <to_lin2> If TRUE the co-ordinate 2 values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <num_restr> The number of restrictions.
    <restr_names> The array of restriction names.
    <restr_values> The restriction values.
    [RETURNS] Nothing.
*/
{
    flag ra_to_linear;
    flag dec_to_linear;
    flag vel_to_linear;
    double *ra;
    double *dec;
    double *vel;
    static char function_name[] = "wcs_astro_transform3";

    VERIFY_AP (ap);
    coords_to_radecvel (ap,
			name0, coords0, to_lin0,
			name1, coords1, to_lin1,
			name2, coords2, to_lin2,
			num_restr, restr_names, restr_values,
			&ra, &ra_to_linear,
			&dec, &dec_to_linear,
			&vel, &vel_to_linear);
    if ( (ra != NULL) || (dec != NULL) || (vel != NULL) )
	wcs_astro_transform (ap, num_coords, ra, ra_to_linear,
			     dec, dec_to_linear, vel, vel_to_linear,
			     num_restr, restr_names, restr_values);
    transform_linear (ap, num_coords, name0, coords0, to_lin0);
    transform_linear (ap, num_coords, name1, coords1, to_lin1);
    transform_linear (ap, num_coords, name2, coords2, to_lin2);
}   /*  End Function wcs_astro_transform3  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_breakdown_hms (int *hours, int *minutes,
			      double *seconds, int *whole_seconds,
			      double *subseconds, double input, double offset)
/*  [SUMMARY] Breakdown a time value into components.
    <hours> The hours are written here. This may be NULL.
    <minutes> The minutes are written here. This may be NULL.
    <seconds> The seconds are written here. This may be NULL.
    <whole_seconds> The integer truncated seconds are written here. This may be
    NULL.
    <subseconds> The fractional seconds are written here. This may be NULL.
    <input> The time value in hours.
    <offset> The offset (in seconds) to apply for rounding purposes.
    [NOTE] All component values are guaranteed to be positive.
    [RETURNS] TRUE if the value was legal, else FALSE.
*/
{
    double h, m, s, trunc_seconds, sub_seconds;

    while (input < 0.0) input += 24.0;
    if (fabs (input) > TIME_LIMIT)
    {
	return (FALSE);
    }
    input += 1e-10;  /*  To avoid roundoff  */
    input += offset / 3600.0;
    /*  Extract hours, minutes and seconds  */
    h = floor (input);
    if (hours != NULL) *hours = h;
    input -= h;
    input *= 60.0;
    m = floor (input);
    if (minutes != NULL) *minutes = m;
    input -= m;
    input *= 60.0;
    input -= offset;
    s = input;
    if (s < 0) s = 0.0;
    if (seconds != NULL) *seconds = s;
    trunc_seconds = floor (s);
    if (whole_seconds != NULL) *whole_seconds = trunc_seconds;
    sub_seconds = s - trunc_seconds;
    if (subseconds != NULL) *subseconds = sub_seconds;
    return (TRUE);
}   /*  End Function wcs_astro_breakdown_hms  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_format_hms (char *string, double hours)
/*  [SUMMARY] Format a time value into a string.
    <string> The string to write to. Sufficient storage must be available.
    <hours> The time value in hours.
    [RETURNS] Nothing.
*/
{
    int h, m;
    double s;

    if ( !wcs_astro_breakdown_hms (&h, &m, &s, NULL, NULL, hours, 0.005) )
    {
	strcpy (string, "ERROR");
	return;
    }
    sprintf (string, "%.2dh %.2dm %6.3fs", h, m, s);
}   /*  End Function wcs_astro_format_hms  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_breakdown_dms (flag *negative, int *degrees, int *minutes,
			      double *seconds, int *whole_seconds,
			      double *subseconds, double input, double offset)
/*  [SUMMARY] Break down an angular value into degrees, minutes and seconds.
    <negative> If the value is negative the value TRUE is written here, else
    FALSE is written here. This may be NULL.
    <degrees> The degrees are written here. This may be NULL.
    <minutes> The minutes are written here. This may be NULL.
    <seconds> The seconds are written here. This may be NULL.
    <whole_seconds> The integer truncated seconds are written here. This may be
    NULL.
    <subseconds> The fractional seconds are written here. This may be NULL.
    <input> The input value in degrees.
    <offset> The offset (in arcseconds) to apply for rounding purposes.
    [NOTE] All component values are guaranteed to be positive.
    [RETURNS] TRUE if the value was legal, else FALSE.
*/
{
    double d, m, s, trunc_seconds, sub_seconds;

    if (fabs (input) > ANGLE_LIMIT)
    {
	return (FALSE);
    }
    if (input < 0.0)
    {
	input = fabs (input);
	if (negative != NULL) *negative = TRUE;
    }
    else if (negative != NULL) *negative = FALSE;
    input += 1e-10;  /*  To avoid roundoff  */
    input += offset / 3600.0;
    d = floor (input);
    if (degrees != NULL) *degrees = d;
    input -= fabs (d);
    input *= 60.0;
    m = floor (input);
    if (minutes != NULL) *minutes = m;
    input -= m;
    input *= 60.0;
    input -= offset;
    s = input;
    if (s < 0) s = 0.0;
    if (seconds != NULL) *seconds = s;
    trunc_seconds = floor (s);
    if (whole_seconds != NULL) *whole_seconds = trunc_seconds;
    sub_seconds = s - trunc_seconds;
    if (subseconds != NULL) *subseconds = sub_seconds;
    return (TRUE);
}   /*  End Function wcs_astro_breakdown_dms  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_format_dms (char *string, double deg)
/*  [SUMMARY] Format an angle value into a string.
    <string> The string to write to. Sufficient storage must be available.
    <deg> The input value in degrees.
    [RETURNS] Nothing.
*/
{
    flag negative;
    int degrees, minutes;
    double seconds;

    if ( !wcs_astro_breakdown_dms (&negative, &degrees, &minutes, &seconds,
				   NULL, NULL, deg, 0.005) )
    {
	strcpy (string, "ERROR");
	return;
    }
    sprintf (string, negative ? "-%.2dd %.2dm %5.2fs" : "%.2dd %.2dm %5.2fs",
	     degrees, minutes, seconds);
}   /*  End Function wcs_astro_format_dms  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_format_vel (KwcsAstro ap, char *string, double vel)
/*  [SUMMARY] Format a velocity value into a string.
    <ap> The KwcsAstro object.
    <string> The string to write to. Sufficient storage must be available.
    <vel> The velocity value.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "wcs_astro_format_vel";

    VERIFY_AP (ap);
    switch (ap->vel.type)
    {
      case SPECTRAL_VELn_FreqOptical:
      case SPECTRAL_VELl_FreqRadio:
      case SPECTRAL_VELl_WaveOptical:
	sprintf (string, "Vel: %.2f km/s", vel * 1e-3);
	break;
      case SPECTRAL_FREQ:
	sprintf (string, "Freq: %.3f MHz", vel * 1e-6);
	break;
      case SPECTRAL_UNKNOWN:
	sprintf (string, "v/f: %e", vel);
	break;
    }
}   /*  End Function wcs_astro_format_vel  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_breakdown_time (unsigned int *year, unsigned int *month,
			       unsigned int *day_of_month,
			       unsigned int *day_of_year,
			       unsigned int *hours, unsigned int *minutes,
			       double *seconds, unsigned int *whole_seconds,
			       double *subseconds, double time_val)
/*  [SUMMARY] Breakdown a time value into components.
    <time_val> The time in Modified Julian Date (units of seconds).
    <year> The year is written here.
    <month> The month is written here.
    <day_of_month> The day of the month is written here.
    <hours> The hours are written here. This may be NULL.
    <minutes> The minutes are written here. This may be NULL.
    <seconds> The seconds are written here. This may be NULL.
    <whole_seconds> The integer truncated seconds are written here. This may be
    NULL.
    <subseconds> The fractional seconds are written here. This may be NULL.
    [RETURNS] Nothing.
*/
{
    double s, trunc_seconds, sub_seconds;
    time_t utc_time;
    struct tm *bt;

    /*  The time is in MJD (Modified Julian Date) in seconds, so first convert
	to seconds since midnight, 1-JAN-1970  */
    time_val -= (2440587.5 - 2400000.5) * 3600.0 * 24.0;
    time_val += 1e-10;  /*  To avoid roundoff  */
    utc_time = floor (time_val);
    s = fmod (time_val, 60.0);
    bt = gmtime (&utc_time);
    if (year != NULL) *year = bt->tm_year + 1900;
    if (month != NULL) *month = bt->tm_mon + 1;
    if (day_of_year != NULL) *day_of_year = bt->tm_yday;
    if (day_of_month != NULL) *day_of_month = bt->tm_mday;
    if (hours != NULL) *hours = bt->tm_hour;
    if (minutes != NULL) *minutes = bt->tm_min;
    if (seconds != NULL) *seconds = s;
    trunc_seconds = floor (s);
    if (whole_seconds != NULL) *whole_seconds = trunc_seconds;
    sub_seconds = s - trunc_seconds;
    if (subseconds != NULL) *subseconds = sub_seconds;
}   /*  End Function wcs_astro_breakdown_time  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_format_time (char *string, double time_val)
/*  [SUMMARY] Format a time value into a string.
    <string> The string to write to. Sufficient storage must be available.
    <time_val> The time in seconds.
    [RETURNS] Nothing.
*/
{
    double seconds;
    time_t utc_time;
    struct tm *bt;

    /*  The time is in MJD (Modified Julian Date) in seconds, so first convert
	to seconds since midnight, 1-JAN-1970  */
    time_val -= (2440587.5 - 2400000.5) * 3600.0 * 24.0;
    utc_time = time_val;
    seconds = fmod (time_val, 60.0);
    bt = gmtime (&utc_time);
    sprintf (string, "%d/%d/%d %d:%d:%5.2f",
	     bt->tm_year + 1900, bt->tm_mon + 1, bt->tm_mday,
	     bt->tm_hour, bt->tm_min, seconds);
}   /*  End Function wcs_astro_format_time  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_format (KwcsAstro ap, CONST char *dim_name,
		       char string[STRING_LENGTH], double value)
/*  [SUMMARY] Format a value. No transformation is applied.
    <ap> The KwcsAstro object.
    <dim_name> The name of the dimension.
    <string> The string to write to.
    <value> The value.
    [RETURNS] Nothing.
*/
{
    int str_len;
    char stokes_char;
    double unit_scale;
    char unit_string[STRING_LENGTH], unit_format[STRING_LENGTH];
    char format[STRING_LENGTH];
    char txt[STRING_LENGTH];
    static char function_name[] = "wcs_astro_format";

    VERIFY_AP (ap);
    str_len = strlen (dim_name);
    if (strcmp (dim_name, ap->sky.ra.dim_name) == 0)
    {
	if ( DOUBLE_IS_VALID (value) )
	{
	    while (value <= -360.0) value += 360.0;
	    while (value >= 360.0) value -= 360.0;
	}
	if (ap->sky.ra.is_sky)
	{
	    strcpy (string, "Ra ");
	    wcs_astro_format_hms (string + strlen (string), value / 15.0);
	}
	else
	{
#ifdef GALACTIC_DMS
	    wcs_astro_format_dms (txt, value);
	    sprintf (string, "Glon %s", txt);
#else
	    sprintf (string, "Glon %.8g", value);
#endif
	}
	return;
    }
    if (strcmp (dim_name, ap->sky.dec.dim_name) == 0)
    {
	if ( DOUBLE_IS_VALID (value) )
	{
	    while (value <= -360.0) value += 360.0;
	    while (value >= 360.0) value -= 360.0;
	}
	wcs_astro_format_dms (txt, value);
	if (ap->sky.dec.is_sky) sprintf (string, "Dec %s", txt);
	else
	{
#ifdef GALACTIC_DMS
	    sprintf (string, "Glat %s", txt);
#else
	    sprintf (string, "Glat %.8g", value);
#endif
	}
	return;
    }
    if (strcmp (dim_name, ap->vel.dim_name) == 0)
    {
	wcs_astro_format_vel (ap, string, value);
	return;
    }
    if (strcmp (dim_name, "STOKES") == 0)
    {
	switch ( (int) value )
	{
	  case 1:
	    stokes_char = 'I';
	    break;
	  case 2:
	    stokes_char = 'Q';
	    break;
	  case 3:
	    stokes_char = 'U';
	    break;
	  case 4:
	    stokes_char = 'V';
	    break;
	  default:
	    stokes_char = '?';
	    break;
	}
	sprintf (string, "Stokes %c", stokes_char);
	return;
    }
    if ( (strcmp (dim_name, "ANGLE") == 0) ||
	      (strcmp (dim_name + str_len - 5, "(deg)") == 0) )
    {
	/*  Offset angle  */
	wcs_astro_format_dms (txt, value);
	sprintf (string, "Offset %s", txt);
	return;
    }
    if (strcmp (dim_name, "TIME") == 0)
    {
	wcs_astro_format_time (txt, value);
	sprintf (string, "Time %s", txt);
	return;
    }
    /*  Don't know how to deal with this one  */
    ds_format_unit (unit_string, unit_format, &unit_scale, dim_name, value);
    sprintf (format, "%s: %s", unit_string, unit_format);
    sprintf (string, format, value * unit_scale);
}   /*  End Function wcs_astro_format  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_format_all (KwcsAstro ap, char coord_string[STRING_LENGTH],
			   CONST char *name0, double coord0,
			   CONST char *name1, double coord1,
			   CONST char *name2, double coord2,
			   unsigned int num_restr, CONST char **restr_names,
			   CONST double *restr_values,
			   char other_string[STRING_LENGTH])
/*  [SUMMARY] Transform and format all available information.
    [PURPOSE] This routine will transform and format all available information.
    The data is transformed from linear to non-linear values.
    <ap> The KwcsAstro object.
    <coord_string> Formatted co-ordinate information is written here.
    <name0> The name of co-ordinate 0.
    <coord0> Co-ordinate 0.
    <name1> The name of co-ordinate 1. If this is NULL, it is ignored.
    <coord1> Co-ordinate 1.
    <name2> The name of co-ordinate 2. If this is NULL, it is ignored.
    <coord2> Co-ordinate 2.
    <num_restr> The number of restrictions.
    <restr_names> The array of restriction names.
    <restr_values> The restriction values.
    <other_string> Formatted extra information is written here.
    [RETURNS] Nothing.
*/
{
    flag ra_to_linear;
    flag dec_to_linear;
    flag vel_to_linear;
    unsigned int count;
    double radec_value, value;
    double *ra;
    double *dec;
    double *vel;
    CONST char *unspec_name = NULL;
    struct extra_type *curr;
    char txt[STRING_LENGTH];
    static char function_name[] = "wcs_astro_format_all";

    VERIFY_AP (ap);
    *coord_string = '\0';
    *other_string = '\0';
    coords_to_radecvel (ap,
			name0, &coord0, FALSE,
			name1, &coord1, FALSE,
			name2, &coord2, FALSE,
			num_restr, restr_names, restr_values,
			&ra, &ra_to_linear,
			&dec, &dec_to_linear,
			&vel, &vel_to_linear);
    if ( (ra == NULL) && (dec != NULL) )
    {
	/*  Have to get unspecified DEC value  */
	if ( ( unspec_name =
	       find_unspecified (ap, ap->sky.ra.dim_name, &radec_value,
				 num_restr, restr_names, restr_values) )
	     == NULL )
	{
	    fprintf (stderr, "%s: missing DEC information\n", function_name);
	    return;
	}
	ra = &radec_value;
    }
    else if ( (ra != NULL) && (dec == NULL) )
    {
	/*  Have to get unspecified RA value  */
	if ( ( unspec_name =
	       find_unspecified (ap, ap->sky.dec.dim_name, &radec_value,
				 num_restr, restr_names, restr_values) )
	     == NULL )
	{
	    fprintf (stderr, "%s: missing DEC information\n", function_name);
	    return;
	}
	dec = &radec_value;
    }
    wcs_astro_transform (ap, 1, ra, FALSE, dec, FALSE, vel, FALSE,
			 num_restr, restr_names, restr_values);
    transform_linear (ap, 1, name0, &coord0, FALSE);
    wcs_astro_format (ap, name0, coord_string, coord0);
    if (name1 != NULL)
    {
	transform_linear (ap, 1, name1, &coord1, FALSE);
	wcs_astro_format (ap, name1, txt, coord1);
	strcat (coord_string, "  ");
	strcat (coord_string, txt);
    }
    if (name2 != NULL)
    {
	transform_linear (ap, 1, name2, &coord2, FALSE);
	wcs_astro_format (ap, name2, txt, coord2);
	strcat (coord_string, "  ");
	strcat (coord_string, txt);
    }
    /*  Add any restriction information  */
    for (count = 0; count < num_restr; ++count)
    {
	if (unspec_name == restr_names[count])
	{
	    value = radec_value;
	}
	else
	{
	    value = restr_values[count];
	    if (strcmp (restr_names[count], ap->vel.dim_name) == 0)
	    {
		transform_vel (ap, 1, &value, FALSE);
	    }
	    else transform_linear (ap, 1, restr_names[count], &value, FALSE);
	}
	wcs_astro_format (ap, restr_names[count], txt, value);
	strcat (coord_string, "  ");
	strcat (coord_string, txt);
    }
    for (curr = ap->first_extra; curr != NULL; curr = curr->next)
    {
	if (unspec_name == curr->ctype) value = radec_value;
	else
	{
	    value = 0.0;
	    if (strcmp (curr->ctype, ap->vel.dim_name) == 0)
	    {
		transform_vel (ap, 1, &value, FALSE);
	    }
	    else transform_linear (ap, 1, curr->ctype, &value, FALSE);
	}
	wcs_astro_format (ap, curr->ctype, txt, value);
	strcat (other_string, txt);
	if (curr->next != NULL) strcat (other_string, "  ");
    }
}   /*  End Function wcs_astro_format_all  */

/*EXPERIMENTAL_FUNCTION*/
flag wcs_astro_get_preferred_units (char new_units[STRING_LENGTH],
				    char format_str[STRING_LENGTH],
				    double *scale, CONST char *old_units)
/*  [SUMMARY] Get preferred units for data.
    <new_units> The new units string is written here.
    <format_str> The format string is written here.
    <scale> The scale value to convert to the new units will be written here.
    <old_units> The old units string.
    [RETURNS] TRUE if preferred units were found, else FALSE.
*/
{
    if ( (strncmp (old_units, "M/S", 3) == 0) ||
	 (strncmp (old_units, "METR/SEC", 8) == 0) )
    {
	strcpy (new_units, "km/s");
	strcpy (format_str, "%.2f");
	*scale = 1e-3;
    }
    else if (strncmp (old_units, "KM/S", 4) == 0)
    {
	strcpy (new_units, "km/s");
	strcpy (format_str, "%.1f");
	*scale = 1.0;
    }
    else if (strncmp (old_units, "JY/BEAM", 7) == 0)
    {
	strcpy (new_units, "mJy/Beam");
	strcpy (format_str, "%.1f");
	*scale = 1e+3;
    }
    else if (strncmp (old_units, "FREQ", 4) == 0)
    {
	strcpy (new_units, "MHz");
	strcpy (format_str, "%.3f");
	*scale = 1e-6;
    }
    else if (strcmp (old_units, "HZ") == 0)
    {
	strcpy (new_units, "MHz");
	strcpy (format_str, "%.3f");
	*scale = 1e-6;
    }
    else if (strncmp (old_units, "FELO", 4) == 0)
    {
	strcpy (new_units, "km/s");
	strcpy (format_str, "%.2f");
	*scale = 1e-3;
    }
    else if (strncmp (old_units, "VELO", 4) == 0)
    {
	strcpy (new_units, "km/s");
	strcpy (format_str, "%.2f");
	*scale = 1e-3;
    }
    else return (FALSE);
    return (TRUE);
}   /*  End Function wcs_astro_get_preferred_units  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int wcs_astro_get_equinox (KwcsAstro ap)
/*  [SUMMARY] Get the Equinox.
    <ap> The KwcsAstro object.
    [RETURNS] The year of the Equinox (typically 1950 or 2000).
*/
{
    static char function_name[] = "wcs_astro_get_equinox";

    VERIFY_AP (ap);
    switch (ap->sky.equinox)
    {
      case EQUINOX_J2000:
	return (2000);
	/*break;*/
      case EQUINOX_B1950:
	return (1950);
	/*break;*/
      default:
	break;
    }
    return (0);
}   /*  End Function wcs_astro_get_equinox  */

/*EXPERIMENTAL_FUNCTION*/
double wcs_astro_get_reference (KwcsAstro ap, CONST char *name, flag linear)
/*  [SUMMARY] Get reference value for an axis.
    <ap> The KwcsAstro object.
    <name> The axis name.
    <linear> If TRUE, the linear reference position is given, else the
    non-linear reference value is given.
    [RETURNS] The reference value.
*/
{
    struct linear_coord_type *lin;
    struct extra_type *extra;
    static char function_name[] = "wcs_astro_get_reference";

    VERIFY_AP (ap);
    if (strcmp (name, ap->sky.ra.dim_name) == 0)
    {
	return linear ? ap->sky.ra.ref_pos : ap->sky.ra.reference;
    }
    if (strcmp (name, ap->sky.dec.dim_name) == 0)
    {
	return linear ? ap->sky.dec.ref_pos : ap->sky.dec.reference;
    }
    if (strcmp (name, ap->vel.dim_name) == 0)
    {
	return linear ? ap->vel.ref_pos : ap->vel.reference;
    }
    for (lin = ap->linear_axes; lin != NULL; lin = lin->next)
    {
	if (strcmp (name, lin->dim_name) == 0)
	{
	    return linear ? lin->ref_pos : lin->reference;
	}
    }
    for (extra = ap->first_extra; extra != NULL; extra = extra->next)
    {
	if (strcmp (name, extra->ctype) == 0)
	{
	    return linear ? extra->refpos : extra->refval;
	}
    }
    return (TOOBIG);
}   /*  End Function wcs_astro_get_reference  */

/*EXPERIMENTAL_FUNCTION*/
void wcs_astro_update_location (KwcsAstro ap, unsigned int axis_number,
				CONST packet_desc *pack_desc,
				CONST char *packet)
/*  [SUMMARY] Update the location ("CRVALn" and "CRPIXn") values.
    <ap> The KwcsAstro object to update.
    <axis_number> The FITS axis number to update.
    <pack_desc> The packet descriptor to extract the new values from.
    <packet> The packet to extract the new values from.
    [RETURNS] Nothing.
*/
{
    char txt[STRING_LENGTH];
    double crval, crpix, tmp[2];
    struct sky_coord_type *sky = NULL;
    struct linear_coord_type *linear;
    struct extra_type *extra;
    static char function_name[] = "wcs_astro_update_location";

    VERIFY_AP (ap);
    if (axis_number < 1) return;
    sprintf (txt, "CRVAL%u", axis_number);
    if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL, tmp) )
	crval = tmp[0];
    else crval = TOOBIG;
    sprintf (txt, "CRPIX%u", axis_number);
    if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL, tmp) )
	crpix = tmp[0];
    else crpix = TOOBIG;
    if (ap->sky.ra.axis_number == axis_number) sky = &ap->sky.ra;
    else if (ap->sky.dec.axis_number == axis_number) sky = &ap->sky.dec;
    if (sky != NULL)
    {
	if (crval < TOOBIG)
	{
	    sky->reference = crval;
	    sky->sin_ref = sin (crval * PION180);
	    sky->cos_ref = cos (crval * PION180);
	}
	if (crpix < TOOBIG)
	{
	    sky->ref_pos = crpix - 1.0;
	}
	return;
    }
    if (ap->vel.axis_number == axis_number)
    {
	if (crval < TOOBIG) ap->vel.reference = crval;
	if (crpix < TOOBIG) ap->vel.ref_pos = crpix - 1.0;
	return;
    }
    for (linear = ap->linear_axes; linear != NULL; linear = linear->next)
    {
	if (linear->axis_number != axis_number) continue;
	if (crval < TOOBIG) linear->reference = crval;
	if (crpix < TOOBIG) linear->ref_pos = crpix - 1.0;
	return;
    }
    for (extra = ap->first_extra; extra != NULL; extra = extra->next)
    {
	if (extra->axis_number != axis_number) continue;
	if (crval < TOOBIG) extra->refval = crval;
	if (crpix < TOOBIG) extra->refpos = crpix - 1.0;
	return;
    }
}   /*  End Function wcs_astro_update_location  */


/*  Private functions follow  */

static KwcsAstro create_new ()
/*  [SUMMARY] Create new KwcsAstro object.
    [RETURNS] A KwcsAstro object on success, else NULL.
*/
{
    KwcsAstro new;
    static char function_name[] = "_wcs_astro_create_new";

    if ( ( new = m_calloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "KwcsAstro object");
	return (NULL);
    }
    /*  Initialise  */
    new->sky.projection = PROJ_INIT;
    new->sky.dss = NULL;
    new->sky.LMtoXY.defined = FALSE;
    new->sky.LMtoXY.cd_11 = TOOBIG;
    new->sky.LMtoXY.cd_12 = TOOBIG;
    new->sky.LMtoXY.cd_21 = TOOBIG;
    new->sky.LMtoXY.cd_22 = TOOBIG;
    new->sky.XMtoLY.defined = FALSE;
    new->sky.XMtoLY.cd_11 = TOOBIG;
    new->sky.XMtoLY.cd_12 = TOOBIG;
    new->sky.XMtoLY.cd_21 = TOOBIG;
    new->sky.XMtoLY.cd_22 = TOOBIG;
    new->sky.LYtoXM.defined = FALSE;
    new->sky.LYtoXM.cd_11 = TOOBIG;
    new->sky.LYtoXM.cd_12 = TOOBIG;
    new->sky.LYtoXM.cd_21 = TOOBIG;
    new->sky.LYtoXM.cd_22 = TOOBIG;
    new->sky.XYtoLM.defined = FALSE;
    new->sky.XYtoLM.cd_11 = TOOBIG;
    new->sky.XYtoLM.cd_12 = TOOBIG;
    new->sky.XYtoLM.cd_21 = TOOBIG;
    new->sky.XYtoLM.cd_22 = TOOBIG;
    new->sky.sin_rotation = TOOBIG;
    new->sky.cos_rotation = TOOBIG;
    new->sky.equinox = EQUINOX_J2000;
    new->sky.transform_func = NULL;
    new->vel.type = SPECTRAL_UNKNOWN;
    new->linear_axes = NULL;
    new->first_extra = NULL;
    new->sky.radec_buffer = NULL;
    new->sky.radec_buf_len = 0;
    new->sky.ra.axis_number = 0;
    new->sky.ra.is_sky = TRUE;
    new->sky.dec.axis_number = 0;
    new->sky.ra.reference = TOOBIG;
    new->sky.dec.reference = TOOBIG;
    new->sky.dec.is_sky = TRUE;
    new->vel.axis_number = 0;
    new->vel.radec_scale_buffer = NULL;
    new->vel.radec_scale_buf_len = 0;
    return (new);
}   /*  End Function create_new  */

static flag process_axis (KwcsAstro ap,
			  CONST packet_desc *pack_desc, CONST char *packet,
			  CONST char *axis_name, unsigned int index)
/*  [SUMMARY] Process a single axis.
    <ap> The KwcsAstro object.
    <axis_name> The name of the axis.
    <index> The FITS index of the axis.
    <pack_desc> The packet descriptor containing the FITS-style keywords.
    <packet> The packet data containing the FITS-stype keyword data.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int proj;
    unsigned int item_index;
    double deg_to_rad = PION180;
    struct sky_coord_type *sky = NULL;
    struct spectral_coord_type *vel = NULL;
    struct linear_coord_type *linear = NULL;
    char *encls_desc;
    char txt[STRING_LENGTH];
    double crval, crpix, cdelt;
    double value[2];
    static char function_name[] = "_wcs_astro_process_axis";

    sprintf (txt, "CRVAL%u", index);
    if ( !ds_get_unique_named_value (pack_desc, packet, txt, NULL, value) )
    {
	fprintf(stderr,
		"Keyword \"CTYPE%u\" found but %s not found:defaulting to 1\n",
		index, txt);
	crval = 1.0;
    }
    else crval = value[0];
    sprintf (txt, "CRPIX%u", index);
    if ( !ds_get_unique_named_value (pack_desc, packet, txt, NULL, value) )
    {
	fprintf(stderr,
		"Keyword \"CTYPE%u\" found but %s not found:defaulting to 1\n",
		index, txt);
	crpix = 1.0;
    }
    else crpix = value[0];
    sprintf (txt, "CDELT%u", index);
    if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL, value) )
    {
	cdelt = value[0];
    }
    else cdelt = TOOBIG;
    if (strncmp (axis_name, "RA", 2) == 0)
    {
	/*  Right ascension  */
	sky = &ap->sky.ra;
	sky->is_sky = TRUE;
    }
    else if (strncmp (axis_name, "GLON", 4) == 0)
    {
	/*  Right ascension  */
	sky = &ap->sky.ra;
	sky->is_sky = FALSE;
    }
    else if (strncmp (axis_name, "DEC", 3) == 0)
    {
	/*  Declination  */
	sky = &ap->sky.dec;
	sky->is_sky = TRUE;
    }
    else if (strncmp (axis_name, "GLAT", 4) == 0)
    {
	/*  Declination  */
	sky = &ap->sky.dec;
	sky->is_sky = FALSE;
    }
    else if (strncmp (axis_name, "FREQ", 4) == 0)
    {
	/*  Frequency  */
	vel = &ap->vel;
	ap->vel.type = SPECTRAL_FREQ;
    }
    else if (strncmp (axis_name, "VELO", 4) == 0)
    {
	/*  Velocity  */
	vel = &ap->vel;
	if ( !ds_get_unique_named_value (pack_desc, packet, "VELREF", NULL,
					 value) ) value[0] = 256.0;
	if (value[0] < 256.0) ap->vel.type = SPECTRAL_VELl_WaveOptical;
	else ap->vel.type = SPECTRAL_VELl_FreqRadio;
    }
    else if (strncmp (axis_name, "FELO", 4) == 0)
    {
	/*  Velocity  */
	vel = &ap->vel;
	ap->vel.type = SPECTRAL_VELn_FreqOptical;
    }
    /*  Always record axis in linear co-ordinate types  */
    if ( ( linear = m_alloc (sizeof *linear) ) == NULL )
    {
	m_abort (function_name, "linear axis");
    }
    linear->reference = crval;
    linear->ref_pos = crpix - 1.0;
    linear->increment = DOUBLE_IS_VALID (cdelt) ? cdelt : 1.0;
    strcpy (linear->dim_name, axis_name);
    linear->axis_number = index;
    linear->next = ap->linear_axes;
    ap->linear_axes = linear;
    if (sky != NULL)
    {
	char *ptr;

	/*  Right ascension or declination  */
	sky->reference = crval;
	sky->ref_pos = crpix - 1.0;
	if (cdelt == 0.0)
	{
	    /*  If degenerate axis, fake an increment so that projections
		still work  */
	    sprintf (txt, "NAXIS%u", index);
	    if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL,
					    value) )
	    {
		if (value[0] <= 1.0) cdelt = 1.0;
	    }
	}
	sky->increment = cdelt;
	sky->pix_to_rad = sky->increment * deg_to_rad;
	sky->rad_to_pix = 180.0 / PI / sky->increment;
	sky->cos_ref = cos (sky->reference * deg_to_rad);
	sky->sin_ref = sin (sky->reference * deg_to_rad);
	sky->axis_number = index;
	/*  Determine the projection system  */
	ptr = strchr (axis_name, '-');
	if (ptr != NULL) while (*ptr == '-') ++ptr;
	if (ptr == NULL) proj = PROJ_FLT;
	else if (strcmp (ptr, "SIN") == 0) proj = PROJ_SIN;
	else if (strcmp (ptr, "TAN") == 0) proj = PROJ_TAN;
	else if (strcmp (ptr, "NCP") == 0) proj = PROJ_NCP;
	else if (strcmp (ptr, "FLT") == 0) proj = PROJ_FLT;
	else if (strcmp (ptr, "ARC") == 0) proj = PROJ_ARC;
	else if (strcmp (ptr, "CAR") == 0) proj = PROJ_FLT;
	else if (strcmp (ptr, "ZEA") == 0) proj = PROJ_ZEA;
	else if (strcmp (ptr, "GLS") == 0) proj = PROJ_GLS;
	else
	{
	    proj = PROJ_UNKNOWN;
	    fprintf (stderr,
		     "%s: unknown projection: \"%s\" for axis: \"%s\"\n",
		     function_name, ptr, axis_name);
	}
	if (ap->sky.projection == PROJ_INIT)
	{
	    ap->sky.projection = proj;
	}
	else if (ap->sky.projection != proj)
	{
	    fputs ("Projections do not match\n", stderr);
	    return (FALSE);
	}
	sprintf (txt, "CROTA%u", index);
	if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL, value) &&
	     (value[0] != 0.0) )
	{
	    if ( DOUBLE_IS_VALID (ap->sky.sin_rotation) &&
		 fabs (sin (deg_to_rad * value[0]) -
		       ap->sky.sin_rotation) > 1e-8 )
	    {
		fprintf (stderr,
			 "Rotation multiply and inconsistently defined\n");
		return (FALSE);
	    }
	    ap->sky.sin_rotation = sin (deg_to_rad * value[0]);
	    ap->sky.cos_rotation = cos (deg_to_rad * value[0]);
	}
	strcpy (sky->dim_name, axis_name);
    }
    if (vel != NULL)
    {
	char *cellscal;

	vel->reference = crval;
	vel->ref_pos = crpix - 1.0;
	vel->increment = DOUBLE_IS_VALID (cdelt) ? cdelt : 1.0;
	vel->axis_number = index;
	strcpy (ap->vel.dim_name, axis_name);
	if ( ds_get_unique_named_value (pack_desc, packet, "RESTFREQ", NULL,
					value) )
	{
	    vel->restfreq = value[0];
	}
	else vel->restfreq = TOOBIG;
	vel->chromatic_aberration = FALSE;
	if ( ( cellscal = ds_get_unique_named_string (pack_desc, packet,
						      "CELLSCAL") ) != NULL )
	{
	    if (st_icmp (cellscal, "1/f") == 0)
		vel->chromatic_aberration = TRUE;
	    m_free (cellscal);
	}
	if ( ds_get_unique_named_value (pack_desc, packet, "VOBS", NULL,
					value) )
	{
	    vel->vobs = value[0];
	}
	else vel->vobs = 0.0;
    }
    /*  Check if this axis is a dimension  */
    /*  Search for specified name first  */
    switch ( ds_f_name_in_packet (pack_desc, axis_name, &encls_desc,
				  &item_index) )
    {
      case IDENT_NOT_FOUND:
	add_extra (ap, axis_name, crval, crpix, index);
	break;
      default:
	break;
    }
    return (TRUE);
}   /*  End Function process_axis  */

static void add_extra (KwcsAstro ap, CONST char *ctype, double refval,
		       double refpos, unsigned int axis_number)
/*  [SUMMARY] Add extra information.
    <ap> The KwcsAstro object.
    <ctype> The value of the "CTYPEn" keyword.
    <refval> The value of the "CRVALn" keyword.
    <refpos> The value of the "CRPIXn" keyword.
    <axis_number> The <<n>> value
    [RETURNS] Nothing.
*/
{
    struct extra_type *new, *curr;
    static char function_name[] = "_wcs_astro_add_extra";

    if ( ( new = m_alloc (sizeof *new) ) == NULL )
    {
	m_abort (function_name, "extra entry");
    }
    strcpy (new->ctype, ctype);
    new->refval = refval;
    new->refpos = refpos - 1.0;
    new->axis_number = axis_number;
    new->next = NULL;
    /*  Append to list  */
    if (ap->first_extra == NULL)
    {
	ap->first_extra = new;
	return;
    }
    for (curr = ap->first_extra; curr->next != NULL; curr = curr->next);
    curr->next = new;
}   /*  End Function add_extra  */

static void get_axis_names (CONST packet_desc *pack_desc, CONST char *packet,
			    char axis_names[MAX_AXES][STRING_LENGTH])
/*  [SUMMARY] Get CTYPEn strings and fix duplicates.
    <pack_desc> The packet descriptor.
    <packet> The packet data.
    [RETURNS] Nothing.
*/
{
    flag keep_going = TRUE;
    flag duplicate;
    unsigned int axis = 0;
    unsigned int count;
    char *ctype;
    char txt[STRING_LENGTH], name[STRING_LENGTH];

    while (keep_going)
    {
	sprintf (txt, "CTYPE%u", axis + 1);
	if ( (ctype = ds_get_unique_named_string (pack_desc, packet,
						  txt) ) == NULL )
	{
	    axis_names[axis][0] = '\0';
	    keep_going = FALSE;
	}
	else
	{
	    strcpy (axis_names[axis], ctype);
	    m_free (ctype);
	}
	++axis;
    }
    for (axis = 0; axis_names[axis][0] != '\0'; ++axis)
    {
	/*  A first pass to check if duplicates exist  */
	duplicate = FALSE;
	for (count = 0; !duplicate && (axis_names[count][0] != '\0'); ++count)
	{
	    if (count == axis) continue;
	    if (strcmp (axis_names[count], axis_names[axis]) == 0)
		duplicate = TRUE;
	}
	if (!duplicate) continue;
	strcpy (name, axis_names[axis]);
	/*  Now modify all duplicates  */
	for (count = 0; axis_names[count][0] != '\0'; ++count)
	{
	    if (strcmp (axis_names[count], name) != 0) continue;
	    sprintf (txt, "%s:%u", name, count + 1);
	    strcpy (axis_names[count], txt);
	}
    }
}   /*  End Function get_axis_names  */

static flag process_sky (KwcsAstro ap, CONST packet_desc *pack_desc,
			 CONST char *packet)
/*  [SUMMARY] Process sky information.
    <ap> The KwcsAstro object.
    <pack_desc> The packet descriptor containing the FITS-style keywords.
    <packet> The packet data containing the FITS-stype keyword data.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    double bot;
    struct cd_matrix_type *mat = &ap->sky.XYtoLM;
    char txt[STRING_LENGTH];
    double value[2];
    static char function_name[] = "_wcs_astro_process_sky";

    if (ap->sky.projection == PROJ_INIT)
    {
	/*  No conventional RA-DEC projection found yet: try DSS plate soln. */
	if ( !scan_dss_header (ap, pack_desc, packet) )
	{
	    return (FALSE);
	}
    }
    if (ap->sky.projection == PROJ_INIT)
    {
	/*  Still no sky projection: none available  */
	ap->sky.projection = PROJ_UNKNOWN;
	return (TRUE);
    }
    if (!ap->sky.ra.is_sky || !ap->sky.dec.is_sky)
	ap->sky.equinox = EQUINOX_NONE;
    if (ap->sky.projection != PROJ_DSS)
    {
	if ( (ap->sky.ra.axis_number < 1) || (ap->sky.dec.axis_number < 1) )
	{
	    /*  Only one axis found  */
	    ap->sky.projection = PROJ_UNKNOWN;
	    return (TRUE);
	}
    }
    switch (ap->sky.projection)
    {
      case PROJ_DSS:
	ap->sky.transform_func = transform_ra_dec_dss;
	return (TRUE);
	/*break;*/
      case PROJ_SIN:
	ap->sky.transform_func = transform_ra_dec_sin;
	break;
      case PROJ_TAN:
	ap->sky.transform_func = transform_ra_dec_tan;
	break;
      case PROJ_NCP:
	ap->sky.transform_func = transform_ra_dec_ncp;
	break;
      case PROJ_FLT:
	ap->sky.transform_func = transform_ra_dec_flat;
	break;
      case PROJ_ARC:
	ap->sky.transform_func = transform_ra_dec_arc;
	break;
      case PROJ_ZEA:
	ap->sky.transform_func = transform_ra_dec_zea;
	break;
      case PROJ_GLS:
	ap->sky.transform_func = transform_ra_dec_gls;
	break;
      case PROJ_UNKNOWN:
	fprintf (stderr, "%s: unknown projection\n", function_name);
	break;
      default:
	fprintf (stderr, "Unknown projection: %u\n", ap->sky.projection);
	a_prog_bug (function_name);
	break;
    }
    if (ap->sky.ra.is_sky != ap->sky.dec.is_sky)
    {
	fprintf (stderr, "%s: mixed RA/DEC and GLON/GLAT co-ordinates\n",
		 function_name);
	return (FALSE);
    }
    /*  Check for matrix-style co-ordinate description  */
    sprintf (txt, "CD%.3u%.3u",
	     ap->sky.ra.axis_number, ap->sky.ra.axis_number);
    if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL,
				    value) ) mat->cd_11 = value[0];
    else if ( ds_get_unique_named_value (pack_desc, packet, "CD1_1", NULL,
					 value) ) mat->cd_11 = value[0];
    sprintf (txt, "CD%.3u%.3u",
	     ap->sky.ra.axis_number, ap->sky.dec.axis_number);
    if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL,
				    value) ) mat->cd_12 = value[0];
    else if ( ds_get_unique_named_value (pack_desc, packet, "CD1_2", NULL,
					 value) ) mat->cd_12 = value[0];
    sprintf (txt, "CD%.3u%.3u",
	     ap->sky.dec.axis_number, ap->sky.ra.axis_number);
    if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL,
				    value) ) mat->cd_21 = value[0];
    else if ( ds_get_unique_named_value (pack_desc, packet, "CD2_1", NULL,
					 value) ) mat->cd_21 = value[0];
    sprintf (txt, "CD%.3u%.3u",
	     ap->sky.dec.axis_number, ap->sky.dec.axis_number);
    if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL,
				    value) ) mat->cd_22 = value[0];
    else if ( ds_get_unique_named_value (pack_desc, packet, "CD2_2", NULL,
					 value) ) mat->cd_22 = value[0];
    if ( DOUBLE_IS_BLANK (mat->cd_11) || DOUBLE_IS_BLANK (mat->cd_12) ||
	 DOUBLE_IS_BLANK (mat->cd_21) || DOUBLE_IS_BLANK (mat->cd_22) )
    {
	/* CD matrix incomplete: check for old-style co-ordinate description */
	if ( DOUBLE_IS_BLANK (ap->sky.ra.increment) ||
	     DOUBLE_IS_BLANK (ap->sky.dec.increment) )
	{
	    fprintf (stderr, "%s: co-ordinate increments missing\n",
		     function_name);
	    return (FALSE);
	}
	if ( DOUBLE_IS_BLANK (ap->sky.sin_rotation) )
	{
	    ap->sky.sin_rotation = 0.0;
	    ap->sky.cos_rotation = 1.0;
	}
	/*  Construct CD matrix from "CDELTn" and "CROTAn"  */
	/*  This is what I think it should be  */
	mat->cd_11 = ap->sky.ra.increment * ap->sky.cos_rotation;
	mat->cd_12 = -ap->sky.dec.increment * ap->sky.sin_rotation;
	mat->cd_21 = ap->sky.ra.increment * ap->sky.sin_rotation;
	mat->cd_22 = ap->sky.dec.increment * ap->sky.cos_rotation;
	mat->defined = TRUE;
    }
    /*  At least partially defined  */
    if ( DOUBLE_IS_BLANK (mat->cd_11) || DOUBLE_IS_BLANK (mat->cd_12) ||
	 DOUBLE_IS_BLANK (mat->cd_21) || DOUBLE_IS_BLANK (mat->cd_22) )
    {
	fprintf (stderr, "%s: incomplete co-ordinate description matrix\n",
		 function_name);
	return (FALSE);
    }
    mat->defined = TRUE;
    /*  Convert to radians  */
    mat->cd_11 *= PION180;
    mat->cd_12 *= PION180;
    mat->cd_21 *= PION180;
    mat->cd_22 *= PION180;
    /*  Compute LMtoXY conversion  */
    bot = mat->cd_11 * mat->cd_22 - mat->cd_21 * mat->cd_12;
    if (bot == 0.0)
    {
	fprintf (stderr, "%s: RA,DEC -> x,y transformation not possible\n",
		 function_name);
	return (FALSE);
    }
    ap->sky.LMtoXY.cd_11 = mat->cd_22 / bot;
    ap->sky.LMtoXY.cd_12 = -mat->cd_12 / bot;
    ap->sky.LMtoXY.cd_21 = -mat->cd_21 / bot;
    ap->sky.LMtoXY.cd_22 = mat->cd_11 / bot;
    ap->sky.LMtoXY.defined = TRUE;
    /*  Compute LYtoXM conversion  */
    if (mat->cd_11 != 0.0)
    {
	ap->sky.LYtoXM.cd_11 = 1.0 / mat->cd_11;
	ap->sky.LYtoXM.cd_12 = -mat->cd_12 / mat->cd_11;
	ap->sky.LYtoXM.cd_21 = mat->cd_21 / mat->cd_11;
	ap->sky.LYtoXM.cd_22 = mat->cd_22 - mat->cd_12 *mat->cd_21 /mat->cd_11;
	ap->sky.LYtoXM.defined = TRUE;
    }
    /*  Compute XMtoLY conversion  */
    if (mat->cd_22 != 0.0)
    {
	ap->sky.XMtoLY.cd_11 = mat->cd_11 -mat->cd_21 * mat->cd_12 /mat->cd_22;
	ap->sky.XMtoLY.cd_12 = mat->cd_12 / mat->cd_22;
	ap->sky.XMtoLY.cd_21 = -mat->cd_21 / mat->cd_22;
	ap->sky.XMtoLY.cd_22 = 1.0 / mat->cd_22;
	ap->sky.XMtoLY.defined = TRUE;
    }
    return (TRUE);
}   /*  End Function process_sky  */

static flag fill_radec_buffer (KwcsAstro ap, CONST char *dim_name,
			       unsigned int num_coords,
			       unsigned int num_restr,CONST char **restr_names,
			       CONST double *restr_values)
/*  [SUMMARY] Fill the RA/DEC buffer with RA or DEC values.
    <ap> The KwcsAstro object.
    <dim_name> The name of the dimension to fill for.
    <num_coords> The number of values to place into the buffer.
    <num_restr> The number of restrictions.
    <restr_names> The array of restriction names.
    <restr_values> The restriction values.
    [RETURNS] TRUE if data was available to fill, else FALSE.
*/
{
    unsigned int count;
    double value;
    static char function_name[] = "_wcs_astro_fill_radec_buffer";

    /*  First make sure buffer is large enough  */
    if (num_coords > ap->sky.radec_buf_len)
    {
	if (ap->sky.radec_buffer != NULL) m_free (ap->sky.radec_buffer);
	if ( ( ap->sky.radec_buffer =
	       m_alloc (num_coords * sizeof *ap->sky.radec_buffer) ) == NULL )
	{
	    m_abort (function_name, "RA/DEC buffer");
	}
	ap->sky.radec_buf_len = num_coords;
    }
    /*  Search for it  */
    if (find_unspecified (ap, dim_name, &value, num_restr, restr_names,
			  restr_values) == NULL) return (FALSE);
    for (count = 0; count < num_coords; ++count)
	ap->sky.radec_buffer[count] = value;
    return (TRUE);
}   /*  End Function fill_radec_buffer  */

static void get_radec_scale (KwcsAstro ap, unsigned int num_coords,
			     double *scales, CONST double *velocities)
/*  [SUMMARY] Get the RA/DEC scaling array by converting velocities.
    <ap> The KwcsAstro object.
    <num_coords> The number of values to place into the buffer.
    <scale> The array to write the scaling values to.
    <velocities> The array of velocity or frequency values.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double vobs = ap->vel.vobs;
    double refval = ap->vel.reference;
    double c = LIGHTSPEED;
    static char function_name[] = "_wcs_astro_get_radec_scale";

    switch (ap->vel.type)
    {
      case SPECTRAL_VELn_FreqOptical:
      case SPECTRAL_VELl_WaveOptical:
	/*  "Optical" definition: v = c(f0/f - 1)
	    where v = radial velocity, f0 = rest frequency,
	    f = observed frequency, c = speed of light  */
	refval += vobs;
	for (count = 0; count < num_coords; ++count)
	{
	    scales[count] = (c + velocities[count] + vobs) / (c + refval);
	}
	break;
      case SPECTRAL_VELl_FreqRadio:
	/*  "Radio" definition: v = c(1 - f/f0)
	    where v = radial velocity, f0 = rest frequency,
	    f = observed frequency, c = speed of light  */
	refval += vobs;
	for (count = 0; count < num_coords; ++count)
	{
	    scales[count] = (c - refval) / (c - velocities[count] - vobs);
	}
	break;
      case SPECTRAL_FREQ:
	for (count = 0; count < num_coords; ++count)
	{
	    scales[count] = refval / velocities[count];
	}
	break;
      default:
	fprintf (stderr, "Unknown velocity type\n");
	a_prog_bug (function_name);
	break;
    }
}   /*  End Function get_radec_scale  */

static CONST char *find_unspecified (KwcsAstro ap, CONST char *dim_name,
				     double *value, unsigned int num_restr,
				     CONST char **restr_names,
				     CONST double *restr_values)
/*  [SUMMARY] Find an unspecified value.
    <ap> The KwcsAstro object.
    <dim_name> The name of the dimension to fill for.
    <value> The unspecified value is written here. This is in linear
    co-ordinates.
    <num_restr> The number of restrictions.
    <restr_names> The array of restriction names.
    <restr_values> The restriction values.
    [RETURNS] A pointer to the found name on success, else NULL.
*/
{
    unsigned int count;
    struct extra_type *curr;
    /*static char function_name[] = "_wcs_astro_find_unspecified";*/

    /*  Search for it in extra information  */
    for (curr = ap->first_extra; curr != NULL; curr = curr->next)
    {
	if (strcmp (curr->ctype, dim_name) == 0)
	{
	    *value = 0.0;
	    return (curr->ctype);
	}
    }
    /*  Search for it in restrictions  */
    for (count = 0; count < num_restr; ++count)
    {
	if (strcmp (restr_names[count], dim_name) == 0)
	{
	    *value = restr_values[count];
	    return (restr_names[count]);
	}
    }
    return (NULL);
}   /*  End Function find_unspecified  */

static void coords_to_radecvel (KwcsAstro ap,
				CONST char *name0,double *coords0,flag to_lin0,
				CONST char *name1,double *coords1,flag to_lin1,
				CONST char *name2,double *coords2,flag to_lin2,
				unsigned int num_restr,
				CONST char **restr_names,
				CONST double *restr_values,
				double **ra, flag *ra_to_linear,
				double **dec, flag *dec_to_linear,
				double **vel, flag *vel_to_linear)
/*  [SUMMARY] Convert named co-ordinates to RA, DEC and VEL.
    <ap> The KwcsAstro object.
    <name0> The name of co-ordinate array 0.
    <coords0> Co-ordinate array 0.
    <to_lin0> If TRUE the co-ordinate 0 values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <name1> The name of co-ordinate array 1.
    <coords1> Co-ordinate array 1.
    <to_lin1> If TRUE the co-ordinate 1 values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <name2> The name of co-ordinate array 2.
    <coords2> Co-ordinate array 2.
    <to_lin2> If TRUE the co-ordinate 2 values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    <num_restr> The number of restrictions.
    <restr_names> The array of restriction names.
    <restr_values> The restriction values.
    <ra> The pointer to the RA co-ordinate array is written here.
    <ra_to_linear> The to_linear flag is written here.
    <dec> The pointer to the DEC co-ordinate array is written here.
    <dec_to_linear> The to_linear flag is written here.
    <vel> The pointer to the VEL co-ordinate array is written here.
    <vel_to_linear> The to_linear flag is written here.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    flag to_lin[3];
    CONST char *names[3];
    double *coords[3];
    static char function_name[] = "_wcs_astro_coords_to_radecvel";

    VERIFY_AP (ap);
    names[0] = name0;
    names[1] = name1;
    names[2] = name2;
    to_lin[0] = to_lin0;
    to_lin[1] = to_lin1;
    to_lin[2] = to_lin2;
    coords[0] = coords0;
    coords[1] = coords1;
    coords[2] = coords2;
    *ra = NULL;
    *dec = NULL;
    *vel = NULL;
    for (count = 0; count < 3; ++count)
    {
	if (names[count] == NULL) continue;
	if ( (ap->sky.projection != PROJ_UNKNOWN) &&
	     (strcmp (names[count], ap->sky.ra.dim_name) == 0) )
	{
	    *ra = coords[count];
	    *ra_to_linear = to_lin[count];
	}
	else if ( (ap->sky.projection != PROJ_UNKNOWN) &&
		  (strcmp (names[count], ap->sky.dec.dim_name) == 0) )
	{
	    *dec = coords[count];
	    *dec_to_linear = to_lin[count];
	}
	else if ( (ap->vel.type != SPECTRAL_UNKNOWN) &&
		  (strcmp (names[count], ap->vel.dim_name) == 0) )
	{
	    *vel = coords[count];
	    *vel_to_linear = to_lin[count];
	}
    }
}   /*  End Function coords_to_radecvel  */

static void transform_ra_dec_sin (KwcsAstro ap, unsigned int num_coords,
				  double *ra, double *dec,
				  unsigned int direction)
/*  [SUMMARY] Orthographic projection of co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    <direction> The direction of the transform.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double deg_to_rad = PION180;
    double rad_to_deg = 180.0 / PI;
    double l, m, x, y, tmp, in_ra, in_dec;
    double toobig = TOOBIG;
    double delta, sin_delta, cos_delta;
    double alpha, sin_alpha, cos_alpha;
    double diff_alpha, sin_diff_alpha, cos_diff_alpha;
    double zero = 0.0;
    double one = 1.0;
    /*static char function_name[] = "_wcs_astro_transform_ra_dec_sin";*/

    if (direction == DIR_ADtoXY)
    {
	/*  Convert from RA,DEC to x,y  */
	struct cd_matrix_type *mat = &ap->sky.LMtoXY;

	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert degrees to offset radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    diff_alpha = (in_ra - ap->sky.ra.reference) * deg_to_rad;
	    delta = in_dec * deg_to_rad;
	    cos_delta = cos (delta);
	    l = cos_delta * sin (diff_alpha);
	    m = sin (delta) * ap->sky.dec.cos_ref -
		cos_delta * ap->sky.dec.sin_ref * cos (diff_alpha);
	    ra[count] = ap->sky.ra.ref_pos + l * mat->cd_11 + m * mat->cd_12;
	    dec[count] = ap->sky.dec.ref_pos + l * mat->cd_21 + m * mat->cd_22;
	}
    }
    else if (direction == DIR_XDtoAY)
    {
	/*  Convert from x,DEC to RA,y  */
	double b, c, s;

	if ( DOUBLE_IS_BLANK (ap->sky.cos_rotation) )
	{
	    generic_xdec2ray (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    alpha = ap->sky.cos_rotation;
	    x = (ra[count] - ap->sky.ra.ref_pos) * ap->sky.ra.pix_to_rad;
	    b = ap->sky.dec.sin_ref * ap->sky.sin_rotation;
	    delta = deg_to_rad * dec[count];
	    cos_delta = cos (delta);
	    sin_delta = sin (delta);
	    s = x - sin_delta * ap->sky.dec.cos_ref * ap->sky.sin_rotation;
	    tmp = cos_delta * sqrt (alpha * alpha + b * b);
	    diff_alpha = atan2 (b, alpha) + asin (s / tmp);
	    cos_alpha = cos (diff_alpha);
	    sin_alpha = sin (diff_alpha);
	    alpha = sin_delta * ap->sky.dec.cos_ref * ap->sky.cos_rotation;
	    b = cos_delta * ap->sky.dec.sin_ref * cos_alpha * ap->sky.cos_rotation;
	    c = cos_delta * sin_alpha * ap->sky.sin_rotation;
	    ra[count] = ap->sky.ra.reference + rad_to_deg * diff_alpha;
	    dec[count] = ap->sky.dec.ref_pos + ap->sky.dec.rad_to_pix * (alpha - b -c);
	}
    }
    else if (direction == DIR_AYtoXD)
    {
	/*  Convert from RA,y to x,DEC  */
	double b, c;

	if ( DOUBLE_IS_BLANK (ap->sky.sin_rotation) )
	{
	    generic_ray2xdec (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
    	    diff_alpha = deg_to_rad * (ra[count] - ap->sky.ra.reference);
	    cos_diff_alpha = cos (diff_alpha);
	    sin_diff_alpha = sin (diff_alpha);
	    y = (dec[count] - ap->sky.dec.ref_pos) * ap->sky.dec.pix_to_rad;
	    alpha = ap->sky.dec.cos_ref * ap->sky.cos_rotation;
	    b = ap->sky.dec.sin_ref * cos_diff_alpha * ap->sky.cos_rotation +
		sin_diff_alpha * ap->sky.sin_rotation;
	    delta = atan (b / alpha) + asin ( y / sqrt (alpha *alpha + b *b) );
	    cos_delta = cos (delta);
	    sin_delta = sin (delta);
	    alpha = cos_delta * sin_diff_alpha * ap->sky.cos_rotation;
	    b = sin_delta * ap->sky.dec.cos_ref * ap->sky.sin_rotation;
	    c = cos_delta * ap->sky.dec.sin_ref * cos_diff_alpha *ap->sky.sin_rotation;
	    ra[count] = ap->sky.ra.ref_pos + ap->sky.ra.rad_to_pix * (alpha + b - c);
	    dec[count] = rad_to_deg * delta;
	}
    }
    else if (direction == DIR_XYtoAD)
    {
	/*  Convert from x,y to RA,DEC  */
	struct cd_matrix_type *mat = &ap->sky.XYtoLM;

	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert pixels to offset radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    x = in_ra - ap->sky.ra.ref_pos;
	    y = in_dec - ap->sky.dec.ref_pos;
	    l = x * mat->cd_11 + y * mat->cd_12;
	    m = x * mat->cd_21 + y * mat->cd_22;
	    tmp = one - l * l - m * m;
	    if (tmp < zero)
	    {
		ra[count] = toobig;
		dec[count] = toobig;
		continue;
	    }
	    tmp = sqrt (tmp);
	    diff_alpha = atan2 (l,
				ap->sky.dec.cos_ref * tmp -
				m * ap->sky.dec.sin_ref);
	    /*  Convert back to degrees and add to reference  */
	    ra[count] = ap->sky.ra.reference + diff_alpha * rad_to_deg;
	    dec[count] = rad_to_deg * asin (m * ap->sky.dec.cos_ref +
					    ap->sky.dec.sin_ref * tmp);
	}
    }
}   /*  End Function transform_ra_dec_sin  */

static void transform_ra_dec_tan (KwcsAstro ap, unsigned int num_coords,
				  double *ra, double *dec,
				  unsigned int direction)
/*  [SUMMARY] Gnomonic projection of co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    <direction> The direction of the transform.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double deg_to_rad = PION180;
    double rad_to_deg = 180.0 / PI;
    double l, m, x, y, tmp, in_ra, in_dec;
    double delta, sin_delta, cos_delta;
    double alpha;
    double diff_alpha, sin_diff_alpha, cos_diff_alpha;

    if (direction == DIR_ADtoXY)
    {
	/*  Convert from RA,DEC to x,y  */
	struct cd_matrix_type *mat = &ap->sky.LMtoXY;

	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert degrees to offset radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    diff_alpha = (in_ra - ap->sky.ra.reference) * deg_to_rad;
	    cos_diff_alpha = cos (diff_alpha);
	    delta = in_dec * deg_to_rad;
	    sin_delta = sin (delta);
	    cos_delta = cos (delta);
	    tmp = sin_delta * ap->sky.dec.sin_ref +
		cos_delta * ap->sky.dec.cos_ref * cos_diff_alpha;
	    l = cos_delta * sin (diff_alpha) / tmp;
	    m = (sin_delta * ap->sky.dec.cos_ref -
		 cos_delta * ap->sky.dec.sin_ref * cos_diff_alpha) / tmp;
	    ra[count] = ap->sky.ra.ref_pos + l * mat->cd_11 + m * mat->cd_12;
	    dec[count] = ap->sky.dec.ref_pos + l * mat->cd_21 + m * mat->cd_22;
	}
    }
    else if (direction == DIR_XDtoAY)
    {
	/*  Convert from x,DEC to RA,y  */
	double b, c, s, tan_delta;
	/*struct cd_matrix_type *mat = &ap->sky.XMtoLY;*/

	if ( DOUBLE_IS_BLANK (ap->sky.cos_rotation) )
	{
	    generic_xdec2ray (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    x = ra[count] - ap->sky.ra.ref_pos;
	    in_dec = dec[count];
	    m = in_dec - ap->sky.dec.reference;
	    l = x * ap->sky.ra.pix_to_rad;
	    alpha = ap->sky.cos_rotation;
	    b = l * ap->sky.dec.cos_ref + ap->sky.dec.sin_ref * ap->sky.sin_rotation;
	    c = l * ap->sky.dec.sin_ref - ap->sky.dec.cos_ref * ap->sky.sin_rotation;
	    delta = deg_to_rad * in_dec;
	    cos_delta = cos (delta);
	    sin_delta = sin (delta);
	    tan_delta = sin_delta / cos_delta;
	    diff_alpha = atan (b / alpha) +
		asin ( tan_delta * c / sqrt (alpha * alpha + b * b) );
	    cos_diff_alpha = cos (diff_alpha);
	    sin_diff_alpha = sin (diff_alpha);
	    alpha = sin_delta * ap->sky.dec.cos_ref * ap->sky.cos_rotation;
	    b = cos_delta * ap->sky.dec.sin_ref * cos_diff_alpha *ap->sky.cos_rotation;
	    c = cos_delta * sin_diff_alpha * ap->sky.sin_rotation;
	    s = sin_delta * ap->sky.dec.sin_ref;
	    tmp = cos_delta * ap->sky.dec.cos_ref * cos_diff_alpha;
	    ra[count] = ap->sky.ra.reference + rad_to_deg * diff_alpha;
	    dec[count] = ap->sky.dec.ref_pos +
		(alpha - b - c) / (s + tmp) * ap->sky.dec.rad_to_pix;
	}
    }
    else if (direction == DIR_AYtoXD)
    {
	/*  Convert from RA,y to x,DEC  */
	double b, c, s, tan_delta;

	if ( DOUBLE_IS_BLANK (ap->sky.sin_rotation) )
	{
	    generic_ray2xdec (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    diff_alpha = deg_to_rad * (ra[count] - ap->sky.ra.reference);
	    cos_diff_alpha = cos (diff_alpha);
	    sin_diff_alpha = sin (diff_alpha);
	    y = (dec[count] - ap->sky.dec.ref_pos) * ap->sky.dec.pix_to_rad;
	    alpha = ap->sky.dec.sin_ref * cos_diff_alpha * ap->sky.cos_rotation;
	    b = sin_diff_alpha * ap->sky.sin_rotation;
	    c = y * ap->sky.dec.cos_ref * cos_diff_alpha;
	    s = ap->sky.dec.cos_ref * ap->sky.cos_rotation;
	    tmp = y * ap->sky.dec.sin_ref;
	    tan_delta = (alpha + b + c) / (s - tmp);
	    alpha = sin_diff_alpha * ap->sky.cos_rotation;
	    b = tan_delta * ap->sky.dec.cos_ref * ap->sky.sin_rotation;
	    c = ap->sky.dec.sin_ref * cos_diff_alpha * ap->sky.sin_rotation;
	    s = tan_delta * ap->sky.dec.sin_ref;
	    tmp = ap->sky.dec.cos_ref * cos_diff_alpha;
	    ra[count] = ap->sky.ra.ref_pos +
		(alpha + b - c) / (s + tmp) * ap->sky.ra.rad_to_pix;
	    dec[count] = rad_to_deg * atan (tan_delta);
	}
    }
    else if (direction == DIR_XYtoAD)
    {
	/*  Convert from x,y to RA,DEC  */
	double s;
	struct cd_matrix_type *mat = &ap->sky.XYtoLM;

	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert pixels to offset radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    x = in_ra - ap->sky.ra.ref_pos;
	    y = in_dec - ap->sky.dec.ref_pos;
	    l = mat->cd_11 * x + mat->cd_12 * y;
	    m = mat->cd_21 * x + mat->cd_22 * y;
	    s = m * ap->sky.dec.cos_ref + ap->sky.dec.sin_ref;
	    tmp = ap->sky.dec.cos_ref - m * ap->sky.dec.sin_ref;
	    diff_alpha = atan2 (l, tmp);
	    /*  Convert back to degrees and add to reference  */
	    ra[count] = ap->sky.ra.reference + diff_alpha * rad_to_deg;
	    dec[count] = rad_to_deg * atan2 (cos (diff_alpha) * s, tmp);
	}
    }
}   /*  End Function transform_ra_dec_tan  */

static void transform_ra_dec_ncp (KwcsAstro ap, unsigned int num_coords,
				  double *ra, double *dec,
				  unsigned int direction)
/*  [SUMMARY] North Celestial Pole projection of co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    <direction> The direction of the transform.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double deg_to_rad = PION180;
    double rad_to_deg = 180.0 / PI;
    double l, m, x, y, tmp, in_ra, in_dec;
    double delta, cos_delta;
    double alpha;
    double diff_alpha, sin_diff_alpha, cos_diff_alpha;
    double zero = 0.0;
    double one = 1.0;
    double minus_one = -1.0;

    if (direction == DIR_ADtoXY)
    {
	/*  Convert from RA,DEC to x,y  */
	struct cd_matrix_type *mat = &ap->sky.LMtoXY;

	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert degrees to offset radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    diff_alpha = (in_ra - ap->sky.ra.reference) * deg_to_rad;
	    delta = in_dec * deg_to_rad;
	    cos_delta = cos (delta);
	    l = cos_delta * sin (diff_alpha);
	    m = ( ap->sky.dec.cos_ref - cos_delta * cos (diff_alpha) ) /
		ap->sky.dec.sin_ref;
	    ra[count] = ap->sky.ra.ref_pos + l * mat->cd_11 + m * mat->cd_12;
	    dec[count] = ap->sky.dec.ref_pos + l * mat->cd_21 + m * mat->cd_22;
	}
    }
    else if (direction == DIR_XDtoAY)
    {
	/*  Convert from x,DEC to RA,y  */
	double b, c, s, out_ra;

	if ( DOUBLE_IS_BLANK (ap->sky.cos_rotation) )
	{
	    generic_xdec2ray (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    x = (ra[count] - ap->sky.ra.ref_pos) * ap->sky.ra.pix_to_rad;
	    cos_delta = cos (deg_to_rad * dec[count]);
	    alpha = ap->sky.cos_rotation;
	    b = ap->sky.sin_rotation / ap->sky.dec.sin_ref;
	    s = x * ap->sky.dec.sin_ref - ap->sky.dec.cos_ref * ap->sky.sin_rotation;
	    tmp = cos_delta * ap->sky.dec.sin_ref * sqrt (alpha * alpha + b * b);
	    diff_alpha = atan (b / alpha) + asin (s / tmp);
	    alpha = ap->sky.cos_rotation * ap->sky.dec.cos_ref / ap->sky.dec.sin_ref;
	    b = ap->sky.cos_rotation *cos_delta *cos (diff_alpha) /ap->sky.dec.sin_ref;
	    c = ap->sky.sin_rotation * cos_delta * sin (diff_alpha);
	    out_ra = ap->sky.ra.reference + rad_to_deg * diff_alpha;
	    while (out_ra < zero) out_ra += 360.0;
	    while (out_ra > 360.0) out_ra -= 360.0;
	    ra[count] = out_ra;
	    dec[count] = ap->sky.dec.ref_pos + (alpha - b -c) * ap->sky.dec.rad_to_pix;
	}
    }
    else if (direction == DIR_AYtoXD)
    {
	/*  Convert from RA,y to x,DEC  */
	double b, c, s, q;

	if ( DOUBLE_IS_BLANK (ap->sky.sin_rotation) )
	{
	    generic_ray2xdec (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    diff_alpha = deg_to_rad * (ra[count] - ap->sky.ra.reference);
	    y = (dec[count] - ap->sky.dec.ref_pos) * ap->sky.dec.pix_to_rad;
	    cos_diff_alpha = cos (diff_alpha);
	    sin_diff_alpha = sin (diff_alpha);
	    s = ap->sky.dec.cos_ref * ap->sky.cos_rotation - y * ap->sky.dec.sin_ref;
	    tmp = cos_diff_alpha * ap->sky.cos_rotation +
		sin_diff_alpha * ap->sky.sin_rotation * ap->sky.dec.sin_ref;
	    q = s / tmp;
	    if (q > one) q = one;
	    else if (q < minus_one) q = minus_one;
	    delta = acos (q);
	    if (ap->sky.dec.reference > zero) delta = fabs (delta);
	    else delta = -fabs (delta);
	    cos_delta = cos (delta);
	    alpha = cos_delta * sin_diff_alpha * ap->sky.cos_rotation;
	    b = ap->sky.dec.cos_ref * ap->sky.sin_rotation / ap->sky.dec.sin_ref;
	    c = cos_delta * cos_diff_alpha * ap->sky.sin_rotation /ap->sky.dec.sin_ref;
	    ra[count] = ap->sky.ra.ref_pos + (alpha + b - c) * ap->sky.ra.rad_to_pix;
	    dec[count] = rad_to_deg * delta;
	}
    }
    else if (direction == DIR_XYtoAD)
    {
	/*  Convert from x,y to RA,DEC  */
	double out_ra;
	struct cd_matrix_type *mat = &ap->sky.XYtoLM;

	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert pixels to offset radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    x = in_ra - ap->sky.ra.ref_pos;
	    y = in_dec - ap->sky.dec.ref_pos;
	    l = mat->cd_11 * x + mat->cd_12 * y;
	    m = mat->cd_21 * x + mat->cd_22 * y;
	    diff_alpha = atan2 (l, ap->sky.dec.cos_ref - m * ap->sky.dec.sin_ref);
	    cos_diff_alpha = cos (diff_alpha);
	    delta = acos ( (ap->sky.dec.cos_ref - m * ap->sky.dec.sin_ref) /
			   cos_diff_alpha );
	    if (ap->sky.dec.reference > zero) delta = fabs (delta);
	    else delta = -fabs (delta);
	    /*  Convert back to degrees and add to reference  */
	    out_ra = ap->sky.ra.reference + diff_alpha * rad_to_deg;
	    while (out_ra < zero) out_ra += 360.0;
	    while (out_ra > 360.0) out_ra -= 360.0;
	    ra[count] = out_ra;
	    dec[count] = rad_to_deg * delta;
	}
    }
}   /*  End Function transform_ra_dec_ncp  */

static void transform_ra_dec_flat (KwcsAstro ap, unsigned int num_coords,
				   double *ra, double *dec,
				   unsigned int direction)
/*  [SUMMARY] Flat projection of co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    <direction> The direction of the transform.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double in_ra, in_dec, l, m, x, y;
    double deg_to_rad = PION180;
    double rad_to_deg = 180.0 / PI;

    if (direction == DIR_ADtoXY)
    {
	/*  Convert from RA,DEC to x,y  */
	struct cd_matrix_type *mat = &ap->sky.LMtoXY;

	for (count = 0; count < num_coords; ++count)
	{
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    l = (in_ra - ap->sky.ra.reference) * deg_to_rad;
	    m = (in_dec - ap->sky.dec.reference) * deg_to_rad;
	    ra[count] = ap->sky.ra.ref_pos + l * mat->cd_11 + m * mat->cd_12;
	    dec[count] = ap->sky.dec.ref_pos + l * mat->cd_21 + m * mat->cd_22;
	}
    }
    else if (direction == DIR_XDtoAY)
    {
	/*  Convert from x,DEC to RA,y  */
	struct cd_matrix_type *mat = &ap->sky.XMtoLY;

	for (count = 0; count < num_coords; ++count)
	{
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    x = in_ra - ap->sky.ra.ref_pos;
	    m = (in_dec - ap->sky.dec.reference) * deg_to_rad;
	    l = x * mat->cd_11 + m * mat->cd_12;
	    y = x * mat->cd_21 + m * mat->cd_22;
	    ra[count] = ap->sky.ra.reference + l * rad_to_deg;
	    dec[count] = ap->sky.dec.ref_pos + y;
	}
    }
    else if (direction == DIR_AYtoXD)
    {
	/*  Convert from RA,y to x,DEC  */
	struct cd_matrix_type *mat = &ap->sky.LYtoXM;

	for (count = 0; count < num_coords; ++count)
	{
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    l = (in_ra - ap->sky.ra.reference) * deg_to_rad;
	    y = in_dec - ap->sky.dec.ref_pos;
	    x = l * mat->cd_11 + y * mat->cd_12;
	    m = l * mat->cd_21 + y * mat->cd_22;
	    ra[count] = ap->sky.ra.ref_pos + x;
	    dec[count] = ap->sky.dec.reference + m * rad_to_deg;
	}
    }
    else if (direction == DIR_XYtoAD)
    {
	/*  Convert from x,y to RA,DEC  */
	struct cd_matrix_type *mat = &ap->sky.XYtoLM;

	for (count = 0; count < num_coords; ++count)
	{
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    x = in_ra - ap->sky.ra.ref_pos;
	    y = in_dec - ap->sky.dec.ref_pos;
	    l = x * mat->cd_11 + y * mat->cd_12;
	    m = x * mat->cd_21 + y * mat->cd_22;
	    ra[count] = ap->sky.ra.reference + l * rad_to_deg;
	    dec[count] = ap->sky.dec.reference + m * rad_to_deg;
	}
    }
}   /*  End Function transform_ra_dec_flat  */

static void transform_ra_dec_arc (KwcsAstro ap, unsigned int num_coords,
				  double *ra, double *dec,
				  unsigned int direction)
/*  [SUMMARY] Rectangular projection of co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    <direction> The direction of the transform.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double deg_to_rad = PION180;
    double rad_to_deg = 180.0 / PI;
    double l, m, x, y, in_ra, in_dec;
    double delta, sin_delta, cos_delta;
    double diff_alpha, sin_diff_alpha, cos_diff_alpha;
    double zero = 0.0;
    double one = 1.0;
    static char function_name[] = "_wcs_astro_transform_ra_dec_arc";

    if (direction == DIR_ADtoXY)
    {
	/*  Convert from RA,DEC to x,y  */
	double a, s, t;
	struct cd_matrix_type *mat = &ap->sky.LMtoXY;

	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert degrees to offset radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    diff_alpha = (in_ra - ap->sky.ra.reference) * deg_to_rad;
	    delta = in_dec * deg_to_rad;
	    cos_diff_alpha = cos (diff_alpha);
	    sin_diff_alpha = sin (diff_alpha);
	    cos_delta = cos (delta);
	    sin_delta = sin (delta);
	    s = sin_delta * ap->sky.dec.sin_ref +
		cos_delta * ap->sky.dec.cos_ref * cos_diff_alpha;
	    if (s > one) s = one;
	    else if (s < -one) s = -one;
	    t = acos (s);
	    if (t == zero) a = one;
	    else a = t / sin (t);
	    l = a * cos_delta * sin_diff_alpha;
	    m = a * (sin_delta * ap->sky.dec.cos_ref -
		     cos_delta * ap->sky.dec.sin_ref * cos_diff_alpha);
	    ra[count] = ap->sky.ra.ref_pos + l * mat->cd_11 + m * mat->cd_12;
	    dec[count] = ap->sky.dec.ref_pos + l * mat->cd_21 + m * mat->cd_22;
	}
    }
    else if (direction == DIR_XDtoAY)
    {
	/*  Convert from x,DEC to RA,y  */
	int timeout;
	double a, b, f, t, yold;

	if (ap->sky.cos_rotation == zero) return;
	if ( DOUBLE_IS_BLANK (ap->sky.cos_rotation) )
	{
	    generic_xdec2ray (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    x = (ra[count] - ap->sky.ra.ref_pos) * ap->sky.ra.pix_to_rad;
	    delta = deg_to_rad * dec[count];
	    cos_delta = cos (delta);
	    sin_delta = sin (delta);
	    a = x * ap->sky.sin_rotation * ap->sky.dec.cos_ref;
	    b = ap->sky.cos_rotation * ap->sky.dec.cos_ref;
	    y = zero;
	    timeout = 0;
	    do
	    {
		yold = y;
		t = sqrt (x * x + y * y);
		if (t == zero) f = one;
		else f = sin (t) / t;
		y = (sin_delta - ap->sky.dec.sin_ref * cos (t) - a * f) / b /f;
	    }
	    while ( (fabs (y - yold) > 1e-10) && (++timeout < 100) );
	    if (timeout >= 100)
	    {
		fprintf (stderr, "%s: TIMEOUT: x: %g  DEC: %g  err: %g\n",
			 function_name, ra[count], dec[count],
			 (y - yold) * ap->sky.dec.rad_to_pix);
		abort ();
	    }
	    t = sqrt (x * x + y * y);
	    if (t == zero) f = one;
	    else f = sin (t) / t;
	    diff_alpha = asin (f * (x * ap->sky.cos_rotation -
				    y * ap->sky.sin_rotation) / cos_delta);
	    ra[count] = ap->sky.ra.reference + rad_to_deg * diff_alpha;
	    dec[count] = ap->sky.dec.ref_pos + y * ap->sky.dec.rad_to_pix;
	}
    }
    else if (direction == DIR_AYtoXD)
    {
	/*  Convert from RA,y to x,DEC  */
	int timeout;
	double a, b, c, s, t, xold;

	if (ap->sky.cos_rotation == zero) return;
	if ( DOUBLE_IS_BLANK (ap->sky.sin_rotation) )
	{
	    generic_ray2xdec (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
    	    diff_alpha = deg_to_rad * (ra[count] - ap->sky.ra.reference);
	    cos_diff_alpha = cos (diff_alpha);
	    sin_diff_alpha = sin (diff_alpha);
	    y = (dec[count] - ap->sky.dec.ref_pos) * ap->sky.dec.pix_to_rad;
	    a = atan (ap->sky.dec.sin_ref / ap->sky.dec.cos_ref / cos_diff_alpha);
	    b = sqrt (one - ap->sky.dec.cos_ref * ap->sky.dec.cos_ref *
		      sin_diff_alpha * sin_diff_alpha);
	    x = zero;
	    timeout = 0;
	    do
	    {
		xold = x;
		t = sqrt (x * x + y * y);
		s = cos (t) / b;
		if (s > one) s = one;
		else if (s < -one) s = one;
		c = acos (s);
		if (y < zero) delta = a - c;
		else delta = a + c;
		x = ( y * ap->sky.sin_rotation +
		      cos (delta) * sin_diff_alpha * t / sin (t) ) /
		    ap->sky.cos_rotation;
	    }
	    while ( (fabs (x - xold) > 1e-10) && (++timeout < 100) );
	    if (timeout >= 100)
	    {
		fprintf (stderr, "%s: TIMEOUT: RA: %g  y: %g  err: %g\n",
			 function_name, ra[count], dec[count],
			 (x - xold) * ap->sky.ra.rad_to_pix);
	    }
	    ra[count] = ap->sky.ra.ref_pos + x * ap->sky.ra.rad_to_pix;
	    dec[count] = rad_to_deg * delta;
	}
    }
    else if (direction == DIR_XYtoAD)
    {
	/*  Convert from x,y to RA,DEC  */
	double a, t, sin_delta, one_on_dec_sin_ref;
	struct cd_matrix_type *mat = &ap->sky.XYtoLM;

	one_on_dec_sin_ref = one / ap->sky.dec.sin_ref;
	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert pixels to offset radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    x = in_ra - ap->sky.ra.ref_pos;
	    y = in_dec - ap->sky.dec.ref_pos;
	    l = mat->cd_11 * x + mat->cd_12 * y;
	    m = mat->cd_21 * x + mat->cd_22 * y;
	    t = sqrt (l * l + m * m);
	    if (t == zero) a = one;
	    else a = sin (t) / t;
	    sin_delta = a * m * ap->sky.dec.cos_ref +
		ap->sky.dec.sin_ref * cos (t);
	    delta = asin (sin_delta);
	    /*  Convert back to degrees and add to reference  */
	    ra[count] = ap->sky.ra.reference +
		atan2 (l * a,
		       (sin_delta * ap->sky.dec.cos_ref - m * a) *
		       one_on_dec_sin_ref) * rad_to_deg;
	    dec[count] = rad_to_deg * delta;
	}
    }
}   /*  End Function transform_ra_dec_arc  */

static void transform_ra_dec_zea (KwcsAstro ap, unsigned int num_coords,
				  double *ra, double *dec,
				  unsigned int direction)
/*  [SUMMARY] Zenithal Equal Area projection of co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    <direction> The direction of the transform.
    [NOTE] The method of doing the projection draws on the new FITS coordinates
    proposal of Greisen & Calabretta (1996), so the equations look somewhat
    different.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double l, m, x, y, in_ra, in_dec, t, alpha, delta, R_theta, z;
    double deg_to_rad = PION180;
    double rad_to_deg = 180.0 / PI;
    double diff_alpha;
    double longpole, latpole, phi_p, theta_p;
    double alpha_p = 0.0;  /*  Initialised to keep compiler happy  */
    double delta_p = 0.0;  /*  Initialised to keep compiler happy  */
    double delta1, delta2;
    double theta_ref, alpha_ref, delta_ref, theta, phi, diff_phi;
    CONST double epsilon = 1e-10;

    /*  Initial setup for wcslib-style conversions. This stuff should
	eventually go into wcs_astro_setup ()
	Most of the work is getting the ra & dec of the n.pole of the 'native'
	spherical coordinate system so that the appropriate rotations can be
	done.
	To calculate those, need the native long & lat of the n.pole, which
	depends on the projection geometry being used.
	For zenithal projections phi_P = 180deg and normally theta_P = 90.0
	These are stored in the header keywords LONGPOLE & LATPOLE, in the new
	system.
	Also need to know the the native-system coords of the standard-system
	n.pole, phi_ref and theta_ref (phi_0 & theta_0 in GC96).
	The calculation of delta_p is inherently double-valued, there are a
	number of tests to be made that decide which is correct  */

    /* LONGPOLE defaults to 0 degrees if the celestial latitude of the
       reference point of the projection is greater than the native
       latitude, otherwise 180 degrees. */
    longpole = 180.0; /* for zenithal proj since native lat of ref pt is +90.*/
    latpole  = 90.0;
    phi_p    = longpole * deg_to_rad;
    theta_p  = latpole * deg_to_rad;

    /* the reference point for a zenithal projection is the N. pole of the
       native spherical coordinate system */
    /*phi_ref   = 0.0;*/
    theta_ref = PI_ON_2;

    alpha_ref = ap->sky.ra.reference * deg_to_rad;
    delta_ref = ap->sky.dec.reference * deg_to_rad;

    /* calculate delta_p = atan2(thing) + or - acos( otherthing ) */
    /* get `otherthing' */
    t = sin (theta_ref) * sin (theta_ref) + 
        cos (theta_ref) * cos (theta_ref) * cos (phi_p) * cos (phi_p);

    /* this trap comes the from wcslib code. */
    if ( t == 0.0 )
    {
	if ( ap->sky.dec.sin_ref != 0.0 )
	{
	    fprintf(stderr,
		    "wcslib_setup: Invalid coordinate transformation parameters." );
	    return;
	}
	theta_p = latpole * deg_to_rad;
    }
    else
    {
	t = ap->sky.dec.sin_ref / sqrt (t);
	/*  GC condition 3  */
	if (fabs (t) > 1)
	{
	    fprintf (stderr,
		     "wcslib_setup: delta_p undefined - bad argument to acos!\n");
	    return;
	}
	t = acos (t);
	/* now `thing'. This is a little obscured, for the sake of speed */
	delta1 = atan2 ( sin (theta_ref), cos (theta_ref) * cos (phi_p) );
	delta2 = delta1 - t;
	delta1 += t;

	/* make sure delta is well-conditioned */
	if ( delta1 > PI )
	{
	    delta1 -= TWO_PI;
	}
	else if ( delta1 < -PI )
	{
	    delta1 += TWO_PI;
	}
	if ( delta2 > PI )
	{
	    delta2 -= TWO_PI;
	}
	else if ( delta2 < -PI )
	{
	    delta2 += TWO_PI;
	}

	/* ATAN2 sometimes gives slightly larger than pi/2 when it means pi/2 */
	if (fabs (delta1 - PI_ON_2) < epsilon) delta1 = PI_ON_2;
	else if (fabs (delta1 + PI_ON_2) < epsilon) delta1 = -PI_ON_2;
	if (fabs (delta2 - PI_ON_2) < epsilon) delta2 = PI_ON_2;
	else if (fabs (delta2 + PI_ON_2) < epsilon) delta2 = -PI_ON_2;

	/* if latpole keyword exists use it to choose between the northerly or
	   southerly soln for delta_p: GC condition 4. */
	if ( fabs(theta_p - delta1) < fabs (theta_p - delta2) )
	{
	    /* valid delta_p is in range +/- pi/2, GC condition 3 */
	    if ( fabs(delta1) <= PI_ON_2 )
	    {
		delta_p = delta1;
	    }	else delta_p = delta2;
	}
	else
	{
	    if ( fabs(delta2) <= PI_ON_2 )
	    {
		delta_p = delta2;
	    }	else delta_p = delta1;
	}
	/* it is not obvious to me from the GC text that this should be done
	   but it is done in wcslib; */
	theta_p = delta_p;
    }
    /* I have not computed 90-delta_p as well, I don't think it's needed. */

    /* so now we know delta_p. Next: figure out alpha_p */

    z = cos(delta_p)*ap->sky.dec.cos_ref;
    if ( fabs(z) < epsilon )
    {
	/* we're near a pole. */
	if ( fabs(ap->sky.dec.cos_ref) < epsilon )
	{
	    /* celestial pole at the reference point, GC condition 1. */
	    alpha_p = alpha_ref;
	}
	else if ( delta_p > 0.0 )
	{
	    /* celestial pole at native n. pole, GC condition 2. */
	    alpha_p = alpha_ref + phi_p - PI ;
	}
	else if ( delta_p < 0.0 )
	{
	    /* celestial pole at native s. pole, GC condition 2. */
	    alpha_p = alpha_ref - phi_p;
	}
    }
    else
    {
	x = ( sin(theta_ref) - sin(delta_p)*sin(delta_ref) ) / z;
	y = sin(phi_p) * cos(theta_ref) * cos(delta_p) / ap->sky.dec.cos_ref;
	/* GC condition 6 */
	if ( x == 0.0 && y == 0.0 )
	{
	    fprintf(stderr,
		    "wcslib_setup: Invalid coordinate transformation parameters." );
	    return;
	}
	alpha_p = alpha_ref - atan2(y,x);
	/* make sure alpha is well-conditioned */
	if ( alpha_p < 0.0 )
	{
	    alpha_p += TWO_PI;
	}
	else if ( alpha_p > TWO_PI )
	{
	    alpha_p -= TWO_PI;
	}
    }

    alpha_p *= rad_to_deg;
    delta_p *= rad_to_deg;
    phi_p   *= rad_to_deg;
    theta_p *= rad_to_deg;
    /* end of setup, now to apply all this stuff. */

    if (direction == DIR_ADtoXY)
    {
	/*  Convert from RA,DEC to x,y  */
	struct cd_matrix_type *mat = &ap->sky.LMtoXY;

	alpha_p    *= deg_to_rad;
	delta_p    *= deg_to_rad;
	phi_p      *= deg_to_rad;

	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert degrees to offset angle in radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) )continue;
	    alpha       = in_ra  * deg_to_rad;
	    delta       = in_dec * deg_to_rad;
	    diff_alpha  = alpha - alpha_p;

	    /* get the native spherical coords */
	    /* native longitude */
	    x = sin(delta) * sin(PI_ON_2-delta_p) -
	        cos(delta) * cos(PI_ON_2-delta_p) * cos(diff_alpha);
	    y = cos(delta) * sin(diff_alpha) * -1 ;
	    if (fabs(x) < epsilon)
	    {
		x = -1*cos(delta+PI_ON_2-delta_p) +
		    cos(delta)*cos(PI_ON_2-delta_p)*(1.0 - cos(diff_alpha)) ;
	    }
	    if ( x != 0.0 || y != 0.0 )
	    {
		diff_phi = atan2( y, x );
	    }
	    else
	    {
		diff_phi = diff_alpha - PI;
	    }
	    phi = diff_phi + phi_p;
	    if ( phi > PI )
	    {
		phi -= TWOPI;
	    }
	    else if ( phi < -PI )
	    {
		phi += TWOPI;
	    }

	    /* native latitude */
	    if ( fmod(diff_phi, PI) == 0.0 )
	    {
		theta = delta + cos(diff_alpha)*(PI_ON_2-delta_p);
		if (theta >  PI_ON_2) theta =  PI - theta;
		if (theta < -PI_ON_2) theta = -PI - theta;
	    }
	    else
	    {
		z = sin(delta) * cos(PI_ON_2-delta_p) +
		    cos(delta) * sin(PI_ON_2-delta_p) * cos(diff_alpha);
		if ( fabs(z) <= 0.99 )
		{
		    theta = asin (z);
		}
		else
		{
		    /* use this formula for better numerical accuracy */
		    theta = copysign( acos(sqrt(x*x+y*y)), z );
		}
	    }

	    /* project the native spherical coords */
	    /* There are two forms,
	       R_theta = sqrt( 2.0*(1.0-sin(theta)) );
	       and
	       R_theta = 2*sin( (PI_ON_2-theta)/2 ) ;
	       */
	    R_theta = 2.0 * sin( 0.5*(PI_ON_2 - theta) );
	    l = R_theta * sin(phi);
	    m = -1.0 * R_theta * cos(phi);
	    ra[count] = ap->sky.ra.ref_pos + l * mat->cd_11 + m * mat->cd_12;
	    dec[count] = ap->sky.dec.ref_pos + l * mat->cd_21 + m * mat->cd_22;
	}
    }
    else if (direction == DIR_XDtoAY)
    {
	/*  Convert from x,DEC to RA,y  */
	generic_xdec2ray (ap, num_coords, ra, dec);
	return;
    }
    else if (direction == DIR_AYtoXD)
    {
	/*  Convert from RA,y to x,DEC  */
	generic_ray2xdec (ap, num_coords, ra, dec);
	return;
    }
    else if (direction == DIR_XYtoAD)
    {
	/*  Convert from x,y to RA,DEC  */
	struct cd_matrix_type *mat = &ap->sky.XYtoLM;

	alpha_p *= deg_to_rad;
	delta_p *= deg_to_rad;
	phi_p   *= deg_to_rad;
	theta_p *= deg_to_rad;
	for (count = 0; count < num_coords; ++count)
	{
	    /*  Convert pixels to offset angle in radians  */
	    in_ra = ra[count];
	    in_dec = dec[count];
	    x = in_ra - ap->sky.ra.ref_pos;
	    y = in_dec - ap->sky.dec.ref_pos;
	    l = mat->cd_11 * x + mat->cd_12 * y;
	    m = mat->cd_21 * x + mat->cd_22 * y;
	    /* compute 'projected native spherical' coords */
	    R_theta = sqrt( l*l + m*m );
	    if ( R_theta == 0)
	    {
		phi = 0.0;
	    }
	    else
	    {
		phi = atan2 (l, -m);
	    }
	    /* this formula may suffer from numerical errors when close to
	       the point opposite the projection center */
	    theta = PI_ON_2 - 2.0 * asin( R_theta * 0.5 );

	    /* transform to standard spherical coords (ra, dec) */
	    diff_phi = phi - phi_p;

	    /* first get the right ascension */
	    x =  sin(theta) * sin(PI_ON_2-theta_p) -
		cos(theta) * cos(PI_ON_2-theta_p) * cos(diff_phi);
	    y = -cos(theta) * sin(diff_phi);
	    if ( fabs(x) < epsilon )
	    {
		/* for small x, this is more accurate */
		x = -cos(theta+PI_ON_2-theta_p) +
		    cos(theta)*cos(PI_ON_2-theta_p)*(1.0-cos(diff_phi));
	    }
	    if ( x != 0.0 || y != 0.0 )
	    {
		diff_alpha = atan2( y, x );
	    }
	    else
	    {
		/* change of origin of longitude */
		diff_alpha = diff_phi + PI;
	    }

            alpha = alpha_p + diff_alpha;
	    /* ensure alpha and alpha_p have same sign - this is belt&braces */
	    if ( alpha_p >= 0.0 )
	    {
		if ( alpha < 0.0 )
		{
		    alpha += TWO_PI;
		}
	    }
	    else if ( alpha < 0.0 )
	    {
		alpha -= TWO_PI;
	    }
	    /* now make it well-behaved. This IS necessary. */
	    if ( alpha > TWO_PI )
	    {
		alpha -= TWO_PI;
	    }
	    else if ( alpha < -TWO_PI )
	    {
		alpha += TWO_PI;
	    }

	    /* now get the declination */
	    if (fmod(diff_phi,PI) == 0.0)
	    {
		/* handle ra difference = 180deg explicitly */
		delta = theta + cos(phi)*(PI_ON_2-theta_p);
		if ( delta >  PI_ON_2 )
		{
		    delta =  PI - delta;
		}
		else if ( delta < -PI_ON_2 )
		{
		    delta = -PI - delta;
		}
	    }
	    else
	    {
		/* normal case */
		z = sin(theta)*cos(PI_ON_2-theta_p) +
		    cos(theta)*sin(PI_ON_2-theta_p)*cos(diff_phi);
		if ( fabs(z) <= 0.99 )
		{
		    delta = asin (z);
		}
		else
		{
		    /* use this formula for better numerical accuracy */
		    delta = copysign( acos(sqrt(x*x+y*y)), z );
		}
	    }

            /* whew! */
	    ra[count]  = alpha * rad_to_deg;
            dec[count] = delta * rad_to_deg;
        }
    }
}   /*  End Function transform_ra_dec_zea  */

static void transform_ra_dec_gls (KwcsAstro ap, unsigned int num_coords,
				  double *ra, double *dec,
				  unsigned int direction)
/*  [SUMMARY] Global Sinusoidal projection of co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    <direction> The direction of the transform.
    [RETURNS] Nothing.
*/
/*
    Coordinate convention notes:

	1. The ra[] and dec[] arrays are used for input and output of BOTH
		celestial coordinate and pixel coordinate values.
	2. The type of celestial coordinate (RA+DEC, GLON+GLAT) is transparent.
	3. All celestial coordinates are in DEGREES at input and output.
	4. All pixel coordinates are in PIXEL UNITS at input and output.
	5. All "difference" coordinates (offsets from references), both pixel
		and celestial, are in RADIANS for mathematical simplicity,
		unless they carry a _pix suffix.
*/
{
    unsigned int count, iter;
    double deg_to_rad = PION180, rad_to_deg = 180.0 / PI, toobig = TOOBIG;
    double tmp, l, m, in_x, in_y, out_x, out_y, diff_x, diff_y, diff_x_pix, 
	diff_y_pix, in_ra, in_dec, out_ra, out_dec, cos_dec, sin_rho, cos_rho, 
	diff_alpha, diff_delta, dold;
    /*static char function_name[] = "_wcs_astro_transform_ra_dec_gls";*/

    if (direction == DIR_ADtoXY)
    {    /*  Convert from RA,DEC to x,y  */
	struct cd_matrix_type *mat = &ap->sky.LMtoXY;
	for (count = 0; count < num_coords; ++count)
	{
	    in_ra  = ra[count];
	    in_dec = dec[count];
	    if (DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec)) continue;
	    diff_alpha = (in_ra  - ap->sky.ra.reference ) * deg_to_rad;
	    diff_delta = (in_dec - ap->sky.dec.reference) * deg_to_rad;
	    l = diff_alpha * cos (in_dec * deg_to_rad);
	    m = diff_delta;
	    diff_x_pix = l * mat->cd_11 + m * mat->cd_12;
	    diff_y_pix = l * mat->cd_21 + m * mat->cd_22;
	    out_x = ap->sky.ra.ref_pos  + diff_x_pix;
	    out_y = ap->sky.dec.ref_pos + diff_y_pix;
	    ra[count]  = out_x;
	    dec[count] = out_y;
	}
    }
    else if (direction == DIR_XDtoAY)
    {    /*  Convert from x,DEC to RA,y  */
	if ( DOUBLE_IS_BLANK (ap->sky.cos_rotation) )
	{
	    generic_xdec2ray (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    in_x   = ra[count];
	    in_dec = dec[count];
	    diff_x     = (in_x   - ap->sky.ra.ref_pos   ) * ap->sky.ra.pix_to_rad;
	    diff_delta = (in_dec - ap->sky.dec.reference) * deg_to_rad;
	    sin_rho = ap->sky.sin_rotation;
	    cos_rho = ap->sky.cos_rotation;
	    if (fabs (cos_rho) < 1.0e-10)
	    {
		ra[count]  = toobig;
		dec[count] = toobig;
		continue;
	    }
	    cos_dec = cos (in_dec * deg_to_rad);
	    if (fabs (cos_dec) < 1.0e-10)
	    {
		ra[count]  = toobig;
		dec[count] = toobig;
		continue;
	    }
	    diff_y     = (diff_delta       - diff_x * sin_rho) / cos_rho;
	    diff_alpha = (diff_x * cos_rho - diff_y * sin_rho) / cos_dec;
	    out_ra = ap->sky.ra.reference + diff_alpha * rad_to_deg;
	    out_y  = ap->sky.dec.ref_pos  + diff_y     * ap->sky.dec.rad_to_pix;
	    ra[count]  = out_ra;
	    dec[count] = out_y;
	}
    }
    else if (direction == DIR_AYtoXD)
    {    /*  Convert from RA,y to x,DEC  */
	if ( DOUBLE_IS_BLANK (ap->sky.sin_rotation) )
	{
	    generic_ray2xdec (ap, num_coords, ra, dec);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    in_ra = ra[count];
	    in_y  = dec[count];
	    diff_alpha = (in_ra - ap->sky.ra.reference) * deg_to_rad;
	    diff_y     = (in_y  - ap->sky.dec.ref_pos ) * ap->sky.dec.pix_to_rad;
	    sin_rho = ap->sky.sin_rotation;
	    cos_rho = ap->sky.cos_rotation;
	    if (fabs (cos_rho) < 1.0e-10)
	    {
		ra[count]  = toobig;
		dec[count] = toobig;
		continue;
	    }
	    out_dec = ap->sky.dec.reference + rad_to_deg * diff_y / cos_rho;
	    iter = 0;
	    do
	    {
		dold = out_dec;
		diff_delta = (diff_y + diff_alpha * cos (out_dec * deg_to_rad) 
			      * sin_rho) / cos_rho;
		out_dec = ap->sky.dec.reference + diff_delta * rad_to_deg;
	    } while (fabs (out_dec - dold) > 1.0e-10 && ++iter < 1000);
	    diff_x = diff_alpha * cos (out_dec * deg_to_rad) * cos_rho
		+ diff_delta * sin_rho;
	    out_x = ap->sky.ra.ref_pos + diff_x * ap->sky.ra.rad_to_pix;
	    ra[count]  = out_x;
	    dec[count] = out_dec;
	}
    }
    else if (direction == DIR_XYtoAD)
    {    /*  Convert from x,y to RA,DEC  */
	struct cd_matrix_type *mat = &ap->sky.XYtoLM;
	for (count = 0; count < num_coords; ++count)
	{
	    in_x = ra[count];
	    in_y = dec[count];
	    diff_x_pix = in_x - ap->sky.ra.ref_pos;
	    diff_y_pix = in_y - ap->sky.dec.ref_pos;
	    l = diff_x_pix * mat->cd_11 + diff_y_pix * mat->cd_12;
	    m = diff_x_pix * mat->cd_21 + diff_y_pix * mat->cd_22;
	    diff_delta = m;
	    out_dec = ap->sky.dec.reference + diff_delta * rad_to_deg;
	    tmp = cos (out_dec * deg_to_rad);
	    if (fabs (tmp) < 1.0e-10)
	    {
		ra[count]  = toobig;
		dec[count] = toobig;
		continue;
	    }
	    diff_alpha = l / tmp;
	    ra[count]  = ap->sky.ra.reference + diff_alpha * rad_to_deg;
	    dec[count] = out_dec;
	}
    }
}   /*  End Function transform_ra_dec_gls  */

static void transform_vel (KwcsAstro ap, unsigned int num_coords,
			   double *vel, flag to_linear)
/*  [SUMMARY] Velocity/frequency transformation.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <vel> A pointer to the velocity/frequency values. These will be modified.
    <to_linear> If TRUE the values are transformed to linear
    co-ordinates, else they are transformed to non-linear co-ordinates.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double n, val;
    double c = LIGHTSPEED;

    switch (ap->vel.type)
    {
      case SPECTRAL_UNKNOWN:
	break;
      case SPECTRAL_VELn_FreqOptical:
	/*  Channels are spaced linearly in frequency, using the "optical"
	    doppler convention, hence velocities are non-linearly spaced
	    "Optical" definition: v = c(f0/f - 1)
	    where v = radial velocity, f0 = rest frequency,
	    f = observed frequency, c = speed of light  */
	for (count = 0; count < num_coords; ++count)
	{
	    val = vel[count];
	    if ( DOUBLE_IS_BLANK (val) ) continue;
	    if (to_linear)
	    {
		/*  TODO: this should really be a non-linear transformation  */
		vel[count] = ap->vel.ref_pos + (val - ap->vel.reference)
		    / ap->vel.increment;
	    }
	    else
	    {
		n = val - ap->vel.ref_pos;
		vel[count] = 1.0 / ( ( n / (ap->vel.reference + ap->vel.increment + c) ) + (1.0 - n) / (ap->vel.reference + c) ) - c;
	    }
	}
	break;
      case SPECTRAL_VELl_FreqRadio:
      case SPECTRAL_VELl_WaveOptical:
	/*  Velocities are linearly spaced  */
      case SPECTRAL_FREQ:
	for (count = 0; count < num_coords; ++count)
	{
	    val = vel[count];
	    if ( DOUBLE_IS_BLANK (val) ) continue;
	    if (to_linear)
	    {
		vel[count] = ap->vel.ref_pos + (val - ap->vel.reference)
		    / ap->vel.increment;
	    }
	    else
	    {
		vel[count] = ap->vel.reference + (val - ap->vel.ref_pos)
		    * ap->vel.increment;
	    }
	}
	break;
      default:
	break;
    }
}   /*  End Function transform_vel  */

static void transform_linear (KwcsAstro ap, unsigned int num_coords,
			      CONST char *name, double *coords, flag to_pix)
/*  [SUMMARY] Linear transformation.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <name> The axis name. This may be NULL.
    <coords> A pointer to the co-ordinate values. These will be modified.
    <to_pix> If TRUE the values are transformed to pixel
    co-ordinates, else they are transformed to world co-ordinates.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double val;
    struct linear_coord_type *curr;
    struct linear_coord_type *found = NULL;

    if (name == NULL) return;
    /*  Abort transformation if it is done better elsewhere  */
    if ( (ap->sky.projection != PROJ_UNKNOWN) &&
	 (strcmp (name, ap->sky.ra.dim_name) == 0) ) return;
    if ( (ap->sky.projection != PROJ_UNKNOWN) &&
	 (strcmp (name, ap->sky.dec.dim_name) == 0) ) return;
    if ( (ap->vel.type != SPECTRAL_UNKNOWN) &&
	 (strcmp (name, ap->vel.dim_name) == 0) ) return;
    /*  First search for matching linear co-ordinate axis  */
    for (curr = ap->linear_axes; (curr != NULL) && (found == NULL);
	 curr = curr->next)
    {
	if (strcmp (curr->dim_name, name) == 0) found = curr;
    }
    if (found == NULL) return;
    for (count = 0; count < num_coords; ++count)
    {
	val = coords[count];
	if ( DOUBLE_IS_BLANK (val) ) continue;
	if (to_pix)
	{
	    coords[count] = found->ref_pos + (val - found->reference)
		/ found->increment;
	}
	else
	{
	    coords[count] = found->reference + (val - found->ref_pos)
		* found->increment;
	}
    }
}   /*  End Function transform_linear  */

static flag scan_dss_header (KwcsAstro ap, CONST packet_desc *pack_desc,
			     CONST char *packet)
/*  [SUMMARY] Attempt to scan DSS header information.
    <ap> The KwcsAstro object.
    <pack_desc> The packet descriptor containing the FITS-style keywords.
    <packet> The packet data containing the FITS-stype keyword data.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag scan;
    double pltdecsn;
    struct dss_plate_type *new;
    char *ptr;
    char txt[STRING_LENGTH];
    double cnpix1[2], cnpix2[2], pltscale[2], xpixelsz[2], ypixelsz[2];
    double pltrah[2], pltram[2], pltras[2], pltdecd[2], pltdecm[2], pltdecs[2];
    double ppo3[2], ppo6[2];
    double value[2];
    static char function_name[] = "scan_dss_header";

    if ( !ds_get_unique_named_value (pack_desc, packet, "CNPIX1", NULL,
				     cnpix1) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "CNPIX2", NULL,
				     cnpix2) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "PLTSCALE", NULL,
				     pltscale) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "XPIXELSZ", NULL,
				     xpixelsz) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "YPIXELSZ", NULL,
				     ypixelsz) ) return (TRUE);
    /*  Some fool decided it was better to put the plate centre RA and DEC
	values into seven (!) separate header keywords. Unbelievable!  */
    if ( !ds_get_unique_named_value (pack_desc, packet, "PLTRAH", NULL,
				     pltrah) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "PLTRAM", NULL,
				     pltram) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "PLTRAS", NULL,
				     pltras) ) return (TRUE);
    /*  And to hammer home the stupidity, a string is used to give the sign  */
    if ( ( ptr = ds_get_unique_named_string (pack_desc, packet, "PLTDECSN") )
	 == NULL ) return (TRUE);
    pltdecsn = (*ptr == '-') ? -1.0 : 1.0;
    m_free (ptr);
    if ( !ds_get_unique_named_value (pack_desc, packet, "PLTDECD", NULL,
				     pltdecd) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "PLTDECM", NULL,
				     pltdecm) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "PLTDECS", NULL,
				     pltdecs) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "PPO3", NULL,
				     ppo3) ) return (TRUE);
    if ( !ds_get_unique_named_value (pack_desc, packet, "PPO6", NULL,
				     ppo6) ) return (TRUE);
    /*  Now have all the "essentials" (except for some number of Xi and Eta
	co-efficients  */
    if ( ( new = (struct dss_plate_type *) m_alloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "plate solution structure");
	return (FALSE);
    }
    new->corner_pixel_x1 = cnpix1[0] + 0.5;
    new->corner_pixel_y1 = cnpix2[0] + 0.5;
    new->scale = pltscale[0];
    new->pixel_size_x = xpixelsz[0];
    new->pixel_size_y = ypixelsz[0];
    /*  Now have to convert those seven lovely RA,DEC values into one RA and
	one DEC value. I like having to do all this work  */
    new->centre_ra = (pltrah[0] + pltram[0] / 60.0 + pltras[0] / 3600.0) *15.0;
    new->centre_dec = pltdecd[0] + pltdecm[0] / 60.0 + pltdecs[0] / 3600.0;
    new->centre_dec *= pltdecsn;
    new->centre_off_x = ppo3[0];
    new->centre_off_y = ppo6[0];
    /*  Now get the Xi co-efficients  */
    for (scan = TRUE, new->num_amdx = 0; scan; )
    {
	sprintf (txt, "AMDX%u", new->num_amdx + 1);
	if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL, value) )
	{
	    new->amdx[new->num_amdx++] = value[0];
	}
	else scan = FALSE;
    }
    /*  Now get the Eta co-efficients  */
    for (scan = TRUE, new->num_amdy = 0; scan; )
    {
	sprintf (txt, "AMDY%u", new->num_amdy + 1);
	if ( ds_get_unique_named_value (pack_desc, packet, txt, NULL, value) )
	{
	    new->amdy[new->num_amdy++] = value[0];
	}
	else scan = FALSE;
    }
    ap->sky.dss = new;
    ap->sky.projection = PROJ_DSS;
    strcpy (ap->sky.ra.dim_name, "Axis 1");
    strcpy (ap->sky.dec.dim_name, "Axis 0");
    return (TRUE);
}   /*  End Function scan_dss_header  */

static void transform_ra_dec_dss (KwcsAstro ap, unsigned int num_coords,
				  double *ra, double *dec,
				  unsigned int direction)
/*  [SUMMARY] DSS projection of co-ordinates.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    <direction> The direction of the transform.
    [RETURNS] Nothing.
*/
{
    if (direction == DIR_ADtoXY)                          /*  RA,DEC to x,y  */
	dss_radec2xy (ap, num_coords, ra, dec);
    else if (direction == DIR_XDtoAY)                     /*  x,DEC to RA,y  */
	generic_xdec2ray (ap, num_coords, ra, dec);
    else if (direction == DIR_AYtoXD)                     /*  RA,y to x,DEC  */
	generic_ray2xdec (ap, num_coords, ra, dec);
    else if (direction == DIR_XYtoAD)                     /*  x,y to RA,DEC  */
	dss_xy2radec (ap, num_coords, ra, dec);
}   /*  End Function transform_ra_dec_dss  */

static void dss_xy2radec (KwcsAstro ap, unsigned int num_coords,
			  double *ra, double *dec)
/*  [SUMMARY] Convert x,y (pixels) on a DSS plate to RA,DEC.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double x, y, xx, yy, xy, xxPyy;
    double xi, eta, diff_ra;
    double arcsec2rad = PION180 / 3600.0;
    double cos_dec, tan_dec, f;
    double *a, *b;

    a = ap->sky.dss->amdx - 1;
    b = ap->sky.dss->amdy - 1;
    cos_dec = cos (ap->sky.dss->centre_dec * PION180);
    tan_dec = tan (ap->sky.dss->centre_dec * PION180);
    for (count = 0; count < num_coords; ++count)
    {
	/*  Convert pixel positions to offset from plate centre in mm  */
	x = 1e-3 * (ap->sky.dss->centre_off_x -
		    (ra[count] + ap->sky.dss->corner_pixel_x1) *
		    ap->sky.dss->pixel_size_x);
	/*  Y is flipped!  */
	y = -1e-3 * (ap->sky.dss->centre_off_y -
		     (dec[count] + ap->sky.dss->corner_pixel_y1) *
		     ap->sky.dss->pixel_size_y);
	xx = x * x;
	yy = y * y;
	xy = x * y;
	xxPyy = xx + yy;
	/*  Note how co-efficients start from 1, not 0. This makes it
	    easier to relate back to the equations  */
	xi = a[1] * x + a[2] * y + a[3] + a[4] * xx + a[5] * xy + a[6] * yy +
	    a[7] * xxPyy + a[8] * x * xx + a[9] * x * xy + a[10] * x * yy +
	    a[11] * y * yy + a[12] * x * xxPyy + a[13] * x * xxPyy * xxPyy;
	eta = b[1] * y + b[2] * x + b[3] + b[4] * yy + b[5] * xy +
	    b[6] * xx + b[7] * xxPyy + b[8] * y * yy + b[9] * x * yy +
	    b[10] * x * xy + b[11] * x * xx + b[12] * y * xxPyy +
	    b[13] * y * xxPyy * xxPyy;
	xi *= arcsec2rad;
	eta *= arcsec2rad;
	/*  Compute RA,DEC in J2000  */
	f = 1.0 - eta * tan_dec;
	diff_ra = atan2 (xi / cos_dec, f) / PION180;
	ra[count] = diff_ra + ap->sky.dss->centre_ra;
	dec[count] = atan ( ( (eta + tan_dec) *
			      cos (diff_ra * PION180) ) / f ) / PION180;
    }
}   /*  End Function dss_xy2radec  */

static void dss_radec2xy (KwcsAstro ap, unsigned int num_coords,
			  double *ra, double *dec)
/*  [SUMMARY] Convert RA,DEC to x,y (pixels) on a DSS plate.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra> A pointer to the right ascension values. These will be modified.
    <dec> A pointer to the declination values. These will be modified.
    [NOTE] This routine must compute the inverse of the plate solution, which
    requires iteration, so this may be a little slow.
    [RETURNS] Nothing.
*/
{
    int iter_count;
    int max_iter = 50;
    int total_iters = 0;
    unsigned int count;
    double tolerance, z, f, cos_dec, sin_dec, cos_dec0, sin_dec0;
    double in_ra, in_dec;
    double cos_dra, sin_dra;
    double xi0, eta0;
    double x, y, xx, yy, xy, xxPyy, xT2, yT2;
    double Dx = 0.0;  /*  Initialised to keep compiler happy  */
    double Dy = 0.0;  /*  Initialised to keep compiler happy  */
    double xi, eta, Dxi_Dx, Dxi_Dy, Deta_Dx, Deta_Dy, delta_xi, delta_eta;
    double *a, *b;

    a = ap->sky.dss->amdx - 1;
    b = ap->sky.dss->amdy - 1;
    /*  Compute tolerance (10% of pixel size)  */
    tolerance = ap->sky.dss->pixel_size_x;
    if (tolerance > ap->sky.dss->pixel_size_y) tolerance = ap->sky.dss->pixel_size_y;
    tolerance *= 1e-4;
    /*  Cache a few values  */
    z = ap->sky.dss->centre_dec * PION180;
    cos_dec0 = cos (z);
    sin_dec0 = sin (z);
    for (count = 0; count < num_coords; ++count)
    {
	/*  Convert RA, DEC to Xi, Eta co-rdinates  */
	in_ra = ra[count];
	in_dec = dec[count];
	if ( DOUBLE_IS_BLANK (in_ra) || DOUBLE_IS_BLANK (in_dec) ) continue;
	z = in_dec * PION180;
	cos_dec = cos (z);
	sin_dec = sin (z);
	z = (in_ra - ap->sky.dss->centre_ra) * PION180;
	cos_dra = cos (z);
	sin_dra = sin (z);
	f = 180.0 * 3600.0 / PI / (sin_dec * sin_dec0 +
				   cos_dec * cos_dec0 * cos_dra);
	xi0 = cos_dec * sin_dra * f;
	eta0 = (sin_dec * cos_dec0 - cos_dec * sin_dec0 * cos_dra) * f;
	/*  Begin iterating. Make a first guess of offset in mm  */
	x = xi0 / ap->sky.dss->scale;
	y = eta0 / ap->sky.dss->scale;
	for (iter_count = 0, Dx = 9.0 * tolerance, Dy = 9.0 * tolerance;
	     (iter_count < max_iter) && (fabs (Dx) > tolerance) &&
		 (fabs (Dy) > tolerance);
	     ++iter_count)
	{
	    /*  Compute plate co-ordinates and their derivatives  */
	    /*  Note how co-efficients start from 1, not 0. This makes it
		easier to relate back to the equations  */
	    xx = x * x;
	    yy = y * y;
	    xy = x * y;
	    xxPyy = xx + yy;
	    xT2 = 2.0 * x;
	    yT2 = 2.0 * y;
	    /*  Compute Xi  */
	    xi = a[1] * x + a[2] * y + a[3] + a[4] * xx + a[5] * xy +
		a[6] * yy + a[7] * xxPyy + a[8] * x * xx + a[9] * x * xy +
		a[10] * x * yy + a[11] * y * yy + a[12] * x * xxPyy +
		a[13] * x * xxPyy * xxPyy;
	    /*  Compute derivative of Xi wrt x  */
	    Dxi_Dx = ( a[1] + a[4] * xT2 + a[5] * y + a[7] * xT2 +
		       a[8] * 3.0 * xx + a[9] * 2.0 * xy + a[10] * yy +
		       a[12] * (3.0 * xx + yy) +
		       a[13] * (5.0 * xx * xx + 6.0 * xx * yy + yy * yy) );
	    /*  Compute derivative of Xi wrt y  */
	    Dxi_Dy = ( a[2] + a[5] * x + a[6] * yT2 + a[7] * yT2 +
		       a[9] * xx + a[10] * x * yT2 + a[11] * 3.0 * yy +
		       a[12] * x * yT2 +
		       a[13] * (4.0 * xx * xy + 4.0 * xy * yy) );
	    /*  Compute Eta  */
	    eta = b[1] * y + b[2] * x + b[3] + b[4] * yy + b[5] * xy +
		b[6] * xx + b[7] * xxPyy + b[8] * y * yy + b[9] * x * yy +
		b[10] * x * xy + b[11] * x * xx + b[12] * y * xxPyy +
		b[13] * y * xxPyy * xxPyy;
	    /*  Compute derivative of Eta wrt x  */
	    Deta_Dx = ( b[2] + b[5] * y + b[6] * xT2 + b[7] * xT2 + b[9] * yy +
			b[10] * xT2 * y + b[11] * 3.0 * xx + b[12] * y * xT2 +
			b[13] * y * (4.0 * x * xx + 4.0 * x * yy) );
	    /*  Compute derivative of Eta wrt y  */
	    Deta_Dy = ( b[1] + b[4] * yT2 + b[5] * x + b[7] * yT2 +
			b[8] * 3.0 * yy + b[9] * x * yT2 + b[10] * xx +
			b[12] * (xx + 3.0 * yy) +
			b[13] * (5.0 * yy * yy + 6.0 * yy * xx + xx * xx) );
	    /*  Compute error terms  */
	    delta_xi = xi0 - xi;
	    delta_eta = eta0 - eta;
	    /*  Compute correction  */
	    z = 1.0 / (Dxi_Dx * Deta_Dy - Dxi_Dy * Deta_Dx);
	    Dx = (delta_xi * Deta_Dy - delta_eta * Dxi_Dy) * z;
	    Dy = (delta_eta * Dxi_Dx - delta_xi * Deta_Dx) * z;
	    /*  Apply correction  */
	    x += Dx;
	    y += Dy;
	}
	total_iters += iter_count;
	/*  Convert offsets in mm to pixels  */
	ra[count] = (ap->sky.dss->centre_off_x - 1e3 * x) / ap->sky.dss->pixel_size_x -
	    ap->sky.dss->corner_pixel_x1;
	dec[count] = (ap->sky.dss->centre_off_y + 1e3 *y) / ap->sky.dss->pixel_size_y -
	    ap->sky.dss->corner_pixel_y1;
    }
}   /*  End Function dss_radec2xy  */

static void generic_xdec2ray (KwcsAstro ap, unsigned int num_coords,
			      double *ra_arr, double *dec_arr)
/*  [SUMMARY] Convert x,DEC to RA,y using an iterative (generic) solution.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra_arr> A pointer to the right ascension values. These will be modified.
    <dec_arr> A pointer to the declination values. These will be modified.
    [RETURNS] Nothing.
*/
{
    unsigned int count, iters;
    double true_x, true_dec, y, ra, dec;
    double old_y[2], old_ra[2], old_dec[2];
    static char function_name[] = "_wcs_astro_generic_xdec2ray";

    for (count = 0; count < num_coords; ++count)
    {
	true_x = ra_arr[count];
	true_dec = dec_arr[count];
	/*  Get initial y guesses  */
	old_y[0] = 0.0;
	old_y[1] = 100.0;
	/*  Convert x,y to RA,DEC  */
	old_ra[0] = true_x;
	old_ra[1] = true_x;
	old_dec[0] = old_y[0];
	old_dec[1] = old_y[1];
	(*ap->sky.transform_func) (ap, 2, old_ra, old_dec, DIR_XYtoAD);
	for (iters = 0; iters < 100; ++iters)
	{
	    /*  Linear interpolate to new Y guess  */
	    y = (true_dec - old_dec[0]) / (old_dec[1] - old_dec[0]) *
		(old_y[1] - old_y[0]) + old_y[0];
	    /*  Transform to DEC  */
	    ra = true_x;
	    dec = y;
	    (*ap->sky.transform_func) (ap, 1, &ra, &dec, DIR_XYtoAD);
	    if (fabs (true_dec - dec) < 1e-6) break;
	    /*  Not close enough yet: update furthest point  */
	    if ( fabs (old_dec[0] - true_dec) > fabs (old_dec[1] - true_dec) )
	    {
		old_dec[0] = dec;
		old_y[0] = y;
	    }
	    else
	    {
		old_dec[1] = dec;
		old_y[1] = y;
	    }
	}
	if (iters >= 100)
	{
	    fprintf (stderr, "%s: solution does not converge\n",function_name);
	    return;
	}
	ra_arr[count] = ra;
	dec_arr[count] = y;
    }
}   /*  End Function generic_xdec2ray  */

static void generic_ray2xdec (KwcsAstro ap, unsigned int num_coords,
			      double *ra_arr, double *dec_arr)
/*  [SUMMARY] Convert RA,y to x,DEC using an iterative (generic) solution.
    <ap> The KwcsAstro object.
    <num_coords> The number of co-ordinates to transform.
    <ra_arr> A pointer to the right ascension values. These will be modified.
    <dec_arr> A pointer to the declination values. These will be modified.
    [RETURNS] Nothing.
*/
{
    unsigned int count, iters;
    double true_ra, true_y, x, ra, dec;
    double old_x[2], old_ra[2], old_dec[2];
    static char function_name[] = "_wcs_astro_generic_ray2xdec";

    for (count = 0; count < num_coords; ++count)
    {
	true_ra = ra_arr[count];
	true_y = dec_arr[count];
	/*  Get initial x guesses  */
	old_x[0] = 0.0;
	old_x[1] = 100.0;
	/*  Convert x,y to RA,DEC  */
	old_ra[0] = old_x[0];
	old_ra[1] = old_x[1];
	old_dec[0] = true_y;
	old_dec[1] = true_y;
	(*ap->sky.transform_func) (ap, 2, old_ra, old_dec, DIR_XYtoAD);
	for (iters = 0; iters < 100; ++iters)
	{
	    /*  Linear interpolate to new X guess  */
	    x = (true_ra - old_ra[0]) / (old_ra[1] - old_ra[0]) *
		(old_x[1] - old_x[0]) + old_x[0];
	    /*  Transform to RA  */
	    ra = x;
	    dec = true_y;
	    (*ap->sky.transform_func) (ap, 1, &ra, &dec, DIR_XYtoAD);
	    if (fabs (true_ra - ra) < 1e-6) break;
	    /*  Not close enough yet: update furthest point  */
	    if ( fabs (old_ra[0] - true_ra) > fabs (old_ra[1] - true_ra) )
	    {
		old_ra[0] = ra;
		old_x[0] = x;
	    }
	    else
	    {
		old_ra[1] = ra;
		old_x[1] = x;
	    }
	}
	if (iters >= 100)
	{
	    fprintf (stderr, "%s: solution does not converge\n",function_name);
	    return;
	}
	ra_arr[count] = x;
	dec_arr[count] = dec;
    }
}   /*  End Function generic_ray2xdec  */

static void job_func (void *pool_info, void *call_info1, void *call_info2,
		      void *call_info3, void *call_info4, void *thread_info)
/*  [SUMMARY] Perform a job.
    <pool_info> The arbitrary pool information pointer.
    <call_info1> The first arbitrary call information pointer.
    <call_info2> The second arbitrary call information pointer.
    <call_info3> The third arbitrary call information pointer.
    <call_info4> The fourth arbitrary call information pointer.
    <thread_info> A pointer to arbitrary, per thread, information. This
    information is private to the thread.
    [RETURNS] Nothing.
*/
{
    CommonJobInfo *common_info = (CommonJobInfo *) call_info1;
    KwcsAstro ap = common_info->ap;

    (*ap->sky.transform_func) (ap, common_info->num_coords,
			       call_info2, call_info3, common_info->direction);
}   /*  End Function job_func  */
