/*LINTLIBRARY*/
/*  main.c

    Descriptor Management routines for the X Intrinsics toolkit

    Copyright (C) 1996-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for the management of file
    descriptors using the Xt toolkit.


    Written by      Richard Gooch   16-DEC-1996: Copied from
  source/lib/karma/packages/dm/native.c and
  source/lib/karmaXt/packages/chx/main.c

    Updated by      Richard Gooch   9-SEP-1997: Print error message if unable
  to get number of bytes readable.

    Updated by      Richard Gooch   31-DEC-1997: Add messages about bugs in
  IRIX 6.2 and Solaris 2.x in function <_dmx_read_fd>.

    Updated by      Richard Gooch   18-FEB-1998: Changed interfaces to support
  new KManagedDescriptor object.

    Updated by      Richard Gooch   19-FEB-1998: Added magic number for
  freelist.

    Updated by      Richard Gooch   28-JUN-1999: Took account of interface
  change for manage function.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Last updated by Richard Gooch   14-FEB-2000: Fixed bug in allocating last
  entry of freelist.


*/
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
/*  Do not include  karma.h  first, since CONST may be screwed otherwise. This
    applies to broken versions of  X11/Intrinsic.h  */
#include <karma_dmx.h>
#include <karma_dm.h>
#include <karma_r.h>
#include <karma_m.h>
#include <karma_a.h>


#define ALLOC_SIZE 64
#define USED_MAGIC_NUMBER 1342832171
#define FREE_MAGIC_NUMBER 1345345349

#define VERIFY_USED(entry) if (entry == NULL) \
{fprintf (stderr, "NULL entry passed\n"); \
 a_prog_bug (function_name); } \
if ((uaddr) entry % sizeof (void *) != 0) \
{fprintf (stderr, "Entry pointer not aligned properly\n"); \
 a_prog_bug (function_name); } \
if (entry->magic_number != USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid entry object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FREE(entry) if (entry == NULL) \
{fprintf (stderr, "NULL entry passed\n"); \
 a_prog_bug (function_name); } \
if ((uaddr) entry % sizeof (void *) != 0) \
{fprintf (stderr, "Entry pointer not aligned properly\n"); \
 a_prog_bug (function_name); } \
if (entry->magic_number != FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed entry object\n"); \
 a_prog_bug (function_name); }


/*  Structure and private data to manage descriptors  */
struct managed_descriptor_type
{
    unsigned int magic_number;
    int fd;
    void *info;
    flag (*input_func) ();
    void (*close_func) ();
    flag (*output_func) ();
    flag (*exception_func) ();
    XtInputId input_id;
    XtInputId output_id;
    XtInputId exception_id;
    flag auto_close;
    struct managed_descriptor_type *next;
    struct managed_descriptor_type *prev;
};
static KManagedDescriptor first_managed = NULL;
static KManagedDescriptor first_free = NULL;
static XtAppContext app_context = NULL;


/*  Private functions  */
STATIC_FUNCTION (KManagedDescriptor _dmx_manage_func,
		 (int fd, void *info, flag auto_close, va_list argp) );
STATIC_FUNCTION (void *_dmx_unmanage_func, (int fd, KManagedDescriptor md) );
STATIC_FUNCTION (void xt_input_handler,
		 (XtPointer client_data, int *fd, XtInputId *id) );
STATIC_FUNCTION (void xt_output_handler,
		 (XtPointer client_data, int *fd, XtInputId *id) );
STATIC_FUNCTION (void xt_exception_handler,
		 (XtPointer client_data, int *fd, XtInputId *id) );
STATIC_FUNCTION (flag _dmx_read_fd, (KManagedDescriptor entry) );
STATIC_FUNCTION (void _dmx_close_fd, (KManagedDescriptor entry) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
void dmx_register_app_context (XtAppContext context)
/*  [SUMMARY] Register Xt application context for event management.
    [PURPOSE] This routine will register the application context to use for
    descriptor management with the Xt toolkit. Once this is called the
    application may then call routines such as [<dm_manage>] and [<chm_manage>]
    <context> The application context.
    [RETURNS] Nothing.
*/
{
    app_context = context;
    dm_register_managers (_dmx_manage_func, _dmx_unmanage_func);
}   /*  End Function dmx_register_app_context  */


/*  Private functions follow  */

static KManagedDescriptor _dmx_manage_func (int fd, void *info,
					    flag auto_close, va_list argp)
/*  [SUMMARY] Manage a file descriptor for events.
    [PURPOSE] This routine will manage a file descriptor for activity by
    registering callback routines. Only one set of callbacks may be registered
    per file descriptor.
    <fd> The file descriptor to manage.
    <info> The initial value of the arbitrary pointer associated with the
    managed descriptor. This pointer may be modified by the callback routines.
    <auto_close> If TRUE, the descriptor is automatically closed upon a closure
    event.
    <argp> The optional list of parameter event-key event-function pairs.
    [RETURNS] A KManagedDescriptorEntry object on success, else NULL.
*/
{
    unsigned int count, att_key;
    XtInputId id;
    KManagedDescriptor entry;
    KManagedDescriptor new;
    void **ptr;
    static char function_name[] = "_dmx_manage_func";

    if (app_context == NULL)
    {
	fprintf (stderr, "Must register the Xt application context first\n");
	a_prog_bug (function_name);
    }
    /*  See if descriptor is already managed  */
    for (entry = first_managed; entry != NULL; entry = entry->next)
    {
	if (fd == entry->fd)
	{
	    fprintf (stderr, "Descriptor: %d is already managed\n", fd);
	    a_prog_bug (function_name);
	}
    }
    /*  Descriptor is not already managed  */
    /*  Get new entry  */
    if (first_free == NULL)
    {
	/*  Have to allocate some more for the free list  */
	if ( ( new = m_alloc (ALLOC_SIZE * sizeof *new) ) == NULL )
	{
	    m_error_notify (function_name, "new managed descriptor entries");
	    return (FALSE);
	}
	m_clear (new, ALLOC_SIZE * sizeof *new);
	/*  Add them to the free list  */
	first_free = new;
	for (count = 0; count < ALLOC_SIZE - 1; ++count)
	{
	    new[count].magic_number = FREE_MAGIC_NUMBER;
	    new[count].next = new + count + 1;
	}
	new[count].magic_number = FREE_MAGIC_NUMBER;
    }
    /*  Take one from the free list  */
    new = first_free;
    VERIFY_FREE (new);
    first_free = new->next;
    /*  Fill in entry  */
    new->magic_number = USED_MAGIC_NUMBER;
    new->next = NULL;
    new->fd = fd;
    new->info = info;
    while ( ( att_key = va_arg (argp, unsigned int) ) != KARMA_FD_EVENT_END )
    {
	switch (att_key)
	{
	  case KARMA_FD_EVENT_INPUT:
	    ptr = (void **) &new->input_func;
	    *ptr = va_arg (argp, void *);
	    break;
	  case KARMA_FD_EVENT_OUTPUT:
	    ptr = (void **) &new->output_func;
	    *ptr = va_arg (argp, void *);
	    break;
	  case KARMA_FD_EVENT_EXCEPTION:
	    ptr = (void **) &new->exception_func;
	    *ptr = va_arg (argp, void *);
	    break;
	  case KARMA_FD_EVENT_CLOSE:
	    ptr = (void **) &new->close_func;
	    *ptr = va_arg (argp, void *);
	    break;
	}
    }
    new->auto_close = auto_close;
    new->next = NULL;
    new->prev = NULL;
    /*  Tell the Xt event system  */
    if ( (new->input_func != NULL) || (new->close_func != NULL) )
    {
	id = XtAppAddInput (app_context, fd, (XtPointer) XtInputReadMask,
			    (XtInputCallbackProc) xt_input_handler,
			    (XtPointer) new);
	new->input_id = id;
    }
    if (new->output_func != NULL)
    {
	id = XtAppAddInput (app_context, fd, (XtPointer) XtInputWriteMask,
			    (XtInputCallbackProc) xt_output_handler,
			    (XtPointer) new);
	new->output_id = id;
    }
    if (new->exception_func != NULL)
    {
	id = XtAppAddInput (app_context, fd, (XtPointer) XtInputExceptMask,
			    (XtInputCallbackProc) xt_exception_handler,
			    (XtPointer) new);
	new->exception_id = id;
    }
    /*  Everything fine: add to list  */
    /*  Add to list  */
    if (first_managed != NULL) first_managed->prev = new;
    new->next = first_managed;
    first_managed = new;
    return (new);
}   /*  End Function _dmx_manage_func  */

static void *_dmx_unmanage_func (int fd, KManagedDescriptor md)
/*  [SUMMARY] Terminate the management of a file descriptor for activity.
    [NOTE] The routine will NOT close the descriptor (nor does it assume the
    descriptor is open).
    <fd> The descriptor to unmanage. This must always be valid.
    <md> The entry to unmanage. If this is NULL then <<fd>> is used.
    [RETURNS] The arbitrary pointer passed when the descriptor was managed.
*/
{
    void *info;
    static char function_name[] = "_dmx_unmanage_func";

    if (md == NULL)
    {
	for (md = first_managed; md != NULL; md = md->next)
	{
	    if (fd == md->fd) break;
	}
	if (md == NULL)
	{
	    /*  Descriptor not found  */
	    fprintf (stderr, "Descriptor: %d not managed\n", fd);
	    a_prog_bug (function_name);
	}
    }
    VERIFY_USED (md);
    /*  Unregister Xt event handlers  */
    if ( (md->input_func != NULL) || (md->close_func != NULL) )
    {
	XtRemoveInput (md->input_id);
    }
    if (md->output_func != NULL) XtRemoveInput (md->output_id);
    if (md->exception_func != NULL) XtRemoveInput (md->exception_id);
    /*  Remove entry from managed list  */
    if (md->prev == NULL) first_managed = md->next;
    else md->prev->next = md->next;
    if (md->next != NULL) md->next->prev = md->prev;
    info = md->info;
    /*  Clear and add to the free list  */
    m_clear (md, sizeof *md);
    md->magic_number = FREE_MAGIC_NUMBER;
    md->next = first_free;
    first_free = md;
    return (info);
}   /*  End Function _dmx_unmanage_func  */

static void xt_input_handler (XtPointer client_data, int *fd, XtInputId *id)
/*  This routine will handle input activity on a file descriptor.
    The Xt client data is given by  client_data  .
    The file descriptor is pointed to by  fd  .
    The XtInputId is pointed to by  id  .
    The routine returns nothing.
*/
{
    KManagedDescriptor entry = (KManagedDescriptor ) client_data;
    static char function_name[] = "_dmx_xt_input_handler";

    VERIFY_USED (entry);
    if (entry->input_id != *id)
    {
	fprintf (stderr, "IDs do not match!\n");
	a_prog_bug (function_name);
    }
    /*  Input/ closure occurred  */
    if ( !_dmx_read_fd (entry) )
    {
	/*  Close and unmanage  */
	_dmx_close_fd (entry);
    }
}   /*  End Function xt_input_handler  */

static void xt_output_handler (XtPointer client_data, int *fd, XtInputId *id)

/*  This routine will handle output activity on a file descriptor.
    The Xt client data is given by  client_data  .
    The file descriptor is pointed to by  fd  .
    The XtInputId is pointed to by  id  .
    The routine returns nothing.
*/
{
    KManagedDescriptor entry = (KManagedDescriptor ) client_data;
    static char function_name[] = "_dmx_xt_output_handler";

    VERIFY_USED (entry);
    if (entry->output_id != *id)
    {
	fprintf (stderr, "IDs do not match!\n");
	a_prog_bug (function_name);
    }
    /*  Output occurred  */
    if ( !(*entry->output_func) (entry->fd, &entry->info) )
    {
	/*  Descriptor to be unmanaged and closed  */
	_dmx_close_fd (entry);
    }
}   /*  End Function xt_output_handler  */

static void xt_exception_handler (XtPointer client_data, int *fd,XtInputId *id)

/*  This routine will handle exception activity on a file descriptor.
    The Xt client data is given by  client_data  .
    The file descriptor is pointed to by  fd  .
    The XtInputId is pointed to by  id  .
    The routine returns nothing.
*/
{
    KManagedDescriptor entry = (KManagedDescriptor ) client_data;
    static char function_name[] = "_dmx_xt_exception_handler";

    VERIFY_USED (entry);
    if (entry->exception_id != *id)
    {
	fprintf (stderr, "IDs do not match!\n");
	a_prog_bug (function_name);
    }
    /*  Exception occurred  */
    if ( !(*entry->exception_func) (entry->fd, &entry->info) )
    {
	/*  Descriptor to be unmanaged and closed  */
	_dmx_close_fd (entry);
    }
}   /*  End Function xt_exception_handler  */

static flag _dmx_read_fd (KManagedDescriptor entry)
/*  [PURPOSE] This routine will read a descriptor (if appropriate) until all
    data is drained from the descriptor.
    <entry> The descriptor entry.
    [RETURNS] TRUE if the descriptor is to remain managed, else FALSE.
*/
{
    int bytes_readable;
    char drain_buffer[1];
    static char function_name[] = "_dmx_read_fd";

    if (entry->close_func == NULL)
    {
	if (entry->input_func == NULL)
	{
	    fprintf (stderr,
		     "Input on asynchronous descriptor but no callback\n");
	    a_prog_bug (function_name);
	}
	return (*entry->input_func) (entry->fd, &entry->info);
    }
    /*  Can test for input/ closure  */
    if ( ( bytes_readable = r_get_bytes_readable (entry->fd) ) < 0 )
    {
	fprintf (stderr,"Error getting bytes readable on descriptor: %d\t%s\n",
		 entry->fd, ERRSTRING);
	a_prog_bug (function_name);
    }
    if (bytes_readable < 1)
    {
	/*  Closure occurred  */
	/*  Descriptor drainable  */
	/*  Drain data (close event)  */
	switch ( r_read (entry->fd, drain_buffer, 1) )
	{
	  case -1:
	    /*  Error  */
	    fprintf (stderr, "%s: error draining descriptor\t%s\n",
		     function_name, ERRSTRING);
	    break;
	  case 0:
	    /*  This is what we expect  */
	    break;
	  default:
	    fputs ("Data readable on closed descriptor\n", stderr);
#ifdef OS_IRIX
	    fputs ("There is a known bug in IRIX 6.2 that can cause this: contact SGI!\n",
		   stderr);
#endif
#ifdef OS_Solaris
	    fputs ("There is a known bug in Solaris 2.x that can cause this: contact Sun!\n",
		   stderr);
#endif
	    a_prog_bug (function_name);
	    break;
	}
	return (FALSE);
    }
    /*  Input/ closure event occurred  */
    if (entry->input_func == NULL)
    {
	fprintf (stderr, "Input on descriptor not being drained\n");
	a_prog_bug (function_name);
    }
    while (bytes_readable > 0)
    {
	/*  More input  */
	if ( !(*entry->input_func) (entry->fd, &entry->info) )
	{
	    /*  Descriptor to be unmanaged and closed  */
	    return (FALSE);
	}
	if ( ( bytes_readable = r_get_bytes_readable (entry->fd) ) < 0 )
	{
	    fprintf (stderr,
		     "Error getting bytes readable on descriptor: %d\t%s\n",
		     entry->fd, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
    return (TRUE);
}   /*  End Function _dmx_read_fd  */

static void _dmx_close_fd (KManagedDescriptor entry)
/*  [PURPOSE] This routine will call the registered <<close_func>> for a
    descriptor, and will then close and unmanage the descriptor.
    <entry> The descriptor entry.
    [RETURNS] Nothing.
*/
{
    /*static char function_name[] = "_dmx_close_fd";*/

    if (entry->close_func != NULL)
	(*entry->close_func) (entry->fd, entry->info);
    if (entry->auto_close) close (entry->fd);
    _dmx_unmanage_func (entry->fd, entry);
}   /*  End Function _dmx_close_fd  */
