/*LINTLIBRARY*/
/*  init.c

    This code provides miscellaneous routines for the X Intrinsics toolkit

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains miscellaneous routines needed for the using the Xt
  toolkit.


    Written by      Richard Gooch   1-DEC-1996

    Updated by      Richard Gooch   7-DEC-1996: Fixed bug in passing argc.
  Caused problems on alpha_OSF1.

    Updated by      Richard Gooch   18-DEC-1996: Added error message when
  display could not be opened.

    Updated by      Richard Gooch   27-DEC-1996: Added communications support.

    Updated by      Richard Gooch   30-DEC-1996: Added <<controlled>> return.
  Added processing of "-num_colours #" switch.

    Updated by      Richard Gooch   18-FEB-1997: Cope with PseudoColour visual
  not being root visual.

    Updated by      Richard Gooch   9-MAR-1997: Added verbose parameter.

    Updated by      Richard Gooch   24-MAR-1997: Send Glish "initialised" event
  if running under Glish.

    Updated by      Richard Gooch   15-APR-1997: Add handler for Glish
  "setXtResource" events.

    Updated by      Richard Gooch   19-JUN-1997: Removed call to test if Glish
  is available: <gls_initialise> does it all.

    Updated by      Richard Gooch   22-JUN-1997: Changed to double varargs list
  for <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   24-JUL-1997: Added 10 to minimum number of
  colours required so that enough room is likely to be left for named colours.

    Updated by      Richard Gooch   17-AUG-1997: Changed to explicit
  allocation of some named colours, rather than increasing minimum number of
  colours. Should be more robust.

    Updated by      Richard Gooch   7-NOV-1997: Fixed bug if application did
  not specify minimum number of colour cells.

    Updated by      Richard Gooch   14-FEB-1998: Forced application shell to be
  non-transient.

    Updated by      Richard Gooch   5-SEP-1998: Added BECOME_SERVER and TITLE
  attributes.

    Updated by      Richard Gooch   3-MAR-1999: Added automatic call to
  <im_register_module_name> if <module_name> is not yet defined.

    Updated by      Richard Gooch   4-MAR-1999: Added two calls to <ext_call>.

    Updated by      Richard Gooch   6-MAR-1999: Fixed bug in processing
  "-num_colours" option.

    Updated by      Richard Gooch   10-JUN-1999: Created experimental function
  <xtmisc_init_post_realise_arg_process>.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   13-SEP-1999: Added "kstopper" client
  protocol.

    Updated by      Richard Gooch   26-JAN-2000: Fixed bug in "kstopper"
  client read function <read_func>: was not reading data on interrupt.

    Last updated by Richard Gooch   23-MAR-2001: Created experimental function
  <xtmisc_init_get_pc_colourmap>.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/types.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
/*  Do not include  karma.h  first, since CONST may be screwed otherwise. This
    applies to broken versions of  X11/Intrinsic.h  */
#ifndef X11
#  define X11
#endif
#include <karma_xtmisc.h>
#include <karma_event.h>
#include <karma_conn.h>
#include <karma_chm.h>
#include <karma_dmx.h>
#include <karma_gls.h>
#include <karma_ext.h>
#include <karma_xv.h>
#include <karma_im.h>
#include <karma_ch.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_r.h>
#include <karma_s.h>


#define EXTRA_CCELLS 1  /*  MUST be at least 1, else bug if application does
			    not specify minimum number of cells  */


/*  Private functions  */
STATIC_FUNCTION (Boolean work_proc, (XtPointer client_data) );
STATIC_FUNCTION (flag resource_func,
		 (void *object, CONST char *name, CONST char *data,
		  unsigned int type, unsigned int num_elements,
		  unsigned int num_axes, CONST int *axis_lengths) );
STATIC_FUNCTION (flag allocate_named_colours, (Display *dpy, Colormap cmap) );
STATIC_FUNCTION (flag open_func, (Connection connection, void **info) );
STATIC_FUNCTION (flag read_func,
		 (Connection connection, void **info, Channel channel) );
STATIC_FUNCTION (void close_func, (Connection connection, void *info) );
STATIC_FUNCTION (void timer_proc, (XtPointer client_data, XtIntervalId *id) );


/*  Private data  */
static XtAppContext app_ctx = NULL;
static XtIntervalId xt_interval_id = 0;


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
Widget xtmisc_init_app_initialise (XtAppContext *app_context_return,
				   CONST char* application_class,
				   XrmOptionDescList options,
				   Cardinal num_options,
				   int *argc_in_out, String *argv_in_out,
				   String *fallback_resources,
				   ...)
/*  [SUMMARY] Initialise an application for Xt.
    [PURPOSE] This routine will initialise an application to use Xt. This
    routine is similar to the <<XtVaAppInitialize>> routine, except that a new
    colourmap for the application may be created.
    <app_context_return> As for <<XtVaAppInitialize>>.
    <application_class> As for <<XtVaAppInitialize>>.
    <options> As for <<XtVaAppInitialize>>.
    <num_options> As for <<XtVaAppInitialize>>.
    <argc_in_out> As for <<XtVaAppInitialize>>.
    <argv_in_out> As for <<XtVaAppInitialize>>.
    <fallback_resources> As for <<XtVaAppInitialize>>.
    [VARARGS] Two separate lists of optional pairs of attribute-key
    attribute-value. The first list is processed by this routine and must be
    terminated with XTMISC_INIT_ATT_END. The second list is processed by
    <<XtVaAppInitialize>> and must be terminated by NULL.
    [RETURNS] The application shell.
*/
{
    va_list argp;
    flag under_control = FALSE;
    flag verbose_local = FALSE;
    flag user_specified_min_ccells = FALSE;
    flag comms_setup = FALSE;
    flag become_server = FALSE;
    flag am_serv = FALSE;
    flag set_module_name = FALSE;
    unsigned int att_key, count;
    unsigned int min_ccells = 0;
    unsigned int server_port_number = 0;  /*  Initialised to keep GCC happy  */
    Colormap xcmap;
    Widget w;
    String attr;
#if XT_REVISION > 5
    XtTypedArgList typed_args;
#else
    ArgList args;
#endif
    struct sigaction new_action;
    flag *controlled = NULL;
    flag *verbose = NULL;
    char *title = NULL;
    Display *dpy;
    Screen *screen;
    char hostname[STRING_LENGTH];
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];
    static char function_name[] = "xtmisc_init_app_initialise";

    va_start (argp, fallback_resources);
    if (strcmp (module_name, "<<Unknown>>") == 0) set_module_name = TRUE;
    /*  First process my varargs  */
    while ( ( att_key = va_arg (argp, unsigned int) ) != XTMISC_INIT_ATT_END )
    {
	switch (att_key)
	{
	  case XTMISC_INIT_ATT_MIN_CCELLS:
	    min_ccells = va_arg (argp, unsigned int);
	    break;
	  case XTMISC_INIT_ATT_COMMS_SETUP:
	    comms_setup = va_arg (argp, flag);
	    break;
	  case XTMISC_INIT_ATT_CONTROLLED:
	    controlled = va_arg (argp, flag *);
	    break;
	  case XTMISC_INIT_ATT_VERBOSE:
	    verbose = va_arg (argp, flag *);
	    break;
	  case XTMISC_INIT_ATT_BECOME_SERVER:
	    become_server = va_arg (argp, flag);
	    if (become_server) comms_setup = TRUE;
	    break;
	  case XTMISC_INIT_ATT_TITLE:
	    title = va_arg (argp, char *);
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    /*  Next search for special options or for tapped options  */
    for (count = 1; count < *argc_in_out; ++count)
    {
	if (strcmp (argv_in_out[count], "-num_colours") == 0)
	{
	    if (++count >= *argc_in_out) continue;
	    min_ccells = atoi (argv_in_out[count]);
	    user_specified_min_ccells = TRUE;
	}
	else if (strcmp (argv_in_out[count], "-verbose") == 0)
	{
	    verbose_local = TRUE;
	}
	else if (strcmp (argv_in_out[count], "-cmap_master") == 0)
	{
	    if (!user_specified_min_ccells) min_ccells = 2;
	}
	else if (strcmp (argv_in_out[count], "-name") == 0)
	{
	    if (!set_module_name) continue;
	    if (++count >= *argc_in_out) continue;
	    im_register_module_name (argv_in_out[count]);
	    set_module_name = FALSE;
	}
    }
    if (verbose != NULL) *verbose = verbose_local;
    if (set_module_name)
    {
	char *ptr;

	if ( ( ptr = strrchr (argv_in_out[0], '/') ) == NULL )
	    ptr = argv_in_out[0];
	else ++ptr;
	im_register_module_name (ptr);
    }
    if (strcmp (module_name, "kstopper") != 0)
    {
	/*  Set up control_c handler  */
	sigemptyset (&new_action.sa_mask);
	new_action.sa_flags = 0;
	new_action.sa_handler = s_int_handler;
	if (sigaction (SIGINT, &new_action, NULL) != 0)
	{
	    fprintf (stderr, "%s: error setting control_c handler\t%s\n",
		     function_name, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
    XtToolkitInitialize ();
    app_ctx = XtCreateApplicationContext ();
    if (app_context_return != NULL) *app_context_return = app_ctx;
    ext_call (module_name, "Xt_AppInitialise", 0, NULL, app_ctx, NULL, NULL,
	      NULL, NULL, NULL);
    XtAppSetFallbackResources (app_ctx, fallback_resources);
    dpy = XtOpenDisplay (app_ctx, NULL, NULL, (String) application_class,
			 options, num_options, argc_in_out, argv_in_out);
    if (dpy == NULL)
    {
	fprintf (stderr,
		 "Can't open display. Please check your DISPLAY environment variable.\n");
	exit (RV_UNDEF_ERROR);
    }
    screen = DefaultScreenOfDisplay (dpy);
    xcmap = xtmisc_init_get_pc_colourmap (screen, min_ccells + EXTRA_CCELLS,
					  verbose_local);
#if XT_REVISION > 5
    if ( ( typed_args = (XtTypedArgList) malloc (sizeof *typed_args) )
	 == NULL )
    {
	m_abort (function_name, "initial ArgList");
    }
    for ( attr = va_arg (argp, String), count = 0; attr != NULL;
	  attr = va_arg (argp, String) )
    {
        if (strcmp (attr, XtVaTypedArg) == 0)
	{
            typed_args[count].name = va_arg (argp, String);
            typed_args[count].type = va_arg (argp, String);
            typed_args[count].value = va_arg (argp, XtArgVal);
            typed_args[count].size = va_arg (argp, int);
        }
	else
	{
	    typed_args[count].name = attr;
	    typed_args[count].type = NULL;
	    typed_args[count].value = va_arg (argp, XtArgVal);
	    typed_args[count].size = 0;
        }
	count++;
	if ( ( typed_args = (XtTypedArgList) 
	       realloc ( (char *) typed_args, 
			 (count + 1) * sizeof *typed_args ) ) == NULL )
	{
	    m_abort (function_name, "ArgList");
	}
    }
    typed_args[count].name = NULL;
    w = XtVaAppCreateShell (NULL, application_class, 
			    applicationShellWidgetClass, dpy,
			    XtNscreen, (XtArgVal) screen,
			    XtNcolormap, xcmap,
			    XtNtransient, False,
			    XtNargc, (XtArgVal) *argc_in_out,
			    XtNargv, (XtArgVal) argv_in_out,
			    XtVaNestedList, (XtVarArgsList) typed_args,
			    NULL);
    free ( (XtPointer) typed_args );
#else
    if ( ( args = (ArgList) malloc (5 * sizeof *args) ) == NULL )
    {
	m_abort (function_name, "initial ArgList");
    }
    XtSetArg (args[0], XtNscreen, (XtArgVal) screen);
    XtSetArg (args[1], XtNcolormap, xcmap);
    XtSetArg (args[2], XtNtransient, False);
    XtSetArg (args[3], XtNargc, (XtArgVal) *argc_in_out);
    XtSetArg (args[4], XtNargv, (XtArgVal) argv_in_out);
    count = 5;
    for ( attr = va_arg (argp, String); attr != NULL;
	  attr = va_arg (argp, String) )
    {
	XtSetArg ( args[count], attr, va_arg (argp, XtArgVal) );
	count++;
	if ( ( args = (ArgList) realloc ( (char *) args, 
					  (count + 1) * sizeof *args ) )
	     == NULL )
	{
	    m_abort (function_name, "ArgList");
	}
    }
    w = XtAppCreateShell (NULL, application_class, 
			  applicationShellWidgetClass, dpy, args, count);
    free ( (XtPointer) args );
#endif
    va_end (argp);
    if (comms_setup)
    {
	/*  Initialise communications  */
	dmx_register_app_context (*app_context_return);
	conn_initialise ( ( void (*) () ) NULL );
	if (strcmp (module_name, "kstopper") != 0)
	{
	    conn_register_client_protocol ("kstopper", 0, 1,
					   ( flag (*) () ) NULL,
					   open_func, read_func, close_func);
	    conn_attempt_connection ("localhost",
				     r_get_def_port ( "kstopper",
						      r_getenv ("DISPLAY") ) ,
				     "kstopper");
	}
	event_initialise ();
	if ( conn_controlled_by_cm_tool () ) under_control = TRUE;
	else
	{
	    conn_register_client_protocol ("hog_request", 0, 0,
					   ( flag (*) () ) NULL,
					   ( flag (*) () ) NULL,
					   ( flag (*) () ) NULL,
					   ( void (*) () ) NULL);
	}
	if ( gls_initialise (argc_in_out, argv_in_out) )
	{
	    under_control = TRUE;
	    gls_register_plain_event_handler (resource_func, "setXtResource",
					      w);
	}
    }
    if (controlled != NULL) *controlled = under_control;
    if (become_server)
    {
	int def_port_number;

	/*  Get default port number  */
	if ( ( def_port_number =
	       r_get_def_port ( module_name,
				DisplayString (XtDisplay (w) ) ) ) < 0 )
	{
	    fprintf (stderr, "Could not get default port number\n");
	    exit (RV_UNDEF_ERROR);
	}
	server_port_number = def_port_number;
	am_serv = conn_become_server (&server_port_number, CONN_MAX_INSTANCES);
	if (am_serv) fprintf (stderr, "Port allocated: %u\n",
			      server_port_number);
	else fprintf (stderr, "Module not operating as Karma server\n");
    }
    if (title != NULL)
    {
	r_gethostname (hostname, STRING_LENGTH);
	if (am_serv)
	    sprintf (title, "%s v%s/Karma v%s @%s:%u",
		     module_name, module_version_date, karma_library_version,
		     hostname, server_port_number);
	else
	    sprintf (title, "%s v%s/Karma v%s @%s",
		     module_name, module_version_date, karma_library_version,
		     hostname);
	XtVaSetValues (w, XtNtitle, title, NULL);
    }
    XtAppAddWorkProc (*app_context_return, work_proc, NULL);
    return (w);
}   /*  End Function xtmisc_init_app_initialise  */

/*EXPERIMENTAL_FUNCTION*/
void xtmisc_init_post_realise_arg_process (int *argc_in_out,
					   String *argv_in_out)
/*  [SUMMARY] Process and remove arguments after main widget realisation.
    <argc_in_out> As for <<XtVaAppInitialize>>.
    <argv_in_out> As for <<XtVaAppInitialize>>.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    for (count = 1; count < *argc_in_out; ++count)
    {
	int args_to_remove, count1;

	if ( (strcmp (argv_in_out[count], "-connect") == 0) &&
	     (count + 3 < *argc_in_out) )
	{
	    int port;
	    char *host, *protocol;

	    host = argv_in_out[count + 1];
	    if ( isdigit (argv_in_out[count + 2][0]) )
		port = atoi (argv_in_out[count + 2]);
	    else
	    {
		port = r_get_def_port ( argv_in_out[count + 2],
					r_getenv ("DISPLAY") );
	    }
	    protocol = argv_in_out[count + 3];
	    if ( !conn_attempt_connection (host, port, protocol) )
	    {
		fprintf (stderr, "Error adding \"%s\" connection\n", protocol);
	    }
	    args_to_remove = 4;
	}
	else continue;
	/*  Strip these arguments  */
	for (count1 = count + args_to_remove; count1 < *argc_in_out; ++count1)
	{
	    argv_in_out[count1 - args_to_remove] = argv_in_out[count1];
	}
	*argc_in_out -= args_to_remove;
    }
}   /*  End Function xtmisc_init_post_realise_arg_process  */

/*EXPERIMENTAL_FUNCTION*/
Colormap xtmisc_init_get_pc_colourmap (Screen *screen, unsigned int min_ccells,
				       flag verbose)
/*  [SUMMARY] Get PseudoColour colourmap with a minimum of free colourcells.
    <screen> The screen.
    <min_ccells> The minimum number of free colourcells.
    <verbose> If TRUE, verbose messages are given.
    [RETURNS] The colourmap.
*/
{
    unsigned long dummy;
    Colormap xcmap;
    unsigned long *pixel_values;
    Display *dpy;
    Visual *pc_visual, *root_visual;
    XVisualInfo *pc_vinfo, *root_vinfo;
    static char function_name[] = "xtmisc_init_get_pc_colourmap";

    dpy = XDisplayOfScreen (screen);
    xcmap = DefaultColormapOfScreen (screen);
    root_visual = XDefaultVisualOfScreen (screen);
    xv_get_visuals (screen, &pc_visual, NULL, NULL);
    if (pc_visual != NULL)
    {
	root_vinfo = xv_get_visinfo_for_visual (dpy, root_visual);
	pc_vinfo = xv_get_visinfo_for_visual (dpy, pc_visual);
	/*  If the PseudoColour visual is not the default, leave colourmap
	    creation to the toolkit  */
	if (root_vinfo->visualid != pc_vinfo->visualid) pc_visual = NULL;
	XFree ( (char *) root_vinfo );
	XFree ( (char *) pc_vinfo );
    }
    if (pc_visual != NULL)
    {
	/*  Try to allocate colours  */
	allocate_named_colours (dpy, xcmap);
	if ( ( pixel_values = m_alloc (min_ccells * sizeof *pixel_values) )
	     == NULL )
	{
	    m_abort (function_name, "pixel values");
	}
	if (XAllocColorCells (dpy, xcmap, False, &dummy, 0, pixel_values,
			      min_ccells) == 0)
	{
	    /*  Not enough colourcells left: create new colourmap  */
	    if (verbose)
		fprintf (stderr, "%s: creating colourmap for application\n",
			 function_name);
	    if ( ( xcmap = XCreateColormap (dpy, XRootWindowOfScreen (screen),
					    pc_visual, AllocNone) )
		 == (Colormap) NULL )
	    {
		fprintf (stderr, "Could not create colourmap\n");
		exit (RV_UNDEF_ERROR);
	    }
	    XSync (dpy, False);
	    allocate_named_colours (dpy, xcmap);
	}
	else
	{
	    /*  Was able to allocate colourcells: free them now  */
	    XFreeColors (dpy, xcmap, pixel_values, min_ccells, 0);
	}
	m_free (pixel_values);
    }
    return (xcmap);
}   /*  End Function xtmisc_init_get_pc_colourmap  */


/*  Private functions follow  */

static Boolean work_proc (XtPointer client_data)
/*  [SUMMARY] Xt work function.
    <client_data> The client data.
    [RETURNS] True if the procedure should not be called again, else False.
*/
{
    int dummy = 0;
    extern char module_name[STRING_LENGTH + 1];

    ext_call (module_name, "Xt_InEventLoop", 0, NULL, NULL, NULL, NULL, NULL,
	      NULL, NULL);
    if ( !gls_test_available () ) return (True);
    gls_send_plain_event ("initialised", (CONST char *) &dummy,
			  K_INT, 0, 0, (CONST int *) NULL);
    return (True);
}   /*  End Function work_proc  */

static flag resource_func (void *object, CONST char *name, CONST char *data,
			   unsigned int type, unsigned int num_elements,
			   unsigned int num_axes, CONST int *axis_lengths)
/*  [SUMMARY] Process a plain Glish event.
    <object> The arbitrary object.
    <name> The name of the Glish event.
    <data> The pointer to the data.
    <type> The type of the data.
    <num_elements> The number of elements.
    <num_axes> The number of axes.
    <axis_lengths> The lengths of the axes.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that the
    event is still to be processed.
*/
{
    char *wname, *resource, *value;
    char string[STRING_LENGTH];
    Widget w;
    Widget main_shell = (Widget) object;

    if (type != K_VSTRING) return (FALSE);
    if (num_elements != 1) return (FALSE);
    strcpy (string, *(CONST char **) data);
    wname = string;
    if ( ( value = strchr (string, ':') ) == NULL )
    {
	fprintf (stderr, "No value found\n");
	return (TRUE);
    }
    *value++ = '\0';
    while ( isspace (*value) ) ++value;
    resource = strrchr (string, '.');
    *resource++ = '\0';
    w = XtNameToWidget (main_shell, wname);
    if (w == NULL)
    {
	fprintf (stderr, "Widget: \"%s\" not found\n", wname);
	return (TRUE);
    }
#ifdef DEBUG
    fprintf (stderr, "\"%s\" widget: %p  resource: \"%s\"\n",
	     wname, w, resource);
#endif
    XtVaSetValues (w,
		   "callCallbacksOnSet", TRUE,
		   XtVaTypedArg, resource, XtRString, value, strlen (value) +1,
		   NULL);
    return (TRUE);
}   /*  End Function resource_func  */

static flag allocate_named_colours (Display *dpy, Colormap cmap)
/*  [SUMMARY] Allocate some common named colours.
    <dpy> The X11 display handle.
    <cmap> The X11 colourmap.
    [RETURNS] TRUE if all the colours could be allocated, else FALSE.
*/
{
    XColor dum1, dum2;
    char **colour;
    static char *required_colours[] =
    {
	"black",
	"white",
	"red",
	"green",
	"blue",
	"turquoise",
	"yellow4",
	NULL
    };

    for (colour = required_colours; *colour != NULL; ++colour)
    {
	if ( XAllocNamedColor (dpy, cmap, *colour, &dum1, &dum2) ) continue;
	/*  Failed: don't bother to unallocate: everybody would want these
	    colours anyway :-)  */
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function allocate_named_colours  */

static flag open_func (Connection connection, void **info)
/*  [SUMMARY] Connection open event callback.
    [PURPOSE] This routine is called when a connection opens.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    [RETURNS] TRUE on successful registration, else FALSE (indicating the
    connection should be closed).
    [NOTE] The <<close_func>> will not be called if this routine returns
    FALSE.
*/
{
    Channel channel;
    pid_t pid;
    static char function_name[] = "_xtmisc_init_open_func";

    channel = conn_get_channel (connection);
    pid = getpid ();
    if ( (ch_write (channel, (char *) &pid, sizeof pid) < sizeof pid) ||
	 !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error writing PID\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function open_func  */

static flag read_func (Connection connection, void **info, Channel channel)
/*  [SUMMARY] Connection read event callback.
    [PURPOSE] This routine is called when data is ready to be read from a
    connection.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    <channel> The channel object on which the read event occurred. This may be
    the normal or datagram channel.
    [RETURNS] TRUE on successful reading, else FALSE (indicating the connection
    should be closed).
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
    unsigned long linterval, bytes_read, pos;
    unsigned long *interval = *info;
    static char function_name[] = "_xtmisc_init_read_func";

    pos = 0;
    while ( ( bytes_read = ch_read (channel, pos + (char *) &linterval,
				    sizeof linterval - pos) )
	    < sizeof linterval - pos )
    {
	if (errno != EINTR)
	{
	    fprintf (stderr, "%s: error reading interval\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	pos += bytes_read;
    }
    if (interval == NULL)
    {
	if ( ( interval = m_alloc (sizeof *interval) ) == NULL )
	{
	    m_error_notify (function_name, "interval");
	    return (FALSE);
	}
	*interval = linterval;
	*info = interval;
	xt_interval_id = XtAppAddTimeOut (app_ctx, *interval, timer_proc,
					  interval);
    }
    else
    {
	*interval = linterval;
	XtRemoveTimeOut (xt_interval_id);
	xt_interval_id = XtAppAddTimeOut (app_ctx, *interval, timer_proc,
					  interval);
    }
    return (TRUE);
}   /*  End Function read_func  */

static void close_func (Connection connection, void *info)
/*  [SUMMARY] Connection close event callback.
    [PURPOSE] This routine is called when a connection closed.
    When this routine is called, this is the last chance to read any
    buffered data from the channel associated with the connection object.
    <connection> The connection object.
    <info> The arbitrary connection information pointer.
    [RETURNS] Nothing.
*/
{
    if (info != NULL) m_free (info);
}   /*  End Function close_func  */

static void timer_proc (XtPointer client_data, XtIntervalId *id)
/*  [SUMMARY] Xt interval procedure.
*/
{
    Connection connection;
    Channel channel;
    char dummy = 0;
    unsigned long *interval = client_data;

    connection = conn_get_client_connection ("kstopper", 0);
    if (connection == NULL) return;
    xt_interval_id = XtAppAddTimeOut (app_ctx, *interval, timer_proc,
				      client_data);
    channel = conn_get_channel (connection);
    ch_write (channel, &dummy, 1);
    ch_flush (channel);
}   /*  End Function timer_proc  */
