/*LINTLIBRARY*/
/*  dressing.c

    This code provides dressing support for KWorldCanvas objects.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for supporting a dressing on a
    World Canvas (window with world co-ordinates).


    Written by      Richard Gooch   12-FEB-1997

    Updated by      Richard Gooch   17-FEB-1997: First implementation of RA/DEC
  scales.

    Updated by      Richard Gooch   21-FEB-1997: First implementation of BEAM,
  Frequency and Velocity axes. First implementation of grid.

    Updated by      Richard Gooch   23-FEB-1997: Improved implementation of
  grid. Support rotated tick marks. Simple implementation for "ANGLE".

    Updated by      Richard Gooch   24-FEB-1997: First implementation of "TIME"
  axis.

    Updated by      Richard Gooch   25-FEB-1997: Minor formatting changes for
  "RA", "DEC" and "TIME" axes: took out spaces in scale values.

    Updated by      Richard Gooch   27-FEB-1997: Moved Equinox into RA,DEC
  labels. Fixed bug in <draw_scale>: did not cope with wrapped RA axis.

    Updated by      Richard Gooch   1-MAR-1997: Improved "ANGLE" support.

    Updated by      Richard Gooch   2-MAR-1997: Added "M/S" support. Added 5
  pixel offset at right.

    Updated by      Richard Gooch   25-MAR-1997: Added GLON,GLAT support.

    Updated by      Richard Gooch   15-APR-1997: Fixed drawing labels only.

    Updated by      Richard Gooch   5-AUG-1997: Auto scaling for Jy/mJy.

    Updated by      Richard Gooch   20-AUG-1997: Made use of new functions for
  testing if axes are longitude or latitude.

    Updated by      Richard Gooch   21-NOV-1997: Fixed bug in <draw_scale>
  which prevented showing minutes&seconds if showing an hour and
  <<show_minutes>> was FALSE and <<show_seconds>> was TRUE.

    Updated by      Richard Gooch   12-FEB-1998: Took account of new rotation
  definition.

    Updated by      Richard Gooch   14-APR-1998: Took account of new <wcs>
  functions to breakdown values.

    Updated by      Richard Gooch   23-APR-1998: Bug fix in dms display.

    Updated by      Richard Gooch   14-JUL-1998: Display Galactic co-ordinates
  in decimal notation.

    Updated by      Richard Gooch   23-JUL-1998: Cope with negative RA values.

    Updated by      Richard Gooch   16-APR-1999: Changed to prefer canvas axis
  name rather than dressing label when drawing scale.

    Updated by      Richard Gooch   22-AUG-1999: Reduced space wastage by
  computing and using size of strings.

    Updated by      Richard Gooch   26-AUG-1999: Split line gap value into
  axis line gap and restriction line gap values.

    Updated by      Richard Gooch   11-SEP-1999: Gracefully handle failure of
  <kwin_hershey_draw_string> when computing title size.

    Updated by      Richard Gooch   26-APR-2000: Removed large if-elseif-else
  clauses for computing interval steps and use log10-based algorithm.

    Updated by      Richard Gooch   27-APR-2000: Added if-elseif-else clauses
  for special latitude intervals.

    Updated by      Richard Gooch   5-MAY-2000: Improved algorithm for
  selecting tick boundaries for <show_linear_scale> and added minor tick marks.

    Updated by      Richard Gooch   6-MAY-2000: Supported external tick marks.

    Updated by      Richard Gooch   9-MAY-2000: Improved tick selection
  algorithm for angle and time axes. Created <get_time_step>, <get_angle_step>,
  <get_ms_step> and <get_decimal_step>. Removed <get_real_step>.

    Updated by      Richard Gooch   8-JUN-2000: Fixed bugs in computing
  precision: "0.25" only had one digit of precision.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Updated by      Richard Gooch   23-JUN-2000: Created <show_generic_scale>.

    Updated by      Richard Gooch   26-JUN-2000: Converted remaining units to
  use new <show_generic_scale> function.

    Updated by      Richard Gooch   27-JUN-2000: Created unpublished function
  <canvas_free_dressing>.

    Updated by      Richard Gooch   6-JUL-2000: Cosmetic changes.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>

    Updated by      Richard Gooch   14-JUL-2000: Removed unused variable.

    Updated by      Richard Gooch   10-OCT-2000: Added primitive wrap-around
  detection code.

    Updated by      Richard Gooch   2-MAR-2001: Added experimental code to
  <draw_scale> for generic linear scales so that exponential notation may be
  used.

    Updated by      Richard Gooch   9-MAR-2001: Fixed <linear_format> so that
  tiny offset is also modified by multiplier.

    Updated by      Richard Gooch   21-MAR-2001: Treat "HZ" unit like "FREQ".

    Updated by      Richard Gooch   20-AUG-2001: Trap canvas < 2 pixels in
  <canvas_get_dressing_size>.

    Last updated by Richard Gooch   29-AUG-2001: Ignore zero-range axis in
  <draw_scale> so that degenerate dimensions can be supported.


*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>


#define BORDER_LINEWIDTH    1.9
#define MAJOR_TICK_PIXELS   10
#define MINOR_TICK_PIXELS   4
#define MAX_MAJOR_TICKS     11
#define MAX_MINOR_TICKS     55
#define SCALE_FONT_SIZE     12.0
#define LABEL_FONT_SIZE     14.0
#define AXIS_LINE_GAP       12.0
#define RESTR_LINE_GAP      4.0
#define TITLE_LOWER_GAP     2.0
#define DEFAULT_DRESSING_COLOUR "green"
/*
#define MAX_LINEAR_SPACINGS (unsigned int) 3
#define LIN_SCALE_CUTOFF1         (double) 4.0
#define LIN_SCALE_CUTOFF2         (double) -2.0
#define ABS_SCALE_SPACING_PIXELS  (unsigned int) 40
#define TICK_BUF_SIZE             256
*/
#define GRID_STEPS                20
#define BINARY_ITERATIONS         10
#define ANGLE_LIMIT               3600.0  /*  Degrees (10 rotations)  */

#define HORIZONTAL_SCALE_LINES  1
#define VERTICAL_SCALE_CHARS    12


struct label_struct
{
    double width;
    double height;
    double ascent;
    double descent;
    char string[STRING_LENGTH];
};

struct scale_struct
{
    double size;
};

struct dressing_info_struct
{
    struct label_struct title;
    struct label_struct hlabel;
    struct label_struct vlabel;
    struct scale_struct hscale;
    struct scale_struct vscale;
    struct label_struct *top_left;
    unsigned int num_top_left;
    unsigned int max_top_left;
};

/*  Dressing structure: WARNING!!! Also change in main.c  */
typedef struct
{
    flag drawn;
    flag display;
    flag show_top_ticks;
    flag show_bottom_ticks;
    flag show_left_ticks;
    flag show_right_ticks;
    flag internal_ticks;
    flag grid;
    char *grid_colour;
    char *scale_colour;
    char *hlabel;
    char *vlabel;
    char *label_colour;
    char *title;
    char *title_colour;
    flag show_labels;
    double font_scale;
    struct dressing_info_struct *info;
} dressing_type;

typedef struct
{
    double x, y;
} coord_type;

typedef struct
{
    coord_type bottom_left;
    coord_type bottom_right;
    coord_type top_left;
    coord_type top_right;
} corners_type;

typedef struct
{
    double multiplier;
    double step;
    CONST char *format_string;
} linear_format_info_type;

typedef struct
{
    flag first_time;
    int num_hour_trans;
    int num_min_trans;
    int precision;
    double ra_step_hours, ra_step_minutes, ra_step_seconds;
} hms_format_info_type;

typedef struct
{
    flag first_time;
    int num_deg_trans;
    int num_min_trans;
    int precision;
    double dec_step_degrees, dec_step_minutes, dec_step_seconds;
} dms_format_info_type;

typedef struct
{
    flag first_time;
    int num_day_trans;
    int num_hour_trans;
    int num_min_trans;
    int precision;
    double time_step_days, time_step_hours, time_step_minutes;
    double time_step_seconds;
} time_format_info_type;


/*  Private functions  */
STATIC_FUNCTION (void setup_labels,
		 (KWorldCanvas canvas, dressing_type *dressing) );
STATIC_FUNCTION (void setup_top_left,
		 (KWorldCanvas canvas, dressing_type *dressing) );
STATIC_FUNCTION (void draw_top_left,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  dressing_type *dressing, int x_pixels, int y_pixels,
		  int x_offset, int y_offset) );
STATIC_FUNCTION (void draw_title,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  dressing_type *dressing, int x_pixels, int y_pixels,
		  int x_offset, int y_offset) );
STATIC_FUNCTION (void draw_labels,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  dressing_type *dressing, int x_pixels, int y_pixels,
		  int x_offset, int y_offset) );
STATIC_FUNCTION (void format_axis_name,
		 (char axis_string[STRING_LENGTH],
		  CONST char *axis_name, KwcsAstro ap) );
STATIC_FUNCTION (void draw_hscale,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  dressing_type *dressing, int x_pixels, int y_pixels,
		  int x_offset, int y_offset) );
STATIC_FUNCTION (void draw_vscale,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  dressing_type *dressing, int x_pixels, int y_pixels,
		  int x_offset, int y_offset) );
STATIC_FUNCTION (void draw_scale,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  dressing_type *dressing,
		  int x_pixels, int y_pixels,int x_offset, int y_offset,
		  CONST char *name, flag vertical, double *max_size) );
STATIC_FUNCTION (void _canvas_dressing_show_generic_scale,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  dressing_type *dressing, CONST corners_type *wcorners,
		  flag vertical, double wrap_modulus, double *max_size,
		  int x_pixels, int y_pixels, int x_offset, int y_offset,
		  int canvas_width, int canvas_height,
		  double left_lx, double right_lx,
		  double bottom_ly, double top_ly,
		  unsigned long scale_pixel, unsigned long grid_pixel,
		  double step, unsigned int num_minor,
		  double (*rescale) (double value, void *info),
		  void (*format) (char *string, double value, void *info),
		  void *info) );
STATIC_FUNCTION (void linear_format, (char *string, double value,void *info) );
STATIC_FUNCTION (double _canvas_dressing_hms_rescale,
		 (double value, void *info) );
STATIC_FUNCTION (void hms_format, (char *string, double value, void *info) );
STATIC_FUNCTION (void dms_format, (char *string, double value, void *info) );
STATIC_FUNCTION (void time_format, (char *string, double value, void *info) );
STATIC_FUNCTION (void draw_x_marker,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  CONST dressing_type *dressing, CONST corners_type *wcorners,
		  int x_pixels, int y_pixels, int x_offset, int y_offset,
		  double left_lx, double right_lx,
		  double bottom_ly, double top_ly,
		  unsigned long pixel_value, double wx, flag is_major,
		  double wrap_modulus, double *bottom_x, double *top_x) );
STATIC_FUNCTION (void draw_y_marker,
		 (KWorldCanvas canvas, KPixCanvas pixcanvas,
		  CONST dressing_type *dressing, CONST corners_type *wcorners,
		  int x_pixels, int y_pixels, int x_offset, int y_offset,
		  double left_lx, double right_lx,
		  double bottom_ly, double top_ly,
		  unsigned long pixel_value, double wy, flag is_major,
		  double wrap_modulus, double *left_y, double *right_y) );
STATIC_FUNCTION (double get_time_step, (double range) );
STATIC_FUNCTION (double get_angle_step, (double range) );
STATIC_FUNCTION (double get_ms_step, (double range) );
STATIC_FUNCTION (double get_decimal_step, (double range) );
STATIC_FUNCTION (flag test_between_limits,
		 (double val, double limit1, double limit2) );


/*  Public functions follow  */

/*UNPUBLISHED_FUNCTION*/
void canvas_get_dressing_size (KWorldCanvas canvas, int *p_left, int *p_right,
			       int *p_top, int *p_bottom,
			       dressing_type *dressing)
/*  [SUMMARY] This routine will determine the size of the dressing.
    <canvas> The world canvas object.
    <p_left> The number of horizontal pixels required to the left will be
    written here.
    <p_right> The number of horizontal pixels required to the right will be
    written here.
    <p_top> The number of vertical pixels required at the top will be written
    here.
    <p_bottom> The number of vertical pixels required at the bottom will be
    written here.
    <dressing> The 
    [RETURNS] Nothing.
*/
{
    int x_pixels, y_pixels, x_offset, y_offset;
    char *xlabel, *ylabel;
    struct dressing_info_struct *info;
    static char function_name[] = "canvas_get_dressing_size";

    if (!dressing->display)
    {
	*p_left = 0;
	*p_right = 0;
	*p_top = 0;
	*p_bottom = 0;
	return;
    }
    if (dressing->info == NULL)
    {
	if ( ( info = m_alloc (sizeof *info) ) == NULL )
	{
	    m_abort (function_name, "info structure");
	}
	m_clear (info, sizeof *info);
	dressing->info = info;
    }
    else info = dressing->info;
    canvas_get_attributes (canvas,
			   CANVAS_ATT_X_PIXELS, &x_pixels,
			   CANVAS_ATT_Y_PIXELS, &y_pixels,
			   CANVAS_ATT_X_OFFSET, &x_offset,
			   CANVAS_ATT_Y_OFFSET, &y_offset,
			   CANVAS_ATT_END);
    if ( (x_pixels < 2) || (y_pixels < 2) )
    {
	fprintf (stderr, "Canvas size: %d*%d too small\n", x_pixels, y_pixels);
	a_prog_bug (function_name);
    }
    setup_labels (canvas, dressing);
    setup_top_left (canvas, dressing);
    draw_hscale (canvas, NULL, dressing, x_pixels, y_pixels,
		 x_offset, y_offset);
    draw_vscale (canvas, NULL, dressing, x_pixels, y_pixels,
		 x_offset, y_offset);
    canvas_get_specification (canvas, &xlabel, &ylabel, NULL, NULL, NULL);
    /*  Make room for box  */
    *p_left = 1;
    *p_right = 1;
    *p_top = 1;
    *p_bottom = 1;
    if (!dressing->internal_ticks)
    {
	/*  External ticks  */
	if (dressing->show_top_ticks) *p_top += MAJOR_TICK_PIXELS;
	if (dressing->show_bottom_ticks) *p_bottom += MAJOR_TICK_PIXELS;
	if (dressing->show_left_ticks) *p_left += MAJOR_TICK_PIXELS;
	if (dressing->show_right_ticks) *p_right += MAJOR_TICK_PIXELS;
    }
    if (dressing->show_top_ticks || dressing->show_bottom_ticks)
    {
	/*  Show horizontal scale  */
	*p_bottom += info->hscale.size + AXIS_LINE_GAP * dressing->font_scale;
    }
    if (dressing->show_left_ticks || dressing->show_right_ticks)
    {
	/*  Show vertical scale  */
	*p_left += info->vscale.size + AXIS_LINE_GAP * dressing->font_scale;
    }
    if (dressing->show_labels)
    {
	unsigned int count;

	if ( (dressing->hlabel != NULL) || (xlabel != NULL) )
	{
	    /*  Show horizontal label  */
	    *p_bottom += info->hlabel.height;
	}
	if ( (dressing->vlabel != NULL) || (ylabel != NULL) )
	{
	    /*  Show vertical label  */
	    *p_left += info->vlabel.height;
	}
	if (dressing->title != NULL)
	{
	    /*  Show title  */
	    *p_top += info->title.height +
		TITLE_LOWER_GAP * dressing->font_scale;
	}
	/*  Compute space needed for restrictions  */
	for (count = 0; count < info->num_top_left; ++count)
	    *p_top += info->top_left[count].height +
		RESTR_LINE_GAP * dressing->font_scale;
    }
}   /*  End Function canvas_get_dressing_size  */

/*UNPUBLISHED_FUNCTION*/
void canvas_draw_dressing_internal (KWorldCanvas canvas,
				    dressing_type *dressing)
/*  [SUMMARY] Draw the dressing (axes, etc.) for a world canvas.
    [NOTE] This routine is called automatically each time the canvas is
    refreshed.
    <canvas> The world canvas object.
    <dressing> The dressing information.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    int x_pixels, y_pixels, x_offset, y_offset;
    unsigned long pixel_value;
    double old_linewidth;
    char *colourname;
    char *def_colourname = DEFAULT_DRESSING_COLOUR;
    double x[5], y[5];
    static char function_name[] = "canvas_draw_dressing_internal";

    dressing->drawn = TRUE;
    if (!dressing->display) return;
    /*  Draw axes  */
    if ( (colourname = dressing->grid_colour) == NULL )
    {
	colourname = def_colourname;
    }
    pixcanvas = canvas_get_pixcanvas (canvas);
    if ( !kwin_get_colour (pixcanvas, colourname, &pixel_value,
			   NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "%s: could not allocate colour: \"%s\" for dressing\n",
		 function_name, colourname);
	return;
    }
    canvas_get_attributes (canvas,
			   CANVAS_ATT_X_PIXELS, &x_pixels,
			   CANVAS_ATT_Y_PIXELS, &y_pixels,
			   CANVAS_ATT_X_OFFSET, &x_offset,
			   CANVAS_ATT_Y_OFFSET, &y_offset,
			   CANVAS_ATT_END);
    x[0] = x_offset - 1;
    y[0] = y_offset - 1;
    x[1] = x_offset + x_pixels;
    y[1] = y[0];
    x[2] = x[1];
    y[2] = y_offset + y_pixels;
    x[3] = x[0];
    y[3] = y[2];
    x[4] = x[0];
    y[4] = y[0];
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_LINEWIDTH, &old_linewidth,
			 KWIN_ATT_END);
    kwin_set_attributes (pixcanvas,
			 KWIN_ATT_LINEWIDTH, BORDER_LINEWIDTH,
			 KWIN_ATT_END);
    kwin_draw_lines_TRANSITION (pixcanvas, x, y, 5, pixel_value);
    kwin_set_attributes (pixcanvas,
			 KWIN_ATT_LINEWIDTH, old_linewidth,
			 KWIN_ATT_END);
    draw_top_left (canvas, pixcanvas, dressing, x_pixels, y_pixels,
		   x_offset, y_offset);
    draw_title (canvas, pixcanvas, dressing, x_pixels, y_pixels,
		x_offset, y_offset);
    draw_hscale (canvas, pixcanvas, dressing, x_pixels, y_pixels,
		 x_offset, y_offset);
    draw_vscale (canvas, pixcanvas, dressing, x_pixels, y_pixels,
		 x_offset, y_offset);
    draw_labels (canvas, pixcanvas, dressing, x_pixels, y_pixels,
		 x_offset, y_offset);
}   /*  End Function canvas_draw_dressing_internal  */

/*UNPUBLISHED_FUNCTION*/
void canvas_free_dressing (dressing_type *dressing)
/*  [SUMMARY] Free the dressing information for a world canvas.
    <dressing> The dressing information.
    [RETURNS] Nothing.
*/
{
    /*static char function_name[] = "canvas_free_dressing";*/

    if (dressing->info == NULL) return;
    if (dressing->info->top_left != NULL) m_free (dressing->info->top_left);
    m_free (dressing->info);
}   /*  End Function canvas_free_dressing  */


/*  Private functions follow  */

static void setup_labels (KWorldCanvas canvas, dressing_type *dressing)
/*  [SUMMARY] Setup the axis label strings.
    <canvas> The world canvas.
    <dressing> The dressing information.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    unsigned int num_restr;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    struct dressing_info_struct *info = dressing->info;
    static char function_name[] = "_canvas_setup_labels";

    if (dressing->title == NULL) m_clear (&info->title, sizeof info->title);
    else
    {
	if ( !kwin_hershey_draw_string (NULL, dressing->title,
					0.0, 0.0, 0.0,
					LABEL_FONT_SIZE * dressing->font_scale,
					0,
					&info->title.width,
					&info->title.height,
					&info->title.ascent,
					&info->title.descent) )
	{
	    fprintf (stderr, "%s: could not get title size\n", function_name);
	    m_clear (&info->title, sizeof info->title);
	}
    }
    info->hlabel.string[0] = '\0';
    info->vlabel.string[0] = '\0';
    if (!dressing->show_labels) return;
    canvas_get_specification (canvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    if (dressing->hlabel != NULL) xlabel = dressing->hlabel;
    if (dressing->vlabel != NULL) ylabel = dressing->vlabel;
    if ( (xlabel == NULL) || (ylabel == NULL) ) return;
    canvas_get_attributes (canvas, CANVAS_ATT_WCS_ASTRO, &ap, CANVAS_ATT_END);
    /*  First do the horizontal axis  */
    format_axis_name (info->hlabel.string, xlabel, ap);
    if ( !kwin_hershey_draw_string (NULL, info->hlabel.string,
				    0.0, 0.0, 0.0,
				    LABEL_FONT_SIZE * dressing->font_scale, 0,
				    &info->hlabel.width,
				    &info->hlabel.height,
				    &info->hlabel.ascent,
				    &info->hlabel.descent) )
    {
	fprintf (stderr, "Could not get hlabel size\n");
	a_prog_bug (function_name);
    }
    /*  Now do the vertical axis  */
    format_axis_name (info->vlabel.string, ylabel, ap);
    if ( !kwin_hershey_draw_string (NULL, info->vlabel.string,
				    0.0, 0.0, 0.0,
				    LABEL_FONT_SIZE * dressing->font_scale, 0,
				    &info->vlabel.width, &info->vlabel.height,
				    &info->vlabel.ascent,
				    &info->vlabel.descent) )
    {
	fprintf (stderr, "Could not get vlabel size\n");
	a_prog_bug (function_name);
    }
}   /*  End Function setup_labels  */

static void setup_top_left (KWorldCanvas canvas, dressing_type *dressing)
/*  [SUMMARY] Setup for the top-left corner (Equinox and restrictions).
    <canvas> The world canvas.
    <dressing> The dressing information.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    flag is_ra, is_glon, is_dec, is_glat, is_lon, is_lat;
    unsigned int num_restr, count;
    char *xlabel, *ylabel, *name;
    char **restr_names;
    double *restr_values;
    struct dressing_info_struct *info = dressing->info;
    struct label_struct *label;
    char txt[STRING_LENGTH];
    char append_string[STRING_LENGTH];
    static char function_name[] = "_canvas_setup_top_left";

    info->num_top_left = 0;
    if (!dressing->show_labels) return;
    canvas_get_specification (canvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    if (num_restr > info->max_top_left)
    {
	if (info->top_left != NULL) m_free (info->top_left);
	if ( ( info->top_left = m_alloc (num_restr * sizeof *info->top_left) )
	     == NULL )
	    m_abort (function_name, "top_left structures");
	info->max_top_left = num_restr;
    }
    canvas_get_attributes (canvas, CANVAS_ATT_WCS_ASTRO, &ap, CANVAS_ATT_END);
    for (count = 0; count < num_restr; ++count)
    {
	double value, radec_value;

	name = restr_names[count];
	value = restr_values[count];
	append_string[0] = '\0';
	label = info->top_left + count;
	is_ra = wcs_astro_test_if_ra (ap, name);
	is_lon = wcs_astro_test_if_lon (ap, name);
	is_glon = wcs_astro_test_if_glon (ap, name);
	is_dec = wcs_astro_test_if_dec (ap, name);
	is_glat = wcs_astro_test_if_glat (ap, name);
	is_lat = wcs_astro_test_if_lat (ap, name);
	if (is_lon)
	{
	    /*  This restriction is a RA value: get DEC value if needed  */
	    if (ap == NULL) continue;  /*  Should never happen  */
	    switch ( wcs_astro_get_equinox (ap) )
	    {
	      case 2000:
		strcat (append_string, " (J2000)");
		break;
	      case 1950:
		strcat (append_string, " (B1950)");
		break;
	    }
	    if ( wcs_astro_test_if_lat (ap, xlabel) )
	    {
		radec_value = wcs_astro_get_reference (ap, xlabel, TRUE);
		wcs_astro_transform3 (ap, 1,
				      name, &value, FALSE,
				      xlabel, &radec_value, FALSE,
				      NULL, NULL, FALSE,
				      num_restr, (CONST char **) restr_names,
				      restr_values);
	    }
	    else if ( wcs_astro_test_if_lat (ap, ylabel) )
	    {
		radec_value = wcs_astro_get_reference (ap, ylabel, TRUE);
		wcs_astro_transform3 (ap, 1,
				      name, &value, FALSE,
				      ylabel, &radec_value, FALSE,
				      NULL, NULL, FALSE,
				      num_restr, (CONST char **) restr_names,
				      restr_values);
	    }
	    else
	    {
		/*  DEC is not in the axes, perhaps it's another restriction */
		wcs_astro_transform3 (ap, 1,
				      name, &value, FALSE,
				      NULL, NULL, FALSE,
				      NULL, NULL, FALSE,
				      num_restr, (CONST char **) restr_names,
				      restr_values);
	    }
	}
	else if (is_lat)
	{
	    /*  This restriction is a DEC value: get RA value if needed  */
	    if (ap == NULL) continue;  /*  Should never happen  */
	    switch ( wcs_astro_get_equinox (ap) )
	    {
	      case 2000:
		strcat (append_string, " (J2000)");
		break;
	      case 1950:
		strcat (append_string, " (B1950)");
		break;
	    }
	    if ( wcs_astro_test_if_lon (ap, xlabel) )
	    {
		radec_value = wcs_astro_get_reference (ap, xlabel, TRUE);
		wcs_astro_transform3 (ap, 1,
				      name, &value, FALSE,
				      xlabel, &radec_value, FALSE,
				      NULL, NULL, FALSE,
				      num_restr, (CONST char **) restr_names,
				      restr_values);
	    }
	    else if ( wcs_astro_test_if_lon (ap, ylabel) )
	    {
		radec_value = wcs_astro_get_reference (ap, ylabel, TRUE);
		wcs_astro_transform3 (ap, 1,
				      name, &value, FALSE,
				      ylabel, &radec_value, FALSE,
				      NULL, NULL, FALSE,
				      num_restr, (CONST char **) restr_names,
				      restr_values);
	    }
	    else
	    {
		/*  RA is not in the axes, perhaps it's another restriction */
		wcs_astro_transform3 (ap, 1,
				      name, &value, FALSE,
				      NULL, NULL, FALSE,
				      NULL, NULL, FALSE,
				      num_restr, (CONST char **) restr_names,
				      restr_values);
	    }
	}
	else if (ap != NULL)
	{
	    wcs_astro_transform3 (ap, 1,
				  name, &value, FALSE,
				  NULL, NULL, FALSE,
				  NULL, NULL, FALSE,
				  num_restr, (CONST char **) restr_names,
				  restr_values);
	}
	/*  Now format the restriction value  */
	if (is_ra)
	{
	    int hours, minutes;
	    double seconds;

	    wcs_astro_breakdown_hms (&hours, &minutes, &seconds, NULL, NULL,
				     value / 15.0, 0.005);
	    sprintf (label->string, "Ra: %.2u\\uh %.2u\\um %5.2f\\us",
		     hours, minutes, seconds);
	}
	else if (is_glon)
	{
#ifdef GALACTIC_DMS
	    int degrees, minutes;
	    double seconds;

	    wcs_astro_breakdown_dms (NULL, &degrees, &minutes, &seconds,
				     NULL, NULL, value, 0.005);
	    sprintf (label->string, "Glon: %.2u\\uh %.2u\\um %5.2f\\us",
		     degrees, minutes, seconds);
#else
	    sprintf (label->string, "Glon: %.8g", value);
#endif
	}
	else if (is_dec)
	{
	    flag negative;
	    int degrees, minutes;
	    double seconds;

	    wcs_astro_breakdown_dms (&negative, &degrees, &minutes, &seconds,
				     NULL, NULL, value, 0.005);
	    sprintf (label->string, "Dec: %s%.2u` %.2u' %5.2f\"",
		     negative ? "-" : "", degrees, minutes, seconds);
	}
	else if (is_glat)
	{
#ifdef GALACTIC_DMS
	    flag negative;
	    int degrees, minutes;
	    double seconds;

	    wcs_astro_breakdown_dms (&negative, &degrees, &minutes, &seconds,
				     NULL, NULL, value, 0.005);
	    sprintf (label->string, "Glat: %s%.2u` %.2u' %5.2f\"",
		     negative ? "-" : "", degrees, minutes, seconds);
#else
	    sprintf (label->string, "Glat: %.8g", value);
#endif
	}
	else if ( (strncmp (name, "VELO", 4) == 0) ||
		  (strncmp (name, "FELO", 4) == 0) ||
		  (strcmp (name, "M/S") == 0) )
	{
	    sprintf (label->string, "Velocity: %.2f km/s", value * 1e-3);
	}
	else if ( (strncmp (name, "FREQ", 4) == 0) ||
		  (strcmp (name, "HZ") == 0) )
	{
	    sprintf (label->string, "Frequency: %.3f MHz", value * 1e-6);
	}
	else if (strncmp (name, "TIME", 4) == 0)
	{
	    wcs_astro_format_time (txt, value);
	    sprintf (label->string, "Time: %s", txt);
	}
	else if (strcmp (name, "BEAM") == 0)
	{
	    sprintf (label->string, "Beam Number: %.0f", value);
	}
	else sprintf (label->string, "%s: %e", name, value);
	strcat (label->string, append_string);
	if ( !kwin_hershey_draw_string (NULL, label->string,
					0.0, 0.0, 0.0,
					SCALE_FONT_SIZE * dressing->font_scale,
					0,
					&label->width, &label->height,
					&label->ascent, &label->descent) )
	{
	    fprintf (stderr, "Could not get top_left size\n");
	    a_prog_bug (function_name);
	}
    }
    info->num_top_left = num_restr;
}   /*  End Function setup_top_left  */

static void draw_top_left (KWorldCanvas canvas, KPixCanvas pixcanvas,
			   dressing_type *dressing, int x_pixels, int y_pixels,
			   int x_offset, int y_offset)
/*  [SUMMARY] Draw the top-left corner (Equinox and restrictions).
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas.
    <dressing> The dressing information.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    [RETURNS] Nothing.
*/
{
    signed int count;
    unsigned long pixel_value;
    double vpos;
    char *colourname;
    char *def_colourname = DEFAULT_DRESSING_COLOUR;
    struct dressing_info_struct *info = dressing->info;
    struct label_struct *label;
    static char function_name[] = "_canvas_draw_top_left";

    if (!dressing->show_labels) return;
    if ( (colourname = dressing->label_colour) == NULL )
    {
	colourname = def_colourname;
    }
    if ( !kwin_get_colour (pixcanvas, colourname, &pixel_value,
			   NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "%s: could not allocate colour: \"%s\" for dressing\n",
		 function_name, colourname);
	return;
    }
    vpos = y_offset - info->title.height -
	(TITLE_LOWER_GAP - RESTR_LINE_GAP) * dressing->font_scale;
    if (!dressing->internal_ticks) vpos -= MAJOR_TICK_PIXELS;
    for (count = info->num_top_left - 1; count >= 0; --count)
    {
	label = info->top_left + count;
	kwin_hershey_draw_string (pixcanvas, label->string,
				  0.0, vpos + label->descent,
				  0.0, SCALE_FONT_SIZE * dressing->font_scale,
				  pixel_value, NULL, NULL, NULL, NULL);
	vpos -= label->height + RESTR_LINE_GAP * dressing->font_scale;
    }
}   /*  End Function draw_top_left  */

static void draw_title (KWorldCanvas canvas, KPixCanvas pixcanvas,
			dressing_type *dressing, int x_pixels, int y_pixels,
			int x_offset, int y_offset)
/*  [SUMMARY] Draw the title.
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas.
    <dressing> The dressing information.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    [RETURNS] Nothing.
*/
{
    int tick_offset = 0;
    unsigned long pixel_value;
    char *colourname;
    char *def_colourname = DEFAULT_DRESSING_COLOUR;
    struct dressing_info_struct *info = dressing->info;
    static char function_name[] = "_canvas_draw_title";

    if (dressing->title == NULL) return;
    if ( (colourname = dressing->title_colour) == NULL )
    {
	colourname = def_colourname;
    }
    if ( !kwin_get_colour (pixcanvas, colourname, &pixel_value,
			   NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "%s: could not allocate colour: \"%s\" for dressing\n",
		 function_name, colourname);
	return;
    }
    if (!dressing->internal_ticks) tick_offset = MAJOR_TICK_PIXELS;
    kwin_hershey_draw_string (pixcanvas, dressing->title,
			      x_offset + (x_pixels - info->title.width) / 2,
			      y_offset - tick_offset -
			      TITLE_LOWER_GAP * dressing->font_scale +
			      info->title.descent,
			      0.0, LABEL_FONT_SIZE * dressing->font_scale,
			      pixel_value, NULL, NULL, NULL, NULL);
}   /*  End Function draw_title  */

static void draw_labels (KWorldCanvas canvas, KPixCanvas pixcanvas,
			 dressing_type *dressing, int x_pixels, int y_pixels,
			 int x_offset, int y_offset)
/*  [SUMMARY] Draw the horizontal and vertical labels.
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas.
    <dressing> The dressing information.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    [RETURNS] Nothing.
*/
{
    int tick_offset = 0;
    int canvas_width, canvas_height, tmp;
    unsigned long pixel_value;
    char *colourname;
    char *def_colourname = DEFAULT_DRESSING_COLOUR;
    struct dressing_info_struct *info = dressing->info;
    static char function_name[] = "_canvas_draw_labels";

    if (!dressing->show_labels) return;
    if ( (info->hlabel.string[0] == '\0') || (info->vlabel.string[0] == '\0') )
	return;
    kwin_get_size (pixcanvas, &canvas_width, &canvas_height);
    if ( (colourname = dressing->label_colour) == NULL )
    {
	colourname = def_colourname;
    }
    if ( !kwin_get_colour (pixcanvas, colourname, &pixel_value,
			   NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "%s: could not allocate colour: \"%s\" for dressing\n",
		 function_name, colourname);
	return;
    }
    if (!dressing->internal_ticks) tick_offset = MAJOR_TICK_PIXELS;
    /*  First do the horizontal axis  */
    tmp = info->hlabel.ascent;
    if (dressing->show_top_ticks || dressing->show_bottom_ticks)
	tmp += info->hscale.size + AXIS_LINE_GAP * dressing->font_scale;
    kwin_hershey_draw_string (pixcanvas, info->hlabel.string,
			      x_offset + (x_pixels - info->hlabel.width) / 2,
			      y_offset + y_pixels + tick_offset + tmp,
			      0.0, LABEL_FONT_SIZE * dressing->font_scale,
			      pixel_value, NULL, NULL, NULL, NULL);
    /*  Now do the vertical axis  */
    tmp = -info->vlabel.descent;
    if (dressing->show_left_ticks || dressing->show_right_ticks)
	tmp += info->vscale.size + AXIS_LINE_GAP * dressing->font_scale;
    kwin_hershey_draw_string (pixcanvas, info->vlabel.string,
			      x_offset - tick_offset - tmp,
			      y_offset + (y_pixels + info->vlabel.width) / 2,
			      -90.0, LABEL_FONT_SIZE * dressing->font_scale,
			      pixel_value, NULL, NULL, NULL, NULL);
}   /*  End Function draw_labels  */

static void format_axis_name (char axis_string[STRING_LENGTH],
			      CONST char *axis_name, KwcsAstro ap)
/*  [SUMMARY] Format an axis label.
    <axis_string> The formatted string is written here.
    <axis_name> The name of the axis.
    <ap> The KwcsAstro object. This may be NULL.
    [RETURNS] Nothing.
*/
{
    flag is_ra, is_glon, is_dec, is_glat;
    int str_len;
    char equinox_string[STRING_LENGTH];

    str_len = strlen (axis_name);
    equinox_string[0] = '\0';
    if (ap != NULL)
    {
	switch ( wcs_astro_get_equinox (ap) )
	{
	  case 2000:
	    strcpy (equinox_string, " (J2000)");
	    break;
	  case 1950:
	    strcpy (equinox_string, " (B1950)");
	    break;
	}
    }
    is_ra = wcs_astro_test_if_ra (ap, axis_name);
    is_glon = wcs_astro_test_if_glon (ap, axis_name);
    is_dec = wcs_astro_test_if_dec (ap, axis_name);
    is_glat = wcs_astro_test_if_glat (ap, axis_name);
    if (is_ra)
    {
	strcpy (axis_string, "Right Ascension");
	strcat (axis_string, equinox_string);
    }
    else if (is_glon)
    {
	strcpy (axis_string, "Galactic Longitude");
	strcat (axis_string, equinox_string);
    }
    else if (is_dec)
    {
	strcpy (axis_string, "Declination");
	strcat (axis_string, equinox_string);
    }
    else if (is_glat)
    {
	strcpy (axis_string, "Galactic Latitude");
	strcat (axis_string, equinox_string);
    }
    else if (strncmp (axis_name, "VELO", 4) == 0)
	strcpy (axis_string, "Velocity (km/s)");
    else if (strncmp (axis_name, "FELO", 4) == 0)
	strcpy (axis_string, "Velocity (km/s)");
    else if (strcmp (axis_name, "M/S") == 0)
	strcpy (axis_string, "Velocity (km/s)");
    else if (strncmp (axis_name, "FREQ", 4) == 0)
	strcpy (axis_string, "Frequency (MHz)");
    else if (strcmp (axis_name, "HZ") == 0)
	strcpy (axis_string, "Frequency (MHz)");
    else if (strcmp (axis_name, "TIME") == 0) strcpy (axis_string, "Time");
    else if ( (strcmp (axis_name, "ANGLE") == 0) ||
	      (strcmp (axis_name + str_len - 5, "(deg)") == 0) )
	strcpy (axis_string, "Angular Offset");
    else if (strcmp (axis_name, "BEAM") == 0) strcpy (axis_string,
						      "Beam Number");
    else if (strcmp (axis_name, "JY/BEAM") == 0) strcpy (axis_string,
							 "mJy/Beam");
    else if (strcmp (axis_name, "JY/PIX") == 0) strcpy (axis_string,
							"mJy/Pixel");
    else if (strcmp (axis_name, "JY") == 0) strcpy (axis_string, "mJy");
    else strcpy (axis_string, axis_name);
}   /*  End Function format_axis_name  */

static void draw_hscale (KWorldCanvas canvas, KPixCanvas pixcanvas,
			 dressing_type *dressing,
			 int x_pixels, int y_pixels,int x_offset, int y_offset)
/*  [SUMMARY] Draw the horizontal scale.
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas. If this is NULL, sizing is peformed but
    nothing is drawn.
    <dressing> The dressing information.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    [RETURNS] Nothing.
*/
{
    unsigned int num_restr;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    /*static char function_name[] = "_canvas_draw_hscale";*/

    dressing->info->hscale.size = 0.0;
    if (!dressing->show_top_ticks && !dressing->show_bottom_ticks) return;
    canvas_get_specification (canvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    if (xlabel == NULL) xlabel = dressing->hlabel;
    draw_scale (canvas, pixcanvas, dressing, x_pixels, y_pixels,
		x_offset, y_offset, xlabel, FALSE,
		&dressing->info->hscale.size);
}   /*  End Function draw_hscale  */

static void draw_vscale (KWorldCanvas canvas, KPixCanvas pixcanvas,
			 dressing_type *dressing,
			 int x_pixels, int y_pixels,int x_offset, int y_offset)
/*  [SUMMARY] Draw the vertical scale.
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas. If this is NULL, sizing is peformed but
    nothing is drawn.
    <dressing> The dressing information.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    [RETURNS] Nothing.
*/
{
    unsigned int num_restr;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    /*static char function_name[] = "_canvas_draw_vscale";*/

    dressing->info->vscale.size = 0.0;
    if (!dressing->show_left_ticks && !dressing->show_right_ticks) return;
    canvas_get_specification (canvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    if (ylabel == NULL) ylabel = dressing->vlabel;
    draw_scale (canvas, pixcanvas, dressing, x_pixels, y_pixels,
		x_offset, y_offset, ylabel, TRUE,
		&dressing->info->vscale.size);
}   /*  End Function draw_vscale  */

static void draw_scale (KWorldCanvas canvas, KPixCanvas pixcanvas,
			dressing_type *dressing,
			int x_pixels, int y_pixels,int x_offset, int y_offset,
			CONST char *name, flag vertical, double *max_size)
/*  [SUMMARY] Draw the scale for an axis.
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas. If this is NULL, sizing is peformed but
    nothing is drawn.
    <dressing> The dressing information.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    <name> The axis name.
    <vertical> If TRUE, the axis is vertical.
    <max_size> The maximum size taken by the scale text is written here
    (vertical pixels for the horizontal scale and horizontal pixels for the
    vertical scale). If this is NULL, nothing is written here.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    flag is_ra, is_glon, is_dec, is_glat;
    int canvas_width, canvas_height;
    unsigned long scale_pixel, grid_pixel;
    double left_lx, right_lx, bottom_ly, top_ly;
    double dummy_size;
    corners_type wcorners;
    char *colourname, *axis_string;
    char *def_colourname = DEFAULT_DRESSING_COLOUR;
    double *warr;
    double seconds[2];
    double xarr[4], yarr[4];
    static char function_name[] = "_canvas_draw_scale";

    if (max_size == NULL) max_size = &dummy_size;
    *max_size = 0.0;
    if (name == NULL) return;
    axis_string = vertical ?
	dressing->info->vlabel.string : dressing->info->hlabel.string;
    canvas_get_attributes (canvas,
			   CANVAS_ATT_WCS_ASTRO, &ap,
			   CANVAS_ATT_END);
    /*  Get linear co-ordinates of box  */
    left_lx = (double) x_offset - 0.5;
    right_lx = (double) (x_offset + x_pixels) - 0.5;
    bottom_ly = (double) (y_offset + y_pixels) - 0.5;
    top_ly = (double) y_offset - 0.5;
    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &left_lx, &bottom_ly,
				     &left_lx, &bottom_ly, NULL, NULL);
    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &right_lx, &top_ly,
				     &right_lx, &top_ly, NULL, NULL);
    if (vertical)
    {
	if (bottom_ly == top_ly) return;
    }
    else
    {
	if (left_lx == right_lx) return;
    }
    if (pixcanvas == NULL)
    {
	canvas_width = 0;
	canvas_height = 0;
    }
    else kwin_get_size (pixcanvas, &canvas_width, &canvas_height);
    /*  Get world co-ordinates of box corners  */
    wcorners.bottom_left.x = (double) (x_offset - 1);
    wcorners.bottom_left.y = (double) (y_offset + y_pixels);
    wcorners.bottom_right.x = (double) (x_offset + x_pixels);
    wcorners.bottom_right.y = (double) (y_offset + y_pixels);
    wcorners.top_left.x = (double) (x_offset - 1);
    wcorners.top_left.y = (double) (y_offset - 1);
    wcorners.top_right.x = (double) (x_offset + x_pixels);
    wcorners.top_right.y = (double) (y_offset - 1);
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &wcorners.bottom_left.x,
				     &wcorners.bottom_left.y,
				     NULL, NULL,
				     &wcorners.bottom_left.x,
				     &wcorners.bottom_left.y);
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &wcorners.bottom_right.x,
				     &wcorners.bottom_right.y,
				     NULL, NULL,
				     &wcorners.bottom_right.x,
				     &wcorners.bottom_right.y);
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &wcorners.top_left.x,
				     &wcorners.top_left.y,
				     NULL, NULL,
				     &wcorners.top_left.x,
				     &wcorners.top_left.y);
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &wcorners.top_right.x,
				     &wcorners.top_right.y,
				     NULL, NULL,
				     &wcorners.top_right.x,
				     &wcorners.top_right.y);
    if ( (colourname = dressing->scale_colour) == NULL )
    {
	colourname = def_colourname;
    }
    if (pixcanvas == NULL)
    {
	scale_pixel = 0;
	grid_pixel = 0;
    }
    else
    {
	if ( !kwin_get_colour (pixcanvas, colourname, &scale_pixel,
			       NULL, NULL, NULL) )
	{
	    fprintf (stderr,
		     "%s: could not allocate colour: \"%s\" for dressing\n",
		     function_name, colourname);
	    return;
	}
	if ( (colourname = dressing->grid_colour) == NULL )
	{
	    colourname = def_colourname;
	}
	if ( !kwin_get_colour (pixcanvas, colourname, &grid_pixel,
			       NULL, NULL, NULL) )
	{
	    fprintf (stderr,
		     "%s: could not allocate colour: \"%s\" for dressing\n",
		     function_name, colourname);
	    return;
	}
    }
    is_ra = wcs_astro_test_if_ra (ap, name);
    is_glon = wcs_astro_test_if_glon (ap, name);
    is_dec = wcs_astro_test_if_dec (ap, name);
    is_glat = wcs_astro_test_if_glat (ap, name);
    if (vertical)
    {
	/*  Get pixel co-ordinates of bottom-left, top-left and middle-left  */
	xarr[0] = x_offset;
	yarr[0] = y_offset + y_pixels - 1;
	xarr[1] = x_offset;
	yarr[1] = y_offset;
	xarr[2] = x_offset;
	yarr[2] = y_offset + y_pixels / 2;
	/*  And bottom-left + 1V  */
	xarr[3] = x_offset;
	yarr[3] = y_offset + y_pixels - 2;
	warr = yarr;
    }
    else
    {
	/*  Get pixel co-ordinates of bottom-left, bottom-right and
	    bottom-middle  */
	xarr[0] = x_offset;
	yarr[0] = y_offset + y_pixels - 1;
	xarr[1] = x_offset + x_pixels - 1;
	yarr[1] = yarr[0];
	xarr[2] = x_offset + x_pixels / 2;
	yarr[2] = yarr[0];
	/*  And bottom-left + 1H  */
	xarr[3] = x_offset + 1;
	yarr[3] = y_offset + y_pixels - 1;
	warr = xarr;
    }
    /*  Convert to world co-ordinates  */
    canvas_convert_to_canvas_coords (canvas, FALSE, 4, xarr, yarr, NULL, NULL,
				     xarr, yarr);
    if (is_ra)
    {
	unsigned int num_minor;
	double ra_step, min_ra, max_ra, sign, ra_diff;
	hms_format_info_type finfo;
	int hours[2], minutes[2], whole_seconds[2];

	while (warr[0] < 0.0) warr[0] += 360.0;
	while (warr[1] < 0.0) warr[1] += 360.0;
	while (warr[2] < 0.0) warr[2] += 360.0;
	if (warr[0] < warr[1])
	{
	    if ( (warr[0] < warr[2]) && (warr[2] < warr[1]) )
	    {
		/*  Not wrapping around 24h (360 deg)  */
		min_ra = warr[0];
		max_ra = warr[1];
	    }
	    else
	    {
		/*  Wrapping around 24h (360 deg)  */
		warr[0] += 360.0;
		min_ra = warr[1];
		max_ra = warr[0];
	    }
	}
	else
	{
	    if ( (warr[1] < warr[2]) && (warr[2] < warr[0]) )
	    {
		/*  Not wrapping around 24h (360 deg)  */
		min_ra = warr[1];
		max_ra = warr[0];
	    }
	    else
	    {
		/*  Wrapping around 24h (360 deg)  */
		warr[1] += 360.0;
		min_ra = warr[0];
		max_ra = warr[1];
	    }
	}
	ra_diff = (max_ra - min_ra) / 15.0;
	wcs_astro_breakdown_hms (hours, minutes, seconds, whole_seconds, NULL,
				 min_ra / 15.0, 0.0);
	wcs_astro_breakdown_hms (hours + 1, minutes + 1, seconds + 1,
				 whole_seconds + 1, NULL,
				 max_ra / 15.0 , 0.0);
	finfo.num_hour_trans = hours[1] - hours[0];
	finfo.num_min_trans = minutes[1] - minutes[0] +
	    finfo.num_hour_trans * 60;
#ifdef DEBUG
	fprintf (stderr, "%s: %uh %um %gs  to %uh %um %gs\n",
		 name, hours[0], minutes[0], seconds[0],
		 hours[1], minutes[1], seconds[1]);
	fprintf (stderr, "num_hour_trans: %d  num_min_trans: %d\n",
		 num_hour_trans, finfo.num_min_trans);
	fprintf (stderr, "ra_diff: hours: %g min: %g sec: %g\n",
		 ra_diff, ra_diff * 60.0, ra_diff_seconds * 3600.0);
#endif
	/*  Determine RA step (in hours)  */
	ra_step = get_time_step (ra_diff);
	num_minor = ra_step / get_time_step (ra_step * 1.001) + 1e-10;
	finfo.ra_step_hours = ra_step;
	finfo.ra_step_minutes = finfo.ra_step_hours * 60.0;
	finfo.ra_step_seconds = finfo.ra_step_hours * 3600.0;
	if (finfo.ra_step_seconds >= 1.0) finfo.precision = 0;
	else
	{
	    double tmp;

	    finfo.precision = ( -floor ( log10 (finfo.ra_step_seconds) ) );
	    tmp = finfo.ra_step_seconds * pow (10.0, (double) finfo.precision);
	    if (floor (tmp) < tmp) ++finfo.precision;
	}
#ifdef DEBUG
	fprintf (stderr, "step_hours: %g  step_mins: %g  step_secs: %g\n",
		 finfo.ra_step_hours, finfo.ra_step_minutes,
		 finfo.ra_step_seconds);
#endif
	/*  Convert back to real units (degrees)  */
	ra_step *= 15.0;
	if (vertical) sign = (warr[0] < warr[3]) ? -1.0 : 1.0;
	else sign = (warr[3] < warr[0]) ? -1.0 : 1.0;
	finfo.first_time = TRUE;
	_canvas_dressing_show_generic_scale
	    (canvas, pixcanvas, dressing, &wcorners, vertical, 360.0,
	     max_size, x_pixels, y_pixels, x_offset, y_offset,
	     canvas_width, canvas_height,
	     left_lx, right_lx, bottom_ly, top_ly,
	     scale_pixel, grid_pixel, ra_step * sign, num_minor,
	     _canvas_dressing_hms_rescale, hms_format, &finfo);
	/*  Nothing more to do for RA axis  */
	return;
#undef DEBUG
    }
#ifdef GALACTIC_DMS
    if ( is_dec || is_glon || is_glat || (strcmp (name, "ANGLE") == 0) )
#else
    if ( is_dec || (strcmp (name, "ANGLE") == 0) )
#endif
    {
	flag negative0, negative1;
	unsigned int num_minor;
	double min_dec, max_dec, dec_step, dec_diff, sign;
	dms_format_info_type finfo;
	int degrees[2], minutes[2], whole_seconds[2];

	if (warr[0] < warr[1])
	{
	    min_dec = warr[0];
	    max_dec = warr[1];
	}
	else
	{
	    min_dec = warr[1];
	    max_dec = warr[0];
	}
	dec_diff = max_dec - min_dec;
	wcs_astro_breakdown_dms (&negative0, degrees, minutes, seconds,
				 whole_seconds,
				 NULL, min_dec, 0.0);
	wcs_astro_breakdown_dms (&negative1,
				 degrees + 1, minutes + 1, seconds + 1,
				 whole_seconds + 1, NULL, max_dec, 0.0);
#ifdef DEBUG
	fprintf (stderr, "%s: %s%dd %dm %gs  to %s%dd %dm %gs\n",
		 name,
		 negative0 ? "-" : "", degrees[0], minutes[0], seconds[0],
		 negative1 ? "-" : "", degrees[1], minutes[1], seconds[1]);
#endif
	if (negative0)
	{
	    degrees[0] = -degrees[0];
	    minutes[0] = -minutes[0];
	    whole_seconds[0] = -whole_seconds[0];
	}
	if (negative1)
	{
	    degrees[1] = -degrees[1];
	    minutes[1] = -minutes[1];
	    whole_seconds[1] = -whole_seconds[1];
	}
	finfo.num_deg_trans = degrees[1] - degrees[0];
	finfo.num_min_trans = minutes[1] - minutes[0] +
	    finfo.num_deg_trans * 60;
	/*  Determine DEC step, with some special rules for full-sky images  */
	dec_step = get_angle_step (dec_diff);
	num_minor = dec_step / get_angle_step (dec_step * 1.001) + 1e-10;
	finfo.dec_step_degrees = dec_step;
	finfo.dec_step_minutes = dec_step * 60.0;
	finfo.dec_step_seconds = dec_step * 3600.0;
	if (finfo.dec_step_seconds >= 1.0) finfo.precision = 0;
	else
	{
	    double tmp;

	    finfo.precision = ( -floor ( log10 (finfo.dec_step_seconds) ) );
	    tmp = finfo.dec_step_seconds * pow (10.0, (double)finfo.precision);
	    if (floor (tmp) < tmp) ++finfo.precision;
	}
	if (vertical) sign = (warr[0] < warr[1]) ? -1.0 : 1.0;
	else sign = (warr[1] < warr[0]) ? -1.0 : 1.0;
	finfo.first_time = (is_dec || is_glon || is_glat) ? TRUE : FALSE;
	_canvas_dressing_show_generic_scale
	    (canvas, pixcanvas, dressing, &wcorners, vertical,
	     360.0, max_size,
	     x_pixels, y_pixels, x_offset, y_offset,
	     canvas_width, canvas_height,
	     left_lx, right_lx, bottom_ly, top_ly,
	     scale_pixel, grid_pixel, dec_step * sign,num_minor,
	     NULL, dms_format, &finfo);
	/*  Nothing more to do for DEC axis  */
	return;
    }
    if (strcmp (name, "TIME") == 0)
    {
	unsigned int num_minor;
	double time_step, min_time, max_time, time_diff, sign;
	time_format_info_type finfo;
	unsigned int year[2], month[2], day_of_month[2], day_of_year[2];
	unsigned int hours[2], minutes[2], whole_seconds[2];

	if (warr[0] < warr[1])
	{
	    min_time = warr[0];
	    max_time = warr[1];
	}
	else
	{
	    min_time = warr[1];
	    max_time = warr[0];
	}
	time_diff = (max_time - min_time) / 3600.0;  /*  In hours  */
	wcs_astro_breakdown_time (year, month, day_of_month, day_of_year,
				  hours, minutes, seconds, whole_seconds,
				  NULL, min_time);
	wcs_astro_breakdown_time (year + 1, month + 1, day_of_month + 1,
				  day_of_year + 1, hours + 1, minutes + 1,
				  seconds + 1, whole_seconds + 1, NULL,
				  max_time);
#ifdef DEBUG
	fprintf (stderr, "%s: %uyd %uh %um %.2fs  to %uyd %uh %um %.2fs\n",
		 name, day_of_year[0], hours[0], minutes[0], seconds[0],
		 day_of_year[1], hours[1], minutes[1], seconds[1]);
#endif
	finfo.num_day_trans = (year[1] - year[0]) * 365;
	finfo.num_day_trans += day_of_year[1] - day_of_year[0];
	finfo.num_hour_trans = (hours[1] - hours[0]) +
	    finfo.num_day_trans * 24;
	finfo.num_min_trans = minutes[1] - minutes[0] +
	    finfo.num_hour_trans * 60;
	/*  Determine TIME step (units are in hours)  */
	time_step = get_time_step (time_diff);
	num_minor = time_step / get_time_step (time_step * 1.001) + 1e-10;
	finfo.time_step_hours = time_step;
	finfo.time_step_days = finfo.time_step_hours / 24.0;
	finfo.time_step_minutes = finfo.time_step_hours * 60.0;
	finfo.time_step_seconds = finfo.time_step_hours * 3600.0;
	if (finfo.time_step_seconds >= 1.0) finfo.precision = 0;
	else
	{
	    double tmp;

	    finfo.precision = ( -floor ( log10 (finfo.time_step_seconds) ) );
	    tmp = finfo.time_step_seconds * pow (10.0, (double) finfo.precision);
	    if (floor (tmp) < tmp) ++finfo.precision;
	}
	if (vertical) sign = (warr[0] < warr[1]) ? -1.0 : 1.0;
	else sign = (warr[1] < warr[0]) ? -1.0 : 1.0;
	finfo.first_time = TRUE;
	_canvas_dressing_show_generic_scale
	    (canvas, pixcanvas, dressing, &wcorners, vertical, 360.0,
	     max_size, x_pixels, y_pixels, x_offset, y_offset,
	     canvas_width, canvas_height,
	     left_lx, right_lx, bottom_ly, top_ly,
	     scale_pixel, grid_pixel,
	     finfo.time_step_seconds * sign, num_minor,
	     NULL, time_format, &finfo);
	/*  Nothing more to do for TIME axis  */
	return;
    }
    if (strcmp (name, "BEAM") == 0)
    {
	double sign;
	linear_format_info_type format_info;

	/*  Beam number axis: should be integer values  */
	/*  Determine beam step  */
	if (vertical) sign = (warr[0] < warr[1]) ? -1.0 : 1.0;
	else sign = (warr[1] < warr[0]) ? -1.0 : 1.0;
	format_info.multiplier = 1.0;
	format_info.step = 1.0;
	format_info.format_string = "%.0f";
	_canvas_dressing_show_generic_scale
	    (canvas, pixcanvas, dressing, &wcorners, vertical, 360.0,
	     max_size, x_pixels, y_pixels, x_offset, y_offset,
	     canvas_width, canvas_height,
	     left_lx, right_lx, bottom_ly, top_ly,
	     scale_pixel, grid_pixel, sign, 1,
	     NULL, linear_format, &format_info);
	/*  Nothing more to do for BEAM axis  */
	return;
    }
    /*  Must be some other kind of axis  */
    if ( (strncmp (name, "VELO", 4) == 0) || (strncmp (name, "FELO", 4) == 0)
	 || (strcmp (name, "M/S") == 0) || (strncmp (name, "FREQ", 4) == 0) ||
	 (strcmp (name, "HZ") == 0) || (strcmp (name, "JY/BEAM") == 0) ||
	 (strcmp (name, "JY/PIX") == 0) || (strcmp (name, "JY") == 0) )
    {
	int precision;
	unsigned int num_minor;
	double range = fabs (warr[1] - warr[0]);
	double step, sign, scale;
	linear_format_info_type format_info;
	char format[STRING_LENGTH];

	if (vertical) sign = (warr[0] < warr[1]) ? -1.0 : 1.0;
	else sign = (warr[1] < warr[0]) ? -1.0 : 1.0;
	step = get_decimal_step (range);
	num_minor = step / get_decimal_step (step * 1.001) + 1e-10;
	scale = 1.0;
	if ( (strncmp (name, "VELO", 4) == 0) ||
	     (strncmp (name, "FELO", 4) == 0) ||
	     (strcmp (name, "M/S") == 0) ) scale = 1e-3;
	else if ( (strncmp (name, "FREQ", 4) == 0) ||
		  (strcmp (name, "HZ") == 0) ) scale = 1e-6;
	else if (strcmp (name, "JY/BEAM") == 0)
	{
	    if (step >= 1.0) strcpy (axis_string, "Jy/Beam");
	    else scale = 1e3;
	}
	else if (strcmp (name, "JY/PIX") == 0)
	{
	    if (step >= 1.0) strcpy (axis_string, "Jy/Pixel");
	    else scale = 1e3;
	}
	else if (strcmp (name, "JY") == 0)
	{
	   if (step >= 1.0) strcpy (axis_string, "Jy");
	   else scale = 1e3;
	}
	precision = floor ( log10 (step * scale) );
	if (precision < 0)
	{
	    double tmp;

	    precision = -precision;
	    tmp = step * pow (10.0, (double) precision);
	    if (floor (tmp) < tmp) ++precision;
	}
	else precision = 0;
	sprintf (format, "%%.%df", precision);
	format_info.multiplier = scale;
	format_info.step = step;
	format_info.format_string = format;
	_canvas_dressing_show_generic_scale
	    (canvas, pixcanvas, dressing, &wcorners, vertical, 360.0,
	     max_size, x_pixels, y_pixels, x_offset, y_offset,
	     canvas_width, canvas_height,
	     left_lx, right_lx, bottom_ly, top_ly,
	     scale_pixel, grid_pixel,
	     step * sign, num_minor,
	     NULL, linear_format, &format_info);
	/*  Nothing more to do for Velocity/Frequency axis  */
	return;
    }
    else
    {   /*  Generic linear axis  */
	int precision;
	unsigned int num_minor;
	double range = fabs (warr[1] - warr[0]);
	double step, sign, maxabs;
	linear_format_info_type format_info;
	char format[STRING_LENGTH];

	if (vertical) sign = (warr[0] < warr[1]) ? -1.0 : 1.0;
	else sign = (warr[1] < warr[0]) ? -1.0 : 1.0;
	maxabs = fabs (warr[0]);
	step = fabs (warr[1]);
	if (step > maxabs) maxabs = step;
	if ( (is_glon || is_glat) && (range >= 120.0) )
	{
	    /*  Special case for large images with galactic co-ordinates
		displayed in decimal degrees  */
	    if      (range >= 420.0) step = 120.0;
	    else if (range >= 210.0) step =  60.0;
	    else                     step =  30.0;
	    num_minor = 3;
	}
	else
	{
	    step = get_decimal_step (range);
	    num_minor = step / get_decimal_step (step * 1.001) + 1e-10;
	}
	format_info.multiplier = 1.0;
	precision = floor ( log10 (step) );
	if (precision < -3)
	{
	    double tmp;

	    format_info.multiplier = pow (10.0, -(double) precision);
	    tmp = step * pow (10.0, (double) -precision);
	    if (floor (tmp) < tmp) sprintf (format, "%%.1fe%d", precision);
	    else sprintf (format, "%%.0fe%d", precision);
	}
	else if (precision < 0)
	{
	    double tmp;

	    precision = -precision;
	    tmp = step * pow (10.0, (double) precision);
	    if (floor (tmp) < tmp) ++precision;
	    sprintf (format, "%%.%df", precision);
	}
	else if (precision < 4) strcpy (format, "%.0f");
	else
	{
	    format_info.multiplier = pow (10.0, -(double) precision);
	    sprintf (format, "%%.0fe%d", precision);
	}
	format_info.step = step;
	format_info.format_string = format;
	_canvas_dressing_show_generic_scale
	    (canvas, pixcanvas, dressing, &wcorners, vertical, 360.0,
	     max_size, x_pixels, y_pixels, x_offset, y_offset,
	     canvas_width, canvas_height,
	     left_lx, right_lx, bottom_ly, top_ly,
	     scale_pixel, grid_pixel,
	     step * sign, num_minor,
	     NULL, linear_format, &format_info);
	/*  Nothing more to do for unknown axis  */
	return;
    }
}   /*  End Function draw_scale  */

static void _canvas_dressing_show_generic_scale
    (KWorldCanvas canvas, KPixCanvas pixcanvas, dressing_type *dressing,
     CONST corners_type *wcorners, flag vertical, double wrap_modulus,
     double *max_size, int x_pixels, int y_pixels, int x_offset, int y_offset,
     int canvas_width, int canvas_height,
     double left_lx, double right_lx, double bottom_ly, double top_ly,
     unsigned long scale_pixel, unsigned long grid_pixel,
     double step, unsigned int num_minor,
     double (*rescale) (double value, void *info),
     void (*format) (char *string, double value, void *info),
     void *info)
/*  [SUMMARY] Draw a linear scale.
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas.
    <dressing> The dressing information.
    <wcorners> The corners in world co-ordinates.
    <vertical> If TRUE, the scale is drawn vertically.
    <wrap_modulus> The wrapping modulus.
    <max_size> The maximum size taken by the scale text is written here
    (vertical pixels for the horizontal scale and horizontal pixels for the
    vertical scale). If this is NULL, nothing is written here.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    <canvas_width> The width of the canvas in pixels.
    <canvas_height> The height of the canvas in pixels.
    <left_lx> The linear co-ordinate of the left boundary.
    <right_lx> The linear co-ordinate of the right boundary.
    <bottom_ly> The linear co-ordinate of the bottom boundary.
    <top_ly> The linear co-ordinate of the top boundary.
    <scale_pixel> The scale pixel value.
    <grid_pixel> The grid pixel value.
    <step> The major tick step.
    <num_minor> The number of minor ticks per major.
    <rescale> A function that is called to rescale values. If this is NULL, no
    rescaling is done.
    <format> A function that is called to format the value into a string.
    <info> An arbitrary pointer that is passed to the functions.
    [RETURNS] Nothing.
*/
{
    int tick_offset = 0;
    unsigned int minor_count;
    double start, value, width, height, ascent, descent, tick_pos, far_pos;
    double minor_step = step / (double) num_minor;
    char string[STRING_LENGTH];

    if (!dressing->internal_ticks) tick_offset = MAJOR_TICK_PIXELS;
    /*  Compute starting position: need to check both sides of bounding box  */
    if (vertical)  /*  Start from top  */
    {
	start = wcorners->top_left.y;
	if (step < 0.0)
	{
	    if (wcorners->top_right.y > start) start = wcorners->top_right.y;
	}
	else
	    if (wcorners->top_right.y < start) start = wcorners->top_right.y;
    }
    else  /*  Start from left  */
    {
	start = wcorners->bottom_left.x;
	if (step < 0.0)
	{
	    if (wcorners->top_left.x > start) start = wcorners->top_left.x;
	}
	else
	    if (wcorners->top_left.x < start) start = wcorners->top_left.x;
    }
    start = floor (start / step) * step;
    for (value = start, minor_count = 0; ; value += minor_step, ++minor_count)
    {
	flag is_major = (minor_count % num_minor == 0) ? TRUE : FALSE;
	double scaled = (rescale == NULL) ? value : (*rescale) (value, info);

	if (vertical)
	{
	    draw_y_marker (canvas, pixcanvas, dressing, wcorners,
			   x_pixels, y_pixels, x_offset, y_offset,
			   left_lx, right_lx, bottom_ly, top_ly,
			   grid_pixel, scaled, is_major,
			   wrap_modulus, &tick_pos, &far_pos);
	    if ( (tick_pos > y_offset + y_pixels) &&
		 (far_pos > y_offset + y_pixels) ) break; /* End of the line */
	    if ( (tick_pos < y_offset - 1) ||
		 (tick_pos > y_offset + y_pixels) ) continue;
	}
	else  /*  Horizontal axis  */
	{
	    draw_x_marker (canvas, pixcanvas, dressing, wcorners,
			   x_pixels, y_pixels, x_offset, y_offset,
			   left_lx, right_lx, bottom_ly, top_ly,
			   grid_pixel, scaled, is_major,
			   wrap_modulus, &tick_pos, &far_pos);
	    if ( (tick_pos > x_offset + x_pixels) &&
		 (far_pos > x_offset + x_pixels) ) break; /* End of the line */
	    if ( (tick_pos < x_offset - 1) ||
		 (tick_pos > x_offset + x_pixels) ) continue;
	}
	if (!is_major) continue;
	(*format) (string, scaled, info);
	if ( !kwin_hershey_draw_string (NULL, string,
					0.0, 0.0, 0.0,
					SCALE_FONT_SIZE * dressing->font_scale,
					0,
					&width, &height, &ascent, &descent) )
	    return;
	++ascent;
	if (vertical)
	{
	    if (width > *max_size) *max_size = width;
	    if (pixcanvas == NULL) continue;
	    kwin_hershey_draw_string (pixcanvas, string,
				      x_offset - width - 3 - tick_offset,
				      tick_pos,
				      0.0,
				      SCALE_FONT_SIZE * dressing->font_scale,
				      scale_pixel, NULL, NULL, NULL, NULL);
	}
	else  /*  Horizontal axis  */
	{
	    if (height > *max_size) *max_size = height;
	    if (pixcanvas == NULL) continue;
	    if (tick_pos + width / 2 > canvas_width)
		tick_pos = canvas_width - width;
	    else tick_pos -= width / 2;
	    kwin_hershey_draw_string (pixcanvas, string,
				      tick_pos,
				      y_offset + y_pixels + tick_offset + 1 +
				      ascent,
				      0.0,
				      SCALE_FONT_SIZE * dressing->font_scale,
				      scale_pixel, NULL, NULL, NULL, NULL);
	}
    }
}   /*  End Function _canvas_dressing_show_generic_scale  */

static void linear_format (char *string, double value, void *info)
{
    linear_format_info_type *finfo = info;

    sprintf (string, finfo->format_string,
	     (value + fabs (finfo->step) * 1e-10) * finfo->multiplier);
}   /*  End Function linear_format  */

static double _canvas_dressing_hms_rescale (double value, void *info)
{
    static char function_name[] = "_canvas_dressing_hms_rescale";

    if (fabs (value) > ANGLE_LIMIT)
    {
	fprintf (stderr, "Illegal angle value: %g\n", value);
	a_prog_bug (function_name);
    }
    while (value >= 360.0) value -= 360.0;
    return (value);
}   /*  End Function _canvas_dressing_hms_rescale  */

static void hms_format (char *string, double value, void *info)
{
    flag show_hour, show_minute, show_second;
    int curr_hour, curr_minute, fwidth;
    double sec;
    hms_format_info_type *finfo = info;

    fwidth = (finfo->precision < 1) ? 2 : finfo->precision + 3;
    wcs_astro_breakdown_hms (&curr_hour, &curr_minute, &sec,
			     NULL, NULL, value / 15.0, 0.005);
    show_hour = FALSE;
    show_minute = FALSE;
    show_second = FALSE;
    if (finfo->num_hour_trans > 0) show_hour = TRUE;
    else if (finfo->num_min_trans > 0) show_minute = TRUE;
    else show_second = TRUE;
    if (finfo->ra_step_hours >= 1.0) show_hour = TRUE;
    else if (finfo->ra_step_minutes >= 1.0) show_minute = TRUE;
    else show_second = TRUE;
    if (finfo->first_time)
    {
	finfo->first_time = FALSE;
	show_hour = TRUE;
    }
    if (show_hour)
    {
	if (show_minute || show_second)
	{
	    if (show_second)
		sprintf (string, "%u\\uh%02u\\um%0*.*f\\us",
			 curr_hour, curr_minute, fwidth, finfo->precision,sec);
	    else sprintf (string, "%u\\uh%02u\\um",
			  curr_hour, curr_minute);
	}
	else sprintf (string, "%u\\uh", curr_hour);
    }
    else if (show_minute)
    {
	if (show_second)
	    sprintf (string, "%02u\\um%0*.*f\\us",
		     curr_minute, fwidth, finfo->precision, sec);
	else sprintf (string, "%02u\\um", curr_minute);
    }
    else sprintf (string, "%0*.*f\\us", fwidth, finfo->precision, sec);
}   /*  End Function hms_format  */

static void dms_format (char *string, double value, void *info)
{
    flag show_degree, show_minute, show_second, negative;
    int curr_degree, curr_minute, fwidth;
    double sec;
    dms_format_info_type *finfo = info;
    char txt[STRING_LENGTH];

    fwidth = (finfo->precision < 1) ? 2 : finfo->precision + 3;
    wcs_astro_breakdown_dms (&negative, &curr_degree, &curr_minute,
			     &sec, NULL, NULL,
			     value,
			     pow (10.0, - (double) finfo->precision) * 0.5);
    show_degree = FALSE;
    show_minute = FALSE;
    show_second = FALSE;
    if (finfo->num_deg_trans > 0) show_degree = TRUE;
    else if (finfo->num_min_trans > 0) show_minute = TRUE;
    else show_second = TRUE;
    if (finfo->dec_step_degrees >= 1.0) show_degree = TRUE;
    else if (finfo->dec_step_minutes >= 1.0) show_minute = TRUE;
    else show_second = TRUE;
    if (finfo->first_time)
    {
	finfo->first_time = FALSE;
	show_degree = TRUE;
    }
    if (show_degree)
    {
	if (show_minute || show_second)
	{
	    if (show_second)
		sprintf (txt, "%d`%02d'%0*.*f\"",
			 curr_degree, curr_minute, fwidth, finfo->precision,
			 sec);
	    else sprintf (txt, "%d`%02d'", curr_degree, curr_minute);
	}
	else sprintf (txt, "%d`", curr_degree);
    }
    else if (show_minute)
    {
	if (show_second)
	    sprintf (txt, "%02d'%0*.*f\"", curr_minute, fwidth,
		     finfo->precision, sec);
	else sprintf (txt, "%02d'", curr_minute);
    }
    else sprintf (txt, "%0*.*f\"", fwidth, finfo->precision,sec);
    if (negative)
    {
	strcpy (string, "-");
	strcat (string, txt);
    }
    else strcpy (string, txt);
}   /*  End Function dms_format  */

static void time_format (char *string, double value, void *info)
{
    flag show_day, show_hour, show_minute, show_second;
    unsigned int curr_day_of_month;
    unsigned int curr_hour, curr_minute, curr_whole_second;
    double sec, subsec;
    time_format_info_type *finfo = info;
    char txt[STRING_LENGTH];

    wcs_astro_breakdown_time (NULL, NULL, &curr_day_of_month, NULL,
			      &curr_hour, &curr_minute, &sec,
			      &curr_whole_second, &subsec, value);
    show_day = FALSE;
    show_hour = FALSE;
    show_minute = FALSE;
    show_second = FALSE;
    if (finfo->num_day_trans > 0) show_day = TRUE;
    else if (finfo->num_hour_trans > 0) show_hour = TRUE;
    else if (finfo->num_min_trans > 0) show_minute = TRUE;
    else show_second = TRUE;
    if (finfo->time_step_days >= 1.0) show_day = TRUE;
    else if (finfo->time_step_hours >= 1.0) show_hour = TRUE;
    else if (finfo->time_step_minutes >= 1.0) show_minute = TRUE;
    else show_second = TRUE;
    if (finfo->first_time)
    {
	finfo->first_time = FALSE;
	show_hour = TRUE;
    }
    if (show_day)
    {
	if (show_second) show_minute = TRUE;
	if (show_minute) show_hour = TRUE;
    }
    else if (show_hour)
    {
	if (show_second) show_minute = TRUE;
    }
    if (show_day) sprintf (string, "%u\\ud", curr_day_of_month);
    else string[0] = '\0';
    if (show_hour)
    {
	sprintf (txt, "%u\\uh", curr_hour);
	strcat (string, txt);
    }
    if (show_minute)
    {
	sprintf (txt, "%u\\um", curr_minute);
	strcat (string, txt);
    }
    if (show_second)
    {
	sprintf (txt, "%.*f\\us", finfo->precision, sec);
	strcat (string, txt);
    }
}   /*  End Function time_format  */

static void draw_x_marker (KWorldCanvas canvas, KPixCanvas pixcanvas,
			   CONST dressing_type *dressing,
			   CONST corners_type *wcorners,
			   int x_pixels, int y_pixels,
			   int x_offset, int y_offset,
			   double left_lx, double right_lx,
			   double bottom_ly, double top_ly,
			   unsigned long pixel_value,
			   double wx, flag is_major, double wrap_modulus,
			   double *bottom_x, double *top_x)
/*  [SUMMARY] Draw a marker/gridline at constant X.
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas. If this is NULL nothing is drawn.
    <dressing> The dressing information.
    <wcorners> The corners in world co-ordinates.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    <left_lx> The linear co-ordinate of the left boundary.
    <right_lx> The linear co-ordinate of the right boundary.
    <bottom_ly> The linear co-ordinate of the bottom boundary.
    <top_ly> The linear co-ordinate of the top boundary.
    <pixel_value> The pixel value.
    <wx> The horizontal tick position in world co-ordinates.
    <is_major> If TRUE, this is a major tick mark.
    <wrap_modulus> The wrapping modulus.
    <bottom_x> The horizontal tick position of the bottom tick in pixel
    co-ordinates is written here.
    <top_x> The horizontal tick position of the top tick in pixel co-ordinates
    is written here.
    [RETURNS] Nothing.
*/
{
    flag is_increasing = TRUE;
    int tick_size, tick_offset;
    unsigned int count, num_segs;
    double x, y, wy, ystep, dummy;
    double bottom_px, top_px;
    double bottom_wy, top_wy;
    double min_px, max_px, min_py, max_py;
    double x0[GRID_STEPS], y0[GRID_STEPS], x1[GRID_STEPS], y1[GRID_STEPS];

    tick_size = is_major ? MAJOR_TICK_PIXELS : MINOR_TICK_PIXELS;
    if (dressing->internal_ticks) tick_offset = 1;
    else
    {
	tick_offset = -1;
	tick_size = -tick_size;
    }
    /*  Determine lower X linear and Y world co-ordinate  */
    x = wx;
    bottom_wy = bottom_ly;
    canvas_coords_transform (canvas, 1, &x, TRUE, &bottom_wy, FALSE);
    /*  Determine pixel co-ordinate at bottom  */
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &x, &bottom_ly, &bottom_px, &dummy);
    bottom_px = floor (bottom_px + 1e-3);
    *bottom_x = bottom_px;
    /*  Determine upper X linear co-ordinate  */
    x = wx;
    top_wy = top_ly;
    canvas_coords_transform (canvas, 1, &x, TRUE, &top_wy, FALSE);
    /*  Determine pixel co-ordinate at top  */
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &x, &top_ly, &top_px, &dummy);
    top_px = floor (top_px + 1e-3);
    if (top_x != NULL) *top_x = top_px;
    if (pixcanvas == NULL) return;
    if ( dressing->show_bottom_ticks && (!dressing->grid || !is_major) &&
	 (bottom_px >= x_offset - 1) && (bottom_px <= x_offset + x_pixels) )
    {
	kwin_draw_line (pixcanvas,
			bottom_px, y_offset + y_pixels - tick_offset,
			bottom_px, y_offset + y_pixels - tick_offset-tick_size,
			pixel_value);
    }
    /*  Draw upper tick if necessary  */
    if ( dressing->show_top_ticks && (!dressing->grid || !is_major) &&
	 (top_px >= x_offset - 1) && (top_px <= x_offset + x_pixels) )
    {
	kwin_draw_line (pixcanvas,
			top_px, y_offset - 1 + tick_offset,
			top_px, y_offset - 1 + tick_offset + tick_size,
			pixel_value);
    }
    if (!dressing->grid || !is_major) return;
    /*  Determine if co-ordinate is increasing bottom->top  */
    x0[0] = bottom_px;
    y0[0] = y_offset + y_pixels;
    x0[1] = bottom_px;
    y0[1] = y_offset + y_pixels - 1;
    x0[2] = bottom_px;
    y0[2] = y_offset + y_pixels - 2;
    canvas_convert_to_canvas_coords (canvas, FALSE, 3, x0, y0, NULL, NULL,
				     x0, y0);
    if ( fabs (y0[1] - y0[0]) > fabs (y0[2] - y0[1]) )
    {
	if (y0[2] < y0[1]) is_increasing = FALSE;
    }
    else
    {
	if (y0[1] < y0[0]) is_increasing = FALSE;
    }
    min_px = x_offset;
    max_px = (double) (x_offset + x_pixels - 1);
    min_py = (double) y_offset;
    max_py = (double) (y_offset + y_pixels - 1);
    /*  Determine where this grid line leaves box  */
    if ( test_between_limits (wx, wcorners->top_left.x,wcorners->top_right.x) )
    {
	/*  Determine upper Y world co-ordinate  */
	x = wx;
	top_wy = top_ly;
	canvas_coords_transform (canvas, 1, &x, TRUE, &top_wy, FALSE);
    }
    else
    {
	if ( fabs (bottom_wy - wcorners->top_left.y) >
	     fabs (bottom_wy - wcorners->top_right.y) )
	    top_wy = wcorners->top_left.y;
	else top_wy = wcorners->top_right.y;
	/*  Iterate a bit closer to the real departure point  */
	count = 0;
	wy = bottom_wy;
	while (count < BINARY_ITERATIONS)
	{
	    x = wx;
	    y = (top_wy + wy) * 0.5;
	    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					       &x, &y, &x, &y);
	    if ( (x >= min_px) && (x <= max_px) ) wy = (top_wy + wy) * 0.5;
	    else top_wy = (top_wy + wy) * 0.5;
	    ++count;
	}
    }
    min_px -= 1.1;
    max_px += 1.1;
    min_py -= 1.1;
    max_py += 1.1;
    if ( (  is_increasing && (top_wy < bottom_wy) ) ||
	 ( !is_increasing && (bottom_wy < top_wy) ) )
    {
	/*  We've got wrap-around  */
	top_wy += wrap_modulus * (is_increasing ? 1.0 : -1.0);
    }
    ystep = (top_wy - bottom_wy) / (double) GRID_STEPS;
    for (count = 0, wy = bottom_wy; count < GRID_STEPS;
	 wy += ystep, ++count)
    {
	x0[count] = wx;
	y0[count] = wy;
	x1[count] = wx;
	y1[count] = wy + ystep;
    }
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, GRID_STEPS,
				       x0, y0, x0, y0);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, GRID_STEPS,
				       x1, y1, x1, y1);
    for (count = 0, num_segs = 0; count < GRID_STEPS; ++count)
    {
	if ( (x0[count] >= min_px) && (x0[count] < max_px) &&
	     (y0[count] >= min_py) && (y0[count] < max_py) &&
	     (x1[count] >= min_px) && (x1[count] < max_px) &&
	     (y1[count] >= min_py) && (y1[count] < max_py) )
	{
	    if (num_segs != count)
	    {
		x0[num_segs] = x0[count];
		y0[num_segs] = y0[count];
		x1[num_segs] = x1[count];
		y1[num_segs] = y1[count];
	    }
	    ++num_segs;
	}
    }
    kwin_draw_segments_TRANSITION (pixcanvas, x0, y0, x1, y1, num_segs,
				   pixel_value);
}   /*  End Function draw_x_marker  */

static void draw_y_marker (KWorldCanvas canvas, KPixCanvas pixcanvas,
			   CONST dressing_type *dressing,
			   CONST corners_type *wcorners,
			   int x_pixels, int y_pixels,
			   int x_offset, int y_offset,
			   double left_lx, double right_lx,
			   double bottom_ly, double top_ly,
			   unsigned long pixel_value,
			   double wy, flag is_major, double wrap_modulus,
			   double *left_y, double *right_y)
/*  [SUMMARY] Draw a marker/gridline at constant Y.
    <canvas> The world canvas.
    <pixcanvas> The pixel canvas. If this is NULL nothing is drawn.
    <dressing> The dressing information.
    <wcorners> The corners in world co-ordinates.
    <x_pixels> The number of horizontal pixels for the world canvas.
    <y_pixels> The number of vertical pixels for the world canvas.
    <x_offset> The horizontal offset of the world canvas.
    <y_offset> The vertical offset of the world canvas.
    <left_lx> The linear co-ordinate of the left boundary.
    <right_lx> The linear co-ordinate of the right boundary.
    <bottom_ly> The linear co-ordinate of the bottom boundary.
    <top_ly> The linear co-ordinate of the top boundary.
    <pixel_value> The pixel value.
    <wy> The vertical tick position in world co-ordinates.
    <is_major> If TRUE, this is a major tick mark.
    <wrap_modulus> The wrapping modulus.
    <left_y> The vertical tick position of the left tick in pixel co-ordinates
    is written here.
    <right_y> The vertical tick position of the right tick in pixel
    co-ordinates is written here.
    [RETURNS] Nothing.
*/
{
    flag is_increasing = TRUE;
    int tick_size, tick_offset;
    unsigned int count, num_segs;
    double x, y, wx, xstep, dummy;
    double left_py, right_py;
    double left_wx, right_wx;
    double min_py, max_py, min_px, max_px;
    double x0[GRID_STEPS], y0[GRID_STEPS], x1[GRID_STEPS], y1[GRID_STEPS];

    tick_size = is_major ? MAJOR_TICK_PIXELS : MINOR_TICK_PIXELS;
    if (dressing->internal_ticks) tick_offset = 1;
    else
    {
	tick_offset = -1;
	tick_size = -tick_size;
    }
    /*  Determine left X world and Y linear co-ordinate  */
    left_wx = left_lx;
    y = wy;
    canvas_coords_transform (canvas, 1, &left_wx, FALSE, &y, TRUE);
    /*  Determine pixel co-ordinate at left  */
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &left_lx, &y, &dummy, &left_py);
    left_py = floor (left_py + 1e-3);
    *left_y = left_py;
    /*  Determine right Y linear co-ordinate  */
    right_wx = right_lx;
    y = wy;
    canvas_coords_transform (canvas, 1, &right_wx, FALSE, &y, TRUE);
    /*  Determine pixel co-ordinate at right  */
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &right_lx, &y, &dummy, &right_py);
    right_py = floor (right_py + 1e-3);
    if (right_y != NULL) *right_y = right_py;
    if (pixcanvas == NULL) return;
    if ( dressing->show_left_ticks && (!dressing->grid || !is_major) &&
	 (left_py >= y_offset - 1) && (left_py <= y_offset + y_pixels) )
    {
	kwin_draw_line (pixcanvas,
			x_offset - 1 + tick_offset, left_py,
			x_offset - 1 + tick_offset + tick_size, left_py,
			pixel_value);
    }
    /*  Draw right tick if necessary  */
    if ( dressing->show_right_ticks && (!dressing->grid || !is_major) &&
	 (right_py >= y_offset - 1) && (right_py <= y_offset + y_pixels) )
    {
	kwin_draw_line (pixcanvas,
			x_offset + x_pixels - tick_offset, right_py,
			x_offset + x_pixels - tick_offset -tick_size, right_py,
			pixel_value);
    }
    if (!dressing->grid || !is_major) return;
    /*  Determine if co-ordinate is increasing left->right  */
    x0[0] = x_offset;
    y0[0] = left_py;
    x0[1] = x_offset + 1;
    y0[1] = left_py;
    x0[2] = x_offset + 2;
    y0[2] = left_py;
    canvas_convert_to_canvas_coords (canvas, FALSE, 3, x0, y0, NULL, NULL,
				     x0, y0);
    if ( fabs (x0[1] - x0[0]) > fabs (x0[2] - x0[1]) )
    {
	if (x0[2] < x0[1]) is_increasing = FALSE;
    }
    else
    {
	if (x0[1] < x0[0]) is_increasing = FALSE;
    }
    min_px = x_offset;
    max_px = (double) (x_offset + x_pixels - 1);
    min_py = (double) y_offset;
    max_py = (double) (y_offset + y_pixels - 1);
    /*  Determine where this grid line leaves box  */
    if ( test_between_limits (wy, wcorners->bottom_right.y,
			      wcorners->top_right.y) )
    {
	/*  Determine right X world co-ordinate  */
	right_wx = right_lx;
	y = wy;
	canvas_coords_transform (canvas, 1, &right_wx, FALSE, &y, TRUE);
    }
    else
    {
	if ( fabs (left_wx - wcorners->bottom_right.x) >
	     fabs (left_wx - wcorners->top_right.x) )
	    right_wx = wcorners->bottom_right.x;
	else right_wx = wcorners->top_right.x;
	/*  Iterate a bit closer to the real departure point  */
	count = 0;
	wx = left_wx;
	while (count < BINARY_ITERATIONS)
	{
	    x = (right_wx + wx) * 0.5;
	    y = wy;
	    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					       &x, &y, &x, &y);
	    if ( (y >= min_py) && (y <= max_py) )
		wx = (right_wx + wx) * 0.5;
	    else right_wx = (right_wx + wx) * 0.5;
	    ++count;
	}
    }
    min_px -= 1.1;
    max_px += 1.1;
    min_py -= 1.1;
    max_py += 1.1;
    if ( (  is_increasing && (right_wx < left_wx) ) ||
	 ( !is_increasing && (left_wx < right_wx) ) )
    {
	/*  We've got wrap-around  */
	right_wx += wrap_modulus * (is_increasing ? 1.0 : -1.0);
    }
    xstep = (right_wx - left_wx) / (double) GRID_STEPS;
    for (count = 0, wx = left_wx; count < GRID_STEPS; wx += xstep, ++count)
    {
	x0[count] = wx;
	y0[count] = wy;
	x1[count] = wx + xstep;
	y1[count] = wy;
    }
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, GRID_STEPS,
				       x0, y0, x0, y0);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, GRID_STEPS,
				       x1, y1, x1, y1);
    for (count = 0, num_segs = 0; count < GRID_STEPS; ++count)
    {
	if ( (x0[count] >= min_px) && (x0[count] < max_px) &&
	     (y0[count] >= min_py) && (y0[count] < max_py) &&
	     (x1[count] >= min_px) && (x1[count] < max_px) &&
	     (y1[count] >= min_py) && (y1[count] < max_py) )
	{
	    if (num_segs != count)
	    {
		x0[num_segs] = x0[count];
		y0[num_segs] = y0[count];
		x1[num_segs] = x1[count];
		y1[num_segs] = y1[count];
	    }
	    ++num_segs;
	}
    }
    kwin_draw_segments_TRANSITION (pixcanvas, x0, y0, x1, y1, num_segs,
				   pixel_value);
}   /*  End Function draw_y_marker  */

static double get_time_step (double range)
/*  [SUMMARY] Get time step.
    <range> The range of the data in hours.
    [RETURNS] The number of hours per step.
*/
{
    double step;

    if      (range <  1.0) step = get_ms_step (range);
    else if (range <  1.5) step =  0.25;
    else if (range <  3.0) step =  0.5;
    else if (range <  6.0) step =  1.0;
    else if (range < 12.0) step =  2.0;
    else if (range < 24.0) step =  4.0;
    else if (range < 36.0) step =  6.0;
    else                   step = get_decimal_step (range / 24.0) * 24.0;
    return (step);
}   /*  End Function get_time_step  */

static double get_angle_step (double range)
/*  [SUMMARY] Get angle step.
    <range> The range of the data in degrees.
    [RETURNS] The number of degrees per step.
*/
{
    double step;

    if      (range <  1.0) step = get_ms_step (range);
    else if (range < 60.0) step = get_ms_step (range / 60.0) * 60.0;
    else                   step = get_ms_step (range / 3600.0) * 3600.0;
    return (step);
}   /*  End Function get_angle_step  */

static double get_ms_step (double range)
/*  [SUMMARY] Get minutes and seconds step.
    <range> The range of the data in degrees/hours.
    [RETURNS] The number of degrees/hours per step.
*/
{
    double step;
    double factor = 60.0;

    if (range >= 1.0) return (0.5);
    if (range * factor < 1.0)
    {
	factor = 3600.0;
	if (range * factor < 1.0)
	    return get_decimal_step (range * factor) / factor;
    }
    range *= factor;
    if      (range >= 60.0) step = 20.0;
    else if (range >= 30.0) step = 10.0;
    else if (range >= 15.0) step =  5.0;
    else if (range >=  6.0) step =  2.0;
    else if (range >=  3.0) step =  1.0;
    else if (range >=  1.5) step =  0.5;
    else                    step =  0.25;
    return (step / factor);
}   /*  End Function get_ms_step  */

static double get_decimal_step (double range)
/*  [SUMMARY] Get a decimal step.
    <range> The range of the data.
    [RETURNS] The step value.
*/
{
    double factor, step;

    factor = pow ( 10.0, -floor ( log10 (range) ) );
    range *= factor;
    if (range >= 6.0)      step = 2.0;
    else if (range >= 3.0) step = 1.0;
    else if (range >= 1.5) step = 0.5;
    else                   step = 0.2;
    return (step / factor);
}   /*  End Function get_decimal_step  */

static flag test_between_limits (double val, double limit1, double limit2)
/*  [SUMMARY] Test if a value lies between two limits.
    <val> The value to test.
    <limit1> One of the limits.
    <limit2> The other limit.
    [RETURNS] TRUE if the value lies between the limits, else FALSE.
*/
{
    if (limit1 < limit2)
    {
	return ( (val >= limit1) && (val <= limit2) ) ? TRUE : FALSE;
    }
    return ( (val >= limit2) && (val <= limit1) ) ? TRUE : FALSE;
}   /*  End Function test_between_limits  */
