/*LINTLIBRARY*/
/*  main.c

    This code provides KContourImage objects.

    Copyright (C) 1996-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating contour images.


    Written by      Richard Gooch   20-JUL-1996

    Updated by      Richard Gooch   22-JUL-1996: Created
  <contour_create_sequence_from_iarray>.

    Updated by      Richard Gooch   4-SEP-1996: More than one contour image may
  be active per canvas.

    Updated by      Richard Gooch   5-SEP-1996: Optimisations when number of
  segments less than NUM_SECTIONS. Fixed bug where number of pixel segments
  less than NUM_SECTIONS was not trapped.

    Updated by      Richard Gooch   13-OCT-1996: Trapped NULL <<pixel_values>>
  in <contour_set_levels>.

    Updated by      Richard Gooch   15-OCT-1996: Trapped NULL
  <<contour_levels>> in several places.

    Updated by      Richard Gooch   15-JAN-1997: Switched to
  <wcs_astro_transform3> in order to deal with non RA/DEC axes.

    Updated by      Richard Gooch   17-JAN-1997: Created
  <contour_set_attributes>.

    Updated by      Richard Gooch   27-JAN-1997: Make use of restriction
  information when doing astronomical projections. Do quick check to see if
  canvas X and Y labels match.

    Updated by      Richard Gooch   29-JAN-1997: Only create a single KwcsAstro
  object for a sequence.

    Updated by      Richard Gooch   1-FEB-1997: Made use of new
  <wcs_astro_get_associated> function.

    Updated by      Richard Gooch   2-FEB-1997: Removed quick check for
  matching X and Y labels: this would reject similar axes which have different
  names (i.e. "RA---SIN" and "RA---TAN" or "Axis 1").

    Updated by      Richard Gooch   11-JUL-1997: Separated caching and drawing
  of levels to support later definition of level attributes (i.e. line styles).

    Updated by      Richard Gooch   21-JUL-1997: Made use of new precompute
  event handlers.

    Updated by      Richard Gooch   31-JUL-1997: Took account of change to
  precompute function interface: pointer to <honoured_areas> flag dropped.

    Updated by      Richard Gooch   11-OCT-1997: Added dashing support.

    Updated by      Richard Gooch   29-OCT-1997: Hacked dashing support to
  use grey instead.

    Updated by      Richard Gooch   18-NOV-1997: Fixed bug where segments
  outside canvas could be drawn when number of segments was low.

    Updated by      Richard Gooch   19-NOV-1997: Fixed bug where segments with
  one end outside canvas could be drawn.

    Updated by      Richard Gooch   7-DEC-1997: Added linewidth support.

    Updated by      Richard Gooch   5-MAR-1998: Removed unnecessary cast in
  <_contour_precompute_image>.

    Updated by      Richard Gooch   6-MAR-1998: Removed casts for memory
  functions. Fixed bug in <contour_destroy> which was deallocating levels which
  are also deallocated in <dealloc_levels>.

    Updated by      Richard Gooch   16-SEP-1998: Added grouping key.

    Updated by      Richard Gooch   17-SEP-1998: Added contour image resource
  CONTOUR_CIMAGE_ATT_COLOURNAME. Changed from inserting new image at front of
  list to appending to end of list.

    Updated by      Richard Gooch   5-JUL-1999: Removed sequence holder.
  Created freelist.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   11-SEP-1999: Check for failure return from
  <ds_contour>.

    Updated by      Richard Gooch   27-JUN-2000: Made use of new function
  <canvas_register_destroy_func>. Bugfix in <contour_set_active>: was not
  forcing refresh if other images deactivated and <<refresh_if_changed>> TRUE.

    Updated by      Richard Gooch   7-JUL-2000: Took account of interface
  change to <canvas_register_destroy_func>.

    Last updated by Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>


*/
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdarg.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <karma_contour.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>


#define CIMAGE_USED_MAGIC_NUMBER (unsigned int) 1342908534
#define CIMAGE_FREE_MAGIC_NUMBER (unsigned int) 1349879874
#define HOLDER_MAGIC_NUMBER (unsigned int) 1653453498
#define DEFAULT_COLOUR "green"
#define NUM_SECTIONS 64

#define VERIFY_HOLDER(holder) {if (holder == NULL) \
{fprintf (stderr, "NULL canvas holder passed\n"); \
 a_prog_bug (function_name); } \
if (holder->magic_number != HOLDER_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas holder object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_USED_CIMAGE(cimage) if (cimage == NULL) \
{fprintf (stderr, "NULL contourable image passed\n"); \
 a_prog_bug (function_name); } \
if (cimage->magic_number != CIMAGE_USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid contourable image object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FREE_CIMAGE(cimage) if (cimage == NULL) \
{fprintf (stderr, "NULL contourable image passed\n"); \
 a_prog_bug (function_name); } \
if (cimage->magic_number != CIMAGE_FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed contourable image object\n"); \
 a_prog_bug (function_name); }


typedef struct canvas_holder_type * CanvasHolder;

struct canvas_holder_type
{
    unsigned int magic_number;
    KWorldCanvas canvas;
    KContourImage first_image;
    KContourImage last_image;
    CanvasHolder prev;
    CanvasHolder next;
    KCallbackList position_list;
    char *colourname;
};

struct section_limits_type
{
    double xmin[NUM_SECTIONS];
    double xmax[NUM_SECTIONS];
    double ymin[NUM_SECTIONS];
    double ymax[NUM_SECTIONS];
};

struct levelinfo_type
{
    uaddr world_coord_buf_size;
    uaddr num_segments;
    unsigned long pixel_value;
    flag pixel_defined;
    flag dash;
    double linewidth;
    uaddr cache_buf_size;
    double *linear_x0;
    double *linear_y0;
    double *linear_x1;
    double *linear_y1;
    struct section_limits_type linear_limits;
    unsigned int num_pixel_segments;
    double *pix_x0;
    double *pix_y0;
    double *pix_x1;
    double *pix_y1;
    struct section_limits_type pix_limits;
    unsigned int num_area_segments;
    double *pix_area_x0;
    double *pix_area_y0;
    double *pix_area_x1;
    double *pix_area_y1;
};

struct contourableimage_type
{
    unsigned int magic_number;
    CanvasHolder canvas_holder;
    flag active;
    /*  Image data  */
    multi_array *multi_desc;
    array_desc *arr_desc;
    char *slice;
    unsigned int hdim;
    unsigned int vdim;
    unsigned int elem_index;
    KwcsAstro astro_projection;
    /*  Contour levels  */
    unsigned int num_levels;
    struct levelinfo_type *levels;
    double *contour_levels;
    /*  Other information  */
    char *colourname;
    flag world_segs_valid; /*Whether the world co-ordinate segments are valid*/
    unsigned int num_restrictions;
    char **restriction_names;
    double *restriction_values;
    char *inline_restr_names[1];
    double inline_restr_values[1];
    uaddr key;                      /*  Key for grouping  */
    /*  Cached segments  */
    double **world_x0;
    double **world_y0;
    double **world_x1;
    double **world_y1;
    int old_x_offset;
    int old_y_offset;
    int old_x_pixels;
    int old_y_pixels;
    double old_left_x;
    double old_right_x;
    double old_bottom_y;
    double old_top_y;
    double test_x[5];
    double test_y[5];
    KContourImage prev;
    KContourImage next;
};


/*  Private data  */
static CanvasHolder first_canvas_holder = NULL;
static KContourImage first_free = NULL;


/*  External functions  */
EXTERN_FUNCTION (KCallbackFunc canvas_register_precompute_func,
		 (KWorldCanvas canvas, void (*refresh_func) (), void *info) );


/*  Private functions  */
STATIC_FUNCTION (KContourImage create_restr,
		 (CanvasHolder holder, multi_array *multi_desc, flag try_ap,
		  array_desc *arr_desc, char *slice,
		  unsigned int hdim, unsigned int vdim,
		  unsigned int elem_index,
		  unsigned int num_levels,
		  CONST double *contour_levels,
		  unsigned num_restr, char **restr_names, double *restr_values,
		  uaddr key) );
STATIC_FUNCTION (unsigned int get_gen_index,
		 (CONST multi_array *multi_desc, CONST array_desc *arr_desc) );
STATIC_FUNCTION (CanvasHolder get_canvas_holder,
		 (KWorldCanvas canvas, char *func_name) );
STATIC_FUNCTION (CanvasHolder alloc_canvas_holder, (KWorldCanvas canvas) );
STATIC_FUNCTION (void _contour_free_canvas_holder, (CanvasHolder holder) );
STATIC_FUNCTION (void process_canvas_attributes,
		 (CanvasHolder holder, va_list argp) );
STATIC_FUNCTION (void _contour_worldcanvas_precompute_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _contour_precompute_image,
		 (KContourImage cimage, struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void precompute_level,
		 (KContourImage cimage, unsigned int level_index,
		  struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag same_trans, flag same_world) );
STATIC_FUNCTION (void _contour_compute_lin_segs,
		 (KContourImage cimage, unsigned int level_index,
		  struct win_scale_type *win_scale,
		  double win_x_min, double win_y_min,
		  double win_x_max, double win_y_max) );
STATIC_FUNCTION (void _contour_compute_pix_segs,
		 (KContourImage cimage, unsigned int level_index,
		  struct win_scale_type *win_scale,
		  double win_x_min, double win_y_min,
		  double win_x_max, double win_y_max) );
STATIC_FUNCTION (void _contour_worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void refresh_image,
		 (KContourImage cimage, struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void draw_level,
		 (KContourImage cimage, unsigned int level_index,
		  struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void compute_pixel_coords,
		 (KWorldCanvas canvas, struct levelinfo_type *level,
		  struct win_scale_type *win_scale,
		  unsigned int num_pix_segs) );
STATIC_FUNCTION (void _contour_dealloc_levels, (KContourImage cimage) );


/* Public functions follow */

/*PUBLIC_FUNCTION*/
void contour_init (KWorldCanvas canvas, ...)
/*  [SUMMARY] Initialise the package for a particular canvas.
    [PURPOSE] This routine will initialise the [<contour>] package for a
    particular world canvas. Calling this routine causes a number of callback
    routines internal to the package to be registered with the canvas (such
    as refresh and position event callbacks). This routine must be called
    before creating contour images.
    <canvas> The world canvas object.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the CONTOUR_CANVAS_ATT_END.
    See [<contour_CANVAS_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    CanvasHolder holder;
    static char function_name[] = "contour_init";

    va_start (argp, canvas);
    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (get_canvas_holder (canvas, NULL) != NULL)
    {
	fprintf (stderr, "Canvas already initialised\n");
	a_prog_bug (function_name);
    }
    if ( ( holder = alloc_canvas_holder (canvas) ) == NULL )
	m_abort (function_name, "canvas holder");
    process_canvas_attributes (holder, argp);
    va_end (argp);
}   /*  End Function contour_init  */

/*PUBLIC_FUNCTION*/
KContourImage contour_create_restr (KWorldCanvas canvas,
				    multi_array *multi_desc,
				    array_desc *arr_desc, char *slice,
				    unsigned int hdim, unsigned int vdim,
				    unsigned int elem_index,
				    unsigned int num_levels,
				    CONST double *contour_levels,
				    unsigned num_restr,
				    char **restr_names, double *restr_values,
				    uaddr key)
/*  [SUMMARY] Create contour image from 2D slice with restrictions.
    [PURPOSE] This routine will create a contour image object from a
    2-dimensional slice of a Karma data structure. At a later time, this
    contour image may be made visible. This routine will not cause the canvas
    to be refreshed.
    <canvas> The world canvas onto which the contour image may be drawn. Many
    contour images may be associated with a single canvas.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <slice> The start of the slice data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index in the data packets.
    <num_levels> The number of contour levels. This may be 0.
    <contour_levels> The array of contour levels.
    <num_restr> The number of matched restrictions.
    <restr_names> The restriction names.
    <restr_values> The restriction values.
    [NOTE] Restriction information is automatically deallocated when
    [<contour_destroy>] is called.
    <key> The key to associate with this contour image.
    [NOTE] The KContourImage will be automatically destroyed if the canvas is
    destroyed.
    [RETURNS] A KContourImage object on success, NULL.
*/
{
    CanvasHolder holder;
    static char function_name[] = "contour_create_restr";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (slice == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Sanity checks  */
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates != NULL)
    {
	fprintf (stderr, "hdim: %u not regularly spaced\n", hdim);
	a_prog_bug (function_name);
    }
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates != NULL)
    {
	fprintf (stderr, "vdim: %u not regularly spaced\n", vdim);
	a_prog_bug (function_name);
    }
    if (elem_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    holder = get_canvas_holder (canvas, function_name);
    return create_restr (holder, multi_desc, TRUE, arr_desc, slice,
			 hdim, vdim, elem_index, num_levels, contour_levels,
			 num_restr, restr_names, restr_values, key);
}   /*  End Function contour_create_restr  */

/*PUBLIC_FUNCTION*/
KContourImage contour_create_from_iarray (KWorldCanvas canvas, iarray array,
					  flag swap, unsigned int num_levels,
					  CONST double *contour_levels,
					  uaddr key)
/*  [SUMMARY] Create a contour image from an Intelligent Array.
    [PURPOSE] This routine will create a contour image object from a
    2-dimensional Intelligant Array. At a later time, this contour image may
    be made visible. This routine will not cause the canvas to be refreshed.
    Many contour images may be associated with a single canvas.
    <canvas> The world canvas object.
    <array> The Intelligent Array. The underlying <<multi_array>> data
    structure will have its attachment count incremented upon successful
    completion.
    <swap> If TRUE the y axis will be displayed horizontally.
    <num_levels> The number of contour levels. This may be 0.
    <contour_levels> The array of contour levels.
    <key> The key to associate with this contour image.
    [NOTE] The KContourImage will be automatically destroyed if the canvas is
    destroyed.
    [RETURNS] A KContourImage object on success, else NULL.
*/
{
    CanvasHolder holder;
    unsigned int num_restr;
    char **restr_names;
    double *restr_values;
    static char function_name[] = "contour_create_from_iarray";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (array == NULL)
    {
	fprintf (stderr, "NULL Intelligent Array passed\n");
	a_prog_bug (function_name);
    }
    FLAG_VERIFY (swap);
    if (iarray_num_dim (array) != 2)
    {
	fprintf (stderr, "Intelligent Array must have exactly 2 dimensions\n");
	return (NULL);
    }
    if ( (array->offsets[0] != array->arr_desc->offsets[0]) ||
	 (array->offsets[1] != array->arr_desc->offsets[1]) )
    {
	fprintf (stderr, "Intelligent Array must not be a sub-array\n");
	return (NULL);
    }
    holder = get_canvas_holder (canvas, function_name);
    num_restr = iarray_get_restrictions (array, &restr_names, &restr_values);
    if (swap)
    {
	return create_restr (holder, array->multi_desc, TRUE,
			     array->arr_desc, array->data,
			     array->orig_dim_indices[0],
			     array->orig_dim_indices[1],
			     array->elem_index,
			     num_levels, contour_levels,
			     num_restr, restr_names, restr_values, key);
    }
    else
    {
	return create_restr (holder, array->multi_desc, TRUE,
			     array->arr_desc, array->data,
			     array->orig_dim_indices[1],
			     array->orig_dim_indices[0],
			     array->elem_index,
			     num_levels, contour_levels,
			     num_restr, restr_names, restr_values, key);
    }
}   /*  End Function contour_create_from_iarray  */

/*PUBLIC_FUNCTION*/
KContourImage *contour_create_sequence (KWorldCanvas canvas,
					multi_array *multi_desc,
					array_desc *arr_desc, char *cube,
					unsigned int hdim, unsigned int vdim,
					unsigned int fdim,
					unsigned int elem_index,
					unsigned int num_levels,
					CONST double *contour_levels,
					uaddr key)
/*  [SUMMARY] Create a sequence of contour images from a 3D slice.
    [PURPOSE] This routine will create a sequence of contour image objects
    from a 3-dimensional cube of a Karma data structure. At a later time, this
    sequence of contour images may be made visible in any order.
    This routine will not cause the canvas to be refreshed.
    <canvas> The world canvas object.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <cube> The start of the cube data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <fdim> The dimension index of the frame dimension (dimension containing the
    sequence). The number of frames is the same as the length of this
    dimension.
    <elem_index> The element index in the data packets.
    <num_levels> The number of contour levels. This may be 0.
    <contour_levels> The array of contour levels.
    <key> The key to associate with this contour image.
    [NOTE] The KContourImage objects will be automatically destroyed if the
    canvas is destroyed.
    [RETURNS] A pointer to a dynamically allocated array of contour image
    objects on success, else NULL.
*/
{
    KwcsAstro ap;
    KContourImage cimage;
    CanvasHolder holder;
    unsigned int num_frames, count, index;
    KContourImage *cimages;
    uaddr *foffsets;
    static char function_name[] = "contour_create_sequence";

    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (cube == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Sanity checks  */
    if (fdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "fdim: %u greater than number of dimensions: %u\n",
		 fdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    /*  Sanity checks  */
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates != NULL)
    {
	fprintf (stderr, "hdim: %u not regularly spaced\n", hdim);
	a_prog_bug (function_name);
    }
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates != NULL)
    {
	fprintf (stderr, "vdim: %u not regularly spaced\n", vdim);
	a_prog_bug (function_name);
    }
    if (elem_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    holder = get_canvas_holder (canvas, function_name);
    foffsets = arr_desc->offsets[fdim];
    num_frames = arr_desc->dimensions[fdim]->length;
    if ( ( cimages = m_alloc (sizeof *cimages * num_frames) ) == NULL )
    {
	m_error_notify (function_name, "array of contourable images");
	return (NULL);
    }
    if (multi_desc == NULL) ap = NULL;
    else
    {
	index = get_gen_index (multi_desc, arr_desc);
	ap = wcs_astro_get_associated (multi_desc, index);
    }
    for (count = 0; count < num_frames; ++count)
    {
	if ( ( cimage =
	       create_restr (holder, multi_desc, FALSE, arr_desc,
			     cube + foffsets[count],
			     hdim, vdim, elem_index,
			     num_levels, contour_levels,
			     1, NULL, NULL, key) )
	    == NULL )
	{
	    for (; count > 0; --count)
	    {
		contour_destroy (cimages[count - 1]);
	    }
	    m_free (cimages);
	    return (NULL);
	}
	cimages[count] = cimage;
	cimages[count]->astro_projection = ap;
	cimage->num_restrictions = 1;
	cimage->inline_restr_names[0] = arr_desc->dimensions[fdim]->name;
	cimage->inline_restr_values[0] =
	    ds_get_coordinate (arr_desc->dimensions[fdim], count);
	cimage->restriction_names = cimage->inline_restr_names;
	cimage->restriction_values = cimage->inline_restr_values;
    }
    return (cimages);
}   /*  End Function contour_create_sequence  */

/*PUBLIC_FUNCTION*/
KContourImage *contour_create_sequence_from_iarray
    (KWorldCanvas canvas, iarray array,
     unsigned int hdim, unsigned int vdim, unsigned int fdim,
     unsigned int num_levels, CONST double *contour_levels, uaddr key)
/*  [SUMMARY] Create a sequence of contour images from an Intelligent Array.
    [PURPOSE] This routine will create a sequence of contour image objects
    from a 3-dimensional Intelligent Array. At a later time, this sequence of
    contour images may be made visible in any order.
    This routine will not cause the canvas to be refreshed.
    <canvas> The world canvas object.
    <array> The Intelligent Array. The underlying <<multi_array>> data
    structure will have its attachment count incremented upon successful
    completion.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <fdim> The dimension index of the frame dimension (dimension containing the
    sequence). The number of frames is the same as the length of this
    dimension.
    <num_levels> The number of contour levels. This may be 0.
    <contour_levels> The array of contour levels.
    <key> The key to associate with this contour image.
    [NOTE] The KContourImage objects will be automatically destroyed if the
    canvas is destroyed.
    [RETURNS] A pointer to a dynamically allocated array of contour image
    objects on success, else NULL.
*/
{
    static char function_name[] = "contour_create_sequence_from_iarray";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (array == NULL)
    {
	fprintf (stderr, "NULL Intelligent Array passed\n");
	a_prog_bug (function_name);
    }
    if (iarray_num_dim (array) != 3)
    {
	fprintf (stderr, "Intelligent Array must have exactly 3 dimensions\n");
	return (NULL);
    }
    if ( hdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "hdim: %u too large\n", hdim);
	a_prog_bug (function_name);
    }
    if ( vdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "vdim: %u too large\n", vdim);
	a_prog_bug (function_name);
    }
    if ( fdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "fdim: %u too large\n", fdim);
	a_prog_bug (function_name);
    }
    if ( (array->offsets[fdim] != array->arr_desc->offsets[fdim]) ||
	 (array->offsets[hdim] != array->arr_desc->offsets[hdim]) ||
	 (array->offsets[hdim] != array->arr_desc->offsets[hdim]) )
    {
	fprintf (stderr, "Intelligent Array must not be a sub-array\n");
	return (NULL);
    }
    return contour_create_sequence (canvas, array->multi_desc,
				    array->arr_desc, array->data,
				    array->orig_dim_indices[hdim],
				    array->orig_dim_indices[vdim],
				    array->orig_dim_indices[fdim],
				    array->elem_index,
				    num_levels, contour_levels, key);
}   /*  End Function contour_create_sequence_from_iarray  */

/*PUBLIC_FUNCTION*/
flag contour_set_active (KContourImage cimage, flag active, flag force_refresh,
			 flag refresh_if_changed, flag exclusive, uaddr key)
/*  [SUMMARY] Set active state for a contour image controlled refresh.
    [PURPOSE] This routine will make a contour image active or inactive.
    <cimage> The contour image.
    <active> If TRUE, the contourable image is made active, else it is made
    inactive.
    <force_refresh> If TRUE, the canvas is always refreshed.
    <refresh_if_changed> If TRUE, the canvas is refreshed if the contourable
    image active state changed.
    <exclusive> If TRUE and <<active>> is TRUE, make this contour image the
    only active one for this canvas. Other contour images for this canvas will
    be made inactive.
    <key> The key to use for inactivation of other contour images. If this is 0
    then all other contour images will be made inactive.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    KContourImage curr;
    static char function_name[] = "contour_set_active";

    VERIFY_USED_CIMAGE (cimage);
    FLAG_VERIFY (active);
    FLAG_VERIFY (force_refresh);
    FLAG_VERIFY (refresh_if_changed);
    holder = cimage->canvas_holder;
    if (active != cimage->active)
    {
	cimage->active = active;
	if (refresh_if_changed) force_refresh = TRUE;
    }
    if (exclusive && active)
    {
	/*  Make others inactive  */
	for (curr = holder->first_image; curr != NULL; curr = curr->next)
	{
	    if (curr == cimage) continue;
	    if ( (key == 0) || (key == curr->key) )
	    {
		if (curr->active && refresh_if_changed) force_refresh = TRUE;
		curr->active = FALSE;
	    }
	}
    }
    if (force_refresh) return canvas_resize (holder->canvas, NULL, FALSE);
    return (TRUE);
}   /*  End Function contour_set_active  */

/*PUBLIC_FUNCTION*/
flag contour_set_all_inactive (KWorldCanvas canvas, flag force_refresh,
			       flag refresh_if_changed, uaddr key)
/*  [SUMMARY] Clear active state all contour images controlled refresh.
    [PURPOSE] This routine will make all contour images associated with a
    world canvas inactive.
    <canvas> The world canvas.
    <force_refresh> If TRUE, the canvas is always refreshed.
    <refresh_if_changed> If TRUE, the canvas is refreshed if the active state
    of any contourable image changed.
    <key> The key to use for inactivation of contour images. If this is 0 then
    all contour images will be made inactive.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    KContourImage curr;
    static char function_name[] = "contour_set_all_inactive";

    holder = get_canvas_holder (canvas, function_name);
    FLAG_VERIFY (force_refresh);
    FLAG_VERIFY (refresh_if_changed);
    /*  Make all inactive  */
    for (curr = holder->first_image; curr != NULL; curr = curr->next)
    {
	if ( (key != 0) && (key != curr->key) ) continue;
	if (curr->active && refresh_if_changed) force_refresh = TRUE;
	curr->active = FALSE;
    }
    if (force_refresh) return canvas_resize (holder->canvas, NULL, FALSE);
    return (TRUE);
}   /*  End Function contour_set_all_inactive  */

/*PUBLIC_FUNCTION*/
flag contour_register_data_change (KContourImage cimage)
/*  [SUMMARY] Notify data for contour image has changed.
    [PURPOSE] This routine will register a change in the Karma data structure
    associated with a contour image. If the contour image is active, it will
    be immediately redrawn on its canvas.
    <cimage> The contour image.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    static char function_name[] = "contour_register_data_change";

    VERIFY_USED_CIMAGE (cimage);
    holder = cimage->canvas_holder;
    cimage->world_segs_valid = FALSE;
    if (cimage->active)
    {
	/*  Active image: refresh  */
	return canvas_resize (holder->canvas, NULL, FALSE);
    }
    return (TRUE);
}   /*  End Function contour_register_data_change  */

/*PUBLIC_FUNCTION*/
void contour_destroy (KContourImage cimage)
/*  [SUMMARY] Destroy contour image.
    [PURPOSE] This routine will destroy a contour image.
    <cimage> The contour image.
    [NOTE] The associated <<multi_array>> descriptor is also deallocated (or
    at least, the attachment count is decreased).
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    unsigned int count;
    static char function_name[] = "contour_destroy";

    VERIFY_USED_CIMAGE (cimage);
    holder = cimage->canvas_holder;
    ds_dealloc_multi (cimage->multi_desc);
    _contour_dealloc_levels (cimage);
    /*  Remove entry  */
    if (cimage->prev == NULL) holder->first_image = cimage->next;
    else cimage->prev->next = cimage->next;
    if (cimage->next == NULL) holder->last_image = cimage->prev;
    else cimage->next->prev = cimage->prev;
    if ( (cimage->restriction_names != NULL) &&
	 (cimage->restriction_names != cimage->inline_restr_names) )
    {
	for (count = 0; count < cimage->num_restrictions; ++count)
	{
	    if (cimage->restriction_names[count] != NULL)
		m_free (cimage->restriction_names[count]);
	}
	m_free (cimage->restriction_names);
    }
    if ( (cimage->restriction_values != NULL ) &&
	 (cimage->restriction_values != cimage->inline_restr_values) )
    {
	m_free (cimage->restriction_values);
    }
    cimage->magic_number = CIMAGE_FREE_MAGIC_NUMBER;
    cimage->next = first_free;
    first_free = cimage;
}   /*  End Function contour_destroy  */

/*PUBLIC_FUNCTION*/
void contour_set_canvas_attributes (KWorldCanvas canvas, ...)
/*  [SUMMARY] Set the contour image attributes for a world canvas.
    <canvas> The world canvas.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the CONTOUR_CANVAS_ATT_END.
    See [<contour_CANVAS_ATTRIBUTES>] for the list of attributes.
    [NOTE] The canvas is not refreshed by this operation.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    va_list argp;
    static char function_name[] = "contour_set_canvas_attributes";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL canvas passed\n");
	a_prog_bug (function_name);
    }
    /*  Dummy operation to ensure canvas is OK  */
    canvas_get_attributes (canvas, CANVAS_ATT_END);
    holder = get_canvas_holder (canvas, function_name);
    va_start (argp, canvas);
    process_canvas_attributes (holder, argp);
    va_end (argp);
}   /*  End Function contour_set_canvas_attributes  */

/*PUBLIC_FUNCTION*/
void contour_set_attributes (KContourImage cimage, ...)
/*  [SUMMARY] Set the attributes for a KContourImage object.
    <cimage> The KContourImage object.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the CONTOUR_CIMAGE_ATT_END.
    See [<contour_CONTOUR_ATTRIBUTES>] for the list of attributes.
    [NOTE] The canvas is not refreshed by this operation.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    flag bool;
    unsigned int att_key;
    char *colour;
    static char function_name[] = "contour_set_attributes";

    VERIFY_USED_CIMAGE (cimage);
    va_start (argp, cimage);
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    CONTOUR_CIMAGE_ATT_END )
    {
	switch (att_key)
	{
	  case CONTOUR_CIMAGE_ATT_ACTIVE:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    cimage->active = bool;
	    break;
	  case CONTOUR_CIMAGE_ATT_COLOURNAME:
	    colour = va_arg (argp, char *);
	    if (colour != NULL)
	    {
		colour = st_dup (colour);
		if (colour == NULL) m_abort (function_name, "colour name");
	    }
	    if (cimage->colourname != NULL) m_free (cimage->colourname);
	    cimage->colourname = colour;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (argp);
}   /*  End Function contour_set_attributes  */

/*PUBLIC_FUNCTION*/
void contour_set_levels (KContourImage cimage, unsigned int num_levels,
			 CONST double *contour_levels,
			 CONST unsigned long *contour_pixels, CONST flag *dash,
			 CONST double *linewidths)
/*  [SUMMARY] Set/update the contour levels for a KContourImage object.
    [PURPOSE] This routine will set/update the contour levels for a
    KContourImage object. The canvas is not refreshed by this operation.
    <cimage> The KContourImage object.
    <num_levels> The number of contour levels.
    <contour_levels> The array of contour levels.
    <contour_pixels> The array of contour pixels. This may be NULL.
    <dash> The array of dash flags. This may be NULL.
    <linewidths> The array of line widths. This may be NULL.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    static char function_name[] = "contour_set_levels";

    VERIFY_USED_CIMAGE (cimage);
    if (num_levels == cimage->num_levels)
    {
	if ( (contour_levels == NULL) || (cimage->contour_levels == NULL) ||
	     !m_cmp (cimage->contour_levels, contour_levels,
		     sizeof *contour_levels * num_levels) )
	    cimage->world_segs_valid = FALSE;
	m_copy (cimage->contour_levels, contour_levels,
		sizeof *contour_levels * num_levels);
    }
    else
    {
	_contour_dealloc_levels (cimage);
	if ( ( cimage->contour_levels =
	       m_dup (contour_levels, sizeof *contour_levels * num_levels) )
	     == NULL )
	{
	    m_abort (function_name, "contour levels");
	}
	if ( ( cimage->world_x0 =
	       m_alloc (sizeof *cimage->world_x0 * num_levels) ) == NULL )
	{
	    m_abort (function_name, "contour world x0 array");
	}
	if ( ( cimage->world_y0 =
	       m_alloc (sizeof *cimage->world_y0 * num_levels) ) == NULL )
	{
	    m_abort (function_name, "contour world y0 array");
	}
	if ( ( cimage->world_x1 =
	       m_alloc (sizeof *cimage->world_x1 * num_levels) ) == NULL )
	{
	    m_abort (function_name, "contour world x1 array");
	}
	if ( ( cimage->world_y1 =
	       m_alloc (sizeof *cimage->world_y1 * num_levels) ) == NULL )
	{
	    m_abort (function_name, "contour world y1 array");
	}
	if ( ( cimage->levels =
	       m_alloc (sizeof *cimage->levels * num_levels) ) == NULL )
	{
	    m_abort (function_name, "contour level structures");
	}
	for (count = 0; count < num_levels; ++count)
	{
	    cimage->levels[count].world_coord_buf_size = 0;
	    cimage->world_x0[count] = NULL;
	    cimage->world_y0[count] = NULL;
	    cimage->world_x1[count] = NULL;
	    cimage->world_y1[count] = NULL;
	    cimage->levels[count].cache_buf_size = 0;
	    cimage->levels[count].linear_x0 = NULL;
	    cimage->levels[count].linear_y0 = NULL;
	    cimage->levels[count].linear_x1 = NULL;
	    cimage->levels[count].linear_y1 = NULL;
	    cimage->levels[count].pix_x0 = NULL;
	    cimage->levels[count].pix_y0 = NULL;
	    cimage->levels[count].pix_x1 = NULL;
	    cimage->levels[count].pix_y1 = NULL;
	    cimage->levels[count].pix_area_x0 = NULL;
	    cimage->levels[count].pix_area_y0 = NULL;
	    cimage->levels[count].pix_area_x1 = NULL;
	    cimage->levels[count].pix_area_y1 = NULL;
	    cimage->levels[count].pix_limits.xmin[0] = TOOBIG;
	}
	cimage->num_levels = num_levels;
	cimage->world_segs_valid = FALSE;
    }
    for (count = 0; count < num_levels; ++count)
    {
	if (contour_pixels == NULL)
	{
	    cimage->levels[count].pixel_defined = FALSE;
	}
	else
	{
	    cimage->levels[count].pixel_value = contour_pixels[count];
	    cimage->levels[count].pixel_defined = TRUE;
	}
	if (dash == NULL) cimage->levels[count].dash = FALSE;
	else cimage->levels[count].dash = dash[count];
	if (linewidths == NULL) cimage->levels[count].linewidth = 0.0;
	else cimage->levels[count].linewidth = linewidths[count];
    }
}   /*  End Function contour_set_levels  */

/*PUBLIC_FUNCTION*/
KWorldCanvas contour_get_worldcanvas (KContourImage cimage)
/*  [SUMMARY] Get the world canvas for a contour image.
    <cimage> The KContourImage object.
    [RETURNS] The KWorldCanvas object.
*/
{
    static char function_name[] = "contour_get_worldcanvas";

    VERIFY_USED_CIMAGE (cimage);
    return (cimage->canvas_holder->canvas);
}   /*  End Function contour_get_worldcanvas  */


/*  Private functions follow  */

static KContourImage create_restr (CanvasHolder holder,
				   multi_array *multi_desc, flag try_ap,
				   array_desc *arr_desc, char *slice,
				   unsigned int hdim, unsigned int vdim,
				   unsigned int elem_index,
				   unsigned int num_levels,
				   CONST double *contour_levels,
				   unsigned num_restr,
				   char **restr_names, double *restr_values,
				   uaddr key)
/*  [SUMMARY] Create contour image from 2D slice with restrictions.
    [PURPOSE] This routine will create a contour image object from a
    2-dimensional slice of a Karma data structure. At a later time, this
    contour image may be made visible. This routine will not cause the canvas
    to be refreshed.
    <holder> The canvas holder.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <try_ap> If TRUE, a KwcsAstro object is created.
    <arr_desc> The array descriptor.
    <slice> The start of the slice data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index in the data packets.
    <num_levels> The number of contour levels. This may be 0.
    <contour_levels> The array of contour levels.
    <num_restr> The number of matched restrictions.
    <restr_names> The restriction names.
    <restr_values> The restriction values.
    [NOTE] Restriction information is automatically deallocated when
    [<contour_destroy>] is called.
    <key> The key to associate with this contour image.
    [RETURNS] A KContourImage object on success, NULL.
*/
{
    KContourImage cimage;
    unsigned int index;
    static char function_name[] = "create_restr";

    if (first_free == NULL)
    {
	if ( ( cimage = m_alloc (sizeof *cimage) ) == NULL ) return (NULL);
    }
    else
    {
	cimage = first_free;
	first_free = cimage->next;
	VERIFY_FREE_CIMAGE (cimage);
    }
    cimage->num_levels = 0;
    cimage->contour_levels = NULL;
    cimage->magic_number = CIMAGE_USED_MAGIC_NUMBER;
    cimage->active = FALSE;
    cimage->canvas_holder = holder;
    cimage->multi_desc = multi_desc;
    cimage->arr_desc = arr_desc;
    cimage->slice = slice + ds_get_element_offset (arr_desc->packet,
						   elem_index);
    cimage->hdim = hdim;
    cimage->vdim = vdim;
    cimage->elem_index = elem_index;
    if ( try_ap && (multi_desc != NULL) )
    {
	index = get_gen_index (multi_desc, arr_desc);
	cimage->astro_projection = wcs_astro_get_associated (multi_desc,index);
    }
    else cimage->astro_projection = NULL;
    cimage->world_segs_valid = FALSE;
    cimage->num_restrictions = num_restr;
    cimage->restriction_names = restr_names;
    cimage->restriction_values = restr_values;
    cimage->key = key;
    cimage->world_x0 = NULL;
    cimage->world_y0 = NULL;
    cimage->world_x1 = NULL;
    cimage->world_y1 = NULL;
    cimage->old_x_offset = -1;
    cimage->old_y_offset = -1;
    cimage->old_x_pixels = -1;
    cimage->old_y_pixels = -1;
    cimage->old_left_x = TOOBIG;
    cimage->old_right_x = TOOBIG;
    cimage->old_bottom_y = TOOBIG;
    cimage->old_top_y = TOOBIG;
    cimage->test_x[0] = TOOBIG;
    /*  Append to list  */
    cimage->prev = holder->last_image;
    cimage->next = NULL;
    if (holder->first_image == NULL) holder->first_image = cimage;
    else holder->last_image->next = cimage;
    holder->last_image = cimage;
    contour_set_levels (cimage, num_levels, contour_levels, NULL, NULL, NULL);
    /*  Do not make the first contour image for this canvas contour, else
	the  contour_make_active  routine will not be able to detect a change
	in the active contour image and will do nothing if called with this
	newly created contour image.
	*/
    /*  Attach  */
    if (multi_desc != NULL) ++multi_desc->attachments;
    return (cimage);
}   /*  End Function create_restr  */

static unsigned int get_gen_index (CONST multi_array *multi_desc,
				   CONST array_desc *arr_desc)
/*  [SUMMARY] Find general data structure which has an array descriptor.
    <multi_desc> The multi_array header.
    <arr_desc> The array descriptor to search for.
    [RETURNS] The index of the general data structure in the multi_array header
*/
{
    unsigned int count;
    char *encls_desc;
    static char function_name[] = "get_gen_index";

    for (count = 0; count < multi_desc->num_arrays; ++count)
    {
	switch ( ds_f_name_in_packet (multi_desc->headers[count],
				      arr_desc->dimensions[0]->name,
				      &encls_desc, NULL) )
	{
	  case IDENT_NOT_FOUND:
	    break;
	  case IDENT_DIMENSION:
	    if (arr_desc != (array_desc *) encls_desc)
	    {
		fprintf (stderr, "Found array descriptor: %p is not: %p\n",
			 encls_desc, arr_desc);
		a_prog_bug (function_name);
	    }
	    return (count);
	    /*break;*/
	  default:
	    fprintf (stderr, "Illegal return from <ds_f_name_in_packet>\n");
	    a_prog_bug (function_name);
	    break;
	}
    }
    fprintf (stderr, "Array descriptor not found\n");
    a_prog_bug (function_name);
    return (0);
}   /*  End Function get_gen_index  */

static CanvasHolder get_canvas_holder (KWorldCanvas canvas, char *func_name)
/*  [SUMMARY] Get the canvas holder for a world canvas.
    <canvas> The canvas.
    <func_name> The name of the calling function. If this is not NULL, the
    process will abort if a canvas holder is not found.
    [RETURNS] The canvas holder on success, else NULL.
*/
{
    CanvasHolder canvas_holder;

    /*  First search for existing canvas holder  */
    for (canvas_holder = first_canvas_holder; canvas_holder != NULL;
	 canvas_holder = canvas_holder->next)
    {
	if (canvas == canvas_holder->canvas) return (canvas_holder);
    }
    if (func_name != NULL)
    {
	fprintf (stderr, "CanvasHolder not found!\n");
	a_prog_bug (func_name);
    }
    return (NULL);
}   /*  End Function get_canvas_holder  */

static CanvasHolder alloc_canvas_holder (KWorldCanvas canvas)
/*  This routine will allocate a canvas holder for a world canvas.
    The canvas must be given by  canvas  .
    The routine returns the canvas holder on success, else it returns NULL.
*/
{
    CanvasHolder holder;
    static char function_name[] = "alloc_canvas_holder";

    /*  Must allocate holder  */
    if ( ( holder = m_calloc (sizeof *holder) ) == NULL )
    {
	m_error_notify (function_name, "canvas holder");
	return (NULL);
    }
    holder->magic_number = HOLDER_MAGIC_NUMBER;
    holder->canvas = canvas;
    /*  Insert at beginning of list  */
    if (first_canvas_holder != NULL)
	first_canvas_holder->prev = holder;
    holder->next = first_canvas_holder;
    first_canvas_holder = holder;
    canvas_register_precompute_func (canvas,
				     _contour_worldcanvas_precompute_func,
				     holder);
    canvas_register_refresh_func (canvas, _contour_worldcanvas_refresh_func,
				  holder);
    canvas_register_destroy_func (canvas, FALSE, _contour_free_canvas_holder,
				  holder);
    return (holder);
}   /*  End Function alloc_canvas_holder  */

static void _contour_free_canvas_holder (CanvasHolder holder)
/*  [SUMMARY] Free a canvas holder (and any associated KContourImage objects).
    <holder> The canvas holder.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_contour_free_canvas_holder";

    VERIFY_HOLDER (holder);
    while (holder->first_image != NULL)
	contour_destroy (holder->first_image);
    c_destroy_list (holder->position_list);
    if (holder->next != NULL) holder->next->prev = holder->prev;
    if (holder->prev != NULL) holder->prev->next = holder->next;
    if (holder == first_canvas_holder) first_canvas_holder = holder->next;
    if (holder->colourname != NULL) m_free (holder->colourname);
    holder->magic_number = 0;
    m_free (holder);
}   /*  End Function _contour_free_canvas_holder  */

static void process_canvas_attributes (CanvasHolder holder, va_list argp)
/*  [PURPOSE] This routine will set the attributes for a canvas.
    <holder> The canvas holder.
    <argp> The optional list of parameter attribute-key attribute-value
    pairs.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int att_key;
    char *colour;
    static char function_name[] = "_contour_process_canvas_attributes";

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    CONTOUR_CANVAS_ATT_END )
    {
	switch (att_key)
	{
	  case CONTOUR_CANVAS_ATT_COLOURNAME:
	    colour = va_arg (argp, char *);
	    if (colour == NULL) colour = st_dup (DEFAULT_COLOUR);
	    else colour = st_dup (colour);
	    if (colour == NULL) m_abort (function_name, "colour name");
	    if (holder->colourname != NULL) m_free (holder->colourname);
	    holder->colourname = colour;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
}   /*  End Function process_canvas_attributes  */

static void _contour_worldcanvas_precompute_func
    (KWorldCanvas canvas, int width, int height,
     struct win_scale_type *win_scale, Kcolourmap cmap, flag cmap_resize,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KContourImage cimage;
    static char function_name[] = "_contour_worldcanvas_precompute_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    /*  Show each active image  */
    for (cimage = holder->first_image; cimage != NULL; cimage = cimage->next)
    {
	if (cimage->active) _contour_precompute_image (cimage, win_scale,
						       num_areas, areas);
    }
}   /*  End Function _contour_worldcanvas_precompute_func  */

static void _contour_precompute_image (KContourImage cimage,
				       struct win_scale_type *win_scale,
				       unsigned int num_areas,
				       KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <cimage> The KContourImage object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    flag same_trans = FALSE;
    flag same_world = FALSE;
    KWorldCanvas canvas = cimage->canvas_holder->canvas;
    unsigned int count;
    double test_x[5], test_y[5];
    CONST packet_desc *pack_desc;
    CONST uaddr *hoffsets, *voffsets;
    CONST dim_desc *hdim, *vdim;
    /*static char function_name[] = "_contour_precompute_image";*/

    if (cimage->num_levels < 1) return;
    pack_desc = cimage->arr_desc->packet;
    hdim = cimage->arr_desc->dimensions[cimage->hdim];
    vdim = cimage->arr_desc->dimensions[cimage->vdim];
    hoffsets = cimage->arr_desc->offsets[cimage->hdim];
    voffsets = cimage->arr_desc->offsets[cimage->vdim];
    ap = cimage->astro_projection;
    if (!cimage->world_segs_valid)
    {
	/*  Compute segments  */
	if ( !ds_contour (cimage->slice,
			  pack_desc->element_types[cimage->elem_index],
			  hdim, hoffsets, vdim, voffsets,
			  cimage->num_levels, cimage->contour_levels,
			  (uaddr *) &cimage->levels[0].world_coord_buf_size,
			  sizeof cimage->levels[0],
			  cimage->world_x0, cimage->world_y0,
			  cimage->world_x1, cimage->world_y1,
			  &cimage->levels[0].num_segments,
			  sizeof cimage->levels[0]) ) return;
	/*  Convert from data pixel to world co-ordinates  */
	for (count = 0; count < cimage->num_levels; ++count)
	{
	    if (ap == NULL) continue;
	    wcs_astro_transform3 (ap, cimage->levels[count].num_segments,
				  hdim->name, cimage->world_x0[count], FALSE,
				  vdim->name, cimage->world_y0[count], FALSE,
				  NULL, NULL, FALSE,
				  cimage->num_restrictions,
				  (CONST char **) cimage->restriction_names,
				  cimage->restriction_values);
	    wcs_astro_transform3 (ap, cimage->levels[count].num_segments,
				  hdim->name, cimage->world_x1[count], FALSE,
				  vdim->name, cimage->world_y1[count], FALSE,
				  NULL, NULL, FALSE,
				  cimage->num_restrictions,
				  (CONST char **) cimage->restriction_names,
				  cimage->restriction_values);
	}
    }
    /*  At this point the contour segments have been computed in world
	co-ordinates. To speed the redisplay, check if the non-linear to linear
	world co-ordinate transformation has changed  */
    test_x[0] = hdim->first_coord;
    test_y[0] = vdim->first_coord;
    test_x[1] = hdim->last_coord;
    test_y[1] = vdim->first_coord;
    test_x[2] = (hdim->first_coord + hdim->last_coord) / 2.0;
    test_y[2] = (vdim->first_coord + vdim->last_coord) / 2.0;
    test_x[3] = hdim->first_coord;
    test_y[3] = vdim->last_coord;
    test_x[4] = hdim->last_coord;
    test_y[4] = vdim->last_coord;
    canvas_coords_transform (canvas, 5, test_x, FALSE, test_y, FALSE);
    if ( m_cmp (cimage->test_x, test_x, sizeof *test_x * 5) &&
	 m_cmp (cimage->test_y, test_y, sizeof *test_y * 5) )
    {
	/*  The world->linear transformation has not changed  */
	same_trans = TRUE;
	if ( (cimage->old_x_offset == win_scale->x_offset) &&
	     (cimage->old_y_offset == win_scale->y_offset) &&
	     (cimage->old_x_pixels == win_scale->x_pixels) &&
	     (cimage->old_y_pixels == win_scale->y_pixels) &&
	     (cimage->old_left_x == win_scale->left_x) &&
	     (cimage->old_right_x == win_scale->right_x) &&
	     (cimage->old_bottom_y == win_scale->bottom_y) &&
	     (cimage->old_top_y == win_scale->top_y) )
	{
	    /*  The linear->pixel transformation has not changed  */
	    same_world = TRUE;
	}
    }
    if (!cimage->world_segs_valid)
    {
	same_trans = FALSE;
	same_world = FALSE;
    }
    /*  Now go ahead and process each level separately  */
    for (count = 0; count < cimage->num_levels; ++count)
    {
	precompute_level (cimage, count, win_scale,
			  num_areas, areas, same_trans, same_world);
    }
    cimage->world_segs_valid = TRUE;
    m_copy (cimage->test_x, test_x, sizeof *test_x * 5);
    m_copy (cimage->test_y, test_y, sizeof *test_y * 5);
    cimage->old_x_offset = win_scale->x_offset;
    cimage->old_y_offset = win_scale->y_offset;
    cimage->old_x_pixels = win_scale->x_pixels;
    cimage->old_y_pixels = win_scale->y_pixels;
    cimage->old_left_x = win_scale->left_x;
    cimage->old_right_x = win_scale->right_x;
    cimage->old_bottom_y = win_scale->bottom_y;
    cimage->old_top_y = win_scale->top_y;
}   /*  End Function _contour_precompute_image  */

static void precompute_level (KContourImage cimage, unsigned int level_index,
			      struct win_scale_type *win_scale,
			      unsigned int num_areas,
			      KPixCanvasRefreshArea *areas,
			      flag same_trans, flag same_world)
/*  [PURPOSE] This routine will precompute a single level.
    <cimage> The KContourImage object.
    <level_index> The contour level to draw.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <same_trans> If TRUE the world->linear co-ordinate transformation has not
    changed since the last refresh.
    <same_world> If TRUE the linear->pixel co-ordinate transformation has not
    changed since the last refresh.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas;
    CanvasHolder holder;
    flag section_visible, segment_visible;
    unsigned int count, num_pix_segs, area_count;
    unsigned int section_count, subsection_count;
    unsigned int pix_section_size;
    double win_x_min, win_y_min, win_x_max, win_y_max, tmp;
    double area_startx, area_endx, area_starty, area_endy;
    double x0, y0, x1, y1;
    double offset = 1.0;
    struct levelinfo_type *level = &cimage->levels[level_index];
    static char function_name[] = "_contour_precompute_level";

    holder = cimage->canvas_holder;
    canvas = holder->canvas;
    level->num_area_segments = 0;
    /*  Find edge of canvas in linear co-ordinates  */
    win_x_min = win_scale->x_offset;
    win_y_min = win_scale->y_offset;
    win_x_max = win_scale->x_offset + win_scale->x_pixels - 1;
    win_y_max = win_scale->y_offset + win_scale->y_pixels - 1;
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &win_x_min, &win_y_min,
				     &win_x_min, &win_y_min,
				     NULL, NULL);
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &win_x_max, &win_y_max,
				     &win_x_max, &win_y_max,
				     NULL, NULL);
    if (win_x_min > win_x_max) swap (win_x_min, win_x_max, tmp);
    if (win_y_min > win_y_max) swap (win_y_min, win_y_max, tmp);
    if (level->num_segments > level->cache_buf_size)
    {
	/*  Need to reallocate linear and pixel co-ordinate buffers  */
	if (level->linear_x0 != NULL)
	{
	    m_free (level->linear_x0);
	    m_free (level->linear_y0);
	    m_free (level->linear_x1);
	    m_free (level->linear_y1);
	    m_free (level->pix_x0);
	    m_free (level->pix_y0);
	    m_free (level->pix_x1);
	    m_free (level->pix_y1);
	    m_free (level->pix_area_x0);
	    m_free (level->pix_area_y0);
	    m_free (level->pix_area_x1);
	    m_free (level->pix_area_y1);
	}
	if ( ( level->linear_x0 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "linear x0 array");
	}
	if ( ( level->linear_y0 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "linear y0 array");
	}
	if ( ( level->linear_x1 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "linear x1 array");
	}
	if ( ( level->linear_y1 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "linear y1 array");
	}
	if ( ( level->pix_x0 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "pixel x0 array");
	}
	if ( ( level->pix_y0 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "pixel y0 array");
	}
	if ( ( level->pix_x1 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "pixel x1 array");
	}
	if ( ( level->pix_y1 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "pixel y1 array");
	}
	if ( ( level->pix_area_x0 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "pixel area x0 array");
	}
	if ( ( level->pix_area_y0 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "pixel area y0 array");
	}
	if ( ( level->pix_area_x1 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "pixel area x1 array");
	}
	if ( ( level->pix_area_y1 =
	       m_alloc (sizeof (double) * level->num_segments) ) == NULL )
	{
	    m_abort (function_name, "pixel area y1 array");
	}
	level->cache_buf_size = level->num_segments;
    }
    if (!same_trans)
    {
	_contour_compute_lin_segs (cimage, level_index, win_scale,
				   win_x_min, win_y_min, win_x_max, win_y_max);
	/*  The pixel segments have now also been computed  */
	same_world = TRUE;
    }
    /*  Now the linear co-ordinates of the segments are valid  */
    if (!same_world)
    {
	/*  The world->linear transformation has changed since the last draw */
	_contour_compute_pix_segs (cimage, level_index, win_scale,
				   win_x_min, win_y_min, win_x_max, win_y_max);
    }
    if ( (num_areas < 1) || (level->num_pixel_segments < NUM_SECTIONS) )
    {
	/*  Either whole canvas is to be refreshed (hence all cached pixel
	    segments should be drawn) or there are not enough segments to be
	    sectioned, so they are all drawn too  */
	return;
    }
    /*  Several refresh areas requested: draw only those areas  */
    pix_section_size = level->num_pixel_segments / NUM_SECTIONS;
    for (count = 0, section_count = 0, subsection_count = 0, num_pix_segs = 0;
	 count < level->num_pixel_segments; ++count, ++subsection_count)
    {
	if (subsection_count >= pix_section_size)
	{
	    if (++section_count < NUM_SECTIONS) subsection_count = 0;
	    else
	    {
		subsection_count = 0;  /* Pretend last section not processed */
		section_count = NUM_SECTIONS - 1;
	    }
	}
	if ( (subsection_count == 0) &&
	     (level->num_pixel_segments >= NUM_SECTIONS) )
	{
	    /*  First time in this section: use the limits to quickly check
		if any segments should be drawn  */
	    section_visible = FALSE;
	    for (area_count = 0; area_count < num_areas; ++area_count)
	    {
		area_startx = (double) areas[area_count].startx - offset;
		area_endx = (double) areas[area_count].endx + offset;
		area_starty = (double) areas[area_count].starty - offset;
		area_endy = (double) areas[area_count].endy + offset;
		if ( (level->pix_limits.xmin[section_count] < area_startx) &&
		     (level->pix_limits.xmax[section_count] < area_startx) )
		{
		    continue;
		}
		if ( (level->pix_limits.xmin[section_count] > area_endx) &&
		     (level->pix_limits.xmax[section_count] > area_endx) )
		{
		    continue;
		}
		if ( (level->pix_limits.ymin[section_count] < area_starty) &&
		     (level->pix_limits.ymax[section_count] < area_starty) )
		{
		    continue;
		}
		if ( (level->pix_limits.ymin[section_count] > area_endy) &&
		     (level->pix_limits.ymax[section_count] > area_endy) )
		{
		    continue;
		}
		/*  Section is visible in this area  */
		section_visible = TRUE;
		area_count = num_areas;
	    }
	    if (!section_visible)
	    {
		count += pix_section_size - 1;
		subsection_count += pix_section_size - 1;
		continue;
	    }
	}
	/*  Segment is in a visible section: test if segment itself is
	    visible  */
	x0 = level->pix_x0[count];
	y0 = level->pix_y0[count];
	x1 = level->pix_x1[count];
	y1 = level->pix_y1[count];
	segment_visible = FALSE;
	for (area_count = 0; area_count < num_areas; ++area_count)
	{
	    area_startx = (double) areas[area_count].startx - offset;
	    area_endx = (double) areas[area_count].endx + offset;
	    area_starty = (double) areas[area_count].starty - offset;
	    area_endy = (double) areas[area_count].endy + offset;
	    if ( (x0 < area_startx) && (x1 < area_startx) ) continue;
	    if ( (x0 > area_endx) && (x1 > area_endx) ) continue;
	    if ( (y0 < area_starty) && (y1 < area_starty) ) continue;
	    if ( (y0 > area_endy) && (y1 > area_endy) ) continue;
	    /*  Segment is visible in this area  */
	    segment_visible = TRUE;
	    area_count = num_areas;
	}
	if (segment_visible)
	{
	    level->pix_area_x0[num_pix_segs] = x0;
	    level->pix_area_y0[num_pix_segs] = y0;
	    level->pix_area_x1[num_pix_segs] = x1;
	    level->pix_area_y1[num_pix_segs] = y1;
	    ++num_pix_segs;
	}
    }
    level->num_area_segments = num_pix_segs;
}   /*  End Function precompute_level  */

static void _contour_compute_lin_segs (KContourImage cimage,
				       unsigned int level_index,
				       struct win_scale_type *win_scale,
				       double win_x_min, double win_y_min,
				       double win_x_max, double win_y_max)
/*  [SUMMARY] Compute linear world canvas segments.
    <cimage> The KContourImage object.
    <level_index> The contour level to draw.
    <win_x_min> The lower horizontal linear co-orinate.
    <win_x_max> The upper horizontal linear co-orinate.
    <win_y_min> The lower vertical linear co-orinate.
    <win_y_max> The upper vertical linear co-orinate.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = cimage->canvas_holder->canvas;
    unsigned int count, num_pix_segs;
    unsigned int section_count, subsection_count, lin_section_size;
    double sec_xmin, sec_ymin, sec_xmax, sec_ymax, x0, y0, x1, y1;
    struct levelinfo_type *level = &cimage->levels[level_index];

    /*  Have to compute linear co-ordinates  */
    for (count = 0; count < level->num_segments; ++count)
	level->linear_x0[count] = cimage->world_x0[level_index][count];
    for (count = 0; count < level->num_segments; ++count)
	level->linear_y0[count] = cimage->world_y0[level_index][count];
    for (count = 0; count < level->num_segments; ++count)
	level->linear_x1[count] = cimage->world_x1[level_index][count];
    for (count = 0; count < level->num_segments; ++count)
	level->linear_y1[count] = cimage->world_y1[level_index][count];
    canvas_coords_transform (canvas, level->num_segments,
			     level->linear_x0, TRUE,
			     level->linear_y0, TRUE);
    canvas_coords_transform (canvas, level->num_segments,
			     level->linear_x1, TRUE,
			     level->linear_y1, TRUE);
    if (level->num_segments < NUM_SECTIONS)
    {
	/*  Not enough segments: the sectioning algorithm will hang  */
	/*  Convert linear world segments to pixel segments  */
	for (count = 0, num_pix_segs = 0; count < level->num_segments; ++count)
	{
	    x0 = level->linear_x0[count];
	    y0 = level->linear_y0[count];
	    x1 = level->linear_x1[count];
	    y1 = level->linear_y1[count];
	    /*  Add segment to pixel list if inside world canvas  */
	    if ( (x0 < win_x_min) || (x0 > win_x_max) ||
		 (x1 < win_x_min) || (x1 > win_x_max) ||
		 (y0 < win_y_min) || (y0 > win_y_max) ||
		 (y1 < win_y_min) || (y1 > win_y_max) ) continue;
	    level->pix_x0[num_pix_segs] = x0;
	    level->pix_y0[num_pix_segs] = y0;
	    level->pix_x1[num_pix_segs] = x1;
	    level->pix_y1[num_pix_segs] = y1;
	    ++num_pix_segs;
	}
	level->num_pixel_segments = num_pix_segs;
	canvas_convert_from_canvas_coords
	    (canvas, FALSE, TRUE, level->num_pixel_segments,
	     level->pix_x0, level->pix_y0,
	     level->pix_x0, level->pix_y0);
	canvas_convert_from_canvas_coords
	    (canvas, FALSE, TRUE, level->num_pixel_segments,
	     level->pix_x1, level->pix_y1,
	     level->pix_x1, level->pix_y1);
	/*  Pixel segments are now ready to be drawn  */
	return;
    }
    /*  Divide the segment arrays into a number of sections and compute the
	limits of each section. This can speed later operations. Since all
	the segments have to be processed anyway, may as well perform the
	world->pixel selection and conversion as well, even though it would
	have been done later anyway. It should be faster to combine the
	operations in one loop, rather than processing all the segments
	twice  */
    lin_section_size = level->num_segments / NUM_SECTIONS;
    section_count = 0;
    sec_xmin = TOOBIG;
    sec_xmax = -TOOBIG;
    sec_ymin = TOOBIG;
    sec_ymax = -TOOBIG;
    for (count = 0, num_pix_segs = 0, subsection_count = 0;
	 count < level->num_segments; ++count, ++subsection_count)
    {
	if (subsection_count >= lin_section_size)
	{
	    /*  End of section: save current section and go to next  */
	    level->linear_limits.xmin[section_count] = sec_xmin;
	    level->linear_limits.xmax[section_count] = sec_xmax;
	    level->linear_limits.ymin[section_count] = sec_ymin;
	    level->linear_limits.ymax[section_count] = sec_ymax;
	    if (++section_count < NUM_SECTIONS)
	    {
		sec_xmin = TOOBIG;
		sec_xmax = -TOOBIG;
		sec_ymin = TOOBIG;
		sec_ymax = -TOOBIG;
		subsection_count = 0;
	    }
	    else
	    {
		subsection_count = 1; /* Carry on with the last section */
		section_count = NUM_SECTIONS - 1;
	    }
	}
	x0 = level->linear_x0[count];
	y0 = level->linear_y0[count];
	x1 = level->linear_x1[count];
	y1 = level->linear_y1[count];
	/*  Update section limits  */
	if (x0 < sec_xmin) sec_xmin = x0;
	if (x0 > sec_xmax) sec_xmax = x0;
	if (y0 < sec_ymin) sec_ymin = y0;
	if (y0 > sec_ymax) sec_ymax = y0;
	if (x1 < sec_xmin) sec_xmin = x1;
	if (x1 > sec_xmax) sec_xmax = x1;
	if (y1 < sec_ymin) sec_ymin = y1;
	if (y1 > sec_ymax) sec_ymax = y1;
	/*  Add segment to pixel list if inside world canvas  */
	if ( (x0 < win_x_min) || (x0 > win_x_max) ||
	     (x1 < win_x_min) || (x1 > win_x_max) ||
	     (y0 < win_y_min) || (y0 > win_y_max) ||
	     (y1 < win_y_min) || (y1 > win_y_max) ) continue;
	level->pix_x0[num_pix_segs] = x0;
	level->pix_y0[num_pix_segs] = y0;
	level->pix_x1[num_pix_segs] = x1;
	level->pix_y1[num_pix_segs] = y1;
	++num_pix_segs;
    }
    /*  Save last section limits  */
    level->linear_limits.xmin[section_count] = sec_xmin;
    level->linear_limits.xmax[section_count] = sec_xmax;
    level->linear_limits.ymin[section_count] = sec_ymin;
    level->linear_limits.ymax[section_count] = sec_ymax;
    compute_pixel_coords (canvas, level, win_scale, num_pix_segs);
}   /*  End Function _contour_compute_lin_segs  */

static void _contour_compute_pix_segs (KContourImage cimage,
				       unsigned int level_index,
				       struct win_scale_type *win_scale,
				       double win_x_min, double win_y_min,
				       double win_x_max, double win_y_max)
/*  [SUMMARY] Compute pixel canvas segments.
    <cimage> The KContourImage object.
    <level_index> The contour level to draw.
    <win_x_min> The lower horizontal linear co-orinate.
    <win_x_max> The upper horizontal linear co-orinate.
    <win_y_min> The lower vertical linear co-orinate.
    <win_y_max> The upper vertical linear co-orinate.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = cimage->canvas_holder->canvas;
    unsigned int count, num_pix_segs;
    unsigned int section_count, subsection_count, lin_section_size;
    double x0, y0, x1, y1;
    struct levelinfo_type *level = &cimage->levels[level_index];

    if (level->num_segments < NUM_SECTIONS)
    {
	/*  Not enough segments: the sectioning algorithm will hang  */
	/*  Convert linear world segments to pixel segments  */
	for (count = 0, num_pix_segs = 0; count < level->num_segments; ++count)
	{
	    x0 = level->linear_x0[count];
	    y0 = level->linear_y0[count];
	    x1 = level->linear_x1[count];
	    y1 = level->linear_y1[count];
	    /*  Add segment to pixel list if inside world canvas  */
	    if ( (x0 < win_x_min) || (x0 > win_x_max) ||
		 (x1 < win_x_min) || (x1 > win_x_max) ||
		 (y0 < win_y_min) || (y0 > win_y_max) ||
		 (y1 < win_y_min) || (y1 > win_y_max) ) continue;
	    level->pix_x0[num_pix_segs] = x0;
	    level->pix_y0[num_pix_segs] = y0;
	    level->pix_x1[num_pix_segs] = x1;
	    level->pix_y1[num_pix_segs] = y1;
	    ++num_pix_segs;
	}
	level->num_pixel_segments = num_pix_segs;
	canvas_convert_from_canvas_coords
	    (canvas, FALSE, TRUE, level->num_pixel_segments,
	     level->pix_x0, level->pix_y0,
	     level->pix_x0, level->pix_y0);
	canvas_convert_from_canvas_coords
	    (canvas, FALSE, TRUE, level->num_pixel_segments,
	     level->pix_x1, level->pix_y1,
	     level->pix_x1, level->pix_y1);
	/*  Pixel segments are now ready to be drawn  */
	return;
    }
    /*  Process only those segments which lie inside the canvas  */
    lin_section_size = level->num_segments / NUM_SECTIONS;
    for (count = 0, num_pix_segs = 0, section_count = 0,
	     subsection_count = 0;
	 count < level->num_segments; ++count, ++subsection_count)
    {
	if (subsection_count >= lin_section_size)
	{
	    if (++section_count < NUM_SECTIONS) subsection_count = 0;
	    else
	    {
		subsection_count = 0; /* Pretend last section not processed */
		section_count = NUM_SECTIONS - 1;
	    }
	}
	if (subsection_count == 0)
	{
	    /*  First time in this section: use the limits to quickly check
		if any segments should be drawn  */
	    if ( (level->linear_limits.xmin[section_count] < win_x_min) &&
		 (level->linear_limits.xmax[section_count] < win_x_min) )
	    {
		count += lin_section_size - 1;
		subsection_count += lin_section_size - 1;
		continue;
	    }
	    if ( (level->linear_limits.xmin[section_count] > win_x_max) &&
		 (level->linear_limits.xmax[section_count] > win_x_max) )
	    {
		count += lin_section_size - 1;
		subsection_count += lin_section_size - 1;
		continue;
	    }
	    if ( (level->linear_limits.ymin[section_count] < win_y_min) &&
		 (level->linear_limits.ymax[section_count] < win_y_min) )
	    {
		count += lin_section_size - 1;
		subsection_count += lin_section_size - 1;
		continue;
	    }
	    if ( (level->linear_limits.ymin[section_count] > win_y_max) &&
		 (level->linear_limits.ymax[section_count] > win_y_max) )
	    {
		count += lin_section_size - 1;
		subsection_count += lin_section_size - 1;
		continue;
	    }
	}
	x0 = level->linear_x0[count];
	y0 = level->linear_y0[count];
	x1 = level->linear_x1[count];
	y1 = level->linear_y1[count];
	if ( (x0 < win_x_min) || (x0 > win_x_max) ||
	     (x1 < win_x_min) || (x1 > win_x_max) ||
	     (y0 < win_y_min) || (y0 > win_y_max) ||
	     (y1 < win_y_min) || (y1 > win_y_max) ) continue;
	level->pix_x0[num_pix_segs] = x0;
	level->pix_y0[num_pix_segs] = y0;
	level->pix_x1[num_pix_segs] = x1;
	level->pix_y1[num_pix_segs] = y1;
	++num_pix_segs;
    }
    compute_pixel_coords (canvas, level, win_scale, num_pix_segs);
}   /*  End Function _contour_compute_pix_segs  */

static void _contour_worldcanvas_refresh_func(KWorldCanvas canvas,
					      int width, int height,
					      struct win_scale_type *win_scale,
					      Kcolourmap cmap,
					      flag cmap_resize,
					      void **info,
					      PostScriptPage pspage,
					      unsigned int num_areas,
					      KPixCanvasRefreshArea *areas,
					      flag *honoured_areas)
/*  [PURPOSE] This routine processes a refresh event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KContourImage cimage;
    static char function_name[] = "_contour_worldcanvas_refresh_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    /*  Show each active image  */
    for (cimage = holder->first_image; cimage != NULL; cimage = cimage->next)
    {
	if (cimage->active) refresh_image (cimage, win_scale, num_areas,areas);
    }
    *honoured_areas = TRUE;
}   /*  End Function _contour_worldcanvas_refresh_func  */

static void refresh_image (KContourImage cimage,
			   struct win_scale_type *win_scale,
			   unsigned int num_areas,KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a refresh event for a world canvas.
    <cimage> The KContourImage object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    flag dash;
    KPixCanvas pixcanvas;
    unsigned int count;
    double linewidth;
    /*static char function_name[] = "_contour_refresh_image";*/

    if (cimage->num_levels < 1) return;
    pixcanvas = canvas_get_pixcanvas (cimage->canvas_holder->canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_LINE_DASH, &dash,
			 KWIN_ATT_LINEWIDTH, &linewidth,
			 KWIN_ATT_END);
    for (count = 0; count < cimage->num_levels; ++count)
    {
	draw_level (cimage, count, win_scale, num_areas, areas);
    }
    kwin_set_attributes (pixcanvas,
			 KWIN_ATT_LINE_DASH, dash,
			 KWIN_ATT_LINEWIDTH, linewidth,
			 KWIN_ATT_END);
}   /*  End Function refresh_image  */

static void draw_level (KContourImage cimage, unsigned int level_index,
			struct win_scale_type *win_scale,
			unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine will draw a single level.
    <cimage> The KContourImage object.
    <level> The contour level to draw.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <same_trans> If TRUE the world->linear co-ordinate transformation has not
    changed since the last refresh.
    <same_world> If TRUE the linear->pixel co-ordinate transformation has not
    changed since the last refresh.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas;
    KPixCanvas pixcanvas;
    CanvasHolder holder;
    unsigned long pixel_value;
    struct levelinfo_type *level = &cimage->levels[level_index];
    static char function_name[] = "_contour_draw_level";

    holder = cimage->canvas_holder;
    canvas = holder->canvas;
    if (level->pixel_defined)
    {
	pixel_value = level->pixel_value;
    }
    else if (cimage->colourname != NULL)
    {
	/*  Get colour, with a hack to use grey rather than real dashing  */
	if ( !canvas_get_colour (canvas,
				 level->dash ? "grey70" : cimage->colourname,
				 &pixel_value, NULL, NULL, NULL) )
	{
	    fprintf (stderr, "%s: error getting colour\n", function_name);
	    return;
	}
    }
    else
    {
	/*  Get colour, with a hack to use grey rather than real dashing  */
	if ( !canvas_get_colour (canvas,
				 level->dash ? "grey70" : holder->colourname,
				 &pixel_value, NULL, NULL, NULL) )
	{
	    fprintf (stderr, "%s: error getting colour\n", function_name);
	    return;
	}
    }
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_set_attributes (pixcanvas,
#ifdef DISABLED
			 KWIN_ATT_LINE_DASH, level->dash,
#endif
			 KWIN_ATT_LINEWIDTH, level->linewidth,
			 KWIN_ATT_END);
    if ( (num_areas < 1) || (level->num_pixel_segments < NUM_SECTIONS) )
    {
	/*  Either whole canvas is to be refreshed (hence all cached pixel
	    segments should be drawn) or there are not enough segments to be
	    sectioned, so they are all drawn too  */
	kwin_draw_segments_TRANSITION (pixcanvas,
				       level->pix_x0, level->pix_y0,
				       level->pix_x1, level->pix_y1,
				       level->num_pixel_segments,
				       pixel_value);
	return;
    }
    if (level->num_area_segments < 1) return;
    kwin_draw_segments_TRANSITION (canvas_get_pixcanvas (canvas),
				   level->pix_area_x0, level->pix_area_y0,
				   level->pix_area_x1, level->pix_area_y1,
				   level->num_area_segments, pixel_value);
}   /*  End Function draw_level  */

static void compute_pixel_coords (KWorldCanvas canvas,
				  struct levelinfo_type *level,
				  struct win_scale_type *win_scale,
				  unsigned int num_pix_segs)
/*  [SUMMARY] Compute pixel co-ordinates.
    <cimage> The world canvas.
    <level> The contour level.
    <win_scale> The window scaling information.
    <num_pix_segs> The number of pixel segments.
    [RETURNS] Nothing.
*/
{
    unsigned int count, section_count, subsection_count, pix_section_size;
    double sec_xmin, sec_ymin, sec_xmax, sec_ymax;
    double x0, y0, x1, y1;
    /*static char function_name[] = "compute_pixel_coords";*/

    level->num_pixel_segments = num_pix_segs;
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE,
				       level->num_pixel_segments,
				       level->pix_x0, level->pix_y0,
				       level->pix_x0, level->pix_y0);
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE,
				       level->num_pixel_segments,
				       level->pix_x1, level->pix_y1,
				       level->pix_x1, level->pix_y1);
    pix_section_size = level->num_pixel_segments / NUM_SECTIONS;
    if (num_pix_segs == level->num_segments)
    {
	/*  Since all segments are visible, assume that the order of the
	    segments in sections will be the same for linear world and
	    pixel co-ordinates  */
	canvas_convert_from_canvas_coords (canvas, FALSE, TRUE,
					   NUM_SECTIONS,
					   level->linear_limits.xmin,
					   level->linear_limits.ymin,
					   level->pix_limits.xmin,
					   level->pix_limits.ymin);
	canvas_convert_from_canvas_coords (canvas, FALSE, TRUE,
					   NUM_SECTIONS,
					   level->linear_limits.xmax,
					   level->linear_limits.ymax,
					   level->pix_limits.xmax,
					   level->pix_limits.ymax);
    }
    else if (level->num_pixel_segments >= NUM_SECTIONS)
    {
	/*  Have to compute limits on pixel segments  */
	sec_xmin = TOOBIG;
	sec_xmax = -TOOBIG;
	sec_ymin = TOOBIG;
	sec_ymax = -TOOBIG;
	for (count = 0, section_count = 0, subsection_count = 0;
	     count < num_pix_segs; ++count, ++subsection_count)
	{
	    if (subsection_count >= pix_section_size)
	    {
		/*  End of section: save current section and go to next  */
		level->pix_limits.xmin[section_count] = sec_xmin;
		level->pix_limits.xmax[section_count] = sec_xmax;
		level->pix_limits.ymin[section_count] = sec_ymin;
		level->pix_limits.ymax[section_count] = sec_ymax;
		if (++section_count < NUM_SECTIONS)
		{
		    sec_xmin = TOOBIG;
		    sec_xmax = -TOOBIG;
		    sec_ymin = TOOBIG;
		    sec_ymax = -TOOBIG;
		    subsection_count = 0;
		}
		else
		{
		    subsection_count = 1; /* Carry one with the last section */
		    section_count = NUM_SECTIONS - 1;
		}
	    }
	    x0 = level->pix_x0[count];
	    y0 = level->pix_y0[count];
	    x1 = level->pix_x1[count];
	    y1 = level->pix_y1[count];
	    /*  Update section limits  */
	    if (x0 < sec_xmin) sec_xmin = x0;
	    if (x0 > sec_xmax) sec_xmax = x0;
	    if (y0 < sec_ymin) sec_ymin = y0;
	    if (y0 > sec_ymax) sec_ymax = y0;
	    if (x1 < sec_xmin) sec_xmin = x1;
	    if (x1 > sec_xmax) sec_xmax = x1;
	    if (y1 < sec_ymin) sec_ymin = y1;
	    if (y1 > sec_ymax) sec_ymax = y1;
	}
	/*  Save last section limits  */
	level->pix_limits.xmin[section_count] = sec_xmin;
	level->pix_limits.xmax[section_count] = sec_xmax;
	level->pix_limits.ymin[section_count] = sec_ymin;
	level->pix_limits.ymax[section_count] = sec_ymax;
    }
}   /*  End Function compute_pixel_coords  */

static void _contour_dealloc_levels (KContourImage cimage)
/*  [SUMMARY] Dealloc the contour levels information.
    <cimage> The KContourImage object.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    /*static char function_name[] = "_contour_dealloc_levels";*/

    if (cimage->num_levels > 0)
    {
	m_free (cimage->contour_levels);
	for (count = 0; count < cimage->num_levels; ++count)
	{
	    if (cimage->world_x0[count] != NULL)
		m_free (cimage->world_x0[count]);
	    if (cimage->world_y0[count] != NULL)
		m_free (cimage->world_y0[count]);
	    if (cimage->world_x1[count] != NULL)
		m_free (cimage->world_x1[count]);
	    if (cimage->world_y1[count] != NULL)
		m_free (cimage->world_y1[count]);
	    if (cimage->levels[count].linear_x0 != NULL)
		m_free (cimage->levels[count].linear_x0);
	    if (cimage->levels[count].linear_y0 != NULL)
		m_free (cimage->levels[count].linear_y0);
	    if (cimage->levels[count].linear_x1 != NULL)
		m_free (cimage->levels[count].linear_x1);
	    if (cimage->levels[count].linear_y1 != NULL)
		m_free (cimage->levels[count].linear_y1);
	    if (cimage->levels[count].pix_x0 != NULL)
		m_free (cimage->levels[count].pix_x0);
	    if (cimage->levels[count].pix_y0 != NULL)
		m_free (cimage->levels[count].pix_y0);
	    if (cimage->levels[count].pix_x1 != NULL)
		m_free (cimage->levels[count].pix_x1);
	    if (cimage->levels[count].pix_y1 != NULL)
		m_free (cimage->levels[count].pix_y1);
	    if (cimage->levels[count].pix_area_x0 != NULL)
		m_free (cimage->levels[count].pix_area_x0);
	    if (cimage->levels[count].pix_area_y0 != NULL)
		m_free (cimage->levels[count].pix_area_y0);
	    if (cimage->levels[count].pix_area_x1 != NULL)
		m_free (cimage->levels[count].pix_area_x1);
	    if (cimage->levels[count].pix_area_y1 != NULL)
		m_free (cimage->levels[count].pix_area_y1);
	}
	m_free (cimage->world_x0);
	m_free (cimage->world_y0);
	m_free (cimage->world_x1);
	m_free (cimage->world_y1);
	m_free (cimage->levels);
    }
    cimage->num_levels = 0;
    cimage->colourname = NULL;
    cimage->world_segs_valid = FALSE;
    cimage->contour_levels = NULL;
    cimage->levels = NULL;
    cimage->world_x0 = NULL;
    cimage->world_y0 = NULL;
    cimage->world_x1 = NULL;
    cimage->world_y1 = NULL;
    cimage->levels = NULL;
}   /*  End Function _contour_dealloc_levels  */
