/*LINTLIBRARY*/
/*  misc.c

    This code provides KContourImage objects.

    Copyright (C) 1998  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating contour images.


    Written by      Richard Gooch   17-SEP-1998

    Last updated by Richard Gooch   17-SEP-1998


*/
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <karma.h>
#include <karma_contour.h>
#include <karma_ex.h>


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int contour_parse_levels (double *contour_levels,
				   unsigned int max_levels, CONST char *string,
				   double min, double max)
/*  [SUMMARY] Parse a textual contour level specification.
    <contour_levels> The raw contour levels are written here.
    <max_levels> The maximum number of levels.
    <string> The string to parse.
    <min> The minimum value in the data range.
    <max> The maximum value in the data range.
    [RETURNS] The number of levels. On error 0 is returned.
*/
{
    flag relative, geometric;
    unsigned int num_levels;
    double value, start, stop, increment, first_inc;
    CONST char *p = string;

    for (num_levels = 0; p != NULL; )
    {
	if (*p == '%')
	{
	    relative = TRUE;
	    ++p;
	}
	else relative = FALSE;
	/*  Get a value  */
	value = ex_float (p, &p);
	if ( (p == NULL) || isdigit (*p) || (*p == '.') || (*p == '%') ||
	     isspace (p[-1]) )
	{
	    /*  End of the line or separate value follows  */
	    if (num_levels >= max_levels)
	    {
		fprintf (stderr, "Too many contour levels: maximum is %u\n",
			 max_levels);
		return (0);
	    }
	    if (relative) value = min + value * 1e-2 * (max - min);
	    contour_levels[num_levels++] = value;
	    continue;
	}
	/*  Special character  */
	if (*p != ':')
	{
	    fprintf (stderr, "Illegal control character: '%c'\n", *p);
	    return (0);
	}
	start = value;
	/*  Get stop and increment values  */
	stop = ex_float (p, &p);
	if (p == NULL)
	{
	    fprintf (stderr, "Increment character, found end of line\n");
	    return (0);
	}
	switch (*p)
	{
	  case '+':
	    geometric = FALSE;
	    break;
	  case '*':
	    geometric = TRUE;
	    break;
	  default:
	    fprintf (stderr,
		     "Missing increment character, found: '%c' instead\n", *p);
	    return (0);
	    /*break;*/
	}
	++p;
	increment = ex_float (p, &p);
	first_inc = TOOBIG;
	for (value = start; value <= stop; )
	{
	    if (num_levels >= max_levels)
	    {
		fprintf (stderr, "Too many contour levels: maximum is %u\n",
			 max_levels);
		return (0);
	    }
	    contour_levels[num_levels++] = relative ?
		( min + value * 1e-2 * (max - min) ) : value;
	    if (geometric) value *= increment;
	    else value += increment;
	    if ( DOUBLE_IS_BLANK (first_inc) ) first_inc = value - start;
	    if (fabs (stop - value) < first_inc * 1e-5) value = stop;
	}
    }
    return (num_levels);
}   /*  End Function contour_parse_levels  */

/*EXPERIMENTAL_FUNCTION*/
void contour_unscale_levels (double *contour_levels, flag *dash,
			     double *linewidths, unsigned int num_levels,
			     double scale, double offset,
			     double neg_width, double pos_width)
/*  [SUMMARY] Unscale contour levels.
    <contour_levels> The scaled contour levels. This is updated with the raw
    (unscaled) contour levels.
    <dash> An array of booleans will be written here. TRUE is written if the
    corresponding scaled contour level is negative, else TRUE is written.
    <linewidths> An array of linewidths will be written here.
    <num_levels> The number of levels.
    <scale> The scale factor.
    <offset> The offset.
    [NOTE] scaled = raw * <<scale>> + <<offset>>.
    <neg_width> The linewidth for negative contours.
    <pos_width> The linewidth for positive contours.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    for (count = 0; count < num_levels; ++count)
    {
	if (contour_levels[count] < 0.0)
	{
	    dash[count] = TRUE;
	    linewidths[count] = neg_width;
	}
	else
	{
	    dash[count] = FALSE;
	    linewidths[count] = pos_width;
	}
	contour_levels[count] = (contour_levels[count] - offset) / scale;
    }
}   /*  End Function contour_unscale_levels  */
