/*LINTLIBRARY*/
/*  main.c

    This code provides 1-D plotting facilities

    Copyright (C) 2000-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed plotting 1-D arrays.


    Written by      Richard Gooch   15-JUN-2000

    Updated by      Richard Gooch   17-JUN-2000: Support NULL <<x>> arrays:
  functions assume equally spaced co-ordinates.

    Updated by      Richard Gooch   20-JUN-2000: Renamed to <graph1d> package.

    Last updated by Richard Gooch   17-MAY-2001: Created <graph1d_points>
  function.


*/
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdarg.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <karma_graph1d.h>
#include <karma_canvas.h>
#include <karma_kwin.h>


#define BUF_SIZE 16384  /*  Number of points  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_lines (KWorldCanvas canvas,
		    unsigned int num_points, CONST double *x, CONST double *y,
		    unsigned long pixel_value,
		    unsigned int num_pixels, unsigned long *pixel_values)
/*  [SUMMARY] Plot a 1-dimensional graph by connecting the vertices.
    <canvas> The world canvas to draw to.
    <num_points> The number of data points to plot.
    <x> The array of horizontal world co-ordinates. These must be in left to
    right order. If this is NULL, the routine assumes the horizontal
    co-ordinates are equally spaced across the world canvas.
    <y> The array of vertical world co-ordinates.
    <pixel_value> The pixel value to draw with.
    <num_pixels> The number of pixels values. If this is 0, <<pixel_value>> is
    used instead.
    <pixel_values> The array of pixel values.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double x0, x1, y0, y1, min_wx, max_wx, min_wy, max_wy, xscale;
    double xb[2], yb[2];

    /*  Compute the boundaries  */
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, xb,
			   CANVAS_ATT_BOTTOM_Y, yb,
			   CANVAS_ATT_RIGHT_X, xb + 1,
			   CANVAS_ATT_TOP_Y, yb + 1,
			   CANVAS_ATT_END);
    canvas_coords_transform (canvas, 2, xb, FALSE, yb, FALSE);
    if (xb[0] < xb[1])
    {
	min_wx = xb[0];
	max_wx = xb[1];
    }
    else
    {
	min_wx = xb[1];
	max_wx = xb[0];
    }
    if (yb[0] < yb[1])
    {
	min_wy = yb[0];
	max_wy = yb[1];
    }
    else
    {
	min_wy = yb[1];
	max_wy = yb[0];
    }
    xscale = (xb[1] - xb[0]) / (double) (num_points - 1);
    for (count = 0; count < num_points - 1; ++count)
    {
	y0 = y[count];
	y1 = y[count + 1];
	if ( DOUBLE_IS_BLANK (y0) || DOUBLE_IS_BLANK (y1) ) continue;
	if (x == NULL)
	{
	    x0 = xb[0] + (double) count * xscale;
	    x1 = x0 + xscale;
	}
	else
	{
	    x0 = x[count];
	    x1 = x[count + 1];
	}
	if ( (x0 < min_wx) || (x0 > max_wx) ||
	     (x1 < min_wx) || (x1 > max_wx) ||
	     (y0 < min_wy) || (y0 > max_wy) ||
	     (y1 < min_wy) || (y1 > max_wy) ) continue;
	if (num_pixels > 0)
	    pixel_value = pixel_values[count * (num_pixels - 1) /
				      (num_points - 1)];
	canvas_draw_line_p (canvas, x0, y0, x1, y1, pixel_value);
    }
}   /*  End Function graph1d_lines  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_steps (KWorldCanvas canvas,
		    unsigned int num_points, CONST double *x, CONST double *y,
		    unsigned long pixel_value,
		    unsigned int num_pixels, unsigned long *pixel_values)
/*  [SUMMARY] Plot a 1-dimensional graph by drawing steps.
    <canvas> The world canvas to draw to.
    <num_points> The number of data points to plot.
    <x> The array of horizontal world co-ordinates. These must be in left to
    right order. If this is NULL, the routine assumes the horizontal
    co-ordinates are equally spaced across the world canvas.
    <y> The array of vertical world co-ordinates.
    <pixel_value> The pixel value to draw with.
    <num_pixels> The number of pixels values. If this is 0, <<pixel_value>> is
    used instead.
    <pixel_values> The array of pixel values.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    double min_wx, max_wx, min_wy, max_wy, last_x, next_x, xscale;
    double xb[2], yb[2];

    /*  Compute the boundaries  */
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, xb,
			   CANVAS_ATT_BOTTOM_Y, yb,
			   CANVAS_ATT_RIGHT_X, xb + 1,
			   CANVAS_ATT_TOP_Y, yb + 1,
			   CANVAS_ATT_END);
    canvas_coords_transform (canvas, 2, xb, FALSE, yb, FALSE);
    if (xb[0] < xb[1])
    {
	min_wx = xb[0];
	max_wx = xb[1];
    }
    else
    {
	min_wx = xb[1];
	max_wx = xb[0];
    }
    if (yb[0] < yb[1])
    {
	min_wy = yb[0];
	max_wy = yb[1];
    }
    else
    {
	min_wy = yb[1];
	max_wy = yb[0];
    }
    last_x = xb[0];
    xscale = (xb[1] - xb[0]) / (double) (num_points - 1);
    for (count = 0; count < num_points; ++count, last_x = next_x)
    {
	flag draw_h = TRUE;
	double x0, y0, y1;

	x0 = x ? x[count] : (xb[0] + (double) count * xscale);
	if (count + 1 >= num_points) next_x = x0;
	else next_x = x ? ( (x0 + x[count + 1]) * 0.5 ) : (x0 + xscale * 0.5);
	if ( (x0 < min_wx) || (x0 > max_wx) ) continue;
	y0 = y[count];
	y1 = y[count + 1];
	if ( DOUBLE_IS_BLANK (y0) ) continue;
	if (y0 < min_wy) y0 = min_wy, draw_h = FALSE;
	if (y0 > max_wy) y0 = max_wy, draw_h = FALSE;
	if (num_pixels > 0)
	    pixel_value = pixel_values[count * (num_pixels - 1) /
				      (num_points - 1)];
	if (draw_h)
	    canvas_draw_line_p (canvas, last_x, y0, next_x, y0, pixel_value);
	if ( DOUBLE_IS_BLANK (y1) ) continue;
	if (y1 < min_wy)
	{
	    if (!draw_h) continue;
	    y1 = min_wy;
	}
	if (y1 > max_wy)
	{
	    if (!draw_h) continue;
	    y1 = max_wy;
	}
	canvas_draw_line_p (canvas, next_x, y0, next_x, y1, pixel_value);
    }
}   /*  End Function graph1d_steps  */

/*EXPERIMENTAL_FUNCTION*/
void graph1d_points (KWorldCanvas canvas,
		     unsigned int num_points, CONST double *x, CONST double *y,
		     unsigned long pixel_value)
/*  [SUMMARY] Plot a 1-dimensional graph by drawing crosses.
    <canvas> The world canvas to draw to.
    <num_points> The number of data points to plot.
    <x> The array of horizontal world co-ordinates. If this is NULL, the
    routine assumes the horizontal co-ordinates are equally spaced across the
    world canvas.
    <y> The array of vertical world co-ordinates.
    <pixel_value> The pixel value to draw with.
    [RETURNS] Nothing.
*/
{
    flag have_x = x ? TRUE : FALSE;
    int x_offset, y_offset, x_pixels, y_pixels;
    double min_wx, max_wx, min_wy, max_wy, xscale;
    double x_left, y_bottom, x_right, y_top;
    double xb[2], yb[2];

    /*  Compute the boundaries  */
    canvas_get_attributes (canvas,
			   CANVAS_ATT_X_OFFSET, &x_offset,
			   CANVAS_ATT_Y_OFFSET, &y_offset,
			   CANVAS_ATT_X_PIXELS, &x_pixels,
			   CANVAS_ATT_Y_PIXELS, &y_pixels,
			   CANVAS_ATT_LEFT_X, xb,
			   CANVAS_ATT_BOTTOM_Y, yb,
			   CANVAS_ATT_RIGHT_X, xb + 1,
			   CANVAS_ATT_TOP_Y, yb + 1,
			   CANVAS_ATT_END);
    x_left = x_offset;
    y_top = y_offset;
    x_right = x_offset + x_pixels - 1;
    y_bottom = y_offset + y_pixels - 1;
    canvas_coords_transform (canvas, 2, xb, FALSE, yb, FALSE);
    if (xb[0] < xb[1])
    {
	min_wx = xb[0];
	max_wx = xb[1];
    }
    else
    {
	min_wx = xb[1];
	max_wx = xb[0];
    }
    if (yb[0] < yb[1])
    {
	min_wy = yb[0];
	max_wy = yb[1];
    }
    else
    {
	min_wy = yb[1];
	max_wy = yb[0];
    }
    xscale = (xb[1] - xb[0]) / (double) (num_points - 1);
    while (num_points > 0)
    {   /*  Process in blocks  */
	unsigned int count, npoints;
	double xbuf[BUF_SIZE], ybuf[BUF_SIZE];
	double x0[BUF_SIZE], y0[BUF_SIZE], x1[BUF_SIZE], y1[BUF_SIZE];

	/*  Extract a list of points which lie inside the boundary  */
	for (count = 0; (count < BUF_SIZE) && (num_points > 0);
	     --num_points, ++x, ++y)
	{
	    double xt, yt;

	    xt = have_x ? *x : (xb[0] + (double) count * xscale);
	    yt = *y;
	    if ( DOUBLE_IS_BLANK (xt) || (xt < min_wx) || (xt > max_wx) )
		continue;
	    if ( DOUBLE_IS_BLANK (yt) || (yt < min_wy) || (yt > max_wy) )
		continue;
	    xbuf[count] = xt;
	    ybuf[count] = yt;
	    ++count;
	}
	npoints = count;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, npoints,
					   xbuf, ybuf, xbuf, ybuf);
	/*  Generate cross shapes  */
	for (count = 0; count < npoints; ++count)
	{
	    double two = 2.0;

	    x0[count] = xbuf[count] - two;
	    y0[count] = ybuf[count] - two;
	    x1[count] = xbuf[count] + two;
	    y1[count] = ybuf[count] + two;
	}
	kwin_draw_segments_TRANSITION (canvas_get_pixcanvas (canvas),
				       x0, ybuf, x1, ybuf, npoints,
				       pixel_value);
	kwin_draw_segments_TRANSITION (canvas_get_pixcanvas (canvas),
				       xbuf, y0, xbuf, y1, npoints,
				       pixel_value);
    }
}   /*  End Function graph1d_points  */
