/*LINTLIBRARY*/
#ifdef HAS_X11
/*  X11.c

    This code provides the X11 driver for KPixCanvas objects.

    Copyright (C) 1995-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating a simple pixel
    canvas (window) independent of the graphics system in use. This file
    contains the X11 code.


    Written by      Richard Gooch   20-SEP-1995

    Updated by      Richard Gooch   22-SEP-1995: Completed passing of RGB masks
  and offsets to <kwin_create_generic>.

    Updated by      Richard Gooch   19-FEB-1996: Modified interface for
  <clear_area> to support width and height parameters.

    Updated by      Richard Gooch   23-FEB-1996: Clip (sub)image to cache size
  in <draw_cached_image> and added <<wait>> parameter.

    Updated by      Richard Gooch   26-FEB-1996: Trap NULL GraphicsContext in
  <kwin_set_gc_x>.

    Updated by      Richard Gooch   13-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   19-APR-1996: Added support for
  KWIN_FUNC_SET_LINEWIDTH.

    Updated by      Richard Gooch   22-APR-1996: Added support for non-1:1
  image display with 16bit PseudoColour software colourmap.

    Updated by      Richard Gooch   4-MAY-1996: Documentation improvement for
  <draw_rectangle>.

    Updated by      Richard Gooch   21-MAY-1996: Fixed bug in
  <draw_cached_image> when offset in image exceeds image size.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   10-JUN-1996: Added clipping code to
  <draw_cached_image>.

    Updated by      Richard Gooch   15-SEP-1996: Made use of new <xv_*>
  routines.

    Updated by      Richard Gooch   10-OCT-1996: Made colour query in chunks.

    Updated by      Richard Gooch   12-DEC-1996: Fixed bug in <get_colour>
  where colours in colourcell_type structure were not filled in.

    Updated by      Richard Gooch   19-FEB-1997: Created <set_background_pixel>
  function.

    Updated by      Richard Gooch   16-MAR-1997: Took account of change to
  interface for <draw_arc>.

    Updated by      Richard Gooch   18-MAY-1997: Supported PseudoColour images
  on 24 bit TrueColour canvases.

    Updated by      Richard Gooch   25-MAY-1997: Supported PseudoColour images
  on canvases of any depth.

    Updated by      Richard Gooch   30-MAY-1997: Supported 24 bit TrueColour
  images on TrueColour and DirectColour canvases of any depth.

    Updated by      Richard Gooch   23-JUN-1997: Fixed 24 bit TrueColour images
  displayed on TrueColour and DirectColour canvases with depth != 24.

    Updated by      Richard Gooch   3-JUL-1997: Avoid black and white colours
  allocated by X for DirectColour canvases.

    Updated by      Richard Gooch   25-JUL-1997: Created subimage cache compute
  functions.

    Updated by      Richard Gooch   4-AUG-1997: Added offsets for rounding
  corrections.

    Updated by      Richard Gooch   3-SEP-1997: Added GC cache.

    Updated by      Richard Gooch   11-OCT-1997: Added preliminary line dash
  support.

    Updated by      Richard Gooch   12-AUG-1998: Trap kill by Xlib when Pixmap
  allocation fails.

    Updated by      Richard Gooch   2-SEP-1998: Changed to <draw_points> for
  <draw_arc>.

    Updated by      Richard Gooch   15-OCT-1998: Added <warp_pointer> function.

    Updated by      Richard Gooch   23-OCT-1998: Fixed bug introduced by
  previous change: PseudoColour <query_colourmap> function was not registered.

    Updated by      Richard Gooch   4-DEC-1998: Supported XImages with 24 bpp.

    Updated by      Richard Gooch   12-JUL-1999: Support different magnify
  functions.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   2-SEP-1999: Took account of interface
  change to KPixFuncWarpPointer.

    Updated by      Richard Gooch   13-MAY-2000: Took account of interface
  change to <kwin_create_generic>. Dropped <draw_point> method.

    Updated by      Richard Gooch   18-MAY-2000: Took account of interface
  changes to <imw_scmap_to24_lossy> and <imw_to8_lossy2>.

    Updated by      Richard Gooch   26-JUN-2000: Took account of interface
  change to <kwin_create_generic>: new destroy method.

    Updated by      Richard Gooch   7-JUL-2000: Fixed memory leak: colour names
  were not being freed in <_kwin_X11_destroy>.

    Updated by      Richard Gooch   11-MAR-2001: Cope with min>max after iscale
  transform.

    Updated by      Richard Gooch   26-APR-2001: Improved diagnostic in
  <_kwin_X11_get_colour>.

    Updated by      Richard Gooch   31-MAY-2001: Trap for thin ellipses in
  <_kwin_draw_arc>.

    Last updated by Richard Gooch   18-JUN-2001: Added test for circles in
  <_kwin_draw_arc> to avoid using rasteriser unless necessary. Took account of
  interface change to <kwin_rasterise_draw_arc>.


*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <stdarg.h>
#include <karma.h>
#define X11
#include <karma_kwin.h>
#include <karma_psw_def.h>
#include <karma_imw.h>
#include <karma_xv.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_xi.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include <os.h>

#define CANVAS_MAGIC_NUMBER (unsigned int) 32987453
#define CACHE_DATA_MAGIC_NUMBER (unsigned int) 924723489
#define FONT_MAGIC_NUMBER (unsigned int) 298732498

#define VERIFY_CANVAS(canvas) if (canvas == NULL) \
{fprintf (stderr, "NULL canvas passed\n"); \
 a_prog_bug (function_name); } \
if (canvas->magic_number != CANVAS_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FONT(font) if (font == NULL) \
{fprintf (stderr, "NULL font passed\n"); \
 a_prog_bug (function_name); } \
if (font->magic_number != FONT_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid font object\n"); \
 a_prog_bug (function_name); }

#define MAX_POINTS 10000
#define MAX_ARCS 10000
#define MAX_SEGMENTS 10000
#define PRIMITIVES_PER_GC_CHANGE_THRESHOLD 8
#define GC_CACHE_MAX_SIZE 256

typedef struct X11canvas_type
{
    unsigned int magic_number;
    KPixCanvas pixcanvas;
    Display *display;
    Window window;
    Colormap cmap;
    XVisualInfo vinfo;
    flag shm_available;
    unsigned int gc_cache_size;
    unsigned int primitive_count;
    GC *gc_arr;
    unsigned long *pixel_values;
    XGCValues gcvalues;
    long max_request_size;
    struct colourcell_type *colours;
    XImage *common_ximage;       /*  Only used for dumping to Pixmaps  */
    uaddr pix_red_offset;
    uaddr pix_green_offset;
    uaddr pix_blue_offset;
} *X11Canvas;

#define KPixHookCanvas X11Canvas
#include <karma_kwin_hooks.h>

/*  Structure declarations  */

struct cache_data_type
{
    KPixCanvas pixcanvas;
    /*  Specific routines define bigger strucures  */
    X11Canvas x11canvas;
    unsigned int magic_number;
    unsigned int width;          /*  The width of the image in the cache   */
    unsigned int height;         /*  The height of the image in the cache  */
    XImage *ximage;
    flag shared;
    Pixmap pixmap;
    unsigned int p_width;        /*  The width of the Pixmap               */
    unsigned int p_height;       /*  The height of the Pixmap              */
};

struct colourcell_type
{
    char *name;
    unsigned long pixel_value;
    unsigned short red;
    unsigned short green;
    unsigned short blue;
    unsigned short pad;
    struct colourcell_type *next;
};


/*  Private data  */
static int num_xpoints_allocated = 0;
static XPoint *xpoints = NULL;
static int num_xarcs_allocated = 0;
static XArc *xarcs = NULL;
static int num_xsegments_allocated = 0;
static XSegment *xsegments = NULL;
static flag caught_error = FALSE;


/*  Mandatory functions  */
STATIC_FUNCTION (X11Canvas create_child,
		 (X11Canvas x11parent, KPixCanvas child) );
STATIC_FUNCTION (flag clear_area, (X11Canvas x11canvas, int x, int y,
				   int width, int height) );
STATIC_FUNCTION (void _kwin_X11_destroy, (X11Canvas x11canvas) );

/*  Optional hook functions  */
STATIC_FUNCTION (flag draw_points,
		 (X11Canvas x11canvas, double *x_array, double *y_array,
		  unsigned int num_points, unsigned long pixel_value) );
STATIC_FUNCTION (flag _kwin_X11_compute_pc_image,
		 (X11Canvas x11canvas,
		  int x_off, int y_off, int x_pixels, int y_pixels,
		  int cache_width, int cache_height,
		  CONST char *slice,
		  CONST uaddr *hoffsets, CONST uaddr *voffsets,
		  unsigned int image_width, unsigned int image_height,
		  unsigned int type, unsigned int conv_type,
		  unsigned int num_pixels, unsigned long *pixel_values,
		  unsigned long blank_pixel,
		  unsigned long min_sat_pixel, unsigned long max_sat_pixel,
		  double i_min, double i_max,
		  flag (*iscale_func) (), void *iscale_info,
		  KPixCanvasImageCache *cache_ptr) );
STATIC_FUNCTION (flag _kwin_X11_compute_rgb_image,
		 (X11Canvas x11canvas,
		  int x_off, int y_off, int x_pixels, int y_pixels,
		  int cache_width, int cache_height,
		  CONST unsigned char *red_slice,
		  CONST unsigned char *green_slice,
		  CONST unsigned char *blue_slice,
		  CONST uaddr *hoffsets, CONST uaddr *voffsets,
		  unsigned int image_width, unsigned int image_height,
		  KPixCanvasImageCache *cache_ptr) );
STATIC_FUNCTION (flag draw_cached_image,
		 (KPixCanvasImageCache cache, flag wait,
		  int parent_x_off, int parent_y_off,
		  int image_width, int image_height,
		  int image_x_off, int image_y_off,
		  int canvas_x_off, int canvas_y_off,
		  int canvas_width, int canvas_height) );
STATIC_FUNCTION (void free_cache_data, (KPixCanvasImageCache cache) );
STATIC_FUNCTION (flag draw_line, (X11Canvas x11canvas,
				  double x0, double y0, double x1, double y1,
				  unsigned long pixel_value) );
STATIC_FUNCTION (flag draw_arc,
		 (X11Canvas x11canvas,
		  double centre_x, double centre_y,
		  double radius_x, double radius_y,
		  int angle1, int angle2, int rotation,
		  unsigned long pixel_value, flag fill) );
STATIC_FUNCTION (flag draw_polygon,
		 (X11Canvas x11canvas, double *x_arr, double *y_arr,
		  unsigned int num_vertices, unsigned long pixel_value,
		  flag convex, flag fill) );
STATIC_FUNCTION (flag draw_string,
		 (X11Canvas x11canvas, double x, double y, CONST char *string,
		  unsigned long pixel_value, flag clear_under) );
STATIC_FUNCTION (flag draw_rectangle,
		 (X11Canvas x11canvas,
		  double x, double y, double width, double height,
		  unsigned long pixel_value, flag fill) );
STATIC_FUNCTION (flag draw_lines,
		 (X11Canvas x11canvas, CONST double *x_arr,CONST double *y_arr,
		  unsigned int num_points, unsigned long pixel_value) );
STATIC_FUNCTION (flag draw_arcs,
		 (X11Canvas x11canvas,
		  double *x, double *y, double *width, double *height,
		  int *angle1, int *angle2, unsigned int num_ellipses,
		  unsigned long pixel_value, flag fill) );
STATIC_FUNCTION (flag draw_segments,
		 (X11Canvas x11canvas,
		  double *x0, double *y0, double *x1, double *y1,
		  unsigned int num_segments, unsigned long pixel_value) );
STATIC_FUNCTION (flag set_linewidth, (X11Canvas x11canvas, double linewidth) );
STATIC_FUNCTION (flag set_background_pixel,
		 (X11Canvas x11canvas, unsigned long pixel_value) );
STATIC_FUNCTION (flag set_line_dash, (X11Canvas x11canvas, flag dash) );

/*  Private functions  */
STATIC_FUNCTION (flag get_colour,
		 (X11Canvas x11canvas, CONST char *colourname,
		  unsigned long *pixel_value, unsigned short *red,
		  unsigned short *green, unsigned short *blue) );
STATIC_FUNCTION (XFontStruct *load_font,
		 (X11Canvas x11canvas, CONST char *fontname) );
STATIC_FUNCTION (void unload_font,
		 (X11Canvas x11canvas, XFontStruct *x11font) );
STATIC_FUNCTION (flag get_string_size,
		 (XFontStruct *x11font, CONST char *string, va_list argp) );
STATIC_FUNCTION (void set_font, (X11Canvas x11canvas, XFontStruct *x11font) );
STATIC_FUNCTION (flag query_colourmap,
		 (X11Canvas canvas, unsigned long *pixels,
		  unsigned short *reds, unsigned short *greens,
		  unsigned short *blues, unsigned int num_colours) );
STATIC_FUNCTION (void warp_pointer,
		 (X11Canvas, int x, int y, flag relative, flag discard) );
STATIC_FUNCTION (void set_pixel_in_gc,
		 (X11Canvas x11canvas, unsigned long pixel_value) );
STATIC_FUNCTION (flag size_cache,
		 (X11Canvas x11canvas, KPixCanvasImageCache *cache_ptr,
		  unsigned int width, unsigned int height) );
STATIC_FUNCTION (void alloc_points, (X11Canvas x11canvas, int num_points) );
STATIC_FUNCTION (void alloc_arcs, (X11Canvas x11canvas, int num_arcs) );
STATIC_FUNCTION (void alloc_segments, (X11Canvas x11canvas,int num_segments) );
STATIC_FUNCTION (flag pc_draw_lossy,
		 (XImage *ximage, int x_off, int y_off,
		  int out_width, int out_height,
		  CONST char *inp_image,
		  CONST iaddr *inp_hoffsets, CONST iaddr *inp_voffsets,
		  int inp_width, int inp_height,
		  unsigned int inp_type, unsigned int conv_type,
		  unsigned int cmap_size, CONST unsigned long *pixels,
		  unsigned int blank_pixel,
		  unsigned int min_sat_pixel, unsigned int max_sat_pixel,
		  double i_min, double i_max,
		  flag (*iscale_func) (), void *iscale_info) );
STATIC_FUNCTION (flag rgb_draw_lossy,
		 (XImage *ximage, int x_off, int y_off,
		  int out_width, int out_height,
		  CONST unsigned char *inp_red_image,
		  CONST unsigned char *inp_green_image,
		  CONST unsigned char *inp_blue_image,
		  CONST iaddr *inp_hoffsets,
		  CONST iaddr *inp_voffsets,
		  int inp_width, int inp_height,
		  unsigned long red_mask, unsigned long green_mask,
		  unsigned long blue_mask) );
STATIC_FUNCTION (int x11_err_handler, (Display *dpy, XErrorEvent *event) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
KPixCanvas kwin_create_x (Display *display, Window window, GC gc,
			  int xoff, int yoff, int width, int height)
/*  [SUMMARY] Create a pixel canvas from an X window.
    [PURPOSE] This routine will create a pixel canvas, ready for drawing, from
    an X window. Note that the origin of a KPixCanvas is the upper-left corner.
    [NOTE] This routine is only available with the X window system.
    <display> The X display.
    <window> The window ID in which the canvas lies.
    <gc> The graphics context.
    <xoff> The horizontal offset of the canvas origin (upper-left corner)
    relative to the parent window.
    <yoff> The vertical offset of the canvas origin (upper-left corner)
    relative to the parent window.
    <width> The width (vertical extent) of the canvas.
    <height> The height (horizontal extent) of the canvas.
    [RETURNS] A pixel canvas on success, else NULL.
*/
{
    KPixCanvas canvas;
    KPixCanvasFont font;
    X11Canvas x11canvas;
    uaddr sz = sizeof (long);
    uaddr im_red_offset = 0;      /*  Initialised to keep compiler happy  */
    uaddr im_green_offset = 0;    /*  Initialised to keep compiler happy  */
    uaddr im_blue_offset = 0;     /*  Initialised to keep compiler happy  */
    uaddr pix_red_offset = 0;     /*  Initialised to keep compiler happy  */
    uaddr pix_green_offset = 0;   /*  Initialised to keep compiler happy  */
    uaddr pix_blue_offset = 0;    /*  Initialised to keep compiler happy  */
    XWindowAttributes window_attributes;
    int im_byte_order, pix_byte_order;
    unsigned int visual = 0;  /*  Initialised to keep compiler happy  */
    unsigned long im_red_mask, im_green_mask, im_blue_mask;
    unsigned char *ch_ptr;
    XVisualInfo *vinfo;
    static char function_name[] = "kwin_create_x";

    im_byte_order = ImageByteOrder (display);
#ifdef MACHINE_BIG_ENDIAN
    pix_byte_order = MSBFirst;
#else
    pix_byte_order = LSBFirst;
#endif
    /*  Get X colourmap  */
    XGetWindowAttributes (display, window, &window_attributes);
    if (window_attributes.colormap == None)
    {
	fprintf (stderr, "Error: window has no colourmap\n");
	a_prog_bug (function_name);
    }
    if ( ( x11canvas = m_alloc (sizeof *x11canvas) ) == NULL )
    {
	m_error_notify (function_name, "X pixel canvas");
	return (NULL);
    }
    m_clear (x11canvas, sizeof *x11canvas);
    x11canvas->display = display;
    x11canvas->window = window;
    x11canvas->cmap = window_attributes.colormap;
    /*  Get GCValues  */
    m_clear (&x11canvas->gcvalues, sizeof x11canvas->gcvalues);
    if (XGetGCValues (display, gc,
		      GCFunction | GCPlaneMask |
		      GCForeground | GCBackground | GCLineWidth | GCLineStyle |
		      GCCapStyle | GCJoinStyle | GCFillStyle | GCFillRule |
		      GCStipple | GCTileStipXOrigin |
		      GCTileStipYOrigin | GCFont | GCSubwindowMode |
		      GCGraphicsExposures | GCClipXOrigin | GCClipYOrigin |
		      GCDashOffset | GCArcMode,
		      &x11canvas->gcvalues) == 0)
    {
	fprintf (stderr, "Error getting GC values\n");
	m_free (x11canvas);
	return (NULL);
    }
    if ( ( x11canvas->gc_arr = m_alloc (sizeof *x11canvas->gc_arr) ) == NULL )
    {
	m_error_notify (function_name, "GC array");
	m_free (x11canvas);
	return (NULL);
    }
    if ( ( x11canvas->pixel_values =
	   m_alloc (sizeof *x11canvas->pixel_values) ) == NULL )
    {
	m_error_notify (function_name, "GC array");
	m_free (x11canvas->gc_arr);
	m_free (x11canvas);
	return (NULL);
    }
    x11canvas->gc_arr[0] =
	XCreateGC (display, window,
		   GCFunction | GCPlaneMask |
		   GCForeground | GCBackground | GCLineWidth | GCLineStyle|
		   GCCapStyle | GCJoinStyle | GCFillStyle | GCFillRule |
		   GCTileStipXOrigin |
		   GCTileStipYOrigin | GCFont | GCSubwindowMode |
		   GCGraphicsExposures | GCClipXOrigin | GCClipYOrigin |
		   GCDashOffset | GCArcMode,
		   &x11canvas->gcvalues);
    x11canvas->pixel_values[0] = x11canvas->gcvalues.foreground;
    x11canvas->gc_cache_size = 1;
    x11canvas->primitive_count = 0;
    x11canvas->shm_available = xi_check_shared_images_available (display);
    x11canvas->max_request_size = XMaxRequestSize (display);
    x11canvas->common_ximage = NULL;
    x11canvas->colours = NULL;
    vinfo = xv_get_visinfo_for_visual (display, window_attributes.visual);
    if (window_attributes.depth != vinfo->depth)
    {
	fprintf (stderr, "Window depth: %d is not visual depth: %d\n",
		 window_attributes.depth, vinfo->depth);
	a_prog_bug (function_name);
    }
    m_copy (&x11canvas->vinfo, vinfo, sizeof *vinfo);
    XFree ( (void *) vinfo );
    switch (x11canvas->vinfo.class)
    {
      case DirectColor:
      case TrueColor:
	if (window_attributes.depth != 24) break;
	if (x11canvas->vinfo.colormap_size != 256)
	{
	    fprintf (stderr,
		     "%s: colourmap size: %d for %s visual is not 256\n",
		     function_name,
		     x11canvas->vinfo.colormap_size,
		     (x11canvas->vinfo.class ==
		      DirectColor) ? "DirectColour" : "TrueColour");
	    m_free (x11canvas);
	    return (NULL);
	}
	/*  Compute the byte offsets in a word required for drawing images
	    (images are stored in server native format for efficiency).
	    Note that the masks in the vinfo structure are in client order.  */
	switch (x11canvas->vinfo.red_mask)
	{
	  case 0xff:
	    im_red_offset = (im_byte_order == MSBFirst) ? 3 : 0;
	    pix_red_offset = (pix_byte_order == MSBFirst) ? sz - 1 : 0;
	    break;
	  case 0xff00:
	    im_red_offset = (im_byte_order == MSBFirst) ? 2 : 1;
	    pix_red_offset = (pix_byte_order == MSBFirst) ? sz - 2 : 1;
	    break;
	  case 0xff0000:
	    im_red_offset = (im_byte_order == MSBFirst) ? 1 : 2;
	    pix_red_offset = (pix_byte_order == MSBFirst) ? sz - 3 : 2;
	    break;
	  default:
	    fprintf (stderr, "Red mask: 0x%lx for %s visual is not valid\n",
		     x11canvas->vinfo.red_mask,
		     (x11canvas->vinfo.class ==
		      DirectColor) ? "DirectColour" : "TrueColour");
	    m_free (x11canvas);
	    return (NULL);
	    /*break;*/
	}
	switch (x11canvas->vinfo.green_mask)
	{
	  case 0xff:
	    im_green_offset = (im_byte_order == MSBFirst) ? 3 : 0;
	    pix_green_offset = (pix_byte_order == MSBFirst) ? sz - 1 : 0;
	    break;
	  case 0xff00:
	    im_green_offset = (im_byte_order == MSBFirst) ? 2 : 1;
	    pix_green_offset = (pix_byte_order == MSBFirst) ? sz - 2 : 1;
	    break;
	  case 0xff0000:
	    im_green_offset = (im_byte_order == MSBFirst) ? 1 : 2;
	    pix_green_offset = (pix_byte_order == MSBFirst) ? sz - 3 : 2;
	    break;
	  default:
	    fprintf (stderr, "Green mask: 0x%lx for %s visual is not valid\n",
		     x11canvas->vinfo.green_mask,
		     (x11canvas->vinfo.class ==
		      DirectColor) ? "DirectColour" : "TrueColour");
	    m_free (x11canvas);
	    return (NULL);
	    /*break;*/
	}
	switch (x11canvas->vinfo.blue_mask)
	{
	  case 0xff:
	    im_blue_offset = (im_byte_order == MSBFirst) ? 3 : 0;
	    pix_blue_offset = (pix_byte_order == MSBFirst) ? sz - 1 : 0;
	    break;
	  case 0xff00:
	    im_blue_offset = (im_byte_order == MSBFirst) ? 2 : 1;
	    pix_blue_offset = (pix_byte_order == MSBFirst) ? sz - 2 : 1;
	    break;
	  case 0xff0000:
	    im_blue_offset = (im_byte_order == MSBFirst) ? 1 : 2;
	    pix_blue_offset = (pix_byte_order == MSBFirst) ? sz - 3 : 2;
	    break;
	  default:
	    fprintf (stderr, "Blue mask: 0x%lx for %s visual is not valid\n",
		     x11canvas->vinfo.blue_mask,
		     (x11canvas->vinfo.class ==
		      DirectColor) ? "DirectColour" : "TrueColour");
	    m_free (x11canvas);
	    return (NULL);
	    /*break;*/
	}
	/*  Compute the masks required for drawing images (images are stored in
	    server native format for efficiency)  */
	im_red_mask = 0;
	ch_ptr = (unsigned char *) &im_red_mask;
	ch_ptr[im_red_offset] = 0xff;
	im_green_mask = 0;
	ch_ptr = (unsigned char *) &im_green_mask;
	ch_ptr[im_green_offset] = 0xff;
	im_blue_mask = 0;
	ch_ptr = (unsigned char *) &im_blue_mask;
	ch_ptr[im_blue_offset] = 0xff;
	x11canvas->pix_red_offset = pix_red_offset;
	x11canvas->pix_green_offset = pix_green_offset;
	x11canvas->pix_blue_offset = pix_blue_offset;
	break;
      default:
	break;
    }
    switch (x11canvas->vinfo.class)
    {
      case PseudoColor:
	visual = KWIN_VISUAL_PSEUDOCOLOUR;
	break;
      case DirectColor:
	visual = KWIN_VISUAL_DIRECTCOLOUR;
	break;
      case TrueColor:
	visual = KWIN_VISUAL_TRUECOLOUR;
	break;
      case GrayScale:
	visual = KWIN_VISUAL_GRAYSCALE;
	break;
      case StaticGray:
	visual = KWIN_VISUAL_STATICGRAY;
	break;
      case StaticColor:
	visual = KWIN_VISUAL_STATICCOLOUR;
	break;
      default:
	fprintf (stderr, "Illegal visual class: %u\n", x11canvas->vinfo.class);
	a_prog_bug (function_name);
	break;
    }
    if ( (visual == KWIN_VISUAL_DIRECTCOLOUR) ||
	 (visual == KWIN_VISUAL_TRUECOLOUR ) )
    {
	canvas = kwin_create_generic
	    (x11canvas, xoff, yoff, width, height,
	     window_attributes.depth, visual, TRUE, create_child, clear_area,
	     _kwin_X11_destroy,
	     KWIN_ATT_PIX_RED_MASK, x11canvas->vinfo.red_mask,
	     KWIN_ATT_PIX_GREEN_MASK, x11canvas->vinfo.green_mask,
	     KWIN_ATT_PIX_BLUE_MASK, x11canvas->vinfo.blue_mask,
	     KWIN_ATT_IM_RED_MASK, im_red_mask,
	     KWIN_ATT_IM_GREEN_MASK, im_green_mask,
	     KWIN_ATT_IM_BLUE_MASK, im_blue_mask,
	     KWIN_ATT_IM_RED_OFFSET, im_red_offset,
	     KWIN_ATT_IM_GREEN_OFFSET, im_green_offset,
	     KWIN_ATT_IM_BLUE_OFFSET, im_blue_offset,
	     KWIN_ATT_BACKGROUND_PIXEL, x11canvas->gcvalues.background,
	     KWIN_FUNC_COMPUTE_PC_IMAGE, _kwin_X11_compute_pc_image,
	     KWIN_FUNC_COMPUTE_RGB_IMAGE, _kwin_X11_compute_rgb_image,
	     KWIN_FUNC_DRAW_POINTS, draw_points,
	     KWIN_FUNC_DRAW_CACHED_IMAGE, draw_cached_image,
	     KWIN_FUNC_DRAW_LINE, draw_line,
	     KWIN_FUNC_DRAW_ARC, draw_arc,
	     KWIN_FUNC_DRAW_POLYGON, draw_polygon,
	     KWIN_FUNC_DRAW_STRING, draw_string,
	     KWIN_FUNC_DRAW_RECTANGLE, draw_rectangle,
	     KWIN_FUNC_DRAW_LINES, draw_lines,
	     KWIN_FUNC_DRAW_ARCS, draw_arcs,
	     KWIN_FUNC_DRAW_SEGMENTS, draw_segments,
	     KWIN_FUNC_SET_LINEWIDTH, set_linewidth,
	     KWIN_FUNC_SET_BACKGROUND_PIXEL, set_background_pixel,
	     KWIN_FUNC_SET_LINE_DASH, set_line_dash,
	     KWIN_FUNC_FREE_CACHE_DATA, free_cache_data,
	     KWIN_FUNC_GET_COLOUR, get_colour,
	     KWIN_FUNC_LOAD_FONT, load_font,
	     KWIN_FUNC_UNLOAD_FONT, unload_font,
	     KWIN_FUNC_GET_STRING_SIZE, get_string_size,
	     KWIN_FUNC_SET_FONT, set_font,
	     KWIN_FUNC_QUERY_COLOURMAP, query_colourmap,
	     KWIN_FUNC_WARP_POINTER, warp_pointer,
	     KWIN_ATT_END);
    }
    else
    {
	canvas = kwin_create_generic
	    (x11canvas, xoff, yoff, width, height, window_attributes.depth,
	     visual, TRUE, create_child, clear_area,
	     _kwin_X11_destroy,
	     KWIN_ATT_BACKGROUND_PIXEL, x11canvas->gcvalues.background,
	     KWIN_FUNC_COMPUTE_PC_IMAGE, _kwin_X11_compute_pc_image,
	     KWIN_FUNC_COMPUTE_RGB_IMAGE, _kwin_X11_compute_rgb_image,
	     KWIN_FUNC_DRAW_CACHED_IMAGE, draw_cached_image,
	     KWIN_FUNC_DRAW_POINTS, draw_points,
	     KWIN_FUNC_DRAW_LINE, draw_line,
	     KWIN_FUNC_DRAW_ARC, draw_arc,
	     KWIN_FUNC_DRAW_POLYGON, draw_polygon,
	     KWIN_FUNC_DRAW_STRING, draw_string,
	     KWIN_FUNC_DRAW_RECTANGLE, draw_rectangle,
	     KWIN_FUNC_DRAW_LINES, draw_lines,
	     KWIN_FUNC_DRAW_ARCS, draw_arcs,
	     KWIN_FUNC_DRAW_SEGMENTS, draw_segments,
	     KWIN_FUNC_SET_LINEWIDTH, set_linewidth,
	     KWIN_FUNC_SET_BACKGROUND_PIXEL, set_background_pixel,
	     KWIN_FUNC_SET_LINE_DASH, set_line_dash,
	     KWIN_FUNC_FREE_CACHE_DATA, free_cache_data,
	     KWIN_FUNC_GET_COLOUR, get_colour,
	     KWIN_FUNC_LOAD_FONT, load_font,
	     KWIN_FUNC_UNLOAD_FONT, unload_font,
	     KWIN_FUNC_GET_STRING_SIZE, get_string_size,
	     KWIN_FUNC_SET_FONT, set_font,
	     KWIN_FUNC_QUERY_COLOURMAP, query_colourmap,
	     KWIN_FUNC_WARP_POINTER, warp_pointer,
	     KWIN_ATT_END);
    }
    if (canvas == NULL) m_abort (function_name, "pixel canvas");
    x11canvas->magic_number = CANVAS_MAGIC_NUMBER;
    x11canvas->pixcanvas = canvas;
    if ( ( font = kwin_load_font (canvas, "fixed") ) == NULL )
    {
	m_abort (function_name, "font");
    }
    kwin_set_attributes (canvas,
			 KWIN_ATT_FONT, font,
			 KWIN_ATT_END);
    return (canvas);
}   /*  End Function kwin_create_x  */

/*PUBLIC_FUNCTION*/
void kwin_set_gc_x (KPixCanvas canvas, GC gc)
/*  [SUMMARY] Set Graphics Context for an X-based pixel canavas.
    [PURPOSE] This routine will register a new Graphics Context to be used
    when drawing into the pixel canvas. Subsequent drawing operations will use
    the new Graphics Context.
    [NOTE] This routine is only available with the X window system.
    <canvas> The canvas.
    <gc> The Graphics Context.
    [RETURNS] Nothing.
*/
{
    X11Canvas x11canvas;
    unsigned int count;
    static char function_name[] = "kwin_set_gc_x";

    kwin_get_attributes (canvas,
			 KWIN_ATT_LOWER_HANDLE, &x11canvas,
			 KWIN_ATT_END);
    VERIFY_CANVAS (x11canvas);
    if (gc == NULL)
    {
	fprintf (stderr, "NULL GraphicsContext passed\n");
	a_prog_bug (function_name);
    }
    /*  Get GCValues  */
    if (XGetGCValues (x11canvas->display, gc,
		      GCFunction | GCPlaneMask |
		      GCForeground | GCBackground | GCLineWidth | GCLineStyle |
		      GCCapStyle | GCJoinStyle | GCFillStyle | GCFillRule |
		      GCTile | GCStipple | GCTileStipXOrigin |
		      GCTileStipYOrigin | GCFont | GCSubwindowMode |
		      GCGraphicsExposures | GCClipXOrigin | GCClipYOrigin |
		      GCDashOffset | GCArcMode,
		      &x11canvas->gcvalues) == 0)
    {
	fprintf (stderr, "Error getting GC values\n");
	return;
    }
    for (count = 0; count < x11canvas->gc_cache_size; ++count)
    {
	XChangeGC (x11canvas->display, x11canvas->gc_arr[count],
		   GCFunction | GCPlaneMask |
		   GCBackground | GCLineWidth | GCLineStyle |
		   GCCapStyle | GCJoinStyle | GCFillStyle | GCFillRule |
		   GCTile | GCStipple | GCTileStipXOrigin |
		   GCTileStipYOrigin | GCFont | GCSubwindowMode |
		   GCGraphicsExposures | GCClipXOrigin | GCClipYOrigin |
		   GCDashOffset | GCArcMode,
		   &x11canvas->gcvalues);
    }
    x11canvas->primitive_count = 0;
}   /*  End Function kwin_set_gc_x  */

/*PUBLIC_FUNCTION*/
GC kwin_get_gc_x (KPixCanvas canvas)
/*  [SUMMARY] Get Graphics Context for an X-based pixel canavas.
    [PURPOSE] This routine will get the Graphics Context used when drawing into
    the pixel canvas. If the Graphics Context is modified, it should be
    registered prior to drawing on the canvas by calling [<kwin_set_gc_x>].
    [NOTE] This routine is only available with the X window system.
    <canvas> The canvas.
    [RETURNS] The Graphics Context.
*/
{
    X11Canvas x11canvas;
    static char function_name[] = "kwin_get_gc_x";

    kwin_get_attributes (canvas,
			 KWIN_ATT_LOWER_HANDLE, &x11canvas,
			 KWIN_ATT_END);
    VERIFY_CANVAS (x11canvas);
    return (x11canvas->gc_arr[0]);
}   /*  End Function kwin_get_gc_x  */

/*UNPUBLISHED_FUNCTION*/
Window kwin_x11_get_window (KPixCanvas canvas)
/*  [SUMMARY] Get the X window for a pixel canvas.
    <canvas> The pixel canvas.
    [RETURNS] The X window ID.
*/
{
    X11Canvas x11canvas;
    static char function_name[] = "kwin_x11_get_window";

    kwin_get_attributes (canvas,
			 KWIN_ATT_LOWER_HANDLE, &x11canvas,
			 KWIN_ATT_END);
    VERIFY_CANVAS (x11canvas);
    return (x11canvas->window);
}   /*  End Function kwin_x11_get_window  */

/*UNPUBLISHED_FUNCTION*/
Colormap kwin_x11_get_cmap (KPixCanvas canvas)
/*  [SUMMARY] Get the X colourmap for a pixel canvas.
    <canvas> The pixel canvas.
    [RETURNS] The X colourmap.
*/
{
    X11Canvas x11canvas;
    static char function_name[] = "kwin_x11_get_cmap";

    kwin_get_attributes (canvas,
			 KWIN_ATT_LOWER_HANDLE, &x11canvas,
			 KWIN_ATT_END);
    VERIFY_CANVAS (x11canvas);
    return (x11canvas->cmap);
}   /*  End Function kwin_x11_get_cmap  */

/*UNPUBLISHED_FUNCTION*/
void kwin_x11_get_vinfo (KPixCanvas canvas, XVisualInfo *vinfo)
/*  [SUMMARY] Get the X visual information for a pixel canvas.
    <canvas> The pixel canvas.
    <vinfo> The XVisualInfo information is written here.
    [RETURNS] Nothing.
*/
{
    X11Canvas x11canvas;
    static char function_name[] = "kwin_x11_get_vinfo";

    kwin_get_attributes (canvas,
			 KWIN_ATT_LOWER_HANDLE, &x11canvas,
			 KWIN_ATT_END);
    VERIFY_CANVAS (x11canvas);
    m_copy (vinfo, &x11canvas->vinfo, sizeof *vinfo);
}   /*  End Function kwin_x11_get_vinfo  */


/*  Mandatory hook functions follow  */

static X11Canvas create_child (X11Canvas x11parent, KPixCanvas child)
/*  [PURPOSE] This routine will create a child X11 canvas.
    <x11parent> The parent X11 canvas.
    [RETURNS] The child X11 canvas on success, else NULL.
*/
{
    X11Canvas x11child;
    static char function_name[] = "_kwin_X11_create_child";

    VERIFY_CANVAS (x11parent);
    if ( ( x11child = m_alloc (sizeof *x11child) ) == NULL )
    {
	m_error_notify (function_name, "X pixel canvas");
	return (NULL);
    }
    m_copy (x11child, x11parent, sizeof *x11child);
    x11child->pixcanvas = child;
    /*  Create new Graphics Context  */
    if ( ( x11child->gc_arr = m_alloc (sizeof *x11child->gc_arr) ) == NULL )
    {
	m_error_notify (function_name, "GC array");
	m_free (x11child);
	return (NULL);
    }
    if ( ( x11child->pixel_values =
	   m_alloc (sizeof *x11child->pixel_values) ) == NULL )
    {
	m_error_notify (function_name, "GC array");
	m_free (x11child->gc_arr);
	m_free (x11child);
	return (NULL);
    }
    x11child->gcvalues.foreground = x11parent->pixel_values[0];
    x11child->gc_arr[0] =
	XCreateGC (x11child->display, x11child->window,
		   GCFunction | GCPlaneMask |
		   GCForeground | GCBackground | GCLineWidth |
		   GCLineStyle | GCCapStyle | GCJoinStyle |
		   GCFillStyle | GCFillRule |
		   GCTileStipXOrigin | GCTileStipYOrigin | GCFont |
		   GCSubwindowMode | GCGraphicsExposures |
		   GCClipXOrigin | GCClipYOrigin | GCDashOffset |
		   GCArcMode,
		   &x11child->gcvalues);
    x11child->pixel_values[0] = x11child->gcvalues.foreground;
    x11child->gc_cache_size = 1;
    x11child->primitive_count = 0;
    return (x11child);
}   /*  End Function create_child  */

static flag clear_area (X11Canvas x11canvas, int x, int y,
			int width, int height)
/*  [PURPOSE] This routine will clear an area in an X11 canvas.
    <x11canvas> The X11 canvas.
    <x> The horizontal offset of the area.
    <y> The vertical offset of the area.
    <width> The width of the area.
    <height> The height of the area.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_kwin_X11_clear_area";

    VERIFY_CANVAS (x11canvas);
    XClearArea (x11canvas->display, x11canvas->window,
		x, y, (unsigned int) width, (unsigned int) height, False);
    return (TRUE);
}   /*  End Function clear_area  */

static void _kwin_X11_destroy (X11Canvas x11canvas)
/*  [PURPOSE] This routine will destroy an X11 canvas.
    <x11canvas> The X11 canvas.
    [RETURNS] Nothing.
*/
{
    struct colourcell_type *curr, *next;
    static char function_name[] = "_kwin_X11_destroy";

    VERIFY_CANVAS (x11canvas);
    m_free (x11canvas->gc_arr);
    m_free (x11canvas->pixel_values);
    for (curr = x11canvas->colours; curr != NULL; curr = next)
    {
	next = curr->next;
	XFreeColors (x11canvas->display, x11canvas->cmap,
		     &curr->pixel_value, 1, 0);
	m_free (curr->name);
	m_free (curr);
    }
    if (x11canvas->common_ximage != NULL)
	xi_destroy_image (x11canvas->display, x11canvas->common_ximage, FALSE);
    x11canvas->magic_number = 0;
    m_free (x11canvas);
}   /*  End Function destroy  */


/*  Optional hook functions follow  */

static flag draw_points (X11Canvas x11canvas, double *x_array, double *y_array,
			 unsigned int num_points, unsigned long pixel_value)
/*  [PURPOSE] This routine will draw multiple points onto an X11 canvas.
    <x11canvas> The X11 canvas.
    <x_array> The horizontal co-ordinates of the points.
    <y_array> The vertical co-ordinates of the points.
    <num_points> The number of points.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int ipoint_count, xpoint_count;
    static char function_name[] = "_kwin_X11_draw_points";

    VERIFY_CANVAS (x11canvas);
    set_pixel_in_gc (x11canvas, pixel_value);
    if (num_points == 1)
    {
	XDrawPoint ( x11canvas->display, x11canvas->window,
		     x11canvas->gc_arr[0],
		     (int) (x_array[0] + 1e-3), (int) (y_array[0] + 1e-3) );
	return (TRUE);
    }
    alloc_points (x11canvas, num_points);
    for (ipoint_count = 0, xpoint_count = 0; ipoint_count < num_points;
	 ++ipoint_count, ++xpoint_count)
    {
	if (xpoint_count >= num_xpoints_allocated)
	{
	    /*  Send some points now  */
	    XDrawPoints (x11canvas->display, x11canvas->window,
			 x11canvas->gc_arr[0], xpoints, xpoint_count,
			 CoordModeOrigin);
	    xpoint_count = 0;
	}
	xpoints[xpoint_count].x = x_array[ipoint_count] + 1e-3;
	xpoints[xpoint_count].y = y_array[ipoint_count] + 1e-3;
    }
    /*  Draw remaining points  */
    if (xpoint_count > 0)
	XDrawPoints (x11canvas->display, x11canvas->window,
		     x11canvas->gc_arr[0], xpoints, xpoint_count,
		     CoordModeOrigin);
    return (TRUE);
}   /*  End Function draw_points  */

static flag _kwin_X11_compute_pc_image (X11Canvas x11canvas,
					int x_off, int y_off,
					int x_pixels, int y_pixels,
					int cache_width, int cache_height,
					CONST char *slice,
					CONST uaddr *hoffsets,
					CONST uaddr *voffsets,
					unsigned int image_width,
					unsigned int image_height,
					unsigned int type,
					unsigned int conv_type,
					unsigned int num_pixels,
					unsigned long *pixel_values,
					unsigned long blank_pixel,
					unsigned long min_sat_pixel,
					unsigned long max_sat_pixel,
					double i_min, double i_max,
					flag (*iscale_func) (),
					void *iscale_info,
					KPixCanvasImageCache *cache_ptr)
/*  [PURPOSE] This routine will compute a subimage cache for a 2-dimensional
    slice of a Karma array for an X11 canvas. This slice may be tiled. The
    slice is a PseudoColour subimage.
    <x11canvas> The X11 canvas.
    <x_off> The horizontal offset of the subimage, relative to the top-left
    corner of the image cache.
    <y_off> The vertical offset of the subimage, relative to the top-left
    corner of the image cache.
    <x_pixels> The number of horizontal pixels to generate.
    <y_pixels> The number of vertical pixels to generate.
    <cache_width> The width of the image cache.
    <cache_height> The height of the image cache.
    <slice> The start of the slice image data.
    <hoffsets> The array of horizontal byte offsets for the input image.
    <voffsets> The array of vertical byte offsets for the input image.
    <image_width> The width of the input image (in values).
    <image_height> The height of the input image (in values).
    <type> The type of the slice image data.
    <conv_type> The input conversion type (when the input is complex).
    <num_pixels> The number of pixels in the pixel array.
    <pixel_values> The array of pixel values.
    <blank_pixel> The pixel value to be used when the intensity value is an
    undefined value.
    <min_sat_pixel> The pixel value to be used when the intensity value is
    below the minimum value.
    <max_sat_pixel> The pixel value to be used when the intensity value is
    above the maximum value.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <iscale_func> The function to be called when non-linear intensity scaling
    is required. If NULL, linear intensity scaling is used. The prototype
    function is [<kwin_PROTO_iscale_func>].
    <iscale_info> A pointer to arbitrary information for <<iscale_func>>.
    <cache_ptr> Cache data is written here. The routine will produce cache data
    which will vastly increase the speed of subsequent operations on this data.
    Prior to process exit, a call MUST be made to <<kwin_free_cache_data>>,
    otherwise shared memory segments could remain after the process exits.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KPixCanvasImageCache cache;
    MagnifyFunc magnify_func;
    iaddr bytes_per_pixel;
    unsigned int count;
    uaddr im_red_offset, im_green_offset, im_blue_offset;
    unsigned char *ub_ptr = NULL;
    unsigned char *out_image;
    unsigned char ub_pixels[256];
    XImage *ximage;
    static char function_name[] = "_kwin_X11_compute_pc_image";

    VERIFY_CANVAS (x11canvas);
    if ( !size_cache (x11canvas, cache_ptr, cache_width, cache_height) )
    {
	return (FALSE);
    }
    cache = *cache_ptr;
    ximage = (cache->ximage ==NULL) ? x11canvas->common_ximage : cache->ximage;
    if (ximage->bits_per_pixel % 8 != 0)
    {
	fprintf (stderr, "bits_per_pixel: %d is not a multiple of 8\n",
		 ximage->bits_per_pixel);
	a_prog_bug (function_name);
    }
    bytes_per_pixel = ximage->bits_per_pixel / 8;
    out_image = (unsigned char *) ximage->data;
    out_image += bytes_per_pixel * x_off;
    out_image += ximage->bytes_per_line * y_off;
    /*  Now have an XImage big enough  */
    if ( (ximage->depth == 8) && (num_pixels <= 256) )
    {
	/*  Convert pixel values  */
	for (count = 0; count < num_pixels; ++count)
	{
	    if (pixel_values[count] > 255)
	    {
		fprintf (stderr, "%s: pixel value: %lu  over 255\n",
			 function_name, pixel_values[count]);
		return (FALSE);
	    }
	    ub_pixels[count] = pixel_values[count];
	}
    }
    if (ximage->depth == 24)
    {
	kwin_get_attributes (x11canvas->pixcanvas,
			     KWIN_ATT_IM_RED_OFFSET, &im_red_offset,
			     KWIN_ATT_IM_GREEN_OFFSET, &im_green_offset,
			     KWIN_ATT_IM_BLUE_OFFSET, &im_blue_offset,
			     KWIN_ATT_END);
	ub_ptr = (unsigned char *) pixel_values;
    }
    kwin_get_attributes (x11canvas->pixcanvas,
			 KWIN_ATT_MAGNIFY_FUNC, &magnify_func,
			 KWIN_ATT_END);
    ub_ptr = (unsigned char *) pixel_values;
    if ( (ximage->depth == 8) &&
	 (image_width == x_pixels) && (image_height == y_pixels) )
    {
	/*  New drawing routines  */
	/*  Note the casts from (uaddr *) to (iaddr *) for the offset
	    arrays. This is dodgy, but it should work.  */
	if ( !imw_to8_oi (out_image, bytes_per_pixel, ximage->bytes_per_line,
			  x_pixels, y_pixels,
			  slice, (iaddr *) hoffsets, (iaddr *) voffsets,
			  type, conv_type,
			  num_pixels, ub_pixels,
			  blank_pixel, min_sat_pixel, max_sat_pixel,
			  i_min, i_max, iscale_func, iscale_info ) )
	{
	    fprintf (stderr, "Error drawing image into X11Canvas\n");
	    return (FALSE);
	}
    }
    else if ( (ximage->depth == 24) &&
	      (image_width == x_pixels) && (image_height == y_pixels) &&
	      (num_pixels == 65536) && (type == K_USHORT) &&
	      (0 == (int) i_min) && (65535 == (int) i_max) )
    {
	if ( !imw_scmap_16to24_o
	     (out_image + im_red_offset,
	      out_image + im_green_offset,
	      out_image + im_blue_offset,
	      FALSE, bytes_per_pixel, ximage->bytes_per_line,
	      x_pixels, y_pixels,
	      (CONST unsigned short *) slice,
	      (iaddr *) hoffsets, (iaddr *) voffsets,
	      ub_ptr + x11canvas->pix_red_offset,
	      ub_ptr + x11canvas->pix_green_offset,
	      ub_ptr + x11canvas->pix_blue_offset,
	      sizeof *pixel_values) )
	{
	    fprintf (stderr, "Error drawing image into X11Canvas\n");
	    return (FALSE);
	}
    }
    else if ( (ximage->depth == 24) &&
	      (num_pixels == 65536) && (type == K_USHORT) &&
	      (0 == (int) i_min) && (65535 == (int) i_max) )
    {
	if ( !imw_scmap_16to24_lossy
	     (out_image + im_red_offset,
	      out_image + im_green_offset,
	      out_image + im_blue_offset,
	      FALSE, bytes_per_pixel, ximage->bytes_per_line,
	      x_pixels, y_pixels,
	      (CONST unsigned short *) slice, image_width, image_height,
	      (iaddr *) hoffsets, (iaddr *) voffsets,
	      ub_ptr + x11canvas->pix_red_offset,
	      ub_ptr + x11canvas->pix_green_offset,
	      ub_ptr + x11canvas->pix_blue_offset,
	      sizeof *pixel_values) )
	{
	    fprintf (stderr, "Error drawing image into X11Canvas\n");
	    return (FALSE);
	}
    }
    else if (ximage->depth == 24)
    {
	if ( !imw_scmap_to24_lossy
	     (out_image + im_red_offset,
	      out_image + im_green_offset,
	      out_image + im_blue_offset,
	      FALSE, TRUE, bytes_per_pixel, ximage->bytes_per_line,
	      x_pixels, y_pixels,
	      (CONST char *) slice, (iaddr *) hoffsets, (iaddr *) voffsets,
	      image_width, image_height,
	      type, conv_type, num_pixels, pixel_values,
	      x11canvas->pix_red_offset, x11canvas->pix_green_offset,
	      x11canvas->pix_blue_offset,
	      blank_pixel, min_sat_pixel, max_sat_pixel,
	      i_min, i_max, iscale_func, iscale_info ) )
	{
	    fprintf (stderr, "Error drawing image into X11Canvas\n");
	    return (FALSE);
	}
    }
    else if (ximage->depth == 8)
    {
	if ( !imw_to8_lossy2 (out_image, TRUE, bytes_per_pixel,
			      ximage->bytes_per_line, x_pixels, y_pixels,
			      slice, (iaddr *) hoffsets, (iaddr *) voffsets,
			      image_width, image_height, type, conv_type,
			      num_pixels, ub_pixels,
			      blank_pixel, min_sat_pixel, max_sat_pixel,
			      i_min, i_max, iscale_func, iscale_info,
			      magnify_func) )
	{
	    fprintf (stderr, "Error drawing image into X11Canvas\n");
	    return (FALSE);
	}
    }
    else
    {
	if ( !pc_draw_lossy (ximage, x_off, y_off, x_pixels, y_pixels,
			     (CONST char *) slice,
			     (iaddr *) hoffsets, (iaddr *) voffsets,
			     image_width, image_height,
			     type, conv_type, num_pixels, pixel_values,
			     blank_pixel, min_sat_pixel, max_sat_pixel,
			     i_min, i_max, iscale_func, iscale_info ) )
	{
	    fprintf (stderr, "Error drawing image into X11Canvas\n");
	    return (FALSE);
	}
    }
    cache->width = cache_width;
    cache->height = cache_height;
    if (cache->pixmap != (Pixmap) NULL)
    {
	xi_put_image (x11canvas->display, cache->pixmap, x11canvas->gc_arr[0],
		      ximage, x_off, y_off, x_off, y_off, x_pixels, y_pixels,
		      cache->shared, TRUE);
    }
    return (TRUE);
}   /*  End Function _kwin_X11_compute_pc_image  */

static flag _kwin_X11_compute_rgb_image (X11Canvas x11canvas,
					 int x_off, int y_off,
					 int x_pixels,int y_pixels,
					 int cache_width, int cache_height,
					 CONST unsigned char *red_slice,
					 CONST unsigned char *green_slice,
					 CONST unsigned char *blue_slice,
					 CONST uaddr *hoffsets,
					 CONST uaddr *voffsets,
					 unsigned int image_width,
					 unsigned int image_height,
					 KPixCanvasImageCache *cache_ptr)
/*  [PURPOSE] This routine will compute a subimage cache for a 2-dimensional
    slice of a Karma array for an X11 canvas. This slice may be tiled. The
    slice is an RGB subimage.
    <x11canvas> The X11 canvas.
    <x_off> The horizontal offset of the subimage, relative to the top-left
    corner of the image cache.
    <y_off> The vertical offset of the subimage, relative to the top-left
    corner of the image cache.
    <x_pixels> The number of horizontal pixels to generate.
    <y_pixels> The number of vertical pixels to generate.
    <cache_width> The width of the image cache.
    <cache_height> The height of the image cache.
    <red_slice> The start of the red slice data.
    <green_slice> The start of the green slice data.
    <blue_slice> The start of the blue slice data.
    [NOTE] The 3 colour components must be of type  K_UBYTE  .
    <hoffsets> The array of horizontal byte offsets for the input image.
    <voffsets> The array of vertical byte offsets for the input image.
    <image_width> The width of the input image (in values).
    <image_height> The height of the input image (in values).
    <cache_ptr> Cache data is written here. The routine will produce cache data
    which will vastly increase the speed of subsequent operations on this data.
    Prior to process exit, a call MUST be made to <<kwin_free_cache_data>>,
    otherwise shared memory segments could remain after the process exits.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KPixCanvasImageCache cache;
    MagnifyFunc magnify_func;
    iaddr pstride;
    uaddr im_red_offset, im_green_offset, im_blue_offset;
    unsigned int count, visual;
    unsigned long red_mask, green_mask, blue_mask;
    unsigned char *out_image;
    XImage *ximage;
    unsigned char pixels[256];
    static char function_name[] = "_kwin_X11_compute_rgb_image";

    VERIFY_CANVAS (x11canvas);
    if ( !size_cache (x11canvas, cache_ptr, cache_width, cache_height) )
    {
	return (FALSE);
    }
    cache = *cache_ptr;
    kwin_get_attributes (x11canvas->pixcanvas,
			 KWIN_ATT_PIX_RED_MASK, &red_mask,
			 KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			 KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    ximage = (cache->ximage ==NULL) ? x11canvas->common_ximage : cache->ximage;
    /*  Now have an XImage big enough  */
    pstride = ximage->bits_per_pixel / 8;  /*  Typically 4  */
    if ( (ximage->depth == 24) && (pstride != 4) )
    {
	fprintf (stderr, "%s: pstride: %ld\n", function_name, pstride);
    }
    out_image = (unsigned char *) ximage->data;
    out_image += ximage->bits_per_pixel / 8 * x_off;
    out_image += ximage->bytes_per_line * y_off;
    /*  Set up pixel value array  */
    for (count = 0; count < 256; ++count) pixels[count] = count;
    if (visual == KWIN_VISUAL_DIRECTCOLOUR)
    {
	/*  A little bit of trickery to get around the fact that X allocates
	    black and white pixels always. Assume the first two pixels have
	    been allocated in this way, so clip the bottom pixel values to
	    what's in the colourmap. The correct way to do this is to grab the
	    pixel values in the Kcolourmap (if the world canvas has one, which
	    may not be the case anyway). Tricky  */
	pixels[0] = 2;
	pixels[1] = 2;
    }
    /*  Note the casts from (uaddr *) to (iaddr *) for the offset
	arrays. This is dodgy, but it should work.  */
    if ( (ximage->depth == 24) &&
	 (x_pixels == image_width) && (y_pixels == image_height) )
    {
	/*  This is what I like to see  */
	kwin_get_attributes (x11canvas->pixcanvas,
			     KWIN_ATT_IM_RED_OFFSET, &im_red_offset,
			     KWIN_ATT_IM_GREEN_OFFSET, &im_green_offset,
			     KWIN_ATT_IM_BLUE_OFFSET, &im_blue_offset,
			     KWIN_ATT_END);
	if ( !imw_to8_o (out_image + im_red_offset, pstride,
			 ximage->bytes_per_line,
			 x_pixels, y_pixels, (CONST char *) red_slice,
			 (iaddr *) hoffsets, (iaddr *) voffsets,
			 K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			 256, pixels, 0, 0, 255, 0.0, 255.0) )
	{
	    fprintf (stderr, "Error drawing red image into X11Canvas\n");
	    return (FALSE);
	}
	if ( !imw_to8_o (out_image + im_green_offset, pstride,
			 ximage->bytes_per_line,
			 x_pixels, y_pixels, (CONST char *) green_slice,
			 (iaddr *) hoffsets, (iaddr *) voffsets,
			 K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			 256, pixels, 0, 0, 255, 0.0, 255.0) )
	{
	    fprintf (stderr, "Error drawing green image into X11Canvas\n");
	    return (FALSE);
	}
	if ( !imw_to8_o (out_image + im_blue_offset, pstride,
			 ximage->bytes_per_line,
			 x_pixels, y_pixels, (CONST char *) blue_slice,
			 (iaddr *) hoffsets, (iaddr *) voffsets,
			 K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			 256, pixels, 0, 0, 255, 0.0, 255.0) )
	{
	    fprintf (stderr, "Error drawing blue image into X11Canvas\n");
	    return (FALSE);
	}
    }
    else if (ximage->depth == 24)
    {
	/*  But we can cope with differing input/output image sizes  */
	kwin_get_attributes (x11canvas->pixcanvas,
			     KWIN_ATT_IM_RED_OFFSET, &im_red_offset,
			     KWIN_ATT_IM_GREEN_OFFSET, &im_green_offset,
			     KWIN_ATT_IM_BLUE_OFFSET, &im_blue_offset,
			     KWIN_ATT_MAGNIFY_FUNC, &magnify_func,
			     KWIN_ATT_END);
	if ( !imw_to8_lossy2 (out_image + im_red_offset, TRUE, pstride,
			      ximage->bytes_per_line,
			      x_pixels, y_pixels, (CONST char *) red_slice,
			      (iaddr *) hoffsets, (iaddr *) voffsets,
			      image_width, image_height,
			      K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			      256, pixels, 0, 0, 255, 0.0, 255.0,
			      ( flag (*) () ) NULL, NULL, magnify_func) )
	{
	    fprintf (stderr, "Error drawing red image into X11Canvas\n");
	    return (FALSE);
	}
	if ( !imw_to8_lossy2 (out_image + im_green_offset, TRUE, pstride,
			      ximage->bytes_per_line,
			      x_pixels, y_pixels, (CONST char *) green_slice,
			      (iaddr *) hoffsets, (iaddr *) voffsets,
			      image_width, image_height,
			      K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			      256, pixels, 0, 0, 255, 0.0, 255.0,
			      ( flag (*) () ) NULL, NULL, magnify_func) )
	{
	    fprintf (stderr, "Error drawing green image into X11Canvas\n");
	    return (FALSE);
	}
	if ( !imw_to8_lossy2 (out_image + im_blue_offset, TRUE, pstride,
			      ximage->bytes_per_line,
			      x_pixels, y_pixels, (CONST char *) blue_slice,
			      (iaddr *) hoffsets, (iaddr *) voffsets,
			      image_width, image_height,
			      K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			      256, pixels, 0, 0, 255, 0.0, 255.0,
			      ( flag (*) () ) NULL, NULL, magnify_func) )
	{
	    fprintf (stderr, "Error drawing blue image into X11Canvas\n");
	    return (FALSE);
	}
    }
    else
    {
	if ( !rgb_draw_lossy (ximage, x_off, y_off, x_pixels, y_pixels,
			      (CONST unsigned char *) red_slice,
			      (CONST unsigned char *) green_slice,
			      (CONST unsigned char *) blue_slice,
			      (iaddr *) hoffsets, (iaddr *) voffsets,
			      image_width, image_height,
			      red_mask, green_mask, blue_mask) )
	{
	    fprintf (stderr, "Error drawing image into X11Canvas\n");
	    return (FALSE);
	}
    }
    cache->width = cache_width;
    cache->height = cache_height;
    if (cache->pixmap != (Pixmap) NULL)
    {
	xi_put_image (x11canvas->display, cache->pixmap, x11canvas->gc_arr[0],
		      ximage, x_off, y_off, x_off, y_off, x_pixels, y_pixels,
		      cache->shared, TRUE);
    }
    return (TRUE);
}   /*  End Function _kwin_X11_compute_rgb_image  */

static flag draw_cached_image (KPixCanvasImageCache cache, flag wait,
			       int parent_x_off, int parent_y_off,
			       int image_width, int image_height,
			       int image_x_off, int image_y_off,
			       int canvas_x_off, int canvas_y_off,
			       int canvas_width, int canvas_height)
/*  [PURPOSE] This routine will draw a previously computed image cache data
    (computed by <<kwin_draw_pc_image>>) onto the canvas on which the original
    image was drawn.
    <cache> The cache data.
    <wait> If TRUE and the cache data is accessable by the graphics system, the
    routine waits for the drawing request to complete.
    <parent_x_off> The horizontal offset, relative to the top-left corner of
    the parent object, where the image will be drawn.
    <parent_y_off> The vertical offset, relative to the top-left corner of the
    parent object, where the image will be drawn.
    <image_width> The width of the image segment to draw. If this is less than
    1 the entire image is drawn.
    <image_height> The height of the image segment to draw. If this is less
    than 1 the entire image is drawn.
    <image_x_off> The horizontal offset, relative to the top-left corner of the
    image.
    <image_y_off> The vertical offset, relative to the top-left corner of the
    image.
    <canvas_x_off> The horizontal offset, relative to the top-left corner of
    the parent object, of the pixel canvas. Used for clipping purposes.
    <canvas_y_off> The vertical offset, relative to the top-left corner of
    the parent object, of the pixel canvas. Used for clipping purposes.
    <canvas_width> The width of the pixel canvas. Used for clipping purposes.
    <canvas_height> The height of the pixel canvas. Used for clipping purposes.
    [RETURNS] TRUE on success if there is valid cache data, else FALSE
    indicating that the image must be recomputed and drawn using
    <<kwin_draw_pc_image>>.
*/
{
    X11Canvas x11canvas;
    static char function_name[] = "_kwin_X11_draw_cached_image";

    if (cache == NULL) return (FALSE);
    if (cache->magic_number != CACHE_DATA_MAGIC_NUMBER)
    {
	fprintf (stderr, "Invalid cache data\n");
	a_prog_bug (function_name);
    }
    x11canvas = cache->x11canvas;
    if ( (image_width < 1) || (image_height < 1) )
    {
	image_width = cache->width;
	image_height = cache->height;
	image_x_off = 0;
	image_y_off = 0;
    }
    /*  Discard if subimage is outside of cache size  */
    if (image_x_off >= cache->width) return (TRUE);
    if (image_y_off >= cache->height) return (TRUE);
    /*  Trim subimage to cache size  */
    if (image_x_off + image_width > cache->width)
	image_width = cache->width - image_x_off;
    if (image_y_off + image_height > cache->height)
	image_height = cache->height - image_y_off;
    /*  Discard if subimage is outside pixel canvas  */
    if (parent_x_off >= canvas_x_off + canvas_width) return (TRUE);
    if (parent_y_off >= canvas_y_off + canvas_height) return (TRUE);
    if (parent_x_off + canvas_width <= canvas_x_off) return (TRUE);
    if (parent_y_off + canvas_height <= canvas_y_off) return (TRUE);
    /*  Trim subimage to pixel canvas size  */
    if (parent_x_off < canvas_x_off)
    {
	parent_x_off = canvas_x_off;
	image_width -= canvas_x_off - parent_x_off;
    }
    if (parent_x_off + image_width > canvas_x_off + canvas_width)
    {
	image_width = canvas_x_off + canvas_width - parent_x_off;
    }
    if (parent_y_off + image_height > canvas_y_off + canvas_height)
    {
	image_height = canvas_y_off + canvas_height - parent_y_off;
    }
    /*  Subimage should now be clipped to be within cache size and pixel canvas
	boundaries  */
    if ( (cache->ximage != NULL) && cache->shared )
    {
	xi_put_image (x11canvas->display, x11canvas->window,
		      x11canvas->gc_arr[0], cache->ximage,
		      image_x_off, image_y_off,
		      parent_x_off, parent_y_off, image_width, image_height,
		      cache->shared, wait);
	return (TRUE);
    }
    if (cache->pixmap != (Pixmap) NULL)
    {
	XCopyArea (x11canvas->display, cache->pixmap, x11canvas->window,
		   x11canvas->gc_arr[0],
		   image_x_off, image_y_off,
		   image_width, image_height, parent_x_off, parent_y_off);
	return (TRUE);
    }
    if (cache->ximage == NULL) return (FALSE);
    /*  Call it again with non-pixmap, non-shared XImage  */
    xi_put_image (x11canvas->display, x11canvas->window, x11canvas->gc_arr[0],
		  cache->ximage,
		  image_x_off, image_y_off,
		  parent_x_off, parent_y_off, image_width, image_height,
		  cache->shared, FALSE);
    return (TRUE);
}   /*  End Function draw_cached_image  */

static void free_cache_data (KPixCanvasImageCache cache)
/*  [PURPOSE] This routine will free some cache data allocated by
    <<kwin_draw_pc_image>>.
    <cache> The cache data.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_kwin_X11_free_cache_data";

    if (cache == NULL) return;
    if (cache->magic_number != CACHE_DATA_MAGIC_NUMBER)
    {
	fprintf (stderr, "Invalid cache data\n");
	a_prog_bug (function_name);
    }
    if (cache->ximage != NULL)
    {
	xi_destroy_image (cache->x11canvas->display, cache->ximage,
			  cache->shared);
    }
    if (cache->pixmap != (Pixmap) NULL)
    {
	XFreePixmap (cache->x11canvas->display, cache->pixmap);
    }
    cache->magic_number = 0;
    m_free (cache);
}   /*  End Function free_cache_data  */

static flag draw_line (X11Canvas x11canvas,
		       double x0, double y0, double x1, double y1,
		       unsigned long pixel_value)
/*  [PURPOSE] This routine will draw a line on an X11 canvas.
    <x11canvas> The X11 canvas.
    <x0> The horizontal offset of the first point.
    <y0> The vertical offset of the first point.
    <x1> The horizontal offset of the second point.
    <y1> The vertical offset of the second point.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    double off = 1e-3;
    static char function_name[] = "_kwin_X11_draw_line";

    VERIFY_CANVAS (x11canvas);
    set_pixel_in_gc (x11canvas, pixel_value);
    XDrawLine ( x11canvas->display, x11canvas->window, x11canvas->gc_arr[0],
		(int) (x0 + off), (int) (y0 + off),
		(int) (x1 + off), (int) (y1 + off) );
    return (TRUE);
}   /*  End Function draw_line  */

static flag draw_arc (X11Canvas x11canvas,
		      double centre_x, double centre_y,
		      double radius_x, double radius_y,
		      int angle1, int angle2, int rotation,
		      unsigned long pixel_value, flag fill)
/*  [PURPOSE] This routine will draw an arc on an X11 canvas.
    <x11canvas> The X11 canvas.
    <centre_x> The horizontal centre of the arc.
    <centre_y> The vertical centre of the arc.
    <radius_x> The horizontal radius of the arc.
    <radius_y> The vertical radius of the arc.
    <angle1> See <<XDrawArc>>.
    <angle2> See <<XDrawArc>>.
    <rotation> The rotation of the arc, in units of 1/64 degree.
    <pixel_value> The pixel value to use.
    <fill> If TRUE, the arc is filled, else only the outside is drawn.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int x, y, width, height;
    static char function_name[] = "_kwin_X11_draw_arc";

    VERIFY_CANVAS (x11canvas);
    FLAG_VERIFY (fill);
    width = 2 * radius_x + 1e-3;
    height = 2 * radius_y + 1e-3;
    if ( (width != height) && (rotation != 0) )
	return kwin_rasterise_draw_arc (x11canvas, draw_points, draw_segments,
					centre_x, centre_y, radius_x, radius_y,
					angle1, angle2, rotation,
					pixel_value, fill);
    set_pixel_in_gc (x11canvas, pixel_value);
    x = centre_x - radius_x + 1e-3;
    y = centre_y - radius_y + 1e-3;
    if (width < 1)
    {
	if (height < 1)
	    XDrawPoint (x11canvas->display, x11canvas->window,
			x11canvas->gc_arr[0], x, y);
	else XDrawLine (x11canvas->display, x11canvas->window,
			x11canvas->gc_arr[0], x, y, x, y + height - 1);
	return (TRUE);
    }
    else if (height < 1)
    {
	XDrawLine (x11canvas->display, x11canvas->window, x11canvas->gc_arr[0],
		   x, y, x + width - 1, y);
	return (TRUE);
    }
    if (fill)
	XFillArc (x11canvas->display, x11canvas->window, x11canvas->gc_arr[0],
		  x, y, width, height, angle1, angle2);
    else
	XDrawArc (x11canvas->display, x11canvas->window, x11canvas->gc_arr[0],
		  x, y, width, height, angle1, angle2);
    return (TRUE);
}   /*  End Function draw_arc  */

static flag draw_polygon (X11Canvas x11canvas, double *x_arr, double *y_arr,
			  unsigned int num_vertices, unsigned long pixel_value,
			  flag convex, flag fill)
/*  [PURPOSE] This routine will draw a polygon onto an X11 canvas.
    <x11canvas> The X11 canvas.
    <x_arr> The array of x co-ordinates of vertices of the polygon.
    <y_arr> The array of y co-ordinates of vertices of the polygon.
    <num_vertices> The number of vertices in the polygon.
    <pixel_value> The pixel value to use.
    <convex> If TRUE, then the points must form a convex polygon.
    <fill> If TRUE, the polygon will be filled.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int coord_count;
    static unsigned int num_points_allocated = 0;
    static XPoint *points = NULL;
    static char function_name[] = "_kwin_X11_draw_polygon";

    VERIFY_CANVAS (x11canvas);
    FLAG_VERIFY (convex);
    FLAG_VERIFY (fill);
    if (!fill)
    {
	fprintf (stderr, "%s: unfilled polygons not supported\n",
		 function_name);
	return (FALSE);
    }
    if (num_vertices > num_points_allocated)
    {
	if (points != NULL)
	{
	    m_free (points);
	    points = NULL;
	}
	num_points_allocated = 0;
	if ( ( points = m_alloc (sizeof *points * num_vertices) ) == NULL )
	{
	    m_error_notify (function_name, "array of XPoint structures");
	    return (FALSE);
	}
	num_points_allocated = num_vertices;
    }
    set_pixel_in_gc (x11canvas, pixel_value);
    /*  Convert world co-ordinates to pixel co-ordinates  */
    for (coord_count = 0; coord_count < num_vertices; ++coord_count)
    {
	points[coord_count].x = x_arr[coord_count] + 1e-3;
	points[coord_count].y = y_arr[coord_count] + 1e-3;
    }
    XFillPolygon (x11canvas->display, x11canvas->window, x11canvas->gc_arr[0],
		  points, (int) num_vertices, convex ? Convex : Complex,
		  CoordModeOrigin);
    return (TRUE);
}   /*  End Function draw_polygon  */

static flag draw_string (X11Canvas x11canvas, double x, double y,
			 CONST char *string, unsigned long pixel_value,
			 flag clear_under)
/*  [PURPOSE] This routine will draw a NULL terminated string onto an X11
    canvas, using the default font for the canvas.
    <x11canvas> The X11 canvas.
    <x> The horizontal offset of the string origin.
    <y> The vertical offset of the string origin.
    <string> The string.
    <pixel_value> The pixel value to use.
    <clear_under> If TRUE, then the routine will draw both the foreground and
    background of the characters.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int length;
    static char function_name[] = "_kwin_X11_draw_string";

    VERIFY_CANVAS (x11canvas);
    FLAG_VERIFY (clear_under);
    set_pixel_in_gc (x11canvas, pixel_value);
    length = strlen (string);
    if (clear_under)
    {
	XDrawImageString (x11canvas->display, x11canvas->window,
			  x11canvas->gc_arr[0],
			  (int) (x + 1e-3), (int) (y + 1e-3), string, length);
    }
    else
    {
	XDrawString (x11canvas->display, x11canvas->window,
		     x11canvas->gc_arr[0],
		     (int) (x + 1e-3), (int) (y + 1e-3), string, length);
    }
    return (TRUE);
}   /*  End Function draw_string  */

static flag draw_rectangle (X11Canvas x11canvas,
			    double x, double y, double width, double height,
			    unsigned long pixel_value, flag fill)
/*  [PURPOSE] This routine will draw a single rectangle onto an X11 canvas.
    <x11canvas> The X11 canvas.
    <x> The horizontal offset of the rectangle.
    <y> The vertical offset of the rectangle.
    <width> The width of the rectangle. The point <<x + width>> is a vertex.
    <height> The height of the rectangle. The point <<y + height>> is a vertex.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_kwin_X11_draw_rectangle";

    VERIFY_CANVAS (x11canvas);
    FLAG_VERIFY (fill);
    set_pixel_in_gc (x11canvas, pixel_value);
    if (fill)
    {
	XFillRectangle ( x11canvas->display, x11canvas->window,
			 x11canvas->gc_arr[0],
			 (int) (x + 1e-3), (int) (y + 1e-3),
			 (unsigned int) (width + 1.001),
			 (unsigned int) (height + 1.001) );
    }
    else
    {
	XDrawRectangle ( x11canvas->display, x11canvas->window,
			 x11canvas->gc_arr[0],
			 (int) (x + 1e-3), (int) (y + 1e-3),
			 (unsigned int) (width + 1e-3),
			 (unsigned int) (height + 1e-3) );
    }
    return (TRUE);
}   /*  End Function draw_rectangle  */

static flag draw_lines (X11Canvas x11canvas,
			CONST double *x_arr, CONST double *y_arr,
			unsigned int num_points, unsigned long pixel_value)
/*  [PURPOSE] This routine will draw multiple connected lines onto an X11
    canvas.
    <x11canvas> The X11 canvas.
    <x_array> The horizontal co-ordinates of the points.
    <y_array> The vetical co-ordinates of the points.
    <num_points> The number of points. The number of lines drawn is 1 less than
    this value.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int ipoint_count, xpoint_count;
    static char function_name[] = "_kwin_X11_draw_lines";

    VERIFY_CANVAS (x11canvas);
    set_pixel_in_gc (x11canvas, pixel_value);
    alloc_points (x11canvas, num_points);
    for (ipoint_count = 0, xpoint_count = 0; ipoint_count < num_points;
	 ++ipoint_count, ++xpoint_count)
    {
	if (xpoint_count >= num_xpoints_allocated)
	{
	    /*  Send some points now  */
	    XDrawLines (x11canvas->display, x11canvas->window,
			x11canvas->gc_arr[0], xpoints, xpoint_count,
			CoordModeOrigin);
	    xpoints[0].x = x_arr[ipoint_count - 1] + 1e-3;
	    xpoints[0].y = y_arr[ipoint_count - 1] + 1e-3;
	    xpoint_count = 1;
	}
	xpoints[xpoint_count].x = x_arr[ipoint_count] + 1e-3;
	xpoints[xpoint_count].y = y_arr[ipoint_count] + 1e-3;
    }
    /*  Draw remaining points  */
    if (xpoint_count > 0)
    {
	XDrawLines (x11canvas->display, x11canvas->window,
		    x11canvas->gc_arr[0], xpoints, xpoint_count,
		    CoordModeOrigin);
    }
    return (TRUE);
}   /*  End Function draw_lines  */

static flag draw_arcs (X11Canvas x11canvas,
		       double *x, double *y, double *width, double *height,
		       int *angle1, int *angle2, unsigned int num_ellipses,
		       unsigned long pixel_value, flag fill)
/*  [PURPOSE] This routine will draw multiple arcs onto an X11 canvas.
    <x11canvas> The X11 canvas.
    <x> The array of horizontal co-orinates of the arcs.
    <y> The array of vertical co-ordinates of the arcs.
    <width> The array of widths of the arcs.
    <height> The array of heights of the arcs.
    <angle1> Another array. See <<XDrawArc>>.
    <angle2> Another array. See <<XDrawArc>>.
    <num_ellipses> The number of ellipses.
    <pixel_value> The pixel value to use.
    <fill> If TRUE, the arcs are filled, else only the outside is drawn.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int iarc_count, xarc_count;
    static char function_name[] = "_kwin_X11_draw_arcs";

    VERIFY_CANVAS (x11canvas);
    FLAG_VERIFY (fill);
    set_pixel_in_gc (x11canvas, pixel_value);
    alloc_arcs (x11canvas, num_ellipses);
    for (iarc_count = 0, xarc_count = 0; iarc_count < num_ellipses;
	 ++iarc_count, ++xarc_count)
    {
	if (xarc_count >= num_xarcs_allocated)
	{
	    /*  Send some arcs now  */
	    if (fill) XFillArcs (x11canvas->display, x11canvas->window,
				 x11canvas->gc_arr[0], xarcs, xarc_count);
	    else XDrawArcs (x11canvas->display, x11canvas->window,
			    x11canvas->gc_arr[0], xarcs, xarc_count);
	    xarc_count = 0;
	}
	xarcs[xarc_count].x = x[iarc_count] + 1e-3;
	xarcs[xarc_count].y = y[iarc_count] + 1e-3;
	xarcs[xarc_count].width = width[iarc_count] + 1e-3;
	xarcs[xarc_count].height = height[iarc_count] + 1e-3;
	xarcs[xarc_count].angle1 = angle1[iarc_count];
	xarcs[xarc_count].angle2 = angle2[iarc_count];
    }
    /*  Draw remaining arcs  */
    if (xarc_count > 0)
    {
	if (fill) XFillArcs (x11canvas->display, x11canvas->window,
			     x11canvas->gc_arr[0], xarcs, xarc_count);
	else XDrawArcs (x11canvas->display, x11canvas->window,
			x11canvas->gc_arr[0], xarcs, xarc_count);
    }
    return (TRUE);
}   /*  End Function draw_arcs  */

static flag draw_segments (X11Canvas x11canvas,
			   double *x0, double *y0, double *x1, double *y1,
			   unsigned int num_segments,
			   unsigned long pixel_value)
/*  [PURPOSE] This routine will draw multiple disjoint lines onto an X11
    canvas.
    <x11canvas> The X11 canvas.
    <x0> The horizontal start co-ordinates of the segments.
    <y0> The vetical start co-ordinates of the segments.
    <x1> The horizontal end co-ordinates of the segments.
    <y1> The vetical end co-ordinates of the segments.
    <num_segments> The number of segments.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int isegment_count, xsegment_count;
    static char function_name[] = "_kwin_X11_draw_segments";

    VERIFY_CANVAS (x11canvas);
    set_pixel_in_gc (x11canvas, pixel_value);
    alloc_segments (x11canvas, num_segments);
    for (isegment_count = 0, xsegment_count = 0;
	 isegment_count < num_segments;
	 ++isegment_count, ++xsegment_count)
    {
	if (xsegment_count >= num_xsegments_allocated)
	{
	    /*  Send some segments now  */
	    XDrawSegments (x11canvas->display, x11canvas->window,
			   x11canvas->gc_arr[0], xsegments, xsegment_count);
	    xsegment_count = 0;
	}
	xsegments[xsegment_count].x1 = x0[isegment_count] + 1e-3;
	xsegments[xsegment_count].y1 = y0[isegment_count] + 1e-3;
	xsegments[xsegment_count].x2 = x1[isegment_count] + 1e-3;
	xsegments[xsegment_count].y2 = y1[isegment_count] + 1e-3;
    }
    /*  Draw remaining segments  */
    if (xsegment_count > 0)
    {
	XDrawSegments (x11canvas->display, x11canvas->window,
		       x11canvas->gc_arr[0], xsegments, xsegment_count);
    }
    return (TRUE);
}   /*  End Function draw_segments  */

static flag set_linewidth (X11Canvas x11canvas, double linewidth)
/*  [SUMMARY] Set the linewidth for a canvas.
    <x11canvas> The X11 canvas.
    <linewidth> The linewidth, in pixels.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int lw = linewidth;
    unsigned int count;
    static char function_name[] = "_kwin_X11_set_linewidth";

    VERIFY_CANVAS (x11canvas);
    if (lw == x11canvas->gcvalues.line_width) return (TRUE);
    x11canvas->gcvalues.line_width = lw;
    for (count = 0; count < x11canvas->gc_cache_size; ++count)
    {
	XChangeGC (x11canvas->display, x11canvas->gc_arr[count], GCLineWidth,
		   &x11canvas->gcvalues);
    }
    return (TRUE);
}   /*  End Function set_linewidth  */

static flag set_background_pixel (X11Canvas x11canvas,
				  unsigned long pixel_value)
/*  [SUMMARY] Set the background pixel for a canvas.
    <x11canvas> The X11 canvas.
    <pixel_value> The pixel value.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    static char function_name[] = "_kwin_X11_set_background_pixel";

    VERIFY_CANVAS (x11canvas);
    x11canvas->gcvalues.background = pixel_value;
    for (count = 0; count < x11canvas->gc_cache_size; ++count)
	XChangeGC (x11canvas->display, x11canvas->gc_arr[count], GCBackground,
		   &x11canvas->gcvalues);
    XSetWindowBackground (x11canvas->display, x11canvas->window, pixel_value);
    return (TRUE);
}   /*  End Function set_background_pixel  */

static flag set_line_dash (X11Canvas x11canvas, flag dash)
/*  [SUMMARY] Set the line dash for a canvas.
    <x11canvas> The X11 canvas.
    <dash> If TRUE, the line will be dashed.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int style = dash ? LineOnOffDash : LineSolid;
    unsigned int count;
    static char function_name[] = "_kwin_X11_set_line_dash";

    VERIFY_CANVAS (x11canvas);
    if (style == x11canvas->gcvalues.line_style) return (TRUE);
    x11canvas->gcvalues.line_style = style;
    for (count = 0; count < x11canvas->gc_cache_size; ++count)
    {
	XChangeGC (x11canvas->display, x11canvas->gc_arr[count], GCLineStyle,
		   &x11canvas->gcvalues);
    }
    return (TRUE);
}   /*  End Function set_line_dash  */


/*  Private functions follow  */

static flag get_colour (X11Canvas x11canvas, CONST char *colourname,
			unsigned long *pixel_value, unsigned short *red,
			unsigned short *green, unsigned short *blue)
/*  [PURPOSE] This routine will allocate a colour for an X11 canvas.
    <x11canvas> The X11 canvas.
    <colourname> The name of the colour.
    <pixel_value> The pixel value will be written here.
    [NOTE] The pixel value is valid ONLY for this canvas. ALSO: the pixel value
    becomes invalid EVERY time the canvas is refreshed/ resized. If the canvas
    is refreshed/ resized, this routine MUST be called again.
    <red> The red intensity in the hardware colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    <green> The green intensity in the hardware colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    <blue> The blue intensity in the hardware colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    [RETURNS] TRUE if the colourcell was allocated, else FALSE.
*/
{
    XColor colourcell_def;
    XColor rgb_db_def;
    struct colourcell_type *colour;
    static char function_name[] = "_kwin_X11_get_colour";

    VERIFY_CANVAS (x11canvas);
    if (colourname == NULL)
    {
	fprintf (stderr, "NULL colourname pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Search canvas for existing allocation  */
    for (colour = x11canvas->colours; colour != NULL; colour = colour->next)
    {
	if (strcmp (colourname, colour->name) == 0)
	{
	    *pixel_value = colour->pixel_value;
	    if (red != NULL) *red = colour->red;
	    if (green != NULL) *green = colour->green;
	    if (blue != NULL) *blue = colour->blue;
	    return (TRUE);
	}
    }
    /*  Need to allocate a new colour  */
    if ( ( colour = m_alloc (sizeof *colour) ) == NULL )
    {
	m_abort (function_name, "colourcell structure");
    }
    if ( ( colour->name = st_dup (colourname) ) == NULL )
    {
	m_abort (function_name, "colour name");
    }
    if (XAllocNamedColor (x11canvas->display, x11canvas->cmap, colourname,
			  &colourcell_def, &rgb_db_def) == 0)
    {
	fprintf (stderr,
		 "%s: error allocating colour: \"%s\" in visualID: 0x%lx\n",
		 function_name, colourname,
		 (unsigned long) x11canvas->vinfo.visualid);
	m_free (colour->name);
	m_free (colour);
	return (FALSE);
    }
    colour->pixel_value = colourcell_def.pixel;
    *pixel_value = colour->pixel_value;
    colour->red = colourcell_def.red;
    colour->green = colourcell_def.green;
    colour->blue = colourcell_def.blue;
    if (red != NULL) *red = colourcell_def.red;
    if (green != NULL) *green = colourcell_def.green;
    if (blue != NULL) *blue = colourcell_def.blue;
    colour->next = x11canvas->colours;
    x11canvas->colours = colour;
    return (TRUE);
}   /*  End Function get_colour  */

static XFontStruct *load_font (X11Canvas x11canvas, CONST char *fontname)
/*  [PURPOSE] This routine will load an X11 font.
    <x11canvas> The X11 canvas for which the font is valid.
    <fontname> The fontname.
    [RETURNS] A pointer to the X11 font structure.
*/
{
    XFontStruct *x11font;
    static char function_name[] = "_kwin_X11_load_font";

    VERIFY_CANVAS (x11canvas);
    if ( ( x11font = XLoadQueryFont (x11canvas->display, fontname) ) == NULL )
    {
	fprintf (stderr, "Error loading font: \"%s\"\n", fontname);
	return (NULL);
    }
    return (x11font);
}   /*  End Function load_font  */

static void unload_font (X11Canvas x11canvas, XFontStruct *x11font)
/*  [PURPOSE] This routine will unload an X11 font.
    <x11canvas> The X11 canvas for which the font is valid.
    <font> The font.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_kwin_X11_unload_font";

    VERIFY_CANVAS (x11canvas);
    XFreeFont (x11canvas->display, x11font);
}   /*  End Function unload_font  */

static flag get_string_size (XFontStruct *x11font, CONST char *string,
			     va_list argp)
/*  [PURPOSE] This routine will get the size of a string.
    <x11font> The pointer to the X11 font structure.
    <string> The string.
    [VARARGS] The optional list of parameter attribute-key attribute-value
    pairs must follow. See the header file for details on defined attributes.
    This list must be terminated with the value  KWIN_STRING_END  .
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int length;
    unsigned int att_key;
    static char function_name[] = "_kwin_X11_get_string_size";

    length = strlen (string);
    while ( ( att_key = va_arg (argp, unsigned int) ) != KWIN_STRING_END )
    {
	switch (att_key)
	{
	  case KWIN_STRING_WIDTH:
	    *( va_arg (argp, int *) ) = XTextWidth (x11font, string, length);
	    break;
	  case KWIN_STRING_HEIGHT:
	    *( va_arg (argp, int *) ) = x11font->ascent + x11font->descent;
	    break;
	  case KWIN_STRING_ASCENT:
	    *( va_arg (argp, int *) ) = x11font->ascent;
	    break;
	  case KWIN_STRING_DESCENT:
	    *( va_arg (argp, int *) ) = x11font->descent;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    return (TRUE);
}   /*  End Function get_string_size  */

static void set_font (X11Canvas x11canvas, XFontStruct *x11font)
/*  [PURPOSE] This routine will set the font for an X11 canvas.
    <x11canvas> The X11 canvas.
    <x11font> The X11 font.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    static char function_name[] = "_kwin_X11_set_font";

    VERIFY_CANVAS (x11canvas);
    if (x11canvas->gcvalues.font != x11font->fid)
    {
	x11canvas->gcvalues.font = x11font->fid;
	for (count = 0; count < x11canvas->gc_cache_size; ++count)
	{
	    XChangeGC (x11canvas->display, x11canvas->gc_arr[count], GCFont,
		       &x11canvas->gcvalues);
	}
    }
}   /*  End Function set_x11font  */

static flag query_colourmap (X11Canvas x11canvas, unsigned long *pixels,
			     unsigned short *reds, unsigned short *greens,
			     unsigned short *blues, unsigned int num_colours)
/*  [PURPOSE] This routine will determine the RGB components of an array of
    colourmap entries.
    <canvas> The X11  canvas.
    <pixels> The array of colourmap indices.
    <reds> The red components will be written to this array.
    <greens> The green components will be written to this array.
    <blues> The blue components will be written to this array.
    <num_colours> The number of colours in the arrays.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count, query_size;
    XColor *xcolours;
    static char function_name[] = "_kwin_X11_query_colourmap";

    VERIFY_CANVAS (x11canvas);
    if ( ( xcolours = m_alloc (num_colours * sizeof *xcolours) ) == NULL )
    {
	m_error_notify (function_name, "XColor array");
	return (FALSE);
    }
    for (count = 0; count < num_colours; ++count)
    {
	xcolours[count].pixel = pixels[count];
	xcolours[count].flags = DoRed | DoGreen | DoBlue;  /*  Paranoia  */
    }
    /*  Break up request into 256-size chunks, since some servers seem to have
	a problem with large queries (such as 65536:-)  */
    for (count = 0; count < num_colours; count += query_size)
    {
	query_size = num_colours - count;
	if (query_size > 256) query_size = 256;
	XQueryColors (x11canvas->display, x11canvas->cmap, xcolours + count,
		      (int) query_size);
    }
    for (count = 0; count < num_colours; ++count)
    {
	reds[count] = xcolours[count].red;
	greens[count] = xcolours[count].green;
	blues[count] = xcolours[count].blue;
    }
    m_free (xcolours);
    return (TRUE);
}   /*  End Function query_colourmap  */

static void warp_pointer (X11Canvas x11canvas, int x, int y, flag relative,
			  flag discard)
/*  [SUMMARY] Warp the pointer to a new location.
    <x11canvas> The X11 canvas.
    <x> The horizontal position to warp to.
    <y> The vertical position to warp to.
    <relative> If TRUE, the pointer is warped relative to it's current
    position.
    <discard> If TRUE, any pending events are discarded.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_kwin_X11_warp_pointer";

    VERIFY_CANVAS (x11canvas);
    XWarpPointer (x11canvas->display,
		  None, relative ? None : x11canvas->window,
		  0, 0, 0, 0, x, y);
    XSync (x11canvas->display, discard ? True : False);
}   /*  End Function warp_pointer  */

static void set_pixel_in_gc (X11Canvas x11canvas, unsigned long pixel_value)
/*  [PURPOSE] Set the foreground pixel value in the Graphics Context.
    <x11canvas> The X11 canvas.
    <pixel_value> The pixel value.
    [RETURNS] Nothing.
*/
{
    unsigned int count, new_size;
    GC tmp_gc;
    unsigned long *pixels;
    GC *GCs;
    static char function_name[] = "_kwin_X11_set_pixel_in_gc";

    VERIFY_CANVAS (x11canvas);
    ++x11canvas->primitive_count;
    if (pixel_value == x11canvas->pixel_values[0]) return;
    for (count = 1; count < x11canvas->gc_cache_size; ++count)
    {
	if (pixel_value == x11canvas->pixel_values[count]) break;
    }
    if (count < x11canvas->gc_cache_size)
    {
	tmp_gc = x11canvas->gc_arr[count];
	for (; count > 0; --count)
	{
	    x11canvas->gc_arr[count] = x11canvas->gc_arr[count - 1];
	    x11canvas->pixel_values[count] = x11canvas->pixel_values[count -1];
	}
	x11canvas->gc_arr[0] = tmp_gc;
	x11canvas->pixel_values[0] = pixel_value;
	return;
    }
    if ( (x11canvas->primitive_count > PRIMITIVES_PER_GC_CHANGE_THRESHOLD) ||
	 (x11canvas->gc_cache_size >= GC_CACHE_MAX_SIZE) )
    {
	tmp_gc = x11canvas->gc_arr[x11canvas->gc_cache_size - 1];
	for (count = x11canvas->gc_cache_size - 1; count > 0; --count)
	{
	    x11canvas->gc_arr[count] = x11canvas->gc_arr[count - 1];
	    x11canvas->pixel_values[count] = x11canvas->pixel_values[count -1];
	}
	x11canvas->gc_arr[0] = tmp_gc;
	x11canvas->pixel_values[0] = pixel_value;
	x11canvas->gcvalues.foreground = pixel_value;
	XChangeGC (x11canvas->display, tmp_gc, GCForeground,
		   &x11canvas->gcvalues);
	x11canvas->primitive_count = 0;
	return;
    }
    new_size = x11canvas->gc_cache_size + 1;
    if ( ( GCs = m_alloc (new_size * sizeof *GCs) ) == NULL )
    {
	m_abort (function_name, "GC array");
    }
    if ( ( pixels = m_alloc (new_size * sizeof *pixels) ) == NULL )
    {
	m_abort (function_name, "pixel array");
    }
    for (count = 0; count < x11canvas->gc_cache_size; ++count)
    {
	GCs[count + 1] = x11canvas->gc_arr[count];
	pixels[count + 1] = x11canvas->pixel_values[count];
    }
    m_free (x11canvas->gc_arr);
    x11canvas->gc_arr = GCs;
    m_free (x11canvas->pixel_values);
    x11canvas->pixel_values = pixels;
    x11canvas->gcvalues.foreground = pixel_value;
    x11canvas->gc_arr[0] =
	XCreateGC (x11canvas->display, x11canvas->window,
		   GCFunction | GCPlaneMask |
		   GCForeground | GCBackground | GCLineWidth | GCLineStyle |
		   GCCapStyle | GCJoinStyle | GCFillStyle | GCFillRule |
		   GCTileStipXOrigin |
		   GCTileStipYOrigin | GCFont | GCSubwindowMode |
		   GCGraphicsExposures | GCClipXOrigin | GCClipYOrigin |
		   GCDashOffset | GCArcMode,
		   &x11canvas->gcvalues);
    x11canvas->pixel_values[0] = pixel_value;
    x11canvas->gc_cache_size = new_size;
}   /*  End Function set_pixel_in_gc  */

static flag size_cache (X11Canvas x11canvas, KPixCanvasImageCache *cache_ptr,
			unsigned int width, unsigned int height)
/*  [PURPOSE] This routine will allocate image cache data.
    <x11canvas> The X11 canvas.
    <cache_ptr> Cache data is written here. The routine may produce cache data
    which will vastly increase the speed of subsequent operations on this data.
    Prior to process exit, a call MUST be made to <<kwin_free_cache_data>>,
    otherwise shared memory segments could remain after the process exits.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KPixCanvasImageCache cache;
    int (*old_handler) ();
    static flag first_time = TRUE;
    static flag enable_pixmaps = TRUE;
    static char function_name[] = "_kwin_X11_size_cache";

    VERIFY_CANVAS (x11canvas);
    if (first_time)
    {
	first_time = FALSE;
	if (r_getenv ("KWIN_DISABLE_PIXMAPS") != NULL)
	{
	    enable_pixmaps = FALSE;
	    fprintf (stderr,
		     "KWIN_DISABLE_PIXMAPS environment variable found\n");
	    fprintf (stderr, "No pixmap image caching used.\n");
	}
    }
    cache = *cache_ptr;
    if (cache == NULL)
    {
	if ( ( cache = m_alloc (sizeof *cache) ) == NULL )
	{
	    m_error_notify (function_name, "cache data structure");
	    return (FALSE);
	}
	*cache_ptr = cache;
	cache->pixcanvas = x11canvas->pixcanvas;
	cache->x11canvas = x11canvas;
	cache->magic_number = CACHE_DATA_MAGIC_NUMBER;
	cache->ximage = NULL;
	cache->pixmap = (Pixmap) NULL;
	cache->shared = FALSE;
	cache->width = 0;
	cache->height = 0;
	cache->p_width = 0;
	cache->p_height = 0;
    }
    if (cache->magic_number != CACHE_DATA_MAGIC_NUMBER)
    {
	fprintf (stderr, "Invalid cache data\n");
	a_prog_bug (function_name);
    }
    /*  Check if existing XImage is big enough  */
    /*  Here we assume that if an XImage is big enough, everything is fine.
	This is of course always true for SHM XImages. For ordinary XImages,
	this is true if we make the assumption that there is no Pixmap.  */
    if (cache->ximage != NULL)
    {
	if ( (width <= cache->ximage->width) &&
	     (height <= cache->ximage->height) )
	{
	    /*  XImage is big enough: great!  */
	    if (cache->pixmap != (Pixmap) NULL)
	    {
		fprintf (stderr, "Cache has XImage and Pixmap!\n");
		a_prog_bug (function_name);
	    }
	    return (TRUE);
	}
	/*  XImage is too small  */
	xi_destroy_image (x11canvas->display, cache->ximage, cache->shared);
	cache->ximage = NULL;
	cache->shared = FALSE;
    }
    /*  No appropriate XImage  */
    if (x11canvas->shm_available)
    {
	/*  No XImage but SHM XImages supported: give it a go  */
	if ( ( cache->ximage = xi_create_shm_image (x11canvas->display,
						    x11canvas->vinfo.visual,
						    x11canvas->vinfo.depth,
						    width, height, FALSE) )
	     != NULL )
	{
	    /*  Created a SHM XImage  */
	    cache->shared = TRUE;
	    if (cache->pixmap != (Pixmap) NULL)
	    {
		/*  Don't need the pixmap anymore  */
		XFreePixmap (x11canvas->display, cache->pixmap);
		cache->pixmap = (Pixmap) NULL;
	    }
	    return (TRUE);
	}
    }
    /*  Have to rely on Pixmaps or ordinary XImages  */
    if (cache->pixmap != (Pixmap) NULL)
    {
	/*  We have an existing Pixmap  */
	if ( (width > cache->p_width) || (height > cache->p_height) )
	{
	    /*  Pixmap is too small  */
	    XFreePixmap (x11canvas->display, cache->pixmap);
	    cache->pixmap = (Pixmap) NULL;
	}
    }
    /*  No appropriate Pixmap already: can we make one?  */
    if ( (cache->pixmap == (Pixmap) NULL) && enable_pixmaps )
    {
	cache->p_width = width;
	cache->p_height = height;
	old_handler = XSetErrorHandler (x11_err_handler);
	caught_error = FALSE;
	cache->pixmap = XCreatePixmap (x11canvas->display, x11canvas->window,
				       width, height, x11canvas->vinfo.depth);
	XSync (x11canvas->display, False);
	XSetErrorHandler (old_handler);
	if (caught_error) cache->pixmap = (Pixmap) NULL;
    }
    if (cache->pixmap != (Pixmap) NULL)
    {
	/*  Have a good Pixmap: is common XImage big enough?  */
	if ( (x11canvas->common_ximage != NULL) &&
	     ( (x11canvas->common_ximage->width < width) ||
	       (x11canvas->common_ximage->height < height) ) )
	{
	    xi_destroy_image (x11canvas->display, x11canvas->common_ximage,
			      FALSE);
	    x11canvas->common_ximage = NULL;
	}
	if (x11canvas->common_ximage == NULL)
	{
	    cache->shared = FALSE;
	    if ( ( x11canvas->common_ximage =
		   xi_create_image (x11canvas->display, x11canvas->window,
				    width, height, &cache->shared) )
		 == NULL )
	    {
		kwin_free_cache_data (cache);
		*cache_ptr = NULL;
		return (FALSE);
	    }
	}
	/*  Have Pixmap and common XImage big enough  */
	return (TRUE);
    }
    /*  No Pixmap and no XImage  */
    cache->shared = FALSE;
    if ( ( cache->ximage = xi_create_image (x11canvas->display,
					    x11canvas->window, width, height,
					    &cache->shared) ) == NULL )
    {
	kwin_free_cache_data (cache);
	*cache_ptr = NULL;
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function size_cache  */

static void alloc_points (X11Canvas x11canvas, int num_points)
/*  This routine will (re)allocate the array of XPoint structures.
    The canvas must be given by  canvas  .The  max_request_size  field is used.
    The number of points requested must be given by  num_points  .There is no
    guarantee that the number of points requested will be allocated.
    The routine returns nothing.
*/
{
    int max_points;
    static char function_name[] = "alloc_points";

    max_points = (x11canvas->max_request_size - 3) / 2 + 1;
    if (max_points > MAX_POINTS) max_points = MAX_POINTS;
    if (num_points > max_points) num_points = max_points;
    /*  Determine if XPoint array is big enough  */
    if (num_points > num_xpoints_allocated)
    {
	if (xpoints != NULL)
	{
	    m_free (xpoints);
	    xpoints = NULL;
	}
	num_xpoints_allocated = 0;
	if ( ( xpoints = m_alloc (sizeof *xpoints * num_points) ) == NULL )
	{
	    m_abort (function_name, "array of XPoint structures");
	}
	num_xpoints_allocated = num_points;
    }
}   /*  End Function alloc_points  */

static void alloc_arcs (X11Canvas x11canvas, int num_arcs)
/*  This routine will (re)allocate the array of XArc structures.
    The canvas must be given by  canvas  .The  max_request_size  field is used.
    The number of arcs requested must be given by  num_arcs  .There is no
    guarantee that the number of arcs requested will be allocated.
    The routine returns nothing.
*/
{
    int max_arcs;
    static char function_name[] = "alloc_arcs";

    max_arcs = (x11canvas->max_request_size - 3) / 3;
    if (max_arcs > MAX_ARCS) max_arcs = MAX_ARCS;
    if (num_arcs > max_arcs) num_arcs = max_arcs;
    /*  Determine if XArc array is big enough  */
    if (num_arcs > num_xarcs_allocated)
    {
	if (xarcs != NULL)
	{
	    m_free (xarcs);
	    xarcs = NULL;
	}
	num_xarcs_allocated = 0;
	if ( ( xarcs = m_alloc (sizeof *xarcs * num_arcs) ) == NULL )
	{
	    m_abort (function_name, "array of XArc structures");
	}
	num_xarcs_allocated = num_arcs;
    }
}   /*  End Function alloc_arcs  */

static void alloc_segments (X11Canvas x11canvas, int num_segments)
/*  This routine will (re)allocate the array of XSegment structures.
    The canvas must be given by  canvas  .The  max_request_size  field is used.
    The number of segments requested must be given by  num_segments  .There is
    no guarantee that the number of segments requested will be allocated.
    The routine returns nothing.
*/
{
    int max_segments;
    static char function_name[] = "alloc_segments";

    max_segments = (x11canvas->max_request_size - 3) / 2;
    if (max_segments > MAX_SEGMENTS) max_segments = MAX_SEGMENTS;
    if (num_segments > max_segments) num_segments = max_segments;
    /*  Determine if XSegment array is big enough  */
    if (num_segments > num_xsegments_allocated)
    {
	if (xsegments != NULL)
	{
	    m_free (xsegments);
	    xsegments = NULL;
	}
	num_xsegments_allocated = 0;
	if ( ( xsegments = m_alloc (sizeof *xsegments * num_segments) )
	     == NULL )
	{
	    m_abort (function_name, "array of XSegment structures");
	}
	num_xsegments_allocated = num_segments;
    }
}   /*  End Function alloc_segments  */

STATIC_FUNCTION (double *alloc_inp_values_buffer, (unsigned int num_values) );
STATIC_FUNCTION (unsigned long *alloc_out_pixels_buffer,
		 (unsigned int num_pixels) );

static flag pc_draw_lossy (XImage *ximage, int x_off, int y_off,
			   int out_width, int out_height,
			   CONST char *inp_image,
			   CONST iaddr *inp_hoffsets,CONST iaddr *inp_voffsets,
			   int inp_width, int inp_height,
			   unsigned int inp_type, unsigned int conv_type,
			   unsigned int cmap_size, CONST unsigned long *pixels,
			   unsigned int blank_pixel,
			   unsigned int min_sat_pixel,
			   unsigned int max_sat_pixel,
			   double i_min, double i_max,
			   flag (*iscale_func) (), void *iscale_info)
/*  [SUMMARY] Write generic PseudoColour subimage.
    [PURPOSE] This routine will write a generic PseudoColour image. The input
    and output sizes may be different. The output image is flipped vertically
    relative to the input image.
    <ximage> The XImage to write to.
    <x_off> The horizontal offset of the subimage, relative to the top-left
    corner of the image cache.
    <y_off> The vertical offset of the subimage, relative to the top-left
    corner of the image cache.
    <out_width> The width of the output subimage.
    <out_height> The height of the output subimage.
    <inp_image> The input image data.
    <inp_hoffsets> The array of horizontal byte offsets.
    <inp_voffsets> The array of vertical byte offsets.
    <inp_width> The width of the input image.
    <inp_height> The height of the input image.
    <inp_type> The type of the input data.
    <conv_type> The input conversion type (when the input is complex).
    <cmap_size> The number of colourmap entries.
    <pixels> The array of pixel values.
    <blank_pixel> The pixel value to be used when the intensity value is an
    undefined value.
    <min_sat_pixel> The pixel value to be used when the intensity value is
    below the minimum value.
    <max_sat_pixel> The pixel value to be used when the intensity value is
    above the maximum value.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <iscale_func> The function to be called when non-linear intensity scaling
    is required. If NULL, linear intensity scaling is used. The prototype
    function is [<imw_PROTO_iscale_func>].
    <iscale_info> A pointer to arbitrary information for <<iscale_func>>.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag complex;
    int out_x, out_y, inp_x, inp_y;
    float h_factor, v_factor;
    float offset = 0.5;
    float tiny_offset = 1e-6;
    double d_mul, d_data;
    double d_toobig = TOOBIG;
    double scaled_min, scaled_max, tmp;
    CONST char *inp_line;
    CONST char *last_inp_line = NULL;
    unsigned long *out_pixels;
    double *inp_values;
    static char function_name[] = "_kwin_X11_pc_draw_lossy";

    if ( (inp_hoffsets == NULL) || (inp_voffsets == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if (i_min >= i_max)
    {
	fprintf (stderr, "i_max: %e  is not greater than i_min: %e\n",
		 i_max, i_min);
	a_prog_bug (function_name);
    }
    if ( !ds_element_is_atomic (inp_type) )
    {
	fprintf (stderr, "Input image must be atomic\n");
	a_prog_bug (function_name);
    }
    if ( imw_test_verbose () )
    {
	 fprintf (stderr, "%s started...\n", function_name);
	 fprintf (stderr, "out_size: %d %d  inp_size: %d %d\n",
		  out_width, out_height, inp_width, inp_height);
    }
    /*  Allocate value buffers  */
    if ( ( inp_values = alloc_inp_values_buffer (inp_width) ) == NULL )
    {
	return (FALSE);
    }
    if ( ( out_pixels = alloc_out_pixels_buffer (out_width) ) == NULL )
    {
	return (FALSE);
    }
    h_factor = (float) inp_width / (float) out_width;
    v_factor = (float) inp_height / (float) out_height;
    if (iscale_func == NULL)
    {
	scaled_min = i_min;
	scaled_max = i_max;
    }
    else
    {
	if ( !(*iscale_func) (&scaled_min, 0, &i_min, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if ( !(*iscale_func) (&scaled_max, 0, &i_max, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if (scaled_max < scaled_min) swap (scaled_min, scaled_max, tmp);
    }
    d_mul = (cmap_size - 1) / (scaled_max - scaled_min);
    for (out_y = 0; out_y < out_height; ++out_y)
    {
	inp_y = (int) (v_factor * (float) out_y + tiny_offset);
	inp_line = inp_image + inp_voffsets[inp_y];
	/*  The following test will prevent the reading and conversion of
	    duplicated input image lines. A time saver  */
	if (inp_line != last_inp_line)
	{
	    /*  Convert line of input data to generic data type  */
	    /*  Note the cast from (iaddr *) to (uaddr *) for the offset
		arrays. This is dodgy, but it should work.  */
	    if ( !ds_get_scattered_elements (inp_line, inp_type,
					     (uaddr *) inp_hoffsets,
					     inp_values, &complex,
					     (unsigned int) inp_width) )
	    {
		fprintf (stderr, "Error converting data\n");
		return (FALSE);
	    }
	    if (complex) ds_complex_to_real_1D (inp_values, 2, inp_values,
						(unsigned int) inp_width,
						conv_type);
	    if (iscale_func != NULL)
	    {
		if ( !(*iscale_func) (inp_values, 2, inp_values, 2,
				      (unsigned int) inp_width, i_min, i_max,
				      iscale_info) )
		{
		    fprintf (stderr, "Error applying intensity scale\n");
		    return (FALSE);
		}
	    }
	    /*  Now take converted input data and generate an output line. This
		algorithm should work for subsampled data (zoomed out) and
		pixel replicated data (zoomed in).  */
	    for (out_x = 0; out_x < out_width; ++out_x)
	    {
		inp_x = (int) (h_factor * (float) out_x + tiny_offset);
		if ( (d_data = inp_values[inp_x * 2]) < scaled_min )
		{
		    out_pixels[out_x] = min_sat_pixel;
		}
		else if (d_data >= d_toobig) out_pixels[out_x] = blank_pixel;
		else if (d_data >scaled_max) out_pixels[out_x] = max_sat_pixel;
		else out_pixels[out_x] = pixels[(int) ( (d_data - scaled_min)
							* d_mul + offset )];
	    }
	}
	last_inp_line = inp_line;
	/*  Now ready to write the line  */
	/*  Loop for each pixel  */
	for (out_x = 0; out_x < out_width; ++out_x)
	{
	    XPutPixel (ximage, x_off + out_x, y_off + out_height - out_y - 1,
		       out_pixels[out_x]);
	}
    }
    return (TRUE);
}   /*  End Function pc_draw_lossy  */

static flag rgb_draw_lossy (XImage *ximage, int x_off, int y_off,
			    int out_width, int out_height,
			    CONST unsigned char *inp_red_image,
			    CONST unsigned char *inp_green_image,
			    CONST unsigned char *inp_blue_image,
			    CONST iaddr *inp_hoffsets,
			    CONST iaddr *inp_voffsets,
			    int inp_width, int inp_height,
			    unsigned long red_mask, unsigned long green_mask,
			    unsigned long blue_mask)
/*  [SUMMARY] Write generic 24 bit TrueColour subimage.
    [PURPOSE] This routine will write a generic 24 bit TrueColour image. The
    input and output sizes may be different. The output image is flipped
    vertically relative to the input image.
    <ximage> The XImage to write to.
    <x_off> The horizontal offset of the subimage, relative to the top-left
    corner of the image cache.
    <y_off> The vertical offset of the subimage, relative to the top-left
    corner of the image cache.
    <out_width> The width of the output subimage.
    <out_height> The height of the output subimage.
    <inp_red_image> The input red component image data.
    <inp_green_image> The input green component image data.
    <inp_blue_image> The input blue component image data.
    <inp_hoffsets> The array of horizontal byte offsets.
    <inp_voffsets> The array of vertical byte offsets.
    <inp_width> The width of the input image.
    <inp_height> The height of the input image.
    <red_mask> The mask for red pixels.
    <green_mask> The mask for green pixels.
    <blue_mask> The mask for blue pixels.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int out_x, out_y, inp_x, inp_y;
    unsigned long red, green, blue;
    uaddr inp_line_offset, last_inp_line_offset, off;
    float h_factor, v_factor;
    float tiny_offset = 1e-6;
    float rmask = (float) red_mask / 255.0;
    float gmask = (float) green_mask / 255.0;
    float bmask = (float) blue_mask / 255.0;
    unsigned long *out_pixels;
    static char function_name[] = "_kwin_X11_rgb_draw_lossy";

    if ( (inp_hoffsets == NULL) || (inp_voffsets == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if ( imw_test_verbose () )
    {
	 fprintf (stderr, "%s started...\n", function_name);
	 fprintf (stderr, "out_size: %d %d  inp_size: %d %d\n",
		  out_width, out_height, inp_width, inp_height);
    }
    /*  Allocate value buffers  */
    if ( ( out_pixels = alloc_out_pixels_buffer (out_width) ) == NULL )
    {
	return (FALSE);
    }
    h_factor = (float) inp_width / (float) out_width;
    v_factor = (float) inp_height / (float) out_height;
    last_inp_line_offset = 0;
    --last_inp_line_offset;
    for (out_y = 0; out_y < out_height; ++out_y)
    {
	inp_y = (int) (v_factor * (float) out_y + tiny_offset);
	inp_line_offset = inp_voffsets[inp_y];
	/*  The following test will prevent the reading and conversion of
	    duplicated input image lines. A time saver  */
	if (inp_line_offset != last_inp_line_offset)
	{
	    /*  Now take converted input data and generate an output line. This
		algorithm should work for subsampled data (zoomed out) and
		pixel replicated data (zoomed in).  */
	    for (out_x = 0; out_x < out_width; ++out_x)
	    {
		inp_x = (int) (h_factor * (float) out_x + tiny_offset);
		off = inp_line_offset + inp_hoffsets[inp_x];
		red = (unsigned long) ( (float) inp_red_image[off] * rmask ) &
		    red_mask;
		green = (unsigned long) ( (float)inp_green_image[off]*gmask ) &
		    green_mask;
		blue = (unsigned long) ( (float) inp_blue_image[off] *bmask ) &
		    blue_mask;
		out_pixels[out_x] = red | green | blue;
	    }
	}
	last_inp_line_offset = inp_line_offset;
	/*  Now ready to write the line  */
	/*  Loop for each pixel  */
	for (out_x = 0; out_x < out_width; ++out_x)
	{
	    XPutPixel (ximage, x_off + out_x, y_off + out_height - out_y - 1,
		       out_pixels[out_x]);
	}
    }
    return (TRUE);
}   /*  End Function rgb_draw_lossy  */

static int x11_err_handler (Display *dpy, XErrorEvent *event)
{
    caught_error = TRUE;
    return (0);
}   /*  End Function x11_err_handler  */

static double *alloc_inp_values_buffer (unsigned int num_values)
/*  This routine will allocate a buffer space for generic data values.
    The number of (DCOMPLEX) values to allocate must be given by  num_values  .
    The routine will return a pointer to the buffer space on success,
    else it returns NULL. The buffer is global and must NOT be deallocated.
*/
{
    static unsigned int value_buf_len = 0;
    static double *values = NULL;
    static char function_name[] = "alloc_inp_values_buffer";

    /*  Make sure value buffer is big enough  */
    if (value_buf_len < num_values)
    {
	/*  Buffer too small  */
	if (values != NULL)
	{
	    m_free (values);
	}
	value_buf_len = 0;
	if ( ( values = m_alloc (sizeof *values * 2 * num_values) ) == NULL )
	{
	    m_error_notify (function_name, "values buffer");
	    return (NULL);
	}
	value_buf_len = num_values;
    }
    return (values);
}   /*  End Function alloc_inp_values_buffer  */

static unsigned long *alloc_out_pixels_buffer (unsigned int num_pixels)
/*  This routine will allocate a buffer space for output pixels.
    The number of pixels to allocate must be given by  num_pixels  .
    The routine will return a pointer to the buffer space on success,
    else it returns NULL. The buffer is global and must NOT be deallocated.
*/
{
    static unsigned int pixel_buf_len = 0;
    static unsigned long *pixels = NULL;
    static char function_name[] = "alloc_out_pixels_buffer";

    /*  Make sure pixel buffer is big enough  */
    if (pixel_buf_len < num_pixels)
    {
	/*  Buffer too small  */
	if (pixels != NULL)
	{
	    m_free (pixels);
	}
	pixel_buf_len = 0;
	if ( ( pixels = m_alloc (sizeof *pixels * num_pixels) ) == NULL )
	{
	    m_error_notify (function_name, "pixels buffer");
	    return (NULL);
	}
	pixel_buf_len = num_pixels;
    }
    return (pixels);
}   /*  End Function alloc_out_pixels_buffer  */

#else  /*  HAS_X11  */

struct unused
{
    int a;
};

#endif  /*  !HAS_X11  */
