/*LINTLIBRARY*/
/*  editor.c

    This code provides a pixel canvas object creator.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for interactively creating geometric
    objects.


    Written by      Richard Gooch   29-JUL-1997

    Updated by      Richard Gooch   1-AUG-1997: Created functions
  <kwin_edit_set_attributes> and <kwin_edit_get_attributes>.

    Updated by      Richard Gooch   4-AUG-1997: Drop non-click event if
  creating rectangle.

    Updated by      Richard Gooch   6-AUG-1997: Drop non-click event if
  creating ellipse. Drop click event if modifying rectangle or ellipse.

    Updated by      Richard Gooch   10-AUG-1997: Moved attributes into sub
  structure.

    Updated by      Richard Gooch   19-AUG-1997: Added circle object.

    Updated by      Richard Gooch   19-SEP-1997: Added rotated circle object.

    Updated by      Richard Gooch   6-FEB-1998: Took account of rotation change

    Updated by      Richard Gooch   14-SEP-1999: Ignore enter and leave events.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Updated by      Richard Gooch   14-JUN-2000: Ignore keysym events.

    Last updated by Richard Gooch   27-JUN-2000: Created unpublished functions
  <kwin_editor_destroy_context> and <kwin_editor_register_destroy_func>.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#define NEW_WIN_SCALE
#include <karma.h>
#include <k_event_codes.h>
#include <karma_kwin.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>

#define MAGIC_NUMBER 34290432

#define VERIFY_CONTEXT(context) if (context == NULL) \
{fprintf (stderr, "NULL context passed\n"); \
 a_prog_bug (function_name); } \
if (context->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid context object\n"); \
 a_prog_bug (function_name); }


struct attribute_type
{
    unsigned int mouse_mask;
    unsigned int object_type;
    unsigned long pixel_value;
    flag clear_object;
};

struct kwin_editor_type
{
    unsigned int magic_number;
    KPixCanvas canvas;
    flag in_create;
    struct attribute_type att;
    unsigned int coord_buf_size;
    unsigned int num_coords;
    int *x;
    int *y;
    double *x_d;
    double *y_d;
    unsigned int ellipse_rotate;
    double angle;
    /*  Callback lists where others register an interest in this context   */
    KCallbackList event_list;
    KCallbackList destroy_list;
    /*  Callback functions this context has registered with canvas  */
    KCallbackFunc position_func;
    KCallbackFunc destroy_func;
    unsigned int num_areas;
    KPixCanvasRefreshArea areas[4];
};

struct event_struct
{
    unsigned int object_type;
    unsigned int num_coords;
    int *x;
    int *y;
    unsigned int event_code;
    void *e_info;
};


/*  Private functions  */
STATIC_FUNCTION (void process_attributes,
		 (KPixCanvasEditorContext context, va_list argp) );
STATIC_FUNCTION (void refresh_func,
		 (KPixCanvas canvas, int width, int height,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (flag position_func,
		 (KPixCanvas canvas, int x, int y,
		  unsigned int event_code, void *e_info, void **f_info) );
STATIC_FUNCTION (void reallocate_coords,
		 (KPixCanvasEditorContext context,
		  unsigned int min_size, flag copy_old) );
STATIC_FUNCTION (flag _kwin_editor_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (void compute_clear_areas,
		 (KPixCanvasEditorContext context, flag cancel) );
STATIC_FUNCTION (flag clear_canvas, (KPixCanvasEditorContext context) );


/*  Public Functions follow  */

/*UNPUBLISHED_FUNCTION*/
KPixCanvasEditorContext kwin_editor_create_context (KPixCanvas canvas, ...)
/*  [SUMMARY] Create a geometry editing context for a pixel canvas list.
    <canvas> The pixel canvas on which to receive events.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    KWIN_EDITOR_CONTEXT_ATT_END. See [<kwin_EDITOR_CONTEXT_ATTRIBUTES>]
    for the list of attributes.
    [NOTE] The context will be automatically destroyed if the pixel canvas is
    destroyed.
    [RETURNS] A KPixCanvasEditorContext object on success, else NULL.
*/
{
    va_list argp;
    KPixCanvasEditorContext context;
    static char function_name[] = "kwin_editor_create_context";

    va_start (argp, canvas);
    if ( ( context = m_calloc (sizeof *context) ) == NULL )
    {
	m_error_notify (function_name, "KPixCanvasEditorContext object");
	return (NULL);
    }
    context->magic_number = MAGIC_NUMBER;
    context->canvas = canvas;
    context->att.mouse_mask = K_CANVAS_EVENT_LEFT_MOUSE_MASK;
    context->att.object_type = KWIN_EDIT_OBJECT_NONE;
    context->in_create = FALSE;
    context->att.clear_object = TRUE;
    process_attributes (context, argp);
    va_end (argp);
    reallocate_coords (context, 256, FALSE);
    context->num_coords = 0;
    kwin_register_refresh_func (canvas, refresh_func, context);
    context->position_func =
	kwin_register_position_event_func (canvas, position_func, context);
    context->destroy_func =
	kwin_register_destroy_func (canvas, kwin_editor_destroy_context,
				    context);
    return (context);
}   /*  End Function kwin_editor_create_context  */

/*UNPUBLISHED_FUNCTION*/
void kwin_editor_destroy_context (KPixCanvasEditorContext context)
/*  [SUMMARY] Destroy a KPixCanvasEditorContext object.
    <context> The context to destroy.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "kwin_editor_destroy_context";

    VERIFY_CONTEXT (context);
    c_unregister_callback (context->position_func);
    c_unregister_callback (context->destroy_func);
    c_call_callbacks (context->destroy_list, NULL);
    c_destroy_list (context->event_list);
    c_destroy_list (context->destroy_list);
    if (context->coord_buf_size > 0)
    {
	m_free (context->x);
	m_free (context->y);
	m_free (context->x_d);
	m_free (context->y_d);
    }
    context->magic_number = 0;
    m_free (context);
}   /*  End Function kwin_editor_destroy_context  */

/*UNPUBLISHED_FUNCTION*/
KCallbackFunc kwin_editor_register_event_func (KPixCanvasEditorContext context,
					       flag (*func) (), void *f_info)
/*  [SUMMARY] Register object event callback.
    [PURPOSE] This routine will register an object event function for a pixel
    canvas editor context. The object event function will be called whenever a
    new object is created. Many object event functions may be registered per
    context. The first function registered is the first function called upon an
    object event.
    <context> The context.
    <func> The function that is called when an object event occurs. The
    prototype function is [<kwin_PROTO_editor_func>].
    <f_info> The initial arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "kwin_editor_register_event_func";

    VERIFY_CONTEXT (context);
    return c_register_callback (&context->event_list,
				_kwin_editor_event_func,
				context, f_info, TRUE, (void *) func, FALSE,
				TRUE);
}   /*  End Function kwin_editor_register_event_func  */

/*UNPUBLISHED_FUNCTION*/
KCallbackFunc kwin_editor_register_destroy_func (KPixCanvasEditorContext ctx,
						 void (*func) (), void *info)
/*  [SUMMARY] Register destroy callback.
    [PURPOSE] This routine will register a destroy function for a pixel
    canvas editor context. The destroy function will be called when the editor
    context is destroyed. Many destroy functions may be registered per
    context. The first function registered is the first function called upon
    destruction.
    <context> The context.
    <func> The function that is called upon destruction. The prototype function
    is [<kwin_PROTO_destroy_func>].
    <info> The arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    static char function_name[] = "kwin_editor_register_destroy_func";

    VERIFY_CONTEXT (ctx);
    return c_register_callback (&ctx->destroy_list, ( flag (*) () ) func,
				info, NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function kwin_editor_register_destroy_func  */

/*UNPUBLISHED_FUNCTION*/
void kwin_editor_set_attributes (KPixCanvasEditorContext context, ...)
/*  [SUMMARY] Set attributes for a geometry editing context.
    <context> The context.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    KWIN_EDITOR_CONTEXT_ATT_END. See [<kwin_EDITOR_CONTEXT_ATTRIBUTES>]
    for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    static char function_name[] = "kwin_editor_set_attributes";

    VERIFY_CONTEXT (context);
    va_start (argp, context);
    process_attributes (context, argp);
    va_end (argp);
}   /*  End Function kwin_editor_set_attributes  */

/*UNPUBLISHED_FUNCTION*/
void kwin_editor_get_attributes (KPixCanvasEditorContext context, ...)
/*  [SUMMARY] Get attributes for a geometry editing context.
    <context> The context.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    KWIN_EDITOR_CONTEXT_ATT_END. See [<kwin_EDITOR_CONTEXT_ATTRIBUTES>]
    for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key;
    static char function_name[] = "kwin_editor_get_attributes";

    VERIFY_CONTEXT (context);
    va_start (argp, context);
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KWIN_EDITOR_CONTEXT_ATT_END )
    {
	switch (att_key)
	{
	  case KWIN_EDITOR_CONTEXT_ATT_MOUSE:
	    *( va_arg (argp, unsigned int *) ) = context->att.mouse_mask;
	    break;
	  case KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT:
	    *( va_arg (argp, unsigned int *) ) = context->att.object_type;
	    break;
	  case KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE:
	    *( va_arg (argp, unsigned long *) ) = context->att.pixel_value;
	    break;
	  case KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT:
	    *( va_arg (argp, flag *) ) = context->att.clear_object;
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
}   /*  End Function kwin_editor_get_attributes  */

/*UNPUBLISHED_FUNCTION*/
KPixCanvas kwin_editor_get_pixcanvas (KPixCanvasEditorContext context)
/*  [SUMMARY] Get the pixel canvas for an editing context.
    <context> The editing context.
    [RETURNS] The KPixCanvas.
*/
{
    static char function_name[] = "kwin_editor_get_pixcanvas";

    VERIFY_CONTEXT (context);
    return (context->canvas);
}   /*  End Function kwin_editor_get_pixcanvas  */


/*  Private functions follow  */

static void process_attributes (KPixCanvasEditorContext context, va_list argp)
/*  [SUMMARY] Process attributes for a KPixCanvasEditorContext.
    <context> The pixel canvas editor context.
    <argp> The varargs parameters.
    [RETURNS] Nothing.
*/
{
    unsigned int att_key, mask;
    static char function_name[] = "_kwin_editor_process_attributes";

    if (context->in_create)
    {
	fprintf (stderr, "Cannot modify attributes while editing\n");
	a_prog_bug (function_name);
    }
    mask = K_CANVAS_EVENT_LEFT_MOUSE_MASK | K_CANVAS_EVENT_MIDDLE_MOUSE_MASK |
	K_CANVAS_EVENT_RIGHT_MOUSE_MASK;
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KWIN_EDITOR_CONTEXT_ATT_END )
    {
	switch (att_key)
	{
	  case KWIN_EDITOR_CONTEXT_ATT_MOUSE:
	    context->att.mouse_mask = va_arg (argp, unsigned int) & mask;
	    break;
	  case KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT:
	    context->att.object_type = va_arg (argp, unsigned int);
	    break;
	  case KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE:
	    context->att.pixel_value = va_arg (argp, unsigned long);
	    break;
	  case KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT:
	    context->att.clear_object = va_arg (argp, flag);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
}   /*  End Function process_attributes  */

static void refresh_func (KPixCanvas canvas, int width, int height,
			  void **info, PostScriptPage pspage,
			  unsigned int num_areas,
			  KPixCanvasRefreshArea *areas,
			  flag *honoured_areas)
/*  [SUMMARY] Process a refresh event for a pixel canvas.
    <canvas> The pixel canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire pixel canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KPixCanvasEditorContext context = (KPixCanvasEditorContext) *info;
    int x1, y1, x2, y2;
    double tmp;
    static char function_name[] = "_kwin_editor_refresh_func";

    VERIFY_CONTEXT (context);
    if (!context->in_create) return;
    switch (context->att.object_type)
    {
      case KWIN_EDIT_OBJECT_RECTANGLE:
	if (context->x[0] < context->x[1])
	{
	    x1 = context->x[0];
	    x2 = context->x[1] - context->x[0];
	}
	else
	{
	    x1 = context->x[1];
	    x2 = context->x[0] - context->x[1];
	}
	if (context->y[0] < context->y[1])
	{
	    y1 = context->y[0];
	    y2 = context->y[1] - context->y[0];
	}
	else
	{
	    y1 = context->y[1];
	    y2 = context->y[0] - context->y[1];
	}
	kwin_draw_rectangle (canvas, x1, y1, x2, y2, context->att.pixel_value);
	break;
      case KWIN_EDIT_OBJECT_ELLIPSE:
	kwin_draw_point (canvas, context->x[0], context->y[0],
			 context->att.pixel_value);
	kwin_draw_arc (canvas, context->x[0], context->y[0],
		       fabs (context->x[1]), fabs (context->y[1]), 0.0, 360.0,
		       context->angle / PION180, context->att.pixel_value,
		       FALSE);
	break;
      case KWIN_EDIT_OBJECT_CIRCLE:
	kwin_draw_point (canvas, context->x[0], context->y[0],
			 context->att.pixel_value);
	kwin_draw_ellipse (canvas, context->x[0], context->y[0],
			   context->x[1], context->y[1],
			   context->att.pixel_value);
	break;
      case KWIN_EDIT_OBJECT_ROTATED_CIRCLE:
	kwin_draw_line (canvas, context->x[0], context->y[0],
			context->x[1], context->y[1],
			context->att.pixel_value);
	tmp = sqrt (context->x[1] * context->x[1] +
		    context->y[1] * context->y[1]);
	kwin_draw_ellipse (canvas, context->x[0], context->y[0], tmp, tmp,
			   context->att.pixel_value);
	break;
      case KWIN_EDIT_OBJECT_POLYGON:
	if (context->num_coords < 2) return;
	kwin_draw_lines_TRANSITION (canvas, context->x_d, context->y_d,
				    context->num_coords,
				    context->att.pixel_value);
	break;
      default:
	fprintf (stderr, "Illegal edit object: %u\n",context->att.object_type);
	a_prog_bug (function_name);
	break;
    }
}   /*  End Function refresh_func  */

static flag position_func (KPixCanvas canvas, int x, int y,
			   unsigned int event_code, void *e_info,
			   void **f_info)
/*  [SUMMARY] Process a position event on a pixel canvas.
    <canvas> The pixel canvas on which the event occurred.
    <x> The horizontal position of the event, relative to the canvas origin
    <y> The vertical position of the event, relative to the canvas origin.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    flag click = FALSE;
    flag release = FALSE;
    flag this_mouse = FALSE;
    KPixCanvasEditorContext context = (KPixCanvasEditorContext) *f_info;
    int tmp;
    int str_len = 0;
    unsigned int num;
    unsigned long mask = 0;
    double dx, dy;
    struct event_struct data;
    char *string = NULL;
    static char function_name[] = "_kwin_editor_position_func";

    VERIFY_CONTEXT (context);
    if (context->canvas != canvas)
    {
	fprintf (stderr, "Spurious event\n");
	a_prog_bug (function_name);
    }
    if (context->att.object_type == KWIN_EDIT_OBJECT_NONE) return (FALSE);
    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
	click = TRUE;
	if (context->att.mouse_mask & K_CANVAS_EVENT_LEFT_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
	if (context->att.mouse_mask & K_CANVAS_EVENT_LEFT_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
	release = TRUE;
	if (context->att.mouse_mask & K_CANVAS_EVENT_LEFT_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
	click = TRUE;
	if (context->att.mouse_mask & K_CANVAS_EVENT_MIDDLE_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
	if (context->att.mouse_mask & K_CANVAS_EVENT_MIDDLE_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	release = TRUE;
	if (context->att.mouse_mask & K_CANVAS_EVENT_MIDDLE_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
	click = TRUE;
	if (context->att.mouse_mask & K_CANVAS_EVENT_RIGHT_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
	if (context->att.mouse_mask & K_CANVAS_EVENT_RIGHT_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
	release = TRUE;
	if (context->att.mouse_mask & K_CANVAS_EVENT_RIGHT_MOUSE_MASK)
	    this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	string = (char *) e_info;
	break;
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	string = (char *) e_info;
	str_len = strlen (string);
	m_copy (&mask, string + str_len + 1, sizeof mask);
	if (context->att.mouse_mask & mask) this_mouse = TRUE;
	break;
      case K_CANVAS_EVENT_POINTER_ENTER:
      case K_CANVAS_EVENT_POINTER_LEAVE:
      case K_CANVAS_EVENT_KEYSYM:
	return (FALSE);
	/*break;*/
    }
    if (!context->in_create && !this_mouse) return (FALSE);
    if (!context->in_create && this_mouse)
    {
	if (!click && !release) return (FALSE);
	/*  YES! Start a new object  */
	switch (context->att.object_type)
	{
	  case KWIN_EDIT_OBJECT_RECTANGLE:
	    if (!click) return (FALSE);
	    reallocate_coords (context, 2, FALSE);
	    context->x[0] = x;
	    context->y[0] = y;
	    context->x[1] = x;
	    context->y[1] = y;
	    break;
	  case KWIN_EDIT_OBJECT_ELLIPSE:
	    if (!click) return (FALSE);
	    reallocate_coords (context, 3, FALSE);
	    context->x[0] = x;
	    context->y[0] = y;
	    context->ellipse_rotate = FALSE;
	    context->angle = 0.0;
	    break;
	  case KWIN_EDIT_OBJECT_CIRCLE:
	    if (!click) return (FALSE);
	    reallocate_coords (context, 2, FALSE);
	    context->x[0] = x;
	    context->y[0] = y;
	    break;
	  case KWIN_EDIT_OBJECT_ROTATED_CIRCLE:
	    if (!click) return (FALSE);
	    reallocate_coords (context, 2, FALSE);
	    context->x[0] = x;
	    context->y[0] = y;
	    break;
	  case KWIN_EDIT_OBJECT_POLYGON:
	    if (click) return (TRUE);
	    reallocate_coords (context, 2, FALSE);
	    context->x[0] = x;
	    context->y[0] = y;
	    context->x_d[0] = x;
	    context->y_d[0] = y;
	    context->x[1] = x;
	    context->y[1] = y;
	    context->x_d[1] = x;
	    context->y_d[1] = y;
	    break;
	  default:
	    fprintf (stderr, "Illegal edit object: %u\n",
		     context->att.object_type);
	    a_prog_bug (function_name);
	    break;
	}
	context->in_create = TRUE;
	return (TRUE);
    }
    /*  Modifying object  */
    switch (context->att.object_type)
    {
      case KWIN_EDIT_OBJECT_RECTANGLE:
	if (!this_mouse || click)
	{
	    context->in_create = FALSE;
	    compute_clear_areas (context, TRUE);
	    clear_canvas (context);
	    return (FALSE);
	}
	compute_clear_areas (context, FALSE);
	context->x[1] = x;
	context->y[1] = y;
	if ( release || (string != NULL) ) break;
	clear_canvas (context);
	return (TRUE);
	/*break;*/
      case KWIN_EDIT_OBJECT_ELLIPSE:
	if (!this_mouse || click)
	{
	    context->in_create = FALSE;
	    compute_clear_areas (context, TRUE);
	    clear_canvas (context);
	    return (FALSE);
	}
	compute_clear_areas (context, FALSE);
	if ( this_mouse && (string != NULL) && (*string == 'r') )
	{
	    context->ellipse_rotate = context->ellipse_rotate ? FALSE : TRUE;
	    return (TRUE);
	}
	if (context->ellipse_rotate)
	{
	    double ang1, ang2;

	    ang1 = atan2 (y - context->y[0], x - context->x[0]);
	    ang2 = atan2 (context->y[2] - context->y[0],
			  context->x[2] - context->x[0]);
	    context->angle += ang1 - ang2;
	    while (context->angle < 0.0) context->angle += TWOPI;
	    while (context->angle > TWOPI) context->angle -= TWOPI;
	}
	else
	{
	    double x1, y1;

	    dx = x - context->x[0];
	    dy = context->y[0] - y;
	    x1 = dx * cos (context->angle) - dy * sin (context->angle);
	    y1 = dy * cos (context->angle) + dx * sin (context->angle);
	    context->x[1] = fabs (x1);
	    context->y[1] = fabs (y1);
	}
	context->x[2] = x;
	context->y[2] = y;
	if ( release || (string != NULL) ) break;
	clear_canvas (context);
	return (TRUE);
	/*break;*/
      case KWIN_EDIT_OBJECT_CIRCLE:
	if (!this_mouse || click)
	{
	    context->in_create = FALSE;
	    compute_clear_areas (context, TRUE);
	    clear_canvas (context);
	    return (FALSE);
	}
	compute_clear_areas (context, FALSE);
	dx = x - context->x[0];
	dy = context->y[0] - y;
	dx = sqrt (dx * dx + dy * dy);
	context->x[1] = dx;
	context->y[1] = dx;
	if ( release || (string != NULL) ) break;
	clear_canvas (context);
	return (TRUE);
	/*break;*/
      case KWIN_EDIT_OBJECT_ROTATED_CIRCLE:
	if (!this_mouse || click)
	{
	    context->in_create = FALSE;
	    compute_clear_areas (context, TRUE);
	    clear_canvas (context);
	    return (FALSE);
	}
	compute_clear_areas (context, FALSE);
	context->x[1] = x;
	context->y[1] = y;
	if ( release || (string != NULL) ) break;
	clear_canvas (context);
	return (TRUE);
	/*break;*/
      case KWIN_EDIT_OBJECT_POLYGON:
	if (click) return (TRUE);
	num = context->num_coords;
	if ( release || (event_code == K_CANVAS_EVENT_POINTER_MOVE) )
	{
	    compute_clear_areas (context, FALSE);
	    context->x[num - 1] = x;
	    context->y[num - 1] = y;
	    context->x_d[num - 1] = x;
	    context->y_d[num - 1] = y;
	}
	else compute_clear_areas (context, TRUE);
	if (event_code == K_CANVAS_EVENT_POINTER_MOVE)
	{
	    clear_canvas (context);
	    return (FALSE);
	}
	if (!release)
	{
	    context->in_create = FALSE;
	    clear_canvas (context);
	    return (FALSE);
	}
	if ( ( (x == context->x[0]) && (y == context->y[0]) ) ||
	     ( (x == context->x[num - 2]) && (y == context->y[num - 2]) ) )
	    break;
	reallocate_coords (context, num + 1, TRUE);
	context->x[num] = x;
	context->y[num] = y;
	context->x_d[num] = x;
	context->y_d[num] = y;
	clear_canvas (context);
	return (TRUE);
	/*break;*/
      default:
	fprintf (stderr, "Illegal edit object: %u\n",context->att.object_type);
	a_prog_bug (function_name);
	break;
    }
    /*  Object completed  */
    context->in_create = FALSE;
    if (context->att.clear_object)
    {
	compute_clear_areas (context, TRUE);
	clear_canvas (context);
    }
    switch (context->att.object_type)
    {
      case KWIN_EDIT_OBJECT_RECTANGLE:
	data.num_coords = 2;
	if (context->x[0] > context->x[1]) swap (context->x[0], context->x[1],
						 tmp);
	if (context->y[0] > context->y[1]) swap (context->y[0], context->y[1],
						 tmp);
	break;
      case KWIN_EDIT_OBJECT_ELLIPSE:
	data.num_coords = 3;
	context->x[1] = abs (context->x[1]);
	context->y[1] = abs (context->y[1]);
	context->x[2] = (64.0 * context->angle / PION180);
	context->y[2] = 0;
	break;
      case KWIN_EDIT_OBJECT_POLYGON:
	data.num_coords = context->num_coords - 1;
	break;
      case KWIN_EDIT_OBJECT_CIRCLE:
	data.num_coords = 2;
	break;
      case KWIN_EDIT_OBJECT_ROTATED_CIRCLE:
	data.num_coords = 2;
	break;
      default:
	fprintf (stderr, "Illegal edit object: %u\n",context->att.object_type);
	a_prog_bug (function_name);
	break;
    }
    data.object_type = context->att.object_type;
    data.x = context->x;
    data.y = context->y;
    data.event_code = event_code;
    data.e_info = e_info;
    c_call_callbacks (context->event_list, &data);
    return (TRUE);
}   /*  End Function position_func  */

static void reallocate_coords (KPixCanvasEditorContext context,
			       unsigned int min_size, flag copy_old)
/*  [SUMMARY] Reallocate the co-ordinate buffers.
    <context> The KPixCanvasEditorContext.
    <min_size> The minimum size of the buffers.
    <copy_old> If TRUE, old values are copied on reallocation.
    [RETURNS] Nothing. On failure the process aborts.
*/
{
    int *x, *y;
    double *x_d, *y_d;
    static char function_name[] = "_kwin_editor_reallocate_coords";

    context->num_coords = min_size;
    if (context->coord_buf_size >= min_size) return;
    if ( ( x = m_alloc (min_size * sizeof *x) ) == NULL )
	m_abort (function_name, "x co-ordinate array");
    if ( ( y = m_alloc (min_size * sizeof *y) ) == NULL )
	m_abort (function_name, "y co-ordinate array");
    if ( ( x_d = m_alloc (min_size * sizeof *x_d) ) == NULL )
	m_abort (function_name, "x co-ordinate array");
    if ( ( y_d = m_alloc (min_size * sizeof *y_d) ) == NULL )
	m_abort (function_name, "y co-ordinate array");
    if (context->coord_buf_size > 0)
    {
	if (copy_old)
	{
	    m_copy (x, context->x, context->coord_buf_size * sizeof *x);
	    m_copy (y, context->y, context->coord_buf_size * sizeof *y);
	    m_copy (x_d, context->x_d, context->coord_buf_size * sizeof *x_d);
	    m_copy (y_d, context->y_d, context->coord_buf_size * sizeof *y_d);
	}
	m_free (context->x);
	m_free (context->y);
	m_free (context->x_d);
	m_free (context->y_d);
    }
    context->x = x;
    context->y = y;
    context->x_d = x_d;
    context->y_d = y_d;
    context->coord_buf_size = min_size;
}   /*  End Function reallocate_coords  */

static flag _kwin_editor_event_func (void *object, void *client1_data,
				     void *call_data, void *client2_data)
/*  [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KPixCanvasEditorContext context = (KPixCanvasEditorContext) object;
    flag (*func) ();
    struct event_struct *data = call_data;
    static char function_name[] = "_kwin_editor_event_func";

    VERIFY_CONTEXT (context);
    func = ( flag (*) () ) client2_data;
    return (*func) (context, context->canvas, data->object_type,
		    data->num_coords, data->x, data->y, data->event_code,
		    data->e_info, client1_data);
}   /*  End Function _kwin_editor_event_func  */

static void compute_clear_areas (KPixCanvasEditorContext context, flag cancel)
/*  [SUMMARY] Compute the clear areas for a canvas.
    <context> The KPixCanvasEditorContext object.
    <cancel> If TRUE, the object is being cancelled.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int x1, y1, x2, y2, tmp, count;
    int thickness = 1;
    static char function_name[] = "_kwin_editor_clear_canvas";

    m_clear (context->areas, sizeof context->areas);
    context->num_areas = 0;
    switch (context->att.object_type)
    {
      case KWIN_EDIT_OBJECT_RECTANGLE:
	context->num_areas = 4;
	x1 = context->x[0];
	x2 = context->x[1];
	y1 = context->y[0];
	y2 = context->y[1];
	if (x1 > x2) swap (x1, x2, tmp);
	if (y1 > y2) swap (y1, y2, tmp);
	/*  Top line  */
	context->areas[0].startx = x1 - thickness;
	context->areas[0].endx = x2 + thickness;
	context->areas[0].starty = y1 - thickness;
	context->areas[0].endy = y1 + thickness;
	context->areas[0].clear = TRUE;
	/*  Bottom line  */
	context->areas[1].startx = x1 - thickness;
	context->areas[1].endx = x2 + thickness;
	context->areas[1].starty = y2 - thickness;
	context->areas[1].endy = y2 + thickness;
	context->areas[1].clear = TRUE;
	/*  Left line  */
	context->areas[2].startx = x1 - thickness;
	context->areas[2].endx = x1 + thickness;
	context->areas[2].starty = y1 - thickness;
	context->areas[2].endy = y2 + thickness;
	context->areas[2].clear = TRUE;
	/*  Right line  */
	context->areas[3].startx = x2 - thickness;
	context->areas[3].endx = x2 + thickness;
	context->areas[3].starty = y1 - thickness;
	context->areas[3].endy = y2 + thickness;
	context->areas[3].clear = TRUE;
	break;
      case KWIN_EDIT_OBJECT_ELLIPSE:
	context->num_areas = 1;
	tmp = (context->x[1] > context->y[1]) ? context->x[1] : context->y[1];
	context->areas[0].startx = context->x[0] - tmp - 1;
	context->areas[0].endx = context->x[0] + tmp + 1;
	context->areas[0].starty = context->y[0] - tmp - 1;
	context->areas[0].endy = context->y[0] + tmp + 1;
	context->areas[0].clear = TRUE;
	break;
      case KWIN_EDIT_OBJECT_CIRCLE:
	context->num_areas = 1;
	context->areas[0].startx = context->x[0] - context->x[1] - 1;
	context->areas[0].endx = context->x[0] + context->x[1] + 1;
	context->areas[0].starty = context->y[0] - context->x[1] - 1;
	context->areas[0].endy = context->y[0] + context->x[1] + 1;
	context->areas[0].clear = TRUE;
	break;
      case KWIN_EDIT_OBJECT_ROTATED_CIRCLE:
	context->num_areas = 1;
	x1 = context->x[0] - context->x[1];
	y1 = context->y[0] - context->y[1];
	x2 = sqrt (x1 * x1 + y1 * y1) + 0.5;
	context->areas[0].startx = context->x[0] - x2 - 1;
	context->areas[0].endx = context->x[0] + x2 + 1;
	context->areas[0].starty = context->y[0] - x2 - 1;
	context->areas[0].endy = context->y[0] + x2 + 1;
	context->areas[0].clear = TRUE;
	break;
      case KWIN_EDIT_OBJECT_POLYGON:
	context->num_areas = 1;
	if (cancel)
	{
	    /*  Search for min&max x&y  */
	    x1 = context->x[0];
	    x2 = context->x[0];
	    y1 = context->y[0];
	    y2 = context->y[0];
	    for (count = 1; count < context->num_coords; ++count)
	    {
		if (context->x[count] < x1) x1 = context->x[count];
		if (context->x[count] > x2) x2 = context->x[count];
		if (context->y[count] < y1) y1 = context->y[count];
		if (context->y[count] > y2) y2 = context->y[count];
	    }
	}
	else
	{
	    x1 = context->x[context->num_coords - 2];
	    y1 = context->y[context->num_coords - 2];
	    x2 = context->x[context->num_coords - 1];
	    y2 = context->y[context->num_coords - 1];
	    if (x1 > x2) swap (x1, x2, tmp);
	    if (y1 > y2) swap (y1, y2, tmp);
	}
	context->areas[0].startx = x1 - 1;
	context->areas[0].endx = x2 + 1;
	context->areas[0].starty = y1 - 1;
	context->areas[0].endy = y2 + 1;
	context->areas[0].clear = TRUE;
	break;
      default:
	fprintf (stderr, "Illegal edit object: %u\n",context->att.object_type);
	a_prog_bug (function_name);
	break;
    }
}   /*  End Function compute_clear_canvas  */

static flag clear_canvas (KPixCanvasEditorContext context)
/*  [SUMMARY] Clear the canvas in an efficient way.
    <context> The KPixCanvasEditorContext object.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    /*static char function_name[] = "_kwin_editor_clear_canvas";*/

    if (context->num_areas < 1)
    {
	return kwin_refresh_if_visible (context->canvas, TRUE);
    }
    return kwin_partial_refresh (context->canvas,
				 context->num_areas, context->areas, FALSE);
}   /*  End Function clear_canvas  */
