/*LINTLIBRARY*/
/*  memory.c

    This code provides the memory driver for KPixCanvas objects.

    Copyright (C) 2000-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating a simple pixel
    canvas (window) independent of the graphics system in use. This file
    contains the memory driver code.


    Written by      Richard Gooch   13-MAY-2000

    Updated by      Richard Gooch   18-MAY-2000: Finished initial
  implementation which supports PC and RGB images. Magnification function
  supported for RGB images (PC images not supported yet: requires changes to
  <imw> package).

    Updated by      Richard Gooch   27-MAY-2000: Took account of interface
  change to <kwin_new_driver_refresh>.

    Updated by      Richard Gooch   1-JUN-2000: Fixed bugs in image drawing
  functions: Y offset was not being flipped.

    Last updated by Richard Gooch   19-JUN-2001: Documentation fix for
  <draw_arc>.


*/
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_kwin.h>
#include <karma_imw.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include <karma_s.h>


#define KPixHookCanvas MemCanvas

typedef struct
{
    unsigned char red;
    unsigned char green;
    unsigned char blue;
} rgb_type;

typedef struct memcanvas_type
{
    unsigned int magic_number;
    KPixCanvas pixcanvas;
    unsigned char *red, *green, *blue;
    uaddr stride;
    unsigned int mem_width, mem_height;
    int pixcanvas_width, pixcanvas_height;
    flag pixel_valid;
    unsigned long pixel_value;
    rgb_type rgb;
    double linewidth;
} *MemCanvas;

#include <karma_kwin_hooks.h>


#define MAX_INTENSITY 65535.0
#define CANVAS_MAGIC_NUMBER 937976328
#define VERIFY_CANVAS(canvas) if (canvas == NULL) \
{fprintf (stderr, "NULL canvas passed\n"); \
 a_prog_bug (function_name); } \
if (canvas->magic_number != CANVAS_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas object\n"); \
 a_prog_bug (function_name); }

/*  Structure declarations  */


/*  Private data  */


/*  Mandatory functions  */
STATIC_FUNCTION (flag draw_points, (MemCanvas memcanvas,
				    double *x_array, double *y_array,
				    unsigned int num_points,
				    unsigned long pixel_value) );
/*  Optional hook functions  */
STATIC_FUNCTION (flag draw_pc_image,
		 (MemCanvas memcanvas,
		  int x_off, int y_off, int x_pixels, int y_pixels,
		  CONST char *slice,
		  CONST uaddr *hoffsets, CONST uaddr *voffsets,
		  unsigned int width, unsigned int height,
		  unsigned int type, unsigned int conv_type,
		  unsigned int num_pixels,unsigned long *pixel_values,
		  unsigned long blank_pixel,
		  unsigned long min_sat_pixel, unsigned long max_sat_pixel,
		  double i_min, double i_max,
		  flag (*iscale_func) (), void *iscale_info,
		  KPixCanvasImageCache *cache_ptr) );
STATIC_FUNCTION (flag draw_rgb_image,
		 (MemCanvas memcanvas,
		  int x_off, int y_off, int x_pixels, int y_pixels,
		  CONST unsigned char *red_slice,
		  CONST unsigned char *green_slice,
		  CONST unsigned char *blue_slice,
		  CONST uaddr *hoffsets, CONST uaddr *voffsets,
		  unsigned int width, unsigned int height,
		  KPixCanvasImageCache *cache_ptr) );
STATIC_FUNCTION (flag draw_line, (MemCanvas memcanvas,
				  double x0, double y0, double x1, double y1,
				  unsigned long pixel_value) );
STATIC_FUNCTION (flag draw_arc,
		 (MemCanvas memcanvas,
		  double centre_x, double centre_y,
		  double radius_x, double radius_y,
		  int angle1, int angle2, int rotation,
		  unsigned long pixel_value, flag fill) );
STATIC_FUNCTION (flag set_linewidth, (MemCanvas memcanvas, double linewidth) );
STATIC_FUNCTION (flag set_line_dash, (MemCanvas memcanvas, flag dash) );


/*  Private functions  */
STATIC_FUNCTION (void get_rgb_of_pixel, (MemCanvas memcanvas, rgb_type *rgb,
					 unsigned long pixel_value) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag kwin_mem_write_rgb (KPixCanvas canvas, void *red, void *green, void *blue,
			 uaddr stride, unsigned int width, unsigned int height)
/*  [SUMMARY] Refresh a pixel canvas into a memory buffer.
    [PURPOSE] This routine will refresh a pixel canvas, redirecting output to a
    memory buffer.
    <canvas> The pixel canvas.
    <red> The red buffer pointer.
    <green> The green buffer pointer.
    <blue> The blue buffer pointer.
    <stride> The stride in bytes between adjacent pixels.
    <width> The width of the memory buffer in pixels.
    <height> The height of the memory buffer in pixels.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok;
    struct memcanvas_type memcanvas;
    double resolution, tmp;
    /*static char function_name[] = "kwin_mem_write_rgb";*/

    memcanvas.magic_number = CANVAS_MAGIC_NUMBER;
    memcanvas.pixcanvas = canvas;
    memcanvas.red = red;
    memcanvas.green = green;
    memcanvas.blue = blue;
    memcanvas.stride = stride;
    memcanvas.mem_width = width;
    memcanvas.mem_height = height;
    memcanvas.pixel_valid = FALSE;
    kwin_get_attributes (canvas,
			 KWIN_ATT_LINEWIDTH, &memcanvas.linewidth,
			 KWIN_ATT_END);
    kwin_get_size (canvas,
		   &memcanvas.pixcanvas_width, &memcanvas.pixcanvas_height);
    resolution = (double) memcanvas.pixcanvas_width / (double) width;
    tmp = (double) memcanvas.pixcanvas_height / (double) height;
    if (tmp < resolution) resolution = tmp;
    tmp = floor (resolution + 1e-3);
    if (floor (resolution) != tmp) resolution = tmp;
    ok = kwin_new_driver_refresh (canvas, NULL, resolution, &memcanvas,
				  KWIN_FUNC_DRAW_POINTS, draw_points,
				  KWIN_FUNC_DRAW_PC_IMAGE, draw_pc_image,
				  KWIN_FUNC_DRAW_RGB_IMAGE, draw_rgb_image,
				  KWIN_FUNC_DRAW_LINE, draw_line,
				  KWIN_FUNC_DRAW_ARC, draw_arc,
				  KWIN_FUNC_SET_LINEWIDTH, set_linewidth,
				  KWIN_FUNC_SET_LINE_DASH, set_line_dash,
				  KWIN_ATT_END);
    m_clear (&memcanvas, sizeof memcanvas);
    return (ok);
}   /*  End Function kwin_mem_write_rgb  */


/*  Mandatory functions follow  */

static flag draw_points (MemCanvas memcanvas, double *x_array, double *y_array,
			 unsigned int num_points, unsigned long pixel_value)
/*  [SUMMARY] Draw a point in a memory canvas.
    <memcanvas> The memory canvas.
    <x_array> The horizontal co-ordinates of the points.
    <y_array> The vertical co-ordinates of the points.
    <num_points> The number of points.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    rgb_type rgb;
    static char function_name[] = "_kwin_mem_draw_points";

    VERIFY_CANVAS (memcanvas);
    get_rgb_of_pixel (memcanvas, &rgb, pixel_value);
    for (count = 0; count < num_points; ++count)
    {
	int x, y;
	uaddr offset;

	x = ( (double) x_array[count] * (double) (memcanvas->mem_width - 1) /
	      (double) (memcanvas->pixcanvas_width - 1) + 1e-3 );
	y = memcanvas->pixcanvas_height - 1 - y_array[count];
	y = ( (double) y * (double) (memcanvas->mem_height - 1) /
	      (double) (memcanvas->pixcanvas_height - 1) + 1e-3 );
	if ( (x < 0) || (x >= memcanvas->mem_width) ) continue;
	if ( (y < 0) || (y >= memcanvas->mem_height) ) continue;
	offset = (y * memcanvas->mem_width + x) * memcanvas->stride;
	memcanvas->red[offset] = rgb.red;
	memcanvas->green[offset] = rgb.green;
	memcanvas->blue[offset] = rgb.blue;
    }
    return (TRUE);
}   /*  End Function draw_points  */


/*  Optional hook functions follow  */

static flag draw_pc_image (MemCanvas memcanvas, int x_off, int y_off,
			   int x_pixels, int y_pixels, CONST char *slice,
			   CONST uaddr *hoffsets, CONST uaddr *voffsets,
			   unsigned int width, unsigned int height,
			   unsigned int type, unsigned int conv_type,
			   unsigned int num_pixels,unsigned long *pixel_values,
			   unsigned long blank_pixel,
			   unsigned long min_sat_pixel,
			   unsigned long max_sat_pixel,
			   double i_min, double i_max,
			   flag (*iscale_func) (), void *iscale_info,
			   KPixCanvasImageCache *cache_ptr)
/*  [PURPOSE] This routine will draw a 2-dimensional slice of a Karma array
    into a memory canvas. This slice may be tiled. The slice is a
    PseudoColour image.
    <memcanvas> The memory canvas.
    <x_off> The horizontal offset, relative to the top-left corner of the
    canvas.
    <y_off> The vertical offset, relative to the top-left corner of the canvas.
    <x_pixels> The number of horizontal pixels to draw.
    <y_pixels> The number of vertical pixels to draw.
    <slice> The start of the slice image data.
    <hoffsets> The array of horizontal byte offsets.
    <voffsets> The array of vertical byte offsets.
    <width> The width of the input image (in values).
    <height> The height of the input image (in values).
    <type> The type of the slice image data.
    <conv_type> The input conversion type (when the input is complex).
    <num_pixels> The number of pixels in the pixel array.
    <pixel_values> The array of pixel values.
    <blank_pixel> The pixel value to be used when the intensity value is an
    undefined value.
    <min_sat_pixel> The pixel value to be used when the intensity value is
    below the minimum value.
    <max_sat_pixel> The pixel value to be used when the intensity value is
    above the maximum value.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <iscale_func> The function to be called when non-linear intensity scaling
    is required. If NULL, linear intensity scaling is used. The interface to
    this function is as follows:
    [<pre>]
    flag iscale_func (double *out, unsigned int out_stride,
                      double *inp, unsigned int inp_stride,
		      unsigned int num_values, double i_min, double i_max,
		      void *info)
    *   [PURPOSE] This routine will perform an arbitrary intensity scaling on
        an array of values. This routine may be called many times to scale an
	image.
        <out> The output array.
	<out_stride> The stride (in doubles) of the output array.
	<inp> The input array.
	<inp_stride> The stride (in doubles) of the input array.
	<num_values> The number of values to scale.
	<i_min> The minimum intensity value.
	<i_max> The maximum intensity value.
	<info> A pointer to arbitrary information.
	[RETURNS] TRUE on success, else FALSE.
    *
    [</pre>]
    <iscale_info> A pointer to arbitrary information for <<iscale_func>>.
    <cache_ptr> Cache data is written here. The routine may produce cache data
    which will vastly increase the speed of subsequent operations on this data.
    Prior to process exit, a call MUST be made to <<kwin_free_cache_data>>,
    otherwise shared memory segments could remain after the process exits.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag retval;
    iaddr offset;
    int pix_red_shift, pix_green_shift, pix_blue_shift;
    int red_bits, green_bits, blue_bits;
    unsigned int count, num_colours, visual;
    unsigned long pix_red_mask, pix_green_mask, pix_blue_mask;
    unsigned long tmp_mask, red, green, blue;
    uaddr red_offset, green_offset, blue_offset;
    unsigned short *cmap_reds, *cmap_greens, *cmap_blues;
    unsigned long *cmap_pixels;
    static char function_name[] = "_kwin_mem_draw_pc_image";

    VERIFY_CANVAS (memcanvas);
    if ( !ds_element_is_atomic (type) )
    {
	fprintf (stderr, "Element must be atomic\n");
	a_prog_bug (function_name);
    }
    /*  First copy pixel values into a temporary array which also includes
	the special pixels  */
    num_colours = num_pixels + 3;
    if ( ( cmap_pixels = m_alloc (num_colours * sizeof *cmap_pixels) )
	 == NULL )
    {
	m_error_notify (function_name, "temporary pixel array");
	return (FALSE);
    }
    cmap_pixels[0] = 1;  /*  Test endianness  */
    if (*(unsigned char *) cmap_pixels == 0)  /*  Big endian  */
    {
	red_offset = sizeof *cmap_pixels - 1;
	green_offset = sizeof *cmap_pixels - 2;
	blue_offset = sizeof *cmap_pixels - 3;
    }
    else  /*  Little endian  */
    {
	red_offset = 0;
	green_offset = 1;
	blue_offset = 2;
    }
    for (count = 0; count < num_pixels; ++count)
	cmap_pixels[count] = pixel_values[count];
    cmap_pixels[num_pixels] = min_sat_pixel;
    cmap_pixels[num_pixels + 1] = max_sat_pixel;
    cmap_pixels[num_pixels + 2] = blank_pixel;
    kwin_get_attributes (memcanvas->pixcanvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    switch (visual)
    {
      case KWIN_VISUAL_PSEUDOCOLOUR:
      case KWIN_VISUAL_GRAYSCALE:
      case KWIN_VISUAL_STATICCOLOUR:
      case KWIN_VISUAL_STATICGRAY:
	if ( ( cmap_reds = m_alloc (num_colours * sizeof *cmap_reds * 3) )
	     == NULL )
	{
	    m_free (cmap_pixels);
	    m_error_notify (function_name, "temporary colourmap");
	    return (FALSE);
	}
	cmap_greens = cmap_reds + num_colours;
	cmap_blues = cmap_greens + num_colours;
	if ( !kwin_get_pixel_RGB_values (memcanvas->pixcanvas, cmap_pixels,
					 cmap_reds, cmap_greens, cmap_blues,
					 num_colours) )
	{
	    m_free (cmap_pixels);
	    m_free (cmap_reds);
	    fprintf (stderr, "%s: cannot find RGB values\n", function_name);
	    return (FALSE);
	}
	/*  Convert 16bit RGB values to 8bit and put back into pixel values  */
	for (count = 0; count < num_colours; ++count)
	{
	    red = (cmap_reds[count] >> 8) & 0xff;
	    green = (cmap_greens[count] >> 8) & 0xff;
	    blue = (cmap_blues[count] >> 8) & 0xff;
	    cmap_pixels[count] = red | (green << 8) | (blue << 16);
	}
	m_free (cmap_reds);
	break;
      case KWIN_VISUAL_TRUECOLOUR:
	kwin_get_attributes (memcanvas->pixcanvas,
			     KWIN_ATT_PIX_RED_MASK, &pix_red_mask,
			     KWIN_ATT_PIX_GREEN_MASK, &pix_green_mask,
			     KWIN_ATT_PIX_BLUE_MASK, &pix_blue_mask,
			     KWIN_ATT_END);
	/*  Find size of each mask and position of LSb  */
	for (count = 0, tmp_mask = 1,
		 red_bits = 0, green_bits = 0, blue_bits = 0,
		 pix_red_shift = -1, pix_green_shift = -1, pix_blue_shift = -1;
	     count < sizeof tmp_mask * 8;
	     ++count, tmp_mask = tmp_mask << 1)
	{
	    if (pix_red_mask & tmp_mask)
	    {
		++red_bits;
		if (pix_red_shift < 0) pix_red_shift = count;
	    }
	    if (pix_green_mask & tmp_mask)
	    {
		++green_bits;
		if (pix_green_shift < 0) pix_green_shift = count;
	    }
	    if (pix_blue_mask & tmp_mask)
	    {
		++blue_bits;
		if (pix_blue_shift < 0) pix_blue_shift = count;
	    }
	}
	/*  Convert pixel RGB values to 8bit and put back into pixel values  */
	for (count = 0; count < num_colours; ++count)
	{
	    int shift;

	    /*  First construct byte values which are left-shifted  */
	    shift = 8 - red_bits - pix_red_shift;
	    red = cmap_pixels[count] & pix_red_mask;
	    if (shift < 0) red = red >> -shift;
	    else red = red << shift;
	    shift = 8 - green_bits - pix_green_shift;
	    green = cmap_pixels[count] & pix_green_mask;
	    if (shift < 0) green = green >> -shift;
	    else green = green << shift;
	    shift = 8 - blue_bits - pix_blue_shift;
	    blue = cmap_pixels[count] & pix_blue_mask;
	    if (shift < 0) blue = blue >> -shift;
	    else blue = blue << shift;
	    cmap_pixels[count] = red | (green << 8) | (blue << 16);
	}
	break;
    }
    min_sat_pixel = cmap_pixels[num_pixels];
    max_sat_pixel = cmap_pixels[num_pixels + 1];
    blank_pixel = cmap_pixels[num_pixels + 2];
    y_off = memcanvas->mem_height - y_off - y_pixels;  /*  Flip  */
    offset = (x_off + y_off * memcanvas->mem_width) * memcanvas->stride;
    retval = imw_scmap_to24_lossy (memcanvas->red + offset,
				   memcanvas->green + offset,
				   memcanvas->blue + offset,
				   TRUE, FALSE, memcanvas->stride,
				   memcanvas->stride * memcanvas->mem_width,
				   x_pixels, y_pixels,
				   slice,
				   (CONST iaddr *) hoffsets,
				   (CONST iaddr *) voffsets,
				   width, height,
				   type, conv_type,
				   num_pixels, cmap_pixels,
				   red_offset, green_offset, blue_offset,
				   blank_pixel, min_sat_pixel, max_sat_pixel,
				   i_min, i_max, iscale_func, iscale_info);
    m_free (cmap_pixels);
    return (retval);
}   /*  End Function draw_pc_image  */

static flag draw_rgb_image (MemCanvas memcanvas,
			    int x_off, int y_off, int x_pixels, int y_pixels,
			    CONST unsigned char *red_slice,
			    CONST unsigned char *green_slice,
			    CONST unsigned char *blue_slice,
			    CONST uaddr *hoffsets, CONST uaddr *voffsets,
			    unsigned int width, unsigned int height,
			    KPixCanvasImageCache *cache_ptr)
/*  [PURPOSE] This routine will draw a 2-dimensional slice of a Karma array
    onto a memory canvas. This slice may be tiled. The slice is a RGB image
    <memcanvas> The memory canvas.
    <x_off> The horizontal offset, relative to the top-left corner of the
    canvas.
    <y_off> The vertical offset, relative to the top-left corner of the canvas.
    <x_pixels> The number of horizontal pixels to draw.
    <y_pixels> The number of vertical pixels to draw.
    <red_slice> The start of the red slice data.
    <green_slice> The start of the green slice data.
    <blue_slice> The start of the blue slice data.
    [NOTE] The 3 colour components must be of type  K_UBYTE  .
    <hoffsets> The array of horizontal byte offsets.
    <voffsets> The array of vertical byte offsets.
    <width> The width of the input image (in values).
    <height> The height of the input image (in values).
    <cache_ptr> Cache data is written here. The routine may produce cache data
    which will vastly increase the speed of subsequent operations on this data.
    Prior to process exit, a call MUST be made to <<kwin_free_cache_data>>,
    otherwise shared memory segments could remain after the process exits.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    MagnifyFunc magnify_func;
    iaddr offset;
    unsigned int count, visual;
    unsigned char imap_red[256], imap_green[256], imap_blue[256];
    static char function_name[] = "_kwin_mem_draw_rgb_image";

    VERIFY_CANVAS (memcanvas);
    kwin_get_attributes (memcanvas->pixcanvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    y_off = memcanvas->mem_height - y_off - y_pixels;  /*  Flip  */
    offset = (x_off + y_off * memcanvas->mem_width) * memcanvas->stride;
    if (visual == KWIN_VISUAL_DIRECTCOLOUR)
    {
	unsigned char red, green, blue;
	unsigned long red_mask, green_mask, blue_mask;
	float red_max, green_max, blue_max, col_scale;
	unsigned short cmap_reds[256], cmap_greens[256], cmap_blues[256];
	unsigned long cmap_pixels[256];

	/*  Is DirectColour visual: need to worry about the state of
	    the colourmap. Note that even if the visual is PseudoColour,
	    there is no need to worry about the colourmap, since in this
	    case the RGB image is not really destined for a PseudoColour
	    visual, but the <viewimg> package is drawing the RGB image
	    instead of the PseudoColour image because PostScript is being
	    generated  */
	kwin_get_attributes (memcanvas->pixcanvas,
			     KWIN_ATT_PIX_RED_MASK, &red_mask,
			     KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			     KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			     KWIN_ATT_END);
	red_max = red_mask;
	green_max = green_mask;
	blue_max = blue_mask;
	for (count = 0; count < 256; ++count)
	{
	    /*  Have to construct pixel value from each mask  */
	    col_scale = (float) count / 256.0;
	    cmap_pixels[count] = ( (unsigned long) (red_max * col_scale) &
				   red_mask ) |
		( (unsigned long) (green_max * col_scale) & green_mask ) |
		( (unsigned long) (blue_max * col_scale) & blue_mask );
	}
	if ( !kwin_get_pixel_RGB_values (memcanvas->pixcanvas, cmap_pixels,
					 cmap_reds, cmap_greens, cmap_blues,
					 256) )
	{
	    fprintf (stderr, "%s: cannot find RGB values\n", function_name);
	    return (FALSE);
	}
	/*  Convert 16bit RGB values to 8bit  */
	for (count = 0; count < 256; ++count)
	{
	    red = (int) cmap_reds[count] >> 8;
	    green = (int) cmap_greens[count] >> 8;
	    blue = (int) cmap_blues[count] >> 8;
	    imap_red[count] = red;
	    imap_green[count] = green;
	    imap_blue[count] = blue;
	}
    }
    else
    {
	/*  Set up default colourmap  */
	for (count = 0; count < 256; ++count)
	{
	    imap_red[count] = count;
	    imap_green[count] = count;
	    imap_blue[count] = count;
	}
    }
    kwin_get_attributes (memcanvas->pixcanvas,
			 KWIN_ATT_MAGNIFY_FUNC, &magnify_func,
			 KWIN_ATT_END);
    if ( !imw_to8_lossy2 (memcanvas->red + offset, FALSE,
			  memcanvas->stride,
			  memcanvas->stride * memcanvas->mem_width,
			  x_pixels, y_pixels, (CONST char *) red_slice,
			  (iaddr *) hoffsets, (iaddr *) voffsets,
			  width, height, K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			  256, imap_red, 0, 0, 255, 0.0, 255.0,
			  ( flag (*) () ) NULL, NULL, magnify_func) )
    {
	fprintf (stderr, "Error drawing red image into X11Canvas\n");
	return (FALSE);
    }
    if ( !imw_to8_lossy2 (memcanvas->green + offset, FALSE,
			  memcanvas->stride,
			  memcanvas->stride * memcanvas->mem_width,
			  x_pixels, y_pixels, (CONST char *) green_slice,
			  (iaddr *) hoffsets, (iaddr *) voffsets,
			  width, height, K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			  256, imap_green, 0, 0, 255, 0.0, 255.0,
			  ( flag (*) () ) NULL, NULL, magnify_func) )
    {
	fprintf (stderr, "Error drawing green image into X11Canvas\n");
	return (FALSE);
    }
    if ( !imw_to8_lossy2 (memcanvas->blue + offset, FALSE,
			  memcanvas->stride,
			  memcanvas->stride * memcanvas->mem_width,
			  x_pixels, y_pixels, (CONST char *) blue_slice,
			  (iaddr *) hoffsets, (iaddr *) voffsets,
			  width, height, K_UBYTE, KIMAGE_COMPLEX_CONV_REAL,
			  256, imap_blue, 0, 0, 255, 0.0, 255.0,
			  ( flag (*) () ) NULL, NULL, magnify_func) )
    {
	fprintf (stderr, "Error drawing blue image into X11Canvas\n");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function draw_rgb_image  */

static flag draw_line (MemCanvas memcanvas,
		       double x0, double y0, double x1, double y1,
		       unsigned long pixel_value)
/*  [PURPOSE] This routine will draw a line on a memory canvas.
    <memcanvas> The memory canvas.
    <x0> The horizontal offset of the first point.
    <y0> The vertical offset of the first point.
    <x1> The horizontal offset of the second point.
    <y1> The vertical offset of the second point.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_kwin_mem_draw_line";

    VERIFY_CANVAS (memcanvas);
    return kwin_rasterise_draw_line (memcanvas, draw_points,
				     x0, y0, x1, y1, pixel_value);
}   /*  End Function draw_line  */

static flag draw_arc (MemCanvas memcanvas,
		      double centre_x, double centre_y,
		      double radius_x, double radius_y,
		      int angle1, int angle2, int rotation,
		      unsigned long pixel_value, flag fill)
/*  [SUMMARY] Draw an arc on a memory canvas.
    <memcanvas> The memory canvas.
    <centre_x> The horizontal centre of the arc.
    <centre_y> The vertical centre of the arc.
    <radius_x> The horizontal radius of the arc.
    <radius_y> The vertical radius of the arc.
    <angle1> See <<XDrawArc>>.
    <angle2> See <<XDrawArc>>.
    <rotation> The rotation of the arc, in units of 1/64 degree. A rotation of
    0 degrees corresponds to (x > 0, y = 0). A rotation of 90 degrees
    corresponds to (x = 0, y > 0). In other words, the angle increases
    anticlockwise from the right.
    <pixel_value> The pixel value to use.
    <fill> If TRUE, the arc is filled, else only the outside is drawn.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_kwin_mem_draw_arc";

    VERIFY_CANVAS (memcanvas);
    FLAG_VERIFY (fill);
    return kwin_rasterise_draw_arc (memcanvas, draw_points, NULL,
				    centre_x, centre_y, radius_x, radius_y,
				    angle1, angle2, rotation,
				    pixel_value, fill);
}   /*  End Function draw_arc  */

static flag set_linewidth (MemCanvas memcanvas, double linewidth)
/*  [SUMMARY] Set the linewidth for a canvas.
    <memcanvas> The memory canvas.
    <linewidth> The linewidth, in pixels.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_kwin_mem_set_linewidth";

    VERIFY_CANVAS (memcanvas);
    if (linewidth >= 2.0)
	fprintf (stderr, "%s: function not yet implemented: ignoring\n",
		 function_name);
    memcanvas->linewidth = linewidth;
    return (TRUE);
}   /*  End Function set_linewidth  */

static flag set_line_dash (MemCanvas memcanvas, flag dash)
/*  [SUMMARY] Set the line dash for a canvas.
    <memcanvas> The memory canvas.
    <dash> If TRUE, the line will be dashed.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_kwin_mem_set_line_dash";

    VERIFY_CANVAS (memcanvas);
    if (dash) fprintf (stderr, "%s: function not yet implemented: ignoring\n",
		       function_name);
    return (TRUE);
}   /*  End Function set_line_dash  */


/*  Private functions follow  */

static void get_rgb_of_pixel (MemCanvas memcanvas, rgb_type *rgb,
			      unsigned long pixel_value)
/*  [SUMMARY] Get the RGB value of a pixel value.
    <memcanvas> The memory canvas.
    <rgb> The RGB value is written here.
    <pixel_value> The pixel value.
    [RETURNS] Nothing.
*/
{
    if ( !memcanvas->pixel_valid || (pixel_value != memcanvas->pixel_value) )
    {
	unsigned short red, green, blue;

	kwin_get_pixel_RGB_values (memcanvas->pixcanvas, &pixel_value,
				   &red, &green, &blue, 1);
	memcanvas->rgb.red = red >> 8;
	memcanvas->rgb.green = green >> 8;
	memcanvas->rgb.blue = blue >> 8;
	memcanvas->pixel_value = pixel_value;
	memcanvas->pixel_valid = TRUE;
    }
    *rgb = memcanvas->rgb;
}   /*  End Function get_rgb_of_pixel  */
