/*LINTLIBRARY*/
/*  ps.c

    This code provides the PostScript driver for KPixCanvas objects.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating a simple pixel
    canvas (window) independent of the graphics system in use. This file
    contains the PostScript code.


    Written by      Richard Gooch   25-AUG-1996: Moved embedded PostScript
  functionality in generic.c here and made use of <kwin_new_driver_refresh>.

    Updated by      Richard Gooch   14-SEP-1996: Created <draw_point> and
  <set_linewidth>.

    Updated by      Richard Gooch   10-OCT-1996: Support huge pixel arrays in
  <draw_pc_image>.

    Updated by      Richard Gooch   22-NOV-1996: Fixed bug in <draw_rgb_image>
  when drawing onto a PseudoColour canvas. Fixed bug in <draw_pc_image> when
  getting RGB values from pixel values.

    Updated by      Richard Gooch   16-MAR-1997: Fixed bug in <draw_arc> and
  took account of interface change.

    Updated by      Richard Gooch   17-MAR-1997: Now call <psw_rgb_arc> in
  <draw_arc>.

    Updated by      Richard Gooch   4-JUN-1997: Fix in <draw_pc_image> for
  depth != 24. Still need to do depth==24 properly.

    Updated by      Richard Gooch   25-JUL-1997: Took account of changes to
  KPixDrawFuncs structure.

    Updated by      Richard Gooch   11-OCT-1997: Added preliminary line dash
  support.

    Updated by      Richard Gooch   14-APR-1997: Took acount of change in
  rotation definition on 6-FEB-1998.

    Updated by      Richard Gooch   19-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   13-SEP-1998: Cope with different mask
  widths when drawing to TrueColour in <draw_pc_image>.

    Updated by      Richard Gooch   16-APR-1999: Created <draw_segments> and
  <get_rgb_of_pixel> to speed up drawing.

    Updated by      Richard Gooch   30-JUN-1999: Added crude text support.

    Updated by      Richard Gooch   10-AUG-1999: Removed crude text support, as
  a slightly better method is now available from the <overlay> package.

    Updated by      Richard Gooch   10-SEP-1999: Added call to function
  <s_check_for_int> to <draw_pc_image>.

    Updated by      Richard Gooch   13-MAY-2000: Took account of interface
  change to <kwin_new_driver_refresh>. Switch <draw_point> to <draw_points>
  method.

    Updated by      Richard Gooch   27-MAY-2000: Took account of interface
  change to <kwin_new_driver_refresh>.

    Updated by      Richard Gooch   11-MAR-2001: Cope with min>max after iscale
  transform.

    Last updated by Richard Gooch   19-JUN-2001: Documentation fix for
  <draw_arc>.


*/
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_kwin.h>
#include <karma_psw.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include <karma_s.h>


#define KPixHookCanvas PSCanvas

typedef struct
{
    double red;
    double green;
    double blue;
} rgb_type;

typedef struct PScanvas_type
{
    unsigned int magic_number;
    PostScriptPage pspage;
    KPixCanvas pixcanvas;
    flag pixel_valid;
    unsigned long pixel_value;
    rgb_type rgb;
} *PSCanvas;

#include <karma_kwin_hooks.h>


#define MAX_INTENSITY 65535.0
#define CANVAS_MAGIC_NUMBER 893453289
#define VERIFY_CANVAS(canvas) if (canvas == NULL) \
{fprintf (stderr, "NULL canvas passed\n"); \
 a_prog_bug (function_name); } \
if (canvas->magic_number != CANVAS_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas object\n"); \
 a_prog_bug (function_name); }

/*  Structure declarations  */


/*  Private data  */


/*  Mandatory functions  */
STATIC_FUNCTION (flag draw_points, (PSCanvas pscanvas,
				    double *x_array, double *y_array,
				    unsigned int num_points,
				    unsigned long pixel_value) );
/*  Optional hook functions  */
STATIC_FUNCTION (flag draw_pc_image,
		 (PSCanvas pscanvas,
		  int x_off, int y_off, int x_pixels, int y_pixels,
		  CONST char *slice,
		  CONST uaddr *hoffsets, CONST uaddr *voffsets,
		  unsigned int width, unsigned int height,
		  unsigned int type, unsigned int conv_type,
		  unsigned int num_pixels,unsigned long *pixel_values,
		  unsigned long blank_pixel,
		  unsigned long min_sat_pixel, unsigned long max_sat_pixel,
		  double i_min, double i_max,
		  flag (*iscale_func) (), void *iscale_info,
		  KPixCanvasImageCache *cache_ptr) );
STATIC_FUNCTION (flag draw_rgb_image,
		 (PSCanvas pscanvas,
		  int x_off, int y_off, int x_pixels, int y_pixels,
		  CONST unsigned char *red_slice,
		  CONST unsigned char *green_slice,
		  CONST unsigned char *blue_slice,
		  CONST uaddr *hoffsets, CONST uaddr *voffsets,
		  unsigned int width, unsigned int height,
		  KPixCanvasImageCache *cache_ptr) );
STATIC_FUNCTION (flag draw_line, (PSCanvas pscanvas,
				  double x0, double y0, double x1, double y1,
				  unsigned long pixel_value) );
STATIC_FUNCTION (flag draw_arc,
		 (PSCanvas pscanvas,
		  double centre_x, double centre_y,
		  double radius_x, double radius_y,
		  int angle1, int angle2, int rotation,
		  unsigned long pixel_value, flag fill) );
STATIC_FUNCTION (flag draw_polygon,
		 (PSCanvas pscanvas, double *x_arr, double *y_arr,
		  unsigned int num_vertices, unsigned long pixel_value,
		  flag convex, flag fill) );
#if 0
STATIC_FUNCTION (flag draw_string,
		 (PSCanvas pscanvas, double x, double y, CONST char *string,
		  unsigned long pixel_value, flag clear_under) );
#endif
STATIC_FUNCTION (flag draw_segments,
		 (PSCanvas pscanvas,
		  double *x0, double *y0, double *x1, double *y1,
		  unsigned int num_segments, unsigned long pixel_value) );
STATIC_FUNCTION (flag set_linewidth, (PSCanvas pscanvas, double linewidth) );
STATIC_FUNCTION (flag set_line_dash, (PSCanvas pscanvas, flag dash) );


/*  Private functions  */
STATIC_FUNCTION (void get_rgb_of_pixel, (PSCanvas pscanvas, rgb_type *rgb,
					 unsigned long pixel_value) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
flag kwin_write_ps (KPixCanvas canvas, PostScriptPage pspage)
/*  [SUMMARY] Refresh a pixel canvas onto a PostScriptPage object.
    [PURPOSE] This routine will refresh a pixel canvas, redirecting output to a
    PostScriptPage object.
    <canvas> The pixel canvas.
    <pspage> The PostScriptPage object.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok, dash;
    int canvas_width, canvas_height;
    double linewidth, d_tmp;
    struct PScanvas_type pscanvas;
    static char function_name[] = "kwin_write_ps";

    if (pspage == NULL)
    {
	fprintf (stderr, "NULL PostScriptPage object passed\n");
	a_prog_bug (function_name);
    }
    pscanvas.magic_number = CANVAS_MAGIC_NUMBER;
    pscanvas.pspage = pspage;
    pscanvas.pixcanvas = canvas;
    pscanvas.pixel_valid = FALSE;
    kwin_get_attributes (canvas,
			 KWIN_ATT_LINEWIDTH, &linewidth,
			 KWIN_ATT_LINE_DASH, &dash,
			 KWIN_ATT_END);
    kwin_get_size (canvas, &canvas_width, &canvas_height);
    if (linewidth > 0.0)
    {
	d_tmp = (canvas_width * canvas_width) + (canvas_height *canvas_height);
	d_tmp = sqrt (d_tmp) / 1.414213562;
	psw_set_attributes (pspage,
			    PSW_ATT_LINEWIDTH_RELATIVE, linewidth / d_tmp,
			    PSW_ATT_LINE_DASH, dash,
			    PSW_ATT_END);
    }
    psw_set_attributes (pspage,
			PSW_ATT_LINE_DASH, dash,
			PSW_ATT_END);
    ok = kwin_new_driver_refresh (canvas, pspage, 0.0, &pscanvas,
				  KWIN_FUNC_DRAW_POINTS, draw_points,
				  KWIN_FUNC_DRAW_PC_IMAGE, draw_pc_image,
				  KWIN_FUNC_DRAW_RGB_IMAGE, draw_rgb_image,
				  KWIN_FUNC_DRAW_LINE, draw_line,
				  KWIN_FUNC_DRAW_ARC, draw_arc,
				  KWIN_FUNC_DRAW_POLYGON, draw_polygon,
#if 0
				  KWIN_FUNC_DRAW_STRING, draw_string,
#endif
				  KWIN_FUNC_DRAW_SEGMENTS, draw_segments,
				  KWIN_FUNC_SET_LINEWIDTH, set_linewidth,
				  KWIN_FUNC_SET_LINE_DASH, set_line_dash,
				  KWIN_ATT_END);
    m_clear (&pscanvas, sizeof pscanvas);
    return (ok);
}   /*  End Function kwin_write_ps  */


/*  Mandatory functions follow  */

static flag draw_points (PSCanvas pscanvas, double *x_array, double *y_array,
			 unsigned int num_points, unsigned long pixel_value)
/*  [PURPOSE] This routine will draw a point on a PostScript canvas.
    <pscanvas> The PostScript canvas.
    <x_array> The horizontal co-ordinates of the points.
    <y_array> The vertical co-ordinates of the points.
    <num_points> The number of points.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int canvas_width, canvas_height;
    unsigned int count;
    rgb_type rgb;
    static char function_name[] = "_kwin_ps_draw_points";

    VERIFY_CANVAS (pscanvas);
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    get_rgb_of_pixel (pscanvas, &rgb, pixel_value);
    for (count = 0; count < num_points; ++count)
    {
	/*  Flip vertical  */
	if ( !(*psw_rgb_ellipse)
	     (pscanvas->pspage, rgb.red, rgb.green, rgb.blue,
	      x_array[count] / (double) (canvas_width - 1),
	      (canvas_height - 1.0 -
	       y_array[count]) / (double) (canvas_height - 1),
	       1e-2 / (double) canvas_width,
	       1e-2 / (double) canvas_height, TRUE) )
	    return (FALSE);
    }
    return (TRUE);
}   /*  End Function draw_points  */


/*  Optional hook functions follow  */

static flag draw_pc_image (PSCanvas pscanvas, int x_off, int y_off,
			   int x_pixels, int y_pixels, CONST char *slice,
			   CONST uaddr *hoffsets, CONST uaddr *voffsets,
			   unsigned int width, unsigned int height,
			   unsigned int type, unsigned int conv_type,
			   unsigned int num_pixels,unsigned long *pixel_values,
			   unsigned long blank_pixel,
			   unsigned long min_sat_pixel,
			   unsigned long max_sat_pixel,
			   double i_min, double i_max,
			   flag (*iscale_func) (), void *iscale_info,
			   KPixCanvasImageCache *cache_ptr)
/*  [PURPOSE] This routine will draw a 2-dimensional slice of a Karma array
    onto a PostScript canvas. This slice may be tiled. The slice is a
    PseudoColour image.
    <pscanvas> The PostScript canvas.
    <x_off> The horizontal offset, relative to the top-left corner of the
    canvas.
    <y_off> The vertical offset, relative to the top-left corner of the canvas.
    <x_pixels> The number of horizontal pixels to draw.
    <y_pixels> The number of vertical pixels to draw.
    <slice> The start of the slice image data.
    <hoffsets> The array of horizontal byte offsets.
    <voffsets> The array of vertical byte offsets.
    <width> The width of the input image (in values).
    <height> The height of the input image (in values).
    <type> The type of the slice image data.
    <conv_type> The input conversion type (when the input is complex).
    <num_pixels> The number of pixels in the pixel array.
    <pixel_values> The array of pixel values.
    <blank_pixel> The pixel value to be used when the intensity value is an
    undefined value.
    <min_sat_pixel> The pixel value to be used when the intensity value is
    below the minimum value.
    <max_sat_pixel> The pixel value to be used when the intensity value is
    above the maximum value.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <iscale_func> The function to be called when non-linear intensity scaling
    is required. If NULL, linear intensity scaling is used. The interface to
    this function is as follows:
    [<pre>]
    flag iscale_func (double *out, unsigned int out_stride,
                      double *inp, unsigned int inp_stride,
		      unsigned int num_values, double i_min, double i_max,
		      void *info)
    *   [PURPOSE] This routine will perform an arbitrary intensity scaling on
        an array of values. This routine may be called many times to scale an
	image.
        <out> The output array.
	<out_stride> The stride (in doubles) of the output array.
	<inp> The input array.
	<inp_stride> The stride (in doubles) of the input array.
	<num_values> The number of values to scale.
	<i_min> The minimum intensity value.
	<i_max> The maximum intensity value.
	<info> A pointer to arbitrary information.
	[RETURNS] TRUE on success, else FALSE.
    *
    [</pre>]
    <iscale_info> A pointer to arbitrary information for <<iscale_func>>.
    <cache_ptr> Cache data is written here. The routine may produce cache data
    which will vastly increase the speed of subsequent operations on this data.
    Prior to process exit, a call MUST be made to <<kwin_free_cache_data>>,
    otherwise shared memory segments could remain after the process exits.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag complex, retval;
    flag greyscale = FALSE;
    int canvas_width, canvas_height, index, shift;
    int pix_red_shift, pix_green_shift, pix_blue_shift;
    int im_red_shift, im_green_shift, im_blue_shift;
    unsigned char red_bits, green_bits, blue_bits, min_bits;
    unsigned int x, y;
    unsigned int count, num_colours, visual, depth;
    unsigned long pix_red_mask, pix_green_mask, pix_blue_mask;
    unsigned long im_red_mask, im_green_mask, im_blue_mask;
    unsigned long tmp_mask, red, green, blue;
    unsigned long grey_mask = 0xff;
    unsigned char *ubimage;
    double y0, y1;
    double val, factor;
    double scaled_min, scaled_max, tmp;
    unsigned short *cmap_reds, *cmap_greens, *cmap_blues, *cmap_greys;
    unsigned long *cmap_pixels;
    double *values, *val_ptr;
    static char function_name[] = "_kwin_ps_draw_pc_image";

    VERIFY_CANVAS (pscanvas);
    if ( !ds_element_is_atomic (type) )
    {
	fprintf (stderr, "Element must be atomic\n");
	a_prog_bug (function_name);
    }
    /*  First copy pixel values into a temporary array which also includes
	the special pixels  */
    num_colours = num_pixels + 3;
    if ( ( cmap_pixels = m_alloc (num_colours * sizeof *cmap_pixels) )
	 == NULL )
    {
	m_error_notify (function_name, "temporary pixel array");
	return (FALSE);
    }
    for (count = 0; count < num_pixels; ++count)
    {
	cmap_pixels[count] = pixel_values[count];
    }
    cmap_pixels[num_pixels] = min_sat_pixel;
    cmap_pixels[num_pixels + 1] = max_sat_pixel;
    cmap_pixels[num_pixels + 2] = blank_pixel;
    if ( ( cmap_reds = m_alloc (num_colours * sizeof *cmap_reds * 3) )
	 == NULL )
    {
	m_free (cmap_pixels);
	m_error_notify (function_name, "temporary colourmap");
	return (FALSE);
    }
    cmap_greens = cmap_reds + num_colours;
    cmap_blues = cmap_greens + num_colours;
    cmap_greys = cmap_reds;
    kwin_get_attributes (pscanvas->pixcanvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_DEPTH, &depth,
			 KWIN_ATT_END);
    switch (visual)
    {
      case KWIN_VISUAL_PSEUDOCOLOUR:
      case KWIN_VISUAL_GRAYSCALE:
      case KWIN_VISUAL_STATICCOLOUR:
      case KWIN_VISUAL_STATICGRAY:
	if ( !kwin_get_pixel_RGB_values (pscanvas->pixcanvas, cmap_pixels,
					 cmap_reds, cmap_greens, cmap_blues,
					 num_colours) )
	{
	    m_free (cmap_pixels);
	    m_free (cmap_reds);
	    fprintf (stderr, "%s: cannot find RGB values\n", function_name);
	    return (FALSE);
	}
	/*  Convert 16bit RGB values to 8bit and check if GreyScale (i.e.
	    R=G=B)  */
	for (count = 0, greyscale = TRUE; count < num_colours; ++count)
	{
	    red = cmap_reds[count] >> 8;
	    green = cmap_greens[count] >> 8;
	    blue = cmap_blues[count] >> 8;
	    cmap_reds[count] = red;
	    cmap_greens[count] = green;
	    cmap_blues[count] = blue;
	    red &= grey_mask;
	    green &= grey_mask;
	    blue &= grey_mask;
	    if ( (red != green) || (red != blue) ) greyscale = FALSE;
	}
	break;
      case KWIN_VISUAL_TRUECOLOUR:
	kwin_get_attributes (pscanvas->pixcanvas,
			     KWIN_ATT_PIX_RED_MASK, &pix_red_mask,
			     KWIN_ATT_PIX_GREEN_MASK, &pix_green_mask,
			     KWIN_ATT_PIX_BLUE_MASK, &pix_blue_mask,
			     KWIN_ATT_END);
	/*  Find size of each mask and position of LSb  */
	for (count = 0, tmp_mask = 1,
		 red_bits = 0, green_bits = 0, blue_bits = 0,
		 pix_red_shift = -1, pix_green_shift = -1, pix_blue_shift = -1;
	     count < sizeof tmp_mask * 8;
	     ++count, tmp_mask = tmp_mask << 1)
	{
	    if (pix_red_mask & tmp_mask)
	    {
		++red_bits;
		if (pix_red_shift < 0) pix_red_shift = count;
	    }
	    if (pix_green_mask & tmp_mask)
	    {
		++green_bits;
		if (pix_green_shift < 0) pix_green_shift = count;
	    }
	    if (pix_blue_mask & tmp_mask)
	    {
		++blue_bits;
		if (pix_blue_shift < 0) pix_blue_shift = count;
	    }
	}
	min_bits = (red_bits < green_bits) ? red_bits : green_bits;
	if (blue_bits < min_bits) min_bits = blue_bits;
	if (green_bits > red_bits)
	{
	    cmap_greys = (blue_bits > green_bits) ? cmap_blues : cmap_greens;
	}
	else if (blue_bits > red_bits) cmap_greys = cmap_blues;
	for (count = 0, tmp_mask = 0x80, grey_mask = 0; count < min_bits;
	     ++count, tmp_mask = tmp_mask >> 1) grey_mask |= tmp_mask;
	/*  Check if GreyScale (i.e. R=G=B)  */
	for (count = 0, greyscale = TRUE; count < num_colours; ++count)
	{
	    /*  First construct byte values which are left-shifted  */
	    shift = 8 - red_bits - pix_red_shift;
	    red = cmap_pixels[count] & pix_red_mask;
	    if (shift < 0) red = red >> -shift;
	    else red = red << shift;
	    cmap_reds[count] = red;
	    shift = 8 - green_bits - pix_green_shift;
	    green = cmap_pixels[count] & pix_green_mask;
	    if (shift < 0) green = green >> -shift;
	    else green = green << shift;
	    cmap_greens[count] = green;
	    shift = 8 - blue_bits - pix_blue_shift;
	    blue = cmap_pixels[count] & pix_blue_mask;
	    if (shift < 0) blue = blue >> -shift;
	    else blue = blue << shift;
	    cmap_blues[count] = blue;
	    red &= grey_mask;
	    green &= grey_mask;
	    blue &= grey_mask;
	    if ( (red != green) || (red != blue) ) greyscale = FALSE;
	}
	if (depth != 24) break;
	/*  Note that the pixel values in the array are valid for drawing
	    images, not pixels. Because the <kwin_get_pixel_RGB_values>
	    routine expects regular pixel values (i.e. NOT image pixel values),
	    the array of pixel values needs to be fiddled  */
	/*  HACK: this is here because of the broken way <khuei> generates it's
	    pixel array, which in turn is due to the broken way the X11 PC
	    image draw is done. This should be fixed  */
	kwin_get_attributes (pscanvas->pixcanvas,
			     KWIN_ATT_IM_RED_MASK, &im_red_mask,
			     KWIN_ATT_IM_GREEN_MASK, &im_green_mask,
			     KWIN_ATT_IM_BLUE_MASK, &im_blue_mask,
			     KWIN_ATT_END);
	if (red_bits != 8)
	{
	    fprintf (stderr, "Shifted red_bits: %d is not 8\n", red_bits);
	    a_prog_bug (function_name);
	}
	if (green_bits != 8)
	{
	    fprintf (stderr, "Shifted green_bits: %d is not 8\n", green_bits);
	    a_prog_bug (function_name);
	}
	if (blue_bits != 8)
	{
	    fprintf (stderr, "Shifted blue_bits: %d is not 8\n", blue_bits);
	    a_prog_bug (function_name);
	}
	for (im_red_shift = 0, tmp_mask = im_red_mask; !(tmp_mask & 1);
	     tmp_mask = tmp_mask >> 1, ++im_red_shift);
	if (tmp_mask != 0xff)
	{
	    fprintf (stderr, "Shifted red_mask: %lx is not 0xff\n", tmp_mask);
	    a_prog_bug (function_name);
	}
	for (im_green_shift = 0, tmp_mask = im_green_mask; !(tmp_mask & 1);
	    tmp_mask = tmp_mask >> 1, ++im_green_shift);
	if (tmp_mask != 0xff)
	{
	    fprintf (stderr, "Shifted green_mask: %lx is not 0xff\n",tmp_mask);
	    a_prog_bug (function_name);
	}
	for (im_blue_shift = 0, tmp_mask = im_blue_mask; !(tmp_mask & 1);
	     tmp_mask = tmp_mask >> 1, ++im_blue_shift);
	if (tmp_mask != 0xff)
	{
	    fprintf (stderr, "Shifted blue_mask: %lx is not 0xff\n", tmp_mask);
	    a_prog_bug (function_name);
	}
	for (count = 0; count < num_colours; ++count)
	{
	    red = cmap_pixels[count];
	    green = red;
	    blue = red;
	    red = (red & im_red_mask) >> im_red_shift << pix_red_shift;
	    green = (green & im_green_mask) >>im_green_shift <<pix_green_shift;
	    blue = (blue & im_blue_mask) >> im_blue_shift << pix_blue_shift;
	    cmap_pixels[count] = red | green | blue;
	}
	break;
    }
    m_free (cmap_pixels);
    /*  Allocate temporary image  */
    if (greyscale)
    {
	/*  Just 1 byte per pixel is enough  */
	if ( ( ubimage = m_alloc (sizeof *ubimage * width * height) ) == NULL )
	{
	    m_free (cmap_reds);
	    m_error_notify (function_name, "ubarray");
	    return (FALSE);
	}
    }
    else
    {
	/*  Need 3 bytes per pixel  */
	if ( ( ubimage = m_alloc (sizeof *ubimage * width * height * 3) )
	     == NULL )
	{
	    m_free (cmap_reds);
	    m_error_notify (function_name, "ubarray");
	    return (FALSE);
	}
    }
    /*  Convert image  */
    if (iscale_func == NULL)
    {
	scaled_min = i_min;
	scaled_max = i_max;
    }
    else
    {
	if ( !(*iscale_func) (&scaled_min, 0, &i_min, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if ( !(*iscale_func) (&scaled_max, 0, &i_max, 0, 1, i_min, i_max,
			      iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if (scaled_max < scaled_min) swap (scaled_min, scaled_max, tmp);
    }
    factor = (double) (num_pixels - 1) / (scaled_max - scaled_min);
    if ( ( values = m_alloc (sizeof *values * 2 * width) ) == NULL )
    {
	m_error_notify (function_name, "values array");
	m_free (cmap_reds);
	m_free (ubimage);
	return (FALSE);
    }
    for (y = 0; y < height; ++y)
    {
	/*  Convert values to generic data type  */
	if ( !ds_get_scattered_elements (slice + voffsets[y], type,
					 hoffsets, values, &complex,
					 width) )
	{
	    fprintf (stderr, "Error converting data\n");
	    m_free (cmap_reds);
	    m_free (ubimage);
	    m_free (values);
	    return (FALSE);
	}
	if (complex) ds_complex_to_real_1D (values, 2, values, width,
					    conv_type);
	if (iscale_func != NULL)
	{
	    if ( !(*iscale_func) (values, 2, values, 2, width,
				  i_min, i_max, iscale_info) )
	    {
		fprintf (stderr, "Error scaling data\n");
		m_free (cmap_reds);
		m_free (ubimage);
		m_free (values);
		return (FALSE);
	    }
	}
	/*  Convert to unsigned bytes  */
	for (x = 0, val_ptr = values; x < width; ++x, val_ptr += 2)
	{
	    val = *val_ptr;
	    if ( DOUBLE_IS_BLANK (val) ) index = num_pixels + 2;
	    else if (val < scaled_min) index = num_pixels;
	    else if (val > scaled_max) index = num_pixels + 1;
	    else
	    {
		val = (val - scaled_min) * factor;
		index = (int) val;
	    }
	    if (greyscale) ubimage[y * width + x] = cmap_greys[index];
	    else
	    {
		ubimage[(y * width + x) * 3] = cmap_reds[index];
		ubimage[(y * width + x) * 3 + 1] = cmap_greens[index];
		ubimage[(y * width + x) * 3 + 2] = cmap_blues[index];
	    }
	}
	if ( s_check_for_int () ) 
	{
	    fprintf (stderr, "%s: interrupt detected\n", function_name);
	    m_free (cmap_reds);
	    m_free (ubimage);
	    m_free (values);
	    return (FALSE);
	}
    }
    m_free (values);
    m_free (cmap_reds);
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    /*  Flip  */
    y0 = (double) (canvas_height - 1 - y_off);
    y1 = (double) ( canvas_height - 1 - (y_off + y_pixels - 1) );
    if (greyscale)
    {
	retval = psw_mono_image
	    ( pscanvas->pspage, ubimage, width, height, NULL, NULL, NULL,
	      (double) x_off / (double) (canvas_width - 1),
	      (double) y1 / (double) (canvas_height - 1),
	      (double) (x_off + x_pixels - 1) / (double) (canvas_width - 1),
	      (double) y0 / (double) (canvas_height - 1) );
    }
    else
    {
	retval = psw_rgb_image
	    ( pscanvas->pspage, ubimage, ubimage + 1, ubimage + 2,
	      width, height, NULL, NULL, NULL, NULL, NULL, NULL, 3,
	      (double) x_off / (double) (canvas_width - 1),
	      (double) y1 / (double) (canvas_height - 1),
	      (double) (x_off + x_pixels - 1) / (double) (canvas_width - 1),
	      (double) y0 / (double) (canvas_height - 1) );
    }
    m_free (ubimage);
    return (retval);
}   /*  End Function draw_pc_image  */

static flag draw_rgb_image (PSCanvas pscanvas,
			    int x_off, int y_off, int x_pixels, int y_pixels,
			    CONST unsigned char *red_slice,
			    CONST unsigned char *green_slice,
			    CONST unsigned char *blue_slice,
			    CONST uaddr *hoffsets, CONST uaddr *voffsets,
			    unsigned int width, unsigned int height,
			    KPixCanvasImageCache *cache_ptr)
/*  [PURPOSE] This routine will draw a 2-dimensional slice of a Karma array
    onto a PostScript canvas. This slice may be tiled. The slice is a RGB image
    <pscanvas> The PostScript canvas.
    <x_off> The horizontal offset, relative to the top-left corner of the
    canvas.
    <y_off> The vertical offset, relative to the top-left corner of the canvas.
    <x_pixels> The number of horizontal pixels to draw.
    <y_pixels> The number of vertical pixels to draw.
    <red_slice> The start of the red slice data.
    <green_slice> The start of the green slice data.
    <blue_slice> The start of the blue slice data.
    [NOTE] The 3 colour components must be of type  K_UBYTE  .
    <hoffsets> The array of horizontal byte offsets.
    <voffsets> The array of vertical byte offsets.
    <width> The width of the input image (in values).
    <height> The height of the input image (in values).
    <cache_ptr> Cache data is written here. The routine may produce cache data
    which will vastly increase the speed of subsequent operations on this data.
    Prior to process exit, a call MUST be made to <<kwin_free_cache_data>>,
    otherwise shared memory segments could remain after the process exits.
    [MT-LEVEL] Unsafe.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok;
    int canvas_width, canvas_height;
    unsigned char red, green, blue;
    unsigned int count, visual;
    unsigned long red_mask, green_mask, blue_mask;
    float red_max, green_max, blue_max, col_scale;
    double y0, y1;
    unsigned char imap_red[256], imap_green[256], imap_blue[256];
    unsigned short cmap_reds[256], cmap_greens[256], cmap_blues[256];
    unsigned long cmap_pixels[256];
    static char function_name[] = "_kwin_ps_draw_rgb_image";

    VERIFY_CANVAS (pscanvas);
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    kwin_get_attributes (pscanvas->pixcanvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    /*  Flip  */
    y0 = (double) (canvas_height - 1 - y_off);
    y1 = (double) ( canvas_height - 1 - (y_off + y_pixels - 1) );
    if (visual != KWIN_VISUAL_DIRECTCOLOUR)
    {
	/*  Not DirectColour visual: no need to worry about the state of
	    the colourmap. Note that even if the visual is PseudoColour,
	    there is no need to worry about the colourmap, since in this
	    case the RGB image is not really destined for a PseudoColour
	    visual, but the <viewimg> package is drawing the RGB image
	    instead of the PseudoColour image because PostScript is being
	    generated  */
	ok = psw_rgb_image
	    ( pscanvas->pspage, red_slice, green_slice, blue_slice,
	      width, height,
	      hoffsets, voffsets, hoffsets, voffsets, hoffsets, voffsets, 0,
	      (double) x_off / (double) (canvas_width - 1),
	      (double) y1 / (double) (canvas_height - 1),
	      (double) (x_off + x_pixels - 1) / (double) (canvas_width - 1),
	      (double) y0 / (double) (canvas_height - 1) );
	return (ok);
    }
    /*  DirectColour visual: have to get colourmap values  */
    kwin_get_attributes (pscanvas->pixcanvas,
			 KWIN_ATT_PIX_RED_MASK, &red_mask,
			 KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			 KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			 KWIN_ATT_END);
    red_max = red_mask;
    green_max = green_mask;
    blue_max = blue_mask;
    for (count = 0; count < 256; ++count)
    {
	/*  Have to construct pixel value from each mask  */
	col_scale = (float) count / 256.0;
	cmap_pixels[count] = ( (unsigned long) (red_max * col_scale) &
			       red_mask ) |
	    ( (unsigned long) (green_max * col_scale) & green_mask ) |
	    ( (unsigned long) (blue_max * col_scale) & blue_mask );
    }
    if ( !kwin_get_pixel_RGB_values (pscanvas->pixcanvas, cmap_pixels,
				     cmap_reds, cmap_greens, cmap_blues,
				     256) )
    {
	fprintf (stderr, "%s: cannot find RGB values\n", function_name);
	return (FALSE);
    }
    /*  Convert 16bit RGB values to 8bit  */
    for (count = 0; count < 256; ++count)
    {
	red = (int) cmap_reds[count] >> 8;
	green = (int) cmap_greens[count] >> 8;
	blue = (int) cmap_blues[count] >> 8;
	imap_red[count] = red;
	imap_green[count] = green;
	imap_blue[count] = blue;
    }
    ok = psw_directcolour_image
	( pscanvas->pspage, red_slice, green_slice, blue_slice, width, height,
	  hoffsets, voffsets, hoffsets, voffsets, hoffsets, voffsets, 0,
	  imap_red, imap_green, imap_blue,
	  (double) x_off / (double) (canvas_width - 1),
	  (double) y1 / (double) (canvas_height - 1),
	  (double) (x_off + x_pixels - 1) / (double) (canvas_width - 1),
	  (double) y0 / (double) (canvas_height - 1) );
    return (ok);
}   /*  End Function draw_rgb_image  */

static flag draw_line (PSCanvas pscanvas,
		       double x0, double y0, double x1, double y1,
		       unsigned long pixel_value)
/*  [PURPOSE] This routine will draw a line on a PostScript canvas.
    <pscanvas> The PostScript canvas.
    <x0> The horizontal offset of the first point.
    <y0> The vertical offset of the first point.
    <x1> The horizontal offset of the second point.
    <y1> The vertical offset of the second point.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int canvas_width, canvas_height;
    rgb_type rgb;
    static char function_name[] = "_kwin_ps_draw_line";

    VERIFY_CANVAS (pscanvas);
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    get_rgb_of_pixel (pscanvas, &rgb, pixel_value);
    /*  Flip vertical  */
    y0 = (double) canvas_height - 1.0 - y0;
    y1 = (double) canvas_height - 1.0 - y1;
    return psw_rgb_line ( pscanvas->pspage, rgb.red, rgb.green, rgb.blue,
			  x0 / (double) (canvas_width - 1),
			  y0 / (double) (canvas_height - 1),
			  x1 / (double) (canvas_width - 1),
			  y1 / (double) (canvas_height - 1) );
}   /*  End Function draw_line  */

static flag draw_arc (PSCanvas pscanvas,
		      double centre_x, double centre_y,
		      double radius_x, double radius_y,
		      int angle1, int angle2, int rotation,
		      unsigned long pixel_value, flag fill)
/*  [PURPOSE] This routine will draw an arc on a PostScript canvas.
    <pscanvas> The PostScript canvas.
    <centre_x> The horizontal centre of the arc.
    <centre_y> The vertical centre of the arc.
    <radius_x> The horizontal radius of the arc.
    <radius_y> The vertical radius of the arc.
    <angle1> See <<XDrawArc>>.
    <angle2> See <<XDrawArc>>.
    <rotation> The rotation of the arc, in units of 1/64 degree. A rotation of
    0 degrees corresponds to (x > 0, y = 0). A rotation of 90 degrees
    corresponds to (x = 0, y > 0). In other words, the angle increases
    anticlockwise from the right.
    <pixel_value> The pixel value to use.
    <fill> If TRUE, the arc is filled, else only the outside is drawn.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int canvas_width, canvas_height;
    rgb_type rgb;
    static char function_name[] = "_kwin_ps_draw_arc";

    VERIFY_CANVAS (pscanvas);
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    get_rgb_of_pixel (pscanvas, &rgb, pixel_value);
    /*  Flip vertical  */
    centre_y = canvas_height - 1.0 - centre_y;
    return psw_rgb_arc (pscanvas->pspage, rgb.red, rgb.green, rgb.blue,
			centre_x / (double) (canvas_width - 1),
			centre_y / (double) (canvas_height - 1),
			radius_x / (double) (canvas_width - 1),
			radius_y / (double) (canvas_height - 1),
			angle1 / 64, angle2 / 64, -rotation / 64,
			fill);
}   /*  End Function draw_arc  */

static flag draw_polygon (PSCanvas pscanvas, double *x_arr, double *y_arr,
			  unsigned int num_vertices, unsigned long pixel_value,
			  flag convex, flag fill)
/*  [PURPOSE] This routine will draw a polygon onto a PostScript canvas.
    <pscanvas> The PostScript canvas.
    <x_arr> The array of x co-ordinates of vertices of the polygon.
    <y_arr> The array of y co-ordinates of vertices of the polygon.
    <num_vertices> The number of vertices in the polygon.
    <pixel_value> The pixel value to use.
    <convex> If TRUE, then the points must form a convex polygon.
    <fill> If TRUE, the polygon will be filled.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag retval;
    int canvas_width, canvas_height;
    int tmp;
    unsigned int coord_count;
    double xscale, yscale;
    rgb_type rgb;
    double *x, *y;
    static char function_name[] = "_kwin_ps_draw_polygon";

    VERIFY_CANVAS (pscanvas);
    FLAG_VERIFY (convex);
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    if ( ( x = m_alloc (num_vertices * sizeof *x) ) == NULL )
    {
	m_error_notify (function_name, "x array");
	return (FALSE);
    }
    if ( ( y = m_alloc (num_vertices * sizeof *y) ) == NULL )
    {
	m_error_notify (function_name, "y array");
	m_free (x);
	return (FALSE);
    }
    /*  Do a PostScript draw  */
    get_rgb_of_pixel (pscanvas, &rgb, pixel_value);
    xscale = 1.0 / (double) (canvas_width - 1);
    yscale = 1.0 / (double) (canvas_height - 1);
    for (coord_count = 0; coord_count < num_vertices; ++coord_count)
    {
	x[coord_count] = (double) x_arr[coord_count] * xscale;
	/*  Flip vertical  */
	tmp = canvas_height - 1 - y_arr[coord_count];
	y[coord_count] = (double) tmp * yscale;
    }
    retval = psw_rgb_polygon (pscanvas->pspage, rgb.red, rgb.green, rgb.blue,
			      x, y, num_vertices, TRUE);
    m_free (x);
    m_free (y);
    return (retval);
}   /*  End Function draw_polygon  */

#if 0
static flag draw_string (PSCanvas pscanvas, double x, double y,
			 CONST char *string, unsigned long pixel_value,
			 flag clear_under)
/*  [PURPOSE] This routine will draw a NULL terminated string onto an X11
    canvas, using the default font for the canvas.
    <pscanvas> The PostScript canvas.
    <x> The horizontal offset of the string origin.
    <y> The vertical offset of the string origin.
    <string> The string.
    <pixel_value> The pixel value to use.
    <clear_under> If TRUE, then the routine will draw both the foreground and
    background of the characters.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int length;
    int canvas_width, canvas_height;
    rgb_type rgb;
    static char function_name[] = "_kwin_ps_draw_string";

    VERIFY_CANVAS (pscanvas);
    FLAG_VERIFY (clear_under);
    if (clear_under)
    {
	fprintf (stderr, "%s: clear_under not supported\n", function_name);
	return (FALSE);
    }
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    get_rgb_of_pixel (pscanvas, &rgb, pixel_value);
    length = strlen (string);
    /*  Flip vertical  */
    y = canvas_height - 1.0 - y;
    return psw_rgb_text (pscanvas->pspage, rgb.red, rgb.green, rgb.blue,
			 string, NULL, 5,
			 x / (double) (canvas_width - 1),
			 y / (double) (canvas_height - 1),
			 0.0);
}   /*  End Function draw_string  */
#endif

static flag draw_segments (PSCanvas pscanvas,
			   double *x0, double *y0, double *x1, double *y1,
			   unsigned int num_segments,
			   unsigned long pixel_value)
/*  [PURPOSE] This routine will draw multiple disjoint lines onto a PostScript
    canvas.
    <pscanvas> The PostScript canvas.
    <x0> The horizontal start co-ordinates of the segments.
    <y0> The vetical start co-ordinates of the segments.
    <x1> The horizontal end co-ordinates of the segments.
    <y1> The vetical end co-ordinates of the segments.
    <num_segments> The number of segments.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int canvas_width, canvas_height;
    unsigned int count;
    double y0_l, y1_l;
    rgb_type rgb;
    static char function_name[] = "_kwin_ps_draw_segments";

    VERIFY_CANVAS (pscanvas);
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    get_rgb_of_pixel (pscanvas, &rgb, pixel_value);
    for (count = 0; count < num_segments; ++count)
    {
	/*  Flip vertical  */
	y0_l = (double) canvas_height - 1.0 - y0[count];
	y1_l = (double) canvas_height - 1.0 - y1[count];
	if ( !psw_rgb_line ( pscanvas->pspage, rgb.red, rgb.green, rgb.blue,
			     x0[count] / (double) (canvas_width - 1),
			     y0_l / (double) (canvas_height - 1),
			     x1[count] / (double) (canvas_width - 1),
			     y1_l / (double) (canvas_height - 1) ) )
	    return (FALSE);
    }
    return (TRUE);
}   /*  End Function draw_segments  */

static flag set_linewidth (PSCanvas pscanvas, double linewidth)
/*  [SUMMARY] Set the linewidth for a canvas.
    <pscanvas> The PostScript canvas.
    <linewidth> The linewidth, in pixels.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int canvas_width, canvas_height;
    double d_tmp;
    static char function_name[] = "_kwin_ps_set_linewidth";

    VERIFY_CANVAS (pscanvas);
    kwin_get_size (pscanvas->pixcanvas, &canvas_width, &canvas_height);
    d_tmp = (canvas_width * canvas_width) + (canvas_height * canvas_height);
    d_tmp = sqrt (d_tmp) / 1.414213562;
    psw_set_attributes (pscanvas->pspage,
			PSW_ATT_LINEWIDTH_RELATIVE, linewidth / d_tmp,
			PSW_ATT_END);
    return (TRUE);
}   /*  End Function set_linewidth  */

static flag set_line_dash (PSCanvas pscanvas, flag dash)
/*  [SUMMARY] Set the line dash for a canvas.
    <pscanvas> The PostScript canvas.
    <dash> If TRUE, the line will be dashed.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "_kwin_ps_set_line_dash";

    VERIFY_CANVAS (pscanvas);
    psw_set_attributes (pscanvas->pspage,
			PSW_ATT_LINE_DASH, dash,
			PSW_ATT_END);
    return (TRUE);
}   /*  End Function set_line_dash  */


/*  Private functions follow  */

static void get_rgb_of_pixel (PSCanvas pscanvas, rgb_type *rgb,
			      unsigned long pixel_value)
/*  [SUMMARY] Get the RGB value of a pixel value.
    <pscanvas> The PostScript canvas.
    <rgb> The RGB value is written here.
    <pixel_value> The pixel value.
    [RETURNS] Nothing.
*/
{
    if ( !pscanvas->pixel_valid || (pixel_value != pscanvas->pixel_value) )
    {
	unsigned short red, green, blue;

	kwin_get_pixel_RGB_values (pscanvas->pixcanvas, &pixel_value,
				   &red, &green, &blue, 1);
	pscanvas->rgb.red = (double) red / MAX_INTENSITY;
	pscanvas->rgb.green = (double) green / MAX_INTENSITY;
	pscanvas->rgb.blue = (double) blue / MAX_INTENSITY;
	pscanvas->pixel_value = pixel_value;
	pscanvas->pixel_valid = TRUE;
    }
    *rgb = pscanvas->rgb;
}   /*  End Function get_rgb_of_pixel  */
