/*LINTLIBRARY*/
/*  rasterise.c

    This code provides rasterise support functions for KPixCanvas objects.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains various rasterisation functions to convert geometric
    primitives to points.


    Written by      Richard Gooch   16-MAR-1997: Created
  <kwin_rasterise_draw_arc> function.

    Updated by      Richard Gooch   28-MAY-1997: Fixed bug when radius < 1.0.

    Updated by      Richard Gooch   6-AUG-1997: Added rounding offset.

    Updated by      Richard Gooch   6-FEB-1998: Redefined rotation.

    Updated by      Richard Gooch   2-SEP-1998: Changed to KPixFuncDrawPoints.

    Updated by      Richard Gooch   17-MAY-2000: Created new function
  <kwin_rasterise_draw_line>: grabbed code from generic.c:<kwin_draw_line>.

    Updated by      Richard Gooch   18-JUN-2001: Supported complete filled
  ellipses.

    Last updated by Richard Gooch   19-JUN-2001: Supported partial filled
  ellipses.


*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <karma_kwin.h>
#include <karma_kwin_hooks.h>
#include <karma_m.h>


#define BUF_SIZE 4096
#define YSELECT_ALL      0
#define YSELECT_LESS     1
#define YSELECT_GREATER  2


/*  Private structures  */
struct extrema_struct
{
    double *xmin;
    double *xmax;
    int yoffset;
    int len;
    unsigned int yselect;
    flag switch_if_selected;
    double ycentre;
};


/*  Private functions  */
STATIC_FUNCTION (flag _kwin_rasterise_draw_arc,
		 (KPixHookCanvas canvas, KPixFuncDrawPoints draw_points,
		  double centre_x, double centre_y,
		  double radius_x, double radius_y,
		  int angle1, int angle2, int rotation,
		  unsigned long pixel_value,
		  double *xbuf, double *ybuf, unsigned int buflen) );
STATIC_FUNCTION (flag do_points,
		 (KPixHookCanvas canvas, double *x_arr, double *y_arr,
		  unsigned int num_points, unsigned long pixel_value) );
STATIC_FUNCTION (flag draw_segs,
		 (KPixHookCanvas canvas, KPixFuncDrawSegments draw_segments,
		  struct extrema_struct *extrema, double *ybuffer,
		  unsigned int num_lines, unsigned long pixel_value) );


/*  Public functions follow  */

/*UNPUBLISHED_FUNCTION*/
flag kwin_rasterise_draw_arc (KPixHookCanvas canvas,
			      KPixFuncDrawPoints draw_points,
			      KPixFuncDrawSegments draw_segments,
			      double centre_x, double centre_y,
			      double radius_x, double radius_y,
			      int angle1, int angle2, int rotation,
			      unsigned long pixel_value, flag fill)
/*  [SUMMARY] Rasterise an arc.
    <canvas> The device-dependent canvas.
    <draw_points> The function to draw points.
    <draw_segments> The function to draw line segments.
    <centre_x> The horizontal centre of the arc.
    <centre_y> The vertical centre of the arc.
    <radius_x> The horizontal radius of the arc.
    <radius_y> The vertical radius of the arc.
    <angle1> See <<XDrawArc>>.
    <angle2> See <<XDrawArc>>.
    <rotation> The rotation of the arc, in units of 1/64 degree. A rotation of
    0 degrees corresponds to (x > 0, y = 0). A rotation of 90 degrees
    corresponds to (x = 0, y > 0). In other words, the angle increases
    anticlockwise from the right.
    <pixel_value> The pixel value to use.
    <fill> If TRUE, the arc is filled, else only the outside is drawn.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int count1, iy;
    double start_angle, stop_angle, rot;
    double x, y, x1, y1, x2, y2;
    double cos_rot, sin_rot;
    struct extrema_struct extrema;
    double buffer[BUF_SIZE * 4];
    double *x0arr, *y0arr;
    static char function_name[] = "kwin_rasterise_draw_arc";

    if (!fill)
	return _kwin_rasterise_draw_arc (canvas, draw_points,
					 centre_x, centre_y,
					 radius_x, radius_y,
					 angle1, angle2, rotation,
					 pixel_value,
					 buffer, buffer + BUF_SIZE * 2,
					 BUF_SIZE * 2);
    if (draw_segments == NULL)
    {
	fprintf (stderr,
		 "%s: cannot draw filled arcs without draw_segments method\n",
		 function_name);
	return (FALSE);
    }
    /*  Draw a filled arc by "drawing" the boundary to compute the extrema for
	each horizontal line, then drawing those lines  */
    while (angle2 <= angle1) angle2 += 360 * 64;
    start_angle = (double) angle1 / 64.0 * PION180;
    stop_angle = (double) angle2 / 64.0 * PION180;
    rot = (double) rotation / 64.0 * PION180;
    cos_rot = cos (rot);
    sin_rot = sin (rot);
    centre_x = floor (centre_x + 1e-3);
    centre_y = floor (centre_y + 1e-3);
    radius_x = floor (fabs (radius_x) + 1e-3);
    radius_y = floor (fabs (radius_y) + 1e-3);
    /*  First compute Y range  */
    iy = (radius_x > radius_y) ? radius_x : radius_y;
    iy = iy << 1;
    if (iy <= BUF_SIZE)
    {
	x0arr = buffer;
	y0arr = x0arr + BUF_SIZE;
	extrema.xmin = y0arr + BUF_SIZE;
	extrema.xmax = extrema.xmin + BUF_SIZE;
    }
    else
    {
	static unsigned int buflen = 0;  /*  Buffer length is 4 times this  */
	static double *buffer = NULL;

	if (iy > buflen)
	{
	    if (buffer != NULL) m_free (buffer);
	    if ( ( buffer = m_alloc (iy * 4 * sizeof *buffer) ) == NULL )
		m_abort (function_name, "co-ordinate buffer");
	    buflen = iy;
	}
	x0arr = buffer;
	y0arr = x0arr + iy;
	extrema.xmin = y0arr + iy;
	extrema.xmax = extrema.xmin + iy;
    }
    extrema.yoffset = centre_y - (iy >> 1);
    extrema.len = iy;
    for (count1 = 0; count1 < iy; ++count1)
    {
	extrema.xmin[count1] =  1000000;
	extrema.xmax[count1] = -1000000;
    }
    x = radius_x * cos (start_angle);
    y = radius_y * sin (start_angle);
    x1 = centre_x + x * cos_rot - y * sin_rot + 0.5;
    y1 = centre_y + y * cos_rot + x * sin_rot + 0.5;
    x = radius_x * cos (stop_angle);
    y = radius_y * sin (stop_angle);
    x2 = centre_x + x * cos_rot - y * sin_rot + 0.5;
    y2 = centre_y + y * cos_rot + x * sin_rot + 0.5;
    if ( (angle2 >= angle1 + 23040) || (angle2 <= angle1 + 11520) ||
	 ( (y1 < centre_y) && (centre_y < y2) ) ||
	 ( (y2 < centre_y) && (centre_y < y1) ) )
    {   /*  Simply filled ellipse: only one segment per line  */
	extrema.yselect = YSELECT_ALL;
	_kwin_rasterise_draw_arc (&extrema, do_points,
				  centre_x, centre_y, radius_x, radius_y,
				  angle1, angle2, rotation, 0,
				  x0arr, y0arr, iy);
	if (angle2 < angle1 + 23040)
	{   /*  Not full ellipse: need line boundaries  */
	    kwin_rasterise_draw_line (&extrema, do_points,
				      centre_x, centre_y, x1, y1, 0);
	    kwin_rasterise_draw_line (&extrema, do_points,
				      centre_x, centre_y, x2, y2, 0);
	}
	return draw_segs (canvas, draw_segments, &extrema, y0arr, iy,
			  pixel_value);
    }
    /*  Some lines require two segments: draw ellipse in two parts  */
    extrema.ycentre = centre_y;
    /*  The first part  */
    extrema.yselect = (y1 < centre_y) ? YSELECT_LESS : YSELECT_GREATER;
    extrema.switch_if_selected = FALSE;
    _kwin_rasterise_draw_arc (&extrema, do_points,
			      centre_x, centre_y, radius_x, radius_y,
			      angle1, angle2, rotation, 0, x0arr, y0arr, iy);
    kwin_rasterise_draw_line (&extrema, do_points, centre_x, centre_y, x1, y1,
			      0);
    if ( !draw_segs (canvas, draw_segments, &extrema, y0arr, iy, pixel_value) )
	return (FALSE);
    /*  Now the second part  */
    extrema.yselect = (y1 < centre_y) ? YSELECT_GREATER : YSELECT_LESS;
    extrema.switch_if_selected = TRUE;
    for (count1 = 0; count1 < iy; ++count1)
    {
	extrema.xmin[count1] =  1000000;
	extrema.xmax[count1] = -1000000;
    }
    _kwin_rasterise_draw_arc (&extrema, do_points,
			      centre_x, centre_y, radius_x, radius_y,
			      angle1, angle2, rotation, 0, x0arr, y0arr, iy);
    kwin_rasterise_draw_line (&extrema, do_points, centre_x, centre_y, x2, y2,
			      0);
    return draw_segs (canvas, draw_segments, &extrema, y0arr, iy, pixel_value);
}   /*  End Function kwin_rasterise_draw_arc  */

/*UNPUBLISHED_FUNCTION*/
flag kwin_rasterise_draw_line (KPixHookCanvas canvas,
			       KPixFuncDrawPoints draw_points,
			       double x0, double y0, double x1, double y1,
			       unsigned long pixel_value)
/*  [SUMMARY] Rasterise a line.
    <canvas> The device-dependent canvas.
    <draw_points> The function to draw points.
    <x0> The horizontal offset of the first point.
    <y0> The vertical offset of the first point.
    <x1> The horizontal offset of the second point.
    <y1> The vertical offset of the second point.
    <pixel_value> The pixel value to use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int i_x0 = x0;
    int i_y0 = y0;
    int i_x1 = x1;
    int i_y1 = y1;
    int sum;
    int inc;
    int x, y;
    int tmp;
    int delta_x, delta_y;
    unsigned int num_points = 0;
    double xarr[BUF_SIZE], yarr[BUF_SIZE];

    /*  First do horizontal draw  */
    if (i_x1 < i_x0)
    {
	tmp = i_x1;  i_x1 = i_x0;  i_x0 = tmp;
	tmp = i_y1;  i_y1 = i_y0;  i_y0 = tmp;
    }
    delta_x = i_x1 - i_x0;
    if (i_y1 < i_y0)
    {
	/*  Line goes up  */
	delta_y = i_y0 - i_y1;
	inc = -1;
    }
    else
    {
	/*  Line goes down  */
	delta_y = i_y1 - i_y0;
	inc = 1;
    }
    x = i_x0;
    y = i_y0;
    sum = delta_x / 2;
    xarr[num_points] = x;
    yarr[num_points] = y;
    ++num_points;
    while (x < i_x1)
    {
	++x;
	sum += delta_y;
	while (sum >= delta_x)
	{
	    y += inc;
	    sum -= delta_x;
	}
	xarr[num_points] = x;
	yarr[num_points] = y;
	if (++num_points >= BUF_SIZE)
	{
	    /*  Draw en-masse  */
	    (*draw_points) (canvas, xarr, yarr, num_points, pixel_value);
	    num_points = 0;
	}
    }
    /*  Now do vertical draw  */
    if (i_y1 < i_y0)
    {
	tmp = i_y1;  i_y1 = i_y0;  i_y0 = tmp;
	tmp = i_x1;  i_x1 = i_x0;  i_x0 = tmp;
    }
    delta_y = i_y1 - i_y0;
    if (i_x1 < i_x0)
    {
	/*  Line goes left  */
	delta_x = i_x0 - i_x1;
	inc = -1;
    }
    else
    {
	/*  Line goes right  */
	delta_x = i_x1 - i_x0;
	inc = 1;
    }
    y = i_y0;
    x = i_x0;
    sum = delta_y / 2;
    xarr[num_points] = x;
    yarr[num_points] = y;
    if (++num_points >= BUF_SIZE)
    {
	/*  Draw en-masse  */
	(*draw_points) (canvas, xarr, yarr, num_points, pixel_value);
	num_points = 0;
    }
    while (y < i_y1)
    {
	++y;
	sum += delta_x;
	while (sum >= delta_y)
	{
	    x += inc;
	    sum -= delta_y;
	}
	xarr[num_points] = x;
	yarr[num_points] = y;
	if (++num_points >= BUF_SIZE)
	{
	    /*  Draw en-masse  */
	    (*draw_points) (canvas, xarr, yarr, num_points, pixel_value);
	    num_points = 0;
	}
    }
    if (num_points > 0)
    {
	/*  Draw remaining points en-masse  */
	(*draw_points) (canvas, xarr, yarr, num_points, pixel_value);
    }
    return (TRUE);
}   /*  End Function kwin_rasterise_draw_line  */


/*  Private functions follow  */
flag _kwin_rasterise_draw_arc (KPixHookCanvas canvas,
			       KPixFuncDrawPoints draw_points,
			       double centre_x, double centre_y,
			       double radius_x, double radius_y,
			       int angle1, int angle2, int rotation,
			       unsigned long pixel_value,
			       double *xbuf, double *ybuf, unsigned int buflen)
/*  [SUMMARY] Rasterise an unfilled arc.
    <canvas> The device-dependent canvas.
    <draw_points> The function to draw points.
    <centre_x> The horizontal centre of the arc.
    <centre_y> The vertical centre of the arc.
    <radius_x> The horizontal radius of the arc.
    <radius_y> The vertical radius of the arc.
    <angle1> See <<XDrawArc>>.
    <angle2> See <<XDrawArc>>.
    <rotation> The rotation of the arc, in units of 1/64 degree. A rotation of
    0 degrees corresponds to (x > 0, y = 0). A rotation of 90 degrees
    corresponds to (x = 0, y > 0). In other words, the angle increases
    anticlockwise from the right.
    <pixel_value> The pixel value to use.
    <xbuf> The X buffer the function may use.
    <ybuf> The Y buffer the function may use.
    <buflen> The length of the buffer the function may use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int ix, iy;
    int old_x = -1;
    int old_y = -1;
    unsigned int num_points = 0;
    double start_angle, stop_angle, angle, step, x, y, x1, y1, rot;
    double cos_rot, sin_rot, one_on_radius;
    /*static char function_name[] = "_kwin_rasterise_draw_arc";*/

    while (angle2 <= angle1) angle2 += 360 * 64;
    start_angle = (double) angle1 / 64.0 * PION180;
    stop_angle = (double) angle2 / 64.0 * PION180;
    rot = (double) rotation / 64.0 * PION180;
    cos_rot = cos (rot);
    sin_rot = sin (rot);
    centre_x = floor (centre_x + 1e-3);
    centre_y = floor (centre_y + 1e-3);
    radius_x = floor (fabs (radius_x) + 1e-3);
    radius_y = floor (fabs (radius_y) + 1e-3);
    one_on_radius = 1.0 / ( (radius_x > radius_y) ? radius_x : radius_y );
    if (one_on_radius <= 1.0) step = asin (one_on_radius) * 0.5;
    else step = 0.75;
    for (angle = start_angle; angle <= stop_angle; angle += step)
    {
	x1 = radius_x * cos (angle);
	y1 = radius_y * sin (angle);
	x = centre_x + x1 * cos_rot - y1 * sin_rot + 0.5;
	y = centre_y + y1 * cos_rot + x1 * sin_rot + 0.5;
	ix = x;
	iy = y;
	if ( (ix == old_x) && (iy == old_y) ) continue;
	old_x = ix;
	old_y = iy;
	xbuf[num_points] = x;
	ybuf[num_points] = y;
	if (++num_points >= buflen)
	{
	    /*  Draw en-masse  */
	    if ( !(*draw_points) (canvas, xbuf, ybuf, num_points,
				  pixel_value) ) return (FALSE);
	    num_points = 0;
	}
    }
    if (num_points > 0)  /*  Draw remaining points en-masse  */
	return (*draw_points) (canvas, xbuf, ybuf, num_points, pixel_value);
    return (TRUE);
}   /*  End Function _kwin_rasterise_draw_arc  */

static flag do_points (KPixHookCanvas canvas, double *x_arr, double *y_arr,
		       unsigned int num_points, unsigned long pixel_value)
/*  [PURPOSE] This routine will compute the horizontal range of multiple points
    <canvas> The overloaded canvas pointer.
    <x_array> The horizontal co-ordinates of the points.
    <y_array> The vertical co-ordinates of the points.
    <num_points> The number of points.
    <pixel_value> The pixel value to use. Ignored.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int ix, iy;
    unsigned int count;
    struct extrema_struct *extrema = (struct extrema_struct *) canvas;
    static char function_name[] = "_kwin_rasterise_do_points";

    for (count = 0; count < num_points; ++count)
    {
	flag add = TRUE;

	ix = x_arr[count];
	iy = (int) y_arr[count] - extrema->yoffset;
	if ( (iy < 0) || (iy >= extrema->len) )
	{
	    fprintf (stderr, "%s: y: %d len: %d\n",
		     function_name, iy, extrema->len);
	    continue;
	}
	if (extrema->yselect == YSELECT_LESS)
	{
	    if (y_arr[count] <= extrema->ycentre)
	    {
		if (extrema->switch_if_selected) extrema->yselect =YSELECT_ALL;
	    }
	    else add = FALSE;
	}
	else if (extrema->yselect == YSELECT_GREATER)
	{
	    if (y_arr[count] >= extrema->ycentre)
	    {
		if (extrema->switch_if_selected) extrema->yselect =YSELECT_ALL;
	    }
	    else add = FALSE;
	}
	if (add)
	{
	    if (ix < extrema->xmin[iy]) extrema->xmin[iy] = ix;
	    if (ix > extrema->xmax[iy]) extrema->xmax[iy] = ix;
	}
	else if (!extrema->switch_if_selected) return (FALSE);
    }
    return (TRUE);
}   /*  End Function do_points  */

static flag draw_segs (KPixHookCanvas canvas,
		       KPixFuncDrawSegments draw_segments,
		       struct extrema_struct *extrema, double *ybuffer,
		       unsigned int num_lines, unsigned long pixel_value)
{
    signed int count, num_segs;

    for (count = 0, num_segs = 0; count < num_lines; ++count)
    {
	if (extrema->xmax[count] < extrema->xmin[count]) continue;
	extrema->xmin[num_segs] = extrema->xmin[count];
	extrema->xmax[num_segs] = extrema->xmax[count];
	ybuffer[num_segs] = count + extrema->yoffset;
	++num_segs;
    }
    return (*draw_segments) (canvas,
			     extrema->xmin, ybuffer, extrema->xmax, ybuffer,
			     num_segs, pixel_value);
}   /*  End Function draw_segs  */
