/*LINTLIBRARY*/
#ifdef HAS_X11
/*  xgl.c

    This code provides KPixCanvas objects.

    Copyright (C) 1995-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating a simple pixel
    canvas (window) independent of the graphics system in use. This file
    contains the XGL code.


    Written by      Richard Gooch   22-SEP-1995

    Updated by      Richard Gooch   22-SEP-1995

    Updated by      Richard Gooch   20-OCT-1995: Created <resize> function.

    Updated by      Richard Gooch   28-JAN-1996: Added test for
  KWIN_XGL_FORCE_STEREO environment variable.

    Updated by      Richard Gooch   1-FEB-1996: Created <convert_colour>
  routine to do colour conversion right.

    Updated by      Richard Gooch   4-FEB-1996: Worked around bug in XGL in
  <size_cache>.

    Updated by      Richard Gooch   19-FEB-1996: Modified interface for
  <clear_area> to support width and height parameters.

    Updated by      Richard Gooch   21-FEB-1996: Pass <resize> func when
  creating generic pixel canvas: forgot to do this when <resize> created.

    Updated by      Richard Gooch   23-FEB-1996: Clip (sub)image to cache size
  in <draw_cached_image> and added <<wait>> parameter.

    Updated by      Richard Gooch   13-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   26-APR-1996: Moved stub functions in.

    Updated by      Richard Gooch   26-JUL-1996: Created <draw_string>.

    Updated by      Richard Gooch   15-SEP-1996: Made use of new <xv_*>
  routines.

    Updated by      Richard Gooch   12-DEC-1996: Use dlopen() where available
  to detect at runtime if XGL is available. Graphics library is no longer
  linked with -lxgl

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <r> package.

    Last updated by Richard Gooch   9-JUN-2000: Removed unused variable.


*/
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <os.h>
#ifdef HAS_DL  /*  XGL requires dlopen() and friends  */
#  include <dlfcn.h>
#endif
#include <karma.h>
#define X11
#include <karma_kwin.h>
#include <karma_dl.h>
#include <karma_r.h>


/*  Private data  */


/*  Private functions  */
STATIC_FUNCTION (flag initialise, () );


/*  Functions in the shared object  */
STATIC_FUNCTION (unsigned int (*FPTR_kwin_xgl_test_stereo),
		 (Display *display, Window window) );
STATIC_FUNCTION (flag (*FPTR_kwin_xgl_create_stereo),
		 (Display *display, Window window,
		  int xoff, int yoff, int width, int height,
		  KPixCanvas *mono,
		  KPixCanvas *left, KPixCanvas *right) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
unsigned int kwin_xgl_test_stereo (Display *display, Window window)
/*  [SUMMARY] Test if stereo supported using XGL.
    [PURPOSE] This routine will determine if stereoscopic display is supported
    using XGL on a particular display.
    [NOTE] This routine is only available with the X window system.
    <display> The X display.
    <window> The window ID of a window of the same class as the desired stereo
    window.
    [RETURNS] A status code. See [<kwin_XGL_STATUS>] for a list of values.
*/
{
    /*static char function_name[] = "kwin_xgl_test_stereo";*/

    if ( !initialise () ) return (KWIN_XGL_NOT_AVAILABLE);
    return (*FPTR_kwin_xgl_test_stereo) (display, window);
}   /*  End Function kwin_xgl_test_stereo  */

/*PUBLIC_FUNCTION*/
flag kwin_xgl_create_stereo (Display *display, Window window,
			     int xoff, int yoff, int width, int height,
			     KPixCanvas *mono,
			     KPixCanvas *left, KPixCanvas *right)
/*  [SUMMARY] Create a stereo pixel canvas using XGL.
    [PURPOSE] This routine will create a stereo pixel canvas, ready for
    drawing, from an X window, using the XGL library. Note that the origin of
    a KPixCanvas is the upper-left corner.
    [NOTE] This routine is only available with the X window system.
    <display> The X display.
    <window> The window ID in which the canvas lies.
    <gc> The graphics context.
    <xoff> The horizontal offset of the canvas origin (upper-left corner)
    relative to the parent window.
    <yoff> The vertical offset of the canvas origin (upper-left corner)
    relative to the parent window.
    <width> The width (vertical extent) of the canvas.
    <height> The height (horizontal extent) of the canvas.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    /*static char function_name[] = "kwin_xgl_create_stereo";*/

    if ( !initialise () ) return (KWIN_XGL_NOT_AVAILABLE);
    return (*FPTR_kwin_xgl_create_stereo) (display, window,
					   xoff, yoff, width, height,
					   mono, left, right);
}   /*  End Function kwin_xgl_create_stereo  */


/*  Private functions follow  */

static flag initialise ()
/*  [SUMMARY] Initialise. May be called many times.
    [RETURNS] TRUE if XGL is available, else FALSE.
*/
{
#ifdef HAS_DL
    void *handle;
    CONST char *xglhome;
    char txt[STRING_LENGTH], failed_object[STRING_LENGTH];
#endif
    static flag initialised = FALSE;
    static flag ok = FALSE;

    if (initialised) return (ok);
    initialised = TRUE;
#ifdef HAS_DL
    /*  Load kwin_xgl.so  */
    if ( ( xglhome = r_getenv ("XGLHOME") ) != NULL )
    {
	sprintf (txt, "%s/lib", xglhome);
    }
    handle = dl_load_object ("kwin_xgl", txt, failed_object);
    if ( (handle == NULL) && (strcmp (failed_object, "libxgl") == 0) )
    {
	fprintf (stderr, "Error loading XGL\n");
	fprintf (stderr,
		 "Please ask your system administrator to install the XGL");
	fprintf (stderr,
		 " library (part of the\nSUNWits package). Stereo not ");
	fprintf (stderr, "available, continuing anyway.\n");
	return (FALSE);
    }
    if (handle == NULL) return (FALSE);
    if ( ( FPTR_kwin_xgl_test_stereo = ( unsigned int (*) () )
	   dlsym (handle, "DL_kwin_xgl_test_stereo") ) == NULL ) return FALSE;
    if ( ( FPTR_kwin_xgl_create_stereo = ( flag (*) () )
	   dlsym (handle, "DL_kwin_xgl_create_stereo") ) == NULL )
	return (FALSE);
    ok = TRUE;
#endif
    return (ok);
}   /*  End Function initialise  */


#else  /*  HAS_X11  */

struct unused
{
    int a;
};

#endif  /*  !HAS_X11  */
