/*LINTLIBRARY*/
/*  editor.c

    This code provides an interactive KOverlayList object editor.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for interactively creating geometric
    objects.


    Written by      Richard Gooch   9-AUG-1997: Moved from OverlayEditorControl
  widget.

    Updated by      Richard Gooch   21-AUG-1997: Completed support for ellipses

    Updated by      Richard Gooch   6-FEB-1998: Cosmetic function name change.

    Updated by      Richard Gooch   10-MAY-1998: Added support for linear
  co-ordinates.

    Updated by      Richard Gooch   30-MAY-1998: Ensure negative radii cannot
  cause problems. Should not actually happen.

    Updated by      Richard Gooch   9-JUN-1998: Removed unnecessary check for
  raw event code in <geom_func>.

    Updated by      Richard Gooch   28-JUN-2000: Made use of new function
  <canvas_register_destroy_func>.

    Updated by      Richard Gooch   7-JUL-2000: Took account of interface
  change to <canvas_register_destroy_func>.

    Last updated by Richard Gooch   4-SEP-2000: Added attribute
  OVERLAY_EDITOR_CONTEXT_ATT_COLOUR.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#define NEW_WIN_SCALE
#include <karma.h>
#include <k_event_codes.h>
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_c.h>


#define BUF_SIZE 16384
#define MAGIC_NUMBER 498532903

#define VERIFY_CONTEXT(context) if (context == NULL) \
{fprintf (stderr, "NULL context passed\n"); \
 a_prog_bug (function_name); } \
if (context->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid context object\n"); \
 a_prog_bug (function_name); }


struct bind_struct
{
    KOverlayListEditorContext olist_ctx;
    KWorldCanvas canvas;
    KPixCanvasEditorContext pix_context;
    struct bind_struct *prev;
    struct bind_struct *next;
};

struct pix_att_type
{
    unsigned int mouse_mask;
    unsigned int object_type;
    flag clear_object;
};

struct attribute_type
{
    unsigned int edit_mode;
    unsigned int first_coord_type;
    unsigned int other_coord_type;
    char *colour;
};

struct olist_editor_type
{
    unsigned int magic_number;
    KOverlayList olist;
    flag in_edit;
    struct attribute_type att;
    struct pix_att_type pix_att;
    struct bind_struct *first_binding;
};


/*  Private functions  */
STATIC_FUNCTION (void process_attributes, (KOverlayListEditorContext context,
					   va_list argp) );
STATIC_FUNCTION (flag _overlay_editor_geom_func,
		 (KPixCanvasEditorContext context, KPixCanvas canvas,
		  unsigned int object_type, unsigned int num_coords,
		  CONST int *x, CONST int *y,
		  unsigned int raw_event_code, void *raw_e_info,
		  void **f_info) );
STATIC_FUNCTION (void _overlay_editor_wcanvas_destroy_func,
		 (struct bind_struct *binding) );


/*  Public Functions follow  */

/*UNPUBLISHED_FUNCTION*/
KOverlayListEditorContext overlay_editor_create_context
    (KOverlayList olist, ...)
/*  [SUMMARY] Create a geometry editing context for an overlay list.
    <olist> The KOverlayList object.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    OVERLAY_EDITOR_CONTEXT_ATT_END. See [<overlay_EDITOR_CONTEXT_ATTRIBUTES>]
    for the list of attributes.
    [RETURNS] A KPixCanvasEditorContext object on success, else NULL.
*/
{
    va_list argp;
    KOverlayListEditorContext context;
    static char function_name[] = "overlay_editor_create_context";

    va_start (argp, olist);
    if ( ( context = m_calloc (sizeof *context) ) == NULL )
    {
	m_error_notify (function_name, "KOverlayListEditorContext object");
	return (NULL);
    }
    context->magic_number = MAGIC_NUMBER;
    context->olist = olist;
    process_attributes (context, argp);
    va_end (argp);
    return (context);
}   /*  End Function overlay_editor_create_context  */

/*UNPUBLISHED_FUNCTION*/
void overlay_editor_bind (KOverlayListEditorContext context,
			  KWorldCanvas canvas, KPixCanvasEditorContext pix_ctx)
/*  [SUMMARY] Bind a canvas to an overlay editing context.
    <context> The overlay editing context.
    <canvas> The world canvas.
    <pix_ctx> The KPixCanvasEditorContext object. If this is NULL one is
    created internally.
    [NOTE] If the canvas is destroyed, the binding is automatically removed.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    unsigned int visual;
    unsigned long pixel_value, green, blue;
    struct bind_struct *new;
    static char function_name[] = "overlay_editor_bind";

    VERIFY_CONTEXT (context);
    pixcanvas = canvas_get_pixcanvas (canvas);
    if ( ( new = m_calloc (sizeof *new) ) == NULL )
	m_abort (function_name, "bind structure");
    if (pix_ctx == NULL)
    {
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_VISUAL, &visual,
			     KWIN_ATT_END);
	if (visual == KWIN_VISUAL_PSEUDOCOLOUR)
	{
	    if ( !kwin_get_colour (pixcanvas, "cyan", &pixel_value,
				   NULL, NULL, NULL) )
	    {
		/*  Can't get green: try white instead  */
		if ( !kwin_get_colour (pixcanvas, "white", &pixel_value,
				       NULL, NULL, NULL) )
		{
		    /*  Can't even get white (this should never actually happen
			with X), so just pick 0  */
		    pixel_value = 0;
		}
	    }
	}
	else
	{
	    kwin_get_attributes (pixcanvas,
				 KWIN_ATT_PIX_GREEN_MASK, &green,
				 KWIN_ATT_PIX_GREEN_MASK, &blue,
				 KWIN_ATT_END);
	    pixel_value = green | blue;
	}
	if ( ( pix_ctx =
	       kwin_editor_create_context (pixcanvas,
					   KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE,
					   pixel_value,
					   KWIN_EDITOR_CONTEXT_ATT_END) )
	     == NULL ) m_abort (function_name, "KPixCanvasEditorContext");
	canvas_register_destroy_func (canvas, TRUE,
				      kwin_editor_destroy_context, pix_ctx);
    }
    else
    {
	if (kwin_editor_get_pixcanvas (pix_ctx) != pixcanvas)
	{
	    fprintf (stderr, "KPixCanvas missmatch\n");
	    a_prog_bug (function_name);
	}
    }
    kwin_editor_set_attributes (pix_ctx,
				KWIN_EDITOR_CONTEXT_ATT_MOUSE,
				context->pix_att.mouse_mask,
				KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
				context->pix_att.object_type,
				KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
				context->pix_att.clear_object,
				KWIN_EDITOR_CONTEXT_ATT_END);
    new->olist_ctx = context;
    new->canvas = canvas;
    new->pix_context = pix_ctx;
    if (context->first_binding != NULL) context->first_binding->prev = new;
    new->next = context->first_binding;
    context->first_binding = new;
    kwin_editor_register_event_func (pix_ctx, _overlay_editor_geom_func, new);
    canvas_register_destroy_func (canvas, FALSE,
				  _overlay_editor_wcanvas_destroy_func, new);
}   /*  End Function overlay_editor_bind  */

/*UNPUBLISHED_FUNCTION*/
void overlay_editor_set_attributes (KOverlayListEditorContext context, ...)
/*  [SUMMARY] Set attributes for a geometry editing context.
    <context> The context.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    OVERLAY_EDITOR_CONTEXT_ATT_END. See [<overlay_EDITOR_CONTEXT_ATTRIBUTES>]
    for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    static char function_name[] = "overlay_editor_set_attributes";

    VERIFY_CONTEXT (context);
    va_start (argp, context);
    process_attributes (context, argp);
    va_end (argp);
}   /*  End Function overlay_editor_set_attributes  */

/*UNPUBLISHED_FUNCTION*/
void overlay_editor_get_attributes (KOverlayListEditorContext context, ...)
/*  [SUMMARY] Get attributes for a geometry editing context.
    <context> The context.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    OVERLAY_EDITOR_CONTEXT_ATT_END. See [<overlay_EDITOR_CONTEXT_ATTRIBUTES>]
    for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key;
    char **colour;
    static char function_name[] = "overlay_editor_get_attributes";

    VERIFY_CONTEXT (context);
    va_start (argp, context);
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    OVERLAY_EDITOR_CONTEXT_ATT_END )
    {
	switch (att_key)
	{
	  case OVERLAY_EDITOR_CONTEXT_ATT_COLOUR:
	    colour = va_arg (argp, char **);
	    *colour = context->att.colour;
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
}   /*  End Function overlay_editor_get_attributes  */

/*UNPUBLISHED_FUNCTION*/
void overlay_editor_set_pixcontext_attributes
    (KOverlayListEditorContext context, ...)
/*  [SUMMARY] Set attributes for all bound KPixCanvasEditorContext objects.
    <context> The context.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    KWIN_EDITOR_CONTEXT_ATT_END. See [<KWIN_EDITOR_CONTEXT_ATTRIBUTES>]
    for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key, mask;
    struct bind_struct *curr;
    static char function_name[] = "overlay_editor_set_pixcontext_attributes";

    VERIFY_CONTEXT (context);
    va_start (argp, context);
    mask = K_CANVAS_EVENT_LEFT_MOUSE_MASK | K_CANVAS_EVENT_MIDDLE_MOUSE_MASK |
	K_CANVAS_EVENT_RIGHT_MOUSE_MASK;
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    KWIN_EDITOR_CONTEXT_ATT_END )
    {
	switch (att_key)
	{
	  case KWIN_EDITOR_CONTEXT_ATT_MOUSE:
	    context->pix_att.mouse_mask = va_arg (argp, unsigned int) & mask;
	    break;
	  case KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT:
	    context->pix_att.object_type = va_arg (argp, unsigned int);
	    break;
	  case KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT:
	    context->pix_att.clear_object = va_arg (argp, flag);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    for (curr = context->first_binding; curr != NULL; curr = curr->next)
    {
	kwin_editor_set_attributes (curr->pix_context,
				    KWIN_EDITOR_CONTEXT_ATT_MOUSE,
				    context->pix_att.mouse_mask,
				    KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
				    context->pix_att.object_type,
				    KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
				    context->pix_att.clear_object,
				    KWIN_EDITOR_CONTEXT_ATT_END);
    }
}   /*  End Function overlay_editor_set_pixcontext_attributes  */


/*  Private functions follow  */

static void process_attributes (KOverlayListEditorContext context,
				va_list argp)
/*  [SUMMARY] Process attributes for a KOverlayListEditorContext.
    <context> The editor context.
    <argp> The varargs parameters.
    [RETURNS] Nothing.
*/
{
    unsigned int att_key;
    char *colour;
    static char function_name[] = "_overlay_editor_process_attributes";

    if (context->in_edit)
    {
	fprintf (stderr, "Cannot modify attributes while editing\n");
	a_prog_bug (function_name);
    }
    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    OVERLAY_EDITOR_CONTEXT_ATT_END )
    {
	switch (att_key)
	{
	  case OVERLAY_EDITOR_CONTEXT_ATT_EDIT_MODE:
	    context->att.edit_mode = va_arg (argp, unsigned int);
	    break;
	  case OVERLAY_EDITOR_CONTEXT_ATT_FIRST_COORD_TYPE:
	    context->att.first_coord_type = va_arg (argp, unsigned int);
	    break;
	  case OVERLAY_EDITOR_CONTEXT_ATT_OTHER_COORD_TYPE:
	    context->att.other_coord_type = va_arg (argp, unsigned int);
	    break;
	  case OVERLAY_EDITOR_CONTEXT_ATT_COLOUR:
	    colour = va_arg (argp, char *);
	    if (colour != NULL)
	    {
		if ( ( colour = st_dup (colour) ) == NULL )
		{
		    m_error_notify (function_name, "colour name");
		    break;
		}
	    }
	    if (context->att.colour != NULL) m_free (context->att.colour);
	    context->att.colour = colour;
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
}   /*  End Function process_attributes  */

static flag _overlay_editor_geom_func (KPixCanvasEditorContext pix_ctx,
				       KPixCanvas canvas,
				       unsigned int object_type,
				       unsigned int num_coords,
				       CONST int *x, CONST int *y,
				       unsigned int raw_event_code,
				       void *raw_e_info, void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <pix_ctx> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas worldcanvas;
    unsigned int count;
    unsigned int num_dcoords = 0;
    double win_left_x, win_right_x, win_bottom_y, win_top_y;
    struct bind_struct *binding = *f_info;
    char *colour = "green";
    unsigned int types[BUF_SIZE];
    double dx[BUF_SIZE], dy[BUF_SIZE];
    static char function_name[] = "_overlay_editor_geom_func";

    if (num_coords + 1 >= BUF_SIZE)
    {
	fprintf (stderr, "Too many co-ordinates: %u in object\n", num_coords);
	a_prog_bug (function_name);
    }
    worldcanvas = binding->canvas;
    if (canvas_get_pixcanvas (worldcanvas) != canvas)
    {
	fprintf (stderr, "World canvas missmatch!\n");
	a_prog_bug (function_name);
    }
    if (binding->olist_ctx->att.colour != NULL)
	colour = binding->olist_ctx->att.colour;
    switch (object_type)
    {
      case KWIN_EDIT_OBJECT_RECTANGLE:
	num_dcoords = 5;
	dx[0] = x[0];
	dy[0] = y[0];
	dx[1] = x[1];
	dy[1] = y[0];
	dx[2] = x[1];
	dy[2] = y[1];
	dx[3] = x[0];
	dy[3] = y[1];
	dx[4] = x[0];
	dy[4] = y[0];
	break;
      case KWIN_EDIT_OBJECT_ELLIPSE:
	num_dcoords = 0;
	dx[0] = x[0];
	dy[0] = y[0];
	dx[1] = abs (x[1]);
	dy[1] = abs (y[1]);
	dx[2] = (double) x[2] / 64.0;
	types[1] = binding->olist_ctx->att.other_coord_type;
	break;
      case KWIN_EDIT_OBJECT_POLYGON:
	num_dcoords = num_coords + 1;
	for (count = 0; count < num_coords; ++count)
	{
	    dx[count] = x[count];
	    dy[count] = y[count];
	}
	dx[num_dcoords - 1] = x[0];
	dy[num_dcoords - 1] = y[0];
	break;
    }
    canvas_get_attributes (worldcanvas,
			   CANVAS_ATT_LEFT_X, &win_left_x,
			   CANVAS_ATT_RIGHT_X, &win_right_x,
			   CANVAS_ATT_BOTTOM_Y, &win_bottom_y,
			   CANVAS_ATT_TOP_Y, &win_top_y,
			   CANVAS_ATT_END);
    if (num_dcoords > 0)
    {
	switch (binding->olist_ctx->att.first_coord_type)
	{
	  case OVERLAY_COORD_PIXEL:
	    break;
	  case OVERLAY_COORD_RELATIVE:
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE, 1,
					     dx, dy, dx, dy, NULL, NULL);
	    dx[0] = (dx[0] - win_left_x) / (win_right_x - win_left_x);
	    dy[0] = (dy[0] - win_bottom_y) / (win_top_y - win_bottom_y);
	    break;
	  case OVERLAY_COORD_LINEAR:
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE, 1,
					     dx, dy, dx, dy, NULL, NULL);
	    break;
	  case OVERLAY_COORD_WORLD:
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE, 1,
					     dx, dy, NULL, NULL, dx, dy);
	    break;
	}
    }
    types[0] = binding->olist_ctx->att.first_coord_type;
    if (num_dcoords > 1)
    {
	switch (binding->olist_ctx->att.other_coord_type)
	{
	  case OVERLAY_COORD_PIXEL:
	    break;
	  case OVERLAY_COORD_RELATIVE:
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE,num_dcoords - 1,
					     dx + 1, dy + 1, dx + 1, dy + 1,
					     NULL, NULL);
	    for (count = 1; count < num_dcoords; ++count)
	    {
		dx[count] = (dx[count] - win_left_x) / (win_right_x -
							win_left_x);
		dy[count] = (dy[count] - win_bottom_y) / (win_top_y -
							  win_bottom_y);
	    }
	    break;
	  case OVERLAY_COORD_LINEAR:
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE,num_dcoords - 1,
					     dx + 1, dy + 1, dx + 1, dy + 1,
					     NULL, NULL);
	    break;
	  case OVERLAY_COORD_WORLD:
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE,num_dcoords - 1,
					     dx + 1, dy + 1, NULL, NULL,
					     dx + 1, dy + 1);
	    break;
	}
    }
    for (count = 1; count < num_dcoords; ++count)
	types[count] = binding->olist_ctx->att.other_coord_type;
    switch (object_type)
    {
      case KWIN_EDIT_OBJECT_RECTANGLE:
	overlay_lines (binding->olist_ctx->olist, num_dcoords, types, dx, dy,
		       colour);
	break;
      case KWIN_EDIT_OBJECT_ELLIPSE:
	if ( (types[0] == OVERLAY_COORD_PIXEL) &&
	     (types[1] == OVERLAY_COORD_PIXEL) )
	{
	    /*  Nothing need be done  */
	}
	else if ( (types[0] == OVERLAY_COORD_PIXEL) &&
		  (types[1] == OVERLAY_COORD_LINEAR) )
	{
	    canvas_convert_ellipse_coords (worldcanvas, TRUE, TRUE,
					   dx, dy, dx + 1, dy + 1, dx + 2);
	    dx[0] = x[0];
	    dy[0] = y[0];
	}
	else if ( (types[0] == OVERLAY_COORD_PIXEL) &&
		  (types[1] == OVERLAY_COORD_WORLD) )
	{
	    canvas_convert_ellipse_coords (worldcanvas, TRUE, FALSE,
					   dx, dy, dx + 1, dy + 1, dx + 2);
	    dx[0] = x[0];
	    dy[0] = y[0];
	}
	else if ( (types[0] == OVERLAY_COORD_LINEAR) &&
		  (types[1] == OVERLAY_COORD_PIXEL) )
	{
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE, 1,
					     dx, dy, dx, dy, NULL, NULL);
	}
	else if ( (types[0] == OVERLAY_COORD_LINEAR) &&
		  (types[1] == OVERLAY_COORD_LINEAR) )
	{
	    canvas_convert_ellipse_coords (worldcanvas, TRUE, TRUE,
					   dx, dy, dx + 1, dy + 1, dx + 2);
	}
	else if ( (types[0] == OVERLAY_COORD_LINEAR) &&
		  (types[1] == OVERLAY_COORD_WORLD) )
	{
	    canvas_convert_ellipse_coords (worldcanvas, TRUE, FALSE,
					   dx, dy, dx + 1, dy + 1, dx + 2);
	    dx[0] = x[0];
	    dy[0] = y[0];
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE, 1,
					     dx, dy, dx, dy, NULL, NULL);
	}
	else if ( (types[0] == OVERLAY_COORD_WORLD) &&
		  (types[1] == OVERLAY_COORD_PIXEL) )
	{
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE, 1,
					     dx, dy, NULL, NULL, dx, dy);
	}
	else if ( (types[0] == OVERLAY_COORD_WORLD) &&
		  (types[1] == OVERLAY_COORD_LINEAR) )
	{
	    canvas_convert_ellipse_coords (worldcanvas, TRUE, TRUE,
					   dx, dy, dx + 1, dy + 1, dx + 2);
	    dx[0] = x[0];
	    dy[0] = y[0];
	    canvas_convert_to_canvas_coords (worldcanvas, TRUE, 1,
					     dx, dy, NULL, NULL, dx, dy);
	}
	else if ( (types[0] == OVERLAY_COORD_WORLD) &&
		  (types[1] == OVERLAY_COORD_WORLD) )
	{
	    canvas_convert_ellipse_coords (worldcanvas, TRUE, FALSE,
					   dx, dy, dx + 1, dy + 1, dx + 2);
	}
	else
	{
	    fprintf (stderr, "%s: co-ordinate types not supported\n",
		     function_name);
	    return (TRUE);
	}
	canvas_resize (worldcanvas, NULL, FALSE);
	overlay_arc (binding->olist_ctx->olist,
		     types[0], dx[0], dy[0], types[1], dx[1], dy[1],
		     dx[2] + 180.0, colour, FALSE);
	break;
      case KWIN_EDIT_OBJECT_POLYGON:
	overlay_lines (binding->olist_ctx->olist, num_dcoords, types, dx, dy,
		       colour);
	break;
    }
    return (TRUE);
}   /*  End Function _overlay_editor_geom_func  */

static void _overlay_editor_wcanvas_destroy_func (struct bind_struct *binding)
/*  [SUMMARY] Process the destruction of a world canvas.
    <binding> The binding.
    [RETURNS] Nothing.
*/
{
    KOverlayListEditorContext context = binding->olist_ctx;
    static char function_name[] = "_overlay_editor_wcanvas_destroy_func";

    VERIFY_CONTEXT (context);
    if (binding->next != NULL) binding->next->prev = binding->prev;
    if (binding->prev != NULL) binding->prev->next = binding->next;
    if (binding == context->first_binding)
	context->first_binding = binding->next;
    m_free (binding);
}   /*  End Function _overlay_editor_wcanvas_destroy_func  */
