/*LINTLIBRARY*/
/*  read_ascii.c

    This code provides ASCII reading for KOverlayList objects.

    Copyright (C) 1997-2001 Richard Gooch  API design and general maintenance
    Copyright (C) 1997-1998 Steven Gibson  Parsing code (the interesting stuff)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Steven Gibson may be reached my email at  gibson@ras.ucalgary.ca
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for reading ASCII files and adding
    to overlay lists.


    Written by      Richard Gooch   9-AUG-1997: Moved <overlay_read_ascii> from
  OverlayMenu widget.

    Updated by      Richard Gooch   25-NOV-1997: Added call to <ext_call> so
  that Steve can experiment with having to bother with linking to <kview>.

    Updated by      Richard Gooch   23-DEC-1997: Merged in Steve's latest code.
  Declared internal functions static.

    Updated by      Richard Gooch   13-FEB-1998: Merged in Steve's latest code.
  Declared internal functions static.

    Updated by      Richard Gooch   5-MAR-1998: Merged in Steve's latest code.

    Updated by      Richard Gooch   30-MAR-1998: Merged in Steve's latest code.
  Added variable expansion of filenames.

    Updated by      Richard Gooch   31-MAR-1998: Merged in Steve's latest code.

    Updated by      Richard Gooch   28-MAY-1998: Always use <overlay_arc>
  instead of <overlay_ellipse>.

    Updated by      Richard Gooch   3-NOV-1998: Removed definition of
  sys_errlist.

    Updated by      Richard Gooch   8-JUN-1999: Took account of change to
  <st_expr_expand>.

    Updated by      Richard Gooch   18-AUG-1999: Changed default font from
  "fixed" to "hershey14".

    Updated by      Richard Gooch   18-OCT-1999: Fixed call to <ch_getl>:
  length given was less than string length.

    Last updated by Antoine Bouchard 18-JUN-2001: Added support for "filled"
  option for circle and ellipse commands.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#define NEW_WIN_SCALE
#include <karma_overlay.h>
#include <karma_chs.h>
#include <karma_ext.h>
#include <karma_ch.h>
#include <karma_st.h>


STATIC_FUNCTION (flag read_ascii, (KOverlayList olist, CONST char *filename) );


/*EXPERIMENTAL_FUNCTION*/
flag overlay_read_ascii (KOverlayList olist, CONST char *filename)
/*  [SUMMARY] Read overlay instructions from an ASCII file.
    [PURPOSE] This routine will read overlay instructions from an ASCII file
    and will add the instructions to an overlay list.
    <olist> The KOverlayList to add instructions to.
    <filename> The name of the ASCII file to read instructions from.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int num_ok;
    /*static char function_name[] = "overlay_read_ascii";*/

    if (ext_call ("overlay", "read_ascii", 1, &num_ok, olist, filename) > 0)
    {
	/*  Managed to call an extension function  */
	return (num_ok > 0) ? TRUE : FALSE;
    }
    return read_ascii (olist, filename);
}   /*  End Function overlay_read_ascii  */

	/* A few ASCII definitions */

#define NULL_CHAR		  0
#define END_OF_STRING		  0
#define TAB			  9
#define LINE_FEED		 10
#define CARRIAGE_RETURN		 13
#define ESC			 27
#define SPACE			 32
#define COMMA			 44

	/* Math definitions */

#ifndef PI
#define PI			3.14159265358979323846
#endif
#define DEG2RAD		  	(PI/180.0)
#define RAD2DEG		  	(180.0/PI)

	/* annotation command descriptor values */

#define AC_BAD			-1	/* Illegal command */

#define AC_INCLUDE		 0	/* Include given file contents */

#define AC_COLOR		10	/* Change color attribute */
#define AC_FONT			11	/* Change text font attribute */
#define AC_COORD_DEFAULT	12	/* Change default coordinate type */
#define AC_PA_CONVENTION	13	/* Change position angle convention */
#define AC_LTYPE		14	/* Change line type attribute */
#define AC_LWEIGHT		15	/* Change line weight attribute */

#define AC_DOT			20	/* Draw Dot */
#define AC_DOTS			21	/* Draw Dots */
#define AC_LINE			22	/* Draw Line */
#define AC_CLINES		23	/* Draw Connected Line Segments */
#define AC_DLINES		24	/* Draw Disconnected Line Segments */
#define AC_VECTOR		25	/* Draw Vector */
#define AC_VECTOR_ARROW		26	/* Draw Vector with Arrowhead */
#define AC_BOX			27	/* Draw Box described by corners */
#define AC_CBOX			28	/* Draw Box desc. by center, w, h */
#define AC_CROSS		29	/* Draw Cross */
#define AC_CIRCLE		30	/* Draw Circle */
#define AC_ELLIPSE		31	/* Draw Ellipse */

#define AC_TEXT			40	/* Draw Text String */

#define AC_RGB_BOX		50	/* Draw RGB transfer function box */

	/* position angle convention flag values */

#define STANDARD_PA		 0	/*(X>0,Y=0) = 0deg, (X=0,Y>0) = 90deg*/
#define SKY_PA			 1	/* (UP) = 0 deg, (LEFT) = 90 deg */

	/* extra overlay coordinate type */

#define OVERLAY_COORD_BAD	99	/* Invalid Overlay Coordinate Type */

	/* maximum total length of a command line (including continued parts)*/

#define MAX_LINE_LENGTH		(20*STRING_LENGTH)

	/* maximum number of tokens allowed in a command line */

#define MAX_NTOKENS	        (MAX_LINE_LENGTH/2)


STATIC_FUNCTION (int trim_line, (char line[], flag *LINE_CONTINUED) );
STATIC_FUNCTION (int parse_line, (char line[], int token[]) );
STATIC_FUNCTION (flag wspace, (char c) );
STATIC_FUNCTION (double convert_sky_pa_to_standard_pa,
		 (unsigned int ctype, double pa_sky) );
STATIC_FUNCTION (unsigned int getct, (char c) );


/* Attributes are implemented as GLOBAL VARIABLES to make them common between
   any loaded annotation files (including those included by others).  They are 
   initialized here outside the parser.  As a result, this initialization is 
   only carried out when the application first starts up. */

char 		overlay_color_str[STRING_LENGTH] = "green",
     		overlay_font_str[STRING_LENGTH]  = "hershey14", 
     		overlay_ltype_str[STRING_LENGTH] = "solid";
unsigned int 	OVERLAY_CTYPE_DEFAULT 		 = OVERLAY_COORD_WORLD;
flag 		OVERLAY_PA_CONVENTION 		 = STANDARD_PA;
double 		overlay_lweight 		 = 1.0;

  /*  Note: A list of possible colors is given in /usr/lib/X11/rgb.txt.
            The command "xlsfonts" lists all available X fonts. 
	    LTYPE and LWEIGHT attributes are not yet supported.  */

static flag read_ascii (KOverlayList olist, CONST char *filename)
/*  [SUMMARY] Read overlay instructions from an ASCII file.
    [PURPOSE] This routine will read overlay instructions from an ASCII file
    and will add the instructions to an overlay list.
    <olist> The KOverlayList to add instructions to.
    <filename> The name of the ASCII file to read instructions from.
    [RETURNS] TRUE on success, else FALSE.
*/
{
  flag GOOD_PARSE, CONTINUE_PARSE, LINE_CONTINUED, /*OLD_LINE_CONTINUED, */
	END_REACHED, COMMAND, ALLCAPS, FILLED;
  unsigned int ctype1 = 0;
  unsigned int ctype2 = 0;
  unsigned int ctypes_1[MAX_NTOKENS], ctypes_2[MAX_NTOKENS];
  Channel inp;
  /*static char function_name[] = "read_ascii";*/

  char c, ct1, ct2, line[MAX_LINE_LENGTH], file_line[MAX_LINE_LENGTH], 
	junk[STRING_LENGTH], tmp_color_str[STRING_LENGTH], 
	text_str[STRING_LENGTH], included_file_name[MAX_LINE_LENGTH];
  int i, j, t, tc, ntokens, token[MAX_NTOKENS], 
	len_line, len_file_line, vertex, nvertices, nsegments;
  double x1, x2, y1, y2, xc, yc, xw, yw, xr, yr, pa, pa_rad, r, 
	x_arr_1[MAX_NTOKENS], y_arr_1[MAX_NTOKENS], 
	x_arr_2[MAX_NTOKENS], y_arr_2[MAX_NTOKENS];

  if ((inp = ch_open_file (filename, "r")) == NULL) {
    fprintf (stderr, "read_ascii: error opening file: \"%s\"\t%s\n",
	     filename, ERRSTRING);
    return (FALSE);
  }

  /* read file line by line until reach end */
  END_REACHED = FALSE;
  while (END_REACHED == FALSE) {
    file_line[0] = line[0] = END_OF_STRING;
    len_line = len_file_line = 0;
    LINE_CONTINUED = /*OLD_LINE_CONTINUED =*/ FALSE;
    FILLED = FALSE;

    /* read & join multiple file lines if continued */
    do {
      if ( !ch_getl (inp, file_line, MAX_LINE_LENGTH) ) {
        if (file_line[0] == '\0') END_REACHED = TRUE;
      }
      /*OLD_LINE_CONTINUED = LINE_CONTINUED;*/
      if (END_REACHED == FALSE) {
#ifdef DEBUG
	fprintf (stderr, "raw     : \"%s\"\n", file_line);
#endif
	len_file_line = trim_line (file_line, &LINE_CONTINUED);
#ifdef DEBUG
	fprintf (stderr, "trimmed : \"%s\"\n", file_line);
#endif
      } else {
	LINE_CONTINUED = FALSE;
      }
      if (len_line + len_file_line > MAX_LINE_LENGTH - 1) {
	len_file_line = MAX_LINE_LENGTH - 1 - len_line;
	fprintf (stderr, "read_ascii: string length exceeds maximum");
	fprintf (stderr, " of %d for line \"%s%s...\"\n", MAX_LINE_LENGTH-1,
		line, file_line);
	fprintf (stderr, "Will attempt to parse line as far as possible.\n");
	LINE_CONTINUED = FALSE;
      }
      for (i=0; i<len_file_line; i++) line[len_line+i] = file_line[i];
      len_line += len_file_line;
      line[len_line] = END_OF_STRING;
    } while (LINE_CONTINUED == TRUE);

    /* if completed line is empty, skip to next */
    if (len_line == 0) {
/*
      fprintf (stderr, "empty or comment line detected\n");
*/
      continue;
    }

    /* parse completed line into tokens */
    ntokens = parse_line (line, token);

    /* make command lower case for identification */
    if (ntokens > 1) {
      j = token[1];
    } else {
      j = len_line;
    }
    for (i=0; i<j; i++) {
      if (line[i] >= 'A' && line[i] <= 'Z') line[i] += 'a' - 'A';
    }

    /* identify command */
    COMMAND = AC_BAD;
    if (   strncmp (line, "include"     , 7) == 0) COMMAND = AC_INCLUDE;
    if (   strncmp (line, "color"       , 5) == 0 
	|| strncmp (line, "colour"      , 6) == 0) COMMAND = AC_COLOR;
    if (   strncmp (line, "font"        , 4) == 0) COMMAND = AC_FONT;
    if (   strncmp (line, "coord"       , 5) == 0) COMMAND = AC_COORD_DEFAULT;
    if (   strncmp (line, "pa"          , 2) == 0) COMMAND = AC_PA_CONVENTION;
    if (   strncmp (line, "ltype"       , 5) == 0) COMMAND = AC_LTYPE;
    if (   strncmp (line, "lweight"     , 7) == 0) COMMAND = AC_LWEIGHT;
    if (   strncmp (line, "dot"         , 3) == 0 
        || strncmp (line, "point"       , 5) == 0) {
      if (   strncmp (line, "dots"      , 4) == 0 
          || strncmp (line, "points"    , 6) == 0) {
						   COMMAND = AC_DOTS;
      } else {
						   COMMAND = AC_DOT;
      }
    }
    if (   strncmp (line, "line"        , 4) == 0) COMMAND = AC_LINE;
    if (   strncmp (line, "clines"      , 6) == 0) COMMAND = AC_CLINES;
    if (   strncmp (line, "dlines"      , 6) == 0) COMMAND = AC_DLINES;
    if (   strncmp (line, "vector"      , 6) == 0) COMMAND = AC_VECTOR;
    if (   strncmp (line, "vector_arrow",12) == 0
	|| strncmp (line, "avector"     , 7) == 0
	|| strncmp (line, "arrow"       , 5) == 0) COMMAND = AC_VECTOR_ARROW;
    if (   strncmp (line, "box"         , 3) == 0) COMMAND = AC_BOX;
    if (   strncmp (line, "cbox"        , 4) == 0) COMMAND = AC_CBOX;
    if (   strncmp (line, "cross"       , 5) == 0) COMMAND = AC_CROSS;
    if (   strncmp (line, "circle"      , 6) == 0) COMMAND = AC_CIRCLE;
    if (   strncmp (line, "ellipse"     , 7) == 0) COMMAND = AC_ELLIPSE;
    if (   strncmp (line, "text"        , 4) == 0) COMMAND = AC_TEXT;
    if (   strncmp (line, "rgb_box"     , 7) == 0) COMMAND = AC_RGB_BOX;

    /* If text command, set first two tokens to lower case but leave actual
       text string as-is; if include, color, or font command, leave all tokens 
       after 1st (command) as-is, unless they contain no lower case letters, 
       in which case, they should be de-capitalized; if any other command, set 
       everything to lower case. 

       NOTE: X seems to accept color names even if they are the wrong case.
       However I'll leave this mechanism in place in case some other system
       is less forgiving. */

    if (ntokens > 1) {
      ALLCAPS = TRUE;
      for (i=token[1]; i<len_line; i++) {
	if (line[i] >= 'a' && line[i] <= 'z') ALLCAPS = FALSE;
      }
      if (COMMAND == AC_TEXT && ntokens > 2) {
	j = token[3];
      } else {
	if (   (   COMMAND == AC_INCLUDE 
	        || COMMAND == AC_COLOR 
	        || COMMAND == AC_FONT) 
	    && ALLCAPS == FALSE) {
	  j = token[1];
	} else {
	  j = len_line;
	}
      }
      for (i=token[1]; i<j; i++) {
	if (line[i] >= 'A' && line[i] <= 'Z') line[i] += 'a' - 'A';
      }
    }

    /* find token with 1st coordinate value, if possible */
    tc = 0;
    if (COMMAND >= AC_DOT && ntokens >= 2) {
      /* if token[1] is coord type, set 1st ctype; 
	 set tc to 1st presumed coord token, based on result */
      if ((ctype1 = getct(line[token[1]])) != OVERLAY_COORD_BAD) {
	tc = 2;
      } else {
	ctype1 = OVERLAY_CTYPE_DEFAULT;
	tc = 1;
      }
    }

    /* Convert all "white space" (including commas) into actual space
	characters if it occurs between coordinate tokens.  This is
	a kludge to allow stuff like sscanf ("%le %le", &x, &y) to work 
	when there are commas between the numbers in the string. */
    if (tc > 0) {
      if (COMMAND == AC_TEXT && ntokens-1 >= tc+2) {
	j = token[tc+2];
      } else {
	j = len_line;
      }
      for (i=token[tc]; i<j; i++) {
	if (wspace (line[i]) == TRUE) line[i] = SPACE;
      }
    }

/*
    fprintf (stderr, "complete: \"%s\"\n", line);
    for (t=0; t<ntokens; t++) {
      fprintf (stderr, "token %3d = %3d = \"", t, token[t]);
      if (t+1<ntokens) {
	j = token[t+1];
      } else {
	j = len_line;
      }
      for (i=token[t]; i<j; i++) fprintf (stderr, "%c", line[i]);
      printf ("\"\n");
    }
*/

    GOOD_PARSE = FALSE;

    /************ Non-Graphical Commands ************/

    if (COMMAND == AC_INCLUDE && ntokens >= 2) {
      strcpy (included_file_name, &line[token[1]]);
      GOOD_PARSE = TRUE;
      /* try to trap recursive includes -- no attempt made to consider path */
      if (strcmp (filename, included_file_name) != 0) {
	st_expr_expand (included_file_name, MAX_LINE_LENGTH,
			included_file_name, NULL, NULL, stderr);
	read_ascii (olist, included_file_name);
      } else {
	fprintf (stderr, "read_ascii: annotation file %s", included_file_name);
	fprintf (stderr, " is not allowed to include itself.\n");
      }
    }

    /************* Attribute Commands ***************/

    if (COMMAND == AC_COLOR && ntokens >= 2) {
      strcpy (overlay_color_str, &line[token[1]]);
      GOOD_PARSE = TRUE;
    }
    if (COMMAND == AC_FONT && ntokens >= 2) {
      strcpy (overlay_font_str, &line[token[1]]);
      GOOD_PARSE = TRUE;
    }
    if (COMMAND == AC_COORD_DEFAULT && ntokens >= 2) {
      if ((ctype1 = getct(line[token[1]])) != OVERLAY_COORD_BAD) {
	GOOD_PARSE = TRUE;
	OVERLAY_CTYPE_DEFAULT = ctype1;
      }
    }
    if (COMMAND == AC_PA_CONVENTION && ntokens >= 2) {
      if (strncmp (&line[token[1]], "standard", 8) == 0) {
	OVERLAY_PA_CONVENTION = STANDARD_PA;
	GOOD_PARSE = TRUE;
      } else {
	if (strncmp (&line[token[1]], "sky", 3) == 0) {
	  OVERLAY_PA_CONVENTION = SKY_PA;
	  GOOD_PARSE = TRUE;
	}
      }
    }
    if (COMMAND == AC_LTYPE && ntokens >= 2) {
      if (sscanf (&line[token[1]], "%s", overlay_ltype_str) == 1) {
        fprintf (stderr, "read_ascii: LTYPE command not available.\n");
	GOOD_PARSE = TRUE;
      }
    }
    if (COMMAND == AC_LWEIGHT && ntokens >= 2) {
      if (sscanf (&line[token[1]], "%le", &overlay_lweight) == 1) {
	fprintf (stderr, 
		"read_ascii: LWEIGHT command not available.\n");
	GOOD_PARSE = TRUE;
      }
    }

    /*************** Object Commands ***************/

    /** Complete syntax list:

      DOT     [ct] <x> <y>
      DOTS    [ct] <x1> <y1> [[ct] <x2> <y2>] ...
      LINE    [ct] <x1> <y1> [ct] <x2> <y2>
      CLINES  [ct] <x1> <y1> [ct] <x2> <y2> [[ct] <x3> <y3>] ...
      DLINES  [ct] <x1> <y1> [ct] <x2> <y2> [[ct] <x3> <y3> [ct] <x4> <y4>] ...
      VECTOR  [ct] <xb> <yb> [ct] <rx> <ry>
      BOX     [ct] <x1> <y1> <x2> <y2>
      CBOX    [ct] <xc> <yc> <wx> <wy> [pa]
      CROSS   [ct] <xc> <yc> <wx> <wy> [pa]
      CIRCLE  [ct] <xc> <yc> <r>
      ELLIPSE [ct] <xc> <yc> <a> <b> [pa]
      TEXT    [ct] <x> <y> <string including white space>

     **/

    if (COMMAND == AC_TEXT && ntokens-tc >= 3) {
      if (sscanf (&line[token[tc]], "%le %le %s", &x1, &y1, junk) == 3) {
	GOOD_PARSE = TRUE;
	strcpy (text_str, &line[token[tc+2]]);
	overlay_text (olist, text_str, ctype1, x1, y1, overlay_color_str,
		      overlay_font_str, FALSE);
      }
    }
    if (COMMAND == AC_DOT && ntokens-tc >= 2) {
      if (sscanf (&line[token[tc]], "%le %le", &x1, &y1) == 2) {
	GOOD_PARSE = TRUE;
	overlay_line (olist, ctype1, x1, y1, ctype1, x1, y1,overlay_color_str);
      }
    }
    if ((COMMAND == AC_LINE || COMMAND == AC_VECTOR) && ntokens-tc >= 4) {
      if ((ctype2 = getct(line[token[tc+2]])) != OVERLAY_COORD_BAD) {
	if (   ntokens >= 5 
	    && sscanf (&line[token[tc]], "%le %le %c %le %le", 
				&x1, &y1, &ct2, &x2, &y2) == 5) {
	  GOOD_PARSE = TRUE;
	}
      } else {
        if (sscanf (&line[token[tc]],"%le %le %le %le",&x1,&y1,&x2,&y2) == 4) {
	  GOOD_PARSE = TRUE;
	  ctype2 = ctype1;
	}
      }
      if (GOOD_PARSE == TRUE) {
	if (COMMAND == AC_LINE) {
	  overlay_line (olist, ctype1, x1, y1, ctype2,x2,y2,overlay_color_str);
	} else {
	  overlay_vector (olist, ctype1, x1,y1,ctype2,x2,y2,overlay_color_str);
	}
      }
    }
    if (COMMAND == AC_VECTOR_ARROW) {
    }
    if (COMMAND == AC_BOX && ntokens-tc >= 4) {
      if (sscanf (&line[token[tc]], "%le %le %le %le", 
				&x1, &y1, &x2, &y2) == 4) {
	GOOD_PARSE = TRUE;
	for (i=0; i<=4; i++) ctypes_1[i] = ctype1;
	x_arr_1[0] = x1;  	y_arr_1[0] = y1;
	x_arr_1[1] = x2;  	y_arr_1[1] = y1;
	x_arr_1[2] = x2;  	y_arr_1[2] = y2;
	x_arr_1[3] = x1;  	y_arr_1[3] = y2;
	x_arr_1[4]=x_arr_1[0];  y_arr_1[4]=y_arr_1[0];
	overlay_lines (olist, 5, ctypes_1, x_arr_1, y_arr_1,overlay_color_str);
      }
    }
    if ((COMMAND == AC_CBOX || COMMAND == AC_CROSS) && ntokens-tc >= 4) {
      if (sscanf (&line[token[tc]], "%le %le %le %le %le", 
				&xc, &yc, &xw, &yw, &pa) == 5) {
	GOOD_PARSE = TRUE;
      } else {
	if (sscanf (&line[token[tc]],"%le %le %le %le",&xc,&yc,&xw,&yw) == 4) {
	  GOOD_PARSE = TRUE;
	  pa = 0.0;
	}	
      }
      if (GOOD_PARSE == TRUE) {
	xr = 0.5 * xw;
	yr = 0.5 * yw;
	pa_rad = pa * DEG2RAD;
	if (OVERLAY_PA_CONVENTION == SKY_PA) {
	  pa_rad = convert_sky_pa_to_standard_pa (ctype1, pa_rad);
	}
	if (COMMAND == AC_CBOX) {
	  for (i=0; i<=4; i++) ctypes_1[i] = ctype1;
	  x_arr_1[0] = xc - xr*cos(pa_rad) + yr*sin(pa_rad);
	  y_arr_1[0] = yc - xr*sin(pa_rad) - yr*cos(pa_rad);
	  x_arr_1[1] = xc + xr*cos(pa_rad) + yr*sin(pa_rad);
	  y_arr_1[1] = yc + xr*sin(pa_rad) - yr*cos(pa_rad);
	  x_arr_1[2] = xc + xr*cos(pa_rad) - yr*sin(pa_rad);
	  y_arr_1[2] = yc + xr*sin(pa_rad) + yr*cos(pa_rad);
	  x_arr_1[3] = xc - xr*cos(pa_rad) - yr*sin(pa_rad);
	  y_arr_1[3] = yc - xr*sin(pa_rad) + yr*cos(pa_rad);
	  x_arr_1[4] = x_arr_1[0];			
	  y_arr_1[4] = y_arr_1[0];
	  overlay_lines (olist, 5, ctypes_1,x_arr_1,y_arr_1,overlay_color_str);
	} else {
	  ctypes_1[0] = ctypes_1[1] = ctype1;
	  x_arr_1[0] = xc - xr*cos(pa_rad);  y_arr_1[0] = yc - xr*sin(pa_rad);
	  x_arr_2[0] = xc + xr*cos(pa_rad);  y_arr_2[0] = yc + xr*sin(pa_rad);
	  x_arr_1[1] = xc - yr*sin(pa_rad);  y_arr_1[1] = yc + yr*cos(pa_rad);
	  x_arr_2[1] = xc + yr*sin(pa_rad);  y_arr_2[1] = yc - yr*cos(pa_rad);
	  overlay_segments (olist, 2, ctypes_1, x_arr_1, y_arr_1, 
			    ctypes_1, x_arr_2, y_arr_2, overlay_color_str);
	}
      }
    }
    if (COMMAND == AC_CIRCLE && ntokens-tc >= 3) {
      if (sscanf (&line[token[tc]], "%le %le %le", &xc, &yc, &r) == 3) {
	GOOD_PARSE = TRUE;
	if (ntokens-tc >= 4 && strncmp (&line[token[tc+3]], "filled", 6) == 0)
	    FILLED = TRUE;
	xr = yr = r;
	overlay_arc (olist, ctype1, xc, yc, ctype1, xr, yr, 0.0,
		     overlay_color_str, FILLED);
      }
    }
    if (COMMAND == AC_ELLIPSE && ntokens-tc >= 4) {
      if (sscanf (&line[token[tc]], "%le %le %le %le %le", 
		  &xc, &yc, &xr, &yr, &pa) == 5) {
	GOOD_PARSE = TRUE;
	if (ntokens-tc >= 6 && strncmp (&line[token[tc+5]], "filled", 6) == 0)
	    FILLED = TRUE;
	if (OVERLAY_PA_CONVENTION == SKY_PA) {
	  pa = RAD2DEG * convert_sky_pa_to_standard_pa (ctype1, pa*DEG2RAD);
	}
	overlay_arc (olist, ctype1, xc, yc, ctype1, xr, yr, pa,
		     overlay_color_str, FILLED);
      } else {
	if (sscanf (&line[token[tc]],"%le %le %le %le",&xc,&yc,&xr,&yr) == 4) {
	  GOOD_PARSE = TRUE;
	  if (ntokens-tc >= 5 && strncmp (&line[token[tc+4]], "filled", 6) == 0)
	      FILLED = TRUE;
	  overlay_arc (olist, ctype1, xc, yc, ctype1, xr, yr, 0.0,
		       overlay_color_str, FILLED);
	}
      }
    }
    if (COMMAND == AC_CLINES && ntokens-tc >= 4) {
      CONTINUE_PARSE = GOOD_PARSE = TRUE;
      vertex = 0;
      t = 1;
      do {
	if (sscanf (&line[token[t]], "%le %le", &x1, &y1) == 2) {
	  if (vertex > 0) {
	    ctypes_1[vertex] = ctypes_1[vertex-1];
	  } else {
	    ctypes_1[vertex] = ctype1;
	  }
	  x_arr_1[vertex] = x1;
	  y_arr_1[vertex] = y1;
	  t += 2;
	} else {
	  if (sscanf (&line[token[t]], "%c %le %le", &ct1, &x1, &y1) == 3) {
	    if ((ctypes_1[vertex]=getct(ct1)) != OVERLAY_COORD_BAD) {
	      x_arr_1[vertex] = x1;
	      y_arr_1[vertex] = y1;
	      t += 3;
	    }
	  } else {
	    CONTINUE_PARSE = FALSE;
	  }
	}
	if (GOOD_PARSE == TRUE) vertex++;
	if (t >= ntokens) CONTINUE_PARSE = FALSE;
      } while (CONTINUE_PARSE == TRUE);
      nvertices = vertex;
      if (GOOD_PARSE == TRUE) {
	overlay_lines (olist, nvertices, ctypes_1, x_arr_1, y_arr_1, 
							overlay_color_str);
      }
    }
    if (COMMAND == AC_DLINES && ntokens-tc >= 4) {

      /* first accumulate list of vertices */
      CONTINUE_PARSE = GOOD_PARSE = TRUE;
      vertex = 0;
      t = 1;
      do {
	if (sscanf (&line[token[t]], "%le %le", &x1, &y1) == 2) {
	  if (vertex > 0) {
	    ctypes_1[vertex] = ctypes_1[vertex-1];
	  } else {
	    ctypes_1[vertex] = ctype1;
	  }
	  x_arr_1[vertex] = x1;
	  y_arr_1[vertex] = y1;
	  t += 2;
	} else {
	  if (sscanf (&line[token[t]], "%c %le %le", &ct1, &x1, &y1) == 3) {
	    if ((ctypes_1[vertex]=getct(ct1)) != OVERLAY_COORD_BAD) {
	      x_arr_1[vertex] = x1;
	      y_arr_1[vertex] = y1;
	      t += 3;
	    }
	  } else {
	    CONTINUE_PARSE = FALSE;
	  }
	}
	if (GOOD_PARSE == TRUE) vertex++;
	if (t >= ntokens) CONTINUE_PARSE = FALSE;
      } while (CONTINUE_PARSE == TRUE);
      nvertices = vertex;

      /* then separate alternate vertices into two lists & pass arguments */
      if ((nsegments = nvertices / 2) * 2 == nvertices) {
	for (i=0; i<nsegments; i++) {
	  ctypes_1[i] = ctypes_1[2*i];  ctypes_2[i] = ctypes_1[2*i+1];
	  x_arr_1[i]  = x_arr_1[2*i];   x_arr_2[i]  = x_arr_1[2*i+1];
	  y_arr_1[i]  = y_arr_1[2*i];   y_arr_2[i]  = y_arr_1[2*i+1];
	}
	if (GOOD_PARSE == TRUE) {
	  overlay_segments (olist, nsegments, ctypes_1, x_arr_1, y_arr_1, 
				ctypes_2, x_arr_2, y_arr_2, overlay_color_str);
	}
      } else {
	fprintf (stderr, "read_ascii: last line in DLINES list\n");
	fprintf (stderr, "\tis not fully specified.\n");
      }
    }
    if (COMMAND == AC_DOTS && ntokens-tc >= 2) {

      /* accumulate list of vertices */
      CONTINUE_PARSE = GOOD_PARSE = TRUE;
      vertex = 0;
      t = 1;
      do {
	if (sscanf (&line[token[t]], "%le %le", &x1, &y1) == 2) {
	  if (vertex > 0) {
	    ctypes_1[vertex] = ctypes_1[vertex-1];
	  } else {
	    ctypes_1[vertex] = ctype1;
	  }
	  x_arr_1[vertex] = x1;
	  y_arr_1[vertex] = y1;
	  t += 2;
	} else {
	  if (sscanf (&line[token[t]], "%c %le %le", &ct1, &x1, &y1) == 3) {
	    if ((ctypes_1[vertex]=getct(ct1)) != OVERLAY_COORD_BAD) {
	      x_arr_1[vertex] = x1;
	      y_arr_1[vertex] = y1;
	      t += 3;
	    }
	  } else {
	    CONTINUE_PARSE = FALSE;
	  }
	}
	if (GOOD_PARSE) vertex++;
	if (t >= ntokens) CONTINUE_PARSE = FALSE;
      } while (CONTINUE_PARSE == TRUE);
      nvertices = vertex;

      /* copy vertex list & pass arguments */
      for (i=0; i<nvertices; i++) {
	ctypes_2[i] = ctypes_1[i];
	x_arr_2[i]  = x_arr_1[i];
	y_arr_2[i]  = y_arr_1[i];
      }
      if (GOOD_PARSE == TRUE) {
	overlay_segments (olist, nvertices, ctypes_1, x_arr_1, y_arr_1, 
				ctypes_2, x_arr_2, y_arr_2, overlay_color_str);
      }
    }
    if (COMMAND == AC_RGB_BOX) {
      /* required argument format:
		RGB_BOX <xc> <yc> <xw> <yw> 
		<char_token (R)> <xr1> <yr1> <xr2> <yr2> [xr3> <yr3> ...]
		<char_token (G)> <xg1> <yg1> <xg2> <yg2> [xg3> <yg3> ...]
		<char_token (B)> <xb1> <yb1> <xb2> <yb2> [xb3> <yb3> ...] */
      if (ntokens >= 20) {
	if (sscanf (line, "%s %le %le %le %le", 
				junk, &xc, &yc, &xw, &yw) == 5) {
	  ctype1 = OVERLAY_COORD_RELATIVE;
	  xr = 0.5 * xw;
	  yr = 0.5 * yw;
	  t = 5;
	  for (i=1; i<=3; i++) {
	    if (   sscanf (&line[token[t]], "%s", tmp_color_str) == 1
		&& ((c=tmp_color_str[0]) == 'r' || c == 'g' || c == 'b')
		&& sscanf (&line[token[t+1]], "%le", &x1) == 1
		&& sscanf (&line[token[t+2]], "%le", &y1) == 1
		&& sscanf (&line[token[t+3]], "%le", &x2) == 1
		&& sscanf (&line[token[t+4]], "%le", &y2) == 1) {
	      vertex = 0;
	      t++;
	      while (   sscanf (&line[token[t]], "%le %le", &x1, &y1) == 2
		     && t < ntokens-1) {
		x_arr_1[vertex]  = xc - xr + x1 * xw;
		y_arr_1[vertex]  = yc - yr + y1 * yw;
		ctypes_1[vertex] = ctype1;
		vertex++;
		t += 2;
	      }
	      nvertices = vertex;
	      switch (i) {
	        case 1: strcpy (tmp_color_str, "red");   break;
	        case 2: strcpy (tmp_color_str, "green"); break;
	        case 3: strcpy (tmp_color_str, "blue");  break;
	      }
	      overlay_lines (olist, nvertices, ctypes_1, x_arr_1, y_arr_1, 
								tmp_color_str);
	    } else {
	      GOOD_PARSE = FALSE;
	      break;
	    }
	  }
	  ctypes_1[0] = ctype1;  x_arr_1[0] = xc-xr;  y_arr_1[0] = yc-yr;
	  ctypes_1[1] = ctype1;  x_arr_1[1] = xc+xr;  y_arr_1[1] = yc-yr;
	  ctypes_1[2] = ctype1;  x_arr_1[2] = xc+xr;  y_arr_1[2] = yc+yr;
	  ctypes_1[3] = ctype1;  x_arr_1[3] = xc-xr;  y_arr_1[3] = yc+yr;
	  ctypes_1[4] = ctype1;  x_arr_1[4]=x_arr_1[0];  y_arr_1[4]=y_arr_1[0];
	  overlay_lines (olist, 5, ctypes_1, x_arr_1, y_arr_1, "gray");
	  GOOD_PARSE = TRUE;
	}
      }
    }
    if (GOOD_PARSE == FALSE) {
      fprintf (stderr, "read_ascii: unable to parse string \"%s\"\n",
	       							line);
    }
    for (i=0; i<STRING_LENGTH; i++) line[i] = END_OF_STRING;
  }
  ch_close (inp);
  return (TRUE);
}   /*  End Function read_ascii  */

/*************************************************************************/
static int trim_line (
	char line[], 		/* text string to be parsed */              
	flag *LINE_CONTINUED)	/* flag indicating this line is continued */
{
  int 	i, j, k, 	/* counters */
	slen; 		/* string length */
  flag	CONTINUATION;	/* flag */

  /* trim comments from string (defined as '#' and anything following) */
  slen = 0;
  while (   slen < MAX_LINE_LENGTH 	/* stay within array boundary */
	 && line[slen] != END_OF_STRING	/* end-of-string marker */
	 && line[slen] != '#'		/* '#' is comment character */
	) slen++;
  line[slen] = END_OF_STRING;
/*
  fprintf (stderr, "trim 1  : \"%s\"  slen=%d\n", line, slen);
*/

  /* trim any leading or trailing white space */
  if (slen > 0) {
    i = 0;
    while (i < slen && wspace(line[i]) == TRUE) i++;
    /* final value of i marks first instance of non-whitespace character */
    j = slen-1;
    while (j > i && wspace(line[j]) == TRUE) j--;
    /* final value of j marks last instance of non-whitespace character */
    for (k=i; k<=j; k++) line[k-i] = line[k];  
    slen = j+1 - i;
    line[slen] = END_OF_STRING;
  }
/*
  fprintf (stderr, "trim 2  : \"%s\"  slen=%d\n", line, slen);
*/

  /* recompute string length to chop at continuation character if present */
  slen = 0;
  while (   slen < MAX_LINE_LENGTH 	/* stay within array boundary */
	 && line[slen] != END_OF_STRING	/* end-of-string marker */
	 && line[slen] != '\\'		/* '\' is continuation character */
	) slen++;

  CONTINUATION = FALSE;
  if (line[slen] == '\\') {
    CONTINUATION = TRUE;
    line[slen] = END_OF_STRING;
  }

  *LINE_CONTINUED = CONTINUATION;
  return slen;		/* a return string length; 0 indicates a blank line */
}

/*************************************************************************/
static int parse_line (
	char line[], 		/* text string to be parsed */              
	int token[]) 		/* index location of each token found */    
{
  int 	i, 		/* counter */
	slen, 		/* string length */
	ntokens;	/* number of tokens found */
  flag	WHITE_SPACE,	/* flags */
	OLD_WHITE_SPACE;  

  /* find line length */
  slen = 0;
  while (   slen < MAX_LINE_LENGTH 	/* stay within array boundary */
	 && line[slen] != END_OF_STRING	/* end-of-string marker */
	) slen++;

  ntokens = 0;
  WHITE_SPACE = TRUE;		/* imaginary white space before string */
  for (i=0; i<slen; i++) {
    OLD_WHITE_SPACE = WHITE_SPACE;
    WHITE_SPACE = wspace(line[i]);
    /* detect beginning of non-whitespace region */
    if (WHITE_SPACE == FALSE && OLD_WHITE_SPACE == TRUE) {
      if (ntokens >= MAX_NTOKENS) {
	fprintf (stderr, "parse_line: Warning: maximum number of tokens");
	fprintf (stderr, " exceeded; further text ignored.\n");
	break;
      }
      token[ntokens] = i;	/* record token start position */
      ntokens++;		/* count token */
    }
  }
  return ntokens;	/* return number of tokens as function value */
}

/*************************************************************************/
static flag wspace (char c)
{
  flag WHITE_SPACE;

  if (   c == END_OF_STRING || c == SPACE           || c == TAB
      || c == LINE_FEED     || c == CARRIAGE_RETURN || c == COMMA) {
    WHITE_SPACE = TRUE;
  } else {
    WHITE_SPACE = FALSE;
  }
  return WHITE_SPACE;
}

/*************************************************************************/
/** NOTE: all angles are in radians **/
static double convert_sky_pa_to_standard_pa (unsigned int ctype, double pa_sky)
{
  double pa_std;

  /* sky PA is ALWAYS measured CCW from Up (E from N), regardless of ctype */
  switch (ctype) {
    case OVERLAY_COORD_WORLD    :
      /* standard PA is measured  CW from Left on an "astronomical" grid */
      pa_std = 0.5 * PI - pa_sky; 
      break;
    case OVERLAY_COORD_RELATIVE : 
      /* standard PA is measured CCW from Right */
      pa_std = 0.5 * PI + pa_sky; 
      break;
    case OVERLAY_COORD_PIXEL    :
      /* standard PA is measured  CW from Right */
      pa_std = 1.5 * PI - pa_sky; 
      break;
    default			: 
      fprintf (stderr, "convert_sky_pa_to_standard_pa: illegal ctype.\n:");
      pa_std = pa_sky;
  }
  return pa_std;
}

/*************************************************************************/
static unsigned int getct (char c)
{
  unsigned int ctype;

  switch (c) {
    case 'r' : ctype = OVERLAY_COORD_RELATIVE; break;
    case 'w' : ctype = OVERLAY_COORD_WORLD;    break;
    case 'p' : ctype = OVERLAY_COORD_PIXEL;    break;
    default  : ctype = OVERLAY_COORD_BAD;
  }
  return ctype;
}
