/*LINTLIBRARY*/
/*  drag_n_zoom.c

    This code provides ViewableImage objects.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains convenience routines which implement a drag and zoom
    user interface.


    Written by      Richard Gooch   26-MAY-1996

    Updated by      Richard Gooch   4-JUN-1996: Switched to left-right
  bottom-top co-ordinate specification instead of min-max x and y.

    Updated by      Richard Gooch   17-JUN-1996: Made use of new
  <canvas_convert_to_canvas_coords> routine so that linear to non-linear
  transformation could be avoided.

    Updated by      Richard Gooch   21-OCT-1996: Added zoom-in 'i' and zoom-out
  'o' feature.

    Updated by      Richard Gooch   25-OCT-1996: Use fallback when green could
  not be allocated.

    Updated by      Richard Gooch   27-OCT-1996: Compute and display statistics
  when 's' key is pressed during drag.

    Updated by      Richard Gooch   1-JUN-1997: Added pan on left click.

    Updated by      Richard Gooch   5-JUN-1997: Created <viewimg_zoom> function

    Updated by      Richard Gooch   29-JUL-1997: Fixed bug where if no active
  viewable image, events were incorrectly consumed.

    Updated by      Richard Gooch   1-AUG-1997: Made use of new editor context
  in <kwin> package for zoom box.

    Updated by      Richard Gooch   16-SEP-1997: Test for infinite zooms in
  <geom_func>.

    Updated by      Richard Gooch   23-SEP-1997: Made use of new experimental
  VIEWIMG_ATT_FILL_CANVAS attribute in order to reduce "letterboxing" effect.

    Updated by      Richard Gooch   4-FEB-1998: Bug fix in <geom_func> which
  did not trap box of zero size in linear co-ordinates.

    Updated by      Richard Gooch   28-MAR-1998: Bug fixes in <viewimg_zoom>:
  various zoom errors fixed by converting to image pixel co-ordinates early.

    Updated by      Richard Gooch   15-OCT-1998: Made use of new function
  <canvas_warp_pointer>.

    Updated by      Richard Gooch   15-AUG-1999: Support changing magnification
  when in panning mode.

    Updated by      Richard Gooch   2-SEP-1999: Took account of interface
  change to function <canvas_warp_pointer> and discard events.

    Updated by      Richard Gooch   29-SEP-1999: Moved <viewimg_zoom> to main.c
  where more trickery can be done.

    Updated by      Richard Gooch   14-JUN-2000: Display values when 'v' key is
  pressed during drag.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>.

    Updated by      Richard Gooch   12-JUL-2000: Added
  #include <karma_canvas.h>

    Updated by      Richard Gooch   13-JUL-2000: Took account of interface
  change to function <viewimg_values_display>.

    Last updated by Richard Gooch   8-OCT-2001: Fixed documentation links.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>


/*  Private functions  */
STATIC_FUNCTION (flag _viewimg_drag_n_zoom_position_func,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin,
		  unsigned int value_type) );
STATIC_FUNCTION (flag _viewimg_drag_n_zoom_geom_func,
		 (KPixCanvasEditorContext context, KPixCanvas canvas,
		  unsigned int object_type, unsigned int num_coords,
		  CONST int *x, CONST int *y,
		  unsigned int raw_event_code, void *raw_e_info,
		  void **f_info) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void viewimg_create_drag_and_zoom_interface (KWorldCanvas canvas)
/*  [SUMMARY] Create a drag and zoom user interface for a canvas.
    [PURPOSE] This routine will create a drag and zoom user interface for a
    world canvas. When the user presses the left mouse button in the canvas and
    moves the mouse with the button still pressed a rubber-band box will be
    drawn on the canvas. When the left mouse button is released the image in
    the rubber-banded box is zoomed to fill as much of the canvas as is
    possible, depending on the zoom policy.
    <canvas> The world canvas object.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    KPixCanvasEditorContext ctx;
    unsigned int visual;
    unsigned long pixel_value;
    static char function_name[] = "viewimg_create_drag_and_zoom_interface";

    pixcanvas = canvas_get_pixcanvas (canvas);
    viewimg_register_position_event_func (canvas,
					  _viewimg_drag_n_zoom_position_func,
					  NULL);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    if (visual == KWIN_VISUAL_PSEUDOCOLOUR)
    {
	if ( !kwin_get_colour (pixcanvas, "green", &pixel_value,
			       NULL, NULL, NULL) )
	{
	    /*  Can't get green: try white instead  */
	    if ( !kwin_get_colour (pixcanvas, "white", &pixel_value,
				   NULL, NULL, NULL) )
	    {
		/*  Can't even get white, so just pick 0  */
		pixel_value = 0;
	    }
	}
    }
    else
    {
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_PIX_GREEN_MASK, &pixel_value,
			     KWIN_ATT_END);
    }
    /*  Create the dragzoom editor context  */
    ctx = kwin_editor_create_context (pixcanvas,
				      KWIN_EDITOR_CONTEXT_ATT_MOUSE,
				      K_CANVAS_EVENT_LEFT_MOUSE_MASK,
				      KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
				      KWIN_EDIT_OBJECT_RECTANGLE,
				      KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE,
				      pixel_value,
				      KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
				      FALSE,
				      KWIN_EDITOR_CONTEXT_ATT_END);
    if (ctx == NULL) m_abort (function_name, "editor context");
    kwin_editor_register_event_func (ctx, _viewimg_drag_n_zoom_geom_func,
				     canvas);
    canvas_register_destroy_func (canvas, TRUE,
				  kwin_editor_destroy_context, ctx);
}   /*  End Function viewimg_create_drag_and_zoom_interface  */


/*  Private functions follow  */

static flag _viewimg_drag_n_zoom_position_func (ViewableImage vimage,
						double x, double y,
						void *value,
						unsigned int event_code,
						void *e_info, void **f_info,
						double x_lin, double y_lin,
						unsigned int value_type)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transforms_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas canvas;
    flag panned;
    CONST char *string = e_info;
    static char function_name[] = "__viewimg_drag_and_zoom_position_func";

    /*  Discard mouse move events quickly  */
    if (event_code == K_CANVAS_EVENT_POINTER_MOVE) return (FALSE);
    canvas = viewimg_get_worldcanvas (vimage);
    viewimg_get_canvas_attributes (canvas,
				   VIEWIMG_ATT_ENABLE_PANNING, &panned,
				   VIEWIMG_ATT_END);
    if (event_code == K_CANVAS_EVENT_PLAIN_KEY_PRESS)
    {
	if (strcmp (string, "u") == 0)
	{
	    /*  Unzoom  */
	    if (panned)
		viewimg_set_canvas_attributes (canvas,
					       VIEWIMG_ATT_PAN_MAGNIFICATION,1,
					       VIEWIMG_ATT_END);
	    else viewimg_set_canvas_attributes (canvas,
						VIEWIMG_ATT_AUTO_X, TRUE,
						VIEWIMG_ATT_AUTO_Y, TRUE,
						VIEWIMG_ATT_END);
	    /*  Redraw the canvas  */
	    if ( !canvas_resize (canvas, NULL, FALSE) )
	    {
		fprintf (stderr, "%s: error refreshing window\n",
			 function_name);
	    }
	    return (TRUE);
	}
	else if (strcmp (string, "i") == 0)
	{
	    if (panned)
	    {
		unsigned int magnification;

		viewimg_get_canvas_attributes (canvas,
					       VIEWIMG_ATT_PAN_MAGNIFICATION,
					       &magnification,
					       VIEWIMG_ATT_END);
		++magnification;
		viewimg_set_canvas_attributes (canvas,
					       VIEWIMG_ATT_PAN_MAGNIFICATION,
					       magnification,
					       VIEWIMG_ATT_END);
		/*  Redraw the canvas  */
		if ( !canvas_resize (canvas, NULL, FALSE) )
		{
		    fprintf (stderr, "%s: error refreshing window\n",
			     function_name);
		}
		return (TRUE);
	    }
	    /*  Zoom in 2x  */
	    viewimg_zoom (vimage, x_lin, y_lin, FALSE, 2.0);
	    canvas_warp_pointer (canvas, x_lin, y_lin, TRUE, TRUE);
	    return (TRUE);
	}
	else if (strcmp (string, "o") == 0)
	{
	    if (panned)
	    {
		unsigned int magnification;

		viewimg_get_canvas_attributes (canvas,
					       VIEWIMG_ATT_PAN_MAGNIFICATION,
					       &magnification,
					       VIEWIMG_ATT_END);
		if (magnification < 2) return (TRUE);
		--magnification;
		viewimg_set_canvas_attributes (canvas,
					       VIEWIMG_ATT_PAN_MAGNIFICATION,
					       magnification,
					       VIEWIMG_ATT_END);
		/*  Redraw the canvas  */
		if ( !canvas_resize (canvas, NULL, FALSE) )
		{
		    fprintf (stderr, "%s: error refreshing window\n",
			     function_name);
		}
		return (TRUE);
	    }
	    /*  Zoom out 2x  */
	    viewimg_zoom (vimage, 0.5, 0.5, TRUE, 0.5);
	    canvas_warp_pointer (canvas, x_lin, y_lin, TRUE, TRUE);
	    return (TRUE);
	}
	else return (FALSE);
    }
    return (FALSE);
}   /*  End Function _viewimg_drag_n_zoom_position_func  */

static flag _viewimg_drag_n_zoom_geom_func (KPixCanvasEditorContext context,
					    KPixCanvas canvas,
					    unsigned int object_type,
					    unsigned int num_coords,
					    CONST int *x, CONST int *y,
					    unsigned int raw_event_code,
					    void *raw_e_info, void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas worldcanvas = (KWorldCanvas) *f_info;
    ViewableImage vimage;
    unsigned long mask;
    CONST char *string = raw_e_info;
    double lx[2], ly[2];
    static char function_name[] = "_viewimg_drag_n_zoom_geom_func";

    vimage = viewimg_get_active (worldcanvas);
    if (vimage == NULL) return (FALSE);
    lx[0] = x[0];
    ly[0] = y[0];
    lx[1] = x[1];
    ly[1] = y[1];
    canvas_convert_to_canvas_coords (worldcanvas, TRUE, 2,
				     lx, ly, lx, ly, NULL, NULL);
    if (raw_event_code == K_CANVAS_EVENT_MODIFIED_KEY_PRESS)
    {
	m_copy (&mask, string + strlen (string) + 1, sizeof mask);
	if ( (strcmp (string, "s") == 0) &&
	     (mask == K_CANVAS_EVENT_LEFT_MOUSE_MASK) )
	{
	    /*  Only compute statistics for the box  */
	    viewimg_statistics_compute (vimage, lx[0], ly[0], lx[1], ly[1]);
	    return (TRUE);
	}
	else if ( (strcmp (string, "v") == 0) &&
		  (mask == K_CANVAS_EVENT_LEFT_MOUSE_MASK) )
	{
	    /*  Only show values in the box  */
	    viewimg_values_display (vimage, TRUE, TRUE,
				    lx[0], ly[0], lx[1], ly[1]);
	    return (TRUE);
	}
	else if ( (strcmp (string, "V") == 0) &&
		  (mask == K_CANVAS_EVENT_LEFT_MOUSE_MASK) )
	{
	    /*  Only show values in the box  */
	    viewimg_values_display (vimage, FALSE, TRUE,
				    lx[0], ly[0], lx[1], ly[1]);
	    return (TRUE);
	}
	else return (FALSE);
    }
    if (raw_event_code != K_CANVAS_EVENT_LEFT_MOUSE_RELEASE) return (FALSE);
    if ( (x[0] == x[1]) && (y[0] == y[1]) )
    {
	/*  Pan  */
	viewimg_zoom (vimage, lx[0], ly[0], FALSE, 1.0);
	return (TRUE);
    }
    if ( (x[0] == x[1]) || (y[0] == y[1]) ) return (TRUE);
    if ( (lx[0] == lx[1]) || (ly[0] == ly[1]) ) return (TRUE);
    viewimg_set_canvas_attributes (worldcanvas,
				   VIEWIMG_ATT_AUTO_X, FALSE,
				   VIEWIMG_ATT_AUTO_Y, FALSE,
				   VIEWIMG_ATT_END);
    canvas_set_attributes (worldcanvas,
			   CANVAS_ATT_LEFT_X, lx[0],
			   CANVAS_ATT_RIGHT_X, lx[1],
			   CANVAS_ATT_BOTTOM_Y, ly[1],
			   CANVAS_ATT_TOP_Y, ly[0],
			   CANVAS_ATT_END);
    if ( !kwin_resize (canvas, TRUE, 0, 0, 0, 0) )
    {
	fprintf (stderr, "%s: Error refreshing window\n", function_name);
    }
    return (TRUE);
}   /*  End Function _viewimg_drag_n_zoom_geom_func  */
