/*LINTLIBRARY*/
/*  main.c

    This code provides ViewableImage objects.

    Copyright (C) 1993-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating viewable images.


    Written by      Richard Gooch   15-APR-1993

    Updated by      Richard Gooch   26-APR-1993

    Updated by      Richard Gooch   30-APR-1993: Fixed bug in
  viewimg_fill_ellipse  and  viewimg_fill_polygon  routines.

    Updated by      Richard Gooch   3-MAY-1993: Added  viewimg_get_active  .

    Updated by      Richard Gooch   2-JUL-1993: Improved error trapping.

    Updated by      Richard Gooch   18-JUL-1993: Added  maintain_aspect_ratio
  parameter to  viewimg_control_autoscaling  and fixed some scale change code.

    Updated by      Richard Gooch   8-AUG-1993: Documented effect on the
  attachment count of  multi_array  data structures by some routines.

    Updated by      Richard Gooch   16-AUG-1993: Fixed bug in  aspect_zoom
  which could hang trying to compute an impossible zoom factor.

    Updated by      Richard Gooch   31-AUG-1993: Fixed declaration of
  first_canvas_holder  to be static.

    Updated by      Richard Gooch   18-NOV-1993: Added  viewimg_create_sequence

    Updated by      Richard Gooch   23-NOV-1993: Made use of routine
  canvas_register_convert_func  .

    Updated by      Richard Gooch   2-DEC-1993: Added  viewimg_init  routine.
  The use of this will become mandatory in version 2.0 of Karma.

    Updated by      Richard Gooch   4-DEC-1993: Added  viewimg_create_restr  .

    Updated by      Richard Gooch   6-DEC-1993: Made use of  canvas_specify
  routine.

    Updated by      Richard Gooch   11-JAN-1994: Fixed bug in  aspect_zoom  .

    Updated by      Richard Gooch   22-APR-1994: Added  viewimg_test_active  .

    Updated by      Richard Gooch   22-JUL-1994: Dropped message in
  worldcanvas_position_func  which complained about no viewable image active.

    Updated by      Richard Gooch   31-JUL-1994: Fixed bug in
  <worldcanvas_size_control_func> where intensity scaling was not correctly set
  for compressed TrueColour images.

    Updated by      Richard Gooch   8-AUG-1994: Added support for 24 bit
  displays.

    Updated by      Richard Gooch   15-SEP-1994: Trapped uniform valued images.

    Updated by      Richard Gooch   24-SEP-1994: Filled in trap for failure to
  compress RGB images in <worldcanvas_size_control_func>.

    Updated by      Richard Gooch   2-OCT-1994: Made use of
  canvas_coord_transform  routine.

    Updated by      Richard Gooch   4-OCT-1994: Changed to  m_cmp  routine in
  order to avoid having to link with buggy UCB compatibility library in
  Slowaris 2.3

    Updated by      Richard Gooch   6-OCT-1994: Fixed
  <worldcanvas_size_control_func> to fit world canvas to pixel canvas when no
  ViewableImage is active for world canvas.

    Updated by      Richard Gooch   24-OCT-1994: Created
  <viewimg_get_canvas_attributes> and <viewimg_set_canvas_attributes> routines.

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   18-NOV-1994: Added attribute documentation.

    Updated by      Richard Gooch   22-NOV-1994: Created
  viewimg_unregister_position_event_func  .

    Updated by      Richard Gooch   26-NOV-1994: Moved to
  packages/viewimg/main.c

    Updated by      Richard Gooch   29-NOV-1994: Made use of  c_  package.

    Updated by      Richard Gooch   14-DEC-1994: Created  viewimg_set_active
  function.

    Updated by      Richard Gooch   27-DEC-1994: Allow NULL multi_array
  pointers to be passed.

    Updated by      Richard Gooch   16-MAR-1995: Generate PostScript directly
  from TrueColour images if available, rather than compressed PseudoColour.

    Updated by      Richard Gooch   31-MAR-1995: Allow RGB movies on
  PseudoColour canvas, but print a warning.

    Updated by      Richard Gooch   15-JUN-1995: Made use of IS_ALIGNED macro.

    Updated by      Richard Gooch   13-JUL-1995: Added
  VIEWIMG_ATT_ALLOW_TRUNCATION attribute.

    Updated by      Richard Gooch   26-JUL-1995: Added support for position
  events on TrueColour ViewableImages.

    Updated by      Richard Gooch   6-SEP-1995: Eased restrictions on tiled
  arrays and created <viewimg_get_attributes>.

    Updated by      Richard Gooch   19-SEP-1995: Ensure canvas size is at least
  3*3 in <worldcanvas_refresh_func> and <worldcanvas_size_control_func>

    Updated by      Richard Gooch   5-JAN-1996: Switched to
  <canvas_register_d_convert_func> routine.

    Updated by      Richard Gooch   21-JAN-1996: Removed warnings about tiled
  arrays for PseudoColour and added for TrueColour.

    Updated by      Richard Gooch   19-FEB-1996: Created
  <viewimg_partial_refresh> routine.

    Updated by      Richard Gooch   28-FEB-1996: Prevent computing image twice
  in some circumstances where not needed and initialise win_scale field of
  viewable image properly. Uninitialised data caused problems in
  <worldcanvas_size_control_func> on alpha_OSF1.

    Updated by      Richard Gooch   27-MAR-1996: Check for genuine data change
  in <worldcanvas_size_control_func> and clear changed flag and set recompute
  flag to prevent message about spurious data change under some conditions.

    Updated by      Richard Gooch   15-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   12-MAY-1996: Added full support for
  non-integral zooming.

    Updated by      Richard Gooch   19-MAY-1996: Changed from using window
  scale structure to using <canvas_get_attributes>.

    Updated by      Richard Gooch   23-MAY-1996: Worked around problem with
  floating point comparisons in <worldcanvas_size_control_func> when running
  under i386_Linux with gcc 2.7.2

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   28-MAY-1996: Fixed bug in
  <worldcanvas_refresh_func> which caused image to be drawn twice if cache is
  invalidated within the <kwin> package (i.e. PostScriptPage refresh).

    Updated by      Richard Gooch   31-MAY-1996: Fixed bug in <determine_size>
  when zooming out without aspect ratio fixed: image could be smaller than
  necessary.

    Updated by      Richard Gooch   4-JUN-1996: Switched to left-right
  bottom-top co-ordinate specification instead of min-max x and y.

    Updated by      Richard Gooch   10-JUN-1996: Added panning support.

    Updated by      Richard Gooch   12-JUN-1996: Optimised panning support.

    Updated by      Richard Gooch   16-JUN-1996: Fixed handling of negative
  co-ordinate systems in <worldcanvas_position_func>.

    Updated by      Richard Gooch   21-JUN-1996: Created
  <viewimg_get_worldcanvas>.

    Updated by      Richard Gooch   18-JUL-1996: Added offset of half a data
  value when converting from world to pixel co-ordinates. This means that a
  point drawn at the co-ordinate for a data value will be drawn in the centre
  of the box of pixels representing that data value, rather than the
  bottom-left corner.

    Updated by      Richard Gooch   20-JUL-1996: Switched to
  <canvas_register_coords_convert_func>.

    Updated by      Richard Gooch   22-JUL-1996: Created
  <viewimg_create_sequence_from_iarray>.

    Updated by      Richard Gooch   6-AUG-1996: Added VIEWIMG_VATT_MULTI_ARRAY
  attribute.

    Updated by      Richard Gooch   31-AUG-1996: Made <viewimg_partial_refresh>
  more efficient when clearing small areas outside the image.

    Updated by      Richard Gooch   5-OCT-1996: Detect changes to intensity
  scaling function.

    Updated by      Richard Gooch   11-OCT-1996: Fixed bug in previous change
  when image minimum and maximum not available (i.e. no auto intensity scaling)

    Updated by      Richard Gooch   2-NOV-1996: Added VIEWIMG_VATT_DATA_SCALE
  and VIEWIMG_VATT_DATA_OFFSET attributes. Created <viewimg_set_attributes> and
  <viewimg_set_sequence_scale> routines.

    Updated by      Richard Gooch   10-NOV-1996: Created
  <viewimg_set_array_attributes> and removed <viewimg_set_sequence_scale>
  routines.

    Updated by      Richard Gooch   27-NOV-1996: Changed <viewimg_set_active>
  to not refresh canvas if it's not visible.

    Updated by      Richard Gooch   8-DEC-1996: Created <copy_restrictions> to
  fix restriction copying: was just copying pointers incorrectly.

    Updated by      Richard Gooch   23-JAN-1997: Added CONST declarations to
  <viewimg_create_rgb_sequence> function.

    Updated by      Richard Gooch   1-FEB-1997: Made use of new
  <wcs_astro_get_associated> function and now call <canvas_use_astro_transform>
  function whenever active ViewableImage changes.

    Updated by      Richard Gooch   10-FEB-1997: Added VIEWIMG_ATT_AUTO_WCS
  attribute.

    Updated by      Richard Gooch   15-FEB-1997: Added VIEWIMG_VATT_WCS_ASTRO.

    Updated by      Richard Gooch   26-MAR-1997: Automatically get data scaling
  if available.

    Updated by      Richard Gooch   17-MAY-1997: Do not redraw TrueColour image
  when colourmap resizes.

    Updated by      Richard Gooch   7-JUN-1997: Attempt to display larger
  images in <_viewimg_trunc_zoom> function.

    Updated by      Richard Gooch   25-JUL-1997: Made use of new precompute
  callback mechanism and world canvas image cache compute functions.

    Updated by      Richard Gooch   31-JUL-1997: Took account of change to
  precompute function interface: pointer to <honoured_areas> flag dropped.

    Updated by      Richard Gooch   4-AUG-1997: No longer add tiny offset when
  converting from pixel->world co-ordinates in <_viewimg_coord_convert_func>.

    Updated by      Richard Gooch   27-AUG-1997: Improved diagnostics in
  <_viewimg_recompute_image>.

    Updated by      Richard Gooch   23-SEP-1997: Added VIEWIMG_ATT_FILL_CANVAS
  experimental attribute and made preliminary use of it.

    Updated by      Richard Gooch   24-SEP-1997: Removed factor test in
  <_viewimg_determine_size> so that zoom in and out are supported for canvas
  filling.

    Updated by      Richard Gooch   12-NOV-1997: Trapped all-blank images in
  <_viewimg_worldcanvas_size_control_func> when auto intensity scaling enabled.

    Updated by      Richard Gooch   24-NOV-1997: Tried a few experimental
  hacks to make function pointer passing more C++ friendly.

    Updated by      Richard Gooch   5-DEC-1997: Added VIEWIMG_ATT_FAST_PANNER
  attribute.

    Updated by      Richard Gooch   7-DEC-1997: Added flag to check if fast
  panner was used last.

    Updated by      Richard Gooch   17-DEC-1997: Fixed intensity scale change
  detection algorithm: now cope with 25 log cycles.

    Updated by      Richard Gooch   28-MAR-1998: Fixed bug in <trunc_zoom>:
  could induce creeping zooming and other problems.

    Updated by      Richard Gooch   29-MAR-1998: Added image and screen pixel
    co-ordinates to world canvas position event functions.

    Updated by      Richard Gooch   31-MAR-1998: Fixed bug in function
  <_viewimg_worldcanvas_position_func> when input co-ordinate out of bounds.

    Updated by      Richard Gooch   31-AUG-1998: Supported more attribute
  getting.

    Updated by      Richard Gooch   5-JUL-1999: Fixed restriction memory leak
  in <viewimg_create_sequence>. Created freelist. Removed sequence holder.
  Removed <copy_restrictions>.

    Updated by      Richard Gooch   12-JUL-1999: Recompute image if magnify
  function changed.

    Updated by      Richard Gooch   15-AUG-1999: Supported more attribute
  getting. Recompute panned images if magnification changes. Made egcs happy.

    Updated by      Richard Gooch   3-OCT-1999: Created <<set_size>> function
  types for normal and panned displays. Moved <viewimg_zoom> from drag_n_zoom.c
  so that magnification factors can be manipulated correctly.

    Updated by      Richard Gooch   12-OCT-1999: Fixed stupid "optimisation" in
  <viewimg_zoom> which discarded unitary zooms (forgot that centre might
  change).

    Updated by      Richard Gooch   13-OCT-1999: Fixed bug in function
  <_viewimg_draw_subcache>: had >= test rather than > test for Y. Disabled the
  entire test since I think it may no longer be necessary.

    Updated by      Richard Gooch   18-OCT-1999: Fixed bug in function
  <viewimg_zoom>: existing magnification was not correctly computed.

    Updated by      Richard Gooch   19-OCT-1999: Fixed <_viewimg_draw_subcache>
  so that world canvas is properly filled (no blank areas).

    Updated by      Richard Gooch   25-DEC-1999: Fixed bug in function
  <_viewimg_set_size_pan>: image drift.

    Updated by      Richard Gooch   11-APR-2000: Fixed bug in function
  <_viewimg_worldcanvas_position_func>: image pixel co-ordinates were not
  clipped.

    Updated by      Richard Gooch   13-MAY-2000: Replaced use of PS page with
  KWIN_FUNC_DRAW_CACHED_IMAGE in order to deterime if cached images are valid.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variables.

    Updated by      Richard Gooch   27-JUN-2000: Made use of new function
  <canvas_register_destroy_func>.

    Updated by      Richard Gooch   6-JUL-2000: Fixed bug in function
  <_viewimg_set_size_normal> when separating values so that image boundary is
  not exceeded. Added sanity checks in function <_viewimg_coord_convert_func>.

    Updated by      Richard Gooch   7-JUL-2000: Took account of interface
  change to <canvas_register_destroy_func>.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>
  and <karma_kcmap.h>.

    Updated by      Richard Gooch   12-JUL-2000: Added
  #include <karma_canvas.h>

    Updated by      Richard Gooch   20-AUG-2001: Fixed <_viewimg_trunc_zoom>
  to ensure npix >= 2 for the case of fixed aspect and zooming out.

    Updated by      Richard Gooch   21-AUG-2001: No longer update pan centre in
  <_viewimg_set_size_pan> because the next size negotiation iteration can
  reduce the canvas size, and thus the edge of the image cannot be seen.
  Compensated for this change in function <_viewimg_draw_subcache>.

    Updated by      Richard Gooch   22-AUG-2001: Fixed bugs in
  <_viewimg_aspect_zoom> and <_viewimg_trunc_zoom> to force integer zoom
  factors if either X or Y requires them and fixed aspect ratio is enabled.
  Fall back to <_viewimg_trunc_zoom> if preferred method fails.

    Updated by      Richard Gooch   27-AUG-2001: Fixed <_viewimg_trunc_zoom> so
  that truncation is no longer performed when zooming in both axes. Truncation
  in this case can cause "multiple exposure" effect for axis labels (boundary
  keeps shifting inwards for successive refreshes).

    Updated by      Richard Gooch   29-AUG-2001: Trap for degenerate dimensions
  in <_viewimg_coord_convert_func> and generate a valid value.

    Updated by      Richard Gooch   26-SEP-2001: Switched to SAFE_ALIGNMENT.

    Updated by      Richard Gooch   4-OCT-2001: Fixed bug in recent changes to
  <_viewimg_trunc_zoom> which always passed TRUE for <<int_zoom>> to
  <_viewimg_aspect_zoom>.

    Updated by      Richard Gooch   8-OCT-2001: Fixed documentation links.

    Last updated by Richard Gooch   12-OCT-2001: Clear canvas specification in
  <viewimg_destroy> so that axis label code doesn't get confused.


*/
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdarg.h>
#include <karma.h>
#define NEW_WIN_SCALE
#define KARMA_VIEWIMG_MAIN_C
#include <karma_viewimg.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_kwin_hooks.h>
#include <karma_imc.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>
#include <os.h>

/*#define DEBUG_SIZE_CONTROL_FUNC*/
/*#define DEBUG_REFRESH_FUNC*/


#define VIMAGE_USED_MAGIC_NUMBER (unsigned int) 229573367
#define VIMAGE_FREE_MAGIC_NUMBER (unsigned int) 142734985
#define HOLDER_MAGIC_NUMBER (unsigned int) 1654545154

#define DEFAULT_NUMBER_OF_COLOURS 200
#define BUF_SIZE  1024

#define VERIFY_HOLDER(holder) {if (holder == NULL) \
{fprintf (stderr, "NULL canvas holder passed\n"); \
 a_prog_bug (function_name); } \
if (holder->magic_number != HOLDER_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas holder object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_USED_VIMAGE(vimage) {if (vimage == NULL) \
{fprintf (stderr, "NULL viewable image passed\n"); \
 a_prog_bug (function_name); } \
if (vimage->magic_number != VIMAGE_USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid viewable image object\n"); \
 a_prog_bug (function_name); } }

#define VERIFY_FREE_VIMAGE(vimage) {if (vimage == NULL) \
{fprintf (stderr, "NULL viewable image passed\n"); \
 a_prog_bug (function_name); } \
if (vimage->magic_number != VIMAGE_FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed viewable image object\n"); \
 a_prog_bug (function_name); } }

#define NUM_ISCALE_VALUES 18
static double iscale_factors[NUM_ISCALE_VALUES] =
{1e-25, 1e-20, 1e-15, 1e-10, 1e-7, 1e-5, 1e-4, 1e-3,
 0.01, 0.1, 0.2, 0.5, 0.51, 0.55, 0.8, 0.9, 0.99, 1.0};


typedef struct canvas_holder_type * CanvasHolder;
typedef flag (*set_size_func_type)
     (CanvasHolder holder, int width, int height,
      struct win_scale_type *win_scale,
      long *hstart, long *hend, CONST dim_desc *hdim,
      long *vstart, long *vend, CONST dim_desc *vdim);

struct canvas_holder_type
{
    unsigned int magic_number;
    KWorldCanvas canvas;
    ViewableImage first_image;
    ViewableImage active_image;
    CanvasHolder prev;
    CanvasHolder next;
    KCallbackList position_list;
    flag auto_x;
    flag auto_y;
    flag auto_v;
    flag int_x;
    flag int_y;
    flag maintain_aspect_ratio;
    flag allow_truncation;
    flag enable_panning;
    flag fast_panner;
    long pan_centre_x;
    long pan_centre_y;
    flag auto_wcs;
    set_size_func_type set_size;
    double pan_magnification_x;
    double pan_magnification_y;
};

struct canvas_override_type
{
    double value_min;
    double value_max;
};

struct viewableimage_type
{
    double value_min;
    double value_max;
    unsigned int magic_number;
    CanvasHolder canvas_holder;
    /*  PseudoColour (single channel) image data  */
    multi_array *pc_multi_desc;
    array_desc *pc_arr_desc;
    char *pc_slice;
    unsigned int pc_hdim;
    unsigned int pc_vdim;
    unsigned int pc_elem_index;
    unsigned int pc_hstride;
    unsigned int pc_vstride;
    double pc_data_scale;
    double pc_data_offset;
    /*  TrueColour image data  */
    multi_array *tc_multi_desc;
    array_desc *tc_arr_desc;  /*  If this is not NULL, the original  */
    char *tc_slice;           /*  image data is TrueColour           */
    unsigned int tc_hdim;
    unsigned int tc_vdim;
    unsigned int tc_red_index;
    unsigned int tc_green_index;
    unsigned int tc_blue_index;
    unsigned int tc_hstride;
    unsigned int tc_vstride;
    /*  Other information  */
    flag recompute;        /* Whether the image cache needs to be recomputed */
    flag changed;          /* Whether the array data has changed             */
    int pixcanvas_width;
    int pixcanvas_height;
    KPixCanvasImageCache cache;
    unsigned int num_restrictions;
    char **restriction_names;
    double *restriction_values;
    char *inline_restr_names[1];
    double inline_restr_values[1];
    double iscale_test[NUM_ISCALE_VALUES];
    struct canvas_override_type override;
    KwcsAstro astro_projection;
    unsigned int last_pan_magnification;
    struct win_scale_type win_scale;
    void *magnify_func;
    ViewableImage prev;
    ViewableImage next;
};

struct position_struct
{
    double x;
    double y;
    void *value;
    int event_code;
    void *e_info;
    double x_lin;
    double y_lin;
    unsigned int type;
    double x_im;
    double y_im;
    int x_pix;
    int y_pix;
};

typedef flag (*IscaleFunc) (double *out, unsigned int out_stride,
			    CONST double *inp, unsigned int inp_stride,
			    unsigned int num_values,
			    double i_min, double i_max, void *info);
typedef flag (*PositionFunc) (ViewableImage vimage, double x, double y,
			      void *value, unsigned int event_code,
			      void *e_info, void **f_info,
			      double x_lin, double y_lin,
			      unsigned int value_type,
			      double x_im, double y_im,
			      int x_pix, int y_pix);


/*  Private data  */
static CanvasHolder first_canvas_holder = NULL;
static ViewableImage first_free = NULL;


/*  External functions  */
EXTERN_FUNCTION (flag kwin_compute_pc_subimage,
		 (KPixCanvas canvas,
		  int x_off, int y_off, int x_pixels, int y_pixels,
		  int cache_width, int cache_height,
		  CONST char *slice,
		  CONST uaddr *hoffsets, CONST uaddr *voffsets,
		  unsigned int image_width, unsigned int image_height,
		  unsigned int type, unsigned int conv_type,
		  unsigned int num_pixels, unsigned long *pixel_values,
		  unsigned long blank_pixel,
		  unsigned long min_sat_pixel, unsigned long max_sat_pixel,
		  double i_min, double i_max,
		  flag (*iscale_func) (), void *iscale_info,
		  KPixCanvasImageCache *cache) );
EXTERN_FUNCTION (flag kwin_compute_rgb_subimage,
		 (KPixCanvas canvas,
		  int x_off, int y_off, int x_pixels, int y_pixels,
		  int cache_width, int cache_height,
		  CONST unsigned char *red_slice,
		  CONST unsigned char *green_slice,
		  CONST unsigned char *blue_slice,
		  CONST uaddr *hoffsets, CONST uaddr *voffsets,
		  unsigned int image_width, unsigned int image_height,
		  KPixCanvasImageCache *cache) );
EXTERN_FUNCTION (KCallbackFunc canvas_register_precompute_func,
		 (KWorldCanvas canvas, void (*refresh_func) (), void *info) );
EXTERN_FUNCTION (flag canvas_compute_pc_image,
		 (KWorldCanvas canvas, array_desc *arr_desc,
		  char *slice, unsigned int hdim, unsigned int vdim,
		  unsigned int elem_index, KPixCanvasImageCache *cache) );
EXTERN_FUNCTION (flag canvas_compute_rgb_image,
		 (KWorldCanvas canvas, array_desc *arr_desc,
		  CONST char *slice, unsigned int hdim, unsigned int vdim,
		  unsigned int red_index, unsigned int green_index,
		  unsigned int blue_index, KPixCanvasImageCache *cache) );


/*  Private functions  */
STATIC_FUNCTION (CanvasHolder _viewimg_get_canvas_holder,
		 (KWorldCanvas canvas, flag alloc, char *func_name) );
STATIC_FUNCTION (CanvasHolder _viewimg_alloc_canvas_holder,
		 (KWorldCanvas canvas) );
STATIC_FUNCTION (void _viewimg_free_canvas_holder, (CanvasHolder holder) );
STATIC_FUNCTION (unsigned int _viewimg_get_gen_index,
		 (CONST multi_array *multi_desc, CONST array_desc *arr_desc) );
STATIC_FUNCTION (void _viewimg_worldcanvas_size_control_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, void **info,
		  flag *boundary_clear) );
STATIC_FUNCTION (flag _viewimg_set_size_normal,
		 (CanvasHolder holder, int width, int height,
		  struct win_scale_type *win_scale,
		  long *hstart, long *hend, CONST dim_desc *hdim,
		  long *vstart, long *vend, CONST dim_desc *vdim) );
STATIC_FUNCTION (flag _viewimg_set_size_pan,
		 (CanvasHolder holder, int width, int height,
		  struct win_scale_type *win_scale,
		  long *hstart, long *hend, CONST dim_desc *hdim,
		  long *vstart, long *vend, CONST dim_desc *vdim) );
STATIC_FUNCTION (flag _viewimg_determine_size,
		 (CanvasHolder holder, int width, int height,
		  struct win_scale_type *win_scale,
		  long *hstart, long *hend, long *vstart, long *vend) );
STATIC_FUNCTION (flag _viewimg_aspect_zoom,
		 (long hlength, int *hpixels, long vlength, int *vpixels,
		  flag int_zoom) );
STATIC_FUNCTION (void _viewimg_trunc_zoom,
		 (CONST struct canvas_holder_type *holder,
		  long *hstart, long *hend, int *hpixels,
		  long *vstart, long *vend, int *vpixels) );
STATIC_FUNCTION (void _viewimg_worldcanvas_precompute_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _viewimg_worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void _viewimg_recompute_image,
		 (CanvasHolder holder, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  ViewableImage vimage) );
STATIC_FUNCTION (void _viewimg_draw_subcache,
		 (CanvasHolder holder, ViewableImage vimage,
		  struct win_scale_type *win_scale) );
STATIC_FUNCTION (flag _viewimg_worldcanvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info,
		  double x_lin, double y_lin, int x_pix, int y_pix) );
STATIC_FUNCTION (flag _viewimg_position_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (flag _viewimg_coord_convert_func,
		 (KWorldCanvas canvas, unsigned int num_coords,
		  CONST double *xin, CONST double *yin,
		  double *xout, double *yout, flag to_world, void **info) );
STATIC_FUNCTION (ViewableImage _viewimg_alloc_vimage, () );
STATIC_FUNCTION (void _viewimg_free_vimage, (ViewableImage vimage) );


/* Public functions follow */

/*PUBLIC_FUNCTION*/
void viewimg_init (KWorldCanvas canvas)
/*  [SUMMARY] Initialise the package for a particular canvas.
    [PURPOSE] This routine will initialise the [<viewimg>] package for a
    particular world canvas. Calling this routine causes a number of callback
    routines internal to the package to be registered with the canvas (such
    as refresh and position event callbacks). The use of this routine is
    optional at the moment: the routines which create viewable images perform
    this function automatically. In version 2.0 of Karma, this use of this
    routine before creating viewable images will become mandatory.
    <canvas> The world canvas object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "viewimg_init";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (_viewimg_get_canvas_holder (canvas, FALSE, function_name) != NULL)
    {
	fprintf (stderr, "Canvas already initialised\n");
	a_prog_bug (function_name);
    }
    if (_viewimg_alloc_canvas_holder (canvas) == NULL)
	m_abort (function_name, "canvas holder");
}   /*  End Function viewimg_init  */

/*PUBLIC_FUNCTION*/
ViewableImage viewimg_create_restr (KWorldCanvas canvas,
				    multi_array *multi_desc,
				    array_desc *arr_desc, char *slice,
				    unsigned int hdim, unsigned int vdim,
				    unsigned int elem_index,
				    unsigned num_restr,
				    CONST char **restr_names,
				    CONST double *restr_values)
/*  [SUMMARY] Create viewable image from 2D slice with restrictions.
    [PURPOSE] This routine will create a PseudoColour viewable image object
    from a 2-dimensional slice of a Karma data structure. At a later time, this
    viewable image may be made visible. This routine will not cause the canvas
    to be refreshed.
    <canvas> The world canvas onto which the viewable image may be drawn. Many
    viewable images may be associated with a single canvas.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <slice> The start of the slice data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index in the data packets.
    <num_restr> The number of matched restrictions. If this is 0, no
    restrictions are recorded (this is the same as calling [<viewimg_create>]).
    <restr_names> The restriction names.
    <restr_values> The restriction values.
    [NOTE] Restriction information is automatically deallocated when
    [<viewimg_destroy>] is called.
    [NOTE] The routine may produce cache data which will vastly increase the
    speed of subsequent operations on this data. Prior to process exit, a call
    MUST be made to [<viewimg_destroy>], otherwise shared memory segments could
    remain after the process exits.
    [NOTE] The ViewableImage will be automatically destroyed if the canvas is
    destroyed.
    [RETURNS] A viewable image on success, else NULL.
*/
{
    CanvasHolder holder;
    ViewableImage vimage;
    unsigned int dim_count, index;
    static char function_name[] = "viewimg_create_restr";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (slice == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Sanity checks  */
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates != NULL)
    {
	fprintf (stderr, "hdim: %u not regularly spaced\n", hdim);
	a_prog_bug (function_name);
    }
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates != NULL)
    {
	fprintf (stderr, "vdim: %u not regularly spaced\n", vdim);
	a_prog_bug (function_name);
    }
    if (elem_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    if ( ( holder = _viewimg_get_canvas_holder (canvas, TRUE, function_name) )
	 == NULL )
    {
	return (NULL);
    }
    /*  Create new viewable image  */
    if ( ( vimage = _viewimg_alloc_vimage () ) == NULL )
    {
	m_error_notify (function_name, "viewable image");
	return (NULL);
    }
    vimage->canvas_holder = holder;
    vimage->pc_multi_desc = multi_desc;
    vimage->pc_arr_desc = arr_desc;
    vimage->pc_slice = slice;
    vimage->pc_hdim = hdim;
    vimage->pc_vdim = vdim;
    vimage->pc_elem_index = elem_index;
    /*  Compute strides  */
    vimage->pc_hstride = ds_get_packet_size (arr_desc->packet);
    for (dim_count = arr_desc->num_dimensions - 1; dim_count > hdim;
	 --dim_count)
    {
	vimage->pc_hstride *= arr_desc->dimensions[dim_count]->length;
    }
    vimage->pc_vstride = ds_get_packet_size (arr_desc->packet);
    for (dim_count = arr_desc->num_dimensions - 1; dim_count > vdim;
	 --dim_count)
    {
	vimage->pc_vstride *= arr_desc->dimensions[dim_count]->length;
    }
    vimage->num_restrictions = num_restr;
    vimage->restriction_names = (char **) restr_names;
    vimage->restriction_values = (double *) restr_values;
    if (multi_desc == NULL) vimage->astro_projection = NULL;
    else
    {
	index = _viewimg_get_gen_index (multi_desc, arr_desc);
	vimage->astro_projection = wcs_astro_get_associated (multi_desc,index);
	ds_get_data_scaling (arr_desc->packet->element_desc[elem_index],
			     multi_desc->headers[index],
			     multi_desc->data[index],
			     &vimage->pc_data_scale, &vimage->pc_data_offset);
    }
    vimage->last_pan_magnification = 0;
    /*  Do not make the first viewable image for this canvas viewable, else
	the  viewimg_make_active  routine will not be able to detect a change
	in the active viewable image and will do nothing if called with this
	newly created viewable image.
	*/
    if (holder->first_image != NULL)
    {
	/*  Insert at beginning of list  */
	holder->first_image->prev = vimage;
    }
    vimage->next = holder->first_image;
    holder->first_image = vimage;
    /*  Attach  */
    if (multi_desc != NULL) ++multi_desc->attachments;
    return (vimage);
}   /*  End Function viewimg_create_restr  */

/*PUBLIC_FUNCTION*/
ViewableImage viewimg_create (KWorldCanvas canvas, multi_array *multi_desc,
			      array_desc *arr_desc, char *slice,
			      unsigned int hdim, unsigned int vdim,
			      unsigned int elem_index)
/*  [SUMMARY] Create viewable image from 2D slice.
    [PURPOSE] This routine will create a PseudoColour viewable image object
    from a 2-dimensional slice of a Karma data structure. At a later time, this
    viewable image may be made visible. This routine will not cause the canvas
    to be refreshed.
    <canvas> The world canvas onto which the viewable image may be drawn. Many
    viewable images may be associated with a single canvas.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <slice> The start of the slice data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index in the data packets.
    [NOTE] The routine may produce cache data which will vastly increase the
    speed of subsequent operations on this data. Prior to process exit, a call
    MUST be made to [<viewimg_destroy>], otherwise shared memory segments could
    remain after the process exits.
    [NOTE] The ViewableImage will be automatically destroyed if the canvas is
    destroyed.
    [RETURNS] A viewable image on success, else NULL.
*/
{
    /*static char function_name[] = "viewimg_create";*/

    return viewimg_create_restr (canvas, multi_desc, arr_desc, slice,
				 hdim, vdim, elem_index,
				 0, NULL, (double *) NULL);
}   /*  End Function viewimg_create  */

/*PUBLIC_FUNCTION*/
ViewableImage viewimg_create_from_iarray (KWorldCanvas canvas, iarray array,
					  flag swap)
/*  [SUMMARY] Create a viewable image from an Intelligent Array.
    [PURPOSE] This routine will create a viewable image object from a
    2-dimensional Intelligent Array. At a later time, this viewable image may
    be made visible. This routine will not cause the canvas to be refreshed.
    Many viewable images may be associated with a single canvas.
    <canvas> The world canvas object.
    <array> The Intelligent Array. The underlying <<multi_array>> data
    structure will have its attachment count incremented upon successful
    completion.
    <swap> If TRUE the y axis will be displayed horizontally.
    [NOTE] The routine may produce cache data which will vastly increase the
    speed of subsequent operations on this data. Prior to process exit, a call
    MUST be made to [<viewimg_destroy>], otherwise shared memory segments could
    remain after the process exits.
    [NOTE] The ViewableImage will be automatically destroyed if the canvas is
    destroyed.
    [RETURNS] A viewable image on success, else NULL.
*/
{
    unsigned int num_restr;
    char **restr_names;
    double *restr_values;
    static char function_name[] = "viewimg_create_from_iarray";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (array == NULL)
    {
	fprintf (stderr, "NULL Intelligent Array passed\n");
	a_prog_bug (function_name);
    }
    FLAG_VERIFY (swap);
    if (iarray_num_dim (array) != 2)
    {
	fprintf (stderr,
			"Intelligent Array must have exactly 2 dimensions\n");
	return (NULL);
    }
    if ( (array->offsets[0] != array->arr_desc->offsets[0]) ||
	 (array->offsets[1] != array->arr_desc->offsets[1]) )
    {
	fprintf (stderr, "Intelligent Array must not be a sub-array\n");
	return (NULL);
    }
    num_restr = iarray_get_restrictions (array, &restr_names, &restr_values);
    if (swap)
    {
	return viewimg_create_restr (canvas, array->multi_desc,
				     array->arr_desc, array->data,
				     array->orig_dim_indices[0],
				     array->orig_dim_indices[1],
				     array->elem_index,
				     num_restr, (CONST char **) restr_names,
				     restr_values);
    }
    else
    {
	return viewimg_create_restr (canvas, array->multi_desc,
				     array->arr_desc, array->data,
				     array->orig_dim_indices[1],
				     array->orig_dim_indices[0],
				     array->elem_index,
				     num_restr, (CONST char **) restr_names,
				     restr_values);
    }
}   /*  End Function viewimg_create_from_iarray  */

/*PUBLIC_FUNCTION*/
ViewableImage *viewimg_create_sequence (KWorldCanvas canvas,
					multi_array *multi_desc,
					array_desc *arr_desc, char *cube,
					unsigned int hdim, unsigned int vdim,
					unsigned int fdim,
					unsigned int elem_index)
/*  [SUMMARY] Create a sequence of viewable images from a 3D slice.
    [PURPOSE] This routine will create a sequence of viewable image objects
    from a 3-dimensional cube of a Karma data structure. At a later time, this
    sequence of viewable images may be made visible in any order.
    This routine will not cause the canvas to be refreshed.
    <canvas> The world canvas object.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <cube> The start of the cube data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <fdim> The dimension index of the frame dimension (dimension containing the
    sequence). The number of frames is the same as the length of this
    dimension.
    <elem_index> The element index in the data packets.
    [NOTE] The routine may produce cache data which will vastly increase the
    speed of subsequent operations on this data. Prior to process exit, a call
    MUST be made to [<viewimg_destroy>], otherwise shared memory segments could
    remain after the process exits.
    [NOTE] The ViewableImage objects will be automatically destroyed if the
    canvas is destroyed.
    [RETURNS] A pointer to a dynamically allocated array of viewable image
    objects on success, else NULL.
*/
{
    CanvasHolder holder;
    ViewableImage vimage;
    KwcsAstro ap;
    unsigned int num_frames, count;
    unsigned int hstride, vstride;
    double scale, offset;
    ViewableImage *vimages;
    uaddr *foffsets;
    static char function_name[] = "viewimg_create_sequence";

    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (cube == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Sanity checks  */
    if (fdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "fdim: %u greater than number of dimensions: %u\n",
		 fdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates != NULL)
    {
	fprintf (stderr, "hdim: %u not regularly spaced\n", hdim);
	a_prog_bug (function_name);
    }
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates != NULL)
    {
	fprintf (stderr, "vdim: %u not regularly spaced\n", vdim);
	a_prog_bug (function_name);
    }
    if (elem_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    /*  Compute strides  */
    hstride = vstride = ds_get_packet_size (arr_desc->packet);
    for (count = arr_desc->num_dimensions - 1; count > hdim; --count)
    {
	hstride *= arr_desc->dimensions[count]->length;
    }
    for (count = arr_desc->num_dimensions - 1; count > vdim; --count)
    {
	vstride *= arr_desc->dimensions[count]->length;
    }
    if (multi_desc == NULL)
    {
	ap = NULL;
	scale = 1.0;
	offset = 0.0;
    }
    else
    {
	count = _viewimg_get_gen_index (multi_desc, arr_desc);
	ap = wcs_astro_get_associated (multi_desc,count);
	ds_get_data_scaling (arr_desc->packet->element_desc[elem_index],
			     multi_desc->headers[count],
			     multi_desc->data[count], &scale, &offset);
    }
    if ( ( holder = _viewimg_get_canvas_holder (canvas, TRUE, function_name) )
	 == NULL )
    {
	return (NULL);
    }
    foffsets = arr_desc->offsets[fdim];
    num_frames = arr_desc->dimensions[fdim]->length;
    if ( ( vimages = m_alloc (sizeof *vimages * num_frames) ) == NULL )
    {
	m_error_notify (function_name, "array of viewable images");
	return (NULL);
    }
    for (count = 0; count < num_frames; ++count)
    {
	/*  Do not use <viewimg_create_restr> as there is a lot of duplicated
	    work  */
	if ( ( vimage = _viewimg_alloc_vimage () ) == NULL )
	{
	    m_error_notify (function_name, "viewable image");
	    for (; count > 0; --count) viewimg_destroy (vimages[count - 1]);
	    m_free (vimages);
	    return (NULL);
	}
	vimages[count] = vimage;
	vimage->canvas_holder = holder;
	vimage->pc_multi_desc = multi_desc;
	vimage->pc_arr_desc = arr_desc;
	vimage->pc_slice = cube + foffsets[count];
	vimage->pc_hdim = hdim;
	vimage->pc_vdim = vdim;
	vimage->pc_elem_index = elem_index;
	vimage->pc_hstride = hstride;
	vimage->pc_vstride = vstride;
	vimage->astro_projection = ap;
	vimage->pc_data_scale = scale;
	vimage->pc_data_offset = offset;
	vimage->num_restrictions = 1;
	vimage->inline_restr_names[0] = arr_desc->dimensions[fdim]->name;
	vimage->inline_restr_values[0] =
	    ds_get_coordinate (arr_desc->dimensions[fdim], count);
	vimage->restriction_names = vimage->inline_restr_names;
	vimage->restriction_values = vimage->inline_restr_values;
	vimage->last_pan_magnification = 0;
	/*  Do not make the first viewable image for this canvas viewable, else
	    the  viewimg_make_active  routine will not be able to detect a
	    change in the active viewable image and will do nothing if called
	    with this newly created viewable image
	    */
	if (holder->first_image != NULL)
	{
	    /*  Insert at beginning of list  */
	    holder->first_image->prev = vimage;
	}
	vimage->next = holder->first_image;
	holder->first_image = vimage;
	/*  Attach  */
	if (multi_desc != NULL) ++multi_desc->attachments;
    }
    return (vimages);
}   /*  End Function viewimg_create_sequence  */

/*PUBLIC_FUNCTION*/
ViewableImage *viewimg_create_sequence_from_iarray (KWorldCanvas canvas,
						    iarray array,
						    unsigned int hdim,
						    unsigned int vdim,
						    unsigned int fdim)
/*  [SUMMARY] Create a sequence of viewable images from an Intelligent Array.
    [PURPOSE] This routine will create a sequence of viewable image objects
    from a 3-dimensional Intelligent Array. At a later time, this sequence of
    viewable images may be made visible in any order.
    This routine will not cause the canvas to be refreshed.
    <canvas> The world canvas object.
    <array> The Intelligent Array. The underlying <<multi_array>> data
    structure will have its attachment count incremented upon successful
    completion.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <fdim> The dimension index of the frame dimension (dimension containing the
    sequence). The number of frames is the same as the length of this
    dimension.
    [NOTE] The routine may produce cache data which will vastly increase the
    speed of subsequent operations on this data. Prior to process exit, a call
    MUST be made to [<viewimg_destroy>], otherwise shared memory segments could
    remain after the process exits.
    [NOTE] The ViewableImage objects will be automatically destroyed if the
    canvas is destroyed.
    [RETURNS] A pointer to a dynamically allocated array of viewable image
    objects on success, else NULL.
*/
{
    static char function_name[] = "viewimg_create_sequence_from_iarray";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (array == NULL)
    {
	fprintf (stderr, "NULL Intelligent Array passed\n");
	a_prog_bug (function_name);
    }
    if (iarray_num_dim (array) != 3)
    {
	fprintf (stderr, "Intelligent Array must have exactly 3 dimensions\n");
	return (NULL);
    }
    if ( hdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "hdim: %u too large\n", hdim);
	a_prog_bug (function_name);
    }
    if ( vdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "vdim: %u too large\n", vdim);
	a_prog_bug (function_name);
    }
    if ( fdim >= iarray_num_dim (array) )
    {
	fprintf (stderr, "fdim: %u too large\n", fdim);
	a_prog_bug (function_name);
    }
    if ( (array->offsets[fdim] != array->arr_desc->offsets[fdim]) ||
	 (array->offsets[hdim] != array->arr_desc->offsets[hdim]) ||
         (array->offsets[hdim] != array->arr_desc->offsets[hdim]) )
    {
	fprintf (stderr, "Intelligent Array must not be a sub-array\n");
	return (NULL);
    }
    return viewimg_create_sequence (canvas, array->multi_desc,
				    array->arr_desc, array->data,
				    array->orig_dim_indices[hdim],
				    array->orig_dim_indices[vdim],
				    array->orig_dim_indices[fdim],
				    array->elem_index);
}   /*  End Function viewimg_create_sequence_from_iarray  */

/*PUBLIC_FUNCTION*/
ViewableImage viewimg_create_rgb (KWorldCanvas canvas, multi_array *multi_desc,
				  array_desc *arr_desc, char *slice,
				  unsigned int hdim, unsigned int vdim,
				  unsigned int red_index,
				  unsigned int green_index,
				  unsigned int blue_index, unsigned num_restr,
				  CONST char **restr_names,
				  CONST double *restr_values)
/*  [SUMMARY] Create a TrueColour viewable image.
    [PURPOSE] This routine will create a TrueColour viewable image object from
    a 2-dimensional slice of a Karma data structure. At a later time, this
    viewable image may be made visible. This routine will not cause the canvas
    to be refreshed.
    <canvas> The world canvas onto which the viewable image may be drawn. Many
    viewable images may be associated with a single canvas.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor on
    successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <slice> The start of the slice data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <red_index> The element index of the red component in the data packets.
    <green_index> The element index of the green component in the data packets.
    <blue_index> The element index of the blue component in the data packets.
    [NOTE] The 3 colour components must be of type  K_UBYTE  .
    <num_restr> The number of matched restrictions. If this is 0, no
    restrictions are recorded.
    <restr_names> The restriction names.
    <restr_values> The restriction values.
    [NOTE] Restriction information is automatically deallocated when
    [<viewimg_destroy>] is called.
    [NOTE] The routine may produce cache data which will vastly increase the
    speed of subsequent operations on this data. Prior to process exit, a call
    MUST be made to [<viewimg_destroy>], otherwise shared memory segments could
    remain after the process exits.
    [NOTE] The ViewableImage will be automatically destroyed if the canvas is
    destroyed.
    [RETURNS] A viewable image on success, else NULL.
*/
{
    CanvasHolder holder;
    ViewableImage vimage;
    unsigned int dim_count;
    unsigned int depth, index;
    static char function_name[] = "viewimg_create_rgb";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (slice == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    if (arr_desc->num_levels > 0)
    {
	fprintf (stderr, "%s: Tiled array. May cause problems.\n",
		 function_name);
    }
    /*  Sanity checks  */
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates != NULL)
    {
	fprintf (stderr, "hdim: %u not regularly spaced\n", hdim);
	a_prog_bug (function_name);
    }
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates != NULL)
    {
	fprintf (stderr, "vdim: %u not regularly spaced\n", vdim);
	a_prog_bug (function_name);
    }
    if (red_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr, "red_index: %u greater than number of elements: %u\n",
		 red_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->packet->element_types[red_index] != K_UBYTE)
    {
	fprintf (stderr, "Red component type: %u is not K_UBYTE\n",
		 arr_desc->packet->element_types[red_index]);
	return (NULL);
    }
    if (green_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,
		 "green_index: %u greater than number of elements: %u\n",
		 green_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->packet->element_types[green_index] != K_UBYTE)
    {
	fprintf (stderr, "Green component type: %u is not K_UBYTE\n",
		 arr_desc->packet->element_types[green_index]);
	return (NULL);
    }
    if (blue_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,
		 "blue_index: %u greater than number of elements: %u\n",
		 blue_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->packet->element_types[blue_index] != K_UBYTE)
    {
	fprintf (stderr, "Blue component type: %u is not K_UBYTE\n",
		 arr_desc->packet->element_types[blue_index]);
	return (NULL);
    }
    kwin_get_attributes (canvas_get_pixcanvas (canvas),
			 KWIN_ATT_DEPTH, &depth,
			 KWIN_ATT_END);
    if ( (arr_desc->num_levels > 0) && (depth != 24) )
    {
	fprintf (stderr, "%s: Tiling not supported for non 24 bit canvases.\n",
		 function_name);
	return (NULL);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    if ( ( holder = _viewimg_get_canvas_holder (canvas, TRUE, function_name) )
	 == NULL )
    {
	return (NULL);
    }
    /*  Create new viewable image  */
    if ( ( vimage = _viewimg_alloc_vimage () ) == NULL )
    {
	m_error_notify (function_name, "viewable image");
	return (NULL);
    }
    vimage->canvas_holder = holder;
    vimage->tc_multi_desc = multi_desc;
    vimage->tc_arr_desc = arr_desc;
    vimage->tc_slice = slice;
    vimage->tc_hdim = hdim;
    vimage->tc_vdim = vdim;
    vimage->tc_red_index = red_index;
    vimage->tc_green_index = green_index;
    vimage->tc_blue_index = blue_index;
    /*  Compute strides  */
    vimage->tc_hstride = ds_get_packet_size (arr_desc->packet);
    for (dim_count = arr_desc->num_dimensions - 1; dim_count > hdim;
	 --dim_count)
    {
	vimage->tc_hstride *= arr_desc->dimensions[dim_count]->length;
    }
    vimage->tc_vstride = ds_get_packet_size (arr_desc->packet);
    for (dim_count = arr_desc->num_dimensions - 1; dim_count > vdim;
	 --dim_count)
    {
	vimage->tc_vstride *= arr_desc->dimensions[dim_count]->length;
    }
    vimage->num_restrictions = num_restr;
    vimage->restriction_names = (char **) restr_names;
    vimage->restriction_values = (double *) restr_values;
    if (multi_desc == NULL) vimage->astro_projection = NULL;
    else
    {
	index = _viewimg_get_gen_index (multi_desc, arr_desc);
	vimage->astro_projection = wcs_astro_get_associated (multi_desc,index);
    }
    /*  Do not make the first viewable image for this canvas viewable, else
	the  viewimg_make_active  routine will not be able to detect a change
	in the active viewable image and will do nothing if called with this
	newly created viewable image.
	*/
    if (holder->first_image != NULL)
    {
	/*  Insert at beginning of list  */
	holder->first_image->prev = vimage;
    }
    vimage->next = holder->first_image;
    holder->first_image = vimage;
    /*  Attach  */
    if (multi_desc != NULL) ++multi_desc->attachments;
    return (vimage);
}   /*  End Function viewimg_create_rgb  */

/*PUBLIC_FUNCTION*/
ViewableImage *viewimg_create_rgb_sequence (KWorldCanvas canvas,
					    multi_array *multi_desc,
					    array_desc *arr_desc, char *cube,
					    unsigned int hdim,
					    unsigned int vdim,
					    unsigned int fdim,
					    unsigned int red_index,
					    unsigned int green_index,
					    unsigned int blue_index,
					    unsigned num_restr,
					    CONST char **restr_names,
					    CONST double *restr_values)
/*  [SUMMARY] Create sequence of TrueColour viewable images from 3D slice.
    [PURPOSE] This routine will create a sequence of TrueColour viewable image
    objects from a 3-dimensional cube of a Karma data structure. At a later
    time, this sequence of viewable images may be made visible in any order.
    This routine will not cause the canvas to be refreshed.
    <canvas> The world canvas onto which the viewable image may be drawn. Many
    viewable images may be associated with a single canvas.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure must. The routine increments the attachment count on the
    descriptor on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <cube> The start of the cube data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <fdim> The dimension index of the frame dimension (dimension containing the
    sequence). The number of frames is the same as the length of this dimension
    <red_index> The element index of the red component in the data packets.
    <green_index> The element index of the green component in the data packets.
    <blue_index> The element index of the blue component in the data packets.
    [NOTE] The 3 colour components must be of type K_UBYTE.
    <num_restr> The number of matched restrictions. If this is 0, no
    restrictions are recorded.
    <restr_names> The restriction names.
    <restr_values> The restriction values.
    [NOTE] Restriction information is copied into internally allocated
    storage.
    [NOTE] The routine may produce cache data which will vastly increase the
    speed of subsequent operations on this data. Prior to process exit, a call
    MUST be made to [<viewimg_destroy>], otherwise shared memory segments could
    remain after the process exits.
    [NOTE] The ViewableImage objects will be automatically destroyed if the
    canvas is destroyed.
    [RETURNS] A viewable image on success, else NULL.
*/
{
    unsigned int frame_count;
    unsigned int num_frames;
    unsigned int depth;
    unsigned int r_count;
    ViewableImage *vimages;
    uaddr *foffsets;
    char **r_names;
    double *r_values;
    static char function_name[] = "viewimg_create_rgb_sequence";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (cube == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    if (arr_desc->num_levels > 0)
    {
	fprintf (stderr, "%s: Tiled array. May cause problems.\n",
		 function_name);
    }
    /*  Sanity checks  */
    if (fdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "fdim: %u greater than number of dimensions: %u\n",
		 fdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates != NULL)
    {
	fprintf (stderr, "hdim: %u not regularly spaced\n", hdim);
	a_prog_bug (function_name);
    }
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates != NULL)
    {
	fprintf (stderr, "vdim: %u not regularly spaced\n", vdim);
	a_prog_bug (function_name);
    }
    if (red_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr, "red_index: %u greater than number of elements: %u\n",
		 red_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->packet->element_types[red_index] != K_UBYTE)
    {
	fprintf (stderr, "Red component type: %u is not K_UBYTE\n",
		 arr_desc->packet->element_types[red_index]);
	return (NULL);
    }
    if (green_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,
		 "green_index: %u greater than number of elements: %u\n",
		 green_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->packet->element_types[green_index] != K_UBYTE)
    {
	fprintf (stderr, "Green component type: %u is not K_UBYTE\n",
		 arr_desc->packet->element_types[green_index]);
	return (NULL);
    }
    if (blue_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,
		 "blue_index: %u greater than number of elements: %u\n",
		 blue_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->packet->element_types[blue_index] != K_UBYTE)
    {
	fprintf (stderr, "Blue component type: %u is not K_UBYTE\n",
		 arr_desc->packet->element_types[blue_index]);
	return (NULL);
    }
    kwin_get_attributes (canvas_get_pixcanvas (canvas),
			 KWIN_ATT_DEPTH, &depth,
			 KWIN_ATT_END);
    if ( (arr_desc->num_levels > 0) && (depth != 24) )
    {
	fprintf (stderr, "%s: Tiling not supported for non 24 bit canvases.\n",
		 function_name);
	return (NULL);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    foffsets = arr_desc->offsets[fdim];
    num_frames = arr_desc->dimensions[fdim]->length;
    if ( ( vimages = m_alloc (sizeof *vimages * num_frames) ) == NULL )
    {
	m_error_notify (function_name, "array of viewable images");
	return (NULL);
    }
    for (frame_count = 0; frame_count < num_frames; ++frame_count)
    {
	/*  Allocate restrictions  */
	if ( ( r_names = m_alloc ( sizeof *r_names * (num_restr + 1) ) )
	     == NULL )
	    m_abort (function_name, "array of restriction name pointers");
	if ( ( r_values = m_alloc ( sizeof *restr_values * (num_restr + 1) ) )
	     == NULL )
	    m_abort (function_name, "array of restriction values");
	for (r_count = 0; r_count < num_restr; ++r_count)
	{
	    if ( ( r_names[r_count] = st_dup (restr_names[r_count]) ) == NULL )
		m_abort (function_name, "restriction name");
	    r_values[r_count] = restr_values[r_count];
	}
	r_values[num_restr] = ds_get_coordinate (arr_desc->dimensions[fdim],
						 frame_count);
	if ( ( r_names[num_restr] =
	       st_dup (arr_desc->dimensions[fdim]->name) ) == NULL )
	    m_abort (function_name, "restriction name");
	if ( ( vimages[frame_count] =
	       viewimg_create_rgb (canvas, multi_desc, arr_desc,
				   cube + foffsets[frame_count],
				   hdim, vdim,
				   red_index, green_index, blue_index,
				   num_restr + 1, (CONST char **) r_names,
				   r_values) )
	     == NULL )
	{
	    for (; frame_count > 0; --frame_count)
		viewimg_destroy (vimages[frame_count - 1]);
	    m_free (vimages);
	    return (NULL);
	}
    }
    return (vimages);
}   /*  End Function viewimg_create_rgb_sequence  */

/*PUBLIC_FUNCTION*/
flag viewimg_make_active (ViewableImage vimage)
/*  [SUMMARY] Make viewable image active and possibly refresh.
    [PURPOSE] This routine will make a viewable image the active image for its
    associated world canvas. The canvas is then refreshed (possibly resized),
    provided that the new viewable image was not already active.
    <vimage> The viewable image.
    [RETURNS] TRUE on success, else FALSE.
    [SEE ALSO] [<viewimg_set_active>].
*/
{
    CanvasHolder holder;
    static char function_name[] = "viewimg_make_active";

    VERIFY_USED_VIMAGE (vimage);
    holder = vimage->canvas_holder;
    if (vimage == holder->active_image) return (TRUE);
    return viewimg_set_active (vimage, TRUE);
}   /*  End Function viewimg_make_active  */

/*PUBLIC_FUNCTION*/
flag viewimg_set_active (ViewableImage vimage, flag refresh)
/*  [SUMMARY] Make viewable image active with controlled refresh.
    [PURPOSE] This routine will make a viewable image the active image for its
    associated world canvas.
    <vimage> The viewable image.
    <refresh> If TRUE, the canvas is always refreshed, if FALSE, the canvas is
    not refreshed.
    [RETURNS] TRUE on success, else FALSE.
    [SEE ALSO] [<viewimg_make_active>].
*/
{
    CanvasHolder holder;
    flag visible;
    unsigned int hdim, vdim;
    dim_desc **dimensions;
    static char function_name[] = "viewimg_set_active";

    VERIFY_USED_VIMAGE (vimage);
    FLAG_VERIFY (refresh);
    holder = vimage->canvas_holder;
    kwin_get_attributes (canvas_get_pixcanvas (holder->canvas),
			 KWIN_ATT_VISIBLE, &visible,
			 KWIN_ATT_END);
    if (!visible) refresh = FALSE;
    if (vimage == holder->active_image)
    {
	if (refresh)
	{
	    return canvas_resize (holder->canvas, NULL, FALSE);
	}
	return (TRUE);
    }
    holder->active_image = vimage;
    if (holder->auto_wcs) canvas_use_astro_transform(holder->canvas,
						     vimage->astro_projection);
    if (vimage->tc_arr_desc == NULL)
    {
	dimensions = vimage->pc_arr_desc->dimensions;
	hdim = vimage->pc_hdim;
	vdim = vimage->pc_vdim;
    }
    else
    {
	dimensions = vimage->tc_arr_desc->dimensions;
	hdim = vimage->tc_hdim;
	vdim = vimage->tc_vdim;
    }
    if ( !canvas_specify (holder->canvas,
			  dimensions[hdim]->name,
			  dimensions[vdim]->name,
			  vimage->num_restrictions,
			  (CONST char **) vimage->restriction_names,
			  vimage->restriction_values) )
    {
	return (FALSE);
    }
    if (refresh)
    {
	return canvas_resize (holder->canvas, NULL, FALSE);
    }
    return (TRUE);
}   /*  End Function viewimg_set_active  */

/*OBSOLETE_FUNCTION*/
void viewimg_control_autoscaling (KWorldCanvas canvas,
				  flag auto_x, flag auto_y, flag auto_v,
				  flag int_x, flag int_y,
				  flag maintain_aspect_ratio)
/*  This routine will control the autoscaling options used when viewable images
    are displayed on their associated world canvas.
    THIS FUNCTION IS OBSOLETE. USE  VIEWIMG_SET_CANVAS_ATTRIBUTES  INSTEAD.
    <canvas> The world canvas object.
    If  auto_x  is TRUE, then the horizontal window scaling information for the
    canvas is set to the horizontal range of the active viewable image. The
    default is TRUE.
    If  auto_y  is TRUE, then the vertical window scaling information for the
    canvas is set to the vertical range of the active viewable image. The
    default is TRUE.
    If  auto_v  is TRUE, then the value/ colourmap window scaling information
    for the canvas is set to the data range of the active viewable image. The
    default is TRUE.
    If  int_x  is TRUE, then when the number of horizontal data values in a
    viewable image is greater than the number of horizontal pixels (zoom out),
    as many pixels are used so that there are an integral number of data values
    per pixel; if the number of data values is less than the number of pixels
    (zoom in), as many pixels are used so that there are an integral number of
    pixels per data value. The default is TRUE.
    The  int_y  parameter is similar to the  int_x  parameter, except that it
    applies to vertical pixels. The default is TRUE.
    If  maintain_aspect_ratio  is TRUE, then any zoom in or out factors will be
    the same (ie. the image aspect ratio is preserved). The default is FALSE.
    The canvas is not refreshed by this operation.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "viewimg_control_autoscaling";

    fprintf (stderr,
		    "Function: %s will be removed in Karma version 2.0\n",
		    function_name);
    fprintf (stderr, "Use:  viewimg_set_canvas_attributes  instead.\n");
/*  This will be activated in Karma version 2.0
    a_prog_bug (func_name);
*/
    viewimg_set_canvas_attributes (canvas,
				   VIEWIMG_ATT_AUTO_X, auto_x,
				   VIEWIMG_ATT_AUTO_Y, auto_y,
				   VIEWIMG_ATT_AUTO_V, auto_v,
				   VIEWIMG_ATT_INT_X, int_x,
				   VIEWIMG_ATT_INT_Y, int_y,
				   VIEWIMG_ATT_MAINTAIN_ASPECT,
				   maintain_aspect_ratio,
				   VIEWIMG_ATT_END);
}   /*  End Function viewimg_control_autoscaling  */

/*PUBLIC_FUNCTION*/
flag viewimg_register_data_change (ViewableImage vimage)
/*  [SUMMARY] Notify data for viewable image has changed.
    [PURPOSE] This routine will register a change in the Karma data structure
    associated with a viewable image. If the viewable image is active, it will
    be immediately redrawn on its canvas.
    <vimage> The viewable image.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    static char function_name[] = "viewimg_register_data_change";

    VERIFY_USED_VIMAGE (vimage);
    holder = vimage->canvas_holder;
    vimage->changed = TRUE;
    vimage->value_min = TOOBIG;
    vimage->value_max = TOOBIG;
    if (vimage == holder->active_image)
    {
	/*  Active image: refresh  */
	return canvas_resize (holder->canvas,
			      (struct win_scale_type *) NULL, FALSE);
    }
    return (TRUE);
}   /*  End Function viewimg_register_data_change  */

/*PUBLIC_FUNCTION*/
void viewimg_destroy (ViewableImage vimage)
/*  [SUMMARY] Destroy viewable image.
    [PURPOSE] This routine will destroy a viewable image. If this is not called
    prior to process exit, shared memory segments could remain after the
    process exits.
    <vimage> The viewable image.
    [NOTE] The associated <<multi_array>> descriptor is also deallocated (or
    at least, the attachment count is decreased).
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    unsigned int count;
    static char function_name[] = "viewimg_destroy";

    VERIFY_USED_VIMAGE (vimage);
    holder = vimage->canvas_holder;
    if (holder->active_image == vimage)
	canvas_specify (holder->canvas, NULL, NULL, 0, NULL, NULL);
    kwin_free_cache_data (vimage->cache);
    if ( (vimage == holder->active_image) &&
	 (vimage->astro_projection != NULL) && holder->auto_wcs )
    {
	canvas_use_astro_transform (holder->canvas, NULL);
    }
    ds_dealloc_multi (vimage->pc_multi_desc);
    ds_dealloc_multi (vimage->tc_multi_desc);
    /*  Remove entry  */
    if (vimage->next != NULL) vimage->next->prev = vimage->prev;
    if (vimage->prev != NULL) vimage->prev->next = vimage->next;
    if (vimage == holder->first_image)
    {
	/*  Remove from start of list  */
	holder->first_image = vimage->next;
    }
    if (vimage == holder->active_image) holder->active_image = NULL;
    if ( (vimage->restriction_names != NULL) &&
	 (vimage->restriction_names != vimage->inline_restr_names) )
    {
	for (count = 0; count < vimage->num_restrictions; ++count)
	{
	    if (vimage->restriction_names[count] != NULL)
		m_free (vimage->restriction_names[count]);
	}
	m_free (vimage->restriction_names);
    }
    if ( (vimage->restriction_values != NULL) &&
	 (vimage->restriction_values != vimage->inline_restr_values) )
	m_free (vimage->restriction_values);
    _viewimg_free_vimage (vimage);
}   /*  End Function viewimg_destroy  */

/*PUBLIC_FUNCTION*/
ViewableImage viewimg_get_active (KWorldCanvas canvas)
/*  [SUMMARY] Get the active ViewableImage associated with a world canvas.
    <canvas> The world canvas object.
    [RETURNS] The active viewable image on success, else NULL (indicating no
    viewable image is active for the canvas).
*/
{
    CanvasHolder holder;
    static char function_name[] = "viewimg_get_active";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL canvas passed\n");
	a_prog_bug (function_name);
    }
    /*  Dummy operation to ensure canvas is OK  */
    canvas_get_attributes (canvas, CANVAS_ATT_END);
    if ( ( holder = _viewimg_get_canvas_holder (canvas, TRUE, function_name) )
	 == NULL )
    {
	m_abort (function_name, "canvas holder");
    }
    return (holder->active_image);
}   /*  End Function viewimg_get_active  */

/*PUBLIC_FUNCTION*/
flag viewimg_test_active (ViewableImage vimage)
/*  [SUMMARY] Test if viewable image is active.
    [PURPOSE] This routine will test if a viewable image is the active image
    for its associated world canvas.
    <vimage> The viewable image.
    [RETURNS] TRUE if the viewable image is actice, else FALSE.
*/
{
    CanvasHolder holder;
    static char function_name[] = "viewimg_test_active";

    VERIFY_USED_VIMAGE (vimage);
    holder = vimage->canvas_holder;
    if (vimage == holder->active_image) return (TRUE);
    return (FALSE);
}   /*  End Function viewimg_test_active  */

/*PUBLIC_FUNCTION*/
KCallbackFunc viewimg_register_position_event_func (KWorldCanvas canvas,
						    flag (*func) (),
						    void *f_info)
/*  [SUMMARY] Register position event callback.
    [PURPOSE] This routine will register a position event function for a world
    canvas which has a number of ViewableImage objects associated with it.
    The position event function will be called whenever a position event on the
    canvas has not been consumed. Many position event functions may be
    registered per canvas. The first function registered is the first function
    called upon a position event.
    <canvas> The world canvas object.
    <func> The function that is called when a position event occurs. The
    prototype function is [<viewimg_PROTO_position_func>].
    <f_info> The initial arbitrary function information pointer.
    [RETURNS] A handle to a KCallbackFunc object.
*/
{
    CanvasHolder holder;
    static char function_name[] = "viewimg_register_position_func";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL canvas passed\n");
	a_prog_bug (function_name);
    }
    if ( ( holder = _viewimg_get_canvas_holder (canvas, TRUE, function_name) )
	 == NULL )
    {
	m_abort (function_name, "canvas holder");
    }
    return c_register_callback (&holder->position_list,
				_viewimg_position_event_func,
				holder, f_info, TRUE, (void *) func, FALSE,
				TRUE);
}   /*  End Function viewimg_register_position_func  */

/*PUBLIC_FUNCTION*/
flag viewimg_fill_ellipse (ViewableImage vimage,
			   double centre_x, double centre_y,
			   double radius_x, double radius_y, double value[2])
/*  [SUMMARY] Draw an ellipse into an array associated with a viewable image.
    [PURPOSE] This routine will draw a filled ellipse into a 2 dimensional
    slice of data associated with a viewable image.
    <vimage> The viewable image.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius_x> The horizontal radius.
    <radius_y> The vertical radius.
    <value> The complex value to fill the ellipse. This must be of type
    K_DCOMPLEX.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    packet_desc *pack_desc;
    array_desc *arr_desc;
    static char function_name[] = "viewimg_fill_ellipse";

    VERIFY_USED_VIMAGE (vimage);
    if ( !SAFE_ALIGNMENT ( value, sizeof (double) ) )
    {
	fprintf (stderr, "value  address: %p not on a required boundary\n",
		 value);
	a_prog_bug (function_name);
    }
    if (vimage->tc_arr_desc != NULL)
    {
	fprintf (stderr, "%s: TrueColour images not supported yet\n",
		 function_name);
	return (FALSE);
    }
    arr_desc = vimage->pc_arr_desc;
    pack_desc = arr_desc->packet;
    return ds_draw_ellipse
	(vimage->pc_slice,
	 pack_desc->element_types[vimage->pc_elem_index],
	 arr_desc->dimensions[vimage->pc_hdim], vimage->pc_hstride,
	 arr_desc->dimensions[vimage->pc_vdim], vimage->pc_vstride,
	 centre_x, centre_y, radius_x, radius_y, value);
}   /*  End Function viewimg_fill_ellipse  */

/*PUBLIC_FUNCTION*/
flag viewimg_fill_polygon (ViewableImage vimage, edit_coord *coords,
			   unsigned int num_vertices, double value[2])
/*  [SUMMARY] Draw a polygon into an array associated with a viewable image.
    [PURPOSE] This routine will draw a filled polygon into a 2 dimensional
    slice of data associated with a viewable image.
    <vimage> The viewable image.
    <coords> The array of world co-ordinates of vertices of the polygon.
    <num_vertices> The number of vertices in the polygon.
    <value> The complex value to fill the polygon with. This must be of type
    K_DCOMPLEX.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    edit_coord lcoords[BUF_SIZE];
    static char function_name[] = "viewimg_fill_polygon";

    VERIFY_USED_VIMAGE (vimage);
    if ( !SAFE_ALIGNMENT ( value, sizeof (double) ) )
    {
	fprintf (stderr, "value  address: %p not on a required boundary\n",
		 value);
	a_prog_bug (function_name);
    }
    if (vimage->tc_arr_desc != NULL)
    {
	fprintf (stderr, "%s: TrueColour images not supported yet\n",
			function_name);
	return (FALSE);
    }
    if (num_vertices > BUF_SIZE)
    {
	fprintf (stderr, "%s: too many vertices: %u\n",
		 function_name, num_vertices);
	return (FALSE);
    }
    m_copy (lcoords, coords, num_vertices * sizeof *coords);
    for (count = 0; count < num_vertices; ++count)
    {
	canvas_coords_transform (vimage->canvas_holder->canvas, 1,
				 &lcoords[count].abscissa, TRUE,
				 &lcoords[count].ordinate, TRUE);
    }
    arr_desc = vimage->pc_arr_desc;
    pack_desc = arr_desc->packet;
    return ds_draw_polygon
	(vimage->pc_slice,
	 pack_desc->element_types[vimage->pc_elem_index],
	 arr_desc->dimensions[vimage->pc_hdim], vimage->pc_hstride,
	 arr_desc->dimensions[vimage->pc_vdim], vimage->pc_vstride,
	 lcoords, num_vertices, value);
}   /*  End Function viewimg_fill_polygon  */

/*PUBLIC_FUNCTION*/
void viewimg_get_canvas_attributes (KWorldCanvas canvas, ...)
/*  [SUMMARY] Get the viewable image attributes for a world canvas.
    <canvas> The world canvas object.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the VIEWIMG_ATT_END.
    See [<viewimg_CANVAS_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    va_list argp;
    unsigned int att_key;
    static char function_name[] = "viewimg_get_canvas_attributes";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL canvas passed\n");
	a_prog_bug (function_name);
    }
    /*  Dummy operation to ensure canvas is OK  */
    canvas_get_attributes (canvas, CANVAS_ATT_END);
    if ( ( holder = _viewimg_get_canvas_holder (canvas, TRUE, function_name) )
	 == NULL )
    {
	m_abort (function_name, "canvas holder");
    }
    va_start (argp, canvas);
    while ( ( att_key = va_arg (argp, unsigned int) ) != VIEWIMG_ATT_END )
    {
	switch (att_key)
	{
	  case VIEWIMG_ATT_AUTO_X:
	    *( va_arg (argp, unsigned int *) ) = holder->auto_x;
	    break;
	  case VIEWIMG_ATT_AUTO_Y:
	    *( va_arg (argp, unsigned int *) ) = holder->auto_y;
	    break;
	  case VIEWIMG_ATT_AUTO_V:
	    *( va_arg (argp, unsigned int *) ) = holder->auto_v;
	    break;
	  case VIEWIMG_ATT_INT_X:
	    *( va_arg (argp, unsigned int *) ) = holder->int_x;
	    break;
	  case VIEWIMG_ATT_INT_Y:
	    *( va_arg (argp, unsigned int *) ) = holder->int_y;
	    break;
	  case VIEWIMG_ATT_MAINTAIN_ASPECT:
	    *( va_arg (argp, unsigned int *) ) = holder->maintain_aspect_ratio;
	    break;
	  case VIEWIMG_ATT_ALLOW_TRUNCATION:
	    *( va_arg (argp, unsigned int *) ) = holder->allow_truncation;
	    break;
	  case VIEWIMG_ATT_ENABLE_PANNING:
	    *( va_arg (argp, flag *) ) = holder->enable_panning;
	    break;
	  case VIEWIMG_ATT_PAN_CENTRE_X:
	    *( va_arg (argp, unsigned long *) ) = holder->pan_centre_x;
	    break;
	  case VIEWIMG_ATT_PAN_CENTRE_Y:
	    *( va_arg (argp, unsigned long *) ) = holder->pan_centre_y;
	    break;
	  case VIEWIMG_ATT_PAN_MAGNIFICATION:
	    *( va_arg (argp, unsigned int *) ) = holder->pan_magnification_x;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (argp);
}   /*  End Function viewimg_get_canvas_attributes  */

/*PUBLIC_FUNCTION*/
void viewimg_set_canvas_attributes (KWorldCanvas canvas, ...)
/*  [SUMMARY] Set the viewable image attributes for a world canvas.
    [PURPOSE] This routine will control the autoscaling options used when
    viewable images are displayed on their associated world canvas.
    <canvas> The world canvas.
    [VARARGS] The list of parameter attribute-key attribute-value pairs
    must follow. This list must be terminated with the VIEWIMG_ATT_END.
    See [<viewimg_CANVAS_ATTRIBUTES>] for the list of attributes.
    [NOTE] The canvas is not refreshed by this operation.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder;
    va_list argp;
    flag bool;
    unsigned int att_key;
    static char function_name[] = "viewimg_set_canvas_attributes";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL canvas passed\n");
	a_prog_bug (function_name);
    }
    /*  Dummy operation to ensure canvas is OK  */
    canvas_get_attributes (canvas, CANVAS_ATT_END);
    if ( ( holder = _viewimg_get_canvas_holder (canvas, TRUE, function_name) )
	 == NULL )
    {
	m_abort (function_name, "canvas holder");
    }
    va_start (argp, canvas);
    while ( ( att_key = va_arg (argp, unsigned int) ) != VIEWIMG_ATT_END )
    {
	switch (att_key)
	{
	  case VIEWIMG_ATT_AUTO_X:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->auto_x = bool;
	    break;
	  case VIEWIMG_ATT_AUTO_Y:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->auto_y = bool;
	    break;
	  case VIEWIMG_ATT_AUTO_V:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->auto_v = bool;
	    break;
	  case VIEWIMG_ATT_INT_X:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->int_x = bool;
	    break;
	  case VIEWIMG_ATT_INT_Y:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->int_y = bool;
	    break;
	  case VIEWIMG_ATT_MAINTAIN_ASPECT:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->maintain_aspect_ratio = bool;
	    break;
	  case VIEWIMG_ATT_ALLOW_TRUNCATION:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->allow_truncation = bool;
	    break;
	  case VIEWIMG_ATT_ENABLE_PANNING:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->enable_panning = bool;
	    holder->set_size = bool ?
		_viewimg_set_size_pan : _viewimg_set_size_normal;
	    break;
	  case VIEWIMG_ATT_FAST_PANNER:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->fast_panner = bool;
	    break;
	  case VIEWIMG_ATT_PAN_CENTRE_X:
	    holder->pan_centre_x = va_arg (argp, unsigned long);
	    break;
	  case VIEWIMG_ATT_PAN_CENTRE_Y:
	    holder->pan_centre_y = va_arg (argp, unsigned long);
	    break;
	  case VIEWIMG_ATT_PAN_MAGNIFICATION:
	    holder->pan_magnification_x = va_arg (argp, unsigned int);
	    holder->pan_magnification_y = holder->pan_magnification_x;
	    break;
	  case VIEWIMG_ATT_AUTO_WCS:
	    bool = va_arg (argp, flag);
	    FLAG_VERIFY (bool);
	    holder->auto_wcs = bool;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (argp);
}   /*  End Function viewimg_set_canvas_attributes  */

/*PUBLIC_FUNCTION*/
void viewimg_get_attributes (ViewableImage vimage, ...)
/*  [SUMMARY] Get the attributes for a viewable image.
    <vimage> The ViewableImage.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the VIEWIMG_VATT_END. See
    [<viewimg_VIMAGE_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    flag truecolour;
    unsigned int att_key;
    unsigned int hdim, vdim;
    char *slice;
    array_desc *arr_desc;
    multi_array *multi_desc;
    static char function_name[] = "viewimg_get_attributes";

    VERIFY_USED_VIMAGE (vimage);
    va_start (argp, vimage);
    truecolour = (vimage->tc_arr_desc == NULL) ? FALSE : TRUE;
    arr_desc = truecolour ? vimage->tc_arr_desc : vimage->pc_arr_desc;
    slice = truecolour ? vimage->tc_slice : vimage->pc_slice;
    hdim = truecolour ? vimage->tc_hdim : vimage->pc_hdim;
    vdim = truecolour ? vimage->tc_vdim : vimage->pc_vdim;
    multi_desc = truecolour ? vimage->tc_multi_desc : vimage->pc_multi_desc;
    while ( ( att_key = va_arg (argp, unsigned int) ) != VIEWIMG_VATT_END )
    {
	switch (att_key)
	{
	  case VIEWIMG_VATT_TRUECOLOUR:
	    *( va_arg (argp, flag *) ) = truecolour;
	    break;
	  case VIEWIMG_VATT_ARRAY_DESC:
	    *( va_arg (argp, array_desc **) ) = arr_desc;
	    break;
	  case VIEWIMG_VATT_SLICE:
	    *( va_arg (argp, char **) ) = slice;
	    break;
	  case VIEWIMG_VATT_HDIM:
	    *( va_arg (argp, unsigned int *) ) = hdim;
	    break;
	  case VIEWIMG_VATT_VDIM:
	    *( va_arg (argp, unsigned int *) ) = vdim;
	    break;
	  case VIEWIMG_VATT_PSEUDO_INDEX:
	    if (truecolour)
	    {
		fprintf (stderr, "TrueColour image!\n");
		a_prog_bug (function_name);
	    }
	    *( va_arg (argp, unsigned int *) ) = vimage->pc_elem_index;
	    break;
	  case VIEWIMG_VATT_RED_INDEX:
	    if (!truecolour)
	    {
		fprintf (stderr, "PseudoColour image!\n");
		a_prog_bug (function_name);
	    }
	    *( va_arg (argp, unsigned int *) ) = vimage->tc_red_index;
	    break;
	  case VIEWIMG_VATT_GREEN_INDEX:
	    if (!truecolour)
	    {
		fprintf (stderr, "PseudoColour image!\n");
		a_prog_bug (function_name);
	    }
	    *( va_arg (argp, unsigned int *) ) = vimage->tc_green_index;
	    break;
	  case VIEWIMG_VATT_BLUE_INDEX:
	    if (!truecolour)
	    {
		fprintf (stderr, "PseudoColour image!\n");
		a_prog_bug (function_name);
	    }
	    *( va_arg (argp, unsigned int *) ) = vimage->tc_blue_index;
	    break;
	  case VIEWIMG_VATT_MULTI_ARRAY:
	    *( va_arg (argp, multi_array **) ) = multi_desc;
	    break;
	  case VIEWIMG_VATT_DATA_SCALE:
	    if (truecolour)
	    {
		fprintf (stderr,
			 "Data scale not defined for TrueColour image\n");
		a_prog_bug (function_name);
	    }
	    *( va_arg (argp, double *) ) = vimage->pc_data_scale;
	    break;
	  case VIEWIMG_VATT_DATA_OFFSET:
	    if (truecolour)
	    {
		fprintf (stderr,
			 "Data offset not defined for TrueColour image\n");
		a_prog_bug (function_name);
	    }
	    *( va_arg (argp, double *) ) = vimage->pc_data_offset;
	    break;
	  case VIEWIMG_VATT_VALUE_MIN:
	    if (truecolour)
	    {
		fprintf (stderr,
			 "Override min value not defined for TrueColour image\n");
		a_prog_bug (function_name);
	    }
	    *( va_arg (argp, double *) ) = vimage->override.value_min;
	    break;
	  case VIEWIMG_VATT_VALUE_MAX:
	    if (truecolour)
	    {
		fprintf (stderr,
			 "Override max value not defined for TrueColour image\n");
		a_prog_bug (function_name);
	    }
	    *( va_arg (argp, double *) ) = vimage->override.value_max;
	    break;
	  case VIEWIMG_VATT_WCS_ASTRO:
	    *( va_arg (argp, KwcsAstro *) ) = vimage->astro_projection;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (argp);
}   /*  End Function viewimg_get_attributes  */

/*PUBLIC_FUNCTION*/
void viewimg_set_attributes (ViewableImage vimage, ...)
/*  [SUMMARY] Set the attributes for a viewable image.
    <vimage> The ViewableImage.
    [VARARGS] The list of parameter attribute-key attribute-value pairs
    must follow. This list must be terminated with the VIEWIMG_VATT_END. See
    [<viewimg_VIMAGE_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    flag truecolour;
    unsigned int att_key;
    static char function_name[] = "viewimg_set_attributes";

    VERIFY_USED_VIMAGE (vimage);
    va_start (argp, vimage);
    truecolour = (vimage->tc_arr_desc == NULL) ? FALSE : TRUE;
    while ( ( att_key = va_arg (argp, unsigned int) ) != VIEWIMG_VATT_END )
    {
	switch (att_key)
	{
	  case VIEWIMG_VATT_DATA_SCALE:
	    if (truecolour)
	    {
		fprintf (stderr,
			 "Data scale not defined for TrueColour image\n");
		a_prog_bug (function_name);
	    }
	    vimage->pc_data_scale = va_arg (argp, double);
	    break;
	  case VIEWIMG_VATT_DATA_OFFSET:
	    if (truecolour)
	    {
		fprintf (stderr,
			 "Data offset not defined for TrueColour image\n");
		a_prog_bug (function_name);
	    }
	    vimage->pc_data_offset = va_arg (argp, double);
	    break;
	  case VIEWIMG_VATT_VALUE_MIN:
	    if (truecolour)
	    {
		fprintf (stderr,
			 "Override min value not defined for TrueColour image\n");
		a_prog_bug (function_name);
	    }
	    vimage->override.value_min = va_arg (argp, double);
	    break;
	  case VIEWIMG_VATT_VALUE_MAX:
	    if (truecolour)
	    {
		fprintf (stderr,
			 "Override max value not defined for TrueColour image\n");
		a_prog_bug (function_name);
	    }
	    vimage->override.value_max = va_arg (argp, double);
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (argp);
}   /*  End Function viewimg_set_attributes  */

/*PUBLIC_FUNCTION*/
void viewimg_set_array_attributes (ViewableImage *vimages, unsigned int len,
				   ...)
/*  [SUMMARY] Set the attributes for an array of viewable images.
    <vimages> The array of ViewableImage objects.
    <len> The length of the array.
    [VARARGS] The list of parameter attribute-key attribute-value pairs
    must follow. This list must be terminated with the VIEWIMG_VATT_END. See
    [<viewimg_VIMAGE_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key;
    unsigned int count;
    double dval;
    static char function_name[] = "viewimg_set_attributes";

    if (len < 1) return;
    if (vimages == NULL)
    {
	fprintf (stderr, "NULL vimages pointer passed\n");
	a_prog_bug (function_name);
    }
    for (count = 0; count < len; ++count) VERIFY_USED_VIMAGE (vimages[count]);
    va_start (argp, len);
    while ( ( att_key = va_arg (argp, unsigned int) ) != VIEWIMG_VATT_END )
    {
	switch (att_key)
	{
	  case VIEWIMG_VATT_DATA_SCALE:
	    dval = va_arg (argp, double);
	    for (count = 0; count < len; ++count)
	    {
		if (vimages[count]->tc_arr_desc != NULL)
		{
		    fprintf (stderr,
			     "Data scale not defined for TrueColour image\n");
		    a_prog_bug (function_name);
		}
		vimages[count]->pc_data_scale = dval;
	    }
	    break;
	  case VIEWIMG_VATT_DATA_OFFSET:
	    dval = va_arg (argp, double);
	    for (count = 0; count < len; ++count)
	    {
		if (vimages[count]->tc_arr_desc != NULL)
		{
		    fprintf (stderr,
			     "Data offset not defined for TrueColour image\n");
		    a_prog_bug (function_name);
		}
		vimages[count]->pc_data_offset = dval;
	    }
	    break;
	  case VIEWIMG_VATT_VALUE_MIN:
	    dval = va_arg (argp, double);
	    for (count = 0; count < len; ++count)
	    {
		if (vimages[count]->tc_arr_desc != NULL)
		{
		    fprintf (stderr,
			     "Override min value not defined for TrueColour image\n");
		    a_prog_bug (function_name);
		}
		vimages[count]->override.value_min = dval;
	    }
	    break;
	  case VIEWIMG_VATT_VALUE_MAX:
	    dval = va_arg (argp, double);
	    for (count = 0; count < len; ++count)
	    {
		if (vimages[count]->tc_arr_desc != NULL)
		{
		    fprintf (stderr,
			     "Override max value not defined for TrueColour image\n");
		    a_prog_bug (function_name);
		}
		vimages[count]->override.value_max = dval;
	    }
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (argp);
}   /*  End Function viewimg_set_array_attributes  */

/*EXPERIMENTAL_FUNCTION*/
flag viewimg_partial_refresh (KWorldCanvas canvas, unsigned int num_areas,
			      KPixCanvasRefreshArea *areas)
/*  [SUMMARY] Perform a partial refresh of a canvas.
    [PURPOSE] This routine will perform a partial refresh of a canvas. This
    call is similar to <<kwin_partial_refresh>> except that areas are not
    cleared prior to drawing if they lie within the image boundary.
    <canvas> The world canvas.
    <num_areas> The number of areas in the pixel canvas to refresh.
    <areas> The list of areas to refresh. The values here are updated to ensure
    all points lie within the boundaries of the underlaying pixel canvas.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    CanvasHolder holder;
    KPixCanvas pixcanvas;
    int x_offset, y_offset, x_pixels, y_pixels;
    int min_x_clear, min_y_clear, max_x_clear, max_y_clear;
    long clear_area, boundary_area;
    unsigned int count;
    static char function_name[] = "viewimg_partial_refresh";

    if (canvas == NULL)
    {
	fprintf (stderr, "NULL canvas passed\n");
	a_prog_bug (function_name);
    }
    canvas_get_attributes (canvas,
			   CANVAS_ATT_X_OFFSET, &x_offset,
			   CANVAS_ATT_Y_OFFSET, &y_offset,
			   CANVAS_ATT_X_PIXELS, &x_pixels,
			   CANVAS_ATT_Y_PIXELS, &y_pixels,
			   CANVAS_ATT_END);
    if ( ( holder = _viewimg_get_canvas_holder (canvas, TRUE, function_name) )
	 == NULL )
    {
	m_abort (function_name, "canvas holder");
    }
    pixcanvas = canvas_get_pixcanvas (canvas);
    if (holder->active_image == NULL)
    {
	return kwin_partial_refresh (pixcanvas, num_areas, areas, FALSE);
    }
    clear_area = 0;
    min_x_clear = x_offset + x_pixels;
    max_x_clear = -1;
    min_y_clear = y_offset + y_pixels;
    max_y_clear = -1;
    for (count = 0; count < num_areas; ++count)
    {
	if (!areas[count].clear) continue;
	/*  Area should be cleared  */
	if ( (areas[count].startx >= x_offset) &&
	     (areas[count].endx < x_offset + x_pixels) &&
	     (areas[count].starty >= y_offset) &&
	     (areas[count].endy < y_offset + y_pixels) )
	{
	    /*  This area is entirely within image: no need to clear  */
	    areas[count].clear = FALSE;
	    continue;
	}
	/*  This area is at least partly outside the image and must be cleared:
	    accumulate clear area and update clear boundary  */
	clear_area += (areas[count].endx - areas[count].startx + 1) *
	    (areas[count].endy - areas[count].starty + 1);
	if (areas[count].startx < min_x_clear) min_x_clear=areas[count].startx;
	if (areas[count].endx > max_x_clear) max_x_clear = areas[count].endx;
	if (areas[count].starty < min_y_clear) min_y_clear=areas[count].starty;
	if (areas[count].endy > max_y_clear) max_y_clear = areas[count].endy;
    }
    if (clear_area < 1)
    {
	/*  No areas require clearing: simple  */
	return kwin_partial_refresh (pixcanvas, num_areas, areas, FALSE);
    }
    /*  Some areas require clearing: check if it would be faster to simply
	clear the requisite area around the image  */
    boundary_area = 0;
    if (min_x_clear < x_offset)
    {
	/*  Left of image requires a clear  */
	boundary_area += (x_offset - min_x_clear) * (max_y_clear - min_y_clear
						     + 1);
    }
    if (max_x_clear >= x_offset + x_pixels)
    {
	/*  Right of image requires a clear  */
	boundary_area += (max_x_clear - x_offset - x_pixels + 1) *
	    (max_y_clear - min_y_clear + 1);
    }
    if (min_y_clear < y_offset)
    {
	/*  Top of image requires a clear  */
	boundary_area += (y_offset - min_y_clear) * x_pixels;
    }
    if (max_y_clear >= y_offset + y_pixels)
    {
	/*  Bottom of image requires a clear  */
	boundary_area += (max_y_clear - y_offset - y_pixels + 1) * x_pixels;
    }
    if (boundary_area > clear_area)
    {
	/*  Faster to clear areas  */
	return kwin_partial_refresh (pixcanvas, num_areas, areas, FALSE);
    }
    /*  Faster to clear boundary: do not clear areas  */
    for (count = 0; count < num_areas; ++count) areas[count].clear = FALSE;
    /*  Clear boundary  */
    if (min_x_clear < x_offset)
    {
	/*  Left of image requires a clear  */
	kwin_clear (pixcanvas, min_x_clear, min_y_clear,
		    x_offset - min_x_clear, max_y_clear - min_y_clear + 1);
    }
    if (max_x_clear >= x_offset + x_pixels)
    {
	/*  Right of image requires a clear  */
	kwin_clear (pixcanvas, x_offset + x_pixels, min_y_clear,
		    max_x_clear - x_offset - x_pixels + 1,
		    max_y_clear - min_y_clear + 1);
    }
    if (min_y_clear < y_offset)
    {
	/*  Top of image requires a clear  */
	kwin_clear (pixcanvas, x_offset, min_y_clear,
		    x_pixels, y_offset - min_y_clear);
    }
    if (max_y_clear >= y_offset + y_pixels)
    {
	/*  Bottom of image requires a clear  */
	kwin_clear (pixcanvas, x_offset, y_offset + y_pixels,
		    x_pixels, max_y_clear - y_offset - y_pixels + 1);
    }
    return kwin_partial_refresh (pixcanvas, num_areas, areas, FALSE);
}   /*  End Function viewimg_partial_refresh  */

/*PUBLIC_FUNCTION*/
KWorldCanvas viewimg_get_worldcanvas (ViewableImage vimage)
/*  [SUMMARY] Get the world canvas for a viewable image.
    <vimage> The ViewableImage object.
    [RETURNS] The KWorldCanvas object.
*/
{
    static char function_name[] = "viewimg_get_worldcanvas";

    VERIFY_USED_VIMAGE (vimage);
    return (vimage->canvas_holder->canvas);
}   /*  End Function viewimg_get_worldcanvas  */

/*UNPUBLISHED_FUNCTION*/
void viewimg_zoom (ViewableImage vimage, double x, double y, flag relative,
		   double zoom)
/*  [SUMMARY] Zoom in or out.
    <vimage> The viewable image.
    <x> The the zoom centre horizontal co-ordinate.
    <y> The the zoom centre vertical co-ordinate.
    <relative> If TRUE, the co-ordinates are scaled from 0.0 to 1.0, otherwise
    they are linear co-ordinates.
    <zoom> The zoom factor. If this is greater than 1 the image is zoomed in
    from its current zoom factor.
    [RETURNS] Nothing.
*/
{
    flag ok;
    KPixCanvas pixcanvas;
    CanvasHolder holder;
    set_size_func_type old_func;
    int x_pixels, y_pixels;
    unsigned int hdim_index, vdim_index;
    signed long num_image_pixels, left_x, right_x, bottom_y, top_y;
    double win_left_x, win_right_x, win_bottom_y, win_top_y;
    dim_desc *hdim, *vdim;
    array_desc *arr_desc;
    static char function_name[] = "viewimg_zoom";

    VERIFY_USED_VIMAGE (vimage);
    if (zoom <= 0.0)
    {
	fprintf (stderr, "Illegal zoom factor: %g\n", zoom);
	a_prog_bug (function_name);
    }
#ifdef DEBUG
    fprintf (stderr, "%s: %g %g  zoom: %g\n", function_name, x, y, zoom);
#endif
    holder = vimage->canvas_holder;
    pixcanvas = canvas_get_pixcanvas (holder->canvas);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hdim_index,
			    VIEWIMG_VATT_VDIM, &vdim_index,
			    VIEWIMG_VATT_END);
    hdim = arr_desc->dimensions[hdim_index];
    vdim = arr_desc->dimensions[vdim_index];
    canvas_get_attributes (holder->canvas,
			   CANVAS_ATT_X_PIXELS, &x_pixels,
			   CANVAS_ATT_LEFT_X, &win_left_x,
			   CANVAS_ATT_RIGHT_X, &win_right_x,
			   CANVAS_ATT_Y_PIXELS, &y_pixels,
			   CANVAS_ATT_BOTTOM_Y, &win_bottom_y,
			   CANVAS_ATT_TOP_Y, &win_top_y,
			   CANVAS_ATT_END);
    /*  Convert all values to image pixel co-ordinates  */
    win_left_x = ds_get_coord_num (hdim, win_left_x, SEARCH_BIAS_CLOSEST);
    win_right_x = ds_get_coord_num (hdim, win_right_x, SEARCH_BIAS_CLOSEST);
    win_bottom_y = ds_get_coord_num (vdim, win_bottom_y, SEARCH_BIAS_CLOSEST);
    win_top_y = ds_get_coord_num (vdim, win_top_y, SEARCH_BIAS_CLOSEST);
    /*  Compute new centre  */
    if (relative)
    {
	x = floor ( win_left_x + x * (win_right_x - win_left_x) );
	y = floor ( win_bottom_y + y * (win_top_y - win_bottom_y) );
    }
    else
    {
	x = ds_get_coord_num (hdim, x, SEARCH_BIAS_CLOSEST);
	y = ds_get_coord_num (vdim, y, SEARCH_BIAS_CLOSEST);
    }
    /*  Compute magnification  */
    holder->pan_magnification_x = x_pixels / (win_right_x - win_left_x + 1.0);
    holder->pan_magnification_y = y_pixels / (win_top_y - win_bottom_y + 1.0);
    holder->pan_magnification_x *= zoom;
    holder->pan_magnification_y *= zoom;
    num_image_pixels = x_pixels / holder->pan_magnification_x;
    if (num_image_pixels < 2) return;
    left_x = x - num_image_pixels * 0.5;
    right_x = left_x + num_image_pixels - 1;
#ifdef DEBUG
    fprintf (stderr, "  num_image_pixels: %ld  left_x: %ld  right_x: %ld\n",
	     num_image_pixels, left_x, right_x);
#endif
    if ( (left_x < 0) || (left_x >= hdim->length) )
    {
	left_x = 0;
	right_x = num_image_pixels - 1;
    }
    if ( (right_x < 0) || (right_x >= hdim->length) )
    {
	right_x = hdim->length - 1;
	left_x = right_x - (num_image_pixels - 1);
	if (left_x < 0) left_x = 0;
    }
    num_image_pixels = y_pixels / holder->pan_magnification_y;
    if (num_image_pixels < 2) return;
    bottom_y = y - num_image_pixels * 0.5;
    top_y = bottom_y + num_image_pixels - 1;
    if ( (bottom_y < 0) || (bottom_y >= vdim->length) )
    {
	bottom_y = 0;
	top_y = num_image_pixels - 1;
    }
    if ( (top_y < 0) || (top_y >= vdim->length) )
    {
	top_y = vdim->length - 1;
	bottom_y = top_y - (num_image_pixels - 1);
	if (bottom_y < 0) bottom_y = 0;
    }
    if ( (left_x >= right_x) || (bottom_y >= top_y) ) return;
    viewimg_set_canvas_attributes (holder->canvas,
				   VIEWIMG_ATT_AUTO_X, FALSE,
				   VIEWIMG_ATT_AUTO_Y, FALSE,
				   VIEWIMG_ATT_END);
    holder->pan_centre_x = x;
    holder->pan_centre_y = y;
    /*  Set up canvas with approximate boundary  */
    win_left_x = ds_get_coordinate (hdim, left_x);
    win_right_x = ds_get_coordinate (hdim, right_x);
    win_bottom_y = ds_get_coordinate (vdim, bottom_y);
    win_top_y = ds_get_coordinate (vdim, top_y);
    canvas_set_attributes (holder->canvas,
			   CANVAS_ATT_LEFT_X, win_left_x,
			   CANVAS_ATT_RIGHT_X, win_right_x,
			   CANVAS_ATT_BOTTOM_Y, win_bottom_y,
			   CANVAS_ATT_TOP_Y, win_top_y,
			   CANVAS_ATT_END);
    old_func = holder->set_size;
    if ( (left_x > 0) || (right_x < hdim->length - 1) ||
	 (bottom_y > 0) || (top_y < vdim->length - 1) )
	holder->set_size = _viewimg_set_size_pan;
    ok = kwin_resize (pixcanvas, TRUE, 0, 0, 0, 0);
    holder->set_size = old_func;
    if (!ok)
    {
	fprintf (stderr, "%s: error refreshing window\n", function_name);
	return;
    }
}   /*  End Function viewimg_zoom  */


/*  Private functions follow  */

static CanvasHolder _viewimg_get_canvas_holder (KWorldCanvas canvas,
						flag alloc, char *func_name)
/*  This routine will get the canvas holder for a world canvas.
    The canvas must be given by  canvas  .
    If the value of  alloc  is TRUE, then if a canvas holder is not found for
    the specified canvas, a new one is allocated.
    The name of the calling function must be pointed to by  func_name  .
    The routine returns the canvas holder on success, else it returns NULL.
*/
{
    CanvasHolder canvas_holder;
    /*static char function_name[] = "_viewimg_get_canvas_holder";*/

    /*  First search for existing canvas holder  */
    for (canvas_holder = first_canvas_holder; canvas_holder != NULL;
	 canvas_holder = canvas_holder->next)
    {
	if (canvas == canvas_holder->canvas) return (canvas_holder);
    }
    if (!alloc) return (NULL);
    fprintf (stderr, "**WARNING**:  %s  called before:  viewimg_init\n",
	     func_name);
    fprintf (stderr,"for this canvas. This will break in Karma version 2.0\n");
/*  This will be activated in Karma version 2.0
    a_prog_bug (func_name);
*/
    return _viewimg_alloc_canvas_holder (canvas);
}   /*  End Function _viewimg_get_canvas_holder  */

static CanvasHolder _viewimg_alloc_canvas_holder (KWorldCanvas canvas)
/*  This routine will allocate a canvas holder for a world canvas.
    The canvas must be given by  canvas  .
    The routine returns the canvas holder on success, else it returns NULL.
*/
{
    CanvasHolder holder;
    static char function_name[] = "_viewimg_alloc_canvas_holder";

    /*  Must allocate holder  */
    if ( ( holder = m_calloc (sizeof *holder) ) == NULL )
    {
	m_error_notify (function_name, "canvas holder");
	return (NULL);
    }
    holder->magic_number = HOLDER_MAGIC_NUMBER;
    holder->canvas = canvas;
    holder->auto_x = TRUE;
    holder->auto_y = TRUE;
    holder->auto_v = TRUE;
    holder->int_x = TRUE;
    holder->int_y = TRUE;
    holder->fast_panner = TRUE;
    holder->pan_centre_x = 0;
    holder->pan_centre_y = 0;
    holder->set_size = _viewimg_set_size_normal;
    holder->pan_magnification_x = 4.0;
    holder->pan_magnification_y = 4.0;
    /*  Insert at beginning of list  */
    if (first_canvas_holder != NULL) first_canvas_holder->prev = holder;
    holder->next = first_canvas_holder;
    first_canvas_holder = holder;
    canvas_register_size_control_func (canvas,
				       _viewimg_worldcanvas_size_control_func,
				       holder);
    canvas_register_precompute_func (canvas,
				     _viewimg_worldcanvas_precompute_func,
				     holder);
    canvas_register_refresh_func (canvas, _viewimg_worldcanvas_refresh_func,
				  holder);
    canvas_register_position_event_func (canvas,
					 _viewimg_worldcanvas_position_func,
					 holder);
    canvas_register_coords_convert_func (canvas, _viewimg_coord_convert_func,
					 holder);
    canvas_register_destroy_func (canvas, FALSE,
				  _viewimg_free_canvas_holder, holder);
    return (holder);
}   /*  End Function _viewimg_alloc_canvas_holder  */

static void _viewimg_free_canvas_holder (CanvasHolder holder)
/*  [SUMMARY] Free a canvas holder (and any associated ViewableImage objects).
    <holder> The canvas holder.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_viewimg_free_canvas_holder";

    VERIFY_HOLDER (holder);
    while (holder->first_image != NULL)
	viewimg_destroy (holder->first_image);
    c_destroy_list (holder->position_list);
    if (holder->prev != NULL) holder->prev->next = holder->next;
    if (holder->next != NULL) holder->next->prev = holder->prev;
    if (holder == first_canvas_holder) first_canvas_holder = holder->next;
    holder->magic_number = 0;
    m_free (holder);
}   /*  End Function _viewimg_free_canvas_holder  */

static unsigned int _viewimg_get_gen_index (CONST multi_array *multi_desc,
					    CONST array_desc *arr_desc)
/*  [SUMMARY] Find general data structure which has an array descriptor.
    <multi_desc> The multi_array header.
    <arr_desc> The array descriptor to search for.
    [RETURNS] The index of the general data structure in the multi_array header
*/
{
    unsigned int count;
    char *encls_desc;
    static char function_name[] = "_viewimg_get_gen_index";

    for (count = 0; count < multi_desc->num_arrays; ++count)
    {
	switch ( ds_f_name_in_packet (multi_desc->headers[count],
				      arr_desc->dimensions[0]->name,
				      &encls_desc, NULL) )
	{
	  case IDENT_NOT_FOUND:
	    break;
	  case IDENT_DIMENSION:
	    if (arr_desc != (array_desc *) encls_desc)
	    {
		fprintf (stderr, "Found array descriptor: %p is not: %p\n",
			 encls_desc, arr_desc);
		a_prog_bug (function_name);
	    }
	    return (count);
	    /*break;*/
	  case IDENT_MULTIPLE:
	    fprintf (stderr, "Multiple instances of: \"%s\"\n",
		     arr_desc->dimensions[0]->name);
	    a_prog_bug (function_name);
	    break;
	  default:
	    fprintf (stderr,
		     "Illegal return from <ds_f_name_in_packet> for: \"%s\"\n",
		     arr_desc->dimensions[0]->name);
	    a_prog_bug (function_name);
	    break;
	}
    }
    fprintf (stderr, "Array descriptor not found\n");
    a_prog_bug (function_name);
    return (0);
}   /*  End Function _viewimg_get_gen_index  */

static void _viewimg_worldcanvas_size_control_func
    (KWorldCanvas canvas, int width, int height,
     struct win_scale_type *win_scale, void **info, flag *boundary_clear)
/*  [SUMMARY] Modify the window scaling information for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> The window scaling information. This may be modified.
    <info> A pointer to arbitrary canvas information.
    <boundary_clear> If the value TRUE is written here then the canvas_resize
    routine will attempt to clear only the boundary between the pixel canvas
    and the world canvas. If the value FALSE is written here or nothing is
    written here, the canvas_resize routine will clear the entire pixel canvas
    as appropriate.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    ViewableImage vimage;
    Kcolourmap cmap;
    iarray pseudo_arr;
    long hstart, hend, vstart, vend;  /*  Inclusive co-ordinates  */
    unsigned int canvas_visual, canvas_depth;
    flag scale_changed = FALSE;
    uaddr *hoffsets, *voffsets;
    char *cmap_packet;
    array_desc *arr_desc;
    packet_desc *pack_desc;
    dim_desc *hdim;
    dim_desc *vdim;
    packet_desc *cmap_pack_desc;
    static char function_name[] = "_viewimg_worldcanvas_size_control_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    win_scale->x_pixels = width;
    win_scale->y_pixels = height;
    if (holder->active_image == NULL)
    {
	win_scale->x_offset = 0;
	win_scale->y_offset = 0;
	return;
    }
    if ( (width < 3) || (height < 3) ) return;
    *boundary_clear = TRUE;
    vimage = holder->active_image;
    /*  Work on changing window scaling if need be  */
    if (vimage->tc_arr_desc == NULL)
    {
	arr_desc = vimage->pc_arr_desc;
	hdim = arr_desc->dimensions[vimage->pc_hdim];
	vdim = arr_desc->dimensions[vimage->pc_vdim];
	hoffsets = arr_desc->offsets[vimage->pc_hdim];
	voffsets = arr_desc->offsets[vimage->pc_vdim];
    }
    else
    {
	arr_desc = vimage->tc_arr_desc;
	hdim = arr_desc->dimensions[vimage->tc_hdim];
	vdim = arr_desc->dimensions[vimage->tc_vdim];
	hoffsets = arr_desc->offsets[vimage->tc_hdim];
	voffsets = arr_desc->offsets[vimage->tc_vdim];
    }
    pack_desc = arr_desc->packet;
#ifdef DEBUG_SIZE_CONTROL_FUNC
    fprintf (stderr, "%s: BEFORE: %g %g  %g %g xp: %d yp: %d\n",
	     function_name,
	     win_scale->left_x, win_scale->right_x,
	     win_scale->bottom_y, win_scale->top_y,
	     win_scale->x_pixels, win_scale->y_pixels);
#endif
    /*  Determine subimage boundary, based on canvas linear boundary, and then
	determine number of screen pixels required  */
    (*holder->set_size) (holder, width, height, win_scale,
			 &hstart, &hend, hdim, &vstart, &vend, vdim);
    /*  Determine number of pixels required  */
    /*  Centre image  */
    win_scale->x_offset = (width - win_scale->x_pixels) / 2;
    win_scale->y_offset = (height - win_scale->y_pixels) / 2;
    win_scale->left_x = ds_get_coordinate (hdim, hstart);
    win_scale->right_x = ds_get_coordinate (hdim, hend);
    win_scale->bottom_y = ds_get_coordinate (vdim, vstart);
    win_scale->top_y = ds_get_coordinate (vdim, vend);
#ifdef DEBUG_SIZE_CONTROL_FUNC
    fprintf (stderr, "%s:  AFTER: %g %g  %g %g xp: %d yp: %d\n",
	     function_name,
	     win_scale->left_x, win_scale->right_x,
	     win_scale->bottom_y, win_scale->top_y,
	     win_scale->x_pixels, win_scale->y_pixels);
#endif
#ifdef MACHINE_i386
    /*  Something wrong the following 4 floating point tests under i386_Linux,
	gcc 2.7.2. Problem exists for both i486 and i586, so it looks like a
	compiler bug. The following dummy call fixes things!
	*/
    ds_element_is_atomic (K_FLOAT);
#endif
    /*  Now know the number of horizontal and vertical pixels required  */
    /*  Check if window scaling changed. Since the intensity scale may change
	with the horizontal and vertical scale, this needs to be done prior to
	any refresh functions being called  */
    /*  First check for world co-ordinate change  */
    if (win_scale->left_x != vimage->win_scale.left_x) scale_changed = TRUE;
    if (win_scale->right_x != vimage->win_scale.right_x) scale_changed = TRUE;
    if (win_scale->bottom_y != vimage->win_scale.bottom_y) scale_changed =TRUE;
    if (win_scale->top_y != vimage->win_scale.top_y) scale_changed = TRUE;
    if (scale_changed && holder->enable_panning && holder->fast_panner &&
	!holder->auto_v)
    {
	/*  Fake it so that the image is not recomputed when panning and auto
	    intensity scale is not enabled. The entire image should already be
	    computed (i.e. moving around in x,y does not require recomputation
	    unless auto intensity scaling is enabled)  */
	if (vimage->last_pan_magnification ==
	    (unsigned int) holder->pan_magnification_x) scale_changed = FALSE;
	vimage->win_scale.left_x = win_scale->left_x;
	vimage->win_scale.right_x = win_scale->right_x;
	vimage->win_scale.bottom_y = win_scale->bottom_y;
	vimage->win_scale.top_y = win_scale->top_y;
	vimage->win_scale.x_pixels = win_scale->x_pixels;
	vimage->win_scale.y_pixels = win_scale->y_pixels;
    }
    if ( !holder->enable_panning && (vimage->last_pan_magnification != 0) )
	scale_changed = TRUE;
    if (scale_changed)
    {
	/*  Precomputed intensity range no longer valid  */
	vimage->value_min = TOOBIG;
	vimage->value_max = TOOBIG;
	vimage->recompute = TRUE;
    }
    if ( holder->auto_v && (vimage->tc_arr_desc == NULL) )
    {
	/*  Automatic intensity scaling with PseudoColour image  */
	if ( DOUBLE_IS_BLANK (vimage->value_min) ||
	     DOUBLE_IS_BLANK (vimage->value_max) )
	{
	    /*  Must compute minimum and maximum values  */
	    vimage->value_min = TOOBIG;
	    vimage->value_max = -TOOBIG;
	    if ( !ds_find_2D_extremes
		 (vimage->pc_slice,
		  vend - vstart + 1, voffsets + vstart,
		  hend - hstart + 1, hoffsets + hstart,
		  pack_desc->element_types[vimage->pc_elem_index],
		  win_scale->conv_type,
		  &vimage->value_min, &vimage->value_max) )
	    {
		fprintf (stderr, "Error getting data range\n");
		a_prog_bug (function_name);
	    }
	    vimage->changed = FALSE;
	    /*Image intensity range has changed: cached image no longer valid*/
	    vimage->recompute = TRUE;
	    /*  Ensure ranges are sensible  */
	    if (vimage->value_min > vimage->value_max)
	    {
		/*  Image has only blanks  */
		vimage->value_min = -1.0;
		vimage->value_max = 1.0;
	    }
	    else if (vimage->value_min == vimage->value_max)
	    {
		/*  Image has only one value  */
		if (vimage->value_min < 0.0)
		{
		    vimage->value_min *= 2.0;
		    vimage->value_max = 0.0;
		}
		else if (vimage->value_min > 0.0)
		{
		    vimage->value_min = 0.0;
		    vimage->value_max *= 2.0;
		}
		else
		{
		    vimage->value_min = -1.0;
		    vimage->value_max = 1.0;
		}
	    }
	}
	/*  World canvas intensity scale has changed  */
	win_scale->z_min = vimage->value_min;
	win_scale->z_max = vimage->value_max;
    }
    if ( !holder->auto_v && vimage->changed && (vimage->tc_arr_desc == NULL) )
    {
	/*  No automatic intensity scaling, data changed and PseudoColour
	    image  */
	vimage->changed = FALSE;
	vimage->recompute = TRUE;
    }
    if ( !holder->auto_v && DOUBLE_IS_VALID (vimage->override.value_min) &&
	 (vimage->tc_arr_desc == NULL) )
    {
	/*  No automatic intensity scaling, override min value provided and
	    PseudoColour image  */
	win_scale->z_min = vimage->override.value_min;
    }
    if ( !holder->auto_v && DOUBLE_IS_VALID (vimage->override.value_max) &&
	 (vimage->tc_arr_desc == NULL) )
    {
	/*  No automatic intensity scaling, override max value provided and
	    PseudoColour image  */
	win_scale->z_max = vimage->override.value_max;
    }
    if ( vimage->changed && (vimage->tc_arr_desc != NULL) )
    {
	/*  Data changed and TrueColour image  */
	if (vimage->tc_arr_desc->num_levels > 0)
	{
	    fprintf (stderr, "%s: Tiling not supported.\n", function_name);
	    return;
	}
	kwin_get_attributes (canvas_get_pixcanvas (canvas),
			     KWIN_ATT_VISUAL, &canvas_visual,
			     KWIN_ATT_DEPTH, &canvas_depth,
			     KWIN_ATT_END);
	if (canvas_visual != KWIN_VISUAL_PSEUDOCOLOUR)
	{
	    vimage->changed = FALSE;
	    vimage->recompute = TRUE;
	    return;
	}
	/*  If TrueColour image and PseudoColour canvas, recompress image  */
	vimage->recompute = TRUE;
	/*  Have RGB component arrays: create PseudoColour array  */
	ds_dealloc_multi (vimage->pc_multi_desc);
	if ( ( pseudo_arr = iarray_create_2D (vdim->length, hdim->length,
					      K_UBYTE) ) == NULL )
	{
	    m_abort (function_name, "PseudoColour image");
	}
	iarray_set_world_coords (pseudo_arr, 1,
				 hdim->first_coord, hdim->last_coord);
	iarray_set_world_coords (pseudo_arr, 0,
				 vdim->first_coord, vdim->last_coord);
	vimage->pc_multi_desc = pseudo_arr->multi_desc;
	vimage->pc_arr_desc = pseudo_arr->arr_desc;
	vimage->pc_slice = pseudo_arr->data;
	vimage->pc_hdim = 1;
	vimage->pc_vdim = 0;
	vimage->pc_elem_index = 0;
	vimage->pc_hstride = 1;
	vimage->pc_vstride = hdim->length;
	++pseudo_arr->multi_desc->attachments;
	iarray_dealloc (pseudo_arr);
	if ( !imc_24to8 (vdim->length * hdim->length,
			 (unsigned char *) vimage->tc_slice +
			 ds_get_element_offset (pack_desc,
						vimage->tc_red_index),
			 (unsigned char *) vimage->tc_slice +
			 ds_get_element_offset (pack_desc,
						vimage->tc_green_index),
			 (unsigned char *) vimage->tc_slice +
			 ds_get_element_offset (pack_desc,
						vimage->tc_blue_index),
			 ds_get_packet_size (pack_desc),
			 (unsigned char *) vimage->pc_slice, 1,
			 DEFAULT_NUMBER_OF_COLOURS,
			 0, &cmap_pack_desc, &cmap_packet) )
	{
	    fprintf (stderr, "Error compressing image\n");
	    return;
	}
	putc ('\n', stderr);
	cmap = canvas_get_cmap (canvas);
	if ( !kcmap_copy_from_struct (cmap, cmap_pack_desc, cmap_packet) )
	{
	    m_abort (function_name, "Compressed image");
	}
	win_scale->z_min = 0.0;
	win_scale->z_max = (kcmap_get_pixels (cmap, (unsigned long **) NULL)
			    -1);
	ds_dealloc_packet (cmap_pack_desc, cmap_packet);
	vimage->changed = FALSE;
	/*  Need to add a mechanism to update the colourmap when running a
	    movie where each frame is individually compressed. Right now the
	    colourmap for the first image in the sequence made active is used,
	    which is clearly wrong  */
    }
}   /*  End Function _viewimg_worldcanvas_size_control_func  */

static flag _viewimg_set_size_normal (CanvasHolder holder,
				      int width, int height,
				      struct win_scale_type *win_scale,
				      long *hstart, long *hend,
				      CONST dim_desc *hdim,
				      long *vstart, long *vend,
				      CONST dim_desc *vdim)
/*  [SUMMARY] Determine size of image to be drawn.
    <holder> The canvas holder.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> The window scaling information. This is modified.
    <hstart> The horizontal start index is written here.
    <hend> The horizontal end index is written here.
    <hdim> The horizontal dimension descriptor.
    <vstart> The vertical start index is written here.
    <vend> The vertical end index is written here.
    <vdim> The vertical dimension descriptor.
    [RETURNS] TRUE if a size could be determined, else FALSE indicating the
    sizing policy is not flexible enough.
*/
{
    static char function_name[] = "_viewimg_set_size_normal";

    if (holder->auto_x)
    {
	*hstart = 0;
	*hend = hdim->length - 1;
    }
    else
    {
	*hstart = ds_get_coord_num (hdim, win_scale->left_x,
				    SEARCH_BIAS_CLOSEST);
	*hend = ds_get_coord_num (hdim, win_scale->right_x,
				  SEARCH_BIAS_CLOSEST);
	if (*hstart > *hend)
	{
	    fprintf (stderr, "hstart: %ld is larger than hend: %ld\n",
		     *hstart, *hend);
	    a_prog_bug (function_name);
	}
	if (*hend == *hstart)
	{
	    /*  Same dimension co-ordinate: separate  */
	    --*hstart;
	    ++*hend;
	    if (*hstart < 0)
	    {
		*hstart = 0;
		*hend = 2;
	    }
	    else if (*hend >= hdim->length) *hend = hdim->length - 1;
	}
    }
    if (holder->auto_y)
    {
	*vstart = 0;
	*vend = vdim->length - 1;
    }
    else
    {
	*vstart = ds_get_coord_num (vdim, win_scale->bottom_y,
				    SEARCH_BIAS_CLOSEST);
	*vend = ds_get_coord_num (vdim, win_scale->top_y, SEARCH_BIAS_CLOSEST);
	if (*vstart > *vend)
	{
	    fprintf (stderr, "vstart: %ld is larger than vend: %ld\n",
		     *vstart, *vend);
	    a_prog_bug (function_name);
	}
	if (*vend == *vstart)
	{
	    /*  Same dimension co-ordinate: separate  */
	    --*vstart;
	    ++*vend;
	    if (*vstart < 0)
	    {
		*vstart = 0;
		*vend = 2;
	    }
	    else if (*vend >= vdim->length) *vend = vdim->length - 1;
	}
    }
    _viewimg_determine_size (holder, width, height, win_scale,
			     hstart, hend, vstart, vend);
    return (TRUE);
}   /*  End Function _viewimg_set_size_normal  */

static flag _viewimg_set_size_pan (CanvasHolder holder, int width, int height,
				   struct win_scale_type *win_scale,
				   long *hstart, long *hend,
				   CONST dim_desc *hdim,
				   long *vstart, long *vend,
				   CONST dim_desc *vdim)
/*  [SUMMARY] Determine size of image to be drawn.
    <holder> The canvas holder.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> The window scaling information. This is modified.
    <hstart> The horizontal start index is written here.
    <hend> The horizontal end index is written here.
    <hdim> The horizontal dimension descriptor.
    <vstart> The vertical start index is written here.
    <vend> The vertical end index is written here.
    <vdim> The vertical dimension descriptor.
    [RETURNS] TRUE if a size could be determined, else FALSE indicating the
    sizing policy is not flexible enough.
*/
{
    double mag_x = holder->pan_magnification_x;
    double mag_y = holder->pan_magnification_y;
    /*static char function_name[] = "_viewimg_set_size_pan";*/

    /*  Apply any policy restrictions (i.e. aspect, integer zoom, etc.)  */
#ifdef DEBUG
    fprintf (stderr,
	     "  %s: mags: %g %g  width: %d height: %d Cx: %ld Cy: %ld\n",
	     function_name, mag_x, mag_y, width, height,
	     holder->pan_centre_x, holder->pan_centre_y);
#endif
    if (holder->maintain_aspect_ratio)
    {
	if (mag_x > mag_y) mag_y = mag_x;
	else if (mag_y > mag_x) mag_x = mag_y;
    }
    if (holder->int_x)
    {
	if (mag_x < 1.0) mag_x = 1.0 / floor (1.0 / mag_x + 0.5);
	else mag_x = floor (mag_x + 0.5);
    }
    if (holder->int_y)
    {
	if (mag_y < 1.0) mag_y = 1.0 / floor (1.0 / mag_y + 0.5);
	else mag_y = floor (mag_y + 0.5);
    }
    if (holder->maintain_aspect_ratio)  /*  Need to check again  */
    {
	if (mag_x > mag_y) mag_y = mag_x;
	else if (mag_y > mag_x) mag_x = mag_y;
    }
    holder->pan_magnification_x = mag_x;
    holder->pan_magnification_y = mag_y;
#ifdef DEBUG
    fprintf (stderr, "    new mags: %g %g\n", mag_x, mag_y);
#endif
    *hstart = holder->pan_centre_x + 0.5 - width / mag_x * 0.5;
    if (*hstart < 0) *hstart = 0;
    *hend = *hstart + width / mag_x - 1;
    if (*hend >= hdim->length)
    {
	*hend = hdim->length - 1;
	*hstart = *hend - (int) (width / mag_x) + 1;
    }
    if (*hstart < 0) *hstart = 0;
    win_scale->x_pixels = (*hend - *hstart + 1) * mag_x;
    *vstart = holder->pan_centre_y + 0.5 - height / mag_y * 0.5;
    if (*vstart < 0) *vstart = 0;
    *vend = *vstart + height / mag_y - 1;
    if (*vend >= vdim->length)
    {
	*vend = vdim->length - 1;
	*vstart = *vend - (int) (height / mag_y) + 1;
    }
    if (*vstart < 0) *vstart = 0;
    win_scale->y_pixels = (*vend - *vstart + 1) * mag_y;
#ifdef DEBUG2
    fprintf (stderr, "    new Cx: %ld xp: %d",
	     holder->pan_centre_x, win_scale->x_pixels);
    fprintf (stderr, " Cy: %ld yp: %d Vs: %ld Ve: %ld Vl: %lu\n",
	     holder->pan_centre_y, win_scale->y_pixels,
	     *vstart, *vend, vdim->length);
#endif
    return (TRUE);
}   /*  End Function _viewimg_set_size_pan  */

static flag _viewimg_determine_size (CanvasHolder holder, int width,int height,
				     struct win_scale_type *win_scale,
				     long *hstart, long *hend,
				     long *vstart, long *vend)
/*  [SUMMARY] Determine size of image to be drawn.
    <holder> The canvas holder.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> The window scaling information. This is modified.
    <hstart> The horizontal start index. This may be modified inwards.
    <hend> The horizontal end index. This may be modified inwards.
    <vstart> The vertical start index. This may be modified inwards.
    <vend> The vertical end index. This may be modified inwards.
    [RETURNS] TRUE if a size could be determined, else FALSE indicating the
    sizing policy is not flexible enough.
*/
{
    long hlength, vlength, factor;

    hlength = *hend - *hstart + 1;
    vlength = *vend - *vstart + 1;
    /*  First test if 1:1 zooming is possible  */
    if ( (width == hlength) && (height == vlength) )
    {
	win_scale->x_pixels = width;
	win_scale->y_pixels = height;
	return (TRUE);
    }
    /*  Determine number of pixels required  */
    if (holder->allow_truncation)
    {
	/*  Permit a small number of data values at the end of each dimension
	    to be ignored in order to allow images whose dimensions are
	    mutually prime to be shrunk. The aspect ratio may need to be
	    preserved
	    */
	_viewimg_trunc_zoom (holder,
			     hstart, hend, &win_scale->x_pixels,
			     vstart, vend, &win_scale->y_pixels);
	return (TRUE);
    }
    if (holder->maintain_aspect_ratio)
    {
	/*  Must display all specified image values and also must maintain
	    aspect ratio: think harder
	    */
	if ( !_viewimg_aspect_zoom (hlength, &win_scale->x_pixels,
				    vlength, &win_scale->y_pixels,
				    (holder->int_x ||
				     holder->int_y) ? TRUE : FALSE) )
	    _viewimg_trunc_zoom (holder,
				 hstart, hend, &win_scale->x_pixels,
				 vstart, vend, &win_scale->y_pixels);
	return (TRUE);
    }
    /*  It is not necessary to maintain the aspect ratio: it is permissible
	to shrink each dimension independently
	*/
    /*  Determine number of horizontal pixels required  */
    if (holder->int_x)
    {
	if (hlength <= width)
	{
	    /*  Zoom in or 1:1  */
	    win_scale->x_pixels = (width / hlength) * hlength;
	}
	else
	{
	    /*  Zoom out  */
	    factor = hlength / width;
	    if (holder->allow_truncation)
	    {
		while (hlength / factor > width) ++factor;
		*hend = *hstart + (hlength / factor) * factor - 1;
	    }
	    else
	    {
		while ( (hlength / factor > width) || (hlength % factor != 0) )
		{
		    if (++factor > hlength)
		    {
			/*  Too much: fail  */
			fprintf (stderr,
				 "Cannot do integral zoom out without truncating\n");
			_viewimg_trunc_zoom (holder,
					     hstart, hend,&win_scale->x_pixels,
					     vstart,vend,&win_scale->y_pixels);
			return (TRUE);
		    }
		}
	    }
	    win_scale->x_pixels = hlength / factor;
	}
    }
    else
    {
	win_scale->x_pixels = width;
    }
    /*  Determine number of vertical pixels required  */
    if (holder->int_y)
    {
	if (vlength <= height)
	{
	    /*  Zoom in  */
	    win_scale->y_pixels = (height / vlength) * vlength;
	}
	else
	{
	    /*  Zoom out or 1:1  */
	    factor = vlength / height;
	    if (holder->allow_truncation)
	    {
		while (vlength / factor > height) ++factor;
		*vend = *vstart + (vlength / factor) * factor - 1;
	    }
	    else
	    {
		while ( (vlength / factor > height) || (vlength % factor != 0))
		{
		    if (++factor > vlength)
		    {
			/*  Too much: fail  */
			fprintf (stderr,
				 "Cannot do integral zoom out without truncating\n");
			_viewimg_trunc_zoom (holder,
					     hstart, hend,&win_scale->x_pixels,
					     vstart,vend,&win_scale->y_pixels);
			return (TRUE);
		    }
		}
	    }
	    win_scale->y_pixels = vlength / factor;
	}
	win_scale->y_offset = (height - win_scale->y_pixels) / 2;
    }
    else
    {
	win_scale->y_pixels = height;
    }
    return (TRUE);
}   /*  End Function _viewimg_determine_size  */

static flag _viewimg_aspect_zoom (long hlength, int *hpixels,
				  long vlength, int *vpixels, flag int_zoom)
/*  [PURPOSE] This routine will compute the number of horizontal and vertical
    pixels to use whilst maintaining the image data aspect ratio.
    <hlength> The number of horizontal data values.
    <hpixels> The number of horizontal pixels available. This should be
    modified to reflect the number of pixels required.
    <vlength> The number of vertical data values.
    <vpixels> The number of vertical pixels available. This should be modified
    to reflect the number of pixels required.
    <int_zoom> If TRUE, integral zooming is required.
    [RETURNS] TRUE if a size could be determined, else FALSE indicating the
    sizing policy is not flexible enough.
*/
{
    long ifactor;
    float hfactor, vfactor, factor;
    float tiny_offset = 1e-3;
    /*static char function_name[] = "_viewimg_aspect_zoom";*/

#ifdef DUMMY
    if (getuid () == 465)
	fprintf (stderr,
		 "h_values: %ld  h_pixels: %d  v_values: %ld v_pixels: %d\n",
		 hlength, *hpixels, vlength, *vpixels);
#endif
    /*  NOTE:  a 1:1 zoom ratio is termed a zoom in of factor 1  */
    /*  Zoom in: replicate image data.    Zoom out: shrink image data  */
    if ( (*hpixels >= hlength) && (*vpixels >= vlength) )
    {
	/*  Zoom in (expand) both axes  */
	hfactor = (float) *hpixels / (float) hlength;
	vfactor = (float) *vpixels / (float) vlength;
	/*  Use smallest zoom in factor  */
	factor = (hfactor > vfactor) ? vfactor : hfactor;
	if (int_zoom) factor = floor (factor);
	*hpixels = (factor * (float) hlength + tiny_offset);
	*vpixels = (factor * (float) vlength + tiny_offset);
	return (TRUE);
    }
    /*  At least one axis must be zoomed out (shrunk)  */
    if (*hpixels < hlength)
    {
	/*  Horizontal axis must be zoomed out  */
	hfactor = (float) hlength / (float) *hpixels;
    }
    else
    {
	/*  Zoom out 1:1  */
	hfactor = 1.0;
    }
    if (*vpixels < vlength)
    {
	/*  Vertical axis must be zoomed out  */
	vfactor = (float) vlength / (float) *vpixels;
    }
    else
    {
	/*  Zoom out 1:1  */
	vfactor = 1.0;
    }
    /*  Start with largest zoom out factor  */
    factor = (hfactor > vfactor) ? hfactor : vfactor;
    if (!int_zoom)
    {
	/*  Do not need integral factor: easy  */
	*hpixels = (float) hlength / factor + tiny_offset;
	*vpixels = (float) vlength / factor + tiny_offset;
	return (TRUE);
    }
    /*  Must have an integral factor: search for one  */
    ifactor = ceil (factor);
    while ( (hlength % ifactor != 0) || (vlength % ifactor != 0) )
    {
	if (++ifactor > hlength)
	{
	    /*  Too much: fail  */
	    fprintf (stderr,
		     "Cannot maintain aspect ratio without truncating\n");
	    return (FALSE);
	}
    }
    *hpixels = hlength / ifactor + tiny_offset;
    *vpixels = vlength / ifactor + tiny_offset;
    return (TRUE);
}   /*  End Function _viewimg_aspect_zoom  */

static void _viewimg_trunc_zoom (CONST struct canvas_holder_type *holder,
				 long *hstart, long *hend, int *hpixels,
				 long *vstart, long *vend, int *vpixels)
/*  [PURPOSE] This routine will compute the number of horizontal and vertical
    pixels to use, permitting the input image to be truncated.
    <holder> The canvas holder.
    <hstart> The starting horizontal image value co-ordinate. This may be
    modified inwards.
    <hend> The end horizontal image value co-ordinate. This may be modified
    inwards.
    <hpixels> The number of horizontal pixels available. This may be modified
    inwards.
    <vstart> The starting vertical image value co-ordinate. This may be
    modified inwards.
    <vend> The end vertical image value co-ordinate. This may be modified
    inwards.
    <vpixels> The number of vertical pixels available. This may be modified
    inwards.
    [RETURNS] Nothing.
*/
{
    long hlength, vlength;
    float hfactor, vfactor, factor;
    float tiny_offset = 1e-3;
    /*static char function_name[] = "_viewimg_trunc_zoom";*/

    hlength = *hend - *hstart + 1;
    vlength = *vend - *vstart + 1;
    if (holder->maintain_aspect_ratio)
    {
	if ( (*hpixels >= hlength) && (*vpixels >= vlength) )
	{
	    /*  Zoom in (expand) both axes. Don't truncate in this case since
		data would be lost. Since the canvas is big enough to see the
		whole image, it would be inappropriate to loose data. Also,
		truncation can result in smaller input image sizes for simple
		refresh, due to interaction with axis labelling size
		negotiation
	     */
	    _viewimg_aspect_zoom (hlength, hpixels, vlength, vpixels,
				  (holder->int_x ||
				   holder->int_y) ? TRUE : FALSE);
	    return;
	}
	/*  At least one axis must be zoomed out (shrunk)  */
	if (*hpixels < hlength)
	{   /*  Horizontal axis must be zoomed out  */
	    hfactor = (float) hlength / (float) *hpixels;
	}
	else hfactor = 1.0;  /*  Zoom out 1:1  */
	if (*vpixels < vlength)
	{   /*  Vertical axis must be zoomed out  */
	    vfactor = (float) vlength / (float) *vpixels;
	}
	else vfactor = 1.0;  /*  Zoom out 1:1  */
	/*  Start with largest zoom out factor  */
	factor = (hfactor > vfactor) ? hfactor : vfactor;
	if (holder->int_x || holder->int_y)
	{
	    factor = floor (factor);
	    /*  Find a zoom factor big enough  */
	    while ( (hlength / (long) factor > *hpixels) ||
		    (vlength / (long) factor > *vpixels) ) ++factor;
	}
	/*  Modify output pixels, making sure npix >= 2  */
	while ( ( *hpixels = ( (float) hlength / factor + tiny_offset ) ) < 2 )
	    ++hlength;
	while ( ( *vpixels = ( (float) vlength / factor + tiny_offset ) ) < 2 )
	    ++vlength;
	if (!holder->int_x && !holder->int_y)
	{   /*  Do not need integral zoom factor: easy, no truncation needed */
	    return;
	}
	/*  It seems we do need to truncate, since integral zoom is required,
	    so modify input image co-ordinates  */
	*hend = *hstart + (hlength / (long) factor) * (long) factor - 1;
	*vend = *vstart + (vlength / (long) factor) * (long) factor - 1;
	return;
    }
    /*  Scale for each axis is independant  */
    if (holder->int_x)
    {
	/*  Must use integral zoom factor for horizontal dimension  */
	if (hlength > *hpixels)
	{
	    /*  Must shrink horizontally  */
	    factor = floor ( (float) hlength / (float) *hpixels );
	    while (hlength / (int) factor > *hpixels) ++factor;
	    /*  Modify input image co-ordinates  */
	    *hend = *hstart + (hlength / (int) factor) * (int) factor - 1;
	    /*  Modify output pixels  */
	    *hpixels = ( (float) hlength / factor + tiny_offset );
	}
	else
	{
	    /*  Possible horizontal expansion  */
	    *hpixels = (*hpixels / hlength) * hlength;
	}
    }
    if (holder->int_y)
    {
	if (vlength > *vpixels)
	{
	    /*  Must shrink vertically  */
	    factor = floor ( (float) vlength / (float) *vpixels );
	    while (vlength / (int) factor > *vpixels) ++factor;
	    /*  Modify input image co-ordinates  */
	    *vend = *vstart + (vlength / (int) factor) * (int) factor - 1;
	    /*  Modify output pixels  */
	    *vpixels = ( (float) vlength / factor + tiny_offset );
	}
	else
	{
	    /*  Possible horizontal expansion  */
	    *vpixels = (*vpixels / vlength) * vlength;
	}
    }
}   /*  End Function _viewimg_trunc_zoom  */

static void _viewimg_worldcanvas_precompute_func
    (KWorldCanvas canvas, int width, int height,
     struct win_scale_type *win_scale, Kcolourmap cmap, flag cmap_resize,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    ViewableImage vimage;
    IscaleFunc iscale_func;
    int x_offset, y_offset, x_pixels, y_pixels;
    unsigned int count;
    void *magnify_func, *draw_cached_image_func;
    double iscale_values[NUM_ISCALE_VALUES];
    static char function_name[] = "_viewimg_worldcanvas_precompute_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    if (holder->active_image == NULL) return;
    if ( (width < 3) || (height < 3) ) return;
    vimage = holder->active_image;
    kwin_get_attributes (canvas_get_pixcanvas (holder->canvas),
			 KWIN_FUNC_DRAW_CACHED_IMAGE, &draw_cached_image_func,
			 KWIN_ATT_END);
    if (draw_cached_image_func == NULL) return;
    if ( cmap_resize && (vimage->tc_arr_desc != NULL) )
    {
	fprintf (stderr, "%s: discarding cmap_resize and TrueColour image\n",
		 function_name);
	return;
    }
    canvas_get_attributes (canvas,
			   CANVAS_ATT_X_OFFSET, &x_offset,
			   CANVAS_ATT_Y_OFFSET, &y_offset,
			   CANVAS_ATT_X_PIXELS, &x_pixels,
			   CANVAS_ATT_Y_PIXELS, &y_pixels,
			   CANVAS_ATT_END);
    for (count = 0; count < num_areas; ++count)
    {
	if (!areas[count].clear && !areas[count].geom_clear) continue;
	/*  Area should be cleared  */
	if ( (areas[count].startx >= x_offset) &&
	     (areas[count].endx < x_offset + x_pixels) &&
	     (areas[count].starty >= y_offset) &&
	     (areas[count].endy < y_offset + y_pixels) )
	{
	    /*  This area is entirely within image: no need to clear  */
	    areas[count].clear = FALSE;
	    areas[count].geom_clear = FALSE;
	}
    }
    /*  Do tests on image  */
    if (vimage->cache == NULL)
    {
	/*  No image cache yet  */
	_viewimg_recompute_image (holder, width, height, win_scale, cmap,
				  vimage);
	return;
    }
    if (vimage->changed)
    {
	/*  Data has changed  */
	fprintf (stderr, "%s: spurious data change\n", function_name);
	vimage->changed = FALSE;
	vimage->recompute = TRUE;
	_viewimg_recompute_image (holder, width, height, win_scale, cmap,
				  vimage);
	return;
    }
    if ( (vimage->tc_arr_desc == NULL) && (win_scale->iscale_func != NULL) &&
	 !vimage->recompute )
    {
	/*  PseudoColour image and intensity scaling function used: check that
	    the mapping function has not changed. Do this by comparing a set
	    of scaled values with a set previously computed. If there is a
	    difference, the intensity scaling has changed. This is only an
	    approximate algorithm  */
	iscale_func = (IscaleFunc) win_scale->iscale_func;
	/*  Compute test values  */
	for (count = 0; count < NUM_ISCALE_VALUES; ++count)
	{
	    iscale_values[count] = win_scale->z_min +
		(win_scale->z_max - win_scale->z_min) * iscale_factors[count];
	}
	if ( !(*iscale_func) (iscale_values, 1, iscale_values, 1,
			      NUM_ISCALE_VALUES,
			      win_scale->z_min, win_scale->z_max,
			      win_scale->iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling test values\n", function_name);
	    return;
	}
	/*  Now compare  */
	if ( DOUBLE_IS_BLANK (vimage->iscale_test[0]) ||
	     !m_cmp (vimage->iscale_test, iscale_values,
		     sizeof *iscale_values * NUM_ISCALE_VALUES ) )
	{
	    /*  Old values not consistent: copy and force recompute  */
	    m_copy (vimage->iscale_test, iscale_values,
		    sizeof *iscale_values * NUM_ISCALE_VALUES );
	    vimage->recompute = TRUE;
	}
    }
    if (vimage->recompute)
    {
	/*  Image cache is no longer valid for some reason  */
	_viewimg_recompute_image (holder, width, height, win_scale, cmap,
				  vimage);
	return;
    }
    if (cmap_resize)
    {
	/*  Colourmap has changed  */
	_viewimg_recompute_image (holder, width, height, win_scale, cmap,
				  vimage);
	return;
    }
#ifdef DEBUG0
    fprintf (stderr, "%s win_scale: %e %e  %e %e\n",
	     function_name,
	     win_scale->left_x, win_scale->right_x,
	     win_scale->bottom_y, win_scale->top_y);
    fprintf (stderr, "%s vimage.win_scale: %e %e  %e %e\n",
	     function_name,
	     vimage->win_scale.left_x, vimage->win_scale.right_x,
	     vimage->win_scale.bottom_y, vimage->win_scale.top_y);
#endif
    /*  More tests  */
    vimage->win_scale.x_offset = win_scale->x_offset;
    vimage->win_scale.y_offset = win_scale->y_offset;
    if (!m_cmp (win_scale, &vimage->win_scale, sizeof *win_scale) )
    {
	/*  Window scaling information has changed  */
	_viewimg_recompute_image (holder, width, height, win_scale, cmap,
				  vimage);
	return;
    }
    kwin_get_attributes (canvas_get_pixcanvas (canvas),
			 KWIN_ATT_MAGNIFY_FUNC, &magnify_func,
			 KWIN_ATT_END);
    if (magnify_func != vimage->magnify_func)
    {
	_viewimg_recompute_image (holder, width, height, win_scale, cmap,
				  vimage);
	return;
    }
    /*  No significant changes: use image cache  */
}   /*  End Function worldcanvas_precompute_func  */

static void _viewimg_worldcanvas_refresh_func
    (KWorldCanvas canvas, int width, int height,
     struct win_scale_type *win_scale, Kcolourmap cmap, flag cmap_resize,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas,flag *honoured_areas)
/*  [PURPOSE] This routine registers a refresh event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    ViewableImage vimage;
    unsigned int visual;
    void *draw_cached_image_func;
    static char function_name[] = "_viewimg_worldcanvas_refresh_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
#ifdef DEBUG_REFRESH_FUNC
    fprintf (stderr, "%s: starting refresh\n\n", function_name);
#endif
    if (holder->active_image == NULL) return;
    if ( (width < 3) || (height < 3) ) return;
    vimage = holder->active_image;
    kwin_get_attributes (canvas_get_pixcanvas (holder->canvas),
			 KWIN_FUNC_DRAW_CACHED_IMAGE, &draw_cached_image_func,
			 KWIN_ATT_END);
    if ( (draw_cached_image_func == NULL) && (vimage->tc_arr_desc != NULL) )
    {
	kwin_get_attributes (canvas_get_pixcanvas (holder->canvas),
			     KWIN_ATT_VISUAL, &visual,
			     KWIN_ATT_END);
	if (visual == KWIN_VISUAL_PSEUDOCOLOUR)
	{
	    fprintf (stderr,
		     "%s: drawing TrueColour image directly to PostScriptPage\n",
		     function_name);
	}
	/*  Drawing TrueColour image to PostScript page  */
	if ( !canvas_draw_rgb_image (holder->canvas,
				     vimage->tc_arr_desc, vimage->tc_slice,
				     vimage->tc_hdim, vimage->tc_vdim,
				     vimage->tc_red_index,
				     vimage->tc_green_index,
				     vimage->tc_blue_index,
				     NULL) )
	{
	    fprintf (stderr, "%s: error drawing image onto world canvas\n",
		     function_name);
	}
	return;
    }
    if (draw_cached_image_func == NULL)
    {
	if ( !canvas_draw_image (holder->canvas,
				 vimage->pc_arr_desc, vimage->pc_slice,
				 vimage->pc_hdim, vimage->pc_vdim,
				 vimage->pc_elem_index, NULL) )
	{
	    fprintf (stderr, "%s: error drawing image onto world canvas\n",
		     function_name);
	}
	return;
    }
    if ( cmap_resize && (vimage->tc_arr_desc != NULL) ) return;
    /*  Do tests on image  */
    if (vimage->cache == NULL)
    {
	/*  No image cache!  */
	fprintf (stderr, "%s: no image cache!\n", function_name);
	return;
    }
    if (vimage->changed)
    {
	/*  Data has changed  */
	fprintf (stderr, "%s: spurious data change\n", function_name);
	return;
    }
    if (vimage->recompute)
    {
	/*  Image cache is no longer valid for some reason  */
	fprintf (stderr, "%s: spurious image recompute change\n",
		 function_name);
	return;
    }
    if (holder->enable_panning && holder->fast_panner && !holder->auto_v)
    {
	_viewimg_draw_subcache (holder, vimage, win_scale);
	return;
    }
    vimage->last_pan_magnification = 0;
    if (num_areas < 1)
    {
	if ( kwin_draw_cached_image (vimage->cache,
				     win_scale->x_offset,
				     win_scale->y_offset) )
	{
	    /*  Cache OK  */
	    return;
	}
	fprintf (stderr, "%s: spurious corrupted cache\n", function_name);
	return;
    }
    /*  Only draw parts of the image  */
    if ( kwin_draw_cached_subimages (vimage->cache,
				     win_scale->x_offset, win_scale->y_offset,
				     num_areas, areas) )
    {
	*honoured_areas = TRUE;
	return;
    }
    /*  Something wrong with cache  */
    fprintf (stderr, "%s: spurious corrupted cache\n", function_name);
}   /*  End Function _viewimg_worldcanvas_refresh_func  */

static void _viewimg_recompute_image (CanvasHolder holder,
				      int width, int height,
				      struct win_scale_type *win_scale,
				      Kcolourmap cmap, ViewableImage vimage)
/*  This routine will recompute a viewable image and then redraw it onto a
    world canvas.
    The canvas holder must be given by  holder  .
    The width of the canvas in pixels must be given by  width  .
    The height of the canvas in pixels must be given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas must be given by  cmap  .
    <vimage> The viewable image.
*/
{
    Kcolourmap kcmap;
    unsigned int num_pixels, elem_index;
    array_desc *arr_desc;
    packet_desc *pack_desc;
    unsigned long *pixel_values;
    dim_desc *hdim, *vdim;
    static char function_name[] = "_viewimg_recompute_image";

    m_copy (&vimage->win_scale, win_scale, sizeof *win_scale);
    vimage->recompute = TRUE;
    kwin_get_attributes (canvas_get_pixcanvas (holder->canvas),
			 KWIN_ATT_MAGNIFY_FUNC, &vimage->magnify_func,
			 KWIN_ATT_END);
    /*  If a PseudoColour array exists, either the image is PseudoColour or the
	canvas is PseudoColour  */
    if (holder->enable_panning && holder->fast_panner && !holder->auto_v)
    {
	/*  Compute a different kind of image cache which is the entire image
	    magnified, irrespective of the canvas size. At a later stage the
	    image cache will be drawn
	    */
	if (vimage->pc_arr_desc != NULL)
	{
	    kcmap = canvas_get_cmap (holder->canvas);
	    num_pixels = kcmap_get_pixels (kcmap, &pixel_values);
	    arr_desc = vimage->pc_arr_desc;
	    hdim = arr_desc->dimensions[vimage->pc_hdim];
	    vdim = arr_desc->dimensions[vimage->pc_vdim];
	    pack_desc = arr_desc->packet;
	    elem_index = vimage->pc_elem_index;
	    if ( !kwin_compute_pc_subimage
		 (canvas_get_pixcanvas (holder->canvas),
		  0, 0,
		  hdim->length * holder->pan_magnification_x,
		  vdim->length * holder->pan_magnification_y,
		  hdim->length * holder->pan_magnification_x,
		  vdim->length * holder->pan_magnification_y,
		  vimage->pc_slice + ds_get_element_offset (pack_desc,
							    elem_index),
		  arr_desc->offsets[vimage->pc_hdim],
		  arr_desc->offsets[vimage->pc_vdim],
		  hdim->length, vdim->length,
		  pack_desc->element_types[elem_index],
		  win_scale->conv_type,
		  num_pixels, pixel_values,
		  win_scale->blank_pixel,
		  win_scale->min_sat_pixel, win_scale->max_sat_pixel,
		  win_scale->z_min, win_scale->z_max,
		  win_scale->iscale_func, win_scale->iscale_info,
		  &vimage->cache) )
	    {
		fprintf (stderr, "%s: error computing PseudoColour image\n",
			 function_name);
		return;
	    }
	}
	else
	{
	    arr_desc = vimage->tc_arr_desc;
	    hdim = arr_desc->dimensions[vimage->tc_hdim];
	    vdim = arr_desc->dimensions[vimage->tc_vdim];
	    pack_desc = arr_desc->packet;
	    if ( !kwin_compute_rgb_subimage
		 (canvas_get_pixcanvas (holder->canvas),
		  0, 0,
		  hdim->length * holder->pan_magnification_x,
		  vdim->length * holder->pan_magnification_y,
		  hdim->length * holder->pan_magnification_x,
		  vdim->length * holder->pan_magnification_y,
		  (CONST unsigned char *)vimage->tc_slice +
		  ds_get_element_offset (pack_desc,
					 vimage->tc_red_index),
		  (CONST unsigned char *)vimage->tc_slice +
		  ds_get_element_offset (pack_desc,
					 vimage->tc_green_index),
		  (CONST unsigned char *)vimage->tc_slice +
		  ds_get_element_offset (pack_desc,
					 vimage->tc_blue_index),
		  arr_desc->offsets[vimage->tc_hdim],
		  arr_desc->offsets[vimage->tc_vdim],
		  hdim->length, vdim->length,
		  &vimage->cache) )
	    {
		fprintf (stderr, "%s: error computing RGB image\n",
			 function_name);
		return;
	    }
	}
    }
    else
    {
	/*  Compute image over region defined by world canvas  */
	if (vimage->pc_arr_desc != NULL)
	{
	    if ( !canvas_compute_pc_image (holder->canvas,
					   vimage->pc_arr_desc,
					   vimage->pc_slice,
					   vimage->pc_hdim,
					   vimage->pc_vdim,
					   vimage->pc_elem_index,
					   &vimage->cache) )
	    {
		fprintf (stderr, "%s: error computing PseudoColour image\n",
			 function_name);
		return;
	    }
	}
	else
	{
	    if ( !canvas_compute_rgb_image
		 (holder->canvas, vimage->tc_arr_desc, vimage->tc_slice,
		  vimage->tc_hdim, vimage->tc_vdim,
		  vimage->tc_red_index, vimage->tc_green_index,
		  vimage->tc_blue_index, &vimage->cache) )
	    {
		fprintf (stderr, "%s: error computing RGB image\n",
			 function_name);
		return;
	    }
	}
    }
    vimage->recompute = FALSE;
    /*  Also need to clear the changed flag because the size control function
	will only clear when auto intensity scaling is on, and it will not be
	cleared on the first canvas refresh (i.e. when no cache is valid).  */
    vimage->changed = FALSE;
}   /*  End Function _viewimg_recompute_image  */

static void _viewimg_draw_subcache (CanvasHolder holder, ViewableImage vimage,
				    struct win_scale_type *win_scale)
/*  [SUMMARY] Draw an image subcache for the panning mode.
    <holder> The canvas holder object.
    <vimage> The ViewableImage object.
    <win_scale> A pointer to the window scaling information.
    [RETURNS] Nothing.
*/
{
    KPixCanvasRefreshArea area;
    int xstart, ystart;
    unsigned long hstart, vend;
    dim_desc *hdim, *vdim;

    vimage->last_pan_magnification = holder->pan_magnification_x;
    /*  Compute part of image cache to draw  */
    if (vimage->pc_arr_desc != NULL)
    {
	hdim = vimage->pc_arr_desc->dimensions[vimage->pc_hdim];
	vdim = vimage->pc_arr_desc->dimensions[vimage->pc_vdim];
    }
    else
    {
	hdim = vimage->tc_arr_desc->dimensions[vimage->tc_hdim];
	vdim = vimage->tc_arr_desc->dimensions[vimage->tc_vdim];
    }
    hstart = ds_get_coord_num (hdim, win_scale->left_x, SEARCH_BIAS_CLOSEST);
    xstart = (hstart * holder->pan_magnification_x) - win_scale->x_offset;
    vend = ds_get_coord_num (vdim, win_scale->top_y, SEARCH_BIAS_CLOSEST);
    vend = vdim->length - 1 - vend;
    ystart = (vend * holder->pan_magnification_y) - win_scale->y_offset;
#if 0
    fprintf (stderr,
	     "subcache X, C: %ld size: %g start: %d end: %d  xoff: %d xpix: %d\n",
	     holder->pan_centre_x, hdim->length * holder->pan_magnification_x,
	     xstart, xstart + win_scale->x_pixels - 1,
	     win_scale->x_offset, win_scale->x_pixels);
    fprintf (stderr,
	     "         Y, C: %ld size: %g start: %d end: %d  yoff: %d ypix: %d\n",
	     holder->pan_centre_y, vdim->length * holder->pan_magnification_y,
	     ystart, ystart + win_scale->y_pixels - 1,
	     win_scale->y_offset, win_scale->y_pixels);
#endif
    m_clear (&area, sizeof area);
    area.startx = win_scale->x_offset;
    area.endx = win_scale->x_offset + win_scale->x_pixels - 1;
    area.starty = win_scale->y_offset;
    area.endy = win_scale->y_offset + win_scale->y_pixels - 1;
    kwin_draw_cached_subimages (vimage->cache, -xstart, -ystart, 1, &area);
}   /*  End Function _viewimg_draw_subcache  */

static flag _viewimg_worldcanvas_position_func (KWorldCanvas canvas,
						double x, double y,
						unsigned int event_code,
						void *e_info, void **f_info,
						double x_lin, double y_lin,
						int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transforms_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    CanvasHolder holder = *f_info;
    ViewableImage vimage;
    unsigned int hdim_index, vdim_index;
    unsigned long x_coord, y_coord;
    double dx, dy;
    struct position_struct data;
    unsigned char rgb_value[3];
    double value[2];
    array_desc *arr_desc;
    packet_desc *pack_desc;
    dim_desc *hdim;
    dim_desc *vdim;
    static char function_name[] = "_viewimg_worldcanvas_position_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    if (holder->position_list == NULL) return (FALSE);
    if (holder->active_image == NULL) return (FALSE);
    vimage = holder->active_image;
    if (vimage->tc_arr_desc == NULL)
    {
	arr_desc = vimage->pc_arr_desc;
	hdim_index = vimage->pc_hdim;
	vdim_index = vimage->pc_vdim;
	data.value = (void *) value;
	data.type = K_DCOMPLEX;
    }
    else
    {
	arr_desc = vimage->tc_arr_desc;
	hdim_index = vimage->tc_hdim;
	vdim_index = vimage->tc_vdim;
	data.value = (void *) rgb_value;
	data.type = K_UB_RGB;
    }
    pack_desc = arr_desc->packet;
    hdim = arr_desc->dimensions[hdim_index];
    vdim = arr_desc->dimensions[vdim_index];
    dx = x_lin;
    dy = y_lin;
    if (hdim->first_coord < hdim->last_coord)
    {
	if (dx < hdim->first_coord) x_coord = 0;
	else if (dx >= hdim->last_coord) x_coord = hdim->length - 1;
	else
	{
	    ds_convert_coordinates (hdim, 1, &dx, TRUE, FALSE, FALSE);
	    x_coord = floor (dx + 0.5);
	}
    }
    else
    {
	if (dx < hdim->last_coord) x_coord = hdim->length - 1;
	else if (dx >= hdim->first_coord) x_coord = 0;
	else
	{
	    ds_convert_coordinates (hdim, 1, &dx, TRUE, FALSE, FALSE);
	    x_coord = floor (dx + 0.5);
	}
    }
    if (vdim->first_coord < vdim->last_coord)
    {
	if (dy < vdim->first_coord) y_coord = 0;
	else if (dy >= vdim->last_coord) y_coord = vdim->length - 1;
	else
	{
	    ds_convert_coordinates (vdim, 1, &dy, TRUE, FALSE, FALSE);
	    y_coord = floor (dy + 0.5);
	}
    }
    else
    {
	if (dy < vdim->last_coord) y_coord = vdim->length - 1;
	else if (dy >= vdim->first_coord) y_coord = 0;
	else
	{
	    ds_convert_coordinates (vdim, 1, &dy, TRUE, FALSE, FALSE);
	    y_coord = floor (dy + 0.5);
	}
    }
    if (vimage->tc_arr_desc == NULL)
    {
	if ( !ds_get_element (vimage->pc_slice +
			      ds_get_element_offset (pack_desc,
						     vimage->pc_elem_index) +
			      arr_desc->offsets[vdim_index][y_coord] +
			      arr_desc->offsets[hdim_index][x_coord],
			      pack_desc->element_types[vimage->pc_elem_index],
			      value, (flag *) NULL) )
	{
	    fprintf (stderr, "Error converting data\n");
	    return (FALSE);
	}
    }
    else
    {
	if ( !ds_get_element (vimage->tc_slice +
			      ds_get_element_offset (pack_desc,
						     vimage->tc_red_index) +
			      arr_desc->offsets[vdim_index][y_coord] +
			      arr_desc->offsets[hdim_index][x_coord],
			      pack_desc->element_types[vimage->tc_red_index],
			      value, (flag *) NULL) )
	{
	    fprintf (stderr, "Error converting data\n");
	    return (FALSE);
	}
	rgb_value[0] = 0xff & (int) value[0];
	if ( !ds_get_element (vimage->tc_slice +
			      ds_get_element_offset (pack_desc,
						     vimage->tc_green_index) +
			      arr_desc->offsets[vdim_index][y_coord] +
			      arr_desc->offsets[hdim_index][x_coord],
			      pack_desc->element_types[vimage->tc_green_index],
			      value, (flag *) NULL) )
	{
	    fprintf (stderr, "Error converting data\n");
	    return (FALSE);
	}
	rgb_value[1] = 0xff & (int) value[0];
	if ( !ds_get_element (vimage->tc_slice +
			      ds_get_element_offset (pack_desc,
						     vimage->tc_blue_index) +
			      arr_desc->offsets[vdim_index][y_coord] +
			      arr_desc->offsets[hdim_index][x_coord],
			      pack_desc->element_types[vimage->tc_blue_index],
			      value, (flag *) NULL) )
	{
	    fprintf (stderr, "Error converting data\n");
	    return (FALSE);
	}
	rgb_value[2] = 0xff & (int) value[0];
    }
    dx = ds_get_coordinate (hdim, x_coord);
    dy = ds_get_coordinate (vdim, y_coord);
    data.x = dx;
    data.y = dy;
    canvas_coords_transform (canvas, 1, &data.x, FALSE, &data.y, FALSE);
    data.x_lin = dx;
    data.y_lin = dy;
    data.x_im = x_coord;
    data.y_im = y_coord;
    data.x_pix = x_pix;
    data.y_pix = y_pix;
    /*  Call event consumer functions  */
    data.event_code = event_code;
    data.e_info = e_info;
    return c_call_callbacks (holder->position_list, &data);
}   /*  End Function _viewimg_worldcanvas_position_func  */

static flag _viewimg_position_event_func (void *object, void *client1_data,
					  void *call_data, void *client2_data)
/*  This routine is called when object callbacks are called.
    The object information pointer will be given by  object  .
    The first client information pointer will be given by  client1_data  .
    The call information pointer will be given by  call_data  .
    The second client information pointer will be given by  client2_data  .
    The routine returns TRUE if further callbacks should not be called.
*/
{
    CanvasHolder holder = object;
    ViewableImage vimage;
    PositionFunc func = (PositionFunc) client2_data;
    struct position_struct *data = call_data;
    static char function_name[] = "_viewimg_position_event_func";

    vimage = holder->active_image;
    VERIFY_USED_VIMAGE (vimage);
    return (*func) (vimage, data->x, data->y, data->value,
		    data->event_code, data->e_info, client1_data,
		    data->x_lin, data->y_lin, data->type,
		    data->x_im, data->y_im, data->x_pix, data->y_pix);
}   /*  End Function _viewimg_position_event_func  */

static flag _viewimg_coord_convert_func (KWorldCanvas canvas,
					 unsigned int num_coords,
					 CONST double *xin, CONST double *yin,
					 double *xout, double *yout,
					 flag to_linear, void **info)
/*  [SUMMARY] Co-ordinate conversion callback.
    [PURPOSE] This routine will convert between linear co-ordinates and pixel
    co-ordinates for a world canvas.
    <canvas> The canvas.
    <num_coords> The number of co-ordinates to transform.
    <xin> The array of input horizontal co-ordinates.
    <yin> The array of input vertical co-ordinates.
    <xout> The array of output horizontal co-ordinates are written here.
    <yout> The array of output vertical co-ordinates are written here.
    <to_linear> If TRUE, then a pixel to linear co-ordinate transform is
    required, else a linear to pixel co-ordinate transform is required.
    <info> A pointer to the arbitrary canvas information pointer.
    [RETURNS] TRUE if the conversion was completed, else FALSE indicating
    that the default conversions should be used.
*/
{
    CanvasHolder holder = *info;
    ViewableImage vimage;
    int count;
    int coord_num0, coord_num1;
    double coord0, coord1;
    double x_offset, y_offset, x_pixels, y_pixels;
    double x_pix, y_pix, xdelt, ydelt;
    struct win_scale_type win_scale;
    array_desc *arr_desc;
    dim_desc *hdim;
    dim_desc *vdim;
    static char function_name[] = "_viewimg_coord_convert_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    if ( (vimage = holder->active_image) == NULL) return (FALSE);
    if (vimage->tc_arr_desc == NULL)
    {
	arr_desc = vimage->pc_arr_desc;
	hdim = arr_desc->dimensions[vimage->pc_hdim];
	vdim = arr_desc->dimensions[vimage->pc_vdim];
    }
    else
    {
	arr_desc = vimage->tc_arr_desc;
	hdim = arr_desc->dimensions[vimage->tc_hdim];
	vdim = arr_desc->dimensions[vimage->tc_vdim];
    }
    /*  These are the correspondences we have to insure:
	wx = left_x             <-->    px = x_offset
	wx = right_x + x_delta  <-->    px = x_offset + x_pixels
	wy = bottom_y           <-->    py = y_offset + y_pixels - 1
	wy = top_y + y_delta    <-->    py = y_offset - 1

	where x_delta and y_delta are the linear co-ordinate increments.
	*/
    canvas_get_attributes (canvas,
			   CANVAS_ATT_X_OFFSET, &win_scale.x_offset,
			   CANVAS_ATT_X_PIXELS, &win_scale.x_pixels,
			   CANVAS_ATT_Y_OFFSET, &win_scale.y_offset,
			   CANVAS_ATT_Y_PIXELS, &win_scale.y_pixels,
			   CANVAS_ATT_LEFT_X, &win_scale.left_x,
			   CANVAS_ATT_RIGHT_X, &win_scale.right_x,
			   CANVAS_ATT_BOTTOM_Y, &win_scale.bottom_y,
			   CANVAS_ATT_TOP_Y, &win_scale.top_y,
			   CANVAS_ATT_END);
    x_offset = win_scale.x_offset;
    x_pixels = win_scale.x_pixels;
    y_offset = win_scale.y_offset;
    y_pixels = win_scale.y_pixels;
    if (x_pixels < 1)
    {
	fprintf (stderr, "Illegal number of X pixels: %d\n",
		 win_scale.x_pixels);
	a_prog_bug (function_name);
    }
    if (y_pixels < 1)
    {
	fprintf (stderr, "Illegal number of Y pixels: %d\n",
		 win_scale.y_pixels);
	a_prog_bug (function_name);
    }
    /*  Compute increment between dimension indices. This increment is used to
	place a point that lies exactly on a data value's co-ordinate at the
	centre of the box of pixels for the data value  */
    xdelt = ds_get_coordinate (hdim, 1) - ds_get_coordinate (hdim, 0);
    ydelt = ds_get_coordinate (vdim, 1) - ds_get_coordinate (vdim, 0);
    if (to_linear)
    {
	/*  Convert from pixel to linear co-ordinates  */
	/*  Compute x co-ordinates  */
	coord_num0 = ds_get_coord_num (hdim, win_scale.left_x,
				       SEARCH_BIAS_CLOSEST);
	coord0 = ds_get_coordinate (hdim, coord_num0);
	if (hdim->length < 2)
	{
	    for (count = 0; count < num_coords; ++count) xout[count] = coord0;
	}
	else
	{
	    coord_num1 = ds_get_coord_num (hdim, win_scale.right_x,
					   SEARCH_BIAS_CLOSEST);
	    if (coord_num1 + 1 < hdim->length)
		coord1 = ds_get_coordinate (hdim, coord_num1 + 1);
	    else
	    {
		coord1 = ds_get_coordinate (hdim, coord_num1);
		coord1 += (coord1 - coord0) / (coord_num1 - coord_num0);
	    }
	    for (count = 0; count < num_coords; ++count)
	    {
		x_pix = xin[count] - x_offset;
		x_pix = (coord1 - coord0) * x_pix / x_pixels;
		xout[count] = x_pix + coord0 - xdelt * 0.5;
		if ( DOUBLE_IS_BLANK (xout[count]) )
		{
		    fprintf (stderr, "Blank output X value\n");
		    a_prog_bug (function_name);
		}
	    }
	}
	/*  Compute y co-ordinates  */
	coord_num0 = ds_get_coord_num (vdim, win_scale.bottom_y,
				       SEARCH_BIAS_CLOSEST);
	coord0 = ds_get_coordinate (vdim, coord_num0);
	if (vdim->length < 2)
	{
	    for (count = 0; count < num_coords; ++count) yout[count] = coord0;
	}
	else
	{
	    coord_num1 = ds_get_coord_num (vdim, win_scale.top_y,
					   SEARCH_BIAS_CLOSEST);
	    if (coord_num1 + 1 < vdim->length)
		coord1 = ds_get_coordinate (vdim, coord_num1 + 1);
	    else
	    {
		coord1 = ds_get_coordinate (vdim, coord_num1);
		coord1 += (coord1 - coord0) / (coord_num1 - coord_num0);
	    }
	    for (count = 0; count < num_coords; ++count)
	    {
		y_pix = yin[count] - y_offset;
		y_pix = (y_pixels - 1.0) - y_pix;
		/*  Removed offsets and flipped vertical  */
		y_pix = (coord1 - coord0) * y_pix / y_pixels;
		yout[count] = y_pix + coord0 - ydelt * 0.5;
		if ( DOUBLE_IS_BLANK (yout[count]) )
		{
		    fprintf (stderr, "Blank output Y value\n");
		    a_prog_bug (function_name);
		}
	    }
	}
	return (TRUE);
    }
    /*  Convert from linear to pixel co-ordinates  */
    /*  Compute x co-ordinates  */
    if (hdim->length < 2)
    {
	for (count = 0; count < num_coords; ++count)
	    xout[count] = x_offset + x_pixels / 2;
    }
    else
    {
	coord_num0 = ds_get_coord_num (hdim, win_scale.left_x,
				       SEARCH_BIAS_CLOSEST);
	coord0 = ds_get_coordinate (hdim, coord_num0);
	/*  Find co-ordinate one data value beyond canvas maximum  */
	coord_num1 = ds_get_coord_num (hdim, win_scale.right_x,
				       SEARCH_BIAS_CLOSEST);
	if (coord_num1 + 1 < hdim->length)
	    coord1 = ds_get_coordinate (hdim, coord_num1 + 1);
	else
	{
	    coord1 = ds_get_coordinate (hdim, coord_num1);
	    coord1 += (coord1 - coord0) / (coord_num1 - coord_num0);
	}
	for (count = 0; count < num_coords; ++count)
	{
	    x_pix = (xin[count] - coord0 + xdelt * 0.5) / (coord1 - coord0);
	    x_pix *= x_pixels;
	    xout[count] = x_pix + x_offset;
	}
    }
    /*  Compute y co-ordinates  */
    if (vdim->length < 2)
    {
	for (count = 0; count < num_coords; ++count)
	    yout[count] = y_offset + y_pixels / 2;
    }
    else
    {
	coord_num0 = ds_get_coord_num (vdim, win_scale.bottom_y,
				       SEARCH_BIAS_CLOSEST);
	coord0 = ds_get_coordinate (vdim, coord_num0);
	/*  Find co-ordinate one data value beyond canvas maximum  */
	coord_num1 = ds_get_coord_num (vdim, win_scale.top_y,
				       SEARCH_BIAS_CLOSEST);
	coord1 = ds_get_coordinate (vdim, coord_num1);
	if (coord_num1 + 1 < vdim->length)
	    coord1 = ds_get_coordinate (vdim, coord_num1 + 1);
	else
	{
	    coord1 = ds_get_coordinate (vdim, coord_num1);
	    coord1 += (coord1 - coord0) / (coord_num1 - coord_num0);
	}
	for (count = 0; count < num_coords; ++count)
	{
	    y_pix = (yin[count] - coord0 + ydelt * 0.5) / (coord1 - coord0);
	    y_pix *= y_pixels;
	    /*  Flip vertical  */
	    yout[count] = (y_offset + y_pixels - 1.0) - y_pix;
	}
    }
    return (TRUE);
}   /*  End Function _viewimg_coord_convert_func  */

static ViewableImage _viewimg_alloc_vimage ()
/*  [SUMMARY] Allocate and initialise a ViewableImage structure.
    [RETURNS] The ViewimgImage on success, else NULL.
*/
{
    ViewableImage vimage;
    static char function_name[] = "_viewimg_alloc_vimage";

    if (first_free == NULL)
    {
	if ( ( vimage = m_alloc (sizeof *vimage) ) == NULL ) return (NULL);
    }
    else
    {
	vimage = first_free;
	first_free = vimage->next;
	VERIFY_FREE_VIMAGE (vimage);
    }
    vimage->value_min = TOOBIG;
    vimage->value_max = TOOBIG;
    vimage->magic_number = VIMAGE_USED_MAGIC_NUMBER;
    vimage->canvas_holder = NULL;
    vimage->pc_multi_desc = NULL;
    vimage->pc_arr_desc = NULL;
    vimage->pc_slice = NULL;
    vimage->pc_hdim = 0;
    vimage->pc_vdim = 0;
    vimage->pc_elem_index = 0;
    vimage->tc_multi_desc = NULL;
    vimage->tc_arr_desc = NULL;
    vimage->tc_slice = NULL;
    vimage->recompute = TRUE;
    vimage->changed = TRUE;
    vimage->pixcanvas_width = -1;
    vimage->pixcanvas_height = -1;
    vimage->cache = NULL;
    /*  Compute strides  */
    vimage->pc_hstride = 0;
    vimage->pc_vstride = 0;
    vimage->pc_data_scale = 1.0;
    vimage->pc_data_offset = 0.0;
    vimage->num_restrictions = 0;
    vimage->restriction_names = NULL;
    vimage->restriction_values = NULL;
    vimage->iscale_test[0] = TOOBIG;
    canvas_init_win_scale (&vimage->win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    vimage->override.value_min = TOOBIG;
    vimage->override.value_max = TOOBIG;
    vimage->prev = NULL;
    vimage->next = NULL;
    return (vimage);
}   /*  End Function _viewimg_alloc_vimage  */

static void _viewimg_free_vimage (ViewableImage vimage)
/*  [SUMMARY] Free a ViewableImage structure.
    <vimage> The ViewableImage to free.
    [RETURNS] Nothing.
*/
{
    vimage->magic_number = VIMAGE_FREE_MAGIC_NUMBER;
    vimage->next = first_free;
    first_free = vimage;
}   /*  End Function _viewimg_free_vimage  */
