/*LINTLIBRARY*/
/*  track.c

    This code provides ViewableImage objects.

    Copyright (C) 1996-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains convenience routines to support track displays.


    Written by      Richard Gooch   3-NOV-1996

    Updated by      Richard Gooch   1-FEB-1997: Made use of new
  <wcs_astro_get_associated> function and changed interface to
  <viewimg_track_compute> function.

    Updated by      Richard Gooch   24-JUL-1997: Switched to %g formatting.

    Updated by      Richard Gooch   31-AUG-1999: Changed interface to function
  <viewimg_track_compute>.

    Updated by      Richard Gooch   12-JUL-2000: Added
  #include <karma_canvas.h>

    Last updated by Richard Gooch   22-SEP-2000: Took account of interface
  change in <ds_format_value>.


*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <karma_viewimg.h>
#include <karma_canvas.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_a.h>


/*  Private structures  */


/*  Private functions  */
STATIC_FUNCTION (unsigned int get_gen_index,
		 (CONST multi_array *multi_desc, CONST array_desc *arr_desc) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void viewimg_track_compute (ViewableImage vimage,
			    void *value, unsigned int value_type,
			    double x, double y,
			    double x_im, double y_im,
			    char pix_string[STRING_LENGTH],
			    char world_string[STRING_LENGTH],
			    char extra_string[STRING_LENGTH])
/*  [SUMMARY] Compute track strings.
    <vimage> The active ViewableImage object.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <pix_string> The pixel co-ordinates and data value will be written here.
    <world_string> The world co-ordinates and restriction information will be
    written here.
    <extra_string> Extra information is written here. This may be NULL if
    <<world_string>> is also NULL.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    unsigned int hdim, vdim, rdim, elem_index;
    unsigned int num_restr, count;
    unsigned long r_index, index;
    double scale, offset, min, max;
    unsigned char *rgb_ptr = (unsigned char *) value;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    multi_array *multi_desc;
    array_desc *arr_desc;
    char txt[STRING_LENGTH], value_string[STRING_LENGTH];
    static char function_name[] = "viewimg_track_compute";

    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_MULTI_ARRAY, &multi_desc,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_END);
    if (multi_desc == NULL) ap = NULL;
    else
    {
	index = get_gen_index (multi_desc, arr_desc);
	ap = wcs_astro_get_associated (multi_desc, index);
    }
    switch (value_type)
    {
      case K_DCOMPLEX:
	/*  Compute value  */
	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_PSEUDO_INDEX, &elem_index,
				VIEWIMG_VATT_DATA_SCALE, &scale,
				VIEWIMG_VATT_DATA_OFFSET, &offset,
				VIEWIMG_VATT_END);
	canvas_get_attributes (viewimg_get_worldcanvas (vimage),
			       CANVAS_ATT_VALUE_MIN, &min,
			       CANVAS_ATT_VALUE_MAX, &max,
			       CANVAS_ATT_END);
	ds_format_value (value_string, *(double *) value,
			 arr_desc->packet->element_desc[elem_index],
			 scale, offset, fabs (max - min) * 0.5, NULL, NULL);
	break;
      case K_UB_RGB:
	sprintf (value_string, "RGB: %u %u %u",
		 rgb_ptr[0], rgb_ptr[1], rgb_ptr[2]);
	break;
      default:
	fprintf (stderr, "Illegal type: %u\n", value_type);
	a_prog_bug (function_name);
	break;
    }
    canvas_get_specification (viewimg_get_worldcanvas (vimage),
			      &xlabel, &ylabel, &num_restr,
			      &restr_names, &restr_values);
    sprintf ( pix_string, "x: %g  y: %g  ",
	      floor (x_im + 0.5), floor (y_im + 0.5) );
    /*  Add any restriction information  */
    for (count = 0; count < num_restr; ++count)
    {
	if ( ( rdim = ds_f_dim_in_array (arr_desc, restr_names[count]) )
	     >= arr_desc->num_dimensions ) continue;
	r_index = ds_get_coord_num (arr_desc->dimensions[rdim],
				    restr_values[count], SEARCH_BIAS_CLOSEST);
	sprintf (txt, "z%u: %lu  ", rdim, r_index);
	strcat (pix_string, txt);
    }
    strcat (pix_string, value_string);
    if ( (world_string == NULL) && (extra_string == NULL) ) return;
    /*  Now display the world co-ordinate information  */
    if (ap == NULL)
    {
	sprintf (world_string, "%g %s  %g %s  ", x, xlabel, y, ylabel);
	/*  Add any restriction information  */
	for (count = 0; count < num_restr; ++count)
	{
	    sprintf (txt, "%g %s  ", restr_values[count], restr_names[count]);
	    strcat (world_string, txt);
	}
	if (extra_string != NULL) extra_string[0] = '\0';
    }
    else
    {
	wcs_astro_format_all (ap, world_string,
			      xlabel, x_im, ylabel, y_im, NULL, 0.0,
			      num_restr, (CONST char **) restr_names,
			      restr_values, extra_string);
    }
}   /*  End Function viewimg_track_compute  */


/*  Private functions follow  */

static unsigned int get_gen_index (CONST multi_array *multi_desc,
				   CONST array_desc *arr_desc)
/*  [SUMMARY] Find general data structure which has an array descriptor.
    <multi_desc> The multi_array header.
    <arr_desc> The array descriptor to search for.
    [RETURNS] The index of the general data structure in the multi_array header
*/
{
    unsigned int count;
    char *encls_desc;
    static char function_name[] = "get_gen_index";

    for (count = 0; count < multi_desc->num_arrays; ++count)
    {
	switch ( ds_f_name_in_packet (multi_desc->headers[count],
				      arr_desc->dimensions[0]->name,
				      &encls_desc, NULL) )
	{
	  case IDENT_NOT_FOUND:
	    break;
	  case IDENT_DIMENSION:
	    if (arr_desc != (array_desc *) encls_desc)
	    {
		fprintf (stderr, "Found array descriptor: %p is not: %p\n",
			 encls_desc, arr_desc);
		a_prog_bug (function_name);
	    }
	    return (count);
	    /*break;*/
	  default:
	    fprintf (stderr, "Illegal return from <ds_f_name_in_packet>\n");
	    a_prog_bug (function_name);
	    break;
	}
    }
    fprintf (stderr, "Array descriptor not found\n");
    a_prog_bug (function_name);
    return (0);
}   /*  End Function get_gen_index  */
