/*LINTLIBRARY*/
/*  values.c

    This code provides ViewableImage objects.

    Copyright (C) 2000-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains convenience routines which implement a values
    display user interface.


    Written by      Richard Gooch   14-JUN-2000: Copied from statistics.c

    Updated by      Richard Gooch   16-JUN-2000: Took account of interface
  change to function <ds_format_unit>.

    Updated by      Richard Gooch   12-JUL-2000: Added
  #include <karma_canvas.h>

    Updated by      Richard Gooch   13-JUL-2000: Changed interface to function
  <viewimg_values_display> and checked for control modifier in
  <viewimg_values_position_func>.

    Updated by      Richard Gooch   5-OCT-2001: Added #include <string.h>.

    Last updated by Richard Gooch   8-OCT-2001: Fixed documentation links.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_canvas.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>


/*  Private structures  */


/*  Private functions  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag viewimg_values_position_func (ViewableImage vimage,
				   double x, double y,
				   void *value, unsigned int event_code,
				   void *e_info, void **f_info,
				   double x_lin, double y_lin,
				   unsigned int value_type,
				   double x_im, double y_im,
				   int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transforms_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [NOTE] All co-ordinates are clipped to pixel boundaries.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas canvas;
    flag scaled;
    unsigned long mask = 0;
    double left_x, right_x, bottom_y, top_y;
    CONST char *string = e_info;
    /*static char function_name[] = "viewimg_values_position_func";*/

    switch (event_code)
    {
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	m_copy (&mask, string + strlen (string) + 1, sizeof mask);
	if (mask != K_CANVAS_EVENT_CONTROL_KEY_MASK) return (FALSE);
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    if (strcmp (string, "v") == 0) scaled = TRUE;
    else if (strcmp (string, "V") == 0) scaled = FALSE;
    else return (FALSE);
    if (value_type != K_DCOMPLEX) return (FALSE);
    /*  Determine start and stop co-ordinates along each dimension  */
    canvas = viewimg_get_worldcanvas (vimage);
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    if ( !viewimg_values_display (vimage, scaled, mask ? TRUE : FALSE,
				  left_x, bottom_y, right_x, top_y) )
	fprintf (stderr,
		 "You've asked for a large number of values: press control-v instead%c\n",
		 7);
    return (TRUE);
}   /*  End Function viewimg_values_position_func  */

/*EXPERIMENTAL_FUNCTION*/
flag viewimg_values_display (ViewableImage vimage, flag scaled, flag force,
			     double lx0, double ly0, double lx1,double ly1)
/*  [SUMMARY] Compute and display values for a subimage.
    <vimage> The active viewable image.
    <scaled> If TRUE, scaled values are displayed, else raw data are displayed.
    <force> If TRUE, a large number of values will always be displayed. If
    FALSE and a large number of values is requested, the function fails.
    <lx0> The first horizontal linear world co-ordinate.
    <ly0> The first vertical linear world co-ordinate.
    <lx1> The second horizontal linear world co-ordinate.
    <ly1> The second vertical linear world co-ordinate.
    [RETURNS] TRUE on success, else FALSE (too many values).
*/
{
    flag complex;
    unsigned int hstart, hend, vstart, vend, tmp_ui;
    unsigned int hdim, vdim, elem_index, hcount, vcount;
    double scale, offset, unit_scale;
    char *array;
    uaddr *hoffsets, *voffsets;
    array_desc *arr_desc;
    packet_desc *pack_desc;
    dim_desc *hdim_desc, *vdim_desc;
    char unit[STRING_LENGTH], unit_format[STRING_LENGTH];
    double value[2];

    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_SLICE, &array,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_PSEUDO_INDEX, &elem_index,
			    VIEWIMG_VATT_DATA_SCALE, &scale,
			    VIEWIMG_VATT_DATA_OFFSET, &offset,
			    VIEWIMG_VATT_END);
    hoffsets = arr_desc->offsets[hdim];
    voffsets = arr_desc->offsets[vdim];
    hdim_desc = arr_desc->dimensions[hdim];
    vdim_desc = arr_desc->dimensions[vdim];
    pack_desc = arr_desc->packet;
    /*  Determine start and stop co-ordinates along each dimension  */
    hstart = ds_get_coord_num (hdim_desc, lx0, SEARCH_BIAS_CLOSEST);
    hend = ds_get_coord_num (hdim_desc, lx1, SEARCH_BIAS_CLOSEST);
    if (hstart >= hend) swap (hstart, hend, tmp_ui);
    vstart = ds_get_coord_num (vdim_desc, ly0, SEARCH_BIAS_CLOSEST);
    vend = ds_get_coord_num (vdim_desc, ly1, SEARCH_BIAS_CLOSEST);
    if (vstart >= vend) swap (vstart, vend, tmp_ui);
    if ( !force && ( (vend - vstart + 1) * (hend - hstart + 1) > 100 ) )
	return (FALSE);
    array += ds_get_element_offset (pack_desc, elem_index);
    if (scaled)
    {
	ds_get_element (array + hoffsets[hstart] + voffsets[vstart],
			pack_desc->element_types[elem_index], value,
			&complex);
	ds_format_unit (unit, unit_format, &unit_scale,
			pack_desc->element_desc[elem_index],
			value[0] * scale + offset);
    }
    else
    {
	if ( (scale != 1.0) || (offset != 0.0) ) strcpy (unit, "raw");
	else strcpy (unit, pack_desc->element_desc[elem_index]);
	strcpy (unit_format, "%-5g");
	scale = 1.0;
	offset = 0.0;
	unit_scale = 1.0;
    }
    printf ("Displaying pixels: (%u,%u) to (%u,%u)  Units: %s\n",
	    hstart, vstart, hend, vend, unit);
    for (vcount = vstart; vcount <= vend; ++vcount)
    {
	for (hcount = hstart; hcount <= hend; ++hcount)
	{
	    ds_get_element (array + hoffsets[hcount] + voffsets[vcount],
			    pack_desc->element_types[elem_index], value,
			    &complex);
	    if ( DOUBLE_IS_BLANK (value[0]) ) fputs ("BLANK", stdout);
	    else printf (unit_format,
			 (value[0] * scale + offset) * unit_scale);
	    if (complex)
	    {
		putc (' ', stdout);
		if ( DOUBLE_IS_BLANK (value[1]) ) fputs ("BLANK", stdout);
		else printf (unit_format,
			     (value[1] * scale + offset) * unit_scale);
	    }
	    fputs ("  ", stdout);
	}
	putc ('\n', stdout);
    }
    fflush (stdout);
    return (TRUE);
}   /*  End Function viewimg_values_display  */
