/*LINTLIBRARY*/
/*  Dataclip.c

    This code provides a data clip control widget for Xt.

    Copyright (C) 1994-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a data clip control widget for
  Xt.


    Written by      Richard Gooch   23-OCT-1994

    Updated by      Richard Gooch   26-NOV-1994

    Updated by      Richard Gooch   25-DEC-1994: Added autoValueScale resource.

    Uupdated by     Richard Gooch   27-DEC-1994: Delay computing histogram
  until widget popped up.

    Updated by      Richard Gooch   6-JAN-1995: Changed from passing flag
  by value to by reference for intensityScaleCallback.

    Updated by      Richard Gooch   9-APR-1995: Added initialisation of
  conv_type  field in  win_scale  structure.

    Updated by      Richard Gooch   17-APR-1995: Fixed mix of XtRBool and
  sizeof Boolean in resource fields. Use Bool because it is of int size.

    Updated by      Richard Gooch   21-APR-1995: Added Full Range button.

    Updated by      Richard Gooch   30-JUL-1995: Added destroy callback for
  Intelligent Array.

    Updated by      Richard Gooch   28-DEC-1995: Added verbose resource.

    Updated by      Richard Gooch   22-FEB-1996: Allow dragging of end-points.

    Updated by      Richard Gooch   18-MAR-1996: Display clip limits.

    Updated by      Richard Gooch   4-MAY-1996: Switched to KtoggleWidget and
  added showBlankControl resource.

    Updated by      Richard Gooch   24-MAY-1996: Completed save functionality
  and set XtNresizable resource to TRUE for the save dialog widget so that
  longer names may be typed and viewed.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   4-JUN-1996: Switched to left-right
  bottom-top co-ordinate specification instead of min-max x and y.

    Updated by      Richard Gooch   15-JUN-1996: Added scaling information when
  writing integer types.

    Updated by      Richard Gooch   16-JUN-1996: Read input scaling information
  and combine with computed scaling when writing integer types.

    Updated by      Richard Gooch   20-JUN-1996: Fixed bug in <SetValues> where
  destroy callback was not being correctly cleared and set.

    Updated by      Richard Gooch   21-JUN-1996: Fixed bug in
  <iarr_destroy_callback> where destroy callback was not being cleared.

    Updated by      Richard Gooch   15-AUG-1996: Added autoPopdown resource.

    Updated by      Richard Gooch   2-OCT-1996: Added <get_colour> routine to
  make widget more robust when insufficient colours are available. Added
  cursor position label. Apply cube scaling attachments to all value displays.

    Updated by      Richard Gooch   13-OCT-1996: Added Zoom and Unzoom buttons.

    Updated by      Richard Gooch   12-NOV-1996: Fixed bug in <apply_cbk> so
  that when <<num_regions>> is zero, no region callbacks are called.

    Updated by      Richard Gooch   31-DEC-1996: Created <XkwDataclipNewArray>
  function.

    Updated by      Richard Gooch   13-JAN-1997: Fixed bug in <update_limits>:
  did not check if array is NULL.

    Updated by      Richard Gooch   30-JAN-1997: Created
  <XkwDataclipArrayChanged> function.

    Updated by      Richard Gooch   13-APR-1997: Also display cursor position
  when clicking or dragging left or right mouse buttons. Show values with
  preferred units.

    Updated by      Richard Gooch   15-APR-1997: Added axis labels for
  histogram display (no tick marks or scale).

    Updated by      Richard Gooch   1-JUN-1997: Move centre of region using
  horizontal position of middle button.

    Updated by      Richard Gooch   4-JUN-1997: Do not require cursor to be in
  region when moving centre if only one region is defined.

    Updated by      Richard Gooch   9-JUN-1997: Change region width using
  vertical position of middle button.

    Updated by      Richard Gooch   12-AUG-1997: Allow auto apply state to be
  set from resources.

    Updated by      Richard Gooch   30-SEP-1997: Added "Clip Outliers" button.

    Updated by      Richard Gooch   2-OCT-1997: Add right button first click if
  no regions are define: left boundary is set to data minimum.

    Updated by      Richard Gooch   31-OCT-1997: Added new <<reset_hscale>>
  parameter to <XkwDataclipNewArray>.

    Updated by      Richard Gooch   8-DEC-1997: Added range of clip buttons.

    Updated by      Richard Gooch   9-DEC-1997: Don't apply in outlier clip
  if auto apply is off.

    Updated by      Richard Gooch   12-DEC-1997: Removed most border chaining.

    Updated by      Richard Gooch   22-DEC-1997: Turned on "Auto Apply" by
  default.

    Updated by      Richard Gooch   25-MAR-1998: Fixed bug in position handler:
  did not trap zero range when moving middle mouse button.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   19-MAY-1998: Set data range when writing
  scaled output.

    Updated by      Richard Gooch   14-JUL-1998: Fixed some problems with
  persistent invalid states.

    Updated by      Richard Gooch   12-AUG-1998: Now update cursor position on
  any canvas event.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   24-OCT-1998: Added "DataclipWidget::Clip"
  event.

    Updated by      Richard Gooch   25-MAR-1999: Fixed bug in scaling
  information when writing float files.

    Updated by      Richard Gooch   9-APR-1999: Added "Zero Left" button.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   17-JUN-2000: Added support for "join" and
  "steps" display modes.

    Updated by      Richard Gooch   19-JUN-2000: Removed support for "join" and
  "steps" display modes: doesn't really make much sense.

    Updated by      Richard Gooch   27-NOV-2000: Fixed bug in function
  <histcanvas_size_control_func>: use total pixel canvas width to determine
  array length.

    Updated by      Richard Gooch   10-MAR-2001: Added Y position label.

    Updated by      Richard Gooch   11-MAR-2001: Changed Y position label to
  show number of values in bin.

    Last updated by Richard Gooch   20-MAR-2001: Created Destroy method.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/wait.h>
#include <signal.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_iarray.h>
#include <karma_c.h>
#include <karma_xtmisc.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/DataclipP.h>
#include <Xkw/Canvas.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>


/*  External functions  */


/*  Private functions  */
STATIC_FUNCTION (void _Dataclip_ClassInititialise, () );
STATIC_FUNCTION (void _Dataclip_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _Dataclip_Destroy, (Widget W) );
STATIC_FUNCTION (void _Dataclip_cnv_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (Boolean _Dataclip_SetValues,
		 (Widget current, Widget request, Widget new) );
STATIC_FUNCTION (void immediate_apply_cbk, (Widget w, XtPointer client_data,
					    XtPointer call_data) );
STATIC_FUNCTION (void apply_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void histcanvas_size_control_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  void **info, flag *boundary_clear) );
STATIC_FUNCTION (void histogram_canvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );
STATIC_FUNCTION (flag histogram_canvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info,
		  double x_lin, double y_lin, int x_pix, int y_pix) );
STATIC_FUNCTION (void pop_cbk, (Widget w, XtPointer client_data,
				XtPointer call_data) );
STATIC_FUNCTION (void full_range_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (void iarr_destroy_callback, (iarray arr,DataclipWidget top) );
STATIC_FUNCTION (void blank_data_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (void type_menu_cbk, (Widget w, XtPointer client_data,
				      XtPointer call_data) );
STATIC_FUNCTION (void save_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (unsigned long get_colour,
		 (KWorldCanvas canvas, CONST char *colourname) );
STATIC_FUNCTION (void zoom_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void unzoom_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void update_limits, (DataclipWidget w, flag reset_hscale) );
STATIC_FUNCTION (void clip_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void clip_outliers, (DataclipWidget top, double proportion) );
STATIC_FUNCTION (flag _Dataclip_ds_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


#define offset(field) XtOffset(DataclipWidget, dataclip.field)

static XtResource DataclipResources[] = 
{
    {XkwNmaxDataRegions, XkwCMaxDataRegions, XtRInt, sizeof (int),
     offset (maxRegions), XtRImmediate, (XtPointer) 1},
    {XkwNregionCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (regionCallback), XtRCallback, (caddr_t) NULL},
    {XkwNshowBlankControl, XkwCShowBlankControl, XtRBool, sizeof (Bool),
     offset (showBlankControl), XtRImmediate, (XtPointer) False},
    {XkwNfixedOutputType, XkwCFixedOutputType, XtRCardinal, sizeof (Cardinal),
     offset (fixedOutputType), XtRImmediate, (XtPointer) NONE},
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, False},
    {XkwNautoPopdown, XkwCAutoPopdown, XtRBool, sizeof (Bool),
     offset (autoPopdown), XtRImmediate, (XtPointer) FALSE},
};
#undef offset

DataclipClassRec dataclipClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass)&topLevelShellClassRec,  /* superclass */
    "Dataclip",                    /* class_name */
    sizeof (DataclipRec),          /* widget_size */
    _Dataclip_ClassInititialise,   /* class_initialise */
    NULL,                          /* class_part_initialise */
    FALSE,                         /* class_init */
    (XtInitProc) _Dataclip_Initialise,  /* initialise */
    NULL,                          /* initialise_hook */
    XtInheritRealize,              /* realise */
    NULL,                          /* actions */
    0,                             /* num_actions */
    DataclipResources,             /* resources */
    XtNumber (DataclipResources),  /* num_resources */
    NULLQUARK,                     /* xrm_class */
    TRUE,                          /* compress_motion */
    TRUE,                          /* compress_exposure */
    TRUE,                          /* compress_enterleave */
    TRUE,                          /* visible_interest */
    _Dataclip_Destroy,             /* destroy */
    XtInheritResize,               /* resize */
    NULL,                          /* expose */
    (XtSetValuesFunc) _Dataclip_SetValues,   /* set_values */
    NULL,                          /* set_values_hook */
    XtInheritSetValuesAlmost,      /* set_values_almost */
    NULL,                          /* get_values_hook */
    NULL,                          /* accept_focus */
    XtVersion,                     /* version */
    NULL,                          /* callback_private */
    NULL,                          /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* DataClipClassPart */
      {0}, {NULL}, 0      /* empty */
  }
};

WidgetClass dataclipWidgetClass = (WidgetClass) &dataclipClassRec;

static void _Dataclip_ClassInititialise ()
{
    unsigned int count, index;
    extern char *data_type_names[NUMTYPES];

    for (count = 0, index = 0; count < NUMTYPES; ++count)
    {
	if ( !ds_element_is_legal (count) ) continue;
	if ( !ds_element_is_atomic (count) ) continue;
	if ( ds_element_is_complex (count) ) continue;
	dataclipClassRec.dataclip_class.type_index_to_type[index] = count;
	dataclipClassRec.dataclip_class.type_names[index] =
	    data_type_names[count];
	++index;
    }
    dataclipClassRec.dataclip_class.num_types = index;
}   /*  End Function ClassInitialise  */

static void _Dataclip_Initialise (Widget Request, Widget New)
{
    unsigned int count;
    /*DataclipWidget request = (DataclipWidget) Request;*/
    DataclipWidget new = (DataclipWidget) New;
    Widget form;
    Widget close_btn, save_btn;
    Widget cnv;
    Widget w, upper_widget;
    static String clips[] = {"95%", "98%", "99%", "99.5%", "99.9%", "99.99%"};
    static char function_name[] = "DataclipWidget::Initialise";

    new->dataclip.pixcanvas = NULL;
    new->dataclip.worldcanvas = NULL;
    new->dataclip.iarr_destroy_callback = NULL;
    new->dataclip.last_was_left = FALSE;
    new->dataclip.left_pos = TOOBIG;
    new->dataclip.popped_up = FALSE;
    new->dataclip.blank_data = FALSE;
    new->dataclip.array = NULL;
    new->dataclip.reset_canvas_scale = TRUE;
    XtAddCallback (New, XtNpopupCallback, pop_cbk, (XtPointer) TRUE);
    XtAddCallback (New, XtNpopdownCallback, pop_cbk, (XtPointer) FALSE);
    if (new->dataclip.maxRegions < 1)
    {
	fprintf (stderr, "maxRegions: %u  must be greater than 0\n",
		 new->dataclip.maxRegions);
	a_prog_bug (function_name);
    }
    if ( ( new->dataclip.minima =
	   m_alloc (sizeof *new->dataclip.minima * new->dataclip.maxRegions) )
	 == NULL )
    {
	m_abort (function_name, "array of minima");
    }
    if ( ( new->dataclip.maxima =
	   m_alloc (sizeof *new->dataclip.maxima * new->dataclip.maxRegions) )
	 == NULL )
    {
	m_abort (function_name, "array of maxima");
    }
    new->dataclip.num_regions = 0;
    new->dataclip.histogram_array = NULL;
    new->dataclip.hist_arr_length = 0;
    new->dataclip.hist_buf_length = 0;
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNheight, 20,
				 NULL);
    close_btn = w;
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk, New);
    if (new->dataclip.showBlankControl)
    {
	w = XtVaCreateManagedWidget ("blankToggle", ktoggleWidgetClass, form,
				     XtNlabel, "Blank",
				     XtNfromHoriz, w,
				     XtNheight, 20,
				     XtNstate, new->dataclip.blank_data,
				     XkwNcrosses, False,
				     NULL);
	XtAddCallback (w, XtNcallback, blank_data_cbk, (XtPointer) New);
    }
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Zoom",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, zoom_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Unzoom",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, unzoom_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Full Range",
				 XtNfromHoriz, w,
				 /*XtNfromVert, close_btn,*/
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, full_range_cbk, (XtPointer) New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Apply",
				 XtNfromVert, close_btn,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, apply_cbk, New);
    upper_widget = w;
    if (new->dataclip.maxRegions == 1)
    {
	w = XtVaCreateManagedWidget ("autoApplyToggle",ktoggleWidgetClass,form,
				     XtNlabel, "Auto Apply",
				     XtNfromHoriz, w,
				     XtNfromVert, close_btn,
				     XtNheight, 20,
				     XkwNcrosses, False,
				     XtNstate, True,
				     NULL);
	XtAddCallback (w, XtNcallback, immediate_apply_cbk, (XtPointer) New);
	XtVaGetValues (w, XtNstate, &new->dataclip.immediate_apply, NULL);
    }
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Zero Left",
				 XtNfromHoriz, w,
				 XtNfromVert, close_btn,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, clip_cbk, (XtPointer) New);
    w = NULL;
    for (count = 0; count < sizeof clips / sizeof (String); ++count)
    {
	w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				     XtNlabel, clips[count],
				     XtNfromVert, upper_widget,
				     XtNfromHoriz, w,
				     XtNheight, 20,
				     NULL);
	XtAddCallback (w, XtNcallback, clip_cbk, (XtPointer) New);
    }
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Data Minimum:              ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.min_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Data Maximum:              ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.max_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Lower Clip:                      ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.lower_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Upper Clip:                      ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.upper_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Bin value:                       ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.cursor_x_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Number of values in bin:         ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    new->dataclip.cursor_y_label = w;
    save_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					XtNlabel, "Save",
					XtNfromVert, w,
					NULL);
    if (new->dataclip.fixedOutputType == NONE)
    {
	w = XtVaCreateManagedWidget("menuButton", exclusiveMenuWidgetClass,
				    form,
				    XtNmenuName, "typeMenu",
				    XkwNchoiceName, "Output Type",
				    XtNfromVert, w,
				    XtNfromHoriz, save_btn,
				    XkwNnumItems,
				    dataclipClassRec.dataclip_class.num_types,
				    XkwNitemStrings,
				    dataclipClassRec.dataclip_class.type_names,
				    NULL);
	XtAddCallback (w, XkwNselectCallback, type_menu_cbk, New);
	new->dataclip.output_type =
	    dataclipClassRec.dataclip_class.type_index_to_type[0];
    }
    else new->dataclip.output_type = new->dataclip.fixedOutputType;
    w = save_btn;
    w = XtVaCreateManagedWidget ("filename", dialogWidgetClass, form,
				 XtNlabel, "Filename: ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNvalue, "scaled",
				 XtNresizable, TRUE,
				 NULL);
    new->dataclip.dialog = w;
    XtAddCallback (save_btn, XtNcallback, save_cbk, New);
    cnv = XtVaCreateManagedWidget ("pseudoColourCanvas", canvasWidgetClass,
				   form,
				   XtNfromVert, w,
				   XtNleft, XtChainLeft,
				   XtNright, XtChainRight,
				   XtNbottom, XtChainBottom,
				   XtNwidth, 350,
				   XtNheight, 250,
				   XkwNsilenceUnconsumed, True,
				   XtNresizable, True,
				   NULL);
    new->dataclip.cnv = cnv;
    XtAddCallback (cnv, XkwNrealiseCallback, _Dataclip_cnv_realise_cbk, New);
    ds_event_register_func (_Dataclip_ds_event_func, New);
}   /*  End Function Initialise  */

static void _Dataclip_Destroy (Widget W)
{
    DataclipWidget w = (DataclipWidget) W;

    if (w->dataclip.minima != NULL) m_free (w->dataclip.minima);
    if (w->dataclip.maxima != NULL) m_free (w->dataclip.maxima);
    if (w->dataclip.histogram_array != NULL)
	m_free (w->dataclip.histogram_array);
    if (w->dataclip.iarr_destroy_callback != NULL)
	c_unregister_callback (w->dataclip.iarr_destroy_callback);
}   /*  End Function Destroy  */

static void _Dataclip_cnv_realise_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
/*  This is the canvas realise callback.
*/
{
    KPixCanvas pixcanvas = (KPixCanvas) call_data;
    KWorldCanvas wc;
    GC gc;
    XGCValues gcvalues;
    struct win_scale_type win_scale;
    DataclipWidget top = (DataclipWidget) client_data;
    char *hlabel;
    /*static char function_name[] = "DataclipWidget::cnv_realise_cbk";*/

    top->dataclip.pixcanvas = pixcanvas;
    /*  Get the Graphics Context of the pixel canvas  */
    gc = kwin_get_gc_x (pixcanvas);
    if (XGetGCValues (XtDisplay (w), gc, GCForeground | GCBackground,
		      &gcvalues) == 0)
    {
	fprintf (stderr, "Error getting GC values\n");
	exit (RV_UNDEF_ERROR);
    }
    /*  Create the offset control world canvas  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    /*  No histogram data yet  */
    win_scale.min_sat_pixel = gcvalues.background;
    win_scale.max_sat_pixel = gcvalues.foreground;
    win_scale.conv_type = CONV_CtoR_REAL;
    if ( ( wc = canvas_create (pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    top->dataclip.worldcanvas = wc;
    canvas_use_log_scale (wc, FALSE, TRUE);
    canvas_register_size_control_func (wc, histcanvas_size_control_func,
				       (void *) top);
    canvas_register_refresh_func (wc, histogram_canvas_refresh_func,
				  (void *) top);
    canvas_register_position_event_func (wc, histogram_canvas_position_func,
					 (void *) top);
    if (top->dataclip.array == NULL) hlabel = "Data Value";
    else hlabel = iarray_value_name (top->dataclip.array);
    canvas_set_dressing (wc,
			 KCD_ATT_DISPLAY, TRUE,
			 KCD_ATT_HLABEL, hlabel,
			 KCD_ATT_VLABEL, "Log(Incidence)",
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_END);
}   /*  End Function cnv_realise_cbk   */

static Boolean _Dataclip_SetValues (Widget Current, Widget Request, Widget New)
{
    DataclipWidget current = (DataclipWidget) Current;
    /*DataclipWidget request = (DataclipWidget) Request;*/
    DataclipWidget new = (DataclipWidget) New;
    static char function_name[] = "DataclipWidget::SetValues";

    if (new->dataclip.maxRegions != current->dataclip.maxRegions)
    {
	fprintf (stderr, "Cannot change number of regions\n");
	a_prog_bug (function_name);
    }
    return False;
}   /*  End Function SetValues  */

static void immediate_apply_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
{
    flag bool = (uaddr) call_data;
    DataclipWidget top = (DataclipWidget) client_data;

    top->dataclip.immediate_apply = bool;
}   /*  End Function immediate_apply_cbk  */

static void apply_cbk (Widget w, XtPointer client_data, XtPointer call_data)
{
    double scale, offset;
    DataclipWidget top = (DataclipWidget) client_data;
    DataclipRegions regions;
    char txt[STRING_LENGTH], val[STRING_LENGTH];

    if (top->dataclip.array == NULL) return;
    if (top->dataclip.num_regions < 1) return;
    regions.blank_data_outside_regions = top->dataclip.blank_data;
    regions.num_regions = top->dataclip.num_regions;
    if ( ( regions.minima = m_dup (top->dataclip.minima,
				   sizeof *top->dataclip.minima *
				   regions.num_regions) )
	== NULL ) return;
    if ( ( regions.maxima = m_dup (top->dataclip.maxima,
				   sizeof *top->dataclip.maxima *
				   regions.num_regions ) )
	== NULL )
    {
	m_free (regions.minima);
	return;
    }
    iarray_get_data_scaling (top->dataclip.array, &scale, &offset);
    iarray_format_value (top->dataclip.array, val,
			 regions.minima[0] * scale + offset, 1.0, 0.0);
    sprintf (txt, "Lower Clip: %s", val + 7);
    XtVaSetValues (top->dataclip.lower_label,
		   XtNlabel, txt,
		   NULL);
    iarray_format_value (top->dataclip.array, val,
			 regions.maxima[regions.num_regions - 1] * scale +
			 offset, 1.0, 0.0);
    sprintf (txt, "Upper Clip: %s", val + 7);
    XtVaSetValues (top->dataclip.upper_label,
		   XtNlabel, txt,
		   NULL);
    XtCallCallbacks ( (Widget) top, XkwNregionCallback, (XtPointer) &regions );
    m_free (regions.minima);
    m_free (regions.maxima);
}   /*  End Function apply_cbk  */

static void histcanvas_size_control_func (KWorldCanvas canvas,
					  int width, int height,
					  struct win_scale_type *win_scale,
					  void **info, flag *boundary_clear)
/*  [SUMMARY] Size control callback.
    [PURPOSE] This routine will modify the window scaling information for a
    world canvas. While this routine is running, colourmap resize events are
    ignored. Hence this routine may safely cause the associated Kcolourmap
    object to be resized.
    <canvas> The world canvas object.
    <width> The width of the canvas in pixels after removing the space consumed
    by the dressing.
    <height> The height of the canvas in pixels after removing the space
    consumed by the dressing.
    <win_scale> The window scaling information. This may be modified.
    <info> A pointer to the arbitrary canvas information pointer.
    <boundary_clear> If the value TRUE is written here then the
    [<canvas_resize>] routine will attempt to clear only the boundary between
    the pixel canvas and the world canvas. If the value FALSE is written here
    or nothing is written here, the [<canvas_resize>] routine will clear the
    entire pixel canvas as appropriate.
    [RETURNS] Nothing.
*/
{
    flag verbose;
    int array_len;
    unsigned int count;
    unsigned long hpeak;
    double frac;
    double min, max;
    DataclipWidget w = (DataclipWidget) *info;
    unsigned long *hist_array;
    static char function_name[]="DataclipWidget::histcanvas_size_control_func";

    win_scale->x_offset = 0;
    win_scale->y_offset = 0;
    win_scale->x_pixels = width;
    win_scale->y_pixels = height;
    verbose = w->dataclip.verbose;
    /*  Use real width, otherwise histogram will be computed multiple times  */
    kwin_get_size (canvas_get_pixcanvas (canvas), &array_len, NULL);
    if (w->dataclip.array == NULL) return;
    if ( DOUBLE_IS_BLANK (w->dataclip.data_min) )
    {
	/*  Recompute the minimum and maximum  */
	if ( !iarray_min_max (w->dataclip.array, CONV1_REAL, &min, &max) )
	{
	    fprintf (stderr, "Error getting image range\n");
	    return;
	}
	if (min >= max)
	{
	    w->dataclip.array = NULL;
	    return;
	}
	w->dataclip.data_min = min;
	w->dataclip.data_max = max;
	update_limits (w, TRUE);
    }
    if (w->dataclip.reset_canvas_scale)
    {
	win_scale->left_x = w->dataclip.data_min;
	win_scale->right_x = w->dataclip.data_max;
	w->dataclip.reset_canvas_scale = FALSE;
    }
    if (array_len == w->dataclip.hist_arr_length) return;
    if ( ( hist_array =
	   m_alloc_scratch (sizeof *hist_array * array_len, function_name) )
	 == NULL )
    {
	m_abort (function_name, "temporary array");
    }
    m_clear (hist_array, sizeof *hist_array * array_len);
    if (array_len > w->dataclip.hist_buf_length)
    {
	/*  Increase histogram array size  */
	if (w->dataclip.histogram_array != NULL)
	{
	    m_free (w->dataclip.histogram_array);
	    w->dataclip.hist_buf_length = 0;
	}
	if ( ( w->dataclip.histogram_array =
	       m_alloc (sizeof *w->dataclip.histogram_array * array_len) )
	     == NULL )
	{
	    m_abort (function_name, "histogram array");
	}
	w->dataclip.hist_buf_length = array_len;
    }
    w->dataclip.hist_arr_length = array_len;
    hpeak = 0;
    w->dataclip.index_of_peak = 0;
    if ( !iarray_compute_histogram (w->dataclip.array,
				    win_scale->conv_type,
				    win_scale->left_x, win_scale->right_x,
				    array_len, hist_array, &hpeak,
				    &w->dataclip.index_of_peak) )
    {
	fprintf (stderr, "Error computing histogram\n");
	a_prog_bug (function_name);
    }
    if (verbose) fprintf (stderr, "histogram peak: %lu\n", hpeak);
    for (count = 0; count < array_len; ++count)
	w->dataclip.histogram_array[count] = hist_array[count];
    m_free_scratch ();
    /*  Have histogram data: setup logarithmic scale  */
    win_scale->top_y = hpeak;
    if (hpeak > 1)
    {
	frac = log10 (win_scale->top_y);
	if (frac > 6.0) frac = 6.0;
	else if (frac < 1.0) frac = 1.0;
	win_scale->bottom_y = win_scale->top_y / pow (10.0, frac);
	if (win_scale->bottom_y < 1.0) win_scale->bottom_y = 1.0;
    }
    else win_scale->bottom_y = 0.1;
}   /*  End Function histcanvas_size_control_func  */

static void histogram_canvas_refresh_func (KWorldCanvas canvas,
					   int width, int height,
					   struct win_scale_type *win_scale,
					   Kcolourmap cmap, flag cmap_resize,
					   void **info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
{
    KPixCanvas pixcanvas;
    int count;
    unsigned long pixel_value;
    DataclipWidget w = (DataclipWidget) *info;
    double x, y, x0, x1, left_x, right_x, bottom_y, top_y;
    /*static char function_name[] = "histogram_canvas_refresh_func";*/

    if (w->dataclip.array == NULL) return;
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_MAX_SAT_PIXEL, &pixel_value,
			   CANVAS_ATT_END);
    canvas_draw_lines_p (canvas, (double *) NULL,
			 w->dataclip.histogram_array,
			 w->dataclip.hist_arr_length,
			 pixel_value);
    pixel_value = get_colour (canvas, "yellow4");
    if ( DOUBLE_IS_VALID (w->dataclip.left_pos) )
    {
	x0 = w->dataclip.left_pos;
	canvas_draw_line_p (canvas, x0, bottom_y, x0, top_y, pixel_value);
    }
    if (w->dataclip.num_regions < 1) return;
    pixcanvas = canvas_get_pixcanvas (canvas);
    for (count = 0; count < w->dataclip.num_regions; ++count)
    {
	x0 = w->dataclip.minima[count];
	x1 = w->dataclip.maxima[count];
	/*  Draw the two vertical lines  */
	canvas_draw_line_p (canvas, x0, bottom_y, x0, top_y, pixel_value);
	canvas_draw_line_p (canvas, x1, bottom_y, x1, top_y, pixel_value);
	/*  Draw the diagonal lines  */
	canvas_draw_line_p (canvas, x0, bottom_y, x1, top_y, pixel_value);
	canvas_draw_line_p (canvas, x1, bottom_y, x0, top_y, pixel_value);
	/*  Draw the "width" ellipse  */
	x = (x0 + x1) * 0.5;
	y = log (top_y / bottom_y) * (x1 - x0) / (right_x - left_x);
	y = bottom_y * exp (y);
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1, &x, &y,
					   &x, &y);
	kwin_draw_ellipse (pixcanvas, x, y, 5.0, 5.0, pixel_value);
    }
}   /*  End Function histogram_canvas_refresh_func  */

static flag histogram_canvas_position_func (KWorldCanvas canvas,
					    double x, double y,
					    unsigned int event_code,
					    void *e_info, void **f_info,
					    double x_lin, double y_lin,
					    int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    flag apply = FALSE;
    unsigned int region, index;
    double scale, offset, range, x0, x1, left_x, right_x, bottom_y, top_y;
    DataclipWidget w = (DataclipWidget) *f_info;
    char txt[STRING_LENGTH], val[STRING_LENGTH];

    if (w->dataclip.array == NULL) return (FALSE);
    iarray_get_data_scaling (w->dataclip.array, &scale, &offset);
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    iarray_format_value (w->dataclip.array, val, x * scale + offset, 1.0, 0.0);
    sprintf (txt, "Bin value: %s", val + 7);
    XtVaSetValues (w->dataclip.cursor_x_label,
		   XtNlabel, txt,
		   NULL);
    index = rint ( (x - left_x) / (right_x - left_x) *
		   (double) (w->dataclip.hist_arr_length - 1) );
    sprintf (txt, "Number of values in bin: %g",
	     w->dataclip.histogram_array[index]);
    XtVaSetValues (w->dataclip.cursor_y_label,
		   XtNlabel, txt,
		   NULL);
    switch (event_code)
    {
      case K_CANVAS_EVENT_POINTER_MOVE:
	return (TRUE);
	/*break;*/
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
	if (w->dataclip.last_was_left)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	if ( (w->dataclip.maxRegions == 1) &&
	     (w->dataclip.num_regions == 1) )
	{
	    if (x >= w->dataclip.maxima[0])
	    {
		XBell (XtDisplay (w), 100);
		return (TRUE);
	    }
	    w->dataclip.minima[0] = x;
	    if (w->dataclip.immediate_apply) apply = TRUE;
	}
	else if (w->dataclip.num_regions >= w->dataclip.maxRegions)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	else
	{
	    w->dataclip.left_pos = x;
	    w->dataclip.last_was_left = TRUE;
	}
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
	/*  Search for region  */
	for (region = 0; region < w->dataclip.num_regions; ++region)
	{
	    if ( (x >= w->dataclip.minima[region]) &&
		 (x <= w->dataclip.maxima[region]) ) break;
	}
	if (w->dataclip.num_regions == 1) region = 0;
	if (region >= w->dataclip.num_regions)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	range = log (y / bottom_y) / log (top_y / bottom_y);
	range = (right_x - left_x) * range;
	x0 = x - range * 0.5;
	x1 = x + range * 0.5;
	if ( (x0 < left_x) || (x1 > right_x) || (x0 >= x1) )
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	w->dataclip.minima[region] = x0;
	w->dataclip.maxima[region] = x1;
	if (w->dataclip.immediate_apply) apply = TRUE;
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
	if ( (w->dataclip.maxRegions == 1) &&
	     (w->dataclip.num_regions == 1) )
	{
	    if (x <= w->dataclip.minima[0])
	    {
		XBell (XtDisplay (w), 100);
		return (TRUE);
	    }
	    w->dataclip.maxima[0] = x;
	}
	else if (!w->dataclip.last_was_left)
	{
	    if (w->dataclip.num_regions > 0)
	    {
		XBell (XtDisplay (w), 100);
		return (TRUE);
	    }
	    w->dataclip.minima[0] = w->dataclip.data_min;
	    w->dataclip.maxima[0] = x;
	    w->dataclip.num_regions = 1;
	}
	else if (w->dataclip.num_regions >= w->dataclip.maxRegions)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	else if (x <= w->dataclip.left_pos)
	{
	    XBell (XtDisplay (w), 100);
	    return (TRUE);
	}
	else
	{
	    w->dataclip.minima[w->dataclip.num_regions] = w->dataclip.left_pos;
	    w->dataclip.maxima[w->dataclip.num_regions++] = x;
	    w->dataclip.last_was_left = FALSE;
	    w->dataclip.left_pos = TOOBIG;
	}
	if (w->dataclip.immediate_apply) apply = TRUE;
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    canvas_resize (canvas, NULL, TRUE);
    if (apply)
    {
	XSync (XtDisplay (w), False);
	apply_cbk ( (Widget) NULL, (XtPointer) w, NULL );
    }
    return (TRUE);
}   /*  End Function histogram_canvas_position_func  */

static void pop_cbk (w, client_data, call_data)
/*  This is the generic popup/popdown callback.
*/
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    DataclipWidget top = (DataclipWidget) w;

    top->dataclip.popped_up = (iaddr) client_data;
}   /*  End Function pop_cbk   */

static void full_range_cbk (w, client_data, call_data)
/*  This is the generic popup/popdown callback.
*/
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    DataclipWidget top = (DataclipWidget) client_data;

    top->dataclip.last_was_left = FALSE;
    top->dataclip.left_pos = TOOBIG;
    if (top->dataclip.array == NULL) return;
    top->dataclip.num_regions = 1;
    top->dataclip.minima[0] = top->dataclip.data_min;
    top->dataclip.maxima[0] = top->dataclip.data_max;
    unzoom_cbk (NULL, top, NULL);
    XSync (XtDisplay (w), False);
    apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
}   /*  End Function full_range_cbk   */

static void iarr_destroy_callback (iarray arr, DataclipWidget top)
/*  [PURPOSE] This routine will register the destruction of the Intelligent
    Array.
    <arr> The Intelligent Array.
    <top> The Dataclip widget.
    [RETURNS] Nothing.
*/
{
    top->dataclip.array = NULL;
    top->dataclip.iarr_destroy_callback = NULL;
    if (top->dataclip.autoPopdown) XtPopdown ( (Widget) top );
}   /*  End Function iarr_destroy_callback  */

static void blank_data_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
{
    flag bool = (uaddr) call_data;
    DataclipWidget top = (DataclipWidget) client_data;

    top->dataclip.blank_data = bool;
    apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
}   /*  End Function blank_data_cbk  */

static void type_menu_cbk (Widget w, XtPointer client_data,XtPointer call_data)
{
    unsigned int type;
    unsigned int index = *(int *) call_data;
    DataclipWidget top = (DataclipWidget) client_data;

    type = dataclipClassRec.dataclip_class.type_index_to_type[index];
    top->dataclip.output_type = type;
}   /*  End Function type_menu_cbk  */

static void save_cbk (Widget w, XtPointer client_data, XtPointer call_data)
{
    iarray new, old;
    flag rescale = TRUE;
    unsigned int count, num_dim;
    Widget dialog;
    DataclipWidget top = (DataclipWidget) client_data;
    double inp_scale, inp_offset, min, max, scale, offset;
    char *arrayfile_name;
    unsigned long *dim_lengths;
    CONST char **dim_names;
    static char function_name[] = "DataclipWidget::save_cbk";

    dialog = top->dataclip.dialog;
    if (top->dataclip.num_regions < 1)
    {
	XBell (XtDisplay (w), 100);
	return;
    }
    if (top->dataclip.num_regions > 1)
    {
	fprintf (stderr, "Multiple regions not supported yet\n");
	return;
    }
    arrayfile_name = XawDialogGetValueString (dialog);
    if (arrayfile_name == NULL) arrayfile_name = "scaled";
    if (strlen (arrayfile_name) < 1) arrayfile_name = "scaled";
    old = top->dataclip.array;
    num_dim = iarray_num_dim (old);
    if ( ( dim_names = m_alloc (num_dim * sizeof *dim_names) ) == NULL )
    {
	m_abort (function_name, "array of dimension names pointers");
    }
    if ( ( dim_lengths = m_alloc (num_dim * sizeof *dim_lengths) ) == NULL )
    {
	m_abort (function_name, "array of dimension lengths");
    }
    for (count = 0; count < num_dim; ++count)
    {
	dim_names[count] = iarray_dim_name (old, count);
	dim_lengths[count] = iarray_dim_length (old, count);
    }
    if ( ( new = iarray_create (top->dataclip.output_type, num_dim, dim_names,
				dim_lengths, iarray_value_name (old), old) )
	 == NULL )
    {
	m_error_notify (function_name, "output Intelligent Array");
	m_free (dim_names);
	m_free (dim_lengths);
	return;
    }
    m_free (dim_names);
    m_free (dim_lengths);
    for (count = 0; count < num_dim; ++count)
    {
	iarray_get_world_coords (old, count, &min, &max);
	iarray_set_world_coords (new, count, min, max);
    }
    iarray_get_data_scaling (old, &inp_scale, &inp_offset);
    iarray_set_data_range (new,
			   top->dataclip.minima[0] * inp_scale + inp_offset,
			   top->dataclip.maxima[0] * inp_scale + inp_offset);
    /*  Determine desired new minimum and maximum range  */
    switch (top->dataclip.output_type)
    {
      case K_BYTE:
	min = -127.0;
	max = 127.0;
	break;
      case K_UBYTE:
	min = 0.0;
	max = 255.0;
	break;
      case K_SHORT:
	min = -32767.0;
	max = 32767.0;
	break;
      case K_USHORT:
	min = 0.0;
	max = 65535.0;
	break;
      case K_INT:
	min = -2147483647.0;
	max = 2147483647.0;
	break;
      case K_UINT:
	min = 0.0;
	max = 4294967296.0;
	break;
      default:
	rescale = FALSE;
	break;
    }
    if (rescale)
    {
	/*  Must fit the defined range of data into a range  */
	scale = (max - min) /
	    (top->dataclip.maxima[0] - top->dataclip.minima[0]);
	offset = min - scale * top->dataclip.minima[0];
	iarray_set_data_scaling (new, inp_scale / scale,
				 inp_offset - offset * inp_scale / scale);
    }
    else
    {
	scale = inp_scale;
	offset = inp_offset;
	iarray_set_data_scaling (new, 1.0, 0.0);
    }
    iarray_clip_scale_and_offset (new, old, scale, offset,
				  top->dataclip.minima[0],
				  top->dataclip.maxima[0],
				  top->dataclip.blank_data);
    if ( !iarray_write (new, arrayfile_name) )
    {
	iarray_dealloc (new);
	return;
    }
    iarray_dealloc (new);
    XBell (XtDisplay (w), -90);
}   /*  End Function save_cbk  */

static unsigned long get_colour (KWorldCanvas canvas, CONST char *colourname)
/*  [SUMMARY] Get colour.
    <canvas> The world canvas.
    <colourname> The name of the colour.
    [RETURNS] The specified colour if available, else a default colour.
*/
{
    KPixCanvas pixcanvas;
    unsigned long pixel_value;

    pixcanvas = canvas_get_pixcanvas (canvas);
    if ( kwin_get_colour (pixcanvas, colourname, &pixel_value,
			  NULL, NULL, NULL) )
    {
	return (pixel_value);
    }
    if ( kwin_get_colour (pixcanvas, "white", &pixel_value, NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "Error allocating colour: \"%s\" defaulting to white\n",
		 colourname);
	return (pixel_value);
    }
    fprintf (stderr, "Error allocating colour: \"%s\" defaulting to 0\n",
	     colourname);
    return (0);
}   /*  End Function get_colour  */

static void zoom_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Zoom callback.
    <w> The widget.
    <client_data> Client data pointer.
    <call_data> Call-specific data pointer.
    [RETURNS] Nothing.
*/
{
    double min, max, left, right;
    DataclipWidget top = (DataclipWidget) client_data;

    if (top->dataclip.array == NULL) return;
    if ( DOUBLE_IS_BLANK (top->dataclip.data_min) ) return;
    if (top->dataclip.num_regions < 1) return;
    canvas_get_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, &left,
			   CANVAS_ATT_RIGHT_X, &right,
			   CANVAS_ATT_END);
    min = top->dataclip.minima[0];
    max = top->dataclip.maxima[top->dataclip.num_regions - 1];
    if ( (min == left) && (max == right) ) return;
    canvas_set_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, min,
			   CANVAS_ATT_RIGHT_X, max,
			   CANVAS_ATT_END);
    /*  Force histogram to be recomputed  */
    top->dataclip.hist_arr_length = 0;
    kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
}   /*  End Function zoom_cbk   */

static void unzoom_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Unzoom callback.
    <w> The widget.
    <client_data> Client data pointer.
    <call_data> Call-specific data pointer.
    [RETURNS] Nothing.
*/
{
    double min, max, left, right;
    DataclipWidget top = (DataclipWidget) client_data;

    if (top->dataclip.array == NULL) return;
    if ( DOUBLE_IS_BLANK (top->dataclip.data_min) ) return;
    canvas_get_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, &left,
			   CANVAS_ATT_RIGHT_X, &right,
			   CANVAS_ATT_END);
    min = top->dataclip.data_min;
    max = top->dataclip.data_max;
    if ( (min == left) && (max == right) )
    {
	/*  Refresh in case called from <full_range_cbk>  */
	kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
	return;
    }
    canvas_set_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, min,
			   CANVAS_ATT_RIGHT_X, max,
			   CANVAS_ATT_END);
    /*  Force histogram to be recomputed  */
    top->dataclip.hist_arr_length = 0;
    kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
}   /*  End Function unzoom_cbk   */

static void update_limits (DataclipWidget w, flag reset_hscale)
/*  [SUMMARY] Update the displayed data range.
    <w> The Dataclip widget.
    <reset_hscale> If TRUE the canvas horizontal scale is reset.
    [RETURNS] Nothing.
*/
{
    double scale, offset;
    char txt[STRING_LENGTH], val[STRING_LENGTH];

    if (w->dataclip.array == NULL) return;
    iarray_get_data_scaling (w->dataclip.array, &scale, &offset);
    iarray_format_value (w->dataclip.array, val,
			 w->dataclip.data_min * scale + offset, 1.0, 0.0);
    sprintf (txt, "Data Min: %s", val + 7);
    XtVaSetValues (w->dataclip.min_label,
		   XtNlabel, txt,
		   NULL);
    iarray_format_value (w->dataclip.array, val,
			 w->dataclip.data_max *scale + offset, 1.0, 0.0);
    sprintf (txt, "Data Max: %s", val + 7);
    XtVaSetValues (w->dataclip.max_label,
		   XtNlabel, txt,
		   NULL);
    w->dataclip.reset_canvas_scale = reset_hscale;
}   /*  End Function update_limits  */

static void clip_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Clip callback.
    <w> The widget.
    <client_data> Client data pointer.
    <call_data> Call-specific data pointer.
    [RETURNS] Nothing.
*/
{
    double proportion;
    DataclipWidget top = (DataclipWidget) client_data;
    String label;
    CONST char *p;
    char txt[STRING_LENGTH];

    if (top->dataclip.array == NULL) return;
    /*  Find how much to clip  */
    XtVaGetValues (w, XtNlabel, &label, NULL);
    if (strcmp (label, "Zero Left") == 0)
    {
	double scale, offset, val;

	iarray_get_data_scaling (top->dataclip.array, &scale, &offset);
	val = -offset / scale;
	if (val < top->dataclip.data_min) return;
	top->dataclip.minima[0] = val;
	if (top->dataclip.num_regions < 1)
	{
	    top->dataclip.maxima[0] = top->dataclip.data_max;
	    top->dataclip.num_regions = 1;
	}
	kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
	if (top->dataclip.immediate_apply)
	{
	    XSync (XtDisplay (top), False);
	    apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
	}
	return;
    }
    strcpy (txt, label);
    txt[strlen (txt) - 1] = '\0';
    proportion = ex_float (txt, &p) / 100.0;
    clip_outliers (top, proportion);
}   /*  End Function clip_cbk   */

static void clip_outliers (DataclipWidget top, double proportion)
/*  [SUMMARY] Clip outliers.
    <top> The Dataclip widget.
    <proportion> The proportion of data to clip to.
    [RETURNS] Nothing.
*/
{
    unsigned int length;
    unsigned long index;
    double total, sum, left, right, factor;

    if ( DOUBLE_IS_BLANK (top->dataclip.data_min) ) return;
    length = top->dataclip.hist_arr_length;
    if (length < 2) return;
    canvas_get_attributes (top->dataclip.worldcanvas,
			   CANVAS_ATT_LEFT_X, &left,
			   CANVAS_ATT_RIGHT_X, &right,
			   CANVAS_ATT_END);
    factor = (right - left) / (length - 1);
    /*  Find cutoff points which emcompass most of the data values. Search from
	the peak, as that is optimal for distributions with a long tail, which
	is why the user is using this option in the first place  */
    /*  First the lower cutoff  */
    for (index = top->dataclip.index_of_peak, total = 0.0; index > 0; --index)
	total += top->dataclip.histogram_array[index];
    total *= proportion;
    for (index = top->dataclip.index_of_peak, sum = 0.0;
	 (index > 0) && (sum < total);
	 --index)
	sum += top->dataclip.histogram_array[index];
    top->dataclip.minima[0] = left + index * factor;
    /*  Next the upper cutoff  */
    for (index = top->dataclip.index_of_peak, total = 0.0; index < length;
	 ++index)
    {
	total += top->dataclip.histogram_array[index];
    }
    total *= proportion;
    for (index = top->dataclip.index_of_peak, sum = 0.0;
	 (index < length) && (sum < total);
	 ++index)
	sum += top->dataclip.histogram_array[index];
    top->dataclip.maxima[0] = left + index * factor;
    top->dataclip.num_regions = 1;
    kwin_refresh_if_visible (top->dataclip.pixcanvas, TRUE);
    if (top->dataclip.immediate_apply)
    {
	XSync (XtDisplay (top), False);
	apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
    }
}   /*  End Function clip_outliers   */

static flag _Dataclip_ds_event_func (void *object, multi_array *multi_desc,
				      CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    double min, max;
    DataclipWidget top = (DataclipWidget) object;
    packet_desc *pack_desc;
    char *packet, *d_title;
    double value[2];

    if (strcmp (name, "DataclipWidget::Clip") != 0) return (FALSE);
    if (top->dataclip.array == NULL) return (FALSE);
    if (multi_desc->num_arrays != 1) return (FALSE);
    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0];
    if ( ( d_title = ds_get_unique_named_string (pack_desc, packet,
						 "title") ) == NULL )
	return (FALSE);
    if (strcmp (top->wm.title, d_title) != 0)
    {
	m_free (d_title);
	return (FALSE);
    }
    m_free (d_title);
    if ( ds_get_unique_named_value (pack_desc, packet, "hist_clip", NULL,
				    value) )
    {
	clip_outliers (top, value[0]);
	return (TRUE);
    }
    if ( !ds_get_unique_named_value (pack_desc, packet, "minimum", NULL,
				     value) ) return (FALSE);
    min = value[0];
    if ( !ds_get_unique_named_value (pack_desc, packet, "maximum", NULL,
				     value) ) return (FALSE);
    max = value[0];
    if (min < top->dataclip.data_min) min = top->dataclip.data_min;
    if (max > top->dataclip.data_max) max = top->dataclip.data_max;
    top->dataclip.num_regions = 1;
    top->dataclip.minima[0] = min;
    top->dataclip.maxima[0] = max;
    if (top->dataclip.worldcanvas != NULL)
	canvas_resize (top->dataclip.worldcanvas, NULL, TRUE);
    apply_cbk ( (Widget) NULL, (XtPointer) top, NULL );
    return (TRUE);
}   /*  End Function _Dataclip_ds_event_func  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwDataclipNewArray (Widget W, iarray array, double min, double max,
			  flag reset_hscale)
/*  [SUMMARY] Register new array.
    <W> The Dataclip widget.
    <array> The new array. This may be NULL.
    <min> The minimum data value in the array.
    <max> The maximum data value in the array. If this is less than <<min>>
    then the minimum and maximum values are computed.
    <reset_hscale> If TRUE the horizontal scale is reset.
    [RETURNS] Nothing.
*/
{
    DataclipWidget w = (DataclipWidget) W;
    static char function_name[] = "XkwDataclipNewArray";

    FLAG_VERIFY (reset_hscale);
    /*  Clean up any old array  */
    if (w->dataclip.iarr_destroy_callback != NULL)
    {
	c_unregister_callback (w->dataclip.iarr_destroy_callback);
	w->dataclip.iarr_destroy_callback = NULL;
    }
    w->dataclip.last_was_left = FALSE;
    w->dataclip.left_pos = TOOBIG;
    w->dataclip.data_min = TOOBIG;
    w->dataclip.data_max = -TOOBIG;
    w->dataclip.hist_arr_length = 0;
    w->dataclip.array = array;
    if (max >= min)
    {
	w->dataclip.data_min = min;
	w->dataclip.data_max = max;
	update_limits (w, reset_hscale);
    }
    if (array != NULL)
    {
	/*  Process new array  */
	w->dataclip.iarr_destroy_callback = 
	    iarray_register_destroy_func(w->dataclip.array,
					 ( flag (*) () ) iarr_destroy_callback,
					 w);
	if (w->dataclip.worldcanvas != NULL)
	{
	    canvas_set_dressing (w->dataclip.worldcanvas,
				 KCD_ATT_HLABEL, iarray_value_name (array),
				 KCD_ATT_END);
	}
    }
    /*  Force computation and display of histogram if new data available,
	else clear canvas  */
    if (w->dataclip.popped_up)
    {
	kwin_refresh_if_visible (w->dataclip.pixcanvas, TRUE);
    }
}   /*  End Function XkwDataclipNewArray  */

/*EXPERIMENTAL_FUNCTION*/
void XkwDataclipArrayChanged (Widget W, double min, double max)
/*  [SUMMARY] Register a change in the array.
    <W> The Dataclip widget.
    <min> The minimum data value in the array.
    <max> The maximum data value in the array. If this is less than <<min>>
    then the minimum and maximum values are computed.
    [RETURNS] Nothing.
*/
{
    DataclipWidget w = (DataclipWidget) W;
    /*static char function_name[] = "XkwDataclipArrayChanged";*/

    w->dataclip.data_min = TOOBIG;
    w->dataclip.data_max = -TOOBIG;
    w->dataclip.hist_arr_length = 0;
    if (max >= min)
    {
	w->dataclip.data_min = min;
	w->dataclip.data_max = max;
	update_limits (w, TRUE);
    }
    /*  Force computation and display of histogram if new data available,
	else clear canvas  */
    if (w->dataclip.popped_up)
    {
	kwin_refresh_if_visible (w->dataclip.pixcanvas, TRUE);
    }
}   /*  End Function XkwDataclipArrayChanged  */
