/*LINTLIBRARY*/
/*  ExclusiveMenu.c

    This code provides a simple exclusive menu widget for Xt.

    Copyright (C) 1994-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a simple exclusive menu widget
  for Xt.


    Written by      Richard Gooch   10-DEC-1994

    Updated by      Richard Gooch   10-DEC-1994

    Updated by      Richard Gooch   12-DEC-1994: Fixed bug in creation of
  simple menu widget: use XtVaCreatePopupShell.

    Updated by      Richard Gooch   14-DEC-1994: Removed unneccessary call to
  XtParent in Initialise.

    Updated by      Richard Gooch   4-JAN-1995: Cosmetic changes.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   31-JUL-1996: Added valuePtr resource.

    Updated by      Richard Gooch   31-JAN-1997: Added setChoice resource.

    Updated by      Richard Gooch   31-JAN-1997: Added callCallbacksOnSet
  resource.

    Updated by      Richard Gooch   13-MAY-1997: Allocate fresh copy of
  choiceName string.

    Updated by      Richard Gooch   4-JUN-1997: Added small leftBitmap.

    Updated by      Richard Gooch   26-OCT-1998: Support setting choice at
  initialisation.

    Updated by      Richard Gooch   15-JUN-2000: Write to value pointer at
  initialisation.

    Updated by      Richard Gooch   21-MAR-2001: Created Destroy method.

    Last updated by Richard Gooch   25-SEP-2001: Removed #include for var-args.


*/
#include <Xkw/ExclusiveMenuP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <karma.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>

struct item_data_type
{
    Widget me;
    int index;
};


STATIC_FUNCTION (void _ExclusiveMenu_Initialise,(Widget request, Widget new) );
STATIC_FUNCTION (void _ExclusiveMenu_Destroy, (Widget w) );
STATIC_FUNCTION (Boolean _ExclusiveMenu_SetValues,
		 (Widget current, Widget request, Widget new) );
STATIC_FUNCTION (void item_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void set_item,
		 (ExclusiveMenuWidget top, unsigned int index, flag notify,
		  int width) );
STATIC_FUNCTION (int _ExclusiveMenu_find_choice,
		 (ExclusiveMenuWidget w, CONST char *name) );
STATIC_FUNCTION (void _ExclusiveMenu_set_choice,
		 (ExclusiveMenuWidget w, flag notify, int width) );


#define offset(field) XtOffsetOf(ExclusiveMenuRec, exclusiveMenu.field)

static XtResource ExclusiveMenuResources[] = 
{
    {XkwNchoiceName, XkwCChoiceName, XtRString, sizeof (String),
     offset(choiceName), XtRImmediate, NULL},
    {XkwNitemStrings, XkwCItemStrings, XtRPointer, sizeof (XtPointer),
     offset(itemStrings), XtRImmediate, NULL},
    {XkwNnumItems, XkwCNumItems, XtRInt, sizeof (int),
     offset(numItems), XtRImmediate, 0},
    {XkwNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset(selectCallback), XtRCallback, (caddr_t) NULL},
    {XkwNvaluePtr, XkwCValuePtr, XtRInt, sizeof (XtPointer),
     offset (valuePtr), XtRImmediate, (XtPointer) NULL},
    {XkwNsetChoice, XkwCSetChoice, XtRString, sizeof (String),
     offset (setChoice), XtRImmediate, NULL},
    {XkwNcallCallbacksOnSet, XkwCCallCallbacksOnSet, XtRBool, sizeof (Bool),
     offset (callCallbacksOnSet), XtRImmediate, False},
};

#undef offset

#define bitmap_width 10
#define bitmap_height 10
static unsigned char bitmap_bits[] = {
   0xff, 0x03, 0x01, 0x02, 0x03, 0x03, 0x02, 0x01, 0x86, 0x01, 0x84, 0x00,
   0xcc, 0x00, 0x48, 0x00, 0x48, 0x00, 0x30, 0x00};

ExclusiveMenuClassRec exclusiveMenuClassRec = 
{
  {        /* CoreClassPart */
      (WidgetClass) &menuButtonClassRec,   /* superclass */
      "ExclusiveMenu",                     /* class_name */
      sizeof(ExclusiveMenuRec),            /* widget_size */
      NULL,                                /* class_initialise */
      NULL,                                /* class_part_initialise */
      FALSE,                               /* class_init */
      (XtInitProc)_ExclusiveMenu_Initialise,/* initialise */
      NULL,                                /* initialise_hook */
      XtInheritRealize,                    /* realise */
      NULL,                                /* actions */
      0,                                   /* num_actions */
      ExclusiveMenuResources,              /* resources */
      XtNumber(ExclusiveMenuResources),    /* num_resources */
      NULLQUARK,                           /* xrm_class */
      TRUE,                                /* compress_motion */
      TRUE,                                /* compress_exposure */
      TRUE,                                /* compress_enterleave */
      TRUE,                                /* visible_interest */
      _ExclusiveMenu_Destroy,              /* destroy */
      XtInheritResize,                     /* resize */
      XtInheritExpose,                     /* expose */
      (XtSetValuesFunc)_ExclusiveMenu_SetValues,/* set_values */
      NULL,                                /* set_values_hook */
      XtInheritSetValuesAlmost,            /* set_values_almost */
      NULL,                                /* get_values_hook */
      NULL,                                /* accept_focus */
      XtVersion,                           /* version */
      NULL,                                /* callback_private */
      XtInheritTranslations,               /* tm_translations */
      NULL,
      NULL,
      NULL,
  },
  {  /* SimpleClassPart */
    XtInheritChangeSensitive /* change_sensitive */
  },
  {  /* LabelClassPart */
    0 /* empty */
  },
  {  /* CommandClassPart */
    0 /* empty */
  },
  {  /* MenuButtonClassPart */
    0 /* empty */
  },
  {  /* ExclusiveMenuClassPart */
    0 /* empty */
  },
};

WidgetClass exclusiveMenuWidgetClass = (WidgetClass) &exclusiveMenuClassRec;

static void _ExclusiveMenu_Initialise (Widget Request, Widget New)
{
    int count, width, tmp;
    /*ExclusiveMenuWidget request = (ExclusiveMenuWidget) Request;*/
    ExclusiveMenuWidget new = (ExclusiveMenuWidget) New;
    Pixmap pixmap;
    Widget menu, menu_entry;
    char **names;
    struct item_data_type *item_data;
    Display *dpy;
    Screen *screen;
    static char function_name[] = "ExclusiveMenu::Initialise";

    menu = XtVaCreatePopupShell (new->menuButton.menu_name,
				 simpleMenuWidgetClass, New,
				 NULL);
    names = (char **) new->exclusiveMenu.itemStrings;
    if ( ( item_data =
	   m_alloc (sizeof *item_data * new->exclusiveMenu.numItems) )
	 == NULL )
    {
	m_abort (function_name, "item data");
    }
    new->exclusiveMenu.item_data = item_data;
    width = 0;
    for (count = 0; count < new->exclusiveMenu.numItems; ++count)
    {
	menu_entry = XtVaCreateManagedWidget (names[count], smeBSBObjectClass,
					      menu,
					      XtNlabel, names[count],
					      NULL);
	item_data[count].me = New;
	item_data[count].index = count;
	XtAddCallback ( menu_entry, XtNcallback, item_cbk,
		       (XtPointer) (item_data + count) );
	tmp = strlen (names[count]);
	if (tmp > width) width = tmp;
    }
    if (new->exclusiveMenu.setChoice == NULL)
    {
	new->exclusiveMenu.setChoice = names[0];
	if (new->exclusiveMenu.valuePtr != NULL)
	    *new->exclusiveMenu.valuePtr = 0;
    }
    _ExclusiveMenu_set_choice (new, FALSE, width);
    /*  Change the bitmap after because it uses SetValues which must come after
	_ExclusiveMenu_set_choice  */
    dpy = XtDisplay (New);
    screen = XtScreen (New);
    /*  Create bitmap  */
    if ( ( pixmap = XCreateBitmapFromData (dpy, RootWindowOfScreen (screen),
					   (char *) bitmap_bits, bitmap_width,
					   bitmap_height) ) == 0 )
    {
	m_abort (function_name, "bitmap");
    }
    XtVaSetValues (New,
		   XtNleftBitmap, pixmap,
		   NULL);
    if (new->exclusiveMenu.choiceName != NULL)
    {
	if ( ( new->exclusiveMenu.choiceName =
	       st_dup (new->exclusiveMenu.choiceName) ) == NULL )
	{
	    m_abort (function_name, "choiceName");
	}
    }
}   /*  End Function Initialise  */

static void _ExclusiveMenu_Destroy (Widget W)
{
    ExclusiveMenuWidget w = (ExclusiveMenuWidget) W;

    if (w->exclusiveMenu.choiceName != NULL)
	m_free (w->exclusiveMenu.choiceName);
    m_free (w->exclusiveMenu.item_data);
}   /*  End Function Destroy  */

static Boolean _ExclusiveMenu_SetValues (Widget Current, Widget Request,
					 Widget New)
{
    flag notify;
    ExclusiveMenuWidget current = (ExclusiveMenuWidget) Current;
    /*ExclusiveMenuWidget request = (ExclusiveMenuWidget) Request;*/
    ExclusiveMenuWidget new = (ExclusiveMenuWidget) New;
    static char function_name[] = "ExclusiveMenu::SetValues";

    if (new->exclusiveMenu.choiceName != current->exclusiveMenu.choiceName)
    {
	if (current->exclusiveMenu.choiceName != NULL)
	    m_free (current->exclusiveMenu.choiceName);
	if ( ( new->exclusiveMenu.choiceName =
	       st_dup (new->exclusiveMenu.choiceName) ) == NULL )
	{
	    m_abort (function_name, "choiceName");
	}
    }
    notify = new->exclusiveMenu.callCallbacksOnSet;
    new->exclusiveMenu.callCallbacksOnSet = FALSE;
    _ExclusiveMenu_set_choice (new, notify, 0);
    return True;
}   /*  End Function SetValues  */

static void item_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the item callback.
*/
{
    ExclusiveMenuWidget top;
    struct item_data_type *item_data = (struct item_data_type *) client_data;

    top = (ExclusiveMenuWidget) item_data->me;
    set_item (top, item_data->index, TRUE, 0);
}   /*  End Function item_cbk  */

static void set_item (ExclusiveMenuWidget top, unsigned int index, flag notify,
		      int width)
/*  [SUMMARY] Set an item.
    <top> The ExclusiveMenu widget.
    <index> The index number of the item.
    <notify> If TRUE, callbacks will be called.
    <width> The size to pad to.
    [RETURNS] Nothing.
*/
{
    int val;
    char **names;
    char txt[STRING_LENGTH];

    if (top->exclusiveMenu.valuePtr != NULL)
	*top->exclusiveMenu.valuePtr = index;
    names = (char **) top->exclusiveMenu.itemStrings;
    sprintf (txt, "%s: %s", top->exclusiveMenu.choiceName, names[index]);
    for (width -= strlen (names[index]); width > 0; --width) strcat (txt, " ");
    XtVaSetValues ( (Widget) top,
		   XtNlabel, txt,
		   NULL );
    val = index;
    if (notify) XtCallCallbacks ( (Widget) top, XkwNselectCallback,
				  (XtPointer) &val );
}   /*  End Function set_item  */

static int _ExclusiveMenu_find_choice (ExclusiveMenuWidget w, CONST char *name)
/*  [SUMMARY] Find the index of a choice name.
    <w> The ExclusiveMenu widget.
    <name> The choice name.
    [RETURNS] The index of the choice name on success, -1 on failure.
*/
{
    unsigned int count;
    CONST char **names;

    names = (CONST char **) w->exclusiveMenu.itemStrings;
    for (count = 0; count < w->exclusiveMenu.numItems; ++count)
    {
	if (strcmp (name, names[count]) == 0) return (count);
    }
    return (-1);
}   /*  End Function _ExclusiveMenu_find_choice   */

static void _ExclusiveMenu_set_choice (ExclusiveMenuWidget w, flag notify,
				       int width)
/*  [SUMMARY] Set the choice.
    <w> The ExclusiveMenu widget.
    <notify> If TRUE, callbacks are called.
    <width> The size to pad to.
    [RETURNS] Nothing.
*/
{
    int index;
    static char function_name[] = "_ExclusiveMenu_set_choice";

    if (w->exclusiveMenu.setChoice == NULL) return;
    index = _ExclusiveMenu_find_choice (w, w->exclusiveMenu.setChoice);
    if (index < 0)
    {
	fprintf (stderr, "Unknown choice name: \"%s\"\n",
		 w->exclusiveMenu.setChoice);
	a_prog_bug (function_name);
    }
    w->exclusiveMenu.setChoice = NULL;
    set_item (w, index, notify, width);
}   /*  End Function _ExclusiveMenu_set_choice  */
