/*LINTLIBRARY*/
/*  ExportMenu.c

    This code provides a simple export menu widget for Xt.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a simple export menu widget for
  Xt.


    Written by      Richard Gooch   1-OCT-1996

    Updated by      Richard Gooch   22-OCT-1996: Set dialog widget label and
  increased dialog widget size.

    Updated by      Richard Gooch   27-NOV-1996: Fixed bug where PseudoColour
  colourmap was being used when writing TrueColour images in PPM or Sun
  Rasterfile formats.

    Updated by      Richard Gooch   16-FEB-1997: Added supportSubsets resource
  and now use iarray resource. Added label resource.

    Updated by      Richard Gooch   5-AUG-1997: Fixed bug in getting colourmap
  colours. Implemented PPM movie write.

    Updated by      Richard Gooch   1-SEP-1997: Created <fiddle_fits> to cope
  gracefully with creating subsets from FITS files missing "CRPIXn" keywords.

    Updated by      Richard Gooch   17-SEP-1997: Fixed bug in <fiddle_fits>:
  failed on successful update of "CRPIXn".

    Updated by      Richard Gooch   5-MAY-1998: Took account of changes to
  <foreign_ppm_write_pseudo> and <foreign_ppm_write_rgb>.

    Updated by      Richard Gooch   11-JUL-1998: Improved diagnostics when
  writing PPM files.

    Updated by      Richard Gooch   17-JUL-1998: Compute min&max in function
  <create_sub> and attach to header.

    Updated by      Richard Gooch   4-SEP-1998: Support exporting Karma & etc.
  data formats when no iarray available.

    Updated by      Richard Gooch   17-SEP-1998: Changed menu name from
  "theMenu" to "exportMenu".

    Updated by      Richard Gooch   15-JUN-1999: Pass intensity scaling
  function to <foreign_ppm_write_pseudo>.

    Updated by      Richard Gooch   17-MAY-2000: Support PPM Window output.

    Updated by      Richard Gooch   1-JUN-2000: Fixed bug in PPM Window output.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Updated by      Richard Gooch   28-JUN-2000: Fixed PPM Window output: was
  not initialising buffer to background colour.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kcmap.h>

    Updated by      Richard Gooch   27-MAY-2001: Created Destroy method. Added
  event "ExportMenuWidget::WritePPMImageWindow".

    Updated by      Richard Gooch   28-MAY-2001: Created PPMWindowMovieCallback

    Last updated by Richard Gooch   25-SEP-2001: Removed #include for var-args.


*/
#include <Xkw/ExportMenuP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <karma.h>
#include <karma_viewimg.h>
#include <karma_foreign.h>
#include <karma_kcmap.h>
#include <karma_dsxfr.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_c.h>
#include <Xkw/ChoiceMenu.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/Postscript.h>


#define AUTO_FREE(func,obj) \
	    c_register_callback (&clist, ( flag (*) () ) (func), \
				 (obj), NULL, FALSE, NULL, FALSE, \
				 FALSE)
#define RETURN {if ((clist) != NULL) \
                 {c_call_callbacks (clist, NULL); c_destroy_list (clist);} \
                 return;}


/*  Private functions  */

STATIC_FUNCTION (void _ExportMenu_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _ExportMenu_Destroy, (Widget W) );
STATIC_FUNCTION (Boolean _ExportMenu_SetValues,
		 (Widget current, Widget request, Widget new) );
STATIC_FUNCTION (void export_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void save_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void postscript_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (multi_array *create_sub,
		 (multi_array *old, flag *allocated,
		  array_desc *arr_desc, unsigned int hdim,
		  unsigned int vdim, KWorldCanvas canvas) );
STATIC_FUNCTION (array_desc *find_array,
		 (CONST multi_array *multi_desc, CONST char *name,
		  unsigned int *index) );
STATIC_FUNCTION (void fiddle_fits,
		 (packet_desc *pack_desc, char **packet, dim_desc *dim) );
STATIC_FUNCTION (flag copy_plane,
		 (CONST char *inp_data, CONST array_desc *inp_arr_desc,
		  unsigned int hdim, unsigned int vdim,
		  uaddr hstart, uaddr vstart,
		  char *out_data, CONST array_desc *out_arr_desc) );
STATIC_FUNCTION (void iarr_destroy_callback,
		 (iarray arr, ExportMenuWidget top) );
STATIC_FUNCTION (void write_ppm_data_movie,
		 (multi_array *multi_desc, flag truecolour,
		  array_desc *arr_desc, unsigned int hdim, unsigned int vdim,
		  KWorldCanvas canvas, ViewableImage vimage,
		  unsigned int cmap_size, CONST unsigned short *red,
		  CONST unsigned short *green, CONST unsigned short *blue,
		  CONST char *filename) );
STATIC_FUNCTION (flag _ExportMenu_ds_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


#define offset(field) XtOffset(ExportMenuWidget, exportMenu.field)

static XtResource ExportMenuResources[] = 
{
    {XkwNiarray, XkwCIarray, XtRPointer, sizeof (XtPointer),
     offset (array), XtRImmediate, NULL},
    {XkwNworldCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (worldCanvas), XtRImmediate, NULL},
    {XkwNsupportSubsets, XkwCSupportSubsets, XtRBool, sizeof (Bool),
     offset (supportSubsets), XtRImmediate, (XtPointer) True},
    {XtNlabel, XtCLabel, XtRString, sizeof (String),
     offset (label), XtRString, (XtPointer) "Export"},
    {XkwNPPMWindowMovieCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (ppmWindowMovieCallback), XtRCallback, (caddr_t) NULL},
};

#undef offset

#define EXPORT_POSTSCRIPT          0
#define EXPORT_SUNRAS              1
#define EXPORT_PPM_DATA_IMAGE      2
#define EXPORT_PPM_DATA_MOVIE      3
#define EXPORT_PPM_WINDOW_IMAGE    4
#define EXPORT_PPM_WINDOW_MOVIE    5
#define EXPORT_WHOLE_KARMA         6
#define EXPORT_WHOLE_FITS          7
#define EXPORT_WHOLE_MIRIAD        8
#define EXPORT_WHOLE_GIPSY         9
#define NUM_WHOLE_CHOICES          10   /*  Lower limit  */
#define EXPORT_SUB_KARMA           10
#define EXPORT_SUB_FITS            11
#define EXPORT_SUB_MIRIAD          12
#define EXPORT_SUB_GIPSY           13
#define NUM_EXPORT_CHOICES         14  /*  Upper limit  */

static char *export_choices[NUM_EXPORT_CHOICES] =
{
    "PostScript", "SunRasterfile", "PPM Data Image", "PPM Data Movie",
    "PPM Window Image", "PPM Window Movie", "Karma (whole dataset)",
    "FITS (whole dataset)", "Miriad (whole dataset)", "GIPSY (whole dataset)",
    "Karma (subset)", "FITS (subset)", "Miriad (subset)", "GIPSY (subset)",
};

ExportMenuClassRec exportMenuClassRec = 
{
  {        /* CoreClassPart */
      (WidgetClass) &formClassRec,              /* superclass */
      "ExportMenu",                             /* class_name */
      sizeof(ExportMenuRec),                    /* widget_size */
      NULL,                                     /* class_initialise */
      NULL,                                     /* class_part_initialise */
      FALSE,                                    /* class_init */
      (XtInitProc) _ExportMenu_Initialise,      /* initialise */
      NULL,                                     /* initialise_hook */
      XtInheritRealize,                         /* realise */
      NULL,                                     /* actions */
      0,                                        /* num_actions */
      ExportMenuResources,                      /* resources */
      XtNumber(ExportMenuResources),            /* num_resources */
      NULLQUARK,                                /* xrm_class */
      TRUE,                                     /* compress_motion */
      TRUE,                                     /* compress_exposure */
      TRUE,                                     /* compress_enterleave */
      TRUE,                                     /* visible_interest */
      _ExportMenu_Destroy,                      /* destroy */
      XtInheritResize,                          /* resize */
      XtInheritExpose,                          /* expose */
      (XtSetValuesFunc) _ExportMenu_SetValues,  /* set_values */
      NULL,                                     /* set_values_hook */
      XtInheritSetValuesAlmost,                 /* set_values_almost */
      NULL,                                     /* get_values_hook */
      NULL,                                     /* accept_focus */
      XtVersion,                                /* version */
      NULL,                                     /* callback_private */
      XtInheritTranslations,                    /* tm_translations */
      NULL,
      NULL,
      NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Constraint */
    /* subresourses       */   NULL,
    /* subresource_count  */   0,
    /* constraint_size    */   sizeof (ExportMenuConstraintsRec),
    /* initialise         */   NULL,
    /* destroy            */   NULL,
    /* set_values         */   NULL,
    /* extension          */   NULL
  },{ /* Form */
    /* layout */                XtInheritLayout
  },
  {  /* ExportMenuClassPart */
    0 /* empty */
  },
};

WidgetClass exportMenuWidgetClass = (WidgetClass) &exportMenuClassRec;

static void _ExportMenu_Initialise (Widget Request, Widget New)
{
    unsigned int num_choices = NUM_WHOLE_CHOICES;
    /*ExportMenuWidget request = (ExportMenuWidget) Request;*/
    ExportMenuWidget new = (ExportMenuWidget) New;
    Widget w;
    static unsigned int instance_number = 0;
    /*static char function_name[] = "ExportMenu::Initialise";*/

    new->exportMenu.instance_number = instance_number++;
    new->form.default_spacing = 0;
    new->core.border_width = 0;
    new->exportMenu.array = NULL;
    new->exportMenu.iarr_destroy_callback = NULL;
    if (new->exportMenu.supportSubsets) num_choices = NUM_EXPORT_CHOICES;
    w = XtVaCreateManagedWidget ("menuButton", choiceMenuWidgetClass, New,
				 XtNlabel, new->exportMenu.label,
				 XkwNmenuTitle, "Export Menu",
				 XtNmenuName, "exportMenu",
				 XkwNnumItems, num_choices,
				 XkwNitemStrings, export_choices,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, export_cbk, New);
    w = XtVaCreatePopupShell ("savePopup", dialogpopupWidgetClass, New,
			      XtNlabel, "Output basename         ",
			      NULL);
    new->exportMenu.save_dialog_popup = w;
    XtAddCallback (w, XtNcallback, save_cbk, New);
    w = XtVaCreatePopupShell ("postscriptwinpopup", postscriptWidgetClass, New,
			      XtNtitle, "Postscript Window",
			      XkwNautoIncrement, TRUE,
			      NULL);
    new->exportMenu.pswinpopup = w;
    XtAddCallback (w, XtNcallback, postscript_cbk, New);
    new->exportMenu.ds_event_callback =
	ds_event_register_func (_ExportMenu_ds_event_func, New);
}   /*  End Function Initialise  */

static void _ExportMenu_Destroy (Widget W)
{
    ExportMenuWidget w = (ExportMenuWidget) W;

    c_unregister_callback (w->exportMenu.ds_event_callback);
}   /*  End Function Destroy  */

static Boolean _ExportMenu_SetValues (Widget Current, Widget Request,
				      Widget New)
{
    ExportMenuWidget current = (ExportMenuWidget) Current;
    /*ExportMenuWidget request = (ExportMenuWidget) Request;*/
    ExportMenuWidget new = (ExportMenuWidget) New;

    if (new->exportMenu.array == current->exportMenu.array) return (False);
    /*  Intelligent array has changed  */
    if (new->exportMenu.iarr_destroy_callback != NULL)
    {
	c_unregister_callback (new->exportMenu.iarr_destroy_callback);
	new->exportMenu.iarr_destroy_callback = NULL;
    }
    if (new->exportMenu.array == NULL) return (False);
    new->exportMenu.iarr_destroy_callback = 
	iarray_register_destroy_func (new->exportMenu.array,
				      ( flag (*) () ) iarr_destroy_callback,
				      new);
    return (False);
}   /*  End Function _ExportMenu_SetValues  */

static void export_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the export menu callback.
*/
{
    unsigned int exportcode = *(int *) call_data;
    ExportMenuWidget top = (ExportMenuWidget) client_data;
    static char function_name[] = "ExportMenuWidget::export_cbk";

    if (exportcode == EXPORT_POSTSCRIPT)
    {
	XtPopup (top->exportMenu.pswinpopup, XtGrabNone);
	return;
    }
    if ( (top->exportMenu.worldCanvas == NULL) &&
	 (top->exportMenu.array == NULL) )
    {
	fprintf (stderr, "No visible canvas or Intelligent Array!\n");
	return;
    }
    if ( (top->exportMenu.array == NULL) &&
	 (exportcode != EXPORT_PPM_WINDOW_IMAGE) )
    {
	if (viewimg_get_active (top->exportMenu.worldCanvas) == NULL)
	{
	    fprintf (stderr, "No visible image!\n");
	    return;
	}
    }
    top->exportMenu.export_type = exportcode;
    switch (top->exportMenu.export_type)
    {
      case EXPORT_PPM_DATA_IMAGE:
      case EXPORT_PPM_DATA_MOVIE:
      case EXPORT_PPM_WINDOW_IMAGE:
      case EXPORT_PPM_WINDOW_MOVIE:
	XtVaSetValues (top->exportMenu.save_dialog_popup,
		       XkwNextension, ".ppm",
		       NULL);
	break;
      case EXPORT_SUNRAS:
	XtVaSetValues (top->exportMenu.save_dialog_popup,
		       XkwNextension, ".ras",
		       NULL);
	break;
      case EXPORT_WHOLE_KARMA:
      case EXPORT_SUB_KARMA:
	XtVaSetValues (top->exportMenu.save_dialog_popup,
		       XkwNextension, ".kf",
		       NULL);
	break;
      case EXPORT_WHOLE_FITS:
      case EXPORT_SUB_FITS:
	XtVaSetValues (top->exportMenu.save_dialog_popup,
		       XkwNextension, ".fits",
		       NULL);
	break;
      case EXPORT_WHOLE_MIRIAD:
      case EXPORT_SUB_MIRIAD:
      case EXPORT_WHOLE_GIPSY:
      case EXPORT_SUB_GIPSY:
	XtVaSetValues (top->exportMenu.save_dialog_popup,
		       XkwNextension, "",
		       NULL);
	break;
      default:
	fprintf (stderr, "Illegal export code: %u\n", exportcode);
	a_prog_bug (function_name);
	break;
    }
    XtPopup (top->exportMenu.save_dialog_popup, XtGrabNone);
}   /*  End Function export_cbk   */

static void save_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the save callback.
*/
{
    flag ok;
    flag truecolour;
    iarray array;
    ViewableImage vimage = NULL;
    Kcolourmap cmap = NULL;  /*  Initialised to keep compiler happy  */
    Channel channel;
    KCallbackList clist = NULL;
    unsigned int hdim, vdim;
    unsigned int pseudo_index, red_index, green_index, blue_index;
    unsigned int cmap_size, visual, count;
    ExportMenuWidget top = (ExportMenuWidget) client_data;
    double i_min, i_max;
    array_desc *arr_desc;
    multi_array *multi_desc;
    CONST unsigned char *image;
    unsigned short *intensities = NULL;
    unsigned short *cmap_red = NULL;
    unsigned short *cmap_green = NULL;
    unsigned short *cmap_blue = NULL;
    void *iscale_info;
    flag (*iscale_func) ();
    unsigned short tmp_cmap[256 * 3];
    String fname = (String) call_data;
    static char function_name[] = "ExportMenuWidget::save_cbk";

    array = top->exportMenu.array;
    if ( (top->exportMenu.worldCanvas == NULL) && (array == NULL) )
    {
	fprintf (stderr, "No visible canvas or Intelligent Array!\n");
	RETURN;
    }
    if (top->exportMenu.export_type == EXPORT_PPM_WINDOW_IMAGE)
    {
	ExportMenuWritePPMWindow ( (Widget) top, fname );
	RETURN;
    }
    if (top->exportMenu.export_type == EXPORT_PPM_WINDOW_MOVIE)
    {
	char *ptr;
	char basename[STRING_LENGTH];

	if (top->exportMenu.ppmWindowMovieCallback == NULL)
	{
	    fprintf (stderr, "PPM Window Movie not available\n");
	    XBell (XtDisplay (w), 100);
	    RETURN;
	}
	strcpy (basename, fname);
	if ( ( ptr = strstr (basename, ".ppm") ) != NULL ) *ptr = '\0';
	XtCallCallbacks ( (Widget) top, XkwNPPMWindowMovieCallback,
			  (XtPointer) basename );
	RETURN;
    }
    if (array == NULL)
    {
	/*  No array supplied: get it from the viewable image  */
	flag dealloc;

	if ( ( vimage = viewimg_get_active (top->exportMenu.worldCanvas) )
	     == NULL )
	{
	    fprintf (stderr, "No visible image!\n");
	    RETURN;
	}
	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_TRUECOLOUR, &truecolour,
				VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
				VIEWIMG_VATT_SLICE, &image,
				VIEWIMG_VATT_HDIM, &hdim,
				VIEWIMG_VATT_VDIM, &vdim,
				VIEWIMG_VATT_MULTI_ARRAY, &multi_desc,
				VIEWIMG_VATT_END);
	if (multi_desc == NULL)
	{
	    /*  Create one for the benefit of the Karma & etc. writers  */
	    array_pointer arrayp;
	    array_desc *tmp_arr_desc;

	    if ( ( tmp_arr_desc = ds_copy_array_desc_until (arr_desc, NULL) )
		 == NULL )
	    {
		m_error_notify (function_name, "array descriptor");
		RETURN;
	    }
	    m_clear (&arrayp, sizeof arrayp);
	    arrayp.array = (char *) image;
	    if ( ( multi_desc =
		   ds_easy_alloc_array_from_array_desc (tmp_arr_desc, &arrayp,
							FALSE) ) == NULL )
	    {
		m_error_notify (function_name, "multi_array structure");
		ds_dealloc_array_desc (tmp_arr_desc);
		RETURN;
	    }
	    AUTO_FREE (ds_dealloc_multi, multi_desc);
	}
	switch (top->exportMenu.export_type)
	{
	  case EXPORT_SUB_KARMA:
	  case EXPORT_SUB_FITS:
	  case EXPORT_SUB_MIRIAD:
	  case EXPORT_SUB_GIPSY:
	    /*  Need to create a sub-array  */
	    if ( ( multi_desc = create_sub (multi_desc, &dealloc,
					    arr_desc, hdim, vdim,
					    top->exportMenu.worldCanvas) )
		 == NULL )
	    {
		m_error_notify (function_name, "sub-array");
		RETURN;
	    }
	    if (dealloc) AUTO_FREE (ds_dealloc_multi, multi_desc);
	    break;
	  default:
	    /* Other export types use the original data or the specific image*/
	    break;
	}
	kwin_get_attributes (canvas_get_pixcanvas(top->exportMenu.worldCanvas),
			     KWIN_ATT_VISUAL, &visual,
			     KWIN_ATT_END);
	cmap = canvas_get_cmap (top->exportMenu.worldCanvas);
	if ( (cmap == NULL) ||
	     ( (visual == KWIN_VISUAL_PSEUDOCOLOUR) && truecolour ) )
	{
	    cmap_red = NULL;
	    cmap_green = NULL;
	    cmap_blue = NULL;
	}
	else
	{
	    if ( ( intensities = kcmap_get_rgb_values (cmap, &cmap_size) )
		 == NULL )
	    {
		RETURN;
	    }
	    cmap_red = tmp_cmap;
	    cmap_green = tmp_cmap + 1;
	    cmap_blue = tmp_cmap + 2;
	    if ( (visual == KWIN_VISUAL_DIRECTCOLOUR) && (cmap_size < 256) )
	    {
		/*  Set extra colours at front to first colour in colourmap  */
		for (count = 0; count < (256 - cmap_size);
		     ++count, cmap_red += 3, cmap_green += 3, cmap_blue += 3)
		{
		    *cmap_red = intensities[0];
		    *cmap_green = intensities[1];
		    *cmap_blue = intensities[2];
		}
		/*  Copy remaining colours  */
		for (count = 0; count < cmap_size; ++count)
		{
		    cmap_red[count * 3] = intensities[count * 3];
		    cmap_green[count * 3] = intensities[count * 3 + 1];
		    cmap_blue[count * 3] = intensities[count * 3 + 2];
		}
		cmap_red = tmp_cmap;
		cmap_green = tmp_cmap + 1;
		cmap_blue = tmp_cmap + 2;
	    }
	    else
	    {
		for (count = 0; count < cmap_size; ++count)
		{
		    cmap_red[count * 3] = intensities[count * 3];
		    cmap_green[count * 3] = intensities[count * 3 + 1];
		    cmap_blue[count * 3] = intensities[count * 3 + 2];
		}
	    }
	    m_free (intensities);
	}
	if (truecolour)
	{
	    viewimg_get_attributes (vimage,
				    VIEWIMG_VATT_RED_INDEX, &red_index,
				    VIEWIMG_VATT_GREEN_INDEX, &green_index,
				    VIEWIMG_VATT_BLUE_INDEX, &blue_index,
				    VIEWIMG_VATT_END);
	}
	else
	{
	    if (cmap == NULL)
	    {
		fprintf (stderr, "%s: no colourmap!\n", function_name);
		RETURN;
	    }
	    canvas_get_attributes (top->exportMenu.worldCanvas,
				   CANVAS_ATT_VALUE_MIN, &i_min,
				   CANVAS_ATT_VALUE_MAX, &i_max,
				   CANVAS_ATT_ISCALE_FUNC, &iscale_func,
				   CANVAS_ATT_ISCALE_INFO, &iscale_info,
				   CANVAS_ATT_END);
	    viewimg_get_attributes (vimage,
				    VIEWIMG_VATT_PSEUDO_INDEX, &pseudo_index,
				    VIEWIMG_VATT_END);
	}
    }
    else
    {
	/*  Intelligent array was supplied  */
	multi_desc = array->multi_desc;
    }
    switch (top->exportMenu.export_type)
    {
      case EXPORT_PPM_DATA_IMAGE:
	if ( (array != NULL) && (iarray_num_dim (array) != 2) )
	{
	    fprintf (stderr, "Array is not 2-dimensional\n");
	    RETURN;
	}
	if ( ( channel = ch_open_file (fname, "w") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening: \"%s\"\t%s\n",
		     function_name, fname, ERRSTRING);
	    RETURN;
	}
	if (array != NULL)
	{
	    iarray_min_max (array, CONV_CtoR_REAL, &i_min, &i_max);
	    ok = foreign_ppm_write_pseudo
		(channel, TRUE, array->data, iarray_type (array),
		 array->offsets[1], array->offsets[0],
		 iarray_dim_length (array, 1), iarray_dim_length (array, 0),
		 0, 0, NULL, NULL, NULL, 0, 0, i_min, i_max, NULL, NULL);
	    if (!ok) fprintf (stderr, "%s: error writing iarray\n",
			      function_name);
	}
	else if (truecolour)
	{
	    ok = foreign_ppm_write_rgb
		(channel, TRUE,
		 image + ds_get_element_offset (arr_desc->packet, red_index),
		 image + ds_get_element_offset (arr_desc->packet, green_index),
		 image + ds_get_element_offset (arr_desc->packet, blue_index),
		 arr_desc->offsets[hdim], arr_desc->offsets[vdim],
		 arr_desc->dimensions[hdim]->length,
		 arr_desc->dimensions[vdim]->length,
		 0, 0, cmap_red, cmap_green, cmap_blue, 3);
	    if (!ok) fprintf (stderr, "%s: error writing TrueColour array\n",
			      function_name);
	}
	else
	{
	    ok = foreign_ppm_write_pseudo
		(channel, TRUE,
		 (CONST char *) image +
		 ds_get_element_offset (arr_desc->packet, pseudo_index),
		 arr_desc->packet->element_types[pseudo_index],
		 arr_desc->offsets[hdim], arr_desc->offsets[vdim],
		 arr_desc->dimensions[hdim]->length,
		 arr_desc->dimensions[vdim]->length,
		 0, 0, cmap_red, cmap_green, cmap_blue, cmap_size, 3,
		 i_min, i_max, iscale_func, iscale_info);
	    if (!ok) fprintf (stderr, "%s: error writing PseudoColour array\n",
			      function_name);
	}
	if (!ok)
	{
	    fprintf (stderr, "%s: error writing PPM file: \"%s\"\n",
		     function_name, fname);
	    ch_close (channel);
	    unlink (fname);
	    RETURN;
	}
	ch_close (channel);
	fprintf (stderr, "Wrote PPM file: \"%s\"\n", fname);
	break;
      case EXPORT_PPM_DATA_MOVIE:
	if (multi_desc == NULL)
	{
	    fprintf (stderr, "%s: NULL multi_array\n", function_name);
	    RETURN;
	}
	write_ppm_data_movie (multi_desc, truecolour, arr_desc, hdim, vdim,
			      top->exportMenu.worldCanvas, vimage,
			      cmap_size, cmap_red, cmap_green, cmap_blue,
			      fname);
	break;
      case EXPORT_SUNRAS:
	if ( (array != NULL) && (iarray_num_dim (array) != 2) )
	{
	    fprintf (stderr, "Array is not 2-dimensional\n");
	    RETURN;
	}
	if ( ( channel = ch_open_file (fname, "w") ) == NULL ) RETURN;
	if (array != NULL)
	{
	    iarray_min_max (array, CONV_CtoR_REAL, &i_min, &i_max);
	    ok = foreign_sunras_write_pseudo
		(channel, array->data, iarray_type (array),
		 array->offsets[1], array->offsets[0],
		 iarray_dim_length (array, 1), iarray_dim_length (array, 0),
		 NULL, NULL, NULL, 0, 0, i_min, i_max);
	}
	else if (truecolour)
	{
	    ok = foreign_sunras_write_rgb
		(channel,
		 image + ds_get_element_offset (arr_desc->packet, red_index),
		 image + ds_get_element_offset (arr_desc->packet, green_index),
		 image + ds_get_element_offset (arr_desc->packet, blue_index),
		 arr_desc->offsets[hdim], arr_desc->offsets[vdim],
		 arr_desc->dimensions[hdim]->length,
		 arr_desc->dimensions[vdim]->length,
		 cmap_red, cmap_green, cmap_blue, 3);
	}
	else ok = foreign_sunras_write_pseudo
		 (channel,
		  (CONST char *) image +
		  ds_get_element_offset (arr_desc->packet,pseudo_index),
		  arr_desc->packet->element_types[pseudo_index],
		  arr_desc->offsets[hdim], arr_desc->offsets[vdim],
		  arr_desc->dimensions[hdim]->length,
		  arr_desc->dimensions[vdim]->length,
		  cmap_red, cmap_green, cmap_blue, cmap_size, 3,
		  i_min, i_max);
	if (!ok)
	{
	    ch_close (channel);
	    unlink (fname);
	    RETURN;
	}
	ch_close (channel);
	fprintf (stderr, "Wrote Sun Rasterfile: \"%s\"\n", fname);
	break;
      case EXPORT_WHOLE_KARMA:
      case EXPORT_SUB_KARMA:
	if ( !dsxfr_put_multi (fname, multi_desc) )
	{
	    unlink (fname);
	    RETURN;
	}
	fprintf (stderr, "Wrote Karma file: \"%s\"\n", fname);
	break;
      case EXPORT_WHOLE_FITS:
      case EXPORT_SUB_FITS:
	if ( ( channel = ch_open_file (fname, "w") ) == NULL )
	{
	    RETURN;
	}
	if ( !foreign_fits_write (channel, multi_desc,
				  FA_FITS_WRITE_END) )
	{
	    ch_close (channel);
	    unlink (fname);
	    RETURN;
	}
	ch_close (channel);
	fprintf (stderr, "Wrote FITS file: \"%s\"\n", fname);
	break;
      case EXPORT_WHOLE_MIRIAD:
      case EXPORT_SUB_MIRIAD:
	if ( !foreign_miriad_write (fname, multi_desc,
				    FA_MIRIAD_WRITE_END) )
	{
	    RETURN;
	}
	fprintf (stderr, "Wrote Miriad file: \"%s\"\n", fname);
	break;
      case EXPORT_WHOLE_GIPSY:
      case EXPORT_SUB_GIPSY:
	if ( !foreign_gipsy_write (fname, multi_desc,
				   FA_GIPSY_WRITE_END) )
	{
	    RETURN;
	}
	fprintf (stderr, "Wrote GIPSY file: \"%s\"\n", fname);
	break;
      default:
	fprintf (stderr, "Illegal export code: %u\n",
		 top->exportMenu.export_type);
	a_prog_bug (function_name);
	break;
    }
    RETURN;
}   /*  End Function save_cbk   */

static void postscript_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  [SUMMARY] PostScript callback.
    [PURPOSE] This is the PostScript callback. It is called when the PostScript
    widget needs to know the pixel canvas and image name.
*/
{
    KPixCanvas pixcanvas;
    ExportMenuWidget top = (ExportMenuWidget) client_data;

    pixcanvas = canvas_get_pixcanvas (top->exportMenu.worldCanvas);
    XkwPostscriptRegisterImageAndName (w, pixcanvas, NULL);
}   /*  End Function postscript_cbk   */

static multi_array *create_sub (multi_array *old, flag *allocated,
				array_desc *inp_arr_desc, unsigned int hdim,
				unsigned int vdim, KWorldCanvas canvas)
/*  [SUMMARY] Create a sub-array.
    <old> The old multi_array descriptor.
    <allocated> TRUE is written here if a new data structure was allocated.
    <inp_arr_desc> The input array descriptor.
    <hdim> The horizontal dimension.
    <vdim> The vertical dimension.
    <canvas> The KWorldCanvas from which the subimage parameters are taken.
    [RETURNS] A new multi_array object on success, else NULL.
*/
{
    flag more_to_do;
    int count;
    uaddr inp_offset, out_offset;
    uaddr hstart, hstop, vstart, vstop;
    unsigned int num_dim, index;
    unsigned int elem_index = 0;
    double left_x, right_x, bottom_y, top_y, min, max;
    char *bunit;
    dim_desc *inp_hd, *inp_vd, *out_hd, *out_vd;
    array_desc *out_arr_desc;
    multi_array *new;
    packet_desc *out_pack_desc;
    uaddr *coordinates;
    CONST char *inp_array;
    char *out_array;
    char txt[STRING_LENGTH];
    static char function_name[] = "ExportMenuWidget::create_sub";

    if (old == NULL)
    {
	fprintf (stderr, "NULL old multi_array pointer passed\n");
	a_prog_bug (function_name);
    }
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    num_dim = inp_arr_desc->num_dimensions;
    inp_hd = inp_arr_desc->dimensions[hdim];
    inp_vd = inp_arr_desc->dimensions[vdim];
    /*  Check input for sanity  */
    if ( !ds_packet_all_data (inp_arr_desc->packet) )
    {
	fprintf (stderr, "%s: array packets are not all atomic data\n",
		 function_name);
	return (NULL);
    }
    if ( ( out_arr_desc = find_array (old, inp_hd->name, &index) ) == NULL )
    {
	fprintf (stderr, "Dimension: \"%s\" not found!\n", inp_hd->name);
	a_prog_bug (function_name);
    }
    if (out_arr_desc != inp_arr_desc)
    {
	fprintf (stderr, "Array missmatch!\n");
	a_prog_bug (function_name);
    }
    if (inp_arr_desc != (array_desc *) old->headers[index]->element_desc[0])
    {
	fprintf (stderr,
		 "%s: array is not the first in it's general structure!\n",
		 function_name);
	return (NULL);
    }
    inp_array = *(char **) old->data[index];
    if ( !ds_compute_array_offsets (inp_arr_desc) )
    {
	m_error_notify (function_name, "array offsets");
	return (NULL);
    }
    hstart = ds_get_coord_num (inp_hd, left_x, SEARCH_BIAS_CLOSEST);
    hstop = ds_get_coord_num (inp_hd, right_x, SEARCH_BIAS_CLOSEST);
    if (hstop <= hstart)
    {
	fprintf (stderr,
		 "Horizontal end index: %lu is not greater than start: %lu\n",
		 hstop, hstart);
	return (NULL);
    }
    ++hstop;
    vstart = ds_get_coord_num (inp_vd, bottom_y, SEARCH_BIAS_CLOSEST);
    vstop = ds_get_coord_num (inp_vd, top_y, SEARCH_BIAS_CLOSEST);
    if (vstop <= vstart)
    {
	fprintf (stderr,
		 "Vertical end index: %lu is not greater than start: %lu\n",
		 vstop, vstart);
	return (NULL);
    }
    ++vstop;
    if ( (inp_hd->length == hstop - hstart) &&
	 (inp_vd->length == vstop - vstart) )
    {
	*allocated = FALSE;
	return (old);
    }
    *allocated = TRUE;
    /*  Create new data structure descriptors  */
    /*  Allocate the multi_array  */
    if ( ( new = ds_alloc_multi (old->num_arrays) ) == NULL )
    {
	m_error_notify (function_name, "multi_array");
	return (NULL);
    }
    /*  Copy top-level packet descriptors  */
    for (count = 0; count < old->num_arrays; ++count)
    {
	if ( ( new->headers[count] = ds_copy_desc_until (old->headers[count],
							 NULL) ) == NULL )
	{
	    m_error_notify (function_name,
			    "general data structure descriptors");
	    ds_dealloc_multi (new);
	    return (NULL);
	}
    }
    /*  Find and fiddle the two dimensions that need to be changed  */
    if ( ( out_arr_desc = find_array (new, inp_hd->name, &index) ) == NULL )
    {
	fprintf (stderr, "Dimension: \"%s\" not found!\n", inp_hd->name);
	a_prog_bug (function_name);
    }
    out_hd = out_arr_desc->dimensions[hdim];
    out_vd = out_arr_desc->dimensions[vdim];
    /*  Modify dimension lengths and first and last co-ordinates  */
    out_hd->length = hstop - hstart;
    out_vd->length = vstop - vstart;
    out_arr_desc->lengths[hdim] = out_hd->length;
    out_arr_desc->lengths[vdim] = out_vd->length;
    out_hd->first_coord = ds_get_coordinate (inp_hd, hstart);
    out_hd->last_coord = ds_get_coordinate (inp_hd, hstop - 1);
    out_vd->first_coord = ds_get_coordinate (inp_vd, vstart);
    out_vd->last_coord = ds_get_coordinate (inp_vd, vstop - 1);
    ds_remove_tiling_info (out_arr_desc);
    if ( !ds_compute_array_offsets (out_arr_desc) )
    {
	m_error_notify (function_name, "array offsets");
	ds_dealloc_multi (new);
	return (NULL);
    }
    /*  Allocate and copy data areas. This should not copy the actual array
	data since the array size has changed  */
    for (count = 0; count < old->num_arrays; ++count)
    {
	if ( ( new->data[count] = ds_alloc_data (new->headers[count], FALSE,
						 TRUE) ) == NULL )
	{
	    m_error_notify (function_name, "data");
	    ds_dealloc_multi (new);
	    return (NULL);
	}
	if ( !ds_copy_data (old->headers[count], old->data[count],
			    new->headers[count], new->data[count]) )
	{
	    fprintf (stderr, "%s: error copying data\n", function_name);
	    ds_dealloc_multi (new);
	    return (NULL);
	}
    }
    /*  Fiddle any FITS-style co-ordinates  */
    fiddle_fits (new->headers[index], new->data + index, out_hd);
    fiddle_fits (new->headers[index], new->data + index, out_vd);
    /*  Proceed to copy data  */
    out_array = *(char **) new->data[index];
    /*  Prepare for computing min&max  */
    out_pack_desc = out_arr_desc->packet;
    bunit = ds_get_unique_named_string (new->headers[index], new->data[index],
					"BUNIT");
    if (bunit != NULL)
    {
	strcpy (txt, bunit);
	m_free (bunit);
	bunit = txt;
	elem_index = ds_f_elem_in_packet (out_pack_desc, bunit);
    }
    min = TOOBIG;
    max = -TOOBIG;
    /*  Allocate and initialise co-ordinate array  */
    if ( ( coordinates = m_alloc (sizeof *coordinates * num_dim) ) == NULL )
    {
	m_abort (function_name, "co-ordinate array");
    }
    m_clear (coordinates, sizeof *coordinates * num_dim);
    /*  Iterate over all co-ordinates except that the specified dimensions are
	skipped  */
    more_to_do = TRUE;
    while (more_to_do)
    {
	/*  Process  */
	inp_offset = 0;
	out_offset = 0;
	for (count = 0; count < num_dim; ++count)
	{
	    inp_offset += inp_arr_desc->offsets[count][ coordinates[count] ];
	    out_offset += out_arr_desc->offsets[count][ coordinates[count] ];
	}
	if ( !copy_plane (inp_array + inp_offset, inp_arr_desc, hdim, vdim,
			  hstart, vstart,
			  out_array + out_offset, out_arr_desc) )
	{
	    m_free (coordinates);
	    return (FALSE);
	}
	if (bunit != NULL)
	{
	    ds_find_2D_extremes (out_array + out_offset,
				 out_vd->length, out_arr_desc->offsets[vdim],
				 out_hd->length, out_arr_desc->offsets[hdim],
				 out_pack_desc->element_types[elem_index],
				 CONV_CtoR_REAL, &min, &max);
	}
	/*  Increment co-ordinates  */
	count = num_dim - 1;
	more_to_do = TRUE;
	while ( (count >= 0) && more_to_do )
	{
	    if ( (count == hdim) || (count == vdim) )
	    {
		--count;
		continue;
	    }
	    /*  Increment co-ordinate in this dimension  */
	    if (++coordinates[count] < inp_arr_desc->dimensions[count]->length)
	    {
		/*  Co-ordinate for this dimension still valid  */
		more_to_do = FALSE;
	    }
	    else
	    {
		/*  Co-ordinate has gone past limits: go up to the next
		    dimension  */
		coordinates[count] = 0;
		--count;
	    }
	}
	more_to_do = (count < 0) ? FALSE : TRUE;
    }
    m_free (coordinates);
    if (bunit != NULL)
    {
	double scale, offset;

	ds_get_data_scaling (bunit, new->headers[index], new->data[index],
			     &scale, &offset);
	min = offset + scale * min;
	max = offset + scale * max;
	ds_set_data_range (bunit, new->headers[index], new->data + index,
			   min, max);
    }
    return (new);
}   /*  End Function create_sub  */

static array_desc *find_array (CONST multi_array *multi_desc, CONST char *name,
			       unsigned int *index)
/*  [SUMMARY] Find a multi-dimensional array.
    <multi_desc> The data structure to search.
    <name> The name of a dimension in the array to search for.
    <index> The index of the general data structure.
    [RETURNS] A pointer to the array descriptor on success, else NULL.
*/
{
    unsigned int count;
    unsigned int dim_index;
    array_desc *arr_desc;
    static char function_name[] = "find_array";

    for (count = 0; count < multi_desc->num_arrays; ++count)
    {
	switch ( ds_f_name_in_packet (multi_desc->headers[count], name,
				      (char **) &arr_desc, &dim_index) )
	{
	  case IDENT_NOT_FOUND:
	    continue;
	    /*break;*/
	  case IDENT_ELEMENT:
	    continue;
	    /*break;*/
	  case IDENT_MULTIPLE:
	    fprintf (stderr, "Item \"%s\" found more than once\n", name);
	    return (NULL);
	    /*break;*/
	  case IDENT_DIMENSION:
	    *index = count;
	    return (arr_desc);
	    /*break;*/
	  case IDENT_GEN_STRUCT:
	  default:
	    fprintf (stderr,
		     "Bad return value from function: ds_f_name_in_packet\n");
	    a_prog_bug (function_name);
	    break;
	}
    }
    return (arr_desc);
}   /*  End Function find_array  */

static void fiddle_fits (packet_desc *pack_desc, char **packet, dim_desc *dim)
/*  [SUMMARY] Fiddle FITS information for an axis.
    <pack_desc> The top-level packet descriptor. This is modified.
    <packet> A pointer to the top-level packet pointer. This is modified.
    <dim> The dimension. This is modified.
    [RETURNS] Nothing.
*/
{
    unsigned int fits_axis;
    char txt[STRING_LENGTH];
    double crpix[2];
    static char function_name[] = "ExportMenuWidget::fiddle_fits";

    if ( ( fits_axis = ds_get_fits_axis (pack_desc, *packet, dim->name) )
	 < 1 ) return;
    sprintf (txt, "CRPIX%u", fits_axis);
    if ( !ds_get_unique_named_value (pack_desc, *packet, txt, NULL,
				     crpix) ) return;
    crpix[0] -= dim->first_coord;
    if ( !ds_put_unique_named_value (pack_desc, packet, txt, K_DOUBLE,
				     crpix, TRUE) )
    {
	m_abort (function_name, "new FITS keyword");
    }
    dim->first_coord = 0.0;
    dim->last_coord = dim->length - 1;
}   /*  End Function fiddle_fits  */

static flag copy_plane (CONST char *inp_data, CONST array_desc *inp_arr_desc,
			unsigned int hdim, unsigned int vdim,
			uaddr hstart, uaddr vstart,
			char *out_data, CONST array_desc *out_arr_desc)
/*  [SUMMARY] Copy part of a plane of data.
    <inp_data> The input data.
    <inp_arr_desc> The input array descriptor.
    <out_data> The output data.
    <out_arr_desc> The output array descriptor.
    <dim_index> The dimension to copy.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    uaddr hcount, vcount, hlen, vlen;
    unsigned int pack_size;
    /*static char function_name[] = "copy_plane";*/

    pack_size = ds_get_packet_size (inp_arr_desc->packet);
    hlen = out_arr_desc->dimensions[hdim]->length;
    vlen = out_arr_desc->dimensions[vdim]->length;
    for (vcount = 0; vcount < vlen; ++vcount)
	for (hcount = 0; hcount < hlen; ++hcount)
	{
	    m_copy (out_data + out_arr_desc->offsets[hdim][hcount] +
		    out_arr_desc->offsets[vdim][vcount],
		    inp_data + inp_arr_desc->offsets[hdim][hcount + hstart] +
		    inp_arr_desc->offsets[vdim][vcount + vstart],
		    pack_size);
	}
    return (TRUE);
}   /*  End Function copy_plane  */

static void iarr_destroy_callback (iarray arr, ExportMenuWidget top)
/*  [PURPOSE] This routine will register the destruction of the Intelligent
    Array.
    <arr> The Intelligent Array.
    <top> The ExportMenu widget.
    [RETURNS] Nothing.
*/
{
    top->exportMenu.array = NULL;
    top->exportMenu.iarr_destroy_callback = NULL;
}   /*  End Function iarr_destroy_callback  */

static void write_ppm_data_movie (multi_array *multi_desc, flag truecolour,
				  array_desc *arr_desc,
				  unsigned int hdim, unsigned int vdim,
				  KWorldCanvas canvas, ViewableImage vimage,
				  unsigned int cmap_size,
				  CONST unsigned short *red,
				  CONST unsigned short *green,
				  CONST unsigned short *blue,
				  CONST char *filename)
/*  [SUMMARY] Write a PPM movie.
    <multi_desc> The multi_array descriptor.
    <truecolour> If TRUE a TrueColour dataset is to be written.
    <arr_desc> The array descriptor.
    <hdim> The horizontal dimension.
    <vdim> The vertical dimension.
    <canvas> The KWorldCanvas from which the subimage parameters are taken.
    <vimage> The ViewableImage object.
    <cmap_size> The size of the associated colourmap.
    <red> The array of red intensities.
    <green> The array of green intensities.
    <blue> The array of blue intensities.
    <filename> The filename.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    flag ok;
    int count, precision;
    uaddr image_size;
    uaddr hstart, hstop, vstart, vstop;
    unsigned int num_dim, index, fdim;
    unsigned int pseudo_index, red_index, green_index, blue_index;
    double left_x, right_x, bottom_y, top_y, i_min, i_max;
    dim_desc *hd, *vd, *fd;
    char *ptr;
    CONST unsigned char *array;
    void *iscale_info;
    flag (*iscale_func) ();
    char basename[STRING_LENGTH], fname[STRING_LENGTH];
    static char function_name[] = "ExportMenuWidget::write_ppm_data_movie";

    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    if ( (num_dim = arr_desc->num_dimensions) != 3 )
    {
	fprintf (stderr, "%s: number of dimensions: %u is not 3\n",
		 function_name, num_dim);
	return;
    }
    if ( (hdim != 0) && (vdim != 0) ) fdim = 0;
    else if ( (hdim != 1) && (vdim != 1) ) fdim = 1;
    else fdim = 2;
    hd = arr_desc->dimensions[hdim];
    vd = arr_desc->dimensions[vdim];
    fd = arr_desc->dimensions[fdim];
    /*  Check input for sanity  */
    if ( !ds_packet_all_data (arr_desc->packet) )
    {
	fprintf (stderr, "%s: array packets are not all atomic data\n",
		 function_name);
	return;
    }
    if (find_array (multi_desc, hd->name, &index) != arr_desc)
    {
	fprintf (stderr, "Dimension: \"%s\" not found!\n", hd->name);
	a_prog_bug (function_name);
    }
    if (arr_desc != (array_desc *) multi_desc->headers[index]->element_desc[0])
    {
	fprintf (stderr,
		 "%s: array is not the first in it's general structure!\n",
		 function_name);
	return;
    }
    array = *(unsigned char **) multi_desc->data[index];
    if ( !ds_compute_array_offsets (arr_desc) )
    {
	m_error_notify (function_name, "array offsets");
	return;
    }
    hstart = ds_get_coord_num (hd, left_x, SEARCH_BIAS_CLOSEST);
    hstop = ds_get_coord_num (hd, right_x, SEARCH_BIAS_CLOSEST);
    if (hstop <= hstart)
    {
	fprintf (stderr,
		 "Horizontal end index: %lu is not greater than start: %lu\n",
		 hstop, hstart);
	return;
    }
    ++hstop;
    vstart = ds_get_coord_num (vd, bottom_y, SEARCH_BIAS_CLOSEST);
    vstop = ds_get_coord_num (vd, top_y, SEARCH_BIAS_CLOSEST);
    if (vstop <= vstart)
    {
	fprintf (stderr,
		 "Vertical end index: %lu is not greater than start: %lu\n",
		 vstop, vstart);
	return;
    }
    ++vstop;
    if (truecolour)
    {
	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_RED_INDEX, &red_index,
				VIEWIMG_VATT_GREEN_INDEX, &green_index,
				VIEWIMG_VATT_BLUE_INDEX, &blue_index,
				VIEWIMG_VATT_END);
    }
    else
    {
	canvas_get_attributes (canvas,
			       CANVAS_ATT_VALUE_MIN, &i_min,
			       CANVAS_ATT_VALUE_MAX, &i_max,
			       CANVAS_ATT_ISCALE_FUNC, &iscale_func,
			       CANVAS_ATT_ISCALE_INFO, &iscale_info,
			       CANVAS_ATT_END);
	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_PSEUDO_INDEX, &pseudo_index,
				VIEWIMG_VATT_END);
    }
    strcpy (basename, filename);
    if ( ( ptr = strstr (basename, ".ppm") ) != NULL ) *ptr = '\0';
    image_size = ds_get_packet_size (arr_desc->packet) *hd->length *vd->length;
    precision = ceil ( log10 (fd->length) );
    for (count = 0; count < fd->length; ++count, array += image_size)
    {
	sprintf (fname, "%s.%.*u.ppm", basename, precision, count);
	if ( ( channel = ch_open_file (fname, "w") ) == NULL ) return;
	if (truecolour)
	{
	    ok = foreign_ppm_write_rgb
		(channel, TRUE,
		 array + ds_get_element_offset (arr_desc->packet, red_index),
		 array + ds_get_element_offset (arr_desc->packet, green_index),
		 array + ds_get_element_offset (arr_desc->packet, blue_index),
		 arr_desc->offsets[hdim] + hstart,
		 arr_desc->offsets[vdim] + vstart,
		 hstop - hstart, vstop - vstart, 0, 0,
		 red, green, blue, 3);
	}
	else ok = foreign_ppm_write_pseudo
		 (channel, TRUE,
		  (CONST char *) array +
		  ds_get_element_offset (arr_desc->packet, pseudo_index),
		  arr_desc->packet->element_types[pseudo_index],
		  arr_desc->offsets[hdim] + hstart,
		  arr_desc->offsets[vdim] + vstart,
		  hstop - hstart, vstop - vstart, 0, 0,
		  red, green, blue, cmap_size, 3, i_min, i_max,
		  iscale_func, iscale_info);
	if (!ok)
	{
	    ch_close (channel);
	    unlink (fname);
	    return;
	}
	ch_close (channel);
	fprintf (stderr, "Wrote PPM file: \"%s\"\n", fname);
	
    }
}   /*  End Function write_ppm_data_movie  */

static flag _ExportMenu_ds_event_func (void *object, multi_array *multi_desc,
				       CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    unsigned int instance_number;
    ExportMenuWidget top = (ExportMenuWidget) object;
    packet_desc *pack_desc;
    char *packet, *filename;
    /*static char function_name[] = "ExportMenu::ds_event_func";*/

    if (strcmp (name, "ExportMenuWidget::WritePPMImageWindow") != 0)
	return (FALSE);
    if (multi_desc->num_arrays != 1) return (FALSE);
    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0];
    if (pack_desc->num_elements != 2) return (FALSE);
    if (pack_desc->element_types[0] != K_UINT) return (FALSE);
    if (pack_desc->element_types[1] != K_VSTRING) return (FALSE);
    if (strcmp (pack_desc->element_desc[0], "instance_number") != 0)
	return (FALSE);
    if (strcmp (pack_desc->element_desc[1], "filename") != 0) return (FALSE);
    m_copy (&instance_number, packet, sizeof instance_number);
    if (instance_number != top->exportMenu.instance_number) return (FALSE);
    m_copy (&filename, packet + ds_get_element_offset (pack_desc, 1),
	    sizeof filename);
    ExportMenuWritePPMWindow ( (Widget) top, filename );
    return (TRUE);
}   /*  End Function _ExportMenu_ds_event_func  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void ExportMenuWritePPMWindow (Widget W, String fname)
{
    KPixCanvas pixcanvas;
    Channel channel;
    KCallbackList clist = NULL;
    int width, height;
    unsigned short red, green, blue;
    unsigned int count;
    unsigned long background;
    uaddr *offsets;
    unsigned char *rgbimage;
    ExportMenuWidget top = (ExportMenuWidget) W;
    unsigned char colour[3];
    static char function_name[] = "ExportMenuWidget::WritePPMWindow";

    if ( !XtIsExportMenu (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    if (top->exportMenu.worldCanvas == NULL)
    {
	fprintf (stderr, "No visible canvas!\n");
	RETURN;
    }
    pixcanvas = canvas_get_pixcanvas (top->exportMenu.worldCanvas);
    kwin_get_size (pixcanvas, &width, &height);
    if ( ( offsets = m_alloc ( sizeof *offsets * (width + height) ) ) == NULL )
    {
	m_error_notify (function_name, "offset arrays");
	RETURN;
    }
    AUTO_FREE (m_free, offsets);
    for (count = 0; count < width; ++count) offsets[count] = 3 * count;
    for (count = 0; count < height; ++count)
	offsets[count + width] = 3 * count * width;
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_BACKGROUND_PIXEL, &background,
			 KWIN_ATT_END);
    if ( !kwin_get_pixel_RGB_values (pixcanvas, &background,
				     &red, &green, &blue, 1) )
    {
	fprintf (stderr, "%s: could not get background RGB value\n",
		 function_name);
	RETURN;
    }
    if ( ( rgbimage = m_alloc (width * height * 3) ) == NULL )
    {
	m_error_notify (function_name, "offset arrays");
	RETURN;
    }
    if ( (red == 0) && (green == 0) && (blue == 0) )
	m_clear (rgbimage, width * height * 3);
    else
    {
	colour[0] = (red >> 8) & 0xff;
	colour[1] = (green >> 8) & 0xff;
	colour[2] = (blue >> 8) & 0xff;
	m_fill (rgbimage, 3, colour, 3, width * height);
    }
    AUTO_FREE (m_free, rgbimage);
    if ( ( channel = ch_open_file (fname, "w") ) == NULL )
    {
	fprintf (stderr, "%s: error opening: \"%s\"\t%s\n",
		 function_name, fname, ERRSTRING);
	RETURN;
    }
    if ( !kwin_mem_write_rgb (pixcanvas, rgbimage, rgbimage + 1,
			      rgbimage + 2, 3, width, height) ||
	 !foreign_ppm_write_rgb (channel, TRUE, rgbimage, rgbimage + 1,
				 rgbimage + 2, offsets, offsets + width,
				 width, height, 0, 0,
				 NULL, NULL, NULL, 0) )
    {
	fprintf (stderr, "%s: error writing PPM file: \"%s\"\n",
		 function_name, fname);
	ch_close (channel);
	unlink (fname);
	RETURN;
    }
    ch_close (channel);
    fprintf (stderr, "Wrote PPM file: \"%s\"\n", fname);
    RETURN;
}   /*  End Function WritePPMWindow  */
