/*LINTLIBRARY*/
/*  Filepopup.c

    This code provides a popup file selector widget with close and
    rescan buttons, for Xt.

    Copyright (C) 1993-1994  Patrick Jordan
    Incorporated into Karma by permission.
    Copyright (C) 1994-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Patrick Jordan may be reached by email at  pjordan@rp.csiro.au
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating a popup file
    selector widget with close and rescan buttons, for Xt.


    Written by      Partick Jordan  9-JUL-1993

    Updated by      Richard Gooch   29-SEP-1993

    Updated by      Richard Gooch   8-DEC-1993: Added more patches from Patrick

    Updated by      Richard Gooch   17-JAN-1994: Added more patches from
  Patrick

    Updated by      Richard Gooch   23-OCT-1994: Changed some resource names.

    Updated by      Richard Gooch   14-NOV-1994: Changed some resource names.

    Updated by      Richard Gooch   18-NOV-1994: Added XkwNautoPopdown resource

    Updated by      Richard Gooch   24-DEC-1994: Propagated change of filename
  tester resource.

    Updated by      Richard Gooch   17-APR-1995: Fixed mix of XtRBool and
  sizeof Boolean in resource fields. Use Bool because it is of int size.

    Updated by      Richard Gooch   25-FEB-1996: Made default position top-left
  corner.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-FEB-1997: Added "pin" toggle.

    Updated by      Richard Gooch   20-APR-1997: Added XkwNloadFiles and
  XkwNdomainName resources.

    Updated by      Richard Gooch   13-MAY-1997: Took account of change to
  <XkwLoadAndDecimateGetCube> function and allocate fresh copy of
  domainName string.

    Updated by      Richard Gooch   6-AUG-1997: Added preferFloat resource.

    Updated by      Richard Gooch   15-AUG-1997: Only create filter toggle if
  XkwNloadFiles resource is set.

    Updated by      Richard Gooch   19-AUG-1997: Do not allow XkwNloadFiles
  resource to be changed after create.

    Updated by      Richard Gooch   10-OCT-1997: Added preliminary extension
  support.

    Updated by      Richard Gooch   1-MAR-1998: Changed to box widget for main
  layout. Added boxWidget resource.

    Updated by      Richard Gooch   2-MAR-1998: Created <rescan_cbk>.

    Updated by      Richard Gooch   5-MAR-1998: Take pinned state directly from
  pin toggle.

    Updated by      Richard Gooch   23-OCT-1998: Added "FilepopupWidget::Load"
  event.

    Updated by      Richard Gooch   10-JUN-1999: Added "writable" toggle.

    Updated by      Richard Gooch   28-JUN-1999: No longer call <rescan_cbk>
  on popup, as Filewin widget now does it (ensuring correct ordering).

    Updated by      Richard Gooch   2-AUG-1999: Added "unload" button.

    Updated by      Richard Gooch   13-DEC-1999: Added "pull" and "push"
  buttons.

    Updated by      Richard Gooch   3-JUN-2000: Added dirbutton config file
  support.

    Updated by      Richard Gooch   27-SEP-2000: Do not display "writable" or
  "unload" buttons if XkwNloadFiles is not TRUE.

    Updated by      Richard Gooch   20-MAR-2001: Created Destroy method.

    Last updated by Richard Gooch   25-SEP-2001: Removed #include for var-args.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/param.h>
#include <errno.h>
#include <Xkw/FilepopupP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>    
#include <X11/Xaw/List.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Box.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_xtmisc.h>
#include <karma_ext.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_c.h>
#include <karma_r.h>
#include <karma_a.h>
#include <Xkw/Filewin.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/LoadAndDecimate.h>


/*  Private functions  */
STATIC_FUNCTION (void _Filepopup_ClassInitialise, () );
STATIC_FUNCTION (void _Filepopup_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _Filepopup_Destroy, (Widget W) );
STATIC_FUNCTION (Boolean _Filepopup_SetValues,
		 (Widget current, Widget request, Widget new) );
STATIC_FUNCTION (void _Filepopup_filesel_cbk, (Widget w, XtPointer client_data,
					       XtPointer call_data) );
STATIC_FUNCTION (void _Filepopup_directory_change_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Filepopup_close_cbk, (Widget w, XtPointer client_data,
					     XtPointer call_data) );
STATIC_FUNCTION (void _Filepopup_rescan_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Filepopup_unload_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Filepopup_pull_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Filepopup_push_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag _Filepopup_push_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (flag _Filepopup_ds_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (void _Filepopup_dirbutton_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Filepopup_read_dirbuttons, (CONST char *filename) );


#define offset(field) XtOffset(FilepopupWidget, filepopup.field)
#define coreOffset(field) XtOffset(FilepopupWidget, core.field)

static XtResource FilepopupResources[] = 
{
    {XkwNautoPopdown, XkwCAutoPopdown, XtRBool, sizeof (Bool),
     offset (autoPopdown), XtRImmediate, False},
    {XkwNfileSelectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (fileSelectCallback), XtRCallback, (caddr_t) NULL},
    {XkwNfilenameTester, XtCCallback, XtRPointer, sizeof (caddr_t),
     offset (accept_file), XtRPointer, (caddr_t) NULL},
    {XkwNloadFiles, XkwCLoadFiles, XtRBool, sizeof (Bool),
     offset (loadFiles), XtRImmediate, FALSE},
    {XkwNpreferFloat, XkwCPreferFloat, XtRBool, sizeof (Bool),
     offset (preferFloat), XtRImmediate, FALSE},
    {XkwNdomainName, XkwCDomainName, XtRString, sizeof (String),
     offset (domainName), XtRImmediate, "unknown"},
    {XkwNboxWidget, XkwCBoxWidget, XtRWidget, sizeof (Widget),
     offset (boxWidget), XtRImmediate, NULL},
    {XkwNfilewinWidget, XkwCFilewinWidget, XtRWidget, sizeof (Widget),
     offset (filewinWidget), XtRImmediate, NULL},
    {XtNx, XtCPosition, XtRPosition, sizeof (Position),
     coreOffset (x), XtRPosition, 0},
    {XtNy, XtCPosition, XtRPosition, sizeof (Position),
     coreOffset (y), XtRPosition, 0},
};

#undef offset

static char last_directory[MAXPATHLEN + 1];
static KCallbackList push_list = NULL;

struct push_info
{
    CONST char *dirname;
    void *sender;
};

struct dirbutton_entry_struct
{
    struct dirbutton_entry_struct *next;
    char *directory;  /*  Pointer into directory portion of label     */
    char label[1];    /*  Overallocation to hold label and directory  */
};


static struct dirbutton_entry_struct *first_dirbutton = NULL;
static struct dirbutton_entry_struct *last_dirbutton = NULL;

FilepopupClassRec filepopupClassRec = 
{
  {        /* CoreClassPart */
    (WidgetClass)&topLevelShellClassRec,  /* superclass */
    "Filepopup",                   /* class_name */
    sizeof(FilepopupRec),          /* widget_size */
    _Filepopup_ClassInitialise,    /* class_initialise */
    NULL,                          /* class_part_initialise */
    FALSE,                         /* class_init */
    (XtInitProc)_Filepopup_Initialise,        /* initialise */
    NULL,                          /* initialise_hook */
    XtInheritRealize,              /* realize */
    NULL,                          /* actions */
    0,                             /* num_actions */
    FilepopupResources,            /* resources */
    XtNumber(FilepopupResources),  /* num_resources */
    NULLQUARK,                     /* xrm_class */
    TRUE,                          /* compress_motion */
    TRUE,                          /* compress_exposure */
    TRUE,                          /* compress_enterleave */
    TRUE,                          /* visible_interest */
    _Filepopup_Destroy,            /* destroy */
    XtInheritResize,               /* resize */
    NULL,                          /* expose */
    (XtSetValuesFunc)_Filepopup_SetValues,    /* set_values */
    NULL,                          /* set_values_hook */
    XtInheritSetValuesAlmost,      /* set_values_almost */
    NULL,                          /* get_values_hook */
    NULL,                          /* accept_focus */
    XtVersion,                     /* version */
    NULL,                          /* callback_private */
    NULL,                          /* tm_translations */
    NULL,                          /* query_geometry  */
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* FilepopupClassPart */
    0 /* empty */
  }
};

WidgetClass filepopupWidgetClass = (WidgetClass) &filepopupClassRec;


static void _Filepopup_ClassInitialise ()
{
    CONST char *home, *karmabase;
    char config_file[STRING_LENGTH];

    /*  Now load extra directory-changing buttons configured by the admin and
	user  */
    if ( ( karmabase = r_get_karmabase () ) != NULL )
    {
	sprintf (config_file, "%s/share/browser-dirbuttons", karmabase);
	_Filepopup_read_dirbuttons (config_file);
	sprintf (config_file, "%s/site/share/browser-dirbuttons", karmabase);
	_Filepopup_read_dirbuttons (config_file);
    }
    if ( ( home = r_getenv ("HOME") ) != NULL )
    {
	sprintf (config_file, "%s/.karma/browser-dirbuttons", home);
	_Filepopup_read_dirbuttons (config_file);
    }
}   /*  End Function ClassInitialise  */

static void _Filepopup_Initialise (Widget Request, Widget New)
{
    /*FilepopupWidget request = (FilepopupWidget) Request;*/
    FilepopupWidget new = (FilepopupWidget) New;
    Dimension width, max_width;
    Widget form, box, w, rescan_btn;
    struct dirbutton_entry_struct *dirbutton;
    static char function_name[] = "Filepopup::Initialise";

    new->shell.allow_shell_resize = True;
    if (new->filepopup.domainName != NULL)
    {
	if ( ( new->filepopup.domainName =
	       st_dup (new->filepopup.domainName) ) == NULL )
	{
	    m_abort (function_name, "domainName");
	}
    }
    w = XtVaCreatePopupShell ("loadAndDecimatePopup",
			      loadAndDecimateWidgetClass, New,
			      XtNtitle, "Load And Decimate Control",
			      XtNx, 0,
			      XtNy, 0,
			      NULL);
    new->filepopup.load_and_decimate_popup = w;
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    XtNdefaultDistance, 0,
				    NULL);
    new->filepopup.filewinWidget = XtVaCreateManagedWidget
	("selector", filewinWidgetClass, form,
	 XkwNfilenameTester, new->filepopup.accept_file,
	 XtNdefaultDistance, 0,
	 XtNleft, XtChainLeft,
	 XtNright, XtChainRight,
	 XtNborderWidth, 0,
	 NULL);
    XtAddCallback (new->filepopup.filewinWidget, XkwNfileSelectCallback,
		   _Filepopup_filesel_cbk, new);
    XtAddCallback (new->filepopup.filewinWidget, XkwNdirectoryChangeCallback,
		   _Filepopup_directory_change_cbk, new);
    w = XtNameToWidget (new->filepopup.filewinWidget, "view.list");
    XtVaGetValues (w, XtNwidth, &max_width, NULL);
    w = XtNameToWidget (new->filepopup.filewinWidget, "dirname");
    XtVaGetValues (w, XtNwidth, &width, NULL);
    if (width > max_width) max_width = width;
    box = XtVaCreateManagedWidget ("box", boxWidgetClass, form,
				   XtNfromVert, new->filepopup.filewinWidget,
				   XtNwidth, max_width + 10,
				   XtNborderWidth, 0,
				   NULL);
    new->filepopup.boxWidget = box;
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, box,
				 XtNlabel, "close",
				 XtNhorizDistance, 5,
				 XtNvertDistance, 5,
				 /*XtNwidth, 52,*/
				 /*XtNheight, 48,*/
				 NULL);
    XtAddCallback (w, XtNcallback, _Filepopup_close_cbk, new);
    w = XtVaCreateManagedWidget ("rescanButton", commandWidgetClass, box,
				 XtNlabel, "rescan",
				 XtNhorizDistance, 5,
				 XtNvertDistance, 5,
				 /*XtNwidth, 52,*/
				 /*XtNheight, 48,*/
				 NULL);
    rescan_btn = w;
    XtAddCallback (rescan_btn, XtNcallback, _Filepopup_rescan_cbk,
		   new->filepopup.filewinWidget);
    if (new->filepopup.autoPopdown)
    {
	w = XtVaCreateManagedWidget ("pinToggle", ktoggleWidgetClass, box,
				     XtNlabel, "pin ",
				     XtNhorizDistance, 5,
				     XtNvertDistance, 5,
				     /*XtNwidth, 52,*/
				     /*XtNinternalHeight, 16,*/
				     XkwNcrosses, FALSE,
				     NULL);
	new->filepopup.pin_tgl = w;
    }
    if (new->filepopup.loadFiles)
    {
	w = XtVaCreateManagedWidget ("filterToggle", ktoggleWidgetClass, box,
				     XtNlabel, "filter",
				     XtNhorizDistance, 5,
				     XtNvertDistance, 5,
				     /*XtNwidth, 52,*/
				     /*XtNinternalHeight, 16,*/
				     XkwNcrosses, FALSE,
				     NULL);
	new->filepopup.filter_tgl = w;
	w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, box,
				     XtNlabel, "writable",
				     XtNhorizDistance, 5,
				     XtNvertDistance, 5,
				     /*XtNwidth, 52,*/
				     /*XtNinternalHeight, 16,*/
				     XkwNcrosses, FALSE,
				     NULL);
	new->filepopup.writable_tgl = w;
    }
    if (new->filepopup.loadFiles)
    {
	w = XtVaCreateManagedWidget ("unloadButton", commandWidgetClass, box,
				     XtNlabel, "unload",
				     XtNhorizDistance, 5,
				     XtNvertDistance, 5,
				     /*XtNwidth, 52,*/
				     /*XtNheight, 48,*/
				     NULL);
	XtAddCallback (w, XtNcallback, _Filepopup_unload_cbk, new);
    }
    w = XtVaCreateManagedWidget ("pullButton", commandWidgetClass, box,
				 XtNlabel, "pull",
				 XtNhorizDistance, 5,
				 XtNvertDistance, 5,
				 /*XtNwidth, 52,*/
				 /*XtNheight, 48,*/
				 NULL);
    XtAddCallback (w, XtNcallback, _Filepopup_pull_cbk, new);
    w = XtVaCreateManagedWidget ("pushButton", commandWidgetClass, box,
				 XtNlabel, "push",
				 XtNhorizDistance, 5,
				 XtNvertDistance, 5,
				 /*XtNwidth, 52,*/
				 /*XtNheight, 48,*/
				 NULL);
    XtAddCallback (w, XtNcallback, _Filepopup_push_cbk, new);
    c_register_callback (&push_list, _Filepopup_push_callback, New,
			 NULL, FALSE, NULL, FALSE, FALSE);
    new->filepopup.ds_event_callback =
	ds_event_register_func (_Filepopup_ds_event_func, New);
    for (dirbutton = first_dirbutton; dirbutton != NULL;
	 dirbutton = dirbutton->next)
    {
	w = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
				     XtNlabel, dirbutton->label,
				     NULL);
	XtAddCallback (w, XtNcallback, _Filepopup_dirbutton_cbk,
		       (XtPointer) dirbutton->directory);
    }
    xtmisc_ext_call (New, "WidgetInitialise", Request, New);
}   /*  End Function Initialise  */

static void _Filepopup_Destroy (Widget W)
{
    FilepopupWidget w = (FilepopupWidget) W;

    c_unregister_callback (w->filepopup.ds_event_callback);
    if (w->filepopup.domainName != NULL) m_free (w->filepopup.domainName);
}   /*  End Function Destroy  */

static Boolean _Filepopup_SetValues (Widget Current, Widget Request,Widget New)
{
    FilepopupWidget current = (FilepopupWidget) Current;
    /*FilepopupWidget request = (FilepopupWidget) Request;*/
    FilepopupWidget new = (FilepopupWidget) New;
    static char function_name[] = "Filepopup::SetValues";

    if (new->filepopup.loadFiles != current->filepopup.loadFiles)
    {
	fprintf (stderr, "%s: change of XkwNloadFiles resource ignored\n",
		 function_name);
	new->filepopup.loadFiles = current->filepopup.loadFiles;
    }
    if (new->filepopup.domainName != current->filepopup.domainName)
    {
	if (current->filepopup.domainName != NULL)
	    m_free (current->filepopup.domainName);
	if ( ( new->filepopup.domainName =
	       st_dup (new->filepopup.domainName) ) == NULL )
	{
	    m_abort (function_name, "domainName");
	}
    }
    if (new->filepopup.accept_file != current->filepopup.accept_file)
    {
	XtVaSetValues (new->filepopup.filewinWidget,
		       XkwNfilenameTester, new->filepopup.accept_file,
		       NULL);
    }
    return False;
}   /*  End Function SetValues  */

static void _Filepopup_filesel_cbk (Widget w, XtPointer client_data,
				    XtPointer call_data)
{
    unsigned int ftype;
    Bool pinned;
    Widget FW = (Widget) client_data;
    FilepopupWidget fw = (FilepopupWidget) FW;
    CONST char *filename = (char *) call_data;
    multi_array *multi_desc;
    static char function_name[] = "Filepopup::filesel_cbk";

    if (fw->filepopup.autoPopdown)
    {
	XtVaGetValues (fw->filepopup.pin_tgl, XtNstate, &pinned, NULL);
	if (!pinned)
	{
	    XtPopdown (FW);
	    XSync (XtDisplay (FW), False);
	}
    }
    if (fw->filepopup.loadFiles)
    {
	Bool filter, writable;

	XtVaGetValues (fw->filepopup.filter_tgl, XtNstate, &filter, NULL);
	XtVaGetValues (fw->filepopup.writable_tgl, XtNstate, &writable, NULL);
	if (filter)
	{
	    XkwLoadAndDecimateGetCube (fw->filepopup.load_and_decimate_popup,
				       filename, fw->filepopup.domainName);
	    return;
	}
	if ( ( multi_desc = foreign_guess_and_read(filename, K_CH_MAP_LOCAL,
						   writable, &ftype,
						   FA_GUESS_READ_FITS_TO_FLOAT,
						   fw->filepopup.preferFloat,
						   FA_GUESS_READ_END) )
	     == NULL )
	{
	    fprintf (stderr, "%s: error reading file: \"%s\"\n",
		     function_name, filename);
	    return;
	}
	ds_event_dispatch (multi_desc, fw->filepopup.domainName, filename);
	ds_dealloc_multi (multi_desc);
	return;
    }
    XtCallCallbacks (FW, XkwNfileSelectCallback, call_data);
}   /*  End Function _Filepopup_filesel_cbk  */

static void _Filepopup_directory_change_cbk (Widget w, XtPointer client_data,
					     XtPointer call_data)
{
    CONST char *dirname = call_data;

    strcpy (last_directory, dirname);
}   /*  End Function _Filepopup_directory_change_cbk  */

static void _Filepopup_close_cbk (Widget w, XtPointer client_data,
				  XtPointer call_data)
{
    FilepopupWidget fw = (FilepopupWidget) client_data;

    XtPopdown ( (Widget) fw );
}   /*  End Function close_cbk  */

static void _Filepopup_rescan_cbk (Widget w, XtPointer client_data,
				   XtPointer call_data)
{
    XkwFilewinRescan (client_data);
}   /*  End Function rescan_cbk  */

static void _Filepopup_unload_cbk (Widget w, XtPointer client_data,
				   XtPointer call_data)
{
    FilepopupWidget fw = (FilepopupWidget) client_data;

    ds_event_dispatch (NULL, fw->filepopup.domainName, "UNLOAD");
}   /*  End Function unload_cbk  */

static void _Filepopup_pull_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
{
    FilepopupWidget fw = (FilepopupWidget) client_data;

    if (last_directory[0] == '\0')
    {
	XBell (XtDisplay (w), 100);
	return;
    }
    XkwFilewinChangeDirectory (fw->filepopup.filewinWidget, last_directory);
}   /*  End Function pull_cbk  */

static void _Filepopup_push_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
{
    FilepopupWidget fw = (FilepopupWidget) client_data;
    struct push_info data;

    data.dirname = XkwFilewinCurrentDirectory (fw->filepopup.filewinWidget);
    data.sender = client_data;
    c_call_callbacks (push_list, &data);
}   /*  End Function push_cbk  */

static flag _Filepopup_push_callback (void *object, void *client1_data,
				      void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when a directory name is pushed.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    struct push_info *data = call_data;
    FilepopupWidget fw = (FilepopupWidget) object;

    if (data->sender == object) return (FALSE);
    XkwFilewinChangeDirectory (fw->filepopup.filewinWidget, data->dirname);
    return (FALSE);
}   /*  End Function push_callback  */

static flag _Filepopup_ds_event_func (void *object, multi_array *multi_desc,
				      CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    FilepopupWidget top = (FilepopupWidget) object;
    packet_desc *pack_desc;
    char *packet, *d_domain, *d_filename;

    if (strcmp (name, "FilepopupWidget::Load") != 0) return (FALSE);
    if (top->filepopup.domainName == NULL) return (FALSE);
    if (multi_desc->num_arrays != 1) return (FALSE);
    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0];
    if (pack_desc->num_elements != 2) return (FALSE);
    if ( ( d_domain = ds_get_unique_named_string (pack_desc, packet,
						  "domain") ) == NULL )
	return (FALSE);
    if (strcmp (top->filepopup.domainName, d_domain) != 0)
    {
	m_free (d_domain);
	return (FALSE);
    }
    m_free (d_domain);
    if ( ( d_filename = ds_get_unique_named_string (pack_desc, packet,
						    "filename") ) == NULL )
	return (FALSE);
    _Filepopup_filesel_cbk (NULL, top, d_filename);
    m_free (d_filename);
    return (TRUE);
}   /*  End Function _Filepopup_ds_event_func  */

static void _Filepopup_dirbutton_cbk (Widget w, XtPointer client_data,
				      XtPointer call_data)
{
    Widget filewin, parent, filepopup;
    CONST char *dirname = (CONST char *) client_data;
    static char function_name[] = "Filepopup::dirbutton_cbk";

    for ( parent = XtParent (w), filepopup = NULL;
	  parent != NULL;
	  parent = XtParent (parent) )
	if ( XtIsFilepopup (parent) )
	{
	    filepopup = parent;
	    break;
	}
    if (filepopup == NULL)
    {
	fprintf (stderr, "Could not find a Filepopup ancestor\n");
	a_prog_bug (function_name);
    }
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XkwFilewinChangeDirectory (filewin, dirname);
}   /*  End Function dirbutton_cbk  */

static void _Filepopup_read_dirbuttons (CONST char *filename)
{
    Channel channel;
    char line[STRING_LENGTH];
    static char function_name[] = "_ext_read_config";

    if (access (filename, R_OK) != 0) return;
    if ( ( channel = ch_open_file (filename, "r") ) == NULL )
    {
	fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		 function_name, filename, ERRSTRING);
	return;
    }
    while ( chs_get_line2 (channel, line, STRING_LENGTH, '#') )
    {
	int len_label, len_directory;
	struct dirbutton_entry_struct *new;
	char label[STRING_LENGTH], directory[STRING_LENGTH];

	sscanf (line, "%s %s", label, directory);
	len_label = strlen (label);
	len_directory = strlen (directory);
	if ( ( new = m_alloc (sizeof *new + len_label + len_directory + 1) )
	     == NULL )
	    m_abort (function_name, "dirbutton entry");
	new->directory = new->label + len_label + 1;
	strcpy (new->label, label);
	strcpy (new->directory, directory);
	new->next = NULL;
	if (first_dirbutton == NULL) first_dirbutton = new;
	if (last_dirbutton != NULL) last_dirbutton->next = new;
	last_dirbutton = new;
    }
    ch_close (channel);
}   /*  End Function read_dirbuttons  */
