/*LINTLIBRARY*/
/*  Filewin.c

    This code provides a file selector widget for Xt.

    Original code copyright (C) 1993-1994    Patrick Jordan
    Incorporated into Karma by permission.
    Subsequent code copyright (C) 1994-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Patrick Jordan may be reached by email at  patrick@ariel.com.au
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating a file selector
    widget for Xt.


    Written by      Patrick Jordan  7-JUL-1993

    Updated by      Richard Gooch   29-SEP-1993

    Updated by      Richard Gooch   8-DEC-1993: Added more patches from Patrick

    Updated by      Richard Gooch   17-JAN-1994: Added more patches from
  Patrick

    Updated by      Richard Gooch   28-JUL-1994: Added  #include <karma_m.h>

    Updated by      Richard Gooch   30-JUL-1994: Removed height setting of
  viewport: this should be done externally.

    Updated by      Richard Gooch   5-OCT-1994: Worked around bug in AIX/rs6000
  native compiler.

    Updated by      Richard Gooch   28-SEP-1995: Added directory callbacks.

    Updated by      Richard Gooch   20-OCT-1995: Added minimum list height.

    Updated by      Richard Gooch   3-MAY-1996: Made widget 60 pixels wider.

    Updated by      Richard Gooch   5-MAY-1996: Improved documentation.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-JUN-1996: Created
  <XkwFilewinStandardFileTester_nD> and <XkwFilewinStandardFileTester_3D>
  routines.

    Updated by      Richard Gooch   24-JUN-1996: Fixed
  <XkwFilewinStandardFileTester_nD> and <XkwFilewinStandardFileTester_3D>
  routines to catenate directory name with filename. Required for properly
  testing GIPSY files.

    Updated by      Richard Gooch   25-OCT-1996: Limit height so that it fits
  on the screen.

    Updated by      Richard Gooch   27-OCT-1996: Strip unneccessary components
  from directory names.

    Updated by      Richard Gooch   30-DEC-1996: Created
  <XkwNforwardSyntheticEvents> resource.

    Updated by      Richard Gooch   12-JAN-1997: Created
  <XkwNtrapDirectoryDatasets> resource.

    Updated by      Richard Gooch   16-JAN-1997: Took account of change to
  <gls_register_plain_event_handler>.

    Updated by      Richard Gooch   18-JUL-1997: Supported DRAO datasets.

    Updated by      Richard Gooch   24-JUL-1997: Set width according to widest
  entry at startup.

    Updated by      Richard Gooch   30-JUL-1997: Supported AIPS datasets.

    Updated by      Richard Gooch   8-AUG-1997: No longer show unreadable
  entries or unexecutable directories in file list.

    Updated by      Richard Gooch   3-OCT-1997: Remove leading "./" when
  constructing full pathnames.

    Updated by      Richard Gooch   6-OCT-1997: Added text widget to display
  and change current directory.

    Updated by      Richard Gooch   8-OCT-1997: Switched to MAXPATHLEN because
  mips2_IRIX6 scribbles beyond the '\0' terminator, and MAXPATHLEN=1024 is
  greater than STRING_LENGTH=255.

    Updated by      Richard Gooch   9-OCT-1997: Made use of new experimental
  <st_expr_expand> function.

    Updated by      Richard Gooch   10-OCT-1997: Created function
  <XkwFilewinChangeDirectory>.

    Updated by      Richard Gooch   17-OCT-1997: Added test for filtered
  directory datasets. Fixed bug when going up a directory. Added test for
  AIPS++ directory datasets.

    Updated by      Richard Gooch   2-MAR-1998: Changed <XkwFilewinRescan>
  interface.

    Updated by      Richard Gooch   12-MAR-1998: Added PGM format to exclusion
  list for <XkwFilewinStandardFileTester_3D>. Made use of new
  <foreign_miriad_test2> and changed <foreign_aips2_test> functions.

    Updated by      Richard Gooch   30-JAN-1999: Added support for split Karma
  files (".desc" and ".dat" pairs).

    Updated by      Richard Gooch   25-MAR-1999: Fixed support for compressed
  Karma files broken in previous change.

    Updated by      Richard Gooch   8-JUN-1999: Took account of change to
  <st_expr_expand>.

    Updated by      Richard Gooch   28-JUN-1999: Added support for inode write
  event notification using native <dm> interface.

    Updated by      Richard Gooch   19-AUG-1999: Changed sequence number
  encoding for DRAO datasets and stripped directory component from catalogue
  name.

    Updated by      Richard Gooch   26-AUG-1999: Created <XkwNnumLines>
  resource. Took account of change to DRAO dataset name list.

    Updated by      Richard Gooch   30-AUG-1999: Removed setting of directory
  name resource, because text widget does not recompute height.

    Updated by      Richard Gooch   13-SEP-1999: Adjusted interval and made it
  OS-dependent (can't expect other OS'es to be as efficient as Linux).

    Updated by      Richard Gooch   5-OCT-1999: Renamed all private functions.
  Fixed bug in <Initialise>: text_widget was not initialised before function
  <XkwFilewinChangeDirectory> called.

    Updated by      Richard Gooch   13-DEC-1999: Created callback resource
  <XkwNdirectoryChangeCallback>.

    Updated by      Richard Gooch   23-APR-2000: Documentation fix.

    Updated by      Richard Gooch   20-MAR-2001: Created Destroy method.

    Updated by      Richard Gooch   21-MAR-2001: Removed directory notification
  code since Linux uses signals instead of my proposed poll(2) implementation.

    Updated by      Richard Gooch   3-OCT-2001: Created Realise method and
  remove trailing empty strings. Only create empty strings in <create_list> if
  not yet realised.

    Updated by      Richard Gooch   9-OCT-2001: Fixed <text_cbk> to not use
  <<dirname>> if <st_expr_expand> fails.

    Updated by      Richard Gooch   12-OCT-2001: Added marker character for
  last loaded file.

    Updated by      Richard Gooch   14-OCT-2001: Fixed bug introduced by adding
  marker character.

    Updated by      Richard Gooch   1-NOV-2001: Clear old marker before writing
  new marker, otherwise repeat loading of the same file loses the marker.

    Last updated by Richard Gooch   7-NOV-2001: Fixed bug in <filesel_cbk>
  which didn't test for CWD="/" when creating relative path to CWD.


*/
#include <stdio.h>
#include <math.h>
#include <sys/param.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <pwd.h>
#include <errno.h>
#include <sys/types.h>
#include <Xkw/FilewinP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Shell.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_dir.h>
#include <karma_gls.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_c.h>
#include <karma_a.h>
#include <karma_r.h>


/*  EMERGENCY HACK FOR OSF1: the version on yeti doesn't support realpath(3) */
#ifdef OS_OSF1
#  define realpath(a,b) NULL
#endif

/*  AIX 3.2 does not have realpath(3)  */
#ifdef OS_AIX
#  define realpath(a,b) NULL
#endif

#define MINIMUM_LIST_ROWS      20
#define FILENAME_OFFSET        5
#define STRING_COMPARE_LENGTH  4
#define MARKER_OFFSET          4
#define MARKER_CHARACTER       '*'
#define STRING_DIR             "D    "
#define STRING_FILE            "F    "
#define STRING_DRAO            "DRAO "
#define STRING_AIPS            "AIPS "


/*  Methods  */
STATIC_FUNCTION (void _Filewin_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _Filewin_Realise, (Widget W, Mask *valueMask,
					 XSetWindowAttributes *attributes) );
STATIC_FUNCTION (void _Filewin_Destroy, (Widget W) );
STATIC_FUNCTION (Boolean _Filewin_SetValues,
		 (Widget current, Widget request, Widget new) );
STATIC_FUNCTION (void _Filewin_increase_list_size, (FilewinWidget w) );
STATIC_FUNCTION (void _Filewin_create_list, (FilewinWidget w) );
STATIC_FUNCTION (void _Filewin_filesel_cbk, (Widget w, XtPointer client_data,
					     XtPointer call_data) );
STATIC_FUNCTION (flag _Filewin_clean_dirname, (char dirname[MAXPATHLEN]) );
STATIC_FUNCTION (flag _Filewin_dir_func,
		 (FilewinWidget w, CONST char *dirname) );
STATIC_FUNCTION (flag _Filewin_glish_event_func,
		 (void *object, CONST char *name,
		  CONST char *data, unsigned int type,
		  unsigned int num_elements,
		  unsigned int num_axes, CONST int *axis_lengths) );
STATIC_FUNCTION (void _Filewin_text_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Filewin_popup_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _Filewin_popdown_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


#define TheOffset(field) XtOffset(FilewinWidget, filewin.field)

static XtResource FilewinResources[] = 
{
    {XkwNfileSelectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     TheOffset (fileSelectCallback), XtRCallback, (caddr_t) NULL},
    {XkwNdirectoryChangeCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     TheOffset (directoryChangeCallback), XtRCallback, (caddr_t) NULL},
    {XkwNfilenameTester, XtCCallback, XtRPointer, sizeof (caddr_t),
     TheOffset (accept_file), XtRPointer, (caddr_t) NULL},
    {XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
     TheOffset (foreground), XtRString, XtDefaultForeground},
    {XkwNnumLines, XkwCNumLines, XtRInt, sizeof (int),
     TheOffset (numLines), XtRImmediate, 0},
    {XkwNforwardSyntheticEvents, XkwCForwardSyntheticEvents, XtRBool,
     sizeof (Bool), TheOffset (forwardSyntheticEvents), XtRImmediate, False},
    {XkwNtrapDirectoryDatasets, XkwCTrapDirectoryDatasets, XtRBool,
     sizeof (Bool), TheOffset (trapDirectoryDatasets), XtRImmediate, False},
};

#undef TheOffset

FilewinClassRec filewinClassRec = 
{
  {        /* CoreClassPart */
    /* superclass */              (WidgetClass) &formClassRec,
    /* class_name */              "Filewin",
    /* widget_size */             sizeof (FilewinRec),
    /* class_initialise */        NULL,                          
    /* class_part_initialise */   NULL,                          
    /* class_init */              FALSE,                         
    /* initialise */              (XtInitProc) _Filewin_Initialise,        
    /* initialise_hook */         NULL,                          
    /* realize */                 _Filewin_Realise,              
    /* actions */                 NULL,                          
    /* num_actions */             0,                             
    /* resources */               FilewinResources,              
    /* num_resources */           XtNumber (FilewinResources),
    /* xrm_class */               NULLQUARK,                     
    /* compress_motion */         TRUE,                          
    /* compress_exposure */       TRUE,                          
    /* compress_enterleave */     TRUE,                          
    /* visible_interest */        TRUE,                          
    /* destroy */                 _Filewin_Destroy,
    /* resize */                  XtInheritResize,
    /* expose */                  NULL,
    /* set_values */              (XtSetValuesFunc) _Filewin_SetValues,
    /* set_values_hook */         NULL,
    /* set_values_almost */       XtInheritSetValuesAlmost,      
    /* get_values_hook */         NULL,                          
    /* accept_focus */            NULL,                          
    /* version */                 XtVersion,                     
    /* callback_private */        NULL,                          
    /* tm_translations */         NULL,                          
				  NULL,
				  NULL,
				  NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Constraint */
    /* subresourses       */   NULL,
    /* subresource_count  */   0,
    /* constraint_size    */   sizeof (FilewinConstraintsRec),
    /* initialise         */   NULL,
    /* destroy            */   NULL,
    /* set_values         */   NULL,
    /* extension          */   NULL
  },{ /* Form */
    /* layout */                XtInheritLayout
  },
  {  /* FilewinClassPart */
    0 /* empty */
  }
};

WidgetClass filewinWidgetClass = (WidgetClass) &filewinClassRec;

static void _Filewin_Initialise (Widget Request, Widget New)
{
    unsigned int width, height, screen_height, max_length, count, len;
    /*FilewinWidget request = (FilewinWidget) Request;*/
    FilewinWidget new = (FilewinWidget) New;
    Widget text, view, w, parent = NULL;  /*  Keep gcc happy  */
    Widget shell = NULL;
    static flag first_time = TRUE;
    static char function_name[] = "Filewin::Initialise";

    if (first_time)
    {
	XtAppContext app_context;

	first_time = FALSE;
	app_context = XtWidgetToApplicationContext (New);
    }
    /* Initialise the directory list space */
    new->filewin.listmax = 64;
    new->filewin.list = m_alloc (sizeof (char *) * new->filewin.listmax);
    new->filewin.listcount = 0;
    new->filewin.last_loaded = -1;
    new->filewin.dir_callbacks = NULL;
    new->filewin.synthetic_func = NULL;
    new->filewin.dir_kdir = NULL;
    new->filewin.dir_fd = -1;
    new->filewin.text_widget = NULL;
    if ( !XkwFilewinChangeDirectory (New, ".") )
    {
	fprintf (stderr, "Error initialising Filewin widget\n");
	exit (RV_UNDEF_ERROR);
    }
    strcpy (new->filewin.orig_dir, new->filewin.curdir);
    for ( parent = XtParent (New); parent != NULL; parent = XtParent (parent) )
    {
	if ( XtIsShell (parent) )
	{
	    shell = parent;
	    break;
	}
    }
    if (shell == NULL)
    {
	fprintf (stderr, "No Shell (sub)class ancestor!\n");
	a_prog_bug (function_name);
    }
    if ( !XtIsApplicationShell (shell) )
    {
	_Filewin_popdown_cbk (NULL, New, NULL);
	XtAddCallback (shell, XtNpopupCallback, _Filewin_popup_cbk, New);
	XtAddCallback (shell, XtNpopdownCallback, _Filewin_popdown_cbk, New);
    }
    screen_height = HeightOfScreen ( XtScreen (New) );
    if (new->filewin.numLines > 0)
    {
	height = new->filewin.numLines * 15 + 140;
	if (height > screen_height) height = screen_height - 140;
    }
    else
    {
	if (new->filewin.listcount * 15 + 140 > screen_height)
	{
	    /*  Files will probably not fit on screen: limit height to a little
		less than screen height  */
	    height = screen_height - 140;
	}
	else height = 0;
    }
    /*  Crude estimate of the required width  */
    max_length = strlen (new->filewin.curdir);
    for (count = 0; count < new->filewin.listcount; ++count)
    {
	len = strlen (new->filewin.list[count]);
	if (len > max_length) max_length = len;
    }
    width = 10 * max_length;
    if (width < 300) width = 300;
    text = XtVaCreateManagedWidget ("dirname", asciiTextWidgetClass, New,
				    XtNborderWidth, 0,
				    XtNdefaultDistance, 0,
				    XtNwidth, width,
				    XtNeditType, XawtextEdit,
				    XtNstring, new->filewin.curdir,
				    /*XtNresize, XawtextResizeWidth,*/
				    NULL);
    new->filewin.text_widget = text;
    XtVaGetValues (text, XtNtextSource, &w, NULL);
    XtAddCallback (w, XtNcallback, _Filewin_text_cbk, New);
    view = XtVaCreateManagedWidget ("view", viewportWidgetClass, New,
				    XtNforceBars, True,
				    XtNwidth, width,
				    XtNheight, height,
				    XtNfromVert, text,
				    XtNuseRight, True,
				    XtNallowVert, True,
				    XtNborderWidth, 1,
				    NULL);
    new->filewin.listwidget = XtVaCreateManagedWidget
	("list", listWidgetClass, view,
	 XtNlist, new->filewin.list,
	 XtNnumberStrings, new->filewin.listcount,
	 XtNbackground, new->core.background_pixel,
	 XtNforeground, new->filewin.foreground,
	 XtNdefaultColumns, 1,
	 XtNforceColumns, True,
	 NULL);
    XtAddCallback (new->filewin.listwidget, XtNcallback, _Filewin_filesel_cbk,
		   new);
    if (new->filewin.forwardSyntheticEvents)
    {
	/*  Add handler  */
	new->filewin.synthetic_func =
	    gls_register_plain_event_handler (_Filewin_glish_event_func,
					      "displayFitsFile", new);
    }
}   /*  End Function Initialise  */

static void _Filewin_Realise (Widget W, Mask *valueMask,
			      XSetWindowAttributes *attributes)
{
    FilewinWidget w = (FilewinWidget) W;

    /*  Call the superclass realise method, which will size everything  */
    (*filewinWidgetClass->core_class.superclass->core_class.realize)
	(W, valueMask, attributes);
    /*  Now remove trailing empty lines which are ugly  */
    while (strlen (w->filewin.list[w->filewin.listcount - 1]) == 0)
    {
	m_free (w->filewin.list[w->filewin.listcount - 1]);
	--w->filewin.listcount;
    }
    XawListChange (w->filewin.listwidget, w->filewin.list,
		   w->filewin.listcount, 0, True);
}   /*  End Function Realise  */

static void _Filewin_Destroy (Widget W)
{
    int count;
    FilewinWidget w = (FilewinWidget) W;

    for (count = 0; count < w->filewin.listcount; ++count)
	m_free (w->filewin.list[count]);
    m_free (w->filewin.list);
    if (w->filewin.dir_kdir != NULL) dir_close (w->filewin.dir_kdir);
    if (w->filewin.synthetic_func != NULL)
	c_unregister_callback (w->filewin.synthetic_func);
    c_destroy_list (w->filewin.dir_callbacks);
    if (w->filewin.dir_fd >= 0) close (w->filewin.dir_fd);
}   /*  End Function Destroy  */

static Boolean _Filewin_SetValues (Widget Current, Widget Request, Widget New)
{
    FilewinWidget current = (FilewinWidget) Current;
    /*FilewinWidget request = (FilewinWidget) Request;*/
    FilewinWidget new = (FilewinWidget) New;

    if (new->core.background_pixel != current->core.background_pixel ||
	new->filewin.foreground != current->filewin.foreground)
    {
	XtVaSetValues (new->filewin.listwidget,
		       XtNbackground, new->core.background_pixel,
		       XtNforeground, new->filewin.foreground,
		       NULL);
    }
    if ( new->filewin.forwardSyntheticEvents &&
	 (new->filewin.synthetic_func == NULL) )
    {
	/*  Add handler  */
	new->filewin.synthetic_func =
	    gls_register_plain_event_handler (_Filewin_glish_event_func,
					      "displayFitsFile", new);
    }
    if ( !new->filewin.forwardSyntheticEvents &&
	 (new->filewin.synthetic_func != NULL) )
    {
	/*  Remove handler  */
	c_unregister_callback (new->filewin.synthetic_func);
	new->filewin.synthetic_func = NULL;
    }
    return (True);
}   /*  End Function SetValues  */

static void _Filewin_increase_list_size (FilewinWidget w)
{
    int i;
    char **new;

    w->filewin.listmax *= 2;
    new = m_alloc ( w->filewin.listmax * sizeof (char *) );
    for (i = 0; i < w->filewin.listcount; i++) new[i] = w->filewin.list[i];
    m_free (w->filewin.list);
    w->filewin.list = new;
}   /*  End Function increase_list_size  */

static void _Filewin_create_list (FilewinWidget w)
{
#define W w->filewin
    flag show_aips, show_drao;
    unsigned int num_datasets, count, len;
    unsigned long raw_entry_count = 0; /* Initialised to keep compiler happy */
    time_t start_time = 0;             /* Initialised to keep compiler happy */
    time_t current_time = 0;           /* Initialised to keep compiler happy */
    KFileInfo *finfo;
    char *ptr;
    CONST char *files;
    static flag first_time = TRUE;
    static char function_name[] = "Filewin::create_list";

    W.last_loaded = -1;
    while (W.listcount > 0)
    {
	W.listcount--;
	m_free (W.list[W.listcount]);
    }
    if (first_time) start_time = time (NULL);
    dir_rewind (W.dir_kdir);
    while ( (finfo = dir_read (W.dir_kdir, KDIR_DOTDOT) ) != NULL )
    {
	if (first_time)
	{
	    if ( (raw_entry_count++ & (64 - 1) ) == 0 )
	    {
		current_time = time (NULL);
		if (current_time - start_time > 5)
		{
		    /*  Print a warning  */
		    fputs ("Directory scanning for browser may take a while\n",
			   stderr);
		    first_time = FALSE;
		}
	    }
	}
	/*  Unreadable entries and unexecutable directories are dropped  */
	if (!finfo->can_read) continue;
	if ( (finfo->type == KFILETYPE_DIRECTORY) && !finfo->can_execute )
	    continue;
	if ( (W.accept_file == NULL) || W.accept_file (*finfo) )
	{
	    len = strlen (finfo->filename);
	    ptr = m_alloc (len + 6);
	    W.list[W.listcount] = ptr;
	    /*  The following reference must be done through  ptr  because the
		AIX/rs6000 compiler is fucked  */
	    strcpy (ptr + FILENAME_OFFSET, finfo->filename);
	    if (finfo->type == KFILETYPE_DIRECTORY)
		strncpy (ptr, STRING_DIR, FILENAME_OFFSET);
	    else
		strncpy (ptr, STRING_FILE, FILENAME_OFFSET);
	    W.listcount++;
	    if (W.listcount == W.listmax) _Filewin_increase_list_size (w);
	}
    }
    first_time = FALSE;
    if (W.accept_file == NULL) show_aips = TRUE;
    else
    {
	KFileInfo fake_finfo;
	char fname[MAXPATHLEN], sympath[MAXPATHLEN];

	fake_finfo.filename = fname;
	fake_finfo.sym_path = sympath;
	fake_finfo.type = KFILETYPE_REGULAR;  /*  Bullshit  */
	fake_finfo.dirname = W.curdir;
	strcpy (fname, "fake.aips");
	show_aips = W.accept_file (fake_finfo);
    }
    if (W.accept_file == NULL) show_drao = TRUE;
    else
    {
	KFileInfo fake_finfo;
	char fname[MAXPATHLEN], sympath[MAXPATHLEN];

	fake_finfo.filename = fname;
	fake_finfo.sym_path = sympath;
	fake_finfo.type = KFILETYPE_REGULAR;  /*  Bullshit  */
	fake_finfo.dirname = W.curdir;
	strcpy (fname, "fake.drao");
	show_drao = W.accept_file (fake_finfo);
    }
    if (show_aips)
    {
	files = foreign_aips1_read_dataset_names (W.curdir, &num_datasets);
	for (count = 0; count < num_datasets; ++count, files += len + 1)
	{
	    len = strlen (files);
	    ptr = m_alloc (len + 6);
	    W.list[W.listcount] = ptr;
	    strcpy (ptr + FILENAME_OFFSET, files);
	    strncpy (ptr, STRING_AIPS, FILENAME_OFFSET);
	    W.listcount++;
	    if (W.listcount == W.listmax) _Filewin_increase_list_size (w);
	}
    }
    if (show_drao)
    {
	files = foreign_drao_read_dataset_names (W.curdir, &num_datasets,NULL);
	for (count = 0; count < num_datasets; ++count, files += len + 1)
	{
	    len = strlen (files);
	    ptr = m_alloc (len + 6);
	    W.list[W.listcount] = ptr;
	    strcpy (ptr + FILENAME_OFFSET, files);
	    strncpy (ptr, STRING_DRAO, FILENAME_OFFSET);
	    W.listcount++;
	    if (W.listcount == W.listmax) _Filewin_increase_list_size (w);
	}
    }
    st_qsort (W.list, 0, W.listcount - 1);
    if (show_drao)
    {
	/*  Fiddle strings so that index numbers appear at the front  */
	char txt[STRING_LENGTH];

	for (count = 0; count < W.listcount; ++count)
	{
	    char *entry = W.list[count];

	    if (strncmp (entry, STRING_DRAO, STRING_COMPARE_LENGTH) != 0)
		continue;
	    if ( ( ptr = strrchr (entry + FILENAME_OFFSET, '.') ) == NULL )
	    {
		fprintf (stderr, "Missing '.'\n");
		a_prog_bug (function_name);
	    }
	    if (strncmp (ptr + 1, "f=", 2) != 0)
	    {
		fprintf (stderr, "Missing \"f=\"\n");
		a_prog_bug (function_name);
	    }
	    len = atoi (ptr + 3);
	    *ptr = '\0';
	    sprintf (txt, "%03u  %s", len, entry + FILENAME_OFFSET);
	    strcpy (entry + FILENAME_OFFSET, txt);
	}
    }
    if ( !XtIsRealized ( (Widget) w ) )
    {
	while (W.listcount < MINIMUM_LIST_ROWS)
	{   /*  Add trailing lines to force minimum sizing  */
	    if ( ( W.list[W.listcount] = st_dup ("") ) == NULL )
	    {
		m_abort (function_name, "empty string");
	    }
	    ++W.listcount;
	}
    }
#undef W
}   /*  End Function _Filewin_create_list  */

static void _Filewin_filesel_cbk (Widget w, XtPointer client_data,
				  XtPointer call_data)
{
    flag relative = FALSE;
    int len;
    XawListReturnStruct *lr = (XawListReturnStruct *) call_data;
    FilewinWidget fw = (FilewinWidget) client_data;
    char *entry = fw->filewin.list[lr->list_index];
    CONST char *name = entry + FILENAME_OFFSET;
    char filename[MAXPATHLEN], pathname[MAXPATHLEN];

    /*  First extract name from list into a pseudo-filename  */
    if (strncmp (entry, STRING_DRAO, STRING_COMPARE_LENGTH) == 0)
    {
	unsigned int fnum;
	CONST char *ptr1, *ptr2;

	fnum = atoi (name);
	for (ptr1 = name; !isspace (*ptr1); ++ptr1);
	for (; isspace (*ptr1); ++ptr1);
	if ( ( ptr2 = strrchr (ptr1, '/') ) == NULL ) ptr2 = ptr1;
	else ++ptr2;
	sprintf (filename, "%s.f=%u.drao", ptr2, fnum);
    }
    else if (strncmp (entry, STRING_AIPS, STRING_COMPARE_LENGTH) == 0)
	sprintf (filename, "%s.aips", name);
    else strcpy (filename, name);
    /*  Now massage the directory name  */
    len = strlen (fw->filewin.orig_dir);
    if ( (strcmp (fw->filewin.orig_dir, ".") == 0) ||
	 (strcmp (name, "..") == 0) )
    {   /*  Relative paths: just concatenate  */
	strcpy (pathname, fw->filewin.curdir);
	strcat (pathname, "/");
    }
    else if (strcmp (fw->filewin.curdir, "/") == 0) strcpy (pathname, "/");
    else if (strcmp (fw->filewin.curdir, fw->filewin.orig_dir) == 0)
    {   /*  We're in the process' CWD, so use just the filename  */
	pathname[0] = '\0';
	relative = TRUE;
    }
    else if ( (len > 1) &&
	      (strncmp (fw->filewin.curdir, fw->filewin.orig_dir, len) == 0) )
    {   /*  We're below the process' CWD, so create a path relative to that  */
	strcpy (pathname, fw->filewin.curdir + len + 1);
	strcat (pathname, "/");
	relative = TRUE;
    }
    else
    {
	strcpy (pathname, fw->filewin.curdir);
	strcat (pathname, "/");
    }
    strcat (pathname, filename);
    _Filewin_clean_dirname (pathname);
    if (pathname[0] == '\0') strcpy (pathname, "/");
    if ( (fw->filewin.last_loaded >= 0) &&
	 (fw->filewin.last_loaded < fw->filewin.listcount) )
	fw->filewin.list[fw->filewin.last_loaded][MARKER_OFFSET] = ' ';
    entry[MARKER_OFFSET] = MARKER_CHARACTER;
    fw->filewin.last_loaded = lr->list_index;
    XawListChange (fw->filewin.listwidget, fw->filewin.list,
		   fw->filewin.listcount, 0, True);
    XawListHighlight (fw->filewin.listwidget, lr->list_index);
    XFlush ( XtDisplay (w) );
    if (strncmp (entry, STRING_DIR, STRING_COMPARE_LENGTH) == 0)
    {
	/*  Yep, it's a directory  */
	char path[MAXPATHLEN];

	if ( _Filewin_dir_func (fw, pathname) ) return;
	/*  Change the directory  */
	if (realpath (pathname, path) == NULL)
	{
	    if (relative)
	    {
		sprintf (pathname, "%s/%s", fw->filewin.curdir, name);
		_Filewin_clean_dirname (pathname);
	    }
	    strcpy (path, pathname);
	}
	XkwFilewinChangeDirectory (client_data, path);
	return;
    }
    XtCallCallbacks ( (Widget) fw, XkwNfileSelectCallback, pathname );
    pathname[0] = '\0';
}   /*  End Function _Filewin_filesel_cbk  */

static flag _Filewin_clean_dirname (char dirname[MAXPATHLEN])
/*  [SUMMARY] Remove unneccessary components from a directory name.
    <dirname> The name of the directory to clean.
    [RETURNS] TRUE if string was modified, else FALSE.
*/
{
    flag copy;
    flag modified = FALSE;
    char *curr, *ptr, *p;
    char new[MAXPATHLEN];

    new[0] = '\0';
    curr = dirname;
    while ( ( ptr = strstr (curr, "/.." ) ) != NULL )
    {
	/*  Check if directory name exists to the left  */
	if (*(ptr - 1) == '.')
	{
	    /*  Nope, it's a special  */
	    copy = TRUE;
	}
	else if ( (ptr[3] != '/') && (ptr[3] != '\0') )
	{
	    /*  Silly name like "..a"  */
	    copy = TRUE;
	}
	else copy = FALSE;
	if (copy)
	{
	    ptr += 3;
	    strncat (new, curr, ptr - curr);
	    /*new[ptr - curr] = '\0';*/
	    curr = ptr;
	}
	else
	{
	    /*  Search backwards for '/'  */
	    for (p = ptr - 1; *p != '/'; --p);
	    ptr += 3;
	    if (p - curr > 0)
	    {
		strncat (new, curr, p - curr);
		/*new[p - curr] = '\0';*/
	    }
	    curr = ptr;
	    modified = TRUE;
	}
    }
    /*  Copy remainder  */
    strcat (new, curr);
    strcpy (dirname, new);
    return (modified);
}   /*  End Function clean_dirname  */

static flag _Filewin_dir_func (FilewinWidget w, CONST char *dirname)
/*  [SUMMARY] Process directory name.
    <w> The FilewinWidget.
    <dirname> The name of the directory.
    [RETURNS] TRUE if the directory selection should be ignored, else FALSE.
*/
{
    if (w->filewin.trapDirectoryDatasets)
    {
	if ( foreign_miriad_test2 (dirname, FALSE) ||
	     foreign_aips2_test (dirname, FALSE) ||
	     foreign_filter_test_directory_dataset (dirname) )
	{
	    XtCallCallbacks ( (Widget) w, XkwNfileSelectCallback,
			      (XtPointer) dirname );
	    return (TRUE);
	}
    }
    return c_call_callbacks (w->filewin.dir_callbacks, (void *) dirname);
}   /*  End Function dir_func  */

static flag _Filewin_glish_event_func (void *object, CONST char *name,
				       CONST char *data, unsigned int type,
				       unsigned int num_elements,
				       unsigned int num_axes,
				       CONST int *axis_lengths)
{
    CONST char *string;
    Widget w = (Widget) object;

    if (type != K_VSTRING) return (FALSE);
    if (num_elements != 1) return (FALSE);
    string = *(CONST char **) data;
    XtCallCallbacks (w, XkwNfileSelectCallback, (XtPointer) string);
    return (TRUE);
}   /*  End Function glish_event_func  */

static void _Filewin_text_cbk (Widget w, XtPointer client_data,
			       XtPointer call_data)
{
    flag apply = FALSE;
    String string;
    FilewinWidget fw = client_data;
    char *ptr1, *ptr2;
    char dirname[MAXPATHLEN];
    static char function_name[] = "Filewin::text_cbk";

    XtVaGetValues (fw->filewin.text_widget, XtNstring, &string, NULL);
    if (string[0] == '\0') return;
    if (strlen (string) >= MAXPATHLEN)
    {
	fprintf (stderr, "%s: string too long (%d characters, max %d)\n",
		 function_name, strlen (string), MAXPATHLEN);
	return;
    }
    for (ptr1 = string, ptr2 = dirname; *ptr1 != '\0'; ++ptr1)
    {
	if (*ptr1 == '\n') apply = TRUE;
	else *ptr2++ = *ptr1;
    }
    *ptr2 = '\0';
    if (!apply) return;
    if ( !st_expr_expand (dirname, MAXPATHLEN, dirname, NULL, NULL, NULL) )
    {   /*  <dirname> may be corrupted and may be missing '\0'  */
	fprintf (stderr, "%s: error expanding path: \"%s\"\n",
		 function_name, string);
	return;
    }
    XkwFilewinChangeDirectory ( (Widget) fw, dirname );
}   /*  End Function text_cbk  */

static void _Filewin_popup_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
/*  This is the popup callback.
*/
{
    FilewinWidget fw = client_data;
    char *dirname = fw->filewin.curdir;
    static char function_name[] = "Filewin::popup_cbk";

    if (fw->filewin.dir_kdir == NULL)
	if ( ( fw->filewin.dir_kdir = dir_open (dirname) ) == NULL ) return;
    if (fw->filewin.dir_fd >= 0) return;
    if ( ( fw->filewin.dir_fd = open (dirname, O_RDONLY, 0) ) < 0 )
    {
	fprintf (stderr, "%s: error opening directory: \"%s\"\t%s\n",
		 function_name, dirname, ERRSTRING);
	dir_close (fw->filewin.dir_kdir);
	fw->filewin.dir_kdir = NULL;
	return;
    }
    XkwFilewinRescan (client_data);
}   /*  End Function popup_cbk   */

static void _Filewin_popdown_cbk (Widget w, XtPointer client_data,
				  XtPointer call_data)
/*  This is the popdown callback.
*/
{
    FilewinWidget fw = client_data;

    if (fw->filewin.dir_kdir != NULL) dir_close (fw->filewin.dir_kdir);
    fw->filewin.dir_kdir = NULL;
    if (fw->filewin.dir_fd >= 0) close (fw->filewin.dir_fd);
    fw->filewin.dir_fd = -1;
}   /*  End Function popdown_cbk   */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwFilewinRescan (Widget W)
/*  [SUMMARY] Force a directory rescan in a Filewin widget.
    <W> The widget.
    [RETURNS] Nothing.
*/
{
    FilewinWidget fw = (FilewinWidget) W;

    _Filewin_create_list (fw);
    XawListChange (fw->filewin.listwidget, fw->filewin.list,
		   fw->filewin.listcount, 0, True);
}   /*  End Function XkwFilewinRescan  */

/*EXPERIMENTAL_FUNCTION*/
char *XkwFilewinCurrentDirectory (Widget W)
/*  [SUMMARY] Get the current directory in a Filewin widget.
    <W> The widget.
    [RETURNS] A pointer to a dynamically allocated string containing the
    current directory for the widget. This must be freed with [<m_free>].
*/
{
    FilewinWidget w = (FilewinWidget) W;

    return st_dup (w->filewin.curdir);
}   /*  End Function XkwFilewinCurrentDirectory  */
  
/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc XkwFilewinRegisterDirCbk (Widget w,
					flag (*callback) (Widget w,
							  void *info,
							  CONST char *dirname),
					void *info)
/*  [SUMMARY] Register a callback to capture directory selection events.
    [PURPOSE] This routine will register a callback to capture directory
    selection events prior to the Filewin widget changing directory.
    <w> The Filewin widget.
    <callback> The routine which will be called when directories are selected.
    The prototype function is [<Filewin_PROTO_dir_cbk>].
    <info> A pointer to arbitrary information.
    [RETURNS] A KCallbackFunc object.
*/
{
    FilewinWidget fw = (FilewinWidget) w;

    return c_register_callback (&fw->filewin.dir_callbacks, callback, w,
				info, FALSE, NULL, FALSE, TRUE);
}   /*  End Function XkwFilewinRegisterDirCbk  */

/*EXPERIMENTAL_FUNCTION*/
flag XkwFilewinStandardFileTester_nD (KFileInfo finfo)
/*  [SUMMARY] Standard file tester.
    [PURPOSE] This routine may be passed as the value for the
    XkwNfilenameTester resource on a FilewinWidget or FilepopupWidget. The
    routine will accept all known filetypes.
    <finfo> The file information.
    [RETURNS] TRUE if the file is to be displayed, else FALSE.
*/
{
    unsigned int filetype;
    char *ptr;
    char pathname[MAXPATHLEN];

    /*  Accept all directories  */
    if (finfo.type == KFILETYPE_DIRECTORY) return (TRUE);
    strcpy (pathname, finfo.dirname);
    strcat (pathname, "/");
    strcat (pathname, finfo.filename);
    filetype = foreign_guess_format_from_filename (pathname);
    if (filetype == FOREIGN_FILE_FORMAT_UNKNOWN) return (FALSE);
    /*  Filetype understood: accept it  */
    switch (filetype)
    {
      case FOREIGN_FILE_FORMAT_KARMA:
	/*  It's a Karma file: fake things a bit for ".desc"/".dat" pairs  */
	if ( ( ptr = strrchr (finfo.filename, '.') ) == NULL ) return (FALSE);
	/*  Ignore the ".dat" file  */
	if (strcmp (ptr, ".dat") == 0) return (FALSE);
	/*  Don't molest if not the ".desc" file  */
	if (strcmp (ptr, ".desc") != 0) break;
	strcpy (ptr, ".kf");
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	/*  It's a proper GIPSY file: fake things a bit  */
	if ( ( ptr = strrchr (finfo.filename, '.') ) == NULL ) return (FALSE);
	/*  Ignore anything other than ".descr" (i.e. the ".image" file)  */
	if (strcmp (ptr, ".descr") != 0) return (FALSE);
	strcpy (ptr, ".gipsy");
	break;
      default:
	break;
    }
    return (TRUE);
}   /*  End Function XkwFilewinStandardFileTester_nD  */

/*EXPERIMENTAL_FUNCTION*/
flag XkwFilewinStandardFileTester_3D (KFileInfo finfo)
/*  [SUMMARY] Standard file tester for cube files.
    [PURPOSE] This routine may be passed as the value for the
    XkwNfilenameTester resource on a FilewinWidget or FilepopupWidget. The
    routine will accept all known filetypes which are capable of containing a
    3-dimensional array of data.
    <finfo> The file information.
    [RETURNS] TRUE if the file is to be displayed, else FALSE.
*/
{
    unsigned int filetype;
    char pathname[MAXPATHLEN];

    /*  Accept all directories  */
    if (finfo.type == KFILETYPE_DIRECTORY) return (TRUE);
    strcpy (pathname, finfo.dirname);
    strcat (pathname, "/");
    strcat (pathname, finfo.filename);
    filetype = foreign_guess_format_from_filename (pathname);
    /*  Reject pure image formats  */
    switch (filetype)
    {
      case FOREIGN_FILE_FORMAT_UNKNOWN:
      case FOREIGN_FILE_FORMAT_PPM:
      case FOREIGN_FILE_FORMAT_PGM:
      case FOREIGN_FILE_FORMAT_PNM:
      case FOREIGN_FILE_FORMAT_FILTER_PNM:
      case FOREIGN_FILE_FORMAT_SUNRAS:
	return (FALSE);
	/*break;*/
      default:
	break;
    }
    return XkwFilewinStandardFileTester_nD (finfo);
}   /*  End Function XkwFilewinStandardFileTester_3D  */

/*EXPERIMENTAL_FUNCTION*/
flag XkwFilewinChangeDirectory (Widget W, CONST char *dirname)
/*  [SUMMARY] Change the current directory for a Filewin widget.
    <W> The widget.
    <dirname> The pathname to the directory.
    [RETURNS] TRUE on success, else FALSE. On error the directory is unchanged.
*/
{
    KDir kdir, old_kdir;
    int fd;
    FilewinWidget fw = (FilewinWidget) W;
    char tmp_path[MAXPATHLEN + 1];
    static char function_name[] = "XkwFilewinChangeDirectory";

    if (W == NULL)
    {
	fprintf (stderr, "NULL widget\n");
	a_prog_bug (function_name);
    }
    if ( !XtIsFilewin (W) )
    {
	fprintf (stderr, "Widget is not a Filewin widget\n");
	a_prog_bug (function_name);
    }
    if ( ( kdir = dir_open (dirname) ) == NULL ) return (FALSE);
    if ( ( fd = open (dirname, O_RDONLY, 0) ) < 0 )
    {
	fprintf (stderr, "%s: error opening directory: \"%s\"\t%s\n",
		 function_name, dirname, ERRSTRING);
	dir_close (kdir);
	return (FALSE);
    }
    /*  Change the directory  */
    if (realpath (dirname, fw->filewin.curdir) == NULL)
	strcpy (fw->filewin.curdir, dirname);
    /*  Copy the dirname so callbacks can't mess with it  */
    strcpy (tmp_path, fw->filewin.curdir);
    XtCallCallbacks (W, XkwNdirectoryChangeCallback, tmp_path);
    old_kdir = fw->filewin.dir_kdir;
    fw->filewin.dir_kdir = kdir;
    _Filewin_create_list (fw);
    fw->filewin.dir_kdir = old_kdir;
    if (fw->filewin.text_widget != NULL)
    {
	XawListChange (fw->filewin.listwidget, fw->filewin.list,
		       fw->filewin.listcount, 0, True);
	/*  Have to break it into two sections  */
	XtVaSetValues (fw->filewin.text_widget,
		       XtNstring, fw->filewin.curdir,
		       NULL);
	XtVaSetValues (fw->filewin.text_widget,
		       XtNinsertPosition, strlen (fw->filewin.curdir),
		       NULL);
    }
    if ( (fw->filewin.text_widget != NULL) && (fw->filewin.dir_kdir == NULL) )
    {
	dir_close (kdir);
	close (fd);
    }
    else
    {
	_Filewin_popdown_cbk (NULL, fw, NULL);
	fw->filewin.dir_kdir = kdir;
	fw->filewin.dir_fd = fd;
    }
    return (TRUE);
}   /*  End Function XkwFilewinChangeDirectory  */

#if 0
/*PROTOTYPE_FUNCTION*/
flag Filewin_PROTO_dir_cbk (Widget w, void *info, CONST char *dirname)
/*  [SUMMARY] Callback for directory selection.
    [PURPOSE] This routine is called when the user selects a directory in
    the file selector. Normally the directory is descended and another
    group of files is visible. This callback makes it to possible to
    "capture" a directory if required. This can be useful when a
    conceptual "data file" is in fact implemented as a directory or files,
    such as with Miriad.
    <w> The Filewin widget on which the selection occurred.
    <info> A pointer to arbitrary information.
    <dirname> The name of the directory.
    [RETURNS] TRUE if the selection event is to be consumed, else FALSE
    (meaning the event should be passed on to other callbacks, or, if none
    consume the event, the widget will change it's active directory).
*/
{
}
#endif
