/*LINTLIBRARY*/
/*  ImageEditorControl.c

    This code provides an image editor control widget for Xt.

    Copyright (C) 1999-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for an image editor control widget
    for Xt.


    Written by      Richard Gooch   9-JUN-1999

    Updated by      Richard Gooch   10-JUN-1999: Made use of <m_test_access>.

    Updated by      Richard Gooch   16-JUN-1999: Added "Pixel" drawing object.

    Updated by      Richard Gooch   17-JUN-1999: Added "Minimum", "Middle" and
  "Maximum" buttons.

    Last updated by Richard Gooch   20-MAR-2001: Created Destroy method.


*/
#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_kwin.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/ImageEditorControlP.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/SimpleSlider.h>

#define BUF_SIZE 16384

#define OBJECT_BRUSH        0
#define OBJECT_PIXEL        1
#define OBJECT_RECTANGLE    2
#define OBJECT_POLYGON      3
#define NUM_OBJECT_CHOICES  4


struct canvas_list
{
    KWorldCanvas worldcanvas;
    KPixCanvas pixcanvas;
    KPixCanvasEditorContext ctx;
    ImageEditorControlWidget w;
    struct canvas_list *next;
};

static char *object_choices[NUM_OBJECT_CHOICES] =
{
    "Brush", "Pixel", "Rectangle", "Polygon",
};

/*  Private functions  */
STATIC_FUNCTION (void _ImageEditorControl_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _ImageEditorControl_Destroy, (Widget W) );
STATIC_FUNCTION (void popup_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void popdown_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void undo_last_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void undo_all_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void apply_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void minimum_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void middle_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void maximum_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void ImageEditorControl__set_data_value,
		 (ImageEditorControlWidget w, double scaled_value) );
STATIC_FUNCTION (void ImageEditorControl__add_func,
		 (KImageEditList ilist, list_entry *instruction,
		  void **info) );
STATIC_FUNCTION (void ImageEditorControl__loss_func,
		 (KImageEditList ilist, void **info) );
STATIC_FUNCTION (void ImageEditorControl__apply_func,
		 (KImageEditList ilist, void **info) );
STATIC_FUNCTION (void ImageEditorControl__refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (flag ImageEditorControl__position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code,void *e_info, void **f_info,
		  double x_lin, double y_lin, int x_pix, int y_pix) );
STATIC_FUNCTION (flag ImageEditorControl__geom_func,
		 (KPixCanvasEditorContext context, KPixCanvas canvas,
		  unsigned int object_type, unsigned int num_coords,
		  CONST int *x, CONST int *y, unsigned int raw_event_code,
		  void *raw_e_info, void **f_info) );


#define offset(field) XtOffset(ImageEditorControlWidget, imageEditorControl.field)

static XtResource ImageEditorControlResources[] = 
{
    {"dummy", "Dummy", XtRInt, sizeof (int),
     offset (dummy), XtRImmediate, 0},
};

#undef offset


ImageEditorControlClassRec imageEditorControlClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,          /* superclass */
    "ImageEditorControl",                          /* class_name */
    sizeof (ImageEditorControlRec),                /* widget_size */
    NULL,                                          /* class_initialise */
    NULL,                                          /* class_part_initialise */
    FALSE,                                         /* class_init */
    (XtInitProc) _ImageEditorControl_Initialise,   /* initialise */
    NULL,                                          /* initialise_hook */
    XtInheritRealize,                              /* realise */
    NULL,                                          /* actions */
    0,                                             /* num_actions */
    ImageEditorControlResources,                   /* resources */
    XtNumber (ImageEditorControlResources),        /* num_resources */
    NULLQUARK,                                     /* xrm_class */
    TRUE,                                          /* compress_motion */
    TRUE,                                          /* compress_exposure */
    TRUE,                                          /* compress_enterleave */
    TRUE,                                          /* visible_interest */
    _ImageEditorControl_Destroy,                   /* destroy */
    XtInheritResize,                               /* resize */
    NULL,                                          /* expose */
    NULL,                                          /* set_values */
    NULL,                                          /* set_values_hook */
    XtInheritSetValuesAlmost,                      /* set_values_almost */
    NULL,                                          /* get_values_hook */
    NULL,                                          /* accept_focus */
    XtVersion,                                     /* version */
    NULL,                                          /* callback_private */
    NULL,                                          /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* ImageEditorControlClassPart */
    0 /* empty */
  }
};

WidgetClass imageEditorControlWidgetClass = (WidgetClass)
     &imageEditorControlClassRec;

static void _ImageEditorControl_Initialise (Widget Request, Widget New)
{
    /*ImageEditorControlWidget request=(ImageEditorControlWidget)Request;*/
    ImageEditorControlWidget new = (ImageEditorControlWidget) New;
    Widget form, w, upper_widget;
    /*static char function_name[] = "ImageEditorControlWidget::Initialise";*/

    new->imageEditorControl.prev_x = TOOBIG;
    new->imageEditorControl.prev_y = TOOBIG;
    new->imageEditorControl.first_canvas = NULL;
    new->imageEditorControl.elist =
	iedit_create_list (ImageEditorControl__add_func,
			   ImageEditorControl__loss_func,
			   ImageEditorControl__apply_func,
			   new);
    XtAddCallback (New, XtNpopupCallback, popup_cbk, New);
    XtAddCallback (New, XtNpopdownCallback, popdown_cbk, New);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Undo Last",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, undo_last_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Undo All",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, undo_all_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Apply",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, apply_cbk, new);
    w = XtVaCreateManagedWidget ("objectMenu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "objectMenu",
				 XkwNchoiceName, "Object",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_OBJECT_CHOICES,
				 XkwNitemStrings, object_choices,
				 XkwNvaluePtr,
				 &new->imageEditorControl.object_mode,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, popup_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Brush Width",
				 XtNwidth, 220,
				 XkwNminimum, 1,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 1,
				 XkwNvaluePtr,
				 &new->imageEditorControl.brush_width,
				 NULL);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("value", dialogWidgetClass, form,
				 XtNlabel,
				 "Paint Value                     ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNvalue,
				 "0.0                             ",
				 NULL);
    upper_widget = w;
    new->imageEditorControl.value_dlg = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Minimum",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromVert, upper_widget,
				 NULL);
    XtAddCallback (w, XtNcallback, minimum_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Middle",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, middle_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Maximum",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, maximum_cbk, New);
}   /*  End Function Initialise  */

static void _ImageEditorControl_Destroy (Widget W)
{
    ImageEditorControlWidget w = (ImageEditorControlWidget) W;
    struct canvas_list *curr, *next;

    for (curr = w->imageEditorControl.first_canvas; curr != NULL; curr = next)
    {
	next = curr->next;
	m_free (curr);
    }
    iedit_free_list (w->imageEditorControl.elist);
}   /*  End Function Destroy  */

static void popup_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the popup callback.
*/
{
    unsigned int kwin_object = KWIN_EDIT_OBJECT_NONE;
    ImageEditorControlWidget top = client_data;
    struct canvas_list *curr;

    switch (top->imageEditorControl.object_mode)
    {
      case OBJECT_BRUSH:
      case OBJECT_PIXEL:
	kwin_object = KWIN_EDIT_OBJECT_NONE;
	break;
      case OBJECT_RECTANGLE:
	kwin_object = KWIN_EDIT_OBJECT_RECTANGLE;
	break;
      case OBJECT_POLYGON:
	kwin_object = KWIN_EDIT_OBJECT_POLYGON;
	break;
    }
    for (curr = top->imageEditorControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	kwin_editor_set_attributes (curr->ctx,
				    KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
				    kwin_object,
				    KWIN_EDITOR_CONTEXT_ATT_END);
    }
}   /*  End Function popup_cbk  */

static void popdown_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the popdown callback.
*/
{
    ImageEditorControlWidget top = client_data;
    struct canvas_list *curr;

    for (curr = top->imageEditorControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	kwin_editor_set_attributes (curr->ctx,
				    KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
				    KWIN_EDIT_OBJECT_NONE,
				    KWIN_EDITOR_CONTEXT_ATT_END);
    }
}   /*  End Function popdown_cbk  */

static void undo_last_cbk (Widget w, XtPointer client_data,XtPointer call_data)
/*  This is the undo last callback.
*/
{
    ImageEditorControlWidget top = client_data;

    iedit_remove_instructions (top->imageEditorControl.elist, 1);
}   /*  End Function undo_last_cbk  */

static void undo_all_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the undo all callback.
*/
{
    ImageEditorControlWidget top = client_data;

    iedit_remove_instructions (top->imageEditorControl.elist, 0);
}   /*  End Function undo_all_cbk  */

static void apply_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the save callback.
*/
{
    struct canvas_list *curr;
    char *slice;
    ImageEditorControlWidget top = client_data;

    /*  Ensure all local arrays are writable  */
    for (curr = top->imageEditorControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	flag visible;
	ViewableImage vimage;

	kwin_get_attributes (curr->pixcanvas,
			     KWIN_ATT_VISIBLE, &visible,
			     KWIN_ATT_END);
	if (!visible) continue;
	vimage = viewimg_get_active (curr->worldcanvas);
	if (vimage == NULL) continue;
	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_SLICE, &slice,
				VIEWIMG_VATT_END);
	if ( !m_test_access (slice, 1, TRUE) )
	{
	    fprintf (stderr, "Image array is not writable: reload first%c\n",
		     BEL);
	    return;
	}
    }
    iedit_apply_instructions (top->imageEditorControl.elist);
}   /*  End Function apply_cbk  */

static void minimum_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the minimum callback.
*/
{
    ImageEditorControlWidget top = client_data;

    ImageEditorControl__set_data_value (top, 0.0);
}   /*  End Function minimum_cbk  */

static void middle_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the middle callback.
*/
{
    ImageEditorControlWidget top = client_data;

    ImageEditorControl__set_data_value (top, 0.5);
}   /*  End Function middle_cbk  */

static void maximum_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the maximum callback.
*/
{
    ImageEditorControlWidget top = client_data;

    ImageEditorControl__set_data_value (top, 1.0);
}   /*  End Function maximum_cbk  */

static void ImageEditorControl__set_data_value (ImageEditorControlWidget w,
						double scaled_value)
/*  [SUMMARY] Set data value.
    <w> The ImageEditorControl widget.
    <scaled_value> The scaled value.
    [RETURNS] Nothing.
*/
{
    double minimum, maximum;
    struct canvas_list *curr;

    for (curr = w->imageEditorControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	flag visible;
	char txt[STRING_LENGTH];

	kwin_get_attributes (curr->pixcanvas,
			     KWIN_ATT_VISIBLE, &visible,
			     KWIN_ATT_END);
	if (!visible) continue;
	canvas_get_attributes (curr->worldcanvas,
			       CANVAS_ATT_VALUE_MIN, &minimum,
			       CANVAS_ATT_VALUE_MAX, &maximum,
			       CANVAS_ATT_END);
	sprintf (txt, "%g", minimum + (maximum - minimum) * scaled_value);
	XtVaSetValues (w->imageEditorControl.value_dlg,
		       XtNvalue, txt,
		       NULL);
	break;
    }
}   /*  End Function set_data_value  */

static void ImageEditorControl__add_func (KImageEditList ilist,
					  list_entry *instruction, void **info)
/*  [SUMMARY] Image edit instruction add callback.
    [PURPOSE] This routine will process a single edit instruction which has
    been added to a managed image edit instruction list.
    <ilist> The managed image edit instruction list.
    <instruction> The edit instruction.
    <info> A pointer to the arbitrary information pointer for the list.
    [RETURNS] Nothing.
*/
{
    struct canvas_list *curr;
    ImageEditorControlWidget top = *info;

    for (curr = top->imageEditorControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	flag visible;

	kwin_get_attributes (curr->pixcanvas,
			     KWIN_ATT_VISIBLE, &visible,
			     KWIN_ATT_END);
	if (!visible) continue;
	canvas_draw_edit_object (curr->worldcanvas, instruction->data);
    }
}   /*  End Function add_func  */

static void ImageEditorControl__loss_func (KImageEditList ilist, void **info)
/*  [SUMMARY] Image edit instruction loss callback.
    [PURPOSE] This routine will process the loss of an edit instruction from a
    managed image edit instruction list.
    <ilist> The managed image edit instruction list.
    <info> A pointer to the arbitrary information pointer for the list.
    [RETURNS] Nothing.
*/
{
    struct canvas_list *curr;
    ImageEditorControlWidget top = *info;

    for (curr = top->imageEditorControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	kwin_refresh_if_visible (curr->pixcanvas, TRUE);
    }
}   /*  End Function loss_func  */

static void ImageEditorControl__apply_func (KImageEditList ilist, void **info)
/*  [SUMMARY] Image edit instruction apply callback.
    [PURPOSE] This routine will apply (commit) an edit list prior to the list
    being cleared.
    <ilist> The managed image edit instruction list.
    <info> A pointer to the arbitrary information pointer for the list.
    [RETURNS] Nothing.
*/
{
    struct canvas_list *curr;
    char *slice;
    ImageEditorControlWidget top = *info;
    static char function_name[] = "ImageEditorControl::apply_func";

    for (curr = top->imageEditorControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	flag visible;
	ViewableImage vimage;

	kwin_get_attributes (curr->pixcanvas,
			     KWIN_ATT_VISIBLE, &visible,
			     KWIN_ATT_END);
	if (!visible) continue;
	vimage = viewimg_get_active (curr->worldcanvas);
	if (vimage == NULL) continue;
	/*  Check if the array is writable  */
	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_SLICE, &slice,
				VIEWIMG_VATT_END);
	if ( !m_test_access (slice, 1, TRUE) )
	{
	    fprintf (stderr,
		     "%s: image array is not writable: ignoring edits%c\n",
		     function_name, BEL);
	}
	viewimg_draw_edit_list (vimage, ilist);
	viewimg_register_data_change (vimage);
    }
}   /*  End Function apply_func  */

static void ImageEditorControl__refresh_func (KWorldCanvas canvas,
					      int width, int height,
					      struct win_scale_type *win_scale,
					      Kcolourmap cmap,
					      flag cmap_resize, void **info,
					      PostScriptPage pspage,
					      unsigned int num_areas,
					      KPixCanvasRefreshArea *areas,
					      flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KImageEditList edit_list = *info;

    canvas_draw_edit_list (canvas, edit_list);
}   /*  End Function refresh_func  */

static flag ImageEditorControl__position_func (KWorldCanvas canvas,
					       double x, double y,
					       unsigned int event_code,
					       void *e_info,
					       void **f_info,
					       double x_lin, double y_lin,
					       int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    edit_coord coords[4];
    ImageEditorControlWidget top = *f_info;
    String svalue;
    double value[2];

    if (!top->shell.popped_up) return (FALSE);
    switch (top->imageEditorControl.object_mode)
    {
      case OBJECT_BRUSH:
      case OBJECT_PIXEL:
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    switch (event_code)
    {
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	switch (top->imageEditorControl.object_mode)
	{
	  case OBJECT_BRUSH:
	    break;
	  default:
	    return (FALSE);
	    /*break;*/
	}
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    if (top->imageEditorControl.object_mode == OBJECT_PIXEL)
    {
	unsigned int count;

	coords[0].abscissa = floor (x_lin + 0.5) - 0.4;
	coords[0].ordinate = floor (y_lin + 0.5) - 0.4;
	coords[1].abscissa = floor (x_lin + 0.5) + 0.4;
	coords[1].ordinate = floor (y_lin + 0.5) - 0.4;
	coords[2].abscissa = floor (x_lin + 0.5) + 0.4;
	coords[2].ordinate = floor (y_lin + 0.5) + 0.4;
	coords[3].abscissa = floor (x_lin + 0.5) - 0.4;
	coords[3].ordinate = floor (y_lin + 0.5) + 0.4;
	for (count = 0; count < 4; ++count)
	    canvas_coords_transform (canvas, 1,
				     &coords[count].abscissa, FALSE,
				     &coords[count].ordinate, FALSE);
	iedit_add_instruction (top->imageEditorControl.elist,
			       EDIT_INSTRUCTION_FPOLY, coords, 4, value);
	return (TRUE);
    }
    if ( (x == top->imageEditorControl.prev_x) &&
	 (y == top->imageEditorControl.prev_y) ) return (TRUE);
    coords[0].abscissa = x;
    coords[0].ordinate = y;
    /*  Get pixel co-ordinates, add brush width and convert back to world
	co-ordinates  */
    coords[1].abscissa = x_pix + top->imageEditorControl.brush_width;
    coords[1].ordinate = y_pix - top->imageEditorControl.brush_width;
    /*  Add brush width and convert back to world co-ordinates  */
    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
				     &coords[1].abscissa,
				     &coords[1].ordinate,
				     NULL, NULL,
				     &coords[1].abscissa,
				     &coords[1].ordinate);
    coords[1].abscissa = fabs (coords[1].abscissa - x) / 2.0;
    coords[1].ordinate = fabs (coords[1].ordinate - y) / 2.0;
    XtVaGetValues (top->imageEditorControl.value_dlg,
		   XtNvalue, &svalue,
		   NULL);
    sscanf (svalue, "%lf", value);
    value[1] = 0.0;
    iedit_add_instruction (top->imageEditorControl.elist, EDIT_INSTRUCTION_DAB,
			   coords, 2, value);
    if (event_code == K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG)
    {
        canvas_create_stroke_instruction (canvas,
					  top->imageEditorControl.prev_x,
					  top->imageEditorControl.prev_y,
					  x, y,
					  top->imageEditorControl.brush_width,
					  value,
					  top->imageEditorControl.elist);
    }
    top->imageEditorControl.prev_x = x;
    top->imageEditorControl.prev_y = y;
    return (TRUE);
}   /*  End Function position_func  */

static flag ImageEditorControl__geom_func (KPixCanvasEditorContext context,
					   KPixCanvas canvas,
					   unsigned int object_type,
					   unsigned int num_coords,
					   CONST int *x, CONST int *y,
					   unsigned int raw_event_code,
					   void *raw_e_info, void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    unsigned int count;
    unsigned int num_dcoords = 0;  /*  Initialised to keep compiler happy  */
    String svalue;
    struct canvas_list *canvas_entry = *f_info;
    double value[2];
    edit_coord coords[BUF_SIZE];
    static char function_name[] = "ImageEditorControl::geom_func";

    if (num_coords + 1 >= BUF_SIZE)
    {
	fprintf (stderr, "Too many co-ordinates: %u in object\n", num_coords);
	a_prog_bug (function_name);
    }
    switch (object_type)
    {
      case KWIN_EDIT_OBJECT_RECTANGLE:
	num_dcoords = 4;
	coords[0].abscissa = x[0];
	coords[0].ordinate = y[0];
	coords[1].abscissa = x[1];
	coords[1].ordinate = y[0];
	coords[2].abscissa = x[1];
	coords[2].ordinate = y[1];
	coords[3].abscissa = x[0];
	coords[3].ordinate = y[1];
	break;
      case KWIN_EDIT_OBJECT_POLYGON:
	num_dcoords = num_coords + 1;
	for (count = 0; count < num_dcoords; ++count)
	{
	    coords[count].abscissa = x[count];
	    coords[count].ordinate = y[count];
	}
	break;
    }
    for (count = 0; count < num_dcoords; ++count)
    {
	canvas_convert_to_canvas_coords (canvas_entry->worldcanvas, FALSE, 1,
					 &coords[count].abscissa,
					 &coords[count].ordinate,
					 NULL, NULL,
					 &coords[count].abscissa,
					 &coords[count].ordinate);
    }
    XtVaGetValues (canvas_entry->w->imageEditorControl.value_dlg,
		   XtNvalue, &svalue,
		   NULL);
    sscanf (svalue, "%lf", value);
    value[1] = 0.0;
    iedit_add_instruction (canvas_entry->w->imageEditorControl.elist,
			   EDIT_INSTRUCTION_FPOLY, coords, num_dcoords, value);
    return (TRUE);
}   /*  End Function geom_func  */


/*  Public functions follow  */

void XkwImageEditorControlAddCanvas (Widget W, KWorldCanvas canvas)
/*  [SUMMARY] Add a KWorldCanvas to an ImageEditorControl widget canvas list.
    <W> The widget.
    <canvas> The canvas to add to the list. The widget will then manipulate
    this canvas. If this is NULL the routine does nothing.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    ImageEditorControlWidget w;
    struct canvas_list *new;
    static char function_name[] = "XkwImageEditorControlAddCanvas";

    if ( !XtIsImageEditorControl (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (ImageEditorControlWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list structure");
    new->worldcanvas = canvas;
    new->pixcanvas = canvas_get_pixcanvas (canvas);
    new->w = w;
    new->next = w->imageEditorControl.first_canvas;
    w->imageEditorControl.first_canvas = new;
    canvas_register_refresh_func (canvas, ImageEditorControl__refresh_func,
				  w->imageEditorControl.elist);
    canvas_register_position_event_func (canvas,
					 ImageEditorControl__position_func, w);
    if ( !kwin_get_colour (new->pixcanvas, "green", &pixel_value,
			   NULL, NULL, NULL) )
    {
	if ( !kwin_get_colour (new->pixcanvas, "white", &pixel_value,
			       NULL, NULL, NULL) )
	    pixel_value = 0;
    }
    new->ctx = kwin_editor_create_context
	(new->pixcanvas,
	 KWIN_EDITOR_CONTEXT_ATT_MOUSE, K_CANVAS_EVENT_MIDDLE_MOUSE_MASK,
	 KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE, pixel_value,
	 KWIN_EDITOR_CONTEXT_ATT_END);
    kwin_editor_register_event_func (new->ctx, ImageEditorControl__geom_func,
				     new);
}   /*  End Function XkwImageEditorControlAddCanvas  */
