/*LINTLIBRARY*/
/*  IntensityPolicy.c

    This code provides a Intensity Policy control widget for Xt.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for an intensity policy control
    widget for Xt.


    Written by      Richard Gooch   8-DEC-1997: Split off from ZoomPolicy.

    Updated by      Richard Gooch   17-DEC-1997: Use <<i_min>> as zero point
  for log and sqrt transforms.

    Updated by      Richard Gooch   18-DEC-1997: Fixed bug in
  <log_iscale_func> to cope with values near <<i_min>> and many log cycles.

    Updated by      Richard Gooch   22-FEB-1998: Created function
  <XkwIntensityPolicyAddCanvas>.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   17-APR-1999: Support initialisation of
  iscale choice.

    Updated by      Richard Gooch   15-AUG-1999: Added <<refresh>> parameter to
  <XkwIntensityPolicyAddCanvas>.

    Updated by      Richard Gooch   24-AUG-1999: Abandoned use of
  callbackOnDrag resource for SimpleSlider widget.

    Updated by      Richard Gooch   12-SEP-1999: Created intensity range entry
  widgets.

    Updated by      Richard Gooch   20-SEP-1999: Made use of SimpleTextEntry
  widget class.

    Updated by      Richard Gooch   16-JUN-2000: Took account of interface
  change to function <ds_format_unit>.

    Updated by      Richard Gooch   26-FEB-2001: Increased log cycles slider
  width and range. Change minimum label depending on intensity transform.

    Updated by      Richard Gooch   28-FEB-2001: Reduced log cycles slider
  range and increased resolution to 0.1 increments.

    Updated by      Richard Gooch   10-MAR-2001: No longer change minimum label
  and added negative logarithmic transform.

    Last updated by Richard Gooch   20-MAR-2001: Created Destroy method.


*/
#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/IntensityPolicyP.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/SimpleTextEntry.h>


#define MIN_LOG_CYCLES      10    /*  Unscaled  */
#define MAX_LOG_CYCLES      100   /*  Unscaled  */
#define DEFAULT_LOG_CYCLES  40    /*  Unscaled  */

struct canvas_list
{
    KWorldCanvas canvas;
    flag refresh;
    struct canvas_list *next;
};


/*  Private functions  */
STATIC_FUNCTION (void _IntensityPolicy_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _IntensityPolicy_Destroy, (Widget W) );
STATIC_FUNCTION (void apply_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void set_bool,
		 (IntensityPolicyWidget top, unsigned int att, flag val) );
STATIC_FUNCTION (void auto_refresh_cbk, (Widget w, XtPointer client_data,
					 XtPointer call_data) );
STATIC_FUNCTION (void auto_intensity_cbk, (Widget w, XtPointer client_data,
					   XtPointer call_data) );
STATIC_FUNCTION (void iscale_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void set_canvas_iscale,
		 (IntensityPolicyWidget top, KWorldCanvas canvas) );
STATIC_FUNCTION (flag log_p_iscale_func,
		 (double *out, unsigned int out_stride,
		  CONST double *inp, unsigned int inp_stride,
		  unsigned int num_values,
		  double i_min, double i_max, void *info) );
STATIC_FUNCTION (flag log_n_iscale_func,
		 (double *out, unsigned int out_stride,
		  CONST double *inp, unsigned int inp_stride,
		  unsigned int num_values,
		  double i_min, double i_max, void *info) );
STATIC_FUNCTION (void log_cycles_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag sqrt_iscale_func,
		 (double *out, unsigned int out_stride,
		  CONST double *inp, unsigned int inp_stride,
		  unsigned int num_values,
		  double i_min, double i_max, void *info) );
STATIC_FUNCTION (void _IntensityPolicy_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void _IntensityPolicy_text_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffset(IntensityPolicyWidget, intensityPolicy.field)

static XtResource IntensityPolicyResources[] = 
{
    {XkwNautoIntensityScale, XkwCAutoIntensityScale, XtRBool, sizeof (Bool),
     offset (autoIntensityScale), XtRImmediate, (XtPointer) True},
    {XkwNshowIntensityReset, XkwCShowIntensityReset, XtRBool, sizeof (Bool),
     offset (showIntensityReset), XtRImmediate, (XtPointer) False},
    {XkwNintensityRangeCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (intensityRangeCallback), XtRCallback, (caddr_t) NULL},
};

#undef offset

#define ISCALE_LINEAR         0
#define ISCALE_LOGARITHMIC_P  1
#define ISCALE_LOGARITHMIC_N  2
#define ISCALE_SQUARE_ROOT    3
#define NUM_ISCALE_CHOICES    4

static char *iscale_choices[NUM_ISCALE_CHOICES] =
{
    "linear", "log(val-min)", "log(max-val)", "sqrt(val-min)",
};


IntensityPolicyClassRec intensityPolicyClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,         /* superclass */
    "IntensityPolicy",                            /* class_name */
    sizeof (IntensityPolicyRec),                  /* widget_size */
    NULL,                                         /* class_initialise */
    NULL,                                         /* class_part_initialise */
    FALSE,                                        /* class_init */
    (XtInitProc) _IntensityPolicy_Initialise,     /* initialise */
    NULL,                                         /* initialise_hook */
    XtInheritRealize,                             /* realise */
    NULL,                                         /* actions */
    0,                                            /* num_actions */
    IntensityPolicyResources,                     /* resources */
    XtNumber (IntensityPolicyResources),          /* num_resources */
    NULLQUARK,                                    /* xrm_class */
    TRUE,                                         /* compress_motion */
    TRUE,                                         /* compress_exposure */
    TRUE,                                         /* compress_enterleave */
    TRUE,                                         /* visible_interest */
    _IntensityPolicy_Destroy,                     /* destroy */
    XtInheritResize,                              /* resize */
    NULL,                                         /* expose */
    NULL,                                         /* set_values */
    NULL,                                         /* set_values_hook */
    XtInheritSetValuesAlmost,                     /* set_values_almost */
    NULL,                                         /* get_values_hook */
    NULL,                                         /* accept_focus */
    XtVersion,                                    /* version */
    NULL,                                         /* callback_private */
    NULL,                                         /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* IntensityPolicyClassPart */
    0 /* empty */
  }
};

WidgetClass intensityPolicyWidgetClass = (WidgetClass) &intensityPolicyClassRec;

static void _IntensityPolicy_Initialise (Widget Request, Widget New)
{
    /*IntensityPolicyWidget request = (IntensityPolicyWidget) Request;*/
    IntensityPolicyWidget new = (IntensityPolicyWidget) New;
    Widget form;
    Widget w, upper_widget;
    /*static char function_name[] = "IntensityPolicyWidget::Initialise";*/

    new->intensityPolicy.auto_refresh = TRUE;
    new->intensityPolicy.in_callback = FALSE;
    new->intensityPolicy.first_canvas = NULL;
    new->intensityPolicy.last_min = TOOBIG;
    new->intensityPolicy.last_max = TOOBIG;
    new->intensityPolicy.scale = 1.0;
    new->intensityPolicy.offset = 0.0;
    new->intensityPolicy.multiplier = 1.0;
    new->intensityPolicy.last_unit = NULL;
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    w = XtVaCreateManagedWidget ("apply", commandWidgetClass, form,
				 XtNlabel, "Apply",
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, apply_cbk, New);
    w = XtVaCreateManagedWidget ("autoRefreshToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Refresh",
				 XtNstate, True,
				 XkwNcrosses, False,
				 XtNfromHoriz, w,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, auto_refresh_cbk, New);
    w = XtVaCreateManagedWidget ("autoIntensityScaleToggle",
				 ktoggleWidgetClass, form,
				 XtNlabel, "Auto Intensity Scale",
				 XtNfromVert, upper_widget,
				 XtNstate,
				 new->intensityPolicy.autoIntensityScale,
				 XkwNcrosses, False,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, auto_intensity_cbk, New);
    if (new->intensityPolicy.showIntensityReset)
    {
	w = XtVaCreateManagedWidget ("resetIntensityToggle",
				     ktoggleWidgetClass, form,
				     XtNlabel,"Reset Intensity Upon File Load",
				     XtNfromVert, upper_widget,
				     XkwNcrosses, False,
				     NULL);
	upper_widget = w;
    }
    w = XtVaCreateManagedWidget ("iscaleMenu", exclusiveMenuWidgetClass, form,
				 XkwNchoiceName, "Intensity Scale",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_ISCALE_CHOICES,
				 XkwNitemStrings, iscale_choices,
				 XkwNvaluePtr,
				 &new->intensityPolicy.iscale_choice,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XkwNselectCallback, iscale_cbk, New);
    w = XtVaCreateManagedWidget ("logCyclesSlider", simpleSliderWidgetClass,
				 form,
				 XtNfromVert, upper_widget,
				 XtNwidth, 300,
				 XtNlabel, "Log Cycles",
				 XkwNminimum, MIN_LOG_CYCLES,
				 XkwNmaximum, MAX_LOG_CYCLES,
				 XkwNmodifier, 1,
				 XtNvalue, DEFAULT_LOG_CYCLES,
				 XkwNscaledFormat, "%.1f",
				 XkwNscaledValuePtr,
				 &new->intensityPolicy.log_cycles,
				 NULL);
    upper_widget = w;
    XkwSimpleSliderSetScale (w, 0.1, 0.0, FALSE, TRUE);
    XtAddCallback (w, XkwNvalueChangeCallback, log_cycles_cbk, new);
    w = XtVaCreateManagedWidget ("unitLabel", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Unit:",
				 NULL);
    upper_widget = w;
    new->intensityPolicy.unit_lbl = w;
    w = XtVaCreateManagedWidget ("minLabel", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Min:",
				 NULL);
    new->intensityPolicy.min_lbl = w;
    w = XtVaCreateManagedWidget ("minText", simpleTextEntryWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, 240,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->intensityPolicy.min_txt = w;
    upper_widget = w;
    XtAddCallback (w, XkwNnewLineCallback, _IntensityPolicy_text_cbk, New);
    w = XtVaCreateManagedWidget ("maxLabel", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Max:",
				 NULL);
    w = XtVaCreateManagedWidget ("maxText", simpleTextEntryWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNwidth, 240,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->intensityPolicy.max_txt = w;
    upper_widget = w;
    XtAddCallback (w, XkwNnewLineCallback, _IntensityPolicy_text_cbk, New);
}   /*  End Function Initialise  */

static void _IntensityPolicy_Destroy (Widget W)
{
    IntensityPolicyWidget w = (IntensityPolicyWidget) W;
    struct canvas_list *curr, *next;

    for (curr = w->intensityPolicy.first_canvas; curr != NULL; curr = next)
    {
	next = curr->next;
	m_free (curr);
    }
}   /*  End Function Destroy  */

static void apply_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the apply callback.
*/
{
    KPixCanvas pixcanvas;
    flag visible;
    IntensityPolicyWidget top = (IntensityPolicyWidget) client_data;
    struct canvas_list *curr;

    for (curr = top->intensityPolicy.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	if (!curr->refresh) continue;
	pixcanvas = canvas_get_pixcanvas (curr->canvas);
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_VISIBLE, &visible,
			     KWIN_ATT_END);
	if (!visible) continue;
	/*  Resize the canvas  */
	if ( !canvas_resize (curr->canvas, NULL, FALSE) )
	{
	    fprintf (stderr, "Error resizing canvas\n");
	}
    }
}   /*  End Function apply_cbk  */

static void set_bool (IntensityPolicyWidget top, unsigned int att, flag val)
{
    struct canvas_list *curr;

    for (curr = top->intensityPolicy.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	viewimg_set_canvas_attributes (curr->canvas,
				       att, val,
				       VIEWIMG_ATT_END);
    }
    if (top->intensityPolicy.auto_refresh) apply_cbk (NULL, top, NULL);
}   /*  End Function set_bool  */

static void auto_refresh_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  This is the auto refresh callback.
*/
{
    IntensityPolicyWidget top = (IntensityPolicyWidget) client_data;

    top->intensityPolicy.auto_refresh = (uaddr) call_data;
}   /*  End Function auto_refresh_cbk  */

static void auto_intensity_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
/*  This is the auto intensity callback.
*/
{
    flag bool = (uaddr) call_data;
    IntensityPolicyWidget top = (IntensityPolicyWidget) client_data;

    set_bool (top, VIEWIMG_ATT_AUTO_V, bool);
}   /*  End Function auto_intensity_cbk  */

static void iscale_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the iscale callback.
*/
{
    IntensityPolicyWidget top = (IntensityPolicyWidget) client_data;
    struct canvas_list *curr;

    for (curr = top->intensityPolicy.first_canvas; curr != NULL;
	 curr = curr->next)
	set_canvas_iscale (top, curr->canvas);
    if (top->intensityPolicy.auto_refresh) apply_cbk (NULL, top, NULL);
}   /*  End Function iscale_cbk  */

static void set_canvas_iscale (IntensityPolicyWidget top, KWorldCanvas canvas)
/*  [SUMMARY] Set the intensity scale for a canvas.
    <top> The IntensityPolicy widget.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    flag (*iscale_func) () = NULL;
    void *iscale_info = NULL;

    switch (top->intensityPolicy.iscale_choice)
    {
      case ISCALE_LINEAR:
	iscale_func = NULL;
	iscale_info = NULL;
	break;
      case ISCALE_LOGARITHMIC_P:
	iscale_func = log_p_iscale_func;
	iscale_info = top;
	break;
      case ISCALE_LOGARITHMIC_N:
	iscale_func = log_n_iscale_func;
	iscale_info = top;
	break;
      case ISCALE_SQUARE_ROOT:
	iscale_func = sqrt_iscale_func;
	iscale_info = NULL;
	break;
    }
    canvas_set_attributes (canvas,
			   CANVAS_ATT_ISCALE_FUNC, iscale_func,
			   CANVAS_ATT_ISCALE_INFO, iscale_info,
			   CANVAS_ATT_END);
}   /*  End Function set_canvas_iscale  */

static flag log_p_iscale_func (double *out, unsigned int out_stride,
			       CONST double *inp, unsigned int inp_stride,
			       unsigned int num_values,
			       double i_min, double i_max, void *info)
/*  [SUMMARY] Intensity scaling callback.
    [PURPOSE] This routine will perform an arbitrary intensity scaling on
    an array of values. This routine may be called many times to scale an
    image.
    <out> The output array.
    <out_stride> The stride (in doubles) of the output array.
    <inp> The input array.
    <inp_stride> The stride (in doubles) of the input array.
    <num_values> The number of values to scale.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <info> A pointer to arbitrary information.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    double value, low_clip;
    IntensityPolicyWidget top = (IntensityPolicyWidget) info;
    /*static char function_name[] = "IntensityPolicy::log_p_iscale_func";*/

    low_clip = (i_max - i_min) / pow (10.0, top->intensityPolicy.log_cycles);
    for (count = 0; count < num_values;
	 ++count, out += out_stride, inp += inp_stride)
    {
	value = *inp;
	if ( DOUBLE_IS_BLANK (value) )
	{
	    *out = TOOBIG;
	    continue;
	}
	if (value > i_max) value = i_max - i_min;
	else
	{
	    value -= i_min;
	    if (value < low_clip) value = low_clip;
	}
	value = log (value);
	if (value < -TOOBIG)
	{
	    fprintf (stderr, "range: %g to %g in: %g out: %g\n",
		     i_min, i_max, *inp, value);
	    abort ();
	}
	*out = value;
    }
    return (TRUE);
}   /*  End Function log_p_iscale_func  */

static flag log_n_iscale_func (double *out, unsigned int out_stride,
			       CONST double *inp, unsigned int inp_stride,
			       unsigned int num_values,
			       double i_min, double i_max, void *info)
/*  [SUMMARY] Intensity scaling callback.
    [PURPOSE] This routine will perform an arbitrary intensity scaling on
    an array of values. This routine may be called many times to scale an
    image.
    <out> The output array.
    <out_stride> The stride (in doubles) of the output array.
    <inp> The input array.
    <inp_stride> The stride (in doubles) of the input array.
    <num_values> The number of values to scale.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <info> A pointer to arbitrary information.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    double value, low_clip;
    IntensityPolicyWidget top = (IntensityPolicyWidget) info;
    /*static char function_name[] = "IntensityPolicy::log_n_iscale_func";*/

    low_clip = (i_max - i_min) / pow (10.0, top->intensityPolicy.log_cycles);
    for (count = 0; count < num_values;
	 ++count, out += out_stride, inp += inp_stride)
    {
	value = *inp;
	if ( DOUBLE_IS_BLANK (value) )
	{
	    *out = TOOBIG;
	    continue;
	}
	if (value < i_min) value = i_max - i_min;
	else
	{
	    value = i_max - value;
	    if (value < low_clip) value = low_clip;
	}
	value = log (value);
	if (value < -TOOBIG)
	{
	    fprintf (stderr, "range: %g to %g in: %g out: %g\n",
		     i_min, i_max, *inp, value);
	    abort ();
	}
	*out = value;
    }
    return (TRUE);
}   /*  End Function log_n_iscale_func  */

static void log_cycles_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  This is the log_cycles callback.
*/
{
    IntensityPolicyWidget top = (IntensityPolicyWidget) client_data;
    struct canvas_list *curr;

    if (!top->intensityPolicy.auto_refresh) return;
    for (curr = top->intensityPolicy.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	if ( !kwin_refresh_if_visible (canvas_get_pixcanvas (curr->canvas),
				       FALSE) )
	{
	    fprintf (stderr, "Error refreshing canvas\n");
	}
    }
}   /*  End Function log_cycles_cbk  */

static flag sqrt_iscale_func (double *out, unsigned int out_stride,
			      CONST double *inp, unsigned int inp_stride,
			      unsigned int num_values,
			      double i_min, double i_max, void *info)
/*  [SUMMARY] Intensity scaling callback.
    [PURPOSE] This routine will perform an arbitrary intensity scaling on
    an array of values. This routine may be called many times to scale an
    image.
    <out> The output array.
    <out_stride> The stride (in doubles) of the output array.
    <inp> The input array.
    <inp_stride> The stride (in doubles) of the input array.
    <num_values> The number of values to scale.
    <i_min> The minimum intensity value.
    <i_max> The maximum intensity value.
    <info> A pointer to arbitrary information.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    double value;
    /*static char function_name[] = "IntensityPolicy::sqrt_iscale_func";*/

    for (count = 0; count < num_values;
	 ++count, out += out_stride, inp += inp_stride)
    {
	value = *inp;
	if ( DOUBLE_IS_BLANK (value) )
	{
	    *out = TOOBIG;
	    continue;
	}
	if (value < i_min) value = 0.0;
	else value = sqrt (value - i_min);
	*out = value;
    }
    return (TRUE);
}   /*  End Function sqrt_iscale_func  */

static void _IntensityPolicy_refresh_func (KWorldCanvas canvas,
					   int width, int height,
					   struct win_scale_type *win_scale,
					   Kcolourmap cmap, flag cmap_resize,
					   void **info, PostScriptPage pspage,
					   unsigned int num_areas,
					   KPixCanvasRefreshArea *areas,
					   flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    ViewableImage vimage;
    double minimum, maximum;
    IntensityPolicyWidget top = *info;
    char *unit = "none";
    char txt[STRING_LENGTH], dummy[STRING_LENGTH];

    canvas_get_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, &minimum,
			   CANVAS_ATT_VALUE_MAX, &maximum,
			   CANVAS_ATT_END);
    vimage = viewimg_get_active (canvas);
    if (vimage != NULL)
    {
	flag truecolour;

	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_TRUECOLOUR, &truecolour,
				VIEWIMG_VATT_END);
	if (truecolour) vimage = NULL;
    }
    if (vimage == NULL)
    {
	top->intensityPolicy.scale = 1.0;
	top->intensityPolicy.offset = 0.0;
	strcpy (txt + 6, unit);
    }
    else
    {
	unsigned int elem_index;
	array_desc *arr_desc;

	viewimg_get_attributes (vimage,
				VIEWIMG_VATT_DATA_SCALE,
				&top->intensityPolicy.scale,
				VIEWIMG_VATT_DATA_OFFSET,
				&top->intensityPolicy.offset,
				VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
				VIEWIMG_VATT_PSEUDO_INDEX, &elem_index,
				VIEWIMG_VATT_END);
	unit = arr_desc->packet->element_desc[elem_index];
	ds_format_unit (txt + 6, dummy, &top->intensityPolicy.multiplier, unit,
			minimum * top->intensityPolicy.scale +
			top->intensityPolicy.offset);
	minimum = (minimum * top->intensityPolicy.scale +
		   top->intensityPolicy.offset) *
	    top->intensityPolicy.multiplier;
	maximum = (maximum * top->intensityPolicy.scale +
		   top->intensityPolicy.offset) *
	    top->intensityPolicy.multiplier;
    }
    if (unit != top->intensityPolicy.last_unit)
    {
	strncpy (txt, "Unit: ", 6);
	XtVaSetValues (top->intensityPolicy.unit_lbl, XtNlabel, txt, NULL);
	top->intensityPolicy.last_unit = unit;
    }
    if ( (minimum != top->intensityPolicy.last_min) ||
	 top->intensityPolicy.in_callback )
    {
	sprintf (txt, "%g", minimum);
	XtVaSetValues (top->intensityPolicy.min_txt, XtNstring, txt, NULL);
	top->intensityPolicy.last_min = minimum;
    }
    if ( (maximum != top->intensityPolicy.last_max) ||
	 top->intensityPolicy.in_callback )
    {
	sprintf (txt, "%g", maximum);
	XtVaSetValues (top->intensityPolicy.max_txt, XtNstring, txt, NULL);
	top->intensityPolicy.last_max = maximum;
    }
}   /*  End Function refresh_func  */

static void _IntensityPolicy_text_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
{
    String string;
    IntensityPolicyWidget top = client_data;
    double range[2];
    static char function_name[] = "IntensityPolicy::text_cbk";

    XtVaGetValues (top->intensityPolicy.min_txt, XtNstring, &string, NULL);
    sscanf (string, "%lg", range);
    XtVaGetValues (top->intensityPolicy.max_txt, XtNstring, &string, NULL);
    sscanf (string, "%lg", range + 1);
    if (range[0] >= range[1])
    {
	fprintf (stderr, "%s: ignoring as min: %g is not less than max: %g\n",
		 function_name, range[0], range[1]);
	range[0] = top->intensityPolicy.last_min;
	range[1] = top->intensityPolicy.last_max;
    }
    range[0] = (range[0] / top->intensityPolicy.multiplier -
		top->intensityPolicy.offset) / top->intensityPolicy.scale;
    range[1] = (range[1] / top->intensityPolicy.multiplier -
		top->intensityPolicy.offset) / top->intensityPolicy.scale;
    top->intensityPolicy.in_callback = TRUE;
    XtCallCallbacks ( (Widget) top, XkwNintensityRangeCallback, range );
    top->intensityPolicy.in_callback = FALSE;
}   /*  End Function text_cbk  */


/*  Public functions follow  */

void XkwIntensityPolicyAddCanvas (Widget W, KWorldCanvas canvas, flag refresh)
/*  [SUMMARY] Add a KWorldCanvas to an IntensityPolicy widget canvas list.
    <W> The widget.
    <canvas> The canvas to add to the list. The widget will then manipulate
    this canvas. If this is NULL the routine does nothing.
    <refresh> If TRUE, allow automatic refreshes on this canvas.
    [RETURNS] Nothing.
*/
{
    IntensityPolicyWidget w;
    struct canvas_list *new;
    static char function_name[] = "XkwIntensityPolicyAddCanvas";

    if ( !XtIsIntensityPolicy (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (IntensityPolicyWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list structure");
    new->canvas = canvas;
    new->refresh = refresh;
    new->next = w->intensityPolicy.first_canvas;
    w->intensityPolicy.first_canvas = new;
    set_canvas_iscale (w, canvas);
    canvas_register_refresh_func (canvas, _IntensityPolicy_refresh_func, w);
}   /*  End Function XkwIntensityPolicyAddCanvas  */
