/*LINTLIBRARY*/
/*  LoadAndDecimate.c

    This code provides a load and decimate control widget for Xt.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a load and decimate widget for
    Xt.


    Written by      Richard Gooch   20-APR-1997

    Updated by      Richard Gooch   24-APR-1997: Added subcube facility.

    Updated by      Richard Gooch   13-MAY-1997: Changed interface to function
  <XkwLoadAndDecimateGetCube>.

    Updated by      Richard Gooch   14-MAY-1997: Supported compressed FITS
  files.

    Updated by      Richard Gooch   23-JUL-1997: Fixed bug when zsize < 1 and
  zskip > 1. Some problems may still remain with other axes.

    Updated by      Richard Gooch   8-AUG-1997: Took account of change to
  <foreign_miriad_create_data_context>

    Updated by      Richard Gooch   17-NOV-1997: Added support for filtered
  files using FITS.

    Updated by      Richard Gooch   12-DEC-1997: Fixed bug with variable
  initialisation.

    Updated by      Richard Gooch   19-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   31-MAY-1998: Create/update data range in
  header.

    Updated by      Richard Gooch   6-NOV-1998: Made use of new function
  <foreign_fits_copy_special>.

    Updated by      Richard Gooch   18-FEB-1999: Tolerate missing FITS keywords

    Updated by      Richard Gooch   29-NOV-1999: Fixed overflow bug in counter
  array by switching to unsigned ints from unsigned bytes.

    Updated by      Richard Gooch   21-MAY-2000: Added DRAO support. Added
  BLC and TRC labels.

    Updated by      Richard Gooch   22-MAY-2000: Added INC label.

    Updated by      Richard Gooch   26-MAY-2000: Added AIPS support.

    Updated by      Richard Gooch   27-MAY-2000: Added 2D support.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variable.

    Updated by      Richard Gooch   22-SEP-2000: Took account of interface
  change in <ds_format_value>.

    Updated by      Richard Gooch   23-NOV-2000: Fixed bug in <load_cbk>:
  CRPIXn correction calculation had an incorrect sign.

    Updated by      Richard Gooch   25-NOV-2000: Fixed bug in function
  <XkwLoadAndDecimateGetCube>: added "filename" object.

    Updated by      Richard Gooch   26-NOV-2000: Changed function
  <XkwLoadAndDecimateGetCube>: change map state rather than sensitive state.

    Last updated by Richard Gooch   26-NOV-2000: Added IDF support.


*/
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_foreign.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_wcs.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/LoadAndDecimateP.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/SimpleProgressMeter.h>


/*  Private functions  */
STATIC_FUNCTION (void _LoadAndDecimate_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void cancel_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void load_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void abort_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void dealloc_input, (LoadAndDecimateWidget w) );
STATIC_FUNCTION (Boolean _LoadAndDecimate_work_proc, (XtPointer client_data) );
STATIC_FUNCTION (flag _LoadAndDecimate_read_image,
		 (LoadAndDecimateWidget w, float *data) );
STATIC_FUNCTION (flag _LoadAndDecimate_read_blocks,
		 (LoadAndDecimateWidget w, char *data, uaddr num_blocks) );
STATIC_FUNCTION (multi_array *load_header, (CONST char *filename) );
STATIC_FUNCTION (void slider_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );


#define offset(field) XtOffset(LoadAndDecimateWidget, loadAndDecimate.field)

static XtResource LoadAndDecimateResources[] = 
{
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, False},
};

#undef offset


LoadAndDecimateClassRec loadAndDecimateClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,    /* superclass */
    "LoadAndDecimate",                       /* class_name */
    sizeof (LoadAndDecimateRec),             /* widget_size */
    NULL,                                    /* class_initialise */
    NULL,                                    /* class_part_initialise */
    FALSE,                                   /* class_init */
    (XtInitProc) _LoadAndDecimate_Initialise,/* initialise */
    NULL,                                    /* initialise_hook */
    XtInheritRealize,                        /* realise */
    NULL,                                    /* actions */
    0,                                       /* num_actions */
    LoadAndDecimateResources,                /* resources */
    XtNumber (LoadAndDecimateResources),     /* num_resources */
    NULLQUARK,                               /* xrm_class */
    TRUE,                                    /* compress_motion */
    TRUE,                                    /* compress_exposure */
    TRUE,                                    /* compress_enterleave */
    TRUE,                                    /* visible_interest */
    NULL,                                    /* destroy */
    XtInheritResize,                         /* resize */
    NULL,                                    /* expose */
    NULL,                                    /* set_values */
    NULL,                                    /* set_values_hook */
    XtInheritSetValuesAlmost,                /* set_values_almost */
    NULL,                                    /* get_values_hook */
    NULL,                                    /* accept_focus */
    XtVersion,                               /* version */
    NULL,                                    /* callback_private */
    NULL,                                    /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* LoadAndDecimateClassPart */
    0 /* empty */
  }
};

WidgetClass loadAndDecimateWidgetClass = (WidgetClass)&loadAndDecimateClassRec;

static void _LoadAndDecimate_Initialise (Widget Request, Widget New)
{
    /*LoadAndDecimateWidget request = (LoadAndDecimateWidget) Request;*/
    LoadAndDecimateWidget new = (LoadAndDecimateWidget) New;
    Widget form, w, upper_widget;
    /*static char function_name[] = "LoadAndDecimateWidget::Initialise";*/

    new->loadAndDecimate.filename = NULL;
    new->loadAndDecimate.domain_name = NULL;
    new->loadAndDecimate.num_blocks = 0;
    new->loadAndDecimate.current_block = 0;
    new->loadAndDecimate.work_proc_id = 0;
    new->loadAndDecimate.orig_multi_desc = NULL;
    new->loadAndDecimate.new_multi_desc = NULL;
    new->loadAndDecimate.buffer = NULL;
    new->loadAndDecimate.countbuf = NULL;
    new->loadAndDecimate.channel = NULL;
    new->loadAndDecimate.miriad_ctx = NULL;
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Cancel",
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    XtAddCallback (w, XtNcallback, cancel_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Load",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, load_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Abort",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, abort_cbk, New);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Save",
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 XtNstate, TRUE,
				 XkwNcrosses, FALSE,
				 NULL);
    new->loadAndDecimate.auto_save_tgl = w;
#ifdef DISABLED
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Average",
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 XtNstate, TRUE,
				 XkwNcrosses, FALSE,
				 NULL);
    new->loadAndDecimate.average_tgl = w;
#endif
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel,
				 "File size:                    ",
				 XtNfromVert, upper_widget,
				 XtNheight, 20,
				 XtNborderWidth, 0,
				 NULL);
    new->loadAndDecimate.size_lbl = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel,
				 "Memory size:                        ",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 XtNborderWidth, 0,
				 NULL);
    new->loadAndDecimate.mem_lbl = w;
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel,
				 "BLC:                        ",
				 XtNfromVert, upper_widget,
				 XtNheight, 20,
				 XtNborderWidth, 0,
				 NULL);
    new->loadAndDecimate.blc_lbl = w;
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel,
				 "TRC:                        ",
				 XtNfromVert, upper_widget,
				 XtNheight, 20,
				 XtNborderWidth, 0,
				 NULL);
    new->loadAndDecimate.trc_lbl = w;
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel,
				 "INC:                        ",
				 XtNfromVert, upper_widget,
				 XtNheight, 20,
				 XtNborderWidth, 0,
				 NULL);
    new->loadAndDecimate.inc_lbl = w;
    upper_widget = w;
    w = XtVaCreateManagedWidget ("startXSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "X Start",
				 XtNfromVert, upper_widget,
				 XkwNmodifier, 1,
				 NULL);
    new->loadAndDecimate.start_x_sld = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("startYSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "Y Start",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNmodifier, 1,
				 NULL);
    new->loadAndDecimate.start_y_sld = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("startZSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "Z Start",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNmodifier, 1,
				 NULL);
    new->loadAndDecimate.start_z_sld = w;
    upper_widget = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("endXSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "X End",
				 XtNfromVert, upper_widget,
				 XkwNmodifier, 1,
				 NULL);
    new->loadAndDecimate.end_x_sld = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("endYSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "Y End",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNmodifier, 1,
				 NULL);
    new->loadAndDecimate.end_y_sld = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("endZSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "Z End",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNmodifier, 1,
				 NULL);
    new->loadAndDecimate.end_z_sld = w;
    upper_widget = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("skipXSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "X Skip",
				 XtNfromVert, upper_widget,
				 XkwNminimum, 1,
				 XkwNmaximum, 20,
				 XkwNmodifier, 1,
				 XtNvalue, 2,
				 NULL);
    new->loadAndDecimate.skip_x_sld = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("skipYSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "Y Skip",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNminimum, 1,
				 XkwNmaximum, 20,
				 XkwNmodifier, 1,
				 XtNvalue, 2,
				 NULL);
    new->loadAndDecimate.skip_y_sld = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("skipZSlider", simpleSliderWidgetClass, form,
				 XtNlabel, "Z Skip",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNminimum, 1,
				 XkwNmaximum, 20,
				 XkwNmodifier, 1,
				 XtNvalue, 2,
				 NULL);
    new->loadAndDecimate.skip_z_sld = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_cbk, New);
    w = XtVaCreateManagedWidget ("progressMeter",
				 simpleProgressMeterWidgetClass, form,
				 XtNheight, 20,
				 XtNfromVert, w,
				 XkwNmaximum, 0,
				 XtNvalue, 0,
				 NULL);
    new->loadAndDecimate.progress_meter = w;
    w = XtVaCreateManagedWidget ("dialog", dialogWidgetClass, form,
				 XtNlabel, "Output filename",
				 XtNheight, 20,
				 XtNfromVert, w,
				 XtNvalue, "decimated_cube",
				 NULL);
    new->loadAndDecimate.save_filename_dlg = w;
}   /*  End Function Initialise  */

static void cancel_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the cancel callback.
*/
{
    LoadAndDecimateWidget top = (LoadAndDecimateWidget) client_data;

    if (top->loadAndDecimate.work_proc_id != 0)
    {
	XtRemoveWorkProc (top->loadAndDecimate.work_proc_id);
	top->loadAndDecimate.work_proc_id = 0;
    }
    if (top->loadAndDecimate.orig_multi_desc != NULL)
    {
	ds_dealloc_multi (top->loadAndDecimate.orig_multi_desc);
	top->loadAndDecimate.orig_multi_desc = NULL;
    }
    if (top->loadAndDecimate.new_multi_desc != NULL)
    {
	ds_dealloc_multi (top->loadAndDecimate.new_multi_desc);
	top->loadAndDecimate.new_multi_desc = NULL;
    }
    if (top->loadAndDecimate.buffer != NULL)
    {
	m_free (top->loadAndDecimate.buffer);
	top->loadAndDecimate.buffer = NULL;
    }
    if (top->loadAndDecimate.countbuf != NULL)
    {
	m_free (top->loadAndDecimate.countbuf);
	top->loadAndDecimate.countbuf = NULL;
    }
    dealloc_input (top);
}   /*  End Function cancel_cbk  */

static void load_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the load callback.
*/
{
    Channel channel;
    KMiriadDataContext miriad_ctx;
    array_pointer arrayp;
    int xsize, ysize, zsize;
    uaddr xlen_out, ylen_out, skip;
    unsigned int count, axis;
    LoadAndDecimateWidget top = (LoadAndDecimateWidget) client_data;
    multi_array *orig_multi_desc = top->loadAndDecimate.orig_multi_desc;
    multi_array *new_multi_desc = top->loadAndDecimate.new_multi_desc;
    packet_desc *top_pack_desc;
    array_desc *arr_desc;
    char *ptr;
    char *filename = top->loadAndDecimate.filename;
    char txt[STRING_LENGTH];
    char fname[STRING_LENGTH];
    static char function_name[] = "XkwLoadAndDecimate::load_cbk";

    if (top->loadAndDecimate.work_proc_id != 0)
    {
	XBell (XtDisplay (w), 100);
	return;
    }
    top_pack_desc = new_multi_desc->headers[0];
    arr_desc = (array_desc *) top_pack_desc->element_desc[0];
    XtVaGetValues (top->loadAndDecimate.skip_x_sld,
		   XtNvalue, &top->loadAndDecimate.xskip,
		   NULL);
    XtVaGetValues (top->loadAndDecimate.start_x_sld,
		   XtNvalue, &top->loadAndDecimate.start_x,
		   NULL);
    XtVaGetValues (top->loadAndDecimate.end_x_sld,
		   XtNvalue, &top->loadAndDecimate.end_x,
		   NULL);
    xsize = top->loadAndDecimate.end_x - top->loadAndDecimate.start_x + 1;
    XtVaGetValues (top->loadAndDecimate.skip_y_sld,
		   XtNvalue, &top->loadAndDecimate.yskip,
		   NULL);
    XtVaGetValues (top->loadAndDecimate.start_y_sld,
		   XtNvalue, &top->loadAndDecimate.start_y,
		   NULL);
    XtVaGetValues (top->loadAndDecimate.end_y_sld,
		   XtNvalue, &top->loadAndDecimate.end_y,
		   NULL);
    ysize = top->loadAndDecimate.end_y - top->loadAndDecimate.start_y + 1;
    if (arr_desc->num_dimensions == 2)
    {
	top->loadAndDecimate.zskip = 1;
	top->loadAndDecimate.start_z = 0;
	top->loadAndDecimate.end_z = 0;
	zsize = 1;
    }
    else
    {
	XtVaGetValues (top->loadAndDecimate.skip_z_sld,
		       XtNvalue, &top->loadAndDecimate.zskip,
		       NULL);
	XtVaGetValues (top->loadAndDecimate.start_z_sld,
		       XtNvalue, &top->loadAndDecimate.start_z,
		       NULL);
	XtVaGetValues (top->loadAndDecimate.end_z_sld,
		       XtNvalue, &top->loadAndDecimate.end_z,
		       NULL);
	zsize = top->loadAndDecimate.end_z - top->loadAndDecimate.start_z + 1;
	zsize -= zsize % top->loadAndDecimate.zskip;
	if (xsize < 1) xsize = 1;
	if (ysize < 1) ysize = 1;
	if (zsize < 1) zsize = 1;
    }
    top->loadAndDecimate.num_blocks = zsize;
    top->loadAndDecimate.current_block = 0;
    XtVaSetValues (top->loadAndDecimate.progress_meter,
		   XtNvalue, 0,
		   XkwNmaximum, top->loadAndDecimate.num_blocks - 1,
		   NULL);
    xlen_out = xsize / top->loadAndDecimate.xskip;
    ylen_out = ysize / top->loadAndDecimate.yskip;
    m_copy (&arrayp, new_multi_desc->data[0], sizeof arrayp);
    if (arrayp.array != NULL)
    {
	(*arrayp.free) (arrayp.info, &arrayp);
	m_clear (new_multi_desc->data[0], sizeof arrayp);
	fprintf (stderr, "deallocated\n");
    }
    /*  Fiddle descriptor  */
    for (count = 0; count < arr_desc->num_dimensions; ++count)
    {
	uaddr size = 0;
	uaddr start = 0;
	unsigned int index = arr_desc->num_dimensions - 1 - count;
	double skip = 0.0;
	double value[2];

	switch (count)
	{
	  case 0:
	    skip = top->loadAndDecimate.xskip;
	    start = top->loadAndDecimate.start_x;
	    size = xsize;
	    break;
	  case 1:
	    skip = top->loadAndDecimate.yskip;
	    start = top->loadAndDecimate.start_y;
	    size = ysize;
	    break;
	  case 2:
	    skip = top->loadAndDecimate.zskip;
	    start = top->loadAndDecimate.start_z;
	    size = zsize;
	    break;
	}
	arr_desc->dimensions[index]->length = size / (uaddr) skip;
	if (arr_desc->dimensions[index]->length < 1)
	    arr_desc->dimensions[index]->length = 1;
	arr_desc->lengths[index] = arr_desc->dimensions[index]->length;
	arr_desc->dimensions[index]->last_coord = arr_desc->lengths[index] - 1;
	if ( ( axis = ds_get_fits_axis (new_multi_desc->headers[0],
					new_multi_desc->data[0],
					arr_desc->dimensions[index]->name) )
	     < 1 ) continue;
	sprintf (txt, "NAXIS%u", axis);
	value[0] = arr_desc->dimensions[index]->length;
	value[1] = 0.0;
	if ( (ds_f_elem_in_packet (new_multi_desc->headers[0], txt) <
	      new_multi_desc->headers[0]->num_elements) &&
	     !ds_put_unique_named_value (new_multi_desc->headers[0],
					 new_multi_desc->data, txt, K_INT,
					 value, TRUE) )
	{
	    fprintf (stderr, "Error setting keyword %s\n", txt);
	    return;
	}
	sprintf (txt, "CRPIX%u", axis);
	if ( ds_get_unique_named_value (orig_multi_desc->headers[0],
					orig_multi_desc->data[0], txt, NULL,
					value) )
	{
	    double offset;

	    offset = value[0] - 1.0 - (double) start - skip * 0.5 + 0.5;
	    value[0] = offset / skip + 1.0;
	    if ( !ds_put_unique_named_value (new_multi_desc->headers[0],
					     new_multi_desc->data, txt,
					     K_DOUBLE, value, TRUE) )
	    {
		fprintf (stderr, "Error setting keyword %s\n", txt);
		return;
	    }
	}
	sprintf (txt, "CDELT%u", axis);
	if ( ds_get_unique_named_value (orig_multi_desc->headers[0],
					orig_multi_desc->data[0], txt, NULL,
					value) )
	{
	    value[0] *= skip;
	    if ( !ds_put_unique_named_value (new_multi_desc->headers[0],
					     new_multi_desc->data, txt,
					     K_DOUBLE, value, TRUE) )
	    {
		fprintf (stderr, "Error setting keyword %s\n", txt);
		return;
	    }
	}
    }
    /*  Allocate (and clear!) output array  */
    if ( !ds_alloc_array (arr_desc, new_multi_desc->data[0], TRUE, TRUE) )
    {
	m_error_notify (function_name, "array data");
	XBell (XtDisplay (w), 100);
	return;
    }
    switch (top->loadAndDecimate.ftype)
    {
      case FOREIGN_FILE_FORMAT_FITS:
      case FOREIGN_FILE_FORMAT_FILTER_FITS:
	/*  Read FITS file  */
	if (top->loadAndDecimate.ftype == FOREIGN_FILE_FORMAT_FITS)
	    channel = ch_open_file (filename, "rz");
	else channel = foreign_filter_get_channel (filename);
	if (channel == NULL)
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, filename, ERRSTRING);
	    return;
	}
	new_multi_desc = foreign_fits_read_header (channel, FALSE, TRUE, TRUE,
						   FA_FITS_READ_HEADER_END);
	if (new_multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading FITS file header\n",
		     function_name);
	    ch_close (channel);
	    return;
	}
	ds_dealloc_multi (new_multi_desc);
	top->loadAndDecimate.channel = channel;
	break;
      case FOREIGN_FILE_FORMAT_MIRIAD:
	/*  Read Miriad Image file  */
	if ( ( miriad_ctx =
	       foreign_miriad_create_data_context (filename, K_CH_MAP_NEVER,
						   FALSE) ) == NULL )
	{
	    fprintf (stderr, "%s: error opening Miriad data\t%s\n",
		     function_name, ERRSTRING);
	    return;
	}
	top->loadAndDecimate.miriad_ctx = miriad_ctx;
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	/*  Read GIPSY file  */
	strcpy (fname, filename);
	if ( ( ptr = strrchr (fname, '.') ) == NULL ) return;
	strcpy (ptr, ".image");
	if ( ( channel = ch_open_file (fname, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, fname, ERRSTRING);
	    return;
	}
	top->loadAndDecimate.channel = channel;
	break;
      case FOREIGN_FILE_FORMAT_DRAO:
	new_multi_desc =
	    foreign_drao_read_header (filename, FALSE,
				      &top->loadAndDecimate.drao_ctx);
	if (new_multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading DRAO header\n", function_name);
	    return;
	}
	ds_dealloc_multi (new_multi_desc);
	break;
      case FOREIGN_FILE_FORMAT_AIPS1:
	new_multi_desc =
	    foreign_aips1_read_header (filename, FALSE,
				       &top->loadAndDecimate.aips1_ctx);
	if (new_multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading AIPS header\n", function_name);
	    return;
	}
	ds_dealloc_multi (new_multi_desc);
	break;
      case FOREIGN_FILE_FORMAT_IDF:
	strcpy (fname, filename);
	if ( ( ptr = strrchr (fname, '.') ) == NULL ) return;
	strcpy (ptr, ".int2");
	if ( ( channel = ch_open_file (fname, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, fname, ERRSTRING);
	    return;
	}
	top->loadAndDecimate.channel = channel;
	break;
      default:
	fprintf (stderr, "%s: file format unsupported\n", function_name);
	return;
	/*break;*/
    }
    /*  Clear count buffer  */
    m_clear (top->loadAndDecimate.countbuf,
	     sizeof *top->loadAndDecimate.countbuf * xlen_out * ylen_out);
    top->loadAndDecimate.minval = TOOBIG;
    top->loadAndDecimate.maxval = -TOOBIG;
    /*  Skip forward  */
    skip = top->loadAndDecimate.start_x;
    skip += top->loadAndDecimate.xlen_in * top->loadAndDecimate.start_y;
    skip += top->loadAndDecimate.xlen_in * top->loadAndDecimate.ylen_in *
	top->loadAndDecimate.start_z;
    if ( !_LoadAndDecimate_read_blocks (top, NULL, skip) ) exit (1);
    top->loadAndDecimate.work_proc_id =
	XtAppAddWorkProc (XtWidgetToApplicationContext (w),
			  _LoadAndDecimate_work_proc, top);
}   /*  End Function load_cbk  */

static void abort_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the abort callback.
*/
{
    LoadAndDecimateWidget top = (LoadAndDecimateWidget) client_data;

    dealloc_input (top);
    if (top->loadAndDecimate.work_proc_id == 0)
    {
	XBell (XtDisplay (w), 100);
	return;
    }
    XtRemoveWorkProc (top->loadAndDecimate.work_proc_id);
    top->loadAndDecimate.work_proc_id = 0;
}   /*  End Function abort_cbk  */

static void dealloc_input (LoadAndDecimateWidget w)
/*  [SUMMARY] Deallocate input objects.
    <w> The widget.
    [RETURNS] Nothing.
*/
{
    if (w->loadAndDecimate.channel != NULL)
    {
	ch_close (w->loadAndDecimate.channel);
	w->loadAndDecimate.channel = NULL;
    }
    if (w->loadAndDecimate.miriad_ctx != NULL)
    {
	foreign_miriad_close_data_context (w->loadAndDecimate.miriad_ctx);
	w->loadAndDecimate.miriad_ctx = NULL;
    }
    if (w->loadAndDecimate.drao_ctx != NULL)
    {
	foreign_drao_close_data_context (w->loadAndDecimate.drao_ctx);
	w->loadAndDecimate.drao_ctx = NULL;
    }
    if (w->loadAndDecimate.aips1_ctx != NULL)
    {
	foreign_aips1_close_data_context (w->loadAndDecimate.aips1_ctx);
	w->loadAndDecimate.aips1_ctx = NULL;
    }
}   /*  End Function dealloc_input  */

static Boolean _LoadAndDecimate_work_proc (XtPointer client_data)
/*  [SUMMARY] Xt work function.
    <client_data> The client data.
    [RETURNS] True if the procedure should not be called again, else False.
*/
{
    int xsize, ysize;
    uaddr xin, yin, xout, yout, zout, tot;
    uaddr xlen_out, ylen_out, xmax, ymax;
    array_pointer arrayp;
    Bool auto_save;
    float fval;
    float toobig = TOOBIG;
    LoadAndDecimateWidget top = (LoadAndDecimateWidget) client_data;
    String out_filename;
    multi_array *new_multi_desc;
    float *array;
    float *inpline;
    static char function_name[] = "XkwLoadAndDecimate::work_proc";

    new_multi_desc = top->loadAndDecimate.new_multi_desc;
    if (new_multi_desc == NULL)
    {
	fprintf (stderr, "NULL data\n");
	a_prog_bug (function_name);
    }
    /*  Load another block  */
    xsize = top->loadAndDecimate.end_x - top->loadAndDecimate.start_x + 1;
    ysize = top->loadAndDecimate.end_y - top->loadAndDecimate.start_y + 1;
    if (xsize < 1) xsize = 1;
    if (ysize < 1) ysize = 1;
    xmax = xsize - xsize % top->loadAndDecimate.xskip;
    ymax = ysize - ysize % top->loadAndDecimate.yskip;
    xlen_out = xsize / top->loadAndDecimate.xskip;
    ylen_out = ysize / top->loadAndDecimate.yskip;
    zout = top->loadAndDecimate.current_block / top->loadAndDecimate.zskip;
    m_copy (&arrayp, new_multi_desc->data[0], sizeof arrayp);
    array = (float *) arrayp.array + zout * xlen_out * ylen_out;
    if ( !_LoadAndDecimate_read_image (top, top->loadAndDecimate.buffer) )
	exit (1);
    /*  Merge latest block  */
    /*  Accumulate  */
    inpline = top->loadAndDecimate.buffer;
    for (yin = 0; yin < ymax; ++yin, inpline += xsize)
    {
	unsigned int *countline;
	float *outline;

	yout = yin / top->loadAndDecimate.yskip;
	outline = array + yout * xlen_out;
	countline = top->loadAndDecimate.countbuf + yout * xlen_out;
	for (xin = 0; xin < xmax; ++xin)
	{
	    xout = xin / top->loadAndDecimate.xskip;
	    fval = inpline[xin];
	    if ( DOUBLE_IS_BLANK (fval) ) continue;
	    outline[xout] += fval;
	    ++countline[xout];
	}
    }
    if ( (top->loadAndDecimate.current_block + 1) %
	 top->loadAndDecimate.zskip == 0)
    {
	/*  Average  */
	unsigned int uival;

	tot = xlen_out * ylen_out;
	for (xout = 0; xout < tot; ++xout)
	{
	    if ( (uival = top->loadAndDecimate.countbuf[xout]) == 0 )
		fval = toobig;
	    else
	    {
		fval = array[xout] / (float) uival;
		if (fval < top->loadAndDecimate.minval)
		    top->loadAndDecimate.minval = fval;
		if (fval > top->loadAndDecimate.maxval)
		    top->loadAndDecimate.maxval = fval;
	    }
	    array[xout] = fval;
	}
	m_clear (top->loadAndDecimate.countbuf,
		 sizeof *top->loadAndDecimate.countbuf * xlen_out * ylen_out);
    }
    if (++top->loadAndDecimate.current_block >=
	top->loadAndDecimate.num_blocks)
    {
	/*  Conversion complete  */
	double scale, offset;
	array_desc *arr_desc;

	top->loadAndDecimate.work_proc_id = 0;
	XtPopdown ( (Widget) top );
	XtVaSetValues (top->loadAndDecimate.progress_meter,
		       XtNvalue, 0,
		       NULL);
	dealloc_input (top);
	m_free (top->loadAndDecimate.buffer);
	top->loadAndDecimate.buffer = NULL;
	m_free (top->loadAndDecimate.countbuf);
	top->loadAndDecimate.countbuf = NULL;
	/*  Adjust header  */
	arr_desc = (array_desc *) new_multi_desc->headers[0]->element_desc[0];
	ds_set_data_range (arr_desc->packet->element_desc[0],
			   new_multi_desc->headers[0], new_multi_desc->data,
			   top->loadAndDecimate.minval,
			   top->loadAndDecimate.maxval);
	ds_get_data_scaling (arr_desc->packet->element_desc[0],
			     new_multi_desc->headers[0],
			     new_multi_desc->data[0],
			     &scale, &offset);
	if ( (scale != 1.0) || (offset != 0.0) )
	    ds_set_data_scaling (arr_desc->packet->element_desc[0],
				 new_multi_desc->headers[0],
				 new_multi_desc->data,
				 1.0, 0.0);
	/*  Save and dispatch  */
	XtVaGetValues (top->loadAndDecimate.auto_save_tgl,
		       XtNstate, &auto_save,
		       NULL);
	if (auto_save)
	{
	    XtVaGetValues (top->loadAndDecimate.save_filename_dlg,
			   XtNvalue, &out_filename,
			   NULL);
	    dsxfr_put_multi (out_filename,top->loadAndDecimate.new_multi_desc);
	}
	ds_event_dispatch (top->loadAndDecimate.new_multi_desc,
			   top->loadAndDecimate.domain_name,
			   top->loadAndDecimate.filename);
	ds_dealloc_multi (top->loadAndDecimate.orig_multi_desc);
	top->loadAndDecimate.orig_multi_desc = NULL;
	ds_dealloc_multi (top->loadAndDecimate.new_multi_desc);
	top->loadAndDecimate.new_multi_desc = NULL;
	return (True);
    }
    XtVaSetValues (top->loadAndDecimate.progress_meter,
		   XtNvalue, top->loadAndDecimate.current_block,
		   NULL);
    return (False);
}   /*  End Function _LoadAndDecimate_work_proc  */

static flag _LoadAndDecimate_read_image (LoadAndDecimateWidget w, float *data)
/*  [SUMMARY] Read an image.
    <w> The LoadAndDecimate widget.
    <data> The data will be written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int start_x, start_y, end_x, end_y, xsize, ysize;
    int y;
    uaddr block_size;
    uaddr xlen_in, ylen_in;
    /*static char function_name[] = "XkwLoadAndDecimate::read_image";*/

    /*  Load another block  */
    start_x = w->loadAndDecimate.start_x;
    start_y = w->loadAndDecimate.start_y;
    end_x = w->loadAndDecimate.end_x;
    end_y = w->loadAndDecimate.end_y;
    xlen_in = w->loadAndDecimate.xlen_in;
    ylen_in = w->loadAndDecimate.ylen_in;
    xsize = end_x - start_x + 1;
    ysize = end_y - start_y + 1;
    if ( (start_x == 0) && (end_x == xlen_in - 1) )
    {
	block_size = xsize * ysize;
	if ( !_LoadAndDecimate_read_blocks (w, (char *) data, block_size) )
	    return (FALSE);
	if (w->loadAndDecimate.current_block + 1 <
	    w->loadAndDecimate.num_blocks)
	{
	    /*  Skip forward to start of next input plane  */
	    block_size = xlen_in * ylen_in - block_size;
	    return _LoadAndDecimate_read_blocks (w, NULL, block_size);
	}
	return (TRUE);
    }
    for (y = start_y; y <= end_y; ++y, data += xsize)
    {
	if ( !_LoadAndDecimate_read_blocks (w, (char *) data, xsize) )
	    return (FALSE);
	if ( (y < end_y) || (w->loadAndDecimate.current_block + 1 <
			     w->loadAndDecimate.num_blocks) )
	    if ( !_LoadAndDecimate_read_blocks (w, NULL, xlen_in - xsize) )
		return (FALSE);
    }
    if (w->loadAndDecimate.current_block + 1 <
	w->loadAndDecimate.num_blocks)
    {
	/*  Skip forward to start of next input plane  */
	block_size = xlen_in * (ylen_in - ysize);
	return _LoadAndDecimate_read_blocks (w, NULL, block_size);
    }
    return (TRUE);
}   /*  End Function _LoadAndDecimate_read_image  */

static flag _LoadAndDecimate_read_blocks (LoadAndDecimateWidget w, char *data,
					  uaddr num_blocks)
/*  [SUMMARY] Read a number of blocks.
    <w> The LoadAndDecimate widget.
    <data> The data will be written here. If NULL, data are skipped.
    <num_blocks> The number of blocks to be read.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag skip = (data == NULL) ? TRUE : FALSE;
    static char function_name[] = "XkwLoadAndDecimate::read_blocks";

    if (num_blocks < 1) return (TRUE);
    switch (w->loadAndDecimate.ftype)
    {
      case FOREIGN_FILE_FORMAT_FITS:
      case FOREIGN_FILE_FORMAT_FILTER_FITS:
	if ( !foreign_fits_read_data (w->loadAndDecimate.channel,
				      w->loadAndDecimate.new_multi_desc,
				      data, num_blocks,
				      FA_FITS_READ_DATA_SKIP_BLOCKS, skip,
				      FA_FITS_READ_DATA_END) )
	{
	    fprintf (stderr, "%s: error reading data\n", function_name);
	    return (FALSE);
	}
	break;
      case FOREIGN_FILE_FORMAT_MIRIAD:
	if ( !foreign_miriad_read_data (w->loadAndDecimate.miriad_ctx,
					w->loadAndDecimate.new_multi_desc,
					data, num_blocks,
					FA_MIRIAD_READ_DATA_SKIP_BLOCKS, skip,
					FA_MIRIAD_READ_DATA_END) )
	{
	    fprintf (stderr, "%s: error reading data\n", function_name);
	    return (FALSE);
	}
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	if ( !foreign_gipsy_read_data (w->loadAndDecimate.channel,
				       w->loadAndDecimate.new_multi_desc,
				       data, num_blocks,
				       FA_GIPSY_READ_DATA_SKIP_BLOCKS, skip,
				       FA_GIPSY_READ_DATA_END) )
	{
	    fprintf (stderr, "%s: error reading data\n", function_name);
	    return (FALSE);
	}
	break;
      case FOREIGN_FILE_FORMAT_DRAO:
	if ( !foreign_drao_read_data (w->loadAndDecimate.drao_ctx, data,
				      num_blocks) )
	{
	    fprintf (stderr, "%s: error reading data\n", function_name);
	    return (FALSE);
	}
	break;
      case FOREIGN_FILE_FORMAT_AIPS1:
	if ( !foreign_aips1_read_data (w->loadAndDecimate.aips1_ctx, data,
				       num_blocks) )
	{
	    fprintf (stderr, "%s: error reading data\n", function_name);
	    return (FALSE);
	}
	break;
      case FOREIGN_FILE_FORMAT_IDF:
	if ( !foreign_idf_read_data (w->loadAndDecimate.channel,
				     w->loadAndDecimate.new_multi_desc,
				     data, num_blocks,
				     FA_IDF_READ_DATA_SKIP_BLOCKS, skip,
				     FA_IDF_READ_DATA_END) )
	{
	    fprintf (stderr, "%s: error reading data\n", function_name);
	    return (FALSE);
	}
	break;
    }
    return (TRUE);
}   /*  End Function _LoadAndDecimate_read_blocks  */

static multi_array *load_header (CONST char *filename)
/*  [SUMMARY] Load a header.
    <filename> The input filename.
    [RETURNS] The header on success, else NULL.
*/
{
    Channel channel;
    unsigned int elem_type;
    unsigned int ftype;
    char *ptr;
    multi_array *multi_desc;
    packet_desc *top_pack_desc;
    packet_desc *pack_desc;
    array_desc *arr_desc;
    char fname[STRING_LENGTH];
    static char function_name[] = "XkwLoadAndDecimate::load_header";

    /*  Determine format  */
    ftype = foreign_guess_format_from_filename (filename);
    if (ftype == FOREIGN_FILE_FORMAT_UNKNOWN) return (NULL);
    switch (ftype)
    {
      case FOREIGN_FILE_FORMAT_FITS:
      case FOREIGN_FILE_FORMAT_FILTER_FITS:
	/*  Read FITS file  */
	if (ftype == FOREIGN_FILE_FORMAT_FITS)
	    channel = ch_open_file (filename, "rz");
	else channel = foreign_filter_get_channel (filename);
	if (channel == NULL)
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, filename, ERRSTRING);
	    return (NULL);
	}
	multi_desc = foreign_fits_read_header (channel, FALSE, TRUE, TRUE,
					       FA_FITS_READ_HEADER_END);
	ch_close (channel);
	if (multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading FITS file header\n",
		     function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_MIRIAD:
	/*  Read Miriad Image file  */
	sprintf (fname, "%s/header", filename);
	if ( ( channel = ch_open_file (fname, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, fname, ERRSTRING);
	    return (NULL);
	}
	multi_desc = foreign_miriad_read_header (channel, FALSE, TRUE,
						 FA_MIRIAD_READ_HEADER_END);
	ch_close (channel);
	if (multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading Miriad Image header\n",
		     function_name);
	    return (NULL);
	}
	if ( !foreign_miriad_read_history (filename, multi_desc) )
	{
	    ds_dealloc_multi (multi_desc);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	/*  Read GIPSY file  */
	strcpy (fname, filename);
	if ( ( ptr = strrchr (fname, '.') ) == NULL ) return (NULL);
	strcpy (ptr, ".descr");
	if ( ( channel = ch_open_file (fname, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, fname, ERRSTRING);
	    return (NULL);
	}
	multi_desc = foreign_gipsy_read_header (channel, FALSE, TRUE,
						FA_GIPSY_READ_HEADER_END);
	ch_close (channel);
	if (multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading GIPSY header\n",
		     function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_DRAO:
	multi_desc = foreign_drao_read_header (filename, FALSE, NULL);
	if (multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading DRAO header\n", function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_AIPS1:
	multi_desc = foreign_aips1_read_header (filename, FALSE, NULL);
	if (multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading AIPS header\n", function_name);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_IDF:
	if ( ( channel = ch_open_file (filename, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, filename, ERRSTRING);
	    return (NULL);
	}
	multi_desc = foreign_idf_read_header (channel, FALSE, FALSE,
					      FA_IDF_READ_HEADER_END);
	ch_close (channel);
	if (multi_desc == NULL)
	{
	    fprintf (stderr, "%s: error reading IDF header\n", function_name);
	    return (NULL);
	}
	break;
      default:
	fprintf (stderr, "%s: file format unsupported\n", function_name);
	return (NULL);
	/*break;*/
    }
    /*  Sanity checks on data structure  */
    if (multi_desc->num_arrays != 1)
    {
	fprintf (stderr, "%s: multiple data structures found!\n",
		 function_name);
	return (NULL);
    }
    top_pack_desc = multi_desc->headers[0];
    if (top_pack_desc->element_types[0] != K_ARRAY)
    {
	fprintf (stderr, "%s: array not found!\n", function_name);
	return (NULL);
    }
    arr_desc = (array_desc *) top_pack_desc->element_desc[0];
    if ( (arr_desc->num_dimensions < 2) || (arr_desc->num_dimensions > 3) )
    {
	fprintf (stderr, "%s: %u dimensions not supported\n",
		 function_name, arr_desc->num_dimensions);
	ds_dealloc_multi (multi_desc);
	return (NULL);
    }
    pack_desc = arr_desc->packet;
    if (pack_desc->num_elements != 1)
    {
	fprintf (stderr, "%s: %u array packet elements not supported\n",
		 function_name, pack_desc->num_elements);
	ds_dealloc_multi (multi_desc);
	return (NULL);
    }
    elem_type = pack_desc->element_types[0];
    if (elem_type != K_FLOAT)
    {
	fprintf (stderr, "%s: array type: %u not supported\n",
		 function_name, pack_desc->element_types[0]);
	ds_dealloc_multi (multi_desc);
	return (NULL);
    }
    return (multi_desc);
}   /*  End Function load_header  */

static void slider_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the slider callback.
*/
{
    KwcsAstro ap;
    unsigned long size;
    int xstart, ystart, zstart, xend, yend, zend;
    packet_desc *top_pack_desc;
    array_desc *arr_desc;
    char *top_packet;
    LoadAndDecimateWidget top = (LoadAndDecimateWidget) client_data;
    int skips[3];
    char txt1[STRING_LENGTH], txt2[STRING_LENGTH];
    double coords[6];

    top_pack_desc = top->loadAndDecimate.new_multi_desc->headers[0];
    arr_desc = (array_desc *) top_pack_desc->element_desc[0];
    XtVaGetValues (top->loadAndDecimate.start_x_sld, XtNvalue, &xstart, NULL);
    XtVaGetValues (top->loadAndDecimate.start_y_sld, XtNvalue, &ystart, NULL);
    XtVaGetValues (top->loadAndDecimate.start_z_sld, XtNvalue, &zstart, NULL);
    XtVaGetValues (top->loadAndDecimate.end_x_sld, XtNvalue, &xend, NULL);
    XtVaGetValues (top->loadAndDecimate.end_y_sld, XtNvalue, &yend, NULL);
    XtVaGetValues (top->loadAndDecimate.end_z_sld, XtNvalue, &zend, NULL);
    XtVaGetValues (top->loadAndDecimate.skip_x_sld, XtNvalue, skips, NULL);
    XtVaGetValues (top->loadAndDecimate.skip_y_sld, XtNvalue, skips + 1, NULL);
    XtVaGetValues (top->loadAndDecimate.skip_z_sld, XtNvalue, skips + 2, NULL);
    size =  (xend - xstart + 1) / (unsigned long) skips[0];
    size *= (yend - ystart + 1) / (unsigned long) skips[1];
    if (arr_desc->num_dimensions == 3)
	size *= (zend - zstart + 1) / (unsigned long) skips[2];
    size *= ds_get_packet_size (arr_desc->packet);
    if (size < 1024)
	sprintf (txt1, "Memory size: %lu Bytes", size);
    else if (size < 1024 * 1024)
	sprintf (txt1, "Memory size: %.1f kiBytes", (float) size / 1024.0);
    else if (size < 1024 * 1024 * 1024)
	sprintf (txt1, "Memory size: %.1f MiBytes", (float) size / 1048576.0);
    else sprintf (txt1, "Memory size: %.1f GiBytes",
		  (float) size / 1073741824.0);
    sprintf (txt2, "%-28s", txt1);
    XtVaSetValues (top->loadAndDecimate.mem_lbl, XtNlabel, txt2, NULL);
    ap = wcs_astro_get_associated (top->loadAndDecimate.orig_multi_desc, 0);
    coords[0] = xstart;
    coords[1] = xend;
    coords[2] = ystart;
    coords[3] = yend;
    coords[4] = zstart;
    coords[5] = zend;
    strcpy (txt1, "BLC: ");
    if (arr_desc->num_dimensions == 2)
	wcs_astro_format_all (ap, txt1 + 5,
			      arr_desc->dimensions[1]->name, coords[0],
			      arr_desc->dimensions[0]->name, coords[2],
			      NULL, 0.0,
			      0, NULL, NULL, txt2);
    else wcs_astro_format_all (ap, txt1 + 5,
			       arr_desc->dimensions[2]->name, coords[0],
			       arr_desc->dimensions[1]->name, coords[2],
			       arr_desc->dimensions[0]->name, coords[4],
			       0, NULL, NULL, txt2);
    XtVaSetValues (top->loadAndDecimate.blc_lbl, XtNlabel, txt1, NULL);
    strcpy (txt1, "TRC: ");
    if (arr_desc->num_dimensions == 2)
	wcs_astro_format_all (ap, txt1 + 5,
			      arr_desc->dimensions[1]->name, coords[1],
			      arr_desc->dimensions[0]->name, coords[3],
			      NULL, 0.0,
			      0, NULL, NULL, txt2);
    else wcs_astro_format_all (ap, txt1 + 5,
			       arr_desc->dimensions[2]->name, coords[1],
			       arr_desc->dimensions[1]->name, coords[3],
			       arr_desc->dimensions[0]->name, coords[5],
			       0, NULL, NULL, txt2);
    XtVaSetValues (top->loadAndDecimate.trc_lbl, XtNlabel, txt1, NULL);
    strcpy (txt1, "INC: ");
    top_pack_desc = top->loadAndDecimate.orig_multi_desc->headers[0];
    top_packet = top->loadAndDecimate.orig_multi_desc->data[0];
    for (size = 0; size < arr_desc->num_dimensions; ++size)
    {
	dim_desc *dim;
	char *ptr;
	char txt3[STRING_LENGTH];

	dim = arr_desc->dimensions[arr_desc->num_dimensions - 1 - size];
	sprintf (txt3, "CDELT%lu", size + 1);
	if ( ds_get_unique_named_value (top_pack_desc, top_packet, txt3, NULL,
					coords) )
	{
	    coords[0] *= (double) skips[size];
	    if ( wcs_astro_test_if_lon (NULL, dim->name) ||
		 wcs_astro_test_if_lat (NULL, dim->name) )
		wcs_astro_format_dms (txt2, coords[0]);
	    else ds_format_value (txt2, coords[0], dim->name, 1.0, 0.0,
				  coords[0], NULL, NULL);
	    if ( ( ptr = strchr (txt2, ':') ) == NULL ) ptr = txt2;
	    else while ( isspace (*++ptr) );
	    strcat (txt1, ptr);
	    strcat (txt1, "  ");
	}
    }
    XtVaSetValues (top->loadAndDecimate.inc_lbl, XtNlabel, txt1, NULL);
}   /*  End Function slider_cbk  */


/*  Public Functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwLoadAndDecimateGetCube (Widget W, CONST char *filename,
				CONST char *domain_name)
/*  [SUMMARY] Get a cube.
    <W> The LoadAndDecimate widget.
    <filename> The name of the file to load.
    <domain_name> The domain name of the file.
    [RETURNS] Nothing.
*/
{
    int value;
    uaddr xlen_in, ylen_in, zlen_in;
    multi_array *orig, *new;
    packet_desc *top_pack_desc;
    array_desc *arr_desc;
    LoadAndDecimateWidget w = (LoadAndDecimateWidget) W;
    char txt1[STRING_LENGTH], txt2[STRING_LENGTH];
    static char function_name[] = "XkwLoadAndDecimateGetCube";

    if (w->loadAndDecimate.orig_multi_desc != NULL)
    {
	fprintf (stderr, "Old data structure found!\n");
	XBell (XtDisplay (w), 100);
	return;
    }
    if (w->loadAndDecimate.filename != NULL)
	m_free (w->loadAndDecimate.filename);
    if (w->loadAndDecimate.domain_name != NULL)
	m_free (w->loadAndDecimate.domain_name);
    if ( ( w->loadAndDecimate.filename = st_dup (filename) ) == NULL )
    {
	m_abort (function_name, "filename");
    }
    if ( ( w->loadAndDecimate.domain_name = st_dup (domain_name) ) == NULL )
    {
	m_abort (function_name, "domain name");
    }
    w->loadAndDecimate.ftype =
	foreign_guess_format_from_filename (w->loadAndDecimate.filename);
    if ( ( orig = load_header (w->loadAndDecimate.filename) ) == NULL )
    {
	if ( ( orig = foreign_guess_and_read (w->loadAndDecimate.filename,
					      K_CH_MAP_LOCAL, FALSE, NULL,
					      FA_GUESS_READ_END) )
	     == NULL )
	{
	    fprintf (stderr, "Error reading file: \"%s\"\n",
		     w->loadAndDecimate.filename);
	    return;
	}
	ds_event_dispatch (orig, w->loadAndDecimate.domain_name,
			   w->loadAndDecimate.filename);
	ds_dealloc_multi (orig);
	return;
    }
    top_pack_desc = orig->headers[0];
    arr_desc = (array_desc *) top_pack_desc->element_desc[0];
    if (arr_desc->num_dimensions == 2)
    {
	xlen_in = arr_desc->dimensions[1]->length;
	ylen_in = arr_desc->dimensions[0]->length;
	zlen_in = 0;
	sprintf (txt1, "%lu*%lu", xlen_in, ylen_in);
	XtSetMappedWhenManaged (w->loadAndDecimate.start_z_sld, False);
	XtSetMappedWhenManaged (w->loadAndDecimate.end_z_sld, False);
	XtSetMappedWhenManaged (w->loadAndDecimate.skip_z_sld, False);
    }
    else
    {
	xlen_in = arr_desc->dimensions[2]->length;
	ylen_in = arr_desc->dimensions[1]->length;
	zlen_in = arr_desc->dimensions[0]->length;
	sprintf (txt1, "%lu*%lu*%lu", xlen_in, ylen_in, zlen_in);
	XtSetMappedWhenManaged (w->loadAndDecimate.start_z_sld, True);
	XtSetMappedWhenManaged (w->loadAndDecimate.end_z_sld, True);
	XtSetMappedWhenManaged (w->loadAndDecimate.skip_z_sld, True);
    }
    sprintf (txt2, "File size: %-18s", txt1);
    XtVaSetValues (w->loadAndDecimate.size_lbl, XtNlabel, txt2, NULL);
    if ( ( new = ds_alloc_multi (1) ) == NULL )
    {
	m_abort (function_name, "new descriptor");
    }
    if ( ( new->headers[0] = ds_copy_desc_until (orig->headers[0], NULL) )
	 == NULL )
    {
	m_abort (function_name, "new descriptor");
    }
    if ( ( new->data[0] = ds_alloc_data (new->headers[0], FALSE, FALSE) )
	 == NULL )
    {
	m_abort (function_name, "new descriptor");
    }
    switch (w->loadAndDecimate.ftype)
    {
      case FOREIGN_FILE_FORMAT_FITS:
      case FOREIGN_FILE_FORMAT_FILTER_FITS:
	foreign_fits_copy_special (new, orig);
	break;
    }
    ds_copy_data (orig->headers[0], orig->data[0],
		  new->headers[0], new->data[0]);
    /*  Simulate file readers by adding "filename" object  */
    ds_add_named_object (new, new->num_arrays,
			 "filename", st_dup (filename), m_free);
    /*  Allocate buffer (input image converted to floats)  */
    if ( ( w->loadAndDecimate.buffer =
	   m_alloc (sizeof *w->loadAndDecimate.buffer * xlen_in * ylen_in) )
	 == NULL )
    {
	m_error_notify (function_name, "input buffer");
	ds_dealloc_multi (new);
	ds_dealloc_multi (orig);
	return;
    }
    if ( ( w->loadAndDecimate.countbuf =
	   m_alloc (sizeof *w->loadAndDecimate.countbuf * xlen_in *
		    ylen_in) ) == NULL )
    {
	m_error_notify (function_name, "count buffer");
	ds_dealloc_multi (new);
	ds_dealloc_multi (orig);
	return;
    }
    w->loadAndDecimate.orig_multi_desc = orig;
    w->loadAndDecimate.new_multi_desc = new;
    w->loadAndDecimate.xlen_in = xlen_in;
    w->loadAndDecimate.ylen_in = ylen_in;
    w->loadAndDecimate.zlen_in = zlen_in;
    XtVaSetValues (w->loadAndDecimate.start_x_sld,
		   XkwNminimum, 0,
		   XkwNmaximum, xlen_in,
		   NULL);
    XtVaSetValues (w->loadAndDecimate.end_x_sld,
		   XkwNminimum, 0,
		   XkwNmaximum, xlen_in - 1,
		   NULL);
    XtVaSetValues (w->loadAndDecimate.start_y_sld,
		   XkwNminimum, 0,
		   XkwNmaximum, ylen_in,
		   NULL);
    XtVaSetValues (w->loadAndDecimate.end_y_sld,
		   XkwNminimum, 0,
		   XkwNmaximum, ylen_in - 1,
		   NULL);
    if (arr_desc->num_dimensions == 3)
    {
	XtVaSetValues (w->loadAndDecimate.start_z_sld,
		       XkwNminimum, 0,
		       XkwNmaximum, zlen_in,
		       NULL);
	XtVaSetValues (w->loadAndDecimate.end_z_sld,
		       XkwNminimum, 0,
		       XkwNmaximum, zlen_in - 1,
		       NULL);
    }
    XtVaGetValues (w->loadAndDecimate.end_x_sld, XtNvalue, &value, NULL);
    if (value < 1) XtVaSetValues (w->loadAndDecimate.end_x_sld,
				  XtNvalue, xlen_in - 1,
				  NULL);
    XtVaGetValues (w->loadAndDecimate.end_y_sld, XtNvalue, &value, NULL);
    if (value < 1) XtVaSetValues (w->loadAndDecimate.end_y_sld,
				  XtNvalue, ylen_in - 1,
				  NULL);
    if (arr_desc->num_dimensions == 3)
    {
	XtVaGetValues (w->loadAndDecimate.end_z_sld, XtNvalue, &value, NULL);
	if (value < 1) XtVaSetValues (w->loadAndDecimate.end_z_sld,
				      XtNvalue, zlen_in - 1,
				      NULL);
    }
    slider_cbk (NULL, W, NULL);
    sprintf (txt2, "%s-decimated", w->loadAndDecimate.filename);
    XtVaSetValues (w->loadAndDecimate.save_filename_dlg,
		   XtNvalue, txt2,
		   XtNwidth, 370,
		   NULL);
    XtPopup (W, XtGrabNone);
}   /*  End Function XkwLoadAndDecimateGetCube  */
