/*LINTLIBRARY*/
/*  OverlayEditorControl.c

    This code provides an overlay editor control widget for Xt.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for an overlay editor control widget
    for Xt.


    Written by      Richard Gooch   3-AUG-1997

    Updated by      Richard Gooch   4-AUG-1997: Added remove last and
  remove all buttons.

    Updated by      Richard Gooch   10-AUG-1997: Made use of new class
  KOverlayListEditorContext.

    Updated by      Richard Gooch   22-FEB-1998: Created function
  <XkwOverlayEditorControlAddCanvas>.

    Updated by      Richard Gooch   2-MAR-1998: Created private overlay list.

    Updated by      Richard Gooch   10-MAY-1998: Added linear co-ordinate
  support.

    Updated by      Richard Gooch   14-APR-1999: Added support for saving
  overlays in Karma format.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>

    Updated by      Richard Gooch   4-SEP-2000: Added colour control.

    Last updated by Richard Gooch   20-MAR-2001: Created Destroy method.


*/
#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Label.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_xtmisc.h>
#include <karma_kwin.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/OverlayEditorControlP.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/Dialogpopup.h>


#define OBJECT_NONE         0
#define OBJECT_RECTANGLE    1
#define OBJECT_ELLIPSE      2
#define OBJECT_POLYGON      3
#define NUM_OBJECT_CHOICES  4


struct canvas_list
{
    KWorldCanvas canvas;
    struct canvas_list *next;
};

static char *object_choices[NUM_OBJECT_CHOICES] =
{
    "None", "Rectangle", "Ellipse", "Polygon",
};

#define MODE_CREATE      0
#define MODE_CHANGE      1
#define MODE_DELETE      2
#define NUM_MODE_CHOICES 3

static char *mode_choices[NUM_MODE_CHOICES] =
{
    "Create", "Change", "Delete",
};

#define COORDSYS_WORLD       0
#define COORDSYS_LINEAR      1
#define COORDSYS_RELATIVE    2
#define COORDSYS_PIXEL       3
#define NUM_COORDSYS_CHOICES 4

static char *coordsys_choices[NUM_COORDSYS_CHOICES] =
{
    "world", "linear", "relative", "pixel",
};

/*  Private functions  */
STATIC_FUNCTION (void _OverlayEditorControl_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _OverlayEditorControl_Destroy, (Widget W) );
STATIC_FUNCTION (void object_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void edit_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void first_coord_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void other_coord_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void remove_last_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void remove_all_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _OverlayEditorControl_save_dialog_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _OverlayEditorControl_text_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffset(OverlayEditorControlWidget, overlayEditorControl.field)

static XtResource OverlayEditorControlResources[] = 
{
    {"dummy", "Dummy", XtRInt, sizeof (int),
     offset (dummy), XtRImmediate, 0},
};

#undef offset


OverlayEditorControlClassRec overlayEditorControlClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,          /* superclass */
    "OverlayEditorControl",                        /* class_name */
    sizeof (OverlayEditorControlRec),              /* widget_size */
    NULL,                                          /* class_initialise */
    NULL,                                          /* class_part_initialise */
    FALSE,                                         /* class_init */
    (XtInitProc) _OverlayEditorControl_Initialise, /* initialise */
    NULL,                                          /* initialise_hook */
    XtInheritRealize,                              /* realise */
    NULL,                                          /* actions */
    0,                                             /* num_actions */
    OverlayEditorControlResources,                 /* resources */
    XtNumber (OverlayEditorControlResources),      /* num_resources */
    NULLQUARK,                                     /* xrm_class */
    TRUE,                                          /* compress_motion */
    TRUE,                                          /* compress_exposure */
    TRUE,                                          /* compress_enterleave */
    TRUE,                                          /* visible_interest */
    _OverlayEditorControl_Destroy,                 /* destroy */
    XtInheritResize,                               /* resize */
    NULL,                                          /* expose */
    NULL,                                          /* set_values */
    NULL,                                          /* set_values_hook */
    XtInheritSetValuesAlmost,                      /* set_values_almost */
    NULL,                                          /* get_values_hook */
    NULL,                                          /* accept_focus */
    XtVersion,                                     /* version */
    NULL,                                          /* callback_private */
    NULL,                                          /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* OverlayEditorControlClassPart */
    0 /* empty */
  }
};

WidgetClass overlayEditorControlWidgetClass = (WidgetClass)
     &overlayEditorControlClassRec;

static void _OverlayEditorControl_Initialise (Widget Request, Widget New)
{
    KOverlayListEditorContext ctx;
    /*OverlayEditorControlWidget request=(OverlayEditorControlWidget)Request;*/
    OverlayEditorControlWidget new = (OverlayEditorControlWidget) New;
    Widget form, w, upper_widget, save_dialog;
    /*static char function_name[] = "OverlayEditorControlWidget::Initialise";*/

    new->overlayEditorControl.first_canvas = NULL;
    new->overlayEditorControl.olist = overlay_va_create_list (NULL, NULL,
							      OVERLAY_ATT_END);
    save_dialog = XtVaCreatePopupShell ("save_dialog", dialogpopupWidgetClass,
					New,
					XtNlabel, "Save Filename:",
					XkwNdefaultName, "overlays",
					NULL);
    XtAddCallback (save_dialog, XtNcallback,
		   _OverlayEditorControl_save_dialog_cbk, New);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Remove Last",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, remove_last_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Remove All",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, remove_all_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Save",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popup_cbk, save_dialog);
    w = XtVaCreateManagedWidget ("objectMenu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "objectMenu",
				 XkwNchoiceName, "Object",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_OBJECT_CHOICES,
				 XkwNitemStrings, object_choices,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, object_mode_cbk, New);
    w = XtVaCreateManagedWidget ("modeMenu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "modeMenu",
				 XkwNchoiceName, "Mode",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNnumItems, NUM_MODE_CHOICES,
				 XkwNitemStrings, mode_choices,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XkwNselectCallback, edit_mode_cbk, New);
    XtSetSensitive (w, False);
    w = XtVaCreateManagedWidget ("coord1Menu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "coord1Menu",
				 XkwNchoiceName, "First Coordinate Type",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_COORDSYS_CHOICES,
				 XkwNitemStrings, coordsys_choices,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XkwNselectCallback, first_coord_cbk, New);
    w = XtVaCreateManagedWidget ("coord2Menu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "coord2Menu",
				 XkwNchoiceName, "Remaining Coordinate Types",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_COORDSYS_CHOICES,
				 XkwNitemStrings, coordsys_choices,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XkwNselectCallback, other_coord_cbk, New);
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNdefaultDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Colour:",
				 NULL);
    w = XtVaCreateManagedWidget ("colourname", asciiTextWidgetClass, form,
				 /*XtNborderWidth, 0,*/
				 XtNdefaultDistance, 0,
				 XtNwidth, 200,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNeditType, XawtextEdit,
				 XtNstring, "green",
				 /*XtNresize, XawtextResizeWidth,*/
				 NULL);
    new->overlayEditorControl.text_widget = w;
    XtVaGetValues (w, XtNtextSource, &w, NULL);
    XtAddCallback (w, XtNcallback, _OverlayEditorControl_text_cbk, New);
    ctx = overlay_editor_create_context
	(new->overlayEditorControl.olist,
	 OVERLAY_EDITOR_CONTEXT_ATT_FIRST_COORD_TYPE, OVERLAY_COORD_WORLD,
	 OVERLAY_EDITOR_CONTEXT_ATT_OTHER_COORD_TYPE, OVERLAY_COORD_WORLD,
	 OVERLAY_EDITOR_CONTEXT_ATT_END);
    overlay_editor_set_pixcontext_attributes
	(ctx,
	 KWIN_EDITOR_CONTEXT_ATT_MOUSE, K_CANVAS_EVENT_MIDDLE_MOUSE_MASK,
	 KWIN_EDITOR_CONTEXT_ATT_END);
    new->overlayEditorControl.ctx = ctx;
}   /*  End Function Initialise  */

static void _OverlayEditorControl_Destroy (Widget W)
{
    OverlayEditorControlWidget w = (OverlayEditorControlWidget) W;
    struct canvas_list *curr, *next;

    /*ToDo: destroy w->overlayEditorControl.ctx  */
    overlay_destroy_list (w->overlayEditorControl.olist);
    for (curr = w->overlayEditorControl.first_canvas; curr != NULL; curr =next)
    {
	next = curr->next;
	m_free (curr);
    }
}   /*  End Function Destroy  */

static void object_mode_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the object mode (shape) callback.
*/
{
    int mode = *(int *) call_data;
    unsigned int kwin_object = KWIN_EDIT_OBJECT_NONE;
    OverlayEditorControlWidget top = (OverlayEditorControlWidget) client_data;

    switch (mode)
    {
      case OBJECT_NONE:
	kwin_object = KWIN_EDIT_OBJECT_NONE;
	break;
      case OBJECT_RECTANGLE:
	kwin_object = KWIN_EDIT_OBJECT_RECTANGLE;
	break;
      case OBJECT_ELLIPSE:
	kwin_object = KWIN_EDIT_OBJECT_ELLIPSE;
	break;
      case OBJECT_POLYGON:
	kwin_object = KWIN_EDIT_OBJECT_POLYGON;
	break;
    }
    overlay_editor_set_pixcontext_attributes
	(top->overlayEditorControl.ctx,
	 KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT, kwin_object,
	 KWIN_EDITOR_CONTEXT_ATT_END);
}   /*  End Function object_mode_cbk  */

static void edit_mode_cbk (Widget w, XtPointer client_data,
			   XtPointer call_data)
/*  This is the edit mode callback.
*/
{
    int mode = *(int *) call_data;
    /*OverlayEditorControlWidget top = (OverlayEditorControlWidget) client_data;*/

    switch (mode)
    {
      case MODE_CREATE:
	break;
      case MODE_CHANGE:
	fprintf (stderr, "Only create mode supported currently\n");
	break;
      case MODE_DELETE:
	fprintf (stderr, "Only create mode supported currently\n");
	break;
    }
}   /*  End Function edit_mode_cbk  */

static void first_coord_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the first co-odinate type callback.
*/
{
    int coord_type = *(int *) call_data;
    unsigned int overlay_coord = 0;
    OverlayEditorControlWidget top = (OverlayEditorControlWidget) client_data;

    switch (coord_type)
    {
      case COORDSYS_PIXEL:
	overlay_coord = OVERLAY_COORD_PIXEL;
	break;
      case COORDSYS_RELATIVE:
	overlay_coord = OVERLAY_COORD_RELATIVE;
	break;
      case COORDSYS_LINEAR:
	overlay_coord = OVERLAY_COORD_LINEAR;
	break;
      case COORDSYS_WORLD:
	overlay_coord = OVERLAY_COORD_WORLD;
	break;
    }
    overlay_editor_set_attributes
	(top->overlayEditorControl.ctx,
	 OVERLAY_EDITOR_CONTEXT_ATT_FIRST_COORD_TYPE, overlay_coord,
	 OVERLAY_EDITOR_CONTEXT_ATT_END);
}   /*  End Function first_coord_cbk  */

static void other_coord_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the other co-odinate type callback.
*/
{
    int coord_type = *(int *) call_data;
    unsigned int overlay_coord = 0;
    OverlayEditorControlWidget top = (OverlayEditorControlWidget) client_data;

    switch (coord_type)
    {
      case COORDSYS_PIXEL:
	overlay_coord = OVERLAY_COORD_PIXEL;
	break;
      case COORDSYS_RELATIVE:
	overlay_coord = OVERLAY_COORD_RELATIVE;
	break;
      case COORDSYS_LINEAR:
	overlay_coord = OVERLAY_COORD_LINEAR;
	break;
      case COORDSYS_WORLD:
	overlay_coord = OVERLAY_COORD_WORLD;
	break;
    }
    overlay_editor_set_attributes
	(top->overlayEditorControl.ctx,
	 OVERLAY_EDITOR_CONTEXT_ATT_OTHER_COORD_TYPE, overlay_coord,
	 OVERLAY_EDITOR_CONTEXT_ATT_END);
}   /*  End Function other_coord_cbk  */

static void remove_last_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the remove last callback.
*/
{
    OverlayEditorControlWidget top = (OverlayEditorControlWidget) client_data;

    overlay_remove_objects (top->overlayEditorControl.olist, 1);
}   /*  End Function remove_last_cbk  */

static void remove_all_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the remove all callback.
*/
{
    OverlayEditorControlWidget top = (OverlayEditorControlWidget) client_data;

    overlay_remove_objects (top->overlayEditorControl.olist, 0);
}   /*  End Function remove_all_cbk  */

static void _OverlayEditorControl_save_dialog_cbk (Widget w,
						   XtPointer client_data,
						   XtPointer call_data)
{
    char *filename = (char *) call_data;
    OverlayEditorControlWidget top = (OverlayEditorControlWidget) client_data;

    overlay_write_karma (top->overlayEditorControl.olist, filename);
}   /*  End Function save_dialog_cbk  */

static void _OverlayEditorControl_text_cbk (Widget w, XtPointer client_data,
					    XtPointer call_data)
{
    flag apply = FALSE;
    String string;
    OverlayEditorControlWidget top = (OverlayEditorControlWidget) client_data;
    char *ptr1, *ptr2;
    char colour[STRING_LENGTH];
    /*static char function_name[] = "OverlayEditorControl::text_cbk";*/

    XtVaGetValues (top->overlayEditorControl.text_widget,
		   XtNstring, &string,
		   NULL);
    if (string[0] == '\0') return;
    for (ptr1 = string, ptr2 = colour; *ptr1 != '\0'; ++ptr1)
    {
	if (*ptr1 == '\n') apply = TRUE;
	else *ptr2++ = *ptr1;
    }
    *ptr2 = '\0';
    if (!apply) return;
    overlay_editor_set_attributes (top->overlayEditorControl.ctx,
				   OVERLAY_EDITOR_CONTEXT_ATT_COLOUR, colour,
				   OVERLAY_EDITOR_CONTEXT_ATT_END);
    overlay_editor_get_attributes (top->overlayEditorControl.ctx,
				   OVERLAY_EDITOR_CONTEXT_ATT_COLOUR, &ptr1,
				   OVERLAY_EDITOR_CONTEXT_ATT_END);
    /*  Have to break it into two sections  */
    XtVaSetValues (top->overlayEditorControl.text_widget,
		   XtNstring, ptr1,
		   NULL);
    XtVaSetValues (top->overlayEditorControl.text_widget,
		   XtNinsertPosition, ptr1,
		   NULL);
}   /*  End Function text_cbk  */


/*  Public functions follow  */

void XkwOverlayEditorControlAddCanvas (Widget W, KWorldCanvas canvas)
/*  [SUMMARY] Add a KWorldCanvas to an OverlayEditorControl widget canvas list.
    <W> The widget.
    <canvas> The canvas to add to the list. The widget will then manipulate
    this canvas. If this is NULL the routine does nothing.
    [RETURNS] Nothing.
*/
{
    OverlayEditorControlWidget w;
    struct canvas_list *new;
    static char function_name[] = "XkwOverlayEditorControlAddCanvas";

    if ( !XtIsOverlayEditorControl (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (OverlayEditorControlWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list structure");
    new->canvas = canvas;
    new->next = w->overlayEditorControl.first_canvas;
    w->overlayEditorControl.first_canvas = new;
    overlay_associate_display_canvas (w->overlayEditorControl.olist, canvas);
    overlay_editor_bind (w->overlayEditorControl.ctx, canvas, NULL);
}   /*  End Function XkwOverlayEditorControlAddCanvas  */
