/*LINTLIBRARY*/
/*  OverlayMenu.c

    This code provides a simple overlay menu widget for Xt.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a simple overlay menu widget for
  Xt.


    Written by      Richard Gooch   18-JUL-1997

    Updated by      Richard Gooch   21-JUL-1997: Fixed bug in <SetValues> which
  caused SEGV because of NULL pointer access.

    Updated by      Richard Gooch   3-AUG-1997: Added overlay editor choice.

    Updated by      Richard Gooch   4-AUG-1997: Removed remove choice because
  editor has those controls now.

    Updated by      Richard Gooch   9-AUG-1997: Moved <overlay_read_ascii> to
  <overlay> package.

    Updated by      Richard Gooch   10-AUG-1997: Create various popup widgets
  on demand rather than at initialisation depending on XkwNcreateAtInit
  resource.

    Updated by      Richard Gooch   22-FEB-1998: Created function
  <XkwDressingControlAddCanvas>.

    Updated by      Richard Gooch   2-MAR-1998: Created remove buttons on file
  browser.

    Updated by      Richard Gooch   5-MAR-1998: Make file browser pinned by
  default.

    Updated by      Richard Gooch   2-SEP-1998: Create connection overlay list
  and make distinction with file overlay list.

    Updated by      Richard Gooch   16-SEP-1998: Added contour loading support.

    Updated by      Richard Gooch   14-APR-1999: Added support for loading
  overlays in Karma format.

    Updated by      Richard Gooch   21-APR-1999: Fixed <accept_file> to show
  Karma ".desc" and ".dat" files as a single ".kf" file.

    Updated by      Richard Gooch   22-AUG-1999: Include filename in title for
  ContourSimpleControl widget.

    Updated by      Richard Gooch   8-MAY-2000: Added crop toggle.

    Updated by      Richard Gooch   20-MAR-2001: Created Destroy method.

    Updated by      Richard Gooch   7-SEP-2001: Removed initialisation of crop
  toggle to True so that users can change the (new) default of False in their
  resources file.

    Last updated by Richard Gooch   25-SEP-2001: Removed #include for var-args.


*/
#include <Xkw/OverlayMenuP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <stdio.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/ChoiceMenu.h>
#include <Xkw/DressingControl.h>
#include <Xkw/Filepopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/OverlayEditorControl.h>
#include <Xkw/ContourSimpleControl.h>
#include <Xkw/Ktoggle.h>


struct canvas_list
{
    KWorldCanvas canvas;
    struct canvas_list *next;
};


/*  Private functions  */
STATIC_FUNCTION (void _OverlayMenu_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _OverlayMenu_Destroy, (Widget W) );
STATIC_FUNCTION (void overlay_cbk, (Widget W, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (flag _OverlayMenu_accept_file, (KFileInfo finfo) );
STATIC_FUNCTION (void _OverlayMenu_filesel_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void create_dressing_control, (OverlayMenuWidget top) );
STATIC_FUNCTION (void create_annotation_filepopup, (OverlayMenuWidget top) );
STATIC_FUNCTION (void create_editor_control, (OverlayMenuWidget top) );
STATIC_FUNCTION (void create_contour_filepopup, (OverlayMenuWidget top) );
STATIC_FUNCTION (void remove_last_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void remove_all_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void crop_toggle_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag data_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


#define offset(field) XtOffset(OverlayMenuWidget, overlayMenu.field)

static XtResource OverlayMenuResources[] = 
{
    {XtNlabel, XtCLabel, XtRString, sizeof (String),
     offset (label), XtRString, (XtPointer) "Overlays"},
    {XkwNcreateAtInit, XkwCCreateAtInit, XtRBool, sizeof (Bool),
     offset (createAtInit), XtRImmediate, (XtPointer) FALSE},
};

#undef offset

#define OVERLAY_AXIS_LABELS         0
#define OVERLAY_LOAD_ANNOTATIONS    1
#define OVERLAY_EDITOR              2
#define OVERLAY_REMOVE_LAST         3
#define OVERLAY_REMOVE_ALL          4
#define OVERLAY_CONTOURS            5
#define NUM_OVERLAY_CHOICES         6

static char *overlay_choices[NUM_OVERLAY_CHOICES] =
{
    "Axis Labels", "Load Annotations (file)", "Editor (manual)",
    "Remove Last (connection)", "Remove All (connection)",
    "Contours",
};

OverlayMenuClassRec overlayMenuClassRec = 
{
  {        /* CoreClassPart */
      (WidgetClass) &formClassRec,              /* superclass */
      "OverlayMenu",                            /* class_name */
      sizeof(OverlayMenuRec),                   /* widget_size */
      NULL,                                     /* class_initialise */
      NULL,                                     /* class_part_initialise */
      FALSE,                                    /* class_init */
      (XtInitProc) _OverlayMenu_Initialise,     /* initialise */
      NULL,                                     /* initialise_hook */
      XtInheritRealize,                         /* realise */
      NULL,                                     /* actions */
      0,                                        /* num_actions */
      OverlayMenuResources,                     /* resources */
      XtNumber(OverlayMenuResources),           /* num_resources */
      NULLQUARK,                                /* xrm_class */
      TRUE,                                     /* compress_motion */
      TRUE,                                     /* compress_exposure */
      TRUE,                                     /* compress_enterleave */
      TRUE,                                     /* visible_interest */
      _OverlayMenu_Destroy,                     /* destroy */
      XtInheritResize,                          /* resize */
      XtInheritExpose,                          /* expose */
      NULL,                                     /* set_values */
      NULL,                                     /* set_values_hook */
      XtInheritSetValuesAlmost,                 /* set_values_almost */
      NULL,                                     /* get_values_hook */
      NULL,                                     /* accept_focus */
      XtVersion,                                /* version */
      NULL,                                     /* callback_private */
      XtInheritTranslations,                    /* tm_translations */
      NULL,
      NULL,
      NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Constraint */
    /* subresourses       */   NULL,
    /* subresource_count  */   0,
    /* constraint_size    */   sizeof (OverlayMenuConstraintsRec),
    /* initialise         */   NULL,
    /* destroy            */   NULL,
    /* set_values         */   NULL,
    /* extension          */   NULL
  },{ /* Form */
    /* layout */                XtInheritLayout
  },
  {  /* OverlayMenuClassPart */
    0 /* empty */
  },
};

WidgetClass overlayMenuWidgetClass = (WidgetClass) &overlayMenuClassRec;

static void _OverlayMenu_Initialise (Widget Request, Widget New)
{
    /*OverlayMenuWidget request = (OverlayMenuWidget) Request;*/
    OverlayMenuWidget new = (OverlayMenuWidget) New;
    Widget w;
    char olist_name[STRING_LENGTH];
    static unsigned int sequence = 0;
    /*static char function_name[] = "OverlayMenu::Initialise";*/

    new->form.default_spacing = 0;
    new->core.border_width = 0;
    new->overlayMenu.dressing_control_popup = NULL;
    new->overlayMenu.annotation_filepopup = NULL;
    new->overlayMenu.editor_popup = NULL;
    new->overlayMenu.contour_filepopup = NULL;
    new->overlayMenu.first_canvas = NULL;
    sprintf (olist_name, "connectionOverlayList:%u\n", sequence);
    new->overlayMenu.conn_olist = overlay_va_create_list (NULL, olist_name,
							  OVERLAY_ATT_END);
    sprintf (olist_name, "fileOverlayList:%u\n", sequence);
    new->overlayMenu.file_olist = overlay_va_create_list (NULL, olist_name,
							  OVERLAY_ATT_END);
    ++sequence;
    w = XtVaCreateManagedWidget ("menuButton", choiceMenuWidgetClass, New,
				 XtNlabel, new->overlayMenu.label,
				 XkwNmenuTitle, "Overlay Menu",
				 XtNmenuName, "overlayMenu",
				 XkwNnumItems, NUM_OVERLAY_CHOICES,
				 XkwNitemStrings, overlay_choices,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, overlay_cbk, New);
    if (new->overlayMenu.createAtInit)
    {
	create_dressing_control (new);
	create_annotation_filepopup (new);
	create_editor_control (new);
	create_contour_filepopup (new);
    }
}   /*  End Function Initialise  */

static void _OverlayMenu_Destroy (Widget W)
{
    OverlayMenuWidget w = (OverlayMenuWidget) W;
    struct canvas_list *curr, *next;

    overlay_destroy_list (w->overlayMenu.file_olist);
    overlay_destroy_list (w->overlayMenu.conn_olist);
    for (curr = w->overlayMenu.first_canvas; curr != NULL; curr = next)
    {
	next = curr->next;
	m_free (curr);
    }
}   /*  End Function Destroy  */

static void overlay_cbk (Widget W, XtPointer client_data, XtPointer call_data)
/*  This is the overlay menu callback.
*/
{
    unsigned int code = *(int *) call_data;
    OverlayMenuWidget top = (OverlayMenuWidget) client_data;
    static char function_name[] = "OverlayMenuWidget::overlay_cbk";

    switch (code)
    {
      case OVERLAY_AXIS_LABELS:
	create_dressing_control (top);
	XtPopup (top->overlayMenu.dressing_control_popup, XtGrabNone);
	break;
      case OVERLAY_LOAD_ANNOTATIONS:
	create_annotation_filepopup (top);
	XtPopup (top->overlayMenu.annotation_filepopup, XtGrabNone);
	break;
      case OVERLAY_EDITOR:
	create_editor_control (top);
	XtPopup (top->overlayMenu.editor_popup, XtGrabNone);
	break;
      case OVERLAY_REMOVE_LAST:
	overlay_remove_objects (top->overlayMenu.conn_olist, 1);
	break;
      case OVERLAY_REMOVE_ALL:
	overlay_remove_objects (top->overlayMenu.conn_olist, 0);
	break;
      case OVERLAY_CONTOURS:
	create_contour_filepopup (top);
	XtPopup (top->overlayMenu.contour_filepopup, XtGrabNone);
	break;
      default:
	fprintf (stderr, "Illegal overlay code: %u\n", code);
	a_prog_bug (function_name);
	break;
    }
}   /*  End Function overlay_cbk   */

static flag _OverlayMenu_accept_file (KFileInfo finfo)
{
    char *ptr;

    if (finfo.type == KFILETYPE_DIRECTORY) return (TRUE);
    if (foreign_guess_format_from_filename (finfo.filename) ==
	FOREIGN_FILE_FORMAT_KARMA)
    {
	/*  It's a Karma file: fake things a bit for ".desc"/".dat" pairs  */
	if ( ( ptr = strrchr (finfo.filename, '.') ) == NULL ) return (FALSE);
	/*  Ignore the ".dat" file  */
	if (strcmp (ptr, ".dat") == 0) return (FALSE);
	/*  Fiddle if the ".desc" file  */
	if (strcmp (ptr, ".desc") == 0) strcpy (ptr, ".kf");
	return (TRUE);
    }
    if (strcmp (finfo.filename + strlen (finfo.filename) - 4, ".ann") == 0)
	return (TRUE);
    /*  Reject everything else  */
    return (FALSE);
}   /*  End Function accept_file  */

static void _OverlayMenu_filesel_cbk (Widget w, XtPointer client_data,
				      XtPointer call_data)
{
    char *filename = (char *) call_data;
    OverlayMenuWidget top = (OverlayMenuWidget) client_data;

    if (foreign_guess_format_from_filename (filename) ==
	FOREIGN_FILE_FORMAT_KARMA)
    {
	overlay_read_karma (top->overlayMenu.file_olist, filename);
	return;
    }
    overlay_read_ascii (top->overlayMenu.file_olist, filename);
}   /*  End Function filesel_cbk  */

static void create_dressing_control (OverlayMenuWidget top)
/*  [SUMMARY] Create the dressing control widget.
    <top> The OverlayMenu widget.
    [RETURNS] Nothing.
*/
{
    Widget w;
    struct canvas_list *curr;

    if (top->overlayMenu.dressing_control_popup != NULL) return;
    w = XtVaCreatePopupShell ("dressingControlPopup",
			      dressingControlWidgetClass, (Widget) top,
			      NULL);
    top->overlayMenu.dressing_control_popup = w;
    for (curr = top->overlayMenu.first_canvas; curr != NULL; curr = curr->next)
	XkwDressingControlAddCanvas (w, curr->canvas);
}   /*  End Function create_dressing_control  */

static void create_annotation_filepopup (OverlayMenuWidget top)
/*  [SUMMARY] Create the filepopup widget.
    <top> The OverlayMenu widget.
    [RETURNS] Nothing.
*/
{
    Widget w, filewin, box;

    if (top->overlayMenu.annotation_filepopup != NULL) return;
    w = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass,
			      (Widget) top,
			      XtNtitle, "Annotation File Selector",
			      XtNx, 0,
			      XtNy, 0,
			      XkwNfilenameTester,
			      _OverlayMenu_accept_file,
			      XkwNautoPopdown, True,
			      NULL);
    top->overlayMenu.annotation_filepopup = w;
    XtAddCallback (w, XkwNfileSelectCallback, _OverlayMenu_filesel_cbk, top);
    XtVaSetValues (XtNameToWidget (w, "form.box.pinToggle"),
		   XtNstate, True,
		   NULL);
    XtVaGetValues (w, XkwNfilewinWidget, &filewin, NULL);
    XtVaGetValues (w, XkwNboxWidget, &box, NULL);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
				 XtNlabel, "Remove Last",
				 /*XtNheight, 20,*/
				 NULL);
    XtAddCallback (w, XtNcallback, remove_last_cbk,
		   top->overlayMenu.file_olist);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
				 XtNlabel, "Remove All",
				 /*XtNheight, 20,*/
				 NULL);
    XtAddCallback (w, XtNcallback, remove_all_cbk,
		   top->overlayMenu.file_olist);
    w = XtVaCreateManagedWidget ("cropToggle", ktoggleWidgetClass, box,
				 XtNlabel, "Crop",
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, crop_toggle_cbk,
		   top->overlayMenu.file_olist);
}   /*  End Function create_annotation_filepopup  */

static void create_editor_control (OverlayMenuWidget top)
/*  [SUMMARY] Create the editor control widget.
    <top> The OverlayMenu widget.
    [RETURNS] Nothing.
*/
{
    Widget n;
    struct canvas_list *curr;

    if (top->overlayMenu.editor_popup != NULL) return;
    n = XtVaCreatePopupShell ("editorPopup", overlayEditorControlWidgetClass,
			      (Widget) top,
			      XtNtitle, "Annotation Editor",
			      NULL);
    top->overlayMenu.editor_popup = n;
    for (curr = top->overlayMenu.first_canvas; curr != NULL; curr = curr->next)
	XkwOverlayEditorControlAddCanvas (n, curr->canvas);
}   /*  End Function create_editor_control  */

static void create_contour_filepopup (OverlayMenuWidget top)
/*  [SUMMARY] Create the contour control widget.
    <top> The OverlayMenu widget.
    [RETURNS] Nothing.
*/
{
    Widget filepopup, filewin;
    char domain[STRING_LENGTH];

    if (top->overlayMenu.contour_filepopup != NULL) return;
    sprintf (domain, "OverlayMenu::Contour:%p", top);
    filepopup = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass,
				      (Widget) top,
				      XtNtitle, "Contour File Selector",
				      XtNx, 0,
				      XtNy, 0,
				      XkwNfilenameTester,
				      XkwFilewinStandardFileTester_nD,
				      XkwNautoPopdown, True,
				      XkwNdomainName, domain,
				      XkwNloadFiles, TRUE,
				      NULL);
    top->overlayMenu.contour_filepopup = filepopup;
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (data_event_func, top);
}   /*  End Function create_contour_filepopup  */

static void remove_last_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the remove last callback.
*/
{
    overlay_remove_objects (client_data, 1);
}   /*  End Function remove_last_cbk  */

static void remove_all_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the remove all callback.
*/
{
    overlay_remove_objects (client_data, 0);
}   /*  End Function remove_all_cbk  */

static void crop_toggle_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the crop toggle callback.
*/
{
    flag bool = (uaddr) call_data;

    overlay_set_attributes (client_data,
			    OVERLAY_ATT_CROP, bool,
			    OVERLAY_ATT_END);
}   /*  End Function remove_all_cbk  */

static flag data_event_func (void *object, multi_array *multi_desc,
			     CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    iarray array = NULL;
    OverlayMenuWidget top = object;
    Widget w;
    struct canvas_list *curr;
    char txt[STRING_LENGTH];
    static char function_name[] = "OverlayMenu::data_event_func";

    if (top->overlayMenu.contour_filepopup == NULL) return (FALSE);
    sprintf (txt, "OverlayMenu::Contour:%p", top);
    if (strcmp (domain, txt) != 0) return (FALSE);
    if ( !iarray_create_and_setup (&array, multi_desc, TRUE, 2, NONE,
				   NULL, NULL, TRUE, NULL) )
    {
	m_error_notify (function_name, "iarray");
	return (TRUE);
    }
    sprintf (txt, "Contour Control for: %s", name);
    w = XtVaCreatePopupShell ("contourControl",
			      contourSimpleControlWidgetClass, (Widget) top,
			      XtNtitle, txt,
			      XkwNiarray, array,
			      NULL);
    
    for (curr = top->overlayMenu.first_canvas; curr != NULL; curr = curr->next)
	XkwContourSimpleControlAddCanvas (w, curr->canvas);
    XtPopup (w, XtGrabNone);
    return (TRUE);
}   /*  End Function data_event_func  */


/*  Public functions follow  */

void XkwOverlayMenuAddCanvas (Widget W, KWorldCanvas canvas)
/*  [SUMMARY] Add a KWorldCanvas to an OverlayMenu widget canvas list.
    <W> The widget.
    <canvas> The canvas to add to the list. The widget will then manipulate
    this canvas. If this is NULL the routine does nothing.
    [RETURNS] Nothing.
*/
{
    OverlayMenuWidget w;
    struct canvas_list *new;
    static char function_name[] = "XkwOverlayMenuAddCanvas";

    if ( !XtIsOverlayMenu (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (OverlayMenuWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list_entry structure");
    new->canvas = canvas;
    new->next = w->overlayMenu.first_canvas;
    w->overlayMenu.first_canvas = new;
    if (w->overlayMenu.dressing_control_popup != NULL)
	XkwDressingControlAddCanvas (w->overlayMenu.dressing_control_popup,
				     canvas);
    if (w->overlayMenu.editor_popup != NULL)
	XkwOverlayEditorControlAddCanvas (w->overlayMenu.editor_popup, canvas);
    overlay_associate_display_canvas (w->overlayMenu.conn_olist, canvas);
    overlay_associate_display_canvas (w->overlayMenu.file_olist, canvas);
}   /*  End Function XkwOverlayMenuAddCanvas  */
