/*LINTLIBRARY*/
/*  PannerPopup.c

    This code provides a panner display widget for Xt.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a panner display
    widget for Xt.


    Written by      Richard Gooch   6-DEC-1997

    Updated by      Richard Gooch   7-DEC-1997

    Updated by      Richard Gooch   22-FEB-1998: Created function
  <XkwMagnifierPopupAddCanvas>. Improved sequencing and prevent double refresh
  events on panner canvas.

    Updated by      Richard Gooch   10-MAR-1998: Fixed bugs when panner popped
  up and image not loaded.

    Updated by      Richard Gooch   12-AUG-1998: Support new event
  K_CANVAS_EVENT_POINTER_WARP.

    Updated by      Richard Gooch   14-AUG-1999: Register <position_func> in
  <XkwPannerPopupAddCanvas> rather than <realised_cbk>.

    Updated by      Richard Gooch   15-AUG-1999: Undraw old box rather than
  doing a full refresh.

    Updated by      Richard Gooch   29-SEP-1999: Trap and discard pointer enter
  and leave events. Switched to <canvas_resize> so that old axis labelling and
  image data is undrawn correctly.

    Updated by      Richard Gooch   30-SEP-1999: Changed <position_func> from
  KWorldCanvas callback to ViewableImage callback, so that image pixel position
  need not be computed by callback.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>.

    Updated by      Richard Gooch   12-JUL-2000: Removed unnecessary #includes.

    Last updated by Richard Gooch   20-MAR-2001: Created Destroy method.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_kwin.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/PannerPopupP.h>
#include <Xkw/ImageDisplay.h>


struct canvas_list
{
    KWorldCanvas canvas;
    struct canvas_list *next;
};


/*  Private functions  */
STATIC_FUNCTION (void _PannerPopup_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _PannerPopup_Destroy, (Widget W) );
STATIC_FUNCTION (void realised_cbk,
		 (Widget parent, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (unsigned long get_colour,
		 (KWorldCanvas canvas, CONST char *colourname) );
STATIC_FUNCTION (void panner_refresh,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void master_refresh,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (flag position_func,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin,
		  unsigned int value_type,
		  double x_im, double y_im, int x_pix, int y_pix) );
STATIC_FUNCTION (void popup_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void close_cbk,
		 (Widget parent, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffset(PannerPopupWidget, pannerPopup.field)

static XtResource PannerPopupResources[] = 
{
    {"empty", "Empty", XtRInt, sizeof (int),
     offset (dummy), XtRImmediate, False},
};

#undef offset

PannerPopupClassRec pannerPopupClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &slaveImageDisplayPopupClassRec, /* superclass */
    "PannerPopup",                            /* class_name */
    sizeof (PannerPopupRec),                  /* widget_size */
    NULL,                                     /* class_initialise */
    NULL,                                     /* class_part_initialise */
    FALSE,                                    /* class_init */
    (XtInitProc) _PannerPopup_Initialise,     /* initialise */
    NULL,                                     /* initialise_hook */
    XtInheritRealize,                         /* realise */
    NULL,                                     /* actions */
    0,                                        /* num_actions */
    PannerPopupResources,                     /* resources */
    XtNumber (PannerPopupResources),          /* num_resources */
    NULLQUARK,                                /* xrm_class */
    TRUE,                                     /* compress_motion */
    TRUE,                                     /* compress_exposure */
    TRUE,                                     /* compress_enterleave */
    TRUE,                                     /* visible_interest */
    _PannerPopup_Destroy,                     /* destroy */
    XtInheritResize,                          /* resize */
    NULL,                                     /* expose */
    NULL,                                     /* set_values */
    NULL,                                     /* set_values_hook */
    XtInheritSetValuesAlmost,                 /* set_values_almost */
    NULL,                                     /* get_values_hook */
    NULL,                                     /* accept_focus */
    XtVersion,                                /* version */
    NULL,                                     /* callback_private */
    NULL,                                     /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* SlaveImageDisplayPopupClassPart */
      0                   /* empty */
  },
  {  /* PannerPopupClassPart */
      0                   /* empty */
  },
};

WidgetClass pannerPopupWidgetClass = (WidgetClass) &pannerPopupClassRec;


static void _PannerPopup_Initialise (Widget Request, Widget New)
{
    /*PannerPopupWidget request = (PannerPopupWidget) Request;*/
    PannerPopupWidget new = (PannerPopupWidget) New;
    /*static char function_name[] = "PannerPopupWidget::Initialise";*/

    new->pannerPopup.first_canvas = NULL;
    new->pannerPopup.areas = NULL;
    XtAddCallback (New, XkwNrealisedCallback, realised_cbk, New);
}   /*  End Function Initialise  */

static void _PannerPopup_Destroy (Widget W)
{
    PannerPopupWidget w = (PannerPopupWidget) W;
    struct canvas_list *curr, *next;

    for (curr = w->pannerPopup.first_canvas; curr != NULL; curr = next)
    {
	next = curr->next;
	m_free (curr);
    }
}   /*  End Function Destroy  */

static void realised_cbk (Widget parent, XtPointer client_data,
			  XtPointer call_data)
{
    PannerPopupWidget top = (PannerPopupWidget) client_data;
    SlaveImageDisplayPopupPart *slave =
	(SlaveImageDisplayPopupPart *) &top->slaveImageDisplayPopup;

    XtAddCallback ( (Widget) top, XtNpopupCallback, popup_cbk, NULL );
    XtAddCallback (slave->close_btn, XtNcallback, close_cbk, top);
    if (slave->pseudoCanvas != NULL)
    {
	canvas_register_refresh_func (slave->pseudoCanvas, panner_refresh,top);
	viewimg_register_position_event_func (slave->pseudoCanvas,
					      position_func, top);
    }
    if (slave->directCanvas != NULL)
    {
	canvas_register_refresh_func (slave->directCanvas, panner_refresh,top);
	viewimg_register_position_event_func (slave->directCanvas,
					      position_func, top);
    }
    if (slave->trueCanvas != NULL)
    {
	canvas_register_refresh_func (slave->trueCanvas, panner_refresh, top);
	viewimg_register_position_event_func (slave->trueCanvas,
					      position_func, top);
    }
}   /*  End Function realised_cbk  */

static unsigned long get_colour (KWorldCanvas canvas, CONST char *colourname)
/*  [SUMMARY] Get colour.
    <canvas> The world canvas.
    <colourname> The name of the colour.
    [RETURNS] The specified colour if available, else a default colour.
*/
{
    KPixCanvas pixcanvas;
    unsigned long pixel_value;

    pixcanvas = canvas_get_pixcanvas (canvas);
    if ( kwin_get_colour (pixcanvas, colourname, &pixel_value,
			  NULL, NULL, NULL) )
    {
	return (pixel_value);
    }
    if ( kwin_get_colour (pixcanvas, "white", &pixel_value, NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "Error allocating colour: \"%s\" defaulting to white\n",
		 colourname);
	return (pixel_value);
    }
    fprintf (stderr, "Error allocating colour: \"%s\" defaulting to 0\n",
	     colourname);
    return (0);
}   /*  End Function get_colour  */

static void panner_refresh (KWorldCanvas canvas, int width, int height,
			    struct win_scale_type *win_scale,
			    Kcolourmap cmap, flag cmap_resize, void **info,
			    PostScriptPage pspage,
			    unsigned int num_areas,
			    KPixCanvasRefreshArea *areas, flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas master_wc;
    unsigned long pixel_value;
    double left_x, right_x, bottom_y, top_y;
    Widget TOP = (Widget) *info;

    XtVaGetValues (XtParent (TOP), XkwNvisibleCanvas, &master_wc, NULL);
    canvas_get_attributes (master_wc,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &left_x, &bottom_y, &left_x, &bottom_y);
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &right_x, &top_y, &right_x, &top_y);
    pixel_value = get_colour (canvas, "green");
    kwin_draw_rectangle (canvas_get_pixcanvas (canvas), left_x, top_y,
			 right_x - left_x, bottom_y - top_y, pixel_value);
}   /*  End Function panner_refresh  */

static void master_refresh (KWorldCanvas canvas, int width, int height,
			    struct win_scale_type *win_scale,
			    Kcolourmap cmap, flag cmap_resize, void **info,
			    PostScriptPage pspage,
			    unsigned int num_areas,
			    KPixCanvasRefreshArea *areas, flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KPixCanvas panner_pc;
    ViewableImage vimage;
    flag visible;
    PannerPopupWidget top = (PannerPopupWidget) *info;
    SlaveImageDisplayPopupPart *slave =
	(SlaveImageDisplayPopupPart *) &top->slaveImageDisplayPopup;

    panner_pc = canvas_get_pixcanvas (slave->visibleCanvas);
    kwin_get_attributes (panner_pc, KWIN_ATT_VISIBLE, &visible, KWIN_ATT_END);
    if (!visible) return;
    vimage = viewimg_get_active (slave->visibleCanvas);
    if (vimage == NULL) return;
    if (top->pannerPopup.areas == NULL)
	kwin_resize (panner_pc, FALSE, 0, 0, 0, 0);
    else viewimg_partial_refresh (slave->visibleCanvas,
				  4, top->pannerPopup.areas);
}   /*  End Function master_refresh  */

static flag position_func (ViewableImage vimage, double x, double y,
			   void *value, unsigned int event_code,
			   void *e_info, void **f_info,
			   double x_lin, double y_lin,
			   unsigned int value_type,
			   double x_im, double y_im, int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transform_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas master_wc, panner_wc;
    ViewableImage master_vimage;
    double left_x, right_x, bottom_y, top_y;
    PannerPopupWidget top = (PannerPopupWidget) *f_info;
    Widget TOP = (Widget) *f_info;
    KPixCanvasRefreshArea panner_areas[4];

    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
      case K_CANVAS_EVENT_POINTER_WARP:
	break;
      default:
	XBell (XtDisplay (TOP), 33);
	/*  Fall through  */
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_POINTER_ENTER:
      case K_CANVAS_EVENT_POINTER_LEAVE:
	return (TRUE);
	/*break;*/
    }
    XtVaGetValues (XtParent (TOP), XkwNvisibleCanvas, &master_wc, NULL);
    master_vimage = viewimg_get_active (master_wc);
    if (master_vimage == NULL) return (TRUE);
    canvas_get_attributes (master_wc,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    /*  Setup areas describing old rectangle to be undrawn (in panner canvas)
     */
    panner_wc = viewimg_get_worldcanvas (vimage);
    canvas_convert_from_canvas_coords (panner_wc, FALSE, TRUE, 1,
				       &left_x, &bottom_y, &left_x, &bottom_y);
    canvas_convert_from_canvas_coords (panner_wc, FALSE, TRUE, 1,
				       &right_x, &top_y, &right_x, &top_y);
    m_clear (panner_areas, sizeof panner_areas);
    /*  Top line  */
    panner_areas[0].startx = left_x - 1.0;
    panner_areas[0].endx = right_x + 1.0;
    panner_areas[0].starty = top_y - 1.0;
    panner_areas[0].endy = top_y + 1.0;
    panner_areas[0].clear = TRUE;
    /*  Bottom line  */
    panner_areas[1].startx = left_x - 1.0;
    panner_areas[1].endx = right_x + 1.0;
    panner_areas[1].starty = bottom_y - 1.0;
    panner_areas[1].endy = bottom_y + 1.0;
    panner_areas[1].clear = TRUE;
    /*  Left line  */
    panner_areas[2].startx = left_x - 1.0;
    panner_areas[2].endx = left_x + 1.0;
    panner_areas[2].starty = top_y - 1.0;
    panner_areas[2].endy = bottom_y + 1.0;
    panner_areas[2].clear = TRUE;
    /*  Right line  */
    panner_areas[3].startx = right_x - 1.0;
    panner_areas[3].endx = right_x + 1.0;
    panner_areas[3].starty = top_y - 1.0;
    panner_areas[3].endy = bottom_y + 1.0;
    panner_areas[3].clear = TRUE;
    viewimg_set_canvas_attributes (master_wc,
				   VIEWIMG_ATT_PAN_CENTRE_X,
				   (unsigned long) (x_im + 0.5),
				   VIEWIMG_ATT_PAN_CENTRE_Y,
				   (unsigned long) (y_im + 0.5),
				   VIEWIMG_ATT_END);
    top->pannerPopup.areas = panner_areas;
    canvas_resize (master_wc, NULL, FALSE);
    top->pannerPopup.areas = NULL;
    return (TRUE);
}   /*  End Function position_func  */

static void popup_cbk (Widget w, XtPointer client_data, XtPointer call_data)
{
    PannerPopupWidget top = (PannerPopupWidget) w;
    KWorldCanvas master_wc;
    ViewableImage vimage;
    unsigned int hdim, vdim;
    array_desc *arr_desc;
    dim_desc *hd, *vd;
    struct canvas_list *curr;
    Widget TOP = (Widget) top;

    /*  Change attributes on master canvases  */
    for (curr = top->pannerPopup.first_canvas; curr != NULL; curr = curr->next)
    {
	viewimg_set_canvas_attributes (curr->canvas,
				       VIEWIMG_ATT_AUTO_X, FALSE,
				       VIEWIMG_ATT_AUTO_Y, FALSE,
				       VIEWIMG_ATT_ENABLE_PANNING, TRUE,
				       VIEWIMG_ATT_PAN_MAGNIFICATION, 1,
				       VIEWIMG_ATT_END);
    }
    XtVaGetValues (XtParent (TOP), XkwNvisibleCanvas, &master_wc, NULL);
    if (master_wc == NULL) return;
    vimage = viewimg_get_active (master_wc);
    if (vimage == NULL) return;
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_END);
    hd = arr_desc->dimensions[hdim];
    vd = arr_desc->dimensions[vdim];
    viewimg_set_canvas_attributes (master_wc,
				   VIEWIMG_ATT_PAN_CENTRE_X, hd->length / 2,
				   VIEWIMG_ATT_PAN_CENTRE_Y, vd->length / 2,
				   VIEWIMG_ATT_END);
    kwin_resize (canvas_get_pixcanvas (master_wc), TRUE, 0, 0, 0, 0);
}   /*  End Function popup_cbk  */

static void close_cbk (Widget parent, XtPointer client_data,
		       XtPointer call_data)
{
    PannerPopupWidget top = (PannerPopupWidget) client_data;
    Widget TOP = (Widget) client_data;
    struct canvas_list *curr;

    /*  Change attributes on master canvases  */
    for (curr = top->pannerPopup.first_canvas; curr != NULL; curr = curr->next)
    {
	viewimg_set_canvas_attributes (curr->canvas,
				       VIEWIMG_ATT_ENABLE_PANNING, FALSE,
				       VIEWIMG_ATT_END);
    }
    XkwImageDisplayRefresh (XtParent (TOP), TRUE);
}   /*  End Function close_cbk  */


/*  Public functions follow  */

void XkwPannerPopupAddCanvas (Widget W, KWorldCanvas canvas)
/*  [SUMMARY] Add a KWorldCanvas to a PannerPopup widget canvas list.
    <W> The widget.
    <canvas> The parent canvas to add to the list. The widget will then
    manipulate this canvas. If this is NULL the routine does nothing.
    [RETURNS] Nothing.
*/
{
    PannerPopupWidget w;
    struct canvas_list *new;
    static char function_name[] = "XkwPannerPopupAddCanvas";

    if ( !XtIsPannerPopup (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (PannerPopupWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list structure");
    new->canvas = canvas;
    new->next = w->pannerPopup.first_canvas;
    w->pannerPopup.first_canvas = new;
    canvas_register_refresh_func (canvas, master_refresh, W);
}   /*  End Function XkwPannerPopupAddCanvas  */
