/*LINTLIBRARY*/
/*  Postscript.c

    This code provides a postscript control widget for Xt.

    PostScript code Copyright (C) 1994-2001  Richard Gooch

    Original widget code Copyright (C) 1994  Patrick Jordan
    Subsequent widget code Copyright (C) 1994-2001  Richard Gooch
    Incorporated into Karma by permission.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Patrick Jordan may be reached by email at  pjordan@rp.csiro.au
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a postscript control
    widget for Xt.


    Written by      Patrick Jordan  26-JUN-1994: Copied PostScript code from
  print_xview.c  and hard_copy.c  in  source/modules/kview_2d (XView version of
  kview_2d  ).

    Updated by      Patrick Jordan  27-JUL-1994: Removed print_mono button
   as code not implemented yet.

    Updated by      Richard Gooch   23-OCT-1994: Changed some resource names.

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   9-APR-1995: Initialised  arrayfile_name
  field to NULL to avoid freeing a non-existant block.

    Updated by      Richard Gooch   13-JUL-1995: Added resources to change
  orientation, page offset and size.

    Updated by      Richard Gooch   30-JUL-1995: Made use of XkwNvaluePtr
  resource for ValueWidgetClass.

    Updated by      Richard Gooch   23-SEP-1995: Coped with NULL name pointer
  to <XkwPostscriptRegisterImageAndName>

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   7-JUN-1996: Added dialog widget for output
  filename.

    Updated by      Richard Gooch   28-JUN-1996: Created XkwNautoIncrement
  resource.

    Updated by      Richard Gooch   1-JUL-1996: Switched to <psw_va_create> and
  <psw_close> routines.

    Updated by      Richard Gooch   30-OCT-1996: Added #include <sys/types.h>

    Updated by      Richard Gooch   20-DEC-1996: Added dialog for queue name.

    Updated by      Richard Gooch   16-JAN-1997: Added tray selection.

    Updated by      Richard Gooch   5-FEB-1997: Fixed bug where tray member was
  not initialised.

    Updated by      Richard Gooch   20-MAY-1997: Ensure square pixels on screen
  are square on paper.

    Updated by      Richard Gooch   4-JUN-1997: Support paper tray.

    Updated by      Richard Gooch   9-JUN-1997: Ensure square pixels on screen
  are square on paper for landscape orientation as well.

    Updated by      Richard Gooch   2-AUG-1997: Made use of new function
  <r_fork_and_execvp>.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <r> package.

    Updated by      Richard Gooch   23-OCT-1998: Added event
  "PostscriptWidget::WritePS".

    Updated by      Richard Gooch   6-AUG-1999: Added linewidth slider.

    Updated by      Richard Gooch   9-SEP-1999: Added "Lock Filename" toggle.

    Updated by      Richard Gooch   22-MAY-2000: Increased size limit to 1m.

    Updated by      Richard Gooch   9-JUN-2000: Removed unused variables.

    Updated by      Richard Gooch   13-MAR-2001: Allow negative offsets.

    Updated by      Richard Gooch   21-MAR-2001: Created Destroy method.

    Last updated by Richard Gooch   25-SEP-2001: Removed #include for var-args.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>
#include <signal.h>
#include <Xkw/PostscriptP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#include <karma_xtmisc.h>
#include <karma_kwin.h>
#include <karma_psw.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_ch.h>
#include <karma_r.h>
#include <karma_m.h>
#include <karma_c.h>
#include <Xkw/Value.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/SimpleSlider.h>

/* Methods*/
STATIC_FUNCTION (void orientation_cbk, (Widget w, XtPointer client_data,
					XtPointer call_data) );
STATIC_FUNCTION (void _Postscript_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _Postscript_Destroy, (Widget W) );
STATIC_FUNCTION (void auto_increment_cbk, (Widget w, XtPointer client_data,
					   XtPointer call_data) );
STATIC_FUNCTION (void tray_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (flag _Postscript_ds_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


#define TheOffset(field) XtOffset(PostscriptWidget, postscript.field)

static XtResource PostscriptResources[] = 
{
    {XkwNportraitOrientation, XkwCPortraitOrientation, XtRBool, sizeof (Bool),
     TheOffset (portrait), XtRImmediate, (XtPointer) True},
    {XkwNpageHorizontalOffset, XkwCPageHorizontalOffset, XtRInt, sizeof (int),
     TheOffset (hoffset), XtRImmediate, (XtPointer) 10},
    {XkwNpageVerticalOffset, XkwCPageVerticalOffset, XtRInt, sizeof (int),
     TheOffset (voffset), XtRImmediate, (XtPointer) 10},
    {XkwNpageHorizontalSize, XkwCPageHorizontalSize, XtRInt, sizeof (int),
     TheOffset (hsize), XtRImmediate, (XtPointer) 180},
    {XkwNpageVerticalSize, XkwCPageVerticalSize, XtRInt, sizeof (int),
     TheOffset (vsize), XtRImmediate, (XtPointer) 180},
    {XkwNautoIncrement, XkwCAutoIncrement, XtRBool, sizeof (Bool),
     TheOffset (autoIncrement), XtRImmediate, (XtPointer) FALSE},
    {XtNcallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     TheOffset (callback), XtRCallback, (caddr_t) NULL},
};

#undef TheOffset

PostscriptClassRec postscriptClassRec = 
{
  {        /* CoreClassPart */
    (WidgetClass)&topLevelShellClassRec,  /* superclass */
    "Postscript",                     /* class_name */
    sizeof(PostscriptRec),            /* widget_size */
    NULL,                          /* class_initialise */
    NULL,                          /* class_part_initialise */
    FALSE,                         /* class_init */
    (XtInitProc)_Postscript_Initialise,        /* initialise */
    NULL,                          /* initialise_hook */
    XtInheritRealize,              /* realize */
    NULL,                          /* actions */
    0,                             /* num_actions */
    PostscriptResources,              /* resources */
    XtNumber(PostscriptResources),    /* num_resources */
    NULLQUARK,                     /* xrm_class */
    TRUE,                          /* compress_motion */
    TRUE,                          /* compress_exposure */
    TRUE,                          /* compress_enterleave */
    TRUE,                          /* visible_interest */
    _Postscript_Destroy,           /* destroy */
    NULL,                          /* resize */
    NULL,                          /* expose */
    NULL,                          /* set_values */
    NULL,                          /* set_values_hook */
    XtInheritSetValuesAlmost,      /* set_values_almost */
    NULL,                          /* get_values_hook */
    NULL,                          /* accept_focus */
    XtVersion,                     /* version */
    NULL,                          /* callback_private */
    NULL,                          /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* PostscriptClassPart */
    0 /* empty */
  }
};

WidgetClass postscriptWidgetClass = (WidgetClass) &postscriptClassRec;


/*  Orientation choices  */
#define ORIENT_PORTRAIT  0
#define ORIENT_LANDSCAPE 1
#define NUM_ORIENTATIONS 2

static char *orientations[NUM_ORIENTATIONS] =
{"portrait", "landscape"};

/*  Tray choices   */
#define TRAY_DEFAULT      0
#define TRAY_PAPER        1
#define TRAY_TRANSPARENCY 2
#define NUM_TRAYS         3

static char *tray_choices[NUM_TRAYS] =
{
    "default", "paper", "transparency",
};


/*----------------------------------------------------------------------*/
/* check that a widget is actually a PostscriptWidget                      */
/*----------------------------------------------------------------------*/

static int check_type (Widget w, char *function_name)
{
    int fl;
    /*static char func_name[] = "Postscript.check_type";*/

    fl = XtIsSubclass (w, postscriptWidgetClass);
    if (!fl)
    fprintf (stderr,
		    "ERROR: Widget passed to %s is not a PostscriptWidget\n",
		    function_name);
    return fl;
}

/*----------------------------------------------------------------------*/
/*  This routine will print the image to a channel. It first prints the */
/*  PostScript header.                                                  */
/*  The channel to write to must be given by  channel  .                */
/*  If the value of  colour  is TRUE, colour PostScript is written.     */
/*  If the value of  eps  is TRUE, encapsulated PostScript is written.  */
/*  The routine returns TRUE on succes, else it returns FALSE.          */
/*----------------------------------------------------------------------*/

static flag write_ps (PostscriptWidget w, Channel channel,flag colour,flag eps)
{
    PostScriptPage pspage;
    int pix_width, pix_height;
    double hsize, vsize, hoffset, voffset, factor;
    static char function_name[] = "write_ps";

    check_type ( (Widget) w, function_name );
    FLAG_VERIFY (colour);
    if (w->postscript.pixcanvas == NULL)
    {
	fprintf (stderr,
		 "Postscript Widget: NULL Canvas passed - No PS written\n");
	return (FALSE);
    }
    kwin_get_size (w->postscript.pixcanvas, &pix_width, &pix_height);
    hsize = w->postscript.hsize / 10.0;
    vsize = w->postscript.vsize / 10.0;
    hoffset = w->postscript.hoffset / 10.0;
    voffset = w->postscript.voffset / 10.0;
    if (pix_width > pix_height)
    {
	factor = (double) pix_height / (double) pix_width;
	if (w->postscript.portrait)
	{
	    voffset += vsize * (1.0 - factor) * 0.5;
	    vsize *= factor;
	}
	else
	{
	    hoffset += hsize * (1.0 - factor) * 0.5;
	    hsize *= factor;
	}
    }
    else if (pix_height > pix_width)
    {
	factor = (double) pix_width / (double) pix_height;
	if (w->postscript.portrait)
	{
	    hoffset += hsize * (1.0 - factor) * 0.5;
	    hsize *= factor;
	}
	else
	{
	    voffset += vsize * (1.0 - factor) * 0.5;
	    vsize *= factor;
	}
    }
    pspage = psw_va_create (channel, hoffset, voffset, hsize, vsize,
			    w->postscript.portrait, eps,
			    PSW_ATT_THIN_LINEWIDTH_MM, w->postscript.linewidth,
			    PSW_ATT_TRAY, w->postscript.tray,
			    PSW_ATT_END);
    if (pspage == NULL ) return (FALSE);
    if ( kwin_write_ps (w->postscript.pixcanvas, pspage) )
    {
	return psw_close (pspage, FALSE, FALSE);
    }
    psw_close (pspage, FALSE, FALSE);
    return (FALSE);
}   /*  End Function write_ps  */

/*----------------------------------------------------------------------*/
/*  This routine will print the image to the queue named by the queue   */
/*  dialog.                                                             */
/*  If the value of client_data is TRUE, colour PostScript is written.  */
/*  The case where client_data is FALSE is not implemented yet.         */
/*  It is setup as a callback for the "queue" buttons.                  */
/*  The routine returns nothing.                                        */
/*----------------------------------------------------------------------*/

static void send_to_queue (Widget w, XtPointer client_data,XtPointer call_data)
{
    Channel read_ch, write_ch;
    int pid;
    int statusp, write_fd;
    flag colour = (uaddr) client_data;
    PostscriptWidget pw = (PostscriptWidget)XtParent(XtParent(XtParent(w)));
    char *queue;
    char txt[STRING_LENGTH];
    char *argv[3];
    static char function_name[] = "send_to_queue";

    XtCallCallbacks((Widget)pw,XtNcallback,NULL);
    FLAG_VERIFY (colour);
    XtVaGetValues (pw->postscript.queue_dialog,
		   XtNvalue, &queue,
		   NULL);
    if ( (queue == NULL) || (*queue == '\0') )
    {
	fprintf (stderr, "No printer queue specified\n");
    }
    if ( !ch_create_pipe (&read_ch, &write_ch) )
    {
	fprintf (stderr, "Error opening pipe\t%s\n",ERRSTRING);
	return;
    }
    sprintf (txt, "-P%s", queue);
    write_fd = ch_get_descriptor (write_ch);
    argv[0] = "lpr";
    argv[1] = txt;
    argv[2] = NULL;
    if ( ( pid = r_fork_and_execvp (ch_get_descriptor (read_ch), 1, 2,
				    1, &write_fd, argv[0], argv) ) == -1 )
    {
	fprintf (stderr, "%s: error forking\t%s\n",
		 function_name, ERRSTRING);
	ch_close (read_ch);
	ch_close (write_ch);
	return;
    }
    ch_close (read_ch);
    if ( write_ps (pw, write_ch, colour, FALSE) )
    {
	if ( ch_close (write_ch) )
	{
	    /*  Success: wait for child  */
	    if (waitpid (pid, &statusp, 0) < 0)
	    {
		fprintf (stderr, "Error waiting\t%s\n",ERRSTRING);
		kill (pid, SIGKILL);
		return;
	    }
	    if (statusp == 0)
	    {
		fprintf (stderr, "lpr successfully finished.\n");
	    }
	    else
	    {
		fprintf (stderr, "lpr returned with status: %d\n", statusp);
	    }
	    return;
	}
    }
    ch_close (write_ch);
    kill (pid, SIGKILL);
}

/*----------------------------------------------------------------------*/
/*  This routine will print the image to a file.                        */
/*  If the value of client_data is TRUE, encapsulated PS is written.    */
/*  It is setup as a callback for the "write" buttons.                  */
/*----------------------------------------------------------------------*/

static void write_to_file (Widget w, XtPointer client_data,XtPointer call_data)
{
    Channel channel;
    flag check = TRUE;
    flag eps = (uaddr) client_data;
    unsigned int sequence_number = 0;
    struct stat statbuf;
    char filename[STRING_LENGTH];
    char basefile[STRING_LENGTH];
    String dialog_name;
    static char function_name[] = "write_to_file";
    PostscriptWidget pw = (PostscriptWidget) XtParent(XtParent(XtParent(w)));

    XtCallCallbacks((Widget)pw,XtNcallback,NULL);
    FLAG_VERIFY (eps);
    XtVaGetValues (pw->postscript.name_dialog,
		   XtNvalue, &dialog_name,
		   NULL);
    if ( (dialog_name == NULL) || (*dialog_name == '\n') )
    {
	strcpy (basefile, "fred");
    }
    else strcpy (basefile, dialog_name);
    while (check)
    {
	if ( pw->postscript.autoIncrement && (sequence_number > 0) )
	{
	    sprintf (filename, "%s.%u.%s",
			    basefile, sequence_number, eps ? "eps" : "ps");
	}
	else
	{
	    sprintf (filename, "%s.%s", basefile, eps ? "eps" : "ps");
	}
	if (pw->postscript.autoIncrement)
	{
	    ++sequence_number;
	    if (stat (filename, &statbuf) != 0) check = FALSE;
	}
	else check = FALSE;
    }
    if ( ( channel = ch_open_file (filename, "w") ) == NULL )  {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 filename, ERRSTRING);
	return;
    }
    if(write_ps(pw,channel,TRUE,eps)) {
	if ( ch_close (channel) )  {
	    fprintf (stderr, "Wrote PostScript file: \"%s\"\n",filename);
	    return;
	}
    }
    fprintf (stderr, "Error writing PostScript\n");
    ch_close (channel);
    unlink (filename);
}

static void orientation_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
{
    unsigned int orientation = *(int *) call_data;
    PostscriptWidget top = (PostscriptWidget) client_data;

    switch (orientation)
    {
      case ORIENT_PORTRAIT:
	top->postscript.portrait = TRUE;
	break;
      case ORIENT_LANDSCAPE:
	top->postscript.portrait = FALSE;
	break;
      default:
	fprintf (stderr, "Illegal orientation: %u\n", orientation);
	break;
    }
}   /*  End Function orientation_cbk  */

static Widget create_value (Widget parent, char *name, int min,
			    int *valuepointer)
{
    Widget w;

    w = XtVaCreateManagedWidget (name, valueWidgetClass, parent,
				 XkwNmodifier, 1,
				 XkwNminimum, min,
				 XkwNmaximum, 1000,
				 XtNvalue, *valuepointer,
				 XtNborderWidth, 0,
				 XtNlabel, name,
				 XkwNvaluePtr, valuepointer,
				 NULL);
    return w;
}   /*  End Function create_value  */

static void create_text_btn (Widget *btn, char *name,char *cbk_val,
			     XtCallbackProc proc, Widget parent)
{
    *btn = XtVaCreateManagedWidget ("button", commandWidgetClass, parent,
				    XtNlabel, name,
				    XtNhorizDistance, 5,
				    XtNvertDistance, 5,
				    NULL);
    XtAddCallback (*btn, XtNcallback, proc, cbk_val);
}  /*  End Function create_text_btn  */

static void _Postscript_Initialise (Widget Request, Widget New)
{
    /*PostscriptWidget request = (PostscriptWidget) Request;*/
    PostscriptWidget new = (PostscriptWidget) New;
    Widget box1, box2;
    Widget psbtn, epsbtn, qcolourbtn, w, orient_menu, closebtn, dlg, tray_menu;
    CONST char *queue;
    static unsigned int instance_number = 0;

    new->postscript.instance_number = instance_number++;
    new->postscript.tray = PSW_TRAY_DEFAULT;
    box1 = XtVaCreateManagedWidget ("box", boxWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    box2 = XtVaCreateManagedWidget ("box2", boxWidgetClass, (Widget) box1,
				    XtNborderWidth, 0,
				    XtNorientation, XtorientHorizontal,
				    NULL);
    closebtn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass,box2,
					XtNlabel, "close",
					XtNhorizDistance, 5,
					XtNvertDistance, 5,
					NULL);
    XtAddCallback (closebtn, XtNcallback, xtmisc_popdown_cbk, New);
    create_text_btn (&psbtn, "save .ps", (char *) FALSE, write_to_file,box2);
    create_text_btn (&epsbtn, "save .eps", (char *) TRUE, write_to_file,box2);
    create_text_btn (&qcolourbtn, "print", (char *) TRUE, send_to_queue,box2);
    tray_menu = XtVaCreateManagedWidget ("menuButton",
					 exclusiveMenuWidgetClass, box1,
					 XtNmenuName, "trayMenu",
					 XkwNchoiceName, "Tray",
					 XkwNnumItems, NUM_TRAYS,
					 XkwNitemStrings, tray_choices,
					 NULL);
    XtAddCallback (tray_menu, XkwNselectCallback, tray_cbk, New);
    orient_menu = XtVaCreateManagedWidget ("menuButton",
					   exclusiveMenuWidgetClass, box1,
					   XtNmenuName, "orientationMenu",
					   XkwNchoiceName, "Orientation",
					   XkwNnumItems, NUM_ORIENTATIONS,
					   XkwNitemStrings, orientations,
					   NULL);
    XtAddCallback (orient_menu, XkwNselectCallback, orientation_cbk, New);
    w = XtVaCreateManagedWidget ("incToggle", ktoggleWidgetClass, box1,
				 XtNlabel, "Auto Increment",
				 XtNstate, new->postscript.autoIncrement,
				 XkwNcrosses, FALSE,
				 NULL);
    XtAddCallback (w, XtNcallback, auto_increment_cbk, New);
    w = XtVaCreateManagedWidget ("lockToggle", ktoggleWidgetClass, box1,
				 XtNlabel, "Lock Filename",
				 XkwNcrosses, FALSE,
				 NULL);
    new->postscript.lock_tgl = w;
    create_value (box1, "hoffset (mm)", -100, &(new->postscript.hoffset));
    create_value (box1, "voffset (mm)", -100, &(new->postscript.voffset));
    create_value (box1, "hsize (mm)", 0, &(new->postscript.hsize));
    create_value (box1, "vsize (mm)", 0, &(new->postscript.vsize));

    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, box1,
				 /*XtNfromVert, upper_widget,
				 XtNfromHoriz, w,*/
				 XtNlabel, "Thin Line Width (mm)",
				 XtNwidth, 350,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 10,
				 XkwNscaledValuePtr,&new->postscript.linewidth,
				 XkwNscaledFormat, "%.2f",
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, TRUE);

    dlg = XtVaCreateManagedWidget ("nameDialog", dialogWidgetClass, box1,
				   XtNlabel, "Output file:    ",
				   XtNvalue, "fred",
				   NULL);
    new->postscript.name_dialog = dlg;
    if ( ( queue = r_getenv ("PRINTER") ) == NULL ) queue = "lp";
    dlg = XtVaCreateManagedWidget ("nameDialog", dialogWidgetClass, box1,
				   XtNlabel, "Printer queue:    ",
				   XtNvalue, queue,
				   NULL);
    new->postscript.queue_dialog = dlg;
    new->postscript.ds_event_callback =
	ds_event_register_func (_Postscript_ds_event_func, New);
}  /*  End Function Initialise  */

static void _Postscript_Destroy (Widget W)
{
    PostscriptWidget w = (PostscriptWidget) W;

    c_unregister_callback (w->postscript.ds_event_callback);
}   /*  End Function Destroy  */

static void auto_increment_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
{
    flag bool = (uaddr) call_data;
    PostscriptWidget top = (PostscriptWidget) client_data;

    top->postscript.autoIncrement = bool;
}   /*  End Function auto_increment_cbk  */

static void tray_cbk (Widget w, XtPointer client_data, XtPointer call_data)
{
    unsigned int tray = *(int *) call_data;
    PostscriptWidget top = (PostscriptWidget) client_data;

    switch (tray)
    {
      case TRAY_DEFAULT:
	top->postscript.tray = PSW_TRAY_DEFAULT;
	break;
      case TRAY_PAPER:
	top->postscript.tray = PSW_TRAY_PAPER;
	break;
      case TRAY_TRANSPARENCY:
	top->postscript.tray = PSW_TRAY_TRANSPARENCY;
	break;
      default:
	fprintf (stderr, "Illegal tray: %u\n", tray);
	break;
    }
}   /*  End Function tray_cbk  */

static flag _Postscript_ds_event_func (void *object, multi_array *multi_desc,
				       CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    Channel channel;
    unsigned int instance_number;
    PostscriptWidget top = (PostscriptWidget) object;
    packet_desc *pack_desc;
    char *packet, *filename;
    static char function_name[] = "Postscript::ds_event_func";

    if (strcmp (name, "PostscriptWidget::WritePS") != 0) return (FALSE);
    if (multi_desc->num_arrays != 1) return (FALSE);
    pack_desc = multi_desc->headers[0];
    packet = multi_desc->data[0];
    if (pack_desc->num_elements != 2) return (FALSE);
    if (pack_desc->element_types[0] != K_UINT) return (FALSE);
    if (pack_desc->element_types[1] != K_VSTRING) return (FALSE);
    if (strcmp (pack_desc->element_desc[0], "instance_number") != 0)
	return (FALSE);
    if (strcmp (pack_desc->element_desc[1], "filename") != 0) return (FALSE);
    m_copy (&instance_number, packet, sizeof instance_number);
    if (instance_number != top->postscript.instance_number) return (FALSE);
    m_copy (&filename, packet + ds_get_element_offset (pack_desc, 1),
	    sizeof filename);
    if ( ( channel = ch_open_file (filename, "w") ) == NULL )
    {
	fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		 function_name, filename, ERRSTRING);
	return (TRUE);
    }
    XtCallCallbacks ( (Widget) top, XtNcallback, NULL );
    if ( write_ps (top, channel, TRUE, FALSE) && ch_close (channel) )
	return (TRUE);
    fprintf (stderr, "Error writing PostScript\n");
    ch_close (channel);
    unlink (filename);
    return (TRUE);
}   /*  End Function _Postscript_ds_event_func  */


/*  Public functions follow  */

void XkwPostscriptRegisterImageAndName (Widget w, KPixCanvas image, char *name)
{
    PostscriptWidget pw = (PostscriptWidget) w;
    Bool locked;
    static char fn_name[] = "XkwPostscriptRegisterImageAndName";

    check_type (w, fn_name);
    pw->postscript.pixcanvas = image;
    XtVaGetValues (pw->postscript.lock_tgl, XtNstate, &locked, NULL);
    if (locked) return;
    if ( (name == NULL) || (*name == '\0') ) return;
    XtVaSetValues (pw->postscript.name_dialog,
		   XtNvalue, name,
		   NULL);
}   /*  End Function XkwPostscriptRegisterImageAndName  */
