/*LINTLIBRARY*/
/*  ProfileEdit.c

    This code provides a profile editing widget for Xt.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains all routines needed for a profile editing widget for Xt.


    Written by      Richard Gooch   28-FEB-1997

    Updated by      Richard Gooch   3-MAR-1997

    Updated by      Richard Gooch   22-FEB-1998: Switched to
  <XkwDressingControlAddCanvas>.

    Updated by      Richard Gooch   16-APR-1999: Make profile colour adapt to
  background colour.

    Updated by      Richard Gooch   17-APR-1999: Do not draw marker when
  generating PostScript output.

    Updated by      Richard Gooch   18-APR-1999: Do not clear canvas and avoid
  redrawing everything (if possible) when moving vertex.

    Updated by      Richard Gooch   21-APR-1999: Support saving and loading.
  Added "Show Markers" toggle.

    Updated by      Richard Gooch   17-MAY-1999: Pass array to callbacks.

    Updated by      Richard Gooch   24-MAY-1999: Fixed bug in <position_func>
  where middle mouse release event was discarded.

    Last updated by Richard Gooch   9-JUN-2000: Removed unused variables.


*/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#ifndef X11
#  define X11
#endif
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Command.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <k_event_codes.h>
#include <karma_foreign.h>
#include <karma_xtmisc.h>
#include <karma_iarray.h>
#include <karma_dsxfr.h>
#include <karma_dsrw.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/ProfileEditP.h>
#include <Xkw/Canvas.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/DressingControl.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/Filepopup.h>


#define CIRCLE_SIZE    5.0
#define MARKER_BUFFER  20

STATIC_FUNCTION (void ProfileEdit__Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (Boolean ProfileEdit__SetValues,
		 (Widget current, Widget request, Widget new) );
STATIC_FUNCTION (void realise_cbk, (Widget w, XtPointer client_data,
				    XtPointer call_data) );
STATIC_FUNCTION (void refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (unsigned long get_fg_pixel, (KPixCanvas canvas) );
STATIC_FUNCTION (flag position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin) );
STATIC_FUNCTION (void move_vertex,
		 (ProfileEditWidget top, double x, double y) );
STATIC_FUNCTION (void draw_moving_vertex,
		 (ProfileEditWidget top, flag draw_side_markers,
		  unsigned long pixel_value) );
STATIC_FUNCTION (void iarr_destroy_callback,
		 (iarray arr, ProfileEditWidget top) );
STATIC_FUNCTION (void apply_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (flag ProfileEdit__add_vertex,
		 (ProfileEditWidget top, double x, double y) );
STATIC_FUNCTION (vertex ProfileEdit__find_vertex,
		 (ProfileEditWidget top, double x) );
STATIC_FUNCTION (void ProfileEdit__notify, (ProfileEditWidget top) );
STATIC_FUNCTION (void ProfileEdit__save_dialog_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag ProfileEdit__accept_file, (KFileInfo finfo) );
STATIC_FUNCTION (void ProfileEdit__filesel_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void show_markers_toggle_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffsetOf(ProfileEditRec, profileEdit.field)

static XtResource resources[] =
{
    {XkwNiarray, XkwCIarray, XtRPointer, sizeof (XtPointer),
     offset (array), XtRImmediate, NULL},
    {XkwNworldCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (worldCanvas), XtRImmediate, NULL},
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, False},
    {XtNcallback, XtCCallback, XtRCallback, sizeof (XtPointer),
     offset (callback), XtRCallback, (XtPointer) NULL},
};
#undef offset

#define MODE_ADD    0
#define MODE_MOVE   1
#define MODE_DELETE 2
#define NUM_MODE_CHOICES 3

static char *mode_choices[NUM_MODE_CHOICES] =
{
    "add", "move", "delete",
};


ProfileEditClassRec profileEditClassRec =
{
    {
	/* core fields  */
	(WidgetClass) &formClassRec,    /*  superclass             */
	"ProfileEdit",                  /*  class_name             */
	sizeof (ProfileEditRec),        /*  widget_size            */
	NULL,                           /*  class_initialise       */
	NULL,                           /*  class_part_initialise  */
	FALSE,                          /*  class_inited           */
	(XtInitProc) ProfileEdit__Initialise, /*  initialise             */
	NULL,                           /*  initialise_hook        */
	XtInheritRealize,               /*  realise                */
	NULL,                           /*  actions                */
	0,                              /*  num_actions            */
	resources,                      /*  resources              */
	XtNumber (resources),           /*  num_resources          */
	NULLQUARK,                      /*  xrm_class              */
	TRUE,                           /*  compress_motion        */
	TRUE,                           /*  compress_exposure      */
	TRUE,                           /*  compress_enterleave    */
	TRUE,                           /*  visible_interest       */
	NULL,                           /*  destroy                */
	XtInheritResize,                /*  resize                 */
	NULL,                           /*  expose                 */
	(XtSetValuesFunc) ProfileEdit__SetValues, /*  set_values */
	NULL,                           /*  set_values_hook        */
	XtInheritSetValuesAlmost,       /*  set_values_almost      */
	NULL,                           /*  get_values_hook        */
	NULL,                           /*  accept_focus           */
	XtVersion,                      /*  version                */
	NULL,                           /*  callback_private       */
	NULL,                           /*  tm_table               */
	XtInheritQueryGeometry,         /*  query_geometry         */
	XtInheritDisplayAccelerator,    /*  display_accelerator    */
	NULL                            /*  extension              */
    },
    {     /* CompositeClassPart */
	XtInheritGeometryManager        /* geometry_manager        */,
	XtInheritChangeManaged          /* change_managed          */,
	XtInheritInsertChild            /* insert_child            */,
	XtInheritDeleteChild            /* delete_child            */,
	NULL                            /* extension               */
    },
    { /* Constraint */
	NULL                            /* subresourses       */,
	0                               /* subresource_count  */,
	sizeof (ProfileEditConstraintsRec) /* constraint_size    */,
	NULL                            /* initialise         */,
	NULL                            /* destroy            */,
	NULL                            /* set_values         */,
	NULL                            /* extension          */ 
    },
    { /* Form */
	XtInheritLayout                 /* layout             */
    },
    {
	/*  profileEdit fields */
	0                               /*  ignore                 */
    }
};

WidgetClass profileEditWidgetClass = (WidgetClass) &profileEditClassRec;

static void ProfileEdit__Initialise (Widget Request, Widget New)
{
    /*ProfileEditWidget request = (ProfileEditWidget) Request;*/
    ProfileEditWidget new = (ProfileEditWidget) New;
    Widget w, apply_btn, canvas, save_dialog, filepopup;
    /*static char function_name[] = "ProfileEditWidget::Initialise";*/

    new->profileEdit.pixcanvas = NULL;
    new->profileEdit.iarr_destroy_func = NULL;
    new->profileEdit.mode = 0;
    new->profileEdit.left_vertex = NULL;
    new->profileEdit.right_vertex = NULL;
    new->profileEdit.moving_vertex = NULL;
    new->profileEdit.update_on_drag = FALSE;
    new->profileEdit.update_on_drop = FALSE;
    new->profileEdit.show_markers = TRUE;
    /*  Create popups first  */
    w = XtVaCreatePopupShell ("dressingControlPopup",
			      dressingControlWidgetClass, New,
			      NULL);
    new->profileEdit.dressing_control_popup = w;
    w = XtNameToWidget (w, "*displayAxisLabelsToggle");
    XtVaSetValues (w, XtNstate, TRUE, NULL);
    save_dialog = XtVaCreatePopupShell ("save_dialog", dialogpopupWidgetClass,
					New,
					XtNlabel, "Save Filename:",
					XkwNdefaultName, "profile",
					NULL);
    XtAddCallback (save_dialog, XtNcallback, ProfileEdit__save_dialog_cbk,New);
    w = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass, New,
			      XtNtitle, "Profile File Selector",
			      XtNx, 0,
			      XtNy, 0,
			      XkwNfilenameTester,
			      ProfileEdit__accept_file,
			      XkwNautoPopdown, True,
			      NULL);
    filepopup = w;
    XtAddCallback (w, XkwNfileSelectCallback, ProfileEdit__filesel_cbk, New);
    /*  Now create the rest  */
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, New,
				 XtNlabel, "Apply",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNheight, 20,
				 NULL);
    apply_btn = w;
    XtAddCallback (w, XtNcallback, apply_cbk, New);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, New,
				 XtNlabel, "Update on Drag",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_toggle_cbk,
		   &new->profileEdit.update_on_drag);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, New,
				 XtNlabel, "Update on Drop",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_toggle_cbk,
		   &new->profileEdit.update_on_drop);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, New,
				 XtNlabel, "Axis Labels",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popup_cbk,
		   new->profileEdit.dressing_control_popup);
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, New,
				 XtNmenuName, "modeMenu",
				 XkwNchoiceName, "Mode",
				 XtNfromVert, apply_btn,
				 XkwNnumItems, NUM_MODE_CHOICES,
				 XkwNitemStrings, mode_choices,
				 XkwNvaluePtr, &new->profileEdit.mode,
				 NULL);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, New,
				 XtNlabel, "Save",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromVert, apply_btn,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popup_cbk, save_dialog);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, New,
				 XtNlabel, "Load",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromVert, apply_btn,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popup_cbk, filepopup);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, New,
				 XtNlabel, "Show Markers",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, apply_btn,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 XtNstate, True,
				 NULL);
    XtAddCallback (w, XtNcallback, show_markers_toggle_cbk, New);
    canvas = XtVaCreateManagedWidget ("canvas", canvasWidgetClass, New,
				      XtNfromVert, w,
				      XtNwidth, 512,
				      XtNheight, 512,
				      XtNmappedWhenManaged, True,
				      XkwNsilenceUnconsumed, True,
				      NULL);
    XtAddCallback (canvas, XkwNrealiseCallback, realise_cbk, (XtPointer) new);
}   /*  End Function Initialise  */

static Boolean ProfileEdit__SetValues (Widget Current, Widget Request,
					 Widget New)
{
    iarray array;
    ProfileEditWidget current = (ProfileEditWidget) Current;
    ProfileEditWidget new = (ProfileEditWidget) New;
    double first_coord, last_coord;
    static char function_name[] = "ProfileEditWidget::SetValues";

    array = new->profileEdit.array;
    new->profileEdit.worldCanvas = current->profileEdit.worldCanvas;
    if (array != current->profileEdit.array)
    {
	/*  Change in array data  */
	if (new->profileEdit.iarr_destroy_func != NULL)
	{
	    c_unregister_callback (new->profileEdit.iarr_destroy_func);
	}
	new->profileEdit.iarr_destroy_func = NULL;
	/*  New array data  */
	if (array != NULL)
	{
	    if (iarray_num_dim (array) != 1)
	    {
		fprintf ( stderr, "Illegal number of array dimensions: %u\n",
			  iarray_num_dim (array) );
		a_prog_bug (function_name);
	    }
	    if (iarray_type (array) != K_FLOAT)
	    {
		fprintf ( stderr, "Illegal type of array: %u\n",
			  iarray_type (array) );
		a_prog_bug (function_name);
	    }
	    new->profileEdit.iarr_destroy_func =
	    iarray_register_destroy_func (array,
					  ( flag (*) () )iarr_destroy_callback,
					  new);
	    if (new->profileEdit.worldCanvas != NULL)
	    {
		canvas_specify (new->profileEdit.worldCanvas,
				iarray_dim_name (array, 0),
				iarray_value_name (array),
				0, NULL, NULL);
		iarray_get_world_coords (array, 0, &first_coord, &last_coord);
		canvas_set_attributes (new->profileEdit.worldCanvas,
				       CANVAS_ATT_LEFT_X, first_coord,
				       CANVAS_ATT_RIGHT_X, last_coord,
				       CANVAS_ATT_END);
	    }
	}
    }
    return False;
}   /*  End Function SetValues  */

static void realise_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the canvas realise callback.
*/
{
    KPixCanvas pixcanvas = (KPixCanvas) call_data;
    iarray array;
    ProfileEditWidget top = (ProfileEditWidget) client_data;
    double first_coord, last_coord;
    struct win_scale_type win_scale;
    /*static char function_name[] = "ProfileEditWidget::realise_cbk";*/

    top->profileEdit.pixcanvas = pixcanvas;
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    if ( ( top->profileEdit.worldCanvas =
	   canvas_create (pixcanvas, NULL, &win_scale) ) == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_set_dressing (top->profileEdit.worldCanvas,
			 KCD_ATT_DISPLAY, TRUE,
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_TOP_TICKS, TRUE,
			 KCD_ATT_BOTTOM_TICKS, TRUE,
			 KCD_ATT_LEFT_TICKS, TRUE,
			 KCD_ATT_RIGHT_TICKS, TRUE,
			 KCD_ATT_END);
    canvas_register_refresh_func (top->profileEdit.worldCanvas,
				  refresh_func, top);
    canvas_register_position_event_func (top->profileEdit.worldCanvas,
					 position_func, top);
    XkwDressingControlAddCanvas (top->profileEdit.dressing_control_popup,
				 top->profileEdit.worldCanvas);
    array = top->profileEdit.array;
    if (array != NULL)
    {
	canvas_specify (top->profileEdit.worldCanvas,
			iarray_dim_name (array, 0), iarray_value_name (array),
			0, NULL, NULL);
	iarray_get_world_coords (array, 0, &first_coord, &last_coord);
	canvas_set_attributes (top->profileEdit.worldCanvas,
			       CANVAS_ATT_LEFT_X, first_coord,
			       CANVAS_ATT_RIGHT_X, last_coord,
			       CANVAS_ATT_END);
    }
}   /*  End Function realise_cbk   */

static void refresh_func (KWorldCanvas canvas, int width, int height,
			  struct win_scale_type *win_scale,
			  Kcolourmap cmap, flag cmap_resize, void **info,
			  PostScriptPage pspage,
			  unsigned int num_areas, KPixCanvasRefreshArea *areas,
			  flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    vertex curr;
    unsigned long pixel_value;
    double x, y, prev_x, prev_y;
    ProfileEditWidget top = (ProfileEditWidget) *info;

    if (top->profileEdit.left_vertex == NULL) return;
    pixel_value = get_fg_pixel (top->profileEdit.pixcanvas);
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &prev_x,
			   CANVAS_ATT_END);
    prev_y = top->profileEdit.left_vertex->y;
    canvas_coords_transform (canvas, 1, &prev_x, FALSE, &prev_y, TRUE);
    prev_y = top->profileEdit.left_vertex->y;
    for (curr = top->profileEdit.left_vertex; curr != NULL; curr = curr->right)
    {
	/*  Draw from last to this one  */
	canvas_draw_line_p (canvas, prev_x, prev_y, curr->x, curr->y,
			    pixel_value);
	prev_x = curr->x;
	prev_y = curr->y;
	if (!top->profileEdit.show_markers || (pspage != NULL) ) continue;
	/*  Draw the marker  */
	x = curr->x;
	y = curr->y;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE,
					   1, &x, &y, &x, &y);
	kwin_draw_ellipse (top->profileEdit.pixcanvas, x, y,
			   CIRCLE_SIZE, CIRCLE_SIZE, pixel_value);
    }
    /*  Draw from the right vertex to the end  */
    canvas_get_attributes (canvas,
			   CANVAS_ATT_RIGHT_X, &x,
			   CANVAS_ATT_END);
    y = top->profileEdit.right_vertex->y;
    canvas_coords_transform (canvas, 1, &x, FALSE, &y, TRUE);
    y = top->profileEdit.right_vertex->y;
    canvas_draw_line_p (canvas, prev_x, prev_y, x, y, pixel_value);
}   /*  End Function refresh_func  */

static unsigned long get_fg_pixel (KPixCanvas canvas)
/*  [SUMMARY] Get the foreground pixel value.
    <canvas> The pixel canvas.
    [RETURNS] The pixel value.
*/
{
    unsigned short red, green, blue;
    unsigned long pixel_value;
    double r, g, b, intensity;
    char *colourname;

    kwin_get_attributes (canvas,
			 KWIN_ATT_BACKGROUND_PIXEL, &pixel_value,
			 KWIN_ATT_END);
    kwin_get_pixel_RGB_values (canvas, &pixel_value, &red, &green, &blue, 1);
    r = red;
    g = green;
    b = blue;
    intensity = sqrt (r * r + g * g + b * b) / 65535.0;
    colourname = (intensity >= 0.5) ? "black" : "green";
    if ( kwin_get_colour (canvas, colourname, &pixel_value, NULL, NULL, NULL) )
	return (pixel_value);
    if ( kwin_get_colour (canvas, "white", &pixel_value, NULL, NULL, NULL) )
	return (pixel_value);
    return (0);
}   /*  End Function get_fg_pixel  */

static flag position_func (KWorldCanvas canvas, double x, double y,
			   unsigned int event_code, void *e_info,
			   void **f_info, double x_lin, double y_lin)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    flag update = FALSE;
    flag refresh = TRUE;
    vertex closest, left, right;
    double left_x, right_x, bottom_y, top_y;
    ProfileEditWidget top = (ProfileEditWidget) *f_info;
    Display *display;

    switch (event_code)
    {
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    display = XtDisplay (top);
    canvas_get_attributes (top->profileEdit.worldCanvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    canvas_coords_transform (top->profileEdit.worldCanvas, 1,
			     &left_x, FALSE, &bottom_y, FALSE);
    canvas_coords_transform (top->profileEdit.worldCanvas, 1,
			     &right_x, FALSE, &top_y, FALSE);
    if (left_x < right_x)
    {
	if ( (x < left_x) || (x > right_x) )
	{
	    XBell (display, 100);
	    return (TRUE);
	}
    }
    else
    {
	if ( (x < right_x) || (x > left_x) )
	{
	    XBell (display, 100);
	    return (TRUE);
	}
    }
    switch (top->profileEdit.mode)
    {
      case MODE_ADD:
	top->profileEdit.moving_vertex = NULL;
	if (event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK) return (TRUE);
	if ( !ProfileEdit__add_vertex (top, x, y) ) return (TRUE);
	update = TRUE;
	break;
      case MODE_MOVE:
	switch (event_code)
	{
	  case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
	    top->profileEdit.moving_vertex = ProfileEdit__find_vertex (top, x);
	    update = TRUE;
	    break;
	  case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
	    if (top->profileEdit.update_on_drag) update = TRUE;
	    break;
	  case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	    update = TRUE;
	    break;
	}
	if (top->profileEdit.moving_vertex == NULL) return (TRUE);
	left = top->profileEdit.moving_vertex->left;
	right = top->profileEdit.moving_vertex->right;
	if (left_x < right_x)
	{
	    if ( (left != NULL) && (x <= left->x) )
	    {
		XBell (display, 100);
		return (TRUE);
	    }
	    if ( (right != NULL) && (x >= right->x) )
	    {
		XBell (display, 100);
		return (TRUE);
	    }
	}
	else
	{
	    if ( (left != NULL) && (x <= left->x) )
	    {
		XBell (display, 100);
		return (TRUE);
	    }
	    if ( (right != NULL) && (x >= right->x) )
	    {
		XBell (display, 100);
		return (TRUE);
	    }
	}
	move_vertex (top, x, y);
	refresh = FALSE;
	break;
      case MODE_DELETE:
	top->profileEdit.moving_vertex = NULL;
	if (event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK) return (TRUE);
	if ( ( closest = ProfileEdit__find_vertex (top, x) ) == NULL )
	    return (TRUE);
	if (closest->left != NULL) closest->left->right = closest->right;
	if (closest->right != NULL) closest->right->left = closest->left;
	if (top->profileEdit.left_vertex == closest)
	    top->profileEdit.left_vertex = closest->right;
	if (top->profileEdit.right_vertex == closest)
	    top->profileEdit.right_vertex = closest->left;
	m_free (closest);
	update = TRUE;
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    /*  Assume that something changed  */
    if (refresh) kwin_refresh_if_visible (top->profileEdit.pixcanvas, TRUE);
    if (update && top->profileEdit.update_on_drop)
    {
	ProfileEdit__notify (top);
    }
    return (TRUE);
}   /*  End Function position_func  */

static void move_vertex (ProfileEditWidget top, double x, double y)
/*  [SUMMARY] Move a vertex, undrawing the old and redrawing the new position.
    <top> The ProfileEdit widget.
    <x> The horizontal position to move it to.
    <y> The vertical position to move it to.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    int px, py, x_offset, y_offset, x_pixels, y_pixels;
    vertex moving_vertex = top->profileEdit.moving_vertex;

    /*  First undraw the old position  */
    kwin_get_attributes (top->profileEdit.pixcanvas,
			 KWIN_ATT_BACKGROUND_PIXEL, &pixel_value,
			 KWIN_ATT_END);
    draw_moving_vertex (top, FALSE, pixel_value);
    /*  Now draw the new position  */
    moving_vertex->x = x;
    moving_vertex->y = y;
    if ( (moving_vertex->left == NULL) || (moving_vertex->right == NULL) )
    {
	/*  Could erase axis labels tick marks, so redraw everything  */
	kwin_refresh_if_visible (top->profileEdit.pixcanvas, FALSE);
	return;
    }
    pixel_value = get_fg_pixel (top->profileEdit.pixcanvas);
    canvas_convert_from_canvas_coords (top->profileEdit.worldCanvas,
				       FALSE, FALSE, 1, &x, &y, &x, &y);
    px = x + 0.1;
    py = y + 0.1;
    canvas_get_attributes (top->profileEdit.worldCanvas,
			   CANVAS_ATT_X_OFFSET, &x_offset,
			   CANVAS_ATT_Y_OFFSET, &y_offset,
			   CANVAS_ATT_X_PIXELS, &x_pixels,
			   CANVAS_ATT_Y_PIXELS, &y_pixels,
			   CANVAS_ATT_END);
    if ( (px - MARKER_BUFFER > x_offset) &&
	 (px + MARKER_BUFFER < x_offset + x_pixels) &&
	 (py - MARKER_BUFFER > y_offset) &&
	 (py + MARKER_BUFFER < y_offset + y_pixels) )
	draw_moving_vertex (top, TRUE, pixel_value);
    else kwin_refresh_if_visible (top->profileEdit.pixcanvas, FALSE);
}   /*  End Function move_vertex  */

static void draw_moving_vertex (ProfileEditWidget top, flag draw_side_markers,
				unsigned long pixel_value)
/*  [SUMMARY] Draw the moving vertex.
    <top> The ProfileEdit widget.
    <draw_side_markers> If TRUE, the side markers are drawn.
    <pixel_value> The pixel value.
    [RETURNS] Nothing.
*/
{
    double x, y;
    vertex moving_vertex = top->profileEdit.moving_vertex;

    /*  Draw the marker  */
    canvas_convert_from_canvas_coords (top->profileEdit.worldCanvas,
				       FALSE, FALSE, 1,
				       &moving_vertex->x, &moving_vertex->y,
				       &x, &y);
    kwin_draw_ellipse (top->profileEdit.pixcanvas,
		       x, y, CIRCLE_SIZE, CIRCLE_SIZE, pixel_value);
    /*  Draw the left line  */
    if (moving_vertex->left == NULL)
    {
	canvas_get_attributes (top->profileEdit.worldCanvas,
			       CANVAS_ATT_LEFT_X, &x,
			       CANVAS_ATT_END);
	y = moving_vertex->y;
	canvas_draw_line_p (top->profileEdit.worldCanvas,
			    x, y, moving_vertex->x, moving_vertex->y,
			    pixel_value);
    }
    else
    {
	x = moving_vertex->left->x;
	y = moving_vertex->left->y;
	canvas_draw_line_p (top->profileEdit.worldCanvas,
			    x, y, moving_vertex->x, moving_vertex->y,
			    pixel_value);
	if (draw_side_markers)
	{
	    /*  Draw the left marker  */
	    canvas_convert_from_canvas_coords (top->profileEdit.worldCanvas,
					       FALSE, FALSE, 1,
					       &x, &y, &x, &y);
	    kwin_draw_ellipse (top->profileEdit.pixcanvas,
			       x, y, CIRCLE_SIZE, CIRCLE_SIZE, pixel_value);
	}
    }
    /*  Draw the right line  */
    if (moving_vertex->right == NULL)
    {
	canvas_get_attributes (top->profileEdit.worldCanvas,
			       CANVAS_ATT_RIGHT_X, &x,
			       CANVAS_ATT_END);
	y = moving_vertex->y;
	canvas_draw_line_p (top->profileEdit.worldCanvas,
			    x, y, moving_vertex->x, moving_vertex->y,
			    pixel_value);
    }
    else
    {
	x = moving_vertex->right->x;
	y = moving_vertex->right->y;
	canvas_draw_line_p (top->profileEdit.worldCanvas,
			    x, y, moving_vertex->x, moving_vertex->y,
			    pixel_value);
	if (draw_side_markers)
	{
	    /*  Draw the right marker  */
	    canvas_convert_from_canvas_coords (top->profileEdit.worldCanvas,
					       FALSE, FALSE, 1,
					       &x, &y, &x, &y);
	    kwin_draw_ellipse (top->profileEdit.pixcanvas,
			       x, y, CIRCLE_SIZE, CIRCLE_SIZE, pixel_value);
	}
    }
}   /*  End Function draw_moving_vertex  */

static void iarr_destroy_callback (iarray arr, ProfileEditWidget top)
/*  [PURPOSE] This routine will register the destruction of the Intelligent
    Array.
    <arr> The Intelligent Array.
    <top> The ProfileEdit widget.
    [RETURNS] Nothing.
*/
{
    top->profileEdit.array = NULL;
    top->profileEdit.iarr_destroy_func = NULL;
}   /*  End Function iarr_destroy_callback  */

static void apply_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the apply callback.
*/
{
    ProfileEditWidget top = (ProfileEditWidget) client_data;
    /*static char function_name[] = "ProfileEditWidget::apply_cbk";*/

    ProfileEdit__notify (top);
}   /*  End Function apply_cbk   */

static flag ProfileEdit__add_vertex (ProfileEditWidget top, double x, double y)
/*  [SUMMARY] Add a vertex to the list.
    <top> The ProfileEdit widget.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    double left_x, right_x, bottom_y, top_y;
    vertex new, curr;
    vertex left = NULL;
    static char function_name[] = "ProfileEdit::add_vertex";

    canvas_get_attributes (top->profileEdit.worldCanvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    canvas_coords_transform (top->profileEdit.worldCanvas, 1,
			     &left_x, FALSE, &bottom_y, FALSE);
    canvas_coords_transform (top->profileEdit.worldCanvas, 1,
			     &right_x, FALSE, &top_y, FALSE);
    if (left_x < right_x)
    {
	if (x < left_x) return (FALSE);
	if (x > right_x) return (FALSE);
    }
    else
    {
	if (x < right_x) return (FALSE);
	if (x > left_x) return (FALSE);
    }
    /*  Find the left vertex  */
    for (curr = top->profileEdit.left_vertex; curr != NULL; curr = curr->right)
    {
	if (x == curr->x)
	{
	    fprintf (stderr, "%s: already have a vertex at x = %e\n",
		     function_name, x);
	    return (FALSE);
	}
	if (left_x < right_x)
	{
	    /*  Co-ordinates increase from left to right  */
	    if (curr->x > x) break;
	    if ( (curr->right != NULL) && (curr->right->x < x) ) continue;
	    left = curr;
	}
	else
	{
	    /*  Co-ordinates decrease from left to right  */
	    if (curr->x < x) break;
	    if ( (curr->right != NULL) && (curr->right->x > x) ) continue;
	    left = curr;
	}
    }
    /*  Create new entry  */
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
    {
	m_error_notify (function_name, "new vertex");
	return (FALSE);
    }
    new->x = x;
    new->y = y;
    new->left = NULL;
    new->right = NULL;
    if (top->profileEdit.left_vertex == NULL)
    {
	top->profileEdit.left_vertex = new;
	top->profileEdit.right_vertex = new;
	return (TRUE);
    }
    /*  The list of vertices is not empty  */
    if (left == NULL)
    {
	/*  There is nothing to the left, but there is to the right  */
	new->right = top->profileEdit.left_vertex;
	top->profileEdit.left_vertex->left = new;
	top->profileEdit.left_vertex = new;
	return (TRUE);
    }
    /*  There is something to the left: may or may not be to the right  */
    new->left = left;
    if (top->profileEdit.right_vertex == left)
    {
	/*  Nothing to the right  */
	left->right = new;
	top->profileEdit.right_vertex = new;
	return (TRUE);
    }
    /*  Something to the right  */
    new->right = left->right;
    left->right = new;
    new->right->left = new;
    return (TRUE);
}   /*  End Function add_vertex  */

static vertex ProfileEdit__find_vertex (ProfileEditWidget top, double x)
/*  [SUMMARY] Find the closest vertex in the list.
    <top> The ProfileEdit widget.
    <x> The horizontal world co-ordinate of the event.
    [RETURNS] A pointer to the vertex on success, else NULL.
*/
{
    double left_x, right_x, bottom_y, top_y;
    vertex curr;
    vertex left = NULL;
    /*static char function_name[] = "ProfileEdit::find_vertex";*/

    if (top->profileEdit.left_vertex == NULL) return (NULL);
    /*  The list of vertices is not empty  */
    canvas_get_attributes (top->profileEdit.worldCanvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    canvas_coords_transform (top->profileEdit.worldCanvas, 1,
			     &left_x, FALSE, &bottom_y, FALSE);
    canvas_coords_transform (top->profileEdit.worldCanvas, 1,
			     &right_x, FALSE, &top_y, FALSE);
    /*  Search for closest  */
    for (curr = top->profileEdit.left_vertex; curr != NULL; curr = curr->right)
    {
	if (x == curr->x) return (curr);
	if (left_x < right_x)
	{
	    /*  Co-ordinates increase from left to right  */
	    if (curr->x > x) break;
	    if ( (curr->right != NULL) && (curr->right->x < x) ) continue;
	    left = curr;
	}
	else
	{
	    /*  Co-ordinates decrease from left to right  */
	    if (curr->x < x) break;
	    if ( (curr->right != NULL) && (curr->right->x > x) ) continue;
	    left = curr;
	}
    }
    if (left == NULL)
    {
	/*  There is nothing to the left, but there is to the right  */
	return (top->profileEdit.left_vertex);
    }
    /*  There is something to the left: may or may not be to the right  */
    if (top->profileEdit.right_vertex == left)
    {
	/*  Nothing to the right  */
	return (left);
    }
    /*  Something to the left and right: find closest  */
    if ( fabs (left->x - x) < fabs (left->right->x - x) )
    {
	/*  Closer to the left  */
	return (left);
    }
    return (left->right);
}   /*  End Function find_vertex  */

static void ProfileEdit__notify (ProfileEditWidget top)
/*  [SUMMARY] Notify clients of changes.
    <top> The ProfileEdit widget.
    [RETURNS] Nothing.
*/
{
    iarray array;
    unsigned long count;
    vertex curr, left, right;
    double x, y;
    dim_desc *dim;

    if (top->profileEdit.left_vertex == NULL) return;
    if (top->profileEdit.array == NULL) return;
    dim = iarray_get_dim_desc (top->profileEdit.array, 0);
    curr = top->profileEdit.left_vertex;
    array = top->profileEdit.array;
    for (count = 0; count < dim->length; ++count)
    {
	x = ds_get_coordinate (dim, count);
	left = NULL;
	right = NULL;
	while (curr != NULL)
	{
	    if (dim->first_coord < dim->last_coord)
	    {
		if (curr->x < x)
		{
		    if (curr->right == NULL)
		    {
			left = curr;
			break;
		    }
		    if (x < curr->right->x)
		    {
			left = curr;
			right = curr->right;
			break;
		    }
		    curr = curr->right;
		    continue;
		}
		right = curr;
		break;
	    }
	    else
	    {
		if (curr->x > x)
		{
		    if (curr->right == NULL)
		    {
			left = curr;
			break;
		    }
		    if (x > curr->right->x)
		    {
			left = curr;
			right = curr->right;
			break;
		    }
		    curr = curr->right;
		    continue;
		}
		right = curr;
		break;
	    }
	}
	if (left == NULL) y = right->y;
	else if (right == NULL) y = left->y;
	else
	{
	    y = left->y + (x - left->x) / (right->x - left->x) *
		(right->y - left->y);
	}
	F1 (array, count) = y;
    }
    XtCallCallbacks ( (Widget) top, XtNcallback, array );
}   /*  End Function notify  */

static void ProfileEdit__save_dialog_cbk (Widget w, XtPointer client_data,
					  XtPointer call_data)
{
    Channel channel;
    vertex curr;
    multi_array *multi_desc;
    packet_desc *list_desc;
    list_header *list_head;
    char *filename = (char *) call_data;
    ProfileEditWidget top = (ProfileEditWidget) client_data;
    double packet[2];
    static char function_name[] = "ProfileEdit::save_dialog_cbk";

    if (top->profileEdit.left_vertex == NULL) return;
    /*  First create the descriptors  */
    if ( ( multi_desc = ds_alloc_multi (1) ) == NULL ) return;
    if ( ( multi_desc->headers[0] = ds_alloc_packet_desc (1) ) == NULL )
    {
	ds_dealloc_multi (multi_desc);
	return;
    }
    multi_desc->headers[0]->element_types[0] = LISTP;
    if ( ( list_desc = ds_alloc_packet_desc (2) ) == NULL )
    {
	ds_dealloc_multi (multi_desc);
	return;
    }
    multi_desc->headers[0]->element_desc[0] = (char *) list_desc;
    list_desc->element_types[0] = K_DOUBLE;
    list_desc->element_types[1] = K_DOUBLE;
    if ( ( list_desc->element_desc[0] = st_dup ("x") ) == NULL )
    {
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( ( list_desc->element_desc[1] = st_dup ("y") ) == NULL )
    {
	ds_dealloc_multi (multi_desc);
	return;
    }
    /*  Now allocate the data space (no list entries)  */
    if ( ( multi_desc->data[0] = ds_alloc_data (multi_desc->headers[0], TRUE,
						TRUE) ) == NULL )
    {
	ds_dealloc_multi (multi_desc);
	return;
    }
    /*  Write the header  */
    if ( ( channel = ch_open_file (filename, "w") ) == NULL )
    {
	fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		 function_name, filename, ERRSTRING);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( !dsrw_write_multi_header (channel, multi_desc) )
    {
	ch_close (channel);
	ds_dealloc_multi (multi_desc);
	return;
    }
    list_head = *(list_header **) multi_desc->data[0];
    /*  Count the number of list entries and update the list header  */
    for (curr = top->profileEdit.left_vertex; curr != NULL; curr = curr->right)
	++list_head->length;
    list_head->sort_type = SORT_INCREASING;
    list_head->sort_elem_num = 0;
    if ( !dsrw_write_packet_desc (channel, multi_desc->headers[0]) )
    {
	ch_close (channel);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( !dsrw_write_packet (channel, multi_desc->headers[0],
			     multi_desc->data[0]) )
    {
	ch_close (channel);
	ds_dealloc_multi (multi_desc);
	return;
    }
    /*  Now write the packets piecemeal  */
    for (curr = top->profileEdit.left_vertex; curr != NULL; curr = curr->right)
    {
	packet[0] = curr->x;
	packet[1] = curr->y;
	if ( !dsrw_write_packet (channel, list_desc, (CONST char *) packet) )
	{
	    ch_close (channel);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
    }
    list_head->length = 0;
    dsrw_write_history (channel, NULL);
    ds_dealloc_multi (multi_desc);
    ch_close (channel);
}   /*  End Function save_dialog_cbk  */

static flag ProfileEdit__accept_file (KFileInfo finfo)
{
    char *ptr;

    if (finfo.type == KFILETYPE_DIRECTORY) return (TRUE);
    if (foreign_guess_format_from_filename (finfo.filename) !=
	FOREIGN_FILE_FORMAT_KARMA) return (FALSE);
    /*  It's a Karma file: fake things a bit for ".desc"/".dat" pairs  */
    if ( ( ptr = strrchr (finfo.filename, '.') ) == NULL ) return (FALSE);
    /*  Ignore the ".dat" file  */
    if (strcmp (ptr, ".dat") == 0) return (FALSE);
    /*  Fiddle if the ".desc" file  */
    if (strcmp (ptr, ".desc") == 0) strcpy (ptr, ".kf");
    return (TRUE);
}   /*  End Function accept_file  */

static void ProfileEdit__filesel_cbk (Widget w, XtPointer client_data,
				      XtPointer call_data)
{
    vertex curr, next;
    unsigned int x_index, y_index, x_offset, y_offset;
    multi_array *multi_desc;
    packet_desc *list_desc;
    list_header *header;
    list_entry *entry;
    char *filename = (char *) call_data;
    ProfileEditWidget top = (ProfileEditWidget) client_data;
    double x[2], y[2];
    static char function_name[] = "ProfileEdit::filesel_cbk";

    if ( ( multi_desc = dsxfr_get_multi (filename, FALSE, K_CH_MAP_NEVER,
					 FALSE) ) == NULL )
    {
	fprintf (stderr, "%s: error reading\n", function_name);
	return;
    }
    if (multi_desc->num_arrays != 1)
    {
	fprintf (stderr, "%s: only one general data structure allowed\n",
		 function_name);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if (multi_desc->headers[0]->num_elements != 1)
    {
	fprintf (stderr, "%s: only one element in top level packet allowed\n",
		 function_name);
	ds_dealloc_multi (multi_desc);
	return;
    }
    if (multi_desc->headers[0]->element_types[0] != LISTP)
    {
	fprintf (stderr, "%s: top level element is not a list pointer\n",
		 function_name);
	ds_dealloc_multi (multi_desc);
	return;
    }
    list_desc = (packet_desc *) multi_desc->headers[0]->element_desc[0];
    if ( ( x_index = ds_f_elem_in_packet (list_desc, "x") ) >=
	 list_desc->num_elements )
    {
	fprintf (stderr, "%s: \"x\" element not found\n", function_name);
	ds_dealloc_multi (multi_desc);
	return;
    }
    x_offset = ds_get_element_offset (list_desc, x_index);
    if ( ( y_index = ds_f_elem_in_packet (list_desc, "y") ) >=
	 list_desc->num_elements )
    {
	fprintf (stderr, "%s: \"y\" element not found\n", function_name);
	ds_dealloc_multi (multi_desc);
	return;
    }
    y_offset = ds_get_element_offset (list_desc, y_index);
    header = *(list_header **) multi_desc->data[0];
    if (header->length < 1)
    {
	ds_dealloc_multi (multi_desc);
	return;
    }
    if ( !ds_list_fragment (list_desc, header) )
    {
	fprintf (stderr, "%s: unable to fragment list\n", function_name);
	ds_dealloc_multi (multi_desc);
	return;
    }
    /*  First remove all the old vertices  */
    for (curr = top->profileEdit.left_vertex; curr != NULL; curr = next)
    {
	next = curr->right;
	m_free (curr);
    }
    top->profileEdit.left_vertex = NULL;
    top->profileEdit.right_vertex = NULL;
    for (entry = header->first_frag_entry; entry != NULL; entry = entry->next)
    {
	if ( !ds_get_element (entry->data + x_offset,
			      list_desc->element_types[x_index], x, NULL) )
	{
	    fprintf (stderr, "%s: error converting x value\n", function_name);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	if ( !ds_get_element (entry->data + y_offset,
			      list_desc->element_types[y_index], y, NULL) )
	{
	    fprintf (stderr, "%s: error converting y value\n", function_name);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
	if ( !ProfileEdit__add_vertex (top, x[0], y[0]) )
	{
	    fprintf (stderr, "%s: error adding vertex\n", function_name);
	    ds_dealloc_multi (multi_desc);
	    return;
	}
    }
    ds_dealloc_multi (multi_desc);
    kwin_refresh_if_visible (top->profileEdit.pixcanvas, TRUE);
}   /*  End Function filesel_cbk  */

static void show_markers_toggle_cbk (Widget w, XtPointer client_data,
				     XtPointer call_data)
{
    uaddr bool = (uaddr) call_data;
    flag refresh;
    ProfileEditWidget top = (ProfileEditWidget) client_data;

    refresh = top->profileEdit.show_markers;
    top->profileEdit.show_markers = bool;
    kwin_refresh_if_visible (top->profileEdit.pixcanvas, refresh);
}   /*  End Function show_markers_toggle_cbk  */
