/*LINTLIBRARY*/
/*  TracePopup.c

    This code provides a trace display widget for Xt.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a trace display widget for Xt.


    Written by      Richard Gooch   15-SEP-1996

    Updated by      Richard Gooch   16-SEP-1996

    Updated by      Richard Gooch   19-SEP-1996: Changed to special function
  call interface for setting new array.

    Updated by      Richard Gooch   9-OCT-1996: Store trace details and use
  refresh function. Added print button.

    Updated by      Richard Gooch   5-NOV-1996: Added track display.

    Updated by      Richard Gooch   6-NOV-1996: Added horizontal pixel
  co-ordinate in track display and put both track dislays on their own line.

    Updated by      Richard Gooch   11-NOV-1996: Only refresh canvas if visible

    Updated by      Richard Gooch   30-DEC-1996: Possible bugfix in
  <XkwTracePopupNewArray> when NULL array passed.

    Updated by      Richard Gooch   27-JAN-1997: Changed interface to
  <XkwTracePopupNewArray> function.

    Updated by      Richard Gooch   28-JAN-1997: Changed back to previous
  interface to <XkwTracePopupNewArray> function: now construct restriction
  information from co-ordinate indices.

    Updated by      Richard Gooch   31-JAN-1997: Added drag&zoom functionality.

    Updated by      Richard Gooch   1-FEB-1997: Made use of
  <wcs_astro_get_associated> function.

    Updated by      Richard Gooch   11-FEB-1997: Added "Unzoom" and "Smooth"
  controls.

    Updated by      Richard Gooch   21-FEB-1997: Added "Auto Zoom" toggle.

    Updated by      Richard Gooch   23-FEB-1997: Added "Axis Labels" button.

    Updated by      Richard Gooch   24-FEB-1997: Added simple clipping.

    Updated by      Richard Gooch   5-MAR-1997: Fixed bug when Y scale range
  was set to 0.

    Updated by      Richard Gooch   24-MAR-1997: Fixed bug drawing zero line
  when autoscaling is enabled: line was drawn before scale was changed.

    Updated by      Richard Gooch   27-MAR-1997: Worked around bug in GCC 2.7.2
  on an alpha.

    Updated by      Richard Gooch   10-APR-1997: Copy trace from array to
  internal buffer to deal with problems when data array changes (i.e. scrolling
  buffer).

    Updated by      Richard Gooch   13-APR-1997: Took account of change to
  <ds_format_unit> function.

    Updated by      Richard Gooch   22-APR-1997: Fixed bug in
  <XkwTracePopupNewArray>: trace buffer was being set to NULL.

    Updated by      Richard Gooch   4-JUN-1997: Supported non-float arrays.

    Updated by      Richard Gooch   7-JUN-1997: Add profile save.

    Updated by      Richard Gooch   19-JUN-1997: Switched to <ds_format_value>.

    Updated by      Richard Gooch   26-JUL-1997: Attempt to preserve horizontal
  range if user zooms in and then a new profile is given with
  <XkwTracePopupShowTrace>.

    Updated by      Richard Gooch   31-JUL-1997: Made use of new editor context
  in <kwin> package for zoom box. Created <XkwTracePopupShowTraces> function.

    Updated by      Richard Gooch   1-AUG-1997: Created function
  <XkwTracePopupShowBoxedTrace>.

    Updated by      Richard Gooch   4-AUG-1997: Added rescaling of "JY/PIX" and
  "JY/BEAM" to "JY" when appropriate.

    Updated by      Richard Gooch   6-AUG-1997: Added test to flux rescaling
  to ensure the two axes are sky co-ordinates (don't want one being frequency).
  Use canvas X & Y labels for track display and save format rather than getting
  from iarray (to capture flux rescaling).

    Updated by      Richard Gooch   12-AUG-1997: Fixed bug in Initialise:
  array was not initialised.

    Updated by      Richard Gooch   16-AUG-1997: Forced trace colour to be
  black if drawing to PostScript.

    Updated by      Richard Gooch   20-AUG-1997: Made use of new functions for
  testing if axes are longitude or latitude.

    Updated by      Richard Gooch   16-SEP-1997: More robust initialisation of
  trace_dim. Apply data scaling everywhere.

    Updated by      Richard Gooch   25-NOV-1997: Fixed bug in <geom_func> to
  ensure sensible boundaries passed to <canvas_set_attributes>.

    Updated by      Richard Gooch   17-DEC-1997: Improved auto-zooming.

    Updated by      Richard Gooch   22-FEB-1998: Switched to
  <XkwDressingControlAddCanvas>.

    Updated by      Richard Gooch   27-FEB-1998: Switched to overlay menu.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   12-AUG-1998: Now update cursor position on
  any canvas event.

    Updated by      Richard Gooch   14-OCT-1998: Handle negative data scaling.

    Updated by      Richard Gooch   13-MAY-2000: Show multi-colour trace even
  for PostScript output.

    Updated by      Richard Gooch   14-JUN-2000: Do not show zero line if it
  falls outside canvas boundary.

    Updated by      Richard Gooch   15-JUN-2000: Implemented 'i' and 'o' zoom
  key bindings. Moved profile plotting into new experimental <profile>
  package. Added support for "join" and "steps" profile display modes.

    Updated by      Richard Gooch   20-JUN-2000: Took account of rename of
  <profile> package to <graph1d>. Created function <XkwTracePopupInvalidate>.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kcmap.h>

    Updated by      Richard Gooch   22-SEP-2000: Took account of interface
  change in <ds_format_value>.

    Updated by      Richard Gooch   8-MAR-2001: Make calls to function
  <canvas_use_astro_transform> when canvas is realised and when a new array is
  provided instead of during refresh. This ensures projection information is
  up-to-date before the precompute callbacks for the canvas are called.

    Updated by      Richard Gooch   9-MAR-2001: Added calls to <canvas_specify>
  to clear axis labels when new array is provided or old one is destroyed.

    Updated by      Richard Gooch   10-MAR-2001: Fixed memory leak: function
  <iarr_destroy_callback> was dropping trace buffer without freeing. Fixed
  multiple traces so that output is not blank if at least one input is not
  blank.

    Updated by      Richard Gooch   11-MAR-2001: Set canvas restrictions based
  on midpoint of traces rather than first trace. Added filename, position and
  number of traces to saved ASCII profile output.

    Updated by      Richard Gooch   13-MAR-2001: Always report number of
  profiles used to produce output profile in ASCII output.

    Updated by      Richard Gooch   20-MAR-2001: Created Destroy method.

    Updated by      Richard Gooch   31-MAY-2001: Changed from <kwin_resize> to
  <kwin_clear> in <XkwTracePopupNewArray>.

    Updated by      Richard Gooch   16-AUG-2001: Made use of function
  <canvas_warp_pointer>.

    Updated by      Richard Gooch   20-AUG-2001: Fixed bugs in <do_zoom> when
  canvas boundaries exceeded. Only warp pointer for zoom in and zoom out.

    Updated by      Richard Gooch   4-SEP-2001: Discard events with key presses
  in <geom_func>. Created <_TracePopup_size_control_func> to smooth data and
  determine min&max, so that canvas range can be safely set.

    Last updated by Richard Gooch   10-NOV-2001: Fixed bug introduced in
  previous change which had inconsistent calculation of <<y_arr>> start.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#include <k_event_codes.h>
#include <karma_graph1d.h>
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_kcmap.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_ic.h>
#include <karma_ch.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_c.h>
#include <Xkw/TracePopupP.h>
#include <Xkw/Canvas.h>
#include <Xkw/Postscript.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/OverlayMenu.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/ExclusiveMenu.h>


/*  Private functions  */
STATIC_FUNCTION (void _TracePopup_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _TracePopup_Destroy, (Widget W) );
STATIC_FUNCTION (void iarr_destroy_callback,
		 (iarray arr,TracePopupWidget top) );
STATIC_FUNCTION (void _TracePopup_canvas_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _TracePopup_size_control_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  void **info, flag *boundary_clear) );
STATIC_FUNCTION (void _TracePopup_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void postscript_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (flag _TracePopup_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin,
		  int x_pix, int y_pix) );
STATIC_FUNCTION (flag geom_func,
		 (KPixCanvasEditorContext context, KPixCanvas canvas,
		  unsigned int object_type, unsigned int num_coords,
		  CONST int *x, CONST int *y,
		  unsigned int raw_event_code, void *raw_e_info,
		  void **f_info) );
STATIC_FUNCTION (void unzoom_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (flag do_zoom, (TracePopupWidget w, char action,
				double lx, double wy) );
STATIC_FUNCTION (void smooth_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void auto_zoom_cbk, (Widget w, XtPointer client_data,
				      XtPointer call_data) );
STATIC_FUNCTION (void save_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void refresh_cbk, (Widget w, XtPointer client_data,
				    XtPointer call_data) );


#define offset(field) XtOffset(TracePopupWidget, tracePopup.field)
#define XkwRDouble "Double"

static XtResource TracePopupResources[] = 
{
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, FALSE},
    {XkwNcanvasVisual, XtCVisual, XtRVisual, sizeof (Visual *),
     offset (canvasVisual), XtRImmediate, CopyFromParent},
    {XkwNworldCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (worldCanvas), XtRImmediate, NULL},
    {XkwNrealiseCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (realiseCallback), XtRCallback, (caddr_t) NULL},
    {XkwNkarmaColourmap, XkwCKarmaColourmap, XtRPointer, sizeof (XtPointer),
     offset (karmaColourmap), XtRImmediate, (caddr_t) NULL},
};
#undef offset

#define PROFILE_MODE_JOIN               0
#define PROFILE_MODE_STEP               1
#define NUM_PROFILE_MODE_CHOICES        2

static char *profile_mode_choices[NUM_PROFILE_MODE_CHOICES] =
{
    "join", "hist",
};

TracePopupClassRec tracePopupClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,     /* superclass */
    "TracePopup",                             /* class_name */
    sizeof (TracePopupRec),                   /* widget_size */
    NULL,                                     /* class_initialise */
    NULL,                                     /* class_part_initialise */
    FALSE,                                    /* class_init */
    (XtInitProc) _TracePopup_Initialise,      /* initialise */
    NULL,                                     /* initialise_hook */
    XtInheritRealize,                         /* realise */
    NULL,                                     /* actions */
    0,                                        /* num_actions */
    TracePopupResources,                      /* resources */
    XtNumber (TracePopupResources),           /* num_resources */
    NULLQUARK,                                /* xrm_class */
    TRUE,                                     /* compress_motion */
    TRUE,                                     /* compress_exposure */
    TRUE,                                     /* compress_enterleave */
    TRUE,                                     /* visible_interest */
    _TracePopup_Destroy,                      /* destroy */
    XtInheritResize,                          /* resize */
    NULL,                                     /* expose */
    NULL,                                     /* set_values */
    NULL,                                     /* set_values_hook */
    XtInheritSetValuesAlmost,                 /* set_values_almost */
    NULL,                                     /* get_values_hook */
    NULL,                                     /* accept_focus */
    XtVersion,                                /* version */
    NULL,                                     /* callback_private */
    NULL,                                     /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* TracePopupClassPart */
      0                   /* empty */
  }
};

WidgetClass tracePopupWidgetClass = (WidgetClass) &tracePopupClassRec;


static void _TracePopup_Initialise (Widget Request, Widget New)
{
    /*TracePopupWidget request = (TracePopupWidget) Request;*/
    TracePopupWidget new = (TracePopupWidget) New;
    Widget form, w, print_btn, cnv, upper_widget;
    Widget save_dialog_popup;
    /*static char function_name[] = "TracePopupWidget::Initialise";*/

    new->tracePopup.iarr_destroy_callback = NULL;
    new->tracePopup.ap = NULL;
    new->tracePopup.trace_dim = -1;
    new->tracePopup.trace = NULL;
    new->tracePopup.weights = NULL;
    new->tracePopup.trace_buf_len = 0;
    new->tracePopup.pixcanvas = NULL;
    new->tracePopup.worldCanvas = NULL;
    new->tracePopup.array = NULL;
    new->tracePopup.smooth = FALSE;
    new->tracePopup.auto_zoom = FALSE;
    new->tracePopup.coord_buf_len = 0;
    new->tracePopup.num_coord_values = 0;
    new->tracePopup.coord_buf = NULL;
    new->tracePopup.num_restr_allocated = 0;
    new->tracePopup.restr_names = NULL;
    new->tracePopup.restr_values = NULL;
    new->tracePopup.coord_indices_buf_len = 0;
    new->tracePopup.coord_indices = NULL;
    new->tracePopup.coord_counters_buf_len = 0;
    new->tracePopup.coord_counters = NULL;
    new->tracePopup.valid_data = FALSE;
    new->tracePopup.num_traces = 0;
    xtmisc_set_icon (New, ic_write_kplot_icon);
    w = XtVaCreatePopupShell ("savePopup", dialogpopupWidgetClass, New,
			      XtNlabel, "Filename         ",
			      XkwNextension, "",
			      NULL);
    save_dialog_popup = w;
    XtAddCallback (w, XtNcallback, save_cbk, New);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Zoom",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, auto_zoom_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Unzoom",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, unzoom_cbk, New);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Smooth",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, smooth_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Print",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNheight, 20,
				 NULL);
    print_btn = w;
    w = XtVaCreateManagedWidget ("overlayMenu", overlayMenuWidgetClass, form,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    new->tracePopup.overlay_menu = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Save Profile",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popup_cbk, save_dialog_popup);
    w = XtVaCreateManagedWidget ("traceStyleMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "traceStyleMenu",
				 XkwNchoiceName, "Style",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNnumItems, NUM_PROFILE_MODE_CHOICES,
				 XkwNitemStrings, profile_mode_choices,
				 XkwNvaluePtr, &new->tracePopup.mode,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, refresh_cbk, New);
    w = XtVaCreateManagedWidget ("trackLabel0", labelWidgetClass, form,
				 XtNlabel,
				 "Track Output                               ",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, w,
				 XtNheight, 20,
				 NULL);
    new->tracePopup.track_label0 = w;
    w = XtVaCreateManagedWidget ("trackLabel1", labelWidgetClass, form,
				 XtNlabel,
				 "Track Output                               ",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNfromVert, w,
				 XtNheight, 20,
				 NULL);
    new->tracePopup.track_label1 = w;
    w = XtVaCreateManagedWidget ("pseudoColourCanvas", canvasWidgetClass, form,
				 XtNfromVert, w,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainRight,
				 XtNwidth, 410,
				 XtNheight, 300,
				 XtNvisual, new->tracePopup.canvasVisual,
				 XkwNsilenceUnconsumed, True,
				 XtNresizable, True,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainBottom,
				 NULL);
    cnv = w;
    XtAddCallback (cnv, XkwNrealiseCallback,
		   _TracePopup_canvas_realise_cbk, (XtPointer) New);
    w = XtVaCreatePopupShell ("postscriptwinpopup", postscriptWidgetClass, New,
			      XtNtitle, "Postscript Window",
			      XkwNautoIncrement, TRUE,
			      NULL);
    new->tracePopup.pswinpopup = w;
    XtAddCallback (w, XtNcallback, postscript_cbk, New);
    XtAddCallback (print_btn, XtNcallback, xtmisc_popup_cbk, w);
}   /*  End Function Initialise  */

static void _TracePopup_Destroy (Widget W)
{
    TracePopupWidget w = (TracePopupWidget) W;

    if (w->tracePopup.iarr_destroy_callback != NULL)
	c_unregister_callback (w->tracePopup.iarr_destroy_callback);
    if (w->tracePopup.trace != NULL) m_free (w->tracePopup.trace);
    if (w->tracePopup.weights != NULL) m_free (w->tracePopup.weights);
    if (w->tracePopup.coord_buf != NULL) m_free (w->tracePopup.coord_buf);
    if (w->tracePopup.restr_names != NULL) m_free (w->tracePopup.restr_names);
    if (w->tracePopup.restr_values != NULL) m_free(w->tracePopup.restr_values);
    if (w->tracePopup.coord_indices != NULL)
	m_free (w->tracePopup.coord_indices);
    if (w->tracePopup.coord_counters != NULL)
	m_free (w->tracePopup.coord_counters);
}   /*  End Function Destroy  */

static void iarr_destroy_callback (iarray arr, TracePopupWidget top)
/*  [PURPOSE] This routine will register the destruction of the Intelligent
    Array.
    <arr> The Intelligent Array.
    <top> The TracePopup widget.
    [RETURNS] Nothing.
*/
{
    top->tracePopup.array = NULL;
    top->tracePopup.iarr_destroy_callback = NULL;
    top->tracePopup.trace_dim = -1;
    if (top->tracePopup.worldCanvas != NULL)
    {
	canvas_use_astro_transform (top->tracePopup.worldCanvas, NULL);
	canvas_specify (top->tracePopup.worldCanvas, NULL, NULL,
			0, (CONST char **) NULL, NULL);
    }
    top->tracePopup.ap = NULL;
}   /*  End Function iarr_destroy_callback  */

static void _TracePopup_canvas_realise_cbk (Widget w, XtPointer client_data,
					    XtPointer call_data)
/*  [SUMMARY] Canvas realise callback.
    [PURPOSE] This routine is called when a Canvas widget is realised. A
    Kcolourmap object is created if required, then the KWorldCanvas is created.
    <w> The Canvas widget that has just been realised.
    <client_data> The TracePopup widget.
    <call_data> The pixel canvas(es).
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas = (KPixCanvas) call_data;
    TracePopupWidget top = (TracePopupWidget) client_data;
    KPixCanvasEditorContext ctx;
    unsigned short red, green, blue;
    unsigned int visual;
    unsigned long pixel_value;
    double scale, offset, min, max, tmp;
    static char function_name[] = "TracePopupWidget::canvas_realise_cbk";

    top->tracePopup.pixcanvas = pixcanvas;
    top->tracePopup.worldCanvas = canvas_create (pixcanvas, NULL, NULL);
    canvas_use_astro_transform (top->tracePopup.worldCanvas,
				top->tracePopup.ap);
    XtVaGetValues (w,
		   XtNforeground, &top->tracePopup.canvas_foreground,
		   NULL);
    canvas_set_dressing (top->tracePopup.worldCanvas,
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_TOP_TICKS, TRUE,
			 KCD_ATT_BOTTOM_TICKS, TRUE,
			 KCD_ATT_LEFT_TICKS, TRUE,
			 KCD_ATT_RIGHT_TICKS, TRUE,
			 KCD_ATT_END);
    if (top->tracePopup.array != NULL)
    {
	iarray_get_data_scaling (top->tracePopup.array, &scale, &offset);
	min = top->tracePopup.array_minimum * scale + offset;
	max = top->tracePopup.array_maximum * scale + offset;
	if (min > max) swap (min, max, tmp);
	canvas_set_attributes (top->tracePopup.worldCanvas,
			       CANVAS_ATT_BOTTOM_Y, min,
			       CANVAS_ATT_TOP_Y, max,
			       CANVAS_ATT_END);
    }
    canvas_register_size_control_func (top->tracePopup.worldCanvas,
				       _TracePopup_size_control_func, top);
    canvas_register_refresh_func (top->tracePopup.worldCanvas,
				  _TracePopup_refresh_func, top);
    canvas_register_position_event_func (top->tracePopup.worldCanvas,
					 _TracePopup_position_func, top);
    XtCallCallbacks ( (Widget) top, XkwNrealiseCallback,
		      (XtPointer) top->tracePopup.worldCanvas );
    XkwOverlayMenuAddCanvas (top->tracePopup.overlay_menu,
			     top->tracePopup.worldCanvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_BACKGROUND_PIXEL,
			 &top->tracePopup.previous_background,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    kwin_get_pixel_RGB_values (pixcanvas, &top->tracePopup.previous_background,
			       &red, &green, &blue, 1);
    top->tracePopup.previous_intensity = sqrt (red * red + green * green +
					       blue * blue) / 65535.0;
    if (visual == KWIN_VISUAL_PSEUDOCOLOUR)
    {
	if ( !kwin_get_colour (pixcanvas, "green", &pixel_value,
			       NULL, NULL, NULL) )
	{
	    /*  Can't get green: try white instead  */
	    if ( !kwin_get_colour (pixcanvas, "white", &pixel_value,
				   NULL, NULL, NULL) )
	    {
		/*  Can't even get white (this should never actually happen
		    with X), so just pick 0  */
		pixel_value = 0;
	    }
	}
    }
    else
    {
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_PIX_GREEN_MASK, &pixel_value,
			     KWIN_ATT_END);
    }
    /*  Create the dragzoom editor context  */
    ctx = kwin_editor_create_context (pixcanvas,
				      KWIN_EDITOR_CONTEXT_ATT_MOUSE,
				      K_CANVAS_EVENT_LEFT_MOUSE_MASK,
				      KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
				      KWIN_EDIT_OBJECT_RECTANGLE,
				      KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE,
				      pixel_value,
				      KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
				      FALSE,
				      KWIN_EDITOR_CONTEXT_ATT_END);
    if (ctx == NULL) m_abort (function_name, "editor context");
    kwin_editor_register_event_func (ctx, geom_func, top);
}   /*  End Function canvas_realise_cbk   */

static void _TracePopup_size_control_func (KWorldCanvas canvas,
					   int width, int height,
					   struct win_scale_type *win_scale,
					   void **info, flag *boundary_clear)
/*  [SUMMARY] Size control callback.
    [PURPOSE] This routine will modify the window scaling information for a
    world canvas. While this routine is running, colourmap resize events are
    ignored. Hence this routine may safely cause the associated Kcolourmap
    object to be resized.
    <canvas> The world canvas object.
    <width> The width of the canvas in pixels after removing the space consumed
    by the dressing.
    <height> The height of the canvas in pixels after removing the space
    consumed by the dressing.
    <win_scale> The window scaling information. This may be modified.
    <info> A pointer to the arbitrary canvas information pointer.
    <boundary_clear> If the value TRUE is written here then the
    [<canvas_resize>] routine will attempt to clear only the boundary between
    the pixel canvas and the world canvas. If the value FALSE is written here
    or nothing is written here, the [<canvas_resize>] routine will clear the
    entire pixel canvas as appropriate.
    [RETURNS] Nothing.
*/
{
    unsigned int count, start, length;
    float Vm1, V0, Vp1;
    double left_x, right_x, tmp;
    double min_y = TOOBIG;
    double max_y = -TOOBIG;
    TracePopupWidget w = (TracePopupWidget) *info;
    double *x_arr, *y_arr;
    dim_desc *dim;
    static char function_name[]="TracePopupWidget::size_control_func";

    win_scale->x_offset = 0;
    win_scale->y_offset = 0;
    win_scale->x_pixels = width;
    win_scale->y_pixels = height;
    if (w->tracePopup.num_coord_values != 0) return;
    if (w->tracePopup.worldCanvas == NULL) return;
    if (w->tracePopup.array == NULL) return;
    if (w->tracePopup.trace == NULL) return;
    if (w->tracePopup.trace_dim < 0) return;
    if (!w->tracePopup.valid_data) return;
    canvas_get_attributes (w->tracePopup.worldCanvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_END);
    dim = iarray_get_dim_desc (w->tracePopup.array, w->tracePopup.trace_dim);
    start = ds_get_coord_num (dim, left_x, SEARCH_BIAS_LOWER);
    length = ds_get_coord_num (dim, right_x, SEARCH_BIAS_UPPER) + 1 - start;
    /*  Ensure buffer is big enough  */
    if (w->tracePopup.coord_buf_len < length * 2)
    {
	if (w->tracePopup.coord_buf != NULL) m_free (w->tracePopup.coord_buf);
	if ( ( w->tracePopup.coord_buf =
	       m_alloc (length * sizeof *w->tracePopup.coord_buf * 2) )
	     == NULL )
	{
	    m_abort (function_name, "coord array");
	}
	w->tracePopup.coord_buf_len = length * 2;
    }
    x_arr = w->tracePopup.coord_buf;
    y_arr = w->tracePopup.coord_buf + w->tracePopup.coord_buf_len / 2;
    /*  Compute array of X co-ordinates in world co-ordinates  */
    for (count = 0; count < length; ++count)
    {
	x_arr[count] = iarray_get_coordinate (w->tracePopup.array,
					      w->tracePopup.trace_dim,
					      start + count);
	y_arr[count] = 0.0;
    }
    canvas_coords_transform (w->tracePopup.worldCanvas, length,
			     x_arr, FALSE, y_arr, FALSE);
    /*  Copy values into buffer  */
    if (w->tracePopup.smooth)
    {
	/*  Apply Hanning smoothing  */
	for (count = 1; count + 1 < length; count += 2)
	{
	    Vm1 = w->tracePopup.trace[start + count - 1];
	    V0 = w->tracePopup.trace[start + count];
	    Vp1 = w->tracePopup.trace[start + count + 1];
	    if ( DOUBLE_IS_BLANK (Vm1) ) Vm1 = 0.0;
	    if ( DOUBLE_IS_BLANK (V0) ) V0 = 0.0;
	    if ( DOUBLE_IS_BLANK (Vp1) ) Vp1 = 0.0;
	    x_arr[count / 2] = x_arr[count];
	    tmp = 0.25 * (Vm1 + Vp1) + 0.5 * V0;
	    y_arr[count / 2] = tmp;
	    if (tmp < min_y) min_y = tmp;
	    if (tmp > max_y) max_y = tmp;
	}
	length = count / 2;
    }
    else
    {
	for (count = 0; count < length; ++count)
	{
	    tmp = w->tracePopup.trace[start + count];
	    y_arr[count] = tmp;
	    if ( DOUBLE_IS_VALID (tmp) )
	    {
		if (tmp < min_y) min_y = tmp;
		if (tmp > max_y) max_y = tmp;
	    }
	}
    }
    if ( w->tracePopup.auto_zoom && (min_y < max_y) )
	canvas_set_attributes (w->tracePopup.worldCanvas,
			       CANVAS_ATT_BOTTOM_Y, min_y,
			       CANVAS_ATT_TOP_Y, max_y,
			       CANVAS_ATT_END);
    else canvas_get_attributes (w->tracePopup.worldCanvas,
				CANVAS_ATT_BOTTOM_Y, &min_y,
				CANVAS_ATT_TOP_Y, &max_y,
				CANVAS_ATT_END);
    w->tracePopup.num_coord_values = length;
}   /*  End Function _TracePopup_size_control_func  */

static void _TracePopup_refresh_func (KWorldCanvas canvas,
				      int width, int height,
				      struct win_scale_type *win_scale,
				      Kcolourmap cmap, flag cmap_resize,
				      void **info, PostScriptPage pspage,
				      unsigned int num_areas,
				      KPixCanvasRefreshArea *areas,
				      flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    unsigned int length, num_pixels;
    unsigned long pixel_value, background, *pixel_values;
    double left_x, right_x, wlx, wrx, dummy, min_y, max_y;
    TracePopupWidget w = (TracePopupWidget) *info;
    char *trace_colour, *zero_colour;
    double *x_arr, *y_arr;
    /*static char function_name[] = "TracePopupWidget::refresh_func";*/

    length = w->tracePopup.num_coord_values;
    w->tracePopup.num_coord_values = 0;
    if (w->tracePopup.worldCanvas == NULL) return;
    if (w->tracePopup.array == NULL) return;
    if (w->tracePopup.trace == NULL) return;
    if (w->tracePopup.trace_dim < 0) return;
    if (!w->tracePopup.valid_data) return;
    pixcanvas = canvas_get_pixcanvas (w->tracePopup.worldCanvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_BACKGROUND_PIXEL, &background,
			 KWIN_ATT_END);
    if (background != w->tracePopup.previous_background)
    {
	unsigned short red, green, blue;
	double r, g, b;

	w->tracePopup.previous_background = background;
	kwin_get_pixel_RGB_values (pixcanvas, &background,
				   &red, &green, &blue, 1);
	/*  Convert to doubles because multiplying 65535 * 65535 inside shorts
	    causes problems (SEGV) with GCC 2.7.2 on an alpha  */
	r = red;
	g = green;
	b = blue;
	w->tracePopup.previous_intensity = sqrt (r * r + g * g +
						 b * b) / 65535.0;
    }
    if ( (w->tracePopup.previous_intensity >= 0.5) || (pspage != NULL) )
    {
	/*  Light background  */
	trace_colour = "black";
	zero_colour = "blue";
    }
    else
    {
	/*  Dark background  */
	trace_colour = "white";
	zero_colour = "yellow";
    }
    x_arr = w->tracePopup.coord_buf;
    y_arr = w->tracePopup.coord_buf + w->tracePopup.coord_buf_len / 2;
    canvas_get_attributes (w->tracePopup.worldCanvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &min_y,
			   CANVAS_ATT_TOP_Y, &max_y,
			   CANVAS_ATT_END);
    /*  Convert left&right x values to non-linear co-ordinates  */
    wlx = left_x;
    dummy = 0.0;
    canvas_coords_transform (w->tracePopup.worldCanvas, 1,
			     &wlx, FALSE, &dummy, FALSE);
    wrx = right_x;
    dummy = 0.0;
    canvas_coords_transform (w->tracePopup.worldCanvas, 1,
			     &wrx, FALSE, &dummy, FALSE);
    if ( (min_y <= 0.0) && (0.0 <= max_y) )
    {
	/*  Draw zero line  */
	if ( !canvas_get_colour (w->tracePopup.worldCanvas, zero_colour,
				 &pixel_value, NULL, NULL, NULL) )
	{
	    pixel_value = w->tracePopup.canvas_foreground;
	}
	canvas_draw_line_p (w->tracePopup.worldCanvas, wlx, 0.0, wrx, 0.0,
			    pixel_value);
    }
    if (w->tracePopup.karmaColourmap == NULL)
    {
	canvas_get_colour (w->tracePopup.worldCanvas, trace_colour,
			   &pixel_value, NULL, NULL, NULL);
	num_pixels = 0;
	pixel_values = NULL;
    }
    else
    {
	/*  Use pixels from supplied colourmap  */
	pixel_value = 0;
	num_pixels = kcmap_get_pixels (w->tracePopup.karmaColourmap,
				       &pixel_values);
    }
    if (w->tracePopup.mode == PROFILE_MODE_JOIN)
	graph1d_lines (w->tracePopup.worldCanvas, length, x_arr, y_arr,
		       pixel_value, num_pixels, pixel_values);
    else
	graph1d_steps (w->tracePopup.worldCanvas, length, x_arr, y_arr,
		       pixel_value, num_pixels, pixel_values);
}   /*  End Function _TracePopup_refresh_func  */

static void postscript_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  [SUMMARY] PostScript callback.
    [PURPOSE] This is the PostScript callback. It is called when the PostScript
    widget needs to know the pixel canvas and image name.
*/
{
    TracePopupWidget top = (TracePopupWidget) client_data;

    XkwPostscriptRegisterImageAndName (w, top->tracePopup.pixcanvas, NULL);
}   /*  End Function postscript_cbk   */

static flag _TracePopup_position_func (KWorldCanvas canvas, double x, double y,
				       unsigned int event_code, void *e_info,
				       void **f_info,
				       double x_lin, double y_lin,
				       int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    unsigned int num_restr;
    unsigned long px;
    double scale, offset;
    TracePopupWidget w = (TracePopupWidget) *f_info;
    dim_desc *dim;
    char *ptr, *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    CONST char *string = e_info;
    char txt[STRING_LENGTH];
    char unit_string[STRING_LENGTH];
    /*static char function_name[] = "TracePopupWidget::position_func";*/

    switch (event_code)
    {
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	return do_zoom (w, 'i', x_lin, y);
	/*break;*/
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
	return do_zoom (w, 'o', x_lin, y);
	/*break;*/
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	if (string[1] != '\0') return (FALSE);
	return do_zoom (w, string[0], x_lin, y);
	/*break;*/
    }
    if (w->tracePopup.trace == NULL) return (FALSE);
    if (w->tracePopup.array == NULL) return (FALSE);
    if (w->tracePopup.trace_dim < 0) return (FALSE);
    /*  First display the x co-ordinate information  */
    canvas_get_specification (canvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    dim = iarray_get_dim_desc (w->tracePopup.array, w->tracePopup.trace_dim);
    /*  Convert linear world co-ordinates to array indices and display  */
    px = ds_get_coord_num (dim, x_lin, SEARCH_BIAS_CLOSEST);
    if (w->tracePopup.ap == NULL)
    {
	sprintf (txt, "x: %lu pixel  ", px);
	ptr = strstr (txt, "pixel");
	ds_format_value (ptr + 7, x, xlabel, 1.0, 0.0, x, NULL, NULL);
    }
    else
    {
	wcs_astro_format (w->tracePopup.ap, xlabel, unit_string, x);
	sprintf (txt, "x: %lu pixel  %s", px, unit_string);
    }
    XtVaSetValues (w->tracePopup.track_label0,
		   XtNlabel, txt,
		   NULL);
    /*  Now display the y co-ordinate information. Unscale and pass scale so
	that raw value can be shown  */
    iarray_get_data_scaling (w->tracePopup.array, &scale, &offset);
    y = (y - offset) / scale;
    ds_format_value (txt, y, ylabel, scale, offset, y,
		     w->tracePopup.array->top_pack_desc,
		     *w->tracePopup.array->top_packet);
    XtVaSetValues (w->tracePopup.track_label1,
		   XtNlabel, txt,
		   NULL);
    return (FALSE);
}   /*  End Function _TracePopup_position_func  */

static flag geom_func (KPixCanvasEditorContext context, KPixCanvas canvas,
		       unsigned int object_type, unsigned int num_coords,
		       CONST int *x, CONST int *y,
		       unsigned int raw_event_code, void *raw_e_info,
		       void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    TracePopupWidget top = (TracePopupWidget) *f_info;
    double lx[2], ly[2], wx[2], wy[2];
    static char function_name[] = "TracePopupWidget::geom_func";

    if (top->tracePopup.array == NULL)
    {
	XBell (XtDisplay (top), 100);
	return (FALSE);
    }
    if (raw_event_code == K_CANVAS_EVENT_MODIFIED_KEY_PRESS) return (FALSE);
    lx[0] = x[0];
    ly[0] = y[0];
    lx[1] = x[1];
    ly[1] = y[1];
    canvas_convert_to_canvas_coords (top->tracePopup.worldCanvas, TRUE, 2,
				     lx, ly, lx, ly, wx, wy);
    if ( (x[0] == x[1]) && (y[0] == y[1]) )
	return do_zoom (top, 'p', lx[0], wy[0]);   /*  Pan  */
    if ( (x[0] == x[1]) || (y[0] == y[1]) ) return (TRUE);
    if ( (lx[0] == lx[1]) || (ly[0] == ly[1]) ) return (TRUE);
    canvas_set_attributes (top->tracePopup.worldCanvas,
			   CANVAS_ATT_LEFT_X, lx[0],
			   CANVAS_ATT_RIGHT_X, lx[1],
			   CANVAS_ATT_BOTTOM_Y, ly[1],
			   CANVAS_ATT_TOP_Y, ly[0],
			   CANVAS_ATT_END);
    if ( !kwin_resize (canvas, TRUE, 0, 0, 0, 0) )
	fprintf (stderr, "%s: Error refreshing window\n", function_name);
    return (TRUE);
}   /*  End Function geom_func  */

static void unzoom_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the unzoom callback.
*/
{
    do_zoom ( (TracePopupWidget) client_data, 'u', 0.0, 0.0 );
}   /*  End Function unzoom_cbk   */

static flag do_zoom (TracePopupWidget w, char action, double lx, double wy)
/*  [SUMMARY] Perform a zoom.
    <w> The TracePopup widget.
    <action> The action to take: 'u', 'i', 'o' or 'p'.
    <lx> The horizontal linear co-ordinate of the event.
    <wy> The vertical world co-ordinate of the event.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    double scale, offset, min, max, tmp, xl, xr, yb, yt;
    dim_desc *dim;

    if ( (w->tracePopup.array == NULL) || (w->tracePopup.trace_dim < 0) )
    {
	XBell (XtDisplay (w), 100);
	return (FALSE);
    }
    dim = iarray_get_dim_desc (w->tracePopup.array,
			       w->tracePopup.trace_dim);
    iarray_get_data_scaling (w->tracePopup.array, &scale, &offset);
    min = w->tracePopup.array_minimum * scale + offset;
    max = w->tracePopup.array_maximum * scale + offset;
    if (min > max) swap (min, max, tmp);
    if (action == 'u')
    {
	xl = dim->first_coord;
	xr = dim->last_coord;
	yb = min;
	yt = max;
    }
    else if ( (action == 'i') || (action == 'o') || (action == 'p') )
    {
	double scale = 0.0;  /*  Initialised to keep compiler happy  */
	double range_x, range_y;

	canvas_get_attributes (w->tracePopup.worldCanvas,
			       CANVAS_ATT_LEFT_X, &xl,
			       CANVAS_ATT_RIGHT_X, &xr,
			       CANVAS_ATT_BOTTOM_Y, &yb,
			       CANVAS_ATT_TOP_Y, &yt,
			       CANVAS_ATT_END);
	switch (action)
	{
	  case 'i':
	    scale = 0.5;
	    break;
	  case 'o':
	    scale = 2.0;
	    break;
	  case 'p':
	    scale = 1.0;
	    break;
	}
	range_x = (xr - xl) * scale;
	range_y = (yt - yb) * scale;
	xl = lx - range_x * 0.5;
	xr = xl + range_x;
	yb = wy - range_y * 0.5;
	yt = yb + range_y;
	if (dim->first_coord < dim->last_coord)
	{
	    if (xl < dim->first_coord)
	    {
		xl = dim->first_coord;
		xr = xl + range_x;
	    }
	    if (xr > dim->last_coord)
	    {
		xr = dim->last_coord;
		xl = xr - range_x;
	    }
	    if (xl < dim->first_coord) xl = dim->first_coord;
	}
	else
	{
	    if (xl > dim->first_coord)
	    {
		xl = dim->first_coord;
		xr = xl + range_x;
	    }
	    if (xr < dim->last_coord)
	    {
		xr = dim->last_coord;
		xl = xr - range_x;
	    }
	    if (xl > dim->first_coord) xl = dim->first_coord;
	}
	if (yb < min)
	{
	    yb = min;
	    yt = min + range_y;
	}
	if (yt > max)
	{
	    yt = max;
	    yb = yt - range_y;
	}
	if (yb < min) yb = min;
    }
    else return (FALSE);
    canvas_set_attributes (w->tracePopup.worldCanvas,
			   CANVAS_ATT_LEFT_X, xl,
			   CANVAS_ATT_RIGHT_X, xr,
			   CANVAS_ATT_BOTTOM_Y, yb,
			   CANVAS_ATT_TOP_Y, yt,
			   CANVAS_ATT_END);
    kwin_refresh_if_visible (w->tracePopup.pixcanvas, TRUE);
    if ( (action == 'i') || (action == 'o') )
	canvas_warp_pointer (w->tracePopup.worldCanvas,
			     (xl + xr) * 0.5, (yb + yt) * 0.5, TRUE, TRUE);
    return (TRUE);
}   /*  End Function do_zoom   */

static void smooth_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the smooth callback.
*/
{
    flag bool = (uaddr) call_data;
    TracePopupWidget top = (TracePopupWidget) client_data;

    top->tracePopup.smooth = bool;
    kwin_refresh_if_visible (top->tracePopup.pixcanvas, TRUE);
}   /*  End Function smooth_cbk   */

static void auto_zoom_cbk (Widget w, XtPointer client_data,XtPointer call_data)
/*  This is the auto zoom callback.
*/
{
    flag bool = (uaddr) call_data;
    TracePopupWidget top = (TracePopupWidget) client_data;

    top->tracePopup.auto_zoom = bool;
    kwin_refresh_if_visible (top->tracePopup.pixcanvas, TRUE);
}   /*  End Function auto_zoom_cbk   */

static void save_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the save callback.
*/
{
    Channel ch;
    unsigned int length, count, num_restr;
    double x, y;
    TracePopupWidget top = (TracePopupWidget) client_data;
    multi_array *multi_desc;
    char *filename = NULL;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    String fname = (String) call_data;

    if (top->tracePopup.array == NULL) return;
    if (top->tracePopup.trace == NULL) return;
    length = iarray_dim_length (top->tracePopup.array,
				top->tracePopup.trace_dim);
    canvas_get_specification (top->tracePopup.worldCanvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    if ( ( ch = ch_open_file (fname, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 fname, ERRSTRING);
	return;
    }
    multi_desc = iarray_multi_array (top->tracePopup.array);
    ds_get_named_object (multi_desc, multi_desc->num_arrays,
			 "filename", (void **) &filename);
    if (filename == NULL) ch_printf (ch, "# ASCII dump of profile\n");
    else ch_printf (ch, "# ASCII dump of profile. Original file: %s\n",
		    filename);
    if (num_restr > 1)
    {
	ch_printf (ch, "# Position of profile (%u dimensions):\n", num_restr);
	for (count = 0; count < num_restr; ++count)
	{
	    double value = restr_values[count];
	    char txt[STRING_LENGTH];

	    if (top->tracePopup.ap == NULL)
	    {
		ch_printf (ch, "# Axis: %s  co-ordinate: %g\n",
			   restr_names[count], value);
		continue;
	    }
	    wcs_astro_transform3 (top->tracePopup.ap, 1,
				  restr_names[count], &value, FALSE,
				  NULL, NULL, FALSE,
				  NULL, NULL, FALSE,
				  num_restr, (CONST char **) restr_names,
				  restr_values);
	    wcs_astro_format (top->tracePopup.ap, restr_names[count],
			      txt, value);
	    ch_printf (ch, "# %-40s  linear co-oridinate: %g\n",
		       txt, restr_values[count]);
	}
    }
    ch_printf (ch,
	       "# %u %s to produce this profile\n",
	       top->tracePopup.num_traces,
	       (top->tracePopup.num_traces < 2) ?
	       "individual profile was used" : "individual profiles were combined");
    ch_printf (ch, "# X axis name: \"%s\"  Y axis name: \"%s\"\n",
	       xlabel, ylabel);
    ch_printf (ch, "# %u values follow\n", length);
    for (count = 0; count < length; ++count)
    {
	x = iarray_get_coordinate (top->tracePopup.array,
				   top->tracePopup.trace_dim, count);
	y = top->tracePopup.trace[count];
	if (top->tracePopup.ap != NULL)
	{
	    wcs_astro_transform3 (top->tracePopup.ap, 1,
				  xlabel, &x, FALSE,
				  NULL, NULL, FALSE, NULL, NULL, FALSE,
				  0, NULL, NULL);
	}
	ch_printf (ch, "%-20.13e   %.13e\n", x, y);
    }
    ch_close (ch);
}   /*  End Function save_cbk  */

static void refresh_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the refresh callback.
*/
{
    TracePopupWidget top = (TracePopupWidget) client_data;

    kwin_refresh_if_visible (top->tracePopup.pixcanvas, TRUE);
}   /*  End Function refresh_cbk  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupNewArray (Widget W, iarray array, double min, double max)
/*  [SUMMARY] Register new array.
    <W> The TracePopup widget.
    <array> The new array. This may be NULL.
    <min> The minimum data value in the array.
    <max> The maximum data value in the array. If this is less than <<min>>
    then the minimum and maximum values are computed.
    [RETURNS] Nothing.
*/
{
    double scale, offset, tmp;
    TracePopupWidget w = (TracePopupWidget) W;
    /*static char function_name[] = "XkwTracePopupNewArray";*/

    /*  Clean up any old array  */
    if (w->tracePopup.iarr_destroy_callback != NULL)
    {
	c_unregister_callback (w->tracePopup.iarr_destroy_callback);
	w->tracePopup.iarr_destroy_callback = NULL;
    }
    if (w->tracePopup.worldCanvas != NULL)
    {
	canvas_use_astro_transform (w->tracePopup.worldCanvas, NULL);
	canvas_specify (w->tracePopup.worldCanvas, NULL, NULL,
			0, (CONST char **) NULL, NULL);
    }
    w->tracePopup.ap = NULL;
    w->tracePopup.trace_dim = -1;
    w->tracePopup.array = array;
    if (w->tracePopup.pixcanvas != NULL)
	kwin_clear (w->tracePopup.pixcanvas, 0, 0, -1, -1);
    w->tracePopup.x_uninitialised = TRUE;
    /*  Process new array  */
    if (array == NULL) return;
    if (max < min)
    {
	/*  Compute the minimum and maximum  */
	iarray_min_max (w->tracePopup.array, CONV_CtoR_REAL,
			&w->tracePopup.array_minimum,
			&w->tracePopup.array_maximum);
	min = w->tracePopup.array_minimum;
	max = w->tracePopup.array_maximum;
    }
    else
    {
	w->tracePopup.array_minimum = min;
	w->tracePopup.array_maximum = max;
    }
    w->tracePopup.iarr_destroy_callback = 
	iarray_register_destroy_func (w->tracePopup.array,
				      ( flag (*) () ) iarr_destroy_callback,
				      w);
    w->tracePopup.ap = wcs_astro_get_associated (array->multi_desc,
						 array->array_num);
    iarray_get_data_scaling (w->tracePopup.array, &scale, &offset);
    if (w->tracePopup.worldCanvas != NULL)
    {
	min = min * scale + offset;
	max = max * scale + offset;
	if (min > max) swap (min, max, tmp);
	canvas_set_attributes (w->tracePopup.worldCanvas,
			       CANVAS_ATT_BOTTOM_Y, min,
			       CANVAS_ATT_TOP_Y, max,
			       CANVAS_ATT_END);
	canvas_use_astro_transform (w->tracePopup.worldCanvas,
				    w->tracePopup.ap);
    }
}   /*  End Function XkwTracePopupNewArray  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupShowTrace (Widget W, CONST unsigned int *dim_indices,
			     CONST uaddr *coord_indices)
/*  [SUMMARY] Show a trace.
    <W> The widget.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <coord_indices> The coordinate indices of the trace. The length of this
    array must be one less than the number of dimensions in the array.
    [RETURNS] Nothing.
*/
{
    XkwTracePopupShowTraces (W, dim_indices, TRUE, 1.0, 1, coord_indices,NULL);
}   /*  End Function XkwTracePopupShowTrace  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupShowTraces (Widget W, CONST unsigned int *dim_indices,
			      flag sum, float global_weight,
			      unsigned int num_traces,
			      CONST uaddr *coord_indices, CONST float *weights)
/*  [SUMMARY] Show a weighted, summed trace.
    <W> The widget.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <sum> If TRUE, the data are summed rather than averaged.
    <global_weight> This weight value is applied to each trace as it is
    combined.
    <num_traces> The number of traces to combine.
    <coord_indices> The co-ordinate indices for the traces. This is grouped
    into <<num_traces>> contiguous blocks of co-ordinate traces. Each block
    must contain <<num_dim>> - 1 co-ordinate indices.
    <weights> The weights to apply to each trace as it is combined. If this is
    NULL no weights are applied.
    [RETURNS] Nothing.
*/
{
    int trace_dim = -1;
    unsigned int trace_count, count, dim_count, num_dim, type;
    TracePopupWidget w = (TracePopupWidget) W;
    float weight, val;
    double scale, offset;
    double value[2];
    dim_desc *dim;
    CONST char *trace;
    CONST char *value_name;
    uaddr *offsets;
    static char function_name[] = "XkwTracePopupShowTraces";

    if (w->tracePopup.worldCanvas == NULL) return;
    if (w->tracePopup.array == NULL) return;
    w->tracePopup.num_traces = num_traces;  /*  For <<save_cbk>>  */
    num_dim = iarray_num_dim (w->tracePopup.array);
    /*  Find trace dimension  */
    for (dim_count = 0; (dim_count < num_dim) && (trace_dim < 0); ++dim_count)
    {
	trace_dim = dim_count;
	for (count = 0; count < num_dim - 1; ++count)
	{
	    if (dim_indices[count] == dim_count) trace_dim = -1;
	}
    }
    if (trace_dim < 0)
    {
	fprintf (stderr, "No trace dimension!\n");
	a_prog_bug (function_name);
    }
    dim = iarray_get_dim_desc (w->tracePopup.array, trace_dim);
    offsets = w->tracePopup.array->offsets[trace_dim];
    type = iarray_type (w->tracePopup.array);
    /*  Ensure trace buffer is large enough  */
    if (w->tracePopup.trace_buf_len < dim->length)
    {
	if (w->tracePopup.trace != NULL) m_free (w->tracePopup.trace);
	w->tracePopup.trace = NULL;
	if (w->tracePopup.weights != NULL) m_free (w->tracePopup.weights);
	w->tracePopup.weights = NULL;
	w->tracePopup.trace_buf_len = 0;
	if ( ( w->tracePopup.trace =
	       m_alloc (sizeof *w->tracePopup.trace * dim->length) ) == NULL )
	{
	    m_error_notify (function_name, "trace buffer");
	    return;
	}
	if ( ( w->tracePopup.weights =
	       m_alloc (sizeof *w->tracePopup.weights * dim->length) ) ==NULL )
	{
	    m_error_notify (function_name, "weight buffer");
	    m_free (w->tracePopup.trace);
	    w->tracePopup.trace = NULL;
	    return;
	}
	w->tracePopup.trace_buf_len = dim->length;
    }
    /*  Ensure restriction arrays are long enough  */
    if (num_dim - 1 > w->tracePopup.num_restr_allocated)
    {
	if (w->tracePopup.restr_names != NULL)
	    m_free (w->tracePopup.restr_names);
	if ( ( w->tracePopup.restr_names =
	       m_alloc (num_dim *sizeof *w->tracePopup.restr_names) ) == NULL )
	{
	    m_abort (function_name, "array of name pointers");
	}
	if (w->tracePopup.restr_values != NULL)
	    m_free (w->tracePopup.restr_values);
	if ( ( w->tracePopup.restr_values =
	       m_alloc (num_dim * sizeof *w->tracePopup.restr_values) )
	     == NULL )
	{
	    m_abort (function_name, "array of values");
	}
	w->tracePopup.num_restr_allocated = num_dim;
    }
    /*  Compute restrictions  */
    for (dim_count = 0; dim_count < num_dim - 1; ++dim_count)
    {
	unsigned int min, max;

	w->tracePopup.restr_names[dim_count] =
	    iarray_dim_name (w->tracePopup.array, dim_indices[dim_count]);
	/*  Find minimum and maximum index for this dimension across traces  */
	min = max = 0;
	--min;
	for (trace_count = 0; trace_count < num_traces; ++trace_count)
	{
	    unsigned int index;

	    index = coord_indices[trace_count * (num_dim - 1) + dim_count];
	    if (index < min) min = index;
	    if (index > max) max = index;
	}
	w->tracePopup.restr_values[dim_count] =
	    iarray_get_coordinate (w->tracePopup.array, dim_indices[dim_count],
				   (max + min) >> 1);
    }
    value_name = iarray_value_name (w->tracePopup.array);
    if ( sum && (num_traces > 1) && (strcmp (value_name, "JY/PIX") == 0) )
	value_name = "JY";
    else if ( sum && (num_traces >1) &&
	      (strcmp (value_name, "JY/BEAM") == 0) && (num_dim == 3) )
    {
	flag beam = TRUE;
	flag f1, f2;
	unsigned int axis;
	double bmin, bmaj, cdelt1, cdelt2;
	char txt[STRING_LENGTH];

	/*  Make sure the beam area information is available and that each of
	    the two axes is a sky co-ordinate  */
	if ( !iarray_get_double (w->tracePopup.array, "BMIN", &bmin, FALSE) )
	    beam = FALSE;
	if ( !iarray_get_double (w->tracePopup.array, "BMAJ", &bmaj, FALSE) )
	    beam = FALSE;
	if ( ( axis = iarray_get_fits_axis (w->tracePopup.array,
					    dim_indices[0]) ) == 0 )
	    beam = FALSE;
	else
	{
	    f1 = wcs_astro_test_if_lon ( w->tracePopup.ap,
					 iarray_dim_name (w->tracePopup.array,
							  dim_indices[0]) );
	    f2 = wcs_astro_test_if_lat ( w->tracePopup.ap,
					 iarray_dim_name (w->tracePopup.array,
							  dim_indices[0]) );
	    if (!f1 && !f2) beam = FALSE;
	    sprintf (txt, "CDELT%u", axis);
	    if ( !iarray_get_double (w->tracePopup.array, txt, &cdelt1,
				     FALSE) ) beam = FALSE;
	}
	if ( ( axis = iarray_get_fits_axis (w->tracePopup.array,
					    dim_indices[1]) ) == 0 )
	    beam = FALSE;
	else
	{
	    f1 = wcs_astro_test_if_lon ( w->tracePopup.ap,
					 iarray_dim_name (w->tracePopup.array,
							  dim_indices[1]) );
	    f2 = wcs_astro_test_if_lat ( w->tracePopup.ap,
					 iarray_dim_name (w->tracePopup.array,
							  dim_indices[1]) );
	    if (!f1 && !f2) beam = FALSE;
	    sprintf (txt, "CDELT%u", axis);
	    if ( !iarray_get_double (w->tracePopup.array, txt, &cdelt2,
				     FALSE) ) beam = FALSE;
	}
	if (beam)
	{
	    /*  PI / ( 4 * ln (2) ) == 1.13309  */
	    global_weight = 1.0 / fabs ( 1.13309 * bmin * bmaj /
					 (cdelt1 * cdelt2) );
	    value_name = "JY";
	}
    }
    canvas_specify (w->tracePopup.worldCanvas, dim->name, value_name,
		    num_dim - 1, (CONST char **) w->tracePopup.restr_names,
		    w->tracePopup.restr_values);
    /*  Initialise the trace buffer  */
    for (count = 0; count < dim->length; ++count)
    {
	w->tracePopup.trace[count] = 0.0;
	w->tracePopup.weights[count] = 0.0;
    }
    iarray_get_data_scaling (w->tracePopup.array, &scale, &offset);
    /*  Loop over the traces  */
    for (trace_count = 0; trace_count < num_traces;
	 ++trace_count, coord_indices += num_dim - 1)
    {
	weight = global_weight;
	if (weights != NULL) weight *= weights[trace_count];
	/*  Compute offset to trace  */
	trace = w->tracePopup.array->data;
	for (count = 0; count < num_dim - 1; ++count)
	{
	    dim_count = dim_indices[count];
	    if (coord_indices[count] >=
		w->tracePopup.array->lengths[dim_count])
	    {
		fprintf(stderr,
			"Axis: %u coordinate: %lu not less than length: %lu\n",
			dim_count, coord_indices[count],
			w->tracePopup.array->lengths[dim_count]);
		return;
	    }
	    trace +=
		w->tracePopup.array->offsets[dim_count][coord_indices[count]];
	}
	/*  Step down a trace and add to the trace array  */
	if (type == K_FLOAT)
	{
	    for (count = 0; count < dim->length; ++count)
	    {
		val = *(float *) (trace + offsets[count]);
		if ( DOUBLE_IS_BLANK (val) ) continue;
		val = val * scale + offset;
		w->tracePopup.trace[count] += weight * val;
		w->tracePopup.weights[count] += weight;
	    }
	}
	else
	{
	    /*  Do it the hard way  */
	    for (count = 0; count < dim->length; ++count)
	    {
		if ( !ds_get_element (trace + offsets[count], type, value,
				      NULL) )
		{
		    fprintf (stderr, "Error converting value\n");
		    a_prog_bug (function_name);
		}
		if ( DOUBLE_IS_BLANK (value[0]) ) continue;
		val = value[0] * scale + offset;
		w->tracePopup.trace[count] += weight * val;
		w->tracePopup.weights[count] += weight;
	    }
	}
    }
    for (count = 0; count < dim->length; ++count)
    {
	if (w->tracePopup.weights[count] == 0.0)
	    w->tracePopup.trace[count] = TOOBIG;
	else if (!sum)
	    w->tracePopup.trace[count] /= w->tracePopup.weights[count];
    }
    if (w->tracePopup.trace_dim != trace_dim)
	w->tracePopup.x_uninitialised = TRUE;
    w->tracePopup.trace_dim = trace_dim;
    if ( w->tracePopup.x_uninitialised &&
	 (w->tracePopup.worldCanvas != NULL) )
    {
	canvas_set_attributes (w->tracePopup.worldCanvas,
			       CANVAS_ATT_LEFT_X, dim->first_coord,
			       CANVAS_ATT_RIGHT_X, dim->last_coord,
			       CANVAS_ATT_END);
	w->tracePopup.x_uninitialised = FALSE;
    }
    w->tracePopup.valid_data = TRUE;
    kwin_refresh_if_visible (w->tracePopup.pixcanvas, TRUE);
}   /*  End Function XkwTracePopupShowTraces  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupShowBoxedTrace (Widget W, CONST unsigned int *dim_indices,
				  flag sum,
				  CONST uaddr *start_indices,
				  CONST uaddr *stop_indices)
/*  [SUMMARY] Show a boxed trace.
    <W> The widget.
    <dim_indices> The dimension indicies. The length of this array must be one
    less than the number of dimensions in the array.
    <sum> If TRUE the traces are summed, else they are averaged.
    <start_indices> The starting coordinate indices of the box. The length of
    this array must be one less than the number of dimensions in the array.
    <stop_indices> The stopping coordinate indices of the box. The length of
    this array must be one less than the number of dimensions in the array.
    [RETURNS] Nothing.
*/
{
    flag keep_going;
    int count;
    unsigned int ndim, num_traces, trace_count;
    TracePopupWidget w = (TracePopupWidget) W;
    uaddr *indices, *counters;
    static char function_name[] = "XkwTracePopupShowBoxedTrace";

    if (w->tracePopup.worldCanvas == NULL) return;
    if (w->tracePopup.array == NULL) return;
    ndim = iarray_num_dim (w->tracePopup.array) - 1;
    /*  Ensure co-ordinate indices array is large enough  */
    for (count = 0, num_traces = 1; count < ndim; ++count)
    {
	num_traces *= stop_indices[count] - start_indices[count];
    }
    if (num_traces * ndim > w->tracePopup.coord_indices_buf_len)
    {
	if (w->tracePopup.coord_indices != NULL)
	    m_free (w->tracePopup.coord_indices);
	if ( ( w->tracePopup.coord_indices =
	       m_alloc (sizeof *w->tracePopup.coord_indices * ndim *
			num_traces) ) == NULL )
	{
	    m_error_notify (function_name, "co-ordinate indices buffer");
	    return;
	}
	w->tracePopup.coord_indices_buf_len = num_traces * ndim;
    }
    /*  Ensure co-ordinate counter array is large enough  */
    if (w->tracePopup.coord_counters_buf_len < ndim)
    {
	if (w->tracePopup.coord_counters != NULL)
	    m_free (w->tracePopup.coord_counters);
	if ( ( w->tracePopup.coord_counters =
	       m_alloc (sizeof *w->tracePopup.coord_counters *ndim) ) == NULL )
	{
	    m_error_notify (function_name, "co-ordinate counters buffer");
	    return;
	}
	w->tracePopup.coord_counters_buf_len = ndim;
    }
    counters = w->tracePopup.coord_counters;
    /*  Initialise co-ordinate counters  */
    for (count = 0; count < ndim; ++count)
	counters[count] = start_indices[count];
    /*  Loop over co-ordinates  */
    keep_going = TRUE;
    trace_count = 0;
    for (indices = w->tracePopup.coord_indices; keep_going;
	 indices += ndim, ++trace_count)
    {
	/*  Copy co-ordinate counter contents to indices array  */
	for (count = 0; count <ndim; ++count) indices[count] = counters[count];
	/*  Now increment counter to next co-ordinate  */
	keep_going = TRUE;
	for (count = ndim - 1; keep_going && (count >= 0); )
	{
	    if (++counters[count] < stop_indices[count]) keep_going = FALSE;
	    else
	    {
		counters[count] = start_indices[count];
		--count;
	    }
	}
	if (count < 0) keep_going = FALSE;
	else keep_going = TRUE;
    }
    if (trace_count != num_traces)
    {
	fprintf (stderr, "trace_count: %u is not num_traces: %u\n",
		 trace_count, num_traces);
	a_prog_bug (function_name);
    }
    XkwTracePopupShowTraces (W, dim_indices, sum, 1.0,
			     num_traces, w->tracePopup.coord_indices, NULL);
}   /*  End Function XkwTracePopupShowBoxedTrace  */

/*EXPERIMENTAL_FUNCTION*/
void XkwTracePopupInvalidate (Widget W)
/*  [SUMMARY] Invalidate trace data.
    <W> The widget.
    [RETURNS] Nothing.
*/
{
    TracePopupWidget w = (TracePopupWidget) W;
    char *xlabel;

    if (!w->tracePopup.valid_data) return;
    w->tracePopup.valid_data = FALSE;
    canvas_get_specification (w->tracePopup.worldCanvas, &xlabel, NULL,
			      NULL, NULL, NULL);
    canvas_specify (w->tracePopup.worldCanvas, xlabel, "NO VALID DATA",
		    0, NULL, NULL);
    kwin_refresh_if_visible (w->tracePopup.pixcanvas, TRUE);
}   /*  End Function XkwTracePopupInvalidate  */
