/*LINTLIBRARY*/
/*  View2Datasets.c

    This code provides a twin dataset view control widget for Xt.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a twin dataset view control
  widget for Xt.


    Written by      Richard Gooch   22-JAN-1997

    Updated by      Richard Gooch   29-JAN-1997: Registered PseudoColour arrays
  with Dataclip widgets and implemented <region_cbk>.

    Updated by      Richard Gooch   30-JAN-1997: Registered change in data for
  PseudoColour arrays with Dataclip widgets.

    Updated by      Richard Gooch   1-FEB-1997: Added profile direction menu
  and removed management of KwcsAstro object.

    Updated by      Richard Gooch   2-FEB-1997: Added display of dataset
  filename and object name.

    Updated by      Richard Gooch   3-FEB-1997: Added profiling of 2D datasets:
  limit profile axes to X and Y.

    Updated by      Richard Gooch   6-FEB-1997: Added "Auto Apply" toggle for
  contour levels.

    Updated by      Richard Gooch   15-FEB-1997: Now set canvas dressing title.

    Updated by      Richard Gooch   25-FEB-1997: Use reset intensity toggle in
  Zoom Policy window to control intensity range when loading a dataset.

    Updated by      Richard Gooch   3-MAR-1997: Deal with colourmap associated
  with dataset.

    Updated by      Richard Gooch   16-MAR-1997: Added beam overlay display.

    Updated by      Richard Gooch   22-MAR-1997: Fixed beam overlay display.

    Updated by      Richard Gooch   26-MAR-1997: Manipulate sensitivities of
  Zoom Menu entries. Track display now also updates when dragging zoom box.

    Updated by      Richard Gooch   15-APR-1997: Changed names of exclusive
  menu widgets.

    Updated by      Richard Gooch   18-APR-1997: Created XkwNcursorCallback
  callback.

    Updated by      Richard Gooch   24-APR-1997: No longer update profile in
  <new_frame_cbk> function.

    Updated by      Richard Gooch   14-MAY-1997: Made contour dialog wider.

    Updated by      Richard Gooch   18-MAY-1997: Support display of
  PseudoColour data on TrueColour canvases.

    Updated by      Richard Gooch   20-MAY-1997: Fixed bug in beam display:
  was 90 degrees out.

    Updated by      Richard Gooch   26-MAY-1997: Fixed bug in beam display:
  was twice as large.

    Updated by      Richard Gooch   4-JUN-1997: Fixed another bug in beam
  display: did not apply cos (declination) correction. Pass non-float arrays to
  TracePopup widget now that it supports non-float arrays.

    Updated by      Richard Gooch   5-JUN-1997: Use middle and right mouse
  button releases to zoom in 2x and zoom out 2x.

    Updated by      Richard Gooch   7-JUN-1997: Added toggle for showing beam
  name. Added dialog and toggle for changing axis labelling title.

    Updated by      Richard Gooch   19-JUN-1997: Update colourmap for 2D
  PseudoColour arrays.

    Updated by      Richard Gooch   24-JUN-1997: Fixed bug when auto title was
  disabled: contours were not shown!

    Updated by      Richard Gooch   26-JUN-1997: Display units next to contour
  levels.

    Updated by      Richard Gooch   1-AUG-1997: Added profile mode menu and
  implemented box profiles.

    Updated by      Richard Gooch   3-AUG-1997: Drop middle and right button
  release events (zoom in and out) if previous event wasn't corresponding click

    Updated by      Richard Gooch   4-AUG-1997: Made use of middle and right
  mouse zoom toggles in zoom policy widget.

    Updated by      Richard Gooch   6-AUG-1997: Fixed bug in <geom_func> where
  dimension indices and co-ordinates were not consistent.

    Updated by      Richard Gooch   7-AUG-1997: Added radial profile.

    Updated by      Richard Gooch   10-AUG-1997: Create TracePopup widget on
  demand rather than at initialise depending on XkwNcreateAtInit resource.

    Updated by      Richard Gooch   12-AUG-1997: Now make call to <ds_fitgauss>
  for radial profiles. Always set intensity scale if canvas range is 0.

    Updated by      Richard Gooch   19-AUG-1997: Set axis names for radial
  profile to something useful.

    Updated by      Richard Gooch   20-AUG-1997: Made use of new functions for
  testing if axes are longitude or latitude. Moved code to convert ellipse in
  world co-ordinates -> pixel co-ordinates to <canvas> package.

    Updated by      Richard Gooch   17-SEP-1997: Account for data scaling in
  radial profiles.

    Updated by      Richard Gooch   18-SEP-1997: Account for data scaling in
  absolute contour levels.

    Updated by      Richard Gooch   11-OCT-1997: Added dashing support.

    Updated by      Richard Gooch   21-OCT-1997: Added boxed horizontal and
  vertical profiles and gaussian fitting.

    Updated by      Richard Gooch   28-OCT-1997: Added background estimation
  for boxed horizontal and vertical profiles. Limit weighted mean estimation of
  centroid to quarter intensity points.

    Updated by      Richard Gooch   31-OCT-1997: Made use of new
  <<reset_hscale>> parameter to <XkwDataclipNewArray>.

    Updated by      Richard Gooch   19-NOV-1997: Fixed bug in
  <<apply_levels_cbk>> which fouled up level strings like "-0.2 -0.1" and
  "0.1 .2"

    Updated by      Richard Gooch   1-DEC-1997: Took account of change to
  <ds_find_centroid>.

    Updated by      Richard Gooch   7-DEC-1997: Added panner support and
  removed control of magnifier position (now in MagnifierPopup widget). Added
  linewidth controls for contours. Added intensity link control.

    Updated by      Richard Gooch   8-DEC-1997: Took account of creation of new
  IntensityPolicy widget.

    Updated by      Richard Gooch   11-DEC-1997: Display value for contour
  width sliders.

    Updated by      Richard Gooch   30-DEC-1997: Added trap for radius < 2 in
  <geom_func> for radial profile.

    Updated by      Richard Gooch   1-JAN-1998: Added XkwNwmtitle resource.

    Updated by      Richard Gooch   6-FEB-1998: Took account of ellipse drawing
  changes.

    Updated by      Richard Gooch   11-FEB-1998: Clear magnifier on reload.

    Updated by      Richard Gooch   12-FEB-1998: Bugfix when loading set 1
  before set 0.

    Updated by      Richard Gooch   13-FEB-1998: Refresh magnifier in
  <region_cbk>.

    Updated by      Richard Gooch   23-FEB-1998: Fixed support for editor
  contexts when multiple visuals available.

    Updated by      Richard Gooch   2-MAR-1998: Fixed bug in <update_title> to
  prevent SEGV when no image data loaded.

    Updated by      Richard Gooch   20-MAR-1998: No longer update colourmap in
  <new_frame_cbk>.

    Updated by      Richard Gooch   13-APR-1998: Added call to <event_dispatch>
  in <XkwView2DatasetsCanvasEventHandler>.

    Updated by      Richard Gooch   10-MAY-1998: Took account of change to
  <canvas_convert_ellipse_coords> function.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   30-MAY-1998: Fixed bug in function
  <azimuthally_average>: did not add 0.5 for pixel centre offset. Changed
  <draw_beam> to limit angle to -90,90 degrees for nicer text orientation
  because <canvas_convert_ellipse_coords> no longer limits the angle.

    Updated by      Richard Gooch   1-JUN-1998: Fixed bug in function
  <XkwView2DatasetsCanvasEventHandler> where pixel string was being displayed
  prior to axis name fiddling.

    Updated by      Richard Gooch   2-JUN-1998: Added set number in WM title.
  Bind 'b' key in canvas to blink. Fixed bug so blinking changes number of
  frames in animate widget.

    Updated by      Richard Gooch   3-JUN-1998: Show more information for
  radial and boxed profiles.

    Updated by      Richard Gooch   12-AUG-1998: Trap unsupported event types
  in <XkwView2DatasetsCanvasEventHandler>.

    Updated by      Richard Gooch   17-AUG-1998: Bug fix in <geom_func> for
  computing box corners.

    Updated by      Richard Gooch   26-AUG-1998: Bug fix in <geom_func> where
  raw background value was not scaled for radial profile. Trap circles bigger
  than image.

    Updated by      Richard Gooch   5-SEP-1998: Bug fix in <display_changes>
  where RGB contours were not being make active.

    Updated by      Richard Gooch   16-SEP-1998: Made use of new key grouping
  in <contour> package.

    Updated by      Richard Gooch   17-SEP-1998: Made use of new functions
  <contour_parse_levels> and <contour_unscale_levels>.

    Updated by      Richard Gooch   20-SEP-1998: Removed use of unnecessary
  variables in <viewimg_position_func>.

    Updated by      Richard Gooch   15-OCT-1998: Made use of new function
  <canvas_warp_pointer>.

    Updated by      Richard Gooch   19-OCT-1998: Display centre of boxed
  profiles.

    Updated by      Richard Gooch   26-OCT-1998: Support resource setting of
  display mode.

    Updated by      Richard Gooch   5-APR-1999: Created <trace_position_func>.

    Updated by      Richard Gooch   13-APR-1999: Made use of new function
  <XkwAnimateControlPopdown>.

    Updated by      Richard Gooch   15-APR-1999: Store frame number in each
  direction structure rather than in widget structure.

    Updated by      Richard Gooch   17-APR-1999: Fixed bug in <next_frame_cbk>
  where no direction was available (bad pointer).

    Updated by      Richard Gooch   21-APR-1999: Trap radial profiles where the
  ellipse moves past the image boundary after centroiding.

    Updated by      Richard Gooch   5-JUL-1999: Don't update title if not
  changed.

    Updated by      Richard Gooch   7-AUG-1999: Added "Link Frames" toggle.

    Updated by      Richard Gooch   15-AUG-1999: Refresh magnifier in function
  <new_frame_cbk>.

    Updated by      Richard Gooch   31-AUG-1999: No longer clip to canvas
  boundaries in <geom_func>. Took account of interface change to function
  <viewimg_track_compute>. Added 'l' key binding in function
  <XkwView2DatasetsCanvasEventHandler>.

    Updated by      Richard Gooch   1-SEP-1999: Move link toggles above contour
  control widgets. Added radial profile mode which doesn't do a Gaussian fit.

    Updated by      Richard Gooch   2-SEP-1999: Took account of interface
  change to function <canvas_warp_pointer> and discard events.

    Updated by      Richard Gooch   11-SEP-1999: Replaced title dialog with
  ASCII text entry so that title can be changed as typed.

    Updated by      Richard Gooch   12-SEP-1999: Added <intensity_range_cbk>.

    Updated by      Richard Gooch   20-SEP-1999: Made use of SimpleTextEntry
  widget class for title entry.

    Updated by      Richard Gooch   29-SEP-1999: Took account of interface
  change to <viewimg_zoom>.

    Updated by      Richard Gooch   2-OCT-1999: Copy scaled intensity range
  rather than raw range when intensity scales are linked.

    Updated by      Richard Gooch   28-OCT-1999: Ignore blink state if only one
  dataset is loaded.

    Updated by      Richard Gooch   8-DEC-1999: Added key bindings to step
  through movie.

    Updated by      Richard Gooch   12-MAY-2000: Ignore K_CANVAS_EVENT_KEYSYM.

    Updated by      Richard Gooch   2-JUL-2000: Take contour frame number from
  contour state rather than image state in <display_changes>.

    Updated by      Richard Gooch   5-JUL-2000: Moved frame linking from
  <new_frame_cbk> to <display_changes>, so that all changes cause frame numbers
  to be copied.

    Updated by      Richard Gooch   6-JUL-2000: Link frames using real units if
  possible. Do not set number of frames to largest dataset in function
  <update_animate_widget>.

    Updated by      Richard Gooch   11-JUL-2000: Fixed bug in function
  <update_animate_widget>: do not use contour direction if contour set is NULL.
  Added #include <karma_kcmap.h>.

    Updated by      Richard Gooch   12-JUL-2000: Added
  #include <karma_viewimg.h> and <karma_contour.h>

    Updated by      Richard Gooch   25-NOV-2000: Fixed bug so that "OBJECT"
  value is used and displayed even if "filename" object is not associated with
  multi_array descriptor. Display "OBJECT" value for window title in preference
  to "filename" object.

    Updated by      Richard Gooch   26-NOV-2000: Reverted previous change to
  display "filename" in preference to object "OBJECT" value for window title.
  Steve prefers the filename, for files which are alternate versions of the
  same object.

    Updated by      Richard Gooch   22-FEB-2001: Do nothing when "Home", "PgUp"
  etc. keys are pressed if number of frames in animate control widget is < 2.

    Updated by      Richard Gooch   26-APR-2001: Fixed bug in <display_changes>
  where visual type was used to select contour images. Now compare canvases.

    Updated by      Richard Gooch   7-MAY-2001: Added -ve contour colour
  control. Always pad filename and object name labels with blanks to ensure
  they are not auto-resized too small.

    Updated by      Richard Gooch   28-MAY-2001: Added function
  <save_ppm_movie_cbk>.

    Updated by      Richard Gooch   31-MAY-2001: Finished adding function
  <compute_radial_points>.

    Updated by      Richard Gooch   15-JUN-2001: Disable changing of direction
  menu when box sum or box average profiles are selected, and change direction.

    Last updated by Richard Gooch   18-JUN-2001: Fixed bug in <geom_func>:
  clip co-ordinates to canvas boundary for box profiles and switch to
  <ds_get_coord_num>.


*/
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_contour.h>
#include <karma_graph1d.h>
#include <karma_xtmisc.h>
#include <karma_event.h>
#include <karma_kcmap.h>
#include <karma_wcs.h>
#include <karma_ex.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
/*#define USE_TRANSIENT*/
#include <Xkw/View2DatasetsP.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/TracePopup.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/ChoiceMenu.h>
#include <Xkw/SlaveImageDisplayPopup.h>
#include <Xkw/IntensityPolicy.h>
#include <Xkw/SimpleTextEntry.h>
#include <Xkw/ExportMenu.h>


#define MAX_CONTOUR_LEVELS 100
#define MAX_HISTOGRAM_BINS 256
#define ELLIPSE_RADIUS     5
#define GAUSSCURVE_LENGTH  512

struct edit_context_list
{
    KPixCanvasEditorContext context;
    struct edit_context_list *next;
};


/*  Private functions  */
STATIC_FUNCTION (void _View2Datasets_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void free_viewables, (struct direction_type *info) );
STATIC_FUNCTION (void slice_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void recompute_profile_dir, (View2DatasetsWidget w) );
STATIC_FUNCTION (void profile_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void profile_dir_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _View2Datasets_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _View2Datasets_display_changes,
		 (View2DatasetsWidget top, flag update_profile) );
STATIC_FUNCTION (void _View2Datasets_update_title, (View2DatasetsWidget top) );
STATIC_FUNCTION (void _View2Datasets_update_cmap, (View2DatasetsWidget top) );
STATIC_FUNCTION (void invalidate_planes, (struct direction_type *info) );
STATIC_FUNCTION (void update_animate_widget, (View2DatasetsWidget top) );
STATIC_FUNCTION (void update_profile_axis,
		 (View2DatasetsWidget top, struct dataset_type *set) );
STATIC_FUNCTION (void new_frame_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void blink_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void iarr_destroy_callback,
		 (iarray arr, struct dataset_type *set) );
STATIC_FUNCTION (void set_iscale_range,
		 (struct dataset_type *set, double min, double max,
		  flag scaled) );
STATIC_FUNCTION (flag setup_2D_pseudo,
		 (View2DatasetsWidget w, struct dataset_type *set,
		  iarray array, double lower_clip, double upper_clip,
		  KWorldCanvas pseudo_canvas, KWorldCanvas rgb_canvas,
		  KWorldCanvas mag_pseudo_canvas, KWorldCanvas mag_rgb_canvas,
		  KWorldCanvas pan_pseudo_canvas,
		  KWorldCanvas pan_rgb_canvas) );
STATIC_FUNCTION (flag setup_3D_pseudo,
		 (View2DatasetsWidget w, struct dataset_type *set,
		  iarray array, double lower_clip, double upper_clip,
		  KWorldCanvas pseudo_canvas, KWorldCanvas rgb_canvas,
		  KWorldCanvas mag_pseudo_canvas, KWorldCanvas mag_rgb_canvas,
		  KWorldCanvas pan_pseudo_canvas,
		  KWorldCanvas pan_rgb_canvas) );
STATIC_FUNCTION (flag setup_2D_rgb,
		 (View2DatasetsWidget w, struct dataset_type *set,
		  iarray red, iarray green, iarray blue,
		  KWorldCanvas canvas, KWorldCanvas mag_canvas,
		  KWorldCanvas pan_canvas) );
STATIC_FUNCTION (flag setup_3D_rgb,
		 (View2DatasetsWidget w, struct dataset_type *set,
		  iarray red, iarray green, iarray blue,
		  KWorldCanvas canvas, KWorldCanvas mag_canvas,
		  KWorldCanvas pan_canvas) );
STATIC_FUNCTION (void apply_levels_cbk, (Widget w, XtPointer client_data,
					 XtPointer call_data) );
STATIC_FUNCTION (void _View2Datasets_set_clevels,
		 (KContourImage cimage, unsigned int num_levels,
		  CONST double *contour_levels, CONST flag *dash,
		  CONST double *linewidths,
		  CONST char *negative_colour, CONST char *positive_colour) );
STATIC_FUNCTION (void reset_levels_cbk, (Widget w, XtPointer client_data,
					 XtPointer call_data) );
STATIC_FUNCTION (void auto_apply_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void region_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _View2Datasets_intensity_range_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void display_changes_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void _View2Datasets_text_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void beam_toggle_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (unsigned long get_green, (KPixCanvas canvas) );
STATIC_FUNCTION (void all_realised_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void refresh_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void slider_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void canvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (flag _View2Datasets_viewimg_position_func,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin,
		  unsigned int value_type) );
STATIC_FUNCTION (void get_sets, (View2DatasetsWidget top,
				 struct dataset_type **image_set,
				 struct dataset_type **contour_set,
				 struct dataset_type **hidden_set,
				 struct dataset_type **other_set) );
STATIC_FUNCTION (struct direction_type *get_dir,
		 (View2DatasetsWidget top, struct dataset_type *set) );
STATIC_FUNCTION (void draw_beam,
		 (KWorldCanvas canvas, struct dataset_type *set, flag name) );
STATIC_FUNCTION (flag _View2Datasets_geom_func,
		 (KPixCanvasEditorContext context, KPixCanvas canvas,
		  unsigned int object_type, unsigned int num_coords,
		  CONST int *x, CONST int *y,
		  unsigned int raw_event_code, void *raw_e_info,
		  void **f_info) );
STATIC_FUNCTION (void _View2Datasets_compute_fit_data,
		 (View2DatasetsWidget top,double centre, double peak_height,
		  double background, flag inverted) );
STATIC_FUNCTION (void compute_radial_points,
		 (View2DatasetsWidget top, CONST char *in,unsigned int in_type,
		  uaddr xlen, CONST uaddr *xoffsets,
		  signed int xcentre, signed int xradius,
		  uaddr ylen, CONST uaddr *yoffsets,
		  signed int ycentre, signed int yradius,
		  float scale, float offset) );
STATIC_FUNCTION (void azimuthally_average,
		 (float *out, uaddr out_len, CONST uaddr *ooffsets,
		  CONST char *in, unsigned int in_type,
		  uaddr xlen, CONST uaddr *xoffsets,
		  double xcentre, double xradius,
		  uaddr ylen, CONST uaddr *yoffsets,
		  double ycentre, double yradius, double rotation,
		  float scale, float offset) );
STATIC_FUNCTION (double compute_magnitude,
		 (CONST char *in, unsigned int in_type,
		  uaddr xlen, CONST uaddr *xoffsets,
		  double xcentre, double xradius,
		  uaddr ylen, CONST uaddr *yoffsets,
		  double ycentre, double yradius,
		  double scale, double offset,
		  double background, flag inverted) );
STATIC_FUNCTION (flag collapse_dim,
		 (float *out, CONST uaddr *ooffsets,
		  CONST char *in, unsigned int in_type,
		  uaddr start_x, uaddr end_x, CONST uaddr *xoffsets,
		  uaddr start_y, uaddr end_y, CONST uaddr *yoffsets,
		  float scale, float offset) );
STATIC_FUNCTION (void create_tracewinpopup, (View2DatasetsWidget top) );
STATIC_FUNCTION (void trace_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (flag trace_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin,
		  int x_pix, int y_pix) );
STATIC_FUNCTION (void _View2Datasets_save_ppm_movie_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffsetOf(View2DatasetsRec, view2Datasets.field)

static XtResource View2DatasetsResources[] = 
{
    {XkwNcursorCallback, XtCCallback, XtRCallback, sizeof (XtPointer),
     offset (cursorCallback), XtRCallback, (XtPointer) NULL},
    {XkwNcreateAtInit, XkwCCreateAtInit, XtRBool, sizeof (Bool),
     offset (createAtInit), XtRImmediate, (XtPointer) FALSE},
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, False},
    {XkwNwmtitle, XkwCWMtitle, XtRString, sizeof (String),
     offset (wmtitle), XtRImmediate, NULL},
};

#undef offset

#define SLICE_XY 0
#define SLICE_XZ 1
#define SLICE_ZY 2
#define NUM_SLICE_CHOICES 3

static char *slice_choices[NUM_SLICE_CHOICES] =
{
    "XY", "XZ", "ZY",
};

#define PROFILE_DIR_Z 0
#define PROFILE_DIR_Y 1
#define PROFILE_DIR_X 2
#define NUM_PROFILE_DIR_CHOICES 3

static char *profile_dir_choices[NUM_PROFILE_DIR_CHOICES] =
{
    "Z", "Y", "X",
};

#define MODE_IMAGE0          0
#define MODE_IMAGE1          1
#define MODE_IMAGE0_CONTOUR1 2
#define MODE_IMAGE1_CONTOUR0 3
#define MODE_BLINK           4
#define NUM_MODE_CHOICES     5

static char *mode_choices[NUM_MODE_CHOICES] =
{
    "Set 1->image",
    "Set 2->image",
    "Set 1->image, Set 2->contour",
    "Set 1->contour, Set 2->image",
    "Blink between Set 1 and Set 2",
};

#define PROFILE_MODE_NONE               0
#define PROFILE_MODE_LINE               1
#define PROFILE_MODE_BOX_SUM            2
#define PROFILE_MODE_BOX_AVERAGE        3
#define PROFILE_MODE_RADIAL             4
#define PROFILE_MODE_RADIAL_GAUSS       5
#define PROFILE_MODE_BOXED_HORIZONTAL   6
#define PROFILE_MODE_BOXED_VERTICAL     7
#define NUM_PROFILE_MODE_CHOICES        8

static char *profile_mode_choices[NUM_PROFILE_MODE_CHOICES] =
{
    "None", "Line", "Box Sum", "Box Average", "Radial", "Radial+Fit",
    "Boxed Horiz", "Boxed Vert",
};

View2DatasetsClassRec view2DatasetsClassRec = 
{
  {
      /* CoreClassPart */
#ifdef USE_TRANSIENT
    (WidgetClass) &transientShellClassRec,       /* superclass */
#else
    (WidgetClass) &topLevelShellClassRec,       /* superclass */
#endif
    "View2Datasets",                            /* class_name */
    sizeof (View2DatasetsRec),                  /* widget_size */
    NULL,                                       /* class_initialise */
    NULL,                                       /* class_part_initialise */
    FALSE,                                      /* class_init */
    (XtInitProc) _View2Datasets_Initialise,     /* initialise */
    NULL,                                       /* initialise_hook */
    XtInheritRealize,                           /* realise */
    NULL,                                       /* actions */
    0,                                          /* num_actions */
    View2DatasetsResources,                     /* resources */
    XtNumber (View2DatasetsResources),          /* num_resources */
    NULLQUARK,                                  /* xrm_class */
    TRUE,                                       /* compress_motion */
    TRUE,                                       /* compress_exposure */
    TRUE,                                       /* compress_enterleave */
    TRUE,                                       /* visible_interest */
    NULL,                                       /* destroy */
    XtInheritResize,                            /* resize */
    NULL,                                       /* expose */
    NULL,                                       /* set_values */
    NULL,                                       /* set_values_hook */
    XtInheritSetValuesAlmost,                   /* set_values_almost */
    NULL,                                       /* get_values_hook */
    NULL,                                       /* accept_focus */
    XtVersion,                                  /* version */
    NULL,                                       /* callback_private */
    NULL,                                       /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },
  { /* Shell */
    /* extension	  */	NULL
  },
  { /* WMShell */
    /* extension	  */	NULL
  },
  { /* VendorShell */
    /* extension	  */	NULL
  },
#ifdef USE_TRANSIENT
  { /* TransientShell */
    /* extension	  */	NULL
  },
#else
  { /* TopLevelShell */
    /* extension	  */	NULL
  },
#endif
  {  /* View2DatasetsClassPart */
    0 /* empty */
  }
};

WidgetClass view2DatasetsWidgetClass = (WidgetClass) &view2DatasetsClassRec;

static void _View2Datasets_Initialise (Widget Request, Widget New)
{
    int num_items, tmp;
    /*View2DatasetsWidget request = (View2DatasetsWidget) Request;*/
    View2DatasetsWidget new = (View2DatasetsWidget) New;
    Widget form, parent, w, upper_widget, movie_btn, zoom_policy;
    static char function_name[] = "View2DatasetsWidget::Initialise";

    /*  Search for ImageDisplay widget somewhere in my ancestory  */
    new->view2Datasets.image_display = NULL;
    new->view2Datasets.auto_apply_state = FALSE;
    new->view2Datasets.slice = SLICE_XY;
    new->view2Datasets.mode = MODE_IMAGE0;
    new->view2Datasets.profile_dir = PROFILE_DIR_Z;
    new->view2Datasets.profile_mode = PROFILE_MODE_NONE;
    new->view2Datasets.profile_arr = NULL;
    m_clear (&new->view2Datasets.profile_points,
	     sizeof new->view2Datasets.profile_points);
    new->view2Datasets.gausscurve_xarr = NULL;
    new->view2Datasets.gausscurve_yarr = NULL;
    new->view2Datasets.trace_winpopup = NULL;
    new->view2Datasets.first_edit_context = NULL;
    new->view2Datasets.old_title[0] = '\0';
    m_clear (&new->view2Datasets.set0, sizeof new->view2Datasets.set0);
    m_clear (&new->view2Datasets.set1, sizeof new->view2Datasets.set1);
    new->view2Datasets.set0.beam_xpos = 0.5;
    new->view2Datasets.set0.beam_ypos = 0.5;
    new->view2Datasets.set1.beam_xpos = 0.5;
    new->view2Datasets.set1.beam_ypos = 0.5;
    strcpy (new->view2Datasets.set0.levels, "%10:90+10");
    strcpy (new->view2Datasets.set1.levels, "%10:90+10");
    strcpy (new->view2Datasets.set0.negative_contour_colour, "grey70");
    strcpy (new->view2Datasets.set0.positive_contour_colour, "green");
    strcpy (new->view2Datasets.set1.negative_contour_colour, "grey70");
    strcpy (new->view2Datasets.set1.positive_contour_colour, "green");
    /*  Get the ancestor ImageDisplay widget  */
    for ( parent = XtParent (New); parent != NULL; parent = XtParent (parent) )
    {
	if ( XtIsImageDisplay (parent) )
	    new->view2Datasets.image_display = parent;
    }
    if (new->view2Datasets.image_display == NULL)
    {
	fprintf (stderr, "Could not find an ImageDisplay ancestor\n");
	a_prog_bug (function_name);
    }
    /*  Get the ancestor shell  */
    for (parent = new->view2Datasets.image_display; parent != NULL;
	 parent = XtParent (parent) )
    {
	if ( XtIsWMShell (parent) ) new->view2Datasets.wmshell = parent;
    }
    if (new->view2Datasets.wmshell == NULL)
    {
	fprintf (stderr, "Could not find a WMShell ancestor\n");
	a_prog_bug (function_name);
    }
    /*  Get the intensity menu  */
    w = XtNameToWidget (new->view2Datasets.image_display,
			"intensityMenuButton");
    if (w == NULL)
    {
	fprintf (stderr,
		 "No Intensity Menu widget in parent ImageDisplay widget!\n");
	a_prog_bug (function_name);
    }
    new->view2Datasets.intensity_menu = w;
    XtVaGetValues (w, XkwNnumItems, &num_items, NULL);
    if (num_items < 2)
    {
	fprintf (stderr, "Intensity Menu widget has only %d entries\n",
		 num_items);
	a_prog_bug (function_name);
    }
    new->view2Datasets.first_intensity = num_items - 2;
    /*  Get the ZoomPolicy widget  */
    if ( ( zoom_policy = XtNameToWidget (new->view2Datasets.image_display,
					 "zoomPolicyPopup") ) == NULL )
    {
	fprintf (stderr, "Could not find ZoomPolicy popup\n");
	a_prog_bug (function_name);
    }
    if ( ( w = XtNameToWidget (zoom_policy,"form.middleZoomToggle") ) == NULL )
    {
	fprintf (stderr, "Could not find middleZoomToggle\n");
	a_prog_bug (function_name);
    }
    new->view2Datasets.middle_zoom_tgl = w;
    if ( ( w = XtNameToWidget (zoom_policy, "form.rightZoomToggle") ) == NULL )
    {
	fprintf (stderr, "Could not find rightZoomToggle\n");
	a_prog_bug (function_name);
    }
    new->view2Datasets.right_zoom_tgl = w;
    XkwView2DatasetsRegisterDataset (New, NULL, 0.0, 0.0, NULL, NULL, NULL, 0);
    XkwView2DatasetsRegisterDataset (New, NULL, 0.0, 0.0, NULL, NULL, NULL, 1);
    XtAddCallback (new->view2Datasets.image_display, XkwNallRealisedCallback,
		   all_realised_cbk, new);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("movieButton", commandWidgetClass, form,
				 XtNlabel, "Movie",
				 XtNfromHoriz, w,
				 XtNheight, 20,
				 NULL);
    movie_btn = w;
    w = XtVaCreateManagedWidget ("blinkToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Blink",
				 XkwNcrosses, False,
				 XtNfromHoriz, w,
				 NULL);
    new->view2Datasets.blink_tgl = w;
    XtAddCallback (w, XtNcallback, blink_cbk, New);
    w = XtVaCreateManagedWidget ("profileModeMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "profileModeMenu",
				 XkwNchoiceName, "Profile Mode",
				 XtNfromHoriz, w,
				 XkwNnumItems, NUM_PROFILE_MODE_CHOICES,
				 XkwNitemStrings, profile_mode_choices,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, profile_mode_cbk, New);
    w = XtVaCreateManagedWidget ("sliceMenu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "sliceMenu",
				 XkwNchoiceName, "Slice Direction",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_SLICE_CHOICES,
				 XkwNitemStrings, slice_choices,
				 NULL);
    XtAddCallback (w, XkwNselectCallback, slice_cbk, New);
    w = XtVaCreateManagedWidget ("profileDirMenu", exclusiveMenuWidgetClass,
				 form,
				 XtNmenuName, "profileDirMenu",
				 XkwNchoiceName, "Profile Axis",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNnumItems, NUM_PROFILE_DIR_CHOICES,
				 XkwNitemStrings, profile_dir_choices,
				 NULL);
    upper_widget = w;
    new->view2Datasets.profile_dir_menu = w;
    XtAddCallback (w, XkwNselectCallback, profile_dir_cbk, New);
    w = XtVaCreateManagedWidget ("modeMenu", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "modeMenu",
				 XkwNchoiceName, "Display Mode",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_MODE_CHOICES,
				 XkwNitemStrings, mode_choices,
				 XkwNvaluePtr, &tmp,
				 NULL);
    upper_widget = w;
    new->view2Datasets.mode_menu = w;
    XtAddCallback (w, XkwNselectCallback, _View2Datasets_mode_cbk, New);
    XtVaSetValues (w, XkwNvaluePtr, NULL, NULL);
    new->view2Datasets.mode = tmp;
    w = XtVaCreatePopupShell ("animatepopup", animateControlWidgetClass, New,
			      XtNtitle, "Animation Control",
			      NULL);
    new->view2Datasets.animatepopup = w;
    XtAddCallback (w, XkwNnewFrameCallback, new_frame_cbk, New);
    XtAddCallback (movie_btn, XtNcallback, xtmisc_popup_cbk, w);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Link Intensity Scales",
				 XkwNcrosses, False,
				 XtNfromVert, upper_widget,
				 NULL);
    new->view2Datasets.link_iscale_tgl = w;
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Link Frames",
				 XkwNcrosses, False,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, display_changes_cbk, New);
    new->view2Datasets.link_frames_tgl = w;
    upper_widget = w;
    w = XtVaCreateManagedWidget ("contourLevels", dialogWidgetClass, form,
				 XtNlabel,
				 "Contour Levels                             ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNvalue,
				 "%10:90+10                                  ",
				 NULL);
    upper_widget = w;
    new->view2Datasets.levels_dlg = w;
    w = XtVaCreateManagedWidget ("contourColour", dialogWidgetClass, form,
				 XtNlabel, "-ve Contour Colour     ",
				 XtNborderWidth, 0,
				 XtNfromVert, w,
				 XtNvalue, "grey70",
				 NULL);
    new->view2Datasets.negative_colour_dlg = w;
    w = XtVaCreateManagedWidget ("contourColour", dialogWidgetClass, form,
				 XtNlabel, "+ve Contour Colour    ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNvalue, "green",
				 NULL);
    upper_widget = w;
    new->view2Datasets.positive_colour_dlg = w;
    w = XtVaCreateManagedWidget ("applyButton", commandWidgetClass, form,
				 XtNlabel, "Apply Levels",
				 XtNfromVert, upper_widget,
				 NULL);
    XtAddCallback (w, XtNcallback, apply_levels_cbk, New);
    w = XtVaCreateManagedWidget ("resetButton", commandWidgetClass, form,
				 XtNlabel, "Reset Levels",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, reset_levels_cbk, New);
    w = XtVaCreateManagedWidget ("autoToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Apply",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, auto_apply_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNlabel, "-ve Contour Width",
				 XtNwidth, 220,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNscaledValuePtr,
				 &new->view2Datasets.contour_neg_width,
				 XkwNscaledFormat, "%.1f",
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-1, 0.0, FALSE, TRUE);
    XtAddCallback (w, XkwNvalueChangeCallback, apply_levels_cbk, New);
    XtAddCallback (w, XkwNrealiseCallback, slider_realise_cbk, New);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "+ve Contour Width",
				 XtNwidth, 220,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNscaledValuePtr,
				 &new->view2Datasets.contour_pos_width,
				 XkwNscaledFormat, "%.1f",
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-1, 0.0, FALSE, TRUE);
    XtAddCallback (w, XkwNvalueChangeCallback, apply_levels_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNlabel,
				 "Set 1 Filename:                            ",
				 XtNfromVert, upper_widget,
				 XtNjustify, XtJustifyLeft,
				 NULL);
    upper_widget = w;
    new->view2Datasets.set0.filename_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNlabel,
				 "Set 1 Object:                              ",
				 XtNfromVert, upper_widget,
				 XtNjustify, XtJustifyLeft,
				 NULL);
    upper_widget = w;
    new->view2Datasets.set0.object_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNlabel,
				 "Set 2 Filename:                            ",
				 XtNfromVert, upper_widget,
				 XtNjustify, XtJustifyLeft,
				 NULL);
    upper_widget = w;
    new->view2Datasets.set1.filename_label = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNlabel,
				 "Set 2 Object:                              ",
				 XtNfromVert, upper_widget,
				 XtNjustify, XtJustifyLeft,
				 NULL);
    upper_widget = w;
    new->view2Datasets.set1.object_label = w;
    w = XtVaCreateManagedWidget ("titleName", labelWidgetClass, form,
				 XtNlabel, "Title String:",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 NULL);
    w = XtVaCreateManagedWidget ("titleToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Title",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNhorizDistance, 200,
				 XtNright, XtChainRight,
				 XkwNcrosses, False,
				 XtNstate, TRUE,
				 NULL);
    XtAddCallback (w, XtNcallback, display_changes_cbk, New);
    upper_widget = w;
    new->view2Datasets.auto_title_tgl = w;
    w = XtVaCreateManagedWidget ("title", simpleTextEntryWidgetClass, form,
				 XtNdefaultDistance, 0,
				 XtNfromVert, upper_widget,
				 XtNwidth, 440,
				 XtNeditType, XawtextEdit,
				 NULL);
    new->view2Datasets.title_txt = w;
    upper_widget = w;
    XtAddCallback (w, XkwNnewCharacterCallback, _View2Datasets_text_cbk, New);
    w = XtVaCreateManagedWidget ("beamToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Show Beam",
				 XtNfromVert, upper_widget,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, beam_toggle_cbk, New);
    new->view2Datasets.show_beam_tgl = w;
    w = XtVaCreateManagedWidget ("beamNameToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Show Beam Name",
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XkwNcrosses, False,
				 XtNstate, TRUE,
				 NULL);
    XtAddCallback (w, XtNcallback, beam_toggle_cbk, New);
    upper_widget = w;
    new->view2Datasets.show_beam_name_tgl = w;
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Set 1 Beam Xpos",
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 50,
				 XkwNscaledValuePtr,
				 &new->view2Datasets.set0.beam_xpos,
				 XkwNshowValue, FALSE,
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, FALSE);
    XtAddCallback (w, XkwNvalueChangeCallback, refresh_cbk, New);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "Set 1 Beam Ypos",
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 50,
				 XkwNscaledValuePtr,
				 &new->view2Datasets.set0.beam_ypos,
				 XkwNshowValue, FALSE,
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, FALSE);
    XtAddCallback (w, XkwNvalueChangeCallback, refresh_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNlabel, "Set 2 Beam Xpos",
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 50,
				 XkwNscaledValuePtr,
				 &new->view2Datasets.set1.beam_xpos,
				 XkwNshowValue, FALSE,
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, FALSE);
    XtAddCallback (w, XkwNvalueChangeCallback, refresh_cbk, New);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "Set 2 Beam Ypos",
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 50,
				 XkwNscaledValuePtr,
				 &new->view2Datasets.set1.beam_ypos,
				 XkwNshowValue, FALSE,
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-2, 0.0, FALSE, FALSE);
    XtAddCallback (w, XkwNvalueChangeCallback, refresh_cbk, New);
    upper_widget = w;
    /*  Create other popups  */
    if (new->view2Datasets.createAtInit) create_tracewinpopup (new);
    /*  Add callbacks on Dataclip widgets  */
    if ( ( w = XtNameToWidget (new->view2Datasets.image_display,
			       "izoomwinpopup0") ) == NULL )
    {
	fprintf (stderr, "No \"izoomwinpopup0\" widget\n");
	a_prog_bug (function_name);
    }
    XtAddCallback (w, XkwNregionCallback, region_cbk, New);
    if ( ( w = XtNameToWidget (new->view2Datasets.image_display,
			       "izoomwinpopup1") ) == NULL )
    {
	fprintf (stderr, "No \"izoomwinpopup1\" widget\n");
	a_prog_bug (function_name);
    }
    XtAddCallback (w, XkwNregionCallback, region_cbk, New);
    if ( ( w = XtNameToWidget (new->view2Datasets.image_display,
			       "intensityPolicyPopup") ) == NULL )
    {
	fprintf (stderr, "No \"intensityPolicyPopup\" widget\n");
	a_prog_bug (function_name);
    }
    XtAddCallback (w, XkwNintensityRangeCallback,
		   _View2Datasets_intensity_range_cbk, New);
    if ( ( w = XtNameToWidget (new->view2Datasets.image_display,
			       "exportMenu") ) == NULL )
    {
	fprintf (stderr, "No \"exportMenu\" widget\n");
	a_prog_bug (function_name);
    }
    XtAddCallback (w, XkwNPPMWindowMovieCallback,
		   _View2Datasets_save_ppm_movie_cbk, New);
}   /*  End Function Initialise  */

static void free_viewables (struct direction_type *dir)
/*  [SUMMARY] Free viewable data.
    <dir> The containing structure.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    if (dir->vimages != NULL)
    {
	for (count = 0; count < dir->num_frames; ++count)
	{
	    if (dir->vimages[count] != NULL)
	    {
		viewimg_destroy (dir->vimages[count]);
		dir->vimages[count] = NULL;
	    }
	}
	m_free (dir->vimages);
	dir->vimages = NULL;
    }
    if (dir->mag_vimages != NULL)
    {
	for (count = 0; count < dir->num_frames; ++count)
	{
	    if (dir->mag_vimages[count] != NULL)
	    {
		viewimg_destroy (dir->mag_vimages[count]);
		dir->mag_vimages[count] = NULL;
	    }
	}
	m_free (dir->mag_vimages);
	dir->mag_vimages = NULL;
    }
    if (dir->pan_vimages != NULL)
    {
	for (count = 0; count < dir->num_frames; ++count)
	{
	    if (dir->pan_vimages[count] != NULL)
	    {
		viewimg_destroy (dir->pan_vimages[count]);
		dir->pan_vimages[count] = NULL;
	    }
	}
	m_free (dir->pan_vimages);
	dir->pan_vimages = NULL;
    }
    if (dir->pc_cimages != NULL)
    {
	for (count = 0; count < dir->num_frames; ++count)
	{
	    if (dir->pc_cimages[count] != NULL)
	    {
		contour_destroy (dir->pc_cimages[count]);
		dir->pc_cimages[count] = NULL;
	    }
	}
	m_free (dir->pc_cimages);
	dir->pc_cimages = NULL;
    }
    if (dir->rgb_cimages != NULL)
    {
	for (count = 0; count < dir->num_frames; ++count)
	{
	    if (dir->rgb_cimages[count] != NULL)
	    {
		contour_destroy (dir->rgb_cimages[count]);
		dir->rgb_cimages[count] = NULL;
	    }
	}
	m_free (dir->rgb_cimages);
	dir->rgb_cimages = NULL;
    }
    dir->num_frames = 0;
    dir->frame_number = 0;
}   /*  End Function free_viewables  */

static void slice_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the slice callback. This is called when the slice direction is
    changed.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    int new_slice = *(int *) call_data;

    if (new_slice == top->view2Datasets.slice) return;
    top->view2Datasets.slice = new_slice;
    update_animate_widget (top);
    recompute_profile_dir (top);
    _View2Datasets_display_changes (top, TRUE);
}   /*  End Function slice_cbk  */

static void recompute_profile_dir (View2DatasetsWidget w)
/*  [SUMMARY] Set the profile direction based on the display mode.
    <w> The View2Datasets widget.
    [RETURNS] Nothing.
*/
{
    int new_profile = 0;

    switch (w->view2Datasets.slice)
    {
      case SLICE_XY:
	new_profile = PROFILE_DIR_Z;
	break;
      case SLICE_XZ:
	new_profile = PROFILE_DIR_Y;
	break;
      case SLICE_ZY:
	new_profile = PROFILE_DIR_X;
	break;
    }
    XtVaSetValues (w->view2Datasets.profile_dir_menu,
		   XkwNsetChoice, profile_dir_choices[new_profile],
		   NULL);
    w->view2Datasets.profile_dir = new_profile;
}   /*  End Function recompute_profile_dir  */

static void profile_mode_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  This is the profile mode (shape) callback.
*/
{
    int new_mode = *(int *) call_data;
    Boolean sensitive = True;
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    struct edit_context_list *curr_ctx;

    create_tracewinpopup (top);
    switch (top->view2Datasets.profile_mode = new_mode)
    {
      case PROFILE_MODE_NONE:
	XtPopdown (top->view2Datasets.trace_winpopup);
	for (curr_ctx = top->view2Datasets.first_edit_context;
	     curr_ctx != NULL;
	     curr_ctx = curr_ctx->next)
	    kwin_editor_set_attributes (curr_ctx->context,
					KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					KWIN_EDIT_OBJECT_NONE,
					KWIN_EDITOR_CONTEXT_ATT_END);
	break;
      case PROFILE_MODE_LINE:
	XtPopup (top->view2Datasets.trace_winpopup, XtGrabNone);
	for (curr_ctx = top->view2Datasets.first_edit_context;
	     curr_ctx != NULL;
	     curr_ctx = curr_ctx->next)
	    kwin_editor_set_attributes (curr_ctx->context,
					KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					KWIN_EDIT_OBJECT_NONE,
					KWIN_EDITOR_CONTEXT_ATT_END);
	m_clear (&top->view2Datasets.old_trace_area,
		 sizeof top->view2Datasets.old_trace_area);
	break;
      case PROFILE_MODE_BOX_SUM:
      case PROFILE_MODE_BOX_AVERAGE:
	sensitive = False;
	/*  Fall through  */
      case PROFILE_MODE_BOXED_HORIZONTAL:
      case PROFILE_MODE_BOXED_VERTICAL:
	XtPopup (top->view2Datasets.trace_winpopup, XtGrabNone);
	for (curr_ctx = top->view2Datasets.first_edit_context;
	     curr_ctx != NULL;
	     curr_ctx = curr_ctx->next)
	    kwin_editor_set_attributes (curr_ctx->context,
					KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					KWIN_EDIT_OBJECT_RECTANGLE,
					KWIN_EDITOR_CONTEXT_ATT_END);
	break;
      case PROFILE_MODE_RADIAL:
      case PROFILE_MODE_RADIAL_GAUSS:
	XtPopup (top->view2Datasets.trace_winpopup, XtGrabNone);
	for (curr_ctx = top->view2Datasets.first_edit_context;
	     curr_ctx != NULL;
	     curr_ctx = curr_ctx->next)
	    kwin_editor_set_attributes (curr_ctx->context,
					KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					KWIN_EDIT_OBJECT_CIRCLE,
					KWIN_EDITOR_CONTEXT_ATT_END);
	break;
    }
    if (!sensitive) recompute_profile_dir (top);
    XtSetSensitive (top->view2Datasets.profile_dir_menu, sensitive);
    _View2Datasets_display_changes (top, TRUE);
}   /*  End Function profile_mode_cbk  */

static void profile_dir_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the profile direction callback.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    int new_profile = *(int *) call_data;

    top->view2Datasets.profile_dir = new_profile;
}   /*  End Function profile_dir_cbk  */

static void _View2Datasets_mode_cbk (Widget w, XtPointer client_data,
				     XtPointer call_data)
/*  This is the mode callback. This is called when the mode is changed.
*/
{
    iarray array;
    char *str;
    struct dataset_type *old_set = NULL;
    struct dataset_type *new_set = NULL;
    int new_mode = *(int *) call_data;
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    char txt[STRING_LENGTH];

    if (new_mode == top->view2Datasets.mode) return;
    switch (new_mode)
    {
      case MODE_IMAGE0_CONTOUR1:
	new_set = &top->view2Datasets.set1;
	break;
      case MODE_IMAGE1_CONTOUR0:
	new_set = &top->view2Datasets.set0;
	break;
    }
    switch (top->view2Datasets.mode)
    {
      case MODE_IMAGE0_CONTOUR1:
	old_set = &top->view2Datasets.set1;
	break;
      case MODE_IMAGE1_CONTOUR0:
	old_set = &top->view2Datasets.set0;
	break;
    }
    top->view2Datasets.mode = new_mode;
    if (old_set != NULL)
    {
	/*  Save old strings  */
	XtVaGetValues (top->view2Datasets.levels_dlg, XtNvalue, &str, NULL);
	strcpy (old_set->levels, str);
	XtVaGetValues (top->view2Datasets.negative_colour_dlg,
		       XtNvalue, &str,
		       NULL);
	strcpy (old_set->negative_contour_colour, str);
	XtVaGetValues (top->view2Datasets.positive_colour_dlg,
		       XtNvalue, &str,
		       NULL);
	strcpy (old_set->positive_contour_colour, str);
    }
    if (new_set != NULL)
    {
	/*  Set new strings  */
	array = new_set->array_pseudo;
	if (array == NULL) strcpy (txt, "Contour Levels");
	else sprintf ( txt, "Contour Levels (%s)",
		       iarray_value_name (array) );
	XtVaSetValues (top->view2Datasets.levels_dlg,
		       XtNlabel, txt,
		       XtNvalue, new_set->levels,
		       NULL);
	XtVaSetValues (top->view2Datasets.negative_colour_dlg,
		       XtNvalue, new_set->negative_contour_colour,
		       NULL);
	XtVaSetValues (top->view2Datasets.positive_colour_dlg,
		       XtNvalue, new_set->positive_contour_colour,
		       NULL);
	update_profile_axis (top, new_set);
    }
    update_animate_widget (top);
    _View2Datasets_display_changes (top, TRUE);
    _View2Datasets_update_cmap (top);
    _View2Datasets_update_title (top);
}   /*  End Function _View2Datasets_mode_cbk  */

static void _View2Datasets_display_changes (View2DatasetsWidget top,
					    flag update_profile)
/*  [SUMMARY] Register a change in the displayed data.
    <top> The View2Datasets widget.
    <update_profile> If TRUE the profile display is updated.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = NULL;
    KContourImage cimage = NULL;
    signed int frame = -1;
    Bool auto_title, link_frames;
    struct dataset_type *image_set, *contour_set, *hidden_set;
    struct direction_type *dir;
    static char function_name[] = "View2DatasetsWidget::display_changes";

    get_sets (top, &image_set, &contour_set, &hidden_set, NULL);
    XtVaGetValues (top->view2Datasets.auto_title_tgl,
		   XtNstate, &auto_title,
		   NULL);
    /*  First update the image  */
    dir = get_dir (top, image_set);
    if (image_set->vimage != NULL)
    {
	canvas = viewimg_get_worldcanvas (image_set->vimage);
	contour_set_all_inactive (canvas, FALSE, FALSE, (uaddr) top);
	viewimg_set_active (image_set->vimage, FALSE);
	viewimg_set_active (image_set->mag_vimage, FALSE);
	viewimg_set_active (image_set->pan_vimage, FALSE);
    }
    else if (image_set->xy.vimages != NULL)
    {
	frame = dir->frame_number;
	canvas = viewimg_get_worldcanvas (dir->vimages[0]);
	contour_set_all_inactive (canvas, FALSE, FALSE, (uaddr) top);
	if (frame >= dir->num_frames)
	{
	    fprintf (stderr,
		     "%s: current iframe: %u is not less than number: %u\n",
		     function_name, frame, dir->num_frames);
	    return;
	}
	viewimg_set_active (dir->vimages[frame], FALSE);
	viewimg_set_active (dir->mag_vimages[frame], FALSE);
	viewimg_set_active (dir->pan_vimages[frame], FALSE);
    }
    if (canvas == NULL) return;
    if ( update_profile && (top->view2Datasets.trace_winpopup != NULL) )
    {
	switch (top->view2Datasets.profile_mode)
	{
	  case PROFILE_MODE_LINE:
	  case PROFILE_MODE_BOX_SUM:
	  case PROFILE_MODE_BOX_AVERAGE:
	    if (image_set->array_pseudo != NULL)
		XkwTracePopupNewArray (top->view2Datasets.trace_winpopup,
				       image_set->array_pseudo,
				       image_set->min, image_set->max);
	    else XkwTracePopupNewArray (top->view2Datasets.trace_winpopup,
					NULL, 0.0, 0.0);
	    break;
	}
    }
    if (!auto_title)
    {
	String title;

	XtVaGetValues (top->view2Datasets.title_txt, XtNstring, &title, NULL);
	canvas_set_dressing (canvas,
			     KCD_ATT_TITLE, title,
			     KCD_ATT_END);
	m_clear ( title, strlen (title) );
    }
    /*  Now display the contours if necessary  */
    /*  Determine frame number for hidden dataset, if appropriate  */
    dir = get_dir (top, hidden_set);
    XtVaGetValues (top->view2Datasets.link_frames_tgl,
		   XtNstate, &link_frames,
		   NULL);
    if ( (dir != NULL) && (dir->num_frames >1) && (frame >= 0) && link_frames )
    {
	/*  Copy over frame number, preferably by matching real units  */
	KwcsAstro image_ap, hidden_ap;
	iarray image_arr, hidden_arr;
	flag match = FALSE;
	int dim_num = -1;
	dim_desc *image_dim, *hidden_dim;

	switch (top->view2Datasets.slice)
	{
	  case SLICE_XY:
	    dim_num = 0;
	    break;
	  case SLICE_XZ:
	    dim_num = 1;
	    break;
	  case SLICE_ZY:
	    dim_num = 2;
	    break;
	}
	image_arr = image_set->array_pseudo;
	if (image_arr == NULL) image_arr = image_set->array_red;
	image_dim = iarray_get_dim_desc (image_arr, dim_num);
	hidden_arr = hidden_set->array_pseudo;
	if (hidden_arr == NULL) hidden_arr = hidden_set->array_red;
	hidden_dim = iarray_get_dim_desc (hidden_arr, dim_num);
	if (strcmp (image_dim->name, hidden_dim->name) == 0) match = TRUE;
	image_ap = iarray_get_associated_wcs_astro (image_arr);
	hidden_ap = iarray_get_associated_wcs_astro (hidden_arr);
	if ( wcs_astro_test_if_ra (image_ap, image_dim->name) &&
	     wcs_astro_test_if_ra (hidden_ap, hidden_dim->name) )
	    match = TRUE;
	if ( wcs_astro_test_if_dec (image_ap, image_dim->name) &&
	     wcs_astro_test_if_dec (hidden_ap, hidden_dim->name) )
	    match = TRUE;
	if ( wcs_astro_test_if_glon (image_ap, image_dim->name) &&
	     wcs_astro_test_if_glon (hidden_ap, hidden_dim->name) )
	    match = TRUE;
	if ( wcs_astro_test_if_glat (image_ap, image_dim->name) &&
	     wcs_astro_test_if_glat (hidden_ap, hidden_dim->name) )
	    match = TRUE;
	if ( ( (strncmp (image_dim->name, "VELO", 4) == 0) ||
	       (strncmp (image_dim->name, "FELO", 4) == 0) ) &&
	     ( (strncmp (hidden_dim->name, "VELO", 4) == 0) ||
	       (strncmp (hidden_dim->name, "FELO", 4) == 0) ) )
	    match = TRUE;
	if (match)
	{
	    /*  Match using real units  */
	    unsigned int count;
	    double coords[3];
	    CONST char *names[3];

	    for (count = 0; count < 3; ++count)
	    {
		dim_desc *dim = iarray_get_dim_desc (image_arr, count);

		coords[count] = (count == dim_num) ? frame : (dim->length / 2);
		ds_convert_coordinates (dim, 1, coords + count, FALSE, FALSE,
					FALSE);
		names[count] = dim->name;
	    }
	    if (image_ap != NULL)
		wcs_astro_transform3 (image_ap, 1,
				      names[0], coords, FALSE,
				      names[1], coords + 1, FALSE,
				      names[2], coords + 2, FALSE,
				      0, NULL, NULL);
	    for (count = 0; count < 3; ++count)
		names[count] = iarray_dim_name (hidden_arr, count);
	    if (hidden_ap != NULL)
		wcs_astro_transform3 (hidden_ap, 1,
				      names[0], coords, TRUE,
				      names[1], coords + 1, TRUE,
				      names[2], coords + 2, TRUE,
				      0, NULL, NULL);
	    dir->frame_number = ds_get_coord_num (hidden_dim, coords[dim_num],
						  SEARCH_BIAS_CLOSEST);
	    ds_convert_coordinates (hidden_dim, 1, coords + dim_num,
				    TRUE, TRUE, TRUE);
	    if ( DOUBLE_IS_VALID (coords[dim_num]) ) frame = dir->frame_number;
	    else frame = -1;
	}
	else  /*  Just copy frame numbers  */
	{
	    if (frame >= dir->num_frames) frame = dir->num_frames - 1;
	    dir->frame_number = frame;
	}
    }
    else  /*  Not copying frame number  */
    {
	if (dir != NULL) frame = dir->frame_number;
    }
    dir = get_dir (top, contour_set);  /*  Pedantry  */
    if (contour_set != NULL)
    {
	if (contour_set->pc_cimage != NULL) cimage = contour_set->pc_cimage;
	else if ( (contour_set->xy.pc_cimages != NULL) && (frame >= 0) )
	    cimage = dir->pc_cimages[frame];
	if ( (cimage != NULL) && (contour_get_worldcanvas (cimage) != canvas) )
	    cimage = NULL;
	if (cimage == NULL)
	{
	    if (contour_set->rgb_cimage != NULL)
		cimage = contour_set->rgb_cimage;
	    else if ( (contour_set->xy.rgb_cimages != NULL) && (frame >= 0) )
		cimage = dir->rgb_cimages[frame];
	}
	if ( (cimage != NULL) && (contour_get_worldcanvas (cimage) != canvas) )
	    cimage = NULL;
    }
    if (cimage == NULL)
    {
	if (auto_title)
	    canvas_set_dressing (canvas,
				 KCD_ATT_TITLE, image_set->object_name,
				 KCD_ATT_END);
    }
    else
    {
	char title[STRING_LENGTH];

	if (auto_title)
	{
	    sprintf (title, "%s overlaid with %s",
		     image_set->object_name, contour_set->object_name);
	    canvas_set_dressing (canvas,
				 KCD_ATT_TITLE, title,
				 KCD_ATT_END);
	}
	contour_set_active (cimage, TRUE, FALSE, FALSE, TRUE, (uaddr) top);
    }
    XkwImageDisplayRefresh (top->view2Datasets.image_display, FALSE);
    XtVaSetValues (top->view2Datasets.image_display,
		   XkwNvisibleCanvas, canvas,
		   NULL);
}   /*  End Function display_changes   */

static void _View2Datasets_update_title (View2DatasetsWidget top)
/*  [SUMMARY] Update the title bar.
    <top> The View2Datasets widget.
    [RETURNS] Nothing.
*/
{
    unsigned int set_num;
    void *object = NULL;
    struct dataset_type *set = NULL;
    char title[STRING_LENGTH];

    if (top->view2Datasets.wmtitle == NULL) return;
    get_sets (top, &set, NULL, NULL, NULL);
    if (set->multi_desc == NULL) return;
    if (set != NULL)
    {
	if ( !ds_get_named_object (set->multi_desc,
				   set->multi_desc->num_arrays,
				   "filename", &object) )
	    object = set->object_name;
    }
    if (object == NULL) strcpy (title, top->view2Datasets.wmtitle);
    else
    {
	set_num = (set == &top->view2Datasets.set0) ? 1 : 2;
	sprintf (title, "%s  Set%u: %s\n",
		 top->view2Datasets.wmtitle, set_num, (char *) object);
    }
    if (strcmp (top->view2Datasets.old_title, title) == 0) return;
    XtVaSetValues (top->view2Datasets.wmshell,
		   XtNtitle, title,
		   NULL);
    strcpy (top->view2Datasets.old_title, title);
}   /*  End Function _View2Datasets_update_title  */

static void _View2Datasets_update_cmap (View2DatasetsWidget top)
/*  [SUMMARY] Update the colourmap if appropriate.
    <top> The View2Datasets widget.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas worldcanvas = NULL;
    Kcolourmap cmap;
    struct dataset_type *set;

    get_sets (top, &set, NULL, NULL, NULL);
    if (set->array_pseudo == NULL) return;
    if (set->cmap_index >= set->multi_desc->num_arrays) return;
    if (set->vimage != NULL)
	worldcanvas = viewimg_get_worldcanvas (set->vimage);
    else if (set->xy.vimages != NULL)
	worldcanvas = viewimg_get_worldcanvas (set->xy.vimages[0]);
    if (worldcanvas == NULL) return;
    cmap = canvas_get_cmap (worldcanvas);
    if (cmap == NULL) return;
    if ( !kcmap_copy_from_struct (cmap,
				  set->multi_desc->headers[set->cmap_index],
				  set->multi_desc->data[set->cmap_index]) )
    {
	fprintf (stderr, "Error changing colourmap\n");
    }
}   /*  End Function update_cmap  */

static void invalidate_planes (struct direction_type *info)
/*  [SUMMARY] Free viewable data.
    <info> The containing structure.
    [RETURNS] Nothing.
*/
{
    unsigned int count;

    if (info->vimages != NULL)
    {
	for (count = 0; count < info->num_frames; ++count)
	{
	    if (info->vimages[count] != NULL)
		viewimg_register_data_change (info->vimages[count]);
	}
    }
    if (info->mag_vimages != NULL)
    {
	for (count = 0; count < info->num_frames; ++count)
	{
	    if (info->mag_vimages[count] != NULL)
		viewimg_register_data_change (info->mag_vimages[count]);
	}
    }
    if (info->pan_vimages != NULL)
    {
	for (count = 0; count < info->num_frames; ++count)
	{
	    if (info->pan_vimages[count] != NULL)
		viewimg_register_data_change (info->pan_vimages[count]);
	}
    }
    if (info->pc_cimages != NULL)
    {
	for (count = 0; count < info->num_frames; ++count)
	{
	    if (info->pc_cimages[count] != NULL)
		contour_register_data_change (info->pc_cimages[count]);
	}
    }
    if (info->rgb_cimages != NULL)
    {
	for (count = 0; count < info->num_frames; ++count)
	{
	    if (info->rgb_cimages[count] != NULL)
		contour_register_data_change (info->rgb_cimages[count]);
	}
    }
}   /*  End Function invalidate_planes  */

static void update_animate_widget (View2DatasetsWidget top)
/*  [SUMMARY] Update the animate control widget with the number of frames.
    <top> The widget.
    [RETURNS] Nothing.
*/
{
    struct direction_type *dir = NULL;
    struct dataset_type *image_set, *contour_set;
    /*staticcharfunction_name[]="View2DatasetsWidget::update_animate_widget";*/

    get_sets (top, &image_set, &contour_set, NULL, NULL);
    dir = get_dir (top, image_set);
    if ( (dir->num_frames < 2) && (contour_set != NULL) )
	dir = get_dir (top, contour_set);
    XtVaSetValues (top->view2Datasets.animatepopup,
		   XkwNnumFrames, dir->num_frames,
		   XkwNcurrentFrame, dir->frame_number,
		   NULL);
    if (dir->num_frames < 2)
	XkwAnimateControlPopdown (top->view2Datasets.animatepopup);
}   /*  End Function update_animate_widget  */

static void update_profile_axis (View2DatasetsWidget top,
				 struct dataset_type *set)
/*  [SUMMARY] Update the profile axis.
    <top> The View2Datasets widget.
    <set> The current dataset.
    [RETURNS] Nothing.
*/
{
    int new_profile = 0;

    if ( (set->array_pseudo != NULL) &&
	 (iarray_num_dim (set->array_pseudo) == 2) )
    {
	XtVaSetValues (top->view2Datasets.profile_dir_menu,
		       XkwNsetChoice, profile_dir_choices[PROFILE_DIR_X],
		       NULL);
	top->view2Datasets.profile_dir = PROFILE_DIR_X;
    }
    else if ( (set->array_pseudo != NULL) &&
	      (iarray_num_dim (set->array_pseudo) == 3) )
    {
	switch (top->view2Datasets.slice)
	{
	  case SLICE_XY:
	    new_profile = PROFILE_DIR_Z;
	    break;
	  case SLICE_XZ:
	    new_profile = PROFILE_DIR_Y;
	    break;
	  case SLICE_ZY:
	    new_profile = PROFILE_DIR_X;
	    break;
	}
	XtVaSetValues (top->view2Datasets.profile_dir_menu,
		       XkwNsetChoice, profile_dir_choices[new_profile],
		       NULL);
	top->view2Datasets.profile_dir = new_profile;
    }
}   /*  End Function update_profile_axis  */

static void new_frame_cbk (Widget w, XtPointer client_data,
			   XtPointer call_data)
/*  This is the new frame callback.
*/
{
    int frame_number = *(int *) call_data;
    struct dataset_type *image_set, *hidden_set;
    struct direction_type *dir;
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    Widget magnifier;

    get_sets (top, &image_set, NULL, &hidden_set, NULL);
    dir = get_dir (top, image_set);
    if (dir->num_frames < 2) dir = get_dir (top, hidden_set);
    if ( (dir == NULL) || (dir->num_frames < 2) )
    {
	fprintf (stderr, "No frames!\n");
	return;
    }
    dir->frame_number = frame_number;
    _View2Datasets_display_changes (top, FALSE);
    XtVaGetValues (top->view2Datasets.image_display,
		   XkwNmagnifier, &magnifier,
		   NULL);
    XkwSlaveImageDisplayPopupRefresh (magnifier, FALSE);
}   /*  End Function new_frame_cbk   */

static void blink_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the blink callback.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;

    update_animate_widget (top);
    _View2Datasets_display_changes (top, TRUE);
    _View2Datasets_update_cmap (top);
    _View2Datasets_update_title (top);
}   /*  End Function blink_cbk   */

static void iarr_destroy_callback (iarray arr, struct dataset_type *set)
/*  [PURPOSE] This routine will register the destruction of the Intelligent
    Array.
    <arr> The Intelligent Array.
    <set> The dataset.
    [RETURNS] Nothing.
*/
{
    set->destroy_func = NULL;
    set->array_pseudo = NULL;
    set->array_red = NULL;
    set->array_green = NULL;
    set->array_blue = NULL;
    if (set->vimage != NULL) viewimg_destroy (set->vimage);
    set->vimage = NULL;
    if (set->mag_vimage != NULL) viewimg_destroy (set->mag_vimage);
    set->mag_vimage = NULL;
    if (set->pan_vimage != NULL) viewimg_destroy (set->pan_vimage);
    set->pan_vimage = NULL;
    if (set->pc_cimage != NULL) contour_destroy (set->pc_cimage);
    set->pc_cimage = NULL;
    if (set->rgb_cimage != NULL) contour_destroy (set->rgb_cimage);
    set->rgb_cimage = NULL;
    free_viewables (&set->xy);
    free_viewables (&set->xz);
    free_viewables (&set->zy);
#ifdef NOT_IMPLEMENTED
    XkwChoiceMenuSetEntrySensitivity
	(top->view2Datasets.intensity_menu,
	 top->view2Datasets.first_intensity + set_number,
	 FALSE);
#endif
}   /*  End Function iarr_destroy_callback  */

static void set_iscale_range (struct dataset_type *set, double min, double max,
			      flag scaled)
/*  [SUMMARY] Set the intensity scale range for a dataset.
    <set> The dataset.
    <min> The minimum value.
    <max> The maximum value.
    <scaled> If TRUE, the values are scaled, otherwise they are unscaled (raw).
    [RETURNS] Nothing.
*/
{
    if (set->array_pseudo == NULL) return;
    if (scaled)
    {
	double scale, offset;

	iarray_get_data_scaling (set->array_pseudo, &scale, &offset);
	min = (min - offset) / scale;
	max = (max - offset) / scale;
    }
    if (set->vimage != NULL)
    {
	viewimg_set_attributes (set->vimage,
				VIEWIMG_VATT_VALUE_MIN, min,
				VIEWIMG_VATT_VALUE_MAX, max,
				VIEWIMG_VATT_END);
	viewimg_set_attributes (set->mag_vimage,
				VIEWIMG_VATT_VALUE_MIN, min,
				VIEWIMG_VATT_VALUE_MAX, max,
				VIEWIMG_VATT_END);
	viewimg_set_attributes (set->pan_vimage,
				VIEWIMG_VATT_VALUE_MIN, min,
				VIEWIMG_VATT_VALUE_MAX, max,
				VIEWIMG_VATT_END);
    }
    if (set->xy.vimages != NULL)
    {
	viewimg_set_array_attributes (set->xy.vimages, set->xy.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
	viewimg_set_array_attributes (set->xz.vimages, set->xz.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
	viewimg_set_array_attributes (set->zy.vimages, set->zy.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
	viewimg_set_array_attributes (set->xy.mag_vimages, set->xy.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
	viewimg_set_array_attributes (set->xz.mag_vimages, set->xz.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
	viewimg_set_array_attributes (set->zy.mag_vimages, set->zy.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
	viewimg_set_array_attributes (set->xy.pan_vimages, set->xy.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
	viewimg_set_array_attributes (set->xz.pan_vimages, set->xz.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
	viewimg_set_array_attributes (set->zy.pan_vimages, set->zy.num_frames,
				      VIEWIMG_VATT_VALUE_MIN, min,
				      VIEWIMG_VATT_VALUE_MAX, max,
				      VIEWIMG_VATT_END);
    }
}   /*  End Function set_iscale_range  */

static flag setup_2D_pseudo (View2DatasetsWidget w, struct dataset_type *set,
			     iarray array, double lower_clip,double upper_clip,
			     KWorldCanvas pseudo_canvas,
			     KWorldCanvas rgb_canvas,
			     KWorldCanvas mag_pseudo_canvas,
			     KWorldCanvas mag_rgb_canvas,
			     KWorldCanvas pan_pseudo_canvas,
			     KWorldCanvas pan_rgb_canvas)
/*  [SUMMARY] Setup viewables for a 2D PsuedoColour array.
    <w> The widget.
    <set> The set.
    <array> The array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "View2DatasetsWidget::setup_2D_pseudo";

    if (pseudo_canvas != NULL)
    {
	if ( ( set->vimage =
	       viewimg_create_from_iarray (pseudo_canvas, array, FALSE) )
	     == NULL )
	{
	    m_error_notify (function_name, "vimage");
	    return (FALSE);
	}
	if ( ( set->mag_vimage =
	       viewimg_create_from_iarray (mag_pseudo_canvas, array, FALSE) )
	     == NULL )
	{
	    m_error_notify (function_name, "magnified vimage");
	    return (FALSE);
	}
	if ( ( set->pan_vimage =
	       viewimg_create_from_iarray (pan_pseudo_canvas, array, FALSE) )
	     == NULL )
	{
	    m_error_notify (function_name, "magnified vimage");
	    return (FALSE);
	}
	set_iscale_range (set, lower_clip, upper_clip, FALSE);
	if ( ( set->pc_cimage =
	       contour_create_from_iarray (pseudo_canvas, array, FALSE,
					   0, NULL, (uaddr) w) ) == NULL )
	{
	    m_error_notify (function_name, "cimage");
	    return (FALSE);
	}
    }
    if (rgb_canvas != NULL)
    {
	if ( ( set->rgb_cimage =
	       contour_create_from_iarray (rgb_canvas, array, FALSE,
					   0, NULL, (uaddr) w) ) == NULL )
	{
	    m_error_notify (function_name, "cimage");
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function setup_2D_pseudo  */

static flag setup_3D_pseudo (View2DatasetsWidget w, struct dataset_type *set,
			     iarray array, double lower_clip,double upper_clip,
			     KWorldCanvas pseudo_canvas,
			     KWorldCanvas rgb_canvas,
			     KWorldCanvas mag_pseudo_canvas,
			     KWorldCanvas mag_rgb_canvas,
			     KWorldCanvas pan_pseudo_canvas,
			     KWorldCanvas pan_rgb_canvas)
/*  [SUMMARY] Setup viewables for a 3D PsuedoColour array.
    <w> The widget.
    <set> The set.
    <array> The array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "View2DatasetsWidget::setup_3D_pseudo";

    if (pseudo_canvas != NULL)
    {
	set->xy.num_frames = iarray_dim_length (array, 0);
	set->xz.num_frames = iarray_dim_length (array, 1);
	set->zy.num_frames = iarray_dim_length (array, 2);
	if ( ( set->xy.vimages = viewimg_create_sequence_from_iarray
	       (pseudo_canvas, array, 2, 1, 0) ) == NULL )
	{
	    m_error_notify (function_name, "xy vimages");
	    return (FALSE);
	}
	if ( ( set->xz.vimages = viewimg_create_sequence_from_iarray
	       (pseudo_canvas, array, 2, 0, 1) ) == NULL )
	{
	    m_error_notify (function_name, "xz vimages");
	    return (FALSE);
	}
	if ( ( set->zy.vimages = viewimg_create_sequence_from_iarray
	       (pseudo_canvas, array, 0, 1, 2) ) == NULL )
	{
	    m_error_notify (function_name, "zy vimages");
	    return (FALSE);
	}
	if ( ( set->xy.mag_vimages = viewimg_create_sequence_from_iarray
	       (mag_pseudo_canvas, array, 2, 1, 0) ) == NULL )
	{
	    m_error_notify (function_name, "xy magnified vimages");
	    return (FALSE);
	}
	if ( ( set->xz.mag_vimages = viewimg_create_sequence_from_iarray
	       (mag_pseudo_canvas, array, 2, 0, 1) ) == NULL )
	{
	    m_error_notify (function_name, "xz magnified vimages");
	    return (FALSE);
	}
	if ( ( set->zy.mag_vimages = viewimg_create_sequence_from_iarray
	       (mag_pseudo_canvas, array, 0, 1, 2) ) == NULL )
	{
	    m_error_notify (function_name, "zy magnified vimages");
	    return (FALSE);
	}
	if ( ( set->xy.pan_vimages = viewimg_create_sequence_from_iarray
	       (pan_pseudo_canvas, array, 2, 1, 0) ) == NULL )
	{
	    m_error_notify (function_name, "xy magnified vimages");
	    return (FALSE);
	}
	if ( ( set->xz.pan_vimages = viewimg_create_sequence_from_iarray
	       (pan_pseudo_canvas, array, 2, 0, 1) ) == NULL )
	{
	    m_error_notify (function_name, "xz magnified vimages");
	    return (FALSE);
	}
	if ( ( set->zy.pan_vimages = viewimg_create_sequence_from_iarray
	       (pan_pseudo_canvas, array, 0, 1, 2) ) == NULL )
	{
	    m_error_notify (function_name, "zy magnified vimages");
	    return (FALSE);
	}
	set_iscale_range (set, lower_clip, upper_clip, FALSE);
	if ( ( set->xy.pc_cimages = contour_create_sequence_from_iarray
	       (pseudo_canvas, array, 2, 1, 0, 0, NULL, (uaddr) w) ) == NULL )
	{
	    m_error_notify (function_name, "xy cimages");
	    return (FALSE);
	}
	if ( ( set->xz.pc_cimages = contour_create_sequence_from_iarray
	       (pseudo_canvas, array, 2, 0, 1, 0, NULL, (uaddr) w) ) == NULL )
	{
	    m_error_notify (function_name, "xz cimages");
	    return (FALSE);
	}
	if ( ( set->zy.pc_cimages = contour_create_sequence_from_iarray
	       (pseudo_canvas, array, 0, 1, 2, 0, NULL, (uaddr) w) ) == NULL )
	{
	    m_error_notify (function_name, "zy cimages");
	    return (FALSE);
	}
    }
    if (rgb_canvas != NULL)
    {
	if ( ( set->xy.rgb_cimages = contour_create_sequence_from_iarray
	       (rgb_canvas, array, 2, 1, 0, 0, NULL, (uaddr) w) ) == NULL )
	{
	    m_error_notify (function_name, "xy cimages");
	    return (FALSE);
	}
	if ( ( set->xz.rgb_cimages = contour_create_sequence_from_iarray
	       (rgb_canvas, array, 2, 0, 1, 0, NULL, (uaddr) w) ) == NULL )
	{
	    m_error_notify (function_name, "xz cimages");
	    return (FALSE);
	}
	if ( ( set->zy.rgb_cimages = contour_create_sequence_from_iarray
	       (rgb_canvas, array, 0, 1, 2, 0, NULL, (uaddr) w) ) == NULL )
	{
	    m_error_notify (function_name, "zy cimages");
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function setup_3D_pseudo  */

static flag setup_2D_rgb (View2DatasetsWidget w, struct dataset_type *set,
			  iarray red, iarray green, iarray blue,
			  KWorldCanvas canvas, KWorldCanvas mag_canvas,
			  KWorldCanvas pan_canvas)
/*  [SUMMARY] Setup viewables for a 2D TrueColour array.
    <w> The widget.
    <set> The set.
    <red> The red component array.
    <green> The green component array.
    <blue> The blue component array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "View2DatasetsWidget::setup_2D_rgb";

    if ( ( set->vimage =
	   viewimg_create_rgb (canvas, red->multi_desc, red->arr_desc,
			       red->data, 1, 0,
			       red->elem_index, green->elem_index,
			       blue->elem_index,
			       0, (CONST char **) NULL,
			       (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "vimage");
	return (FALSE);
    }
    if ( ( set->mag_vimage =
	   viewimg_create_rgb (mag_canvas, red->multi_desc, red->arr_desc,
			       red->data, 1, 0,
			       red->elem_index, green->elem_index,
			       blue->elem_index,
			       0, (CONST char **) NULL,
			       (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "vimage");
	return (FALSE);
    }
    if ( ( set->pan_vimage =
	   viewimg_create_rgb (pan_canvas, red->multi_desc, red->arr_desc,
			       red->data, 1, 0,
			       red->elem_index, green->elem_index,
			       blue->elem_index,
			       0, (CONST char **) NULL,
			       (CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "vimage");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function setup_2D_rgb  */

static flag setup_3D_rgb (View2DatasetsWidget w, struct dataset_type *set,
			  iarray red, iarray green, iarray blue,
			  KWorldCanvas canvas, KWorldCanvas mag_canvas,
			  KWorldCanvas pan_canvas)
/*  [SUMMARY] Setup viewables for a 3D TrueColour array.
    <w> The widget.
    <set> The set.
    <red> The red component array.
    <green> The green component array.
    <blue> The blue component array.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "View2DatasetsWidget::setup_3D_rgb";

    set->xy.num_frames = iarray_dim_length (red, 0);
    set->xz.num_frames = iarray_dim_length (red, 1);
    set->zy.num_frames = iarray_dim_length (red, 2);
    if ( ( set->xy.vimages =
	   viewimg_create_rgb_sequence (canvas, red->multi_desc, red->arr_desc,
					red->data, 2, 1, 0,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xy vimages");
	return (FALSE);
    }
    if ( ( set->xz.vimages =
	   viewimg_create_rgb_sequence (canvas, red->multi_desc, red->arr_desc,
					red->data, 2, 0, 1,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xz vimages");
	return (FALSE);
    }
    if ( ( set->zy.vimages =
	   viewimg_create_rgb_sequence (canvas, red->multi_desc, red->arr_desc,
					red->data, 0, 1, 2,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "zy vimages");
	return (FALSE);
    }
    if ( ( set->xy.mag_vimages =
	   viewimg_create_rgb_sequence (mag_canvas, red->multi_desc,
					red->arr_desc,
					red->data, 2, 1, 0,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xy magnified vimages");
	return (FALSE);
    }
    if ( ( set->xz.mag_vimages =
	   viewimg_create_rgb_sequence (mag_canvas, red->multi_desc,
					red->arr_desc,
					red->data, 2, 0, 1,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xz magnified vimages");
	return (FALSE);
    }
    if ( ( set->zy.mag_vimages =
	   viewimg_create_rgb_sequence (mag_canvas, red->multi_desc,
					red->arr_desc,
					red->data, 0, 1, 2,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "zy magnified vimages");
	return (FALSE);
    }
    if ( ( set->xy.pan_vimages =
	   viewimg_create_rgb_sequence (pan_canvas, red->multi_desc,
					red->arr_desc,
					red->data, 2, 1, 0,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xy magnified vimages");
	return (FALSE);
    }
    if ( ( set->xz.pan_vimages =
	   viewimg_create_rgb_sequence (pan_canvas, red->multi_desc,
					red->arr_desc,
					red->data, 2, 0, 1,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "xz magnified vimages");
	return (FALSE);
    }
    if ( ( set->zy.pan_vimages =
	   viewimg_create_rgb_sequence (pan_canvas, red->multi_desc,
					red->arr_desc,
					red->data, 0, 1, 2,
					red->elem_index, green->elem_index,
					blue->elem_index,
					0, (CONST char **) NULL,
					(CONST double *) NULL) ) == NULL )
    {
	m_error_notify (function_name, "zy magnified vimages");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function setup_3D_rgb  */

static void apply_levels_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  [SUMMARY] Contour levels change callback.
    <w> The widget the event occurred on. If this is NULL the display canvas is
    not refreshed.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    unsigned int count, num_levels;
    double scale, offset;
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    struct dataset_type *set;
    char *level_str, *colour;
    flag dash[MAX_CONTOUR_LEVELS];
    double contour_levels[MAX_CONTOUR_LEVELS], linewidths[MAX_CONTOUR_LEVELS];
    /*static char function_name[] = "View2DatasetsWidget::apply_levels_cbk";*/

    get_sets (top, NULL, &set, NULL, NULL);
    if ( (set == NULL) || (set->array_pseudo == NULL) )
    {
	XBell (XtDisplay (w), 100);
	fprintf (stderr, "No mono-variate contour dataset loaded\n");
	return;
    }
    iarray_get_data_scaling (set->array_pseudo, &scale, &offset);
    XtVaGetValues (top->view2Datasets.negative_colour_dlg,
		   XtNvalue, &colour,
		   NULL);
    strcpy (set->negative_contour_colour, colour);
    XtVaGetValues (top->view2Datasets.positive_colour_dlg,
		   XtNvalue, &colour,
		   NULL);
    strcpy (set->positive_contour_colour, colour);
    XtVaGetValues (top->view2Datasets.levels_dlg,
		   XtNvalue, &level_str,
		   NULL);
    strcpy (set->levels, level_str);
    if ( ( num_levels =
	   contour_parse_levels (contour_levels, MAX_CONTOUR_LEVELS, level_str,
				 set->min * scale + offset,
				 set->max * scale + offset) ) < 1 ) return;
    contour_unscale_levels (contour_levels, dash, linewidths, num_levels,
			    scale, offset,
			    top->view2Datasets.contour_neg_width,
			    top->view2Datasets.contour_pos_width);
    if (set->pc_cimage != NULL)
    {
	_View2Datasets_set_clevels (set->pc_cimage, num_levels,
				    contour_levels, dash, linewidths,
				    set->negative_contour_colour,
				    set->positive_contour_colour);
    }
    if (set->rgb_cimage != NULL)
    {
	_View2Datasets_set_clevels (set->rgb_cimage, num_levels,
				    contour_levels, dash, linewidths,
				    set->negative_contour_colour,
				    set->positive_contour_colour);
    }
    for (count = 0; count < set->xy.num_frames; ++count)
    {
	if ( (set->xy.pc_cimages != NULL) &&
	     (set->xy.pc_cimages[count] != NULL) )
	    _View2Datasets_set_clevels (set->xy.pc_cimages[count], num_levels,
					contour_levels, dash, linewidths,
					set->negative_contour_colour,
					set->positive_contour_colour);
	if ( (set->xy.rgb_cimages != NULL) &&
	     (set->xy.rgb_cimages[count] != NULL) )
	    _View2Datasets_set_clevels (set->xy.rgb_cimages[count], num_levels,
					contour_levels, dash, linewidths,
					set->negative_contour_colour,
					set->positive_contour_colour);
    }
    for (count = 0; count < set->xz.num_frames; ++count)
    {
	if ( (set->xz.pc_cimages != NULL) &&
	     (set->xz.pc_cimages[count] != NULL) )
	    _View2Datasets_set_clevels (set->xz.pc_cimages[count], num_levels,
					contour_levels, dash, linewidths,
					set->negative_contour_colour,
					set->positive_contour_colour);
	if ( (set->xz.rgb_cimages != NULL) &&
	     (set->xz.rgb_cimages[count] != NULL) )
	    _View2Datasets_set_clevels (set->xz.rgb_cimages[count], num_levels,
					contour_levels, dash, linewidths,
					set->negative_contour_colour,
					set->positive_contour_colour);
    }
    for (count = 0; count < set->zy.num_frames; ++count)
    {
	if ( (set->zy.pc_cimages != NULL) &&
	     (set->zy.pc_cimages[count] != NULL) )
	    _View2Datasets_set_clevels (set->zy.pc_cimages[count], num_levels,
					contour_levels, dash, linewidths,
					set->negative_contour_colour,
					set->positive_contour_colour);
	if ( (set->zy.rgb_cimages != NULL) &&
	     (set->zy.rgb_cimages[count] != NULL) )
	    _View2Datasets_set_clevels (set->zy.rgb_cimages[count], num_levels,
					contour_levels, dash, linewidths,
					set->negative_contour_colour,
					set->positive_contour_colour);
    }
    if (w != NULL) XkwImageDisplayRefresh (top->view2Datasets.image_display,
					   TRUE);
}   /*  End Function apply_levels_cbk  */

static void _View2Datasets_set_clevels (KContourImage cimage,
					unsigned int num_levels,
					CONST double *contour_levels,
					CONST flag *dash,
					CONST double *linewidths,
					CONST char *negative_colour,
					CONST char *positive_colour)
/*  [SUMMARY] Set contour levels for a KContourImage object.
    <cimage> The KContourImage object.
    <num_levels> The number of contour levels.
    <contour_levels> The array of contour levels.
    <dash> The array of dash flags. This may be NULL.
    <linewidths> The array of line widths. This may be NULL.
    <negative_colour> The colour to use for negative levels.
    <positive_colour> The colour to use for positive levels.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas;
    unsigned int count;
    unsigned long negative_pixel_value, positive_pixel_value;
    unsigned long pixel_values[MAX_CONTOUR_LEVELS];

    canvas = contour_get_worldcanvas (cimage);
    canvas_get_colour (canvas, negative_colour, &negative_pixel_value,
		       NULL, NULL, NULL);
    canvas_get_colour (canvas, positive_colour, &positive_pixel_value,
		       NULL, NULL, NULL);
    for (count = 0; count < num_levels; ++count)
    {
	pixel_values[count] = dash[count] ?
	    negative_pixel_value : positive_pixel_value;
    }
    contour_set_levels (cimage, num_levels, contour_levels, pixel_values, NULL,
			linewidths);
}   /*  End Function set_clevels  */

static void reset_levels_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  [SUMMARY] Contour levels reset callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    struct dataset_type *set;
    /*static char function_name[] = "View2DatasetsWidget::reset_levels_cbk";*/

    get_sets (top, NULL, &set, NULL, NULL);
    if (set == NULL)
    {
	XBell (XtDisplay (w), 100);
	return;
    }
    strcpy (set->levels, "%10:90+10");
    XtVaSetValues (top->view2Datasets.levels_dlg,
		   XtNvalue, set->levels,
		   NULL);
    apply_levels_cbk (w, top, NULL);
}   /*  End Function reset_levels_cbk  */

static void auto_apply_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  This is the auto apply callback.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;

    top->view2Datasets.auto_apply_state = (uaddr) call_data;
}   /*  End Function auto_apply_cbk   */

static void region_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the region callback. This is called when the intensity scaling for
    a dataset is changed.
*/
{
    flag visible = FALSE;
    unsigned int set_number = 0;
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    Bool link, blink;
    DataclipRegions *regions = (DataclipRegions *) call_data;
    struct dataset_type *set = NULL;
    struct dataset_type *other_set = NULL;
    static char function_name[] = "View2DatasetsWidget::region_cbk";

    if (regions->num_regions != 1)
    {
	fprintf (stderr, "num_regions: %u is not 1\n", regions->num_regions);
	a_prog_bug (function_name);
    }
    XtVaGetValues (top->view2Datasets.link_iscale_tgl, XtNstate, &link, NULL);
    if (link) visible = TRUE;
    if (w == NULL)
    {
	get_sets (top, &set, NULL, NULL, &other_set);
	visible = TRUE;
    }
    else
    {
	if (XtNameToWidget (top->view2Datasets.image_display, "izoomwinpopup0")
	    == w)
	{
	    set = &top->view2Datasets.set0;
	    other_set = &top->view2Datasets.set1;
	}
	else
	{
	    set = &top->view2Datasets.set1;
	    other_set = &top->view2Datasets.set0;
	    set_number = 1;
	}
    }
    set_iscale_range (set, regions->minima[0], regions->maxima[0], FALSE);
    if ( link && (other_set->array_pseudo != NULL) )
    {
	double scale, offset;

	iarray_get_data_scaling (set->array_pseudo, &scale, &offset);
	set_iscale_range (other_set,
			  regions->minima[0] * scale + offset,
			  regions->maxima[0] * scale + offset,
			  TRUE);
    }
    XtVaGetValues (top->view2Datasets.blink_tgl, XtNstate, &blink, NULL);
    switch (top->view2Datasets.mode)
    {
      case MODE_IMAGE0:
      case MODE_IMAGE0_CONTOUR1:
	if (set_number == 0) visible = TRUE;
	break;
      case MODE_IMAGE1:
      case MODE_IMAGE1_CONTOUR0:
	if (set_number == 1) visible = TRUE;
	break;
      case MODE_BLINK:
	if ( (top->view2Datasets.set0.multi_desc == NULL) ||
	     (top->view2Datasets.set1.multi_desc == NULL) ) visible = TRUE;
	else if (set_number == blink ? 1 : 0) visible = TRUE;
	break;
    }
    if (visible)
    {
	Widget magnifier;

	XkwImageDisplayRefresh (top->view2Datasets.image_display, FALSE);
	XtVaGetValues (top->view2Datasets.image_display,
		       XkwNmagnifier, &magnifier,
		       NULL);
	XkwSlaveImageDisplayPopupRefresh (magnifier, FALSE);
    }
}   /*  End Function region_cbk   */

static void _View2Datasets_intensity_range_cbk (Widget w,
						XtPointer client_data,
						XtPointer call_data)
{
    DataclipRegions region;
    double *range = call_data;

    region.num_regions = 1;
    region.minima = range;
    region.maxima = range + 1;
    region_cbk (NULL, client_data, &region);
}   /*  End Function intensity_range_cbk  */

static void display_changes_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
/*  This is the display changes callback.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;

    _View2Datasets_display_changes (top, FALSE);
}   /*  End Function display_changes_cbk   */

static void _View2Datasets_text_cbk (Widget w, XtPointer client_data,
				     XtPointer call_data)
/*  This is the title text callback.
*/
{
    Bool auto_title;
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;

    XtVaGetValues (top->view2Datasets.auto_title_tgl,
		   XtNstate, &auto_title,
		   NULL);
    if (auto_title) return;
    _View2Datasets_display_changes (top, FALSE);
}   /*  End Function text_cbk   */

static void beam_toggle_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the beam toggle callback.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;

    XkwImageDisplayRefresh (top->view2Datasets.image_display, FALSE);
}   /*  End Function beam_toggle_cbk   */

static unsigned long get_green (KPixCanvas canvas)
/*  [SUMMARY] Get green pixel colour.
    <canvas> The pixel canvas.
    [RETURNS] The pixel value on success, else the white pixel value on failure
*/
{
    unsigned int visual;
    unsigned long pixel_value;

    kwin_get_attributes (canvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    if (visual == KWIN_VISUAL_PSEUDOCOLOUR)
    {
	if ( !kwin_get_colour (canvas, "green", &pixel_value,
			       NULL, NULL, NULL) )
	{
	    /*  Can't get green: try white instead  */
	    if ( !kwin_get_colour (canvas, "white", &pixel_value,
				   NULL, NULL, NULL) )
	    {
		/*  Can't even get white (this should never actually happen
		    with X), so just pick 0  */
		pixel_value = 0;
	    }
	}
    }
    else
    {
	kwin_get_attributes (canvas,
			     KWIN_ATT_PIX_GREEN_MASK, &pixel_value,
			     KWIN_ATT_END);
    }
    return (pixel_value);
}   /*  End Function get_green  */

static void all_realised_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  This is the all realised callback.
*/
{
    KWorldCanvas pseudo_canvas, direct_canvas, true_canvas;
    KPixCanvas pixcanvas;
    unsigned long pixel_value;
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    KWorldCanvas *canvases = (KWorldCanvas *) call_data;
    struct edit_context_list *new_ctx;
    static char function_name[] = "View2DatasetsWidget::all_realised_cbk";

    for (; *canvases != NULL; ++canvases)
    {
	canvas_register_refresh_func (*canvases, canvas_refresh_func, top);
	pixcanvas = canvas_get_pixcanvas (*canvases);
	pixel_value = get_green (pixcanvas);
	/*  Create the dragzoom editor context  */
	if ( ( new_ctx = m_alloc (sizeof *new_ctx) ) == NULL )
	    m_abort (function_name, "editor context list entry");
	new_ctx->context =
	    kwin_editor_create_context (pixcanvas,
					KWIN_EDITOR_CONTEXT_ATT_MOUSE,
					K_CANVAS_EVENT_MIDDLE_MOUSE_MASK,
					KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
					KWIN_EDIT_OBJECT_NONE,
					KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE,
					pixel_value,
					KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
					FALSE,
					KWIN_EDITOR_CONTEXT_ATT_END);
	if (new_ctx->context == NULL) m_abort (function_name,"editor context");
	kwin_editor_register_event_func (new_ctx->context,
					 _View2Datasets_geom_func, top);
	new_ctx->next = top->view2Datasets.first_edit_context;
	top->view2Datasets.first_edit_context = new_ctx;
    }
    XtVaGetValues (top->view2Datasets.image_display,
		   XkwNpseudoColourCanvas, &pseudo_canvas,
		   XkwNdirectColourCanvas, &direct_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   NULL);
    if (pseudo_canvas != NULL)
	viewimg_register_position_event_func
	    (pseudo_canvas, _View2Datasets_viewimg_position_func, top);
    if (direct_canvas != NULL)
	viewimg_register_position_event_func
	    (direct_canvas, _View2Datasets_viewimg_position_func, top);
    if (true_canvas != NULL)
	viewimg_register_position_event_func
	    (true_canvas, _View2Datasets_viewimg_position_func, top);
}   /*  End Function all_realised_cbk   */

static void refresh_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the refresh callback.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;

    XkwImageDisplayRefresh (top->view2Datasets.image_display, FALSE);
}   /*  End Function refresh_cbk   */

static void slider_realise_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
/*  This is the slider realise callback.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;

    XtVaSetValues (top->view2Datasets.levels_dlg, XtNvalue, "%10:90+10", NULL);
}   /*  End Function slider_realise_cbk   */

static void canvas_refresh_func (KWorldCanvas canvas, int width, int height,
				 struct win_scale_type *win_scale,
				 Kcolourmap cmap, flag cmap_resize,void **info,
				 PostScriptPage pspage,
				 unsigned int num_areas,
				 KPixCanvasRefreshArea *areas,
				 flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KwcsAstro ap;
    unsigned int num_restr;
    View2DatasetsWidget top = (View2DatasetsWidget) *info;
    Bool show_beam, show_beam_name;
    struct dataset_type *image_set;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    /*static char function_name[]="View2DatasetsWidget::canvas_refresh_func";*/

    XtVaGetValues (top->view2Datasets.show_beam_tgl,
		   XtNstate, &show_beam,
		   NULL);
    XtVaGetValues (top->view2Datasets.show_beam_name_tgl,
		   XtNstate, &show_beam_name,
		   NULL);
    if (!show_beam) return;
    get_sets (top, &image_set, NULL, NULL, NULL);
    if ( ( ap = wcs_astro_get_associated (image_set->multi_desc,
					  image_set->array_index) ) == NULL )
	return;
    canvas_get_specification (canvas, &xlabel, &ylabel,
			      &num_restr, &restr_names, &restr_values);
    if ( !wcs_astro_test_if_lon (ap, xlabel) ) return;
    if ( !wcs_astro_test_if_lat (ap, ylabel) ) return;
    draw_beam (canvas, &top->view2Datasets.set0, show_beam_name);
    draw_beam (canvas, &top->view2Datasets.set1, show_beam_name);
}   /*  End Function canvas_refresh_func  */

static flag _View2Datasets_viewimg_position_func (ViewableImage vimage,
						  double x, double y,
						  void *value,
						  unsigned int event_code,
						  void *e_info, void **f_info,
						  double x_lin, double y_lin,
						  unsigned int value_type)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transform_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    Bool bool;
    int frame, num_frames;
    unsigned int old_event, obj;
    View2DatasetsWidget top = (View2DatasetsWidget) *f_info;
    char *key;
    /*static char function_name[] =
      "View2DatasetsWidget::viewimg_position_func";*/

    old_event = top->view2Datasets.last_event_code;
    top->view2Datasets.last_event_code = event_code;
    kwin_editor_get_attributes (top->view2Datasets.first_edit_context->context,
				KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT, &obj,
				KWIN_EDITOR_CONTEXT_ATT_END);
    switch (event_code)
    {
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
	if (obj != KWIN_EDIT_OBJECT_NONE) return (FALSE);
	XtVaGetValues (top->view2Datasets.middle_zoom_tgl,
		       XtNstate, &bool,
		       NULL);
	if (!bool) return (FALSE);
	if (old_event != K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK) return (FALSE);
	viewimg_zoom (vimage, x_lin, y_lin, FALSE, 2.0);
	canvas_warp_pointer (viewimg_get_worldcanvas (vimage),
			     x_lin, y_lin, TRUE, TRUE);
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
	if (old_event != K_CANVAS_EVENT_RIGHT_MOUSE_CLICK) return (FALSE);
	XtVaGetValues (top->view2Datasets.right_zoom_tgl,
		       XtNstate, &bool,
		       NULL);
	if (!bool) return (FALSE);
	viewimg_zoom (vimage, 0.5, 0.5, TRUE, 0.5);
	canvas_warp_pointer (viewimg_get_worldcanvas (vimage),
			     x_lin, y_lin, TRUE, TRUE);
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	key = e_info;
	if (*key != 'b') return (FALSE);
	XtVaGetValues (top->view2Datasets.blink_tgl, XtNstate, &bool, NULL);
	bool = bool ? FALSE : TRUE;
	XtVaSetValues (top->view2Datasets.blink_tgl, XtNstate, bool, NULL);
	blink_cbk (NULL, top, NULL);
	break;
      case K_CANVAS_EVENT_KEYSYM:
	key = e_info;
	if (strncmp (key, "KP_", 3) == 0) key += 3;
	XtVaGetValues (top->view2Datasets.animatepopup,
		       XkwNcurrentFrame, &frame,
		       XkwNnumFrames, &num_frames,
		       NULL);
	if (num_frames < 2) break;
	if (strcmp (key, "Next") == 0) --frame;                    /*  PgUp  */
	else if (strcmp (key, "Prior") == 0) ++frame;              /*  PgDn  */
	else if (strcmp (key, "Home") == 0) frame = 0;
	else if (strcmp (key, "End") == 0) frame = num_frames - 1;
	else return (FALSE);
	XtVaSetValues (top->view2Datasets.animatepopup,
		       XkwNcurrentFrame, frame,
		       NULL);
	XtVaGetValues (top->view2Datasets.animatepopup,
		       XkwNcurrentFrame, &frame,
		       NULL);
	new_frame_cbk (NULL, top, &frame);
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    return (TRUE);
}   /*  End Function _View2Datasets_viewimg_position_func  */

static void get_sets (View2DatasetsWidget top,
		      struct dataset_type **image_set,
		      struct dataset_type **contour_set,
		      struct dataset_type **hidden_set,
		      struct dataset_type **other_set)
/*  [SUMMARY] Get datasets.
    <top> The View2Datasets widget.
    <image_set> The image dataset is written here. This may be NULL.
    <contour_set> The contour dataset is written here. This may be NULL.
    <hidden_set> The hidden dataset is written here. This may be NULL.
    <other_set> The other dataset (may be contoured or hidden) is written here.
    This may be NULL.
    [RETURNS] Nothing.
*/
{
    unsigned int set_num = 2;
    Bool blink;

    XtVaGetValues (top->view2Datasets.blink_tgl, XtNstate, &blink, NULL);
    if ( (image_set != NULL) || (other_set != NULL) )
    {
	switch (top->view2Datasets.mode)
	{
	  case MODE_IMAGE0:
	  case MODE_IMAGE0_CONTOUR1:
	    set_num = 0;
	    break;
	  case MODE_IMAGE1:
	  case MODE_IMAGE1_CONTOUR0:
	    set_num = 1;
	    break;
	  case MODE_BLINK:
	    set_num = blink ? 1 : 0;
	    if ( (set_num == 0) &&
		 (top->view2Datasets.set0.multi_desc == NULL) )
		set_num = 1;
	    else if ( (set_num == 1) &&
		 (top->view2Datasets.set1.multi_desc == NULL) )
		set_num = 0;
	    break;
	}
	if (image_set != NULL)
	{
	    if (set_num == 0) *image_set = &top->view2Datasets.set0;
	    else *image_set = &top->view2Datasets.set1;
	}
	if (other_set != NULL)
	{
	    if (set_num == 0) *other_set = &top->view2Datasets.set1;
	    else *other_set = &top->view2Datasets.set0;
	}
    }
    if (hidden_set != NULL)
    {
	switch (top->view2Datasets.mode)
	{
	  case MODE_IMAGE0:
	    set_num = 0;
	    break;
	  case MODE_IMAGE1:
	    set_num = 1;
	    break;
	  case MODE_BLINK:
	    set_num = blink ? 1 : 0;
	    break;
	}
	if (set_num == 0) *hidden_set = &top->view2Datasets.set1;
	else *hidden_set = &top->view2Datasets.set0;
    }
    if (contour_set != NULL)
    {
	switch (top->view2Datasets.mode)
	{
	  case MODE_IMAGE0_CONTOUR1:
	    *contour_set = &top->view2Datasets.set1;
	    break;
	  case MODE_IMAGE1_CONTOUR0:
	    *contour_set = &top->view2Datasets.set0;
	    break;
	  default:
	    *contour_set = NULL;
	    break;
	}
    }
}   /*  End Function get_sets  */

static struct direction_type *get_dir (View2DatasetsWidget top,
				       struct dataset_type *set)
/*  [SUMMARY] Get direction for a dataset.
    <top> The View2Datasets widget.
    <set> The dataset.
    [RETURNS] The direction pointer on success, else NULL indicating that there
    is no direction.
*/
{
    struct direction_type *dir = NULL;

    if (set == NULL) return (NULL);
    switch (top->view2Datasets.slice)
    {
      case SLICE_XY:
	dir = &set->xy;
	break;
      case SLICE_XZ:
	dir = &set->xz;
	break;
      case SLICE_ZY:
	dir = &set->zy;
	break;
    }
    return (dir);
}   /*  End Function get_dir  */

static void draw_beam (KWorldCanvas canvas, struct dataset_type *set,flag name)
/*  [SUMMARY] Draw a beam on a canvas.
    <canvas> The world canvas.
    <set> The dataset for the beam.
    <name> If TRUE, the beam name is displayed.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    unsigned int array_num;
    unsigned long pixel_value;
    double left_x, right_x, bottom_y, top_y;
    double cx, cy, x, y, width, height;
    multi_array *multi_desc;
    char *colourname;
    double bmaj, bmin, bpa, value[2];

    if (set->multi_desc == NULL) return;
    pixcanvas = canvas_get_pixcanvas (canvas);
    canvas_get_dressing (canvas,
			 KCD_ATT_LABEL_COLOUR, &colourname,
			 KCD_ATT_END);
    if (colourname == NULL) colourname = "green";
    kwin_get_colour (pixcanvas, colourname, &pixel_value, NULL, NULL, NULL);
    canvas_get_attributes (canvas,
			   CANVAS_ATT_LEFT_X, &left_x,
			   CANVAS_ATT_RIGHT_X, &right_x,
			   CANVAS_ATT_BOTTOM_Y, &bottom_y,
			   CANVAS_ATT_TOP_Y, &top_y,
			   CANVAS_ATT_END);
    multi_desc = set->multi_desc;
    array_num = set->array_index;
    if ( !ds_get_unique_named_value (multi_desc->headers[array_num],
				     multi_desc->data[array_num],
				     "BMAJ", NULL, value) ) return;
    bmaj = value[0] * 0.5;
    if ( !ds_get_unique_named_value (multi_desc->headers[array_num],
				     multi_desc->data[array_num],
				     "BMIN", NULL, value) ) return;
    bmin = value[0] * 0.5;
    if ( !ds_get_unique_named_value (multi_desc->headers[array_num],
				     multi_desc->data[array_num],
				     "BPA", NULL, value) ) return;
    bpa = 90.0 - value[0];
    cx = left_x + set->beam_xpos * (right_x - left_x);
    cy = bottom_y + set->beam_ypos * (top_y - bottom_y);
    canvas_coords_transform (canvas, 1, &cx, FALSE, &cy, FALSE);
    canvas_convert_ellipse_coords (canvas, FALSE, FALSE,
				   &cx, &cy, &bmaj, &bmin, &bpa);
    if (bpa <= -90.0) bpa += 180.0;
    else if (bpa >= 90.0) bpa -= 180.0;
    if (fabs (bmaj - bmin) < 1.0) bpa = 0.0;
    /*  Draw the ellipse now that parameters are in pixel canvas  */
    kwin_draw_arc (pixcanvas, cx, cy, bmaj, bmin, 0, 360, bpa,
		   pixel_value, FALSE);
    if (!name) return;
    kwin_hershey_draw_string (NULL, set->object_name, 0.0, 0.0, bpa, 12.0, 0,
			      &width, &height, NULL, NULL);
    x = cos (bpa * PION180 + PI_ON_2) * (bmin + height);
    y = sin (bpa * PION180 + PI_ON_2) * (bmin + height);
    x -= cos (bpa * PION180) * width * 0.5;
    y -= sin (bpa * PION180) * width * 0.5;
    kwin_hershey_draw_string (pixcanvas, set->object_name, cx + x, cy + y,
			      bpa, 12.0, pixel_value, NULL, NULL, NULL, NULL);
}   /*  End Function draw_beam  */

static flag _View2Datasets_geom_func (KPixCanvasEditorContext context,
				      KPixCanvas canvas,
				      unsigned int object_type,
				      unsigned int num_coords,
				      CONST int *x, CONST int *y,
				      unsigned int raw_event_code,
				      void *raw_e_info, void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    View2DatasetsWidget top = (View2DatasetsWidget) *f_info;
    KWorldCanvas worldcanvas;
    ViewableImage vimage;
    iarray pa;
    flag inverted;
    flag horizontal = FALSE;
    unsigned int hdim, vdim, elem_index, prof_len, count, type, low, high;
    unsigned long hist_peak, hist_mode, pixel_value;
    float sum, wsum, val, tmp;
    double radius_x, radius_y, radius, peak_height, background, scale, offset;
    double imin, imax, px, py, mag_raw, mag_fit;
    array_desc *arr_desc;
    dim_desc *hd, *vd;
    struct dataset_type *set;
    CONST char *array;
    unsigned int dim_indices[2];
    unsigned long hist_array[MAX_HISTOGRAM_BINS];
    uaddr start_indices[2], stop_indices[2];
    double lx[3], ly[3];
    static char function_name[] = "View2DatasetsWidget::geom_func";

    get_sets (top, &set, NULL, NULL, NULL);
    if (raw_event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE) return (FALSE);
    if (set->array_pseudo == NULL) return (FALSE);
    XtVaGetValues (top->view2Datasets.image_display,
		   XkwNvisibleCanvas, &worldcanvas,
		   NULL);
    vimage = viewimg_get_active (worldcanvas);
    if (vimage == NULL) return (FALSE);
    canvas_get_attributes (worldcanvas,
			   CANVAS_ATT_VALUE_MIN, &imin,
			   CANVAS_ATT_VALUE_MAX, &imax,
			   CANVAS_ATT_END);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_SLICE, &array,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_PSEUDO_INDEX, &elem_index,
			    VIEWIMG_VATT_END);
    hd = arr_desc->dimensions[hdim];
    vd = arr_desc->dimensions[vdim];
    switch (top->view2Datasets.profile_mode)
    {
      case PROFILE_MODE_BOX_SUM:
      case PROFILE_MODE_BOX_AVERAGE:
	if (iarray_num_dim (set->array_pseudo) != 3)
	{
	    fprintf (stderr, "Box profile not possible with 2D dataset\n");
	    return (FALSE);
	}
	lx[0] = x[0];
	ly[0] = y[1];
	lx[1] = x[1];
	ly[1] = y[0];
	canvas_convert_to_canvas_coords (worldcanvas, TRUE, 2,
					 lx, ly, lx, ly, NULL, NULL);
	lx[0] = ds_get_coord_num (hd, lx[0], SEARCH_BIAS_CLOSEST);
	lx[1] = ds_get_coord_num (hd, lx[1], SEARCH_BIAS_CLOSEST);
	ly[0] = ds_get_coord_num (vd, ly[0], SEARCH_BIAS_CLOSEST);
	ly[1] = ds_get_coord_num (vd, ly[1], SEARCH_BIAS_CLOSEST);
	/*  NOTE: horizontal is index = 0, vertical is index = 1  */
	/*  Convert linear world co-ordinates to array indices and display  */
	start_indices[0] = lx[0];
	start_indices[1] = ly[0];
	stop_indices[0] = lx[1] + 1.0;
	stop_indices[1] = ly[1] + 1.0;
	switch (top->view2Datasets.slice)
	{
	  case SLICE_XY:
	    dim_indices[0] = 2;
	    dim_indices[1] = 1;
	    break;
	  case SLICE_XZ:
	    dim_indices[0] = 2;
	    dim_indices[1] = 0;
	    break;
	  case SLICE_ZY:
	    dim_indices[0] = 0;
	    dim_indices[1] = 1;
	    break;
	}
	XkwTracePopupShowBoxedTrace (top->view2Datasets.trace_winpopup,
				     dim_indices,
				     (top->view2Datasets.profile_mode ==
				      PROFILE_MODE_BOX_SUM) ? TRUE : FALSE,
				     start_indices, stop_indices);
	break;
      case PROFILE_MODE_RADIAL:
      case PROFILE_MODE_RADIAL_GAUSS:
	lx[0] = x[0];
	ly[0] = y[0];
	lx[1] = x[0] + x[1];
	ly[1] = y[0];
	lx[2] = x[0];
	ly[2] = y[0] + y[1];
	canvas_convert_to_canvas_coords (worldcanvas, FALSE, 3,
					 lx, ly, lx, ly, NULL, NULL);
	ds_convert_coordinates (hd, 3, lx, TRUE, FALSE, FALSE);
	ds_convert_coordinates (vd, 3, ly, TRUE, FALSE, FALSE);
	radius_x = lx[1] - lx[0];
	radius_y = ly[0] - ly[2];
	radius = sqrt (radius_x * radius_x + radius_y * radius_y);
	prof_len = 1.0 * radius;
	if (prof_len < 2)
	{
	    fprintf (stderr, "%s: radius: %u too small\n",
		     function_name, prof_len);
	    return (TRUE);
	}
	array += ds_get_element_offset (arr_desc->packet, elem_index);
	if ( !ds_find_centroid (array,
				arr_desc->packet->element_types[elem_index],
				&background, &inverted,
				hd->length, arr_desc->offsets[hdim],
				lx, radius_x,
				vd->length, arr_desc->offsets[vdim],
				ly, radius_y) ) return (TRUE);
	/*  Convert centroid to screen pixel co-ordinates and draw  */
	px = lx[0];
	py = ly[0];
	ds_convert_coordinates (hd, 1, &px, FALSE, FALSE, FALSE);
	ds_convert_coordinates (vd, 1, &py, FALSE, FALSE, FALSE);
	canvas_convert_from_canvas_coords (worldcanvas, FALSE, TRUE, 1,
					   &px, &py, &px, &py);
	kwin_get_colour (canvas, "red", &pixel_value, NULL, NULL, NULL);
	kwin_draw_line (canvas, px - 1, py, px + 1, py, pixel_value);
	kwin_draw_line (canvas, px, py - 1, px, py + 1, pixel_value);
	if (top->view2Datasets.profile_arr != NULL)
	    iarray_dealloc (top->view2Datasets.profile_arr);
	if ( ( pa = iarray_create_1D (prof_len, K_FLOAT) ) == NULL )
	    m_abort (function_name, "profile array");
	iarray_set_dim_name (pa, 0, "Radius (pixels)", TRUE);
	iarray_set_value_name (pa, iarray_value_name (set->array_pseudo),TRUE);
	top->view2Datasets.profile_arr = pa;
	iarray_get_data_scaling (set->array_pseudo, &scale, &offset);
	background = background * scale + offset;
	compute_radial_points (top, array,
			       arr_desc->packet->element_types[elem_index],
			       hd->length, arr_desc->offsets[hdim],
			       lx[0] + 0.5, radius_x,
			       vd->length, arr_desc->offsets[vdim],
			       ly[0] + 0.5, radius_y, scale, offset);
	azimuthally_average ( (float *) pa->data, prof_len, pa->offsets[0],
			      array,
			      arr_desc->packet->element_types[elem_index],
			      hd->length, arr_desc->offsets[hdim],
			      lx[0], radius_x,
			      vd->length, arr_desc->offsets[vdim],
			      ly[0], radius_y, 0.0, scale, offset);
	mag_raw =
	    compute_magnitude (array,
			       arr_desc->packet->element_types[elem_index],
			       hd->length, arr_desc->offsets[hdim],
			       lx[0], radius_x,
			       vd->length, arr_desc->offsets[vdim],
			       ly[0], radius_y, scale, offset,
			       background, inverted);
	if (top->view2Datasets.profile_mode == PROFILE_MODE_RADIAL_GAUSS)
	{
	    if ( !ds_fitgauss ( (float *) pa->data, pa->offsets[0],
				iarray_dim_length (pa, 0), background,
				inverted, 0.0, &peak_height,
				&top->view2Datasets.fit_fwhm, &mag_fit) )
	    {
		fprintf (stderr, "Error fitting gaussian\n");
		return (TRUE);
	    }
	    fprintf (stderr,
		     "Raw peak Raw mag  Fit peak Fit mag  Background    FWHM  Cx      Cy\n");
	    fprintf (stderr,
		     "%-8g %-7.2f  %-8g %-7.2f  %-12g  %-5.2f %-7.1f %-7.1f\n",
		     peak_height, mag_raw, peak_height - background, mag_fit,
		     background, top->view2Datasets.fit_fwhm, lx[0], ly[0]);
	    _View2Datasets_compute_fit_data (top, 0.0, peak_height, background,
					     inverted);
	}
	XkwTracePopupNewArray (top->view2Datasets.trace_winpopup, pa, 1.0,0.0);
	XkwTracePopupShowTrace (top->view2Datasets.trace_winpopup, NULL, NULL);
	break;
      case PROFILE_MODE_BOXED_HORIZONTAL:
	horizontal = TRUE;
	/*  Fall through  */
      case PROFILE_MODE_BOXED_VERTICAL:
	lx[0] = x[0];
	ly[0] = y[1];
	lx[1] = x[1];
	ly[1] = y[0];
	canvas_convert_to_canvas_coords (worldcanvas, TRUE, 2,
					 lx, ly, lx, ly, NULL, NULL);
	lx[0] = ds_get_coord_num (hd, lx[0], SEARCH_BIAS_CLOSEST);
	lx[1] = ds_get_coord_num (hd, lx[1], SEARCH_BIAS_CLOSEST);
	ly[0] = ds_get_coord_num (vd, ly[0], SEARCH_BIAS_CLOSEST);
	ly[1] = ds_get_coord_num (vd, ly[1], SEARCH_BIAS_CLOSEST);
	array += ds_get_element_offset (arr_desc->packet, elem_index);
	if (top->view2Datasets.profile_arr != NULL)
	    iarray_dealloc (top->view2Datasets.profile_arr);
	if (horizontal)
	    prof_len = (unsigned int) lx[1] - (unsigned int) lx[0] + 1;
	else
	    prof_len = (unsigned int) ly[1] - (unsigned int) ly[0] + 1;
	if ( ( pa = iarray_create_1D (prof_len, K_FLOAT) ) == NULL )
	    m_abort (function_name, "profile array");
	iarray_set_value_name (pa, iarray_value_name (set->array_pseudo),TRUE);
	iarray_set_dim_name (pa, 0, "Offset (pixels)", TRUE);
	top->view2Datasets.profile_arr = pa;
	iarray_get_data_scaling (set->array_pseudo, &scale, &offset);
	hist_peak = 0;
	hist_mode = 0;
	m_clear (hist_array, sizeof hist_array);
	type = arr_desc->packet->element_types[elem_index];
	if (horizontal)
	{
	    if ( !collapse_dim ( (float *) pa->data, pa->offsets[0],
				 array, type,
				 lx[0], lx[1], arr_desc->offsets[hdim],
				 ly[0], ly[1], arr_desc->offsets[vdim],
				 scale, offset ) ) return (TRUE);
	    count = (unsigned int) ly[1] - (unsigned int) ly[0] + 1;
	    ds_find_single_histogram (array +
				      arr_desc->offsets[hdim][ (int) lx[0] ],
				      type, CONV_CtoR_REAL, count,
				      arr_desc->offsets[vdim] + (int) ly[0], 0,
				      imin, imax, MAX_HISTOGRAM_BINS,
				      hist_array, &hist_peak, &hist_mode);
	    ds_find_single_histogram (array +
				      arr_desc->offsets[hdim][ (int) lx[1] ],
				      type, CONV_CtoR_REAL, count,
				      arr_desc->offsets[vdim] + (int) ly[0], 0,
				      imin, imax, MAX_HISTOGRAM_BINS,
				      hist_array, &hist_peak, &hist_mode);
	}
	else
	{
	    if ( !collapse_dim ( (float *) pa->data, pa->offsets[0],
				 array, type,
				 ly[0], ly[1], arr_desc->offsets[vdim],
				 lx[0], lx[1], arr_desc->offsets[hdim],
				 scale, offset ) ) return (TRUE);
	    count = (unsigned int) lx[1] - (unsigned int) lx[0] + 1;
	    ds_find_single_histogram (array +
				      arr_desc->offsets[vdim][ (int) ly[0] ],
				      type, CONV_CtoR_REAL, count,
				      arr_desc->offsets[hdim] + (int) lx[0], 0,
				      imin, imax, MAX_HISTOGRAM_BINS,
				      hist_array, &hist_peak, &hist_mode);
	    ds_find_single_histogram (array +
				      arr_desc->offsets[vdim][ (int) ly[1] ],
				      type, CONV_CtoR_REAL, count,
				      arr_desc->offsets[hdim] + (int) lx[0], 0,
				      imin, imax, MAX_HISTOGRAM_BINS,
				      hist_array, &hist_peak, &hist_mode);
	}
	background = (double) hist_mode / MAX_HISTOGRAM_BINS;
	background = imin + (imax - imin) * background;
	background = background * scale + offset;
	/*  Find the peak in the profile  */
	for (count = 0, low = 0, val = -TOOBIG; count < prof_len; ++count)
	{
	    tmp = F1 (pa, count);
	    if ( DOUBLE_IS_VALID (tmp) )
	    {
		if (tmp > val)
		{
		    val = tmp;
		    low = count;
		}
	    }
	}
	val -= background;
	high = low;
	/*  Find the lower 25% point  */
	while ( (low > 0) && (F1 (pa, low) - background > val * 0.25) ) --low;
	/*  Find the upper 25% point  */
	while ( (high < prof_len - 1) &&
		(F1 (pa, high) - background > val * 0.25) ) ++high;
	/*  Now find centre: use a weighted mean algorithm  */
	for (count = low, sum = 0.0, wsum = 0.0; count <= high; ++count)
	{
	    val = F1 (pa, count) - background;
	    sum += val;
	    wsum += val * (float) count;
	}
	wsum = wsum / sum;
	if (horizontal)
	    iarray_set_world_coords (pa, 0, -wsum, lx[1] - lx[0] - wsum);
	else
	    iarray_set_world_coords (pa, 0, -wsum, ly[1] - ly[0] - wsum);
	/*  TODO: remove assumption that profile is positive  */
	if ( !ds_fitgauss ( (float *) pa->data, pa->offsets[0],
			    iarray_dim_length (pa, 0), background, FALSE,
			    wsum, &peak_height, &top->view2Datasets.fit_fwhm,
			    &mag_fit) )
	{
	    fprintf (stderr, "Error fitting gaussian\n");
	    return (TRUE);
	}
	fprintf (stderr,
		 "Raw peak  Peak      Background Opt. mag. FWHM  C%c\n",
		 horizontal ? 'h' : 'v');
	fprintf (stderr, "%-9g %-9g %-10g %-9g %-5.2f %-7.1f\n",
		 peak_height, peak_height - background, background, mag_fit,
		 top->view2Datasets.fit_fwhm,
		 floor (horizontal ? lx[0] : ly[0]) + wsum);
	_View2Datasets_compute_fit_data (top, wsum, peak_height, background,
					 FALSE);
	XkwTracePopupNewArray (top->view2Datasets.trace_winpopup, pa, 1.0,0.0);
	XkwTracePopupShowTrace (top->view2Datasets.trace_winpopup, NULL, NULL);
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    return (TRUE);
}   /*  End Function geom_func  */

static void _View2Datasets_compute_fit_data (View2DatasetsWidget top,
					     double centre, double peak_height,
					     double background, flag inverted)
/*  [SUMMARY] Allocate and compute the Gaussian fit data.
    <top> The View2Datasets widget.
    <centre> The centre of the profile.
    <peak_height> The height of the fitted peak.
    <background> The background level.
    <inverted> If TRUE the object values are inverted (i.e. is negative) else
    the object is positive.
    [RETURNS] Nothing. On failure the process aborts.
*/
{
    unsigned int count;
    double start, scale;
    static char function_name[] = "View2Datasets::compute_fit_data";

    if (top->view2Datasets.gausscurve_xarr == NULL)
    {
	if ( ( top->view2Datasets.gausscurve_xarr =
	       m_alloc (sizeof *top->view2Datasets.gausscurve_xarr *
			GAUSSCURVE_LENGTH) ) == NULL )
	{
	    m_abort (function_name, "x array");
	}
	if ( ( top->view2Datasets.gausscurve_yarr =
	       m_alloc (sizeof *top->view2Datasets.gausscurve_yarr *
			GAUSSCURVE_LENGTH) ) == NULL )
	{
	    m_abort (function_name, "y array");
	}
    }
    /*  Set up the horizontal co-ordinates  */
    iarray_get_world_coords (top->view2Datasets.profile_arr, 0, &start,&scale);
    scale = (scale - start) / (double) (GAUSSCURVE_LENGTH - 1);
    for (count = 0; count < GAUSSCURVE_LENGTH; ++count)
	top->view2Datasets.gausscurve_xarr[count] =
	    start + scale * (double) count;
    ds_gausscurve (top->view2Datasets.gausscurve_yarr, NULL,
		   GAUSSCURVE_LENGTH, centre / scale,
		   peak_height, top->view2Datasets.fit_fwhm / scale,
		   background, inverted);
}   /*  End Function compute_fit_data  */

static void compute_radial_points (View2DatasetsWidget top,
				   CONST char *in, unsigned int in_type,
				   uaddr xlen, CONST uaddr *xoffsets,
				   signed int xcentre, signed int xradius,
				   uaddr ylen, CONST uaddr *yoffsets,
				   signed int ycentre, signed int yradius,
				   float scale, float offset)
/*  [SUMMARY] Compute a list of radial profile points.
    <top> The View2Datasets widget.
    <in> The input image.
    <in_type> The type of the input image.
    <xlen> The horizontal image width.
    <xoffsets> The array of horizontal byte offsets
    <xcentre> The horizontal ellipse centre. 0.0 is the centre of the origin.
    <xradius> The horizontal radius of the ellipse.
    <ylen> The vertical image width.
    <yoffsets> The array of vertical byte offsets
    <ycentre> The vertical ellipse centre. 0.0 is the centre of the origin.
    <yradius> The vertical radius of the ellipse.
    <scale> The data scale.
    <offset> The data offset.
    [RETURNS] Nothing.
*/
{
    signed int x, y, xstart, xstop;
    unsigned int tpoints = (xradius * 2 + 1) * (yradius * 2 + 1);
    CONST char *ptr;
    static char function_name[] = "compute_radial_points";

    if (top->view2Datasets.profile_points.buflen < tpoints)
    {
	if (top->view2Datasets.profile_points.x != NULL)
	    m_free (top->view2Datasets.profile_points.x);
	if (top->view2Datasets.profile_points.y != NULL)
	    m_free (top->view2Datasets.profile_points.y);
	if ( ( top->view2Datasets.profile_points.x =
	       m_alloc (sizeof *top->view2Datasets.profile_points.x *
			tpoints) )
	     == NULL )
	    m_abort (function_name, "X profile array");
	if ( ( top->view2Datasets.profile_points.y =
	       m_alloc (sizeof *top->view2Datasets.profile_points.y *
			tpoints) )
	     == NULL )
	    m_abort (function_name, "Y profile array");
	top->view2Datasets.profile_points.buflen = tpoints;
    }
    xstart = xcentre - xradius;
    xstop = xcentre + xradius + 1;
    tpoints = 0;
    for (y = ycentre - yradius; y < ycentre + yradius + 1; ++y)
	for (x = xstart; x < xstop; ++x)
	{
	    double value[2];

	    if ( (x < 0) || (x >= xlen) || (y < 0) || (y >= ylen) ) continue;
	    ptr = in + xoffsets[x] + yoffsets[y];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) ) continue;
	    top->view2Datasets.profile_points.x[tpoints] =
		sqrt ( (xcentre - x) * (xcentre - x) +
		       (ycentre - y) * (ycentre - y) );
	    top->view2Datasets.profile_points.y[tpoints] =
		value[0] * scale + offset;
	    ++tpoints;
	}
    top->view2Datasets.profile_points.num = tpoints;
}   /*  End Function compute_radial_points  */

static void azimuthally_average (float *out, uaddr out_len,
				 CONST uaddr *ooffsets,
				 CONST char *in, unsigned int in_type,
				 uaddr xlen, CONST uaddr *xoffsets,
				 double xcentre, double xradius,
				 uaddr ylen, CONST uaddr *yoffsets,
				 double ycentre, double yradius,
				 double rotation,
				 float scale, float offset)
/*  [SUMMARY] Azimuthally average an image.
    <out> The output array.
    <out_len> The length of the output array.
    <ooffsets> The byte offsets for the output array.
    <in> The input image.
    <in_type> The type of the input image.
    <xlen> The horizontal image width.
    <xoffsets> The array of horizontal byte offsets
    <xcentre> The horizontal ellipse centre. 0.0 is the centre of the origin.
    <xradius> The horizontal radius of the ellipse.
    <ylen> The vertical image width.
    <yoffsets> The array of vertical byte offsets
    <ycentre> The vertical ellipse centre. 0.0 is the centre of the origin.
    <yradius> The vertical radius of the ellipse.
    <rotation> The rotation of the ellipse.
    <scale> The data scale.
    <offset> The data offset.
    [RETURNS] Nothing.
*/
{
    uaddr rcount, acount, num_angle_steps;
    float val = 0.0;
    double radius, radius_step, angle, angle_step;
    double xmax, ymax, cos_rot, sin_rot, cos_ang, sin_ang;
    CONST char *ptr;
    char *out_ptr = (char *) out;
    static unsigned int count_arr_len = 0;
    static unsigned int *count_arr = NULL;
    static char function_name[] = "azimuthally_average";

#if 0
    xcentre += 0.5;
    ycentre += 0.5;
#endif
    if (count_arr_len < out_len)
    {
	if (count_arr != NULL) m_free (count_arr);
	if ( ( count_arr = m_alloc (sizeof *count_arr * out_len) ) == NULL )
	{
	    m_abort (function_name, "count array");
	}
	count_arr_len = out_len;
    }
    for (rcount = 0; rcount < out_len; ++rcount)
    {
	*(float *) (out_ptr + ooffsets[rcount]) = 0.0;
	count_arr[rcount] = 0;
    }
    radius_step = 1.0 / (double) (out_len - 1);
    num_angle_steps = TWOPI * sqrt (xradius * xradius + yradius * yradius);
    angle_step = TWOPI / (double) num_angle_steps;
    xmax = xlen - 1;
    ymax = ylen - 1;
    cos_rot = cos (rotation * PION180);
    sin_rot = sin (rotation * PION180);
    for (acount = 0, angle = 0.0; acount < num_angle_steps;
	 ++acount, angle += angle_step)
    {
	cos_ang = cos (angle);
	sin_ang = sin (angle);
	for (rcount = 0, radius = 0.0; rcount < out_len;
	     ++rcount, radius += radius_step)
	{
	    int toobig_count;
	    float xg, yg, x0, x1, y0, y1, dx, dy, dx_r, dy_r;
	    float val, val_00, val_10, val_01, val_11;
	    float one = 1.0;
	    double xp, yp;
	    double zero = 0.0;
	    double value[2];

	    xp = xradius * radius * cos_ang;
	    yp = yradius * radius * sin_ang;
	    xg = xcentre + xp * cos_rot - yp * sin_rot;
	    yg = ycentre + yp * cos_rot + xp * sin_rot;
	    if ( (xg < zero) || (xg > xmax) || (yg < zero) || (yg > ymax) )
		continue;
	    /*  Take contribution from 4 neighbouring points  */
	    toobig_count = 0;
	    x0 = floor (xg);
	    x1 = ceil (xg);
	    dx = xg - x0;
	    dx_r = one - dx;
	    y0 = floor (yg);
	    y1 = ceil (yg);
	    dy = yg - y0;
	    dy_r = one - dy;
	    ptr = in + xoffsets[(int) x0] + yoffsets[(int) y0];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) )
	    {
		val_00 = zero;
		++toobig_count;
	    }
	    else val_00 = value[0];
	    ptr = in + xoffsets[(int) x1] + yoffsets[(int) y0];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) )
	    {
		val_10 = zero;
		++toobig_count;
	    }
	    else val_10 = value[0];
	    ptr = in + xoffsets[(int) x0] + yoffsets[(int) y1];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) )
	    {
		val_01 = zero;
		++toobig_count;
	    }
	    else val_01 = value[0];
	    ptr = in + xoffsets[(int) x1] + yoffsets[(int) y1];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) )
	    {
		val_11 = zero;
		++toobig_count;
	    }
	    else val_11 = value[0];
	    if (toobig_count > 3) continue;
	    val = val_00 * dx_r * dy_r + val_10 * dx * dy_r +
		val_01 * dx_r * dy + val_11 * dx * dy;
	    if ( DOUBLE_IS_BLANK (val) ) continue;
	    *(float *) (out_ptr + ooffsets[rcount]) += val;
	    ++count_arr[rcount];
	}
    }
    /*  Average computation  */
    for (rcount = 0; rcount < out_len; ++rcount)
    {
	if (count_arr[rcount] < 1) val = TOOBIG;
	else
	{
	    val = *(float *) (out_ptr + ooffsets[rcount]) /
		(float) count_arr[rcount];
	    val = val * scale + offset;
	}
	*(float *) (out_ptr + ooffsets[rcount]) = val;
    }
    if ( DOUBLE_IS_BLANK (val) ) fprintf (stderr, "Trailing blank(s)\n");
}   /*  End Function azimuthally_average  */

static double compute_magnitude (CONST char *in, unsigned int in_type,
				 uaddr xlen, CONST uaddr *xoffsets,
				 double xcentre, double xradius,
				 uaddr ylen, CONST uaddr *yoffsets,
				 double ycentre, double yradius,
				 double scale, double offset,
				 double background, flag inverted)
/*  [SUMMARY] Compute the magnitude in an ellipse.
    <in> The input image.
    <in_type> The type of the input image.
    <xlen> The horizontal image width.
    <xoffsets> The array of horizontal byte offsets
    <xcentre> The horizontal ellipse centre. 0.0 is the centre of the origin.
    <xradius> The horizontal radius of the ellipse.
    <ylen> The vertical image width.
    <yoffsets> The array of vertical byte offsets
    <ycentre> The vertical ellipse centre. 0.0 is the centre of the origin.
    <yradius> The vertical radius of the ellipse.
    <scale> The data scale.
    <offset> The data offset.
    <background> The background value. This must be in scaled units.
    <inverted> The value TRUE will be written here if the image values are
    inverted (i.e. the object of interested is in a local minimum rather than a
    maximum).
    [RETURNS] The magnitude.
*/
{
    iaddr x, y;
    unsigned int num_values = 0;
    double distance, tmp, xmax, ymax;
    double magnitude = 0.0;
    double zero = 0.0;
    CONST char *ptr;
    double value[2];
    /*static char function_name[] = "compute_magnitude";*/

    xcentre += 0.5;
    ycentre += 0.5;
    xmax = xlen - 1;
    ymax = ylen - 1;
    for (y = ycentre - (iaddr) yradius; y < ycentre + (iaddr) yradius; ++y)
    {
	for (x = xcentre - (iaddr) xradius; x < xcentre + (iaddr) xradius; ++x)
	{
	    if ( (x < zero) || (x > xmax) || (y < zero) || (y > ymax) )
		continue;
	    tmp = x - xcentre;
	    distance = tmp * tmp / (xradius * xradius);
	    tmp = y - ycentre;
	    distance += tmp * tmp / (yradius * yradius);
	    if (distance > 1.0) continue;
	    ptr = in + xoffsets[(int) x] + yoffsets[(int) y];
	    ds_get_element (ptr, in_type, value, NULL);
	    if ( DOUBLE_IS_BLANK (value[0]) ) continue;
	    magnitude += value[0] * scale + offset - background;
	    ++num_values;
	}
    }
    if (inverted) magnitude = -magnitude;
    return 30.0 - 2.5 * log10 (magnitude);
}   /*  End Function compute_magnitude  */

static flag collapse_dim (float *out, CONST uaddr *ooffsets,
			  CONST char *in, unsigned int in_type,
			  uaddr start_x, uaddr end_x, CONST uaddr *xoffsets,
			  uaddr start_y, uaddr end_y, CONST uaddr *yoffsets,
			  float scale, float offset)
/*  [SUMMARY] Collapse a dimension along Y.
    <out> The output array.
    <ooffsets> The byte offsets for the output array.
    <in> The input image.
    <in_type> The type of the input image.
    <start_x> The starting index along X.
    <end_x> The end index along X.
    <xoffsets> The array of horizontal byte offsets
    <start_y> The starting index along Y.
    <end_y> The end index along Y.
    <yoffsets> The array of vertical byte offsets
    <scale> The data scale.
    <offset> The data offset.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag some_data = FALSE;
    uaddr x, y, xo, count;
    float sum, val;
    double dval;
    CONST char *ptr;
    char *out_ptr = (char *) out;
    double value[2];
    static char function_name[] = "View2DatasetsWidget::collapse_dim";

    for (x = start_x, xo = 0; x <= end_x; ++x, ++xo)
    {
	sum = 0.0;
	count = 0;
	ptr = in + xoffsets[x];
	for (y = start_y; y <= end_y; ++y)
	{
	    ds_get_element (ptr + yoffsets[y], in_type, value, NULL);
	    dval = value[0];
	    if ( DOUBLE_IS_VALID (dval) )
	    {
		some_data = TRUE;
		sum += (float) dval;
		++count;
	    }
	}
	if (count < 1) val = TOOBIG;
	else val = sum / (float) count * scale + offset;
	*(float *) (out_ptr + ooffsets[xo]) = val;
    }
    if (some_data) return (TRUE);
    fprintf (stderr, "%s: all blanks!\n", function_name);
    return (FALSE);
}   /*  End Function collapse_dim  */

static void create_tracewinpopup (View2DatasetsWidget top)
/*  [SUMMARY] Create the tracewinpopup widget.
    <top> The View2Datasets widget.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas;
    Widget w;

    if (top->view2Datasets.trace_winpopup != NULL) return;
    w = XtVaCreatePopupShell ("tracewinpopup", tracePopupWidgetClass,
			      (Widget) top,
			      XtNtitle, "Profile Window",
			      NULL);
    top->view2Datasets.trace_winpopup = w;
    XtRealizeWidget (w);
    XtVaGetValues (w, XkwNworldCanvas, &canvas, NULL);
    canvas_register_refresh_func (canvas, trace_refresh_func, top);
    canvas_register_position_event_func (canvas, trace_position_func, top);
}   /*  End Function create_tracewinpopup  */

static void trace_refresh_func (KWorldCanvas canvas, int width, int height,
				struct win_scale_type *win_scale,
				Kcolourmap cmap, flag cmap_resize, void **info,
				PostScriptPage pspage,
				unsigned int num_areas,
				KPixCanvasRefreshArea *areas,
				flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas = canvas_get_pixcanvas (canvas);
    int x_offset, x_pixels, y_offset;
    unsigned long pixel_value;
    double string_width, string_height;
    View2DatasetsWidget top = (View2DatasetsWidget) *info;
    char *label_colour;
    char txt[STRING_LENGTH];

    switch (top->view2Datasets.profile_mode)
    {
      case PROFILE_MODE_RADIAL:
      case PROFILE_MODE_RADIAL_GAUSS:
	if (top->view2Datasets.profile_points.num > 0)
	{
	    kwin_get_colour (pixcanvas, "white", &pixel_value,
			     NULL, NULL, NULL);
	    graph1d_points (canvas,
			    top->view2Datasets.profile_points.num,
			    top->view2Datasets.profile_points.x,
			    top->view2Datasets.profile_points.y,
			    pixel_value);
	}
	/*  Fall through  */
      case PROFILE_MODE_BOXED_HORIZONTAL:
      case PROFILE_MODE_BOXED_VERTICAL:
	break;
      default:
	return;
	/*break;*/
    }
    if (top->view2Datasets.gausscurve_xarr == NULL) return;
    if ( !kwin_get_colour (pixcanvas, "red", &pixel_value, NULL, NULL, NULL) )
	pixel_value = 0;
    canvas_draw_lines_p (canvas, top->view2Datasets.gausscurve_xarr,
			 top->view2Datasets.gausscurve_yarr,
			 GAUSSCURVE_LENGTH, pixel_value);
    canvas_get_dressing (canvas,
			 KCD_ATT_TITLE_COLOUR, &label_colour,
			 KCD_ATT_END);
    if (label_colour == NULL) label_colour = "green";
    if ( !kwin_get_colour (pixcanvas, label_colour, &pixel_value,
			   NULL, NULL, NULL) ) pixel_value = 0;
    sprintf (txt, "FWHM: %.3g pixels", top->view2Datasets.fit_fwhm);
    if ( !kwin_hershey_draw_string (NULL, txt, 0.0, 0.0, 0.0, 12.8,
				    pixel_value, &string_width, &string_height,
				    NULL, NULL) )
	return;
    canvas_get_attributes (canvas,
			   CANVAS_ATT_X_OFFSET, &x_offset,
			   CANVAS_ATT_X_PIXELS, &x_pixels,
			   CANVAS_ATT_Y_OFFSET, &y_offset,
			   CANVAS_ATT_END);
    kwin_hershey_draw_string (pixcanvas, txt,
			      x_offset + x_pixels - string_width - 30,
			      y_offset + string_height + 10,
			      0.0, 14.0, pixel_value, NULL, NULL, NULL, NULL);
}   /*  End Function trace_refresh_func  */

static flag trace_position_func (KWorldCanvas canvas, double x, double y,
				 unsigned int event_code, void *e_info,
				 void **f_info, double x_lin, double y_lin,
				 int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas image_worldcanvas;
    KPixCanvas image_pixcanvas;
    unsigned int trace_num_restr, count;
    double image_x = TOOBIG;
    double image_y = TOOBIG;
    View2DatasetsWidget top = (View2DatasetsWidget) *f_info;
    char *trace_xlabel, *image_xlabel, *image_ylabel;
    char **trace_restr_names;
    double *trace_restr_values;

    if (top->view2Datasets.profile_mode != PROFILE_MODE_LINE) return (FALSE);
    XtVaGetValues (top->view2Datasets.image_display,
		   XkwNvisibleCanvas, &image_worldcanvas,
		   NULL);
    canvas_get_specification (canvas, &trace_xlabel, NULL, &trace_num_restr,
			      &trace_restr_names, &trace_restr_values);
    canvas_get_specification (image_worldcanvas, &image_xlabel, &image_ylabel,
			      NULL, NULL, NULL);
    if (trace_xlabel == NULL) return (FALSE);
    if (strcmp (trace_xlabel, image_xlabel) == 0) image_x = x_lin;
    else if (strcmp (trace_xlabel, image_ylabel) == 0) image_y = x_lin;
    for (count = 0; count < trace_num_restr; ++count)
    {
	if (strcmp (trace_restr_names[count], image_xlabel) == 0)
	    image_x = trace_restr_values[count];
	else if (strcmp (trace_restr_names[count], image_ylabel) == 0)
	    image_y = trace_restr_values[count];
    }
    canvas_convert_from_canvas_coords (image_worldcanvas, FALSE, TRUE,
				       1, &image_x, &image_y,
				       &image_x, &image_y);
    image_pixcanvas = canvas_get_pixcanvas (image_worldcanvas);
    /*  Undraw the old circle  */
    kwin_partial_refresh (image_pixcanvas, 1,
			  &top->view2Datasets.old_trace_area, FALSE);
    /*  Setup undraw parameters for next time  */
    top->view2Datasets.old_trace_area.startx = image_x - ELLIPSE_RADIUS - 1;
    top->view2Datasets.old_trace_area.endx = image_x + ELLIPSE_RADIUS + 1;
    top->view2Datasets.old_trace_area.starty = image_y - ELLIPSE_RADIUS - 1;
    top->view2Datasets.old_trace_area.endy = image_y + ELLIPSE_RADIUS + 1;
    top->view2Datasets.old_trace_area.clear = TRUE;
    /*  And draw this one  */
    kwin_draw_ellipse ( image_pixcanvas, image_x, image_y,
			ELLIPSE_RADIUS, ELLIPSE_RADIUS,
			get_green (image_pixcanvas) );
    return (FALSE);
}   /*  End Function trace_position_func  */

static void _View2Datasets_save_ppm_movie_cbk (Widget w, XtPointer client_data,
					       XtPointer call_data)
/*  [SUMMARY] Save PPM window movie callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    int precision;
    unsigned int frame_number, count;
    View2DatasetsWidget top = (View2DatasetsWidget) client_data;
    struct dataset_type *image_set, *hidden_set;
    struct direction_type *dir;
    CONST char *basename = call_data;
    char filename[STRING_LENGTH];

    get_sets (top, &image_set, NULL, &hidden_set, NULL);
    dir = get_dir (top, image_set);
    if (dir->num_frames < 2) dir = get_dir (top, hidden_set);
    if ( (dir == NULL) || (dir->num_frames < 2) )
    {
	fprintf (stderr, "No frames to display\n");
	return;
    }
    precision = ceil ( log10 (dir->num_frames) );
    frame_number = dir->frame_number;  /*  Save it  */
    for (count = 0; count < dir->num_frames; ++count)
    {
	dir->frame_number = count;
	_View2Datasets_display_changes (top, FALSE);
	sprintf (filename, "%s.%.*u.ppm", basename, precision, count);
	ExportMenuWritePPMWindow (w, filename);
    }
    /*  Put display back where it was  */
    dir->frame_number = frame_number;
    _View2Datasets_display_changes (top, FALSE);
}   /*  End Function save_ppm_movie_cbk  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag XkwView2DatasetsRegisterDataset (Widget W,
				      iarray pseudo, double min, double max,
				      iarray red, iarray green, iarray blue,
				      unsigned int set_number)
/*  [SUMMARY] Register a dataset.
    <W> The widget.
    <pseudo> The PseudoColour dataset (i.e. single-channel). This may be NULL.
    <min> The minimum value in the PseudoColour dataset.
    <max> The maximum value in the PseudoColour dataset.
    <red> The red component dataset. This may be NULL.
    <green> The green component dataset. This may be NULL.
    <blue> The blue component dataset. This may be NULL.
    <set_number> The dataset number. This may be 0 or 1.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KWorldCanvas pseudo_canvas, direct_canvas, true_canvas, rgb_canvas;
    KWorldCanvas mag_pseudo_canvas, mag_direct_canvas, mag_true_canvas;
    KWorldCanvas mag_rgb_canvas, pan_rgb_canvas;
    KWorldCanvas pan_pseudo_canvas, pan_direct_canvas, pan_true_canvas;
    KWorldCanvas canvas, mag_canvas, pan_canvas;
    iarray array;
    flag change_mode = FALSE;
    flag set_levels = FALSE;
    flag set_is_image = FALSE;
    Bool reset_intensity = TRUE;
    double lower_clip, upper_clip;
    struct dataset_type *set = NULL;
    struct dataset_type *other_set = NULL;
    void *object;
    View2DatasetsWidget top = (View2DatasetsWidget) W;
    Widget izoomwinpopup, reset_intensity_tgl, magnifier, panner;
    char txt[STRING_LENGTH];
    static char function_name[] = "XkwView2DatasetsRegisterDataset";

    switch (set_number)
    {
      case 0:
	set = &top->view2Datasets.set0;
	other_set = &top->view2Datasets.set1;
	break;
      case 1:
	set = &top->view2Datasets.set1;
	other_set = &top->view2Datasets.set0;
	break;
      default:
	fprintf (stderr, "Illegal set number: %u\n", set_number);
	a_prog_bug (function_name);
	break;
    }
    /*  Clear out old information  */
    if (set->destroy_func != NULL)
    {
	c_unregister_callback (set->destroy_func);
	set->destroy_func = NULL;
    }
    set->multi_desc = NULL;
    set->array_pseudo = NULL;
    set->array_red = NULL;
    set->array_green = NULL;
    set->array_blue = NULL;
    if (set->vimage != NULL) viewimg_destroy (set->vimage);
    set->vimage = NULL;
    if (set->mag_vimage != NULL) viewimg_destroy (set->mag_vimage);
    set->mag_vimage = NULL;
    if (set->pan_vimage != NULL) viewimg_destroy (set->pan_vimage);
    set->pan_vimage = NULL;
    if (set->pc_cimage != NULL) contour_destroy (set->pc_cimage);
    set->pc_cimage = NULL;
    if (set->rgb_cimage != NULL) contour_destroy (set->rgb_cimage);
    set->rgb_cimage = NULL;
    free_viewables (&set->xy);
    free_viewables (&set->xz);
    free_viewables (&set->zy);
    if ( (pseudo == NULL) && (red == NULL) && (green == NULL) &&
	 (blue == NULL) )
    {
	/*  No data at all  */
	set->object_name[0] = '\0';
	XkwChoiceMenuSetEntrySensitivity
	    (top->view2Datasets.intensity_menu,
	     top->view2Datasets.first_intensity + set_number,
	     FALSE);
	if (set->filename_label == NULL) return (TRUE);
	sprintf (txt, "Set %u Filename:                            ",
		 set_number + 1);
	XtVaSetValues (set->filename_label,
		       XtNlabel, txt,
		       NULL);
	sprintf (txt, "Set %u Object:                              ",
		 set_number + 1);
	XtVaSetValues (set->object_label,
		       XtNlabel, txt,
		       NULL);
	return (TRUE);
    }
    if ( (other_set->array_pseudo == NULL) && (other_set->array_red == NULL) )
    {
	/*  No data for the other set: make sure this set is viewable  */
	switch (top->view2Datasets.mode)
	{
	  case MODE_IMAGE0:
	  case MODE_IMAGE0_CONTOUR1:
	    if (set_number == 1) change_mode = TRUE;
	    break;
	  case MODE_IMAGE1:
	  case MODE_IMAGE1_CONTOUR0:
	    if (set_number == 0) change_mode = TRUE;
	    break;
	}
	if (change_mode)
	{
	    XtVaSetValues (top->view2Datasets.mode_menu,
			   XkwNsetChoice, mode_choices[set_number],
			   NULL);
	    top->view2Datasets.mode = (set_number ==
				       0) ? MODE_IMAGE0 : MODE_IMAGE1;
	}
    }
    switch (top->view2Datasets.mode)
    {
      case MODE_IMAGE0_CONTOUR1:
	if (set_number == 1) set_levels = TRUE;
	/*  Fall through  */
      case MODE_IMAGE0:
	if (set_number == 0) set_is_image = TRUE;
	break;
      case MODE_IMAGE1_CONTOUR0:
	if (set_number == 0) set_levels = TRUE;
	/*  Fall through  */
      case MODE_IMAGE1:
	if (set_number == 1) set_is_image = TRUE;
	break;
    }
    if (!top->view2Datasets.auto_apply_state) set_levels = FALSE;
    XtVaGetValues (top->view2Datasets.image_display,
		   XkwNpseudoColourCanvas, &pseudo_canvas,
		   XkwNdirectColourCanvas, &direct_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   XkwNmagnifier, &magnifier,
		   XkwNpanner, &panner,
		   NULL);
    XtVaGetValues (magnifier,
		   XkwNpseudoColourCanvas, &mag_pseudo_canvas,
		   XkwNdirectColourCanvas, &mag_direct_canvas,
		   XkwNtrueColourCanvas, &mag_true_canvas,
		   NULL);
    XtVaGetValues (panner,
		   XkwNpseudoColourCanvas, &pan_pseudo_canvas,
		   XkwNdirectColourCanvas, &pan_direct_canvas,
		   XkwNtrueColourCanvas, &pan_true_canvas,
		   NULL);
    if (pseudo_canvas == NULL)
    {
	pseudo_canvas = true_canvas;
	mag_pseudo_canvas = mag_true_canvas;
	pan_pseudo_canvas = pan_true_canvas;
    }
    rgb_canvas = (direct_canvas == NULL) ? true_canvas : direct_canvas;
    mag_rgb_canvas = (mag_direct_canvas ==
		      NULL) ? mag_true_canvas : mag_direct_canvas;
    pan_rgb_canvas = (pan_direct_canvas ==
		      NULL) ? pan_true_canvas : pan_direct_canvas;
    if (set_is_image)
    {
	if (mag_pseudo_canvas != NULL)
	    kwin_refresh_if_visible (canvas_get_pixcanvas (mag_pseudo_canvas),
				     TRUE);
	if (mag_rgb_canvas != NULL)
	    kwin_refresh_if_visible (canvas_get_pixcanvas (mag_rgb_canvas),
				     TRUE);
    }
    sprintf (txt, "izoomwinpopup%u", set_number);
    if ( ( izoomwinpopup = XtNameToWidget (top->view2Datasets.image_display,
					   txt) ) == NULL )
    {
	fprintf (stderr, "No \"%s\" widget!\n", txt);
	a_prog_bug (function_name);
    }
    array = (pseudo == NULL) ? red : pseudo;
    set->multi_desc = array->multi_desc;
    set->array_index = array->array_num;
    set->cmap_index = set->multi_desc->num_arrays;
    if ( ( object = iarray_get_named_string (array, "OBJECT") ) == NULL )
    {
	set->object_name[0] = '\0';
	sprintf (txt, "Set %u has no object name", set_number + 1);
	XtVaSetValues (set->object_label,
		       XtNlabel, txt,
		       NULL);
    }
    else
    {
	strcpy (set->object_name, object);
	sprintf (txt, "Set %u Object: %-30s", set_number + 1, (char *) object);
	XtVaSetValues (set->object_label,
		       XtNlabel, txt,
		       NULL);
	m_free (object);
    }
    if ( ds_get_named_object (set->multi_desc, set->multi_desc->num_arrays,
			      "filename", &object) )
    {
	sprintf (txt, "Set %u Filename: %-29s",
		 set_number + 1, (char *) object);
	XtVaSetValues (set->filename_label,
		       XtNlabel, txt,
		       NULL);
	if (set->object_name[0] == '\0')
	    strcpy (set->object_name, (char *) object);
    }
    if (pseudo != NULL)
    {
	unsigned int dummy;

    	/*  PseudoColour dataset  */
	if ( (iarray_num_dim (pseudo) < 2) || (iarray_num_dim (pseudo) > 3) )
	{
	    fprintf ( stderr, "%s: PseudoColour dataset has: %u dimensions!\n",
		      function_name, iarray_num_dim (pseudo) );
	    return (FALSE);
	}
	set->array_pseudo = pseudo;
	set->min = min;
	set->max = max;
	XkwChoiceMenuSetEntrySensitivity
	    (top->view2Datasets.intensity_menu,
	     top->view2Datasets.first_intensity + set_number,
	     TRUE);
	/*  Check for colourmap  */
	switch ( dummy = ds_f_array_name (set->multi_desc, "RGBcolourmap",
					  (char **) NULL, &set->cmap_index) )
	{
	  case IDENT_NOT_FOUND:
	    set->cmap_index = set->multi_desc->num_arrays;
	    break;
	  case IDENT_GEN_STRUCT:
	    /*  Got it!  */
	    break;
	  case IDENT_MULTIPLE:
	    fprintf (stderr, "Multiple RGBcolourmap structures found\n");
	    set->cmap_index = set->multi_desc->num_arrays;
	    return (FALSE);
	    /*break;*/
	  default:
	    fprintf (stderr,
		     "%s: illegal return value: %u from: ds_f_array_name\n",
		     function_name, dummy);
	    a_prog_bug (function_name);
	    break;
	}
	canvas_get_attributes (pseudo_canvas,
			       CANVAS_ATT_VALUE_MIN, &lower_clip,
			       CANVAS_ATT_VALUE_MAX, &upper_clip,
			       CANVAS_ATT_END);
	reset_intensity_tgl =
	    XtNameToWidget (top->view2Datasets.image_display,
			    "intensityPolicyPopup*resetIntensityToggle");
	if (reset_intensity_tgl != NULL)
	    XtVaGetValues (reset_intensity_tgl,
			   XtNstate, &reset_intensity,
			   NULL);
	if (lower_clip == upper_clip) reset_intensity = TRUE;
	if ( reset_intensity || (pseudo_canvas == NULL) )
	{
	    lower_clip = min;
	    upper_clip = max;
	}
	set->destroy_func =
	    iarray_register_destroy_func (pseudo,
					  ( flag (*) () )iarr_destroy_callback,
					  set);
	if (set_is_image) update_profile_axis (top, set);
	if (iarray_num_dim (pseudo) == 2)
	{
	    /*  2-dimensional PsuedoColour array  */
	    if ( !setup_2D_pseudo (top, set, pseudo, lower_clip, upper_clip,
				   pseudo_canvas, rgb_canvas,
				   mag_pseudo_canvas, mag_rgb_canvas,
				   pan_pseudo_canvas, pan_rgb_canvas) )
	    {
		XkwView2DatasetsRegisterDataset (W, NULL, 0.0, 0.0,
						 NULL, NULL, NULL, set_number);
		return (FALSE);
	    }
	    /*  All finished with 2-dimensional PsuedoColour array  */
	    update_animate_widget (top);
	    if (set_levels) apply_levels_cbk (NULL, top, NULL);
	    _View2Datasets_display_changes (top, TRUE);
	    _View2Datasets_update_cmap (top);
	    XkwDataclipNewArray (izoomwinpopup, pseudo, min, max,
				 reset_intensity);
	    _View2Datasets_update_title (top);
	    return (TRUE);
	}
	/*  Must be a 3-dimensional PsuedoColour array  */
	if ( !setup_3D_pseudo (top, set, pseudo, lower_clip, upper_clip,
			       pseudo_canvas, rgb_canvas,
			       mag_pseudo_canvas, mag_rgb_canvas,
			       pan_pseudo_canvas, pan_rgb_canvas) )
	{
	    XkwView2DatasetsRegisterDataset (W, NULL, 0.0, 0.0,
					     NULL, NULL, NULL, set_number);
	    return (FALSE);
	}
	/*  All finished with 3-dimensional PsuedoColour array  */
	update_animate_widget (top);
	if (set_levels) apply_levels_cbk (NULL, top, NULL);
	_View2Datasets_display_changes (top, TRUE);
	_View2Datasets_update_cmap (top);
	XkwDataclipNewArray (izoomwinpopup, pseudo, min, max, reset_intensity);
	_View2Datasets_update_title (top);
	return (TRUE);
    }
    /*  TrueColour dataset  */
    set->destroy_func =
	iarray_register_destroy_func (red,
				      ( flag (*) () ) iarr_destroy_callback,
				      set);
    XkwChoiceMenuSetEntrySensitivity
	(top->view2Datasets.intensity_menu,
	 top->view2Datasets.first_intensity + set_number,
	 FALSE);
    canvas = (rgb_canvas == NULL) ? pseudo_canvas : rgb_canvas;
    mag_canvas = (mag_rgb_canvas == NULL) ? mag_pseudo_canvas : mag_rgb_canvas;
    pan_canvas = (pan_rgb_canvas == NULL) ? pan_pseudo_canvas : pan_rgb_canvas;
    if ( (iarray_num_dim (red) < 2) || (iarray_num_dim (red) > 3) )
    {
	fprintf ( stderr, "%s: Red component dataset has: %u dimensions!\n",
		  function_name, iarray_num_dim (red) );
	XkwView2DatasetsRegisterDataset (W, NULL, 0.0, 0.0, NULL, NULL, NULL,
					 set_number);
	return (FALSE);
    }
    set->array_red = red;
    set->array_green = green;
    set->array_blue = blue;
    if (iarray_num_dim (red) == 2)
    {
	/*  2-dimensional TrueColour array  */
	if ( !setup_2D_rgb (top, set, red, green, blue, canvas, mag_canvas,
			    pan_canvas) )
	{
	    XkwView2DatasetsRegisterDataset (W, NULL, 0.0, 0.0,
					     NULL, NULL, NULL, set_number);
	    return (FALSE);
	}
	/*  All finished with 2-dimensional PsuedoColour array  */
	update_animate_widget (top);
	_View2Datasets_display_changes (top, TRUE);
	_View2Datasets_update_title (top);
	return (TRUE);
    }
    /*  Must be a 3-dimensional TrueColour array  */
    if ( !setup_3D_rgb (top, set, red, green, blue, canvas, mag_canvas,
			pan_canvas) )
    {
	XkwView2DatasetsRegisterDataset (W, NULL, 0.0, 0.0, NULL, NULL, NULL,
					 set_number);
	return (FALSE);
    }
    /*  All finished with 3-dimensional TrueColour array  */
    update_animate_widget (top);
    _View2Datasets_display_changes (top, TRUE);
    _View2Datasets_update_title (top);
    return (TRUE);
}   /*  End Function XkwView2DatasetsRegisterDataset  */

/*EXPERIMENTAL_FUNCTION*/
void XkwView2DatasetsInvalidatePlane (Widget W, unsigned int set_number,
				      unsigned int dim_index,
				      unsigned long plane_index)
/*  [SUMMARY] Register the change in data of a plane.
    <W> The widget.
    <set_number> The dataset number. This may be 0 or 1.
    <dim_index> The dimension index of the plane.
    <plane_index> The index along the specified dimension.
    [RETURNS] Nothing.
*/
{
    struct dataset_type *set = NULL;
    View2DatasetsWidget top = (View2DatasetsWidget) W;
    Widget izoomwinpopup;
    char txt[STRING_LENGTH];
    static char function_name[] = "XkwView2DatasetsInvalidatePlane";

    switch (set_number)
    {
      case 0:
	set = &top->view2Datasets.set0;
	break;
      case 1:
	set = &top->view2Datasets.set1;
	break;
      default:
	fprintf (stderr, "Illegal set number: %u\n", set_number);
	a_prog_bug (function_name);
	break;
    }
    if (set->xy.num_frames < 1) return;
    switch (dim_index)
    {
      case 0:
	viewimg_register_data_change (set->xy.vimages[plane_index]);
	viewimg_register_data_change (set->xy.mag_vimages[plane_index]);
	viewimg_register_data_change (set->xy.pan_vimages[plane_index]);
	if (set->xy.pc_cimages != NULL)
	    contour_register_data_change (set->xy.pc_cimages[plane_index]);
	if (set->xy.rgb_cimages != NULL)
	    contour_register_data_change (set->xy.rgb_cimages[plane_index]);
	invalidate_planes (&set->xz);
	invalidate_planes (&set->zy);
	break;
      case 1:
	viewimg_register_data_change (set->xz.vimages[plane_index]);
	viewimg_register_data_change (set->xz.mag_vimages[plane_index]);
	viewimg_register_data_change (set->xz.pan_vimages[plane_index]);
	if (set->xz.pc_cimages != NULL)
	    contour_register_data_change (set->xz.pc_cimages[plane_index]);
	if (set->xz.rgb_cimages != NULL)
	    contour_register_data_change (set->xz.rgb_cimages[plane_index]);
	invalidate_planes (&set->xy);
	invalidate_planes (&set->zy);
	break;
      case 2:
	viewimg_register_data_change (set->zy.vimages[plane_index]);
	viewimg_register_data_change (set->zy.mag_vimages[plane_index]);
	viewimg_register_data_change (set->zy.pan_vimages[plane_index]);
	if (set->zy.pc_cimages != NULL)
	    contour_register_data_change (set->zy.pc_cimages[plane_index]);
	if (set->zy.rgb_cimages != NULL)
	    contour_register_data_change (set->zy.rgb_cimages[plane_index]);
	invalidate_planes (&set->xy);
	invalidate_planes (&set->xz);
	break;
      default:
	fprintf (stderr, "Illegal dimension index: %u\n", dim_index);
	a_prog_bug (function_name);
	break;
    }
    if (set->array_pseudo != NULL)
    {
	sprintf (txt, "izoomwinpopup%u", set_number);
	izoomwinpopup = XtNameToWidget (top->view2Datasets.image_display, txt);
	XkwDataclipArrayChanged (izoomwinpopup, TOOBIG, -TOOBIG);
    }
}   /*  End Function XkwView2DatasetsInvalidatePlane  */

/*EXPERIMENTAL_FUNCTION*/
void XkwView2DatasetsInvalidateAll (Widget W, unsigned int set_number)
/*  [SUMMARY] Register the change in data of a dataset.
    <W> The widget.
    <set_number> The dataset number. This may be 0 or 1.
    [RETURNS] Nothing.
*/
{
    struct dataset_type *set = NULL;
    View2DatasetsWidget top = (View2DatasetsWidget) W;
    Widget izoomwinpopup;
    char txt[STRING_LENGTH];
    static char function_name[] = "XkwView2DatasetsInvalidateAll";

    switch (set_number)
    {
      case 0:
	set = &top->view2Datasets.set0;
	break;
      case 1:
	set = &top->view2Datasets.set1;
	break;
      default:
	fprintf (stderr, "Illegal set number: %u\n", set_number);
	a_prog_bug (function_name);
	break;
    }
    if (set->vimage != NULL) viewimg_register_data_change (set->vimage);
    if (set->mag_vimage != NULL) viewimg_register_data_change(set->mag_vimage);
    if (set->pan_vimage != NULL) viewimg_register_data_change(set->pan_vimage);
    if (set->pc_cimage != NULL) contour_register_data_change (set->pc_cimage);
    if (set->rgb_cimage != NULL) contour_register_data_change(set->rgb_cimage);
    invalidate_planes (&set->xy);
    invalidate_planes (&set->xz);
    invalidate_planes (&set->zy);
    if (set->array_pseudo != NULL)
    {
	sprintf (txt, "izoomwinpopup%u", set_number);
	izoomwinpopup = XtNameToWidget (top->view2Datasets.image_display, txt);
	XkwDataclipArrayChanged (izoomwinpopup, TOOBIG, -TOOBIG);
    }
}   /*  End Function XkwView2DatasetsInvalidateAll  */

/*EXPERIMENTAL_FUNCTION*/
flag XkwView2DatasetsCanvasEventHandler (ViewableImage vimage,
					 double x, double y,
					 void *value, unsigned int event_code,
					 void *e_info, void **f_info,
					 double x_lin, double y_lin,
					 unsigned int value_type,
					 double x_im, double y_im,
					 int x_pix, int y_pix)
/*  [SUMMARY] Event handler provided by View2Datasets widget.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see  canvas_register_transform_func  ) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    K_UB_RGB.
    [RETURNS] FALSE indicating that the event is still to be processed.
*/
{
    KEvent event;
    uaddr index_x = 0;
    uaddr index_y = 0;
    uaddr index_z = 0;
    struct XkwView2DatasetsCursor cursor;
    View2DatasetsWidget top = (View2DatasetsWidget) *f_info;
    Widget first_track_label, second_track_label, third_track_label;
    Widget image_display;
    KEventCoordinate *coordinate = NULL;
    struct dataset_type *set;
    struct direction_type *dir;
    char *ptr_x, *ptr_y, *ptr_z, *ptr_string;
    uaddr coord_indices[2];
    char pix_string[STRING_LENGTH];
    char world_string[STRING_LENGTH], extra_string[STRING_LENGTH];
    unsigned int dim_indices[2];
    static char function_name[] = "XkwView2DatasetsCanvasEventHandler";

    m_clear (&event, sizeof event);
    event.type = event_code;
    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_LEFT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_RELEASE:
      case K_CANVAS_EVENT_RIGHT_MOUSE_RELEASE:
      case K_CANVAS_EVENT_POINTER_WARP:
	coordinate = &event.data.point;
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	coordinate = &event.data.keypress.position;
	event.data.keypress.modifier_mask = 0;
	event.data.keypress.string = e_info;
	break;
      case K_CANVAS_EVENT_MODIFIED_KEY_PRESS:
	coordinate = &event.data.keypress.position;
	ptr_string = e_info;
	m_copy (&event.data.keypress.modifier_mask,
		ptr_string + strlen (ptr_string) + 1,
		sizeof event.data.keypress.modifier_mask);
	event.data.keypress.string = ptr_string;
	break;
      case K_CANVAS_EVENT_KEYSYM:
      case K_CANVAS_EVENT_POINTER_LEAVE:
      case K_CANVAS_EVENT_POINTER_ENTER:
	return (FALSE);
	/*break;*/
      default:
	fprintf (stderr, "Unsupported event type: %u\n", event_code);
	return (TRUE);
	/*break;*/
    }
    coordinate->screen_pixel.defined = TRUE;
    coordinate->screen_pixel.x = x_pix;
    coordinate->screen_pixel.y = y_pix;
    coordinate->linear.defined = TRUE;
    coordinate->linear.x = x_lin;
    coordinate->linear.y = y_lin;
    coordinate->world.defined = TRUE;
    coordinate->world.x = x;
    coordinate->world.y = y;
    coordinate->image_pixel.defined = TRUE;
    coordinate->image_pixel.x = x_im;
    coordinate->image_pixel.y = y_im;
    event_dispatch (event);
    get_sets (top, &set, NULL, NULL, NULL);
    viewimg_track_compute (vimage, value, value_type, x, y, x_im, y_im,
			   pix_string, world_string, extra_string);
    if ( (event_code == K_CANVAS_EVENT_PLAIN_KEY_PRESS) &&
	 (strcmp (event.data.keypress.string, "l") == 0) )
	puts (world_string);
    if (set->xy.vimages != NULL)
    {
	if ( ( ptr_x = strchr (pix_string, 'x') ) == NULL )
	{
	    fprintf (stderr, "%s: error searching for 'x' in: \"%s\"\n",
		     function_name, pix_string);
	    return (FALSE);
	}
	if ( ( ptr_y = strchr (pix_string, 'y') ) == NULL )
	{
	    fprintf (stderr, "%s: error searching for 'y' in: \"%s\"\n",
		     function_name, pix_string);
	    return (FALSE);
	}
	if ( ( ptr_z = strchr (pix_string, 'z') ) == NULL )
	{
	    fprintf (stderr, "%s: error searching for 'z' in: \"%s\"\n",
		     function_name, pix_string);
	    return (FALSE);
	}
	switch (top->view2Datasets.slice)
	{
	  case SLICE_XY:
	    ptr_z[0] = ' ';
	    ptr_z[1] = 'z';
	    break;
	  case SLICE_XZ:
	    *ptr_y = 'z';
	    ptr_z[0] = ' ';
	    ptr_z[1] = 'y';
	    break;
	  case SLICE_ZY:
	    *ptr_x = 'z';
	    ptr_z[0] = ' ';
	    ptr_z[1] = 'x';
	    break;
	}
    }
    image_display = top->view2Datasets.image_display;
    first_track_label = XtNameToWidget (image_display, "trackLabel0");
    second_track_label = XtNameToWidget (image_display, "trackLabel1");
    third_track_label = XtNameToWidget (image_display, "trackLabel2");
    XtVaSetValues (first_track_label, XtNlabel, pix_string, NULL);
    XtVaSetValues (second_track_label, XtNlabel, world_string, NULL);
    XtVaSetValues (third_track_label, XtNlabel, extra_string, NULL);
    if ( (event_code != K_CANVAS_EVENT_POINTER_MOVE) &&
	 (event_code != K_CANVAS_EVENT_POINTER_WARP) ) return (FALSE);
    if ( (set->array_pseudo != NULL) &&
	 (iarray_num_dim (set->array_pseudo) == 3) )
    {
	/*  Get the full 3-dimensional position of the cursor  */
	dir = get_dir (top, set);
	switch (top->view2Datasets.slice)
	{
	  case SLICE_XY:
	    index_x = floor (x_im + 0.5);
	    index_y = floor (y_im + 0.5);
	    index_z = dir->frame_number;
	    break;
	  case SLICE_XZ:
	    index_x = floor (x_im + 0.5);
	    index_y = dir->frame_number;
	    index_z = floor (y_im + 0.5);
	    break;
	  case SLICE_ZY:
	    index_x = dir->frame_number;
	    index_y = floor (y_im + 0.5);
	    index_z = floor (x_im + 0.5);
	    break;
	}
    }
    if ( (set->array_pseudo != NULL) &&
	 (top->view2Datasets.profile_mode == PROFILE_MODE_LINE) )
    {
	if (iarray_num_dim (set->array_pseudo) == 2)
	{
	    /*  Can only get X or Y profile  */
	    switch (top->view2Datasets.profile_dir)
	    {
	      case PROFILE_DIR_Z:
		fprintf (stderr, "Z profile not possible with 2D dataset\n");
		break;
	      case PROFILE_DIR_Y:
		dim_indices[0] = 1;
		coord_indices[0] = floor (x_im + 0.5);
		XkwTracePopupShowTrace (top->view2Datasets.trace_winpopup,
					dim_indices, coord_indices);
		break;
	      case PROFILE_DIR_X:
		dim_indices[0] = 0;
		coord_indices[0] = floor (y_im + 0.5);
		XkwTracePopupShowTrace (top->view2Datasets.trace_winpopup,
					dim_indices, coord_indices);
		break;
	    }
	}
	else
	{
	    /*  Can get X, Y or Z profile  */
	    switch (top->view2Datasets.profile_dir)
	    {
	      case PROFILE_DIR_Z:
		dim_indices[0] = 2;
		dim_indices[1] = 1;
		coord_indices[0] = index_x;
		coord_indices[1] = index_y;
		break;
	      case PROFILE_DIR_Y:
		dim_indices[0] = 2;
		dim_indices[1] = 0;
		coord_indices[0] = index_x;
		coord_indices[1] = index_z;
		break;
	      case PROFILE_DIR_X:
		dim_indices[0] = 0;
		dim_indices[1] = 1;
		coord_indices[0] = index_z;
		coord_indices[1] = index_y;
		break;
	    }
	    XkwTracePopupShowTrace (top->view2Datasets.trace_winpopup,
				    dim_indices, coord_indices);
	}
    }
    switch (top->view2Datasets.mode)
    {
      case MODE_IMAGE0:
      case MODE_IMAGE0_CONTOUR1:
	cursor.set_number = 0;
	break;
      case MODE_IMAGE1:
      case MODE_IMAGE1_CONTOUR0:
	cursor.set_number = 1;
	break;
    }
    cursor.pixel.x = index_x;
    cursor.pixel.y = index_y;
    cursor.pixel.z = index_z;
    XtCallCallbacks ( (Widget) top, XkwNcursorCallback, (XtPointer) &cursor );
    return (FALSE);
}   /*  End Function XkwView2DatasetsCanvasEventHandler  */
