/*LINTLIBRARY*/
/*  ZoomPolicy.c

    This code provides a Zoom Policy control widget for Xt.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a zoom policy control widget for
    Xt.


    Written by      Richard Gooch   12-MAY-1996

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   29-MAY-1996: Added support for logarithmic
  and square-root intensity scaling.

    Updated by      Richard Gooch   1-JAN-1997: Added XkwNshowIntensityReset
  resource.

    Updated by      Richard Gooch   20-JAN-1997: Created integerXZoom and
  integerYZoom resources.

    Updated by      Richard Gooch   3-FEB-1997: Created fixAspect resource.

    Updated by      Richard Gooch   16-MAR-1997: Fixed bug in <apply_cbk> where
  canvas array pointer was not checked.

    Updated by      Richard Gooch   4-AUG-1997: Added configurable middle and
  right mouse button zoom toggles.

    Updated by      Richard Gooch   8-DEC-1997: Split some controls into
  IntensityPolicy widget.

    Updated by      Richard Gooch   22-FEB-1998: Created function
  <XkwZoomPolicyAddCanvas>.

    Updated by      Richard Gooch   13-JUL-1999: Added magnification menu.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>

    Last updated by Richard Gooch   20-MAR-2001: Created Destroy method.


*/
#include <stdio.h>
#include <math.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_kwin.h>
#include <karma_imw.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/ZoomPolicyP.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>


struct canvas_list
{
    KWorldCanvas canvas;
    struct canvas_list *next;
};


/*  Private functions  */
STATIC_FUNCTION (void _ZoomPolicy_Initialise, (Widget request, Widget new) );
STATIC_FUNCTION (void _ZoomPolicy_Destroy, (Widget W) );
STATIC_FUNCTION (void apply_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void set_bool,
		 (ZoomPolicyWidget top, unsigned int att, flag val) );
STATIC_FUNCTION (void auto_refresh_cbk, (Widget w, XtPointer client_data,
					 XtPointer call_data) );
STATIC_FUNCTION (void aspect_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void trunc_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void int_x_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void int_y_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void magnify_cbk, (Widget w, XtPointer client_data,
				    XtPointer call_data) );


#define offset(field) XtOffset(ZoomPolicyWidget, zoomPolicy.field)

static XtResource ZoomPolicyResources[] = 
{
    {XkwNshowMiddleZoom, XkwCShowMiddleZoom, XtRBool, sizeof (Bool),
     offset (showMiddleZoom), XtRImmediate, (XtPointer) False},
    {XkwNshowRightZoom, XkwCShowRightZoom, XtRBool, sizeof (Bool),
     offset (showRightZoom), XtRImmediate, (XtPointer) False},
    {XkwNintegerXZoom, XkwCIntegerXZoom, XtRBool, sizeof (Bool),
     offset (integerXZoom), XtRImmediate, (XtPointer) True},
    {XkwNintegerYZoom, XkwCIntegerYZoom, XtRBool, sizeof (Bool),
     offset (integerYZoom), XtRImmediate, (XtPointer) True},
    {XkwNfixAspect, XkwCFixAspect, XtRBool, sizeof (Bool),
     offset (fixAspect), XtRImmediate, (XtPointer) True},
};

#undef offset

#define MAGNIFY_REPLICATE      0
#define MAGNIFY_INTERPOLATE    1
#define NUM_MAGNIFY_CHOICES    2

static char *magnify_choices[NUM_MAGNIFY_CHOICES] =
{
    "pixel replication", "bi-linear interpolation",
};


ZoomPolicyClassRec zoomPolicyClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,    /* superclass */
    "ZoomPolicy",                            /* class_name */
    sizeof (ZoomPolicyRec),                  /* widget_size */
    NULL,                                    /* class_initialise */
    NULL,                                    /* class_part_initialise */
    FALSE,                                   /* class_init */
    (XtInitProc) _ZoomPolicy_Initialise,     /* initialise */
    NULL,                                    /* initialise_hook */
    XtInheritRealize,                        /* realise */
    NULL,                                    /* actions */
    0,                                       /* num_actions */
    ZoomPolicyResources,                     /* resources */
    XtNumber (ZoomPolicyResources),          /* num_resources */
    NULLQUARK,                               /* xrm_class */
    TRUE,                                    /* compress_motion */
    TRUE,                                    /* compress_exposure */
    TRUE,                                    /* compress_enterleave */
    TRUE,                                    /* visible_interest */
    _ZoomPolicy_Destroy,                     /* destroy */
    XtInheritResize,                         /* resize */
    NULL,                                    /* expose */
    NULL,                                    /* set_values */
    NULL,                                    /* set_values_hook */
    XtInheritSetValuesAlmost,                /* set_values_almost */
    NULL,                                    /* get_values_hook */
    NULL,                                    /* accept_focus */
    XtVersion,                               /* version */
    NULL,                                    /* callback_private */
    NULL,                                    /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* ZoomPolicyClassPart */
    0 /* empty */
  }
};

WidgetClass zoomPolicyWidgetClass = (WidgetClass) &zoomPolicyClassRec;

static void _ZoomPolicy_Initialise (Widget Request, Widget New)
{
    /*ZoomPolicyWidget request = (ZoomPolicyWidget) Request;*/
    ZoomPolicyWidget new = (ZoomPolicyWidget) New;
    Widget form;
    Widget w, upper_widget;
    /*static char function_name[] = "ZoomPolicyWidget::Initialise";*/

    new->zoomPolicy.auto_refresh = TRUE;
    new->zoomPolicy.magnify_choice = MAGNIFY_REPLICATE;
    new->zoomPolicy.first_canvas = NULL;
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    w = XtVaCreateManagedWidget ("apply", commandWidgetClass, form,
				 XtNlabel, "Apply",
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, apply_cbk, New);
    w = XtVaCreateManagedWidget ("autoRefreshToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Auto Refresh",
				 XtNstate, True,
				 XkwNcrosses, False,
				 XtNfromHoriz, w,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, auto_refresh_cbk, New);
    w = XtVaCreateManagedWidget ("fixAspectToggle", ktoggleWidgetClass, form,
				 XtNlabel, "Fix Aspect",
				 XtNfromVert, upper_widget,
				 XtNstate, new->zoomPolicy.fixAspect,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, aspect_cbk, New);
    w = XtVaCreateManagedWidget ("allowTruncationToggle", ktoggleWidgetClass,
				 form,
				 XtNlabel, "Allow Truncation",
				 XtNfromHoriz, w,
				 XtNfromVert, upper_widget,
				 XtNstate, True,
				 XkwNcrosses, False,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, trunc_cbk, New);
    w = XtVaCreateManagedWidget ("integerXZoomToggle", ktoggleWidgetClass,form,
				 XtNlabel, "Integer X Zoom",
				 XtNfromVert, upper_widget,
				 XtNstate, new->zoomPolicy.integerXZoom,
				 XkwNcrosses, False,
				 NULL);
    XtAddCallback (w, XtNcallback, int_x_cbk, New);
    w = XtVaCreateManagedWidget ("integerYZoomToggle", ktoggleWidgetClass,form,
				 XtNlabel, "Integer Y Zoom",
				 XtNfromHoriz, w,
				 XtNfromVert, upper_widget,
				 XtNstate, new->zoomPolicy.integerYZoom,
				 XkwNcrosses, False,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XtNcallback, int_y_cbk, New);
    if (new->zoomPolicy.showMiddleZoom)
    {
	w = XtVaCreateManagedWidget ("middleZoomToggle", ktoggleWidgetClass,
				     form,
				     XtNlabel, "Zoom In on Middle Mouse Click",
				     XtNfromVert, upper_widget,
				     XkwNcrosses, False,
				     NULL);
	upper_widget = w;
    }
    if (new->zoomPolicy.showRightZoom)
    {
	w = XtVaCreateManagedWidget ("rightZoomToggle", ktoggleWidgetClass,
				     form,
				     XtNlabel, "Zoom Out on Right Mouse Click",
				     XtNfromVert, upper_widget,
				     XkwNcrosses, False,
				     NULL);
	upper_widget = w;
    }
    w = XtVaCreateManagedWidget ("expandMenu", exclusiveMenuWidgetClass, form,
				 XkwNchoiceName, "Expand",
				 XtNfromVert, upper_widget,
				 XkwNnumItems, NUM_MAGNIFY_CHOICES,
				 XkwNitemStrings, magnify_choices,
				 XkwNvaluePtr, &new->zoomPolicy.magnify_choice,
				 NULL);
    upper_widget = w;
    XtAddCallback (w, XkwNselectCallback, magnify_cbk, New);
}   /*  End Function Initialise  */

static void _ZoomPolicy_Destroy (Widget W)
{
    ZoomPolicyWidget w = (ZoomPolicyWidget) W;
    struct canvas_list *curr, *next;

    for (curr = w->zoomPolicy.first_canvas; curr != NULL; curr = next)
    {
	next = curr->next;
	m_free (curr);
    }
}   /*  End Function Destroy  */

static void apply_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the apply callback.
*/
{
    KPixCanvas pixcanvas;
    flag visible;
    ZoomPolicyWidget top = (ZoomPolicyWidget) client_data;
    struct canvas_list *curr;

    for (curr = top->zoomPolicy.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	pixcanvas = canvas_get_pixcanvas (curr->canvas);
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_VISIBLE, &visible,
			     KWIN_ATT_END);
	if (!visible) continue;
	/*  Resize the canvas  */
	if ( !canvas_resize (curr->canvas, NULL, FALSE) )
	{
	    fprintf (stderr, "Error resizing canvas\n");
	}
    }
}   /*  End Function apply_cbk   */

static void set_bool (ZoomPolicyWidget top, unsigned int att, flag val)
{
    struct canvas_list *curr;

    for (curr = top->zoomPolicy.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	viewimg_set_canvas_attributes (curr->canvas,
				       att, val,
				       VIEWIMG_ATT_END);
    }
    if (top->zoomPolicy.auto_refresh) apply_cbk (NULL, top, NULL);
}   /*  End Function set_bool   */

static void auto_refresh_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  This is the auto refresh callback.
*/
{
    ZoomPolicyWidget top = (ZoomPolicyWidget) client_data;

    top->zoomPolicy.auto_refresh = (uaddr) call_data;
}   /*  End Function auto_refresh_cbk   */

static void aspect_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the aspect callback.
*/
{
    flag bool = (uaddr) call_data;
    ZoomPolicyWidget top = (ZoomPolicyWidget) client_data;

    set_bool (top, VIEWIMG_ATT_MAINTAIN_ASPECT, bool);
}   /*  End Function aspect_cbk   */

static void trunc_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the trunc callback.
*/
{
    flag bool = (uaddr) call_data;
    ZoomPolicyWidget top = (ZoomPolicyWidget) client_data;

    set_bool (top, VIEWIMG_ATT_ALLOW_TRUNCATION, bool);
}   /*  End Function trunc_cbk   */

static void int_x_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the int_x callback.
*/
{
    flag bool = (uaddr) call_data;
    ZoomPolicyWidget top = (ZoomPolicyWidget) client_data;

    set_bool (top, VIEWIMG_ATT_INT_X, bool);
}   /*  End Function int_x_cbk   */

static void int_y_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the int_y callback.
*/
{
    flag bool = (uaddr) call_data;
    ZoomPolicyWidget top = (ZoomPolicyWidget) client_data;

    set_bool (top, VIEWIMG_ATT_INT_Y, bool);
}   /*  End Function int_y_cbk   */

static void magnify_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the magnify callback.
*/
{
    ZoomPolicyWidget top = client_data;
    MagnifyFunc func = NULL;
    struct canvas_list *curr;

    switch (top->zoomPolicy.magnify_choice)
    {
      case MAGNIFY_INTERPOLATE:
	func = imw_to8_bilinear;
	break;
    }
    for (curr = top->zoomPolicy.first_canvas; curr != NULL; curr = curr->next)
    {
	kwin_set_attributes (canvas_get_pixcanvas (curr->canvas),
			     KWIN_ATT_MAGNIFY_FUNC, func,
			     KWIN_ATT_END);
    }
    if (top->zoomPolicy.auto_refresh) apply_cbk (NULL, top, NULL);
}   /*  End Function iscale_cbk  */


/*  Public functions follow  */

void XkwZoomPolicyAddCanvas (Widget W, KWorldCanvas canvas)
/*  [SUMMARY] Add a KWorldCanvas to a ZoomPolicy widget canvas list.
    <W> The widget.
    <canvas> The canvas to add to the list. The widget will then manipulate
    this canvas. If this is NULL the routine does nothing.
    [RETURNS] Nothing.
*/
{
    ZoomPolicyWidget w;
    struct canvas_list *new;
    static char function_name[] = "XkwZoomPolicyAddCanvas";

    if ( !XtIsZoomPolicy (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (ZoomPolicyWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list structure");
    new->canvas = canvas;
    new->next = w->zoomPolicy.first_canvas;
    w->zoomPolicy.first_canvas = new;
}   /*  End Function XkwZoomPolicyAddCanvas  */
